/// @file acumon/acquisition_info_linux.c

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#define __STDC_FORMAT_MACROS

#include <stddef.h>
#include <inttypes.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>

#include "os_dependent.hh"
#include "acquisition_info.hh"
#include "am_assert.hh"

acq_handle_t *
open_acquisition_info(const char *acquisition_info_path)
{
    acq_handle_t *acq_handle = acq_create(acquisition_info_path);

    if (acq_handle == NULL) {
        am_printf(AM_Ctrl_Msg_Error,
		  "Error: Failed to create acquisition info file %s: %s\n",
		  acquisition_info_path, strerror(errno));
        force_exit(AM_Exit_ACQ);
    }

    /* Make the socket close itself in child processes */
    int fd = acq_fd(acq_handle);
    int flags = fcntl(fd, F_GETFD);
    AM_ASSERT_WITH_ERRNO(flags != -1);
    flags = fcntl(fd, F_SETFD, flags | FD_CLOEXEC);
    AM_ASSERT_WITH_ERRNO(flags != -1);
    return acq_handle;
}


void
close_acquisition_info(acq_handle_t *acq_handle)
{
    if (acq_handle) {
	int res;
	res = acq_close(acq_handle);
	AM_ASSERT(res == 0);
    }
}

void
append_acquisition_mappings(acq_handle_t *acq_handle, am_process_t *process)
{
    int res;
    acq_mapping_t acq_mapping;

    if (!acq_handle)
        return;

    char maps_fn[PATH_MAX];
    res = snprintf(maps_fn, PATH_MAX, "/proc/%d/maps", process->pid);
    AM_ASSERT(res >= 0 && res < PATH_MAX);
    FILE *maps_file = fopen(maps_fn, "r");
    if (maps_file == NULL) {
        /* Work-around for bug in some Fedora kernels, see ticket #319. */
        am_printf(AM_Ctrl_Msg_Warn,
		  "Warning: Failed to open %s: error %d, %s\n",
		  maps_fn, errno, strerror(errno));
        am_printf(AM_Ctrl_Msg_Warn,
		  "Warning: Source code mappings for shared libraries may be "
                  "incomplete.\n");
        return;
    }


    /* first read the mapping file */

    char *maps = NULL;
    int maps_max = 0;
    int maps_size;
    do {
        maps_max = maps_max == 0 ? 1024 : maps_max * 2;
        maps = (char *)am_realloc(maps, maps_max);
        res = fseek(maps_file, 0, SEEK_SET);
        AM_ASSERT(res == 0);
        maps_size = fread(maps, 1, maps_max, maps_file);
        if (ferror(maps_file)) {
            /* Work-around for bug in some Ubuntu kernels, see ticket #481. */
            am_printf(AM_Ctrl_Msg_Warn,
                      "Warning: Failed to read %s: error %d, %s\n",
                      maps_fn, errno, strerror(errno));
            am_printf(AM_Ctrl_Msg_Warn,
                      "Warning: Source code mappings for shared libraries "
                      "may be incomplete.\n");
            return;
        }
    } while (maps_size == maps_max);

    maps[maps_size] = 0;

    res = fclose(maps_file);
    AM_ASSERT(res == 0);

    /* Now parse the file and package the individual lines */

    VECT(acq_mapping_entry_t *, acq_mapping_entry) mapping_entries = VECT_NULL;

    uint64_t begin, end;
    char r, w, x, s;
    uint64_t offset;
    unsigned char dev_major, dev_minor;
    uint64_t inode;
    char *maps_pos = maps;
    int consumed;
    while ((res = sscanf(maps_pos, "%" SCNx64 "-%" SCNx64
			 " %c%c%c%c %" SCNx64 " %" SCNx8
			 ":%" SCNx8 " %" SCNd64 "%n",
			 &begin, &end, &r, &w, &x, &s,
			 &offset, &dev_major, &dev_minor,
			 &inode, &consumed))
	   != EOF) {

	acq_mapping_entry_t *m =
	    (acq_mapping_entry_t *)am_malloc(sizeof (acq_mapping_entry_t));

	m->pr_vaddr = begin;
	m->pr_size = end - begin;
	m->flag_read = r == 'r';
	m->flag_write = w == 'w';
	m->flag_exec = x == 'x';
	m->flag_shared = s == 's';
	m->pr_offset = offset;
	m->major = dev_major;
	m->minor = dev_minor;
	m->inode = inode;
	m->path = NULL;


	// Advance past leading white spaces
	maps_pos += consumed;
	while (*maps_pos == ' ')
	    maps_pos++;

	// Calculate length of path
	char *start_of_path = maps_pos;
	while (*maps_pos != '\n') {
	    if (*maps_pos == 0) {
		// error
		am_printf(AM_Ctrl_Msg_Error,
		  "Error: Failed to create acquisition info file: %s\n",
		  "Syntax error in maps file");
		goto cleanup;
	    }
	    maps_pos++;
	}

	// copy path
	m->path = (char *)malloc(maps_pos - start_of_path + 1);
	strncpy(m->path, start_of_path, maps_pos - start_of_path + 1);
	m->path[maps_pos-start_of_path] = 0;
	maps_pos++;

	VECT_APPEND(mapping_entries, acq_mapping_entry_t *, m);
    }

    acq_mapping.n_entries = VECT_SIZE(mapping_entries);
    acq_mapping.entries = VECT_RAW_ELEM(mapping_entries);

    /* Write record */

    acq_write_mapping(acq_handle, &acq_mapping);

    /* clean up */
cleanup:
    {
	acq_mapping_entry_t *map;
	VECT_FOREACH(mapping_entries, map) {
	    am_free(map->path);
	    am_free(map);
	}
	VECT_FINI(mapping_entries);
    }
}

void
write_acquisition_start(acq_handle_t *acq_handle, am_process_t *process, tid_t tid)
{
    int res;
    acq_begin_t acq_begin;

    if (!acq_handle)
        return;

    switch (get_process_bitness(process, tid)) {
    case 32:
        acq_begin.arch = Acq_I386;
        break;
    case 64:
        acq_begin.arch = Acq_X86_64;
        break;
    default:
        AM_ILLEGAL_STATE();
    }
    acq_begin.version = 1;
    acq_begin.start_time = get_time();
    const char *user = getenv("USER");
    acq_begin.user = user; /* null is fine */

    /* command line parameters */

    char argv_fn[PATH_MAX];
    res = snprintf(argv_fn, PATH_MAX, "/proc/%d/cmdline", process->pid);
    AM_ASSERT(res >= 0 && res < PATH_MAX);
    FILE *argv_file = fopen(argv_fn, "r");
    AM_ASSERT(argv_file != NULL);

    VECT(char *, charp) argv = VECT_NULL;

    int c;
    while ((c = fgetc(argv_file)) != EOF) {
        char *arg = NULL;
        int arg_len = 0;
        while (c != 0 && c != EOF) {
            arg = (char *)am_realloc(arg, ++arg_len);
            arg[arg_len - 1] = c;
            c = fgetc(argv_file);
        }
        arg = (char *)am_realloc(arg, ++arg_len);
        arg[arg_len - 1] = 0;

	VECT_APPEND(argv, char *, arg);
    }

    acq_begin.argc = VECT_SIZE(argv);
    acq_begin.argv = VECT_RAW_ELEM(argv);

    /* dump record to file */

    acq_write_begin(acq_handle, &acq_begin);

    /* cleanup */

    char *arg;
    VECT_FOREACH(argv, arg)
        am_free(arg);

    VECT_FINI(argv);

    res = fclose(argv_file);
    AM_ASSERT(res == 0);

    append_acquisition_mappings(acq_handle, process);
}

void
append_acquisition_end(acq_handle_t *acq_handle)
{
    if (!acq_handle)
        return;

    acq_end_t acq_end;
    acq_end.end_time = get_time();
    acq_write_end(acq_handle, &acq_end);
}
