/// @file acumon/acumon.c

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

/*
 *  This file is part of ThreadSpotter.
 *
 *  ThreadSpotter is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ThreadSpotter is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ThreadSpotter.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Operating system independent part of acumon.
 */

// TODO: handle unexpected closing of the control socket.
// TODO: error handling (errors from debug API calls)
// TODO: error handling for errors in the bootstrapper/instrumenter/debootstrapper
// TODO: handle EINTR properly in bootstrapper/instrumenter/debootstrapper
// TODO: Check all options before doing anything.

#define _XOPEN_SOURCE 600
/* We need _BSD_SOURCE for MAP_ANONYMOUS on Linux */
#define _BSD_SOURCE
#define ELF_TARGET_ALL

#include <stddef.h>
#include <stdio.h>
#include <limits.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <signal.h>
#include <getopt.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/poll.h>
#include <sys/utsname.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <gelf.h>
#include <libelf.h>

#include "library/version/version.h"
#include "library/am_mem/am_mem.h"
#include "library/exe_path/exe_path.hh"
#include "sampler/src/tls_size.h"
#include "sampler/src/sampler_limits.h"

#include "safe_io.hh"
#include "protocol.h"
#include "pages.hh"
#include "event.hh"
#include "binaries.h"
#include "am_assert.hh"
#include "os_dependent.hh"
#include "signal_handling.hh"
#include "acquisition_info.hh"
#include "instrumenter_loader.h"
#include "thread_instrumenter.h"
#include "thread_deinstrumenter.h"
#include "instrumenter_unloader.h"
#include "thread_predicates.hh"
#include "licensing.hh"
#include "acumon.hh"

namespace fs = boost::filesystem;

// Ugly hack... For some reason ULLONG_MAX is not defined on FC3.
#ifndef ULLONG_MAX
#define ULLONG_MAX ULONG_LONG_MAX
#endif

/* Bring in acumem svn version symbols by forcing linking to the libversion.a
   archive. */
BRING_IN_VERSIONINFO

/* Change to 1 to enable the debug commands in the command line interface. */
#define DEBUG_COMMANDS 0

/* Change to 1 to make acumon print some debug info when receiving a SIGINT. */
#define SIGINT_DEBUG 0

/* Change DEBUG_MODE to 1 to enable debug mode. Acumon will than automatically
   switch between attaching and detaching about every
   DEBUG_MODE_SWITCH_INTERVAL ns on average. */
#define DEBUG_MODE 0
#define DEBUG_MODE_SWITCH_INTERVAL 100000000
static int debug_mode_switches = 0;

/* Size of field f in struct s. */
#define SIZEOF_FIELD(s, f) sizeof(((s*)NULL)->f)

/* Path to our server socket and its file descriptor. The file descriptor is -1
   if the server socket isn't open. */
static am_path_buf_t server_socket_path;
static int server_socket_fd = -1;

sampler_params_t sampler_params = {
    1000000,             /* sample_period */
    0,                   /* sample_count */
    {6, -1},             /* log2linesizes */
    0,                   /* bp */
    {'\0'},              /* trace_file_prefix */
    0,                   /* code_area_start */
    0,                   /* code_area_end */
    ULLONG_MAX,          /* burst sampling length */
    ULLONG_MAX,          /* burst cooldown length */
    0,                   /* safe_stack */
    0,                   /* demo */
    0};                  /* use_stderr */

/* Monitored processes. */
process_vect_t processes = VECT_NULL;

/* Loaded but not relocated instrumenter image. */
//static unsigned long instr_32_image_size;
//static char *instr_32_image;
#if defined(__x86_64__)
static unsigned long instr_64_image_size;
static char *instr_64_image;
#endif

/* Offset of the instrumenter load entry point of the instrumenter image. */
//static unsigned long load_entry_offset_32;
#if defined(__x86_64__)
static unsigned long load_entry_offset_64;
#endif

/* Offset of the thread instrumentation entry point of the instrumenter
   image. */
//static unsigned long instrumentation_entry_offset_32;
#if defined(__x86_64__)
static unsigned long instrumentation_entry_offset_64;
#endif

/* Offset of the instrumenter unload entry point of the instrumenter image. */
//static unsigned long unload_entry_offset_32;
#if defined(__x86_64__)
static unsigned long unload_entry_offset_64;
#endif

/* Offset of the notification entry point of the instrumenter image. */
//static unsigned long notification_entry_offset_32;
#if defined(__x86_64__)
static unsigned long notification_entry_offset_64;
#endif

/* Offset of the notification stack of the instrumenter image. */
//static unsigned long notification_stack_offset_32;
#if defined(__x86_64__)
static unsigned long notification_stack_offset_64;
#endif

/* Offset of the notification args of the instrumenter image. */
//static unsigned long notification_args_offset_32;
#if defined(__x86_64__)
static unsigned long notification_args_offset_64;
#endif

/* A relocation of a field with size bytes at the specified offset in the
   instrumenter image.

   Value contains the value of the symbol or offset referred to by the
   relocation (if any) plus plus the explicit addend (if any).

   If add_base is true the instrumenter load address should be added to the
   value.

   If implicit_addend is true the contents of the field in the image should be
   added to the value.  */
typedef struct {
    unsigned long offset;
    int size;
    long value;
    int add_base;
    int implicit_addend;
} rel_t;

/* Returns a relocation entry with the specified paramters. */
static rel_t
gen_rel(int size, int add_base, int implicit_addend, unsigned long offset, long value) {
    rel_t rel = {offset, size, value, add_base, implicit_addend};
    return rel;
}

/* Array of instrumenter relocations. */
typedef VECT(rel_t, rel) rel_vect_t;
//rel_vect_t instr_32_rels = VECT_NULL;
#if defined(__x86_64__)
rel_vect_t instr_64_rels = VECT_NULL;
#endif

/* Offset and length of the readable-executable and readable-writable
   instrumenter segments, rounded to page boundaries. */
typedef struct {
    unsigned long start;
    unsigned long size;
} instr_seg_t;
//instr_seg_t instr_32_rx_seg = {-1};
//instr_seg_t instr_32_rw_seg = {-1};
#if defined(__x86_64__)
instr_seg_t instr_64_rx_seg = {-1};
instr_seg_t instr_64_rw_seg = {-1};
#endif

/* True when acumon is exiting. */
static int am_exiting = 0;

/* Exit code to return through the control socket. */
static am_exit_status_t am_exit_status;

#if defined(__x86_64__)
#define PREALLOCATION_MIN 0x8000000UL
#define PREALLOCATION_MAX 0x80000000UL
/* Amount of memory to preallocate for x86_64 targets. */
static uint64_t preallocation_size = 0x80000000UL;
#endif

/* True if acumon owns the terminal and could start the command line
   interface. */
int own_terminal = 1;

/* Start sampling after sampling_delay seconds, detach after sampling_duration
   seconds. */
int sampling_delay = 0;
int sampling_duration = 0;

/* start_at_function and start_at_address contains the function or address at
   which the sampling should be started. If a function is specified it is
   converted to an address as soon as the process has been launched or
   attached, and start_at_function is then not used any more.

   start_at_ignore specifies the number of executions of the start location
   that should be ignored before starting the sampling.

   start_at_bp_id is the id of the breakpoint used to detect executions of the
   start location. */
static const char *start_at_function = NULL;
static unsigned long start_at_address = 0;
static unsigned long start_at_ignore = 0;
static uint64_t start_at_bp_id;

/* The same as above, but for stopping the sampling. However, there is the
   additional twist that the instrumented threads will actually execute
   translated code for the specified address, not the specified address
   itself. We therefore keep a vector of translated addresses in
   stop_at_instrum_pcs. */
static const char *stop_at_function = NULL;
static unsigned long stop_at_address = 0;
static unsigned long stop_at_ignore = 0;
static uint64_t stop_at_bp_id;
VECT(unsigned long, long) stop_at_instrum_pcs = VECT_NULL;

/* The interval between the end of a burst and the start of the next in
   nanoseconds, or 0 if burst sampling is disabled. */
uint64_t burst_interval = 0;

/* Arguments to the launched process. */
argument_vect_t arguments = VECT_NULL;

/* Assign breakpoint ids from here. */
uint64_t next_breakpoint_id = 0;

/* Path to control socket and its file descriptor. The file descriptor is -1 if
   the control socket is not open. */
static char *control_socket_path = NULL;
static int control_socket_fd = -1;

/* Path to the acquisition info file. */
static char *acquisition_info_path = NULL;

/* Original atf_file_prefix */
static char *atf_file_prefix = NULL;

/* Acquisition info file generation number, increased after exec */
static int acquisition_generation_number = 0;

/* Make the server socket world accessible and disable UID checking */
static int unsupported_attach = 0;

/* True if an exit should be forced as soon as no threads are being attached or
   detached. */
static int queue_force_exit = 0;

/* Minimum fork generation to sample, default just the initial generation */
static int min_generation = 0;

/* Maximum fork generation to sample, default just the initial generation */
static int max_generation = 0;

/* A flag that says whether we have seen any process worth sampling at
   all. This controls some diagnostic output. TS-2239 */
static int seen_interesting_process = 0;

fs::path tmpdir;

/* Forward declarations. */
static void socket_event(short events, void *process_);
static void sigterm_event(void);
static void thread_send_event(short events, void *process_);
static void destroy_server_socket(void);
static void destroy_control_socket(void);
static void destroy_stdin_handler(void);
static void trigger_attach(am_process_t *process);
static void trigger_exit(am_exit_status_t exit_status);
static void control_exit(am_exit_status_t status);
static void remove_process(am_process_t *process);

const char *
thread_state_name(am_thread_state_t ts)
{
    switch (ts) {
    case AM_TS_Native:
        return "AM_TS_Native";
    case AM_TS_Load_Wait_All_Stopped:
        return "AM_TS_Load_Wait_All_Stopped";
    case AM_TS_Load_Wait_Hello:
        return "AM_TS_Load_Wait_Hello";
    case AM_TS_Load_Wait_Hello_Stopped:
        return "AM_TS_Load_Wait_Hello_Stopped";
    case AM_TS_Load_Wait_Ready:
        return "AM_TS_Load_Wait_Ready";
    case AM_TS_Load_Wait_Ready_Stopped:
        return "AM_TS_Load_Wait_Ready_Stopped";
    case AM_TS_Instr_Wait_All_Stopped:
        return "AM_TS_Instr_Wait_All_Stopped";
    case AM_TS_Instr_Wait_Ready:
        return "AM_TS_Instr_Wait_Ready";
    case AM_TS_Instr_Wait_Ready_Stopped:
        return "AM_TS_Instr_Wait_Ready_Stopped";
    case AM_TS_Instrumented:
        return "AM_TS_Instrumented";
    case AM_TS_Deinstr_Wait_Ready:
        return "AM_TS_Deinstr_Wait_Ready";
    case AM_TS_Deinstr_Wait_Done_1:
        return "AM_TS_Deinstr_Wait_Done_1";
    case AM_TS_Deinstr_Wait_Done_1_Run:
        return "AM_TS_Deinstr_Wait_Done_1_Run";
    case AM_TS_Deinstr_Wait_Done_1_All_Stopped:
        return "AM_TS_Deinstr_Wait_Done_1_All_Stopped";
    case AM_TS_Deinstr_Wait_Done_2:
        return "AM_TS_Deinstr_Wait_Done_2";
    case AM_TS_Deinstr_Wait_Done_2_Stopped:
        return "AM_TS_Deinstr_Wait_Done_2_Stopped";
    case AM_TS_Unload_Wait_All_Stopped:
        return "AM_TS_Unload_Wait_All_Stopped";
    case AM_TS_Unload_Wait_Close:
        return "AM_TS_Unload_Wait_Close";
    case AM_TS_Unload_Wait_Close_Stopped:
        return "AM_TS_Unload_Wait_Close_Stopped";
    case AM_TS_Instrumented_Exiting:
        return "AM_TS_Instrumented_Exiting";
    case AM_TS_Instrumented_Wait_Ready:
        return "AM_TS_Instrumented_Wait_Ready";
    case AM_TS_Notification_Wait:
        return "AM_TS_Notification_Wait";
    case AM_TS_Notification_Wait_Done:
        return "AM_TS_Notification_Wait_Done";
    case AM_TS_Notification_Wait_Done_Stopped:
        return "AM_TS_Notification_Wait_Done_Stopped";
    default:
        AM_ILLEGAL_STATE();
    }
}

const char *
process_state_name(am_process_state_t ps)
{
    switch (ps) {
    case AM_PS_Idle:
        return "AM_PS_Idle";
    case AM_PS_Burst:
        return "AM_PS_Burst";
    case AM_PS_Detach:
        return "AM_PS_Detach";
    default:
        AM_ILLEGAL_STATE();
    }
}

const char *
process_instrumentation_state_name(am_process_instrumentation_state_t pis)
{
    switch (pis) {
    case AM_PIS_Attaching:
        return "AM_PIS_Attaching";
    case AM_PIS_Idling:
        return "AM_PIS_Idling";
    case AM_PIS_Starting_Burst:
        return "AM_PIS_Starting_Burst";
    case AM_PIS_Bursting:
        return "AM_PIS_Bursting";
    case AM_PIS_Detaching:
        return "AM_PIS_Detaching";
    case AM_PIS_Native:
        return "AM_PIS_Native";
    default:
        AM_ILLEGAL_STATE();
    }
}

const char *
process_execution_state_name(am_process_execution_state_t pes)
{
    switch (pes) {
    case AM_PES_Run:
        return "AM_PES_Run";
    case AM_PES_Run_Excl:
        return "AM_PES_Run_Excl";
    case AM_PES_Run_One:
        return "AM_PES_Run_One";
    case AM_PES_Stop:
        return "AM_PES_Stop";
    default:
        AM_ILLEGAL_STATE();
    }
}

const char *
process_ptrace_purpose_name(am_process_ptrace_purpose_t ptp)
{
    switch (ptp) {
    case AM_PPP_Sample_And_Fork:
	return "AM_PPP_Sample_And_Fork";
    case AM_PPP_Fork:
	return "AM_PPP_Fork";
    default:
	AM_ILLEGAL_STATE();
    }
}

static am_process_ptrace_purpose_t
calculate_ptrace_purpose(am_process_t *process)
{
    if (process->generation >= min_generation &&
	process->generation <= max_generation)
	return AM_PPP_Sample_And_Fork;
    
    return AM_PPP_Fork;
}

int
process_may_be_instrumented(am_process_t *process)
{
    return process->ptrace_purpose == AM_PPP_Sample_And_Fork;
}

int
page_hijacked(const am_process_t *process)
{
    return !!process->hijack_page;
}

int
instrumenter_loaded(const am_process_t *process)
{
    return process->instr_addr != 0;
}

void
am_printf(am_control_message_level_t level, const char *format, ...)
{
    va_list ap;

    va_start(ap, format);
    if(control_socket_fd == -1)
	vfprintf(stderr, format, ap);
    else
	control_vprintf(control_socket_fd, 1, level, format, ap);
    va_end(ap);
}

void
am_perror(am_control_message_level_t level, const char *s)
{
    char err_buf[256];
    int err = errno;
    if (strerror_r(err, err_buf, sizeof(err_buf)) == 0) {
	if (s && s[0])
	    am_printf(level, "%s: %s\n", s, err_buf);
	else
	    am_printf(level, "%s\n", err_buf);
    } else {
	if (s && s[0])
	    am_printf(level, "%s: Unknown error %i\n", s, err);
	else
	    am_printf(level, "Unknown error %i\n", err);
    }
}

static unsigned long
lookup_symbol(am_process_t *process, const char *symbol)
{
    int res;

    unsigned long addr = 0;

    int elf_fd = safe_open(process->elf_file, O_RDONLY);
    if (elf_fd == -1) {
        am_printf(AM_Ctrl_Msg_Error, "Error: Failed to open ELF file %s for reading: %s\n"
		  "Can not look up symbol \'%s\'.",
                  process->elf_file, strerror(errno), symbol);
        force_exit(AM_Exit_Target_Binary);
    }

    elf_version(1);

    Elf *elf = elf_begin(elf_fd, ELF_C_READ, NULL);
    AM_ASSERT(elf != NULL);

    GElf_Ehdr ehdr;
    GElf_Ehdr *ehdr_p = gelf_getehdr(elf, &ehdr);
    AM_ASSERT(ehdr_p != NULL);

    size_t shstrndx;
    res = elf_getshstrndx(elf, &shstrndx);
    AM_ASSERT(res != -1);

    Elf_Scn   *sym_scn = NULL;
    GElf_Shdr  sym_shdr = {};
    Elf_Data  *sym_data = NULL;

    GElf_Half  str_idx = 0;
    int str_idx_found = 0;

    for (int i = 0; i < ehdr.e_shnum; i++) {
        Elf_Scn *scn = elf_getscn(elf, i);
        AM_ASSERT(scn != NULL);

        GElf_Shdr shdr;
        GElf_Shdr *shdr_p = gelf_getshdr(scn, &shdr);
        AM_ASSERT(shdr_p != NULL);

        /* Compiler/linker sometimes lays out unused section header table
           entries, e.g., when compiling on Ubuntu. Ignore them. */
        if (shdr.sh_type == SHT_NULL)
            continue;

        Elf_Data *data_p = elf_getdata(scn, NULL);
        AM_ASSERT(data_p != NULL);

        switch (shdr.sh_type) {
        case SHT_SYMTAB:
            AM_ASSERT(sym_scn == NULL);
            sym_scn = scn;
            sym_shdr = shdr;
            sym_data = data_p;
            break;
        case SHT_STRTAB:
            if (strcmp(elf_strptr(elf, shstrndx, shdr.sh_name), ".strtab") == 0) {
                AM_ASSERT(!str_idx_found);
                str_idx_found = 1;
                str_idx = i;
            }
            break;
        }
    }

    /* TODO: How to handle weak symbols? */

    if (sym_scn != NULL) {
        AM_ASSERT(str_idx_found);
        for (int i = 0; i < sym_shdr.sh_size / sym_shdr.sh_entsize; i++) {
            GElf_Sym sym;
            GElf_Sym *symp = gelf_getsym(sym_data, i, &sym);
            if (symp == NULL)
                continue;
            AM_ASSERT(symp != NULL);

            const char *sym_name = elf_strptr(elf, str_idx, sym.st_name);
            if (strcmp(sym_name, symbol) == 0) {
                if (GELF_ST_TYPE(sym.st_info) != STT_FUNC)
                    am_printf(AM_Ctrl_Msg_Warn, "Warning: Symbol \'%s\' does not refer to "
			      "a function.\n", symbol);

                addr = sym.st_value;
                goto done;
            }
        }
    }

    am_printf(AM_Ctrl_Msg_Error, "Error: No function \'%s\' found in %s.\n",
	      symbol, process->elf_file);
    force_exit(AM_Exit_Symbol);

done:
    elf_end(elf);

    safe_close(elf_fd);

    return addr;
}

static void
set_noevpend_on(am_thread_t *thread)
{
    union {
	unsigned int buf[sizeof(long)/sizeof(int)];
	unsigned long _align;
    } u;
    unsigned long noevpend_addr;
    unsigned long laddr;
    unsigned int iix;

    noevpend_addr = thread->noevpend_addr;
    iix = (noevpend_addr / sizeof(int)) & ((1 << ((sizeof(long)/sizeof(int)) - 1)) - 1);
    laddr = noevpend_addr & ~(unsigned long)(sizeof(long) - 1);

    if (sizeof(int) != sizeof(long))
	read_data(thread, laddr, sizeof(u), &u);
    u.buf[iix] = 0;
    write_data(thread, laddr, sizeof(u), &u, 0);
}

void
update_msg_counter(am_thread_t *thread)
{
    AM_ASSERT(thread_stopped(thread, 1));
    long counter;
    read_data(thread, thread->msg_counter_addr, sizeof(counter), &counter);
    counter += thread->pending_msgs;
    write_data(thread, thread->msg_counter_addr, sizeof(counter), &counter, 0);
    thread->pending_msgs = 0;

    set_noevpend_on(thread);
}

/* Add an execution breakpoint to a thread. */
static void
add_bp(am_thread_t *thread, uint64_t bp_id, unsigned long address, bp_cb_t callback)
{
    for (int i = 0; i < NUM_BPS; i++) {
        bp_t *bp = &thread->new_bps[i];
	/* Due to phase-ordering issues it's possible for the "same" breakpoint to
	   be added twice, which breaks e.g. start-at-ignore.  Prevent that.  */
	if (bp->bp_id == bp_id && bp->address == address && bp->callback == callback)
	    return;
        if (bp->callback == 0) {
            bp->bp_id = bp_id;
            bp->address = address;
            bp->callback = callback;
            bp->suspended = 0;
            goto bp_found;
        }
    }
    am_printf(AM_Ctrl_Msg_Warn, "Failed to find free breakpoint register for breakpoint. "
	      "The breakpoint will be ignored.\n");
bp_found:
    thread->update_bps = 1;
}

/* Remove an execution breakpoint from a thread. */
static void
remove_bp(am_thread_t *thread, int bp_num)
{
    thread->new_bps[bp_num].callback = 0;
    thread->update_bps = 1;
}

static void
remove_bp_by_id(am_thread_t *thread, int bp_id, unsigned long addr)
{
    for (int i = 0; i < NUM_BPS; i++) {
        bp_t *bp = &thread->bps[i];
        if (bp->bp_id == bp_id && bp->address == addr)
            remove_bp(thread, i);
    }
}

/* Suspend a breakpoint until we have single stepped past the next instruction
   (used to single step past an instruction that triggered the breakpoint). */
static void
suspend_bp(am_thread_t *thread, int bp_num)
{
    disable_bp(thread, bp_num);
    thread->bps[bp_num].suspended = 1;
}

static void
start_at_callback(am_thread_t *thread, uint64_t bp_id, int bp_num)
{
    am_process_t *process = thread->process;
    if (!process->attach_triggered) {
        if (process->start_at_ignore == 0) {
            am_printf(AM_Ctrl_Msg_Info, "Selected start location triggered, starting sampling.\n");
            trigger_attach(process);
            append_acquisition_mappings(process->acq_handle, process);
            am_thread_t *t;
            VECT_FOREACH(process->threads, t)
                remove_bp_by_id(t, bp_id, start_at_address);
        }
        else {
            suspend_bp(thread, bp_num);
            process->start_at_ignore--;
        }
    }
}

static void
stop_at_callback(am_thread_t *thread, uint64_t bp_id, int bp_num)
{
    if (!thread->process->exit_triggered) {
        if (stop_at_ignore == 0) {
            am_printf(AM_Ctrl_Msg_Info, "Selected stop location triggered, stopping sampling and exiting.\n");
            trigger_exit(AM_Exit_Normal);
            am_thread_t *t;
            VECT_FOREACH(thread->process->threads, t) {
                unsigned long instrum_pc;
                VECT_FOREACH(stop_at_instrum_pcs, instrum_pc)
                    remove_bp_by_id(t, bp_id, instrum_pc);
                remove_bp_by_id(t, bp_id, stop_at_address);
            }
        }
        else {
            suspend_bp(thread, bp_num);
            stop_at_ignore--;
        }
    }
}

/* Start the controlled exit sequence. Acumon continues to run as normal until
   all threads have been detached, and then exits. There is no guarantee that
   acumon will exit within a finite amount of time, since threads may never be
   able to detach because they are stuck in syscalls. */
static void
trigger_exit(am_exit_status_t exit_status)
{
    am_process_t *process;
    VECT_FOREACH(processes, process) {
        process->state = AM_PS_Detach;
	process->exit_triggered = 1;
    }
    am_exiting = 1;
    am_exit_status = exit_status;
    am_printf(AM_Ctrl_Msg_Info, "Waiting for all threads to detach.\n");
}

/* Send the messages in the process's message queue. Use blocking send if block
   is true. Returns 1 if all messages in the queue were sent, false if not. */
static int
send_messages(am_process_t *process, int block)
{
    while (process->msg_head != NULL) {
        am_message_buf_t *msg_buf = process->msg_head;
        am_thread_t *thread = msg_buf->thread;
	struct iovec iov[2] = {
	    {&msg_buf->msg,
	     sizeof(am_message_t) },
	    {msg_buf->extra_data,
	     msg_buf->msg.extra_data_len }
	};

	struct msghdr msg = {
	    NULL,
	    0,
	    &iov[0],
	    msg_buf->msg.extra_data_len > 0 ? 2 : 1,
	    NULL,
	    0,
	    0
	};

        ssize_t sent_bytes;
#if defined(__linux__)
        sent_bytes = safe_sendmsg(process->socket_fd, &msg,
				  (block ? 0 : MSG_DONTWAIT) | MSG_NOSIGNAL);
#else
        /* The MSG_DONTWAIT flag is only valid for recv() on Solaris,
           so if block is false we have to temporarily make the socket
           non-blocking for the send.*/
        int flags = 0; /* Initialize just to avoid compiler warning. */
        int res;
        if (!block) {
            flags = fcntl(process->socket_fd, F_GETFL);
            AM_ASSERT_WITH_ERRNO(flags != -1);
            res = fcntl(process->socket_fd, F_SETFL, flags | O_NONBLOCK);
            AM_ASSERT_WITH_ERRNO(res != -1);
        }
        sent_bytes = safe_sendmsg(process->socket_fd, &msg, 0);
        if (!block) {
            res = fcntl(process->socket_fd, F_SETFL, flags);
            AM_ASSERT_WITH_ERRNO(res != -1);
        }
#endif

        if (!block && sent_bytes == -1 && errno == EAGAIN)
            return 0;
        AM_ASSERT_WITH_ERRNO(sent_bytes != -1);
        AM_ASSERT(sent_bytes == sizeof(am_message_t) + msg_buf->msg.extra_data_len);
        if (msg_buf->inc_msg_counter)
            thread->pending_msgs++;
        am_message_buf_t *next = process->msg_head->next;
	if (msg_buf->extra_data)
	    am_free(msg_buf->extra_data);
        am_free(process->msg_head);
        process->msg_head = next;
        if (process->msg_head == NULL) {
            process->msg_tail = NULL;
            remove_fd_event(process->socket_fd, POLLOUT, thread_send_event, process);
        }
    }

    return 1;
}

static void
kill_if_launched(am_process_t *process)
{
    if (process->binary.size() != 0 && process->pid != 0) {
        int res = kill(process->pid, SIGKILL);
        AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
    }
}

static void
update_open_acquisition_info(am_process_t *process)
{
    char buf[PATH_MAX];
    snprintf(buf, PATH_MAX, "%s.%d", acquisition_info_path,
	     acquisition_generation_number);

    /* update sampler_param.trace_file_prefix */
    AM_ASSERT(atf_file_prefix != NULL);
    AM_ASSERT(strlen(atf_file_prefix) < TRACE_FILE_PREFIX_MAX);
    snprintf(process->trace_file_prefix, TRACE_FILE_PREFIX_MAX,
	     "%s.%d", atf_file_prefix, acquisition_generation_number);

    process->acq_handle = open_acquisition_info(buf);

    acquisition_generation_number++;
}

void
continue_after_exec(am_process_t *process)
{
    /* Close files and clean up the process and thread
     * structure. Note: The process is still PTRACE_ATTACHed from a ptrace
     * point of view. */

    int was_instrumented = process->instrumentation_state != AM_PIS_Native;
    int generation = process->generation;

    am_thread_t *thread;
    VECT_FOREACH(process->threads, thread)
	remove_thread(thread);

    pid_t pid = process->pid;
    pid_t ppid = process->ppid;
    remove_process(process);

    /* Create new process/thread objects. */
    process = add_process(pid, ppid, std::string(), generation);

    /* 0 means build new thread objects, but don't attempt to attach
     * threads to ptrace */
    attach_process(process, 0);
    if (was_instrumented)
        setup_start_conditions(process);

    AM_ASSERT(process->elf_file != NULL);
    if (licensing_check_manifest(process->elf_file) != 0)
        trigger_exit(AM_Exit_License_Exceeded);
}

__attribute__((__noreturn__)) void
force_exit(am_exit_status_t status)
{
    am_process_t *process;
    VECT_FOREACH(processes, process) {
        if (process->socket_fd != -1)
            send_messages(process, 1);

        am_thread_t *thread;
        VECT_FOREACH(process->threads, thread)
            force_detach_thread(thread);

        kill_if_launched(process);

	/* Normally this would be done in remove_process, but that is not called here */
	append_acquisition_end(process->acq_handle);
	close_acquisition_info(process->acq_handle);
	process->acq_handle = 0;
    }


    control_exit(status);
    exit(status);
}

/* Return true if val is a power of 2. */
static int
is_power_of_2(uint64_t val)
{
    return val != 0 && (val & (val - 1)) == 0 ;
}

/* Return 2 logarithm of val rounded down, or 0 for 0. */
static int
am_log2(uint64_t val)
{
    int i = -1;
    do {
        i++;
        val >>= 1;
    } while (val != 0);

    return i;
}

const char *
get_instr_image(const am_process_t *process)
{
    switch (process->bitness) {
//    case 32:
//        return instr_32_image;
#if defined(__x86_64__)
    case 64:
        return instr_64_image;
#endif
    default:
        AM_ILLEGAL_STATE();
    }
}

unsigned long
get_instr_image_size(const am_process_t *process)
{
    switch (process->bitness) {
//    case 32:
//        return instr_32_image_size;
#if defined(__x86_64__)
    case 64:
        return instr_64_image_size;
#endif
    default:
        AM_ILLEGAL_STATE();
    }
}

int
addr_in_instrumenter(const am_process_t *process, unsigned long addr)
{
    return (process->instr_addr != 0 &&
            (addr >= process->instr_addr &&
             addr < process->instr_addr + get_instr_image_size(process)));
}

static unsigned long
get_load_entry_offset(const am_process_t *process)
{
    switch (process->bitness) {
//    case 32:
//        return load_entry_offset_32;
//        break;
#if defined(__x86_64__)
    case 64:
        return load_entry_offset_64;
        break;
#endif
    default:
        AM_ILLEGAL_STATE();
    }
}

static unsigned long
get_instrumentation_entry_offset(const am_process_t *process)
{
    switch (process->bitness) {
//    case 32:
//        return instrumentation_entry_offset_32;
//        break;
#if defined(__x86_64__)
    case 64:
        return instrumentation_entry_offset_64;
        break;
#endif
    default:
        AM_ILLEGAL_STATE();
    }
}

static unsigned long
get_unload_entry_offset(const am_process_t *process)
{
    switch (process->bitness) {
//    case 32:
//        return unload_entry_offset_32;
//        break;
#if defined(__x86_64__)
    case 64:
        return unload_entry_offset_64;
        break;
#endif
    default:
        AM_ILLEGAL_STATE();
    }
}

static void
check_bitness(am_process_t *process, tid_t tid)
{
    if (process->bitness == -1)
        process->bitness = get_process_bitness(process, tid);

#if defined(__i386__)
    switch (process->bitness) {
    case 32:
        break;
    case 64:
        if (process->binary.size() == 0)
            am_printf(AM_Ctrl_Msg_Error, "Error: Process %" PRIpid " is a 64-bit process, "
		      "only sampling of 32-bit processes is supported.\n",
		      process->pid);
        else
            am_printf(AM_Ctrl_Msg_Error, "Error: \"%s\" starts a 64-bit process, only "
		      "sampling of 32-bit processes is supported.\n",
		      process->binary.c_str());
        force_exit(AM_Exit_Unsupported_Arch);
    default:
        AM_ILLEGAL_STATE();
    }
#endif
}

/* PRE: !exit_triggered && stop_at_address != 0 */
static void thread_add_stop_at_bps(am_thread_t *thread)
{
    unsigned long instrum_pc;

    VECT_FOREACH(stop_at_instrum_pcs, instrum_pc)
	add_bp(thread, stop_at_bp_id, instrum_pc, stop_at_callback);

    add_bp(thread, stop_at_bp_id, stop_at_address, stop_at_callback);
}

am_thread_t *
add_thread(am_process_t *process, tid_t tid,
           int debug_stopped, int instrumented)
{
    bp_t null_bp = {0, 0, 0};

    check_bitness(process, tid);

    set_thread_debug_options(tid);

    am_thread_t *thread = (am_thread_t *)am_malloc(sizeof(am_thread_t));
    thread->process = process;
    thread->tid = tid;
    thread->state =
        instrumented ? AM_TS_Instrumented_Wait_Ready : AM_TS_Native;
    thread->debug_stopped = debug_stopped;
    thread->pending_msgs = 0;
    thread->msg_counter_addr = 0;
    thread->noevpend_addr = 0;
    VECT_APPEND(process->threads, am_thread_t *, thread);
    for (int i = 0; i < NUM_BPS; i++) {
        thread->bps[i] = null_bp;
        thread->new_bps[i] = null_bp;
    }
    thread->update_bps = 0;
    thread->exiting_syscall = 0;
    thread->single_step = 0;
    thread->never_instrumented = !instrumented;

#if defined(__linux__)
    thread->sigstop_pending = 0;
    thread->sigtrap_pending = 0;
    thread->syscall_state = 1;
#endif
#if defined(__sun__)
    thread->ctl_fd = -1;
    thread->status_fd = -1;
    thread->abort_syscall = 0;
    thread->pcdstop_pending = 0;
    thread->suspend_state = AM_TSS_Continued;
    thread->suspender_tid = 0;
    thread->suspend_spinlock_addr = 0;
    thread->suspendee_tid = 0;
#endif

    if (start_at_address != 0 &&
        !process->attach_triggered)
        add_bp(thread, start_at_bp_id, start_at_address, start_at_callback);

    if (stop_at_address != 0 &&
        process->attach_triggered &&
        !process->exit_triggered)
        thread_add_stop_at_bps(thread);

    if (!sampler_params.demo && !feature_unlimited_threads &&
        VECT_SIZE(process->threads) > THREAD_LIMIT &&
        !process->exit_triggered) {
        am_printf(AM_Ctrl_Msg_Info, "Number of sampled threads allowed by "
                  "license (%d) exceeded, stopping sampling and exiting.\n",
                  THREAD_LIMIT);
        trigger_exit(AM_Exit_License_Exceeded);
    }

    return thread;
}

void
remove_thread(am_thread_t *thread)
{
    am_process_t *process = thread->process;

    int found;
    VECT_REMOVE(process->threads, thread, found);
    AM_ASSERT(found);

    if (process->msg_head != NULL) {
        am_message_buf_t **msg = &process->msg_head;
        while (*msg != NULL) {
            if ((*msg)->thread == thread) {
                am_message_buf_t *next = (*msg)->next;
		if ((*msg)->extra_data)
		    am_free(((*msg)->extra_data));
                am_free(*msg);
                *msg = next;
            }
            else
                msg = &(*msg)->next;
        }
        if (process->msg_head == NULL) {
            process->msg_tail = NULL;
            remove_fd_event(process->socket_fd, POLLOUT, thread_send_event, process);
        }
        else {
            msg = &process->msg_head;
            while ((*msg)->next != NULL)
                msg = &(*msg)->next;
            process->msg_tail = *msg;
        }
    }

    am_free(thread);
}

am_process_t *
add_process(pid_t pid, pid_t ppid, const std::string &binary, int generation)
{
    am_process_t *process = new am_process_t;
    process->pid = pid;
    process->ppid = ppid;
    process->generation = generation;
    process->require_ownership = !unsupported_attach;
    process->bitness = -1;
    process->binary = binary;
    process->state = AM_PS_Detach;
    process->instrumentation_state = AM_PIS_Native;
    process->execution_state = AM_PES_Run;
    VECT_INIT(process->threads);
    process->hijack_page = 0;
    process->socket_fd = -1;
    process->target_socket_fd = -1;
    process->msg_head = NULL;
    process->msg_tail = NULL;
    process->instr_addr = 0;
    process->elf_file = NULL;
    process->burst_generation = 0;
    process->start_new_burst = 1;
    process->is_execing = 0;

    process->start_at_ignore = start_at_ignore;
    process->stop_at_ignore = stop_at_ignore;
    process->attach_triggered = 0;
    process->exit_triggered = 0;

#if defined(__linux__)
    /* True if we have not seen any syscall yet. */
    process->first_syscall = 1;
#endif
#if defined(__sun__)
    process->ctl_fd = -1;
    process->status_fd = -1;
#endif

    process->acq_handle = 0;
    process->pages = new pages_t;

    process->ptrace_purpose = calculate_ptrace_purpose(process);

    VECT_APPEND(processes, am_process_t *, process);
    return process;
}

void
close_socket(am_process_t *process)
{
    if (process->socket_fd == -1)
        return;

    int res = close(process->socket_fd);
    AM_ASSERT_WITH_ERRNO(res == 0);

    remove_fd_event(process->socket_fd, POLLIN, socket_event, process);
    if (process->msg_head != NULL)
        remove_fd_event(process->socket_fd, POLLOUT, thread_send_event, process);

    process->socket_fd = -1;
    process->target_socket_fd = -1;
}

static void
remove_process(am_process_t *process)
{
    AM_ASSERT(VECT_SIZE(process->threads) == 0);

    int found;
    VECT_REMOVE(processes, process, found);
    AM_ASSERT(found);

    remove_time_events_with_data(process);

    if (process->acq_handle) {
	append_acquisition_end(process->acq_handle);
	close_acquisition_info(process->acq_handle);
	process->acq_handle = 0;
    }

    am_free(process->elf_file);
    VECT_FINI(process->threads);
    close_socket(process);

    delete process->pages;
    delete process;
}

static void
sigabrt_handler(void)
{
    am_process_t *process;
    VECT_FOREACH(processes, process) {
        kill_if_launched(process);

	/* Normally called during remove_process, but that is not called here */
	append_acquisition_end(process->acq_handle);
	close_acquisition_info(process->acq_handle);
	process->acq_handle = 0;
    }

    /* Use the abort(3) to raise SIGABRT and make sure the parent gets
       notified that we failed on a fatal signal. We don't use
       raise(SIGABRT) since abort(3) provides better fallbacks if
       signal delivery fails. */
    abort();
}

/* Called to free all data structures and exit once there are no more
   processes. */
static void __attribute__((__noreturn__))
exit_acumon(void)
{
    AM_ASSERT(VECT_SIZE(processes) == 0);
    VECT_FINI(processes);
//    VECT_FINI(instr_32_rels);
#if defined(__x86_64__)
    VECT_FINI(instr_64_rels);
#endif
//    am_free(instr_32_image);
#if defined(__x86_64__)
    am_free(instr_64_image);
#endif
    unlink(server_socket_path.chars);
    if (SIGINT_DEBUG)
        remove_signal_handler(SIGINT, sigint_debug);
    remove_signal_event(SIGTERM, sigterm_event);
    remove_signal_handler(SIGABRT, sigabrt_handler);
    control_exit(am_exit_status);
    destroy_server_socket();
    destroy_control_socket();
    destroy_stdin_handler();
    os_dependent_fini();
    signal_handling_fini();
    event_fini();
    licensing_fini();
    exit(0);
}

/* Process 'ppid' has no live threads and is about to be removed.
 * If a process 'pid' has 'ppid' as its parent, and 'pid' is not instrumentable,
 * then consider 'pid' an orphan that should not be tracked by acumon any longer.
 * This handles applications that spawn off daemonised helper processes that remain
 * alive even after the initial processes have terminated.  See TS-2250 for one such
 * case involving GIMP.
 */
static void reap_orphans(pid_t ppid)
{
    am_process_t *process;

    VECT_FOREACH(processes, process) {
	if (process->ppid != ppid)
	    continue;
	if (process_may_be_instrumented(process)) {
	    process->ppid = 0;
	    continue;
	}
	while (VECT_SIZE(process->threads) > 0) {
	    am_thread_t *thread;

	    thread = VECT_GET(process->threads, 0);
	    detach_thread(thread);
	    remove_thread(thread);
	}
    }
}

void
prune_processes(void)
{
    am_process_t *process;
restart_processes:
    VECT_FOREACH(processes, process)
        if (VECT_SIZE(process->threads) == 0) {
	    
	    reap_orphans(process->pid);

	    // Suppress this status message for threads in processes
	    // that we are not interested in sampling. See TS-2239 for
	    // rationale.
	    if (process_may_be_instrumented(process))
		am_printf(AM_Ctrl_Msg_Info,
			  "All threads in process %" PRIpid
			  " have terminated.\n",
			  process->pid);
            remove_process(process);
            goto restart_processes;
        }
}

static long
rel_sym_value(GElf_Rel *rel, Elf_Data *sym_scn_data)
{
    int sym_idx = GELF_R_SYM(rel->r_info);
    GElf_Sym sym;
    GElf_Sym *symp = gelf_getsym(sym_scn_data, sym_idx, &sym);
    AM_ASSERT(symp != NULL);
    return sym.st_value;
}

static long
rela_sym_value(GElf_Rela *rela, Elf_Data *sym_scn_data)
{
    int sym_idx = GELF_R_SYM(rela->r_info);
    GElf_Sym sym;
    GElf_Sym *symp = gelf_getsym(sym_scn_data, sym_idx, &sym);
    AM_ASSERT(symp != NULL);
    return sym.st_value;
}

static void
load_instrumenter(char instr_bin[], unsigned int instr_bin_size,
                  char **instr_image, unsigned long *instr_image_size,
                  instr_seg_t *instr_rx_seg, instr_seg_t *instr_rw_seg,
                  unsigned long *load_entry_offset,
                  unsigned long *instrumentation_entry_offset,
                  unsigned long *unload_entry_offset,
                  unsigned long *notification_entry_offset,
                  unsigned long *notification_stack_offset,
                  unsigned long *notification_args_offset,
                  rel_vect_t *instr_rels, int instr_bitness)
{
    typedef struct {
        Elf_Scn   *scn;
        GElf_Shdr  shdr;
        Elf_Data  *data;
    } scn_info_t;

    int res;

    elf_version(1);

    Elf *elf = elf_memory(instr_bin, instr_bin_size);
    AM_ASSERT(elf != NULL);

    GElf_Ehdr ehdr;
    GElf_Ehdr *ehdr_p = gelf_getehdr(elf, &ehdr);
    AM_ASSERT(ehdr_p != NULL);

    /* Get the size of the instrumenter image. */
    *instr_image_size = 0;
    for (int i = 0; i < ehdr.e_phnum; i++) {
        GElf_Phdr phdr;
        GElf_Phdr *phdr_p = gelf_getphdr(elf, i, &phdr);
        AM_ASSERT(phdr_p != NULL);

        if (phdr.p_type == PT_LOAD) {
            unsigned long end = phdr.p_vaddr + phdr.p_memsz;
            if (end > *instr_image_size)
                *instr_image_size = end;
        }
    }
    *instr_image_size = (*instr_image_size + AM_PAGE_SIZE - 1) & ~(AM_PAGE_SIZE - 1);

    /* Allocate the instrumenter image. */
    *instr_image = (char *)am_malloc(*instr_image_size);
    memset(*instr_image, 0, *instr_image_size);

    /* Load the instrumenter segments into the image. */
    for (int i = 0; i < ehdr.e_phnum; i++) {
        GElf_Phdr phdr;
        GElf_Phdr *phdr_p = gelf_getphdr(elf, i, &phdr);
        AM_ASSERT(phdr_p != NULL);

        if (phdr.p_type == PT_LOAD) {
            memcpy(*instr_image + phdr.p_vaddr, instr_bin + phdr.p_offset, phdr.p_filesz);

            instr_seg_t *seg = NULL;
            switch (phdr.p_flags) {
            case PF_R | PF_X:
                seg = instr_rx_seg;
                break;
            case PF_R | PF_W:
            case PF_R | PF_W | PF_X:
                seg = instr_rw_seg;
                break;
            default:
                AM_ILLEGAL_STATE();
            }
            AM_ASSERT(seg->start == -1);

            seg->start = phdr.p_vaddr & ~(AM_PAGE_SIZE - 1);
            seg->size = ((phdr.p_vaddr + phdr.p_memsz + AM_PAGE_SIZE - 1) &
                         ~(AM_PAGE_SIZE - 1)) - seg->start;
        }
    }

    /* Relocation sections. */
    typedef VECT(scn_info_t, scn_info) scn_info_vect_t;
    scn_info_vect_t rel_scns = VECT_NULL;
    scn_info_vect_t rela_scns = VECT_NULL;

    /* Dynamic symbol table. */
    scn_info_t dynsym_scn = {NULL, {}, NULL};

    /* Symbol table. */
    scn_info_t symtab_scn = {NULL, {}, NULL};

    /* Section index of the string table. */
    GElf_Half  str_idx = 0;
    int str_idx_found = 0;

    /* Get the section header string table index. */
    size_t shstrndx;
    res = elf_getshstrndx(elf, &shstrndx);
    AM_ASSERT(res != -1);

    /* Find the relocation and symbol table sections. */
    for (int i = 0; i < ehdr.e_shnum; i++) {
        Elf_Scn *scn = elf_getscn(elf, i);
        AM_ASSERT(scn != NULL);

        GElf_Shdr shdr;
        GElf_Shdr *shdr_p = gelf_getshdr(scn, &shdr);
        AM_ASSERT(shdr_p != NULL);

        /* Compiler/linker sometimes lays out unused section header table
           entries, e.g., when compiling on Ubuntu. Ignore them. */
        if (shdr.sh_type == SHT_NULL)
            continue;

        Elf_Data *data_p = elf_getdata(scn, NULL);
        AM_ASSERT(data_p != NULL);

        switch (shdr.sh_type) {
        case SHT_REL: {
            scn_info_t scn_info = {scn, shdr, data_p};
            VECT_APPEND(rel_scns, scn_info_t, scn_info);
        } break;
        case SHT_RELA: {
            scn_info_t scn_info = {scn, shdr, data_p};
            VECT_APPEND(rela_scns, scn_info_t, scn_info);
        } break;
        case SHT_DYNSYM:
            AM_ASSERT(dynsym_scn.scn == NULL);
            dynsym_scn.scn = scn;
            dynsym_scn.shdr = shdr;
            dynsym_scn.data = data_p;
            break;
        case SHT_SYMTAB:
            AM_ASSERT(symtab_scn.scn == NULL);
            symtab_scn.scn = scn;
            symtab_scn.shdr = shdr;
            symtab_scn.data = data_p;
            break;
        case SHT_STRTAB:
            if (strcmp(elf_strptr(elf, shstrndx, shdr.sh_name), ".strtab") == 0) {
                AM_ASSERT(!str_idx_found);
                str_idx_found = 1;
                str_idx = i;
            }
            break;
        }
    }

    AM_ASSERT(VECT_SIZE(rel_scns) + VECT_SIZE(rela_scns) > 0);
    AM_ASSERT(dynsym_scn.scn != NULL);
    AM_ASSERT(symtab_scn.scn != NULL);
    AM_ASSERT(str_idx_found);

    /* Load the instrumenter relocations. */
    scn_info_t rel_scn;
    VECT_FOREACH(rel_scns, rel_scn) {
        AM_ASSERT(instr_bitness == 32);
        for (int i = 0; i < rel_scn.shdr.sh_size / rel_scn.shdr.sh_entsize; i++) {
            GElf_Rel rel;
            GElf_Rel *rel_p = gelf_getrel(rel_scn.data, i, &rel);
            AM_ASSERT(rel_p != NULL);

            rel_t r;
            switch (GELF_R_TYPE(rel.r_info)) {
            case R_386_32:
                r = gen_rel(4, 1, 1, rel.r_offset, rel_sym_value(rel_p, dynsym_scn.data));
                break;
            case R_386_PC32:
                r = gen_rel(4, 0, 1, rel.r_offset, rel_sym_value(rel_p, dynsym_scn.data) - rel.r_offset);
                break;
            case R_386_GLOB_DAT:
                r = gen_rel(4, 1, 0, rel.r_offset, rel_sym_value(rel_p, dynsym_scn.data));
                break;
            case R_386_JMP_SLOT:
                r = gen_rel(4, 1, 0, rel.r_offset, rel_sym_value(rel_p, dynsym_scn.data));
                break;
            case R_386_RELATIVE:
                r = gen_rel(4, 1, 1, rel.r_offset, 0);
                break;
            default:
                AM_ILLEGAL_STATE();
            }
            VECT_APPEND(*instr_rels, rel_t, r);
        }
    }
    scn_info_t rela_scn;
    VECT_FOREACH(rela_scns, rela_scn) {
        AM_ASSERT(instr_bitness == 64);
        for (int i = 0; i < rela_scn.shdr.sh_size / rela_scn.shdr.sh_entsize; i++) {
            GElf_Rela rela;
            GElf_Rela *rela_p = gelf_getrela(rela_scn.data, i, &rela);
            AM_ASSERT(rela_p != NULL);

            rel_t r;
            switch (GELF_R_TYPE(rela.r_info)) {
            case R_X86_64_64:
                r = gen_rel(8, 1, 0, rela.r_offset, rela_sym_value(rela_p, dynsym_scn.data) + rela.r_addend);
                break;
            case R_X86_64_PC32:
                r = gen_rel(4, 0, 0, rela.r_offset, rela_sym_value(rela_p, dynsym_scn.data) + rela.r_addend - rela.r_offset);
                break;
            case R_X86_64_GLOB_DAT:
                r = gen_rel(8, 1, 0, rela.r_offset, rela_sym_value(rela_p, dynsym_scn.data));
                break;
            case R_X86_64_JUMP_SLOT:
                r = gen_rel(8, 1, 0, rela.r_offset, rela_sym_value(rela_p, dynsym_scn.data));
                break;
            case R_X86_64_RELATIVE:
                r = gen_rel(8, 1, 0, rela.r_offset, rela.r_addend);
                break;
            default:
                AM_ILLEGAL_STATE();
            }
            VECT_APPEND(*instr_rels, rel_t, r);
        }
    }

    int load_entry_found = 0;
    int instrumentation_entry_found = 0;
    int unload_entry_found = 0;
    int notification_entry_found = 0;
    int notification_stack_found = 0;
    int notification_args_found = 0;

    /* Get the offsets of the entry points. */
    for (int i = 0; i < symtab_scn.shdr.sh_size / symtab_scn.shdr.sh_entsize; i++) {
        GElf_Sym sym;
        GElf_Sym *symp = gelf_getsym(symtab_scn.data, i, &sym);
        if (symp == NULL)
            continue;
        AM_ASSERT(symp != NULL);

        const char *sym_name = elf_strptr(elf, str_idx, sym.st_name);
        if (strcmp(sym_name, "load_entry") == 0) {
            AM_ASSERT(!load_entry_found);
            load_entry_found = 1;
            AM_ASSERT(GELF_ST_TYPE(sym.st_info) == STT_FUNC);
            *load_entry_offset = sym.st_value;
        }
        if (strcmp(sym_name, "instrumentation_entry") == 0) {
            AM_ASSERT(!instrumentation_entry_found);
            instrumentation_entry_found = 1;
            AM_ASSERT(GELF_ST_TYPE(sym.st_info) == STT_FUNC);
            *instrumentation_entry_offset = sym.st_value;
        }
        if (strcmp(sym_name, "unload_entry") == 0) {
            AM_ASSERT(!unload_entry_found);
            unload_entry_found = 1;
            AM_ASSERT(GELF_ST_TYPE(sym.st_info) == STT_FUNC);
            *unload_entry_offset = sym.st_value;
        }
         if (strcmp(sym_name, "notification_entry") == 0) {
            AM_ASSERT(!notification_entry_found);
            notification_entry_found = 1;
            AM_ASSERT(GELF_ST_TYPE(sym.st_info) == STT_FUNC);
            *notification_entry_offset = sym.st_value;
        }
        if (strcmp(sym_name, "notification_stack") == 0) {
            AM_ASSERT(!notification_stack_found);
            notification_stack_found = 1;
            AM_ASSERT(GELF_ST_TYPE(sym.st_info) == STT_OBJECT);
            *notification_stack_offset = sym.st_value;
        }
        if (strcmp(sym_name, "notification_args") == 0) {
            AM_ASSERT(!notification_args_found);
            notification_args_found = 1;
            AM_ASSERT(GELF_ST_TYPE(sym.st_info) == STT_OBJECT);
            *notification_args_offset = sym.st_value;
        }
    }

    AM_ASSERT(load_entry_found);
    AM_ASSERT(instrumentation_entry_found);
    AM_ASSERT(unload_entry_found);
    AM_ASSERT(notification_entry_found);
    AM_ASSERT(notification_stack_found);
    AM_ASSERT(notification_args_found);

    elf_end(elf);

    VECT_FINI(rel_scns);
    VECT_FINI(rela_scns);
}

static void
thread_send_event(short events, void *process_)
{
    am_process_t *process = (am_process_t *)process_;
    send_messages(process, 0);
}

/* Send the supplied message to the thread. The message will be freed when it
   has been sent, so the caller must not free or reuse it. */
static void
thread_send(am_thread_t *thread, am_message_buf_t *msg, int inc_msg_counter)
{
    am_process_t *process = thread->process;

    AM_ASSERT(process->socket_fd != -1);

    msg->thread = thread;
    msg->inc_msg_counter = inc_msg_counter;
    msg->next = NULL;
    if (process->msg_tail != NULL) {
        process->msg_tail->next = msg;
        process->msg_tail = msg;
    }
    else {
        process->msg_head = msg;
        process->msg_tail = msg;
        add_fd_event(process->socket_fd, POLLOUT, thread_send_event, process);
    }
}

am_thread_t *
thread_by_tid(am_process_t *process, tid_t tid)
{
    am_thread_t *thread;
    VECT_FOREACH(process->threads, thread)
        if (thread->tid == tid)
            return thread;

    return NULL;
}

static void
relocate_instrumenter(am_process_t *process, am_thread_t *thread)
{
    unsigned long instr_image_size = get_instr_image_size(process);
    const char *instr_image = get_instr_image(process);

    char *instr_buf = (char *)am_malloc(instr_image_size);

    memcpy(instr_buf, instr_image, instr_image_size);

    rel_vect_t *instr_rels;
    switch (process->bitness) {
//    case 32:
//        instr_rels = &instr_32_rels;
//        break;
#if defined(__x86_64__)
    case 64:
        instr_rels = &instr_64_rels;
        break;
#endif
    default:
        AM_ILLEGAL_STATE();
    }

    rel_t rel;
    VECT_FOREACH(*instr_rels, rel)
        switch (rel.size) {
        case 4: {
            uint32_t value = rel.value;
            if (rel.add_base)
                value += process->instr_addr;
            if (rel.implicit_addend)
                value += *(uint32_t *)(instr_buf + rel.offset);
            *(uint32_t *)(instr_buf + rel.offset) = value;
        } break;
        case 8: {
            uint64_t value = rel.value;
            if (rel.add_base)
                value += process->instr_addr;
            if (rel.implicit_addend)
                value += *(uint64_t *)(instr_buf + rel.offset);
            *(uint64_t *)(instr_buf + rel.offset) = value;
        } break;
        default:
            AM_ILLEGAL_STATE();
        }
    AM_ASSERT(instr_image_size % WORD_SIZE == 0);

    /* We know the memory allocated for the instrumenter by the bootstrapper is
       already zeroed, so we set ignore_zeros to optimize the write. */
    write_data(thread, process->instr_addr,  instr_image_size, instr_buf, 1);

    am_free(instr_buf);
}

static void
set_syscall_args_32(
    syscall_args_t *syscall_args,
    uint32_t arg0, uint32_t arg1, uint32_t arg2,
    uint32_t arg3, uint32_t arg4, uint32_t arg5)
{
    syscall_args->args32[0] = arg0;
    syscall_args->args32[1] = arg1;
    syscall_args->args32[2] = arg2;
    syscall_args->args32[3] = arg3;
    syscall_args->args32[4] = arg4;
    syscall_args->args32[5] = arg5;
}

#if defined(__x86_64__)
static void
set_syscall_args_64(
    syscall_args_t *syscall_args,
    uint64_t arg0, uint64_t arg1, uint64_t arg2,
    uint64_t arg3, uint64_t arg4, uint64_t arg5)
{
    syscall_args->args64[0] = arg0;
    syscall_args->args64[1] = arg1;
    syscall_args->args64[2] = arg2;
    syscall_args->args64[3] = arg3;
    syscall_args->args64[4] = arg4;
    syscall_args->args64[5] = arg5;
}
#endif

static void
run_instrumenter_loader(am_thread_t *thread)
{
    am_process_t *process = thread->process;

    unsigned long instr_image_size = get_instr_image_size(process);

    set_hold_signals(thread, 1);
    process->hijack_page = find_hijack_page(thread, 0);
    read_data(thread, process->hijack_page, AM_PAGE_SIZE, &process->hijack_page_buf);

    save_registers(thread);

    instrumenter_loader_page_t il_page;
    switch (process->bitness) {
//    case 32:
//        il_page.code = instrumenter_loader_bin_32;
//        il_page.preallocation_size = 0;
        /* NOTE: Assumes that the PROT_* and MAP_* values are the same for
                 32-bit syscalls as for 64-bit syscalls. */
//        set_syscall_args_32(
//            &il_page.mmap_syscall_args,
//            0, TLS_MMAP_SIZE, PROT_READ | PROT_WRITE,
//            MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);
//        break;
#if defined(__x86_64__)
    case 64:
        il_page.code = instrumenter_loader_bin_64;
        il_page.preallocation_size = preallocation_size;
        break;
#endif
    default:
        AM_ILLEGAL_STATE();
    }
    il_page.instr_size = instr_image_size;
    il_page.socket_path = server_socket_path;
    il_page.load_ready_msg.pid = process->pid;
    il_page.load_ready_msg.tid = thread->tid;
    il_page.load_ready_msg.kind = AM_Msg_Load_Ready;
    il_page.load_ready_msg.extra_data_len = 0;

    write_data(thread, process->hijack_page, AM_PAGE_SIZE, &il_page, 0);

    set_rip(thread, process->hijack_page);
    set_rbp(thread, process->hijack_page);

    thread_run(thread);
}

static void
check_new_burst_generation(am_process_t *process)
{
    if (burst_interval != 0 && process->start_new_burst) {
        process->start_new_burst = 0;
        process->burst_generation++;
        append_burst_start(process->acq_handle,
			   process->burst_generation,
                           sampler_params.burst_sampling_length,
                           sampler_params.burst_cooldown_length);
    }
}

static void
run_thread_instrumenter(am_thread_t *thread)
{
    am_process_t *process = thread->process;

    check_new_burst_generation(process);

    set_hold_signals(thread, 1);
    process->hijack_page = find_hijack_page(thread, 0);
    read_data(thread, process->hijack_page, AM_PAGE_SIZE, &process->hijack_page_buf);

    save_registers(thread);

    am_code_buf_t *thread_instrumenter_bin;
    switch (process->bitness) {
//    case 32:
//        thread_instrumenter_bin = &thread_instrumenter_bin_32;
//        break;
#if defined(__x86_64__)
    case 64:
        thread_instrumenter_bin = &thread_instrumenter_bin_64;
        break;
#endif
    default:
        AM_ILLEGAL_STATE();
    }

    thread_instrumenter_page_t ti_page;
    ti_page.code = *thread_instrumenter_bin;
    ti_page.regs = thread->regs;
    ti_page.new_thread = thread->never_instrumented;
    ti_page.burst_generation = process->burst_generation;
    ti_page.instrumentation_entry =
        process->instr_addr + get_instrumentation_entry_offset(process);
    switch (process->bitness) {
//    case 32:
        /* NOTE: Assumes that the PROT_* and MAP_* values are the same for
                 32-bit syscalls as for 64-bit syscalls. */
//        set_syscall_args_32(
//            &ti_page.mmap_syscall_args,
//            0, TLS_MMAP_SIZE, PROT_READ | PROT_WRITE,
//            MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);
//        break;
#if defined(__x86_64__)
    case 64:
        break;
#endif
    default:
        AM_ILLEGAL_STATE();
    }

    write_data(thread, process->hijack_page, AM_PAGE_SIZE, &ti_page, 0);

    set_rip(thread, process->hijack_page);
    set_rbp(thread, process->hijack_page);

    thread_run(thread);

    thread->never_instrumented = 0;
}

static void
run_thread_deinstrumenter(am_thread_t *thread)
{
    am_process_t *process = thread->process;

    set_hold_signals(thread, 1);
    process->hijack_page = find_hijack_page(thread, 0);
    read_data(thread, process->hijack_page, AM_PAGE_SIZE, &process->hijack_page_buf);

    thread_deinstrumenter_page_t td_page;
    switch (process->bitness) {
//    case 32:
//        td_page.code = thread_deinstrumenter_bin_32;
//        break;
#if defined(__x86_64__)
    case 64:
        td_page.code = thread_deinstrumenter_bin_64;
        break;
#endif
    default:
        AM_ILLEGAL_STATE();
    }
    td_page.stack_addr = thread->stack_addr;
    td_page.deinstr_done_2_msg.pid = process->pid;
    td_page.deinstr_done_2_msg.tid = thread->tid;
    td_page.deinstr_done_2_msg.kind = AM_Msg_Deinstr_Done_2;
    td_page.deinstr_done_2_msg.extra_data_len = 0;
    switch (process->bitness) {
/*    case 32:
        set_syscall_args_32(
            &td_page.sendto_syscall_args,
            process->target_socket_fd,
            process->hijack_page +
            offsetof(thread_deinstrumenter_page_t, deinstr_done_2_msg),
            sizeof(am_message_t), 0, 0, 0);
        break;
*/
#if defined(__x86_64__)
    case 64:
        set_syscall_args_64(
            &td_page.sendto_syscall_args,
            process->target_socket_fd,
            process->hijack_page +
            offsetof(thread_deinstrumenter_page_t, deinstr_done_2_msg),
            sizeof(am_message_t), 0, 0, 0);
        break;
#endif
    default:
        AM_ILLEGAL_STATE();
    }

    write_data(thread, process->hijack_page, AM_PAGE_SIZE, &td_page, 0);

    set_rip(thread, process->hijack_page);
    set_rbp(thread, process->hijack_page);

    thread_run(thread);
}

static void
run_instrumenter_unloader(am_thread_t *thread)
{
    am_process_t *process = thread->process;

    unsigned long instr_image_size = get_instr_image_size (process);

    set_hold_signals(thread, 1);
    process->hijack_page = find_hijack_page(thread, 0);
    read_data(thread, process->hijack_page, AM_PAGE_SIZE, &process->hijack_page_buf);

    save_registers(thread);

    instrumenter_unloader_page_t iu_page;
    switch (process->bitness) {
//    case 32:
//        iu_page.code = instrumenter_unloader_bin_32;
//        break;
#if defined(__x86_64__)
    case 64:
        iu_page.code = instrumenter_unloader_bin_64;
        break;
#endif
    default:
        AM_ILLEGAL_STATE();
    }
    iu_page.unload_entry = process->instr_addr + get_unload_entry_offset(process);
    iu_page.instr_addr = process->instr_addr;
    switch (process->bitness) {
/*    case 32:
        iu_page.instr_size = instr_image_size;
        set_syscall_args_32(
            &iu_page.close_syscall_args,
            process->target_socket_fd, 0, 0, 0, 0, 0);
        set_syscall_args_32(
            &iu_page.mmap_syscall_args,
            0, TLS_MMAP_SIZE, PROT_READ | PROT_WRITE,
            MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);
        break;
*/
#if defined(__x86_64__)
    case 64:
        iu_page.instr_size = preallocation_size;
        set_syscall_args_64(
            &iu_page.close_syscall_args,
            process->target_socket_fd, 0, 0, 0, 0, 0);
        break;
#endif
    default:
        AM_ILLEGAL_STATE();
    }

    write_data(thread, process->hijack_page, AM_PAGE_SIZE, &iu_page, 0);

    set_rip(thread, process->hijack_page);
    set_rbp(thread, process->hijack_page);

    thread_run(thread);

    process->instr_addr = 0;
}

static void
run_notification(am_thread_t *thread)
{
    AM_ASSERT(thread_stopped(thread, 1));
    AM_ASSERT(!thread_in_syscall(thread));

    am_process_t *process = thread->process;

    save_registers(thread);

    unsigned long entry_offset;
    unsigned long stack_offset;
    unsigned long args_offset;

    switch (process->bitness) {
/*    case 32:
        entry_offset = notification_entry_offset_32;
        stack_offset = notification_stack_offset_32;
        args_offset = notification_args_offset_32;
        break;
*/
#if defined(__x86_64__)
    case 64:
        entry_offset = notification_entry_offset_64;
        stack_offset = notification_stack_offset_64;
        args_offset = notification_args_offset_64;
        break;
#endif
    default:
        AM_ILLEGAL_STATE();
    }

    set_rip(thread, process->instr_addr + entry_offset);
    set_rsp(thread, process->instr_addr + stack_offset + NOTIFICATION_STACK_SIZE);

    write_data(thread, process->instr_addr + args_offset, sizeof(notification_args_t), &thread->notification_args, 0);

    thread->state = AM_TS_Notification_Wait_Done;

    thread_run(thread);
}

/* Stop the thread as quickly as possible. It is ok for the thread to be
   stopped in a syscall. Returns 1 if the thread state was changed in any way,
   0 if not. */
static int
update_thread_stop(am_thread_t *thread)
{
    if (!thread_stopped(thread, 1))
        return thread_stop(thread);
    else
        return 0;
}

/* Run the thread. Returns 1 if the thread state was changed in any way, 0 if
   not. */
static int
update_thread_run(am_thread_t *thread)
{
    if (thread->pending_msgs > 0 && thread->msg_counter_addr != 0)
        /* We have pending messages for the thread. Stop it and update its
           message counter. */
        if (thread_stopped(thread, 1)) {
            update_msg_counter(thread);
            return 1;
        }
        else
            return update_thread_stop(thread);
    else
        if (!thread->debug_stopped)
            /* The thread is already running, no need to do anything. */
            return 0;
        else {
            /* The thread is stopped, resume it. */
            thread_run(thread);
            return 1;
        }
}

/* Stop the thread and make sure it is not stopped in a syscall. Returns 1 if
   the thread state was changed in any way, 0 if not.  */
static int
update_thread_stop_not_in_syscall(am_thread_t *thread)
{
    if (!thread_stopped(thread, 1))
        if (thread_exiting_syscall(thread))
            /* The thread is running in order to exit a syscall. No need to do
               anything. */
            return 0;
        else
            /* The thread is running, but not in order to exit a syscall. Stop
               it. */
            return thread_stop(thread);
    else if (thread_in_syscall(thread)) {
        if (thread_suspended(thread))
            /* The thread is suspended, there is nothing we can do. When the
               thread is continued we will trap the SYS_lwp_continue call,
               causing us to re-evaluate this decision. */
            return 0;
        else {
            /* The thread is stopped in a syscall. Make it exit the syscall. */
            thread_exit_syscall(thread);
            return 1;
        }
    }
    else
        /* The thread is stopped and not in a syscall, no need to do
           anything. */
        return 0;
}

/* Update the state of the thread. Returns 1 if the thread changed state (or if
   another thread change state, in the case of thread suspension) in any way, 0
   if not. */
static int
update_thread(am_process_t *process, am_thread_t *thread)
{
#if defined(__sun__)
    /* Check if someone is about to suspend or continue this thread. */
    if (thread->suspender_tid != 0) {
        switch (process->execution_state) {
        case AM_PES_Run:
            /* Ok to proceed with suspend or continue. */
            break;
        case AM_PES_Run_Excl:
        case AM_PES_Run_One:
        case AM_PES_Stop:
            /* The suspender will have to wait until the current attach or
               detach is finished.

               When the process is in "run excl" state it may switch to "stop"
               state at any moment. There must not be any suspend actvity in
               "stop" state, so we can therefore not start a suspend or
               continue in "run excl" state.  This thread could also be the
               thread that is attaching or detaching, and suspending it would
               then cause a deadlock.

               Allowing the suspend or continue in "run one" or "stop" state
               may cause this thread to access the hijacked page, plus the same
               deadlock hazard as in "run excl" state above. */
            goto delay_suspendee;
        }

        /* Get the suspender thread. */
        am_thread_t *suspender_thread = thread_by_tid(process, thread->suspender_tid);
        AM_ASSERT(suspender_thread != NULL);

        /* Check whether it wants to suspend or continue this thread. */
        switch (suspender_thread->suspend_action) {
        case AM_TSA_Suspend:
            /* Check what suspend state this thread is in. */
            switch (thread->suspend_state) {
            case AM_TSS_Continued:
                /* The thread is running. Stop it in order to check if it is in
                   a critical region in the sampler. */
                thread->suspend_state = AM_TSS_Suspending_Wait_Stop;
                return update_thread_stop(thread);
            case AM_TSS_Suspending_Wait_Stop:
                /* We are waiting for the thread to stop in order to check if
                   it is in a critical region in the sampler. */
                if (thread_stopped(thread, 1))
                    if (thread_in_critical_region(thread)) {
                        /* The thread is in a critical region, send a "suspend"
                           message to it and wait for a "suspend ready"
                           reply. */
                        am_message_buf_t *msg = (am_message_buf_t *)am_malloc(sizeof(am_message_buf_t));
                        msg->msg.pid = process->pid;
                        msg->msg.tid = thread->tid;
                        msg->msg.kind = AM_Msg_Suspend;
                        msg->msg.extra_data_len = 0;
			msg->extra_data = NULL;
                        thread_send(thread, msg, 1);
                        thread->suspend_state = AM_TSS_Suspending_Wait_Ready;
                        return 1;
                    }
                    else {
                        /* The thread is stopped and not in a critical
                           region. Wait for the suspender to finish. */
                        thread->suspend_state = AM_TSS_Suspending_Wait_Done;
                        return 1;
                    }
                else
                    /* The thread hasn't stopped yet. */
                    return update_thread_stop(thread);
            case AM_TSS_Suspending_Wait_Ready:
                /* Just run the thread until we get the "suspend ready" reply. */
                return update_thread_run(thread);
            case AM_TSS_Suspending_Wait_Done:
                /* Just wait for the suspender to finish. We must run the
                   thread, as lwp_suspend may hang if the thread is debug
                   stopped. */
                if (thread_stopped(thread, 1) &&
                    !thread_in_syscall(thread))
                    poke_thread(thread);
                else
                    return update_thread_stop_not_in_syscall(thread);
            case AM_TSS_Suspended:
                /* The thread is already suspended so it must be in a safe
                   state. Just wait for this additional suspend to finish. */
                thread->suspend_state = AM_TSS_Suspending_Wait_Done;
                return 1;
            case AM_TSS_Continuing_Wait_Done:
                /* We should not be in a continuing state when someone is
                   trying to suspend us! */
                AM_ILLEGAL_STATE();
            default:
                AM_ILLEGAL_STATE();
            }
            break;
        case AM_TSA_Continue:
            /* Check what suspend state this thread is in. */
            switch (thread->suspend_state) {
            case AM_TSS_Continued:
            case AM_TSS_Suspended:
                /* Continuing the thread is safe regardless of it it is
                   suspended or not. Just wait for the continuer to finish. */
                thread->suspend_state = AM_TSS_Continuing_Wait_Done;
                return 1;
            case AM_TSS_Suspending_Wait_Stop:
            case AM_TSS_Suspending_Wait_Ready:
            case AM_TSS_Suspending_Wait_Done:
                /* We should not be in a suspending state when someone is
                   trying to continue us. */
                AM_ILLEGAL_STATE();
            case AM_TSS_Continuing_Wait_Done:
                /* Just wait for the continuer to finish.  We must run the
                   thread, as lwp_continue may hang if the thread is debug
                   stopped. */
                return update_thread_run(thread);
            default:
                AM_ILLEGAL_STATE();
            }
        default:
            AM_ILLEGAL_STATE();
        }
    }
delay_suspendee:

    /* Check if the thread is about to suspend or continue another thread. */
    if (thread->suspendee_tid != 0) {
        switch (process->execution_state) {
        case AM_PES_Run:
            /* Ok to proceed with suspend or continue. */
            break;
        case AM_PES_Run_Excl:
        case AM_PES_Run_One:
        case AM_PES_Stop:
            /* We will have to wait with the suspend or continue until the
               current attach or detach is finished. Keep the thread stopped in
               the entry to the suspend or continue syscall.

               We know this thread isn't the one attaching or detaching and
               that it isn't holding any instrumenter locks, so keeping it
               stopped will not cause a deadlock.*/
            return update_thread_stop(thread);
        }

        /* Get the suspendee thread. */
        am_thread_t *suspendee_thread = thread_by_tid(process, thread->suspendee_tid);

        /* Proceed with the suspend or resume if the suspendee thread has
           terminated. */
        if (suspendee_thread == NULL)
            return update_thread_run(thread);

        /* Grab the suspendee thread if no one else has already done it. */
        if (suspendee_thread->suspender_tid == 0)
            suspendee_thread->suspender_tid = thread->tid;

        /* Check if we have grabbed the suspendee thread. */
        if (suspendee_thread->suspender_tid == thread->tid)
            /* Check if we are suspending or continuing the thread. */
            switch (thread->suspend_action) {
            case AM_TSA_Suspend:
                /* Wait until the suspendee is in "suspend wait done" state and
                   then continue with the suspend. */
                switch (suspendee_thread->suspend_state) {
                case AM_TSS_Continued:
                case AM_TSS_Suspended:
                case AM_TSS_Suspending_Wait_Stop:
                case AM_TSS_Suspending_Wait_Ready:
                    /* Wait, the suspendee is not ready. */
                    return update_thread_stop(thread);
                case AM_TSS_Suspending_Wait_Done:
                    /* The suspendee is ready, go for it! */
                    return update_thread_run(thread);
                case AM_TSS_Continuing_Wait_Done:
                    /* The suspendee should not be in a continuing state when
                       we are trying to suspend it. */
                    AM_ILLEGAL_STATE();
                default:
                    AM_ILLEGAL_STATE();
                }
            case AM_TSA_Continue:
                /* Wait until the suspendee is in "continue wait done" state
                   and then continue with the continue. */
                switch (suspendee_thread->suspend_state) {
                case AM_TSS_Continued:
                case AM_TSS_Suspended:
                    /* Wait, the suspendee is not ready. */
                    return update_thread_stop(thread);
                case AM_TSS_Continuing_Wait_Done:
                    /* The suspendee is ready, go for it! */
                    return update_thread_run(thread);
                case AM_TSS_Suspending_Wait_Stop:
                case AM_TSS_Suspending_Wait_Ready:
                case AM_TSS_Suspending_Wait_Done:
                    /* The suspendee should not be in a suspending state when
                       we are trying to continue it. */
                    AM_ILLEGAL_STATE();
                default:
                    AM_ILLEGAL_STATE();
                }
            default:
                AM_ILLEGAL_STATE();
            }
        else
            /* We haven't grabbed the suspendee yet, we have to wait for our
               turn. */
            return update_thread_stop(thread);
    }
#endif /* defined(__sun__) */

    /* Stop the process and update it's breakpoints as soon as possible if it
       needs a breakpoint update. */
    if (thread->update_bps && process->execution_state == AM_PES_Run) {
        if (thread_stopped(thread, 1))
            update_bps(thread);
        else
            /* TODO: Check if it is possible to change the breakpoint registers
               while the thread is in a syscall. Then we don't need to check
               for AM_PES_Run. */
            update_thread_stop_not_in_syscall(thread);
    }

    switch (process->execution_state) {
    case AM_PES_Run:
        switch (thread->state) {
        case AM_TS_Native:
            switch (process->instrumentation_state) {
            case AM_PIS_Attaching:
                /* If the thread is stopped, not in a syscall, not suspended
                   and not being suspended or resumed we should proceed with
                   stopping all other threads so that the thread can load the
                   instrumenter. */
                if (thread_stopped(thread, 1) &&
                    !thread_in_syscall(thread) &&
                    !thread_suspended(thread) &&
                    !suspend_activity(process)) {
                    thread->state = AM_TS_Load_Wait_All_Stopped;
                    process->execution_state = AM_PES_Stop;
                    return 1;
                }
                else
                    return update_thread_stop_not_in_syscall(thread);
            case AM_PIS_Idling:
            case AM_PIS_Starting_Burst:
                /* The thread is already native, let it run. */
                return update_thread_run(thread);
            case AM_PIS_Bursting:
                if (sigsegv_pending(thread))
                    /* The thread has a SIGSEGV pending that may be caused by
                       our self modifying code detection. We must let it run
                       and handle the notification if that is the case. */
                    return update_thread_run(thread);
                else if (thread_stopped(thread, 1) &&
                         !thread_in_syscall(thread) &&
                         !thread_suspended(thread) &&
                         !suspend_activity(process)) {
                    /* If the thread is stopped, not in a syscall, not
                       suspended and not being suspended or continued we should
                       proceed with stopping all other threads so that the
                       thread can be instrumented. */
                    thread->state = AM_TS_Instr_Wait_All_Stopped;
                    process->execution_state = AM_PES_Stop;
                    return 1;
                }
                else
                    return update_thread_stop_not_in_syscall(thread);
            case AM_PIS_Detaching:
                /* If the thread is stopped, not in a syscall, not suspended
                   and not being suspended or continued we should proceed with
                   stopping all other threads so that the thread can unload the
                   instrumenter. */
                if (thread_stopped(thread, 1) &&
                    !thread_in_syscall(thread) &&
                    !thread_suspended(thread) &&
                    !suspend_activity(process)) {
                    thread->state = AM_TS_Unload_Wait_All_Stopped;
                    process->execution_state = AM_PES_Stop;
                    return 1;
                }
                else
                    return update_thread_stop_not_in_syscall(thread);
            case AM_PIS_Native:
                if (am_exiting && all_threads_native(process, 1))
#if defined(__linux__)
                    if (thread->sigstop_pending)
                        /* We must let the thread run until it receives the
                           pending SIGSTOP before we can exit. */
                        return update_thread_run(thread);
                    else
#endif
                        /* We are about to exit and all threads are
                           native. Stop the thread so that we can detach it. */
                        return update_thread_stop(thread);
                else
                    /* The thread is already native but there are still other
                       instrumented threads. Let the thread run until they have
                       been deinstrumented. */
                    return update_thread_run(thread);
            default:
                AM_ILLEGAL_STATE();
            }
        case AM_TS_Deinstr_Wait_Ready:
            /* We have sent a "deinstr" message to the instrumenter, continue
               running the thread until we receive a "deinstr ready" reply. */
            if (process->instrumentation_state == AM_PIS_Starting_Burst) {
                /* If we are waiting to start a new burst we need to know if
                   the thread is in a syscall. To check this we stop the thread
                   every now and then. See ticket #986. */
                if (thread_stopped(thread, 1))
                    if (thread_in_syscall(thread))
                        return update_thread_stop_not_in_syscall(thread);
                    else {
                        int change = update_thread_run(thread);
                        usleep(10000);
                        return change;
                    }
                else
                    return update_thread_stop(thread);
            }
            else
                return update_thread_run(thread);
        case AM_TS_Deinstr_Wait_Done_1:
            /* We have sent a "deinstr go" message to the instrumenter and are
               waiting to receive a "deinstr done 1" reply. */
            return update_thread_run(thread);
        case AM_TS_Deinstr_Wait_Done_1_Run: {
            /* We have received a "deinstr done 1" message from the
               instrumenter. Switch to "run excl" execution mode to prevent any
               other threads from proceeding with instrumenting or
               deinstrumenting.

               If the thread is suspended or being suspended or continued we
               must wait until it has been continued, if there is any suspend
               activity going on we must wait until it is finished. */
            if(thread_suspended(thread) ||
               suspend_activity(process))
                return update_thread_run(thread);
            process->execution_state = AM_PES_Run_Excl;
            thread->state = AM_TS_Deinstr_Wait_Done_1_All_Stopped;
            return 1;
        }
        case AM_TS_Instrumented:
            switch (process->instrumentation_state) {
            case AM_PIS_Idling:
            case AM_PIS_Starting_Burst: {
                /* Send a "deinstr" message to the thread and then continue
                   running it until the instrumenter sends "deinstr ready"
                   reply. */
                am_message_buf_t *msg = (am_message_buf_t *)am_malloc(sizeof(am_message_buf_t));
                msg->msg.pid = process->pid;
                msg->msg.tid = thread->tid;
                msg->msg.kind = AM_Msg_Deinstr;
		msg->msg.extra_data_len = 0;
		msg->extra_data = NULL;
                thread_send(thread, msg, 1);
                thread->state = AM_TS_Deinstr_Wait_Ready;
                return 1;
            }
            case AM_PIS_Bursting:
                /* The thread is already instrumented, run it. */
                return update_thread_run(thread);
            default:
                AM_ILLEGAL_STATE();
            }
        case AM_TS_Instrumented_Exiting:
            /* Run the thread until is exits. */
            return update_thread_run(thread);
        case AM_TS_Instrumented_Wait_Ready:
            /* Run the thread until we get the "ready" message. */
            return update_thread_run(thread);
        case AM_TS_Notification_Wait:
            /* Wait until the thread is stopped an no other thread is
               processing a notification, then start processing it. */
            if (thread_stopped(thread, 1) &&
                !any_thread_in_notification(process)) {
#if defined(__sun__)
                /* If a signal is just about to be delivered trying to change
                   the pc of the thread may not work. Let the thread get into
                   the signal handler first. */
                if (signal_imminent(thread)) {
                    poke_thread(thread);
                    return 0;
                }
#endif
                run_notification(thread);
                return 1;
            }
            else
                return update_thread_stop(thread);
        case AM_TS_Notification_Wait_Done:
            /* Run the thread until it is finished. */
            return update_thread_run(thread);
        case AM_TS_Notification_Wait_Done_Stopped:
            /* Wait for the thread to stop and then restore its native
               state. */
            if (thread_stopped(thread, 1) &&
                !thread_in_syscall(thread)) {
                restore_registers(thread);

                if (thread->has_ret_val)
                    set_syscall_ret(thread, thread->ret_val);

                thread->state = AM_TS_Native;
                return 1;
            }
            else
                return update_thread_stop_not_in_syscall(thread);
        default:
            AM_ILLEGAL_STATE();
        }
    case AM_PES_Run_Excl:
        switch (thread->state) {
        case AM_TS_Native:
        case AM_TS_Instrumented:
        case AM_TS_Deinstr_Wait_Ready:
        case AM_TS_Deinstr_Wait_Done_1:
        case AM_TS_Deinstr_Wait_Done_1_Run:
        case AM_TS_Instrumented_Exiting:
        case AM_TS_Instrumented_Wait_Ready:
        case AM_TS_Notification_Wait_Done:
        case AM_TS_Notification_Wait_Done_Stopped:
            /* All threads should just continue running until the thread
               selected for detaching gets its "detach done 1" message. This
               prevents deadlocks with other threads holding locks while
               stopped. */
            return update_thread_run(thread);
        case AM_TS_Notification_Wait:
            /* We can't run threads waiting to process a notification. However,
               we know that they can't be holding any locks, so keeping them
               stopped won't cause any deadlocks. */
            return update_thread_stop(thread);
        case AM_TS_Deinstr_Wait_Done_1_All_Stopped:
            process->execution_state = AM_PES_Stop;
            return 1;
        default:
            AM_ILLEGAL_STATE();
        }
    case AM_PES_Run_One:
        switch (thread->state) {
        case AM_TS_Native:
        case AM_TS_Instrumented:
        case AM_TS_Instrumented_Exiting:
        case AM_TS_Instrumented_Wait_Ready:
        case AM_TS_Notification_Wait:
        case AM_TS_Notification_Wait_Done:
        case AM_TS_Notification_Wait_Done_Stopped:
            /* Another thread is bootstrapping or debootstrapping, so this
               thread must be stopped. */
            AM_ASSERT(thread_stopped(thread, 0));
            return 0;
        case AM_TS_Load_Wait_Hello_Stopped:
            /* We have received a "load hello" message from the instrumenter
               loader. If the thread is stopped we should relocate the
               instrumenter into the memory area the instrumenter loader has
               allocated, send the thread a "load welcome" message and continue
               running the instrumenter loader until we receive a "load ready"
               message. Otherwise we should stop the thread. */
            if (thread_stopped(thread, 1)) {
                /* Get the instrumenter segment info. */
                instr_seg_t *instr_rx_seg;
                instr_seg_t *instr_rw_seg;
                switch (process->bitness) {
//                case 32:
//                    instr_rx_seg = &instr_32_rx_seg;
//                    instr_rw_seg = &instr_32_rw_seg;
//                    break;
#if defined(__x86_64__)
                case 64:
                    instr_rx_seg = &instr_64_rx_seg;
                    instr_rw_seg = &instr_64_rw_seg;
                    break;
#endif
                default:
                    AM_ILLEGAL_STATE();
                }

                /* Relocate the instrumenter into the process. */
                relocate_instrumenter(process, thread);

                /* Send a "load welcome" message to the instrumenter loader
                   with all the info about the instrumenter image. */
                am_message_buf_t *msg = (am_message_buf_t *)am_malloc(sizeof(am_message_buf_t));
                msg->msg.pid = process->pid;
                msg->msg.tid = thread->tid;
                msg->msg.kind = AM_Msg_Load_Welcome;
		msg->msg.extra_data_len = 0;
                msg->msg.u.load_welcome.load_entry =
                    process->instr_addr + get_load_entry_offset(process);
                msg->msg.u.load_welcome.instr_rx_start =
                    process->instr_addr + instr_rx_seg->start;
                msg->msg.u.load_welcome.instr_rx_size = instr_rx_seg->size;
                msg->msg.u.load_welcome.instr_rw_start =
                    process->instr_addr + instr_rw_seg->start;
                msg->msg.u.load_welcome.instr_rw_size = instr_rw_seg->size;
                msg->msg.u.load_welcome.sampler_params = sampler_params;
		/* filename comes from the process */
		memcpy(msg->msg.u.load_welcome.sampler_params.trace_file_prefix,
		       process->trace_file_prefix, TRACE_FILE_PREFIX_MAX);
#if defined(__x86_64__)
		msg->msg.u.load_welcome.sampler_params.code_area_start = process->instr_addr + get_instr_image_size(process);
		msg->msg.u.load_welcome.sampler_params.code_area_end = process->instr_addr + preallocation_size;
#else
		msg->msg.u.load_welcome.sampler_params.code_area_start = 0;
		msg->msg.u.load_welcome.sampler_params.code_area_end = 0;
#endif
		msg->extra_data = NULL;
                thread_send(thread, msg, 0);

                /* Fill in the syscall arguments to the instrumenter loader's
                   sendto syscall, now that we know what fd it uses and have it
                   stopped. */
                syscall_args_t sendto_syscall_args;
                switch (process->bitness) {
/*                case 32:
                    set_syscall_args_32(
                        &sendto_syscall_args,
                        process->target_socket_fd,
                        process->hijack_page +
                        offsetof(instrumenter_loader_page_t, load_ready_msg),
                        sizeof(am_message_t), 0, 0, 0);
                    break;
*/
#if defined(__x86_64__)
                case 64:
                    set_syscall_args_64(
                        &sendto_syscall_args,
                        process->target_socket_fd,
                        process->hijack_page +
                        offsetof(instrumenter_loader_page_t, load_ready_msg),
                        sizeof(am_message_t), 0, 0, 0);
                    break;
#endif
                default:
                    AM_ILLEGAL_STATE();
                }
                write_data(thread, process->hijack_page +
                           offsetof(instrumenter_loader_page_t, sendto_syscall_args),
                           sizeof(syscall_args_t), &sendto_syscall_args, 0);

                /* Continue running the thread until we receive a "load ready"
                   message from it. */
                thread->state = AM_TS_Load_Wait_Ready;
                thread_run(thread);
                return 1;
            }
            else
                return update_thread_stop(thread);
        case AM_TS_Load_Wait_Ready:
            /* Thread is loading the instrumenter but we have not yet received
               a "load ready" message. Continue running the instrumenter
               loader. */
            return update_thread_run(thread);
        case AM_TS_Load_Wait_Ready_Stopped:
            /* The instrumenter loader is done. If the thread is stopped and
               not in a syscall we restore the hijacked page and the thread's
               native state and continue running the process. */
            if (thread_stopped(thread, 1) &&
                !thread_in_syscall(thread)) {
                write_data(thread, process->hijack_page, AM_PAGE_SIZE, &process->hijack_page_buf, 0);
                process->hijack_page = 0;
                set_hold_signals(thread, 0);
                restore_registers(thread);
                thread->state = AM_TS_Native;
                process->instrumentation_state = AM_PIS_Idling;
                process->execution_state = AM_PES_Run;
                return 1;
            }
            else
                return update_thread_stop_not_in_syscall(thread);
        case AM_TS_Instr_Wait_Ready:
            /* Thread is being instrumented but we have not yet received a
               "instr ready" message. Continue running the the thread
               instrumenter. */
            return update_thread_run(thread);
        case AM_TS_Instr_Wait_Ready_Stopped:
            /* We have received an "instr ready" message from the thread
               instrumenter. If the thread is stopped we should restore the
               hijacked page, send the thread an "instr go" message and
               continue running the process. Otherwise we should stop the
               thread. */
            if (thread_stopped(thread, 1)) {
                write_data(thread, process->hijack_page, AM_PAGE_SIZE, &process->hijack_page_buf, 0);
                process->hijack_page = 0;
                set_hold_signals(thread, 0);
                am_message_buf_t *msg = (am_message_buf_t *)am_malloc(sizeof(am_message_buf_t));
                msg->msg.pid = process->pid;
                msg->msg.tid = thread->tid;
                msg->msg.kind = AM_Msg_Instr_Go;
		msg->msg.extra_data_len = 0;
		msg->extra_data = NULL;
                thread_send(thread, msg, 0);
                thread->state = AM_TS_Instrumented;
                process->execution_state = AM_PES_Run;
                return 1;
            }
            else
                return update_thread_stop(thread);
        case AM_TS_Deinstr_Wait_Ready:
        case AM_TS_Deinstr_Wait_Done_1:
        case AM_TS_Deinstr_Wait_Done_1_Run:
            /* Another thread is bootstrapping or debootstrapping, so this
               thread must be stopped. */
            AM_ASSERT(thread_stopped(thread, 0));
            return 0;
        case AM_TS_Deinstr_Wait_Done_2:
            /* Continue running the thread until the debootstrapper is
               finished, i.e., until it sends a "detach done 2" message or
               closes the socket. */
            return update_thread_run(thread);
        case AM_TS_Deinstr_Wait_Done_2_Stopped:
            /* The debootstrapper is done. If the thread is stopped and not in
               a syscall we restore the hijacked page and continue running the
               process. */
            if (thread_stopped(thread, 1) && !thread_in_syscall(thread)) {
                write_data(thread, process->hijack_page, AM_PAGE_SIZE, &process->hijack_page_buf, 0);
                process->hijack_page = 0;
                set_hold_signals(thread, 0);
                restore_registers(thread);
                thread_run(thread);
                thread->state = AM_TS_Native;
                process->execution_state = AM_PES_Run;
                return 1;
            }
            else
                return update_thread_stop_not_in_syscall(thread);
        case AM_TS_Unload_Wait_Close:
            /* Continue running the thread until it closes the socket to
               indicate that it is done. */
            return update_thread_run(thread);
        case AM_TS_Unload_Wait_Close_Stopped:
            /* The thread has closed the socket to indicate that it is done. If
               the thread is stopped we should restore the hijacked page and
               the native state of the thread. Otherwise we should stop the
               thread. */
            if (thread_stopped(thread, 1) && !thread_in_syscall(thread)) {
                write_data(thread, process->hijack_page, AM_PAGE_SIZE, &process->hijack_page_buf, 0);
                process->hijack_page = 0;
                set_hold_signals(thread, 0);
                restore_registers(thread);
                thread->state = AM_TS_Native;
                process->execution_state = AM_PES_Run;
                process->instrumentation_state = AM_PIS_Native;
                return 1;
            }
            else
                return update_thread_stop_not_in_syscall(thread);
        default:
            AM_ILLEGAL_STATE();
        }
    case AM_PES_Stop:
        switch (thread->state) {
        case AM_TS_Native:
            if (!any_thread_instrumented(process) &&
                sigsegv_pending(thread))
                /* We can't unload the instrumenter while there are threads
                   that have SIGSEGVs pending that may be caused by our
                   self-modifying code detection. Run the thread until we get
                   to know the cause for the signal. */
                return update_thread_run(thread);
            else
                return update_thread_stop(thread);
        case AM_TS_Instrumented:
        case AM_TS_Instrumented_Exiting:
        case AM_TS_Instrumented_Wait_Ready:
            /* In stop exection state, all threads except the one waiting to
               bootstrap or debootstrap should stop as soon as possible. */
            return update_thread_stop(thread);
        case AM_TS_Load_Wait_All_Stopped:
            /* The thread is waiting for all other threads to stop to be able
               to load the instrumenter. Check if all other threads have
               stopped, and if so proceed with the instrumenter loading. */
            if (!all_threads_stopped(process))
                return 0;
#if defined(__sun__)
            /* If a signal is just about to be delivered trying to change the
               pc of the thread may not work. Let the thread get into the
               signal handler first. */
            if (signal_imminent(thread)) {
                poke_thread(thread);
                return 0;
            }
#endif
            run_instrumenter_loader(thread);
            thread->state = AM_TS_Load_Wait_Ready;
            process->execution_state = AM_PES_Run_One;
            return 1;
        case AM_TS_Instr_Wait_All_Stopped:
            /* The thread is waiting for all other threads to stop to be able
               to instrument itself. Check if all other threads have stopped,
               and if so proceed with the thread instrumentation. */
            if (!all_threads_stopped(process))
                return 0;
            if (sigsegv_pending(thread))
                /* We can't instrument the thread if it has a SIGSEGV from
                   self-modfying code detection pending. Run it and handle the
                   SIGSEGV first. */
                return update_thread_run(thread);
            else {
#if defined(__sun__)
                /* If a signal is just about to be delivered trying to change
                   the pc of the thread may not work. Let the thread get into
                   the signal handler first. */
                if (signal_imminent(thread)) {
                    poke_thread(thread);
                    return 0;
                }
#endif

                /* We can't instrument the thread if it is running in
                   instrumenter code, e.g., in an instrumenter signal
                   handler. Single step until we get out of the
                   instrumenter. */
                unsigned long rip = get_rip(thread);
                unsigned long instr_image_size = get_instr_image_size(process);
                if (rip >= process->instr_addr &&
                    rip < process->instr_addr + instr_image_size) {
                    thread->single_step = 1;
                    return update_thread_run(thread);
                }

                run_thread_instrumenter(thread);
                thread->state = AM_TS_Instr_Wait_Ready;
                process->execution_state = AM_PES_Run_One;
                return 1;
            }
        case AM_TS_Deinstr_Wait_Ready:
        case AM_TS_Deinstr_Wait_Done_1:
        case AM_TS_Deinstr_Wait_Done_1_Run:
            /* It is some other thread's turn to debootstrap, so we stop this
               thread. */
            return update_thread_stop(thread);
        case AM_TS_Deinstr_Wait_Done_1_All_Stopped:
            /* The thread is waiting for all other threads to stop to be able
               to debootstrap. Check if all threads have stopped and this
               thread is not in a syscall, and if so, proceed with
               debootstrapping. */
            if (all_threads_stopped(process) && thread_stopped(thread, 1) && !thread_in_syscall(thread)) {
#if defined(__sun__)
                /* If a signal is just about to be delivered trying to change
                   the pc of the thread may not work. Let the thread get into
                   the signal handler first. */
                if (signal_imminent(thread)) {
                    poke_thread(thread);
                    return 0;
                }
#endif
                run_thread_deinstrumenter(thread);
                thread->state = AM_TS_Deinstr_Wait_Done_2;
                process->execution_state = AM_PES_Run_One;
                return 1;
            }
            else
                return update_thread_stop_not_in_syscall(thread);
        case AM_TS_Unload_Wait_All_Stopped:
            /* The thread is waiting for all other threads to stop to be able
               to unload the instrumenter. Check if all threads have stopped,
               there are no threads with SIGSEGVs pending, no threads are
               processing or waiting to process notifications and this thread
               is not in a syscall, and if so, proceed with the unloading. */
            if (all_threads_stopped(process) &&
                thread_stopped(thread, 1) &&
                !any_sigsegv_pending(process) &&
                !any_thread_in_notification(process) &&
                !any_thread_in_notification_wait(process) &&
                !thread_in_syscall(thread)) {
#if defined(__sun__)
                /* If a signal is just about to be delivered trying to change
                   the pc of the thread may not work. Let the thread get into
                   the signal handler first. */
                if (signal_imminent(thread)) {
                    poke_thread(thread);
                    return 0;
                }
#endif
                run_instrumenter_unloader(thread);
                thread->state = AM_TS_Unload_Wait_Close;
                process->execution_state = AM_PES_Run_One;
                return 1;
            }
            else
                return update_thread_stop_not_in_syscall(thread);
        case AM_TS_Notification_Wait:
            if (!any_thread_instrumented(process) &&
                thread_stopped(thread, 1) &&
                !any_thread_in_notification(process)) {
#if defined(__sun__)
                /* If a signal is just about to be delivered trying to change
                   the pc of the thread may not work. Let the thread get into
                   the signal handler first. */
                if (signal_imminent(thread)) {
                    poke_thread(thread);
                    return 0;
                }
#endif
                /* We can't unload the instrumenter while there are threads
                   waiting to process notifications. We therefore start
                   processing the notification. Since there are no instrumented
                   threads, we know no locks are taken and no dead-locks can
                   occurs. */
                run_notification(thread);
                return 1;
            }
            else
                return update_thread_stop(thread);
        case AM_TS_Notification_Wait_Done:
            if (any_thread_instrumented(process))
                return update_thread_stop(thread);
            else
                /* We can't unload the instrumenter while there are threads
                   processing notifications. We therefore finish processing the
                   notification before stopping the thread. Since there are no
                   instrumented threads, we know no locks are taken and no
                   dead-locks can occurs. */
                return update_thread_run(thread);
        case AM_TS_Notification_Wait_Done_Stopped:
            if (any_thread_instrumented(process))
                return update_thread_stop(thread);
            else {
                /* We can't unload the instrumenter while there are threads
                   processing notifications. We therefore finish processing the
                   notification before stopping the thread. Since there are no
                   instrumented threads, we know no locks are taken and no
                   dead-locks can occurs. */
                if (thread_stopped(thread, 1) &&
                    !thread_in_syscall(thread)) {
                    restore_registers(thread);
                    thread->state = AM_TS_Native;
                    return 1;
                }
                else
                    return update_thread_stop_not_in_syscall(thread);
            }
        default:
            AM_ILLEGAL_STATE();
        }
    default:
        AM_ILLEGAL_STATE();
    }
}

/* Returns 1 if it is dangerous for the monitor to exit, otherwise returns 0.

   It is dangerous for the monitor to exit if any thread in any process is
   loading the instrumenter, instrumenting itself, deinstrumenting itself or
   unloading the instrumenter. Exiting the monitor in this state will leave the
   thread state and/or process memory in a modified state and therefore most
   likely crash or hang the process. */
static int
dangerous_to_exit()
{
    am_process_t *process;
    VECT_FOREACH(processes, process) {
        am_thread_t *thread;
        VECT_FOREACH(process->threads, thread)
            switch (thread->state) {
            case AM_TS_Load_Wait_Hello:
            case AM_TS_Load_Wait_Hello_Stopped:
            case AM_TS_Load_Wait_Ready:
            case AM_TS_Load_Wait_Ready_Stopped:
            case AM_TS_Instr_Wait_Ready:
            case AM_TS_Instr_Wait_Ready_Stopped:
            case AM_TS_Deinstr_Wait_Done_1:
            case AM_TS_Deinstr_Wait_Done_1_Run:
            case AM_TS_Deinstr_Wait_Done_1_All_Stopped:
            case AM_TS_Deinstr_Wait_Done_2:
            case AM_TS_Deinstr_Wait_Done_2_Stopped:
            case AM_TS_Unload_Wait_Close:
            case AM_TS_Unload_Wait_Close_Stopped:
            case AM_TS_Notification_Wait:
            case AM_TS_Notification_Wait_Done:
            case AM_TS_Notification_Wait_Done_Stopped:
                return 1;
            case AM_TS_Native:
            case AM_TS_Load_Wait_All_Stopped:
            case AM_TS_Instr_Wait_All_Stopped:
            case AM_TS_Instrumented:
            case AM_TS_Instrumented_Exiting:
            case AM_TS_Instrumented_Wait_Ready:
            case AM_TS_Deinstr_Wait_Ready:
            case AM_TS_Unload_Wait_All_Stopped:
                break;
            default:
                AM_ILLEGAL_STATE();
            }
    }
    return 0;
}

static void
sigterm_event(void)
{
    am_printf(AM_Ctrl_Msg_Info, "Received SIGTERM, exiting.\n");
    trigger_exit(AM_Exit_Normal);
}

static void
burst_time_event(void *data)
{
    am_process_t *process = (am_process_t *)data;
    process->state = AM_PS_Burst;
}

static void
detach_time_event(void *nodata)
{
    am_printf(AM_Ctrl_Msg_Info, "Sampling duration passed, stopping sampling and exiting.\n");
    trigger_exit(AM_Exit_Normal);
}

/* Called when the process has been attached to, either due to a
 * timing or breakpoint event, or immediately when the initial process
 * is launched. */
static void
trigger_attach(am_process_t *process)
{

    /* A detach may have been triggered before the attach is triggered, e.g.,
       if the number of threads in the process exceeds the number allowed by
       the license. Do nothing in that case. */
    if (!process->exit_triggered) {
	if (process_may_be_instrumented(process)) {

	    /* Remember that we have seen interesting processes for
	     * later diagnostic output */
	    seen_interesting_process = 1;

	    /* First time? Create an acq file, add first event */
	    if (process->acq_handle == 0) {
		update_open_acquisition_info(process);
		write_acquisition_start(process->acq_handle, process,
					process->pid);
	    }

	    process->state = AM_PS_Burst;
	    
	    if (stop_at_address != 0) {
		am_thread_t *thread;
		VECT_FOREACH(process->threads, thread) {
		    thread_add_stop_at_bps(thread);
		}
	    }
	    else if (sampling_duration != 0)
		add_time_event(sampling_duration * (am_time_t)1000000000,
			       detach_time_event, NULL);
	}
    }
    
    process->attach_triggered = 1;
}

static void
attach_time_event(void *data)
{
    am_process_t *process = (am_process_t *) data;
    am_printf(AM_Ctrl_Msg_Info, "Sampling delay passed, starting sampling.\n");
    trigger_attach(process);
    append_acquisition_mappings(process->acq_handle, process);
}

static void
socket_event(short events, void *process_)
{
    am_thread_t *thread;
    am_process_t *process = (am_process_t *)process_;
    am_message_t msg;
    void *extra_data = NULL;
    ssize_t msg_len = safe_recv(process->socket_fd, &msg, sizeof(msg), 0);
    AM_ASSERT_WITH_ERRNO(msg_len != -1 || errno == ECONNRESET);
    if (msg_len == 0 || msg_len == -1)
	goto socket_closed;

    AM_ASSERT(msg_len == sizeof(msg));
    if (msg.extra_data_len > 0) {
	extra_data = malloc(msg.extra_data_len);
	AM_ASSERT_WITH_ERRNO(extra_data != NULL);
	ssize_t data_len = safe_recv(process->socket_fd, extra_data,
				     msg.extra_data_len, 0);
	AM_ASSERT_WITH_ERRNO(msg_len != -1 || errno == ECONNRESET);
	if (msg_len == 0 || msg_len == -1)
	    goto socket_closed;
	AM_ASSERT(data_len == msg.extra_data_len);
    }

    VECT_FOREACH(process->threads, thread) {
        if (thread->tid == msg.tid) {
            switch (msg.kind) {
            case AM_Msg_Load_Ready:
                /* The thread has finished loading the instrumenter. Wait for
                   it to stop and then restore the hijack page and the thread's
                   native state. */
                thread->state = AM_TS_Load_Wait_Ready_Stopped;
                break;
            case AM_Msg_Instr_Ready: {
                thread->stack_addr = msg.u.instr_ready.stack_addr;
                thread->msg_counter_addr = msg.u.instr_ready.msg_counter_addr;
		thread->noevpend_addr = msg.u.instr_ready.noevpend_addr;
                switch (thread->state) {
                case AM_TS_Instr_Wait_Ready:
                    /* The instrumentation of the thread has finished. Wait for
                       the thread to stop so we can restore the hijacked
                       page. */
                    thread->state = AM_TS_Instr_Wait_Ready_Stopped;
                    break;
                case AM_TS_Instrumented_Wait_Ready: {
                    /* The thread is cloned from an instrumented thread. Now we
                       have all information needed to treat it as a regular
                       instrumented thread. Send it a "go" message so it can
                       continue. */
                    thread->state = AM_TS_Instrumented;
                    am_message_buf_t *msg = (am_message_buf_t *)am_malloc(sizeof(am_message_buf_t));
                    msg->msg.pid = process->pid;
                    msg->msg.tid = thread->tid;
                    msg->msg.kind = AM_Msg_Instr_Go;
		    msg->msg.extra_data_len = 0;
		    msg->extra_data = NULL;
                    thread_send(thread, msg, 0);
                } break;
                default:
                    AM_ILLEGAL_STATE();
                }
            } break;
            case AM_Msg_Deinstr_Ready: {
                am_message_buf_t *reply_msg = (am_message_buf_t *)am_malloc(sizeof(am_message_buf_t));
                reply_msg->msg.pid = process->pid;
                reply_msg->msg.tid = thread->tid;
                reply_msg->msg.extra_data_len = 0;
                reply_msg->extra_data = NULL;

                switch (process->state) {
                case AM_PS_Idle:
                case AM_PS_Detach:
                    thread->state = AM_TS_Deinstr_Wait_Done_1;
                    thread->regs = msg.u.deinstr_ready.regs;
                    reply_msg->msg.kind = AM_Msg_Deinstr_Go;
                    break;
                case AM_PS_Burst:
                    thread->state = AM_TS_Instrumented;
                    check_new_burst_generation(process);
                    reply_msg->msg.kind = AM_Msg_Deinstr_Cancel;
                    reply_msg->msg.u.deinstr_cancel.burst_generation = process->burst_generation;
                    break;
                default:
                    AM_ILLEGAL_STATE();
                }

                thread_send(thread, reply_msg, 0);
            } break;
            case AM_Msg_Deinstr_Done_1:
                thread->state = AM_TS_Deinstr_Wait_Done_1_Run;
                break;
            case AM_Msg_Deinstr_Done_2:
                thread->state = AM_TS_Deinstr_Wait_Done_2_Stopped;
                break;
            case AM_Msg_Exit: {
                thread->state = AM_TS_Instrumented_Exiting;
                thread->msg_counter_addr = 0;
		thread->noevpend_addr = 0;
                am_message_buf_t *msg = (am_message_buf_t *)am_malloc(sizeof(am_message_buf_t));
                msg->msg.pid = process->pid;
                msg->msg.tid = thread->tid;
                msg->msg.kind = AM_Msg_Exit_Go;
		msg->msg.extra_data_len = 0;
		msg->extra_data = NULL;
                thread_send(thread, msg, 0);
            } break;
            case AM_Msg_BP : {
                VECT_APPEND(stop_at_instrum_pcs, unsigned long, msg.u.bp.addr);
                if (process->attach_triggered && !process->exit_triggered) {
                    am_thread_t *t;
                    VECT_FOREACH(process->threads, t)
                        add_bp(t, stop_at_bp_id, msg.u.bp.addr, stop_at_callback);
                }
                am_message_buf_t *msg = (am_message_buf_t *)am_malloc(sizeof(am_message_buf_t));
                msg->msg.pid = process->pid;
                msg->msg.tid = thread->tid;
                msg->msg.kind = AM_Msg_BP_Go;
		msg->msg.extra_data_len = 0;
		msg->extra_data = NULL;
                thread_send(thread, msg, 0);
            } break;
            case AM_Msg_BP_Inv : {
                int found;
                VECT_REMOVE(stop_at_instrum_pcs, msg.u.bp_inv.addr, found);
                AM_ASSERT(found);
                if (process->attach_triggered && !process->exit_triggered) {
                    am_thread_t *t;
                    VECT_FOREACH(process->threads, t)
                        remove_bp_by_id(t, stop_at_bp_id, msg.u.bp_inv.addr);
                }
                am_message_buf_t *msg = (am_message_buf_t *)am_malloc(sizeof(am_message_buf_t));
                msg->msg.pid = process->pid;
                msg->msg.tid = thread->tid;
                msg->msg.kind = AM_Msg_BP_Inv_Go;
		msg->msg.extra_data_len = 0;
		msg->extra_data = NULL;
                thread_send(thread, msg, 0);
            } break;
#if defined(__sun__)
            case AM_Msg_Suspend_Ready:
                thread->suspend_state = AM_TSS_Suspending_Wait_Done;
                thread->suspend_spinlock_addr = msg.u.suspend_ready.spinlock_addr;
                break;
#endif
            case AM_Msg_Protect_Page: {
                process->pages->protect(msg.u.protect_page.addr);

                am_message_buf_t *msg = (am_message_buf_t *)am_malloc(sizeof(am_message_buf_t));
                msg->msg.pid = process->pid;
                msg->msg.tid = thread->tid;
                msg->msg.kind = AM_Msg_Protect_Page_Go;
		msg->msg.extra_data_len = 0;
		msg->extra_data = NULL;
                thread_send(thread, msg, 0);
            } break;
            case AM_Msg_Unprotect_Page: {
                process->pages->unprotect(msg.u.unprotect_page.addr);

                am_message_buf_t *msg = (am_message_buf_t *)am_malloc(sizeof(am_message_buf_t));
                msg->msg.pid = process->pid;
                msg->msg.tid = thread->tid;
                msg->msg.kind = AM_Msg_Unprotect_Page_Go;
		msg->msg.extra_data_len = 0;
		msg->extra_data = NULL;
                thread_send(thread, msg, 0);
            } break;
            case AM_Msg_Notification_Done:
                thread->has_ret_val = msg.u.notification_done.has_ret_val;
                thread->ret_val =  msg.u.notification_done.ret_val;
                thread->state = AM_TS_Notification_Wait_Done_Stopped;
                break;
	    case AM_Msg_Log:
		if (extra_data) {
		    char *message = (char *)am_malloc(msg.extra_data_len + 1);
		    memcpy(message, extra_data, msg.extra_data_len);
		    message[msg.extra_data_len] = '\0';
		    am_printf((am_control_message_level_t)msg.u.log.level, "sampler: %s", message);
		    am_free(message);
		} else
		    am_printf((am_control_message_level_t)msg.u.log.level, "sampler: (NULL)\n");
		break;
	    case AM_Msg_Abort:
		am_printf(AM_Ctrl_Msg_Error, "sampler aborted: pcdiff %" PRId64 "\n",
			  msg.u.abort.pcdiff);
		break;
            case AM_Msg_Burst_End:
                process->state = AM_PS_Idle;
                add_time_event(burst_interval, burst_time_event, process);
                break;
            default:
                AM_ILLEGAL_STATE();
            }
	    if (extra_data)
		free(extra_data);
            return;
        }
    }

    AM_ILLEGAL_STATE();

socket_closed:
    /* The socket to the instrumenter has been closed. This can be caused
       by a number of things:

       * The instrumenter has been unloaded.

       * The last thread is exiting, or exit_group has been called. We should
         have received an "exiting" message, so exit_msg should be set for all
         threads.

       * The process has been killed without getting a chance to notify the
         monitor, e.g., by a SIGKILL. */

    close_socket(process);
    VECT_FOREACH(process->threads, thread)
	if (thread->state == AM_TS_Unload_Wait_Close)
	    thread->state = AM_TS_Unload_Wait_Close_Stopped;
    prune_processes();
    return;
}

static void
orphan_socket_handler(short events, void *data)
{
    AM_ASSERT(events == POLLIN);

    int fd = *(int *)data;

    am_message_t msg;
    int res = safe_recv(fd, &msg, sizeof(msg), 0);
    AM_ASSERT_WITH_ERRNO(res != -1);
    AM_ASSERT(res == sizeof(msg));
    AM_ASSERT(msg.kind == AM_Msg_Load_Hello);
    AM_ASSERT(msg.extra_data_len == 0);
    am_process_t *process;
    VECT_FOREACH(processes, process)
        if (process->pid == msg.pid) {
            am_thread_t *thread;
            VECT_FOREACH(process->threads, thread)
                if (thread->tid == msg.tid) {
                    AM_ASSERT(process->socket_fd == -1);
                    process->socket_fd = fd;
                    process->target_socket_fd = msg.u.load_hello.target_socket_fd;
                    process->instr_addr = msg.u.load_hello.instr_addr;
                    add_fd_event(process->socket_fd, POLLIN, socket_event,
                                 process);
                    thread->state = AM_TS_Load_Wait_Hello_Stopped;

                    remove_fd_event(fd, POLLIN, orphan_socket_handler, data);
                    am_free(data);
                    return;
                }
        }
    AM_ILLEGAL_STATE();
}

static void
server_socket_event(short events, void *data)
{
    AM_ASSERT(data == NULL);
    struct sockaddr_un from;
    socklen_t fromlen = sizeof(from);
    int fd = safe_accept(server_socket_fd, (struct sockaddr *)&from, &fromlen);
    AM_ASSERT_WITH_ERRNO(fd != -1);

    int *fd_data = (int *)am_malloc(sizeof(int));
    *fd_data= fd;
    add_fd_event(fd, POLLIN, orphan_socket_handler, fd_data);
}

static void
destroy_server_socket(void)
{
    if (server_socket_fd != -1) {
        remove_fd_event(server_socket_fd, POLLIN, server_socket_event, NULL);
        int res = safe_close(server_socket_fd);
        AM_ASSERT_WITH_ERRNO(res != -1);
    }
}

static void
setup_server_socket(void)
{
    int res;
    server_socket_fd = socket(PF_UNIX, SOCK_STREAM, 0);
    AM_ASSERT_WITH_ERRNO(server_socket_fd != -1);

    /* Make the socket close itself in child processes */
    int flags = fcntl(server_socket_fd, F_GETFD);
    AM_ASSERT_WITH_ERRNO(flags != -1);
    flags = fcntl(server_socket_fd, F_SETFD, flags | FD_CLOEXEC);
    AM_ASSERT_WITH_ERRNO(flags != -1);

    struct sockaddr_un addr;
    mode_t old_umask;
    old_umask = umask(unsupported_attach ?
		      (S_IXUSR | S_IXGRP | S_IXOTH) :
		      (S_IXUSR |
		       S_IRGRP | S_IWGRP | S_IXGRP |
		       S_IROTH | S_IWOTH | S_IXOTH));

    for (int i = 0; i < 1000000;  i++) {
        addr.sun_family = AF_UNIX;
        res = snprintf(server_socket_path.chars, AM_PATH_MAX,
                       "%s/acumon-%" PRIpid "-%d",
		       tmpdir.c_str(), getpid(), i);
        AM_ASSERT(res >= 0 && res < AM_PATH_MAX);
        AM_ASSERT(strlen(server_socket_path.chars) < sizeof(addr.sun_path));
        strncpy(addr.sun_path, server_socket_path.chars,
                sizeof(addr.sun_path));
        res = bind(server_socket_fd, (struct sockaddr *)&addr,
                   sizeof(addr.sun_family) + strlen(addr.sun_path));
        if (res != -1)
            goto bound;
    }
    am_printf(AM_Ctrl_Msg_Error,
	      "%s:%i: %s: bind %s failed, error %s\n",
	      __FILE__, __LINE__, __FUNCTION__, addr.sun_path, strerror(errno));
    control_exit(AM_Exit_Unknown);
    exit(AM_Exit_Unknown);
    /*NOTREACHED*/
bound:
    umask(old_umask);
    res = listen(server_socket_fd, 1);
    AM_ASSERT_WITH_ERRNO(res != -1);
    add_fd_event(server_socket_fd, POLLIN, server_socket_event, NULL);
}

static void
control_socket_handle_packet(am_control_packet_t *packet)
{
    switch (packet->kind) {
    case AM_Ctrl_Detach:
	switch(packet->u.detach.type) {
	case AM_Ctrl_Detach_Normal:
	    trigger_exit(AM_Exit_Normal);
	    break;
	case AM_Ctrl_Detach_Force:
	    if (!dangerous_to_exit())
		force_exit(AM_Exit_Normal);
	    else
		queue_force_exit = 1;
	    break;
	case AM_Ctrl_Detach_Force_Unsafe:
	    force_exit(AM_Exit_Normal);
	    break;
	default:
	    AM_ILLEGAL_STATE();
	}
	break;

    default:
	/* Unhandled packet, skip it... */
	break;
    }
}

static void
control_socket_event(short events, void *data)
{
    if(events & POLLIN) {
	am_control_packet_t packet;
	int ret;

	/* Try to receive as many packets as possible without
	   blockinging */
	while ((ret = control_receive_packet(control_socket_fd, 0, &packet)) > 0) {
	    control_socket_handle_packet(&packet);
	    control_free_packet_data(&packet);
	}
	if (ret == -1 && errno != EAGAIN && errno != EWOULDBLOCK)
	    /* IO Error, e.g. socket closed */
	    force_exit(AM_Exit_Ctrl_Socket);
	else if (ret == -2)
	    /* Not enough data on the socket (shouldn't happen) */
	    force_exit(AM_Exit_Ctrl_Protocol);
    }

    if(events & POLLHUP) {
	/* The control application shouldn't normally close the
	   control socket. */
	force_exit(AM_Exit_Ctrl_Socket);
    }
}

static void
control_do_handshake()
{
    am_control_packet_t packet;
    int ret;

    ret = control_receive_packet(control_socket_fd, 1, &packet);
    if (ret <= 0)
	force_exit(AM_Exit_Ctrl_Socket);

    if (packet.kind != AM_Ctrl_Hello)
	force_exit(AM_Exit_Ctrl_Protocol);

    if (packet.u.hello.major != AM_CTRL_MAJOR) {
	control_send_protocol_mismatch(control_socket_fd, 1);
	force_exit(AM_Exit_Ctrl_Version);
    } else {
	control_send_hello_response(control_socket_fd, 1,
				    AM_CTRL_MAJOR, AM_CTRL_MINOR);
    }

    control_free_packet_data(&packet);
}

static void
control_exit(am_exit_status_t status)
{
    if(control_socket_fd != -1)
	control_send_exit(control_socket_fd, 1, status);
}

static void
control_do_setup()
{
    am_control_packet_t packet;
    int ret;

    while(1) {
	ret = control_receive_packet(control_socket_fd, 1, &packet);
	if (ret <= 0)
	    force_exit(AM_Exit_Ctrl_Socket);

	switch (packet.kind) {
	case AM_Ctrl_Protocol_Mismatch:
	    force_exit(AM_Exit_Ctrl_Version);
	case AM_Ctrl_Go:
	    control_free_packet_data(&packet);
	    return;
	default:
	    /* Unhandled packet, skip it... */
	    break;
	}

	control_free_packet_data(&packet);
    }
}

static void
destroy_control_socket(void)
{
    if(control_socket_fd != -1) {
        remove_fd_event(control_socket_fd, POLLIN | POLLHUP, control_socket_event, NULL);
	int res = safe_close(control_socket_fd);
	control_socket_fd = -1;
	AM_ASSERT_WITH_ERRNO(res != -1);
    }
}

static void
open_control_socket(void)
{
    int res;
    control_socket_fd = socket(PF_UNIX, SOCK_SEQPACKET, 0);
    AM_ASSERT_WITH_ERRNO(control_socket_fd != -1);
    struct sockaddr_un addr;
    addr.sun_family = AF_UNIX;
    strcpy(addr.sun_path, control_socket_path);
    res = safe_connect(control_socket_fd, (struct sockaddr *)&addr,
                       sizeof(addr.sun_family) + strlen(addr.sun_path));
    if (res == -1)
	force_exit(AM_Exit_Ctrl_Socket);
    add_fd_event(control_socket_fd, POLLIN | POLLHUP, control_socket_event, NULL);
}

static void
attach_processes()
{
    am_process_t *process;
    VECT_FOREACH(processes, process) {
        if (process->pid != 0) {
            if ((!sampler_params.demo && !feature_attach) && !process->exit_triggered) {
                am_printf(AM_Ctrl_Msg_Info, "License does not allow attaching "
                          "to an existing process.\n");
                trigger_exit(AM_Exit_License_Exceeded);
            }

	    /* 1 means to also attach to ptrace */
            attach_process(process, 1);
        }
        if (process->binary.size() != 0)
            launch_process(process);

        AM_ASSERT(process->elf_file != NULL);
        if (licensing_check_manifest(process->elf_file) != 0)
            trigger_exit(AM_Exit_License_Exceeded);
    }
}

static void
update_process(am_process_t *process)
{
    int repeat;
    do {
        repeat = 0;

        switch (process->state) {
        case AM_PS_Idle:
            switch (process->instrumentation_state) {
            case AM_PIS_Attaching:
            case AM_PIS_Idling:
            case AM_PIS_Detaching:
                /* If the process is idling we don't need to do anything. If
                   the process is attaching or detaching we can't do anything
                   now, the state transition will instead be done when the
                   attach or detach is finished. */
                break;
            case AM_PIS_Bursting:
                if (burst_interval != 0)
                    append_burst_end(process->acq_handle,
				     process->burst_generation);
                /* FALL THROUGH */
            case AM_PIS_Starting_Burst:
                /* Switch to idling state. */
                process->instrumentation_state = AM_PIS_Idling;
                repeat = 1;
                break;
            case AM_PIS_Native:
                /* Switch to attaching state. */
                process->instrumentation_state = AM_PIS_Attaching;
                repeat = 1;
                break;
            default:
                AM_ILLEGAL_STATE();
            }
            break;
        case AM_PS_Burst:
            switch (process->instrumentation_state) {
            case AM_PIS_Attaching:
            case AM_PIS_Bursting:
            case AM_PIS_Detaching:
                /* If the process is sampling a burst we don't need to do
                   anything. If the process is attaching or detaching we can't
                   do anything now, the state transition will instead be done
                   when the attach or detach is finished. */
                break;
            case AM_PIS_Starting_Burst:{
                /* If all threads are either uninstrumented or in a syscall we
                   can start a new burst without mixing burst generations. */
                am_thread_t *thread;
                VECT_FOREACH(process->threads, thread)
                    if (!thread_native(thread, 0) &&
                        !(!thread_stopped(thread, 1) &&
                          thread_exiting_syscall(thread)))
                        goto not_ready;
                process->instrumentation_state = AM_PIS_Bursting;
                process->start_new_burst = 1;
                repeat = 1;
                not_ready:;
            } break;
            case AM_PIS_Idling:
                /* Switch to starting burst state. */
                process->instrumentation_state = AM_PIS_Starting_Burst;
                repeat = 1;
                break;
            case AM_PIS_Native:
                /* Switch to attaching state. */
                process->instrumentation_state = AM_PIS_Attaching;
                repeat = 1;
                break;
            default:
                AM_ILLEGAL_STATE();
            }
            break;
        case AM_PS_Detach:
            switch (process->instrumentation_state) {
            case AM_PIS_Attaching:
            case AM_PIS_Detaching:
            case AM_PIS_Native:
                /* If the process is native we don't need to do anything. If
                   the process is attaching or detaching we can't do anything
                   now, the state transition will instead be done when the
                   attach or detach is finished. */
                break;
            case AM_PIS_Idling:
                /* If all threads are uninstrumented we can switch to detaching
                   state to start unloading the instrumenter. Otherwise remain
                   in idling state and wait for the threads to deinstrument
                   themselves. */
                if (all_threads_native(process, 1)) {
                    process->instrumentation_state = AM_PIS_Detaching;
                    repeat = 1;
                }
                break;
            case AM_PIS_Bursting:
                if (burst_interval != 0)
                    append_burst_end(process->acq_handle,
				     process->burst_generation);
                /* FALL THROUGH */
            case AM_PIS_Starting_Burst:
                /* Switch to idling state to start deinstrumenting the
                   threads. */
                process->instrumentation_state = AM_PIS_Idling;
                repeat = 1;
                break;
            default:
                AM_ILLEGAL_STATE();
            }
            break;
        default:
            AM_ILLEGAL_STATE();
        }

        am_thread_t *thread;
        VECT_FOREACH(process->threads, thread)
            if (update_thread(process, thread))
                repeat = 1;
    } while (repeat);

    if ((am_exiting && process->instrumentation_state == AM_PIS_Native) ||
        VECT_SIZE(process->threads) == 0) {
        am_thread_t *thread;
        VECT_FOREACH(process->threads, thread)
            if (!thread_detachable(thread))
                goto thread_not_ready;

        while (VECT_SIZE(process->threads) > 0) {
            am_thread_t *thread = VECT_GET(process->threads, 0);
            detach_thread(thread);
            remove_thread(thread);
        }

        kill_if_launched(process);

        remove_process(process);
    }
thread_not_ready:
    ;
}

static void
update_processes(void)
{
    am_process_t *process;
    VECT_FOREACH(processes, process)
        update_process(process);

    if (VECT_SIZE(processes) == 0) {

	// Suppress this output unless we have actually seen a process
	// that we wanted to sample. See TS-2239 for rationale

	if (seen_interesting_process) {	
	    if (am_exiting)
		am_printf(AM_Ctrl_Msg_Info, "All threads detached.\n");
	    else
		am_printf(AM_Ctrl_Msg_Info, "All monitored processes have exited.\n");
	}

	if (control_socket_fd != -1)
	    control_send_sampling_stopped(control_socket_fd, 1,
					  (am_exiting ?
					   AM_Ctrl_Sampling_Stopped_Detached :
					   AM_Ctrl_Sampling_Stopped_Normal));
        exit_acumon();
    }
    else if (queue_force_exit && dangerous_to_exit())
	force_exit(AM_Exit_Normal);
}

static void
print_prompt(void)
{
    fprintf(stderr, "Type q and enter to stop the sampling> ");
    fflush(stdout);
}

static void
stdin_event(short events, void *data)
{
    AM_ASSERT(data == NULL);

    int reprint_prompt = 0;
    ssize_t res;
    char c;
    while ((res = safe_read(STDIN_FILENO, &c, 1)) == 1) {
        reprint_prompt = 1;
        switch(c) {
        case 'q':
            if (!am_exiting) {
                fprintf(stderr, "Exiting.\n");
                trigger_exit(AM_Exit_Normal);
            }
            else {
                do {
                    res = safe_read(STDIN_FILENO, &c, 1);
                    AM_ASSERT_WITH_ERRNO(res != -1 || errno == EAGAIN);
                } while (res == 1);

                if (dangerous_to_exit())
                    fprintf(stderr, "There are currently threads being attached or detached. Exiting now will leave "
                           "the memory of the instrumented process corrupted and is likely crash it.\n\n");
                else
                    fprintf(stderr, "There are still instrumented threads. These will be automatically "
                           "de-instrumented but may leak some memory in the instrumented process if you "
                           "exit now.\n\n");

                fprintf(stderr, "Are you sure you want to exit anyway? [y/n] ");
                fflush(stdout);
                char c;
                do {
                    res = safe_read(STDIN_FILENO, &c, 1);
                    AM_ASSERT_WITH_ERRNO(res != -1 || errno == EAGAIN);
                } while (res != 1);
                if (c == 'y')
                    force_exit(AM_Exit_Normal);
            }
            break;
        case 'i':
            if (DEBUG_COMMANDS) {
                AM_ASSERT(VECT_SIZE(processes) == 1);
                am_process_t *process = VECT_GET(processes, 0);
                process->state = AM_PS_Idle;
            }
            break;
        case 's':
            if (DEBUG_COMMANDS) {
                AM_ASSERT(VECT_SIZE(processes) == 1);
                am_process_t *process = VECT_GET(processes, 0);
                process->state = AM_PS_Burst;
            }
            break;
        case 'd':
            if (DEBUG_COMMANDS) {
                AM_ASSERT(VECT_SIZE(processes) == 1);
                am_process_t *process = VECT_GET(processes, 0);
                process->state = AM_PS_Detach;
            }
            break;
        }
    }
    AM_ASSERT_WITH_ERRNO(res != -1 || errno == EAGAIN);

    if (reprint_prompt)
        print_prompt();
}

static void
destroy_stdin_handler(void)
{
    if (!own_terminal || control_socket_path != NULL)
        return;

    remove_fd_event(STDIN_FILENO, POLLIN, stdin_event, NULL);
}

static void
setup_stdin_handler(void)
{
    if (!own_terminal || control_socket_path != NULL)
        return;

    int res;

    res = fcntl(STDIN_FILENO, F_SETFL, O_NONBLOCK);
    AM_ASSERT_WITH_ERRNO(res != -1);

    am_process_t *process;
    VECT_FOREACH(processes, process)
        if (process->binary.size() != 0)
            return;

    add_fd_event(STDIN_FILENO, POLLIN, stdin_event, NULL);

    print_prompt();
}

static void
print_usage(const char *argv0)
{
    fprintf(stderr,
            "Usage: %s [OPTION]...\n", argv0);
    fprintf(stderr,
            "  --pid PID                      Attach to and sample the specified process.\n");
    fprintf(stderr,
            "  --binary BINARY                Execute and sample the specified binary.\n");
    fprintf(stderr,
            "  --argument ARGUMENT            Pass ARGUMENT as an argument to the binary\n"
            "                                 specified with --binary. Can occur multiple\n"
            "                                 times, the arguments are then passed to the\n"
            "                                 binary in the order they occur.\n");
    fprintf(stderr,
            "  --sample-period ACCESSES       Sample period in number of memory accesses.\n"
            "                                 The default is 1000000 accesses.\n"
            "                                 Used as initial sample period if automatic\n"
            "                                 sample period adjustment is used.\n");
    fprintf(stderr,
            "  --line-sizes BYTES[,BYTES]...  Line sizes to sample in bytes. The default is\n"
            "                                 1, 32, 64 and 4096 byte lines.\n");
    fprintf(stderr,
            "  --atf-file-prefix PREFIX       Prefix added to the file name of the ATF file.\n"
            "                                 The default is no prefix.\n");
    fprintf(stderr,
            "  --acquisition-info FILENAME    Write acquisition info to the specified file.\n");
    fprintf(stderr,
            "  --sampling-delay DELAY         Start sampling after DELAY seconds.\n");
    fprintf(stderr,
            "  --sampling-duration DURATION   Detach and exit after sampling for DURATION\n"
            "                                 seconds.\n");
    fprintf(stderr,
            "  --sample-count COUNT           Use automatic sample period adjustment and\n"
            "                                 collect the specified number of samples.\n");
    fprintf(stderr,
            "  --start-at-function FUNCTION   Start sampling when the specified function is\n"
            "                                 executed.\n");
    fprintf(stderr,
            "  --start-at-address ADDRESS     Start sampling when the instruction at the\n"
            "                                 specified address is executed.\n");
    fprintf(stderr,
            "  --start-at-ignore COUNT        Ignore the first COUNT times the sampling\n"
            "                                 start location (set with --start-at-function\n"
            "                                 or --start-at-address) is reached.\n");
    fprintf(stderr,
            "  --stop-at-function FUNCTION    Stop sampling when the specified function is\n"
            "                                 executed.\n");
    fprintf(stderr,
            "  --stop-at-address ADDRESS      Stop sampling when the instruction at the\n"
            "                                 specified address is executed.\n");
    fprintf(stderr,
            "  --stop-at-ignore COUNT         Ignore the first COUNT times the sampling\n"
            "                                 stop location (set with --stop-at-function\n"
            "                                 or --stop-at-address) is reached.\n");
    fprintf(stderr,
	    "  --control-socket FILENAME      Path to an existing control socket.\n");
    fprintf(stderr,
            "  --safe-stack                   Disable optimization that may cause problems\n"
            "                                 with programs with non-standard stack handling.\n");
#if defined(__x86_64__)
    fprintf(stderr,
            "  --preallocation SIZE           Changes preallocation size for x86-64 targets\n"
            "                                 to SIZE in MB. Default is 2048 MB.\n");
#endif
    fprintf(stderr,
            "  --burst-sampling ACCESSES      The length of the burst sampling period in\n"
            "                                 accesses. Requires SlowSpotter Pro.\n");
    fprintf(stderr,
            "  --burst-cooldown ACCESSES      The length of the burst cooldown period in\n"
            "                                 accesses.\n");
    fprintf(stderr,
            "  --burst-interval MS            The interval in milliseconds from the end of\n"
            "                                 one burst to the beginning of the next.\n");
    fprintf(stderr,
            "  --use-target-stderr            Allow logging to the target applications\n"
	    "                                 stderr.\n");
    fprintf(stderr,
	    "  --unsupported-attach           Disables UID checking and makes the sampler\n"
            "                                 socket world accessible. This is unsupported\n"
	    "                                 and insecure.\n");
    fprintf(stderr,
	    "  --min-generation               Name the first program generation to\n"
	    "                                 be sampled, default=0.\n");
    fprintf(stderr,
	    "  --max-generation               Name the last program generation to\n"
	    "                                 be sampled, default=0.\n");

    fprintf(stderr,
	    "  --tmpdir                       Directory for storing temporary files.\n");

    /* The demo option should probably be undocumented */
#if 0
    fprintf(stderr,
            "  --demo                         Run the sampler in demo mode. Generated sample\n"
            "                                 files will not be compatible with Acumem VPE.\n");
#endif
    fprintf(stderr,
            "  --help                         Print this help message and exit.\n");
}

static void
check_multi_opt_error(int *seen, const char *opt, const char *argv0)
{
    if (*seen) {
        fprintf(stderr, "%s: duplicate option `%s'\n", argv0, opt);
        force_exit(AM_Exit_Usage);
    }
    *seen = 1;
}

unsigned long long
get_integer_option(const char *val_str, const char *opt, const char *argv0,
                   unsigned long long min_value, unsigned long long max_value)
{
    char *end;
    unsigned long long val = strtoll(val_str, &end, 0);
    if (*end != '\0') {
        fprintf(stderr, "%s: argument '%s' to option %s is not an integer\n",
                argv0, val_str, opt);
        force_exit(AM_Exit_Usage);
    }
    if (val < min_value || val > max_value) {
        fprintf(stderr, "%s: argument '%s' to option %s is out of range\n",
                argv0, val_str, opt);
        force_exit(AM_Exit_Usage);
    }
    return val;
}

static void
parse_args(int argc, char **argv)
{
    enum optvals {
        opt_pid,
        opt_binary,
        opt_argument,
        opt_sample_period,
        opt_line_sizes,
        opt_atf_file_prefix,
        opt_acquisition_info,
        opt_sampling_delay,
        opt_sampling_duration,
        opt_sample_count,
        opt_start_at_function,
        opt_start_at_address,
        opt_start_at_ignore,
        opt_stop_at_function,
        opt_stop_at_address,
        opt_stop_at_ignore,
	opt_control_socket,
        opt_safe_stack,
	opt_preallocation,
        opt_burst_sampling,
        opt_burst_cooldown,
        opt_burst_interval,
	opt_target_stderr,
	opt_unsupported_attach,
	opt_min_generation,
	opt_max_generation,
	opt_tmpdir,
	opt_demo,
        opt_help
    };

    static const struct option opts[] = {
        {"pid",               required_argument, NULL, opt_pid              },
        {"binary",            required_argument, NULL, opt_binary           },
        {"argument",          required_argument, NULL, opt_argument         },
        {"sample-period",     required_argument, NULL, opt_sample_period    },
        {"line-sizes",        required_argument, NULL, opt_line_sizes       },
        {"atf-file-prefix",   required_argument, NULL, opt_atf_file_prefix  },
        {"acquisition-info",  required_argument, NULL, opt_acquisition_info },
        {"sampling-delay",    required_argument, NULL, opt_sampling_delay   },
        {"sampling-duration", required_argument, NULL, opt_sampling_duration},
        {"sample-count",      required_argument, NULL, opt_sample_count     },
        {"start-at-function", required_argument, NULL, opt_start_at_function},
        {"start-at-address",  required_argument, NULL, opt_start_at_address },
        {"start-at-ignore",   required_argument, NULL, opt_start_at_ignore  },
        {"stop-at-function",  required_argument, NULL, opt_stop_at_function },
        {"stop-at-address",   required_argument, NULL, opt_stop_at_address  },
        {"stop-at-ignore",    required_argument, NULL, opt_stop_at_ignore   },
        {"control-socket",    required_argument, NULL, opt_control_socket   },
        {"safe-stack",        no_argument,       NULL, opt_safe_stack       },
        {"preallocation",     required_argument, NULL, opt_preallocation    },
        {"burst-sampling",    required_argument, NULL, opt_burst_sampling   },
        {"burst-cooldown",    required_argument, NULL, opt_burst_cooldown   },
        {"burst-interval",    required_argument, NULL, opt_burst_interval   },
	{"use-target-stderr", no_argument,       NULL, opt_target_stderr    },
	{"unsupported-attach",no_argument,       NULL, opt_unsupported_attach},
	{"min-generation",    required_argument, NULL, opt_min_generation   },
	{"max-generation",    required_argument, NULL, opt_max_generation   },
        {"demo",              no_argument,       NULL, opt_demo             },
	{"tmpdir",	      required_argument, NULL, opt_tmpdir           },
        {"help",              no_argument,       NULL, opt_help             },
        {NULL}};

    int seen_pid = 0;
    int seen_binary = 0;
    int seen_sample_period = 0;
    int seen_line_sizes = 0;
    int seen_atf_file_prefix = 0;
    int seen_acquisition_info = 0;
    int seen_sampling_delay = 0;
    int seen_sampling_duration = 0;
    int seen_sample_count = 0;
    int seen_start_at_function = 0;
    int seen_start_at_address = 0;
    int seen_start_at_ignore = 0;
    int seen_stop_at_function = 0;
    int seen_stop_at_address = 0;
    int seen_stop_at_ignore = 0;
    int seen_safe_stack = 0;
    int seen_burst_sampling = 0;
    int seen_burst_cooldown = 0;
    int seen_burst_interval = 0;
    int seen_target_stderr = 0;
    int seen_unsupported_attach = 0;
    int seen_min_generation = 0;
    int seen_max_generation = 0;
    int seen_demo = 0;
    int seen_control_socket = 0;
    int seen_tmpdir = 0;

    pid_t pid = 0;
    std::string binary;

    tmpdir = acumem::exe_path::get_temp_path();

    int opt;
    while ((opt = getopt_long(argc, argv, "", opts, NULL)) != -1)
        switch (opt) {
        case opt_pid: {
            check_multi_opt_error(&seen_pid, "--pid", argv[0]);
            pid_t tid = get_integer_option(optarg, "--pid", argv[0], 0, INT_MAX);
            pid = get_pid(tid);
            if (pid == -1) {
                fprintf(stderr, "Error: No such process: %" PRIpid "\n", tid);
                force_exit(AM_Exit_PID);
            }
        } break;
        case opt_binary: {
            check_multi_opt_error(&seen_binary, "--binary", argv[0]);
            am_process_t *process;
            VECT_FOREACH(processes, process)
                AM_ASSERT(process == NULL);
	    binary = optarg;
            own_terminal = 0;
        } break;
        case opt_argument: {
            VECT_APPEND(arguments, char *, optarg);
        } break;
        case opt_sample_period: {
            check_multi_opt_error(&seen_sample_period, "--sample-period", argv[0]);
            sampler_params.sample_period = get_integer_option(optarg, "--sample-period", argv[0], 0, MAX_SAMPLE_PERIOD);
        } break;
        case opt_line_sizes: {
            check_multi_opt_error(&seen_line_sizes, "--line-sizes", argv[0]);
            int i;
            char *endptr = optarg;
            for (i = 0; i < MAX_LINE_SIZES && *endptr != '\0'; i++)  {
                unsigned long line_size = strtol(endptr, &endptr, 10);
                if (!is_power_of_2(line_size)) {
                    fprintf(stderr, "Error: Line size %ld is not power of 2.\n", line_size);
                    force_exit(AM_Exit_Usage);
                }
                /* Can not allow sampling of 1-byte or 4k-byte lines. See
                   ticket #694. */
                if (line_size < 2 || line_size > 2048) {
                    fprintf(stderr, "Error: Can not sample line sizes less "
                            "than 2 bytes or greater than 2048 bytes, but %ld "
                            "bytes line size requested.\n", line_size);
                    force_exit(AM_Exit_Usage);
                }
                sampler_params.log2linesizes[i] = am_log2(line_size);
                switch (*endptr) {
                case ',':
                    endptr++;
                    break;
                case '\0':
                    break;
                default:
                    fprintf(stderr, "Error: Unsupporter character '%c' in line "
                            "sizes argument: %s\n", *endptr, optarg);
                    force_exit(AM_Exit_Usage);
                }
            }
            if (*endptr != '\0') {
                fprintf(stderr, "Error: At most 7 line sizes may be sampled "
                        "at the same time.\n");
                force_exit(AM_Exit_Usage);
            }
            sampler_params.log2linesizes[i] = -1;
        } break;
        case opt_atf_file_prefix:
            check_multi_opt_error(&seen_atf_file_prefix, "--atf-file-prefix", argv[0]);
	    atf_file_prefix = optarg;
            break;
        case opt_acquisition_info:
            check_multi_opt_error(&seen_acquisition_info, "--acquisition-info", argv[0]);
            acquisition_info_path = optarg;
            break;
        case opt_sampling_delay:
            check_multi_opt_error(&seen_sampling_delay, "--sampling-delay", argv[0]);
            sampling_delay = get_integer_option(optarg, "--sampling-delay", argv[0], 0, INT_MAX);
            break;
        case opt_sampling_duration:
            check_multi_opt_error(&seen_sampling_duration, "--sampling-duration", argv[0]);
            sampling_duration = get_integer_option(optarg, "--sampling-duration", argv[0], 0, INT_MAX);
            break;
        case opt_sample_count:
            check_multi_opt_error(&seen_sample_count, "--sample-count", argv[0]);
            sampler_params.sample_count = get_integer_option(optarg, "--sample-count", argv[0], 0, INT_MAX);
            break;
        case opt_start_at_function:
            check_multi_opt_error(&seen_start_at_function, "--start-at-function", argv[0]);
            start_at_function = optarg;
            break;
        case opt_start_at_address:
            check_multi_opt_error(&seen_start_at_address, "--start-at-address", argv[0]);
            start_at_address = get_integer_option(optarg, "--start-at-address", argv[0], 1, ULONG_MAX);
            break;
        case opt_start_at_ignore:
            check_multi_opt_error(&seen_start_at_ignore, "--start-at-ignore", argv[0]);
            start_at_ignore = get_integer_option(optarg, "--start-at-ignore", argv[0], 0, INT_MAX);
            break;
        case opt_stop_at_function:
            check_multi_opt_error(&seen_stop_at_function, "--stop-at-function", argv[0]);
            stop_at_function = optarg;
            break;
        case opt_stop_at_address:
            check_multi_opt_error(&seen_stop_at_address, "--stop-at-address", argv[0]);
            stop_at_address = get_integer_option(optarg, "--stop-at-address", argv[0], 1, ULONG_MAX);
            break;
        case opt_stop_at_ignore:
            check_multi_opt_error(&seen_stop_at_ignore, "--stop-at-ignore", argv[0]);
            stop_at_ignore = get_integer_option(optarg, "--stop-at-ignore", argv[0], 0, INT_MAX);
            break;
	case opt_control_socket:
            check_multi_opt_error(&seen_control_socket, "--control-socket", argv[0]);
	    control_socket_path = optarg;
	    break;
	case opt_safe_stack:
            check_multi_opt_error(&seen_safe_stack, "--safe-stack", argv[0]);
            sampler_params.safe_stack = 1;
	    break;
	case opt_preallocation:
#if defined(__x86_64__)
            preallocation_size = get_integer_option(optarg, "--preallocation", argv[0],
						    PREALLOCATION_MIN >> 20,
						    PREALLOCATION_MAX >> 20);
	    preallocation_size = preallocation_size << 20;
#else
	    fprintf(stderr, "%s: Ignoring --preallocation, it has no effect on 32-bit installations.\n",
		    argv[0]);
#endif
	    break;
        case opt_burst_sampling:
            check_multi_opt_error(&seen_burst_sampling, "--burst-sampling", argv[0]);
            sampler_params.burst_sampling_length = get_integer_option(
                optarg, "--burst-sampling", argv[0], 1, 1000000000000000LL);
            break;
        case opt_burst_cooldown:
            check_multi_opt_error(&seen_burst_cooldown, "--burst-cooldown", argv[0]);
            sampler_params.burst_cooldown_length = get_integer_option(
                optarg, "--burst-cooldown", argv[0], 1, 1000000000000000LL);
            break;
        case opt_burst_interval:
            check_multi_opt_error(&seen_burst_interval, "--burst-interval", argv[0]);
            burst_interval = 1000000 * get_integer_option(
                optarg, "--burst-interval", argv[0], 1, INT_MAX);
            break;
	case opt_target_stderr:
	    check_multi_opt_error(&seen_target_stderr, "--use-target-stderr", argv[0]);
	    sampler_params.use_stderr = 1;
	    break;
	case opt_unsupported_attach:
	    check_multi_opt_error(&seen_unsupported_attach, "--unsupported-attach", argv[0]);
	    unsupported_attach = 1;
	    break;
	case opt_min_generation:
	    check_multi_opt_error(&seen_min_generation, "--min-generation", argv[0]);
	    min_generation = get_integer_option(
		optarg, "--min-generation", argv[0], 0, 1000000000000000LL);
	    break;
	case opt_max_generation:
	    check_multi_opt_error(&seen_max_generation, "--max-generation", argv[0]);
	    max_generation = get_integer_option(
		optarg, "--max-generation", argv[0], 0, 1000000000000000LL);
	    break;
	case opt_tmpdir:
	    check_multi_opt_error(&seen_tmpdir, "--tmpdir", argv[0]);
	    tmpdir = optarg;
	    break;
	case opt_demo:
            check_multi_opt_error(&seen_demo, "--demo", argv[0]);
            sampler_params.demo = 1;
	    break;
        case opt_help:
            print_usage(argv[0]);
            force_exit(AM_Exit_Normal);
        default:
            print_usage(argv[0]);
            force_exit(AM_Exit_Usage);
        }
    if (optind != argc) {
        fprintf(stderr, "%s: unsupported non-option argument `%s'\n", argv[0], argv[optind]);
        print_usage(argv[0]);
        force_exit(AM_Exit_Usage);
    }
    if ((seen_pid && seen_binary) || (!seen_pid && !seen_binary)) {
        fprintf(stderr, "%s: must specify exactly one of `--pid' or `--binary'\n", argv[0]);
        force_exit(AM_Exit_Usage);
    }
    if (VECT_SIZE(arguments) != 0 && !seen_binary) {
        fprintf(stderr, "%s: '--argument' can only be used with '--binary'\n", argv[0]);
        force_exit(AM_Exit_Usage);
    }
    if (seen_sampling_delay + seen_start_at_function + seen_start_at_address > 1) {
        fprintf(stderr, "%s: only one of '--sampling-delay', '--start-at-function' and '--start-at-address' can be specified\n", argv[0]);
        force_exit(AM_Exit_Usage);
    }
    if (seen_sampling_duration + seen_stop_at_function + seen_stop_at_address > 1) {
        fprintf(stderr, "%s: only one of '--sampling-duration', '--stop-at-function' and '--stop-at-address' can be specified\n", argv[0]);
        force_exit(AM_Exit_Usage);
    }
    if (control_socket_path && strlen(control_socket_path) + 1 > SIZEOF_FIELD(struct sockaddr_un, sun_path)) {
	fprintf(stderr, "%s: provided control socket path too long.\n", argv[0]);
	force_exit(AM_Exit_Usage);
    }
    if ((seen_burst_sampling && !seen_burst_cooldown) ||
        (seen_burst_cooldown && !seen_burst_interval) ||
        (seen_burst_interval && !seen_burst_sampling)) {
	fprintf(stderr, "%s: either none or all of '--burst-sampling', '--burst-cooldown' and '--burst-interval'  must be specified.\n", argv[0]);
	force_exit(AM_Exit_Usage);
    }
    if (seen_burst_sampling && seen_sample_count) {
        fprintf(stderr, "%s: only one of '--burst-sampling' and '--sample-count' can be specified.\n", argv[0]);
	force_exit(AM_Exit_Usage);
    }
    if (seen_burst_sampling)
        /* Burst cooldown length is meassured from the start of the burst
           internally, but from the end of the sampling in the arguments. */
        sampler_params.burst_cooldown_length +=
            sampler_params.burst_sampling_length;
    if (sampler_params.demo && !seen_control_socket) {
	fprintf(stderr, "%s: '--demo' requires a valid control socket.\n", argv[0]);
	force_exit(AM_Exit_Usage);
    }

    if (seen_pid)
	add_process(pid, 0, std::string(), 0);
    else if (seen_binary)
	add_process(0, 0, binary, 0);

    /* Update the process' ptrace purpose after all command line
       parameters have been read. */
    am_process_t *process;
    VECT_FOREACH(processes, process)
	process->ptrace_purpose = calculate_ptrace_purpose(process);
}

static int
get_vernum_part(const char **p)
{
    int num = 0;
    while (1){
        char c = **p;
        if (c == '.' || c == '\0') {
            (*p)++;
            return num;
        }
        else if (c >= '0' && c <= '9') {
            num = num * 10 + c - '0';
            (*p)++;
        }
        else
            AM_ILLEGAL_STATE();
    }
}

static void
check_os(void)
{
    struct utsname utsname;
    int res = uname(&utsname);
    AM_ASSERT_WITH_ERRNO(res >= 0);
    const char *p = utsname.release;
    int major = get_vernum_part(&p);
    int minor = get_vernum_part(&p);
    check_os_version(major, minor);
}

static void
debug_mode_time_event(void *nodata)
{
    am_process_t *process;
    VECT_FOREACH(processes, process) {
        int r = random() % 1000;
        switch (process->state) {
        case AM_PS_Idle:
            if (r < 900)
                process->state = AM_PS_Burst;
            else
                process->state = AM_PS_Detach;
            break;
        case AM_PS_Burst:
            if (r < 900)
                process->state = AM_PS_Idle;
            else
                process->state = AM_PS_Detach;
            break;
        case AM_PS_Detach:
            if (r < 500)
                process->state = AM_PS_Idle;
            else
                process->state = AM_PS_Burst;
            break;
        default:
            AM_ILLEGAL_STATE();
        }
    }

    if (++debug_mode_switches % 100 == 0)
        am_printf(AM_Ctrl_Msg_Debug, "%d switches.\n", debug_mode_switches);

    add_time_event(random() % (2 * DEBUG_MODE_SWITCH_INTERVAL) + 1,
                   debug_mode_time_event, NULL);
}

void
setup_start_conditions(am_process_t *process)
{
    if (process_may_be_instrumented(process)) {
	if (sampling_delay != 0)
	    add_time_event(sampling_delay * (am_time_t)1000000000,
			   attach_time_event, process);
	else if (start_at_address != 0) {
	    am_thread_t *thread;
	    VECT_FOREACH(process->threads, thread)
		add_bp(thread, start_at_bp_id,
		       start_at_address, start_at_callback);
	}
	else
	    trigger_attach(process);
    }
}

int
main(int argc, char **argv)
{
    event_init();
    signal_handling_init();
    os_dependent_init();
    check_os();
    add_fatal_signal_handler(SIGABRT, sigabrt_handler);
    add_signal_event(SIGTERM, sigterm_event);
    if (SIGINT_DEBUG)
        add_signal_handler(SIGINT, sigint_debug);
    if (DEBUG_MODE)
        add_time_event(0, debug_mode_time_event, NULL);
    parse_args(argc, argv);
    if (!sampler_params.demo) {
        am_exit_status_t exit_status;
	if (licensing_init(&exit_status))
            exit(exit_status);
    }
    if (control_socket_path) {
	open_control_socket();
	control_do_handshake();
	control_do_setup();
    }
/*    load_instrumenter(instr_32_bin, instr_32_bin_size,

                      &instr_32_image, &instr_32_image_size,
                      &instr_32_rx_seg, &instr_32_rw_seg,
                      &load_entry_offset_32,
                      &instrumentation_entry_offset_32,
                      &unload_entry_offset_32,
                      &notification_entry_offset_32,
                      &notification_stack_offset_32,
                      &notification_args_offset_32,
                      &instr_32_rels, 32);
*/
#if defined(__x86_64__)
    load_instrumenter(instr_64_bin, instr_64_bin_size,
                      &instr_64_image, &instr_64_image_size,
                      &instr_64_rx_seg, &instr_64_rw_seg,
                      &load_entry_offset_64,
                      &instrumentation_entry_offset_64,
                      &unload_entry_offset_64,
                      &notification_entry_offset_64,
                      &notification_stack_offset_64,
                      &notification_args_offset_64,
                      &instr_64_rels,  64);
#endif
    setup_server_socket();

    /* If acquisition info files are requested, set those up here */
    am_process_t *process;
    
    attach_processes();
    setup_stdin_handler();

    start_at_bp_id = next_breakpoint_id++;
    stop_at_bp_id = next_breakpoint_id++;

    if (start_at_function != NULL) {
        AM_ASSERT(VECT_SIZE(processes) == 1);
        start_at_address = lookup_symbol(VECT_GET(processes, 0),
                                         start_at_function);
    }
    if (stop_at_function != NULL) {
        AM_ASSERT(VECT_SIZE(processes) == 1);
        stop_at_address = lookup_symbol(VECT_GET(processes, 0),
                                        stop_at_function);
    }
    sampler_params.bp = stop_at_address;

    VECT_FOREACH(processes, process)
	setup_start_conditions(process);

    while (1) {
        update_processes();
        wait_for_event();
    }
}
