/// @file acumon/acumon.h

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#ifndef ACUMON_HH
#define ACUMON_HH

#include "../sampler/src/am_registers.h"

#include "library/control/control.h"

#include "acumon_defs.h"
#include "protocol.h"
#include "os_dependent_types.hh"
#include "vect.hh"
#include "pages.hh"

#include <string>

/* Set to 1 to print some debug info when a monitored process receives
   a SIGSEGV. */
#define SIGSEGV_DEBUG 0

#define ARRAY_SIZE(x) (sizeof(x) /sizeof((x)[0]))

/* Number of execution breakpoints supported. */
#define NUM_BPS 4

/* Sampler parameters. */
extern sampler_params_t sampler_params;

/* Typedefs for commonly used structs. */
typedef struct am_process am_process_t;
typedef struct am_thread am_thread_t;
typedef struct acq_handle_t acq_handle_t;

/* Type of execution breakpoint callback functions. */
typedef void (*bp_cb_t)(am_thread_t *thread, uint64_t bp_id, int bp_num);

/* An execution breakpoint. If callback == 0 the breakpoint is disabled. */
typedef struct {
    /* Breakpoint id. */
    uint64_t bp_id;

    /* Address to break at. */
    unsigned long address;

    /* Function to call when the breakpoint is triggered. */
    bp_cb_t callback;

    /* True if the breakpoint has been suspended in order to step over an
       instruction that triggered the breakpoint. */
    int suspended;
} bp_t;

/* Buffer of words on a page, for use with ptrace. */
typedef struct {
    long words[AM_PAGE_SIZE / WORD_SIZE];
} page_buf_t;

typedef enum {
    /* The thread is not instrumented. */
    AM_TS_Native,



    /* Loading the instrumenter. Waiting for all threads to stop so that we can
       hijack a page for the instrumenter loader code and start the loading. */
    AM_TS_Load_Wait_All_Stopped,

    /* Loading the instrumenter. We are waiting for a "load hello" message from
       the instrumenter loader. */
    AM_TS_Load_Wait_Hello,

    /* Loading the instrumenter. We have received a "load hello" message from
       the instrumenter loader and are waiting for the thread to stop so that
       we can relocate the instrumenter into the area allocated by the
       bootstrapper. */
    AM_TS_Load_Wait_Hello_Stopped,

    /* Loading the instrumenter. Waiting for the "load ready" message from the
       instrumenter loader. */
    AM_TS_Load_Wait_Ready,

    /* Loading the instrumenter. We have received a "load ready" message from
       the instrumenter loader and are waiting for the thread to stop so that
       we can restore the hijacked page and the native state of the thread. */
    AM_TS_Load_Wait_Ready_Stopped,



    /* Instrumenting the thread. Waiting for all threads to stop so that we can
       hijack a page for the thread instrumenter code and start the
       instrumentation. */
    AM_TS_Instr_Wait_All_Stopped,

    /* Instrumenting the thread. Waiting for the "instr ready" message from the
       thread instrumenter. */
    AM_TS_Instr_Wait_Ready,

    /* Instrumenting the thread. We have received a "instr ready" message and
       are waiting for the thread to stop so that we can restore the hijacked
       page. */
    AM_TS_Instr_Wait_Ready_Stopped,



    /* The thread is instrumented. */
    AM_TS_Instrumented,



    /* Deinstrumenting. We have sent the thread a "deinstr" message and are
       waiting for the "deinstr ready" reply. */
    AM_TS_Deinstr_Wait_Ready,

    /* Deinstrumenting. We have sent the thread a "deinstr go" and are waiting
       for a "deinstr done 1" reply. */
    AM_TS_Deinstr_Wait_Done_1,

    /* Deinstrumenting. We have received a "deinstr done 1" message from the
       instrumenter. Wait for "run" execution state before switching to "run
       excl" execution state. This prevents other threads from instrumenting or
       deinstrumenting concurrently with the deinstrumentation of this
       thread. */
    AM_TS_Deinstr_Wait_Done_1_Run,

    /* Deinstrumenting. We have received a "deinstr done 1" message from the
       instrumenter and have switched to "run excl" execution state. We are now
       waiting for all threads to stop so that we can hijack a page for the
       thread deinstrumenter and start running it. */
    AM_TS_Deinstr_Wait_Done_1_All_Stopped,

    /* Deinstrumenting. The thread deinstrumenter is running and we are waiting
       for a "deinstr done 2" message. */
    AM_TS_Deinstr_Wait_Done_2,

    /* Deinstrumenting. We have received a "deinstr done 2 " message from the
       debootstrapper and are waiting for the thread to stop so that we can
       restore the hijacked page and the thread's native state. */
    AM_TS_Deinstr_Wait_Done_2_Stopped,



    /* Unloading the instrumenter. Waiting for all threads to stop so that we
       can hijack a page for the instrumenter unloader code and start the
       unloading. */
    AM_TS_Unload_Wait_All_Stopped,

    /* Unloading the instrumenter. Waiting for the socket to be closed to
       indicate that the unloading has finished. */
    AM_TS_Unload_Wait_Close,

    /* Unloading the instrumenter. The socket has been closed and we are
       waiting for the thread to stop so that we can restore the hijacked page
       and the native state of the thread. */
    AM_TS_Unload_Wait_Close_Stopped,



    /* The thread is instrumented and is about to exit. */
    AM_TS_Instrumented_Exiting,



    /* The thread has been cloned from an already instrumented thread, and we
       are waiting for the "ready" message from it. */
    AM_TS_Instrumented_Wait_Ready,



    /* The thread is waiting for its turn to process a notification. */
    AM_TS_Notification_Wait,

    /* The thread is processing a notification, wait for it to send
       "notification done" message to indicate it has finished. */
    AM_TS_Notification_Wait_Done,

    /* The thread has finished processing a notification. Wait for it to stop
       so that we can restore its native state. */
    AM_TS_Notification_Wait_Done_Stopped
} am_thread_state_t;

#if defined(__sun__)

/* Suspension state of a thread. */
typedef enum {
    /* The thread is not suspended. */
    AM_TSS_Continued,

    /* The thread is being suspended. We are stopping the thread to check if it
       is in a critical region in the sampler.*/
    AM_TSS_Suspending_Wait_Stop,

    /* The thread is being suspended. We have stopped the thread and determined
       it was in a critical region in the sampler, sent it a "suspend" message
       and resumed it, and are now waiting for it to send a "suspend ready"
       message to confirm it is out of the critical region. */
    AM_TSS_Suspending_Wait_Ready,

    /* The thread is ready to be suspended and just waiting for the suspender
       to finish. */
    AM_TSS_Suspending_Wait_Done,

    /* The thread is suspended. */
    AM_TSS_Suspended,

    /* The thread is being continued and just waiting for the continuer to
       finish. */
    AM_TSS_Continuing_Wait_Done
} am_thread_suspend_state_t;

/* Suspend action of the thread. */
typedef enum {
    /* The thread is suspending another thread. */
    AM_TSA_Suspend,

    /* The thread is continuing another thread. */
    AM_TSA_Continue
} am_thread_suspend_action_t;

#endif /* defined(__sun__) */

/* Thread struct, there is one for each ptrace attached thread in each
   monitored process.

   A thread is considered ptrace attached once a PTRACE_ATTACH has returned
   successfully, once we have received a SIGTRAP caused by a PTRACE_TRACEME and
   an exec by a launched thread, or once we get a PTRACE_EVENT_CLONE
   notification to a cloned thread's parent.

   A thread is no longer considered ptrace attached once a PTRACE_DETACH has
   return successfully. */
struct am_thread {
    /* Process the thread belongs to. */
    am_process_t *process;

    /* Thread id. */
    tid_t tid;

    /* Thread state. */
    am_thread_state_t state;

    /* True if the thread has been stopped using the debugger interface. */
    int debug_stopped;

    /* Temporary storage for the registers of the instrumented thread during
       bootstrapping and debootstrapping. */
    am_registers_t regs;

    /* Address at which the thread's stack is allocated. */
    unsigned long stack_addr;

    /* The number of messages we have sent to the thread since we last updated
       its message counter. */
    int pending_msgs;

    /* Address of the thread's message counter in the thread's address space.
       0 if we do not know the location of the message counter, either because
       we have not yet gotten a "ready" mesage from the thread or because the
       thread is about to do an exit syscall and will deallocate its TLS. */
    unsigned long msg_counter_addr;

    /* Address of the "no events pending" flag in the thread's address space.
       0 if we do not know the location of it, either because
       we have not yet gotten a "ready" mesage from the thread or because the
       thread is about to do an exit syscall and will deallocate its TLS. */
    unsigned long noevpend_addr;

    /* Current execution breakpoints. */
    bp_t bps[NUM_BPS];

    /* New execution breakpoints. */
    bp_t new_bps[NUM_BPS];

    /* True if there are changes to the thread's execution breakpoints that
       have not yet been performed (bps and new_bps differ). We should then
       stop the thread an update the breakpoints as soon as possible. */
    int update_bps;

    /* True if we are waiting for the thread to exit from a syscall.

       On Linux this means that we are single stepping the thread. On Solaris
       this means we have set the thread to stop on all syscall exits. */
    int exiting_syscall;

    /* Arguments to the next notification to be run by the thread. */
    notification_args_t notification_args;

    /* Syscall return value from notification, if has_ret_val is true. */
    int has_ret_val;
    unsigned long ret_val;

    /* Set to true if the thread should be single stepped once the next time it
       is resumed. */
    int single_step;

    /* True if this thread has not been instrumented before.

       When burst sampling, the instrumenter has no way of knowing whether a
       thread is the same thread as one with the same id in a previous
       burst. */
    int never_instrumented;

#if defined(__linux__)
    /* We need to keep track of which synthetic signals we are expecting the
       thread to receive to be able to distinguish synthetic signals from real
       signals.

       sigstop_pending is true if we have sent a SIGSTOP using tkill to stop
       the thread.

       sigtrap_pending is true if we have resumed the thread with
       PTRACE_SINGLESTEP and are expecting a SIGTRAP when the thread finishes
       single stepping. */
    int sigstop_pending;
    int sigtrap_pending;

    /* We use PTRACE_SYSCALL to trace system calls in the sampled
       process. PTRACE_SYSCALL (when used with PTRACE_O_TRACESYSGOOD) causes
       the process to get a (SIGTRAP | 0x80) signal when entering and leaving
       system calls.

       In addition to this we can see whether the thread is currently in a
       syscall by looking at the orig_eax/orig_rax field in the struct
       user_regs_struct. orig_eax/orig_rax also gives use the syscall number if
       the thread is in a syscall.

       There is no way to distinguish between syscall entry and syscall exit
       events. The thread is in a syscall both when we get both entry and exit
       notifications from PTRACE_SYSCALL according to orig_eax/orig_rax. This
       means we have to keep track of whether the thread is currently in a
       syscall to know what is an entry and what is an exit.

       When the thread has exited a syscall according to orig_eax/orig_rax we
       can not modify its state. We can change the contents of some registers,
       but after some syscalls (clone?) trying to change eip/rip causes the
       thread to continue execution at an unpredictable address.

       If the thread receives a signal while in a syscall orig_eax/orig_rax may
       indicate the __NR_restart_syscall syscall instead of the actual syscall
       when we get the syscall exit notification.

       Given these conditions we would like to handle syscall tracing like this
       (well, really, we would like to be able to change the state of the
       thread immediately after a syscall exit notification so we didn't have
       to single step over the next instruction):

       1. The thread is not in a syscall. orig_eax/orig_rax is -1. Resume the
          thread with PTRACE_SYSCALL.

       2. The thread executes for a while and then gets a (SIGTRAP | 0x80)
          signal to indicate that it entered a syscall. orig_eax/orig_rax
          contains the syscall number. Read the syscall arguments from the
          thread's registers and the syscall number from
          orig_eax/orig_rax. Resume the thread with PTRACE_SYSCALL.

       3. The thread executes for a while and then gets a (SIGTRAP | 0x80)
          signal to indicate that it exited the syscall. orig_eax/orig_rax
          indicates that we are in a syscall, but the syscall number is
          unreliable in case the thread has gotten a signal. Read the syscall
          return value from the thread's registers. Resume the thread with
          PTRACE_SINGLESTEP.

       4. The thread singlesteps over the instruction after the syscall
          instruction and then gets a SIGTRAP to indicate that it has finished
          single stepping. We are no longer in a syscall according to
          orig_eax/orig_rax, and we can alter the state of the thread if we,
          e.g., want to instrument it or make it run a notification in the
          instrumenter. Otherwise, restart at 1.

       Note that this will not work if the instruction after the initial
       syscall instruction is also a syscall instruction, since we will then
       single step over it and miss that entire syscall. This should be a very
       rare case, though.

       If we happen attach to a thread while it is in a syscall
       orig_eax/orig_rax will indicate that the thread is in a syscall, but we
       will also get a syscall entry notification once we resume it. We
       therefore add another state for this case (threads attached while not in
       a syscall would start at 1):

       5. The thread has been attached while in a syscall. orig_eax/orig_rax
          indicates that we are in a syscall. Resume the thread with
          PTRACE_SYSCALL. Continue to 2.

       This is the way we would like to handle syscall tracing. However, ptrace
       has some strange quirks that makes this impossible:

       * If we send a SIGSTOP that we do not deliver to a thread that is in a
         syscall, PTRACE_SYSCALL will report a spurious syscall exit before
         telling us about the SIGSTOP, and then report the real syscall exit
         once that happens.

       * If the thread receives a signal while in a syscall ptrace will first
         report a normal syscall exit. However, when we then try to single step
         the next instruction the effect is instead to transfer control to the
         signal hander. And, once rip has changed to the signal handler,
         orig_eax/orig_rax will still indicate that we are in a syscall, so we
         will have to single step once again before we get the thread into a
         state where we can modify its state.

       To handle these quirks we use a more robust scheme for syscall tracing:

       1. The thread is not in a syscall (orig_eax/orig_rax == -1), or we have
          just attached to the thread while it is in a syscall
          (orig_eax/orig_rax != -1).

          Resume the thread with PTRACE_SYSCALL.

       2. The thread has been stopped by a (SIGTRAP | 0x80) signal because of a
          syscall entry.

          Save the syscall number in orig_eax/orig_rax. Save the syscall
          arguments in the thread's registers. Save the address of the
          instruction after the syscall if the thread's eip/rip.

          Resume the thread with PTRACE_SYSCALL.

       3. The thread has been stopped by a (SIGTRAP | 0x80) or SIGTRAP signal
          because of a syscall exit, or by a SIGTRAP signal because we have
          single stepped over the instruction after the syscall or into a
          signal handler.

          If the thread's eip/rip is the same as the one saved in step 2, save
          the syscall return value in the thread's eax/rax (possibly replacing
          a previously saved value).

          If orig_eax/orig_rax != -1, resume the thread with PTRACE_SINGLESTEP
          and continue at 3.

          Otherwise treat the syscall as completed and continue at 1.

       In essence; we save the syscall number and arguments while they are
       reliable when entering the syscall, and then repeatedly single step
       until we are out of the syscall and in a state where the state of the
       thread can be modified, regardless of signal delivery quirks, meanwhile
       saving the last syscall return value, which should be the correct one,
       seen before we single step over the instruction after the syscall or
       into a signal handler.

       Finally, there is conflict between our handling of breakpoints set with
       breakpoint registers and the syscall tracing. When we hit a breakpoint
       that we do not want to remove (it has an ignore count), we disable the
       watchpoint, single step over the instruction and then re-enable the
       watchpoint. If the breakpoint happens to be set on a syscall instruction
       we will then single step over it and miss the entire syscall.

       There is no way to fix this without decoding the instruction that will
       be single stepped to determine if it is a syscall instruction and then
       use PTRACE_SYSCALL instead of PTRACE_SINGLESTEP to step over it (perhaps
       this mechanism could then also be used to handle the case of two syscall
       instructions in a row). */

    /* 1, 2 or 3, corresponding to the states described in the comment
       above. */
    int syscall_state;

    /* Info about the syscall we are currently in. */
    int syscall_num;
    unsigned long syscall_args[6];
    unsigned long syscall_ret_val;
    unsigned long syscall_ret_addr;
#endif
#if defined(__sun__)
    /* The ctl and status file descriptors for the thread. -1 if the files have
       not been opened. */
    int ctl_fd;
    int status_fd;

    /* If true, abort the current system call. Used to cancel sigaction and
       lwp_sigmask syscalls in uninstrumented threads, to prevent them from
       intefering with the instrumenters signal handlers and signal mask. */
    int abort_syscall;

    /* True if we have sent a PCDSTOP request to the thread. */
    int pcdstop_pending;

    /* The suspension state of the thread. */
    am_thread_suspend_state_t suspend_state;

    /* Contains the tid of the thread that is waiting to suspend or continue
       this thread if any, otherwise 0. */
    tid_t suspender_tid;

    /* The address of the suspend spinlock if the thread is waiting on the
       spinlock, otherwise 0. */
    unsigned long suspend_spinlock_addr;

    /* Constains the tid of the thread to be suspended or continued if the
       thread is about to suspend or continue another thread, otherwise 0. */
    tid_t suspendee_tid;

    /* Indicates whether the thread is suspending or resuming another
       thread. Only valid when suspendee_tid is not 0. */
    am_thread_suspend_action_t suspend_action;

    /* Original signal mask is stored here while signals are blocked during
       bootstrapping and debootstrapping. */
    sigset_t orig_sigmask;
#endif
};

/* Structure used for queueing messages. */
typedef struct am_message_buf am_message_buf_t;
struct am_message_buf {
    am_message_t msg;
    void *extra_data;
    am_thread_t *thread;
    int inc_msg_counter;
    am_message_buf_t *next;
};

/* Process states. */
typedef enum {
    /* Desired state:
       - instrumenter loaded
       - threads not instrumented
    */
    AM_PS_Idle,

    /* Desired state:
       - instrumenter loaded
       - threads instrumented
    */
    AM_PS_Burst,

    /* Desired state:
       - instrumenter not loaded
    */
    AM_PS_Detach
} am_process_state_t;

/* Process instrumentation state. */
typedef enum {
    /* The instrumenter is not loaded. We will load the instrumenter and then
       transition to AM_PIS_Idle or AM_PIS_Sampling state. */
    AM_PIS_Attaching,

    /* The instrumenter is loaded and there may be both instrumented and
       uninstrumented threads. We will deinstrument all threads and then let
       them run uninstrumented. */
    AM_PIS_Idling,

    /* The instrumenter is loaded and there may be both instrumented and
       uninstrumented threads. We will deinstrument threads until all threads
       are either uninstrumented or stuck in a syscall, and then proceed to
       AM_PIS_Bursting. */
    AM_PIS_Starting_Burst,

    /* The instrumenter is loaded and there may be both instrumented and
       uninstrumented threads. We will instrument all threads and then let them
       run instrumented. */
    AM_PIS_Bursting,

    /* The instrumenter is loaded but there are no instrumented threads. We
       will unload the instrumenter and transition to AM_PIS_Native state. */
    AM_PIS_Detaching,

    /* The instrumenter is not loaded. We will let the threads run
       uninstrumented. */
    AM_PIS_Native
} am_process_instrumentation_state_t;

/* Process execution states. */
typedef enum {
    /* All threads may run and proceed with attaching or detaching. */
    AM_PES_Run,

    /* All threads may run, but no thread except the one selected for detaching
       may proceed with attaching or detaching. */
    AM_PES_Run_Excl,

    /* All threads but the one attaching or detaching must be stopped. */
    AM_PES_Run_One,

    /* All threads most stop as soon as possible. */
    AM_PES_Stop
} am_process_execution_state_t;

/** Process ptrace purpose. With what purpose are we ptracing this process? */
typedef enum {
    /* Track with intent to sample it, and also to track all its forks */
    AM_PPP_Sample_And_Fork,

    /* Track with intent to follow forks only */
    AM_PPP_Fork
} am_process_ptrace_purpose_t;

/* Vector of pids. */
typedef VECT(pid_t, pid) pid_vect_t;

/* Process struct, there is one for monitored process. */
struct am_process {
    /* Process id, 0 if the process has not been launched yet. */
    pid_t pid;

    /* Parent process id, set to 0 for top-level or orphaned processes. */
    pid_t ppid;

    /* Fork-generation. A directly attached or launched process has
     * generation 0, each child gets generation = parent.generation +
     * 1 */
    int generation;

    /* This says whether this process shall be instrumented at all, or
     * just followed to track forks. New processes are evaluated at
     * fork and exec time and have this field set according to the
     * outcome of a predicate.
     */
    am_process_ptrace_purpose_t ptrace_purpose;

    /* require_ownership == !unsupported_attach */
    int require_ownership;

    /* Bitness of the process, 32 or 64, or -1 if it is unknown. */
    int bitness;

    /* Launch binary, or empty string if we attached to a running process. */
    std::string binary;

    /* Process state. */
    am_process_state_t state;
    am_process_instrumentation_state_t instrumentation_state;
    am_process_execution_state_t execution_state;

    /* Threads in the process. */
    VECT(am_thread_t *, thread) threads;

    /* Address of the hijacked page in the monitored process used during
       bootstrapping and debootstrapping, 0 if no page is hijacked. */
    unsigned long hijack_page;

    /* Buffer used to store the original data on the page. */
    page_buf_t hijack_page_buf;

    /* The file descriptors of the communication pipe in the monitor process
       and in the target process. -1 of the socket is not open. */
    int socket_fd;
    int target_socket_fd;

    /* List of messages to send to the process. */
    am_message_buf_t *msg_head;
    am_message_buf_t *msg_tail;

    /* The address of the instrumenter in the process. 0 if no instrumenter is
       currently loaded. */
    unsigned long instr_addr;

    /* ELF file of the process, or NULL if not yet determined. */
    char *elf_file;

    /* Burst generation. */
    unsigned int burst_generation;

    int start_new_burst;

    /* Flag set when singlestepping over an exec. Used to get control
     * in the new image. Due to bug
     * https://bugzilla.redhat.com/show_bug.cgi?id=520206 we have to
     * singlestep once after receiving the PTRACE_EVENT_EXEC */
    int is_execing;

#if defined(__linux__)
    /* True if we have not seen any syscall yet. */
    int first_syscall;
#endif
#if defined(__sun__)
    /* The ctl and status file descriptors for the process. -1 if the files
       have not been opened. */
    int ctl_fd;
    int status_fd;
#endif
    /* Keep track of whether the sampling start and sampling end conditions have
       triggered. */
    int attach_triggered;
    int exit_triggered;

    int start_at_ignore;
    int stop_at_ignore;

    /* The acquisition file associated with this process (if it is sampled) */
    acq_handle_t *acq_handle;

    /* trace file prefix, really in sample_params */
    char trace_file_prefix[TRACE_FILE_PREFIX_MAX];

    /* The page protection bitmap */
    pages_t *pages;
};

/* Monitored processes. */
typedef VECT(am_process_t *, process) process_vect_t;
extern process_vect_t processes;

/* Vector of arguments to the process started with --binary. */
typedef VECT(char *, charp) argument_vect_t;
extern argument_vect_t arguments;

/* Force acumon to exit within a finite amount of time (unless writing to the
   pipe to a process blocks). May leave threads instrumented, or even attaching
   or detaching (the hijacked page may then not be restored, causing memory
   corruption in the process). */
__attribute__((__noreturn__)) void force_exit(am_exit_status_t status);

/* After an exec, reset things to restart sampling in the new image */
void continue_after_exec(am_process_t *process);


/* Add a new process record. */
am_process_t * add_process(pid_t pid, pid_t ppid, const std::string &binary, int generation);


/* set the process state to AM_PS_Burst, and add stop conditions */
void setup_start_conditions(am_process_t *process);

/* Should be called immediately after a thread has been attached.

   On Linux that is either after an explicit PTRACE_ATTACH, when a launched
   process has done PTRACE_TRACEME and an exec, or when we get a
   PTRACE_EVENT_CLONE event to a cloned thread's parent. */
am_thread_t *add_thread(am_process_t *process, pid_t tid,
                        int debug_stopped, int instrumented);

/* Should be called immediately before a thread is detached.

   On Linux that is either before an explicit PTRACE_DETACH or before
   continuing a thread after receiving a PTRACE_EVENT_EXIT notification to
   it. */
void remove_thread(am_thread_t *thread);

/* Close the communication socket to the instrumenter. */
void close_socket(am_process_t *process);

/* Remove all processes where all threads have terminated. */
void prune_processes(void);

/* Update the message counter of the thread. The thread must be stopped to
   avoid race conditions. */
void update_msg_counter(am_thread_t *thread);

/* Returns the thread with the specified tid in the specified process, or NULL
   if there is no such thread. */
am_thread_t *thread_by_tid(am_process_t *process, tid_t tid);

/* Convenience function that either prints a message on STDERR or on
   the control socket. */
void am_printf(am_control_message_level_t level, const char *format, ...);

/* Emulates perror(3) using strerror_r(3) and am_printf. */
void am_perror(am_control_message_level_t level, const char *s);

/* Returns true if a page in the process has been hijacked, false if not. */
int page_hijacked(const am_process_t *process);

/* Returns true if the instrumenter has been loaded into the specified
   process. */
int instrumenter_loaded(const am_process_t *process);

/* Returns true if the address lies within the instrumenter image of the
   process. Returns false if not, or if no instrumenter image is loaded. */
int addr_in_instrumenter(const am_process_t *process, unsigned long addr);

/* Returns the name of the thread state. */
const char *thread_state_name(am_thread_state_t ts);

/* Returns the name of the process state. */
const char *process_state_name(am_process_state_t ps);

/* Returns the name of the process instrumentation state. */
const char *process_instrumentation_state_name(am_process_instrumentation_state_t pis);

/* Returns the name of the process execution state. */
const char *process_execution_state_name(am_process_execution_state_t pes);

#endif /* ACUMON_HH */
