/// @file acumon/event.c

// Copyright (c) 2007-2008 Acumem AB.  All rights reserved.

#define _XOPEN_SOURCE 600

#include <stdlib.h>
#include <stdio.h>
#include <poll.h>
#include <errno.h>
#include <assert.h>
#include <time.h>
#include <sys/time.h>

#include "../library/am_mem/am_mem.h"

#include "safe_io.hh"
#include "am_assert.hh"
#include "event.hh"

/* A registered file descriptor event. */
typedef struct {
    /* The event handler and the data that should be sent to it. */
    void (*handler)(short events, void *data);
    void *data;

    /* We can't delete an event while we are iterating over the array of
       events. Instead we mark the event as deleted and remove it just before
       running poll the next time. */
    int deleted;
} am_fd_event_t;

/* Array of currently active file descriptor events. The current number of
   events is kept in num_fd_events and the number of allocated event structs is
   kept in max_fd_events. We also maintain an array of associated pollfd
   structs, one for each active event, stored in the same order. */
static int num_fd_events;
static int max_fd_events;
static am_fd_event_t *fd_events;
static struct pollfd *event_poll_fds;

/* A registered time event. */
typedef struct am_time_event {
    /* The event handler and the data that should be sent to it. */
    void (*handler)(void *data);
    void *data;

    /* Monotonic time of the event. */
    am_time_t time;

    /* The next event in the time event queue. */
    struct am_time_event *next;
} am_time_event_t;

/* Time event queue. */
static am_time_event_t *time_events;

void
add_fd_event(int fd, short poll_events,
             void (*handler)(short events, void *data), void *data)
{
    int i = num_fd_events++;
    if (num_fd_events > max_fd_events) {
        fd_events = (am_fd_event_t *)realloc(fd_events, num_fd_events * sizeof(am_fd_event_t));
        AM_ASSERT_WITH_ERRNO(fd_events != NULL);
        event_poll_fds = (pollfd *)realloc(event_poll_fds,
                                 num_fd_events * sizeof(struct pollfd));
        AM_ASSERT_WITH_ERRNO(event_poll_fds != NULL);
    }
    fd_events[i].handler = handler;
    fd_events[i].data = data;
    fd_events[i].deleted = 0;
    event_poll_fds[i].fd = fd;
    event_poll_fds[i].events = poll_events;
}

void
remove_fd_event(int fd, short poll_events,
                void (*handler)(short events, void *data), void *data)
{
    for (int i = 0; i < num_fd_events; i++)
        if (event_poll_fds[i].fd == fd &&
            event_poll_fds[i].events == poll_events &&
            fd_events[i].handler == handler &&
            fd_events[i].data == data &&
            !fd_events[i].deleted) {
            fd_events[i].deleted = 1;
            return;
        }
    AM_ILLEGAL_STATE();
}

static am_time_t
get_time(void)
{
    struct timespec timespec;
    int res = clock_gettime(CLOCK_MONOTONIC, &timespec);
    AM_ASSERT_WITH_ERRNO(res == 0);
    return timespec.tv_sec * (am_time_t)1000000000 + timespec.tv_nsec;
}

void
add_time_event(am_time_t timeout,
               void (*handler)(void *data), void *data)
{
    am_time_event_t *event = (am_time_event_t *)am_malloc(sizeof(am_time_event_t));
    event->handler = handler;
    event->data = data;
    event->time = get_time() + timeout;

    am_time_event_t **nextp = &time_events;
    while ((*nextp) != NULL &&
           (*nextp)->time < event->time)
        nextp = &(*nextp)->next;
    event->next = *nextp;
    *nextp = event;
}

void
remove_time_event(void (*handler)(void *data), void *data)
{
    am_time_event_t **eventp = &time_events;
    while (*eventp != NULL) {
        am_time_event_t *event = *eventp;
        if (event->handler == handler &&
            event->data == data) {
            *eventp = event->next;
            am_free(event);
            return;
        }
        else
            eventp = &event->next;
    }
    AM_ILLEGAL_STATE();
}

void
remove_time_events_with_data(void *data)
{
    am_time_event_t **eventp = &time_events;
    while (*eventp != NULL) {
        am_time_event_t *event = *eventp;
        if (event->data == data) {
            *eventp = event->next;
            am_free(event);
        }
        else
            eventp = &event->next;
    }    
}

/* Remove any events that have been marked for deletion. */
static void
prune_events(void)
{
    for (int i = 0; i < num_fd_events; i++)
        if (fd_events[i].deleted) {
            num_fd_events--;
            if (i != num_fd_events) {
                fd_events[i] = fd_events[num_fd_events];
                event_poll_fds[i] = event_poll_fds[num_fd_events];
            }
            i--;
        }
}

void
wait_for_event(void)
{
    /* Removed all events marked as deleted. */
    prune_events();

    /* Handle any time events that have timed out. */
    int timeout;
retry:
    if (time_events != NULL) {
        am_time_t time = get_time();
        if (time_events->time <= time) {
            do {
                time_events->handler(time_events->data);
                am_time_event_t *next_time_events = time_events->next;
                am_free(time_events);
                time_events = next_time_events;
            } while (time_events != NULL &&
                     time_events->time <= time);
            return;
        }

        timeout = (time_events->time - time + 999999) / 1000000;
    }
    else
        timeout = -1;

    /* Poll for events on the file descriptors. */
    int res;
    res = safe_poll(event_poll_fds, num_fd_events, timeout);
    AM_ASSERT_WITH_ERRNO(res != -1);
    if (res == 0)
        goto retry;

    /* Save the number of registered events before we start calling the event
       handlers, so that we don't iterate over any new events that are added by
       the event handlers. */
    int orig_num_fd_events = num_fd_events;
    for (int i = 0; i < orig_num_fd_events; i++)
        if (event_poll_fds[i].revents && !fd_events[i].deleted)
            fd_events[i].handler(event_poll_fds[i].revents, fd_events[i].data);
}

void
event_init(void)
{
    num_fd_events = 0;
    max_fd_events = 0;
    fd_events = NULL;
    event_poll_fds = NULL;
}

void
event_fini(void)
{
    prune_events();
    AM_ASSERT(num_fd_events == 0);
    free(fd_events);
    free(event_poll_fds);
}
