/// @file acumon/licensing.cc

// Copyright (c) 2010 Acumem AB.  All rights reserved.

#include "product.h"

#include "licensing.hh"
#include "acumon.hh"

#include "library/license/license.hh"
#include "library/manifest/manifest.hh"

#include <limits.h>
#include <exception>

using namespace acumem::license;
using namespace std;

/* SSS:
int feature_attach = 0;
int feature_unlimited_threads = 0;
int feature_api = 0;
int feature_burst = 0;
*/
int feature_attach = 1;
int feature_unlimited_threads = 1;
int feature_api = 1;
int feature_burst = 1;

/// License features. Needs to stay alive until end of sampling. It holds the
/// actual license token.
///
/// This is made into a pointer to avoid executing automatic c++
/// destructors in forked processes. See TS-2065.
static feature_set_ptr_t *feature_set_p;

static void
output_callback(const char *msg)
{
    am_printf(AM_Ctrl_Msg_Info, "%s\n", msg);
}

int
licensing_init(am_exit_status_t *exit_status)
{
    try {
	warning_consumer_ptr_t wc(new warning_consumer_callback_t(output_callback));
	license_manager_ptr_t license(new license_manager_t(wc));
	feature_set_p =
	    new feature_set_ptr_t(license->acquire(license_manager_t::tool_kind));
	
	if (!(*feature_set_p)) {
	    am_printf(AM_Ctrl_Msg_Error,
		      "No license token available.\n");
	    *exit_status = AM_Exit_License_Invalid;
	    return 1;
	}

    /* SSS: 
	feature_attach = (*feature_set_p)->attach();
	feature_unlimited_threads = (*feature_set_p)->unlimited_threads();
	feature_api = (*feature_set_p)->api();
	feature_burst = (*feature_set_p)->burst();
    */
	
	if (sampler_params.burst_sampling_length != ULLONG_MAX && !feature_burst) {
	    am_printf(AM_Ctrl_Msg_Error,
		      "Burst-sampling not covered by license.\n");
	    *exit_status = AM_Exit_License_Exceeded;
	    return 1;
	}

	return 0;
    } catch (std::exception &e) {
	am_printf(AM_Ctrl_Msg_Error, "%s\n", e.what());
	*exit_status = AM_Exit_License_Invalid;
	return 1;
    }
}

void
licensing_fini(void)
{
    feature_set_p->reset();
}

int
licensing_check_manifest(const char *elf_file)
{
    if (!(*feature_set_p)) {
	am_printf(AM_Ctrl_Msg_Error, "License error.\n");
	return 1;
    }

/* SSS : */
    if ((*feature_set_p)->require_manifest_check())
        try {
            sha1_t sha1 = verify_manifest(elf_file);

            vector<unsigned char> manifest_hash(
                sha1.bytes, sha1.bytes + sizeof(sha1.bytes));
            if (!(*feature_set_p)->check_manifest_hash(manifest_hash)) {
                am_printf(AM_Ctrl_Msg_Error,
                          "Not authorized to sample application %s. The "
                          "corresponding manifest file's signature is "
                          "invalid.\n", elf_file);
                return 1;
            }
        }
        catch (manifest_exception_t &e) {
            am_printf(AM_Ctrl_Msg_Error, "%s\n", e.what());
            am_printf(AM_Ctrl_Msg_Error,
                      "License does not allow sampling of %s.\n", elf_file);
            //return 1;
            return 0;
        }

    return 0;
}
