/// @file acumon/linux.c

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

/*
 * Linux implementation of operating system dependent functions in
 * os_dependent.h.
 */

// TODO: queue signals during bootstrapping/debootstrapping

/* Needed by libelf.h on Fedora 7 & 8. */
#define _LARGEFILE64_SOURCE

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/user.h>
#include <sys/ptrace.h>
#include <gelf.h>
#include <libelf.h>

#include "../sampler/src/tls_size.h"

//#include "syscall_nums_32.h"
#if defined(__x86_64__)
#include "syscall_nums_64.h"
#endif
#include "os_dependent.hh"
#include "thread_predicates.hh"
#include "pages.hh"
#include "safe_io.hh"
#include "am_assert.hh"
#include "signal_handling.hh"
#include "acquisition_info.hh"
#include "path_utils.hh"

#define MIN(x, y) ((x) < (y) ? (x) : (y))
#define MAX(x, y) ((x) > (y) ? (x) : (y))

#define am_offsetof(type, field) (&((type *)0)->field)

/* PTRACE_GETSIGINFO and PTRACE_SETSIGINFO have been implemented in the kernel
   since 2.3.99-pre6.

   PTRACE_SETOPTIONS has been implemented in the kernel since 2.4.6.

   PTRACE_GETEVENTMSG, PTRACE_O_TRACEFORK, PTRACE_O_TRACEVFORK,
   PTRACE_O_TRACECLONE, PTRACE_O_TRACEEXEC, PTRACE_EVENT_FORK,
   PTRACE_EVENT_VFORK, PTRACE_EVENT_CLONE and PTRACE_EVENT_EXEC have been
   implemented in the kernel since 2.5.46.

   PTRACE_TRACEEXIT and PTRACE_EVENT_EXIT have been implemented in the kernel
   since 2.5.60.

   They have been added to sys/ptrace.h and linux/ptrace.h only recently (they
   are present in Fedora Core 6, but not in Fedora Core 5). */
#ifndef PTRACE_SETOPTIONS
#define PTRACE_SETOPTIONS  0x4200
#endif
#ifndef PTRACE_GETEVENTMSG
#define PTRACE_GETEVENTMSG 0x4201
#endif
#ifndef PTRACE_GETSIGINFO
#define PTRACE_GETSIGINFO  0x4202
#endif
#ifndef PTRACE_SETSIGINFO
#define PTRACE_SETSIGINFO  0x4203
#endif
#ifndef PTRACE_O_TRACEFORK
#define PTRACE_O_TRACEFORK  0x00000002
#endif
#ifndef PTRACE_O_TRACEVFORK
#define PTRACE_O_TRACEVFORK 0x00000004
#endif
#ifndef PTRACE_O_TRACECLONE
#define PTRACE_O_TRACECLONE 0x00000008
#endif
#ifndef PTRACE_O_TRACEEXEC
#define PTRACE_O_TRACEEXEC  0x00000010
#endif
#ifndef PTRACE_O_TRACEEXIT
#define PTRACE_O_TRACEEXIT  0x00000040
#endif
#ifndef PTRACE_EVENT_FORK
#define PTRACE_EVENT_FORK  1
#endif
#ifndef PTRACE_EVENT_VFORK
#define PTRACE_EVENT_VFORK 2
#endif
#ifndef PTRACE_EVENT_CLONE
#define PTRACE_EVENT_CLONE 3
#endif
#ifndef PTRACE_EVENT_EXEC
#define PTRACE_EVENT_EXEC  4
#endif
#ifndef PTRACE_EVENT_EXIT
#define PTRACE_EVENT_EXIT  6
#endif

/* There doesn't seem to be any nice way to get the ptrace event associated
   with a signal. This is how gdb does it. */
#define PTRACE_EVENT(status) ((status) >> 16)

/* PREFIX_REG adds the correct 'e' or 'r' prefix to a register base name
   depending on the architecture.

   SCNO is the offset of the orig_rax/orig_eax field in struct
   user_regs_struct, that contains the number of the system call the thread
   currently is in, or -1.

   SCNO_OFFS is the offset of SCNO in struct user_regs_struct. */
#if defined(__x86_64__)
#define PREFIX_REG(base) r ## base
#define SCNO orig_rax
#elif defined(__i386__)
#define PREFIX_REG(base) e ## base
#define SCNO orig_eax
#else
#error unexpected arch
#endif
#define SCNO_OFFS ((void *)offsetof(struct user_regs_struct, SCNO))

#define PT(x) ((__ptrace_request)x)

static unsigned long
get_rax(const am_thread_t *thread)
{
    errno = 0;
    long res = ptrace(PT(PTRACE_PEEKUSER), thread->tid,
                      (void *)offsetof(struct user, regs.PREFIX_REG(ax)),
                      NULL);
    AM_ASSERT_WITH_ERRNO(errno == 0 || errno == ESRCH);
    return errno == 0 ? res : 0;
}

static void __attribute__((__noreturn__))
sigsegv_debug(am_thread_t *thread, siginfo_t *siginfo)
{
    switch (siginfo->si_signo) {
    case SIGSEGV:
        am_printf(AM_Ctrl_Msg_Debug, "Thread %d got SIGSEGV accessing %p.\n", thread->tid, siginfo->si_addr);
        break;
    default:
        am_printf(AM_Ctrl_Msg_Debug, "Thread %d got signal %d.\n", thread->tid, siginfo->si_signo);
        break;
    }

    struct user_regs_struct regs;
    long res = ptrace(PT(PTRACE_GETREGS), thread->tid, NULL, &regs);
    AM_ASSERT_WITH_ERRNO(res != -1);
    am_printf(AM_Ctrl_Msg_Debug, "IP:  %lx\n", regs.PREFIX_REG(ip));
    am_printf(AM_Ctrl_Msg_Debug, "AX:  %lx\n", regs.PREFIX_REG(ax));
    am_printf(AM_Ctrl_Msg_Debug, "CX:  %lx\n", regs.PREFIX_REG(cx));
    am_printf(AM_Ctrl_Msg_Debug, "DX:  %lx\n", regs.PREFIX_REG(dx));
    am_printf(AM_Ctrl_Msg_Debug, "BX:  %lx\n", regs.PREFIX_REG(bx));
    am_printf(AM_Ctrl_Msg_Debug, "SP:  %lx\n", regs.PREFIX_REG(sp));
    am_printf(AM_Ctrl_Msg_Debug, "BP:  %lx\n", regs.PREFIX_REG(bp));
    am_printf(AM_Ctrl_Msg_Debug, "SI:  %lx\n", regs.PREFIX_REG(si));
    am_printf(AM_Ctrl_Msg_Debug, "DI:  %lx\n", regs.PREFIX_REG(di));
#if defined(__x86_64__)
    am_printf(AM_Ctrl_Msg_Debug, "R8:  %lx\n", regs.r8);
    am_printf(AM_Ctrl_Msg_Debug, "R9:  %lx\n", regs.r9);
    am_printf(AM_Ctrl_Msg_Debug, "R10: %lx\n", regs.r10);
    am_printf(AM_Ctrl_Msg_Debug, "R11: %lx\n", regs.r11);
    am_printf(AM_Ctrl_Msg_Debug, "R12: %lx\n", regs.r12);
    am_printf(AM_Ctrl_Msg_Debug, "R13: %lx\n", regs.r13);
    am_printf(AM_Ctrl_Msg_Debug, "R14: %lx\n", regs.r14);
    am_printf(AM_Ctrl_Msg_Debug, "R15: %lx\n", regs.r15);
#endif

#if 0
    for (int i = regs.PREFIX_REG(ip) & ~0xfffL; i < ((regs.PREFIX_REG(ip) + 0x1000) & (~0xfffL)); i+=16) {
        am_printf(AM_Ctrl_Msg_Debug, "0x%08x:", i);
        for (int j = 0; j < 16; j++)
            am_printf(AM_Ctrl_Msg_Debug, " %02x", (unsigned char)ptrace(PT(PTRACE_PEEKTEXT), thread->tid, (void *)(i + j), NULL));
        am_printf(AM_Ctrl_Msg_Debug, "\n");
    }
#endif

    while (1);
}

static void
get_elf_file(am_process_t *process, tid_t tid)
{
    int res;

    char exe_path[PATH_MAX];
    res = snprintf(exe_path, PATH_MAX, "/proc/%d/exe", tid);
    AM_ASSERT(res >= 0 && res < PATH_MAX);

    char exe_filename[PATH_MAX];
    size_t exe_filename_size;
    exe_filename_size = readlink(exe_path, exe_filename, sizeof(exe_filename));
    if (exe_filename_size == -1) {
	if (errno == EACCES) {
	    am_printf(AM_Ctrl_Msg_Error,
		      "Error: Could not determine process ELF file from %s: %s\n",
		      exe_path, strerror(errno));
	    am_printf(AM_Ctrl_Msg_Error,
		      "This is likely caused by the binary having the setuid or setgid attribute.\n"
		      "Sampling such binaries is not supported.\n");
	    force_exit(AM_Exit_Target_Binary);
	} else {
	    am_perror(AM_Ctrl_Msg_Error,
		      "Fatal error in get_elf_file");
	    force_exit(AM_Exit_Target_Binary);
	}
    }
    AM_ASSERT(exe_filename_size < PATH_MAX);
    exe_filename[exe_filename_size] = '\0';
    process->elf_file = am_strdup(exe_filename);
    return;
}

static unsigned long
read_dr(const am_thread_t *thread, const int reg_num)
{
    AM_ASSERT(thread_stopped(thread, 1));
    errno = 0;
    long res = ptrace(PT(PTRACE_PEEKUSER), thread->tid,
                      am_offsetof(struct user, u_debugreg[reg_num]), NULL);
    AM_ASSERT_WITH_ERRNO(errno == 0 || errno == ESRCH);
    return errno == 0 ? res : 0;
}

static void
write_dr(const am_thread_t *thread, int reg_num, unsigned long val)
{
    AM_ASSERT(thread_stopped(thread, 1));
    errno = 0;
    long res = ptrace(PT(PTRACE_POKEUSER), thread->tid,
                      am_offsetof(struct user, u_debugreg[reg_num]),
                      (void *)val);
    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
}

static void
resume_thread(am_thread_t *thread, int signal)
{
    AM_ASSERT(thread->debug_stopped);

    int suspended_bp = 0;
    for (int i = 0; i < NUM_BPS; i++)
        if (thread->bps[i].suspended)
            suspended_bp = 1;

    __ptrace_request req = PT(PTRACE_CONT);

    int native = 0;
    if (instrumenter_loaded(thread->process)) {
	native = thread_native(thread, 0);

	/*
	 * Catch cases where we need to consider the thread to be in native mode,
	 * even though thread_native(thread, 0) returns false.
	 *
	 * One such case occurs during deinstrumentation: the thread is in state
	 * AM_TS_Deinstr_Wait_Done_2_Stopped, the target has run the deinstrumentation
	 * code, sent an AM_TS_Deinstr_Wait_Done_2 message to the monitor, and is
	 * in a sched_yield() loop waiting to be redirected to native mode.  At this
	 * point a SIGCHLD can be processed before update_thread() is called, causing
	 * us to call resume_thread() before update_thread() has changed its state to
	 * AM_TS_Native.  This made us resume the thread with PTRACE_CONT rather than
	 * PTRACE_SYSCALL, which made us miss mmap()/mprotect() calls in idle phases,
	 * which prevented generated code from being invalidated (TS-2249, TS-2164).
	 */
	if (!native) {
	    switch (thread->state) {
	    case AM_TS_Deinstr_Wait_Done_2_Stopped:
		native = 1;
	    default:
		break;
	  }
	}
    }

    if (native ||
        thread->syscall_state != 1)
        switch (thread->syscall_state) {
        case 1:
        case 2:
            req = PT(PTRACE_SYSCALL);
            break;
        case 3:
            req = PT(PTRACE_SINGLESTEP);
            break;
        default:
            AM_ILLEGAL_STATE();
        }
    else
        if (thread->exiting_syscall)
            req = PT(PTRACE_SINGLESTEP);

    /* TODO: This conflicts with syscall tracing. */
    if (thread->single_step) {
        req = PT(PTRACE_SINGLESTEP);
        thread->single_step = 0;
    }

    /* TODO: This conflicts with syscall tracing, see ticket #756. */
    if (suspended_bp)
        req = PT(PTRACE_SINGLESTEP);

    /* If a signal is delivered while acumon is single stepping a thread, the
       trap flag may be set in the signal frame that the kernel writes to the
       stack. When the thread returns from the signal handler it will continue
       executing with the trap flag set, causing a SIGTRAP to be raised. The
       monitor does not feel responsible for the SIGTRAP so it will deliver it
       to the process, which will be killed or get an unexpected signal. */
    if (req != PTRACE_SINGLESTEP ) {
        struct user_regs_struct regs;
        long res = ptrace(PT(PTRACE_GETREGS), thread->tid, NULL, &regs);
        AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
        if (errno == ESRCH)
            memset(&regs, 0, sizeof(regs));
        if (regs.eflags & 0x100) {
            regs.eflags &= 0xfffffffffffffeff;
            res = ptrace(PT(PTRACE_SETREGS), thread->tid, NULL, &regs);
            AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
        }
    }

    long res = ptrace(req, thread->tid, NULL, (void *)(long)signal);
    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
    thread->debug_stopped = 0;
    thread->sigtrap_pending = req == PT(PTRACE_SINGLESTEP);
}

/* Applies the scanf pattern with field as argument to each line of the
   /proc/<tid>/status file of the thread until a match is found. Returns 0 on
   success, -1 on error.  */
static int
get_status_field(pid_t tid, const char *pattern, void *field)
{
    int res;
    char maps_path[PATH_MAX];
    res = snprintf(maps_path, PATH_MAX, "/proc/%d/status", tid);
    AM_ASSERT(res >= 0 && res < PATH_MAX);
    FILE *file = fopen(maps_path, "r");
    if (file == NULL)
        return -1;

    while (1) {
        char buf[1024];
        char *s = fgets(buf, sizeof(buf), file);
        AM_ASSERT(s != NULL);
        if (strlen(buf) == sizeof(buf) - 1)
            AM_ILLEGAL_STATE();
        res = sscanf(buf, pattern, field);
        if (res == 1) {
            res = fclose(file);
            AM_ASSERT_WITH_ERRNO(res == 0);
            return 0;
        }
    }

    AM_ILLEGAL_STATE();
}

static int
blocking_waitpid(pid_t tid)
{
    pid_t res;
    int status;
    res = safe_waitpid(tid, &status, __WALL);
    AM_ASSERT_WITH_ERRNO(res == tid);
    return status;
}

/* Get the parent pid of the specified thread. Returns -1 on error. */
static pid_t
get_ppid(pid_t tid)
{
    pid_t pid;
    int res = get_status_field(tid, "PPid:\t%d\n", &pid);
    return res == -1 ? -1 : pid;
}

static int
thread_syscall_num(const am_thread_t *thread)
{
    AM_ASSERT(thread->debug_stopped);

    errno = 0;
    long res = ptrace(PT(PTRACE_PEEKUSER), thread->tid, SCNO_OFFS, NULL);
    AM_ASSERT_WITH_ERRNO(errno == 0 || errno == ESRCH);
    /* The returned value is not properly sign extended from 32 to 64
       bits by ptrace for 32-bit processes with on 64-bit hosts. */
    return errno == 0 ? (int32_t)res : 0;
}

static void
syscall_complete(am_thread_t *thread)
{
    if (thread_native(thread, 1) &&
        instrumenter_loaded(thread->process))
        switch (thread->process->bitness) {
/*        case 32:
            if ((uint32_t)thread->syscall_ret_val > 0xfffff000)
                break;;
            switch (thread->syscall_num) {
            case SYS_mmap_32: {
                uint32_t args[6];
                read_data(thread, (uint32_t)thread->syscall_args[0], sizeof(args), args);
                thread->notification_args.type = Notification_Mmap;
                thread->notification_args.u.mmap.addr = thread->syscall_ret_val;
                thread->notification_args.u.mmap.len = args[1];
                thread->notification_args.u.mmap.prot = args[2];
                thread->notification_args.u.mmap.flags = args[3];
                thread->notification_args.u.mmap.fd = args[4];
                thread->notification_args.u.mmap.offs = args[5];
                thread->state = AM_TS_Notification_Wait;
            } break;
            case SYS_mmap2_32:
                thread->notification_args.type = Notification_Mmap;
                thread->notification_args.u.mmap.addr = thread->syscall_ret_val;
                thread->notification_args.u.mmap.len = thread->syscall_args[1];
                thread->notification_args.u.mmap.prot = thread->syscall_args[2];
                thread->notification_args.u.mmap.flags = thread->syscall_args[3];
                thread->notification_args.u.mmap.fd = thread->syscall_args[4];
                thread->notification_args.u.mmap.offs = thread->syscall_args[5];
                thread->state = AM_TS_Notification_Wait;
                break;
            case SYS_munmap_32:
                thread->notification_args.type = Notification_Munmap;
                thread->notification_args.u.munmap.addr = thread->syscall_args[0];
                thread->notification_args.u.munmap.len = thread->syscall_args[1];
                thread->state = AM_TS_Notification_Wait;
                break;
            case SYS_mprotect_32:
                thread->notification_args.type = Notification_Mprotect;
                thread->notification_args.u.mprotect.addr = thread->syscall_args[0];
                thread->notification_args.u.mprotect.len = thread->syscall_args[1];
                thread->notification_args.u.mprotect.prot = thread->syscall_args[2];
                thread->state = AM_TS_Notification_Wait;
                break;
            case SYS_mremap_32:
                thread->notification_args.type = Notification_Mremap;
                thread->notification_args.u.mremap.old_addr = thread->syscall_args[0];
                thread->notification_args.u.mremap.old_size = thread->syscall_args[1];
                thread->notification_args.u.mremap.new_addr = thread->syscall_ret_val;
                thread->notification_args.u.mremap.new_size = thread->syscall_args[2];
                thread->state = AM_TS_Notification_Wait;
                break;
            case SYS_rt_sigaction_32:
                thread->notification_args.type = Notification_Sigaction;
                thread->notification_args.u.sigaction.sig = thread->syscall_args[0];
                thread->notification_args.u.sigaction.nact = thread->syscall_args[1];
                thread->notification_args.u.sigaction.oact = thread->syscall_args[2];
                thread->state = AM_TS_Notification_Wait;
                break;
            default:
                break;
            }
            break; */
#if defined(__x86_64__)
        case 64:
            if (thread->syscall_ret_val > 0xfffffffffffff000)
                break;
            switch (thread->syscall_num) {
            case SYS_mmap_64:
                thread->notification_args.type = Notification_Mmap;
                thread->notification_args.u.mmap.addr = thread->syscall_ret_val;
                thread->notification_args.u.mmap.len = thread->syscall_args[1];
                thread->notification_args.u.mmap.prot = thread->syscall_args[2];
                thread->notification_args.u.mmap.flags = thread->syscall_args[3];
                thread->notification_args.u.mmap.fd = thread->syscall_args[4];
                thread->notification_args.u.mmap.offs = thread->syscall_args[5];
                thread->state = AM_TS_Notification_Wait;
                break;
            case SYS_munmap_64:
                thread->notification_args.type = Notification_Munmap;
                thread->notification_args.u.munmap.addr = thread->syscall_args[0];
                thread->notification_args.u.munmap.len = thread->syscall_args[1];
                thread->state = AM_TS_Notification_Wait;
                break;
            case SYS_mprotect_64:
                thread->notification_args.type = Notification_Mprotect;
                thread->notification_args.u.mprotect.addr = thread->syscall_args[0];
                thread->notification_args.u.mprotect.len = thread->syscall_args[1];
                thread->notification_args.u.mprotect.prot = thread->syscall_args[2];
                thread->state = AM_TS_Notification_Wait;
                break;
            case SYS_mremap_64:
                thread->notification_args.type = Notification_Mremap;
                thread->notification_args.u.mremap.old_addr = thread->syscall_args[0];
                thread->notification_args.u.mremap.old_size = thread->syscall_args[1];
                thread->notification_args.u.mremap.new_addr = thread->syscall_ret_val;
                thread->notification_args.u.mremap.new_size = thread->syscall_args[2];
                thread->state = AM_TS_Notification_Wait;
                break;
            case SYS_rt_sigaction_64:
                thread->notification_args.type = Notification_Sigaction;
                thread->notification_args.u.sigaction.sig = thread->syscall_args[0];
                thread->notification_args.u.sigaction.nact = thread->syscall_args[1];
                thread->notification_args.u.sigaction.oact = thread->syscall_args[2];
                thread->state = AM_TS_Notification_Wait;
                break;
            default:
                break;
            }
            break;
#endif
        default:
            AM_ILLEGAL_STATE();
        }
    thread->syscall_state = 1;
}

static void
sigchld_event(void)
{
    tid_t tid;
    int status;
signal_handled:
    while ((tid = safe_waitpid(-1, &status, WNOHANG | __WALL)) != 0) {
        if (tid == -1) {
            if (errno == EINTR)
                continue;
            else if (errno == ECHILD)
                /* All monitored threads have exited, nothing to do. */
                return;
            else
                AM_ILLEGAL_STATE();
        }

        /* Work-around for
           https://bugzilla.redhat.com/bugzilla/show_bug.cgi?id=242635. */
        if (status == 0x7f)
            status = 0x6057f;

	/* See if we know this thread already. If not we may need to
	   register a new thread or process */
        am_process_t *process;
        am_thread_t *thread;
        VECT_FOREACH(processes, process)
            VECT_FOREACH(process->threads, thread)
                if (thread->tid == tid)
                    goto thread_found;


        if (WIFEXITED(status) ||
            WIFSIGNALED(status ) ||
            WCOREDUMP(status))
            /* A launched process has terminated, nothing to do. */
            continue;

        /* Threads and processes created by a clone or fork syscall that we
           singlestep over may generate SIGTRAPs before we are even told that
           they exist (they are started with the trap flag set). If that is the
           case, just ignore the SIGTRAPs and and clear the thread's trap
           flag. */
        if (WIFSTOPPED(status) && WSTOPSIG(status) == SIGTRAP) {
            struct user_regs_struct regs;
            long res = ptrace(PT(PTRACE_GETREGS), tid, NULL, &regs);
            AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
            if (errno == ESRCH)
                memset(&regs, 0, sizeof(regs));
            regs.eflags &= 0xfffffffffffffeff;
            res = ptrace(PT(PTRACE_SETREGS), tid, NULL, &regs);
            /* Some kernels, e.g., vanilla 2.6.28, don't allow us to write the
               registers here (but on the other hand it doesn't seem to have
               the problem we are trying to work around), so be tolerant of
               errors. */
            /* AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH); */
            res = ptrace(PT(PTRACE_CONT), tid, NULL, (void *)(long)0);
            AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
            goto signal_handled;
        }

        /* Threads and processes created by a clone or fork syscall that we
           trace over may generate syscall exit notifications before we are
           even told that they exist. If that is the case, just ignore the
           notification. */
        if (WIFSTOPPED(status) && WSTOPSIG(status) == (SIGTRAP | 0x80)) {
            long res = ptrace(PT(PTRACE_CONT), tid, NULL, (void *)(long)0);
            AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
            goto signal_handled;
        }

        AM_ASSERT(WIFSTOPPED(status) && WSTOPSIG(status) == SIGSTOP);

        /* We may have received a SIGSTOP to the child thread created by a
           clone syscall. */
        {
            pid_t pid = get_pid(tid);
            AM_ASSERT_WITH_ERRNO(pid != -1);
            VECT_FOREACH(processes, process)
                if (process->pid == pid) {
                    errno = 0;
                    long rip = ptrace(PT(PTRACE_PEEKUSER), tid,
                                      (void *)offsetof(struct user, regs.PREFIX_REG(ip)),
                                      NULL);
                    AM_ASSERT_WITH_ERRNO(errno == 0 || errno == ESRCH);
                    if (errno == ESRCH)
                        rip = 0;

                    add_thread(process, tid, 1, addr_in_instrumenter(process, rip));
                    goto signal_handled;
                }
        }

        /* We may have received a SIGSTOP to the child process created
           by a fork syscall. If we knew about the parent, then we
           should also keep tracking the child. If the original
           process was instrumented it will currently do a standalone
           detach. If it was not instrumented, we may want to track
           it. */
        {
            pid_t ppid = get_ppid(tid);
            AM_ASSERT_WITH_ERRNO(ppid != -1);
            VECT_FOREACH(processes, process)
                if (process->pid == ppid) {
		    am_process_t *child =
			add_process(tid, ppid, std::string(), process->generation + 1);

		    /* Currently, this only works if our predicate
		       doesn't tell us to sample both parent and
		       child. To make this work, the sampler has to be
		       updated to make a decision to do a standalone
		       detach or a sampler restart based on the
		       predicate. This would involve some
		       communication between the child process and the
		       monitor. */
		    AM_ASSERT(child->ptrace_purpose != AM_PPP_Sample_And_Fork ||
			      process->ptrace_purpose != AM_PPP_Sample_And_Fork);

		    /* Track this process, 0 means it is already
		       attached in a ptrace sense */
		    attach_process(child, 0);

		    /* set up --start-at-... or --delay or start immediately  */
		    setup_start_conditions(child);
                    goto signal_handled;
                }
        }

        AM_ILLEGAL_STATE();

    thread_found:
        if (WIFSTOPPED(status)) {
            thread->debug_stopped = 1;

            siginfo_t siginfo;
            long res = ptrace(PT(PTRACE_GETSIGINFO), thread->tid, NULL, &siginfo);
            int have_siginfo = res != -1;

	    if (WSTOPSIG(status) == SIGTRAP && process->is_execing) {
		/* We have completed a single step after exec. Now is
		 * the time to reinitialize the sampler */
		process->is_execing = 0;
	        continue_after_exec(process);
	    }
            else if (WSTOPSIG(status) == SIGTRAP &&
                PTRACE_EVENT(status) == PTRACE_EVENT_EXIT) {

                /* Notification from ptrace that the thread is about to
                   exit. */
                pid_t tid = thread->tid;
                remove_thread(thread);
                res = ptrace(PT(PTRACE_DETACH), tid, NULL, NULL);
                AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);

                prune_processes();
            }
            else if (WSTOPSIG(status) == SIGTRAP &&
                     PTRACE_EVENT(status) == PTRACE_EVENT_CLONE) {
                /* Notification from ptrace that the thread has performed a
                   clone. These notifications are unreliable when tracing
                   syscalls on some kernels, so we don't use them. See ticket
                   #793. */
            }
            else if (WSTOPSIG(status) == SIGTRAP &&
                     PTRACE_EVENT(status) == PTRACE_EVENT_FORK) {
                /* Monitored process forked. We handle the child process when
                   we get the SIGSTOP from it, so there is nothing to do
                   here. */
            }
            else if (WSTOPSIG(status) == SIGTRAP &&
                     PTRACE_EVENT(status) == PTRACE_EVENT_VFORK) {
                am_printf(AM_Ctrl_Msg_Error,
			  "Sampled process vforked, not supported. "
			  "Detaching from sampled process. The sampled process "
			  "is most likely in a corrupted state.\n");
                force_exit(AM_Exit_Syscall);
            }
            else if (WSTOPSIG(status) == SIGTRAP &&
                     PTRACE_EVENT(status) == PTRACE_EVENT_EXEC) {

		/* after exec. Postpone sampler reinitialization by
		 * singlestepping once. This avoids bug
		 * https://bugzilla.redhat.com/show_bug.cgi?id=520206
		 */
		thread->single_step = 1;
		process->is_execing = 1;
            }
            else if (WSTOPSIG(status) == SIGTRAP &&
                     (read_dr(thread, 6) & 0xf) != 0x0) {
                /* A watchpoint set in a watchpoint register was triggered,
                   call the associated callback. */
                unsigned long dr6 = read_dr(thread, 6);
                for (int i = 0; i < NUM_BPS; i++)
                    if ((dr6 & (1 << i)) != 0) {
                        bp_t *bp = &thread->bps[i];
                        /* The breakpoint may have been removed, but
                           the breakpoint register not updated yet. */
                        if (bp->callback != 0)
                            bp->callback(thread, bp->bp_id, i);
                    }
            }
            else if (have_siginfo &&
                     siginfo.si_signo == SIGSTOP &&
                     siginfo.si_code == SI_TKILL &&
                     thread->sigstop_pending) {
                /* A SIGSTOP from tkill() used by the monitor to stop the
                   thread. */
                thread->sigstop_pending = 0;
            }
            else if (WSTOPSIG(status) == SIGTRAP &&
                     /* Here we should really look at the siginfo and expect
                        TRAP_BRKPT in si_code. However, the kernel now and then
                        seems to give us various garbage in si_code, so we
                        don't. */
                     thread->sigtrap_pending) {
                /* A breakpoint SIGTRAP while single stepping means the thread
                   has finished single stepping. */

                /* Re-enable any breakpoints that were suspended. */
                for (int i = 0; i < NUM_BPS; i++) {
                    bp_t *bp = &thread->bps[i];
                    if (bp->suspended) {
                        unsigned long dr7 = read_dr(thread, 7);
                        dr7 |= 1UL << (i * 2);
                        write_dr(thread, 7, dr7);
                        bp->suspended = 0;
                    }
                }

                if (thread->syscall_state == 3) {
                    if (get_rip(thread) == thread->syscall_ret_addr)
                        thread->syscall_ret_val = get_rax(thread);
                    if (!thread_in_syscall(thread))
                        syscall_complete(thread);
                }
            }
            else if (WSTOPSIG(status) == (SIGTRAP | 0x80)) {
                /* The thread is entering or exiting a syscall. */

                switch (thread->syscall_state) {
                case 1: {
                    struct user_regs_struct regs;
                    res = ptrace(PT(PTRACE_GETREGS), thread->tid, NULL, &regs);
                    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
                    if (errno == ESRCH)
                        memset(&regs, 0, sizeof(regs));

                    /* Some kernels give us an execve exit when launching a
                       process, others don't. If we launched this process and
                       the first syscall notification we get is an execve,
                       assume that its from the launch. */
                    if (process->first_syscall) {
                        process->first_syscall = 0;

                        int is_execve;
                        switch (process->bitness) {
/*                        case 32:
                            is_execve = regs.SCNO == SYS_execve_32;
                            break;
*/
#if defined(__x86_64__)
                        case 64:
                            is_execve = regs.SCNO == SYS_execve_64;
                            break;
#endif
                        default:
                            AM_ILLEGAL_STATE();
                        }
                        if (process->binary.size() != 0 && is_execve)
                            goto skip_syscall;
                    }

                    thread->syscall_num = regs.SCNO;
                    switch (process->bitness) {
/*                    case 32:
                        thread->syscall_args[0] = (uint32_t)regs.PREFIX_REG(bx);
                        thread->syscall_args[1] = (uint32_t)regs.PREFIX_REG(cx);
                        thread->syscall_args[2] = (uint32_t)regs.PREFIX_REG(dx);
                        thread->syscall_args[3] = (uint32_t)regs.PREFIX_REG(si);
                        thread->syscall_args[4] = (uint32_t)regs.PREFIX_REG(di);
                        thread->syscall_args[5] = (uint32_t)regs.PREFIX_REG(bp);
                        break;
*/
#if defined(__x86_64__)
                    case 64:
                        thread->syscall_args[0] = regs.rdi;
                        thread->syscall_args[1] = regs.rsi;
                        thread->syscall_args[2] = regs.rdx;
                        thread->syscall_args[3] = regs.r10;
                        thread->syscall_args[4] = regs.r8;
                        thread->syscall_args[5] = regs.r9;
                        break;
#endif
                    default:
                        AM_ILLEGAL_STATE();
                    }
                    thread->syscall_ret_addr = regs.PREFIX_REG(ip);
                    thread->syscall_state = 2;

                    /* We can't allow uninstrumented threads to do rt_sigaction
                       syscalls, since that may replace the instrumenter's
                       signal handlers. Instead we replace them with a harmless
                       getpid syscall and a notification to the instrumenter.

                       We don't handle the obsolete signal and sigaction
                       syscalls, but print a warning.*/
                    if (thread_native(thread, 1) &&
                        instrumenter_loaded(process))
                        switch (process->bitness) {
 /*                       case 32:
                            switch (thread->syscall_num) {
                            case SYS_signal_32:
                                am_printf(AM_Ctrl_Msg_Warn, "Syscall 'signal' not "
                                          "handled, may cause sampling stability "
                                          "problems.");
                                break;
                            case SYS_sigaction_32:
                                am_printf(AM_Ctrl_Msg_Warn, "Syscall 'sigaction' "
                                          "not handled, may cause sampling "
                                          "stability problems.");
                                break;
                            case SYS_rt_sigaction_32: {
                                long res = ptrace(PT(PTRACE_POKEUSER), thread->tid,
                                                  SCNO_OFFS,
                                                  (void *)SYS_getpid_32);
                                AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
                            } break;
                            default:
                                break;
                            }
                            break;
*/
#if defined(__x86_64__)
                        case 64:
                            switch (thread->syscall_num) {
                            case SYS_rt_sigaction_64: {
                                long res = ptrace(PT(PTRACE_POKEUSER), thread->tid,
                                                  SCNO_OFFS,
                                                  (void *)SYS_getpid_64);
                                AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
                            } break;
                            default:
                                break;
                            }
                            break;
#endif
                        default:
                            AM_ILLEGAL_STATE();
                        }
                } break;
                case 2:
                    thread->syscall_ret_val = get_rax(thread);
                    thread->syscall_state = 3;
                    break;
                case 3:
                    if (get_rip(thread) != thread->syscall_ret_addr)
                        syscall_complete(thread);
                    else
                        thread->syscall_ret_val = get_rax(thread);
                    break;
                default:
                    AM_ILLEGAL_STATE();
                }

            skip_syscall:;
            }
            else if (have_siginfo &&
                     siginfo.si_signo == SIGILL &&
                     page_hijacked(process)) {
                /* A SIGILL from the kernel (raised by ud2)
                   is used to signal errors during bootstrapping and
                   debootstrapping. */

                /* TODO: The bootstrapper or debootstrapper signalled an error
                   by raising a SIGILL. Restore the hijacked page and the
                   registers. */

                struct user_regs_struct regs;
                res = ptrace(PT(PTRACE_GETREGS), thread->tid, NULL, &regs);
                AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
                if (errno == ESRCH)
                    memset(&regs, 0, sizeof(regs));

                long errno_ = regs.PREFIX_REG(ax);
                long location = regs.PREFIX_REG(cx);

                struct rlimit rlim;
                int res = getrlimit(RLIMIT_AS, &rlim);
                AM_ASSERT_WITH_ERRNO(res == 0);

                if (location == ERR_MAP_INSTR &&
                    (rlim.rlim_cur != RLIM_INFINITY ||
                     rlim.rlim_max != RLIM_INFINITY)) {
                    am_printf(AM_Ctrl_Msg_Error,
                              "Error: Failed to allocate memory for the "
			      "sampler in the sampled process.\n\n"
                              "If the sampled process has a resource limit on "
                              "virtual memory you may want to try to increase "
                              "or remove it, e.g., \"ulimit -S -v unlimited\". "
                              "Alternatively you can reduce the amount of memory "
                              "allocated by the sampler with the --preallocation "
                              "argument to the sample script.\n");
                }
                else
                    am_printf(AM_Ctrl_Msg_Error,
                              "Error: Thread %d failed to bootstrap, got error %ld "
                              "\"%s\" in operation 0x%lx.\n",
                              thread->tid, errno_, strerror(errno_), location);

                abort();
            }
            else {
                /* The signal was not synthetic. Forward the signal to the
                   thread if it is not bootstrapping or debootstrapping,
                   otherwise drop it and print a warning.

                   TODO: Queue the signal instead of dropping it if the thread
                   is bootstrapping or debootstrapping. */

                if (SIGSEGV_DEBUG &&
                    have_siginfo &&
                    siginfo.si_signo == SIGSEGV &&
                    !process->pages->is_protected((unsigned long)siginfo.si_addr))
                    sigsegv_debug(thread, &siginfo);

                if (have_siginfo &&
                    process->pages->is_protected((unsigned long)siginfo.si_addr) &&
                    thread_native(thread, 1)) {
                    thread->notification_args.type = Notification_Sigsegv;
                    thread->notification_args.u.sigsegv.addr = (unsigned long)siginfo.si_addr;
                    thread->state = AM_TS_Notification_Wait;
                }
                else if (page_hijacked(process))
                    /* TODO: We should only drop the signal if this actually is
                       the thread being bootstrapped/debootstrapped. We can
                       still deliver the signal if there is another thread
                       being bootstrapped/debootstrapped. */
                    am_printf(AM_Ctrl_Msg_Error,
			      "Dropping signal %d received by thread %d during "
			      "bootstrapping/debootstrapping.\n",
			      WSTOPSIG(status), thread->tid);
                else
                    resume_thread(thread, WSTOPSIG(status));
            }
        }
        else if (WIFEXITED(status)) {
            // TODO: We could get here for the launched process if we allow
            // monitoring of multiple processes.

            // We should already have detached from the thread when we got the
            // PT(PTRACE_EVENT_EXIT), so we should never get here.
            //AM_ILLEGAL_STATE();

	    /* Copied from WSTOPSIG(status) == SIGTRAP && PTRACE_EVENT(status) == PTRACE_EVENT_EXIT
	       case above.  See TS-2250.  */
	    pid_t tid = thread->tid;
	    remove_thread(thread);
	    int res = ptrace(PT(PTRACE_DETACH), tid, NULL, NULL);
	    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);

	    prune_processes();
        }
        else if (WIFSIGNALED(status)) {
            // TODO: We could get here for the launched process if we allow
            // monitoring of multiple processes.

            // We should already have detached from the thread when we got the
            // PT(PTRACE_EVENT_EXIT), so we should never get here.

            //AM_ILLEGAL_STATE();

            // We sometimes get here anyway because of buggy kernels, see of
            // ticket #817. Exit with an ok error code so that the ATF file is
            // resampled.

            exit(0);
        }
        else if (WCOREDUMP(status)) {
            // TODO: We could get here for the launched process if we allow
            // monitoring of multiple processes.

            // We should already have detached from the thread when we got the
            // PT(PTRACE_EVENT_EXIT), so we should never get here.
            AM_ILLEGAL_STATE();
        }
        else if (WIFCONTINUED(status)) {
            /* We don't ask for continuation events, so we shouldn't get
               any. */
            AM_ILLEGAL_STATE();
        }
        else
            AM_ILLEGAL_STATE();
    }
}

void
os_dependent_init(void)
{
    add_signal_event(SIGCHLD, sigchld_event);
}

void
os_dependent_fini(void)
{
    remove_signal_event(SIGCHLD, sigchld_event);
}

void
check_os_version(int major, int minor) {
    if (major > 2 || (major == 2 && minor >= 6))
        return;
    am_printf(AM_Ctrl_Msg_Error,
	      "Error: This computer seems to be running a %d.%d kernel, a "
	      "2.6 kernel or newer is required.\n", major, minor);
    force_exit(AM_Exit_OS_Unsupported);
}


pid_t
get_pid(pid_t tid)
{
    pid_t pid;
    int res = get_status_field(tid, "Tgid:\t%d\n", &pid);
    return res == -1 ? -1 : pid;
}

int
get_process_bitness(const am_process_t *process, tid_t tid)
{
    int res;
    int bitness;

    int fd = safe_open(process->elf_file, O_RDONLY);
    AM_ASSERT_WITH_ERRNO(fd != -1);

    elf_version(1);

    Elf *elf = elf_begin(fd, ELF_C_READ, NULL);
    AM_ASSERT(elf != NULL);

    GElf_Ehdr ehdr;
    GElf_Ehdr *ehdr_p = gelf_getehdr(elf, &ehdr);
    AM_ASSERT(ehdr_p != NULL);

    switch (ehdr.e_machine) {
    case EM_386:
        bitness = 32;
        break;
    case EM_X86_64:
        bitness = 64;
        break;
    default:
        AM_ILLEGAL_STATE();
    }

    elf_end(elf);

    res = safe_close(fd);
    AM_ASSERT_WITH_ERRNO(res != -1);

    return bitness;
}

static int
thread_is_zombie(pid_t tid)
{
    char state;
    int res = get_status_field(tid, "State:\t%c", &state);
    return res == -1 || state == 'Z';
}

static uid_t
get_process_uid(const pid_t pid)
{
    struct stat s;
    char process_path[PATH_MAX];
    int res;
    res = snprintf(process_path, PATH_MAX, "/proc/%d", pid);
    AM_ASSERT(res >= 0 && res < PATH_MAX);

    res = stat(process_path, &s);
    AM_ASSERT(res == 0);

    return s.st_uid;
}

static void
check_process_owner(const pid_t pid, int require_ownership)
{
    /* Check if the process executes with the same UID as the
       monitor. We normally don't allow the monitor to attach to other
       uses processes due to security concerns. */
    const uid_t process_uid = get_process_uid(pid);
    const uid_t my_uid = getuid();
    if (process_uid != my_uid) {
	if (require_ownership) {
	    am_printf(AM_Ctrl_Msg_Error,
		      "The target process is running with UID %i, but the monitor is running with\n"
		      "UID %i. Attaching to other uses processes is not supported. Aborting.\n",
		      process_uid, my_uid);
	    force_exit(AM_Exit_Target_Attach);
	} else
	    am_printf(AM_Ctrl_Msg_Warn,
		      "The target process is running with UID %i, but the monitor is running with\n"
		      "UID %i. This is unsupported.\n",
		      process_uid, my_uid);
    }
}

/* Attach to the process. */
int
attach_process(am_process_t *process, int do_debug_attach)
{
    // TODO: make sure that all threads have stopped (no new have been started)

    int new_threads;

    do {
        new_threads = 0;

        long res;
        char task_path[PATH_MAX];
        res = snprintf(task_path, PATH_MAX, "/proc/%d/task", process->pid);
        AM_ASSERT(res >= 0 && res < PATH_MAX);
        DIR *task_dir = opendir(task_path);
        AM_ASSERT_WITH_ERRNO(task_dir != NULL);

        struct dirent *thread_dir;
        while ((thread_dir = readdir(task_dir)) != NULL) {
            char *endptr;
            if (strcmp(thread_dir->d_name, ".") == 0 ||
                strcmp(thread_dir->d_name, "..") == 0)
                continue;
            pid_t tid = strtol(thread_dir->d_name, &endptr, 10);
            AM_ASSERT(*endptr == '\0');

            am_thread_t *thread;
            VECT_FOREACH(process->threads, thread)
                if (thread->tid == tid)
                    /* Already attached to thread. */
                    goto dont_attach;

	    if (do_debug_attach) {
		res = ptrace(PT(PTRACE_ATTACH), tid, NULL, NULL);
		if (res == -1) {
		    if (errno == ESRCH)
			/* The thread exited before we got a chance to attach. */
			continue;
		    else if (errno == EPERM && thread_is_zombie(tid))
			/* Thread is a zombie. */
			continue;
		    else {
			am_printf(AM_Ctrl_Msg_Error,
				  "Failed to attach to thread %d: %s.\n",
				  tid, strerror(errno));
			force_exit(AM_Exit_Target_Attach);
		    }
		}


		int got_sigstop = 0;
		do {
		    int status = blocking_waitpid(tid);
		    if (WIFEXITED(status))
			/* Thread exited before it got the SIGSTOP. */
			goto dont_attach;
		    if (WIFSTOPPED(status) && WSTOPSIG(status) == SIGSTOP)
			/* We got the SIGSTOP from the PTRACE_ATTACH. */
			got_sigstop = 1;
		    else {
			/* We some other signal to the thread. Deliver it and try
			   again. */
			AM_ASSERT(WIFSTOPPED(status));
			res = ptrace(PT(PTRACE_CONT), tid, NULL, (void *)(long)WSTOPSIG(status));
			AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
		    }
		} while (!got_sigstop);
	    } /* if (do_ptrace_attach) */

            if (VECT_SIZE(process->threads) == 0)
                get_elf_file(process, tid);

            new_threads = 1;
            add_thread(process, tid, 1, 0);

        dont_attach:
            ;
        }

        res = closedir(task_dir);
        AM_ASSERT_WITH_ERRNO(res != -1);
    } while (new_threads);

    /* Check if we attach is supported from this UID */
    check_process_owner(process->pid, process->require_ownership);

    return 0;
}

void
launch_process(am_process_t *process)
{
    int num_args = VECT_SIZE(arguments);
    char **argv = (char **)am_malloc((num_args + 2) * sizeof(char **));
    argv[0] = const_cast<char *>(process->binary.c_str());
    for (int i = 0; i < num_args; i++)
        argv[i + 1] = VECT_GET(arguments, i);
    argv[num_args + 1] = NULL;

    process->pid = fork();
    AM_ASSERT_WITH_ERRNO(process->pid != -1);
    if (process->pid == 0) {
        long res;

        signal_prepare_for_launch();

        res = ptrace(PT(PTRACE_TRACEME), 0, NULL, NULL);
        AM_ASSERT_WITH_ERRNO(res != -1);

        res = execv_path(argv[0], &argv[0]);
        if (res == -1) {
            am_printf(AM_Ctrl_Msg_Error,
		      "Error: Failed to exec '%s'.\n", argv[0]);
            exit(1);
        }
	AM_ILLEGAL_STATE();
    }
    else {
        am_free(argv);

        int status = blocking_waitpid(process->pid);
        /* TODO: There are other possibilities here. */
        if (WIFSTOPPED(status) && WSTOPSIG(status) == SIGTRAP) {
            get_elf_file(process, process->pid);
            add_thread(process, process->pid, 1, 0);
        }
        else if (WIFEXITED(status)) {
            am_printf(AM_Ctrl_Msg_Error, "Error: Failed to launch process.\n");
            force_exit(AM_Exit_Target_Launch);
        }
        else
            AM_ILLEGAL_STATE();
    }
}

void
detach_thread(am_thread_t *thread)
{
    long res = ptrace(PT(PTRACE_DETACH), thread->tid, NULL, NULL);
    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
}

unsigned long
find_hijack_page(const am_thread_t *thread, unsigned long lower_limit)
{
    am_process_t *process = thread->process;

    int res;
    char maps_path[PATH_MAX];
    res = snprintf(maps_path, PATH_MAX, "/proc/%" PRItid "/maps", thread->tid);
    AM_ASSERT(res >= 0 && res < PATH_MAX);
    FILE *file = fopen(maps_path, "r");
    AM_ASSERT(file != NULL);

    /* Iterate over all mappings. */
    unsigned long start;
    unsigned long end;
    char r;
    char x;
    char s;
    unsigned int device;
    unsigned int inode;
    while ((res = fscanf(file, "%lx-%lx %c%*c%c%c %*x %x:%x %*i%*[^\n]",
                         &start, &end, &r, &x, &s, &device, &inode)) != EOF) {
        AM_ASSERT(res == 7);

        /* The mapping must be readable, executable and privately mapped. */
        if (r != 'r' || x != 'x' || s != 'p')
            continue;

        /* Make sure that we pick a page about the lower limit. */
        if (end <= lower_limit)
            continue;
        else
            if (lower_limit > start)
                start = lower_limit;

        /* The mapping must be a file mapping. This protects us from choosing a
           page belonging to the instrumenter or the virtual syscall page. */
        if (device == 0 && inode == 0)
            continue;

        while (start < end) {
            /* Make sure no thread is in a system call returning to an
               instruction on the page, or we may single step an instruction on
               the page because of a work-around for ticket #793. */
            am_thread_t *t;
            VECT_FOREACH(process->threads, t)
                if (t->syscall_state != 1 &&
                    start <= t->syscall_ret_addr + 16 &&
                    start + 4096 > t->syscall_ret_addr)
                    goto reject_page;

            res = fclose(file);
            AM_ASSERT(res != EOF);
            return start;

        reject_page:
            start += 4096;
        };
    }

    /* No suitable page found. */
    AM_ILLEGAL_STATE();
    return 0;
}

void
set_thread_debug_options(pid_t tid)
{
    long res = ptrace(PT(PTRACE_SETOPTIONS), tid, NULL,
                      (void *)(PTRACE_O_TRACEFORK | PTRACE_O_TRACEVFORK |
                               PTRACE_O_TRACECLONE | PTRACE_O_TRACEEXEC |
                               PTRACE_O_TRACEEXIT | PTRACE_O_TRACESYSGOOD));
    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
}

void
read_data(const am_thread_t *thread, unsigned long addr, unsigned long len, void *buf)
{
    unsigned long a = addr & ~((unsigned long)sizeof(long) - 1);
    char *b = (char *)buf;

    while (a < addr + len) {
        errno = 0;
        unsigned long d = ptrace(PT(PTRACE_PEEKTEXT), thread->tid, (void *)a, NULL);
        AM_ASSERT_WITH_ERRNO(errno == 0 || errno == ESRCH);
        if (errno == ESRCH)
            d = 0;

        int low_byte = MAX(a, addr) - a;
        int high_byte = MIN(a + sizeof(unsigned long), addr + len) - a;
        for (int i = low_byte; i < high_byte; i++)
            /* NOTE: Assumes little-endian host. */
            *(b++) = d >> (i * 8);

        a += sizeof(long);
    }

    AM_ASSERT(b - (char *)buf == len);
}

void
write_data(const am_thread_t *thread, unsigned long addr, unsigned long len, const void *buf, int ignore_zeros)
{
    unsigned long a = addr & ~((unsigned long)sizeof(long) - 1);
    const char *b = (char *)buf;

    while (a < addr + len) {
        int low_byte = MAX(a, addr) - a;
        int high_byte = MIN(a + sizeof(unsigned long), addr + len) - a;

        unsigned long d = 0;
        if (low_byte != 0 || high_byte != sizeof(unsigned long)) {
            errno = 0;
            d = ptrace(PT(PTRACE_PEEKTEXT), thread->tid, (void *)a, NULL);
            AM_ASSERT_WITH_ERRNO(errno == 0 || errno == ESRCH);
            if (errno == ESRCH)
                d = 0;
        }

        for (int i = low_byte; i < high_byte; i++) {
            /* NOTE: Assumes little-endian host. */
            d &= ~(0xffl << (i * 8));
            d |= ((unsigned long)*(b++)) << (i * 8);
        }

        if (!ignore_zeros || d != 0) {
            long res = ptrace(PT(PTRACE_POKETEXT), thread->tid, (void *)a, (void *)d);
            AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
        }

        a += sizeof(long);
    }

    AM_ASSERT(b - (char *)buf == len);
}

void
save_registers(am_thread_t *thread)
{
    struct user_regs_struct regs;

    long res = ptrace(PT(PTRACE_GETREGS), thread->tid, NULL, &regs);
    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
    if (errno == ESRCH)
        memset(&regs, 0, sizeof(regs));

    // Ptrace returns eflags with the trap flag set after you have single
    // stepped a thread. If you write eflags with the trap flag set the process
    // will continue to single step forever. We therefore have to clear the
    // trap flag.
    thread->regs.rflags = regs.eflags & 0xfffffffffffffeff;
    thread->regs.rip = regs.PREFIX_REG(ip);
    thread->regs.rax = regs.PREFIX_REG(ax);
    thread->regs.rcx = regs.PREFIX_REG(cx);
    thread->regs.rdx = regs.PREFIX_REG(dx);
    thread->regs.rbx = regs.PREFIX_REG(bx);
    thread->regs.rsp = regs.PREFIX_REG(sp);
    thread->regs.rbp = regs.PREFIX_REG(bp);
    thread->regs.rsi = regs.PREFIX_REG(si);
    thread->regs.rdi = regs.PREFIX_REG(di);
#if defined(__i386__)
    thread->regs.segreg = regs.xfs;
#else
    thread->regs.r8 = regs.r8;
    thread->regs.r9 = regs.r9;
    thread->regs.r10 = regs.r10;
    thread->regs.r11 = regs.r11;
    thread->regs.r12 = regs.r12;
    thread->regs.r13 = regs.r13;
    thread->regs.r14 = regs.r14;
    thread->regs.r15 = regs.r15;
    switch (thread->process->bitness) {
/*    case 32:
        thread->regs.segreg = regs.fs;
        break;
*/
    case 64:
        thread->regs.segreg = regs.gs;
        break;
    default:
        AM_ILLEGAL_STATE();
    }
#endif
}

void
restore_registers(const am_thread_t *thread)
{
    long res;

    struct user_regs_struct regs;

    res = ptrace(PT(PTRACE_GETREGS), thread->tid, NULL, &regs);
    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
    if (errno == ESRCH)
        memset(&regs, 0, sizeof(regs));

    regs.eflags = thread->regs.rflags;
    regs.PREFIX_REG(ip) = thread->regs.rip;
    regs.PREFIX_REG(ax) = thread->regs.rax;
    regs.PREFIX_REG(cx) = thread->regs.rcx;
    regs.PREFIX_REG(dx) = thread->regs.rdx;
    regs.PREFIX_REG(bx) = thread->regs.rbx;
    regs.PREFIX_REG(sp) = thread->regs.rsp;
    regs.PREFIX_REG(bp) = thread->regs.rbp;
    regs.PREFIX_REG(si) = thread->regs.rsi;
    regs.PREFIX_REG(di) = thread->regs.rdi;
#if defined(__i386__)
    regs.xfs = thread->regs.segreg;
#else
    regs.r8 = thread->regs.r8;
    regs.r9 = thread->regs.r9;
    regs.r10 = thread->regs.r10;
    regs.r11 = thread->regs.r11;
    regs.r12 = thread->regs.r12;
    regs.r13 = thread->regs.r13;
    regs.r14 = thread->regs.r14;
    regs.r15 = thread->regs.r15;
    switch (thread->process->bitness) {
//    case 32:
//        regs.fs = thread->regs.segreg;

        /* For 32-bit processes on 64-bit kernels we need to mask away the high
           bits of the segment bases (they are sign extended when read, but
           must not be when written). See ticket #606. */
//        regs.fs_base &= 0xffffffff;
//        regs.gs_base &= 0xffffffff;
//        break;
    case 64:
        regs.gs = thread->regs.segreg;
        break;
    default:
        AM_ILLEGAL_STATE();
    }
#endif

    res = ptrace(PT(PTRACE_SETREGS), thread->tid, NULL, &regs);
    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);

    /* Work-around for kernels, e.g.,ubuntu-2.6.15-53-amd64, that do not allow
       us to set eip to the vsyscall page. Push the address on the stack and
       set eip to a ret instruction instead.

       This problem should only occur with 32-bit processes on 64-bit kernels.

       See ticket #987. */
#if defined (__x86_64__)
    if (get_rip(thread) != regs.PREFIX_REG(ip)) {
        am_printf(AM_Ctrl_Msg_Info, "Kernel ptrace bug detected, using workaround.\n");
        AM_ASSERT(thread->process->bitness == 32);
        uint32_t rip = regs.PREFIX_REG(ip);

        /* Find a page with a ret instruction. */
        unsigned int lower_limit = 0;
        while (1) {
            lower_limit = find_hijack_page(thread, lower_limit);

            unsigned char buf[0x1000];
            read_data(thread, lower_limit, 0x1000, buf);

            for (unsigned long offs = 0; offs < 0x1000; ++offs)
                if (buf[offs] == 0xc3) {
                    /* Ret instruction found! */
                    unsigned long new_rsp = get_rsp(thread) - sizeof(rip);
                    set_rip(thread, lower_limit + offs);
                    set_rsp(thread, new_rsp);
                    write_data(thread, new_rsp, sizeof(rip), &rip, 0);
                    goto found_ret;
                }

            lower_limit += 0x1000;
        }
    found_ret:;
    }
#endif
}

void
set_rip(const am_thread_t *thread, unsigned long val)
{
    long res = ptrace(PT(PTRACE_POKEUSER), thread->tid,
                      (void *)offsetof(struct user, regs.PREFIX_REG(ip)),
                      (void *)val);
    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
}

void
set_rbp(const am_thread_t *thread, unsigned long val)
{
    long res = ptrace(PT(PTRACE_POKEUSER), thread->tid,
                      (void *)offsetof(struct user, regs.PREFIX_REG(bp)),
                      (void *)val);
    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
}

void
set_rsp(const am_thread_t *thread, unsigned long val)
{
    long res = ptrace(PT(PTRACE_POKEUSER), thread->tid,
                      (void *)offsetof(struct user, regs.PREFIX_REG(sp)),
                      (void *)val);
    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
}

unsigned long
get_rip(const am_thread_t *thread)
{
    errno = 0;
    long res = ptrace(PT(PTRACE_PEEKUSER), thread->tid,
                      (void *)offsetof(struct user, regs.PREFIX_REG(ip)),
                      NULL);
    AM_ASSERT_WITH_ERRNO(errno == 0 || errno == ESRCH);
    return errno == 0 ? res : 0;
}

unsigned long
get_rsp(const am_thread_t *thread)
{
    errno = 0;
    long res = ptrace(PT(PTRACE_PEEKUSER), thread->tid,
                      (void *)offsetof(struct user, regs.PREFIX_REG(sp)),
                      NULL);
    AM_ASSERT_WITH_ERRNO(errno == 0 || errno == ESRCH);
    return errno == 0 ? res : 0;
}

void
set_syscall_ret(const am_thread_t *thread, unsigned long val)
{
    long res = ptrace(PT(PTRACE_POKEUSER), thread->tid,
                      (void *)offsetof(struct user, regs.PREFIX_REG(ax)),
                      (void *)val);
    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
}

static void
tkill(pid_t tid, int signum)
{
    int res = syscall(SYS_tkill, (long)tid, (long)signum);
    AM_ASSERT_WITH_ERRNO(res != -1);
}

int
thread_stop(am_thread_t *thread)
{
    AM_ASSERT(!thread->debug_stopped);

    /* Don't send another SIGSTOP to a thread that already has one pending. */
    if (thread->sigstop_pending)
        return 0;

    tkill(thread->tid, SIGSTOP);
    thread->sigstop_pending = 1;
    return 1;
}

void
thread_run(am_thread_t *thread)
{
    thread->exiting_syscall = 0;
    resume_thread(thread, 0);
}

void
thread_exit_syscall(am_thread_t *thread)
{
    thread->exiting_syscall = 1;
    resume_thread(thread, 0);
}

void
force_detach_thread(am_thread_t *thread)
{
    long res;

    if (thread->sigstop_pending && thread->debug_stopped) {
        res = ptrace(PT(PTRACE_CONT), thread->tid, NULL, (void *)0);
        AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
        thread->debug_stopped = 0;
    }

    if (!thread->debug_stopped) {
        if (!thread->sigstop_pending)
            tkill(thread->tid, SIGSTOP);

        while (1) {
            int status = blocking_waitpid(thread->tid);

            if (WIFSTOPPED(status)) {
                siginfo_t siginfo;
                int res = ptrace(PT(PTRACE_GETSIGINFO), thread->tid,
                                 NULL, &siginfo);
                int have_siginfo = res != -1;

                long cont_sig = 0;

                if (WSTOPSIG(status) == SIGTRAP &&
                    PTRACE_EVENT(status) == PTRACE_EVENT_EXIT)
                    /* The thread has exited. */
                    return;
                else if (WSTOPSIG(status) == SIGTRAP &&
                         PTRACE_EVENT(status) == PTRACE_EVENT_CLONE) {
                    /* The thread has cloned. We must wait for the SIGSTOP to
                       the new thread and then detach it. */
                    long tid;
                    res = ptrace(PT(PTRACE_GETEVENTMSG), thread->tid, NULL, &tid);
                    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
                    if (errno == ESRCH)
                        tid = 0;

                    int status = blocking_waitpid(tid);
                    AM_ASSERT(WIFSTOPPED(status) &&
			      (WSTOPSIG(status) == SIGSTOP ||
			       WSTOPSIG(status) == SIGTRAP));

                    res = ptrace(PT(PTRACE_DETACH), thread->tid, NULL, NULL);
                    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
                }
                else if (WSTOPSIG(status) == SIGTRAP &&
                         thread->sigtrap_pending)
                    /* Finished single-stepping. */
                    thread->sigtrap_pending = 0;
                else if (have_siginfo &&
                         siginfo.si_signo == SIGSTOP &&
                         siginfo.si_code == SI_TKILL)
                    /* We got the pending stop signal. */
                    break;
                else
                    /* The signal was not synthetic. Forward it to the
                       thread. */
                    cont_sig = WSTOPSIG(status);

                res = ptrace(PT(PTRACE_CONT), thread->tid, NULL, (void *)cont_sig);
                AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
            }
            else
                AM_ILLEGAL_STATE();
        }
    }

    if (thread->pending_msgs > 0 && thread->msg_counter_addr != 0)
        update_msg_counter(thread);

    pid_t tid = thread->tid;
    remove_thread(thread);
    res = ptrace(PT(PTRACE_DETACH), tid, NULL, NULL);
    AM_ASSERT_WITH_ERRNO(res != -1 || errno == ESRCH);
}

void
update_bps(am_thread_t *thread)
{
    unsigned long dr7_enable_mask = 0;

    for (int i = 0; i < NUM_BPS; i++) {
        bp_t *bp = &thread->new_bps[i];
        if (bp->callback != 0) {
            write_dr(thread, i, bp->address);
            dr7_enable_mask |= 1UL << (i * 2);
        }
        thread->bps[i] = *bp;
    }

    unsigned long dr7 = read_dr(thread, 7);
    dr7 &= ~0x55UL;
    dr7 |= dr7_enable_mask;
    write_dr(thread, 7, dr7);

    thread->update_bps = 0;
}

void
disable_bp(am_thread_t *thread, int bp_num)
{
    unsigned long dr7 = read_dr(thread, 7);
    dr7 &= ~(1UL << (bp_num * 2));
    write_dr(thread, 7, dr7);
}

unsigned long long
get_time(void)
{
    struct timeval tv;
    int res = gettimeofday(&tv, NULL);
    AM_ASSERT_WITH_ERRNO(res != -1);
    return (tv.tv_usec * 1000ULL +
            tv.tv_sec * 1000000000ULL);
}

void
set_hold_signals(am_thread_t *thread, int hold)
{
    /* Ptrace offers no way to hold signals. We drop the signals when they are
       received instead. */
}

int
suspend_activity(am_process_t *process)
{
    /* Linux has no thread suspension. */
    return 0;
}

int
sigsegv_pending(const am_thread_t *thread)
{
    /* Threads should never have SIGSEGV pending on Linux. */
    return 0;
}

int
any_sigsegv_pending(const am_process_t *process)
{
    /* Threads should never have SIGSEGV pending on Linux. */
    return 0;
}

/* This function accesses data structures unsafely from a signal
   handler. Acumon may have been updating the data structures when it got the
   SIGINT so they may be in an inconsistent state. But it's only for
   debugging... */
void
sigint_debug(void)
{
    am_process_t *process;
    VECT_FOREACH(processes, process) {
        am_printf(AM_Ctrl_Msg_Debug,
                  "process %" PRIpid ":\n"
                  "    state %s\n"
                  "    instrumentation state %s\n"
                  "    execution state %s\n"
                  "    socket %s\n",
                  process->pid,
                  process_state_name(process->state),
                  process_instrumentation_state_name(process->instrumentation_state),
                  process_execution_state_name(process->execution_state),
                  process->socket_fd == -1 ? "closed" : "open");

        am_thread_t *thread;
        VECT_FOREACH(process->threads, thread)
            am_printf(AM_Ctrl_Msg_Debug, "  thread %" PRItid ":\n"
                      "    state %s\n"
                      "    %s\n"
                      "%s%s%s"
                      "    %d messages pending\n",
                      thread->tid,
                      thread_state_name(thread->state),
                      thread->debug_stopped ? "stopped" : "running",
                      thread_stopped(thread, 1)
                      ? (thread_in_syscall(thread) ? "    in syscall\n" : "")
                      : (thread_exiting_syscall(thread) ? "    exiting syscall\n" : ""),
                      thread->sigstop_pending ? "    SIGSTOP pending\n" : "",
                      thread->sigtrap_pending ? "    SIGTRAP pending\n" : "",
                      thread->pending_msgs);
    }
}

/******************** Functions from thread_predicates.h. ********************/

int
thread_stopped(const am_thread_t *thread, int strict)
{
    return thread->debug_stopped || (!strict && thread->syscall_state != 1 && thread->sigstop_pending);
}

int
thread_in_syscall(const am_thread_t *thread)
{
    /* Older kernels, e.g., 2.6.9-67.0.4.ELxenU, may return any negative number
       when not in a syscall, and not necessarily -1. */
    return thread_syscall_num(thread) >= 0;
}

int
thread_suspended(const am_thread_t *thread)
{
    /* Linux has no thread suspension. */
    return 0;
}

int
thread_detachable(const am_thread_t *thread)
{
    return thread_stopped(thread, 1) && !thread->sigstop_pending;
}
