/// @file acumon/os_dependent.h

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

/*
 * Operating system dependent functions, implemented in linux.c and solaris.c.x
 */

#ifndef OS_DEPENDENT_HH
#define OS_DEPENDENT_HH

#include "acumon.hh"

/* Do any initialization needed by the operating system dependent code. */
void os_dependent_init(void);

/* Do any cleanup needed by the operating system dependent code. */
void os_dependent_fini(void);

/* Check that the host is running a supported operating system version. If not,
   print an error message and exit with an error. */
void check_os_version(int major, int minor);

/* Return the process id of the specified thread if processes and threads share
   namespace, otherwise just return the argument. Returns -1 on error. */
pid_t get_pid(pid_t tid);

/* Return the bitness of the process, 32 or 64.

   The tid argument must be the tid of a live thread on Linux, since
   /proc/<pid>/exe does not work if the initial thread of the process has
   terminated. It is unused on Solaris. */
int get_process_bitness(const am_process_t *process, tid_t tid);

/* Attach to the specified process. Returns 0 on success, prints an error
   message and returns -1 on error. */
int attach_process(am_process_t *process, int do_debug_attach);

/* Launch the specified process. */
void launch_process(am_process_t *process);

/* Detach from the specified thread. */
void detach_thread(am_thread_t *thread);

/* Find a suitable page to hijack for bootstrapping or debootstrapping and
   return its address.

   The thread argument must be a live thread, since /proc/<pid>/maps does not
   work on Linux if the initial thread of the process has terminated. */
unsigned long find_hijack_page(const am_thread_t *thread, unsigned long lower_limit);

/* Set any per-thread debug options in the debug interface. */
void set_thread_debug_options(pid_t pid);

/* Read len bytes from address addr in the monitored process to buf.

   The thread must be specified since PTRACE_PEEKTEXT does not work on the
   initial thread of the process on Linux if the initial thread has
   terminated. */
void read_data(const am_thread_t *thread, unsigned long addr, unsigned long len, void *buf);

/* Write len bytes from buf to address addr in the monitored process.

   The thread must be specified since PTRACE_POKETEXT does not work on the
   initial thread of the process on Linux if the initial thread has terminated.

   If ignore_zeros is true zero bytes do not need to be written (but may still
   be). */
void write_data(const am_thread_t *thread, unsigned long addr, unsigned long len, const void *buf, int ignore_zeros);

/* Save the registers of the thread to the regs struct in the am_thread_t
   struct. */
void save_registers(am_thread_t *thread);

/* Restore the registers of the thread from the regs struct in the am_thread_t
   struct. */
void restore_registers(const am_thread_t *thread);

/* Set rip of the thread to the specified value. */
void set_rip(const am_thread_t *thread, unsigned long val);

/* Set rbp of the thread to the specified value. */
void set_rbp(const am_thread_t *thread, unsigned long val);

/* Set rsp of the thread to the specified value. */
void set_rsp(const am_thread_t *thread, unsigned long val);

/* Return rip of the specified thread. */
unsigned long get_rip(const am_thread_t *thread);

/* Return rsp of the specified thread. */
unsigned long get_rsp(const am_thread_t *thread);

/* Insert the specified system call return value into the registers of the
   thread. */
void set_syscall_ret(const am_thread_t *thread, unsigned long val);

/* Request the thread to stop as soon as possible. The thread may be stopped
   asynchronously, so it may not be stopped when the call returns. Returns 1 if
   a stop request was sent, 0 if no stop request was sent because there already
   was a pending stop request for the thread. Must only be called when the
   thread is running. */
int thread_stop(am_thread_t *thread);

/* Run the thread. Must only be called when the thread is stopped. */
void thread_run(am_thread_t *thread);

/* Run the thread until it exits the syscall it is in. The thread will be
   resumed and will stop asynchronously when it has exited the syscall. Must
   only be called when the thread is stopped in a system call. */
void thread_exit_syscall(am_thread_t *thread);

/* Forcefully detach from the thread, even if it leaves the thread in a
   currupted state. */
void force_detach_thread(am_thread_t *thread);

/* Update the breakpoints of the thread to match bp_addrs and bp_cbs. */
void update_bps(am_thread_t *thread);

/* Disable the specified breakpoint. */
void disable_bp(am_thread_t *thread, int bp_num);

/* Returns the current time in nanoseconds since 1970-01-01 00:00:00 UTC. */
unsigned long long get_time(void);

/* Enable or disable holding of signals to the specified thread in the
   debugging interface, if it is possible. */
void set_hold_signals(am_thread_t *thread, int hold);

/* Return true if there are any thread suspension operations going on. */
int suspend_activity(am_process_t *process);

/* Return true if the thread has a SIGSEGV pending. */
int sigsegv_pending(const am_thread_t *thread);

/* Return true if any thread in the process has a SIGSEGV pending. */
int any_sigsegv_pending(const am_process_t *process);

#if defined(__sun__)
/* Returns true if a signal is just about to be delivered. In that case trying
   to change the pc of the thread is not reliable, it may be overwritten by the
   address of the signal handler. */
int signal_imminent(const am_thread_t *thread);

/* Make the thread runnable for a short while without running any userspace
   code. Gives lwp_suspend from another thread a chance to take effect. */
void poke_thread(am_thread_t *thread);
#endif /* defined(__sun__)  */

/* Print status information for debugging. Called from a SIGINT handler. */
void sigint_debug(void);

#endif /* OS_DEPENDENT_HH */
