/// @file acumon/path_utils.c

// Copyright (c) 2009 Acumem AB.  All rights reserved.

#define _XOPEN_SOURCE 600

#include "path_utils.hh"

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

/* Get the current effective PATH or NULL on error. The returned
   pointer must be free'd with free(3). */
static char *
get_path()
{
    char *env_path = getenv("PATH");

    if (!env_path) {
	size_t path_len = confstr(_CS_PATH, NULL, 0);
	size_t new_path_len;
	char *path;
	if (!path_len)
	    return NULL;

	path = (char *)malloc(path_len);
	new_path_len = confstr(_CS_PATH, path, path_len);
	assert(new_path_len == path_len);

	return path;
    } else
	return strdup(env_path);
}

static int
execv_try_path(const char *file, char *const argv[])
{
    char *path = get_path();
    char *path_end;
    size_t path_len;
    int got_eaccess = 0;

    if (!path) {
	errno = ENOENT;
	return -1;
    }

    path_len = strlen(path);
    path_end = path + path_len;

    char *current = path;
    char *next;
    char *new_file = (char *)malloc(strlen(file) + path_len + 3);
    if (!new_file) {
	errno = ENOMEM;
	return -1;
    }
    do {
	size_t new_len;
	next = strchr(current, ':');
	if (!next)
	    next = path_end;
	*next = '\0';

	new_len = next - current;
	if (new_len > 0)
	    memcpy(new_file, current, new_len);
	else
	    /* If the current path element is the empty string, check
	       the current working directory. Don't know if this is
	       standard, but this is what GLIBC does. */
	    new_file[new_len++] = '.';

	new_file[new_len++] = '/';
	strcpy(&new_file[new_len], file);

	execv(new_file, argv);
	switch (errno) {
	case EACCES:
	    /* Only return EACCESS if we fail to find a suitable
	     * executable. */
	    got_eaccess = 1;
	    break;

	case ENOENT:
	case ESTALE:
	case ENOTDIR:
	    /* File not found or invalid path entry, try the next
	     * one. */
	    break;

	case ENODEV:
	case ETIMEDOUT:
	    /* Stranger error messages that, according to glibc source
	     * comments, may occur on various strange file
	     * systems. Try the next path entry instead... */
	    break;

	default:
	    free(path);
	    free(new_file);
	    return -1;
	}

	current = ++next;
    } while (next < path_end);

    free(path);
    free(new_file);

    errno = got_eaccess ? EACCES : ENOENT;
    return -1;
}

int
execv_path(const char *file, char *const argv[])
{
    if (!file || file[0] == '\0') {
	errno = ENOENT;
	return -1;
    }

    if (!strchr(file, '/'))
	return execv_try_path(file, argv);
    else
	return execv(file, argv);
}
