/// @file acumon/protocol.h

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#ifndef PROTOCOL_H
#define PROTOCOL_H

#include "acumon_defs.h"

#include "sampler/src/instrumenter_stub.h"

#include "library/lang/common_macros.h"

#include <unistd.h>

BEGIN_EXTERN_C;

/* Size of the notification stack. */
#define NOTIFICATION_STACK_SIZE 4096

/* Notification types. */
typedef enum {
    Notification_Sigsegv,
    Notification_Mmap,
    Notification_Munmap,
    Notification_Mprotect,
#if defined(__linux__)
    Notification_Mremap,
#endif
    Notification_Sigaction
} notification_type_t;

/* Arguments to notification.

   Must be identical for 32-bit and 64-bit target. */
typedef struct {
    uint64_t type;

    union {
        struct {
            uint64_t addr;
        } sigsegv;
        struct {
            uint64_t addr;
            uint64_t len;
            uint64_t prot;
            uint64_t flags;
            uint64_t fd;
            uint64_t offs;
        } mmap;
        struct {
            uint64_t addr;
            uint64_t len;
        } munmap;
	struct {
	    uint64_t addr;
	    uint64_t len;
	    uint64_t prot;
	} mprotect;
#if defined(__linux__)
        struct {
            uint64_t old_addr;
            uint64_t old_size;
            uint64_t new_addr;
            uint64_t new_size;
        } mremap;
#endif
        struct {
            uint64_t sig;
            uint64_t nact;
            uint64_t oact;
        } sigaction;
    } u;
} notification_args_t;

/* Message kinds sent on the socket to the instrumenter. */
typedef enum {
    /* The thread loading the instrumenter sends a "load hello" message to the
       monitor. The monitor responds with a "load welcome" message containing
       initialization parameters. The thread loading the instrumenter sends a
       "load ready" message when it has finished the initialization. */
    AM_Msg_Load_Hello,
    AM_Msg_Load_Welcome,
    AM_Msg_Load_Ready,

    /* When a thread has been instrumented and is ready to start running
       instrumented code it sends an "instr ready" message to the monitor. The
       monitor then restores the hijacked page and responds with an "instr go"
       message. */
    AM_Msg_Instr_Ready,
    AM_Msg_Instr_Go,

    /* The monitor sends a "deinstr" message to the instrumenter to start the
       deinstrumentation sequence. When the instrumenter receives the message
       it responds with a "deinstr ready" message containing the native state
       of the thread. The monitor in turn responds with either a "deinstr go"
       message telling the instrumenter to go ahead, or a "deinstr cancel"
       message telling the instrumenter the the deinstrumentation has been
       cancelled.

       If the deintrumentation was not cancelled the instrumenter sends a
       "deinstr done 1" message when it is done with its part of the
       deinstrumentation sequence. The monitor then hijacks a page and runs the
       deinstrumenter, which when it is done sends a "deinstr done 2" message
       to the monitor. Finally the monitor restores the hijacked page and
       continues running the thread. */
    AM_Msg_Deinstr,
    AM_Msg_Deinstr_Ready,
    AM_Msg_Deinstr_Cancel,
    AM_Msg_Deinstr_Go,
    AM_Msg_Deinstr_Done_1,
    AM_Msg_Deinstr_Done_2,

    /* When an instrumented thread is about to do an exit syscall it
       deallocates its TLS to avoid leaking it. If the monitor is not aware
       that this is going on it may try to update the message counter in the
       thread's deallocated TLS and fail, or even worse, overwrite some other
       data that has been allocated where the TLS used to be.

       To prevent this from happening the instrumented thread sends an "exit"
       message to the monitor and then waits for an "exit go" reply before it
       proceeds with the exit syscall.

       If the monitor has simultaneously decided to send a "detach" message to
       the instrumented thread, the thread will ignore that message and the
       monitor should reply with an "exit go" message when it receives the
       "exit" message.

       This mechanism is not needed for exit_group, since the TLS is not
       deallocated prior to the exit_group call. */
    AM_Msg_Exit,
    AM_Msg_Exit_Go,

    /* When an instrumented thread translates an instruction matching a
       breakpoint set by the monitor it sends a "breakpoint" message with the
       translated address to the monitor. It then waits for a "breakpoint go"
       reply before continuing. */
    AM_Msg_BP,
    AM_Msg_BP_Go,

    /* When an instrumented thread invalidates a translated instruction
       matching a breakpoint set by the monitor it sends a "breakpoint
       invalidation" message with the translated address to the monitor. It
       then waits for a "breakpoint invalidation go" reply before
       continuing. */
    AM_Msg_BP_Inv,
    AM_Msg_BP_Inv_Go,

#if defined(__sun__)
    /* When an instrumented thread is about to be suspended we send a "suspend"
       message to it. When the thread is ready to be suspended (does not hold
       any locks in the instrumenter) it responds with a "suspend ready"
       message. */
    AM_Msg_Suspend,
    AM_Msg_Suspend_Ready,
#endif

    /* The instrumenter sends a "protect page" message to the monitor and waits
       for a "protect page go" reply before it protects a page with generated
       code. */
    AM_Msg_Protect_Page,
    AM_Msg_Protect_Page_Go,

    /* The instrumenter sends an "unprotect page" message to the monitor and
       waits for an "unprotect page go" reply before it unprotects a page with
       generated code. */
    AM_Msg_Unprotect_Page,
    AM_Msg_Unprotect_Page_Go,

    /* The thread has finished processing a notification and is waiting for the
       monitor to restore its native state. */
    AM_Msg_Notification_Done,

    /* Log message, used by the instrumented thread to notify users
       about warnings etc. */
    AM_Msg_Log,

    /* The instrumenter has called am_abort().  */
    AM_Msg_Abort,

    /* Sent to the monitor when a burst has ended. */
    AM_Msg_Burst_End,

    /* Not really a message kind, used to accept any kind of message when
       filtering messages based on kind. */
    AM_Msg_Any = -1
} am_message_kind_t;

/* Format of messages sent between the monitor and the
   bootstrapper/instrumenter/debootstrapper.

   Must be identical for 32-bit and 64-bit target. */
typedef struct {
    /* Process id, thread id and message kind. Must always be valid. */
    uint32_t pid;
    uint32_t tid;
    uint32_t kind;

    /* Length of extra data that appearing after the "ordinary" message. */
    uint32_t extra_data_len;

    /* Fields specific to each message kind. */
    union {
        struct {
            uint32_t target_socket_fd;
            uint8_t pad[4];
            uint64_t instr_addr;
        } load_hello;
        struct {
            uint64_t load_entry;
            uint64_t instr_rx_start;
            uint64_t instr_rx_size;
            uint64_t instr_rw_start;
            uint64_t instr_rw_size;
            sampler_params_t sampler_params;
        } load_welcome;
        struct {
            uint64_t stack_addr;
            uint64_t msg_counter_addr;
	    uint64_t noevpend_addr;
        } instr_ready;
        struct {
            am_registers_t regs;
        } deinstr_ready;
        struct {
            uint32_t burst_generation;
        } deinstr_cancel;
        struct {
            uint64_t addr;
        } bp;
        struct {
            uint64_t addr;
        } bp_inv;
#if defined(__sun__)
        struct {
            uint64_t spinlock_addr;
        } suspend_ready;
#endif
        struct {
            uint64_t addr;
        } protect_page;
        struct {
            uint64_t addr;
        } unprotect_page;
        struct {
            uint64_t has_ret_val;
            uint64_t ret_val;
        } notification_done;
	struct {
	    /* The actual text message is sent after the fixed length
	       part of the message and is exactly extra_data_len bytes
	       long. */
	    uint8_t subsystem;
	    int8_t level;
	    uint8_t pad[6];
	} log;
	struct {
	    int64_t pcdiff;	/* caller's RIP relative to send_abort() */
	} abort;
    } u;
} am_message_t;

/* Subsystem IDs for log message */
#define AM_SUBSYSTEM_UNKNOWN 0

/* Message levels for log messages, these shuold be a subset of the
   mesasge levels defined in control.h for communication between
   acumon and the GUI. */
#define AM_LOG_DEBUG   (-1)
#define AM_LOG_VERBOSE   0
#define AM_LOG_INFO     10
#define AM_LOG_WARN     20
#define AM_LOG_ERROR    127

/* Maximum length of paths sent to the bootstrapper. */
#define AM_PATH_MAX  0x100

/* Buffer for bootstrapping or debootstrapping code.

   Must be identical for 32-bit and 64-bit target. */
typedef struct {
    char chars[AM_CODE_SIZE];
} am_code_buf_t;

/* Buffer for paths sent to the bootstrapper.

   Must be identical for 32-bit and 64-bit target. */
typedef struct {
    char chars[AM_PATH_MAX];
} am_path_buf_t;

/* System call arguments. Reserves enough room for 6 64-bit arguments.

   Must be identical for 32-bit and 64-bit target. */
typedef union {
    uint32_t args32[6];
    uint64_t args64[6];
} syscall_args_t;

END_EXTERN_C;

#endif /* PROTOCOL_H */
