/// @file acumon/safe_io.c

// Copyright (c) 2007-2008 Acumem AB.  All rights reserved.

#include <fcntl.h>
#include <sys/wait.h>
#include <errno.h>

#include "safe_io.hh"

int
safe_open(const char *path, int flags)
{
    int res;
    do {
        res = open(path, flags);
    } while (res == -1 && errno == EINTR);
    return res;
}

int
safe_close(int fd)
{
    int res;
    do {
        res = close(fd);
    } while (res == -1 && errno == EINTR);
    return res;
}

ssize_t
safe_read(int fd, void *buf, size_t nbytes)
{
    ssize_t res;
    do {
        res = read(fd, buf, nbytes);
    } while (res == -1 && errno == EINTR);
    return res;
}

ssize_t
safe_write(int fd, const void *buf, size_t nbytes)
{
    ssize_t res;
    do {
        res = write(fd, buf, nbytes);
    } while (res == -1 && errno == EINTR);
    return res;
}

off_t
safe_lseek(int fd, off_t offset, int whence)
{
    off_t res;
    do {
        res = lseek(fd, offset, whence);
    } while (res == -1 && errno == EINTR);
    return res;
}

int
safe_poll(struct pollfd *fds, nfds_t nfds, int timeout)
{
    /* TODO: This is incorrect! If we do repeated polls with the same timeout
       we will get a longer total timeout than specified. */

    int res;
    do {
        res = poll(fds, nfds, timeout);
    } while (res == -1 && errno == EINTR);
    return res;
}

pid_t
safe_waitpid(pid_t pid, int *status, int options)
{
    pid_t res;
    do {
        res = waitpid(pid, status, options);
    } while (res == -1 && errno == EINTR);
    return res;
}

ssize_t
safe_send(int s, const void *buf, size_t len, int flags)
{
    ssize_t res;
    do {
        res = send(s, buf, len, flags);
    } while (res == -1 && errno == EINTR);
    return res;
}

ssize_t
safe_sendmsg(int socket, const struct msghdr *message, int flags)
{
    ssize_t res;
    do {
        res = sendmsg(socket, message, flags);
    } while (res == -1 && errno == EINTR);
    return res;
}

ssize_t
safe_recv(int s, void *buf, size_t len, int flags)
{
    ssize_t received = 0;
    do {
        ssize_t res = recv(s, (char *)buf + received, len - received, flags);
        switch (res) {
        case -1:
            if (errno == EINTR)
                continue;
            else
                return res;
        case 0:
            return 0;
        default:
            received += res;
        }
    } while (received < len);
    return received;
}

int
safe_accept(int s, struct sockaddr *addr, socklen_t *addrlen)
{
    int res;
    do {
        res = accept(s, addr, addrlen);
    } while (res == -1 && errno == EINTR);
    return res;
}

int
safe_connect(int s, const struct sockaddr *addr, socklen_t addrlen)
{
    int res;
    do {
        res = connect(s, addr, addrlen);
    } while (res == -1 && errno == EINTR);
    return res;
}
