/// @file acumon/signal_handling.c

// Copyright (c) 2007-2008 Acumem AB.  All rights reserved.

#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <poll.h>
#include <errno.h>

#include "signal_handling.hh"
#include "safe_io.hh"
#include "am_assert.hh"
#include "vect.hh"
#include "event.hh"

#define PIPE_BUF_SIZE 0x100

typedef enum {
    HT_Ignore,
    HT_Handler,
    HT_Event
} handler_type_t;

typedef struct {
    int signo;
    handler_type_t type;
    void (*handler)(void);
    struct sigaction orig_action;
    int pipe_fds[2];
} handler_info_t;

static sigset_t sigmask_orig;
static sigset_t sigmask_full;

static VECT(handler_info_t *, handle_info) handler_infos;

static void
signal_handler(int signo)
{
    handler_info_t *handler_info;
    VECT_FOREACH(handler_infos, handler_info)
        if (handler_info->signo == signo)
            switch (handler_info->type) {
            case HT_Handler:
                handler_info->handler();
                return;
            case HT_Event: {
                char buf = 0;
                ssize_t res = safe_write(handler_info->pipe_fds[1], &buf, 1);
                AM_ASSERT_WITH_ERRNO(res != -1 || errno == EAGAIN);
                return;
            }
            default:
                AM_ILLEGAL_STATE();
            }
    AM_ILLEGAL_STATE();
}

static void
event_handler(short events, void *data)
{
    AM_ASSERT(events == POLLIN);

    handler_info_t *handler_info = (handler_info_t *)data;

    char buf[PIPE_BUF_SIZE];
    ssize_t res;
    while ((res = safe_read(handler_info->pipe_fds[0], buf, PIPE_BUF_SIZE)) == PIPE_BUF_SIZE)
        ;
    AM_ASSERT_WITH_ERRNO(res != -1 || errno == EAGAIN);

    handler_info->handler();
}

static void
add_handler(int signo, handler_type_t type, int flags, void (*handler)(void))
{
    int res;
    sigset_t sigmask;
    handler_info_t *handler_info;

    res = sigprocmask(SIG_SETMASK, &sigmask_full, &sigmask);
    AM_ASSERT_WITH_ERRNO(res == 0);

    VECT_FOREACH(handler_infos, handler_info)
        AM_ASSERT(handler_info->signo != signo);

    handler_info = (handler_info_t *)am_malloc(sizeof(handler_info_t));
    handler_info->signo = signo;
    handler_info->type = type;
    handler_info->handler = handler;

    struct sigaction sa = {};
    sa.sa_flags = flags;

    switch (handler_info->type) {
    case HT_Ignore:
        sa.sa_handler = SIG_IGN;
        break;
    case HT_Handler:
        sa.sa_handler = signal_handler;
        break;
    case HT_Event:
        sa.sa_handler = signal_handler;
        res = pipe(handler_info->pipe_fds);
        AM_ASSERT_WITH_ERRNO(res != -1);
        res = fcntl(handler_info->pipe_fds[0], F_SETFL, O_NONBLOCK);
        AM_ASSERT_WITH_ERRNO(res != -1);
        res = fcntl(handler_info->pipe_fds[1], F_SETFL, O_NONBLOCK);
        AM_ASSERT_WITH_ERRNO(res != -1);
        add_fd_event(handler_info->pipe_fds[0], POLLIN, event_handler, handler_info);
        break;
    default:
        AM_ILLEGAL_STATE();
    }

    res = sigaction(signo, &sa, &handler_info->orig_action);
    AM_ASSERT_WITH_ERRNO(res == 0);

    VECT_APPEND(handler_infos, handler_info_t *, handler_info);

    res = sigdelset(&sigmask, signo);
    AM_ASSERT_WITH_ERRNO(res == 0);
    res = sigprocmask(SIG_SETMASK, &sigmask, NULL);
    AM_ASSERT_WITH_ERRNO(res == 0);
}

static void
remove_handler(int signo, handler_type_t type, void (*handler)(void))
{
    int res;
    sigset_t sigmask;
    handler_info_t *handler_info;

    res = sigprocmask(SIG_SETMASK, &sigmask_full, &sigmask);
    AM_ASSERT_WITH_ERRNO(res == 0);

    VECT_FOREACH(handler_infos, handler_info)
        if (handler_info->signo == signo) {
            AM_ASSERT(handler_info->type == type);
            AM_ASSERT(handler_info->handler == handler);

            int found;
            VECT_REMOVE(handler_infos, handler_info, found);
            AM_ASSERT(found);

            res = sigaction(signo, &handler_info->orig_action, NULL);
            AM_ASSERT_WITH_ERRNO(res == 0);

            switch (handler_info->type) {
            case HT_Ignore:
                break;
            case HT_Handler:
                break;
            case HT_Event:
                remove_fd_event(handler_info->pipe_fds[0], POLLIN, event_handler, handler_info);
                res = safe_close(handler_info->pipe_fds[0]);
                AM_ASSERT_WITH_ERRNO(res == 0);
                res = safe_close(handler_info->pipe_fds[1]);
                AM_ASSERT_WITH_ERRNO(res == 0);
                break;
            default:
                AM_ILLEGAL_STATE();
            }

            am_free(handler_info);

            res = sigaddset(&sigmask, signo);
            AM_ASSERT_WITH_ERRNO(res == 0);
            res = sigprocmask(SIG_SETMASK, &sigmask, NULL);
            AM_ASSERT_WITH_ERRNO(res == 0);

            return;
        }

    AM_ILLEGAL_STATE();
}

void
add_signal_handler(int signo, void (*handler)(void))
{
    add_handler(signo, HT_Handler, 0, handler);
}

void
add_fatal_signal_handler(int signo, void (*handler)(void))
{
    add_handler(signo, HT_Handler, SA_RESETHAND, handler);
}

void
remove_signal_handler(int signo, void (*handler)(void))
{
    remove_handler(signo, HT_Handler, handler);
}

void
add_signal_event(int signo, void (*handler)(void))
{
    add_handler(signo, HT_Event, 0, handler);
}

void
remove_signal_event(int signo, void (*handler)(void))
{
    remove_handler(signo, HT_Event, handler);
}

void
signal_handling_init(void)
{
    int res;
    res = sigfillset(&sigmask_full);
    AM_ASSERT_WITH_ERRNO(res == 0);

    res = sigprocmask(SIG_SETMASK, &sigmask_full, &sigmask_orig);
    AM_ASSERT_WITH_ERRNO(res == 0);

    VECT_INIT(handler_infos);
}

void
signal_handling_fini(void)
{
    AM_ASSERT(VECT_SIZE(handler_infos) == 0);
    VECT_FINI(handler_infos);
}

void
signal_prepare_for_launch(void)
{
    int res;

    while (VECT_SIZE(handler_infos) > 0) {
        handler_info_t *info = VECT_GET(handler_infos, 0);
        remove_handler(info->signo, info->type, info->handler);
    }

    res = sigprocmask(SIG_SETMASK, &sigmask_orig, NULL);
    AM_ASSERT_WITH_ERRNO(res == 0);
}
