/// @file acumon/thread_predicates.c

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#include "thread_predicates.hh"
#include "acumon.hh"
#include "os_dependent.hh"
#include "am_assert.hh"

#include "sampler/src/tls_size.h"

int
thread_native(const am_thread_t *thread, int strict)
{
    switch (thread->state) {
    case AM_TS_Native:
        return 1;
    case AM_TS_Instr_Wait_All_Stopped:
    case AM_TS_Notification_Wait:
    case AM_TS_Notification_Wait_Done:
    case AM_TS_Notification_Wait_Done_Stopped:
        return !strict;
    case AM_TS_Load_Wait_All_Stopped:
    case AM_TS_Load_Wait_Hello:
    case AM_TS_Load_Wait_Hello_Stopped:
    case AM_TS_Load_Wait_Ready:
    case AM_TS_Load_Wait_Ready_Stopped:
    case AM_TS_Instr_Wait_Ready:
    case AM_TS_Instr_Wait_Ready_Stopped:
    case AM_TS_Instrumented:
    case AM_TS_Deinstr_Wait_Ready:
    case AM_TS_Deinstr_Wait_Done_1:
    case AM_TS_Deinstr_Wait_Done_1_Run:
    case AM_TS_Deinstr_Wait_Done_1_All_Stopped:
    case AM_TS_Deinstr_Wait_Done_2:
    case AM_TS_Deinstr_Wait_Done_2_Stopped:
    case AM_TS_Unload_Wait_All_Stopped:
    case AM_TS_Unload_Wait_Close:
    case AM_TS_Unload_Wait_Close_Stopped:
    case AM_TS_Instrumented_Exiting:
    case AM_TS_Instrumented_Wait_Ready:
        return 0;
    default:
        AM_ILLEGAL_STATE();
    }
}

int
all_threads_native(const am_process_t *process, int strict)
{
    am_thread_t *thread;
    VECT_FOREACH(process->threads, thread)
        if (!thread_native(thread, strict))
            return 0;
    return 1;
}

int
all_other_threads_native(const am_thread_t *thread, int strict)
{
    am_process_t *process = thread->process;

    am_thread_t *t;
    VECT_FOREACH(process->threads, t)
        if (t != thread && !thread_native(t, strict))
            return 0;
    return 1;
}

int
thread_instrumented(const am_thread_t *thread)
{
    switch (thread->state) {
    case AM_TS_Instrumented:
    case AM_TS_Instrumented_Exiting:
    case AM_TS_Instrumented_Wait_Ready:
    case AM_TS_Deinstr_Wait_Ready:
    case AM_TS_Deinstr_Wait_Done_1:
        return 1;
    case AM_TS_Native:
    case AM_TS_Load_Wait_All_Stopped:
    case AM_TS_Load_Wait_Hello:
    case AM_TS_Load_Wait_Hello_Stopped:
    case AM_TS_Load_Wait_Ready:
    case AM_TS_Load_Wait_Ready_Stopped:
    case AM_TS_Instr_Wait_All_Stopped:
    case AM_TS_Instr_Wait_Ready:
    case AM_TS_Instr_Wait_Ready_Stopped:
    case AM_TS_Deinstr_Wait_Done_1_Run:
    case AM_TS_Deinstr_Wait_Done_1_All_Stopped:
    case AM_TS_Deinstr_Wait_Done_2:
    case AM_TS_Deinstr_Wait_Done_2_Stopped:
    case AM_TS_Unload_Wait_All_Stopped:
    case AM_TS_Unload_Wait_Close:
    case AM_TS_Unload_Wait_Close_Stopped:
    case AM_TS_Notification_Wait:
    case AM_TS_Notification_Wait_Done:
    case AM_TS_Notification_Wait_Done_Stopped:
        return 0;
    default:
        AM_ILLEGAL_STATE();
    }
}

int
any_thread_instrumented(const am_process_t *process)
{
    am_thread_t *t;
    VECT_FOREACH(process->threads, t)
        if (thread_instrumented(t))
            return 1;
    return 0;
}

int
all_threads_stopped(const am_process_t *process)
{
    am_thread_t *thread;
    VECT_FOREACH(process->threads, thread)
        if (!thread_stopped(thread, 0))
            return 0;

    return 1;
}

int
thread_exiting_syscall(const am_thread_t *thread)
{
    AM_ASSERT(!thread_stopped(thread, 1));

    return
        thread->exiting_syscall
#if defined(__linux__)
        && (!(instrumenter_loaded(thread->process) &&
              thread_native(thread, 0)) ||
            thread->syscall_state != 1)
#endif
        ;
}

int
thread_in_critical_region(const am_thread_t *thread)
{
    AM_ASSERT(thread_stopped(thread, 1));

    if (!thread_instrumented(thread)) {
        /* If the thread is uninstrumented it is in a critical region if it is
           processing a notification. */

        /* TODO */
        return 0;
    }
    else {
        /* If the thread is instrumented it is in a critical region if the
           stack pointer points to the instrumenter stack. */
        unsigned long rsp = get_rsp(thread);

        return rsp > thread->stack_addr && rsp <= thread->stack_addr + TLS_STACK_SIZE;
    }
}

int
any_thread_in_notification(const am_process_t *process)
{
    am_thread_t *thread;
    VECT_FOREACH(process->threads, thread)
        switch (thread->state) {
        case AM_TS_Notification_Wait_Done:
        case AM_TS_Notification_Wait_Done_Stopped:
            return 1;
        default:
            break;
        }
    return 0;
}

int
any_thread_in_notification_wait(const am_process_t *process)
{
    am_thread_t *thread;
    VECT_FOREACH(process->threads, thread)
        if (thread->state == AM_TS_Notification_Wait)
            return 1;
    return 0;
}
