/// @file acumon/thread_predicates.h

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

/* Some of these functions are implemented in thread_predicates.c, those that
   are operating system specific are implemented in the operating system
   specific code. */

#ifndef THREAD_PREDICATES_HH
#define THREAD_PREDICATES_HH

#include "acumon.hh"

/* Returns true if the specified thread is native.

   If strict is true only the AM_TS_Native state is regarded as native. If
   strict is false other states where the thread is not instrumented, waiting
   to be instrumented and handling notifications, are also regarded as
   native. */
int thread_native(const am_thread_t *thread, int strict);

/* Returns true if all threads in the specified process are native.

   If strict is true only the AM_TS_Native state is regarded as native. If
   strict is false other states where the thread is not instrumented, waiting
   to be instrumented and handling notifications, are also regarded as
   native. */
int all_threads_native(const am_process_t *process, int strict);

/* Returns true if all threads in the process, possibly except the specified
   thread, are native.

   If strict is true only the AM_TS_Native state is regarded as native. If
   strict is false other states where the thread is not instrumented, waiting
   to be instrumented and handling notifications, are also regarded as
   native. */
int all_other_threads_native(const am_thread_t *thread, int strict);

/* Returns true if the thread is instrumented. A thread is counted as
   instrumented when it is in a state where it may take instrumenter locks. */
int thread_instrumented(const am_thread_t *thread);

/* Returns true if any thread in the specified process is instrumented. A
   thread is counted as instrumented when it is in a state where it may take
   instrumenter locks. */
int any_thread_instrumented(const am_process_t *process);

/* Return true if the thread is stopped. This is not necessarily the same thing
   as the thread's debug_stopped flag being true, it is enough that the thread
   is in a state where it will not be able to run any user space code until we
   resume it. */
int thread_stopped(const am_thread_t *thread, int strict);

/* Return true if all threads in the process are stopped. */
int all_threads_stopped(const am_process_t *process);

/* Returns true if the thread is waiting to exit a system call, false if
   not. Must only be called when the thread is running. */
int thread_exiting_syscall(const am_thread_t *thread);

/* Determine if the thread is in a critical region.

   This is the case if an instrumented thread is in the critical region of the
   instrumenter, or if an uninstrumented thread is handling a notification. */
int thread_in_critical_region(const am_thread_t *thread);

/* Return 1 if the thread is in a system call, 0 if not. Must only be called
   when the thread is stopped. */
int thread_in_syscall(const am_thread_t *thread);

/* Returns true if the thread is suspended (if the operating system supports
   thread suspension). */
int thread_suspended(const am_thread_t *thread);

/* Returns true if the thread is in a state where it can safely be detached
   from a debugging interface perspective, i.e., there are no signals synthetic
   signals pending, no breakpoints left etc. */
int thread_detachable(const am_thread_t *thread);

/* Returns true if any thread in the specified process is processing a
   notification. */
int any_thread_in_notification(const am_process_t *process);

/* Returns true if any thread in the specified process is waiting to process a
   notification. */
int any_thread_in_notification_wait(const am_process_t *process);

#endif /* THREAD_PREDICATES_HH */
