/// @file analysis/advice/advice.cc

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#include "advice.hh"

#include "analysis/issues/utilization_issue.hh"
#include "analysis/issues/random_issue.hh"
#include "analysis/issues/hot_spot.hh"
#include "analysis/issues/loop_fusion.hh"
#include "analysis/issues/block.hh"
#include "analysis/issues/prefetch.hh"
#include "analysis/issues/loop_order.hh"
#include "analysis/issues/comm_spot.hh"
#include "analysis/issues/fs.hh"
#include "analysis/issues/wnt_possible.hh"
#include "analysis/issues/prefetchnta.hh"
#include "analysis/filter/filter_module.hh"

#include <iostream>

using namespace acumem::library;
using namespace std;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

// ----- configuration_t ------------------------------------------------------

#define PAGE_SIZE 4096
#define LIMIT_FACTOR 10
#define HW_PF_POS_STREAMS 16      /* Hardware prefetcher supports 16 streams with positive stride. */
#define HW_PF_NEG_STREAMS 4       /* Hardware prefetcher supports 4 streams with negative stride. */
#define HW_PF_REACH 2             /* Hardware prefetcher fetches two lines ahead. */
#define COMM_COST 7               /* Cost of a coherence fetch in replacement fetches. */
#define COMM_UTIL_THRESHOLD 0.8   /* Threshold for communication utilization issues. */
#define LOOP_LENGTH_FACTOR 3.0    /* Loop order loop length factor. */
#define RANDOM_THRESHOLD 0.2      /* Random threshold. */
#define SWPF_COST 0.5             /* Cost of a software prefetch in cycles. */
#define SWPF_MAX_L1_SIZE 0x10000  /* Caches larger than this are considered to be L2 caches. */
#define SWPF_L1_COST 10.0         /* Cost of an L1 miss. */
#define SWPF_L2_COST 100.0        /* Cost of an L2 miss. */
#define L1_CACHE_SIZE 0x8000      /* L1 size for prefetchnta calculation. */

configuration_t::configuration_t(const string &sample_file,
                                 const fnr_stack_depth_t stack_depth,
                                 const fnr_cache_size_t cache_size,
                                 const fnr_line_size_t line_size,
				 const thread_mapping_cptr_t thread_mapping,
				 const cpu_model_t cpu_model,
				 const repl_fr_engine_cptr_t engine,
				 const windowing_engine_cptr_t &windowing,
                                 const sweep_t sweep,
				 const bool sw_prefetch,
				 const bool blocking_expert,
				 const bool mt,
				 const bool last_cache_level,
                                 const bool low_mem,
				 const bool use_old_hwpf_algorithm,
                                 const bool hwpf_enabled,
				 const double minimum_fuwb_fraction,
                                 const double minimum_fuwb_ratio,
                                 const fnr_cache_size_t l1_cache_size)
    : sample_file(sample_file),
      stack_depth(stack_depth),
      cache_size(cache_size),
      line_size(line_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      engine(engine),
      windowing(windowing),
      sweep(sweep),
      page_size(PAGE_SIZE),
      pos_streams(HW_PF_POS_STREAMS),
      neg_streams(HW_PF_NEG_STREAMS),
      prefetch_reach(HW_PF_REACH),
      shared_prefetch_streams(true),
      sw_prefetch(sw_prefetch),
      blocking_expert(blocking_expert),
      mt(mt),
      last_cache_level(last_cache_level),
      low_mem(low_mem),
      use_old_hwpf_algorithm(use_old_hwpf_algorithm),
      hwpf_enabled(hwpf_enabled),
      minimum_fuwb_fraction(minimum_fuwb_fraction),
      minimum_fuwb_ratio(minimum_fuwb_ratio),
      comm_cost(COMM_COST),
      l1_cache_size(l1_cache_size)
{
}

// ----- advisor_t ------------------------------------------------------------

void
advisor_t::populate_all_pcs(const configuration_t &c,
			    const feature_context_ptr_t &context,
			    location_set_t &location_set) const
{
    stack_pc_map_result_cptr_t spm_res(
	stack_pc_map_t(context, c.sample_file, c.stack_depth).get());

    BOOST_FOREACH(const trunc_stack_pc_id_map_t::value_type &v,
		  spm_res->trunc_stack_pc_id_map)
	location_set.insert(v.first);
}

static void
do_sweep_point(const configuration_t &c,
	       const feature_context_ptr_t &context,
	       const location_set_t &all_pcs,
	       drill_down_query_list_t &queries,
    	       const sweep_point_t &sp,
	       int cache_size_index)
{
    stats_slicer_result_cptr_t stats_slicer(
	stats_slicer_t(context, c.sample_file, c.stack_depth, sp.line_size,
		       c.thread_mapping, c.cpu_model, sp.cache_size,
		       c.engine, c.windowing,
		       c.page_size, c.pos_streams,
		       c.neg_streams, c.prefetch_reach,
		       c.use_old_hwpf_algorithm, c.hwpf_enabled).get());


    BOOST_FOREACH(drill_down_query_t &q, queries) {
	drill_down_result_t &result(*q.result);

	result.cache_range[sp.cache_size] = cache_size_index;
	result.stats[cache_size_index] =
	    stats_slicer->get(q.locations ? *q.locations : all_pcs,
			      false);
    }
}

void
advisor_t::prepare_performance_data(const configuration_t &c,
				    const feature_context_ptr_t &context,
				    drill_down_query_list_t &queries) const
{
    const unsigned int sweep_size = c.sweep.size();

    // Setup result structures
    location_set_t all_pcs;
    bool all_pcs_populated = false;

    /// Populate the all PCs set if needed
    BOOST_FOREACH(const drill_down_query_t &query, queries) {
	query.result.reset(new drill_down_result_t(sweep_size));

	if (!query.locations && !all_pcs_populated) {
	    populate_all_pcs(c, context, all_pcs);
	    all_pcs_populated = true;
	}
    }

    // Number of completed minor steps
    int minor_step = 0;

    // Setup the number of minor steps, on for each data point and sweep.
    context->feedback->set_minor_steps(sweep_size);

    // Sweep over the cache configurations.
    int cache_size_index = 0;
    BOOST_FOREACH(sweep_point_t sp, c.sweep) {
	if (c.low_mem) {
	    // Create a temporary workspace, we only cache sweep points
	    // locally to prevent excessive memory usage.
	    scoped_temporary_workspace_t stw(context->workspace_stack);

	    do_sweep_point(c, context, all_pcs, queries,
			   sp, cache_size_index);
	} else
	    do_sweep_point(c, context, all_pcs, queries,
			   sp, cache_size_index);

	cache_size_index++;

	// Update number of completed minor steps
	context->feedback->update(++minor_step);

	// Throw a cancellation exception if the user cancels.
	context->feedback->test_cancel();
    }
}

void
advisor_t::report_analysis_info(const configuration_t &c,
				const feature_context_ptr_t &context,
				const advice_result_ptr_t &result) const
{
    analysis_info_ptr_t ai(new analysis_info_t());

    parser_result_cptr_t line_samples(
	filter_module_t(context, c.sample_file, c.line_size).get());

    thread_mapping_result_cptr_t tm_res(
	c.thread_mapping->get());

    ai->process_info = line_samples->process_info;
    ai->sample_count = line_samples->repl_sample_count();
    ai->num_caches = tm_res->num_caches;

    result->analysis_info = ai;
}

void
advisor_t::report_global_statistics(const configuration_t &c,
				    const feature_context_ptr_t &context,
				    const advice_result_ptr_t &result) const
{
    // Setup result structures
    location_set_t all_pcs;
    populate_all_pcs(c, context, all_pcs);

    stats_slicer_result_cptr_t stats_slicer(
	stats_slicer_t(context, c.sample_file, c.stack_depth, c.line_size,
		       c.thread_mapping, c.cpu_model, c.cache_size,
		       c.engine, c.windowing,
		       c.page_size, c.pos_streams, c.neg_streams,
		       c.prefetch_reach, c.use_old_hwpf_algorithm, c.hwpf_enabled).get());

    result->global_statistics =
	stats_slicer->get(all_pcs, false);

    // Throw a cancellation exception if the user cancels.
    context->feedback->test_cancel();
}

void
advisor_t::report_program_structure(const configuration_t &c,
				    const feature_context_ptr_t &context,
				    const advice_result_ptr_t &result) const
{
    program_structure_ptr_t program_structure(new program_structure_t);

    access_lookup_result_cptr_t al_res =
        access_lookup_t(context, c.sample_file, c.stack_depth).get();

    // Get loops.
    context->feedback->log(feedback_base_t::LOG_INFO, "Identify loops");
    loop_identification_result_cptr_t loops = loop_identification_t(
        context, c.sample_file, c.stack_depth).get();

    // Get dependence groups.
    context->feedback->log(feedback_base_t::LOG_INFO, "Instruction groups");
    loop_depgroups_result_cptr_t loop_depgroups(
	loop_depgroups_t(context, c.sample_file, c.stack_depth, c.line_size,
			 c.page_size, c.thread_mapping, c.cpu_model,
			 Mapped_Line, LIMIT_FACTOR).get());

    // Get sliced statistics.
    context->feedback->log(feedback_base_t::LOG_INFO, "Slice statistics");
    stats_slicer_result_cptr_t stats_slicer(
	stats_slicer_t(context, c.sample_file, c.stack_depth, c.line_size,
		       c.thread_mapping, c.cpu_model, c.cache_size,
		       c.engine, c.windowing, c.page_size,
		       c.pos_streams, c.neg_streams, c.prefetch_reach,
		       c.use_old_hwpf_algorithm, c.hwpf_enabled).get());

    // Collect loops.
    context->feedback->log(feedback_base_t::LOG_INFO, "Loop statistics");
    BOOST_FOREACH(const loop_ptr_t &loop, loops->loops) {
	context->feedback->test_cancel();
        program_structure->loops.push_back(
            prog_loop_t(loop,
                        stats_slicer->get(loop->get_locations(false), false),
                        loop_depgroups->loop_depgroups[loop->id]));
    }

    // Collect dependence groups.
    context->feedback->log(feedback_base_t::LOG_INFO, "Instruction group statistics");
    BOOST_FOREACH(location_set_t location_set, loop_depgroups->depgroups) {
	context->feedback->test_cancel();
        program_structure->depgroups.push_back(
            prog_depgroup_t(location_set,
			    stats_slicer->get(location_set, false)));
    }

    // Collect instructions.
    stack_pc_map_result_cptr_t stack_pc_map =
        stack_pc_map_t(context, c.sample_file, c.stack_depth).get();
    context->feedback->log(feedback_base_t::LOG_INFO, "Location statistics");
    BOOST_FOREACH(const trunc_stack_pc_id_map_t::value_type &v,
                  stack_pc_map->trunc_stack_pc_id_map) {
	context->feedback->test_cancel();
        location_t location = v.first;
        assert(program_structure->locations.size() == location);
        sliced_stats_cptr_t stats = stats_slicer->get(location, true);
        loc_access_info_map_t::const_iterator lai =
            al_res->loc_access_info.find(location);
        fnr_access_classes_t access_classes =
            lai == al_res->loc_access_info.end() ?
            0 :
            lai->second.access_classes;
        program_structure->locations.push_back(
            prog_location_t(stats, access_classes));
    }

    result->program_structure = program_structure;
}

void
advisor_t::report_loop_analysis(const configuration_t &c,
				const feature_context_ptr_t &context,
				const advice_result_ptr_t &result) const
{
    parser_result_cptr_t parser_result =
        parser_t(context, c.sample_file).get();

    sample_mapping_result_cptr_t sample_mapping =
        sample_mapping_t(context,
                         c.sample_file,
                         c.stack_depth,
                         c.line_size,
                         c.page_size,
                         c.thread_mapping,
                         c.cpu_model).get();

    location_set_t all_pcs;
    populate_all_pcs(c, context, all_pcs);

    stats_slicer_result_cptr_t stats_slicer(
	stats_slicer_t(context, c.sample_file, c.stack_depth, c.line_size,
		       c.thread_mapping, c.cpu_model, c.cache_size,
		       c.engine, c.windowing,
		       c.page_size, c.pos_streams, c.neg_streams,
		       c.prefetch_reach, c.use_old_hwpf_algorithm, c.hwpf_enabled).get());

    issues::issue_relevant_cptr_t issue_relevant(
        new issues::issue_relevant_t(
            stats_slicer->get(all_pcs, false)->stats,
            c.minimum_fuwb_fraction,
            c.minimum_fuwb_ratio));

    // set up loop extractor
    issues::loop_driver_t ld(context);
    issues::loop_driver_conf_t ld_c(c.sample_file,
				    c.stack_depth,
				    c.line_size,
				    c.thread_mapping,
				    c.cpu_model,
				    c.cache_size,
				    c.engine,
				    c.windowing,
				    c.page_size,
				    c.pos_streams,
				    c.neg_streams,
				    c.prefetch_reach,
				    c.use_old_hwpf_algorithm,
                                    c.hwpf_enabled,
                                    c.comm_cost,
                                    COMM_UTIL_THRESHOLD,
                                    LIMIT_FACTOR,
                                    LOOP_LENGTH_FACTOR,
                                    RANDOM_THRESHOLD,
				    SWPF_COST,
				    c.cache_size > SWPF_MAX_L1_SIZE ? SWPF_L2_COST : SWPF_L1_COST,
                                    c.blocking_expert,
                                    issue_relevant,
                                    L1_CACHE_SIZE);
    ld.set_configuration(ld_c);

    // Set up loop order engine
    issues::loop_order_t lo(context);
    lo.set_configuration(ld_c);
    ld.add(&lo);

    // Report hot/cold issues.
    issues::utilization_t ui(context);
    ui.set_configuration(ld_c);
    ld.add(&ui);

    // Report hot/cold issues.
    issues::random_t rand(context);
    rand.set_configuration(ld_c);
    ld.add(&rand);

    // Report hot-spot issues.
    issues::hot_spot_t hs(context);
    hs.set_configuration(ld_c);
    ld.add(&hs);

    // Report temporal/spatial blocking opportunities
    issues::block_t block(context);
    block.set_configuration(ld_c);
    ld.add(&block);


    // Report loop fusion issues.
    issues::loop_fusion_t lf(context);
    lf.set_configuration(ld_c);
    ld.add(&lf);

    // Report prefetch issues.
    issues::prefetch_t pf(context);
    if (c.sw_prefetch) {
        pf.set_configuration(ld_c);
        ld.add(&pf);
    }

    // Wnt possible
    issues::wnt_possible_t wnt_possible(context);
    if (parser_result->mt_sampled() && c.last_cache_level) {
	wnt_possible.set_configuration(ld_c);
	ld.add(&wnt_possible);
    }

    // Report prefetchnta issues.
    issues::prefetchnta_t prefetchnta(context);
    if (c.last_cache_level) {
	prefetchnta.set_configuration(ld_c);
	ld.add(&prefetchnta);
    }

    // Communication hot spots
    issues::comm_spot_t comm(context);

    // False sharing
    issues::fs_t fs(context);

    // Only analyse MT related issues if MT analysis has been enabled
    if (c.mt && parser_result->mt_sampled()) {
	comm.set_configuration(ld_c);
	ld.add(&comm);

	fs.set_configuration(ld_c);
	ld.add(&fs);
    }

    // Process all loops and dependecy groups, and evaluate
    // all registered engines for those.
    ld.process();

    // Get results from the engines

    result->hot_spot_issues = hs.get()->hot_spot_issues;
    result->loop_order_issues = lo.get()->loop_order_issues;
    result->utilization_issues = ui.get()->issues;
    result->random_access_issues = rand.get()->issues;
    result->block_issues = block.get()->issues;
    result->loop_fusion_issues = lf.get()->loop_fusion_vector;

    if (c.mt && parser_result->mt_sampled()) {
	issues::comm_spot_result_cptr_t cs_res(comm.get());
	result->comm_spot_issues = cs_res->comm_spots;
	result->comm_utilization_issues = cs_res->comm_utilization;
	result->fs_issues = fs.get()->issues;
	if (c.last_cache_level)
	    result->wnt_possible_issues = wnt_possible.get()->issues;
    }
    if (c.last_cache_level)
        result->prefetchnta_issues = prefetchnta.get()->issues;

    if (c.sw_prefetch)
	result->prefetch_issues = pf.get()->issues;
}

advice_result_cptr_t
advisor_t::process(const configuration_t &c,
		   const feature_context_ptr_t &context,
		   int advisor_mask) const
{
    advice_result_ptr_t result(new advice_result_t());
    // Depending on the advisor mask, dig out different kinds of information
    // and call various callbacks in the advice_consumer.

    // return the following:
    // 1. Global performance information
    //    repl_mr, cold_mr, repl_mr_rd, repl_mr_wr, repl_cold_rd, repl_cold_wr
    //    dito in estimated misses
    //    spatial_util
    // 2. spatial hot-spots based on pc2 mr
    // 3. List of loop contents
    //    loop-enumeration (PC)
    //    interdependency groups within the PC
    // 4. Hot/cold issues.
    // 5. Loop fusion issues.
    // 6. Loop order issues

    feedback_ptr_t feedback = context->feedback;

    int analysis_info = 0;
    int program_structure = 0;
    int loop_analysis = 0;

    // Allocate major step IDs for all major analysis steps.  Report
    // the analysis info step as reading sample file, since thats what
    // actually takes time to complete.

    analysis_info = feedback->add_major_step("Global analysis");

    if (advisor_mask & Advisor_program_structure)
	program_structure = feedback->add_major_step("Examining program structure");

    if (advisor_mask & Advisor_issues)
	loop_analysis = feedback->add_major_step("Advice analysis");

    try {
	// Tell the feedback mechanism that the analysis has started.
	feedback->update_status(feedback_base_t::STATUS_RUNNING);

	feedback->select_major_step(analysis_info);
	context->feedback->log(feedback_base_t::LOG_INFO, "Call stack mapping");
	stack_pc_map_result_cptr_t stack_pc_map =
            stack_pc_map_t(context, c.sample_file, c.stack_depth).get();

	// Report process information
	context->feedback->log(feedback_base_t::LOG_INFO, "Thread mapping");

	report_analysis_info(c, context, result);
	context->feedback->log(feedback_base_t::LOG_INFO, "Global statistics");
	report_global_statistics(c, context, result);

	if (advisor_mask & Advisor_program_structure) {
	    // Report program structure, e.g. loops
	    feedback->select_major_step(program_structure);
	    report_program_structure(c, context, result);
	}

	if (advisor_mask & Advisor_issues) {
	    // Report loop/dependency granularity issues
	    feedback->select_major_step(loop_analysis);
	    report_loop_analysis(c, context, result);
	}

	// Signal that analysis has finished.
	feedback->update_status(feedback_base_t::STATUS_COMPLETED);
	return result;
    } catch (cancellation_exception_t) {
	feedback->update_status(feedback_base_t::STATUS_CANCELED);
	throw;
    } catch (...) {
	feedback->update_status(feedback_base_t::STATUS_FAILED);
	throw;
    }

    return result;
}

void
advisor_t::drill_down(const configuration_t &c,
		      const feature_context_ptr_t &context,
		      drill_down_query_t &query) const
{
    drill_down_query_list_t ql;
    ql.push_back(query);
    drill_down(c, context, ql);
}

void
advisor_t::drill_down(const configuration_t &c,
		      const feature_context_ptr_t &context,
		      drill_down_query_list_t &queries) const
{
    feedback_ptr_t feedback = context->feedback;

    int step_performance = feedback->add_major_step("Issue statistics");

    feedback->select_major_step(step_performance);

    prepare_performance_data(c, context, queries);
}

END_NAMESPACE;
END_NAMESPACE;
