/// @file analysis/advice/advice.hh

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#ifndef ADVICE_HH
#define ADVICE_HH

#include "advice_result.hh"
#include "analysis/misc/sweep.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace acumem::library;

class workspace_t;

/// Struct configuration_t is a collection of parameters to be consumed
/// by the individual steps in an analysis chain.
struct configuration_t {
    friend class advisor_t;

public:
    configuration_t(const string &sample_file,
                    const fnr_stack_depth_t stack_depth,
                    const fnr_cache_size_t cache_size,
                    const fnr_line_size_t line_size,
		    const thread_mapping_cptr_t thread_mapping,
		    const cpu_model_t cpu_model,
                    const repl_fr_engine_cptr_t engine,
		    const windowing_engine_cptr_t &windowing,
                    const sweep_t sweep,
		    const bool sw_prefetch,
		    const bool blocking_expert,
		    const bool mt,
		    const bool last_cache_level,
                    const bool low_mem,
		    const bool use_old_hwpf_algorithm,
                    const bool hwpf_enabled,
		    const double minimum_fuwb_fraction,
                    const double minimum_fuwb_ratio,
                    const fnr_cache_size_t l1_cache_size);

    const string sample_file;
    const fnr_stack_depth_t stack_depth;
    const fnr_cache_size_t cache_size;
    const fnr_line_size_t line_size;

    const thread_mapping_cptr_t thread_mapping;
    const cpu_model_t cpu_model;

    // These are used by the mr_sweep_cache_size_t module
    const repl_fr_engine_cptr_t engine;
    const windowing_engine_cptr_t windowing;

    // These are used by the mr_sweep and mr_by_pc modules. This defines
    // a sequence of cache-size/line-size tuples by varying either the
    // cache-size or the line-size (but not both).

    /// Defines the interesting cache size range.
    /// advice_consumer_t::report_performance will report data for
    /// each point given by this parameter.
    const sweep_t sweep;

    // --- hw prefetcher

    /// Page size that the hardware prefetcher operates on.
    const fnr_line_size_t page_size;

    /// The number of prefetch streams with positive and negative strides,
    /// respectively.
    const int pos_streams;
    const int neg_streams;

    /// The number of sequential cache lines that are prefetched when a
    /// prefetch stream is triggered.
    const fnr_line_count_t prefetch_reach;

    /// Whether or not to divide the prefetch streams among the
    /// instances when calculating the bandwidth need.  For instance,
    /// Intel Core has private prefetchers for the L1 cache but a
    /// shared for the shared L2 cache.
    const bool shared_prefetch_streams;

    /// Whether to perform sw prefetch analysis
    const bool sw_prefetch;

    /// Whether to activate the blocking and fusion expert
    const bool blocking_expert;

    /// Activate mt issues?
    const bool mt;

    /// Activate issues that only applies to the last caches level?
    const bool last_cache_level;

    /// Optimize for low memory usage
    const bool low_mem;

    /// Whether to use the old hwpf algorithm
    const bool use_old_hwpf_algorithm;

    bool hwpf_enabled;

    /// Minimum fraction of all fetches, upgrades or write-backs that must be
    /// related to an issue in order for it to be reported.
    const double minimum_fuwb_fraction;

    /// Minimum combined ratio of fetches, upgrades and write-backs that must
    /// be related to an issue in order for it to be reported.
    const double minimum_fuwb_ratio;

    /// Communication cost in fetch equivalents
    const double comm_cost;

    /// L1 cache size. Currently only used by the prefetchnta analysis, which
    /// needs to determine which samples cause fetches in the L1 cache, even
    /// though the rest of the analysis is done on the last level cache.
    const fnr_cache_size_t l1_cache_size;
};

typedef shared_ptr<configuration_t> configuration_ptr_t;
typedef shared_ptr<const configuration_t> configuration_cptr_t;

enum advisor_mask_t {
    Advisor_program_structure	= 1,
    Advisor_issues		= 2,

    Advisor_all			= 0xffffffff
};

// ----------------------------------------------------------------------

struct drill_down_query_t {
    drill_down_query_t(drill_down_result_ptr_t &result,
		       const location_set_t *locations)
	: result(result), locations(locations) {}

    /// Reference to the result pointer, the contents of the pointer
    /// will be replaced upon execution of the query.
    drill_down_result_ptr_t &result;

    /// location list to consider, or NULL to consider all locations
    const location_set_t *locations;
};

typedef list<drill_down_query_t> drill_down_query_list_t;

// ----------------------------------------------------------------------

/// This is the advisor feature.
///
/// Unlike other exposed high level classes, this one doesn't use the
/// feature_t framework, since it doesn't need to cache any information.
class advisor_t {
public:
    /// Extract all pieces of advice from a sample file. It will based
    /// on the configuration parameter call the the virtual functions
    /// of the advice_consumer_t parameter for each advice it gives.
    ///
    /// @param c configuration including target cache properties
    /// @param context the analysis context
    /// @param advice_consumer the class that gets called back with advice
    /// @param advisor_mask mask selecting the type of information to report
    advice_result_cptr_t process(const configuration_t &c,
				 const feature_context_ptr_t &context,
				 int advisor_mask /*= Advisor_all */) const;

    /// Dig into a particular set of pc:s. It will report low level
    /// performance reports (using
    /// advice_consumer_t::report_performance) for this set, including
    /// sliced statistics for the set and for each instruction.
    ///
    /// @param c configuration including target cache properties
    /// @param context the analysis context
    /// @param query the drill down query to resolve
    void drill_down(const configuration_t &c,
		    const feature_context_ptr_t &context,
		    drill_down_query_t &query) const;

    void drill_down(const configuration_t &c,
		    const feature_context_ptr_t &context,
		    drill_down_query_list_t &queries) const;

private:
    void populate_all_pcs(const configuration_t &c,
			  const feature_context_ptr_t &context,
			  location_set_t &location_set) const;

    void prepare_performance_data(const configuration_t &c,
				  const feature_context_ptr_t &context,
				  drill_down_query_list_t &queries) const;

    void report_analysis_info(const configuration_t &c,
			      const feature_context_ptr_t &context,
			      const advice_result_ptr_t &result) const;

    void report_global_statistics(const configuration_t &c,
				  const feature_context_ptr_t &context,
				  const advice_result_ptr_t &result) const;

    void report_program_structure(const configuration_t &c,
				  const feature_context_ptr_t &context,
				  const advice_result_ptr_t &result) const;

    void report_loop_analysis(const configuration_t &c,
			      const feature_context_ptr_t &context,
			      const advice_result_ptr_t &result) const;
};

typedef shared_ptr<advisor_t> advisor_ptr_t;
typedef shared_ptr<const advisor_t> advisor_cptr_t;

END_NAMESPACE;
END_NAMESPACE;

#endif
