/// @file analysis/advice/advice_result.hh

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#include "location.hh"

#include "analysis/issues/issue.hh"
#include "analysis/misc/process_info.hh"

#include "library/lang/common_macros.h"

#ifndef ADVICE_RESULT_HH
#define ADVICE_RESULT_HH

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;

using boost::shared_ptr;

/// General information about the analysis, number of samples,
/// program, arguments, architecture, etc.
struct analysis_info_t {
    analysis_info_t()
	: sample_count(0)
	{}

    /// Process information as reported by parser_t. Target program and
    /// arguments etc.
    process_info_ptr_t process_info;

    /// Number of samples the analysis was based on.
    fnr_sample_count_t sample_count;

    /// Number of caches in the thread mapping
    tm_cache_id_t num_caches;
};

typedef shared_ptr<analysis_info_t> analysis_info_ptr_t;

struct advice_result_t {
    /// Report general information about the process, arguments etc.
    analysis_info_ptr_t analysis_info;
    /// Report relevant metrics for the program as a whole or for a subset.
    sliced_stats_cptr_t global_statistics;
    /// Describe the loops and dependency groups found.
    program_structure_cptr_t program_structure;

    // --- issues -----------------------------

    /// Prefetch usefulness advice.
    issues::prefetch_issue_vector_t prefetch_issues;
    /// Advice of loop fusion opportunities.
    issues::loop_fusion_issue_vector_t loop_fusion_issues;
    /// Advice of poor fetch/write-back utilization. This could
    /// indicate a case of unused fields, but also sparse data
    /// representation or general sparse usage.
    issues::utilization_issue_vector_t utilization_issues;
    /// Advice of hot-spots.
    issues::hot_spot_issue_vector_t hot_spot_issues;
    /// Advice to correct incorrect loop order.
    issues::loop_order_issue_vector_t loop_order_issues;
    /// Advice to perform temporal/spatial loop blocking.
    issues::block_issue_vector_t block_issues;
    /// Advice about random access patterns
    issues::random_access_issue_vector_t random_access_issues;
    /// Communication hot spots.
    issues::comm_spot_issue_vector_t comm_spot_issues;
    /// Advice on poor communication utilization.
    issues::comm_utilization_issue_vector_t comm_utilization_issues;
    /// False sharing issues.
    issues::fs_issue_vector_t fs_issues;
    /// Wnt possible issues.
    issues::wnt_possible_issue_vector_t wnt_possible_issues;
    /// Prefetchnta issues.
    issues::prefetchnta_issue_vector_t prefetchnta_issues;
};

typedef shared_ptr<advice_result_t> advice_result_ptr_t;
typedef shared_ptr<const advice_result_t> advice_result_cptr_t;

/// Map between cache sizes and indexes in the data vectors.
typedef map<fnr_cache_size_t,
	    unsigned int> cache_sweep_index_map_t;

struct drill_down_result_t {
    drill_down_result_t(size_t sweep_size)
	: cache_range(), stats(sweep_size) {}

    /// Map between cache sizes and vector positions for the cache
    /// sweep.
    cache_sweep_index_map_t cache_range;

    /// Statistics for the cache sweep. The #cache_range map
    /// contains mappings between cache sizes and vector indices.
    sliced_stats_vector_t stats;
};

typedef shared_ptr<drill_down_result_t> drill_down_result_ptr_t;
typedef shared_ptr<const drill_down_result_t> drill_down_result_cptr_t;

ostream &operator<<(ostream &os, const advice_result_t &ar);
ostream &operator<<(ostream &os, const drill_down_result_t &dr);

END_NAMESPACE;
END_NAMESPACE;

#endif
