/// @file analysis/call_stack/stack_pc_map.cc

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#include "stack_pc_map.hh"

#include <boost/lexical_cast.hpp>
#include <boost/foreach.hpp>
#include <iterator>

using namespace std;
using namespace boost;
using namespace acumem::library;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

// ----- trunc_stack_pc_t -----------------------------------------------------

bool
operator==(const trunc_stack_pc_t &x, const trunc_stack_pc_t &y) {
    return x.pc == y.pc && x.ret_addrs == y.ret_addrs;
}
bool
operator<(const trunc_stack_pc_t &x, const trunc_stack_pc_t &y) {
    return (x.pc < y.pc ||
            (x.pc == y.pc && x.ret_addrs < y.ret_addrs));
}

// ----- trunc_stack_pc_id_set_t ----------------------------------------------

ostream &
operator <<(ostream &os, const trunc_stack_pc_id_set_t &s)
{
    os << "(";
    bool first = true;
    BOOST_FOREACH(const trunc_stack_pc_id_t &i, s) {
	if (first)
	    first = false;
	else
	    os << ",";

	os << i;
    }
    os << ")";

    return os;
}

// ----- trunc_stack_pc_id_vector_t -------------------------------------------

ostream &
operator <<(ostream &os, const trunc_stack_pc_id_vector_t &s)
{
    os << "(";
    bool first = true;
    BOOST_FOREACH(const trunc_stack_pc_id_t &i, s) {
	if (first)
	    first = false;
	else
	    os << ",";

	os << i;
    }
    os << ")";

    return os;
}

// ----- stack_pc_map_result_t ------------------------------------------------

location_t
stack_pc_map_result_t::get_location(
    fnr_stack_id_t stack_id, fnr_address_t pc) const
{
    stack_id_pc_map_t::const_iterator id_it =
        stack_id_pc_map.find(stack_id_pc_t(stack_id, pc));
    assert(id_it != stack_id_pc_map.end());
    return id_it->second;
}

location_t
stack_pc_map_result_t::get_location(
    const fnr_sample_access_t *access) const
{
    return get_location(access->stack, access->pc);
}

location_t
stack_pc_map_result_t::get_pc_1_plus_location(
    const fnr_event_sample_t *sample) const
{
    return get_location(sample->pc_1_plus_stack, sample->pc_1_plus);
}

fnr_address_t
stack_pc_map_result_t::get_pc(location_t location) const
{
    trunc_stack_pc_id_map_t::const_iterator trunc_stack_pc_it =
        trunc_stack_pc_id_map.find(location);
    assert(trunc_stack_pc_it != trunc_stack_pc_id_map.end());
    return trunc_stack_pc_it->second.pc;
}

const trunc_stack_pc_t *
stack_pc_map_result_t::get_trunc_stack_pc(location_t location) const
{
    trunc_stack_pc_id_map_t::const_iterator trunc_stack_pc_it =
        trunc_stack_pc_id_map.find(location);
    assert(trunc_stack_pc_it != trunc_stack_pc_id_map.end());
    return &trunc_stack_pc_it->second;
}

// ----- stack_pc_map_t -------------------------------------------------------

/// Maps a truncated stack + pc to its id.
typedef map<trunc_stack_pc_t, trunc_stack_pc_id_t> trunc_stack_pc_map_t;

static trunc_stack_pc_id_t
handle_stack_pc(stack_pc_map_result_ptr_t res,
                parser_result_cptr_t pm_res,
                trunc_stack_pc_map_t &trunc_stack_pc_map,
                fnr_stack_depth_t max_depth,
                fnr_stack_id_t stack_id,
                fnr_address_t pc)
{
    // Generate the truncated stack + pc.
    trunc_stack_pc_t trunc_stack_pc;
    const fnr_event_stack_t *stack = pm_res->stacks[stack_id];
    fnr_stack_depth_t depth = (fnr_stack_depth_t)stack->ret_addrs.size();
    fnr_stack_depth_t start_depth =
        depth < max_depth ? 0 : depth  - max_depth;
    for (fnr_stack_depth_t i = start_depth; i < depth; i++)
        trunc_stack_pc.ret_addrs.push_back(stack->ret_addrs[i]);
    trunc_stack_pc.pc = pc;

    // Lookup the id of the truncated stack + pc.
    trunc_stack_pc_map_t::const_iterator trunc_stack_pc_id_it =
        trunc_stack_pc_map.find(trunc_stack_pc);
    trunc_stack_pc_id_t trunc_stack_pc_id;
    if (trunc_stack_pc_id_it != trunc_stack_pc_map.end())
        // Truncated stack + pc has already been assigned an id.
        trunc_stack_pc_id = trunc_stack_pc_id_it->second;
    else {
        // Truncated stack + pc has not yet been assigned an id. Assign it an
        // id and insert it into the maps.
        trunc_stack_pc_id = (trunc_stack_pc_id_t)trunc_stack_pc_map.size();
        trunc_stack_pc_map[trunc_stack_pc] = trunc_stack_pc_id;
        res->trunc_stack_pc_id_map[trunc_stack_pc_id] = trunc_stack_pc;
    }

    // Create a mapping from the stack id + pc to the truncated stack + pc id.
    res->stack_id_pc_map[stack_id_pc_t(stack_id, pc)] = trunc_stack_pc_id;

    // Populate the full call stack map with the frames that lie above
    // the trunc stack. Update the counts along the way.
    res->full_stack[trunc_stack_pc_id].node.deposit(
	stack->ret_addrs, start_depth);

    return trunc_stack_pc_id;
}

stack_pc_map_result_cptr_t
stack_pc_map_t::calculate(
    feature_context_ptr_t context,
    parser_result_cptr_t parser_result,
    fnr_stack_depth_t stack_depth)
{
    int event_count = 0;

    stack_pc_map_result_ptr_t result(new stack_pc_map_result_t());

    // Maps a truncated stack + pc to its id.
    trunc_stack_pc_map_t trunc_stack_pc_map;

    // Create mappings for all pcs in all samples and danglings.
    BOOST_FOREACH(const parser_burst_t &pb, parser_result->bursts) {
	BOOST_FOREACH(const event_sample_ptr_t &s, pb.samples_by_pc1) {
	    if ((event_count++%100) == 0)
		context->feedback->test_cancel();

	    handle_stack_pc(result, parser_result, trunc_stack_pc_map,
			    stack_depth,
			    s->begin.stack, s->begin.pc);
	    handle_stack_pc(result, parser_result, trunc_stack_pc_map,
			    stack_depth,
			    s->pc_1_plus_stack, s->pc_1_plus);
	    BOOST_FOREACH(const fnr_sample_access_t &access, s->end)
		handle_stack_pc(result, parser_result, trunc_stack_pc_map,
				stack_depth,
				access.stack, access.pc);
	}
    }

    return result;
}

stack_pc_map_t::stack_pc_map_t(feature_context_ptr_t context,
                               const string &sample_file,
                               fnr_stack_depth_t stack_depth)
    : feature_t<stack_pc_map_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth)
{
}

stack_pc_map_result_cptr_t
stack_pc_map_t::get() const
{
    return get_data();
}

string
stack_pc_map_t::name() const
{
    return "stack_pc_map_t(" +
        sample_file + "," +
        lexical_cast<string>(stack_depth) + ")";
}

stack_pc_map_result_cptr_t
stack_pc_map_t::calculate() const
{
    return calculate(
	context,
        parser_t(context, sample_file).get(),
        stack_depth);
}

END_NAMESPACE;
END_NAMESPACE;
