/// @file analysis/call_stack/stack_pc_map.hh

// Copyright (c) 2007-2008, 2010 Acumem AB.  All rights reserved.

/* The stack_pc_map module creates a mapping from <call stack id + pc> to a
   <truncated call stack + pc> id. This id can then be used instead of the pc
   in the rest of the analysis to create a call stack aware analysis.

   It also creates as also a mapping from <truncated call stack + pc> id to the
   actual truncated stack and pc, for presentation in the analysis result. */

#ifndef STACK_PC_MAP_HH
#define STACK_PC_MAP_HH

#include "full_stack.hh"
#include "analysis/misc/parser.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace acumem::library;

using boost::shared_ptr;

// ----- trunc_stack_pc_t -----------------------------------------------------

// Return addresses of the innermost stack frames of a trucated call stack, the
// outermost frame first.
typedef vector<fnr_address_t> trunc_stack_t;

/// A truncated call stack + pc.
struct trunc_stack_pc_t {
    /// The truncated call stack.
    trunc_stack_t ret_addrs;

    /// The pc.
    fnr_address_t pc;
};

// Comparison operators for trunc_stack_pc_t.
bool operator==(const trunc_stack_pc_t &x, const trunc_stack_pc_t &y);
bool operator<(const trunc_stack_pc_t &x, const trunc_stack_pc_t &y);

typedef set<trunc_stack_pc_t> trunc_stack_pc_set_t;

// ----- trunc_stack_pc_id_t --------------------------------------------------

// Id of a truncated stack + pc.
typedef fnr_sample_count_t trunc_stack_pc_id_t;

// ----- trunc_stack_pc_id_set_t ----------------------------------------------

// Set of truncated stack + pc ids.
typedef set<trunc_stack_pc_id_t> trunc_stack_pc_id_set_t;
ostream & operator <<(ostream &os, const trunc_stack_pc_id_set_t &s);

// ----- trunc_stack_pc_id_vector_t -------------------------------------------

// Vector of truncated stack + pc ids.
typedef vector<trunc_stack_pc_id_t> trunc_stack_pc_id_vector_t;
ostream & operator <<(ostream &os,
                           const trunc_stack_pc_id_vector_t &s);

// ----- short type names -----------------------------------------------------

typedef trunc_stack_pc_id_t location_t;
typedef trunc_stack_pc_id_set_t location_set_t;
typedef vector<location_set_t> location_set_vector_t;

// ----- special locations ----------------------------------------------------
#define LOCATION_INVALID ((location_t)-1)

// ----- full call stack info--------------------------------------------------

// Record the preceeding part of callstacks (which is not part of the
// truncated stacks.

struct stack_tree_node_holder_t {
    ~stack_tree_node_holder_t()	{
        node.destroy();
    }
    full_stack_node_t node;
};

typedef map<location_t, stack_tree_node_holder_t> full_call_stack_map_t;

// ----- stack_pc_map_result_t ------------------------------------------------

/// Maps a stack id + pc to the corresponding truncated stack + pc id.
typedef pair<fnr_stack_id_t, fnr_address_t> stack_id_pc_t;
typedef map<stack_id_pc_t, trunc_stack_pc_id_t> stack_id_pc_map_t;

/// Maps a truncated stack + pc id to the truncated stack + pc.
typedef map<trunc_stack_pc_id_t, trunc_stack_pc_t> trunc_stack_pc_id_map_t;

struct stack_pc_map_result_t {
    /// Returns the truncated stack + pc id of the specified stack id and pc.
    location_t get_location(fnr_stack_id_t stack_id,
			    fnr_address_t pc) const;

    location_t get_location(const fnr_sample_access_t *access) const;

    /// Returns the truncated stack + pc id of the specified access of the
    /// sample.
    location_t get_pc_1_plus_location(
        const fnr_event_sample_t *sample) const;

    /// Returns the pc correspending to a truncated stack + pc id.
    fnr_address_t get_pc(location_t location) const;

    /// Returns the truncaded stack and pc of a location id
    /// (corresponding to a "truncated stack + pc" id).
    const trunc_stack_pc_t *get_trunc_stack_pc(location_t location) const;

    /// Maps a stack id + pc to its truncated stack + pc id.
    stack_id_pc_map_t stack_id_pc_map;

    /// Maps a truncated stack + pc id (location_t) to the truncated
    /// stack + pc.
    trunc_stack_pc_id_map_t trunc_stack_pc_id_map;

    /// Maps a location to a record capturing the remaining call stack, beyond
    /// the trunc stack associated with the location.
    full_call_stack_map_t full_stack;
};

typedef shared_ptr<stack_pc_map_result_t> stack_pc_map_result_ptr_t;
typedef shared_ptr<const stack_pc_map_result_t> stack_pc_map_result_cptr_t;

// ----- stack_pc_map_t -------------------------------------------------------

class stack_pc_map_t : feature_t<stack_pc_map_result_cptr_t> {
public:
    static stack_pc_map_result_cptr_t calculate(
	feature_context_ptr_t context,
        parser_result_cptr_t parser_result,
        fnr_stack_depth_t stack_depth);

    stack_pc_map_t(feature_context_ptr_t context,
                   const string &sample_file,
                   fnr_stack_depth_t stack_depth);
    stack_pc_map_result_cptr_t get() const;

protected:
    string name() const;
    stack_pc_map_result_cptr_t calculate() const;

private:
    const string sample_file;
    const fnr_stack_depth_t stack_depth;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
