/// @file analysis/communication/communication.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include <boost/lexical_cast.hpp>

#include "analysis/filter/filter_module.hh"

#include "communication.hh"

using namespace boost;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

// --- communication_result_t -------------------------------------------

communication_info_t
communication_result_t::get_info(const locationx_set_t &lx_set) const
{
    communication_info_t info;

    // XXX: This should be optimized to utilize the known ordering of
    // the location set and the communication event map.
    BOOST_FOREACH(const locationx_t &l, lx_set) {
	communication_consumer_map_t::const_iterator it = comm.find(l);

	if (it != comm.end())
	    info.add(it->second.all);
    }

    return info;
}

communication_info_t
communication_result_t::get_info(const locationx_t &lx) const
{
    communication_consumer_map_t::const_iterator it = comm.find(lx);
    if (it != comm.end())
	return it->second.all;
    else
	return communication_info_t();
}

communication_info_t
communication_result_t::get_info() const
{
    communication_info_t info;

    BOOST_FOREACH(const communication_consumer_map_t::value_type &c, comm)
	info.add(c.second.all);

    return info;
}

result_valid_t
communication_result_t::utilization(const locationx_set_t &lx_set) const
{
    return get_info(lx_set).utilization();
}

result_valid_t
communication_result_t::utilization(const locationx_t lx) const
{
    return get_info(lx).utilization();
}

result_valid_t
communication_result_t::utilization() const
{
    return get_info().utilization();
}


double
communication_result_t::waste(const locationx_set_t &lx_set) const
{
    return get_info(lx_set).waste();
}

double
communication_result_t::waste(const locationx_t lx) const
{
    return get_info(lx).waste();
}

double
communication_result_t::waste() const
{
    return get_info().waste();
}

// --- communication_t --------------------------------------------------

communication_result_cptr_t
communication_t::calculate(const sample_mapping_result_cptr_t &sample_mapping,
			   const time_mapping_result_cptr_t &time_mappings,
			   const locationx_mapper_result_cptr_t &lx_mappings,
			   const repl_fr_engine_result_cptr_t &fr_byte,
			   const repl_fr_engine_result_cptr_t &fr_line,
			   const fnr_line_size_t line_size)
{
    communication_result_ptr_t result(new communication_result_t());

    parser_burst_id_t burst(0);
    time_mapping_burst_vector_t::const_iterator it_time(
	time_mappings->bursts.begin());
    BOOST_FOREACH(const sample_mapping_burst_t &smb, sample_mapping->bursts) {
	const time_mapping_burst_t &time(*it_time++);

	BOOST_FOREACH(const mapped_repl_byte_t &bs, smb.repl_byte) {
	    const mapped_access_t &start(*bs.start());
	    const locationx_t start_lx(lx_mappings->get_locationx(start));

	    BOOST_FOREACH(const mapped_access_t *a, bs.comm_use()) {
		const locationx_t consumer_lx(lx_mappings->get_locationx(*a));
		const tm_local_time_t comm_time(
		    time.get_local_time(start.cache, a->global_time()));
		const double p_repl(
		    fr_byte->get_repl_fr(burst, start.cache,
					 start.local_time,
					 comm_time));

		result->comm[consumer_lx].add_byte(start_lx, p_repl,
						   smb.sample_period);
	    }
	}

	BOOST_FOREACH(const mapped_repl_line_t &ls, smb.repl_line) {
	    const mapped_access_t *modified(ls.modified());

            if (modified == NULL)
                continue;

	    const locationx_t modified_lx(lx_mappings->get_locationx(*modified));

	    BOOST_FOREACH(const mapped_access_t *a, ls.comm()) {
		const locationx_t consumer_lx(lx_mappings->get_locationx(*a));
		const tm_local_time_t comm_time(
		    time.get_local_time(modified->cache, a->global_time()));
		const double p_repl(
		    fr_line->get_repl_fr(burst, modified->cache,
					 modified->local_time,
					 comm_time));

		result->comm[consumer_lx].add_line(modified_lx, p_repl,
						   line_size,
						   smb.sample_period);
	    }
	}

	burst++;
    }

    return result;
}

communication_t::communication_t(feature_context_ptr_t context,
				 const string &sample_file,
				 fnr_stack_depth_t stack_depth,
				 fnr_line_size_t line_size,
				 fnr_line_size_t page_size,
				 fnr_cache_size_t cache_size,
				 const thread_mapping_cptr_t &thread_mapping,
				 const cpu_model_t &cpu_model,
				 const repl_fr_engine_cptr_t &fr_engine,
				 const locationx_mapper_cptr_t &lx_mapper)
    : feature_t<communication_result_cptr_t>(context),
      sample_file(sample_file), stack_depth(stack_depth),
      line_size(line_size), page_size(page_size),
      cache_size(cache_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      fr_engine(fr_engine),
      lx_mapper(lx_mapper)
{
}

string
communication_t::name() const
{
    return "communication_t(" +
        sample_file + "," +
        lexical_cast<string>(stack_depth) + "," +
        lexical_cast<string>(line_size) + "," +
        lexical_cast<string>(page_size) + "," +
        lexical_cast<string>(cache_size) + "," +
        thread_mapping->name() + "," +
	cpu_model.name() + "," +
	fr_engine->name() + "," +
	lx_mapper->name() + ")";
}

communication_result_cptr_t
communication_t::get() const
{
    return get_data();
}

communication_result_cptr_t
communication_t::calculate() const
{
    return calculate(
	sample_mapping_t(context, sample_file, stack_depth, line_size,
			 page_size, thread_mapping, cpu_model).get(),
        time_mapping_t(context, sample_file, thread_mapping).get(),
	lx_mapper->get(),
	fr_engine->get(Mapped_Line, Mapped_Byte,
		       line_size, page_size, cache_size),
	fr_engine->get(Mapped_Line, Mapped_Line,
		       line_size, page_size, cache_size),
	line_size);
}


END_NAMESPACE;
END_NAMESPACE;
