/// @file analysis/communication/communication.hh

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#ifndef COMMUNICATION_HH
#define COMMUNICATION_HH

#include "library/lang/common_macros.h"

#include "analysis/fetch_ratio/repl_fr.hh"
#include "analysis/locationx/locationx.hh"
#include "analysis/thread_mapping/sample_mapping.hh"
#include "analysis/misc/result_valid.hh"

#include <boost/shared_ptr.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;

using boost::shared_ptr;

struct communication_info_t {
    communication_info_t()
	: line_comm(0), byte_comm(0) {}

    void add(const communication_info_t &ci) {
	line_comm += ci.line_comm;
	byte_comm += ci.byte_comm;
    }

    result_valid_t utilization() const {
	if (has_communication())
	    return result_valid_t(true, min(1.0, byte_comm / line_comm));
	else
	    // Communication sites without line samples have no real
	    // communication.
	    return result_valid_t(false, 1.0);
    }

    double waste() const {
	return max(0.0, line_comm - byte_comm);
    }

    bool has_communication() const {
	return line_comm > 0;
    }

    /**
     * Number of bytes communicated with cache line transfers for a
     * given cache size. Each line sample may cause a transmission of
     * up to 1 full cache line (1 - p_{repl}).
     */
    double line_comm;
    /**
     * Number of transferred bytes that were actually used a given
     * cache size. Each byte sample may cause a transmission of up to
     * 1 byte (1 - p_{repl}).
     */
    double byte_comm;
};

typedef map<locationx_t, communication_info_t> communication_producer_map_t;

struct communication_consumer_t {
    void add_line(locationx_t producer, double p_repl,
		  fnr_line_size_t line_size, double sample_period) {
	communication_info_t &p(producers[producer]);
	p.line_comm += (1 - p_repl) * line_size * sample_period;
	all.line_comm += (1 - p_repl) * line_size * sample_period;
    }

    void add_byte(locationx_t producer, double p_repl, double sample_period) {
	communication_info_t &p(producers[producer]);
	p.byte_comm += (1 - p_repl) * sample_period;
	all.byte_comm += (1 - p_repl) * sample_period;
    }

    communication_producer_map_t producers;
    communication_info_t all;
};

typedef map<locationx_t, communication_consumer_t> communication_consumer_map_t;

struct communication_result_t {
    //@{
    /**
     * Return the communication information about a (set of) consumer
     * locations. Used internally by the convenience functions
     * utilization() and waste().
     */
    communication_info_t get_info(const locationx_set_t &lx_set) const;
    communication_info_t get_info(const locationx_t &lx) const;
    communication_info_t get_info() const;
    //@}


    //@{
    /**
     * Calculate communication utilization for a (set of) consumer(s),
     * i.e. PC2 locations.
     */
    result_valid_t utilization(const locationx_set_t &lx_set) const;
    result_valid_t utilization(const locationx_t lx) const;
    result_valid_t utilization() const;
    //@}

    //@{
    /**
     * Calculate communication waste for a (set of) consumer(s),
     * i.e. PC2, locations.
     */
    double waste(const locationx_set_t &lx_set) const;
    double waste(const locationx_t lx) const;
    double waste() const;
    //@}

    communication_consumer_map_t comm;
};

typedef shared_ptr<communication_result_t> communication_result_ptr_t;
typedef shared_ptr<const communication_result_t> communication_result_cptr_t;

/**
 * Calculate the communication efficiency and communication waste for
 * all samples in the byte_samples and line_samples sample sets.
 *
 * The communication utilization between a pair of locations is defined
 * as the "amount of data communicated in byte samples (W-R)" divided
 * by the "amount of data communicated in line samples (W-R)".
 *
 * Communication waste is defined as "the amount of data communicated
 * in byte samples" subtracted from "the amount of data communicated
 * in line samples".
 *
 * The data amounts are defined as expectation values and depend on
 * the distance between the producer and consumer (measured in memory
 * accesses to the consumer cache).
 *
 * \param samples cache mapped samples to work on.
 * \return a pointer shared_ptr to a communication_result_t, always != NULL.
 */
class communication_t : public feature_t<communication_result_cptr_t> {
public:
    static communication_result_cptr_t calculate(
	const sample_mapping_result_cptr_t &sample_mapping,
	const time_mapping_result_cptr_t &time_mappings,
	const locationx_mapper_result_cptr_t &lx_mappings,
	const repl_fr_engine_result_cptr_t &fr_byte,
	const repl_fr_engine_result_cptr_t &fr_line,
	const fnr_line_size_t line_size);
    communication_t(feature_context_ptr_t context,
		    const string &sample_file,
		    fnr_stack_depth_t stack_depth,
		    fnr_line_size_t line_size,
		    fnr_line_size_t page_size,
		    fnr_cache_size_t cache_size,
		    const thread_mapping_cptr_t &thread_mapping,
		    const cpu_model_t &cpu_model,
		    const repl_fr_engine_cptr_t &fr_engine,
		    const locationx_mapper_cptr_t &lx_mapper);
    communication_result_cptr_t get() const;

protected:
    string name() const;
    communication_result_cptr_t calculate() const;

private:
    const string sample_file;
    const fnr_stack_depth_t stack_depth;
    const fnr_line_size_t line_size;
    const fnr_line_size_t page_size;
    const fnr_cache_size_t cache_size;
    const thread_mapping_cptr_t thread_mapping;
    const cpu_model_t cpu_model;
    const repl_fr_engine_cptr_t fr_engine;
    const locationx_mapper_cptr_t lx_mapper;
};


END_NAMESPACE;
END_NAMESPACE;

#endif
