/// @file analysis/communication/false_sharing.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include <algorithm>
#include <boost/lexical_cast.hpp>

#include "analysis/filter/filter_module.hh"

#include "false_sharing.hh"

using namespace boost;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

// --- false_sharing_site_t ---------------------------------------------

double
false_sharing_site_t::total_false_sharing_accesses() const
{
    double total(0);
    BOOST_FOREACH(const false_sharing_offenders_t::value_type &v,
		  offenders)
	total += v.second;

    return total;
}


// --- false_sharing_result_t -------------------------------------------

false_sharing_info_t
false_sharing_result_t::get(const locationx_set_t &lx_set) const
{
    double fs_lines(0);
    double accesses(0);
    bool valid(false);

    BOOST_FOREACH(const locationx_t &lx, lx_set) {
	false_sharing_map_t::const_iterator it(false_sharing.find(lx));

	if (it != false_sharing.end()) {
	    const false_sharing_site_t &fss(it->second);
	    fs_lines += fss.total_false_sharing_accesses();
	    accesses += fss.num_accesses;
	    valid |= !fss.offenders.empty();
	}
    }

    if (valid)
	return false_sharing_info_t(
	    fs_lines * line_size,
	    fs_lines / accesses);
    else
	return false_sharing_info_t();
}

false_sharing_info_t
false_sharing_result_t::get(const locationx_t lx) const
{
    false_sharing_map_t::const_iterator it(false_sharing.find(lx));

    if (it != false_sharing.end() &&
	!it->second.offenders.empty()) {
	const false_sharing_site_t &site(it->second);
	const double fs_lines(site.total_false_sharing_accesses());
	return false_sharing_info_t(
	    fs_lines * line_size,
	    fs_lines / site.num_accesses);
    } else
	return false_sharing_info_t();
}

false_sharing_info_t
false_sharing_result_t::get() const
{
    double fs_lines(0);
    double accesses(0);
    bool valid(false);

    BOOST_FOREACH(const false_sharing_map_t::value_type &v, false_sharing) {
	const false_sharing_site_t &fss(v.second);
	fs_lines += fss.total_false_sharing_accesses();
	accesses += fss.num_accesses;
	valid |= !fss.offenders.empty();
    }

    if (valid)
	return false_sharing_info_t(
	    fs_lines * line_size,
	    fs_lines / accesses);
    else
	return false_sharing_info_t();
}

// --- false_sharing_t --------------------------------------------------

false_sharing_result_cptr_t
false_sharing_t::calculate(
    const cpu_model_t &cpu_model,
    const sample_mapping_result_cptr_t &samples,
    const locationx_mapper_result_cptr_t &lx,
    const time_mapping_result_cptr_t &time_mappings,
    const repl_fr_engine_result_cptr_t &fr_engine,
    const fnr_line_size_t line_size)
{
    false_sharing_result_ptr_t result(
	new false_sharing_result_t(line_size));
    false_sharing_map_t &fsm(result->false_sharing);

    parser_burst_id_t burst(0);
    time_mapping_burst_vector_t::const_iterator it_time(
	time_mappings->bursts.begin());
    BOOST_FOREACH(const sample_mapping_burst_t &smb, samples->bursts) {
	const time_mapping_burst_t &time(*it_time++);

	BOOST_FOREACH(const mapped_repl_line_t &ls, smb.repl_line) {
	    const mapped_access_t &end(*ls.end());
	    const locationx_t end_lx(lx->get_locationx(end));
	    false_sharing_site_t &fss(fsm[end_lx]);
	    const mapped_access_t *fs(ls.false_sharing());

	    fss.num_accesses += smb.sample_period;
	    if (fs) {
		const locationx_t fs_lx(lx->get_locationx(*fs));
		const tm_local_time_t fs_end_time(
		    time.get_local_time(end.cache, fs->global_time()));

                tm_local_time_t fs_start_time;
                if (cpu_model.makes_valid(end) &&
                    ls.devalid() == fs)
                    fs_start_time = ls.valid()->local_time;
                else if (cpu_model.makes_exclusive(end) &&
                         ls.deexclusive() == fs)
                    fs_start_time = ls.exclusive()->local_time;
                else if (cpu_model.makes_modified(end) &&
                         ls.demodified() == fs)
                    fs_start_time = ls.modified()->local_time;
                else if (cpu_model.makes_nontemporal(end) &&
                         ls.denontemporal() == fs)
                    fs_start_time = ls.nontemporal()->local_time;
                else {
                    assert(0);
                    while (1);
                }

                const double p_repl = fr_engine->get_repl_fr(
                    burst, end.cache, fs_start_time, fs_end_time);

                fss.offenders[fs_lx] += (1 - p_repl) * smb.sample_period;
	    }
	}
	burst++;
    }

    return result;
}

false_sharing_t::false_sharing_t(feature_context_ptr_t context,
				 const string &sample_file,
				 fnr_stack_depth_t stack_depth,
				 fnr_line_size_t line_size,
				 fnr_line_size_t page_size,
				 fnr_cache_size_t cache_size,
				 const thread_mapping_cptr_t &thread_mapping,
				 const cpu_model_t &cpu_model,
				 const repl_fr_engine_cptr_t &fr_engine,
				 const locationx_mapper_cptr_t &lx_mapper)
    : feature_t<false_sharing_result_cptr_t>(context),
      sample_file(sample_file), stack_depth(stack_depth),
      line_size(line_size), page_size(page_size),
      cache_size(cache_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      fr_engine(fr_engine),
      lx_mapper(lx_mapper)
{
}

string
false_sharing_t::name() const
{
    return "false_sharing_t(" +
        sample_file + "," +
        lexical_cast<string>(stack_depth) + "," +
        lexical_cast<string>(line_size) + "," +
        lexical_cast<string>(page_size) + "," +
        lexical_cast<string>(cache_size) + "," +
        thread_mapping->name() + "," +
	cpu_model.name() + "," +
	lx_mapper->name() + ")";
}

false_sharing_result_cptr_t
false_sharing_t::get() const
{
    return get_data();
}

false_sharing_result_cptr_t
false_sharing_t::calculate() const
{
    return calculate(
        cpu_model,
	sample_mapping_t(context, sample_file, stack_depth,
			 line_size, page_size,
			 thread_mapping, cpu_model).get(),
	lx_mapper->get(),
	time_mapping_t(context, sample_file, thread_mapping).get(),
	fr_engine->get(Mapped_Line, Mapped_Line,
		       line_size, page_size, cache_size),
        line_size);
}


END_NAMESPACE;
END_NAMESPACE;
