/// @file analysis/communication/false_sharing.hh

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#ifndef FALSE_SHARING_HH
#define FALSE_SHARING_HH

#include "library/lang/common_macros.h"

#include "analysis/fetch_ratio/repl_fr.hh"
#include "analysis/locationx/locationx.hh"
#include "analysis/thread_mapping/sample_mapping.hh"

#include <boost/shared_ptr.hpp>


BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;

using boost::shared_ptr;

/// Map between accesses causing false sharing and the expected number
/// of false sharing events.
typedef map<locationx_t, double> false_sharing_offenders_t;

struct false_sharing_site_t {
    false_sharing_site_t()
	: offenders(), num_accesses(0) {}

    double total_false_sharing_accesses() const;

    false_sharing_offenders_t offenders;
    double num_accesses;
};

struct false_sharing_info_t {
    false_sharing_info_t()
	: waste(0.0), ratio(0.0), valid(false) {}

    false_sharing_info_t(const double waste, const double ratio)
	: waste(waste), ratio(ratio), valid(true) {}
    double waste;
    double ratio;
    bool valid;
};

/// Map between pc_end and false_sharing_site_t strucutres.
typedef map<locationx_t, false_sharing_site_t> false_sharing_map_t;

struct false_sharing_result_t {
    false_sharing_result_t(fnr_line_size_t line_size)
	: line_size(line_size) {}

    //@{
    /**
     * Calculate false sharing waste and ratio for a (set of) pc_end
     * location(s).
     */
    false_sharing_info_t get(const locationx_set_t &lx_set) const;
    false_sharing_info_t get(const locationx_t lx) const;
    false_sharing_info_t get() const;
    //@}

    /// Producer pc_end to false sharing site map
    false_sharing_map_t false_sharing;

    const fnr_line_size_t line_size;
};

typedef shared_ptr<false_sharing_result_t> false_sharing_result_ptr_t;
typedef shared_ptr<const false_sharing_result_t> false_sharing_result_cptr_t;

/**
 * Calculate the false sharing ratio and false sharing communication
 * for a set of samples.
 *
 * False sharing is calculated using line samples and their associated
 * byte samples (using the shadow mapper), the data are stored on a
 * per PC1 (producer PC) basis.
 *
 * For all sum the number of bytes terminated by unique byte samples
 * in every cache. For every cache that touches the line sample add
 * the number of used and unused byte to the fs_site_cache_data_t
 * structure associated with the line PC1 and the target cache. Used
 * and unused bytes are expected bytes, taking replacement (1 -
 * p_repl) and the sample period into account.
 *
 * The false sharing ratio for a false sharing site is calculated as:
 * unused_bytes / (used_bytes + unused_bytes)
 */
class false_sharing_t : public feature_t<false_sharing_result_cptr_t> {
public:
    static false_sharing_result_cptr_t calculate(
        const cpu_model_t &cpu_model,
	const sample_mapping_result_cptr_t &samples,
	const locationx_mapper_result_cptr_t &lx,
	const time_mapping_result_cptr_t &time_mappings,
	const repl_fr_engine_result_cptr_t &fr_engine,
	const fnr_line_size_t line_size);

    false_sharing_t(feature_context_ptr_t context,
		    const string &sample_file,
		    fnr_stack_depth_t stack_depth,
		    fnr_line_size_t line_size,
		    fnr_line_size_t page_size,
		    fnr_cache_size_t cache_size,
		    const thread_mapping_cptr_t &thread_mapping,
		    const cpu_model_t &cpu_model,
		    const repl_fr_engine_cptr_t &fr_engine,
		    const locationx_mapper_cptr_t &lx_mapper);
    false_sharing_result_cptr_t get() const;

protected:
    string name() const;
    false_sharing_result_cptr_t calculate() const;

private:
    const string sample_file;
    const fnr_stack_depth_t stack_depth;
    const fnr_line_size_t line_size;
    const fnr_line_size_t page_size;
    const fnr_cache_size_t cache_size;
    const thread_mapping_cptr_t thread_mapping;
    const cpu_model_t cpu_model;
    const repl_fr_engine_cptr_t fr_engine;
    const locationx_mapper_cptr_t lx_mapper;
};


END_NAMESPACE;
END_NAMESPACE;

#endif
