/// @file analysis/exe_maps/exe_maps.cc

// Copyright (c) 2007-2009 Acumem AB.  All rights reserved.

#include <sstream>

#include <boost/foreach.hpp>

#include "analysis/misc/parser.hh"
#include "analysis/feature_base/warnings.hh"
#include "exe_maps.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace boost;
using namespace acumem::library;

static bool
ignored(exe_map_t &em)
{
    return em.ignore;
}

exe_maps_result_cptr_t
exe_maps_t::calculate(feature_context_ptr_t context,
                      parser_result_cptr_t parser_result)
{
    exe_maps_result_ptr_t result(new exe_maps_result_t());

    BOOST_FOREACH(event_mmap_ptr_t &mmap, parser_result->process_info->mmaps) {
        if ((mmap->attr & FNR_Mmap_Attr_Execute) == 0 ||
            (mmap->attr & FNR_Mmap_Attr_Write) != 0 ||
#if defined(__linux__)
            // This is to avoid trying to find source code for mappings that
            // have a value in the filename column in /proc/<pid>/maps on Linux
            // but aren't really files, e.g., [vdso] and [stack]. On Solaris we
            // haven't seen that problem, and we always assume 00:00 as device.
            mmap->device == 0 ||
#endif
            mmap->filename.length() == 0)
            continue;

        exe_map_t exe_map(mmap->address,
                          mmap->size,
                          mmap->filename,
                          mmap->offset);

        BOOST_FOREACH(exe_map_t &em, result->exe_maps)
            if (exe_map.address == em.address &&
                exe_map.size == em.size &&
                exe_map.filename == em.filename &&
                exe_map.offset == em.offset)
                goto already_mapped;

        BOOST_FOREACH(exe_map_t &em, result->exe_maps) {
            uint64_t exe_map_end = exe_map.address + exe_map.size;
            uint64_t em_end = em.address + em.size;

            if (exe_map.address < em_end && em.address < exe_map_end &&
		exe_map.address - em.address != exe_map.offset - em.offset) {
		ostringstream oss;
                oss << hex
		    << "Overlapping memory mappings:" << endl
		    << "New: "
		    << exe_map.address << "-" << exe_map_end
		    << " off: " << exe_map.offset
		    << " " << exe_map.filename << endl
		    << "Old: "
		    << em.address << "-" << em_end
		    << " off: " << em.offset << " " << em.filename << endl
		    << "No source code mapping will be available for "
		    "this new mapping."
		    << dec;
		context->warn(new feature_warning_t("exe_maps", oss.str()));
                exe_map.ignore = true;
                em.ignore = true;
                break;
            }
        }

        result->exe_maps.push_back(exe_map);

    already_mapped:
        ;
    }

    remove_if(result->exe_maps.begin(),
              result->exe_maps.end(),
              ignored);

    return result;
}

exe_maps_t::exe_maps_t(feature_context_ptr_t context,
                       const string &sample_file)
    : feature_t<exe_maps_result_cptr_t>(context),
      sample_file(sample_file)
{
}

exe_maps_result_cptr_t
exe_maps_t::get() const
{
    return get_data();
}

string
exe_maps_t::name() const
{
    return "exe_maps_t(" + sample_file + ")";
}

exe_maps_result_cptr_t
exe_maps_t::calculate() const
{
    return calculate(context, parser_t(context, sample_file).get());
}

END_NAMESPACE;
END_NAMESPACE;
