/// @file analysis/feature_base/context.hh

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#ifndef CONTEXT_HH
#define CONTEXT_HH

#include <boost/shared_ptr.hpp>

#include "library/warning/warning_consumer.hh"
#include "library/feedback/feedback.hh"
#include "workspace.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;

using boost::shared_ptr;

/**
 * Structure containing context information for an analysis.
 */
struct feature_context_t {
    /// This constructuctor should only be used by tests and internal tools,
    /// since it creates a feedback object that does not do license checks.
    feature_context_t()
	: workspace_stack(new workspace_stack_t()),
	  warning_consumer(new warning_consumer_stderr_t()),
	  feedback(new feedback_base_t())
	{}

    feature_context_t(workspace_ptr_t workspace,
		      warning_consumer_ptr_t warning_consumer,
		      feedback_ptr_t feedback)
	: workspace_stack(new workspace_stack_t(workspace)),
	  warning_consumer(warning_consumer),
	  feedback(feedback)
	{}

    /**
     * Convenience function that wraps the warning_consumer->warn
     * call.
     */
    void warn(const warning_base_t *warning) const {
	warning_consumer->warn(warning);
    }

    /**
     * Explicitly clear out the entire workspace.
     *
     * This is currently needed because statistics_result_cptr_t
     * (which is one of the types stored in a workspace_ptr_t, managed by a
     * feature_context_ptr_t), contains a potential_ptr_t which
     * contains a feature_context_ptr_t.  In other words, there is a
     * shared_ptr cycle involving these objects.  This prevents
     * contexts to be released properly unless we manually break this
     * cycle. See also issue #1143.
     *
     * feature_context_t contains a
     * workspace_ptr_t, which stores a
     * statistics_result_cptr_t, which owns a
     * potential_t, which owns a smart pointer to
     * feature_context_t
     */
    void clear() {
	workspace_stack.reset();
    }

    /// Pointer to the cache area - the workspace.
    workspace_stack_ptr_t workspace_stack;
    /// Pointer to the warning consumer. Should never be null.
    warning_consumer_ptr_t warning_consumer;
    /// Pointer to the feedback consumer. Should never be null.
    feedback_ptr_t feedback;
};

typedef shared_ptr<feature_context_t> feature_context_ptr_t;

END_NAMESPACE;
END_NAMESPACE;

#endif
