/// @file analysis/feature_base/feature_base.hh

// Copyright (c) 2006-2008 Acumem AB.  All rights reserved.

#ifndef FEATURE_BASE_HH
#define FEATURE_BASE_HH

#include "context.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace std;

/// Base class for caching calculation modules.
///
/// The class's cached data type is provided through the DataType template
/// parameter, and items of this type is cached in the workspace.
template <class DataType>
class feature_t {
public:
    feature_t(const feature_context_ptr_t &context)
        : context(context) {}
    virtual ~feature_t()
        {}
    virtual string name() const = 0;

protected:
    /// Calculate the module's data to be cached.
    virtual DataType calculate() const = 0;

    /// Look up or calculate new data for a particular configuration for this
    /// module.
    DataType
    get_data(bool save = true) const {
	string n = name();
	DataType data;

	if (!context->workspace_stack->get(n, data)) {
#ifdef ENABLE_MEMORY_PROFILING
	    scoped_hierarchical_memory_profiler_t hmp(
		context->workspace_stack->top()->memory_profiler_stack);
#endif
#ifdef ENABLE_TIME_PROFILING
	    double inclusive_time, exclusive_time;
#endif
	    {
#ifdef ENABLE_TIME_PROFILING
		scoped_hierarchy_timer_t timer(
		    context->workspace_stack->top()->timer_hierarchy,
		    inclusive_time, exclusive_time);
#endif
		data = calculate();
	    }

	    if (save) {
#if defined(ENABLE_WORKSPACE_PROFILING)
		workspace_entry_t &we(context->workspace_stack->store(n, data));

#ifdef ENABLE_MEMORY_PROFILING
		we.inclusive_memory = hmp.get(true);
		we.exclusive_memory = hmp.get(false);
#endif
#ifdef ENABLE_TIME_PROFILING
		we.inclusive_time = inclusive_time;
		we.exclusive_time = exclusive_time;
#endif

#else // !defined(ENABLE_WORKSPACE_PROFILING)
		context->workspace_stack->store(n, data);
#endif // !defined(ENABLE_WORKSPACE_PROFILING)
	    }
	}

	return data;
    }

    /// Pointer to the analysis context.
    feature_context_ptr_t context;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
