/// @file analysis/feature_base/warnings.hh

// Copyright (c) 2007-2008 Acumem AB.  All rights reserved.

#ifndef WARNINGS_HH
#define WARNINGS_HH

#include "library/fenris/fnr_types.hh"
#include "library/warning/warning.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace std;

/**
 * Abstract virtual base class for all warnings that are produced by
 * the analysis components.
 */
class feature_warning_base_t
    : public warning_base_t
{
public:
    feature_warning_base_t(const string &feature);
    virtual ~feature_warning_base_t();

    /**
     * Returns a human readable string identifying the module that
     * produced the warning.
     */
    string where() const;

private:
    string _where;
};

/**
 * Generic feature warning.
 */
class feature_warning_t
    : public feature_warning_base_t
{
public:
    feature_warning_t(const string &feature,
		      const string &what);
    ~feature_warning_t();

    /**
     * Returns a string describing the warning.
     */
    string what() const;

private:
    string _what;
};

/**
 * Non fatal sample file corruption.
 */
class sample_file_corrupted_warning_t
    : public feature_warning_base_t
{
public:
    sample_file_corrupted_warning_t(const string &feature,
				    const string &why);
    ~sample_file_corrupted_warning_t();

    string what() const;

private:
    string _why;
};

/**
 * Use this warning to report that a step has been skipped because of
 * a missing line size, e.g. page size samples missing. This warning
 * should only be used if the sample file contains no samples for a
 * specific line size, use #cache_line_size_missing_warning_t instead
 * if it only applies to one specific cache.
 */
class line_size_missing_warning_t
    : public feature_warning_base_t
{
public:
    line_size_missing_warning_t(const string &feature,
				const string &why,
				const fnr_line_size_t line_size);
    ~line_size_missing_warning_t();

    /**
     * Returns the line size that caused the warning.
     */
    fnr_line_size_t line_size() const;

    string what() const;

private:
    string _why;
    fnr_line_size_t _line_size;
};


/**
 * Use this warning to report that a step has been skipped because of
 * a missing line size, e.g. page size samples missing. Use this
 * warning if the samples for a certain line size are missing in a
 * specific cache, use #line_size_missing_warning_t instead if they are
 * known to be missing globaly.
 */
class cache_line_size_missing_warning_t
    : public feature_warning_base_t
{
public:
    cache_line_size_missing_warning_t(const string &feature,
				      const string &why,
				      const unsigned int cache,
				      const fnr_line_size_t line_size);
    ~cache_line_size_missing_warning_t();

    /**
     * Returns the line size that caused the warning.
     */
    fnr_line_size_t line_size() const;

    /**
     * Returns the cache id that this warning applies to.
     */
    unsigned int cache() const;

    string what() const;

private:
    string _why;
    unsigned int _cache;
    fnr_line_size_t _line_size;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
