/// @file analysis/feature_base/workspace.hh

// Copyright (c) 2006-2009 Acumem AB.  All rights reserved.

#ifndef WORKSPACE_HH
#define WORKSPACE_HH

#include "library/lang/common_macros.h"
#include <string>
#include <boost/foreach.hpp>
#include <boost/any.hpp>
#include <boost/shared_ptr.hpp>
#include <map>
#include <list>
#include <iostream>

#ifdef ENABLE_MEMORY_PROFILING
#include "analysis/feature_base/memory.hh"
#endif

#ifdef ENABLE_TIME_PROFILING
#include "library/timer/timer.hh"
#endif

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;


#if defined(ENABLE_WORKSPACE_PROFILING)
/// Workspace entry used for profiling contains a boost::any reference
/// to the cached data and additional profiling information.
struct workspace_entry_t {
    template <class T>
    workspace_entry_t(const boost::shared_ptr<T> &value)
	: name(typeid(T).name()), accesses(0), data(value) {}

    template <class T>
    workspace_entry_t(const T &value)
	: name(typeid(T).name()), accesses(0), data(value) {}

    template <class T>
    void get(T &value) const {
	accesses++;
	value = boost::any_cast<T>(data);
    }

    /// Class name (as returned by typeid(type).name()).
    std::string name;
    /// Number of times the item has been retrieved from the cache.
    mutable uint32_t accesses;

#if defined(ENABLE_MEMORY_PROFILING)
    /// Memory usage including cached sub-analyses.
    memory_snapshot_t inclusive_memory;
    /// Memory usage excluding cached sub-analyses.
    memory_snapshot_t exclusive_memory;
#endif

#if defined(ENABLE_TIME_PROFILING)
    /// Run time including cached sub-analyses.
    double inclusive_time;
    /// Run time excluding cached sub-analyses.
    double exclusive_time;
#endif

    /// Data reference.
    boost::any data;
};

#else // !defined(ENABLE_WORKSPACE_PROFILING)

/// Non-profiling workspace entry, contains a boost::any reference to the
/// data cached in the workspace.
struct workspace_entry_t {
    template <class T>
    workspace_entry_t(const T &value)
        : data(value) {}

    template <class T>
    void get(T &value) const {
	value = boost::any_cast<T>(data);
    }

    /// Data reference.
    boost::any data;
};

#endif // !defined(ENABLE_WORKSPACE_PROFILING)


/// Workspace is a structure for caching calculated data from feature modules.
class workspace_t {
    typedef map<string, workspace_entry_t> any_map_t;

public:
    /// Retrieve a cached value from the workspace.
    ///
    /// \param key key the value was filed under.
    /// \param value reference to a destination to copy the data to.
    /// \return true if the value was found.
    template <class T>
    bool get(const string &key, T &value) const {
	any_map_t::const_iterator iter = any_map.find(key);
	if (iter == any_map.end())
	    return false;

	iter->second.get(value);
	return true;
    }

    /// Store a data item in the workspace cache, returns a reference
    /// to the stored workspace entry. Stored values should preferably
    /// be of the type boost::shared_ptr to handle memory deallocation
    /// correctly.
    ///
    /// \param key a lookup key...
    /// \param value a reference to the value to store in the cache.
    /// \return a reference to the stored #workspace_entry_t.
    template <class T>
    workspace_entry_t &store(const string &key, const T &value) {
	std::pair<any_map_t::iterator, bool> ip;

	ip = any_map.insert(make_pair(key, value));
	if (!ip.second)
	    ip.first->second = workspace_entry_t(value);

	return ip.first->second;
    }

    /// Dump the workspace to an output stream, output format depends
    /// on compile time options. See source and tools in
    /// tools/workspace_profiling for details.
    void dump(ostream &os = cerr) const;


#ifdef ENABLE_MEMORY_PROFILING
    /// Local memory profiling stack, used to get hierarchical memory
    /// profiles. See analysis/feature_base/memory.h for more
    /// information.
    memory_profiler_stack_t memory_profiler_stack;
#endif

#ifdef ENABLE_TIME_PROFILING
    /// Local timer hierarchy
    acumem::library::timer_hierarchy_t timer_hierarchy;
#endif

private:
    /// Map of cached values.
    any_map_t any_map;
};

typedef boost::shared_ptr<workspace_t> workspace_ptr_t;

/// Stack of workspaces used to allow creation of temporary workspaces
/// for larger calculations that only need local caching. Use
/// scoped_temporary_workspace_t instead of the push() and pop()
/// functions to assure correct handling in the event of exceptions
/// etc.
class workspace_stack_t {
public:
    workspace_stack_t();
    workspace_stack_t(workspace_ptr_t &w);

    /// Retrieve the first (note that a key should never exist more
    /// than once in the stack) cached value from the workspace
    /// stack. See the documentation for #workspace::get() for more
    /// information.
    ///
    /// \param key key the value was filed under.
    /// \param value reference to a destination to copy the data to.
    /// \return true if the value was found.
    template <class T>
    bool get(const string &key, T &value) const {
	BOOST_FOREACH(const workspace_ptr_t &w, stack) {
	    if (w->get(key, value))
		return true;
	}

	return false;
    }

    /// Store a data item in the top workspace. See the documentation
    /// for #workspace_t::store() for more information.
    ///
    /// \param key a lookup key...
    /// \param value a reference to the value to store in the cache.
    /// \return a reference to the stored #workspace_entry_t.
    template <class T>
    workspace_entry_t &store(const string &key, const T &value) {
	return stack.front()->store(key, value);
    }

    /// Create a new temporary workspace.
    void push();

    /// Add a specific temporary workspace on the top of the workspace
    /// stack.
    ///
    /// \param w workspace to add.
    void push(workspace_ptr_t &w);

    /// Removes the top workspace from the stack.
    void pop();

    /// Retrieves a pointer to the top workspace.
    workspace_ptr_t &top();

    /// Calls dump for all workspace entries in the stack.
    void dump(ostream &os = cerr) const;

private:
    typedef list<workspace_ptr_t> workspace_list_t;
    /// Stack of workspaces, stored as a std::list.
    workspace_list_t stack;
};

typedef boost::shared_ptr<workspace_stack_t> workspace_stack_ptr_t;

/// Class used to provide scoping of temporary workspaces. Safer than
/// calling #workspace_stack_t::push() and #workspace_stack_t::pop()
/// directly in the event of exceptions etc.
class scoped_temporary_workspace_t {
public:
    scoped_temporary_workspace_t(workspace_stack_ptr_t &workspace_stack);

    ~scoped_temporary_workspace_t();

private:
    workspace_stack_ptr_t workspace_stack;
};

END_NAMESPACE;
END_NAMESPACE;

#endif

