/// @file analysis/fetch_ratio/fr_lookup.cc

// Copyright (c) 2007-2009 Acumem AB.  All rights reserved.

#include "analysis/fetch_ratio/fr_lookup.hh"
#include "analysis/fetch_ratio/repl_fr.hh"
#include "analysis/fetch_ratio/repl_fr_line.hh"
#include "analysis/filter/filter_module.hh"
#include "analysis/thread_mapping/sample_mapping.hh"

#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>

using namespace std;
using namespace boost;
using namespace acumem::library;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

// ----- fr_lookup_data_t -----------------------------------------------------

fr_lookup_data_t::fr_lookup_data_t(const locationa_result_cptr_t &la_res,
                                   const sample_mapping_result_cptr_t &sample_mapping)
{
    // Just allocate entries for the pcs in the data structures. All fetch
    // ratio calculations are done in the get function where we know the cache
    // size.
    la_fr_data.resize(la_res->locationas.size());

    BOOST_FOREACH(const sample_mapping_burst_t &smb, sample_mapping->bursts) {
	BOOST_FOREACH(const mapped_repl_line_t &s, smb.repl_line) {
	    if (s.valid()) {
		const locationa_t la1(la_res->find(*s.valid()));
		const locationa_t la2(la_res->find(*s.end()));

		la_fr_data[la2].pc1_pc2_fr_data[la1];
	    }
	}
    }
}

// ----- fr_lookup_result_t ---------------------------------------------------

fr_lookup_result_t::fr_lookup_result_t(const fr_lookup_data_ptr_t &data,
                                       const locationx_mapper_result_cptr_t &lxm_res,
                                       size_t cache_size_index)
    : cache_size_index(cache_size_index),
      la_fr_data(data->la_fr_data),
      data(data),
      lxm_res(lxm_res)
{
}

fr_data_t
fr_lookup_result_t::fr_by_lx_set(const locationx_set_t &lx_set) const
{
    fr_data_t fr_data;

    BOOST_FOREACH(const locationx_t &lx, lx_set) {
        const locationa_set_t &la_set  =
            lxm_res->locationx_locationa.find(lx)->second;
        BOOST_FOREACH(locationa_t la, la_set)
            fr_data.add_fr_data(la_fr_data[la].fr_data[cache_size_index]);
    }

    return fr_data;
}

fr_data_t
fr_lookup_result_t::fr_by_lx(const locationx_t &lx) const
{
    fr_data_t fr_data;

    locationx_locationa_set_map_t::const_iterator it = lxm_res->locationx_locationa.find(lx);
    assert(it != lxm_res->locationx_locationa.end());
    const locationa_set_t &la_set = it->second;
    BOOST_FOREACH(locationa_t la, la_set)
        fr_data.add_fr_data(la_fr_data[la].fr_data[cache_size_index]);

    return fr_data;
}

fr_data_t
fr_lookup_result_t::pc1_pc2_fr(const locationx_t &pc1,
			       const locationx_t &pc2) const
{
    fr_data_t fr_data;

    const locationa_set_t &la1_set  =
        lxm_res->locationx_locationa.find(pc1)->second;
    const locationa_set_t &la2_set  =
        lxm_res->locationx_locationa.find(pc2)->second;

    BOOST_FOREACH(locationa_t la2, la2_set) {
        const la_fr_data_t &la2_fr_data(la_fr_data[la2]);
        BOOST_FOREACH(locationa_t la1, la1_set) {
            pc1_pc2_fr_data_map_t::const_iterator it(
                la2_fr_data.pc1_pc2_fr_data.find(la1));
            if (it != la2_fr_data.pc1_pc2_fr_data.end())
                fr_data.add_fr_data(it->second[cache_size_index]);
        }
    }

    return fr_data;
}

// ----- fr_lookup_t ----------------------------------------------------------

static bool
filter_sample(const cpu_model_t &cpu_model, fr_lookup_type_t filter,
	      const mapped_access_t &a)
{
    fnr_access_classes_t ac(a.access_classes());
    if ((ac & FNR_Access_Class_NTA) && (ac & FNR_Access_Class_WR))
	return false;

    switch (filter) {
    case FRL_Type_All:
	return true;
    case FRL_Type_Stalls:
	return !a.is_prefetch() && !a.is_flush();
    case FRL_Type_Write:
	return a.is_write(true, true);
    case FRL_Type_Write_No_Prefetch:
	return a.is_write(false, true);
    case FRL_Type_Read:
	return a.is_read(true, true);
    case FRL_Type_Read_No_Prefetch:
	return a.is_read(false, true);
    case FRL_Type_Prefetch:
	return a.is_prefetch();
    default:
	assert(0);
        while (1);
    }
}

/// Returns true if the line has been coherence invalidated.
static bool
invalidated(const mapped_repl_line_t *sample)
{
    return
        sample->devalid() != NULL &&
        sample->devalid()->cache != sample->start()->cache;
}

/// Returns true if the line has been coherence downgraded.
static bool
downgraded(const mapped_repl_line_t *sample)
{
    return
        sample->devalid() == NULL &&
        sample->deexclusive() != NULL &&
        sample->deexclusive()->cache != sample->start()->cache;
}

fr_lookup_result_cptr_t
fr_lookup_t::get()
{
    fr_lookup_data_ptr_t data(get_data());

    // Just return the result object if there is already cached fetch ratio
    // data for the requested cache size.
    vector<fnr_cache_size_t>::iterator cs_it(
        find(data->cache_sizes.begin(), data->cache_sizes.end(), cache_size));
    if (cs_it != data->cache_sizes.end())
        return fr_lookup_result_cptr_t(
            new fr_lookup_result_t(data, lx_mapper->get(), cs_it - data->cache_sizes.begin()));

    sample_mapping_result_cptr_t sample_mapping(
        sample_mapping_t(context, sample_file, stack_depth, line_size,
			 page_size, thread_mapping, cpu_model).get());

    // Don't cache the fetch rate values when called from here. This
    // class will call fetch rate for so many different sizes which
    // will only ever be used through this class, so it doesn't make
    // sense to cache the underlying data.
    repl_fr_line_result_cptr_t repl_fr_res(
	repl_fr_line_t(context, sample_file, stack_depth,
                       line_size, page_size, cache_size,
                       thread_mapping, cpu_model, engine).get(false));

    locationa_result_cptr_t la_res(
        locationa_mapper_t(context, sample_file, stack_depth, thread_mapping).get());

    // No cached fetch ratio data for this particular cache size, let's get it.
    const size_t csi(data->cache_sizes.size());
    data->cache_sizes.push_back(cache_size);

    // Allocate slots for the new cache size.
    BOOST_FOREACH(la_fr_data_t &v, data->la_fr_data) {
	assert (v.fr_data.size() == csi);
	v.fr_data.push_back(fr_data_t());

	BOOST_FOREACH(pc1_pc2_fr_data_map_t::value_type &v2,
		      v.pc1_pc2_fr_data) {
	    assert (v2.second.size() == csi);
	    v2.second.push_back(fr_data_t());
	}
    }

    repl_fr_line_burst_vector_t::const_iterator it_repl(
	repl_fr_res->bursts.begin());
    BOOST_FOREACH(const sample_mapping_burst_t &smb, sample_mapping->bursts) {
	const double sample_period(smb.sample_period);

	const ratio_vector_t &mod_to_exc(it_repl->mod_to_exc);
	const ratio_vector_t &exc_to_val(it_repl->exc_to_val);
        const ratio_vector_t &val_to_end(it_repl->val_to_end);
        it_repl++;

	const mapped_repl_line_vector_t &repl_samples(smb.repl_line);
	const fnr_sample_count_t num_repl_samples((fnr_sample_count_t)repl_samples.size());

	// Add the replacement fetch ratios to the data object.
	for (fnr_sample_count_t i = 0; i < num_repl_samples; i++) {
	    const mapped_repl_line_t &s(repl_samples[i]);

	    const mapped_access_t &end(*s.end());
	    const locationa_t la_pc2(la_res->find(end));
	    const bool filter_match(filter_sample(cpu_model, frl_type,
						  end));

            const double repl_fetch(1 - (1 - val_to_end[i]));
            const double repl_upgrade(1 - (1 - exc_to_val[i]) * (1 - val_to_end[i]));
	    const double repl_write_back(1 - (1 - mod_to_exc[i]) * (1 - exc_to_val[i]) * (1 - val_to_end[i]));

            const double end_dirty = cpu_model.makes_modified(end);

            // The regular replacement fetch ratio.
            const double repl_fr = end.is_flush() ? 0.0 : repl_fetch;

            // A coherence fetch occurs if the line has been invalidated and
            // there would have been no replacement in the local cache.
            const double coh_fr = end.is_flush() ? 0.0 :
                (invalidated(&s) ? (1 - repl_fetch) : 0.0);

	    // Set flush_fr to 1 for samples starting with a flush or Wnt, such
	    // samples has repl_fr == 0.0 since they lack a pc_valid.
	    const double flush_fr = end.is_flush() ? 0.0 :
                (s.start()->is_flush() ||
                 (s.start()->is_nta() && s.start()->is_write()) ?
                 1.0 : 0.0);

	    // Upgrade misses occur if we suffer from a downgrade by a foreign
	    // access and the end access is a write.
	    const double upgr(downgraded(&s) &&
                              end.is_write() ?
                              (1 - repl_upgrade) : 0.0);

            // A replacement since the start of the sample causes a write-back.
            const double repl_wbr =  end_dirty ? repl_write_back : 0.0;

            // A coherence write-back occurs if the access makes the line dirty
            // and the line has been invalidated or downgraded.
            const double coh_wbr =
                (end_dirty && (invalidated(&s) ||downgraded(&s))) ?
                (1 - repl_write_back) : 0.0;

	    // Set flush_wbr to 1 for samples starting with a flush or Wnt,
	    // such samples has repl_fr == 0.0 since they lack a pc_valid.
	    const double flush_wbr((s.start()->is_flush() ||
                                    (s.start()->is_nta() && s.start()->is_write())) &&
                                   end_dirty ?
                                   1.0 : 0.0);

	    la_fr_data_t &la_data = data->la_fr_data[la_pc2];
	    la_data.fr_data[csi].add_repl_sample(
		sample_period,
		repl_fr, coh_fr, flush_fr, upgr, repl_wbr, coh_wbr, flush_wbr,
		!filter_match);

	    if (s.valid()) {
		const locationa_t la_pc1(la_res->find(*s.valid()));
		la_data.pc1_pc2_fr_data[la_pc1][csi].add_repl_sample(
		    sample_period,
		    repl_fr, coh_fr, flush_fr, upgr, repl_wbr, coh_wbr, flush_wbr,
		    !filter_match);
	    }
	}

	// Add the cold fetch ratios to the data object.
	const mapped_cold_vector_t &cold_samples(smb.cold_line);
	const fnr_sample_count_t num_cold_samples((fnr_sample_count_t)cold_samples.size());
	for (fnr_sample_count_t i = 0; i < num_cold_samples; i++) {
	    const mapped_cold_t &s(cold_samples[i]);
	    const locationa_t la(la_res->find(s.end));
	    const bool filter_match(filter_sample(cpu_model, frl_type, s.end));

	    // Add cold fetch ratio to the pc1.
	    la_fr_data_t &la_data(data->la_fr_data[la]);
	    la_data.fr_data[csi].add_cold_sample(
                sample_period,
                s.end.is_flush() ? 0.0 : 1.0,
                cpu_model.makes_modified(s.end) ? 1.0 : 0.0,
                !filter_match);
	}
    }

    fr_lookup_result_ptr_t result(
        new fr_lookup_result_t(data, lx_mapper->get(), data->cache_sizes.size() - 1));

    return result;
}

fr_lookup_t::fr_lookup_t(const feature_context_ptr_t &context,
			 const string &sample_file,
			 fnr_stack_depth_t stack_depth,
			 fnr_line_size_t line_size,
			 fnr_line_size_t page_size,
			 const thread_mapping_cptr_t &thread_mapping,
			 const cpu_model_t &cpu_model,
			 const locationx_mapper_cptr_t &lx_mapper,
			 fnr_cache_size_t cache_size,
			 const repl_fr_engine_cptr_t &engine,
			 fr_lookup_type_t frl_type)
    : feature_t<fr_lookup_data_ptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      lx_mapper(lx_mapper),
      cache_size(cache_size),
      engine(engine),
      frl_type(frl_type)
{
}

string
fr_lookup_t::name() const
{
    // We omit cache_size in the name, since results for different cache sizes
    // are cached in the same data object.
    //
    // We omit the locationx mapper in the name, since results for different
    // locationx mappings are cache in the same data object.
    return "fr_lookup_t(" +
	lexical_cast<string>(sample_file) + "," +
        lexical_cast<string>(stack_depth) + "," +
	lexical_cast<string>(line_size) + "," +
	lexical_cast<string>(page_size) + "," +
        thread_mapping->name() + "," +
	cpu_model.name() + "," +
	engine->name() + "," +
	lexical_cast<string>(frl_type) + ")";
}

fr_lookup_data_ptr_t
fr_lookup_t::calculate() const
{
    locationa_result_cptr_t la_res(
        locationa_mapper_t(context, sample_file, stack_depth,
                           thread_mapping).get());
    locationx_mapper_result_cptr_t lxm_res(lx_mapper->get());
    sample_mapping_result_cptr_t sample_mapping(
        sample_mapping_t(context, sample_file, stack_depth, line_size,
			 page_size, thread_mapping, cpu_model).get());

    return fr_lookup_data_ptr_t(new fr_lookup_data_t(la_res, sample_mapping));
}

END_NAMESPACE;
END_NAMESPACE;
