/// @file analysis/fetch_ratio/fr_lookup.hh

// Copyright (c) 2007-2009 Acumem AB.  All rights reserved.

#ifndef FR_LOOKUP_HH
#define FR_LOOKUP_HH

#include "analysis/fetch_ratio/repl_fr.hh"
#include "analysis/locationx/locationx.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace acumem::library;

using boost::shared_ptr;

/// Type to use when storing calculated statistics values. The range and
/// precision of float is more than enough.
typedef float stat_fp_t;

// ----- fr_data_t ------------------------------------------------------------

/// Fetch, upgrade and write-back statistics.
struct by_kind_t {
    by_kind_t(const by_kind_t &bk)
        : access_count(bk.access_count),
          cold_fetch_count(bk.cold_fetch_count),
          replacement_fetch_count(bk.replacement_fetch_count),
          coherence_fetch_count(bk.coherence_fetch_count),
	  flush_fetch_count(bk.flush_fetch_count),
          upgrade_count(bk.upgrade_count),
          cold_write_back_count(bk.cold_write_back_count),
          replacement_write_back_count(bk.replacement_write_back_count),
          coherence_write_back_count(bk.coherence_write_back_count),
	  flush_write_back_count(bk.flush_write_back_count)
        {}

    by_kind_t(double access_count,
              double cold_fetch_count,
              double replacement_fetch_count,
              double coherence_fetch_count,
	      double flush_fetch_count,
              double upgrade_count,
              double cold_write_back_count,
              double replacement_write_back_count,
              double coherence_write_back_count,
	      double flush_write_back_count)
        : access_count((stat_fp_t)access_count),
          cold_fetch_count((stat_fp_t)cold_fetch_count),
          replacement_fetch_count((stat_fp_t)replacement_fetch_count),
          coherence_fetch_count((stat_fp_t)coherence_fetch_count),
	  flush_fetch_count((stat_fp_t)flush_fetch_count),
          upgrade_count((stat_fp_t)upgrade_count),
          cold_write_back_count((stat_fp_t)cold_write_back_count),
          replacement_write_back_count((stat_fp_t)replacement_write_back_count),
          coherence_write_back_count((stat_fp_t)coherence_write_back_count),
	  flush_write_back_count((stat_fp_t)flush_write_back_count)
        {}

    by_kind_t()
        : access_count(0),
          cold_fetch_count(0.0),
          replacement_fetch_count(0.0),
          coherence_fetch_count(0.0),
	  flush_fetch_count(0.0),
          upgrade_count(0.0),
          cold_write_back_count(0.0),
          replacement_write_back_count(0.0),
          coherence_write_back_count(0.0),
	  flush_write_back_count(0.0)
        {}

    // ----- Accessor methods for counts. -------------------------------------

    double cold_fetches() const {
        return cold_fetch_count;
    }

    double replacement_fetches() const {
        return replacement_fetch_count;
    }

    double coherence_fetches() const {
        return coherence_fetch_count;
    }

    double flush_fetches() const {
        return flush_fetch_count;
    }

    double total_fetches() const {
        return
            cold_fetch_count +
            replacement_fetch_count +
            coherence_fetch_count +
            flush_fetch_count;
    }

    double upgrades() const {
        return upgrade_count;
    }

    double total_upgrades() const {
        return upgrade_count;
    }

    double cold_write_backs() const {
        return cold_write_back_count;
    }

    double replacement_write_backs() const {
        return replacement_write_back_count;
    }

    double coherence_write_backs() const {
        return coherence_write_back_count;
    }

    double flush_write_backs() const {
        return flush_write_back_count;
    }

    double total_write_backs() const {
        return
            cold_write_back_count +
            replacement_write_back_count +
            coherence_write_back_count +
            flush_write_back_count;
    }

    // ----- Accessor methods for ratios. -------------------------------------

    double accesses() const {
        return access_count;
    }

    double cold_fetch_ratio() const {
        return access_count == 0.0 ? 0.0 : cold_fetches() / access_count;
    }

    double replacement_fetch_ratio() const {
        return access_count == 0.0 ? 0.0 : replacement_fetches() / access_count;
    }

    double coherence_fetch_ratio() const {
        return access_count == 0.0 ? 0.0 : coherence_fetches() / access_count;
    }

    double flush_fetch_ratio() const {
        return access_count == 0.0 ? 0.0 : flush_fetches() / access_count;
    }

    double total_fetch_ratio() const {
        return access_count == 0.0 ? 0.0 : total_fetches() / access_count;
    }

    double upgrade_ratio() const {
        return access_count == 0.0 ? 0.0 : upgrades() / access_count;
    }

    double total_upgrade_ratio() const {
        return access_count == 0.0 ? 0.0 : upgrades() / access_count;
    }

    double cold_write_back_ratio() const {
        return access_count == 0.0 ? 0.0 : cold_write_backs() / access_count;
    }

    double replacement_write_back_ratio() const {
        return access_count == 0.0 ? 0.0 : replacement_write_backs() / access_count;
    }

    double coherence_write_back_ratio() const {
        return access_count == 0.0 ? 0.0 : coherence_write_backs() / access_count;
    }

    double flush_write_back_ratio() const {
        return access_count == 0.0 ? 0.0 : flush_write_backs() / access_count;
    }

    double total_write_back_ratio() const {
        return access_count == 0.0 ? 0.0 : total_write_backs() / access_count;
    }

    /// TODO: Remove!
    by_kind_t multiply_fetches_upgrades(double multiplier) const {
        return by_kind_t(this->access_count,
                         this->cold_fetch_count * multiplier,
                         this->replacement_fetch_count * multiplier,
                         this->coherence_fetch_count * multiplier,
			 this->flush_fetch_count * multiplier,
                         this->upgrade_count * multiplier,
                         this->cold_write_back_count,
                         this->replacement_write_back_count,
                         this->coherence_write_back_count,
			 this->flush_write_back_count);
    }

    // ----- Comparison operators. --------------------------------------------

    bool operator==(const by_kind_t &that) const {
        return
            access_count == that.access_count &&
            cold_fetch_count == that.cold_fetch_count &&
            replacement_fetch_count == that.replacement_fetch_count &&
            coherence_fetch_count == that.coherence_fetch_count &&
            flush_fetch_count == that.flush_fetch_count &&
            upgrade_count == that.upgrade_count &&
            cold_write_back_count == that.cold_write_back_count &&
            replacement_write_back_count == that.replacement_write_back_count &&
            coherence_write_back_count == that.coherence_write_back_count &&
            flush_write_back_count == that.flush_write_back_count;
    }

private:
    /// Access count.
    stat_fp_t access_count;

    /// Fetches due to cold misses (or other unclassified categories).
    stat_fp_t cold_fetch_count;

    /// Fetches due to replacements.
    stat_fp_t replacement_fetch_count;

    /// Fetches due to coherence.
    stat_fp_t coherence_fetch_count;

    /// Fetches due to cache line flushes.
    stat_fp_t flush_fetch_count;

    /// Upgrades.
    stat_fp_t upgrade_count;

    /// Write-backs due to cold misses (or other unclassified categories).
    stat_fp_t cold_write_back_count;

    /// Write-backs due to replacements.
    stat_fp_t replacement_write_back_count;

    /// Write-backs due to coherence.
    stat_fp_t coherence_write_back_count;

    /// Write-backs due to cache line flushes.
    stat_fp_t flush_write_back_count;
};

/// Aggregate fetch ratio information for a subset of pcs. May contain combined
/// replacement and cold fetch ratio or only replacement fetch ratio depending
/// on the context.
struct fr_data_t {
    fr_data_t() :
        repl_fetches(0.0),
	cold_fetches(0.0),
	coherence_fetches(0.0),
	flush_fetches(0.0),
        upgrades(0.0),
        repl_write_backs(0.0),
	cold_write_backs(0.0),
	coherence_write_backs(0.0),
	flush_write_backs(0.0),
	sample_count(0.0),
	access_count(0.0) {}

    fr_data_t(const fr_data_t &fr_data) :
        repl_fetches(fr_data.repl_fetches),
        cold_fetches(fr_data.cold_fetches),
        coherence_fetches(fr_data.coherence_fetches),
        flush_fetches(fr_data.flush_fetches),
        upgrades(fr_data.upgrades),
        repl_write_backs(fr_data.repl_write_backs),
        cold_write_backs(fr_data.cold_write_backs),
        coherence_write_backs(fr_data.coherence_write_backs),
        flush_write_backs(fr_data.flush_write_backs),
        sample_count(fr_data.sample_count),
        access_count(fr_data.access_count) {}

    /// Add replacement fetch ratio (pc2) sample.
    void add_repl_sample(double sample_period,
			 double repl_fr,
                         double coh_fr,
                         double flush_fr,
			 double upgr,
			 double repl_wbr,
                         double coh_wbr,
                         double flush_wbr,
			 bool sample_count_only) {
	if (!sample_count_only) {
	    repl_fetches = (stat_fp_t)(repl_fetches + sample_period * repl_fr);
	    coherence_fetches = (stat_fp_t)(coherence_fetches + sample_period * coh_fr);
	    flush_fetches = (stat_fp_t)(flush_fetches + sample_period * flush_fr);
	    upgrades = (stat_fp_t)(upgrades + sample_period * upgr);
	    repl_write_backs = (stat_fp_t)(repl_write_backs + sample_period * repl_wbr);
	    coherence_write_backs = (stat_fp_t)(coherence_write_backs + sample_period * coh_wbr);
	    flush_write_backs = (stat_fp_t)(flush_write_backs + sample_period * flush_wbr);
	}
	sample_count += 1.0;
	access_count = (stat_fp_t)(access_count + sample_period);
    }

    /// Add cold fetch ratio (pc1) sample.
    void add_cold_sample(double sample_period,
                         double cold_fr,
                         double cold_wbr,
                         bool sample_count_only) {
	if (!sample_count_only) {
	    cold_fetches = (stat_fp_t)(cold_fetches + sample_period * cold_fr);
	    cold_write_backs = (stat_fp_t)(cold_write_backs + sample_period * cold_wbr);
	}
	sample_count += 1.0;
	access_count = (stat_fp_t)(access_count + sample_period);
    }

    /// Add fetch ratio data.
    void add_fr_data(const fr_data_t &fr_data) {
	repl_fetches += fr_data.repl_fetches;
	cold_fetches += fr_data.cold_fetches;
	coherence_fetches += fr_data.coherence_fetches;
	flush_fetches += fr_data.flush_fetches;
	upgrades += fr_data.upgrades;
	repl_write_backs += fr_data.repl_write_backs;
	cold_write_backs += fr_data.cold_write_backs;
	coherence_write_backs += fr_data.coherence_write_backs;
	flush_write_backs += fr_data.flush_write_backs;
        sample_count += fr_data.sample_count;
        access_count += fr_data.access_count;
    }

    /// Return event ratios.
    by_kind_t by_kind() const {
        if (samples() == 0.0)
            return by_kind_t();
        else
            return by_kind_t(access_count,
                             cold_fetches,
                             repl_fetches,
                             coherence_fetches,
                             flush_fetches,
                             upgrades,
                             cold_write_backs,
                             repl_write_backs,
                             coherence_write_backs,
                             flush_write_backs);
    }

    /// Return the estimated number of samples.
    double samples() const {
	return sample_count;
    }

    /// Return the number of accesses.
    double accesses() const {
        return access_count;
    }

private:
    /// Accumulated replacement fetches
    stat_fp_t repl_fetches;

    /// Accumulated cold fetches
    stat_fp_t cold_fetches;

    /// Accumulated coherence fetches
    stat_fp_t coherence_fetches;

    /// Accumulated flush fetches
    stat_fp_t flush_fetches;

    /// Accumulated upgrades
    stat_fp_t upgrades;

    /// Accumulated replacement write-backs
    stat_fp_t repl_write_backs;

    /// Accumulated cold write-backs
    stat_fp_t cold_write_backs;

    /// Accumulated coherence write-backs
    stat_fp_t coherence_write_backs;

    /// Accumulated flush write-backs
    stat_fp_t flush_write_backs;

    /// Sample count, including both replacement and cold samples.
    stat_fp_t sample_count;

    /// Access count, including contributions from both replacement
    /// and cold samples.
    stat_fp_t access_count;
};

typedef vector<fr_data_t> fr_data_vector_t;

// ----- la_fr_data_t ---------------------------------------------------------

/// Maps a pc1 in the context of a pc2 to fetch ratio information (only
/// replacement fetches) for the samples with that pc1 and pc2.
typedef map<locationa_t, fr_data_vector_t> pc1_pc2_fr_data_map_t;

/// Aggregate fetch ratio information for a pc.
struct la_fr_data_t {
    la_fr_data_t() {}

    /// Fetch ratio information (both cold and replacement fetches)
    /// for the pc.
    fr_data_vector_t fr_data;

    /// Map from a pc1 to fetch ratio information (only replacement fetches)
    /// for all samples with that pc1 and with this pc as pc2.
    pc1_pc2_fr_data_map_t pc1_pc2_fr_data;
};

/// Maps a locationa_t to aggregate information about that pc.
typedef vector<la_fr_data_t> la_fr_data_vector_t;

// ----- fr_lookup_data_t ----------------------------------------------------

/// Cached result data, refered to by the non-cached fr_lookup_result_t.
class fr_lookup_data_t {
    friend class fr_lookup_t;
    friend class fr_lookup_result_t;

public:
    fr_lookup_data_t(const locationa_result_cptr_t &la_res,
		     const sample_mapping_result_cptr_t &sample_mapping);

private:
    /// Cache sizes that have been populated with fetch ratios, indexed
    /// by cache size index.
    vector<fnr_cache_size_t> cache_sizes;

    /// Maps a pc to the number of and aggregate fetch ratio (both cold and
    /// replacement fetches) of all samples with that pc, and to a map
    /// from pc1 to the aggregate fetch ratio (only replacement
    /// fetches) of all samples with that pc1 and the key pc from this
    /// map as pc2.
    la_fr_data_vector_t la_fr_data;

    double sample_period;
};

typedef shared_ptr<fr_lookup_data_t> fr_lookup_data_ptr_t;

// ----- fr_lookup_result_t ---------------------------------------------------

/// Lookup result. Not cached, just refers to structures in the cached
/// fr_lookup_data_t.
class fr_lookup_result_t {
    friend class fr_lookup_t;

public:
    /// Returns the average fetch ratio (both cold and replacement fetches) of
    /// all samples with a pc in the specified pc set.
    fr_data_t fr_by_lx_set(const locationx_set_t &lx_set) const;

    /// Returns the fetch ratio (both cold and replacement fetches) of all
    /// samples with a given pc, regardless of kind.
    fr_data_t fr_by_lx(const locationx_t &lx) const;


    //@{
    /// Get the average fr for a pc2 in the context of a pc1.
    fr_data_t pc1_pc2_fr(const locationx_t &pc1,
			 const locationx_t &pc2) const;

    template <class Cont1, class Cont2>
    fr_data_t pc1_pc2_fr(const Cont1 &pc1_cont,
			 const Cont2 &pc2_cont) const {
	fr_data_t result;
	BOOST_FOREACH(const locationx_t &pc2, pc2_cont) {
            const locationa_set_t &la2_set  =
                lxm_res->locationx_locationa.find(pc2)->second;
            BOOST_FOREACH(locationa_t la2, la2_set) {
                assert(la2 < la_fr_data.size());
                const la_fr_data_t &fr_data(la_fr_data[la2]);

                BOOST_FOREACH(const locationx_t &pc1, pc1_cont) {
                    const locationa_set_t &la1_set  =
                        lxm_res->locationx_locationa.find(pc1)->second;
                    BOOST_FOREACH(locationa_t la1, la1_set) {
                        pc1_pc2_fr_data_map_t::const_iterator it(
                            fr_data.pc1_pc2_fr_data.find(la1));
                        if (it != fr_data.pc1_pc2_fr_data.end())
                            result.add_fr_data(it->second[cache_size_index]);
                    }
                }
            }
        }

	return result;
    }
    //@}

private:
    fr_lookup_result_t(const fr_lookup_data_ptr_t &data,
                       const locationx_mapper_result_cptr_t &lxm_res,
                       size_t cache_size_index);

    /// Cache size index, -1 if no cache size was specified.
    const size_t cache_size_index;

    /// Maps a pc to the number of and aggregate fetch ratio (both cold and
    /// replacement fetches) of all samples with that pc2, and to a
    /// map from pc1 to the aggregate fetch ratio (only replacement
    /// fetches) of all samples with that pc1 and the key pc from this
    /// map as pc2.
    const la_fr_data_vector_t &la_fr_data;

    // Keep a shared pointer to the data object we refer to, so that
    // it isn't garbage collected while we still hold references to it
    // from this object.
    fr_lookup_data_ptr_t data;

    locationx_mapper_result_cptr_t lxm_res;
};

typedef shared_ptr<fr_lookup_result_t> fr_lookup_result_ptr_t;
typedef shared_ptr<const fr_lookup_result_t> fr_lookup_result_cptr_t;

// ----- fr_lookup_t ----------------------------------------------------------

/// Specifies which access types to include in the result.
///
/// TODO & NOTE: Non-temporal stores are currently never included in any of
///              these classes, regardless of which of these categories is
///              selected.
enum fr_lookup_type_t {
    // All accesses.
    FRL_Type_All,

    // All accesses except prefetches and cache line flushes.
    FRL_Type_Stalls,

    // Read accesses (including read-write accesses).
    FRL_Type_Read,

    // Read accesses (including read-write accesses) except prefetches.
    FRL_Type_Read_No_Prefetch,

    // Write accesses (including read-write accesses).
    FRL_Type_Write,

    // Write accesses (including read-write accesses) except prefetches.
    FRL_Type_Write_No_Prefetch,

    // Prefetch accesses only.
    FRL_Type_Prefetch
};

class fr_lookup_t : public feature_t<fr_lookup_data_ptr_t> {
public:
    fr_lookup_t(const feature_context_ptr_t &context,
                const string &sample_file,
                fnr_stack_depth_t stack_depth,
                fnr_line_size_t line_size,
                fnr_line_size_t page_size,
                const thread_mapping_cptr_t &thread_mapping,
		const cpu_model_t &cpu_model,
		const locationx_mapper_cptr_t &lx_mapper,
                fnr_cache_size_t cache_size,
		const repl_fr_engine_cptr_t &engine,
		fr_lookup_type_t frl_type);

    // Return a result for the specified cache size.
    fr_lookup_result_cptr_t get();

protected:
    string name() const;
    fr_lookup_data_ptr_t calculate() const;

private:
    const string sample_file;
    const fnr_stack_depth_t stack_depth;
    const fnr_line_size_t line_size;
    const fnr_line_size_t page_size;
    const thread_mapping_cptr_t thread_mapping;
    const cpu_model_t cpu_model;
    const locationx_mapper_cptr_t lx_mapper;
    const fnr_cache_size_t cache_size;
    const repl_fr_engine_cptr_t engine;
    const fr_lookup_type_t frl_type;
};

END_NAMESPACE;
END_NAMESPACE;

#endif /* FR_LOOKUP_HH */
