/// @file analysis/fetch_ratio/repl_fr.cc

// Copyright (c) 2006-2009 Acumem AB.  All rights reserved.

#include "analysis/fetch_ratio/repl_fr.hh"
#include "analysis/filter/filter_module.hh"
#include "analysis/feature_base/warnings.hh"
#include "analysis/thread_mapping/sample_mapping.hh"
#include "analysis/feature_base/exceptions.hh"

#include <boost/lexical_cast.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace boost;
using namespace acumem::library;

static window_vector_t::size_type
find_window(const window_vector_t &windows,
            window_vector_t::size_type end,
            tm_local_time_t time2)
{
    while (windows[end]->begin_time > time2)
	end--;

    return end;
}

// ----- repl_fr_t ------------------------------------------------------------

repl_fr_t::repl_fr_t(const feature_context_ptr_t &context,
		     const string &sample_file,
		     fnr_stack_depth_t stack_depth,
		     mapped_sample_type_t window_sample_type,
		     mapped_sample_type_t sample_sample_type,
		     fnr_line_size_t line_size,
		     fnr_line_size_t page_size,
		     fnr_cache_size_t cache_size,
		     thread_mapping_cptr_t thread_mapping,
		     const cpu_model_t &cpu_model,
		     repl_fr_engine_cptr_t engine)
    : feature_t<repl_fr_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      window_sample_type(window_sample_type),
      sample_sample_type(sample_sample_type),
      line_size(line_size),
      page_size(page_size),
      cache_size(cache_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      engine(engine)
{
}

repl_fr_t::~repl_fr_t()
{
}

repl_fr_result_cptr_t
repl_fr_t::get(bool cache_results) const
{
    return get_data(cache_results);
}

string
repl_fr_t::name() const
{
    return "repl_fr(" +
	lexical_cast<string>(sample_file) + "," +
	lexical_cast<string>(stack_depth) + "," +
	lexical_cast<string>(window_sample_type) + "," +
	lexical_cast<string>(sample_sample_type) + "," +
	lexical_cast<string>(line_size) + "," +
	lexical_cast<string>(page_size) + "," +
	lexical_cast<string>(cache_size) + "," +
	thread_mapping->name() + "," +
	cpu_model.name() + "," +
	engine->name() + ")";
}

template<class SampleType>
static void
repl_fr_calculate(repl_fr_burst_t &burst_result,
		  const parser_burst_id_t burst_id,
		  const vector<SampleType> &samples,
		  const repl_fr_engine_result_t &engine_data,
		  const tm_cache_id_t num_caches)
{
    const cache_window_vector_t &windows(
	engine_data.windowing->windows[burst_id]);

    // Prepare the result.
    fnr_sample_count_t num_samples((fnr_sample_count_t)samples.size());
    ratio_vector_t &val_to_end(burst_result.val_to_end);
    val_to_end.resize(num_samples, 0.0);

    vector<window_vector_t::size_type> end_window(num_caches);
    for (tm_cache_id_t cache = 0; cache < num_caches; cache++)
	end_window[cache] = windows[cache].size() - 1;

    for (int i = num_samples - 1; i >= 0; i--) {
	const mapped_repl_base_t &sample(samples[i]);
	const tm_cache_id_t cache(sample.valid()->cache);
	const mapped_access_t *valid(sample.valid());
	const mapped_access_t *end(sample.end());

	end_window[cache] = find_window(
	    windows[cache], end_window[cache], end->local_time);

        assert(valid != NULL);
        assert(end != NULL);

	const window_t &w(*windows[cache][end_window[cache]]);
	assert(w.begin_time <= end->local_time);
	assert(end->local_time <= w.end_time);

	val_to_end[i] = engine_data.get_repl_fr(
            burst_id, cache, end_window[cache],
            valid->local_time, end->local_time);
    }
}

repl_fr_result_cptr_t
repl_fr_t::calculate(sample_mapping_result_cptr_t sample_mapping,
		     repl_fr_engine_result_cptr_t engine_data,
		     mapped_sample_type_t sample_type)
{
    repl_fr_result_ptr_t result(
	new repl_fr_result_t(sample_mapping->bursts.size()));

    const tm_cache_id_t num_caches(sample_mapping->num_caches);

    sample_mapping_burst_vector_t::const_iterator it_sm(
	sample_mapping->bursts.begin());
    parser_burst_id_t burst_id(0);
    BOOST_FOREACH(repl_fr_burst_t &burst_res, result->bursts) {
	const sample_mapping_burst_t &sm_burst(*it_sm++);

	switch (sample_type) {
	case Mapped_Byte:
	    repl_fr_calculate(burst_res,
			      burst_id,
			      sm_burst.repl_byte,
			      *engine_data,
			      num_caches);
	    break;
	case Mapped_Page_Thread:
	    repl_fr_calculate(burst_res,
			      burst_id,
			      sm_burst.repl_page_thread,
			      *engine_data,
			      num_caches);
	    break;
	case Mapped_Page_Cache:
	    repl_fr_calculate(burst_res,
			      burst_id,
			      sm_burst.repl_page_cache,
			      *engine_data,
			      num_caches);
	    break;
	default:
	    assert(0);
	}

	burst_id++;
    }

    return result;
}

repl_fr_result_cptr_t
repl_fr_t::calculate() const
{
    return calculate(
        sample_mapping_t(context, sample_file, stack_depth, line_size,
			 page_size, thread_mapping, cpu_model).get(),
	engine->get(window_sample_type, sample_sample_type,
		    line_size, page_size, cache_size),
	sample_sample_type);
}

END_NAMESPACE;
END_NAMESPACE;
