/// @file analysis/fetch_ratio/repl_fr_engine.hh

// Copyright (c) 2009 Acumem AB.  All rights reserved.

#ifndef REPL_FR_ENGINE_HH
#define REPL_FR_ENGINE_HH

#include "analysis/fetch_ratio/windowing.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

// ----- repl_fr_engine_result_t ----------------------------------------------

struct repl_fr_engine_result_t {
    virtual ~repl_fr_engine_result_t() {}

    /**
     * Retrieve the replacement probability for a sample starting at
     * time1 and ending at time2. Both time specifiers must be local
     * to the specified cache.
     *
     * \note If multiple fetch ratios are required, consider using the
     * #get_repl_fr(window_vector_t::size_type, tm_cache_id_t,
     * tm_local_time_t, tm_local_time_t) function instead to avoid
     * doing potentially slow window lookups.
     */
    double get_repl_fr(parser_burst_id_t burst, tm_cache_id_t cache,
		       tm_local_time_t time1, tm_local_time_t time2) const;

    /**
     * Retrieve the replacement probability for a sample starting at
     * time1 and ending at time2. Both time specifiers must be local
     * to the specified cache.
     */
    virtual double get_repl_fr(parser_burst_id_t burst,
			       tm_cache_id_t cache,
			       window_vector_t::size_type pc2_window,
			       tm_local_time_t time1,
			       tm_local_time_t time2) const = 0;

    windowing_result_cptr_t windowing;

protected:
    repl_fr_engine_result_t(const windowing_result_cptr_t &windowing,
			    fnr_line_count_t num_cache_lines);

    const fnr_line_count_t num_cache_lines;
};

typedef shared_ptr<repl_fr_engine_result_t> repl_fr_engine_result_ptr_t;
typedef shared_ptr<const repl_fr_engine_result_t> repl_fr_engine_result_cptr_t;

// ----- repl_fr_engine_t -----------------------------------------------------

typedef shared_ptr<const class repl_fr_engine_t> repl_fr_engine_cptr_t;

class repl_fr_engine_t {
public:
    repl_fr_engine_t(const feature_context_ptr_t &context,
		     const string &sample_file,
		     fnr_stack_depth_t stack_depth,
		     thread_mapping_cptr_t thread_mapping,
		     const windowing_engine_cptr_t &windowing_engine,
		     const cpu_model_t &cpu_model);
    virtual ~repl_fr_engine_t();

    repl_fr_engine_result_cptr_t get(mapped_sample_type_t window_sample_type,
				     mapped_sample_type_t sample_sample_type,
				     fnr_line_size_t line_size,
				     fnr_line_size_t page_size,
				     fnr_cache_size_t cache_size) const;

    virtual string name() const;
    /// Name of the replacement policy (excluding parameters), e.g. LRU
    virtual string short_name() const = 0;


protected:
    virtual repl_fr_engine_result_cptr_t calculate_repl_fr(
	mapped_sample_type_t window_sample_type,
	mapped_sample_type_t sample_sample_type,
	fnr_line_size_t line_size,
	fnr_line_size_t page_size,
	fnr_cache_size_t cache_size,
        sample_mapping_result_cptr_t sample_mapping,
        windowing_result_cptr_t windowing,
        fnr_line_count_t num_cache_lines) const = 0;

    repl_fr_engine_result_cptr_t calculate() const;

    const feature_context_ptr_t context;
    const string sample_file;
    const fnr_stack_depth_t stack_depth;
    const thread_mapping_cptr_t thread_mapping;
    const windowing_engine_cptr_t windowing_engine;
    const cpu_model_t cpu_model;
};

END_NAMESPACE;
END_NAMESPACE;

#endif /* REPL_FR_ENGINE_HH */
