/// @file analysis/fetch_ratio/repl_fr_lru.cc

// Copyright (c) 2006-2009 Acumem AB.  All rights reserved.

#include "repl_fr_lru.hh"

#include <cmath>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace boost;
using namespace acumem::library;

/// This structure records the stack distance for the first slot in a histogram
/// bin, and the correction factor to get the stack distance for subsequent
/// slots in the bin.
struct r2s_t {
    double sd;   // stack distance
    double sd_1; // stack distance for the previous slot
    double c;    // correction factor, count/size/N
    uint64_t width; // width of this bin
};

typedef map<tm_local_time_t, r2s_t> r2s_map_t;
typedef vector<r2s_map_t> r2s_map_vector_t;
typedef vector<r2s_map_vector_t> cache_r2s_map_vector_t;
typedef vector<cache_r2s_map_vector_t> burst_r2s_map_vector_t;

static double
r2s(const r2s_map_t *rdist2sdist, uint64_t rdist)
{
    // We take the base value, plus a correction factor as the value
    // for r2s.
    //
    // This formula is derived to produce the same SD values as if the
    // bin's population were evenly spread out over its entire width.
    //
    // The correction factor 'c' is h/(N*width), where h is the count of
    // the bin, N is the number of samples, and width is the width of
    // the bin. This correction factor enters linearly in the
    // calculation of p_GTD, which is then integrated to give SD.
    //
    // Consider a bin with width 4, h = 100, start=x0 and N = 1000
    // t(x0)   = t(x0-1) + h(x0-1)
    // t(x0+1) = t(x0-1) + h(x0-1) + h(x0)
    // t(x0+2) = t(x0-1) + h(x0-1) + h(x0) + h(x0+1)
    // t(x0+3) = t(x0-1) + h(x0-1) + h(x0) + h(x0+1) + h(x0+2)
    //
    // The recursion formula looks like this
    // SD(x0) = SD(x0-1) + 1-t(x0)/N
    //
    // thus:
    // 1-t(x0)/N = SD(x0)-SD(x0-1)
    //
    // Expressing subsequent iterations using incremental applications of
    // the recursion formula
    //
    // SD(x0 + 1) = SD(x0) + (N-t(x0+1))/N
    //            = SD(x0) + 1*(1-t(x0)/N) - h(x0)/N
    //            = SD(x0) + 1*(SD(x0)-SD(x0-1)) - h(x0)/N
    //
    // SD(x0 + 2) = SD(x0+1) + (N-t(x0+2))/N
    //            = SD(x0) + (2N-t(x0)-h(x0)-t(x0+1)-h(x0+1))/N =
    //            = SD(x0) + (2N-2t(x0)-2h(x0)-h(x0+1))/N
    //            = SD(x0) + 2*(1-t(x0)/N) - (2h(x0)+h(x0+1))/N
    //            = SD(x0) + 2*(SD(x0)-SD(x0-1)) - (2h(x0)+h(x0+1))/N
    //
    // etc.
    //
    // Now, within a histogram bin, these h(x0 + 0) .. h(x0 + 3) are
    // considered to be equal to each other, and sum up to the
    // bin.count. Then SD(x0+i) can be written:
    //
    // SD(x0+i)   = SD(x0) + i*(1-t(x0)/N) - f(i)*h/N
    // where f(i) = 0 if i=0, 1 if i=1, 3 if i=2, ..
    //            = (i*i+i)/2
    //
    // SD(x0 + i) = SD(x0) + i*(SD(x0)-SD(x0-1)) - h*(i*i+i)/2/N
    //
    // with c = count/size/N, and h = count/size, we get the final formula:
    //
    // SD(x0 + i) = SD(x0) + i*(SD(x0)-SD(x0-1)) - c*(i*i+i)/2
    //
    // For each histogram, we record SD(x0), SD(x0-1) and c, and for
    // other purposes, the histogram bin width.

    assert(rdist2sdist->size() != 0);
    r2s_map_t::const_iterator iter = rdist2sdist->upper_bound(rdist);
    if (iter == rdist2sdist->begin()) {
	// If the query is for something before the first histogram,
	// then we return the stack distance, since the stack distance
	// grows with rate 1 when all of its samples are to the right
	// of the current point.

#ifdef DEBUGSTUFF
	cout << "query for " << rdist
	     << " which is to the left of all samples. res = " << rdist
	     << endl;
#endif

	return (double) rdist;
    }

    // move back the iterator to get to the candidate histogram bin
    iter--;
    const r2s_t &r2s = iter->second;
    uint64_t i = rdist - iter->first;

    if (i<r2s.width) {
	// If this reuse distance is within the histogram bin we
	// found, we use this formula.
	double res = r2s.sd + i*(r2s.sd - r2s.sd_1) - r2s.c*(i*i+i)/2;

#ifdef DEBUGSTUFF
	cout << "query for " << rdist
	     << " which is within bin (" << iter->first
	     << "," << r2s.width
	     << ")  i: " << i
	     << "  c: " << r2s.c
	     << "  result:" << res
	     << endl;
#endif
      	return res;
    } else {
	// No - the reuse distance is beyond the width of this
	// histogram bin, and must refer to a reuse distance for which
	// we don't have any samples.
	//
	// We use the formula above for the last item within the bin
	// and add a correction factor that depends linearly on the
	// extra distance outside the histogram bin.
	uint64_t j = r2s.width-1;
      	double res = r2s.sd + j*(r2s.sd - r2s.sd_1) - r2s.c*(j*j+j)/2 +
	    (i-j)*(r2s.sd-r2s.sd_1 - r2s.c*r2s.width);

#ifdef DEBUGSTUFF
	cout << "query for " << rdist
	     << " which is BEYOND bin (" << iter->first
	     << "," << r2s.width
	     << ")  i: " << i
	     << "  j: " << j
	     << "  c: " << r2s.c
	     << "  result:" << res
	     << endl;
#endif

	return res;
    }
}

// ----- repl_fr_result_t -----------------------------------------------------

struct repl_fr_lru_result_t : public repl_fr_engine_result_t {
    friend class repl_fr_lru_t;

    repl_fr_lru_result_t(const windowing_result_cptr_t &windowing,
			 parser_burst_id_t num_bursts,
			 tm_cache_id_t num_caches,
			 fnr_line_count_t num_cache_lines);

    double get_repl_fr(parser_burst_id_t burst,
		       tm_cache_id_t cache,
		       window_vector_t::size_type pc2_window,
		       tm_local_time_t time1,
		       tm_local_time_t time2) const;

private:
    burst_r2s_map_vector_t burst_r2s_map;
};

typedef shared_ptr<repl_fr_lru_result_t> repl_fr_lru_result_ptr_t;

repl_fr_lru_result_t::repl_fr_lru_result_t(
    const windowing_result_cptr_t &windowing,
    parser_burst_id_t num_bursts,
    tm_cache_id_t num_caches,
    fnr_line_count_t num_cache_lines)
    : repl_fr_engine_result_t(windowing, num_cache_lines),
      burst_r2s_map(num_bursts, cache_r2s_map_vector_t(num_caches))
{
}

double
repl_fr_lru_result_t::get_repl_fr(
    parser_burst_id_t burst,
    tm_cache_id_t cache,
    window_vector_t::size_type pc2_window,
    tm_local_time_t time1,
    tm_local_time_t time2) const
{
    // Don't allow time to go backwards.
    assert(time2 >= time1);

    // Always return a hit if time2 and time1 are the same access or if time2
    // is the next access after time1.
    if (time2 == time1 || time2 == time1 + 1)
        return 0.0;

    const tm_local_time_t rdist = time2 - time1 - 1;
    const double sdist = r2s(&burst_r2s_map[burst][cache][pc2_window],
			     rdist);
    const window_t &window = *windowing->windows[burst][cache][pc2_window];

    if (sdist >= num_cache_lines)
	return 1.0 - ((double)window.known_replacements / window.end_accesses);
    else
	return 0.0;
}

// ----- repl_fr_lru_t --------------------------------------------------------

static void
calculate_rdist2sdist(
    r2s_map_vector_t &window_rdist2sdist,
    const window_vector_t &windows)
{
    // This is the formula mats and zeffer came up with.
    //
    // SD(0) = 0
    // SD(k) = SD(k-1) + (n - t(k))/(n - h(k-1))
    //
    // where:
    // t(0) = 0
    // t(i) = t(i-1) + h(i-1)
    //
    // See comments in function r2s() above for calculation of SD for
    // reuse distances within a wide histogram bucket.
    //
    // For comparison, the original formula from the patent text,
    // expressed in the corresponding recursive way reads:
    //
    // SD(k) = SD(k-1) + (n-t(k))/n
    //
    // --------------------------------------------------------------
    //
    // The idea is that we temporarily neglect the histogram bin that
    // corresponds to the reuse distance - 1. That particular reuse
    // distance would correspond to a jump that coincides with the
    // trigger instruction for the reuse distance under consideration.
    //
    // Consider the following set of drawings:
    //          _
    //         / \     no potentially escaping accesses
    //    ----o---o---
    //        A   B
    //
    // This is a reuse with distance 0. Clearly, there can be no other
    // accesses between A and B. Stack distance is defined to be 0.
    //
    //   sd(0) == 0
    //
    //          _____
    //         /    _\___
    //        |    /  |
    //    ----o---+---o---
    //        A   x   B
    //
    // This is a reuse with distance 1. Similarly, there is exactly
    // one distinct memory access between A and B, namely x, and the
    // stack distance is therefore defined to be 1. Despite the
    // acquired sample's reuse distance distribution, the access x
    // cannot have reuse distance 0, since it would then collide with
    // B, which is the trigger access corresponding to A. If we don't
    // consider this condition, the projected stack distance could
    // turn out to be less than 1.
    //
    //   sd(1) == 1
    //
    //          _________
    //         /    _____\____
    //        |    / \    |
    //    ----o---+---+---o---
    //        A   x   y   B
    //
    // This is a reuse with distance 2. There are two accesses within
    // this interval, and the projected average stack distance will be
    // somewhere between 1 and 2. Now, consider access x. The sample
    // histogram offers some information on the distribution of how
    // long reuse x access has. For instance, the probability that its
    // reuse distance equals 0 is h(0)/n, and generally, the
    // probability that the reuse distance is R is given by h(R)/n.
    //
    // From the picture, it is evident that in this scenario, x's
    // reuse distance cannot be 1, since it would then coincide with
    // B, which by definition matches A. When we calculate the reuse
    // distance probability, we therefore temporarily remove the
    // contribution from samples with exactly this reuse distance.
    // This manifests itself by reducing n with the corresponding h.
    //
    //   (An alternative approach would be to distribute the
    //   contribution from h(r-1) into the other bins according to
    //   some distribution. n would stay the same, but all h()
    //   values would get a correction. Yet another way would be to
    //   distribute the h(r-1) into the h(r-2) and h(r) buckets only.)
    //
    // Anyway, in this case, the probability that x has reuse distance
    // 0 is h(0)/(n-h(1)), probability for reuse distance 1 is 0 per
    // definition, and it is larger than 1 with probability
    // 1-h(0)/(n-h(1)).
    //
    // This corresponds directly to stack distance: the probability
    // that stack distance is 1 (meaning that x and y are the same
    // memory position) equals the probability that reuse distance for
    // x is zero, namely h(0)/(n-h(1)). The probability that the stack
    // distance is 2 equals the probability that the reuse distance
    // for x is larger than 1, namely 1-h(0)/(n-h(1)). The expected
    // value for this stack distance is therefore:
    //
    //   sd(2) = 1 * h(0)/(n-h(1)) + 2 * (1-h(0)/(n-h(1)))
    //         = 1 + 1-h(0)/(n-h(1))
    //         = sd(1) + 1-h(0)/(n-h(1))
    //
    //
    //          _____________
    //         /    _________\___
    //        |    / \   \    |
    //    ----o---+---+---+---o---
    //        A   x   y   z   B
    //
    // Now, for sd(3), we have a case where the stack distance should
    // be somewhere between 1 and 3.
    //
    // Considering only y and z for a while. These two accesses
    // contribute to the stack distance with a value between 1 and 2,
    // expected to have average value sd(2), as per previous calculation.
    //
    // x can either have reuse distance 0 or 1, in which case the
    // stack distance is equal to sd(2), or a reuse distance greater
    // than 2, in which case the stack distance will have a value
    // sd(2) + 1. Looking at probabilities, the probabilities for
    // different valies of x's reuse distance are as follows:
    //
    //   reuse dist. =r     p(r)
    //   -------------------------------------------
    //   0                  h(0)/(n-h(2))
    //   1                  h(1)/(n-h(2))
    //   2                  0 (forbidden)
    //   3..inf             1 - (h(0)+h(1))/(n-h(2))
    //
    // Expressing this as an expected value for sd(3):
    //
    //   sd(3) = sd(2) * (p(0)+p(1)) + (sd(2)+1) * (1-(p(0)+p(1)))
    //         = sd(2)(p(0)+p(1)) + sd(2)-sd(2)*(p(0)+p(1)) + 1-(p(0)+p(1))
    //         = sd(2) + 1-(p(0)+p(1))
    //         = sd(2) + 1-(h(0)+h(1))/(n-h(2))
    //         = sd(2) + p(3..inf)
    //
    // We conjecture the following general recursive formula for
    // calculating sd(r):
    //
    //   sd(0) = 0
    //   sd(r) = sd(r-1) + 1 - sum{k=0..r-2; h(k)}/(n-h(r-1))
    //
    // And we can also express it as an explicit non-recursive
    // formula:
    //
    //   sd(r) = {r=0: 0
    //            r>0: sum{s:1..r;  1-sum{k=0..s-2; h(k)}/(n-h(s-1))}
    //
    // which for r=3 expands to
    //
    //   sd(3) = sum{s:1..3; 1-sum{k=0..s-2; h(k)}/(n-h(s-1))}
    //         = 1-sum{k=0..-1; h(k)/(n-h(0))} +
    //           1-sum{k=0..0;  h(k)/(n-h(1))} +
    //           1-sum{k=0..1;  h(k)/(n-h(2))}
    //         = 1  +  1-h(0)/(n-h(1))  +  1-(h(0)+h(1))/(n-h(2))
    //
    // whose parts we recognize from the discussion above

    /// @todo The reuse distance to stack distance functions for the windows do
    /// not depend on the cache size, call stack depth or whether we are
    /// considering cold misses or not. It could be cached in the windows
    /// between calculations where only these parameters differ.

    // Calculate reuse distance to stack distance functions for the windows.
    BOOST_FOREACH(const window_ptr_t &window, windows) {
	window_t *w = &*window;
	double n = w->end_accesses;
	bin_vector_t *bins = &*w->rdist_hist;

	window_rdist2sdist.push_back(r2s_map_t());
	r2s_map_t *rdist2sdist = &window_rdist2sdist.back();

	bin_vector_t::iterator i = bins->begin();
	bin_vector_t::iterator e = bins->end();

#ifdef DEBUGSTUFF
	for (; i != e; i++) {
	    cout << "bin: (" << i->start
		 << "," << i->size
		 << "): " << i->count << endl;
	}
	i = bins->begin();
#endif

	if (i != e) {
	    double   t_i_1  = 0.0;	        // t(0) = 0
	    uint64_t h_i_1  = i->count;	// h(0) = ...

	    r2s_t entry;
	    entry.sd    = (double) i->start;
	    entry.sd_1  = (double) i->start-1;
	    entry.c     = (double) h_i_1 / (double) i->size / n;
	    entry.width = i->size;
	    (*rdist2sdist)[i->start] = entry;

#ifdef DEBUGSTUFF
	    cout << "entry: r: " << i->start
		 << "  sd: " << entry.sd
		 << "  h:" << h_i_1
		 << "  size: " << i->size
		 << "  c: " << entry.c
		 << endl;
#endif
	    i++;

	    for (; i != e; i++) {
		double t_i  = t_i_1 + h_i_1;	// t(i) = t(i-1) + h(i-1)
		double sd_i_1 = r2s(rdist2sdist, i->start - 1);

#define ORIGINAL_LRU_FORMULA
#ifdef ORIGINAL_LRU_FORMULA
		// Erik Hagerstens original formula
		double sd_i = sd_i_1 + (n-t_i) / n;
#else
		// Mats & zeffers modified formula
		double sd_i = sd_i_1 + (n-t_i) / (n-h_i_1);
#endif

		h_i_1 = i->count;			// h(k-1) = ...
		t_i_1 = t_i;

		// This are the correction factor parameters
		entry.sd    = sd_i;
		entry.sd_1  = sd_i_1;
		entry.c     = (double) h_i_1 / (double) i->size / n;
		entry.width = i->size;
		(*rdist2sdist)[i->start] = entry;

#ifdef DEBUGSTUFF
		cout << "entry: r: " << i->start
		     << "  sd: " << sd_i
		     << "  h:" << h_i_1
		     << "  size: " << i->size
		     << "  c: " << entry.c
		     << endl;
#endif
	    }
	}
    }
}

repl_fr_engine_result_cptr_t
repl_fr_lru_t::calculate(
    const sample_mapping_result_cptr_t &sample_mapping,
    const windowing_result_cptr_t &windowing,
    fnr_line_count_t num_cache_lines)
{
    const parser_burst_id_t num_bursts(sample_mapping->bursts.size());
    const tm_cache_id_t num_caches(sample_mapping->num_caches);
    repl_fr_lru_result_ptr_t result(
	new repl_fr_lru_result_t(windowing,
				 num_bursts, num_caches,
				 num_cache_lines));

    for (parser_burst_id_t burst = 0; burst < num_bursts; burst++) {
	cache_r2s_map_vector_t &cache_r2s_map(result->burst_r2s_map[burst]);
	const cache_window_vector_t &cache_windows(windowing->windows[burst]);
	for (tm_cache_id_t cache = 0; cache < num_caches; cache++) {
	    /// Translation tables from reuse distance to stack distance for each
	    /// window.
	    const window_vector_t &windows(cache_windows[cache]);
	    r2s_map_vector_t &window_rdist2sdist(cache_r2s_map[cache]);
	    window_rdist2sdist.reserve(windows.size());

	    calculate_rdist2sdist(cache_r2s_map[cache], windows);
	}
    }

    return result;
}

repl_fr_lru_t::repl_fr_lru_t(const feature_context_ptr_t &context,
                             const string &sample_file,
                             fnr_stack_depth_t stack_depth,
                             thread_mapping_cptr_t thread_mapping,
			     const windowing_engine_cptr_t &windowing,
			     const cpu_model_t &cpu_model)
    : repl_fr_engine_t(context, sample_file, stack_depth, thread_mapping,
		       windowing, cpu_model)
{
}


string
repl_fr_lru_t::name() const
{
    return "repl_fr_lru_t(" + repl_fr_engine_t::name() + ")";
}

string
repl_fr_lru_t::short_name() const
{
    return "lru";
}

repl_fr_engine_result_cptr_t
repl_fr_lru_t::calculate_repl_fr(
    mapped_sample_type_t window_sample_type,
    mapped_sample_type_t sample_sample_type,
    fnr_line_size_t line_size,
    fnr_line_size_t page_size,
    fnr_cache_size_t cache_size,
    sample_mapping_result_cptr_t sample_mapping,
    windowing_result_cptr_t windowing,
    fnr_line_count_t num_cache_lines) const
{
    return calculate(sample_mapping, windowing, num_cache_lines);
}

END_NAMESPACE;
END_NAMESPACE;
