/// @file analysis/fetch_ratio/repl_fr_random.cc

// Copyright (c) 2006-2009 Acumem AB.  All rights reserved.

#include "analysis/fetch_ratio/repl_fr_random.hh"

#include <cmath>
#include <iostream>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace boost;
using namespace acumem::library;

static double
f(fnr_line_count_t num_cache_lines, double exponent)
{
    return 1.0 - pow(1.0 - (1.0 / num_cache_lines), exponent);
}

// ----- repl_fr_random_result_t ----------------------------------------------

struct repl_fr_random_result_t : public repl_fr_engine_result_t {
    repl_fr_random_result_t(const windowing_result_cptr_t &windowing,
			    const window_rr_result_cptr_t &window_rr,
			    fnr_line_count_t num_cache_lines);

    double get_repl_fr(parser_burst_id_t burst,
		       tm_cache_id_t cache,
		       window_vector_t::size_type pc2_window,
		       tm_local_time_t time1,
		       tm_local_time_t time2) const;

private:
    const window_rr_result_cptr_t window_rr;
};

repl_fr_random_result_t::repl_fr_random_result_t(
    const windowing_result_cptr_t &windowing,
    const window_rr_result_cptr_t &window_rr,
    fnr_line_count_t num_cache_lines)
    : repl_fr_engine_result_t(windowing, num_cache_lines),
      window_rr(window_rr)
{
}


double
repl_fr_random_result_t::get_repl_fr(
    parser_burst_id_t burst,
    tm_cache_id_t cache,
    window_vector_t::size_type pc2_window,
    tm_local_time_t time1,
    tm_local_time_t time2) const
{
    // Don't allow time to go backwards.
    assert(time2 >= time1);

    // Always return a hit if time2 and time1 are the same access or if time2
    // is the next access after time1.
    if (time2 == time1 || time2 == time1 + 1)
        return 0.0;

    const window_vector_t &windows =
	windowing->windows[burst][cache];
    const window_rr_vector_t &window_repl =
	window_rr->replacement_ratio[burst][cache];

    tm_local_time_t rdist = time2 - time1 - 1;
    if (rdist > 0) {
	// Sum the number of replacements in the windows that
	// the sample spans over.
	double total_fetches = 0.0;
	tm_local_time_t total_overlap = 0;
	tm_local_time_t overlap_end = time2;

	window_vector_t::size_type local_window = pc2_window;
	do {
	    const window_t &w = *windows[local_window];
	    tm_local_time_t overlap_begin = (time1 >= w.begin_time ?
					time1 :
					w.begin_time - 1);
	    tm_local_time_t overlap = overlap_end - overlap_begin -
		(overlap_end == time2 ? 1 : 0);
	    total_fetches += window_repl[local_window] * overlap;
	    total_overlap += overlap;
	    overlap_end = overlap_begin;
	} while (time1 < windows[local_window--]->begin_time);
	assert(total_overlap == rdist);

	// Calculate the replacement fetch ratio based on the number of
	// fetches between the sample start and end.
	return f(num_cache_lines, total_fetches);
    }
    else
	// Reuse distance 0, can't be a fetch.
	return 0;
}

// ----- repl_fr_random_t -------------------------------------------------

repl_fr_engine_result_cptr_t
repl_fr_random_t::calculate(
    const sample_mapping_result_cptr_t &sample_mapping,
    const windowing_result_cptr_t &windowing,
    const window_rr_result_cptr_t &window_rr,
    fnr_line_count_t num_cache_lines)
{
    repl_fr_engine_result_ptr_t result(
	new repl_fr_random_result_t(windowing, window_rr, num_cache_lines));

    return result;
}

repl_fr_random_t::repl_fr_random_t(const feature_context_ptr_t &context,
                                   const string &sample_file,
                                   fnr_stack_depth_t stack_depth,
                                   thread_mapping_cptr_t thread_mapping,
				   const windowing_engine_cptr_t &windowing,
				   const cpu_model_t &cpu_model)
    : repl_fr_engine_t(context, sample_file, stack_depth, thread_mapping,
		       windowing, cpu_model)
{
}

string
repl_fr_random_t::name() const
{
    return "repl_fr_random_t(" + repl_fr_engine_t::name() + ")";
}

string
repl_fr_random_t::short_name() const
{
    return "random";
}

repl_fr_engine_result_cptr_t
repl_fr_random_t::calculate_repl_fr(
    mapped_sample_type_t window_sample_type,
    mapped_sample_type_t sample_sample_type,
    fnr_line_size_t line_size,
    fnr_line_size_t page_size,
    fnr_cache_size_t cache_size,
    sample_mapping_result_cptr_t sample_mapping,
    windowing_result_cptr_t windowing,
    fnr_line_count_t num_cache_lines) const
{
    // Calculate replacement ratios for the windows.
    window_rr_result_cptr_t window_rr =
        window_rr_t(context, sample_file, stack_depth, line_size, page_size,
                    thread_mapping, cpu_model, window_sample_type,
		    cache_size, windowing_engine).get();

    return calculate(sample_mapping, windowing, window_rr, num_cache_lines);
}

END_NAMESPACE;
END_NAMESPACE;
