/// @file analysis/fetch_ratio/window_rr.cc

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#include "window_rr.hh"
#include "windowing.hh"
#include <boost/lexical_cast.hpp>
#include <cmath>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace boost;
using namespace acumem::library;

// ----- window_rr_t ----------------------------------------------------------

// Whether to use the pow function for each data point (OLD), or to
// use the pow() much less often and use a table of precomputed powers
// to accumulate corresponding values (NEW). There are small round-off
// differences between these methods, but all tests pass.

//#define OLD_CALCULATION
#define NEW_CALCULATION

// Enable comparing the OLD and NEW version
//#define DEBUGSTUFF

static double
sum_rightside(const bin_vector_t &bins,
	      fnr_line_count_t num_cache_lines, double r)
{
    double base = 1.0 - (1.0 / num_cache_lines);

#ifdef OLD_CALCULATION
    double res_old = 0;
#endif

#ifdef NEW_CALCULATION
    double factor = pow(base, r / 2.0);

    // Precompute a table of powers in this base:
    //
    // factor[0] = base^(r/2)
    // factor[1] = base^(r)
    // factor[2] = base^(2r)
    // factor[3] = base^(4r)
    // ...

    double factors[32];
    for (int fi = 0; fi < 32; fi++) {
	factors[fi] = factor;
	factor *= factor;
    }

    uint64_t last_center_2 = 0;
    uint64_t last_diff_2 = 0;
    double val = 1;
    double power = 1;
    double res = 0;
#endif

    BOOST_FOREACH(const bin_t &b, bins) {

#ifdef OLD_CALCULATION
	double c2 = b.start + (b.size - 1) / 2.0;
	double p = pow(base, r * c2);
	res_old += b.count * (1.0 - p);
#endif

#ifdef NEW_CALCULATION

	// Depending whether the width of the bin is even or odd, the
	// center could be a non integer (.5). We avoid using floating
	// point values by multiplying the center with two, and then
	// compensate for this by letting the first slot in factors be
	// base^(r/2).
	uint64_t center_2 = 2 * b.start + (b.size - 1);
	uint64_t diff_2 = center_2 - last_center_2;

	last_center_2 = center_2;

	// First check if the previous bin interval can be reused
	if (diff_2 != last_diff_2) {

	    // Apparently not.
	    last_diff_2 = diff_2;

	    // See if the lookup table can be used for this interval.
	    if (diff_2 > (uint64_t) 1<<31UL) {

		// No, the diff is outside the scope of the lookup
		// table. Fall back to use pow() directly.
		val = pow(base, r * diff_2 / 2.0);

	    } else {

		// Yes, the lookup table can be used. Pick those
		// factors that correspond to set bits in the diff_2
		// variable.

		val = 1;
		for (int f = 0; f < 32 && diff_2 != 0; f++, diff_2 /= 2)
		    if (diff_2 & 1)
			val *= factors[f];
	    }
	}

	power *= val;
	res += b.count * (1.0 - power);
#endif
    }

#ifdef DEBUGSTUFF
    cerr << "r: " << r << "  new: " << res
	 << " old: " << res_old << "  diff: " << (res_old - res) << endl;
#endif

#if defined(OLD_CALCULATION) && !defined(NEW_CALCULATION)
    return res_old;
#else
    return res;
#endif
}

window_rr_result_cptr_t
window_rr_t::calculate(
    fnr_line_count_t num_cache_lines,
    windowing_result_cptr_t windowing)
{
    window_rr_result_ptr_t result(
	new window_rr_result_t(windowing->windows.size()));

    window_burst_vector_t::const_iterator it_win(windowing->windows.begin());
    BOOST_FOREACH(window_rr_cache_vector_t &wrr, result->replacement_ratio) {
	const cache_window_vector_t &cwv(*it_win++);
	const tm_cache_id_t num_caches((tm_cache_id_t)cwv.size());
	wrr.resize(num_caches);

	for (tm_cache_id_t i = 0; i < num_caches; i++) {
	    window_rr_vector_t &cache_result(wrr[i]);
	    const window_vector_t &cache_windows(cwv[i]);
	    cache_result.reserve(cache_windows.size());

	    // Calculate the total replacement ratios of the windows.
	    BOOST_FOREACH(const window_ptr_t &window, cache_windows) {
		window_t &w = *window;
		bin_vector_t &bins = *w.rdist_hist;

		double rmid = 0.5;
		double rhigh = 1.0;
		double rlow = 0.0;

		if (w.end_accesses > 0) {
		    double num_accesses = w.end_accesses;
		    double known_rr = w.known_replacements / num_accesses;

		    while ((rhigh - rlow) > 0.000003) {
			double rs = sum_rightside(bins, num_cache_lines, rmid);
			double ls = (num_accesses * (rmid - known_rr)) / (1.0 - known_rr);

			if (rs < ls)
			    rhigh = rmid;
			else
			    rlow = rmid;

			rmid = (rhigh + rlow) / 2.0;
		    }

		    cache_result.push_back(rmid);
		}
		else
		    cache_result.push_back(0.0);
	    }
	}
    }

    return result;
}

window_rr_t::window_rr_t(feature_context_ptr_t context,
                         const string &sample_file,
                         fnr_stack_depth_t stack_depth,
                         fnr_line_size_t line_size,
                         fnr_line_size_t page_size,
                         thread_mapping_cptr_t thread_mapping,
			 const cpu_model_t &cpu_model,
			 mapped_sample_type_t sample_type,
                         fnr_cache_size_t cache_size,
			 const windowing_engine_cptr_t &windowing)
    : feature_t<window_rr_result_cptr_t>(context),
      context(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      sample_type(sample_type),
      cache_size(cache_size),
      windowing(windowing)
{
}

window_rr_result_cptr_t
window_rr_t::get() const
{
    return get_data();
}

string
window_rr_t::name() const
{
    return "window_rr_t(" +
        sample_file + "," +
	lexical_cast<string>(stack_depth) + "," +
	lexical_cast<string>(line_size) + "," +
	lexical_cast<string>(page_size) + "," +
        thread_mapping->name() + "," +
	cpu_model.name() + "," +
	lexical_cast<string>(sample_type) + "," +
	lexical_cast<string>(cache_size) + "," +
	windowing->name() + ")";
}

window_rr_result_cptr_t
window_rr_t::calculate() const
{
    windowing_result_cptr_t windowing_res(
        windowing_t(context, sample_file, stack_depth, line_size, page_size,
                    thread_mapping, cpu_model, sample_type,
		    windowing).get());

    // Calculate the number of cache lines.
    fnr_line_count_t num_cache_lines;

    switch (sample_type) {
    case Mapped_Line:
	num_cache_lines = cache_size / line_size;
	break;
    case Mapped_Page_Thread:
    case Mapped_Page_Cache:
	num_cache_lines = cache_size / page_size;
	break;
    default:
	abort();
    }

    return calculate(num_cache_lines, windowing_res);
}

END_NAMESPACE;
END_NAMESPACE;
