/// @file analysis/fetch_ratio/windowing.cc

// Copyright (c) 2006-2009 Acumem AB.  All rights reserved.

#include "windowing.hh"
#include "analysis/thread_mapping/sample_mapping.hh"
#include "analysis/misc/parser.hh"
#include "analysis/feature_base/exceptions.hh"
#include <boost/lexical_cast.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace boost;
using namespace acumem::library;

static fnr_time_t
reuse_distance(const mapped_repl_base_t *s)
{
    return s->rdist();
}

template <class SampleType>
static void
calculate_windowing(cache_window_vector_t &cwv,
		    const time_mapping_burst_t &time_mapping,
		    const mapped_cold_vector_t &cold_samples,
		    const vector<SampleType> &repl_samples,
                    const cpu_model_t &cpu_model,
		    const windowing_engine_t &engine)
{
    typedef vector<mapped_cold_cptr_vector_t> cache_cold_vector_t;
    typedef vector<mapped_repl_base_cptr_vector_t> cache_repl_vector_t;

    const tm_cache_id_t num_caches((tm_cache_id_t)cwv.size());
    cache_cold_vector_t cache_cold(num_caches);
    cache_repl_vector_t cache_repl(num_caches);

    BOOST_FOREACH(const mapped_cold_t &sample, cold_samples)
	cache_cold[sample.end.cache].push_back(&sample);

    BOOST_FOREACH(const mapped_repl_base_t &sample, repl_samples)
	cache_repl[sample.end()->cache].push_back(&sample);

    histogram_t< exp_histogram_t<4> > h;

    for (tm_cache_id_t cache = 0; cache < num_caches; cache++) {
	if (!cache_cold[cache].empty() || !cache_repl[cache].empty())
	    engine.do_windowing(
		cwv[cache],
		time_mapping,
		cache_cold[cache], cache_repl[cache],
		cache);

        cold_iterator_t cold_it(cache_cold[cache].begin());
        cold_iterator_t cold_end(cache_cold[cache].end());
        repl_iterator_t repl_it(cache_repl[cache].begin());
        repl_iterator_t repl_end(cache_repl[cache].end());

        BOOST_FOREACH(window_ptr_t &wp, cwv[cache]) {
            window_t *w = &*wp;

            repl_iterator_t repl_it_begin(repl_it);

            tm_local_time_t window_begin_time = w->begin_time;
            tm_local_time_t window_end_time = w->end_time;

            w->end_accesses = 0;
            w->known_replacements = 0;

            while (cold_it != cache_cold[cache].end() &&
                   (*cold_it)->end.local_time <= window_end_time) {
                assert((*cold_it)->end.local_time >= window_begin_time);
                if (cpu_model.makes_valid((*cold_it)->end))
                    w->known_replacements++;
                w->end_accesses++;
                cold_it++;
            }

            while (repl_it != cache_repl[cache].end() &&
                   (*repl_it)->end()->local_time <= window_end_time) {
                assert((*repl_it)->end()->local_time >= window_begin_time);
                w->end_accesses++;
                repl_it++;
            }

            w->rdist_hist = h.get_histogram(
                unknown_replacement_iterator_t(repl_it_begin, repl_it),
                unknown_replacement_iterator_t(repl_it, repl_it),
                reuse_distance);
        }
    }
}

windowing_result_cptr_t
windowing_t::calculate(feature_context_ptr_t context,
		       sample_mapping_result_cptr_t sample_mapping,
		       time_mapping_result_cptr_t time_mapping,
		       fnr_line_size_t line_size,
		       fnr_line_size_t page_size,
                       const cpu_model_t &cpu_model,
		       mapped_sample_type_t sample_type,
		       windowing_engine_cptr_t windowing_engine)
{
    windowing_result_ptr_t result(
	new windowing_result_t(sample_mapping->bursts.size(),
			       sample_mapping->num_caches));
    window_burst_vector_t &wbv(result->windows);

    switch (sample_type) {
    case Mapped_Byte:
	if (!sample_mapping->have_byte_samples())
	    throw feature_byte_samples_missing_exception_t(
		"windowing",
		"No byte sized samples availables, can't do windowing.");
	break;
    case Mapped_Line:
	if (!sample_mapping->have_line_samples())
	    throw feature_line_samples_missing_exception_t(
		"windowing",
		"No line sized samples availables, can't do windowing.",
		line_size);
	break;
    case Mapped_Page_Cache:
    case Mapped_Page_Thread:
	if (!sample_mapping->have_page_samples())
	    throw feature_page_samples_missing_exception_t(
		"windowing",
		"No page sized samples availables, can't do windowing.");
	break;
    default:
	assert(0);
    }

    sample_mapping_burst_vector_t::const_iterator it_smb(
	sample_mapping->bursts.begin());
    time_mapping_burst_vector_t::const_iterator it_time(
	time_mapping->bursts.begin());
    BOOST_FOREACH(cache_window_vector_t &cwv, wbv) {
	const sample_mapping_burst_t &smb(*it_smb++);
	const time_mapping_burst_t &time(*it_time++);
	switch (sample_type) {
	case Mapped_Byte:
	    calculate_windowing(
		cwv,
		time,
		smb.cold_byte, smb.repl_byte,
		cpu_model, *windowing_engine);
	    break;
	case Mapped_Line:
	    calculate_windowing(
		cwv,
		time,
		smb.cold_line, smb.repl_line,
		cpu_model, *windowing_engine);
	    break;
	case Mapped_Page_Thread:
	    calculate_windowing(
		cwv,
		time,
		smb.cold_page, smb.repl_page_thread,
		cpu_model, *windowing_engine);
	    break;
	case Mapped_Page_Cache:
	    calculate_windowing(
		cwv,
		time,
		smb.cold_page, smb.repl_page_cache,
		cpu_model, *windowing_engine);
	    break;
	}
    }

    return result;
}

windowing_t::windowing_t(
    feature_context_ptr_t context,
    const string &sample_file,
    fnr_stack_depth_t stack_depth,
    fnr_line_size_t line_size,
    fnr_line_size_t page_size,
    thread_mapping_cptr_t thread_mapping,
    const cpu_model_t &cpu_model,
    mapped_sample_type_t sample_type,
    windowing_engine_cptr_t windowing_engine)
    : feature_t<windowing_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      sample_type(sample_type),
      windowing_engine(windowing_engine)
{
}

windowing_result_cptr_t
windowing_t::get() const
{
    return get_data();
}

string
windowing_t::name() const
{
    return "windowing_t(" +
        sample_file + "," +
        lexical_cast<string>(stack_depth) + "," +
        lexical_cast<string>(line_size) + "," +
        lexical_cast<string>(page_size) + "," +
        thread_mapping->name() + "," +
	cpu_model.name() + "," +
	lexical_cast<string>(sample_type) + "," +
	windowing_engine->name() + ")";
}

windowing_result_cptr_t
windowing_t::calculate() const
{
    return calculate(context,
		     sample_mapping_t(context, sample_file, stack_depth,
				      line_size, page_size, thread_mapping,
				      cpu_model).get(),
		     time_mapping_t(context, sample_file,
				    thread_mapping).get(),
		     line_size,
		     page_size,
                     cpu_model,
		     sample_type,
		     windowing_engine);
}

END_NAMESPACE;
END_NAMESPACE;
