/// @file analysis/fetch_ratio/windowing.hh

// Copyright (c) 2006-2009 Acumem AB.  All rights reserved.

#ifndef WINDOWING_HH
#define WINDOWING_HH

#include "analysis/thread_mapping/sample_mapping.hh"
#include "analysis/call_stack/stack_pc_map.hh"
#include "analysis/misc/parser.hh"
#include "analysis/misc/histogram.hh"
#include "analysis/feature_base/feature_base.hh"
#include "library/fenris/fenris.hh"

#include <boost/iterator/filter_iterator.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace acumem::library;

using boost::shared_ptr;

// ----- Convenience types ----------------------------------------------------

typedef mapped_cold_cptr_vector_t::const_iterator cold_iterator_t;

typedef mapped_repl_base_cptr_vector_t::const_iterator repl_iterator_t;

// ----- Filtered iterators ---------------------------------------------------

struct is_unknown_replacement_t {
    bool operator()(const mapped_repl_base_t *sample) const {
        return
            !sample->known_nonreplacement() &&
            !sample->known_replacement();
    }
};

/// Iterates over replacement samples that are not known to either cause a
/// replacement or not cause a replacement.
typedef boost::filter_iterator<
    is_unknown_replacement_t,
    repl_iterator_t>
unknown_replacement_iterator_t;

// ----- window_t -------------------------------------------------------------

/// A replacement miss ratio calculation window.
struct window_t {
    // Inclusive begin time of the window. Should be the end_time + 1 of the
    // previous window, or 0 if this is the first window.
    tm_local_time_t begin_time;

    // Inclusive end time of the window. Should be the begin_time - 1 of the
    // next window, or the acquisition end time if this is the last window.
    tm_local_time_t end_time;

    /// Number of end accesses in the window. These can be divided into three
    /// categories:
    ///
    /// * Known replacements. End accesses we know will cause a replacement,
    ///   e.g., cold accesses and accesses to lines that have been flushed or
    ///   invalidated. The number of such accesses is kept in
    ///   \known_replacements.
    ///
    /// * Known non-replacements. End accesses we know will not cause a
    ///   replacement, e.g., cache line flushes. We don't need to keep track of
    ///   these.
    ///
    /// * Unknown replacements. These are the end accesses that we need to run
    ///   through the replacement ratio calculation in order to know the
    ///   probability that they cause a replacement. We keep their reuse
    ///   distances in \rdist_hist.
    fnr_sample_count_t end_accesses;

    /// Number of end accesses that we know will cause a replacement.
    fnr_sample_count_t known_replacements;

    /// Reuse distance histogram for samples to run through the replacement
    /// ratio calculation.
    bin_vector_ptr_t rdist_hist;
};

typedef shared_ptr<window_t> window_ptr_t;
typedef vector<window_ptr_t> window_vector_t;
typedef vector<window_vector_t> cache_window_vector_t;
typedef vector<cache_window_vector_t> window_burst_vector_t;

// ----- windowing_result_t ---------------------------------------------------

/// Windowing result.
struct windowing_result_t {
    windowing_result_t(parser_burst_id_t num_bursts,
		       tm_cache_id_t num_caches)
	: windows(num_bursts, cache_window_vector_t(num_caches)) {}

    /// Windows.
    window_burst_vector_t windows;
};

typedef shared_ptr<windowing_result_t> windowing_result_ptr_t;
typedef shared_ptr<const windowing_result_t> windowing_result_cptr_t;


// ----- windowing_engine_t ---------------------------------------------------

class windowing_engine_t {
public:
    virtual ~windowing_engine_t() {};

    /// Creates window objects and fills in begin_time and end_time. Leaves the
    /// other fields in the windows untouched.
    virtual void do_windowing(
	window_vector_t &windows,
	const time_mapping_burst_t &tm_burst,
	const mapped_cold_cptr_vector_t &cold_samples,
	const mapped_repl_base_cptr_vector_t &repl_samples,
	const tm_cache_id_t cache_id) const = 0;

    virtual string name() const = 0;
};

typedef shared_ptr<windowing_engine_t> windowing_engine_ptr_t;
typedef shared_ptr<const windowing_engine_t> windowing_engine_cptr_t;

// ----- windowing_t ----------------------------------------------------------

class windowing_t : feature_t<windowing_result_cptr_t> {
public:
    static windowing_result_cptr_t calculate(
	feature_context_ptr_t context,
	sample_mapping_result_cptr_t sample_mapping,
	time_mapping_result_cptr_t time_mapping,
	fnr_line_size_t line_size,
	fnr_line_size_t page_size,
        const cpu_model_t &cpu_model,
	mapped_sample_type_t sample_type,
	windowing_engine_cptr_t windowing_engine);

    windowing_t(feature_context_ptr_t context,
		const string &sample_file,
		fnr_stack_depth_t stack_depth,
		fnr_line_size_t line_size,
		fnr_line_size_t page_size,
		thread_mapping_cptr_t thread_mapping,
		const cpu_model_t &cpu_model,
		mapped_sample_type_t sample_type,
		windowing_engine_cptr_t windowing_engine);

    windowing_result_cptr_t get() const;

protected:
    windowing_result_cptr_t calculate() const;
    string name() const;

    const string sample_file;
    const fnr_stack_depth_t stack_depth;
    const fnr_line_size_t line_size;
    const fnr_line_size_t page_size;
    const thread_mapping_cptr_t thread_mapping;
    const cpu_model_t cpu_model;
    const mapped_sample_type_t sample_type;
    const windowing_engine_cptr_t windowing_engine;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
