/// @file analysis/fetch_ratio/windowing_adaptive.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include "windowing_adaptive.hh"
#include <cmath>
#include <boost/lexical_cast.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

//#define DEBUGOUT

#define MIN(a, b) ((a) < (b) ? (a) : (b))

using namespace std;
using namespace boost;
using namespace acumem::library;

typedef histogram_t< exp_histogram_t<4> > histogram_type_t;

/// Returns the reuse distance of a sample.
static fnr_time_t
reuse_distance(const mapped_repl_base_t *s)
{
    return s->rdist();
}

/**
 * Sums all bins in a histogram.
 */
static fnr_sample_count_t
histogram_size(const bin_vector_ptr_t &h)
{
    fnr_sample_count_t c(0);
    BOOST_FOREACH(const bin_t &b, *h)
	c += b.count;

    return c;
}

/**
 * Calculate the distance between two histograms. The function makes
 * the following assumptions:
 *  * count is always positive
 *  * bins are equally distributed, but not all bins have to
 *    be populated in the vectors.
 *
 * \return A value in the [0, 1] range denoting the distance between
 *         h1 and h2.
 */
static double
histogram_distance(const bin_vector_ptr_t &h1,
		   const bin_vector_ptr_t &h2)
{
    const fnr_sample_count_t h1_size(histogram_size(h1)),
	h2_size(histogram_size(h2));

    bin_vector_t::const_iterator
	it_h1(h1->begin()), it_h1e(h1->end()),
	it_h2(h2->begin()), it_h2e(h2->end());

    // Sum using two integers instead of a single double
    // value. Improves numerical stability and reduces the number of
    // divisions.
    int64_t h1_sum(0), h2_sum(0);

    while (it_h1 != it_h1e ||
	   it_h2 != it_h2e) {

	if (it_h1 == it_h1e) {
	    h2_sum += it_h2->count;
	    it_h2++;
	} else if (it_h2 == it_h2e) {
	    h1_sum += it_h1->count;
	    it_h1++;
	} else {
	    if (it_h1->start < it_h2->start) {
		h1_sum += it_h1->count;
		it_h1++;
	    } else if (it_h2->start < it_h1->start) {
		h2_sum += it_h2->count;
		it_h2++;
	    } else {
		assert(it_h1->size == it_h2->size);
		double rel1((double)it_h1->count / h1_size),
		    rel2((double)it_h2->count / h2_size);
		// The following is equivalent with pushing fabs(rel1 - rel2)
		if (rel1 > rel2) {
		    h1_sum += it_h1->count;
		    h2_sum -= it_h2->count;
		} else {
		    h1_sum -= it_h1->count;
		    h2_sum += it_h2->count;
		}
		it_h1++;
		it_h2++;
	    }
	}
    }

    return fabs((double)h1_sum / h1_size + (double)h2_sum / h2_size) / 2;
}

/**
 * Creates a sorted vector of pc2 times from a cold sample vector and
 * a replacement sample vector.
 */
static void
populate_pc2_times(
    fnr_time_vector_t &pc2_times,
    const mapped_cold_cptr_vector_t &cold_samples,
    const mapped_repl_base_cptr_vector_t &repl_samples)
{
    // Create a sorted vector of pc2 times of both replacement and cold
    // samples.
    repl_iterator_t
        rsi(repl_samples.begin()),
        rse(repl_samples.end());
    cold_iterator_t
        csi(cold_samples.begin()),
	cse(cold_samples.end());
    while (rsi != rse || csi != cse)
        if (rsi == rse)
            pc2_times.push_back((*(csi++))->end.local_time);
        else if (csi == cse)
            pc2_times.push_back((*(rsi++))->end()->local_time);
        else {
            fnr_time_t rt((*rsi)->end()->local_time);
            fnr_time_t ct((*csi)->end.local_time);
            if (rt < ct) {
                pc2_times.push_back(rt);
                rsi++;
            }
            else {
                pc2_times.push_back(ct);
                csi++;
            }
        }

}

/**
 * Finds the iterator pointing to the first sample with a local end
 * time that is greater than the specified end time.
 *
 * \param it_begin Iterator pointing to the start of the search
 * \param it_end Iterator pointing to the end of the container.
 * \param end_time End time to look for.
 * \return Iterator pointing to the first element with a local
 *         end time greater than end_time, it_end if no such element
 *         exists.
 */
static repl_iterator_t
find_end_sample(const repl_iterator_t &it_begin,
                const repl_iterator_t &it_end,
		const fnr_time_t end_time)
{
    repl_iterator_t it(it_begin);
    while (it != it_end &&
	   (*it)->end()->local_time <= end_time)
	it++;

    return it;
}

/**
 * Try to grow the current window represented by its seed. Returns the
 * index of the end time in the pc2_times vector.
 *
 * \param h_seed The histogram of the seed window
 * \param pc2_times Sorted vector containing all pc2 times.
 * \param seed_end_time_index Index in the pc2_times vector of the last
 *                            pc2 time included in the seed window.
 * \param seed_end Iterator pointing to the first sample outside the seed
 *                 window.
 * \param samples_end Iterator pointing to the last sample in the sample
 *                    vector.
 * \param growth Number of pc2 events in a growth window.
 * \param max_delta Window difference limit.
 * \return The index of the last pc2 time to include in the window.
 */
static fnr_sample_count_t
grow_window(const bin_vector_ptr_t &h_seed,
	    const fnr_time_vector_t &pc2_times,
	    const fnr_sample_count_t seed_end_time_index,
	    const repl_iterator_t &seed_end,
	    const repl_iterator_t &samples_end,
	    const fnr_sample_count_t growth,
	    const double max_delta)
{
    histogram_type_t h;
    fnr_sample_count_t g_begin_index(seed_end_time_index);
    repl_iterator_t g_begin(seed_end);

    if (growth == 0)
	return seed_end_time_index;

    while (g_begin_index < pc2_times.size()) {
	fnr_sample_count_t g_end_index(
	    (fnr_sample_count_t)MIN(g_begin_index + growth - 1, pc2_times.size() - 1));

        // If there are more than one sample ending at the end time of this
        // window, include them all in the window.
        while (g_end_index + 1 != pc2_times.size() &&
               pc2_times[g_end_index + 1] == pc2_times[g_end_index])
            g_end_index++;

	repl_iterator_t g_end(
	    find_end_sample(g_begin, samples_end,
			    pc2_times[g_end_index]));

	bin_vector_ptr_t h_growth(
	    h.get_histogram(unknown_replacement_iterator_t(g_begin, samples_end),
			    unknown_replacement_iterator_t(g_end, samples_end),
			    reuse_distance));

	double delta(histogram_distance(h_seed, h_growth));

#ifdef DEBUGOUT
	cerr << "Grow: [" << pc2_times[g_begin_index] << ","
	     << pc2_times[g_end_index] << "]: " << delta << endl;
#endif

	// The difference between the seed and the current window was
	// too large, return the
	if (delta >= max_delta)
	    return g_begin_index;

	g_begin = g_end;
	g_begin_index = g_end_index + 1;
    }

    return (fnr_sample_count_t)(pc2_times.size() - 1);
}

windowing_adaptive_t::windowing_adaptive_t(fnr_sample_count_t seed,
					   fnr_sample_count_t growth,
					   double max_delta)
    : windowing_engine_t(),
      seed(seed), growth(growth), max_delta(max_delta)
{
}

string
windowing_adaptive_t::name() const
{
    return "windowing_adaptive_t(" +
	lexical_cast<string>(seed) + "," +
	lexical_cast<string>(growth) + "," +
	lexical_cast<string>(max_delta) + ")";
}


void
windowing_adaptive_t::do_windowing(
    window_vector_t &windows,
    const time_mapping_burst_t &tm_burst,
    const mapped_cold_cptr_vector_t &cold_samples,
    const mapped_repl_base_cptr_vector_t &repl_samples,
    const tm_cache_id_t cache_id) const
{
    // Count the samples.
    fnr_sample_count_t num_samples =
        (fnr_sample_count_t)(repl_samples.size() + cold_samples.size());
    if (num_samples == 0)
	return;

    fnr_time_vector_t pc2_times;
    populate_pc2_times(pc2_times, cold_samples, repl_samples);

    // Create windows.
    histogram_type_t h;
    fnr_time_t next_begin_time = 0;
    repl_iterator_t begin_sample(repl_samples.begin());
    fnr_sample_count_t begin_time_index(0);
    while (begin_time_index < pc2_times.size()) {
	// Create the seed window.
	fnr_sample_count_t seed_end_time_index(
	    (fnr_sample_count_t)MIN(begin_time_index + seed - 1, pc2_times.size() - 1));

        // If there are more than one sample ending at the end time of this
        // window, include them all in the window.
        while (seed_end_time_index + 1 != pc2_times.size() &&
               pc2_times[seed_end_time_index + 1] == pc2_times[seed_end_time_index])
            seed_end_time_index++;

	repl_iterator_t seed_end_sample(
	    find_end_sample(begin_sample, repl_samples.end(),
			    pc2_times[seed_end_time_index]));
	bin_vector_ptr_t h_seed(
	    h.get_histogram(unknown_replacement_iterator_t(begin_sample, repl_samples.end()),
			    unknown_replacement_iterator_t(seed_end_sample, repl_samples.end()),
			    reuse_distance));

	// Grow the window
	fnr_sample_count_t end_time_index(
	    grow_window(h_seed, pc2_times,
			seed_end_time_index,
			seed_end_sample, repl_samples.end(),
			growth, max_delta));

	// Create the window.
	window_ptr_t window(new window_t);

	// If there would be less than SEED pc2 events left after
	// creating the window, include the last pc2 times as well to
	// prevent having a small window on the end.
        window->begin_time = next_begin_time;
	if (pc2_times.size() - end_time_index - 1 < seed) {
	    end_time_index = (fnr_sample_count_t)(pc2_times.size() - 1);
	    window->end_time = tm_burst.get_last_local_time(cache_id);
	} else
	    window->end_time = pc2_times[end_time_index];

        // Add the window to the vector.
	windows.push_back(window);

        // Prepare for the next iteration.
	next_begin_time = window->end_time + 1;
	begin_sample = find_end_sample(begin_sample, repl_samples.end(), window->end_time);
	begin_time_index = end_time_index + 1;
    }
}

END_NAMESPACE;
END_NAMESPACE;
