/// @file analysis/fetch_ratio/windowing_adaptive.hh

// Copyright (c) 2008 Acumem AB.  All rights reserved.

#ifndef WINDOWING_ADAPTIVE_HH
#define WINDOWING_ADAPTIVE_HH

#include "windowing.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace acumem::library;

// ----- windowing_adaptive_t ---------------------------------------------------
/**
 * Adaptive windowing algorithm
 *
 * The basic idea of the adaptive windowing algorithm is to merge
 * similar adjecent windows into larger windows with similar
 * properties. The algortihm works as follows:
 *
 * 0. Current window := 0.
 * 1. Choose a seed window, starting at the last unwindowed pc2 time
 *    and ending SEED pc2 events later. Add the seed window to the
 *    current window.
 * 2. Create a normalized histogram of the seed window from step 1.
 * 3. Create a growth window starting at the first unwindowed pc2 time
 *    and ending GROWTH pc2 events later.
 * 4. Create a histogram for the current growth window.
 * 5. If the distance between the seed window histogram and the growth
 *    window is less than DELTA, add the growth window to the current
 *    window and repeat 3.
 * 6. Add the current window to the list of windows.
 * 7. Goto 0 if there is still unwindowed time.
 *
 * The distance measurement currently in use is the sum of the
 * absolute difference of all bins divided by 2.
 */
class windowing_adaptive_t
    : public windowing_engine_t
{
public:
    /**
     * \param seed Size of the seed window in pc2 accesses.
     * \param growth Size of the growth window in pc2 accesses.
     * \param max_delta Maximum difference between growth and
     *                  seed window.
     */
    windowing_adaptive_t(fnr_sample_count_t seed,
			 fnr_sample_count_t growth,
			 double max_delta);

    string name() const;


    void do_windowing(
	window_vector_t &windows,
	const time_mapping_burst_t &tm_burst,
	const mapped_cold_cptr_vector_t &cold_samples,
	const mapped_repl_base_cptr_vector_t &repl_samples,
	const tm_cache_id_t cache_id) const;

private:
    const fnr_sample_count_t seed;
    const fnr_sample_count_t growth;
    const double max_delta;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
