/// @file analysis/fetch_ratio/windowing_static.cc

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#include "windowing_static.hh"
#include <boost/lexical_cast.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace boost;
using namespace acumem::library;

windowing_static_t::windowing_static_t(
    fnr_sample_count_t samples_per_window)
    : windowing_engine_t(),
      samples_per_window(samples_per_window)
{
}

string
windowing_static_t::name() const
{
    return "windowing_static_t(" +
        lexical_cast<string>(samples_per_window) + ")";
}

void
windowing_static_t::do_windowing(
    window_vector_t &windows,
    const time_mapping_burst_t &tm_burst,
    const mapped_cold_cptr_vector_t &cold_samples,
    const mapped_repl_base_cptr_vector_t &repl_samples,
    const tm_cache_id_t cache_id) const
{
    // Create a sorted vector of pc2 times of both replacement and cold
    // samples.
    fnr_time_vector_t pc2_times;
    repl_iterator_t
        rsi(repl_samples.begin()),
        rse(repl_samples.end());
    cold_iterator_t
        csi(cold_samples.begin()),
	cse(cold_samples.end());
    while (rsi != rse || csi != cse)
        if (rsi == rse)
            pc2_times.push_back((*(csi++))->end.local_time);
        else if (csi == cse)
            pc2_times.push_back((*(rsi++))->end()->local_time);
        else {
            fnr_time_t rt((*rsi)->end()->local_time);
            fnr_time_t ct((*csi)->end.local_time);
            if (rt < ct) {
                pc2_times.push_back(rt);
                rsi++;
            }
            else {
                pc2_times.push_back(ct);
                csi++;
            }
        }

    // Create windows.
    fnr_time_t begin_time = 0;
    fnr_sample_count_t begin_index = 0;
    while (begin_index < pc2_times.size()) {
        window_ptr_t window(new window_t);

        // Calculate the begin and end index of the window in the pc2_times
        // vector.
        fnr_sample_count_t end_index;
        if (begin_index + samples_per_window > pc2_times.size())
            end_index = (fnr_sample_count_t)pc2_times.size();
        else
            end_index = begin_index + samples_per_window;

        // If there are more than one sample ending at the end time of this
        // window, include them all in the window.
        while (end_index != pc2_times.size() &&
               pc2_times[end_index] == pc2_times[end_index - 1])
            end_index++;

        // If there are less than samples_per_window samples left, include the
        // rest of the samples in this window.
        if (pc2_times.size() - end_index < samples_per_window)
            end_index = (fnr_sample_count_t)pc2_times.size();

        // Lookup the begin and end times of the window.
        window->begin_time = begin_time;
        if (end_index == pc2_times.size())
            window->end_time = tm_burst.get_last_local_time(cache_id);
        else
            window->end_time = pc2_times[end_index - 1];

        // Add the window to the vector.
        windows.push_back(window);

        // Prepare for the next iteration.
        begin_time = window->end_time + 1;
        begin_index = end_index;
    }
}

END_NAMESPACE;
END_NAMESPACE;
