/// @file analysis/filter/filter.cc

// Copyright (c) 2006-2008 Acumem AB.  All rights reserved.

#include <boost/lexical_cast.hpp>

#include "filter.hh"
#include "filter_builder.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace boost;
using namespace acumem::library;

// -------------------------------------------------------------------

bool
true_filter_t::filter(const fnr_event_sample_t &sample) const
{
    return true;
}

filter_ptr_t
true_filter_t::create(const string_vector_t &v)
{
    return filter_ptr_t(new true_filter_t);
}

string
true_filter_t::name() const
{
    return "true";
}

// -------------------------------------------------------------------

not_filter_t::not_filter_t()
{
}

not_filter_t::not_filter_t(filter_ptr_t n)
: not_filter(n)
{
}

void
not_filter_t::set_filter(filter_ptr_t n)
{
    not_filter = n;
}

bool
not_filter_t::filter(const fnr_event_sample_t &sample) const
{
    if (!not_filter)
        return false;
    return ! not_filter->filter(sample);
}

filter_ptr_t
not_filter_t::create(const string_vector_t &v)
{
    if (v.size() == 1)
	return filter_ptr_t(new not_filter_t(filter_builder_t().create(v[0])));

    // XXX consider to use not(and(...)) if supplied with a list of arguments

    else
	return filter_ptr_t(new not_filter_t(true_filter_t::create(v)));
}

string
not_filter_t::name() const
{
    if (!not_filter)
        return "not()";
    return "not(" + not_filter->name() + ")";
}

// -------------------------------------------------------------------

list_filter_base_t::list_filter_base_t()
    : filter_t()
{
}

string
list_filter_base_t::name() const
{
    string name_string;
    // TODO: Check if caching could be implemented in some way that
    // avoids the problem us not knowing if a child has been
    // modified. The old way caching was implemented (updating a
    // variable containing the name on every add) would return the
    // wrong name if children are modified.
    name_string = operator_name() + "(";
    bool first = true;
    for (filter_vector_t::const_iterator i=filter_vector.begin(),
	     e=filter_vector.end();
	 i != e;
	 i++)
    {
        if (first)
            first = false;
        else
            name_string += ",";
        name_string += (*i)->name();
    }
    name_string += ")";

    return name_string;
}

void
list_filter_base_t::add(filter_ptr_t filter)
{
    filter_vector.push_back(filter);
}

// -------------------------------------------------------------------

or_filter_t::or_filter_t()
    : list_filter_base_t()
{
}

or_filter_t::or_filter_t(filter_ptr_t f0, filter_ptr_t f1)
    : list_filter_base_t()
{
    add(f0);
    add(f1);
}

bool
or_filter_t::filter(const fnr_event_sample_t &sample) const
{
    for (filter_vector_t::const_iterator i=filter_vector.begin(),
	     e=filter_vector.end();
        i != e;
        i++)
    {
        if ((*i)->filter(sample))
            return true;
    }
    return false;
}

filter_ptr_t
or_filter_t::create(const string_vector_t &v)
{
    string_vector_t::const_iterator i = v.begin(), e = v.end();

    or_filter_t *f = new or_filter_t;
    filter_ptr_t ff(f);

    for (; i != e; i++)
	f->add(filter_builder_t().create(*i));

    return ff;
}


string
or_filter_t::operator_name() const
{
    return "or";
}

// -------------------------------------------------------------------

and_filter_t::and_filter_t()
    : list_filter_base_t()
{
}

and_filter_t::and_filter_t(filter_ptr_t f0, filter_ptr_t f1)
    : list_filter_base_t()
{
    add(f0);
    add(f1);
}

bool
and_filter_t::filter(const fnr_event_sample_t &sample) const
{
    for (filter_vector_t::const_iterator i=filter_vector.begin(),
	     e=filter_vector.end();
        i != e;
        i++)
    {
        if (!(*i)->filter(sample))
            return false;
    }
    return true;
}

filter_ptr_t
and_filter_t::create(const string_vector_t &v)
{
    string_vector_t::const_iterator i = v.begin(), e = v.end();

    and_filter_t *f = new and_filter_t;
    filter_ptr_t ff(f);

    for (; i != e; i++)
	f->add(filter_builder_t().create(*i));

    return ff;
}

string
and_filter_t::operator_name() const
{
    return "and";
}

// -------------------------------------------------------------------

access_size_filter_t::access_size_filter_t()
    : line_size(0)
{
}

string
access_size_filter_t::name() const
{
    return string("line_size(") +
	lexical_cast<string>(line_size) + ")";
}

bool
access_size_filter_t::filter(const fnr_event_sample_t &sample) const
{
    return line_size_from_log2(sample.line_size_log2) == line_size;
}

filter_ptr_t
access_size_filter_t::create(const string_vector_t &v)
{
    return filter_ptr_t(new access_size_filter_t);
}

line_size_filter_t::line_size_filter_t()
    : access_size_filter_t()
{
}

line_size_filter_t::line_size_filter_t(fnr_line_size_t line_size_)
    : access_size_filter_t()
{
    set_line_size(line_size_);
}

filter_ptr_t
line_size_filter_t::create(const string_vector_t &v)
{
    if (v.size() != 1) {
	return filter_ptr_t(new line_size_filter_t(1)); // XXX better non-match value
    }

    return filter_ptr_t(new line_size_filter_t(lexical_cast<int>(v[0])));
}

void
line_size_filter_t::set_line_size(fnr_line_size_t line_size_)
{
    line_size = line_size_;
}

// -------------------------------------------------------------------

address_set_filter_t::address_set_filter_t()
{
    compare_tag = true;
}

address_set_filter_t::address_set_filter_t(const fnr_address_set_t &s)
    : address_set(s)
{
    compare_tag = true;
}

void
address_set_filter_t::add(const fnr_address_t &address)
{
    if (compare_tag)
	address_set.insert(address);
    else {
	fnr_address_t a2;
	a2.virtual_address = address.virtual_address;

	address_set.insert(a2);
    }
}

bool
address_set_filter_t::match(const fnr_address_t &address) const
{
    if (compare_tag)
	return address_set.find(address) != address_set.end();
    else {
	fnr_address_t a2;
	a2.virtual_address = address.virtual_address;

	return address_set.find(a2) != address_set.end();
    }
}

string
address_set_filter_t::set_name() const
{
    bool first = true;
    string n;
    fnr_address_set_t::const_iterator i = address_set.begin(), e = address_set.end();
    for (; i != e; i++) {
	if (!first) {
	    n += ",";
	} else
	    first = false;

	n += lexical_cast<string>(i->virtual_address);
    }
    return n;
}

int64_t my_strtoll(const char *str, char **endptr, int base)
{
  long long val = 0;

  if(base == 10) {
    sscanf(str, "%lld", &val);
  }
  else if (base == 16) {
    sscanf(str, "%llx", &val);
  }

  return val;
}


#define IMPLEMENT_ADDRESS_FILTER(prefix, sample_expr)			\
prefix##_set_filter_t::prefix##_set_filter_t()                          \
    : address_set_filter_t()                                            \
{                                                                       \
}                                                                       \
prefix##_set_filter_t::prefix##_set_filter_t(                           \
    const fnr_address_set_t &s)                                         \
    : address_set_filter_t(s)                                           \
{                                                                       \
}                                                                       \
                                                                        \
string								\
prefix##_set_filter_t::name() const {					\
    return #prefix "(" + set_name() + ")";				\
}									\
									\
bool									\
prefix##_set_filter_t::filter(const fnr_event_sample_t &sample) const		\
{									\
    return sample_expr;                                                 \
}									\
									\
filter_ptr_t								\
prefix##_set_filter_t::create(const string_vector_t &v)			\
{									\
    prefix##_set_filter_t *f = new prefix##_set_filter_t;		\
    filter_ptr_t ff(f);							\
    f->compare_tag = false;						\
    string_vector_t::const_iterator i = v.begin(), e = v.end();		\
									\
    for (; i != e; i++) {						\
	fnr_address_t a;						\
	a.virtual_address = my_strtoll(i->c_str(), NULL, 16);           \
	f->add(a);							\
    }									\
									\
    return ff;								\
									\
}									\

IMPLEMENT_ADDRESS_FILTER(
    pc1, match(sample.begin.pc));
IMPLEMENT_ADDRESS_FILTER(
    pc1p, match(sample.pc_1_plus));
IMPLEMENT_ADDRESS_FILTER(
    pc2, !sample.end.empty() && match(sample.end[0].pc));
IMPLEMENT_ADDRESS_FILTER(
    pc12,
    match(sample.begin.pc) && (sample.end.empty() || match(sample.end[0].pc)))

IMPLEMENT_ADDRESS_FILTER(
    mem1, match(sample.begin.addr));
IMPLEMENT_ADDRESS_FILTER(
    mem2, !sample.end.empty() && match(sample.end[0].addr));

// ----- location_set_filter_t ------------------------------------------------

location_set_filter_t::location_set_filter_t(
    feature_context_ptr_t context,
    string sample_file,
    fnr_stack_depth_t stack_depth,
    location_set_t location_set,
    stack_pc_map_result_cptr_t stack_pc_map)
        : context(context),
          sample_file(sample_file),
          stack_depth(stack_depth),
          location_set(location_set),
          stack_pc_map(stack_pc_map)
{
}

bool
location_set_filter_t::filter(const fnr_event_sample_t &sample) const
{
    location_t pc1_loc = stack_pc_map->get_location(&sample.begin);

    if (!sample.end.empty()) {
	location_t pc2_loc = stack_pc_map->get_location(&sample.end[0]);

	return (location_set.find(pc1_loc) != location_set.end() &&
		location_set.find(pc2_loc) != location_set.end());
    } else
	return location_set.find(pc1_loc) != location_set.end();
}

string
location_set_filter_t::name() const
{
    stringstream locations;
    locations << location_set;

    return "location_set_filter(" +
        sample_file + "," +
        lexical_cast<string>(stack_depth) + "," +
        locations.str() + ")";
}

END_NAMESPACE;
END_NAMESPACE;
