/// @file analysis/filter/filter.hh

// Copyright (c) 2006-2008 Acumem AB.  All rights reserved.

#ifndef FILTER_HH
#define FILTER_HH

#include "analysis/call_stack/stack_pc_map.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace acumem::library;

using boost::shared_ptr;

typedef vector<string> string_vector_t;

/// base class for the filter heirarchy.
class filter_t {
public:
    virtual ~filter_t() {}

    virtual bool filter(const fnr_event_sample_t &sample) const =0;

    virtual string name() const =0;
};
typedef shared_ptr<filter_t> filter_ptr_t;

// Filter vector.
typedef vector<filter_ptr_t> filter_vector_t;

/// Default filter, which always allows all samples
class true_filter_t : public filter_t {
public:
    virtual bool filter(const fnr_event_sample_t &sample) const;
    static filter_ptr_t create(const string_vector_t &s);

    virtual string name() const;
};
typedef shared_ptr<true_filter_t> true_filter_ptr_t;

/// This filter inverts the meaning of the subordinate filter
class not_filter_t : public filter_t {
public:
    not_filter_t();
    not_filter_t(filter_ptr_t filter);

    void set_filter(filter_ptr_t n);

    virtual bool filter(const fnr_event_sample_t &sample) const;
    static filter_ptr_t create(const string_vector_t &s);

    virtual string name() const;

private:
    filter_ptr_t not_filter;
};
typedef shared_ptr<not_filter_t> not_filter_ptr_t;

/// This is a base class for filters that have many subordinate filters
class list_filter_base_t : public filter_t {
public:
    list_filter_base_t();
    virtual string name() const;
    void add(filter_ptr_t filter);

protected:
    virtual string operator_name() const =0;

    filter_vector_t filter_vector;
};


/// This filter calculates the lazy boolean or-operator among the
/// filters in its list
class or_filter_t : public list_filter_base_t {
public:
    or_filter_t();
    or_filter_t(filter_ptr_t f0, filter_ptr_t f1);

    virtual bool filter(const fnr_event_sample_t &sample) const;
    static filter_ptr_t create(const string_vector_t &s);

protected:
    string operator_name() const;
};
typedef shared_ptr<or_filter_t> or_filter_ptr_t;

/// This filter calculates the lazy boolean and-operator among the
/// filters in its list
class and_filter_t : public list_filter_base_t {
public:
    and_filter_t();
    and_filter_t(filter_ptr_t f0, filter_ptr_t f1);

    virtual bool filter(const fnr_event_sample_t &sample) const;
    static filter_ptr_t create(const string_vector_t &s);

protected:
    string operator_name() const;
};
typedef shared_ptr<and_filter_t> and_filter_ptr_t;

class access_size_filter_t : public filter_t {
public:
    access_size_filter_t();
    virtual string name() const;

    virtual bool filter(const fnr_event_sample_t &sample) const;
    static filter_ptr_t create(const string_vector_t &s);

protected:
    fnr_line_size_t line_size;
};
typedef shared_ptr<access_size_filter_t> access_size_filter_ptr_t;

/// This filter matches samples with a specific line-size
class line_size_filter_t : public access_size_filter_t {
public:
    line_size_filter_t();
    line_size_filter_t(fnr_line_size_t line_size_);
    void set_line_size(fnr_line_size_t line_size_);

    static filter_ptr_t create(const string_vector_t &s);
};
typedef shared_ptr<line_size_filter_t> line_size_filter_ptr_t;

// Base class for address set filters.
class address_set_filter_t : public filter_t {
public:
    address_set_filter_t();
    address_set_filter_t(const fnr_address_set_t &s);
    virtual ~address_set_filter_t() {}
    void add(const fnr_address_t &address);
protected:
    string set_name() const;
    bool match(const fnr_address_t &address) const;
    bool compare_tag;
private:
    fnr_address_set_t address_set;
};

#define DECLARE_ADDRESS_FILTER(prefix)                                  \
class prefix##_set_filter_t : public address_set_filter_t {             \
public:                                                                 \
    prefix##_set_filter_t();                                            \
    prefix##_set_filter_t(const fnr_address_set_t &s); \
    virtual string name() const;                                   \
    virtual bool filter(const fnr_event_sample_t &sample) const;        \
    static filter_ptr_t create(const string_vector_t &s);		\
};                                                                      \
typedef shared_ptr<prefix##_set_filter_t> prefix##_set_filter_ptr_t


// This filter matches samples with a specific pc1
DECLARE_ADDRESS_FILTER(pc1);

// dito pc1+
DECLARE_ADDRESS_FILTER(pc1p);

// dito pc2
DECLARE_ADDRESS_FILTER(pc2);

// both pc1 and pc2 must be in set
DECLARE_ADDRESS_FILTER(pc12);

// This filter matches samples with a specific memory-address 1.
DECLARE_ADDRESS_FILTER(mem1);

// This filter matches samples with a specific memory-address 2.
DECLARE_ADDRESS_FILTER(mem2);

/// Filter that matches samples whose pc1 and pc2 are in the specified location
/// set.
class location_set_filter_t : public access_size_filter_t {
public:
    location_set_filter_t(feature_context_ptr_t context,
                          string sample_file,
                          fnr_stack_depth_t stack_depth,
                          location_set_t location_set,
                          stack_pc_map_result_cptr_t stack_pc_map);
    bool filter(const fnr_event_sample_t &sample) const;
    string name() const;

private:
    feature_context_ptr_t context;
    string sample_file;
    fnr_stack_depth_t stack_depth;
    location_set_t location_set;

    stack_pc_map_result_cptr_t stack_pc_map;
};
typedef shared_ptr<location_set_filter_t> location_set_filter_ptr_t;

END_NAMESPACE;
END_NAMESPACE;

#endif /* FILTER_HH */
