/// @file analysis/filter/filter_builder.cc

// Copyright (c) 2006-2008 Acumem AB.  All rights reserved.

#include "filter_builder.hh"

#include <iostream>
#include <algorithm>
#include <assert.h>

using namespace std;
using namespace boost;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

builder_base_t::~builder_base_t()
{
}

void
filter_registry_base_t::add(const string &name, builder_base_t *b)
{
    pair<builder_map_t::iterator, bool> res =
	builder_map.insert(make_pair(name, b));

    bool unique_filter_prefix = res.second;
    assert(unique_filter_prefix);
}

const builder_base_t *
filter_registry_base_t::get_builder(const string &name) const
{
    builder_map_t::const_iterator i = builder_map.find(name);
    if (i == builder_map.end())
	return 0;
    return i->second;
}

filter_registry_t::filter_registry_t()
{
    add("true", new builder_t<true_filter_t>);
    add("not", new builder_t<not_filter_t>);
    add("and", new builder_t<and_filter_t>);
    add("or", new builder_t<or_filter_t>);
    add("line_size", new builder_t<line_size_filter_t>);
    add("access_size", new builder_t<access_size_filter_t>);
    add("pc1", new builder_t<pc1_set_filter_t>);
    add("pc1p", new builder_t<pc1p_set_filter_t>);
    add("pc2", new builder_t<pc2_set_filter_t>);
    add("mem1", new builder_t<mem1_set_filter_t>);
    add("mem2", new builder_t<mem2_set_filter_t>);
}

filter_builder_t::filter_builder_t()
{
    if (filter_registry == 0) {
	// XXX double checked locking here
	if (filter_registry == 0) {
	    filter_registry = new filter_registry_t;
	}
    }
}

filter_ptr_t
filter_builder_t::create(const string &filter) const
{
    bool has_params = false;
    string::size_type pos = filter.find('(');
    if (pos == filter.npos)
	pos = filter.find(',');
    else has_params = true;

    string type = filter.substr(0, pos);

    const builder_base_t *builder = filter_registry->get_builder(type);
    if (builder != 0) {
	string_vector_t sv;
	if (has_params) {
	    int level = 0;
	    string::const_iterator p = filter.begin() + pos + 1;
	    string::const_iterator i = p;
	    string::const_iterator e = filter.end();

	    for ( ; i != e; i++) {
		switch (*i) {
		case '(':
		    level ++;
		    break;
		case ',':
		    if (level == 0) {
			sv.push_back(string(p, i));
			p=i+1;
		    }
		    break;
		case ')':
		    if (level == 0) {
			sv.push_back(string(p, i));
			goto double_break;
		    } else level--;
		}
	    }
	    if (level == 0 && p != i)
		sv.push_back(string(p,i));
	}
    double_break:
	return builder->create(sv);
    } else {
	// XXX throw something
	return filter_ptr_t(new true_filter_t);
    }
}

filter_registry_base_t *
filter_builder_t::filter_registry = 0;

END_NAMESPACE;
END_NAMESPACE;
