/// @file analysis/filter/filter_module.cc

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#include <assert.h>

#include <boost/lexical_cast.hpp>
#include <boost/foreach.hpp>

#include "filter_module.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace boost;
using namespace std;

using namespace acumem::library;

static void
do_filtering(parser_result_t &out_data,
	     const parser_result_cptr_t &in_data,
	     const filter_ptr_t &filter)
{
    BOOST_FOREACH(const parser_burst_t &burst_in, in_data->bursts) {
	out_data.bursts.push_back(parser_burst_t());
	parser_burst_t &burst_out(out_data.bursts.back());

	BOOST_FOREACH(const sample_ptr_t &s, burst_in.samples_by_pc1) {
	    if (filter->filter(*s))
		burst_out.samples_by_pc1.push_back(s);
	}

	burst_out.begin = burst_in.begin;
	burst_out.end = burst_in.end;
	burst_out.begin_time = burst_in.begin_time;
	burst_out.length = burst_in.length;
	burst_out.time_map = burst_in.time_map;
	if (!burst_out.samples_by_pc1.empty())
	    burst_out.sample_period = (double)burst_out.length /
		burst_out.samples_by_pc1.size();
	else
	    burst_out.sample_period = 0;
    }
}

parser_result_cptr_t
filter_module_t::calculate(parser_result_cptr_t in_data,
                           fnr_line_size_t line_size)
{
    static filter_ptr_t true_filter(new true_filter_t());

    return calculate(in_data, line_size, true_filter);
}

parser_result_cptr_t
filter_module_t::calculate(parser_result_cptr_t in_data,
                           fnr_line_size_t line_size,
                           filter_ptr_t filter)
{
    parser_result_ptr_t res(new parser_result_t);
    parser_result_t *out_data = res.get();

    and_filter_t *and_filter = new and_filter_t();
    filter_ptr_t f(and_filter);
    and_filter->add(filter_ptr_t(new line_size_filter_t(line_size)));
    and_filter->add(filter);

    out_data->bursts.reserve(in_data->bursts.size());

    do_filtering(*out_data, in_data, f);

    out_data->nominal_sample_period = in_data->nominal_sample_period;
    out_data->process_info = in_data->process_info;
    out_data->stacks = in_data->stacks;
    out_data->event_store = in_data->event_store;
    out_data->stream_header = in_data->stream_header;
    out_data->burst_sampled = in_data->burst_sampled;
    out_data->fenris_features = in_data->fenris_features;
    out_data->contexts = in_data->contexts;

    return res;
}

filter_module_t::filter_module_t(feature_context_ptr_t context,
                                 const string &sample_file,
                                 fnr_line_size_t line_size)
    : feature_t<parser_result_cptr_t>(context),
      sample_file(sample_file),
      line_size(line_size),
      filter(new true_filter_t())
{
}

filter_module_t::filter_module_t(feature_context_ptr_t context,
                                 const string &sample_file,
                                 fnr_line_size_t line_size,
                                 filter_ptr_t filter)
    : feature_t<parser_result_cptr_t>(context),
      sample_file(sample_file),
      line_size(line_size),
      filter(filter)
{
}

parser_result_cptr_t
filter_module_t::get() const
{
    return get_data();
}

string
filter_module_t::name() const
{
    return "filter_module_t(" +
        sample_file + "," +
        lexical_cast<string>(line_size) + "," +
        filter->name() + ")";
}

parser_result_cptr_t
filter_module_t::calculate() const
{
    return calculate(
        parser_t(context, sample_file).get(),
        line_size, filter);
}

END_NAMESPACE;
END_NAMESPACE;
