/// @file analysis/hw_prefetch/hw_prefetch.cc

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

/* The hw_prefetch_t class estimates for each pc how likely it is that the
   cache lines that the pc accesses have been prefetched into the cache by the
   hardware prefetcher.

   The hw_prefetch_t class currently models a stream prefetcher like the ones
   found in the L2 cache of the Pentium 3/4/D/M, Core, Core 2, Athlon 64 and
   Opteron processors, and in the L1 cache of the upcoming Phenom processor. A
   stream prefetcher detects cache misses to consecutive cache lines, N and
   N+1, and prefetches the following cache lines, N+2 ... N+1+X. A stream
   prefetcher may also detect and prefetch streams with negative strides. How
   many cache lines ahead of the stream the prefetcher works, X, depends both
   on the processor model and many other poorly documented factors such as if
   "not many other load misses are in progress" or if "the bus is not very
   busy".

   Stream prefetchers typically only work within a page. When an access stream
   crosses a page boundary it has to trigger a new prefetch stream on the new
   page. The prefetcher can typically track a limited number of streams and
   only track one stream per page, e.g., the Core Solo processor can track 16
   positive stride streams and 4 negative stride streams.

   The hw_prefetch_t class uses a very simplified model to estimate hardware
   prefetching. Let's start with the most accurate model possible, and then go
   through the simplifications made in the hw_prefetch_t class:

   Assume that we have sampled all accesses with cache line-sized lines and
   that we (as a consequence) have a full access trace, and that we have
   calculated the miss probability of each access.

   To calculate the probability that an access finds its cache line in the
   cache thanks to hardware prefetching, we need to find the last pair of
   consecutive accesses that miss in the L1 cache, to consecutive cache lines,
   close enough (spatially) to the cache line we are looking at that it is
   prefetched.

   We don't know exactly what the spatial "close enough" means since we don't
   know how many cache lines are prefetched when a stream is detected, but
   let's assume that 2 cache lines are prefetched. Assume that our access uses
   cache line M.

   We now go through the access trace looking for the last pair of accesses
   before our access on the same page as M that miss in the L1 cache between
   cache lines M-2 -> M-1 (would trigger prefetch of M and M+1) or M-3 -> M-2
   (would trigger prefetch of M-1 and M) for positive strides, or M+2 -> M+1 or
   M+3 -> M+2 for negative strides, that have no intervening accesses to the
   same page that miss in the L1 cache. If such a pair of accesses exists, the
   second of these accesses is a last time before our access at which cache
   line M would have been prefetched.

   If the prefetch time is later than the start of the cache-line sized sample
   that ends at our access the prefetcher improves the chances of the cache
   line being found in the cache. We can calculate the miss probability of our
   access using the prefetch time instead of the sample start time in the miss
   probability calculation.

   If we want to know how large a difference hardware prefetching makes for the
   access we can simply calculate the miss rate of the instruction using the
   sample start time and using the prefetch time and compare the results.

   So far we have not taken the limited number of prefetch streams into
   account. This can be done by calculating the miss ratio for page-sized
   samples using a cache size that corresponds to the number of pages the
   prefetching streams can follow. We can then calculate the reuse distance at
   which the same number of page replacements have occered as the number of
   prefetch streams. We can then model the limited number of prefetch streams
   by limiting the pairs of accesses that trigger a prefetch stream to those
   with a reuse distance shorter than that.

   Even with this detailed model we make a number of errors:

   - We don't know the exact number of cache lines that will be prefetched when
     a prefetch stream is detected.

   - We underestimate the cache miss rate by ignoring cache pollution. The
     hardware prefetching sometimes replaces lines in the cache with lines that
     are not used before they are replaced, but we don't take that into account
     when we do our miss rate calculation, so we get prefetching "for free".

   - If we model a random replacement cache we don't know exactly which
     accesses will miss in the L1 cache.

   Sampling all accesses isn't practical, so for hw_prefetch_t we instead take
   sparse samples with cache line-sized lines, and also with page-sized
   lines. Now to the implementation of hw_prefetch_t:

   Using the page-sized line samples, we start with building a "histogram" for
   each pc of the stride in cache lines of all samples where the pc is pc2. We
   only need to count the number of times the stride is -1 and 1, and the total
   number of samples.

   Now we want to estimate the probability that the cache line accessed by the
   pc2 of each page-sized sample has been prefetched by the hardware
   prefetcher. Unlike the detailed model we can't locate the last time at which
   the cache line would have been prefetched, since we only have sparse
   samples. Instead we limit ourselves to only looking at the pc1 access and
   its preceding access. This will give us an underestimation of the
   probability for a hardware prefetch.

   Looking at the page-sized sample, we can determine if the pc2 access is
   close enough (spatially) to the pc1 access to be prefetched. If the stride
   in cache lines between the start access and end access of the sample is less
   then or equal to the number of lines prefetched when a stream is detected
   the line could be prefetched, otherwise not.

   If the line is close enough, we want to know the probability that the pc1
   access triggered a prefetch. It would do that if the stride of the
   preceeding access was +1 cache line and the stride of the sample is
   positive, or if the stride of the preceeding access was -1 cache line and
   the stride of the sample is negative. We can get the probability for that
   from the "histogram" we built in the first step. Simply take the number of
   samples with the stride we are interested in divided by the total number of
   samples where our pc1 is pc2.

   Here we lose some precision. The pc1 could be followed by different pc2s at
   different times, and there may be dependencies that make the the stride
   between the pc1 and its preceding access behave differently for different
   pc2s. These dependencies are lost in our processing. However, we assume that
   such dependencies are unusual.

   If we determine that a cache line has likely been prefetched, we can
   calculate the probability that the line is still in the cache at the pc2
   access with an ordinary line-sized miss rate calculation based on the reuse
   distance of the page-sized sample. This is not done in the current
   implementation.

   Finally we can average the prefetch probability of all page-sized samples
   where a pc is pc2 to get a overall prefetch probability for the pc.

   We could also model the limited number of prefetch streams in the same way
   as in the detailed model, but that has not been implemented.

   In addition to the errors made in the detailed model we now also have these
   errors:

   - We only look at the immediately preceding pair of accesses on the page,
     but prior accesses could also have triggered a prefetch of the line.

   - We don't take into account that a line may be replaced between a hardware
     prefetch and the next use.

   - We don't take L1 hits into consideration.

   - We don't take into account the limited number of prefetch streams that can
     be tracked.

   - We average out differing behaviours of a pc when executed in different
     contexts.

   In conclusion, the results produced by the hw_prefetch_t class are not
   really probabilities of the accessed cache lines being found in the cache
   thanks to hardware prefetching, but rather some indication of whether the
   access pattern looks like it could easily be handled by a hardware
   prefetcher. */

#include "analysis/hw_prefetch/hw_prefetch.hh"
#include "analysis/fetch_ratio/repl_fr_lru.hh"
#include "analysis/feature_base/exceptions.hh"
#include "analysis/feature_base/warnings.hh"
#include <boost/lexical_cast.hpp>
#include <cmath>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace boost;
using namespace std;

// ----- hw_prefetch_t --------------------------------------------------------

typedef map<locationx_specifier_t, fnr_sample_count_t> pred_count_map_t;

/// Statistics per location specifier.
struct pf_stat_t {
    pf_stat_t()
        : total(0), zero(0), pos(0), neg(0)
        {
        }

    /// @{
    /// The total number of samples with the pc as pc2, the number of samples
    /// with cache line stride 0, and the number of samples triggering positive
    /// and negative prefetches, respectively.
    fnr_sample_count_t total;
    double zero;
    double pos;
    double neg;
    /// @}

    /// @{
    /// Derived ratio of samples triggering positive and negative prefetches,
    /// respectively. Two sets of the values are used alternatingly as source
    /// and destination by the iterative solver.
    double d_pos[2];
    double d_neg[2];
    /// }@

    /// Map from pc to the number of samples where the pc is pc1 in a sample
    /// where this pc is pc2.
    pred_count_map_t predecessors;
};

typedef map<locationx_specifier_t, pf_stat_t> pf_stat_map_t;

hw_prefetch_result_cptr_t
hw_prefetch_t::calculate(
    const feature_context_ptr_t &context,
    const sample_mapping_result_cptr_t &sample_mapping,
    const locationx_mapper_result_cptr_t &lx_mapping,
    const repl_fr_result_cptr_t &pos_page_fr,
    const repl_fr_result_cptr_t &neg_page_fr,
    const fr_lookup_result_cptr_t &fr_lookup,
    const fnr_line_size_t line_size,
    const fnr_line_size_t page_size,
    const fnr_line_count_t prefetch_reach,
    const bool hwpf_enabled)
{
    // Create the result object.
    hw_prefetch_result_ptr_t result(new hw_prefetch_result_t());

    if (!hwpf_enabled)
        return result;

    // Can't do analysis if there are no page size samples.
    if (!sample_mapping->have_page_samples()) {
	context->warn(new line_size_missing_warning_t(
			  "hw_prefetch",
			  "No page samples available "
			  "in the sample file, skipping hardware "
			  "prefetch analysis.",
			  page_size));
	return result;
    }


    // Count the number of pc2 accesses with different cache line strides of
    // each pc.
    pf_stat_map_t pf_stat_map;
    repl_fr_burst_vector_t::const_iterator it_pos_page_fr(
	pos_page_fr->bursts.begin());
    repl_fr_burst_vector_t::const_iterator it_neg_page_fr(
	neg_page_fr->bursts.begin());
    BOOST_FOREACH(const sample_mapping_burst_t &smb, sample_mapping->bursts) {
	const ratio_vector_t &pos_page_repl_fr((it_pos_page_fr++)->val_to_end);
	const ratio_vector_t &neg_page_repl_fr((it_neg_page_fr++)->val_to_end);
	const mapped_repl_page_vector_t &page_repl(smb.repl_page_cache);
	const fnr_sample_count_t num_page_samples((fnr_sample_count_t)page_repl.size());

	for (fnr_sample_count_t i = 0; i < num_page_samples; i++) {
	    const mapped_access_t start(*page_repl[i].start());
	    const mapped_access_t end(*page_repl[i].end());
	    const int64_t stride(
		end.addr().virtual_address / line_size -
		start.addr().virtual_address / line_size);

	    pf_stat_t &pf_stat(pf_stat_map[locationx_specifier_t(end)]);

	    switch (stride) {
	    case 0:
		pf_stat.zero += 1.0;
		break;
	    case 1:
		pf_stat.pos += 1.0 - pos_page_repl_fr[i];
		break;
	    case -1:
		pf_stat.neg += 1.0 - neg_page_repl_fr[i];
		break;
	    }
	    pf_stat.total++;
	    pf_stat.predecessors[locationx_specifier_t(start)]++;
	}
    }

    // Set up the initial state for the iterative solver.
    BOOST_FOREACH(pf_stat_map_t::value_type &stat_entry, pf_stat_map) {
        pf_stat_t *stat = &stat_entry.second;
        stat->d_pos[0] = stat->pos / stat->total;
        stat->d_neg[0] = stat->neg / stat->total;
    }

    // There may be pcs that always or almost always access the same cache line
    // as the previous pc, e.g., several pcs accessing different fields in the
    // same struct. In this case we assume that the pc has the same stride as
    // the previous pc. As there may be chains of several such pcs we must
    // propagate the strides iteratively to until we reach a stable state.
    int src = 0;
    int dst = 1;
    double max_diff;
    do {
        max_diff = 0.0;

        BOOST_FOREACH(pf_stat_map_t::value_type &dst_stat_entry, pf_stat_map) {
            pf_stat_t *dst_stat = &dst_stat_entry.second;
            double dst_ratio = dst_stat->zero / dst_stat->total;
            dst_stat->d_pos[dst] = dst_stat->pos / dst_stat->total;
            dst_stat->d_neg[dst] = dst_stat->neg / dst_stat->total;

            BOOST_FOREACH(pred_count_map_t::value_type &pred,
                          dst_stat->predecessors) {
                pf_stat_map_t::const_iterator src_stat_it =
                    pf_stat_map.find(pred.first);
                if (src_stat_it != pf_stat_map.end()) {
                    const pf_stat_t *src_stat = &src_stat_it->second;
                    double src_ratio = (double)pred.second / dst_stat->total;
                    dst_stat->d_pos[dst] += src_stat->d_pos[src] * src_ratio * dst_ratio;
                    dst_stat->d_neg[dst] += src_stat->d_neg[src] * src_ratio * dst_ratio;
                }
            }

            double pos_diff = dst_stat->d_pos[dst] - dst_stat->d_pos[src];
            pos_diff = max(pos_diff, -pos_diff);
            double neg_diff = dst_stat->d_neg[dst] - dst_stat->d_neg[src];
            neg_diff = max(neg_diff, -neg_diff);
            max_diff = max(max_diff, max(pos_diff, neg_diff));
        }

        src = 1 - src;
        dst = 1 - dst;
    } while (max_diff > 0.001);

    /* Calculate a prefetch ratio for each pc2 access and accumulate the
       results per pc. */
    BOOST_FOREACH(const sample_mapping_burst_t &smb, sample_mapping->bursts) {
	const mapped_repl_page_vector_t &page_repl(smb.repl_page_cache);
	BOOST_FOREACH(const mapped_repl_page_t &s, page_repl) {
	    const mapped_access_t start(*s.start());
	    const mapped_access_t end(*s.end());
	    const locationx_specifier_t pc2_lxspec(end);

	    const int64_t stride(
		end.addr().virtual_address / line_size -
		start.addr().virtual_address / line_size);
	    pf_stat_t &pf_stat = pf_stat_map[pc2_lxspec];

	    /* Can't analyze the sample if we have no statistics for the pc1 of the
	       sample. Assume no prefetch. */
	    pf_stat_map_t::const_iterator pred_pf_stat_it(
		pf_stat_map.find(locationx_specifier_t(start)));
	    if (pred_pf_stat_it == pf_stat_map.end())
		continue;
	    const pf_stat_t &pred_pf_stat = pred_pf_stat_it->second;

	    /* Get the appropriate ratio of sequential cache line accesses of the
	       pc1 of the sample, depending on if the sample has a positive stride
	       within prefetch reach, a negative stride within prefetch reach, or
	       other stride. */
	    double pred_ratio = 0.0;
	    if (stride == 0)
		pred_ratio = max(pred_pf_stat.d_pos[src], pred_pf_stat.d_neg[src]);
	    if (stride > 0 && (fnr_line_count_t)stride <= prefetch_reach)
		pred_ratio = pred_pf_stat.d_pos[src];
	    if (stride < 0 && (fnr_line_count_t)-stride <= prefetch_reach)
		pred_ratio = pred_pf_stat.d_neg[src];

	    /* Accumulate prefetch ratio. */
	    result->prefetch_info[pc2_lxspec] +=
		1.0 / pf_stat.total * pred_ratio;
	}
    }

    BOOST_FOREACH(hw_prefetch_result_t::prefetch_map_t::value_type &v,
		  result->prefetch_info) {
	locationx_t lx(lx_mapping->get_locationx(v.first));
	v.second *= fr_lookup->fr_by_lx(lx).by_kind().total_fetches();
    }

    return result;
}

hw_prefetch_t::hw_prefetch_t(const feature_context_ptr_t &context,
                             const string &sample_file,
                             const fnr_stack_depth_t stack_depth,
                             const fnr_line_size_t line_size,
                             const fnr_line_size_t page_size,
                             const thread_mapping_cptr_t &thread_mapping,
			     const cpu_model_t &cpu_model,
                             const fnr_cache_size_t cache_size,
                             const repl_fr_engine_cptr_t &engine,
			     const windowing_engine_cptr_t &windowing,
                             const int pos_streams,
                             const int neg_streams,
                             const fnr_line_count_t prefetch_reach,
                             const bool hwpf_enabled)
    : feature_t<hw_prefetch_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      cache_size(cache_size),
      engine(engine),
      windowing(windowing),
      pos_streams(pos_streams),
      neg_streams(neg_streams),
      prefetch_reach(prefetch_reach),
      hwpf_enabled(hwpf_enabled)
{
}

hw_prefetch_result_cptr_t
hw_prefetch_t::get() const
{
    return get_data();
}

string
hw_prefetch_t::name() const
{
    return "hw_prefetch_t(" +
	lexical_cast<string>(sample_file) + "," +
        lexical_cast<string>(stack_depth) + "," +
	lexical_cast<string>(line_size) + "," +
	lexical_cast<string>(page_size) + "," +
        thread_mapping->name() + "," +
	cpu_model.name() + "," +
	lexical_cast<string>(cache_size) + "," +
        engine->name() + "," +
	windowing->name() + "," +
	lexical_cast<string>(pos_streams) + "," +
	lexical_cast<string>(neg_streams) + "," +
	lexical_cast<string>(prefetch_reach) + "," +
        lexical_cast<string>(hwpf_enabled) + ")";
}

hw_prefetch_result_cptr_t
hw_prefetch_t::calculate() const
{
    sample_mapping_result_cptr_t sample_mapping(
	sample_mapping_t(context, sample_file, stack_depth, line_size,
			 page_size, thread_mapping, cpu_model).get());

    locationx_mapper_cptr_t lx_mapper_full(
        new locationx_mapper_t(context, sample_file, stack_depth,
                               thread_mapping, true, true, true));

    locationx_mapper_result_cptr_t lx_mapping = lx_mapper_full->get();

    // Get LRU fetch ratio results for the number of page sized lines
    // corresponding to the number of positive and negative prefetch
    // streams. Used to model a limited number of prefetch streams.
    repl_fr_engine_cptr_t fr_engine(
	new repl_fr_lru_t(context, sample_file, stack_depth, thread_mapping,
			  windowing, cpu_model));

    repl_fr_result_cptr_t pos_page_fr;
    repl_fr_result_cptr_t neg_page_fr;
    try {
	pos_page_fr =
	    repl_fr_t(context, sample_file, stack_depth, Mapped_Page_Cache,
		      Mapped_Page_Cache, line_size, page_size,
		      pos_streams * page_size, thread_mapping,
		      cpu_model, fr_engine).get();

	neg_page_fr =
	    repl_fr_t(context, sample_file, stack_depth, Mapped_Page_Cache,
		      Mapped_Page_Cache, line_size, page_size,
		      pos_streams * page_size, thread_mapping,
		      cpu_model, fr_engine).get();
    } catch (feature_page_samples_missing_exception_t psme) {
	context->warn(new line_size_missing_warning_t(
			  "hw_prefetch",
			  "No page samples available "
			  "in the sample file, skipping hardware "
			  "prefetch analysis.",
			  page_size));
	return hw_prefetch_result_cptr_t(new hw_prefetch_result_t());
    }

    // Get fetch ratio lookup.
    fr_lookup_result_cptr_t fr_lookup(
	fr_lookup_t(context, sample_file, stack_depth, line_size, page_size,
		    thread_mapping, cpu_model, lx_mapper_full, cache_size,
		    engine, FRL_Type_All).get());

    return calculate(context, sample_mapping, lx_mapping, pos_page_fr,
                     neg_page_fr, fr_lookup, line_size, page_size,
                     prefetch_reach, hwpf_enabled);
}

END_NAMESPACE;
END_NAMESPACE;
