/// @file analysis/hw_prefetch/hw_prefetch.hh

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#ifndef HW_PREFETCH_HH
#define HW_PREFETCH_HH

#include "analysis/thread_mapping/sample_mapping.hh"
#include "analysis/thread_mapping/thread_mapping.hh"
#include "analysis/fetch_ratio/fr_lookup.hh"
#include "analysis/locationx/locationx.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace std;

using boost::shared_ptr;

// ----- hw_prefetch_result_t -------------------------------------------------

class hw_prefetch_result_t {
public:
    /// Maps location specifiers to their prefetch counts. Locations
    /// not in the map should be considered to have prefetch count
    /// 0.0.
    typedef map<locationx_specifier_t, double> prefetch_map_t;
    prefetch_map_t prefetch_info;
};

typedef shared_ptr<hw_prefetch_result_t> hw_prefetch_result_ptr_t;
typedef shared_ptr<const hw_prefetch_result_t> hw_prefetch_result_cptr_t;

// ----- hw_prefetch_t --------------------------------------------------------

/// Module for hardware prefetch analysis.
class hw_prefetch_t : public feature_t<hw_prefetch_result_cptr_t> {
public:
    static hw_prefetch_result_cptr_t calculate(
        const feature_context_ptr_t &context,
        const sample_mapping_result_cptr_t &sample_mapping,
        const locationx_mapper_result_cptr_t &lx_mapping,
        const repl_fr_result_cptr_t &pos_page_fr,
        const repl_fr_result_cptr_t &neg_page_fr,
        const fr_lookup_result_cptr_t &fr_lookup,
        const fnr_line_size_t line_size,
        const fnr_line_size_t page_size,
        const fnr_line_count_t prefetch_reach,
        const bool hwpf_enabled);

    hw_prefetch_t(const feature_context_ptr_t &context,
                  const string &sample_file,
                  const fnr_stack_depth_t stack_depth,
                  const fnr_line_size_t line_size,
                  const fnr_line_size_t page_size,
                  const thread_mapping_cptr_t &thread_mapping,
		  const cpu_model_t &cpu_model,
                  const fnr_cache_size_t cache_size,
                  const repl_fr_engine_cptr_t &engine,
		  const windowing_engine_cptr_t &windowing,
                  const int pos_streams,
                  const int neg_streams,
                  const fnr_line_count_t prefetch_reach,
                  const bool hwpf_enabled);

    hw_prefetch_result_cptr_t get() const;

protected:
    string name() const;
    hw_prefetch_result_cptr_t calculate() const;

private:
    /// Sample file.
    const string sample_file;

    /// Call stack depth to use in the analysis.
    const fnr_stack_depth_t stack_depth;

    /// Cache line size.
    const fnr_line_size_t line_size;

    /// Page size that the hardware prefetcher operates on.
    const fnr_line_size_t page_size;

    const thread_mapping_cptr_t thread_mapping;
    const cpu_model_t cpu_model;

    /// Fetch ratio calculation parameters.
    const fnr_cache_size_t cache_size;
    const repl_fr_engine_cptr_t engine;
    const windowing_engine_cptr_t windowing;

    /// The number of prefetch streams with positive and negative strides,
    /// respectively.
    const int pos_streams;
    const int neg_streams;

    /// The number of sequential cache lines that are prefetched when a
    /// prefetch stream is triggered.
    const fnr_line_count_t prefetch_reach;

    const bool hwpf_enabled;
};

/// Module for hardware prefetch analysis.
class hw_prefetch_new_t : public feature_t<hw_prefetch_result_cptr_t> {
public:
    static hw_prefetch_result_cptr_t calculate(
        const feature_context_ptr_t &context,
        const sample_mapping_result_cptr_t &sample_mapping,
        const locationx_mapper_result_cptr_t &lx_mapping,
        const repl_fr_result_cptr_t &pos_page_fr,
        const repl_fr_result_cptr_t &neg_page_fr,
        const fr_lookup_result_cptr_t &fr_lookup,
        const fnr_line_size_t line_size,
        const fnr_line_size_t page_size,
        const fnr_line_count_t prefetch_reach,
        const bool hwpf_enabled);

    hw_prefetch_new_t(const feature_context_ptr_t &context,
		      const string &sample_file,
		      const fnr_stack_depth_t stack_depth,
		      const fnr_line_size_t line_size,
		      const fnr_line_size_t page_size,
		      const thread_mapping_cptr_t &thread_mapping,
		      const cpu_model_t &cpu_model,
		      const fnr_cache_size_t cache_size,
		      const repl_fr_engine_cptr_t &engine,
		      const windowing_engine_cptr_t &windowing,
		      const int pos_streams,
		      const int neg_streams,
		      const fnr_line_count_t prefetch_reach,
                      const bool hwpf_enabled);

    hw_prefetch_result_cptr_t get() const;

protected:
    string name() const;
    hw_prefetch_result_cptr_t calculate() const;

private:
    /// Sample file.
    const string sample_file;

    /// Call stack depth to use in the analysis.
    const fnr_stack_depth_t stack_depth;

    /// Cache line size.
    const fnr_line_size_t line_size;

    /// Page size that the hardware prefetcher operates on.
    const fnr_line_size_t page_size;

    const thread_mapping_cptr_t thread_mapping;
    const cpu_model_t cpu_model;

    /// Fetch ratio calculation parameters.
    const fnr_cache_size_t cache_size;
    const repl_fr_engine_cptr_t engine;
    const windowing_engine_cptr_t windowing;

    /// The number of prefetch streams with positive and negative strides,
    /// respectively.
    const int pos_streams;
    const int neg_streams;

    /// The number of sequential cache lines that are prefetched when a
    /// prefetch stream is triggered.
    const fnr_line_count_t prefetch_reach;

    const bool hwpf_enabled;
};

END_NAMESPACE;
END_NAMESPACE;

#endif /* HW_PREFETCH_HH */
