/// @file analysis/hw_prefetch/hw_prefetch_new.cc

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

// See doc/notes/hwpf.txt for theory

#include "hw_prefetch.hh"
#include "watermark.hh"
#include "analysis/filter/filter_module.hh"
#include "analysis/fetch_ratio/repl_fr_lru.hh"
#include "analysis/fetch_ratio/repl_fr.hh"
#include "analysis/fetch_ratio/fr_lookup.hh"
#include "analysis/feature_base/exceptions.hh"
#include "analysis/feature_base/warnings.hh"

#include <boost/lexical_cast.hpp>
#include <iostream>
#include <fstream>
#include <cmath>

// Define DEBUGSTUFF to turn on some debug printouts.
//#define DEBUGSTUFF

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace boost;
using namespace std;


// ----------------------------------------------------------------------
// Internal classes - Stride histogram
// ----------------------------------------------------------------------

/// Representation of a stride + count
struct stride_t {
    stride_t(int64_t stride, int count)
	: stride(stride), count(count) {}
    int64_t stride;
    int count;
};

/// A vector of stride + count tuples. This is a slot in a stride histogram.
typedef vector<stride_t> stride_vector_t;


/// A stride histogram. Each page sample contributes with increments
/// to one slot in the histogram.
struct stride_hist_t {
    stride_hist_t()
	: total(0) {}

    struct equal_t {
	equal_t(int64_t s) : s(s) {}
	bool operator() (const stride_t &r) const
	    {
		return s == r.stride;
	    }
	int64_t s;
    };

    void add_stride(int64_t stride, int count = 1) {
	stride_vector_t::iterator i = find_if(strides.begin(),
					      strides.end(),
					      equal_t(stride));
	if (i == strides.end()) {
	    strides.push_back(stride_t(stride, count));
	} else {
	    i->count += count;
	}
	total += count;
    }

    /// The total number of strides added to this histogram
    int total;

    /// The histogram vector
    stride_vector_t strides;
};

/// A location -> histogram map
typedef std::map<locationx_specifier_t, stride_hist_t> loc_stride_hist_map_t;

/// A location -> location -> histogram map
typedef std::map<locationx_specifier_t, loc_stride_hist_map_t> loc_loc_stride_hist_map_t;

/// Counter of samples per pc2
typedef std::map<locationx_specifier_t, int> pc2_count_map_t;


// ----------------------------------------------------------------------
// Internal classes - Deposits
// ----------------------------------------------------------------------

/// A record of a deposit - that is a collector of hwpf estimates for
/// a pc
struct deposit_t {
    deposit_t()
	: stride(0), total(0), weight(0){}

    deposit_t(int64_t stride)
	: stride(stride), total(0), weight(0) {}

    /// The stride for which this deposit is collected
    int64_t stride;

    /// The total weight deposited
    double total;

    /// Weights deposited for an eligible stride cycle, and after
    /// finalize() is called, adjusted by total
    double weight;
};

/// Lookup structures supporting the mapping (pc1,pc2,stride) -> deposit
typedef vector<deposit_t> deposit_vector_t;
typedef map<locationx_specifier_t, deposit_vector_t> loc_deposit_map_t;
typedef map<locationx_specifier_t, loc_deposit_map_t> loc_loc_deposit_map_t;


/// Registry for deposits. Lookup using (pc1,pc2,stride) tuples.
class deposit_map_t {
public:
    /// Find the deposit for a (pc1, pc2, stride) tuple, or create one if
    /// necessary
    deposit_t &find_or_create(locationx_specifier_t pc1,
			      locationx_specifier_t pc2, int64_t stride) {
	deposit_vector_t &v = loc_loc_deposit_map[pc1][pc2];
	BOOST_FOREACH(deposit_t &d, v) {
	    if (d.stride == stride)
		return d;
	}
	v.push_back(deposit_t(stride));
	return v.back();
    }

    /// Given a (pc1, pc2, stride) tuple, return the estimated hwpf
    /// likelihood for pc2 of such a sample.
    double lookup(locationx_specifier_t pc1,
		  locationx_specifier_t pc2,
		  int64_t stride) {
	deposit_vector_t &v = loc_loc_deposit_map[pc1][pc2];

	if (v.size() == 0)
	    return 0;

	double tot = 0;
	double num = 0;
	BOOST_FOREACH(deposit_t &d, v) {
	    if (d.stride == stride)
		return d.weight;

	    num += d.total;

	    // XXX I had a comment here about this being wrong. "The
	    // wrong s.total is weighed into this". I can't recall
	    // what I meant with that.
	    tot += d.weight * d.total;
	}

	// If we haven't found a match for the sought stride, lets
	// return an average of all reported strides for this pc1/pc2 combo.
	if (num == 0)
	    return 0;

	return tot/num;
    }

    /// After adding all strides and hwpf estimates, normalize the
    /// values by dividing in the totals.
    void finalize(pc2_count_map_t &pc2_count) {
	BOOST_FOREACH(loc_loc_deposit_map_t::value_type &v1,
		      loc_loc_deposit_map) {
	    BOOST_FOREACH(loc_deposit_map_t::value_type &v2, v1.second) {
		locationx_specifier_t pc2 = v2.first;

		// XXX Idea: before finalizing the weight, we need to
		// artificially add entries for all unaccounted
		// strides. They should reflect the surrounding
		// situation in which they occur, and this is likely
		// similar to the original iterative solver.

		BOOST_FOREACH(deposit_t &deposit, v2.second) {

		    assert (deposit.total != 0);
		    assert (pc2_count[pc2] != 0);

#ifdef DEBUGSTUFFx
		    locationx_specifier_t pc1 = v1.first;
		    cerr << "pc1: " << pc1 << " pc2: " << pc2
			 << " stride: " << deposit.stride
			 << " weight: " << deposit.weight
			 << " total: " << deposit.total;
#endif
		    deposit.weight = deposit.weight / deposit.total / pc2_count[pc2];
#ifdef DEBUGSTUFFx
		    cerr << " ==> weight: " << deposit.weight
			 << endl;
#endif
		}
	    }
	}
    }

    loc_loc_deposit_map_t loc_loc_deposit_map;
};

// ----------------------------------------------------------------------
// Internal classes - Convolution result
// ----------------------------------------------------------------------

#ifndef MAX_INT
#define MAX_INT 0x7fffffff
#endif

/// Represents one resulting discrete stride during the convolution process.
struct convolution_result_t {
    convolution_result_t()
	: weight(0), stride(0), min_num_samples(MAX_INT) {}

    convolution_result_t(double weight, int64_t stride,
			 const watermark_t &pos,
			 const watermark_t &neg,
			 int min_num_samples,
			 const vector<int64_t> &strides_history)
	: weight(weight), stride(stride),
	  pos(pos), neg(neg),
	  min_num_samples(min_num_samples),
	  strides_history(strides_history) {}

    /// Weight of this stride value
    double weight;

    /// Current value of the total stride
    int64_t stride;

    /// Along each chain, track the positive and negative maximum
    /// strides as well as the length of the increment progression.
    watermark_t pos;
    watermark_t neg;

    /// The least number of samples along any edge is a metric used to judge
    /// the relative importance of a particular chain
    int min_num_samples;

    /// Maintains the sequence of strides leading up to this point.
    /// XXX. This vector is repeatedly copied. We should change the
    /// representation to avoid this.
    vector<int64_t> strides_history;
};

/// This represents the entire convolution result.
typedef vector<convolution_result_t> convolution_result_vector_t;


// ----------------------------------------------------------------------
// Internal classes - Cycle enumeration
// ----------------------------------------------------------------------

/// Vector of sequences of locations. Each sequence represents a cycle
/// in the page sample dependency graph. The tarjan_enumerate_cycle_t
/// publishes its result in terms of this type.
typedef std::vector<locationx_specifier_t> locationx_specifier_vector_t;
typedef std::vector<locationx_specifier_vector_t> location_cycle_vector_t;


/// Enumerator of cycles in a directed graph. It's a basic depth-first
/// algorithm, but already exhausted branches are blocked to limit the
/// branch-off. Google "enumerate circuits tarjan"
class tarjan_enumerate_cycles_t {
public:
    static void get_cycles(loc_loc_stride_hist_map_t &adj,
			   location_cycle_vector_t &cycles);
private:
    tarjan_enumerate_cycles_t(loc_loc_stride_hist_map_t &adj,
			      location_cycle_vector_t &cycles)
	: adj(adj), start(0, 0, 0), cycles(cycles)
	{}

    bool cycle_enumerate_backtrack(locationx_specifier_t node, int level=0);
    void add_cycle();

    loc_loc_stride_hist_map_t &adj;

    std::list<locationx_specifier_t> point_stack;
    std::set<locationx_specifier_t> mark_set;
    std::list<locationx_specifier_t> mark_stack;
    locationx_specifier_t start;
    location_cycle_vector_t &cycles;
};


void
tarjan_enumerate_cycles_t::get_cycles(loc_loc_stride_hist_map_t &adj,
				      location_cycle_vector_t &cycles)
{
    tarjan_enumerate_cycles_t t(adj, cycles);
    BOOST_FOREACH(const loc_loc_stride_hist_map_t::value_type &v, adj) {
	t.start = v.first;
	t.mark_set.clear();
	t.mark_stack.clear();
	t.cycle_enumerate_backtrack(t.start);
    }
}

void
tarjan_enumerate_cycles_t::add_cycle()
{
    locationx_specifier_vector_t cycle;
    copy(point_stack.begin(), point_stack.end(),
	 inserter(cycle, cycle.end()));
    cycles.push_back(cycle);
}

bool
tarjan_enumerate_cycles_t::cycle_enumerate_backtrack(locationx_specifier_t node,
						     int level)
{
    bool flag = false;

    point_stack.push_back(node);
    mark_set.insert(node);
    mark_stack.push_back(node);

    loc_loc_stride_hist_map_t::const_iterator m = adj.find(node);

    if (m != adj.end()) {
	BOOST_FOREACH(const loc_stride_hist_map_t::value_type &next_pair,
		      m->second) {
	    const locationx_specifier_t &next = next_pair.first;
	    if (next < start)
		continue;

	    if (next == start) {
		add_cycle();
		flag = true;

	    } else if (mark_set.find(next) == mark_set.end()) {
		if (cycle_enumerate_backtrack(next, level+1))
		    flag = true;
	    }
	}

	if (flag) {
	    while (! (mark_stack.back() == node)) {
		mark_set.erase(mark_stack.back());
		mark_stack.pop_back();
	    }
	    mark_set.erase(mark_stack.back());
	    mark_stack.pop_back();
	}
    }

    point_stack.pop_back();
    return flag;
}

// ----------------------------------------------------------------------
// Internal classes - Averagers
// ----------------------------------------------------------------------

/// Maintain a total and a count
struct average_t {
    average_t() : sum(0), num(0) {}

    void add(double x) {
	sum += x;
	num ++;
    }

    double average() const {
	return num == 0 ? 0 : sum/num;
    }

    double sum;
    int num;
};


/// Two "averagers", one for positive stride fetch ratios, one for
/// negative strides.
struct pos_neg_fr_t {
    void add_pos(double fr) {
	pos.add(fr);
    }

    void add_neg(double fr) {
	neg.add(fr);
    }

    double get_pos() const {
	return pos.average();
    }

    double get_neg() const {
	return neg.average();
    }

    average_t pos;
    average_t neg;
};

/// For each location, maintain two fetch ratio averages, one for each
/// one of positive and negative strides.
typedef std::map<locationx_specifier_t, pos_neg_fr_t> location_pos_neg_fr_t;


// ----------------------------------------------------------------------
// Other internal helper functions
// ----------------------------------------------------------------------

/// Populate a location_pos_neg_fr_t based on page samples. Also
/// populate a loc_loc_stride_hist_map_t and count samples per pc2.
static void
scan_samples(const sample_mapping_result_cptr_t &sample_mapping,
	     repl_fr_result_cptr_t pos_page_fr,
	     repl_fr_result_cptr_t neg_page_fr,
	     fnr_line_size_t line_size,
	     loc_loc_stride_hist_map_t &loc_loc_map,
	     location_pos_neg_fr_t &loc_pos_neg_fr,
	     pc2_count_map_t &pc2_count_map)
{
    repl_fr_burst_vector_t::const_iterator it_pos_page_fr(
	pos_page_fr->bursts.begin());
    repl_fr_burst_vector_t::const_iterator it_neg_page_fr(
	neg_page_fr->bursts.begin());
    BOOST_FOREACH(const sample_mapping_burst_t &smb, sample_mapping->bursts) {
	const ratio_vector_t &pos_page_repl_fr((it_pos_page_fr++)->val_to_end);
	const ratio_vector_t &neg_page_repl_fr((it_neg_page_fr++)->val_to_end);
	const mapped_repl_page_vector_t &page_repl(smb.repl_page_cache);
	const fnr_sample_count_t num_page_samples((fnr_sample_count_t)page_repl.size());

	for (fnr_sample_count_t i = 0; i < num_page_samples; i++) {
	    const mapped_access_t start(*page_repl[i].start());
	    const mapped_access_t end(*page_repl[i].end());
	    const int64_t stride(
		end.addr().virtual_address / line_size -
		start.addr().virtual_address / line_size);

	    locationx_specifier_t pc2(end), pc1(start);

	    // update histogram
	    loc_loc_map[pc1][pc2].add_stride(stride);

	    // update page size replacement rates for pos/neg streams
	    pos_neg_fr_t &pnf = loc_pos_neg_fr[pc2];
	    pnf.add_pos(pos_page_repl_fr[i]);
	    pnf.add_neg(neg_page_repl_fr[i]);

	    // Count samples by pc2
	    pc2_count_map[pc2] ++;
	}
    }
}


/// Remove edges that are insignificant. Insignificant and coincident
/// edges tend to join otherwise irrelated parts of the tree and this
/// causes explosion in the number of circuits of the adjacency graph.
static void
prune_insignificant_edges(loc_loc_stride_hist_map_t &loc_loc_map)
{
#ifdef DEBUGSTUFFx
    cerr << "Samples before pruning" << endl;
    BOOST_FOREACH(loc_loc_stride_hist_map_t::value_type &v1,
		  loc_loc_map) {
	BOOST_FOREACH(const loc_stride_hist_map_t::value_type &v2,
		      v1.second) {
	    cerr << "pc1: " << v1.first
		 << " pc2: " << v2.first
		 << " total: " << v2.second.total
		 << " (";
	    BOOST_FOREACH(const stride_t &s, v2.second.strides)
		cerr << "[" << s.stride << ":" << s.count << "] ";
	    cerr << ")" << endl;
	}
    }
#endif

    BOOST_FOREACH(loc_loc_stride_hist_map_t::value_type &v1,
		  loc_loc_map) {

	// Only consider pruning if there are more than one pc2
	// following this pc1. If there are only one pc2 ever
	// succeeding a pc1, then do not prune it.
	if (v1.second.size() > 1) {

	    // Count number of samples having this pc1
	    int total = 0;
	    BOOST_FOREACH(const loc_stride_hist_map_t::value_type &v2,
			  v1.second) {
		const stride_hist_t &s = v2.second;
		total += s.total;
	    }

	    // Prune pc2 entries having less samples than say 5% of
	    // this number (MAGIC number 5%)
	    int limit = max(2, total/20);
	    vector<locationx_specifier_t> prune;

	    BOOST_FOREACH(loc_stride_hist_map_t::value_type &v2,
			  v1.second) {
		const locationx_specifier_t &pc2 = v2.first;
		stride_hist_t &s = v2.second;

		if (s.total+1 <= limit) {
		    prune.push_back(pc2);
		} else {
		    // as an additional pruning, remove all strides less than 10% of the total.
		    stride_hist_t x;
		    BOOST_FOREACH(const stride_t &stride, s.strides) {
			if (stride.count > total/10)
			    x.add_stride(stride.stride, stride.count);
		    }
		    if (x.total == 0) {
			// all slots removed - noise - likely not an
			// interesting cycle
			prune.push_back(pc2);
		    } else {
			s.total = x.total;
			s.strides.swap(x.strides);
		    }
		}
	    }

	    BOOST_FOREACH(locationx_specifier_t p, prune) {
		v1.second.erase(p);
	    }
	}
    }

#ifdef DEBUGSTUFFx
    cerr << "Samples after pruning" << endl;
    BOOST_FOREACH(loc_loc_stride_hist_map_t::value_type &v1,
		  loc_loc_map) {
	BOOST_FOREACH(const loc_stride_hist_map_t::value_type &v2,
		      v1.second) {
	    cerr << "pc1: " << v1.first
		 << " pc2: " << v2.first
		 << " total: " << v2.second.total
		 << " (";

	    BOOST_FOREACH(const stride_t &s, v2.second.strides)
		cerr << "[" << s.stride << ":" << s.count << "] ";
	    cerr << ")" << endl;
	}
    }
#endif
}


/// Convolute the stride histograms along each circuit. The result is
/// a vector of total strides and their respective "weight".
static void
convolute(const locationx_specifier_vector_t &lv,
	  const loc_loc_stride_hist_map_t &loc_loc_map,
	  fnr_line_count_t prefetch_reach,
	  convolution_result_vector_t &result_vector)
{
    result_vector.clear();

    // This is the discrete convolution unit element, a single unit pulse
    result_vector.push_back(convolution_result_t(1, 0,
						 watermark_t(prefetch_reach),
						 watermark_t(prefetch_reach),
						 MAX_INT,
						 vector<int64_t>()));
    locationx_specifier_t last(lv.back());
    BOOST_FOREACH(locationx_specifier_t next, lv) {

	// Make sure we have data for the locations in the cycle.
	loc_loc_stride_hist_map_t::const_iterator ll = loc_loc_map.find(last);
	if (ll == loc_loc_map.end()) {
	    last = next;
	    continue;
	}
	loc_stride_hist_map_t::const_iterator l = ll->second.find(next);
	if (l == ll->second.end()) {
	    last = next;
	    continue;
	}

	const stride_hist_t &sh = l->second;

	convolution_result_vector_t sum;
	sum.reserve(result_vector.size() * sh.strides.size());

	BOOST_FOREACH(convolution_result_t &l, result_vector) {
	    BOOST_FOREACH(const stride_t &r, sh.strides) {

		// XXX At this point, we could consider adding
		// additional pruning, say to limit the analysis to
		// the 5 most dominant histogram bins.

		// XXX Another idea is to prune everything that is
		// more than half a page away.

		int64_t stride = l.stride + r.stride;
		double weight = l.weight * r.count;
		vector<int64_t> next = l.strides_history;
		next.push_back(r.stride);

		// The 'mark' thing is to maintain the positive
		// and negative stride watermarks and stretches
		sum.push_back(
		    convolution_result_t(weight, stride,
					 l.pos.mark(stride),
					 l.neg.mark(-stride),
					 min(l.min_num_samples, r.count),
					 next));
	    }
	}
	result_vector.swap(sum);
	last = next;
    }
}


// ----------------------------------------------------------------------
// hw_prefetch_new_t
// ----------------------------------------------------------------------

hw_prefetch_result_cptr_t
hw_prefetch_new_t::calculate(
    const feature_context_ptr_t &context,
    const sample_mapping_result_cptr_t &sample_mapping,
    const locationx_mapper_result_cptr_t &lx_mapping,
    const repl_fr_result_cptr_t &pos_page_fr,
    const repl_fr_result_cptr_t &neg_page_fr,
    const fr_lookup_result_cptr_t &fr_lookup,
    const fnr_line_size_t line_size,
    const fnr_line_size_t page_size,
    const fnr_line_count_t prefetch_reach,
    const bool hwpf_enabled)
{
    // Create the result object.
    hw_prefetch_result_ptr_t result(new hw_prefetch_result_t());

    if (!hwpf_enabled)
        return result;

    // Can't do analysis if there are no page size samples.
    if (!sample_mapping->have_page_samples()) {
	context->warn(new line_size_missing_warning_t(
			  "hw_prefetch",
			  "No page samples available "
			  "in the sample file, skipping hardware "
			  "prefetch analysis.",
			  page_size));
	return result;
    }

    // Loop over all bursts and all samples and build an adjancency
    // structure. Also accumulate and average the pos_page_fr and
    // neg_page_fr misses per pc. Count samples per pc2. We need those
    // structures later.

    loc_loc_stride_hist_map_t loc_loc_map;
    location_pos_neg_fr_t loc_pos_neg_fr;
    pc2_count_map_t pc2_count_map;
    scan_samples(sample_mapping, pos_page_fr, neg_page_fr, line_size,
		 loc_loc_map, loc_pos_neg_fr, pc2_count_map);


    // Prune insignificant edges, which are relatively unique samples
    // with low incidence. They just contribute to linking unrelated
    // code sections and expanding the search space.

    prune_insignificant_edges(loc_loc_map);


    // From this adjancency structure, extract all cycles. This only
    // considers the adjacency structure, not stride histograms.

    location_cycle_vector_t cycles;
    tarjan_enumerate_cycles_t::get_cycles(loc_loc_map, cycles);

#ifdef DEBUGSTUFF
    cerr << "Cycles (" << cycles.size() << ")" << endl;
    int n = 0;
#endif

    // Here we attribute all data to (pc1,pc2,stride) tuples.
    //
    // Loop over all cycles:
    //
    //   Convolute the stride histograms along cycles. This gives
    //   several alternative stride totals per cycle, encoded as a
    //   convolution result vector. Each such total is named a "variant".
    //
    //   Loop over all stride variants:
    //     For eligible total strides,
    //       distribute the hwpf likelihood to (pc1,pc2,stride) tuples
    //       according to the weight of the loop.

    deposit_map_t deposits;
    BOOST_FOREACH(const locationx_specifier_vector_t lv, cycles) {
#ifdef DEBUGSTUFF
	cerr << "Cycle " << ++n << ": " << endl;
#endif
#ifdef DEBUGSTUFFx
	cerr << "  " << lv << endl;
#endif

	// Perform the convolution for this cycle. A convolution is
	// done to carry forward the total effect of all variations of
	// strides in this loop. The result is not delivered in an
	// aggregated form, but rather one variant at a time in a
	// list.
	//
	// We also gather information about total stride, and in case
	// we move over more than one cache line, we tell whether we
	// stepped on each cache line before reaching max stride.

	convolution_result_vector_t result_vector;
	convolute(lv, loc_loc_map, prefetch_reach, result_vector);

#ifdef DEBUGSTUFF
	cerr << result_vector.size() << endl;
#endif
	// Sum the weight of non-zero stride loops, to be able to
	// apply some scaling between different cycles. We don't count
	// zero-stride sums, since they don't contribute to
	// understanding if there is a prefetchable stride or not.

	double nonzero_weight  = 0;
	BOOST_FOREACH(const convolution_result_t &r, result_vector)
	    if (r.stride != 0)
		nonzero_weight += r.weight;


	// If there were only zero-stride sums, then we don't have
	// anything to say about this cycle, and assume it will not be
	// prefetchable.

	if (nonzero_weight == 0)
	    continue;

	// Loop over all individual variants
	BOOST_FOREACH(const convolution_result_t &r, result_vector) {

#ifdef DEBUGSTUFFx
	    cerr << "  stride: " << r.stride
		 << "  min_num_samples: " << r.min_num_samples
		 << "  weight: "  << r.weight << " / ";
	    if (r.stride == 0)
		cerr << endl;
#endif
	    // We don't care for zero-stride cycles. They don't tilt the
	    // situation in any direction.

	    if (r.stride == 0)
		continue;


	    // We assign a weight to this variant through the cycle, and
	    // it is the least number of samples along any edge

	    double loop_weight = r.min_num_samples;

	    // Express the fraction of the weight of this variant
	    // among the other non-zero variants of this cycle. This
	    // is important, since the weight is a metric that can't
	    // be compared outside a cycle. By converting it to the
	    // range [0,1] here, it can be used to scale contributions
	    // from different cycles.

	    double conv_fraction = r.weight / nonzero_weight;

#ifdef DEBUGSTUFFx
	    cerr << 100.0*conv_fraction << "%";
	    if (r.stride < 0 && r.neg.stretch >= -r.stride) {
		cerr << " -" << -r.stride << endl;
	    } else if (r.stride > 0 && r.pos.stretch >= r.stride) {
		cerr << " +" << r.stride << endl;
	    } else cerr << endl;
#endif


	    // deposit hwpf ratios for (pc1,pc2,stride) tuples along the cycle

	    locationx_specifier_t pc1 = lv.back();
	    vector<int64_t>::const_iterator stride_history_iter =
		r.strides_history.begin();

	    BOOST_FOREACH(locationx_specifier_t pc2, lv) {
		deposit_t &deposit =
		    deposits.find_or_create(pc1, pc2, *stride_history_iter);

		double tot = loop_weight * conv_fraction;
		double weight = 0;

		// Criteria fulfilled for a negative/positive stride?
		if (r.stride < 0 && r.neg.stretch >= -r.stride) {
#ifdef DEBUGSTUFFx
		    cerr << pc1 << " -> " << pc2
			 << " negstride " << -r.stride << endl;
#endif
		    pos_neg_fr_t &pnf = loc_pos_neg_fr[pc2];
		    weight = tot * (1 - pnf.get_neg());

		} else if (r.stride > 0 && r.pos.stretch >= r.stride) {
#ifdef DEBUGSTUFFx
		    cerr << pc1 << " -> " << pc2
			 << " posstride " << r.stride << endl;
#endif
		    pos_neg_fr_t &pnf = loc_pos_neg_fr[pc2];
		    weight = tot * (1 - pnf.get_pos());
		}
#ifdef DEBUGSTUFFx
		else {
		    cerr << pc1 << " -> " << pc2 << " nostride" << endl;
		}
#endif

		// deposit a total and a weight, both scaled by the loop_weight

		deposit.weight += weight;
		deposit.total += tot;

#ifdef DEBUGSTUFFx
		cerr << "  total += " << tot << " = " << deposit.total
		     << "  weight += " << weight << " = " << deposit.weight << endl;
#endif

		// advance to next (pc1,pc2) pair
		pc1 = pc2;
		stride_history_iter ++;
	    }
	} // BOOST_FOREACH(r, result_vector)
    } // BOOST_FOREACH(lv, cycles)


    // After all cycles have been evaluated, complete the calculation
    // by normalizing the weights.

    deposits.finalize(pc2_count_map);


    // Calculate a prefetch ratio for each pc2 access and accumulate
    // the results per pc.

    BOOST_FOREACH(const sample_mapping_burst_t &smb, sample_mapping->bursts) {
	const mapped_repl_page_vector_t &page_repl(smb.repl_page_cache);
	BOOST_FOREACH(const mapped_repl_page_t &s, page_repl) {
	    const mapped_access_t start(*s.start());
	    const mapped_access_t end(*s.end());
	    const locationx_specifier_t pc1(start);
	    const locationx_specifier_t pc2(end);

	    const int64_t stride(
		end.addr().virtual_address / line_size -
		start.addr().virtual_address / line_size);

	    double pf = deposits.lookup(pc1, pc2, stride);
	    result->prefetch_info[pc2] += pf;
	}
    }


    // Prevent round-off errors that cause the prefetch ratio to add
    // up to something slightly too large.

    // Here we scale the prefetch numbers with the number of fetches
    // from that pc2. When combining hwpf info from several pc2, they
    // need to scale based on number of fetches. This logic was
    // previously in the hw_prefetch_result_t class, but is now
    // located here and in the hwpf_lookup_t class.

    BOOST_FOREACH(hw_prefetch_result_t::prefetch_map_t::value_type &v,
		  result->prefetch_info) {
	locationx_t lx(lx_mapping->get_locationx(v.first));
	if (v.second > 1.0)
	    v.second = 1.0;
	v.second *= fr_lookup->fr_by_lx(lx).by_kind().total_fetches();
    }

    return result;
}

hw_prefetch_new_t::hw_prefetch_new_t(const feature_context_ptr_t &context,
                                     const string &sample_file,
                                     const fnr_stack_depth_t stack_depth,
                                     const fnr_line_size_t line_size,
                                     const fnr_line_size_t page_size,
                                     const thread_mapping_cptr_t &thread_mapping,
                                     const cpu_model_t &cpu_model,
                                     const fnr_cache_size_t cache_size,
                                     const repl_fr_engine_cptr_t &engine,
                                     const windowing_engine_cptr_t &windowing,
                                     const int pos_streams,
                                     const int neg_streams,
                                     const fnr_line_count_t prefetch_reach,
                                     const bool hwpf_enabled)
    : feature_t<hw_prefetch_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      cache_size(cache_size),
      engine(engine),
      windowing(windowing),
      pos_streams(pos_streams),
      neg_streams(neg_streams),
      prefetch_reach(prefetch_reach),
      hwpf_enabled(hwpf_enabled)
{
}

hw_prefetch_result_cptr_t
hw_prefetch_new_t::get() const
{
    return get_data();
}

string
hw_prefetch_new_t::name() const
{
    return "hw_prefetch_new_t(" +
	lexical_cast<string>(sample_file) + "," +
        lexical_cast<string>(stack_depth) + "," +
	lexical_cast<string>(line_size) + "," +
	lexical_cast<string>(page_size) + "," +
        thread_mapping->name() + "," +
	cpu_model.name() + "," +
	lexical_cast<string>(cache_size) + "," +
        engine->name() + "," +
	windowing->name() + "," +
	lexical_cast<string>(pos_streams) + "," +
	lexical_cast<string>(neg_streams) + "," +
	lexical_cast<string>(prefetch_reach) + "," +
        lexical_cast<string>(hwpf_enabled) + ")";
}

hw_prefetch_result_cptr_t
hw_prefetch_new_t::calculate() const
{

#ifdef DEBUGSTUFF
    cerr << name() << endl;
#endif

    sample_mapping_result_cptr_t sample_mapping(
	sample_mapping_t(context, sample_file, stack_depth, line_size,
			 page_size, thread_mapping, cpu_model).get());

    locationx_mapper_cptr_t lx_mapper_full(
        new locationx_mapper_t(context, sample_file, stack_depth,
                               thread_mapping, true, true, true));

    locationx_mapper_result_cptr_t lx_mapping = lx_mapper_full->get();

    // Get LRU fetch ratio results for the number of page sized lines
    // corresponding to the number of positive and negative prefetch
    // streams. Used to model a limited number of prefetch streams.
    repl_fr_engine_cptr_t fr_engine(
	new repl_fr_lru_t(context, sample_file, stack_depth, thread_mapping,
			  windowing, cpu_model));

    repl_fr_result_cptr_t pos_page_fr;
    repl_fr_result_cptr_t neg_page_fr;
    try {
	pos_page_fr =
	    repl_fr_t(context, sample_file, stack_depth, Mapped_Page_Cache,
		      Mapped_Page_Cache, line_size, page_size,
		      pos_streams * page_size, thread_mapping,
		      cpu_model, fr_engine).get();

	neg_page_fr =
	    repl_fr_t(context, sample_file, stack_depth, Mapped_Page_Cache,
		      Mapped_Page_Cache, line_size, page_size,
		      pos_streams * page_size, thread_mapping,
		      cpu_model, fr_engine).get();
    } catch (feature_page_samples_missing_exception_t psme) {
	context->warn(new line_size_missing_warning_t(
			  "hw_prefetch",
			  "No page samples available "
			  "in the sample file, skipping hardware "
			  "prefetch analysis.",
			  page_size));
	return hw_prefetch_result_cptr_t(new hw_prefetch_result_t());
    }

    // Get fetch ratio lookup.
    fr_lookup_result_cptr_t fr_lookup(
	fr_lookup_t(context, sample_file, stack_depth, line_size, page_size,
		    thread_mapping, cpu_model, lx_mapper_full, cache_size,
		    engine, FRL_Type_All).get());

    return calculate(context, sample_mapping, lx_mapping, pos_page_fr,
                     neg_page_fr, fr_lookup, line_size, page_size,
                     prefetch_reach, hwpf_enabled);

}

END_NAMESPACE;
END_NAMESPACE;
