/// @file analysis/hw_prefetch/hwpf_lookup.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include "analysis/hw_prefetch/hwpf_lookup.hh"
#include <boost/lexical_cast.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace boost;
using namespace std;

// ----- hwpf_lookup_result_t -------------------------------------------------

hwpf_lookup_result_t::hwpf_lookup_result_t()
{
}

double
hwpf_lookup_result_t::prefetch_ratio(locationx_t lx) const
{
    return lx_prefetches[lx.id].prefetch_ratio();
}

double
hwpf_lookup_result_t::prefetch_ratio(const locationx_set_t &lx_set) const
{
    hwpf_lookup_info_t info;

    BOOST_FOREACH(const locationx_t &lx, lx_set)
	info.add(lx_prefetches[lx.id]);

    return info.prefetch_ratio();
}

// ----- hwpf_lookup_t --------------------------------------------------------

hwpf_lookup_result_cptr_t
hwpf_lookup_t::calculate(
    const locationx_mapper_result_cptr_t &lx_res,
    const fr_lookup_result_cptr_t &frl_res,
    const access_lookup_result_cptr_t &al_res,
    const hw_prefetch_result_cptr_t &hw_prefetch,
    const bool include_all)
{
    hwpf_lookup_result_ptr_t result(new hwpf_lookup_result_t());
    result->lx_prefetches.resize(lx_res->max_id);

    BOOST_FOREACH(const hw_prefetch_result_t::prefetch_map_t::value_type &v,
		  hw_prefetch->prefetch_info) {
	const locationx_specifier_t &lx_spec(v.first);
	const double hwpfs(v.second);
	const locationx_t lx(lx_res->get_locationx(lx_spec));

        loc_access_info_map_t::const_iterator ai_it =
            al_res->loc_access_info.find(lx_spec.location());
        if (include_all ||
	    !(ai_it != al_res->loc_access_info.end() &&
	      ai_it->second.access_classes & FNR_Access_Class_PF))
	    result->lx_prefetches[lx.id].sum_prefetches += hwpfs;
    }

    locationx_id_t lx_size((locationx_id_t)lx_res->locationx_spec.size());
    for (locationx_id_t i = 0; i < lx_size; i++) {
	const locationx_t lx(i);
	const double fetches = frl_res->fr_by_lx(lx).by_kind().total_fetches();

	result->lx_prefetches[lx.id].sum_fetches = fetches;

        /// Rounding errors can give us a prefetch sum that is slightly higher
        /// than the fetch sum. This causes ugly -0.0% values in the report, so
        /// we set the prefetch sum to the fetch sum in that case. See ticket
        /// #876.
        ///
        /// We also check that the rounding error isn't unreasonably large,
        /// more than 1%.
        const double prefetches = result->lx_prefetches[lx.id].sum_prefetches;
        if (prefetches > fetches) {
            assert(prefetches <= fetches * 1.01);
            result->lx_prefetches[lx.id].sum_prefetches = fetches;
        }
    }

    return result;
}

hwpf_lookup_t::hwpf_lookup_t(
    const feature_context_ptr_t &context,
    const string &sample_file,
    const fnr_stack_depth_t stack_depth,
    const fnr_line_size_t line_size,
    const fnr_line_size_t page_size,
    const thread_mapping_cptr_t &thread_mapping,
    const cpu_model_t &cpu_model,
    const locationx_mapper_cptr_t &lx_mapper,
    const fnr_cache_size_t cache_size,
    const repl_fr_engine_cptr_t &engine,
    const windowing_engine_cptr_t &windowing,
    const int pos_streams,
    const int neg_streams,
    const fnr_line_count_t prefetch_reach,
    const bool include_all,
    const bool use_old_algorithm,
    const bool hwpf_enabled)
    : feature_t<hwpf_lookup_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      lx_mapper(lx_mapper),
      cache_size(cache_size),
      engine(engine),
      windowing(windowing),
      pos_streams(pos_streams),
      neg_streams(neg_streams),
      prefetch_reach(prefetch_reach),
      include_all(include_all),
      use_old_algorithm(use_old_algorithm),
      hwpf_enabled(hwpf_enabled)
{
}

hwpf_lookup_result_cptr_t
hwpf_lookup_t::get() const
{
    return get_data();
}

string
hwpf_lookup_t::name() const
{
    return "hwpf_lookup_t(" +
	lexical_cast<string>(sample_file) + "," +
        lexical_cast<string>(stack_depth) + "," +
	lexical_cast<string>(line_size) + "," +
	lexical_cast<string>(page_size) + "," +
        thread_mapping->name() + "," +
	cpu_model.name() + "," +
        lx_mapper->name() + "," +
	lexical_cast<string>(cache_size) + "," +
        engine->name() + "," +
	windowing->name() + "," +
	lexical_cast<string>(pos_streams) + "," +
	lexical_cast<string>(neg_streams) + "," +
	lexical_cast<string>(prefetch_reach) + "," +
	lexical_cast<string>(include_all) + "," +
	lexical_cast<string>(use_old_algorithm) + "," +
        lexical_cast<string>(hwpf_enabled) + ")";
}

hwpf_lookup_result_cptr_t
hwpf_lookup_t::calculate() const
{
    locationx_mapper_result_cptr_t lx_res(lx_mapper->get());

    fr_lookup_result_cptr_t frl_res(
        fr_lookup_t(context, sample_file, stack_depth, line_size, page_size,
                    thread_mapping, cpu_model, lx_mapper, cache_size, engine,
		    include_all ? FRL_Type_All : FRL_Type_Stalls).get());

    access_lookup_result_cptr_t al_res(
        access_lookup_t(context, sample_file, stack_depth).get());

    hw_prefetch_result_cptr_t hw_prefetch;
    if (use_old_algorithm)
	hw_prefetch =
	    hw_prefetch_t(context, sample_file, stack_depth, line_size,
			  page_size, thread_mapping, cpu_model,
			  cache_size, engine, windowing, pos_streams,
			  neg_streams, prefetch_reach, hwpf_enabled).get();
    else
	hw_prefetch =
	    hw_prefetch_new_t(context, sample_file, stack_depth, line_size,
			      page_size, thread_mapping, cpu_model,
			      cache_size, engine, windowing, pos_streams,
			      neg_streams, prefetch_reach, hwpf_enabled).get();

    return calculate(lx_res, frl_res, al_res, hw_prefetch, include_all);
}

END_NAMESPACE;
END_NAMESPACE;
