/// @file analysis/hw_prefetch/hwpf_lookup.hh

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#ifndef HWPF_LOOKUP_HH
#define HWPF_LOOKUP_HH

#include "analysis/hw_prefetch/hw_prefetch.hh"
#include "analysis/structure/access_lookup.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace std;

using boost::shared_ptr;

// ----- hwpf_lookup_result_t -------------------------------------------------

struct hwpf_lookup_info_t {
    hwpf_lookup_info_t()
	: sum_prefetches(0), sum_fetches(0) {}

    hwpf_lookup_info_t(double sum_prefetches, double sum_fetches)
	: sum_prefetches(sum_prefetches), sum_fetches(sum_fetches) {}

    double prefetch_ratio() const {
	return sum_fetches == 0.0 ? 0.0 : sum_prefetches / sum_fetches;
    }

    void add(const hwpf_lookup_info_t &pi) {
	sum_prefetches += pi.sum_prefetches;
	sum_fetches += pi.sum_fetches;
    }

    double sum_prefetches;
    double sum_fetches;
};

/// Result of the hardware prefetch analysis module.
///
/// The prefetch ratio returned is an estimated ratio of the accesses of a pc
/// or set of pcs that are prefetched by the hardware prefetcher. However, the
/// estimates are very rough, usually an underestimation, and should not be
/// taken literally.
///
/// A ratio of 0.0 means the hardware prefetcher doesn't have a chance, 1.0
/// means the hardware prefetcher trivially prefetches every access, anything
/// above 0.4 means the hardware prefetcher is probably doing a good job.
struct hwpf_lookup_result_t {
    friend class hwpf_lookup_t;

public:
    /// Return the prefetch ratio of one pc.
    double prefetch_ratio(locationx_t lx) const;

    /// Return the average prefetch ratio of a set of pcs, weighted by the
    /// number of fetches of each pc.
    double prefetch_ratio(const locationx_set_t &lx_set) const;

private:
    hwpf_lookup_result_t();

    /// Maps pcs to their prefetch ratios.
    typedef vector<hwpf_lookup_info_t> lx_prefetch_info_t;
    lx_prefetch_info_t lx_prefetches;
};

typedef shared_ptr<hwpf_lookup_result_t> hwpf_lookup_result_ptr_t;
typedef shared_ptr<const hwpf_lookup_result_t> hwpf_lookup_result_cptr_t;

// ----- hwpf_lookup_t --------------------------------------------------------

class hwpf_lookup_t : public feature_t<hwpf_lookup_result_cptr_t>  {
public:
    static hwpf_lookup_result_cptr_t calculate(
        const locationx_mapper_result_cptr_t &lx_res,
        const fr_lookup_result_cptr_t &frl_res,
        const access_lookup_result_cptr_t &al_res,
        const hw_prefetch_result_cptr_t &hw_prefetch,
        const bool include_all);

    hwpf_lookup_t(const feature_context_ptr_t &context,
                  const string &sample_file,
                  const fnr_stack_depth_t stack_depth,
                  const fnr_line_size_t line_size,
                  const fnr_line_size_t page_size,
                  const thread_mapping_cptr_t &thread_mapping,
		  const cpu_model_t &cpu_model,
                  const locationx_mapper_cptr_t &lx_mapper,
                  const fnr_cache_size_t cache_size,
                  const repl_fr_engine_cptr_t &engine,
		  const windowing_engine_cptr_t &windowing,
                  const int pos_streams,
                  const int neg_streams,
                  const fnr_line_count_t prefetch_reach,
                  const bool include_all,
		  const bool use_old_algorithm,
                  const bool hwpf_enabled);
    hwpf_lookup_result_cptr_t get() const;

protected:
    string name() const;
    hwpf_lookup_result_cptr_t calculate() const;

private:
    const string sample_file;
    const fnr_stack_depth_t stack_depth;
    const fnr_line_size_t line_size;
    const fnr_line_size_t page_size;
    const thread_mapping_cptr_t thread_mapping;
    const cpu_model_t cpu_model;
    const locationx_mapper_cptr_t lx_mapper;
    const fnr_cache_size_t cache_size;
    const repl_fr_engine_cptr_t engine;
    const windowing_engine_cptr_t windowing;
    const int pos_streams;
    const int neg_streams;
    const fnr_line_count_t prefetch_reach;
    const bool include_all;
    const bool use_old_algorithm;
    const bool hwpf_enabled;
};

END_NAMESPACE;
END_NAMESPACE;

#endif /* HWPF_LOOKUP_HH */
