/// @file analysis/issues/block.cc

// Copyright (c) 2007-2009 Acumem AB.  All rights reserved.

#include "block.hh"
#include "analysis/structure/dep_tree_search.hh"
#include "analysis/structure/dep_cycle_search.hh"
#include "analysis/feature_base/exceptions.hh"
#include "analysis/thread_mapping/sample_mapping.hh"

#include <boost/foreach.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);
BEGIN_NAMESPACE(issues);

using namespace std;
using namespace boost;
using namespace acumem::library;

// ----------------------------------------------------------------------

block_t::block_t(const feature_context_ptr_t &c)
    : context(c)
{
}

void
block_t::set_configuration(loop_driver_conf_t const &c)
{
    conf = c;
    res.reset(new block_result_t);

    pcl_byte_res = pc_lookup_t(
	context, conf.sample_file, conf.stack_depth, conf.line_size,
	conf.page_size, conf.thread_mapping, conf.cpu_model, Mapped_Byte).get();

    pcl_line_res = pc_lookup_t(
	context, conf.sample_file, conf.stack_depth, conf.line_size,
	conf.page_size, conf.thread_mapping, conf.cpu_model, Mapped_Line).get();
}

block_result_ptr_t
block_t::get()
{
    return res;
}

void
block_t::process_loop(loop_ptr_t &loop,
		      const location_set_t &loop_pc_set,
		      const stats_cptr_t &loop_stat)
{
    // clean depgroups
    depgroups_t d;
    d.swap(depgroups);
}

void
block_t::process_dep(loop_ptr_t &loop,
		     const location_set_t &loop_pc_set,
                     const stats_cptr_t &loop_stat,
                     depgroup_id_t dep_id,
		     const location_set_t &dep_pc_set,
		     const stats_cptr_t &stat)
{
    // We look at spatial samples completely in the dependency group
    // address set. Given that this dependency group has a significant
    // repl mr (handled outside this loop), and that there exists
    // temporal/spatial samples with a reuse larger than the average
    // loop length indicate that we should evaluate blocking with
    // respect to this depgroup.


    // @todo XXX currently we evaluate R->R spatial reuse. (R == PF in
    // this case.)  Other combinations may also be interesting.  Those
    // require changes both in this loop as well as in the
    // dep_tree_search configuration.

    // We might need to compare dist against the
    // average_loop_length using some kind of fuzz factor.
    // @todo XXX Tune these, possibly expose them through conf
    double access_fuzz = 1.0;
    double line_fuzz = 1.0;

    fnr_access_classes_t selection_mask = FNR_Access_Class_RD | FNR_Access_Class_PF;

    fnr_time_t line_rdist = (fnr_time_t) -1;
    fnr_time_t access_rdist = (fnr_time_t) -1;

    // First access samples
    BOOST_FOREACH(const location_t &pc, dep_pc_set) {
	pc_info_map_t::const_iterator i = pcl_byte_res->pc_info.find(pc);
	if (i != pcl_byte_res->pc_info.end()) {
            BOOST_FOREACH(const per_pc1_t &p, i->second.per_pc1) {
                if (dep_pc_set.find(p.pc1) != dep_pc_set.end()) {
                    BOOST_FOREACH(const mapped_repl_base_t *s, p.samples) {
                        if (s->valid() != NULL &&
			    (s->end()->access_classes() & selection_mask) &&
                            (s->valid()->access_classes() & selection_mask)) {
                            fnr_time_t dist = s->rdist();
                            if (dist > loop->average_loop_length * access_fuzz)
				access_rdist = min(dist, access_rdist);
			}
                    }
                }
            }
        }
    }

    // and then line-sized samples
    BOOST_FOREACH(const location_t &pc, dep_pc_set) {
	pc_info_map_t::const_iterator i = pcl_line_res->pc_info.find(pc);
	if (i != pcl_line_res->pc_info.end()) {
            BOOST_FOREACH(const per_pc1_t &p, i->second.per_pc1) {
                if (dep_pc_set.find(p.pc1) != dep_pc_set.end()) {
                    // Traverse the samples for this combination of pc:s
                    BOOST_FOREACH(const mapped_repl_base_t *s, p.samples) {
                        if (s->valid() != NULL &&
			    (s->end()->access_classes() & selection_mask) &&
                            (s->valid()->access_classes() & selection_mask)) {
                            fnr_time_t dist = s->rdist();
                            if (dist > loop->average_loop_length * line_fuzz)
				line_rdist = min(dist, line_rdist);
                        }
                    }
                }
	    }
	}
    }

    // If the max rdist is undisturbed, set to zero to indicate no such reuse
    if (line_rdist == (fnr_time_t) -1)
	line_rdist = 0;

    if (access_rdist == (fnr_time_t) -1)
	access_rdist = 0;

    // At this point, we have a potential blocker. Let's do the
    // remaining checks in the finish() member function. This allows
    // us to conduct the necessary barrier checks for all dependency
    // groups at a time.

    // A required criterion for a blocking condition is that there are
    // large reuses.
    if ((line_rdist != 0 || access_rdist != 0) && stat->all_ratio.total_fetch_ratio() > 0.001)
	depgroups.push_back(interesting_depgroup_t(dep_id, stat, dep_pc_set,
						   line_rdist, access_rdist));
}

static fnr_time_t
loop_iterations(loop_ptr_t loop, fnr_time_t blocking_rdist)
{
    if (!loop)
        return 0;

    return (fnr_time_t)(0.5 + blocking_rdist / loop->average_loop_length);
}

void
block_t::process_finish(loop_ptr_t &loop,
                        const stats_cptr_t &loop_stat,
                        const loop_vector_t &loop_stack)
{
    if (depgroups.size() == 0)
	return;

    fnr_time_t max_rdist = 0;
    BOOST_FOREACH(const interesting_depgroup_t &d, depgroups)
	max_rdist = max(max_rdist, max(d.line_rdist, d.access_rdist));

    // At this point 'max_rdist' is the longest reuse distance for any
    // of the dependency groups of this loop that exhibit reuse
    // potential. This marks the limit for how far we will look for
    // barriers. We terminate the search once the average loop length
    // is larger than this, or if we find write operations on a cycle
    // with less this length for any instruction in this loop
    // (including inner loops).

    double loop_fuzz = 2.0;
    bool possible = true;

    // Proceed through the loops from the current level and towards
    // the root loop.
    loop_vector_t::const_reverse_iterator
	i = loop_stack.rbegin(),
	e = loop_stack.rend();

    loop_ptr_t outer_loop = *i;
    loop_ptr_t next_outer_loop;
    loop_ptr_t target_loop;
    location_set_map_t barriers;
    loop_ptr_t previous_child;

    for (int loop_level = 0; i != e; i++, loop_level ++) {
	loop_ptr_t current_loop = *i;
	target_loop = *i;

	// Apply fuzz here. a comparison between 26.001 and 26
	// should not immediately trigger this comparison.
	if (current_loop->average_loop_length * loop_fuzz > max_rdist) {
	    // No need to search further. If we get here, then there
	    // are no interesting barriers, and we should report the
	    // next inner loop as the likely target loop level (if
	    // there is such a loop)
	    break;
	}

	if (possible) {

	    // Get "inclusive" locations for the loop under consideration,
	    // except the loop which we evaluated the previous
	    // iteration. No need to do it again.
	    location_set_t location_set =
		current_loop->get_locations_excluding_loop(previous_child);

	    previous_child = current_loop;

	    // For each instruction in the loop, look for write operations
	    // in less than the limit length.
	    BOOST_FOREACH(const location_t &loc, location_set) {
		location_set_t start_locations;
		start_locations.insert(loc);

		dep_cycle_search_result_cptr_t cycle_res =
                    dep_cycle_search_t(
			context, conf.sample_file, conf.stack_depth,
			conf.line_size, conf.page_size, conf.thread_mapping,
			conf.cpu_model, Mapped_Byte,
			&fnr_access_filter_wr_wrnta,
			start_locations, max_rdist, false).get();

		if (cycle_res->size() != 0) {
		    BOOST_FOREACH(const dep_cycle_t &dep_cycle, *cycle_res) {
			if (location_set.find(dep_cycle.location) ==
			    location_set.end()) {
			    // If the barrier occurs in the first
			    // loop, we don't bother to loop up
			    // barriers and even to look at the rest
			    // of the depgroups. Just abort.
			    if (loop_level == 0)
				if (!conf.blocking_expert)
				    return;

			    barriers[loc].insert(dep_cycle.location);
			    next_outer_loop = *i;
			    possible = false;
			}
		    }
		}
	    }
	}

	if (possible)
	    outer_loop = *i;
    }

    // Criteria for reporting a block is:
    //   * No barriers have been seen (possible == true), or
    //   * There are barriers but we have identified an outer loop
    if (possible ||
	(next_outer_loop.get() && outer_loop->id != next_outer_loop->id)) {
	// Create issues for the different dependency groups.
	BOOST_FOREACH(const interesting_depgroup_t &d, depgroups) {
            // Ignore the dependency group if it causes to few
            // fetches/upgrades/write-backs to be relevant.
            if (!conf.issue_relevant->issue_relevant(d.stat))
                continue;

            fnr_time_t blocking_rdist = max(d.access_rdist, d.line_rdist);

	    block_issue_ptr_t issue(new block_issue_t());
	    issue->block_type_mask =
		(d.access_rdist > 0 ? Block_Temporal : 0) |
		(d.line_rdist > 0 ? Block_Spatial : 0);
	    issue->loop_id = loop->id;
            issue->depgroup_id = d.id;
	    issue->outer_loop_id =
                outer_loop.get() != NULL ? outer_loop->id : 0;
            issue->outer_loop_iterations = loop_iterations(outer_loop, blocking_rdist);
	    issue->next_outer_loop_id =
                next_outer_loop.get() != NULL ? next_outer_loop->id : 0;
            issue->next_outer_loop_iterations = loop_iterations(next_outer_loop, blocking_rdist);
	    issue->target_loop_id =
                target_loop.get() != NULL ? target_loop->id : 0;
            issue->target_loop_iterations =  loop_iterations(target_loop, blocking_rdist);
	    issue->barriers = barriers;

	    // Search backwards for previous writer
	    dep_tree_search_result_cptr_t span_backwards =
                dep_tree_search_access_t(
                    context, conf.sample_file, conf.stack_depth,
		    conf.line_size, conf.page_size, conf.thread_mapping,
		    conf.cpu_model, Mapped_Byte, fnr_access_filter_wr_wrnta,
		    true, true, d.dep_pc_set, (fnr_time_t)-1).get();

	    // Search forwards for next writer
	    dep_tree_search_result_cptr_t span_forwards =
                dep_tree_search_access_t(
                    context, conf.sample_file, conf.stack_depth,
		    conf.line_size, conf.page_size, conf.thread_mapping,
		    conf.cpu_model, Mapped_Byte, fnr_access_filter_wr_wrnta,
		    false, true, d.dep_pc_set, (fnr_time_t)-1).get();

	    BOOST_FOREACH(const location_time_map_t::value_type &v,
			  span_backwards->location_time_map)
		issue->previous_writers.insert(v.first);

	    BOOST_FOREACH(const location_time_map_t::value_type &v,
			  span_forwards->location_time_map)
		issue->next_writers.insert(v.first);

	    res->issues.push_back(issue);
	}
    }
}

bool
block_t::want_depgroup() const
{
    return true;
}

END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;
