/// @file analysis/issues/block.hh

// Copyright (c) 2007-2009 Acumem AB.  All rights reserved.

#ifndef BLOCK_HH
#define BLOCK_HH

#include "issue.hh"
#include "loop_engine.hh"
#include "analysis/thread_mapping/sample_mapping.hh"
#include "analysis/structure/pc_lookup.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);
BEGIN_NAMESPACE(issues);

using namespace std;
using namespace acumem::library;

using boost::shared_ptr;

/// Report temporal blocking opportunities
struct block_result_t {
    block_issue_vector_t issues;
};

typedef shared_ptr<block_result_t> block_result_ptr_t;

/// Module for detecting temporal or spatial blocking opportunities.
///
/// Idea:
///
/// Only look at loops/depgroups with mr above a certain limit.
///
/// Look if there are temporal(spatial) R->R reuses, completely within
/// the loop/depgroup, that are larger than the average loop
/// length. If so, then this loop would benefit from restructuring
/// with respect to the dependency group.
///
/// Next, look for samples W->R, where pc2 is in the target
/// depgroup. Also look for chains such as W->R", R"->R', R'->R. Mark
/// the W as the likely limit for code motion for this dependency
/// group.
class block_t : public loop_analysis_base_t {
public:
    block_t(const feature_context_ptr_t &c);
    void set_configuration(loop_driver_conf_t const &c);
    block_result_ptr_t get();
    virtual void process_loop(loop_ptr_t &loop,
			      const location_set_t &loop_pc_set,
			      const stats_cptr_t &loop_stat);

    virtual void process_dep(loop_ptr_t &loop,
			     const location_set_t &loop_pc_set,
                             const stats_cptr_t &loop_stat,
                             depgroup_id_t dep_id,
			     const location_set_t &dep_pc_set,
			     const stats_cptr_t &stat);

    virtual void process_finish(loop_ptr_t &loop,
                                const stats_cptr_t &loop_stat,
				const loop_vector_t &loop_stack);

    virtual bool want_depgroup() const;

protected:
    feature_context_ptr_t context;
    loop_driver_conf_t conf;
    pc_lookup_result_cptr_t pcl_byte_res;
    pc_lookup_result_cptr_t pcl_line_res;
    block_result_ptr_t res;

    // collect depgroups that indicate reuse. Process these in finish()
    struct interesting_depgroup_t {
	interesting_depgroup_t(depgroup_id_t id,
                               const stats_cptr_t &stat,
			       const location_set_t &dep_pc_set,
			       fnr_time_t line_rdist,
                               fnr_time_t access_rdist)
	    : id(id),
              stat(stat),
              dep_pc_set(dep_pc_set),
	      line_rdist(line_rdist),
              access_rdist(access_rdist)
	    {}

        depgroup_id_t id;
	stats_cptr_t stat;
	location_set_t dep_pc_set;
	fnr_time_t line_rdist;
	fnr_time_t access_rdist;
    };

    typedef vector<interesting_depgroup_t> depgroups_t;
    depgroups_t depgroups;
};


END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;

#endif
