/// @file analysis/issues/comm_spot.cc

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#include "comm_spot.hh"

#include "analysis/structure/dep_tree_search.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);
BEGIN_NAMESPACE(issues);

using namespace std;
using namespace boost;
using namespace acumem::library;

template<class ISSUE>
static void
copy_writers(ISSUE &issue,
            dep_tree_search_result_cptr_t span_backwards,
	     dep_tree_search_result_cptr_t span_forwards,
             locationx_mapper_result_cptr_t lxm_res)
{
    BOOST_FOREACH(const location_time_map_t::value_type &v,
		  span_backwards->location_time_map)
	issue.previous_writers.insert(v.first);

    BOOST_FOREACH(const location_time_map_t::value_type &v,
		  span_forwards->location_time_map)
        issue.next_writers.insert(v.first);
}

template<class ISSUE>
static void
copy_producers(ISSUE &issue,
	       const locationx_set_t &consumers,
	       const communication_consumer_map_t &comm,
               locationx_mapper_result_cptr_t lxm_res)
{
    BOOST_FOREACH(const locationx_t &lx, consumers) {
	communication_consumer_map_t::const_iterator ci = comm.find(lx);
	if (ci != comm.end()) {
	    const communication_consumer_t &cc = ci->second;
	    BOOST_FOREACH(const communication_producer_map_t::value_type &v,
			  cc.producers)
		issue.producers.insert(lxm_res->get_location(v.first));
	}
    }
}

// ----------------------------------------------------------------------

comm_spot_t::comm_spot_t(const feature_context_ptr_t &c)
    : context(c)
{
}

void
comm_spot_t::set_configuration(const loop_driver_conf_t &c)
{
    conf = c;

    lx_mapper = locationx_mapper_cptr_t(
	new locationx_mapper_t(
	    context, c.sample_file, c.stack_depth, c.thread_mapping,
	    true, false, false));
    lxm_res = lx_mapper->get();

    comm_res = communication_t(context, c.sample_file, c.stack_depth,
			       c.line_size, c.page_size, c.cache_size,
			       c.thread_mapping, c.cpu_model, c.engine,
			       lx_mapper).get();

    res = comm_spot_result_ptr_t(new comm_spot_result_t);
}

comm_spot_result_ptr_t
comm_spot_t::get()
{
    return res;
}

void
comm_spot_t::process_dep(loop_ptr_t &loop,
			 const location_set_t &loop_pc_set,
			 const stats_cptr_t &loop_stat,
			 depgroup_id_t dep_id,
			 const location_set_t &dep_pc_set,
			 const stats_cptr_t &stat)
{
    const locationx_set_t lx_set(lxm_res->get_locationx_set(dep_pc_set));
    communication_info_t comm_info(comm_res->get_info(lx_set));

    if (comm_info.has_communication()) {
	const result_valid_t utilization(comm_info.utilization());
	const double waste(comm_info.waste());

	bool comm_spot = conf.issue_relevant->issue_relevant_fetches(
	    comm_info.line_comm / conf.line_size * conf.comm_cost);

	bool comm_utilization = utilization.valid() &&
	    utilization.value() < conf.comm_util_threshold &&
	    conf.issue_relevant->issue_relevant_fetches(
                waste / conf.line_size * conf.comm_cost);

	dep_tree_search_result_cptr_t span_backwards;
	dep_tree_search_result_cptr_t span_forwards;

	// Only dig out these if we will actually produce an issue.
	if (comm_spot || comm_utilization) {
	    // Search backwards for previous writer
	    span_backwards =
		dep_tree_search_access_t(
		    context, conf.sample_file, conf.stack_depth,
		    conf.line_size, conf.page_size, conf.thread_mapping,
		    conf.cpu_model, Mapped_Byte,
		    &fnr_access_filter_wr_wrnta, true, true, dep_pc_set,
		    (fnr_time_t)-1).get();

	    // Search forwards for next writer
	    span_forwards =
		dep_tree_search_access_t(
		    context, conf.sample_file, conf.stack_depth,
		    conf.line_size, conf.page_size, conf.thread_mapping,
		    conf.cpu_model, Mapped_Byte,
		    &fnr_access_filter_wr_wrnta, false, true, dep_pc_set,
		    (fnr_time_t)-1).get();

	    if (comm_spot) {
		comm_spot_issue_ptr_t spot_issue(
		    new comm_spot_issue_t(loop, dep_id, comm_info.line_comm));
		copy_producers(*spot_issue, lx_set, comm_res->comm, lxm_res);
		copy_writers(*spot_issue, span_backwards, span_forwards, lxm_res);
		res->comm_spots.push_back(spot_issue);
	    }

	    if (comm_utilization) {
		comm_utilization_issue_ptr_t util_issue(
		    new comm_utilization_issue_t(loop, dep_id,
						 utilization.value(), waste));

		copy_producers(*util_issue, lx_set, comm_res->comm, lxm_res);
		copy_writers(*util_issue, span_backwards, span_forwards, lxm_res);

		res->comm_utilization.push_back(util_issue);
	    }
	}
    }
}

bool
comm_spot_t::want_depgroup() const
{
    return true;
}

END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;
