/// @file analysis/issues/issue.hh

// Copyright (c) 2006-2009 Acumem AB.  All rights reserved.

#ifndef ISSUE_HH
#define ISSUE_HH

#include "analysis/statistics/stats_slicer.hh"
#include "analysis/structure/loop_depgroups.hh"
#include "analysis/structure/loop_identification.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);
BEGIN_NAMESPACE(issues);

using namespace acumem::library;
using namespace std;

using boost::shared_ptr;

/// Maps locations to location sets.
typedef map<location_t, location_set_t> location_set_map_t;

// ----------------------------------------------------------------------------

class issue_relevant_t {
    /// An issue must have at least this number of either fetches, upgrades or
    /// write-backs to be reported.
    const double f_threshold;
    const double u_threshold;
    const double wb_threshold;

    /// An issue must have at least this number of total number fetches,
    /// upgrades and write-backs to be reported.
    const double fuwb_threshold;

public:
    issue_relevant_t(stats_cptr_t global_stats,
                     double minimum_fuwb_fraction,
                     double minimum_fuwb_ratio)
        : f_threshold(global_stats->all_ratio.total_fetches() * minimum_fuwb_fraction),
          u_threshold(global_stats->all_ratio.total_upgrades() * minimum_fuwb_fraction),
          wb_threshold(global_stats->all_ratio.total_write_backs() * minimum_fuwb_fraction),
          fuwb_threshold(global_stats->all_ratio.accesses() * minimum_fuwb_ratio)
        {}

    /// Check if an issue is relevant based on its statistics.
    bool issue_relevant(stats_cptr_t stats) const {
        double fetches = stats->all_ratio.total_fetches();
        double upgrades = stats->all_ratio.total_upgrades();
        double write_backs = stats->all_ratio.total_write_backs();

        // NOTE: We use > and not >=, since if we use >= all issues are
        // included if there are no fetches, no upgrades or no write-backs.
        return
            (fetches > f_threshold ||
             upgrades > u_threshold ||
             write_backs > wb_threshold) &&
            fetches + upgrades + write_backs > fuwb_threshold;
    }

    /// Check if an issue is relevant based on its number of fetches.
    bool issue_relevant_fetches(double fetches) const {
        // NOTE: We use > and not >=, since if we use >= all issues are
        // included if there are no fetches, no upgrades or no write-backs.
        return
            fetches > f_threshold &&
            fetches > fuwb_threshold;
    }

    /// Check if an issue is relevant based on its number of write-backs.
    bool issue_relevant_write_backs(double write_backs) const {
        // NOTE: We use > and not >=, since if we use >= all issues are
        // included if there are no fetches, no upgrades or no write-backs.
        return
            write_backs > wb_threshold &&
            write_backs > fuwb_threshold;
    }
};

typedef shared_ptr<const issue_relevant_t> issue_relevant_cptr_t;

// ----------------------------------------------------------------------
// Hot-spots

enum hot_spot_type_t {
    Hot_Spot_Fetch = 0,
    Hot_Spot_Write_Back
};

struct hot_spot_issue_t {
    hot_spot_issue_t(hot_spot_type_t type,
		     loop_id_t loop_id,
		     depgroup_id_t depgroup_id)
	: type(type),
	  loop_id(loop_id),
          depgroup_id(depgroup_id)
        {}

    /// Hot spot type
    hot_spot_type_t type;

    /// Loop id.
    loop_id_t loop_id;

    /// Dependency group id.
    depgroup_id_t depgroup_id;

    /// Tracks last write to this dependency group
    location_set_t previous_writers;
};

typedef shared_ptr<hot_spot_issue_t> hot_spot_issue_ptr_t;
typedef shared_ptr<const hot_spot_issue_t> hot_spot_issue_cptr_t;

typedef vector<hot_spot_issue_cptr_t> hot_spot_issue_vector_t;

// ----------------------------------------------------------------------
// Poor fetch or write-back utilization

enum utilization_type_t {
    Utilization_Legacy = 0,
    Utilization_Fetch,
    Utilization_Write_Back
};

struct utilization_issue_t {
    utilization_issue_t(utilization_type_t type,
			loop_id_t loop_id,
			depgroup_id_t depgroup_id)
	: type(type),
	  loop_id(loop_id),
          depgroup_id(depgroup_id)
        {}

    /// Utilization type (write-back/fetch)
    utilization_type_t type;

    /// Loop id.
    loop_id_t loop_id;

    /// Dependency group id.
    depgroup_id_t depgroup_id;

    /// Tracks last write to this dependency group
    location_set_t previous_writers;
};

typedef shared_ptr<utilization_issue_t> utilization_issue_ptr_t;
typedef shared_ptr<const utilization_issue_t> utilization_issue_cptr_t;

typedef vector<utilization_issue_cptr_t> utilization_issue_vector_t;

// ----------------------------------------------------------------------
// Random access

struct random_access_issue_t {
    random_access_issue_t(loop_id_t loop_id,
			  depgroup_id_t depgroup_id)
	: loop_id(loop_id),
          depgroup_id(depgroup_id)
        {}

    /// Loop id.
    loop_id_t loop_id;

    /// Dependency group id.
    depgroup_id_t depgroup_id;

    /// Tracks last write to this dependency group
    location_set_t previous_writers;
};

typedef shared_ptr<random_access_issue_t> random_access_issue_ptr_t;
typedef shared_ptr<const random_access_issue_t> random_access_issue_cptr_t;

typedef vector<random_access_issue_cptr_t> random_access_issue_vector_t;

// ----------------------------------------------------------------------
// temporal and spatial fusion

/// Return type for fusion related results (loop_fusion_t).
struct loop_fusion_issue_t {
    loop_id_t loop1_id;
    loop_id_t loop2_id;

    /// Instructions that are joined by a temporal or spatial sample.
    /// loop1_pcs and loop2_pcs are the exact start and end locations of
    /// joining samples.
    location_set_t loop1_pcs;
    location_set_t loop2_pcs;

    /// Dependence group containing the loop1_pcs and loop2_pcs.
    depgroup_id_t depgroup1_id;
    depgroup_id_t depgroup2_id;

    /// Is it possible to move loop2 up into loop1, or vice versa, or both?
    bool move_loop2_up;
    bool move_loop1_down;

    /// Report blockers for each instruction in the loop separately.
    location_set_map_t block_for_moving_loop1_down;
    location_set_map_t block_for_moving_loop2_up;

    /// Report instructions involved in 1-to-n split-fusions and
    /// n-to-1 split-fusions
    location_set_t additional_pc1;
    location_set_t additional_pc2;
};

typedef shared_ptr<loop_fusion_issue_t> loop_fusion_issue_ptr_t;
typedef shared_ptr<const loop_fusion_issue_t> loop_fusion_issue_cptr_t;

typedef vector<loop_fusion_issue_cptr_t> loop_fusion_issue_vector_t;

// ----- Loop order analysis. -------------------------------------------------

/// Loop order issue.
struct loop_order_issue_t {
    /// Loop id.
    loop_id_t loop_id;

    /// Dependency group id.
    depgroup_id_t depgroup_id;

    /// Code motion boundary.
    location_set_t previous_writers;
};

typedef shared_ptr<loop_order_issue_t> loop_order_issue_ptr_t;
typedef shared_ptr<const loop_order_issue_t> loop_order_issue_cptr_t;

typedef vector<loop_order_issue_cptr_t> loop_order_issue_vector_t;

// ----------------------------------------------------------------------
// temporal/spatial blocking

enum block_type_mask_t {
    Block_Temporal = 1,
    Block_Spatial  = 2
};

/// Result type for temporal/spatial blocking results (temp_block_t,
/// spat_block_t).
///
/// @todo XXX factor out the loop level, just like for loop_order
struct block_issue_t {
    /// Loop containing the dependency group that has a spatial or
    /// temporal reuse
    loop_id_t loop_id;

    /// List of members in the dependency group for which to block
    depgroup_id_t depgroup_id;

    /// Outermost loop level where it is still possible to block with
    /// respect to the reuse instruction
    loop_id_t outer_loop_id;
    /// Iterations of the outer loop per data reuse.
    fnr_time_t outer_loop_iterations;

    /// If barriers are detected, and if the loop hieararchy is
    /// complete, this is the next outer loop level, the one that
    /// contains the barriers. Otherwise the loop id is 0.
    loop_id_t next_outer_loop_id;
    location_set_map_t barriers;
    /// Iterations of the next outer loop per data reuse.
    fnr_time_t next_outer_loop_iterations;

    /// Target loop denotes the loop that has a length on par with the
    /// trigger reuse distance.
    loop_id_t target_loop_id;
    /// Iterations of the target loop per data reuse.
    fnr_time_t target_loop_iterations;

    /// Last/next writers of the dependency group under consideration.
    location_set_t previous_writers;
    location_set_t next_writers;

    /// whether this is a temp block/spat block or both;
    int block_type_mask;
};

typedef shared_ptr<block_issue_t> block_issue_ptr_t;
typedef shared_ptr<const block_issue_t> block_issue_cptr_t;

typedef vector<block_issue_cptr_t> block_issue_vector_t;

// ----------------------------------------------------------------------
// prefetch analysis

/// Classification of prefetch problems
enum prefetch_problem_t {
    Prefetch_Hits,      ///< Prefetched data already in the cache, the
			///< prefetch is unnecessary.

    Prefetch_Too_Far,   ///< Prefetch too far ahead of the instruction
			///< using the data, data is evicted again
			///< before it is used.

    Prefetch_Too_Close  ///< Prefetch too close the instruction using
			///< the data.
};

/// Return type for prefetch related results.
struct prefetch_issue_t {
    prefetch_issue_t(prefetch_problem_t problem,
		     location_t pc,
                     loop_id_t loop_id,
                     depgroup_id_t depgroup_id,
                     double cost,
                     double fetch_ratio,
                     location_set_t prev_locations,
                     location_set_t next_locations)
	: problem(problem),
          pc(pc),
          loop_id(loop_id),
          depgroup_id(depgroup_id),
          cost(cost),
          fetch_ratio(fetch_ratio),
          prev_locations(prev_locations),
          next_locations(next_locations) {
    }

    prefetch_issue_t(prefetch_problem_t problem,
                     location_t pc,
                     loop_id_t loop_id,
                     depgroup_id_t depgroup_id,
                     double cost,
		     fnr_time_t median_rdist,
                     location_set_t prev_locations,
                     location_set_t next_locations)
	: problem(problem),
          pc(pc),
          loop_id(loop_id),
          depgroup_id(depgroup_id),
          cost(cost),
          median_rdist(median_rdist),
          prev_locations(prev_locations),
          next_locations(next_locations) {
    }

    prefetch_problem_t problem;

    // Address of the prefetch instruction.
    location_t pc;

    /// Loop containing the prefetch.
    loop_id_t loop_id;

    /// Dependency group containing the prefetch.
    depgroup_id_t depgroup_id;

    // Cost of the issue in cache miss equivalents (using some assumption about
    // how many prefetch instructions are equivalent to a cache miss). Compared
    // to threshold to determine whether the issue should by displayed. Can be
    // negative if the prefetch does more good than harm.
    double cost;

    // For Prefetch_Hits    - fetch ratio of prefetch instruction.
    // For Prefetch_Too_Far - fetch ratio of following instruction.
    double fetch_ratio;

    // For Prefetch_Too_Close only.
    fnr_time_t median_rdist;

    // Previous and next locations using the same data.
    location_set_t prev_locations;
    location_set_t next_locations;
};

typedef shared_ptr<prefetch_issue_t> prefetch_issue_ptr_t;
typedef shared_ptr<const prefetch_issue_t> prefetch_issue_cptr_t;

typedef vector<prefetch_issue_cptr_t> prefetch_issue_vector_t;

// ----------------------------------------------------------------------
// write-write analysis

struct ww_issue_t {
    ww_issue_t(const stats_cptr_t &stat)
        : stat(stat)
        {}

    /// Current loop
    loop_ptr_t loop;
    location_t pc;
    location_set_t dep_pc;
    location_set_t next_locations;
    stats_cptr_t stat;
};

typedef shared_ptr<ww_issue_t> ww_issue_ptr_t;
typedef shared_ptr<const ww_issue_t> ww_issue_cptr_t;

typedef vector<ww_issue_cptr_t> ww_issue_vector_t;

// ----------------------------------------------------------------------
// communication hot-spots

struct comm_spot_issue_t {
    comm_spot_issue_t(const loop_ptr_t &loop,
		      depgroup_id_t depgroup_id,
		      double comm_lines)
	: loop(loop),
	  depgroup_id(depgroup_id),
	  comm_lines(comm_lines)
        {}

    /// current loop
    loop_ptr_t loop;

    /// current dependency group
    depgroup_id_t depgroup_id;

    /// Communicated lines
    double comm_lines;

    /// List of data producers
    location_set_t producers;

    location_set_t previous_writers;
    location_set_t next_writers;
};

typedef shared_ptr<comm_spot_issue_t> comm_spot_issue_ptr_t;
typedef shared_ptr<const comm_spot_issue_t> comm_spot_issue_cptr_t;

typedef vector<comm_spot_issue_cptr_t> comm_spot_issue_vector_t;

// ----------------------------------------------------------------------
// poor communication utilization

struct comm_utilization_issue_t {
    comm_utilization_issue_t(const loop_ptr_t &loop,
			     depgroup_id_t depgroup_id,
			     double utilization,
			     double waste)
	: loop(loop),
	  depgroup_id(depgroup_id),
	  utilization(utilization),
	  waste(waste)
        {}

    /// current loop
    loop_ptr_t loop;

    /// current dependency group
    depgroup_id_t depgroup_id;

    /// Communication utilization
    double utilization;

    /// Communication waste
    double waste;

    /// List of data producers
    location_set_t producers;

    location_set_t previous_writers;
    location_set_t next_writers;
};

typedef shared_ptr<comm_utilization_issue_t> comm_utilization_issue_ptr_t;
typedef shared_ptr<const comm_utilization_issue_t> comm_utilization_issue_cptr_t;

typedef vector<comm_utilization_issue_cptr_t> comm_utilization_issue_vector_t;

// ----------------------------------------------------------------------
// false sharing

struct fs_issue_t {
    fs_issue_t(const loop_ptr_t &loop,
	       depgroup_id_t depgroup_id,
	       double ratio,
	       double waste)
	: loop(loop),
	  depgroup_id(depgroup_id),
	  ratio(ratio),
	  waste(waste)
        {}

    /// current loop
    loop_ptr_t loop;

    /// current dependency group
    depgroup_id_t depgroup_id;

    /// False sharing ratio
    double ratio;

    /// False sharing waste
    double waste;

    /// List of possible false data consumers.
    location_set_t offenders;
};

typedef shared_ptr<fs_issue_t> fs_issue_ptr_t;
typedef shared_ptr<const fs_issue_t> fs_issue_cptr_t;

typedef vector<fs_issue_cptr_t> fs_issue_vector_t;

// ----------------------------------------------------------------------
// Write NTA possible

struct wnt_possible_issue_t {
    wnt_possible_issue_t(const loop_ptr_t &loop,
			 depgroup_id_t depgroup_id,
			 const location_set_t &next_readers)
	: loop(loop),
	  depgroup_id(depgroup_id),
	  next_readers(next_readers)
        {}

    /// current loop
    loop_ptr_t loop;

    /// current dependency group
    depgroup_id_t depgroup_id;

    // Next locations reading from the data.
    location_set_t next_readers;
};

typedef shared_ptr<wnt_possible_issue_t> wnt_possible_issue_ptr_t;
typedef shared_ptr<const wnt_possible_issue_t> wnt_possible_issue_cptr_t;

typedef vector<wnt_possible_issue_cptr_t> wnt_possible_issue_vector_t;

// ------ prefetchnta_issue_t -------------------------------------------------

// A location, its fraction of the non-temporal reuses in the instruction group
// and the cache size required to get 80% hits.
struct location_fraction_t {
    location_fraction_t(
        location_t location,
        double fraction,
        double required_cache_size)
        : location(location),
          fraction(fraction),
          required_cache_size(required_cache_size)
        {}

    bool operator==(const location_fraction_t &lf) const {
        return this->fraction == lf.fraction;
    }

    bool operator<(const location_fraction_t &lf) const {
        return this->fraction > lf.fraction;
    }

    location_t location;
    double fraction;
    double required_cache_size;
};

typedef vector<location_fraction_t> location_fraction_vector_t;

struct prefetchnta_issue_t {
    prefetchnta_issue_t(const loop_ptr_t &loop,
                        const depgroup_id_t &depgroup_id,
                        const double accesses,
                        const double fetches,
                        const double required_cache_size,
                        const location_fraction_vector_t &pc1_fractions,
                        const location_fraction_vector_t &pc2_fractions)
        : loop(loop),
	  depgroup_id(depgroup_id),
          accesses(accesses),
          fetches(fetches),
          required_cache_size(required_cache_size),
          pc1_fractions(pc1_fractions),
          pc2_fractions(pc2_fractions)
        {}

    /// Loop.
    loop_ptr_t loop;

    /// Dependency group.
    depgroup_id_t depgroup_id;

    /// Number of accesses and fetches associated with the non-temporal reuses.
    double accesses;
    double fetches;

    /// Cache size required to get 80% hits.
    double required_cache_size;

    /// Pc1 locations, their fraction of reuses and the cache. Sorted by
    /// descending fraction.
    location_fraction_vector_t pc1_fractions;

    /// Pc2 locations and their fraction of reuses. The required cache size is
    /// not valid and should be 0.0. Sorted by descending fraction.
    location_fraction_vector_t pc2_fractions;
};

typedef shared_ptr<prefetchnta_issue_t> prefetchnta_issue_ptr_t;
typedef shared_ptr<const prefetchnta_issue_t> prefetchnta_issue_cptr_t;

typedef vector<prefetchnta_issue_cptr_t> prefetchnta_issue_vector_t;

END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;

#endif
