/// @file analysis/issues/loop_engine.cc

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#include "loop_engine.hh"

#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>

#include <iostream>
#include <iomanip>

using namespace std;
using namespace boost;
using namespace acumem::library;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);
BEGIN_NAMESPACE(issues);

#define LIMIT_FACTOR 10

// ----------------------------------------------------------------------

loop_driver_conf_t::loop_driver_conf_t()
{
}

loop_driver_conf_t::loop_driver_conf_t(const string &sample_file,
                                       const fnr_stack_depth_t stack_depth,
				       const fnr_line_size_t line_size,
                                       const thread_mapping_cptr_t &thread_mapping,
				       const cpu_model_t &cpu_model,
				       const fnr_cache_size_t cache_size,
				       const repl_fr_engine_cptr_t &engine,
				       const windowing_engine_cptr_t &windowing,
				       const fnr_line_size_t page_size,
                                       const int hw_pf_pos_streams,
                                       const int hw_pf_neg_streams,
                                       const fnr_line_count_t hw_pf_reach,
				       const bool use_old_hwpf_algorithm,
                                       const bool hwpf_enabled,
                                       const double comm_cost,
                                       const double comm_util_threshold,
                                       const int depgroup_limit_factor,
                                       const double loop_length_factor,
                                       const double random_threshold,
                                       const double prefetch_cycles,
                                       const double cache_miss_cycles,
                                       const bool blocking_expert,
                                       const issue_relevant_cptr_t issue_relevant,
                                       const fnr_cache_size_t l1_cache_size)
    : sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      cache_size(cache_size),
      engine(engine),
      windowing(windowing),
      page_size(page_size),
      hw_pf_pos_streams(hw_pf_pos_streams),
      hw_pf_neg_streams(hw_pf_neg_streams),
      hw_pf_reach(hw_pf_reach),
      use_old_hwpf_algorithm(use_old_hwpf_algorithm),
      hwpf_enabled(hwpf_enabled),
      comm_cost(comm_cost),
      comm_util_threshold(comm_util_threshold),
      depgroup_limit_factor(depgroup_limit_factor),
      loop_length_factor(loop_length_factor),
      random_threshold(random_threshold),
      prefetch_cycles(prefetch_cycles),
      cache_miss_cycles(cache_miss_cycles),
      blocking_expert(blocking_expert),
      issue_relevant(issue_relevant),
      l1_cache_size(l1_cache_size)
{
}

// ----------------------------------------------------------------------

loop_analysis_base_t::~loop_analysis_base_t()
{
}

void
loop_analysis_base_t::process_loop(loop_ptr_t &loop,
				   const location_set_t &loop_pc_set,
				   const stats_cptr_t &loop_stat)
{
}

void
loop_analysis_base_t::process_dep(loop_ptr_t &loop,
				  const location_set_t &loop_pc_set,
                                  const stats_cptr_t &loop_stat,
                                  depgroup_id_t dep_id,
				  const location_set_t &dep_pc_set,
				  const stats_cptr_t &stat)
{
}

void
loop_analysis_base_t::process_finish(loop_ptr_t &loop,
                                     const stats_cptr_t &loop_stat,
				     const loop_vector_t &loop_stack)
{
}

void
loop_analysis_base_t::process_pair(const loop_wrapper_ptr_t &c1,
				   const loop_wrapper_ptr_t &c2,
				   const loop_ptr_t &parent,
				   const location_set_t &loop_pc_set,
				   const stats_cptr_t &c1_stat,
				   const stats_cptr_t &c2_stat)
{
}

bool
loop_analysis_base_t::want_depgroup() const
{
    return false;
}

bool
loop_analysis_base_t::want_pairs() const
{
    return false;
}

// ----------------------------------------------------------------------

loop_driver_t::loop_driver_t(const feature_context_ptr_t &c)
    : context(c)
{
}

void
loop_driver_t::set_configuration(const loop_driver_conf_t &c)
{
    conf = c;

    locationx_mapper_cptr_t lxm(
	new locationx_mapper_t(context,
			       conf.sample_file,
			       conf.stack_depth,
			       conf.thread_mapping,
			       true, false, false));

    st_res = statistics_t(
        context, conf.sample_file, conf.stack_depth, conf.line_size,
        conf.thread_mapping, conf.cpu_model, lxm, conf.cache_size,
	conf.engine, conf.windowing, conf.page_size,
	conf.hw_pf_pos_streams, conf.hw_pf_neg_streams,
	conf.hw_pf_reach, conf.use_old_hwpf_algorithm, conf.hwpf_enabled).get();

    loop_depgroups = loop_depgroups_t(
        context, conf.sample_file, conf.stack_depth, conf.line_size,
	conf.page_size, conf.thread_mapping, conf.cpu_model, Mapped_Line,
	LIMIT_FACTOR).get();
}

void
loop_driver_t::add(loop_analysis_base_t *e)
{
    loop_analysis.push_back(e);

    if (e->want_depgroup())
        dep_analysis.push_back(e);

    if (e->want_pairs())
	pair_analysis.push_back(e);
}

stats_cptr_t
loop_driver_t::calculate_statistics(const location_set_t &pcs)
{
    return st_res->get(pcs, false);
}

loop_wrapper_list_ptr_t
loop_driver_t::traverse(loop_ptr_t loop, loop_vector_t &loop_stack)
{
    // This function does two major things:
    //
    // 1. look for pairs of "sibling" loops and call the registered
    //    methods that are interested in loop pairs
    // 2. Each individual loop is presented to all those engines that
    //    are interested in single loops.
    //
    // This function does an post-order traversal of loops and child loops,
    // propagating partial results up to previous recursion levels.


    loop_wrapper_list_ptr_t previous(new loop_wrapper_list_t);
    list<loop_wrapper_list_ptr_t> child_lists;

    // Add pc:s from this loop level, exclusive and inclusive loop
    // locations, and inclusive dependence group results (for the
    // benefit of loop fusion).
    loop_wrapper_ptr_t this_level(new loop_wrapper_t(loop));
    location_set_t &pc_set_exclusive = this_level->pc_set_exclusive;
    BOOST_FOREACH(depgroup_id_t depgroup_id,
                  loop_depgroups->loop_depgroups[loop->id])
        this_level->dep_groups.push_back(
            loop_depgroups->depgroups[depgroup_id]);

    // The inclusive set is initialized to being empty, and will be
    // incrementally populated as we recurse through all
    // children. Start populating the empty "inclusive" set by adding
    // the exclusive set for this loop.
    location_set_t &pc_set_inclusive = this_level->pc_set_inclusive;
    copy(pc_set_exclusive.begin(), pc_set_exclusive.end(),
	 insert_iterator<location_set_t>(pc_set_inclusive,
					    pc_set_inclusive.end()));

    // Recurse into children loops. Each recursion returns back a
    // structure with data about each encountered loop.
    loop_vector_t child_loops = loop->get_loops(false);

    // Use the child loops as minor steps if we are in the outer loop.
    int minor_step = 0;
    if (loop->id == 0)
	context->feedback->set_minor_steps((int) child_loops.size());
    BOOST_FOREACH(loop_ptr_t included_loop, child_loops) {
	context->feedback->test_cancel();

	loop_stack.push_back(included_loop);
	loop_wrapper_list_ptr_t child_list = traverse(included_loop,
						      loop_stack);
	loop_stack.pop_back();
	child_lists.push_back(child_list);

	// We maintain the union of all childrens' pc_sets. It will be
	// useful in the code motion block analysis, among other things.
	BOOST_FOREACH(loop_wrapper_ptr_t &c, *child_list) {
	    copy(c->pc_set_inclusive.begin(), c->pc_set_inclusive.end(),
		 insert_iterator<location_set_t>(pc_set_inclusive,
                                                 pc_set_inclusive.end()));
	}
	if (loop->id == 0)
	    context->feedback->update(++minor_step);
    }

    // Perform loop/loop comparison - if such engines are registered
    // and if there are any eligible loops to be compared.
    if (pair_analysis.size()) {
	while (!child_lists.empty()) {
	    // Compare each list from each of the children against the
	    // loops in the 'previous' list - those items from the
	    // children that we have already seen.
	    loop_wrapper_list_ptr_t this_child_list = child_lists.front();
	    child_lists.pop_front();

	    BOOST_FOREACH(loop_wrapper_ptr_t &c1, *previous) {
                loop_statistics_map_t::const_iterator c1_stat_it =
                    loop_statistics_map.find(c1->loop->id);
                assert(c1_stat_it != loop_statistics_map.end());
                const stats_cptr_t &c1_stat = c1_stat_it->second;

		BOOST_FOREACH(loop_wrapper_ptr_t &c2, *this_child_list) {
                    loop_statistics_map_t::const_iterator c2_stat_it =
                        loop_statistics_map.find(c2->loop->id);
                    assert(c2_stat_it != loop_statistics_map.end());
                    const stats_cptr_t &c2_stat = c2_stat_it->second;

		    context->feedback->test_cancel();
		    BOOST_FOREACH(loop_analysis_base_t *e, pair_analysis) {
			e->process_pair(c1, c2, loop, pc_set_exclusive,
					c1_stat, c2_stat);
			e->process_pair(c2, c1, loop, pc_set_exclusive,
					c2_stat, c1_stat);
		    }
		}
	    }

	    // Done with this child. Add its list items to the result, so that
	    // they can be evaluated against the other childrens' lists.
	    previous->splice(previous->end(), *this_child_list);
	}
    }

    // To sum up the loop pair analysis, append the excl pc:s from
    // this level, to appear as just one more child loop when
    // analyzing the situation one level up.  And note that in the
    // context of loop-pair analysis, it doesn't make sense to
    // evaluate this level's pc:s against those of its children.
    previous->push_back(this_level);


    // If this was the artificial top-level loop, we don't want to
    // perform analysis on it.
    if (loop->id == 0)
	return previous;

    // At this point we are done with loop-pair checking. Now to the
    // single loop engines.

    // make sure we haven't analysed this loop already (in some other
    // parent loop context)
    if (!loop_id_set.insert(loop->id).second)
	return previous;

    // Calculate statistics per loop
    stats_cptr_t stat = calculate_statistics(pc_set_exclusive);

    // Remember the statistics until when the loop is involved in a
    // loop pair analysis.
    loop_statistics_map.insert(make_pair(loop->id, stat));

    // loop entry - report to all!
    BOOST_FOREACH(loop_analysis_base_t *e, loop_analysis) {
	e->process_loop(loop, pc_set_exclusive, stat);
    }

    // Extract exclusive dependency groups and report
    BOOST_FOREACH(depgroup_id_t depgroup_id,
                  loop_depgroups->loop_depgroups[loop->id]) {
        const location_set_t &depgroup =
            loop_depgroups->depgroups[depgroup_id];
        stats_cptr_t dep_stat = calculate_statistics(depgroup);
        BOOST_FOREACH(loop_analysis_base_t *e, dep_analysis)
            e->process_dep(loop, pc_set_exclusive, stat,
                           depgroup_id, depgroup, dep_stat);
    }

    // loop exit - for dep-groups only
    BOOST_FOREACH(loop_analysis_base_t *e, dep_analysis)
	e->process_finish(loop, stat, loop_stack);

    return previous;
}

void
loop_driver_t::process()
{
    // Get all loops
    loop_identification_result_cptr_t loop_res = get_loops();

    // Push all toplevel loops under a virtual new all-inclusive top
    // level loop to tie everything together
    loop_ptr_t top(new loop_t(0));
    BOOST_FOREACH(const loop_ptr_t &loop, loop_res->loops) {
	if (loop->top_level)
	    top->cf_edges.push_back(
                cf_edge_t(cf_node_ptr_t(new cf_node_t(-1, loop)), 0.0));
    }

    loop_id_set.clear();

    // Run the recursive traversal. Single loop analysis will be
    // performed post-order.  Loop pair analysis will be performed on
    // each node with children.
    loop_vector_t loop_stack;
    traverse(top, loop_stack);
}

loop_identification_result_cptr_t
loop_driver_t::get_loops()
{
    return loop_identification_t(
        context, conf.sample_file, conf.stack_depth).get();
}

END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;
