/// @file analysis/issues/loop_engine.hh

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#ifndef LOOP_ENGINE_HH
#define LOOP_ENGINE_HH

#include "library/lang/common_macros.h"
#include "library/fenris/fnr_types.hh"
#include "analysis/call_stack/stack_pc_map.hh"
#include "analysis/structure/loop_depgroups.hh"
#include "analysis/structure/loop_identification.hh"
#include "analysis/statistics/statistics.hh"
#include "analysis/issues/issue.hh"

#include <boost/shared_ptr.hpp>

#include <vector>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);
BEGIN_NAMESPACE(issues);

using namespace std;
using namespace acumem::library;

using boost::shared_ptr;

/// Vessel for communicating loop related information during loop_engine runs
struct loop_wrapper_t {
    loop_wrapper_t(loop_ptr_t loop)
	: loop(loop)
	{
	    /// @todo - still copying sets here.
	    pc_set_exclusive = loop->get_locations(false);

	    // pc_set_inclusive will be built incrementally by the
	    // traverse function.

	    // The dep_groups will also be updated by the traverse
	    // logic when the sets are complete
	}

    /// Pointer to the loop structure
    loop_ptr_t loop;

    /// This is the exclusive pc set. It is rather cheap to
    /// compute from the 'loop', and seldom used. Or ?
    ///
    /// @todo XXX remove it from here if not needed
    location_set_t pc_set_exclusive;

    /// This is the inclusive pc list. Each time this loop is part
    /// of an dependency analysis, this set is needed. In total,
    /// it is needed N - 1 - P - C times, where N is number of
    /// loops, and P is number of direct parents and C is number
    /// of children and grand children.
    location_set_t pc_set_inclusive;

    /// Dependence groups in the excl pc set. Used in
    /// loop_driver_t::traverse, among other places
    location_set_vector_t dep_groups;
};

typedef shared_ptr<loop_wrapper_t> loop_wrapper_ptr_t;
typedef list<loop_wrapper_ptr_t> loop_wrapper_list_t;
typedef shared_ptr<loop_wrapper_list_t> loop_wrapper_list_ptr_t;


/// Interface and base class for loop level/dependency group analysis
/// engines.
///
/// The functions process_start_loop, process_finish_loop will be
/// called once for each loop, and process_dep_loop once for each
/// dependency group (found in each loop)
class loop_analysis_base_t {
public:
    virtual ~loop_analysis_base_t();

    /// loop_pc_set is the exclusive pc set of the loop
    virtual void process_loop(loop_ptr_t &loop,
                              const location_set_t &loop_pc_set,
			      const stats_cptr_t &loop_stat);

    /// loop_pc_set is the exclusive pc set of the loop
    virtual void process_dep(loop_ptr_t &loop,
                             const location_set_t &loop_pc_set,
                             const stats_cptr_t &loop_stat,
                             depgroup_id_t dep_id,
                             const location_set_t &dep_pc_set,
			     const stats_cptr_t &stat);

    virtual void process_finish(loop_ptr_t &loop,
                                const stats_cptr_t &loop_stat,
				const loop_vector_t &loop_stack);

    /// loop_pc_set is the exclusive pc set of the loop
    virtual void process_pair(const loop_wrapper_ptr_t &c1,
			      const loop_wrapper_ptr_t &c2,
			      const loop_ptr_t &parent,
                              const location_set_t &loop_pc_set,
			      const stats_cptr_t &c1_stat,
			      const stats_cptr_t &c2_stat);

    /// Declare if the implementation wants to do dependency group
    /// analysis. This enables the process_dep_loop and
    /// process_finish_loop callbacks
    virtual bool want_depgroup() const =0;

    /// Declare if the implementation wants to do analysis on pairs of
    /// loops in the context of their parent loop. This enables the
    /// process_loop_pair callback.
    virtual bool want_pairs() const;
};


/// Configuration for the loop_driver_t module.
class loop_driver_conf_t {
public:
    loop_driver_conf_t();

    loop_driver_conf_t(const string &sample_file,
                       const fnr_stack_depth_t stack_depth,
		       const fnr_line_size_t line_size,
                       const thread_mapping_cptr_t &thread_mapping,
		       const cpu_model_t &cpu_model,
		       const fnr_cache_size_t cache_size,
		       const repl_fr_engine_cptr_t &engine,
		       const windowing_engine_cptr_t &windowing,
		       const fnr_line_size_t page_size,
                       const int hw_pf_pos_streams,
                       const int hw_pf_neg_streams,
                       const fnr_line_count_t hw_pf_reach,
		       const bool use_old_hwpf_algorithm,
                       const bool hwpf_enabled,
                       const double comm_cost,
                       const double comm_util_threshold,
                       const int depgroup_limit_factor,
                       const double loop_length_factor,
                       const double random_threshold,
                       const double prefetch_cycles,
                       const double cache_miss_cycles,
                       const bool blocking_expert,
                       const issue_relevant_cptr_t issue_relevant,
                       const fnr_cache_size_t l1_cache_size);

    /// Sample file.
    string sample_file;

    /// Call stack depth to use in the analysis.
    fnr_stack_depth_t stack_depth;

    //@{
    /// Line size and cache size to optimize for.
    fnr_line_size_t line_size;
    thread_mapping_cptr_t thread_mapping;
    cpu_model_t cpu_model;
    fnr_cache_size_t cache_size;
    //@}

    //@{
    /// Miss ratio calculation parameters.
    repl_fr_engine_cptr_t engine;
    windowing_engine_cptr_t windowing;
    //@}

    /// Page size needed for randomness and hw-prefetch analysis
    fnr_line_size_t page_size;

    ///@{
    /// Hardware prefetch analysis parameters.
    int hw_pf_pos_streams;
    int hw_pf_neg_streams;
    fnr_line_count_t hw_pf_reach;
    ///@}

    /// Use the old hwpf algorithm
    bool use_old_hwpf_algorithm;

    bool hwpf_enabled;

    /// Cost of communication in replacement miss equivalents.
    double comm_cost;

    /// Utilization threshold for reporting communication utilization issues.
    double comm_util_threshold;

    /// Dependence group limit factor.
    int depgroup_limit_factor;

    /// When comparing loop length with average reuse distance, this factor
    /// describes the extra reuse distance allowed too still be considered to
    /// be within the loop limit. Typical values between 1 and 5.
    double loop_length_factor;

    /// The random ratio above which random issues are generated. Random ratios
    /// above this threshold also prevent generation of loop order issues.
    double random_threshold;

    /// Estimated cost of a prefetch instruction in processor cycles.
    double prefetch_cycles;

    /// Estimated cost of a cache miss in processor cycles.
    double cache_miss_cycles;

    /// Flag to tell whether to enable the expert mode, in which the
    /// code motion barriers are disregarded
    bool blocking_expert;

    /// Check if an issue is related to enough fetches, upgrades or write-backs
    /// to be relevant.
    issue_relevant_cptr_t issue_relevant;

    /// L1 cache size. Currently only used by the prefetchnta analysis, which
    /// needs to determine which samples cause fetches in the L1 cache, even
    /// though the rest of the analysis is done on the last level cache.
    fnr_cache_size_t l1_cache_size;
};


/// This class extracts loops and calls registered analysis engines
/// per loop.
class loop_driver_t {
public:
    loop_driver_t(const feature_context_ptr_t &c);
    void set_configuration(const loop_driver_conf_t &c);
    void add(loop_analysis_base_t *e);

    void process();

private:
    feature_context_ptr_t context;
    loop_driver_conf_t conf;

    typedef vector<loop_analysis_base_t *> loop_analysis_vector_t;

    /// Includes all loops
    loop_analysis_vector_t loop_analysis;

    /// Includes those that require dep.group analysis
    loop_analysis_vector_t dep_analysis;

    /// Include those that require pair-wise loop analysis
    loop_analysis_vector_t pair_analysis;

    /// Some loops may appear in more than one location. This set
    /// remembers which loops we have already seen, so we don't
    /// perform analysis more than once.
    typedef set<int> loop_id_set_t;
    loop_id_set_t loop_id_set;

    /// Remember loop statistics when it's calculated (as a leaf)
    /// to be able to reuse it when doing loop pair analysis.
    typedef map<int, stats_cptr_t> loop_statistics_map_t;
    loop_statistics_map_t loop_statistics_map;

    stats_cptr_t calculate_statistics(const location_set_t &pcs);

    loop_wrapper_list_ptr_t traverse(loop_ptr_t loop,
				     loop_vector_t &loop_stack);
    loop_identification_result_cptr_t get_loops();

    // Cached statistics result.
    statistics_result_cptr_t st_res;

    // Cached dependency group result.
    loop_depgroups_result_cptr_t loop_depgroups;
};

END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;

#endif
