/// @file analysis/issues/loop_fusion.cc

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

// Implementation idea:
//
// * Get loops.
//
// * Traverse loops post order, from outer to inner
//
//   - If no children
//       calculate depgroups and
//       return list of (depgroups and pc-set) outwards
//
//   - If children
//       recurse into each child, take output (list of [depgroup, pc-set])
//       do loop-fusion analysis for each pair of [depgroup, pc-set] from
//          different child loops, given the current level as parent.
//       collect pc-set and depgroups from the current level.
//       add together lists of [depgroup and pc-set] from all children and
//         the current level.
//       return this list to the next level up.
//
// * When evaluating loop A against loop B, check that:
//
//   - The average number of samples per access in the two loops is
//     about the same. That should help only comparing loops that are
//     at the same level.
//
//   - Obviously there has to be temporal (or spatial) reuse between
//     the two loops.
//
//   - The reuse distance for these reuses shall be significant. This
//     means that the number of misses for the pc2 shall be over a
//     certain limit
//
//     Evaluate accesses on the path between the two loops for w->r,
//     r->w and w->w dependencies. The goal is to see if all
//     instructions of loop B (including children) can be moved back
//     into A, or if all instructions of loop A (including its
//     children) can be moved forwards into loop B, or that both loops
//     can be moved towards each other and fuse in the middle
//     somewhere.
//
//       First evaluate pc -> pc+1 links from all instructions of loop
//       A. Terminate the search when the accumulated reuse distance
//       exceeds the reuse distance between the two loops. Also
//       consider limiting the search to instructions within siblings
//       of a common parent.
//
//       This gives us a set of pc:s to be used both forwards and
//       backwards. Call the "spanning tree" (bad name) on this set
//       for each address in loop A. This will give us potential
//       limits for the movement of loop A. Same for loop B, except
//       backwards.
//
//       A read operation on the loop cannot be moved past a write
//       operation. A write operation cannot be moved across a read
//       operation on the path towards the other loop.

#include <boost/foreach.hpp>

#include "analysis/structure/access_lookup.hh"
#include "analysis/structure/dep_tree_search.hh"
#include "loop_fusion.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);
BEGIN_NAMESPACE(issues);

using namespace std;
using namespace boost;
using namespace acumem::library;

// ----- loop_fusion_t --------------------------------------------------------

loop_fusion_t::loop_fusion_t(const feature_context_ptr_t &c)
    : context(c)
{
    loop_fusion_result.reset(new loop_fusion_result_t);
}

loop_fusion_result_ptr_t
loop_fusion_t::get()
{
    return loop_fusion_result;
}

void
loop_fusion_t::set_configuration(loop_driver_conf_t const &c)
{
    conf = c;

    // Get pc lookup.
    pcl_res = pc_lookup_t(
        context, conf.sample_file, conf.stack_depth, conf.line_size,
	conf.page_size, conf.thread_mapping, conf.cpu_model, Mapped_Line).get();

    // Get access types.
    al_res = access_lookup_t(context, conf.sample_file,
                             conf.stack_depth).get();

    // Get dependence groups.
    loop_depgroups = loop_depgroups_t(
        context, conf.sample_file, conf.stack_depth, conf.line_size,
	conf.page_size, conf.thread_mapping, conf.cpu_model, Mapped_Line,
	conf.depgroup_limit_factor).get();

    locationx_mapper_cptr_t lxm(
	new locationx_mapper_t(context,
			       conf.sample_file,
			       conf.stack_depth,
			       conf.thread_mapping,
			       true, false, false));

    stat_res = statistics_t(
        context, conf.sample_file, conf.stack_depth, conf.line_size,
        conf.thread_mapping, conf.cpu_model, lxm, conf.cache_size,
	conf.engine, conf.windowing, conf.page_size,
	conf.hw_pf_pos_streams, conf.hw_pf_neg_streams,
	conf.hw_pf_reach, conf.use_old_hwpf_algorithm, conf.hwpf_enabled).get();
}

bool
loop_fusion_t::want_pairs() const
{
    return true;
}

bool
loop_fusion_t::want_depgroup() const
{
    return false;
}

/// Returns true if there are instructions between loop1_locs and loop2_locs
/// that prevent fusing the loops, and adds those instructions to the supplied
/// blockers map.
bool
loop_fusion_t::find_blockers(const location_set_t &loop1_locs,
                             const location_set_t &loop2_locs,
                             bool backwards,
                             fnr_time_t max_rdist,
                             location_set_map_t &blocker_map) const
{
    bool blocker_found = false;

    BOOST_FOREACH(const location_t &loc, loop1_locs) {
        // Calculate which access types would block a fusion.
        loc_access_info_map_t::const_iterator access_classes_it =
            al_res->loc_access_info.find(loc);

        assert(access_classes_it != al_res->loc_access_info.end());
        fnr_access_classes_t access_classes =
	    access_classes_it->second.access_classes;
        fnr_access_filter_t blocker_filter = NULL;

        if (access_classes & FNR_Access_Class_PF)
	    continue;
        else if (access_classes & FNR_Access_Class_WR)
            blocker_filter = &fnr_access_filter_rw;
	else if (access_classes & FNR_Access_Class_RD)
            blocker_filter = &fnr_access_filter_wr_wrnta;
	else
	    continue;

        // Calculate blockers.
        dep_tree_search_result_cptr_t dts_res = dep_tree_search_access_t(
            context, conf.sample_file, conf.stack_depth, conf.line_size,
	    conf.page_size, conf.thread_mapping, conf.cpu_model, Mapped_Byte,
	    blocker_filter, backwards, true, loc, max_rdist).get();

        location_set_t &blockers = blocker_map[loc];
	blockers.clear();
	BOOST_FOREACH(const location_time_map_t::value_type &v,
		      dts_res->location_time_map)
	    blockers.insert(v.first);

        // Remove the loops themselves from the blockers.
	BOOST_FOREACH(const location_t &loop1_loc, loop1_locs)
	    blockers.erase(loop1_loc);
	BOOST_FOREACH(const location_t &loop2_loc, loop2_locs)
	    blockers.erase(loop2_loc);

        // Keep track of if we have seen any blockers.
	if (!blockers.empty())
            blocker_found = true;
    }

    return blocker_found;
}

/// Evaluate if any access in c1 points to c2, and in that case look
/// for code motion blocks in the scope of the common_parent
void
loop_fusion_t::process_pair(const loop_wrapper_ptr_t &c1,
			    const loop_wrapper_ptr_t &c2,
			    const loop_ptr_t &common_parent,
			    const location_set_t &loop_pc_set,
			    const stats_cptr_t &c1_stat,
			    const stats_cptr_t &c2_stat)
{
    // Exclusive location sets of the loops.
    const location_set_t *c1_locs = &c1->pc_set_exclusive;
    const location_set_t *c2_locs = &c2->pc_set_exclusive;

    // Make sure that we haven't already analyzed these two loops
    pair<seen_loop_pair_set_t::iterator, bool> inserted =
	seen_loop_pair_set.insert(make_pair(c1->loop->id,
					    c2->loop->id));
    if (!inserted.second)
	return;

    // Make sure the loops are not the same loop.
    if (c1->loop->id == c2->loop->id)
	return;

    // Make sure that the loops are not (grand-) parent/and child
    if (c1->loop->has_child_loop(c2->loop, true) ||
        c2->loop->has_child_loop(c1->loop, true))
	return;

    // Both loops have non-child instructions, right?
    if (c1_locs->empty() || c2_locs->empty())
	return;

    // Same magnitude of number of iterations?
    double c1_count = c1->loop->average_sample_count;
    double c2_count = c2->loop->average_sample_count;
    if (c1_count == 0.0 || c2_count == 0.0)
	return;
    if (c1_count / c2_count > 10 || c2_count / c1_count > 10)
	return;

    // Find reuses from c1 to c2 and the maximum reuse distance among
    // these reuses. Also look for other locations acting as pc1 in
    // case those conflict with our loop c1.
    typedef pair<location_t, location_t> location_pair_t;
    typedef vector<location_pair_t> reuse_pair_vector_t;
    reuse_pair_vector_t reuse_pairs;
    reuse_pair_vector_t additional_pc1_locations;
    reuse_pair_vector_t additional_pc2_locations;

    fnr_time_t max_rdist = 0;
    BOOST_FOREACH(location_t pc2, *c2_locs) {

	// Get ourself an iterator to all samples targeting pc2
	pc_info_map_t::const_iterator pc2_it = pcl_res->pc_info.find(pc2);
	if (pc2_it != pcl_res->pc_info.end()) {
	    bool seen_proper_pc1 = false;
	    vector<location_t> from_other;

	    // Look at all corresponding pc1 addresses, divide them into
	    // those that occur in loop c1, and those that come from
	    // other places
            BOOST_FOREACH(const per_pc1_t &per_pc1, pc2_it->second.per_pc1) {
                location_t pc1 = per_pc1.pc1;

                if (c1_locs->find(pc1) != c1_locs->end()) {
		    reuse_pairs.push_back(make_pair(pc1, pc2));
                    max_rdist = max(max_rdist, per_pc1.max_rdist);
		    seen_proper_pc1 = true;
                } else if (c2_locs->find(pc1) == c2_locs->end()) {
		    from_other.push_back(pc1);
		}
            }

	    // If 'from_other' contains anything, then this is not a
	    // pure 1-to-1 fusion, but rather n-to-1. If not the
	    // expert flag is set, abort here. Otherwise continue and
	    // record those other places.
	    if (seen_proper_pc1 && from_other.size()) {
		if (!conf.blocking_expert)
		    return;

		BOOST_FOREACH(location_t other_pc1, from_other) {
		    additional_pc1_locations.push_back(make_pair(other_pc1,
								 pc2));
		}
	    }
        }
    }

    // Return if there are no connecting samples.
    if (reuse_pairs.size() == 0)
	return;

    // Now a similar but slightly limited scan (since we have already
    // found some connecting samples) in the opposite direction to
    // find 1-to-n fusions
    BOOST_FOREACH(location_t pc1, *c1_locs) {

	// Get ourself an iterator to all samples targeting pc1
	pc_info_map_t::const_iterator pc1_it = pcl_res->pc_info.find(pc1);
	if (pc1_it != pcl_res->pc_info.end()) {
	    bool seen_proper_pc2 = false;
	    vector<location_t> to_other;

	    // Look at all corresponding pc2 addresses, divide them
	    // into those that occur in loop c2, and those that go to
	    // other places
            BOOST_FOREACH(const per_pc2_t &per_pc2, pc1_it->second.per_pc2) {
                location_t pc2 = per_pc2.pc2;

                if (c2_locs->find(pc2) != c2_locs->end()) {
		    seen_proper_pc2 = true;
                } else if (c1_locs->find(pc1) == c1_locs->end()) {
		    to_other.push_back(pc2);
		}
            }

	    // If 'to_other' contains anything, then this is not a
	    // pure 1-to-1 fusion, but rather 1-to-n. If not the
	    // expert flag is set, abort here. Otherwise continue and
	    // record those other places.
	    if (seen_proper_pc2 && to_other.size()) {
		if (!conf.blocking_expert)
		    return;

		BOOST_FOREACH(location_t other_pc2, to_other) {
		    additional_pc2_locations.push_back(make_pair(pc1,
								 other_pc2));
		}
	    }
        }
    }



    // Look for blockers from c1 forwards to c2.
    location_set_map_t block_for_moving_loop1_down;
    bool move_loop1_down =
	!find_blockers(c1->pc_set_inclusive, c2->pc_set_inclusive,
		       false, max_rdist, block_for_moving_loop1_down);

    // Look for blockers from c2 backwards to c1,
    location_set_map_t block_for_moving_loop2_up;
    bool move_loop2_up =
	!find_blockers(c2->pc_set_inclusive, c1->pc_set_inclusive,
		       true, max_rdist, block_for_moving_loop2_up);

    // If neither loop can be moved, just return here (unless in
    // expert mode)
    //
    /// @todo In case the loops can't be fused right away, look for a
    /// point somewhere in between where a new loop could be located.
    if (!move_loop1_down && !move_loop2_up)
	if (!conf.blocking_expert)
	    return;

    // Report results.

    // This is where we look at the reuse_pairs and try to assign them
    // to the corresponding dependency groups. This is needed to
    // correctly calculate spatial utilization for the loop2 instructions.

    typedef pair<depgroup_id_t, depgroup_id_t> depgroup_pair_t;
    typedef map<depgroup_pair_t, reuse_pair_vector_t> depgroup_pair_map_t;
    depgroup_pair_map_t sep_map;

    // Locate the dependence group pairs containing the instructions with
    // reuses between the loops.
    BOOST_FOREACH(location_pair_t &v, reuse_pairs) {
        depgroup_id_t depgroup1_id;
        for (depgroup1_id = 0;
             depgroup1_id < (depgroup_id_t)loop_depgroups->depgroups.size();
             depgroup1_id++)
            if (loop_depgroups->depgroups[depgroup1_id].find(v.first) !=
                loop_depgroups->depgroups[depgroup1_id].end())
                goto depgroup1_found;
        assert(0);
    depgroup1_found:

        depgroup_id_t depgroup2_id;
        for (depgroup2_id = 0;
             depgroup2_id < (depgroup_id_t)loop_depgroups->depgroups.size();
             depgroup2_id++)
            if (loop_depgroups->depgroups[depgroup2_id].find(v.second) !=
                loop_depgroups->depgroups[depgroup2_id].end())
                goto depgroup2_found;
        assert(0);
    depgroup2_found:

        sep_map[make_pair(depgroup1_id, depgroup2_id)].push_back(v);
    }

    // Loop over all dependence group pairs that we found to be interesting,
    // and create an issue for each one.
    BOOST_FOREACH(depgroup_pair_map_t::value_type &v, sep_map) {
        depgroup_id_t depgroup1_id = v.first.first;
        depgroup_id_t depgroup2_id = v.first.second;
	reuse_pair_vector_t &rpv = v.second;

        // Ignore the issue if it causes to few fetches/upgrades/write-backs to
        // be relevant.
        if (!conf.issue_relevant->issue_relevant(
                stat_res->get(loop_depgroups->depgroups[depgroup2_id], false)))
            continue;

	// Create the issue.
	loop_fusion_issue_ptr_t loop_fusion_issue(new loop_fusion_issue_t());
	loop_fusion_issue->move_loop1_down = move_loop1_down;
	loop_fusion_issue->move_loop2_up = move_loop2_up;
	loop_fusion_issue->loop1_id = c1->loop->id;
	loop_fusion_issue->loop2_id = c2->loop->id;
        loop_fusion_issue->depgroup1_id = depgroup1_id;
        loop_fusion_issue->depgroup2_id = depgroup2_id;
	loop_fusion_issue->block_for_moving_loop1_down =
	    block_for_moving_loop1_down;
	loop_fusion_issue->block_for_moving_loop2_up =
	    block_for_moving_loop2_up;

	// Add all interesting locations for the reuse pairs.
	BOOST_FOREACH(location_pair_t &rp, rpv) {
	    loop_fusion_issue->loop1_pcs.insert(rp.first);
	    loop_fusion_issue->loop2_pcs.insert(rp.second);
	}

	// Add any n-to-1 sources and 1-to-n destinations
	BOOST_FOREACH(location_pair_t &other, additional_pc1_locations) {
	    loop_fusion_issue->additional_pc1.insert(other.first);
	}
	BOOST_FOREACH(location_pair_t &other, additional_pc2_locations) {
	    loop_fusion_issue->additional_pc2.insert(other.second);
	}

	// add the issue
	loop_fusion_result->loop_fusion_vector.push_back(loop_fusion_issue);
    }
}

END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;
