/// @file analysis/issues/loop_fusion.hh

// Copyright (c) 2006-2009 Acumem AB.  All rights reserved.

#ifndef LOOP_FUSION_HH
#define LOOP_FUSION_HH

#include "library/lang/common_macros.h"

#include "analysis/feature_base/feature_base.hh"
#include "analysis/feature_base/workspace.hh"
#include "analysis/structure/access_lookup.hh"
#include "analysis/structure/pc_lookup.hh"
#include "analysis/structure/dependence_group.hh"

#include "issue.hh"
#include "loop_engine.hh"

#include <boost/shared_ptr.hpp>

#include <vector>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);
BEGIN_NAMESPACE(issues);

using namespace std;
using namespace acumem::library;

using boost::shared_ptr;

/// This is the information reported per loop for loop fusion.
struct pc_loop_info_t {
    double spatial_util;
};

typedef map<location_t, pc_loop_info_t> loop_info_map_t;

/// Information about loop fusion results.
struct loop_fusion_result_t {
    /// This map reports spatial usage per loop
    loop_info_map_t loop_info_map;

    /// This vector lists the given advice: For each fusion
    /// opportunity, it names the first and second loop and the address
    /// in the first loop that has a relation to the second loop.
    loop_fusion_issue_vector_t loop_fusion_vector;
};

typedef shared_ptr<loop_fusion_result_t> loop_fusion_result_ptr_t;

/// Module for finding loop fusion issues
class loop_fusion_t : public loop_analysis_base_t {
public:
    loop_fusion_t(const feature_context_ptr_t &c);
    void set_configuration(loop_driver_conf_t const &c);
    loop_fusion_result_ptr_t get();

    virtual void process_pair(const loop_wrapper_ptr_t &c1,
			      const loop_wrapper_ptr_t &c2,
			      const loop_ptr_t &parent,
			      const location_set_t &loop_pc_set,
			      const stats_cptr_t &c1_stat,
			      const stats_cptr_t &c2_stat);

    bool want_pairs() const;
    bool want_depgroup() const;

private:
    /// Returns true if there are instructions between loop1_locs and
    /// loop2_locs that prevent fusing the loops, and adds those instructions
    /// to the supplied blockers map.
    bool find_blockers(const location_set_t &loop1_locs,
                       const location_set_t &loop2_locs,
                       bool backwards,
                       fnr_time_t max_rdist,
                       location_set_map_t &blockers) const;

    feature_context_ptr_t context;
    loop_driver_conf_t conf;

    /// Pre-calculated results from set_configuration.
    access_lookup_result_cptr_t al_res;
    pc_lookup_result_cptr_t pcl_res;
    loop_depgroups_result_cptr_t loop_depgroups;
    statistics_result_cptr_t stat_res;

    /// Total results.
    loop_fusion_result_ptr_t loop_fusion_result;

    /// Remember what pairs of loops we have already analyzed. Loops
    /// may appear several times, due to child loops appearing under
    /// different parent loops.
    typedef pair<int, int> loop_pair_t;
    typedef set<loop_pair_t> seen_loop_pair_set_t;
    seen_loop_pair_set_t seen_loop_pair_set;
};

END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;

#endif
