/// @file analysis/issues/loop_order.cc

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

// Loop nesting analysis, see doc/notes/loop_nesting.txt for details.


#include "loop_order.hh"

#include "analysis/feature_base/exceptions.hh"
#include "analysis/structure/dep_tree_search.hh"

#include <boost/foreach.hpp>

#include <iostream>
#include <iterator>
#include <algorithm>

using namespace std;
using namespace boost;
using namespace acumem::library;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);
BEGIN_NAMESPACE(issues);

// ----- loop_order_t ---------------------------------------------------------

loop_order_t::loop_order_t(const feature_context_ptr_t &c)
    : context(c)
{
}

void
loop_order_t::set_configuration(loop_driver_conf_t const &c)
{
    conf = c;
    loop_order_result.reset(new loop_order_result_t);


    pcl_byte_res = pc_lookup_t(
	context, conf.sample_file, conf.stack_depth, conf.line_size,
	conf.page_size, conf.thread_mapping, conf.cpu_model, Mapped_Byte).get();

    pcl_line_res = pc_lookup_t(
	context, conf.sample_file, conf.stack_depth, conf.line_size,
	conf.page_size, conf.thread_mapping, conf.cpu_model, Mapped_Line).get();
}

loop_order_result_ptr_t
loop_order_t::get()
{
    return loop_order_result;
}

void
loop_order_t::process_dep(loop_ptr_t &loop,
			  const location_set_t &loop_pc_set,
                          const stats_cptr_t &loop_stat,
                          depgroup_id_t dep_id,
			  const location_set_t &dep_pc_set,
			  const stats_cptr_t &stat)
{
    // Return if there is nothing to report, or if the issues causes to few
    // fetches, upgrades and write-backs to be relevant.
    if (!inefficient_loop_nesting(dep_pc_set, loop) ||
        !conf.issue_relevant->issue_relevant(stat))
        return;

    // Don't report the loop nesting problem if the access pattern is too
    // irregular.
    if (stat->random_ratio > conf.random_threshold)
        return;

    // Create the issue.
    loop_order_issue_ptr_t issue(new loop_order_issue_t());
    issue->loop_id = loop->id;
    issue->depgroup_id = dep_id;

    // Dig out code motion boundaries.

    /// @todo XXX consider tracking dependencies for each address
    /// separately, and not for the entire dependency group.
    dep_tree_search_result_cptr_t dts_res = dep_tree_search_access_t(
        context, conf.sample_file, conf.stack_depth, conf.line_size,
	conf.page_size, conf.thread_mapping, conf.cpu_model, Mapped_Byte,
	&fnr_access_filter_wr_wrnta, true, true, dep_pc_set,
	(fnr_time_t)-1).get();

    BOOST_FOREACH(const location_time_map_t::value_type &v,
                  dts_res->location_time_map)
	issue->previous_writers.insert(v.first);

    // Add the issue to the result.
    loop_order_result->loop_order_issues.push_back(issue);
}

bool
loop_order_t::want_depgroup() const
{
    return true;
}

void
loop_order_t::get_long_rdists(const location_set_t &pc_set,
			      pc_lookup_result_cptr_t &pcl_res,
			      fnr_time_t min_rdist,
			      fnr_time_vector_t &rdists)
{

    // Collect the long reuse distances.
    BOOST_FOREACH(location_t pc, pc_set) {
	pc_info_map_t::const_iterator pc2_it = pcl_res->pc_info.find(pc);
	if (pc2_it != pcl_res->pc_info.end()) {
	    BOOST_FOREACH(const per_pc1_t &per_pc1, pc2_it->second.per_pc1)
                if (pc_set.find(per_pc1.pc1) != pc_set.end()) {
                    BOOST_FOREACH(const mapped_repl_base_t *s,
				  per_pc1.samples) {
                        fnr_time_t rdist = s->rdist();
                        if (rdist >= min_rdist)
                            rdists.push_back(rdist);
                    }
                }
        }
    }

    // Sort the long reuse distances.
    sort(rdists.begin(), rdists.end());
}

bool
loop_order_t::inefficient_loop_nesting(const location_set_t &pc_set,
                                       loop_ptr_t loop)
{
    // Minimum reuse distance to classify a sample as a long sample.
    fnr_time_t min_rdist =
        (fnr_time_t)(loop->average_loop_length * conf.loop_length_factor);

    // Get long line size reuse distances;
    fnr_time_vector_t line_rdists;
    get_long_rdists(pc_set, pcl_line_res, min_rdist, line_rdists);

    // Get long byte size reuse distances;
    fnr_time_vector_t byte_rdists;
    get_long_rdists(pc_set, pcl_byte_res, min_rdist, byte_rdists);

    // Check that we have both long line size samples and byte size samples.
    if (line_rdists.empty() || byte_rdists.empty())
        return false;

    // See doc/notes/loop_nesting.txt for details about these conditions. */
    return
        // Number of long line size reuses roughly the same as the number of
        // inner loop iterations. */
        line_rdists.size() > loop->average_sample_count * 0.5 &&
        // Median line size reuse distance, representing the dominating line
        // size reuse distance.
        line_rdists[line_rdists.size() * 5 / 10] <
        // 75% of the 90th percentile byte size reuse distance, representing
        // the length of one execution of the nested loops.
        byte_rdists[byte_rdists.size() * 9 / 10] * 0.75;
}

END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;
