/// @file analysis/issues/prefetch.cc

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#include "analysis/issues/prefetch.hh"
#include "analysis/thread_mapping/sample_mapping.hh"

#include <boost/lexical_cast.hpp>
#include <boost/foreach.hpp>

#include <iostream>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);
BEGIN_NAMESPACE(issues);

using namespace std;
using namespace boost;
using namespace acumem::library;

// Read/Prefetch/Prefetch Write/Read NTA
static bool
access_filter(fnr_access_classes_t ac)
{
    return (ac & (FNR_Access_Class_RD | FNR_Access_Class_WR)) &&
	!((ac & FNR_Access_Class_NTA) && (ac & FNR_Access_Class_WR));
}

// ----------------------------------------------------------------------

prefetch_t::prefetch_t(const feature_context_ptr_t &context)
    : context(context),
      res(new prefetch_result_t)
{
}

void
prefetch_t::set_configuration(loop_driver_conf_t const &c)
{
    conf = c;

    sample_mapping = sample_mapping_t(
        context, conf.sample_file, conf.stack_depth, conf.line_size,
        conf.page_size, conf.thread_mapping, conf.cpu_model).get();

    fr_res = repl_fr_line_t(
	context, conf.sample_file, conf.stack_depth,
	conf.line_size, conf.page_size, conf.cache_size, conf.thread_mapping,
	conf.cpu_model, conf.engine).get();

    // Create maps from pc1s and pc2s of prefetches to the samples begun/ended
    // by the prefetches. Also calculate the replacement fetch ratio of each
    // prefetch instruction.
    burst_info.resize(sample_mapping->bursts.size());
    sample_mapping_burst_vector_t::const_iterator it_smb(
	sample_mapping->bursts.begin());
    repl_fr_line_burst_vector_t::const_iterator it_repl_fr(
	fr_res->bursts.begin());
    BOOST_FOREACH(burst_info_t &bi, burst_info) {
	const mapped_repl_line_vector_t &repl_samples((it_smb++)->repl_line);
	const ratio_vector_t &repl_frs((it_repl_fr++)->val_to_end);
	const fnr_sample_count_t num_samples((fnr_sample_count_t)repl_samples.size());

	for (fnr_sample_count_t i = 0; i < num_samples; i++) {
	    const mapped_repl_line_t &s(repl_samples[i]);
	    if (s.valid() && s.valid()->is_prefetch())
		bi.pc1_pf_samples[s.valid()->pc].push_back(i);
	    const mapped_access_t &pc2(*s.end());
	    if (pc2.is_prefetch()) {
		bi.pc2_pf_samples[pc2.pc].push_back(i);
		bi.pf_repl_fr[pc2.pc] += repl_frs[i];
	    }
	}

	BOOST_FOREACH(loc_repl_map_t::value_type &v, bi.pf_repl_fr)
	    v.second /= bi.pc2_pf_samples[v.first].size();
    }
}

prefetch_result_ptr_t
prefetch_t::get()
{
    return res;
}

void
prefetch_t::process_dep(loop_ptr_t &loop,
			const location_set_t &loop_pc_set,
                        const stats_cptr_t &loop_stat,
                        depgroup_id_t dep_id,
			const location_set_t &dep_pc_set,
			const stats_cptr_t &stat)
{
    BOOST_FOREACH(location_t pc, dep_pc_set) {
	// Find unnecessary prefetches that fetch data that is already in the
	// cache.
	analyze_pf_hits(pc, loop->id, dep_id);

        // Find prefetches that are too far from the instruction using the fetched
	// data.
	analyze_pf_too_far(pc, loop->id, dep_id);

	// Find prefetches that are too close to the instruction using the fetched
	// data.
	analyze_pf_too_close(pc, loop->id, dep_id);
    }
}

bool
prefetch_t::want_depgroup() const
{
    return true;
}

void
prefetch_t::analyze_pf_hits(location_t pc,
			    loop_id_t loop_id,
			    depgroup_id_t depgroup_id)
{
    double prefetches(0.0);
    double repl_fr(0.0);

    sample_mapping_burst_vector_t::const_iterator it_smb(
	sample_mapping->bursts.begin());
    BOOST_FOREACH(const burst_info_t &bi, burst_info) {
	const double sample_period((it_smb++)->sample_period);
	pf_pc_samples_map_t::const_iterator psi(bi.pc2_pf_samples.find(pc));
	if (psi != bi.pc2_pf_samples.end()) {
	    const sample_index_vector_t &indexes(psi->second);
	    prefetches += indexes.size() * sample_period;
	    repl_fr += bi.pf_repl_fr.find(pc)->second;
	}
    }

    if (prefetches > 0.0) {
	repl_fr /= burst_info.size();
	const double misses(prefetches * repl_fr);
	const double cost =
	    (prefetches * conf.prefetch_cycles / conf.cache_miss_cycles -
	     misses);

	if (conf.issue_relevant->issue_relevant_fetches(cost))
	    res->issues.push_back(
		create_pf_issue(Prefetch_Hits,
				pc,
				loop_id,
				depgroup_id,
				cost,
				repl_fr));
    }
}

void
prefetch_t::analyze_pf_too_far(location_t pc,
			       loop_id_t loop_id,
			       depgroup_id_t depgroup_id)
{
    double prefetches(0.0);
    double fetches(0.0);

    sample_mapping_burst_vector_t::const_iterator it_smb(
	sample_mapping->bursts.begin());
    repl_fr_line_burst_vector_t::const_iterator it_repl_fr(
	fr_res->bursts.begin());
    BOOST_FOREACH(const burst_info_t &bi, burst_info) {
	const double sample_period((it_smb++)->sample_period);
	const ratio_vector_t &repl_frs((it_repl_fr++)->val_to_end);
	pf_pc_samples_map_t::const_iterator psi(bi.pc1_pf_samples.find(pc));
	if (psi != bi.pc1_pf_samples.end()) {
	    const sample_index_vector_t &indexes(psi->second);
	    prefetches += indexes.size() * sample_period;
	    BOOST_FOREACH(const fnr_sample_count_t &ci, indexes)
		fetches += repl_frs[ci] * sample_period;
	}
    }

    if (conf.issue_relevant->issue_relevant_fetches(fetches))
	res->issues.push_back(
	    create_pf_issue(Prefetch_Too_Far,
			    pc,
			    loop_id,
			    depgroup_id,
			    fetches /* cost */,
			    fetches / prefetches /* repl_fr */));
}

void
prefetch_t::analyze_pf_too_close(location_t pc,
				 loop_id_t loop_id,
				 depgroup_id_t depgroup_id)
{
    double cost(0.0);
    fnr_time_vector_t rdists;
    sample_mapping_burst_vector_t::const_iterator it_smb(
	sample_mapping->bursts.begin());
    BOOST_FOREACH(const burst_info_t &bi, burst_info) {
	const sample_mapping_burst_t &smb(*it_smb++);
	const double sample_period(smb.sample_period);

	// make sure the pc occurs as pc1...
	pf_pc_samples_map_t::const_iterator psi(bi.pc1_pf_samples.find(pc));
	if (psi != bi.pc1_pf_samples.end()) {

	    // ... as well as pc2
	    loc_repl_map_t::const_iterator loc_repl_i(bi.pf_repl_fr.find(pc));
	    if (loc_repl_i != bi.pf_repl_fr.end()) {

		const sample_index_vector_t &indexes(psi->second);
		const double pf_repl_fr(loc_repl_i->second);
		BOOST_FOREACH(const fnr_sample_count_t &si, indexes) {
		    const mapped_repl_line_t &sample(smb.repl_line[si]);
		    if (sample.valid()) {
			fnr_time_t rdist(sample.rdist());
			rdists.push_back(rdist);

			/// @todo Assumes a memory access every 3 cycles.
			double stall_cost =
			    pf_repl_fr * (conf.cache_miss_cycles - rdist * 3)
			    / conf.cache_miss_cycles * sample_period;
			if (stall_cost > 0.0)
			    cost += stall_cost;
		    }
		}
	    }
	}
    }

    if (conf.issue_relevant->issue_relevant_fetches(cost)) {
	sort(rdists.begin(), rdists.end());
	fnr_time_t median_rdist = rdists[rdists.size() / 2];
	res->issues.push_back(
	    create_pf_issue(Prefetch_Too_Close,
			    pc,
			    loop_id,
			    depgroup_id,
			    cost,
			    median_rdist));
    }
}

prefetch_issue_ptr_t
prefetch_t::create_pf_issue(prefetch_problem_t problem,
			    location_t pc,
			    loop_id_t loop_id,
			    depgroup_id_t depgroup_id,
			    double cost,
			    double fetch_ratio)
{
    dep_tree_search_result_cptr_t prev =
	dep_tree_search_access_t(
	    context, conf.sample_file, conf.stack_depth, conf.line_size,
	    conf.page_size, conf.thread_mapping, conf.cpu_model, Mapped_Line,
	    &access_filter, true, true, pc, (fnr_time_t)-1).get();

    dep_tree_search_result_cptr_t next =
	dep_tree_search_access_t(
	    context, conf.sample_file, conf.stack_depth, conf.line_size,
	    conf.page_size, conf.thread_mapping, conf.cpu_model, Mapped_Line,
	    &access_filter, false, true, pc, (fnr_time_t)-1).get();

    return prefetch_issue_ptr_t(
	new prefetch_issue_t(problem,
			     pc,
			     loop_id,
			     depgroup_id,
			     cost,
			     fetch_ratio,
			     prev->locations(),
			     next->locations()));
}

prefetch_issue_ptr_t
prefetch_t::create_pf_issue(prefetch_problem_t problem,
			    location_t pc,
			    loop_id_t loop_id,
			    depgroup_id_t depgroup_id,
			    double cost,
			    fnr_time_t median_rdist)
{
    dep_tree_search_result_cptr_t prev =
	dep_tree_search_access_t(
	    context, conf.sample_file, conf.stack_depth, conf.line_size,
	    conf.page_size, conf.thread_mapping, conf.cpu_model, Mapped_Line,
	    &access_filter, true, true, pc, (fnr_time_t)-1).get();

    dep_tree_search_result_cptr_t next =
	dep_tree_search_access_t(
	    context, conf.sample_file, conf.stack_depth, conf.line_size,
	    conf.page_size, conf.thread_mapping, conf.cpu_model, Mapped_Line,
	    &access_filter, false, true, pc, (fnr_time_t)-1).get();

    return prefetch_issue_ptr_t(
	new prefetch_issue_t(problem,
			     pc,
			     loop_id,
			     depgroup_id,
			     cost,
			     median_rdist,
			     prev->locations(),
			     next->locations()));
}

END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;
