/// @file analysis/issues/prefetchnta.cc

// Copyright (c) 2009-2010 Acumem AB.  All rights reserved.

#include "analysis/issues/prefetchnta.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);
BEGIN_NAMESPACE(issues);

/// We require the reuses of the data to have at least this fetch ratio in
/// order to generate a prefetchnta issue.
#define REUSE_FR_THRESHOLD 0.80

prefetchnta_t::prefetchnta_t(const feature_context_ptr_t &context)
    : context(context)
{
}

void
prefetchnta_t::set_configuration(const loop_driver_conf_t &c)
{
    conf = c;

    access_lookup = access_lookup_t(context,
                                    c.sample_file,
                                    c.stack_depth).get();

    sample_mapping_result_cptr_t sample_mapping =
        sample_mapping_t(context,
                         c.sample_file,
                         c.stack_depth,
                         c.line_size,
                         c.page_size,
                         c.thread_mapping,
                         c.cpu_model).get();

    repl_fr_line_result_cptr_t repl_fr_line =
        repl_fr_line_t(context,
                       c.sample_file,
                       c.stack_depth,
                       c.line_size,
                       c.page_size,
                       c.cache_size,
                       c.thread_mapping,
                       c.cpu_model,
                       c.engine).get();

    repl_fr_line_result_cptr_t repl_fr_line_l1 =
        repl_fr_line_t(context,
                       c.sample_file,
                       c.stack_depth,
                       c.line_size,
                       c.page_size,
                       c.l1_cache_size,
                       c.thread_mapping,
                       c.cpu_model,
                       c.engine).get();

    assert(sample_mapping->bursts.size() == repl_fr_line->bursts.size());
    assert(repl_fr_line->bursts.size() == repl_fr_line_l1->bursts.size());
    for (unsigned int b = 0; b < sample_mapping->bursts.size(); b++) {
        double sample_period = sample_mapping->bursts[b].sample_period;
        const mapped_repl_line_vector_t &line_samples = sample_mapping->bursts[b].repl_line;
        const ratio_vector_t &line_frs = repl_fr_line->bursts[b].val_to_end;
        const ratio_vector_t &line_frs_l1 = repl_fr_line_l1->bursts[b].val_to_end;

        assert(line_samples.size() == line_frs.size());
        assert(line_frs.size() == line_frs_l1.size());
        for (unsigned int s = 0; s < line_samples.size(); s++) {
            const mapped_repl_line_t &sample = line_samples[s];
            const double fr = line_frs[s];
            const double fr_l1 = line_frs_l1[s];

            const mapped_access_t *valid = sample.valid();
            if (valid == NULL)
                continue;

            const mapped_access_t *end = sample.end();
            assert(end != NULL);

            reuse_stat_t rs(sample_period * fr_l1, sample_period * fr * fr_l1);

            pc1_reuse_stat[valid->pc] += rs;
            pc1_pc2_reuse_stat[valid->pc][end->pc] += rs;

            pc1_thread_reuse_stat[valid->pc][end->thread()] += rs;
        }
    }

    prefetchnta_lru(sample_mapping, repl_fr_line_l1, pc1_sd_sets);

    res = prefetchnta_result_ptr_t(new prefetchnta_result_t);
}

prefetchnta_result_ptr_t
prefetchnta_t::get()
{
    return res;
}

static double
required_cache_size(const sd_set_t &sd_set,
                    fnr_line_size_t line_size)
{
    double fr_l1_sum = 0.0;
    // summing from back to reduce truncation errors
    for (sd_set_t::const_reverse_iterator sdi = sd_set.rbegin(),
	     sde = sd_set.rend(); sdi != sde; sdi++)
        fr_l1_sum += sdi->second;

    double fr_l1_partial = 0.0;
    for (sd_set_t::const_reverse_iterator sdi = sd_set.rbegin(),
	     sde = sd_set.rend(); sdi != sde; sdi++) {
        fr_l1_partial += sdi->second;
        if (fr_l1_partial / fr_l1_sum > 0.8)
            return sdi->first * line_size;
    }

    assert(0);
#ifdef _WIN32
    while(1);
#endif
}

void
prefetchnta_t::process_dep(
    loop_ptr_t &loop,
    const location_set_t &loop_pc_set,
    const stats_cptr_t &loop_stat,
    depgroup_id_t dep_id,
    const location_set_t &dep_pc_set,
    const stats_cptr_t &stat)
{
    typedef map<location_t, double> location_fraction_map_t;

    // Don't generate a prefetchnta issue if the instruction group already
    // contains a prefetchnta instruction.
    fnr_access_classes_t pnta_classes = FNR_Access_Class_PF | FNR_Access_Class_NTA;

    BOOST_FOREACH(location_t pc1, dep_pc_set) {
        loc_access_info_map_t::const_iterator aii = access_lookup->loc_access_info.find(pc1);
        assert(aii != access_lookup->loc_access_info.end());
        if ((aii->second.access_classes & pnta_classes) == pnta_classes)
            return;
    }

    // Extract the pcs that have interesting non-temporal reuses.
    location_set_t pc1s;
    reuse_stat_t pc1s_rs;

    BOOST_FOREACH(location_t pc1, dep_pc_set) {
        const reuse_stat_t &rs = pc1_reuse_stat[pc1];
        if (rs.fetches / rs.accesses >= REUSE_FR_THRESHOLD) {
            pc1s.insert(pc1);
            pc1s_rs += rs;
        }
    }

    // Remove pcs that contribute less than 1% of the non-temporal reuses.
    location_set_t dead_pc1s;

    double fetch_threshold = 0.01 * pc1s_rs.fetches;
    BOOST_FOREACH(location_t pc1, pc1s) {
        const reuse_stat_t &rs = pc1_reuse_stat[pc1];
        if (rs.fetches < fetch_threshold)  {
            dead_pc1s.insert(pc1);
            pc1s_rs -= rs;
        }
    }

    BOOST_FOREACH(location_t pc1, dead_pc1s)
        pc1s.erase(pc1);

    // Don't generate an issue if there are no interesting pc1s, or if there
    // are too few fetches involved for the issue to be relevant.
    if (pc1s.empty() ||
        !conf.issue_relevant->issue_relevant_fetches(pc1s_rs.fetches))
        return;

    // Sum the fraction of the non-temporal temporal reuses of the pc1s that
    // each pc1 and pc2 is involved in.
    location_fraction_map_t pc1_fraction_map;
    location_fraction_map_t pc2_fraction_map;

    BOOST_FOREACH(location_t pc1, pc1s) {
        pc2_reuse_stat_t &pc2_reuse_stat = pc1_pc2_reuse_stat[pc1];
        BOOST_FOREACH(pc2_reuse_stat_t::value_type &v, pc2_reuse_stat) {
            location_t pc2 = v.first;
            double fraction = v.second.fetches / pc1s_rs.fetches;
            pc1_fraction_map[pc1] += fraction;
            pc2_fraction_map[pc2] += fraction;
        }
    }

    // Sum the number of involved accesses and fetches.
    thread_reuse_stat_t thread_rs;

    BOOST_FOREACH(location_t pc1, pc1s) {
        BOOST_FOREACH(thread_reuse_stat_t::value_type v, pc1_thread_reuse_stat[pc1])
            thread_rs[v.first] += v.second;
    }

    // Calculate the required cache size for the entire issue.
    sd_set_t total_sd_set;
    BOOST_FOREACH(location_t pc1, pc1s) {
        BOOST_FOREACH(const sd_set_t::value_type &v, pc1_sd_sets[pc1])
            total_sd_set[v.first] += v.second;
    }
    double rcs = required_cache_size(total_sd_set, conf.line_size);

    // Convert the fraction maps into sorted fraction vectors.
    location_fraction_vector_t pc1_fraction_vector;
    location_fraction_vector_t pc2_fraction_vector;

    BOOST_FOREACH(location_fraction_map_t::value_type &v, pc1_fraction_map)
        if (v.second >= 0.01)
            pc1_fraction_vector.push_back(
                location_fraction_t(
                    v.first, v.second,
                    required_cache_size(pc1_sd_sets[v.first],
                                        conf.line_size)));
    sort(pc1_fraction_vector.begin(), pc1_fraction_vector.end());

    BOOST_FOREACH(location_fraction_map_t::value_type &v, pc2_fraction_map)
        if (v.second >= 0.01)
            pc2_fraction_vector.push_back(
                location_fraction_t(v.first, v.second, 0.0));
    sort(pc2_fraction_vector.begin(), pc2_fraction_vector.end());

    // Generate the issue.
    prefetchnta_issue_cptr_t issue(
        new prefetchnta_issue_t(
            loop, dep_id, pc1s_rs.accesses, pc1s_rs.fetches, rcs,
            pc1_fraction_vector, pc2_fraction_vector));
    res->issues.push_back(issue);
}

bool
prefetchnta_t::want_depgroup() const
{
    return true;
}

typedef map<fnr_time_t, double> hist_t;

void
prefetchnta_t::prefetchnta_lru(sample_mapping_result_cptr_t sample_mapping,
                               repl_fr_line_result_cptr_t repl_fr_line_l1,
                               pc1_sd_set_t &pc1_sd_set)
{
    for (unsigned int b = 0; b < sample_mapping->bursts.size(); b++) {
        const mapped_repl_line_vector_t &line_samples = sample_mapping->bursts[b].repl_line;
        const ratio_vector_t &line_frs_l1 = repl_fr_line_l1->bursts[b].val_to_end;

        hist_t hist;
        double num_rdists = 0.0;

        assert(line_samples.size() == line_frs_l1.size());
        for (unsigned int s = 0; s < line_samples.size(); s++) {
            const mapped_repl_line_t &sample = line_samples[s];

            const mapped_access_t *valid = sample.valid();
            if (valid == NULL)
                continue;

            const mapped_access_t *end = sample.end();
            assert(end != NULL);

            hist[end->local_time - valid->local_time] += 1.0;

            num_rdists += 1.0;
        }

        double rdist_sum = 0.0;
        BOOST_FOREACH(hist_t::value_type &v, hist) {
            rdist_sum += v.second;
            v.second = 1.0 - rdist_sum / num_rdists;
        }

        typedef map<fnr_time_t, double> rd_to_sd_t;
        rd_to_sd_t rd_to_sd;

        fnr_time_t lo_rd = 0;
        double lo_sd = 0.0;
        double lo_ratio = 1.0;
        hist_t::const_iterator hi = hist.begin();
        while (hi != hist.end()) {
            fnr_time_t hi_rd = hi->first;
            double hi_sd = lo_sd + (hi_rd - lo_rd) * lo_ratio;
            double hi_ratio = hi->second;

            rd_to_sd[hi_rd] = hi_sd;

            lo_rd = hi_rd;
            lo_sd = hi_sd;
            lo_ratio = hi_ratio;
            hi++;
        }

        assert(line_samples.size() == line_frs_l1.size());
        for (unsigned int s = 0; s < line_samples.size(); s++) {
            const mapped_repl_line_t &sample = line_samples[s];
            const double fr_l1 = line_frs_l1[s];

            const mapped_access_t *valid = sample.valid();
            if (valid == NULL)
                continue;

            const mapped_access_t *end = sample.end();
            assert(end != NULL);

            rd_to_sd_t::const_iterator i = rd_to_sd.find(end->local_time - valid->local_time);
            assert(i != rd_to_sd.end());
            pc1_sd_sets[valid->pc][i->second] += fr_l1;
        }
    }
}

END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;
