/// @file analysis/issues/wnt_possible.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include "wnt_possible.hh"

#include "analysis/structure/dep_tree_search.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);
BEGIN_NAMESPACE(issues);

/*
 * TODO: We may currently generate WNT possible on Wnt accesses, this
 * is obviously undesirable. Currently, though, our fetch ratio
 * calculation has a "bug" with respect to calculating misses on wnt
 * instructions (they are never considered to miss), and instruction
 * groups involving such instructions can never fulfill the criteria
 * to save a substantial number of misses.
 *
 * TODO: Consider what happens in a dependency group consisting of
 * both WNT and normal W accesses.
 *
 * TODO: Not that we use this here, but calculation of WNT write-back
 * utilization doesn't work since we don't currently calculate miss
 * rates for WNT accesses.
 *
 */

// TODO: These limits are _very_ arbitrary at the moment
#define WRITE_UTIL_LIMIT 0.9
#define FETCH_UTIL_LIMIT 0.1
#define RANDOM_LIMIT 0.1

/*
 * The realtive extra fetch limit is the upper limit for the expected
 *  #introduced_fetches / #avoided_fetches when the user implements the
 * Wnt advice. Break even should be 1. We are a bit conservative here
 * to be well on the safe side.
 */
#define REL_EXTRA_FETCH_LIMIT 0.50

using namespace std;
using namespace boost;
using namespace acumem::library;

// ----------------------------------------------------------------------

wnt_possible_t::wnt_possible_t(const feature_context_ptr_t &c)
    : context(c)
{
}

void
wnt_possible_t::set_configuration(const loop_driver_conf_t &c)
{
    conf = c;

    locationx_mapper_cptr_t lx_mapper(
	new locationx_mapper_t(context,
			       c.sample_file,
			       c.stack_depth,
			       c.thread_mapping,
			       true, false, false));
    lx = lx_mapper->get();

    fr_lookup = fr_lookup_t(context,
			    c.sample_file,
			    c.stack_depth,
			    c.line_size, c.page_size,
			    c.thread_mapping, c.cpu_model,
			    lx_mapper,
			    c.cache_size,
			    c.engine,
			    FRL_Type_Stalls).get();

    fr_lookup_wr = fr_lookup_t(context,
			       c.sample_file,
			       c.stack_depth,
			       c.line_size, c.page_size,
			       c.thread_mapping, c.cpu_model,
			       lx_mapper,
			       c.cache_size,
			       c.engine,
			       FRL_Type_Write_No_Prefetch).get();

    res = wnt_possible_result_ptr_t(new wnt_possible_result_t);
}

wnt_possible_result_ptr_t
wnt_possible_t::get()
{
    return res;
}

void
wnt_possible_t::process_dep(
    loop_ptr_t &loop,
    const location_set_t &loop_pc_set,
    const stats_cptr_t &loop_stat,
    depgroup_id_t dep_id,
    const location_set_t &dep_pc_set,
    const stats_cptr_t &stat)
{
    if (conf.issue_relevant->issue_relevant_fetches(stat->all_ratio.total_fetches()) &&
	(stat->write_back_util.valid() && stat->write_back_util.value() > WRITE_UTIL_LIMIT) &&
        (!stat->fetch_util.valid() || stat->fetch_util.value() < FETCH_UTIL_LIMIT) &&
	stat->random_ratio < RANDOM_LIMIT) {

	// Find the locations reusing the cache lines touched by this
	// dependence group.
	dep_tree_search_result_cptr_t span_forwards(
	    dep_tree_search_access_t(
		context, conf.sample_file, conf.stack_depth,
		conf.line_size, conf.page_size,
		conf.thread_mapping, conf.cpu_model,
		Mapped_Line, &fnr_access_filter_any,
		false, true, dep_pc_set,
		(fnr_time_t)-1).get());
	location_set_t reuse_locs(span_forwards->locations());

        // Remove the locations in this dependence group from the set of reuse
        // locations, so that short reuses within the dependence group do no
        // prevent us from generating non-temporal store issues.
        //
        // TODO: We should really include longer reuses between the dependence
        //       group and itself, reuses "between iterations".
        BOOST_FOREACH(location_t l, dep_pc_set)
            reuse_locs.erase(l);

        // Get locationx sets for this dependence group and the reusers.
        locationx_set_t reads_lx(lx->get_locationx_set(reuse_locs));
	const locationx_set_t dep_lx_set(lx->get_locationx_set(dep_pc_set));

	// Calculate the number of reuses with this dependence group as pc1 and
	// one of the reusers as pc2 that currently hit in the cache.
        const fr_data_t reuse_fr(fr_lookup->pc1_pc2_fr(dep_lx_set, reads_lx));
	const double reuse_hits(reuse_fr.accesses() -
                                reuse_fr.by_kind().total_fetches());

	// Calculate the number of fetches that are caused by writes in this
	// dependence group. We assume that all fetches by writes in this
	// dependence group will disappear and all hits where this dependence
	// group is pc1 will become fetches after transforming the group to use
	// non-temporal stores.
	const fr_data_t wr_fr(fr_lookup_wr->fr_by_lx_set(dep_lx_set));
	const double write_fetches(wr_fr.by_kind().total_fetches());

	// Relate the number of additional fetches to the number of fetches
	// that may be removed by transforming the dependence group. Only
	// create the advice if this value is low.
	const double rel_extra_fetches(reuse_hits / write_fetches);
	if (rel_extra_fetches < REL_EXTRA_FETCH_LIMIT)
	    res->issues.push_back(
		wnt_possible_issue_ptr_t(
		    new wnt_possible_issue_t(loop, dep_id, reuse_locs)));
    }
}

bool
wnt_possible_t::want_depgroup() const
{
    return true;
}

END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;
