/// @file analysis/issues/wnt_possible.hh

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#ifndef WNT_POSSIBLE_HH
#define WNT_POSSIBLE_HH

#include "analysis/locationx/locationx.hh"
#include "analysis/fetch_ratio/fr_lookup.hh"

#include "loop_engine.hh"
#include "issue.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);
BEGIN_NAMESPACE(issues);

using namespace std;
using namespace acumem::library;

using boost::shared_ptr;

/// Result from hot_spot_t::get().
struct wnt_possible_result_t {
    wnt_possible_issue_vector_t issues;
};

typedef shared_ptr<const wnt_possible_result_t> wnt_possible_result_cptr_t;
typedef shared_ptr<wnt_possible_result_t> wnt_possible_result_ptr_t;

/**
 * Module to identify dependency groups in loops that may benefit from
 * non-temporal stores instead of ordinary stores.
 *
 * The following must be true for such a group:
 *
 * 1. High write-back utilization
 *
 *    Most implementations seem to start a write-back of a NTA buffer
 *    as soon as the whole buffer (one cache line) has been written,
 *    or when all NTA buffers have been exhausted. Writing the whole
 *    cache line allows the memory controller to start the write-back
 *    early and prevents stalls due to buffer starvation.
 *
 *    Running out of NTA buffers seems to incur a large performance
 *    penalty, a small test program running on a Core 2 shows almost a
 *    factor 6 slowdown.
 *
 *    The number of NTA buffers is typically very low, e.g. the Intel
 *    Core 2 seems to have 6 buffers and AMD Opteron (2nd & 3rd
 *    generation) seems to have 4 buffers.
 *
 * 2. The next read should cause a fetch
 *
 *    Since the data will be written back to memory it is desirable
 *    that the next read has a high fetch ratio. Otherwise the NTA
 *    write will introduce undesirable fetches.
 *
 *    Some processors seem to implement forwarding from the NTA
 *    buffers when a read occurs just after a write. This might allow
 *    reads just after the writes to hit in the NTA buffers, but the
 *    feature is not documented (?) and we won't model it for now.
 *
 * 3. Low randomness
 *
 *    This condition is related to condition 1, we want to have as few
 *    cache lines in flight as possible in order not to exhaust the
 *    number of available NTA buffers.
 *
 *    As an alternative we could later consider using an LRU fetch
 *    ratio calculation (with a very small "cache" size, four or six
 *    lines) over the write instructions in that loop. If the "fetch
 *    ratio" is high, then we will exhaust the number of buffers, and
 *    that should be a warning bell.
 */
class wnt_possible_t : public loop_analysis_base_t {
public:
    wnt_possible_t(const feature_context_ptr_t &context);

    void set_configuration(const loop_driver_conf_t &c);
    wnt_possible_result_ptr_t get();

    virtual void process_dep(loop_ptr_t &loop,
			     const location_set_t &loop_pc_set,
                             const stats_cptr_t &loop_stat,
			     depgroup_id_t dep_id,
			     const location_set_t &dep_pc_set,
			     const stats_cptr_t &stat);

    bool want_depgroup() const;

private:
    feature_context_ptr_t context;
    loop_driver_conf_t conf;

    locationx_mapper_result_cptr_t lx;
    fr_lookup_result_cptr_t fr_lookup;
    fr_lookup_result_cptr_t fr_lookup_wr;

    wnt_possible_result_ptr_t res;
};


END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;

#endif
