/// @file analysis/locationx/locationx.cc

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#include "locationx.hh"

#include <boost/lexical_cast.hpp>
#include <boost/foreach.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace boost;

// ----- locationx_specifier_t ------------------------------------------------

locationx_specifier_t::locationx_specifier_t(location_t location,
					     fnr_thread_id_t thread,
					     tm_cache_id_t cache)
    : the_location(location), the_cache(cache), the_thread(thread)
{
}

locationx_specifier_t::locationx_specifier_t(const mapped_access_t &access)
    : the_location(access.pc),
      the_cache(access.cache),
      the_thread(access.thread())
{
}

// ----- locationx_mapper_result_t --------------------------------------------

locationx_mapper_result_t::locationx_mapper_result_t(bool use_location,
						     bool use_thread,
						     bool use_cache)
    : use_location(use_location),
      use_thread(use_thread),
      use_cache(use_cache),
      max_id(0)
{
}

locationx_set_t
locationx_mapper_result_t::get_locationx_set(const location_t loc) const
{
    locationx_set_t lx_set;
    location_locationx_multimap_t::const_iterator range_begin, range_end;
    tie(range_begin, range_end) = location_multimap.equal_range(loc);

    copy(range_begin, range_end, locationx_inserter_t(lx_set));
    return lx_set;
}

const locationx_specifier_t &
locationx_mapper_result_t::get_specifier(const locationx_t locationx) const
{
    assert(locationx.id < locationx_spec.size());
    return locationx_spec[locationx.id];
}

location_t
locationx_mapper_result_t::get_location(const locationx_t lx) const
{
    assert(use_location);
    return get_specifier(lx).location();
}

fnr_thread_id_t
locationx_mapper_result_t::get_thread(const locationx_t lx) const
{
    assert(use_location);
    return get_specifier(lx).thread();
}

tm_cache_id_t
locationx_mapper_result_t::get_cache(const locationx_t lx) const
{
    assert(use_location);
    return get_specifier(lx).cache();
}

// ----- locationx_mapper_t ---------------------------------------------------

static void
add_access(
    locationx_specifier_set_t &locationxs,
    const thread_mapping_burst_t &thread_mapping,
    const stack_pc_map_result_cptr_t &stack_pc_map,
    fnr_address_t pc,
    fnr_stack_id_t stack,
    fnr_thread_id_t thread,
    fnr_time_t time,
    bool use_location,
    bool use_thread,
    bool use_cache)
{
    const location_t location =
	use_location ?
        stack_pc_map->get_location(stack, pc) :
        LOCATION_INVALID;
    const locationx_specifier_t spec(
	location,
	use_thread ? thread : 0,
	use_cache ? thread_mapping.get_cache_id(time, thread) : 0);

    locationxs.insert(spec);
}

static void
add_access(
    locationx_specifier_set_t &locationxs,
    const thread_mapping_burst_t &thread_mapping,
    const stack_pc_map_result_cptr_t &stack_pc_map,
    const fnr_sample_access_t &access,
    fnr_time_t time1,
    bool use_location,
    bool use_thread,
    bool use_cache)
{
    add_access(locationxs, thread_mapping, stack_pc_map, access.pc,
               access.stack, access.thread, time1, use_location, use_thread,
               use_cache);
}

locationx_mapper_result_cptr_t
locationx_mapper_t::calculate(parser_result_cptr_t samples,
			      thread_mapping_result_cptr_t thread_mapping,
			      stack_pc_map_result_cptr_t stack_pc_map,
                              locationa_result_cptr_t locationa_mapping,
			      bool use_location, bool use_thread, bool use_cache)
{
    locationx_mapper_result_ptr_t result(
	new locationx_mapper_result_t(use_location, use_thread, use_cache));

    locationx_specifier_set_t locationxs;

    thread_mapping_burst_vector_t::const_iterator it_tmb(
	thread_mapping->bursts.begin());
    BOOST_FOREACH(const parser_burst_t &burst, samples->bursts) {
	const thread_mapping_burst_t &tmb(*it_tmb++);
	BOOST_FOREACH(const sample_ptr_t &sample, burst.samples_by_pc1) {
	    fnr_time_t time1(sample->begin.time);

	    add_access(locationxs, tmb, stack_pc_map, sample->begin, time1,
                       use_location, use_thread, use_cache);
	    add_access(locationxs, tmb, stack_pc_map, sample->pc_1_plus,
                       sample->pc_1_plus_stack, sample->begin.thread, time1,
                       use_location, use_thread, use_cache);

	    BOOST_FOREACH(const fnr_sample_access_t &access, sample->end)
		add_access(locationxs, tmb, stack_pc_map, access, time1,
                           use_location, use_thread, use_cache);
	}
    }

    result->locationx_spec.reserve(locationxs.size());
    BOOST_FOREACH(const locationx_specifier_t &spec, locationxs)
        result->locationx_spec.push_back(spec);

    BOOST_FOREACH(const locationx_specifier_t &spec, result->locationx_spec)
        result->location_multimap.insert(
            make_pair(spec.location(), locationx_t(result->max_id++)));

    locationa_t num_locationas = (locationa_t)locationa_mapping->locationas.size();
    for (locationa_t la = 0; la < num_locationas; la++) {
        const locationa_specifier_t &las =
            locationa_mapping->locationas[la];
        locationx_t lx = result->find(
            use_location ? las.location : LOCATION_INVALID,
            use_thread ? las.thread : 0,
            use_cache ? las.cache : 0);
        result->locationx_locationa[lx].insert(la);
    }

    return result;
}

locationx_mapper_t::locationx_mapper_t(feature_context_ptr_t context,
				       const string &sample_file,
				       fnr_stack_depth_t stack_depth,
				       const thread_mapping_cptr_t &thread_mapping,
				       bool use_location, bool use_thread, bool use_cache)
    : feature_t<locationx_mapper_result_cptr_t>(context),
      sample_file(sample_file),
      thread_mapping(thread_mapping),
      stack_depth(stack_depth),
      use_location(use_location), use_thread(use_thread), use_cache(use_cache)
{
}

locationx_mapper_result_cptr_t
locationx_mapper_t::get() const
{
    return get_data();
}

string
locationx_mapper_t::name() const
{
    return "locationx_mapper_t(" +
        sample_file + "," +
        lexical_cast<string>(stack_depth) + "," +
        thread_mapping->name() + "," +
        lexical_cast<string>(use_location) + "," +
        lexical_cast<string>(use_thread) + "," +
        lexical_cast<string>(use_cache) + ")";
}

locationx_mapper_result_cptr_t
locationx_mapper_t::calculate() const
{
    return calculate(
        parser_t(context, sample_file).get(),
        thread_mapping->get(),
        stack_pc_map_t(context, sample_file, stack_depth).get(),
        locationa_mapper_t(
            context, sample_file, stack_depth, thread_mapping).get(),
	use_location, use_thread, use_cache);
}

END_NAMESPACE;
END_NAMESPACE;
