/// @file analysis/locationx/locationx.hh

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#ifndef LOCATIONX_HH
#define LOCATIONX_HH

#include "analysis/locationa/locationa.hh"

#include <boost/tuple/tuple.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;

using boost::shared_ptr;

// ----- locationx_specifier_t ------------------------------------------------

class locationx_specifier_t {
public:
    friend struct locationx_mapper_result_t;

    locationx_specifier_t(location_t location,
			  fnr_thread_id_t thread,
			  tm_cache_id_t cache);

    locationx_specifier_t(const mapped_access_t &acccess);


    inline bool operator<(const locationx_specifier_t &rhs) const {
	if (location() != rhs.location())
	    return location() < rhs.location();
	else if (cache() != rhs.cache())
	    return cache() < rhs.cache();
	else
	    return thread() < rhs.thread();
    }

    inline bool operator==(const locationx_specifier_t &rhs) const {
	return location() == rhs.location() &&
	    cache() == rhs.cache() &&
	    thread() == rhs.thread();
    }

    inline location_t location() const {
	return the_location;
    }

    inline fnr_thread_id_t thread() const {
	return the_thread;
    }

    inline tm_cache_id_t cache() const {
	return the_cache;
    }

private:
    /// A valid location if use_location, LOCATION_INVALID otherwise.
    location_t the_location;

    /// A valid cache id if use_cache, 0 otherwise.
    tm_cache_id_t the_cache;

    /// A valid thread id if use_thread, 0 otherwise.
    fnr_thread_id_t the_thread;
};

typedef set<locationx_specifier_t> locationx_specifier_set_t;
typedef vector<locationx_specifier_t> locationx_specifier_vector_t;

// ----- locationx_t ----------------------------------------------------------

typedef uint32_t locationx_id_t;

#define LOCATIONX_INVALID ((locationx_id_t)-1)

struct locationx_t
{
    explicit locationx_t(locationx_id_t id)
	: id(id) {}

    bool invalid() {
        return id == LOCATIONX_INVALID;
    }

    inline bool operator<(const locationx_t &rhs) const {
	return id < rhs.id;
    }

    const locationx_id_t id;
};

typedef set<locationx_t> locationx_set_t;
typedef vector<locationx_t> locationx_vector_t;
typedef multimap<location_t, locationx_t> location_locationx_multimap_t;

typedef map<locationx_t, locationa_set_t> locationx_locationa_set_map_t;

// ----- locationx_mapper_result_t --------------------------------------------

struct locationx_mapper_result_t {
    friend class locationx_mapper_t;

    inline locationx_t get_locationx(location_t location,
				     fnr_thread_id_t thread,
				     tm_cache_id_t cache) const {
	return find(use_location ? location : LOCATION_INVALID,
                    use_thread ? thread : 0,
                    use_cache ? cache : 0);
    }

    inline locationx_t get_locationx(const mapped_access_t &access) const {
	return get_locationx(access.pc, access.thread(), access.cache);
    }

    inline locationx_t get_locationx(const locationx_specifier_t &spec) const {
	return get_locationx(spec.location(), spec.thread(), spec.cache());
    }

    locationx_set_t get_locationx_set(const location_t loc) const;

    template <class Cont>
    locationx_set_t get_locationx_set(const Cont &locs) const {
	locationx_set_t lx_set;

	BOOST_FOREACH(const location_t loc, locs) {
	    location_locationx_multimap_t::const_iterator range_begin, range_end;
	    boost::tie(range_begin, range_end) = location_multimap.equal_range(loc);
	    copy(range_begin, range_end, locationx_inserter_t(lx_set));
	}

	return lx_set;
    }

    const locationx_specifier_t &get_specifier(const locationx_t locationx) const;

    location_t get_location(const locationx_t lx) const;
    fnr_thread_id_t get_thread(const locationx_t lx) const;
    tm_cache_id_t get_cache(const locationx_t lx) const;

    const bool use_location;
    const bool use_thread;
    const bool use_cache;

    locationx_specifier_vector_t locationx_spec;
    location_locationx_multimap_t location_multimap;
    locationx_id_t max_id;

    locationx_locationa_set_map_t locationx_locationa;

private:
    class locationx_inserter_t
	: public std::iterator<output_iterator_tag, void, void, void, void>
    {
    public:
	explicit locationx_inserter_t(locationx_set_t &cont)
	    : cont(cont) {}

	locationx_inserter_t &operator=(const pair<location_t, locationx_t> &val) {
	    cont.insert(val.second);
	    return *this;
	}

	locationx_inserter_t &operator*() { return *this; }
	locationx_inserter_t &operator++() { return *this; }

    private:
	locationx_set_t &cont;
    };

    locationx_mapper_result_t(bool use_location,
			      bool use_thread,
			      bool use_cache);

    locationx_t find(location_t location,
                     fnr_thread_id_t thread,
                     tm_cache_id_t cache) const {
        locationx_specifier_t spec(location, thread, cache);

        locationx_specifier_vector_t::const_iterator lower =
            lower_bound(locationx_spec.begin(), locationx_spec.end(), spec);

        if (lower != locationx_spec.end() && *lower == spec)
            return locationx_t((locationx_id_t)(lower - locationx_spec.begin()));
        else
            return locationx_t(LOCATIONX_INVALID);
    }
};

typedef shared_ptr<locationx_mapper_result_t> locationx_mapper_result_ptr_t;
typedef shared_ptr<const locationx_mapper_result_t> locationx_mapper_result_cptr_t;

// ----- locationx_mapper_t ---------------------------------------------------

/**
 * The locationx_mapper_t maps any combination of location, thread and
 * cache to an extended location (#locationx_t). Using only the
 * location part makes the extended location equivalent to the
 * ordinary location.
 *
 * Comparison of #locationx_t entries generated from different
 * #locationx_mapper_t objects is not possible. The result of such a
 * comparison is undefined.
 */
class locationx_mapper_t : public feature_t<locationx_mapper_result_cptr_t> {
public:
    static locationx_mapper_result_cptr_t calculate(
	parser_result_cptr_t samples,
	thread_mapping_result_cptr_t thread_mapping,
	stack_pc_map_result_cptr_t stack_pc_map,
        locationa_result_cptr_t locationa_mapping,
	bool use_location, bool use_thread, bool use_cache);

    locationx_mapper_t(feature_context_ptr_t context,
		       const string &sample_file,
		       fnr_stack_depth_t stack_depth,
		       const thread_mapping_cptr_t &thread_mapping,
		       bool use_location, bool use_thread, bool use_cache);
    locationx_mapper_result_cptr_t get() const;

    string name() const;

protected:
    locationx_mapper_result_cptr_t calculate() const;

private:
    const string sample_file;
    const thread_mapping_cptr_t thread_mapping;
    const fnr_stack_depth_t stack_depth;

    const bool use_location;
    const bool use_thread;
    const bool use_cache;
};

typedef shared_ptr<const locationx_mapper_t> locationx_mapper_cptr_t;

END_NAMESPACE;
END_NAMESPACE;

#endif
