/// @file analysis/misc/histogram.hh

// Copyright (c) 2006-2008 Acumem AB.  All rights reserved.

#ifndef STAT_HISTOGRAM_HH
#define STAT_HISTOGRAM_HH

#include "library/lang/common_macros.h"
#include "library/lang/common_types.h"

#include <boost/shared_ptr.hpp>
#include <map>
#include <memory>
#include <vector>
#include <limits.h>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using boost::shared_ptr;

// ----------------------------------------------------------------------

/// implements a scheme where all bins are of equal size
template <int bin_size>
class equal_size_histogram_t {
public:
    int get_bin(uint64_t value) const;
    uint64_t get_bin_size(int bin) const;
    uint64_t get_bin_start(int bin) const;
};

template <int bin_size>
int
equal_size_histogram_t<bin_size>::get_bin(uint64_t value) const
{
    uint64_t bin = value / bin_size;
    if (bin < INT_MAX)
        return (int) bin;
    else
        return INT_MAX;
}

template <int bin_size>
uint64_t
equal_size_histogram_t<bin_size>::get_bin_size(int bin) const
{
    return bin_size;
}

template <int bin_size>
uint64_t
equal_size_histogram_t<bin_size>::get_bin_start(int bin) const
{
    return bin * bin_size;
}


// ----------------------------------------------------------------------

/// implements a scheme where we have bin sizes that grow
/// exponentially (base 2)
class exp_2_size_histogram_t {
public:
    exp_2_size_histogram_t();
    int get_bin(uint64_t value) const;
    uint64_t get_bin_size(int bin) const;
    uint64_t get_bin_start(int bin) const;
};

// ----------------------------------------------------------------------

/// implements a scheme where we have bin sizes that grow
/// exponentially (general base)
class exp_size_histogram_t {
public:
    exp_size_histogram_t(double base);
    int get_bin(uint64_t value) const;
    uint64_t get_bin_size(int bin) const;
    uint64_t get_bin_start(int bin) const;
private:
    double base_;
};

// ----------------------------------------------------------------------

/// This is a crude scheme, where values 0-255 will be placed in bins
/// of size 1, values 256 - 64k will be placed in bins of size 256
/// etc. In other words, use the most significant non zero 8-bit
/// aligned byte and use as a bin.
class exp_8_bit_t {
public:
    int get_bin(uint64_t value) const;
    uint64_t get_bin_size(int bin) const;
    uint64_t get_bin_start(int bin) const;
};

// ----------------------------------------------------------------------

class exp_histogram_base_t {
protected:
    static uint32_t floor_log2_32(register uint32_t x);
    static uint32_t ones32(register uint32_t x);
    static uint32_t floor_log2_64(register uint64_t x);
    static uint32_t ones64(register uint64_t x);
};

// ----------------------------------------------------------------------

/// Implements a exponentially decaying scheme with a bin size that is
/// constant for all values with the same floor_log2, and the first (1
/// << precision-bits) values have bins of size one.
///
/// If you do the math, there will be 2^(bits-1)*(66-bits) - 1 number
/// of bins (for 2^64 possible values).
///
///  precision	Bins
///    1	64
///    2	127     <- fits in 7 bits
///    3	251     <- fits in 8 bits
///    4	495
///    5	975
///    6	1919
///    7	3775
///    8	7423
///    9	14591
///    10	28671   <- fits in 31 bits
///    11	56319   <- fits in 32 bits
///
///   (larger values for bits would require a change in the return
///   type for bins.)
///
/// One example: to represent a signed delta in a 32-bit integer, use
/// 10 bits for precision (plus a sign bit). The first 1024 slots will
/// be direct mapped to values 0-1023. Values 4092-4095 will share
/// one bin. 4096 - 5004 will share the next one.

template <int bits>
class exp_histogram_t : public exp_histogram_base_t {
public:
    int get_bin(uint64_t value) const;
    uint64_t get_bin_size(int bin) const;
    uint64_t get_bin_start(int bin) const;

private:
    enum {
	//! values up to this have size 1
	first_bin       = (1 << bits),
	//! subsequent sizes used by this many bins
	subsequent_bins = first_bin/2
    };
};

template <int bits>
int
exp_histogram_t<bits>::get_bin(uint64_t value) const
{
    if (value < first_bin)
	return (int) value;

    unsigned int startbit = floor_log2_64(value);
    unsigned int stopbit = startbit - bits + 1;

    return (int)(value >> stopbit) + subsequent_bins * stopbit;
}

template <int bits>
uint64_t
exp_histogram_t<bits>::get_bin_size(int bin) const
{
    if (bin < first_bin)
	return 1;
    return 1LL << ((bin >> (bits - 1)) - 1);
}

template <int bits>
uint64_t
exp_histogram_t<bits>::get_bin_start(int bin) const
{
    if (bin < first_bin)
	return bin;

    uint32_t stopbits = bin/subsequent_bins - 1;
    uint32_t i = bin - subsequent_bins*stopbits;

    return (uint64_t)i << stopbits;
}

// -------------------------------------------------------------------


/// The histogram_t delivers its result in terms of this structure
struct bin_t {
    bin_t(uint64_t start=0, uint64_t size=0, int count=0)
	: start(start), size(size), count(count) {}
    uint64_t start;
    uint64_t size;
    uint32_t count;
};

typedef vector<bin_t> bin_vector_t;
typedef shared_ptr<bin_vector_t> bin_vector_ptr_t;

/// Implements a histogram calculation engine. It uses the
/// histogram_distribution_t object to define bin-sizes.
template <class Distribution>
class histogram_t {
public:
    /// given a pair of iterators, return a vector of bins
    template <typename ForwardIterator, typename Op>
    bin_vector_ptr_t
    get_histogram(ForwardIterator iter,
		  ForwardIterator end,
		  Op o) const;
};

// -------------------------------------------------------------------

// Implementation of the histogram_t::get_histogram function.
// Note: it has to be included in the header file, since it is a
// template member function

template <class Distribution>
template <typename ForwardIterator, typename Op>
bin_vector_ptr_t
histogram_t<Distribution>::get_histogram(ForwardIterator iter,
					 ForwardIterator end,
					 Op op) const
{
    Distribution distribution;

    // Use a map during construction. map is open-ended, which suits
    // us well. It will also handle sparse bins nicely. It records
    // a count per bin
    map<int, int> temp_bin;

    // Increment per-bin-counter. Implicit conversion of *iter to an int.
    for (; iter != end; iter++)
	temp_bin[distribution.get_bin(op(*iter))] ++;

    // build a resulting histogram vector
    bin_vector_t *v = new bin_vector_t;

    // fill in the vector
    map<int,int>::iterator i = temp_bin.begin(), e = temp_bin.end();
    for (; i != e; i++) {
	// Ignore count == 0
	if (i->second == 0)
	    continue;
#if 0
	cout << "start: " << distribution.get_bin_start(i->first)
		  << " size: " << distribution.get_bin_size(i->first)
		  << " count: " << i->second
		  << endl;
#endif
	v->push_back(bin_t(distribution.get_bin_start(i->first),
			   distribution.get_bin_size(i->first),
			   i->second));
    }

    // return the vector, protected by auto_ptr
    return bin_vector_ptr_t(v);
}

// -------------------------------------------------------------------

END_NAMESPACE
END_NAMESPACE

#endif
