/// @file analysis/misc/parser.cc

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#include "analysis/misc/parser.hh"
#include "analysis/feature_base/exceptions.hh"
#include "analysis/feature_base/warnings.hh"
#include "library/fenris/flow_file_reader.hh"
#include "library/stream/header_collector.hh"

#include <boost/foreach.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace boost;
using namespace std;

// ----- parser_result_t ------------------------------------------------------

parser_result_t::parser_result_t()
    : process_info(new process_info_t),
      burst_sampled(false)
{
    /* Insert a fake empty call stack at index 0 of the stack array, to emulate
       an empty stack for pcs that don't have call stack information. */
    stacks.push_back(&empty_stack);
}

fnr_sample_count_t
parser_result_t::repl_sample_count() const
{
    fnr_sample_count_t count(0);
    BOOST_FOREACH(const parser_burst_t &pb, bursts)
	count += (fnr_sample_count_t)pb.samples_by_pc1.size();

    return count;
}

fnr_time_t
parser_result_t::total_length() const
{
    fnr_time_t length(0);
    BOOST_FOREACH(const parser_burst_t &pb, bursts)
	length += pb.length;

    return length;
}

fnr_context_id_t
parser_result_t::get_context_id(fnr_thread_id_t thread) const
{
    assert(thread.id < contexts.size());

    return contexts[thread.id];
}

// ----- event_collector_t ----------------------------------------------------

/// Fenris flow class that collects the results we need from the fenris
/// events. We do this in a separate class, and not in parser_module_t itself,
/// so that parser_module_t doesn't have to contain the sample_begin_map and
/// sample_end_map.
class event_collector_t: public fenris_flow_by_type_t {
    typedef map<fnr_sample_count_t, fnr_event_sample_begin_t> begin_event_map_t;
    typedef map<fnr_sample_count_t, fnr_event_sample_end_t> end_event_map_t;

    typedef map<fnr_context_id_t, fnr_thread_id_t> thread_map_t;

public:
    event_collector_t(feature_context_ptr_t context);
    parser_result_ptr_t get_result(
	const stream_container_header_t &stream_header);

private:
    /// The current context, to check for cancellation
    feature_context_ptr_t context;

    void test_cancel();
    int event_counter;

    /// Result, filled in during event processing.
    parser_result_ptr_t parser_result;

    /// Pointer to the current burst.
    parser_burst_t *current_burst;

    /// Time mapping events, per burst.
    vector<local_time_vector_t> local_times;

    /// Legacy time mapping events, per burst
    vector<time_mapping_vector_t> time_mappings;

    /// Maps sample ids to sample begin events that have not yet gotten a
    /// corresponding sample end event.
    begin_event_map_t sample_begin_map;

    /// Maps sample ids to sample end events that have not yet gotten a
    /// corresponding sample begin event.
    end_event_map_t sample_end_map;

    /// Time of the last seen access
    fnr_time_t end_time;

    /// Fenris features
    fnr_features_t fenris_features;

    /// Vector mapping thread IDs to context IDs.
    fnr_context_id_vector_t thread_to_context;

    /// Map between context IDs and thread IDs for sample files
    /// without the "virtual threads" feature.
    thread_map_t context_to_thread;

    void update_acquisition_end(fnr_time_t time);
    void update_acquisition_end(const fnr_event_sample_t &time);

    /// Emulates virtual thread IDs for sample files without the
    /// virtual thread ID feature.
    fnr_thread_id_t convert_thread_id(fnr_thread_id_t thread);

    void generate_sample(const fnr_event_sample_t &sample);
    void generate_sample(const fnr_event_sample_begin_t &begin,
                         const fnr_event_sample_end_t &end);

    void generate_dangling(const fnr_event_sample_begin_t &begin);

    flow_status_t flow_init(fnr_features_t features);

    flow_status_t read(const fnr_event_read_t &event);
    flow_status_t write(const fnr_event_write_t &event);
    flow_status_t prefetch(const fnr_event_prefetch_t &event);
    flow_status_t sample_begin(const fnr_event_sample_begin_t &event);
    flow_status_t sample_end(const fnr_event_sample_end_t &event);
    flow_status_t sample_period(const fnr_event_sample_period_t &event);
    flow_status_t acquisition_begin(const fnr_event_acquisition_begin_t &event);
    flow_status_t acquisition_end(const fnr_event_acquisition_end_t &event);
    flow_status_t mmap(const fnr_event_mmap_t &event);
    flow_status_t munmap(const fnr_event_munmap_t &event);
    flow_status_t stack(const fnr_event_stack_t &event);
    flow_status_t sample(const fnr_event_sample_t &event);
    flow_status_t time_mapping(const fnr_event_time_mapping_t &event);
    flow_status_t local_time(const fnr_event_local_time_t &event);
    flow_status_t burst_begin(const fnr_event_burst_begin_t &event);
    flow_status_t burst_end(const fnr_event_burst_end_t &event);
    flow_status_t contexts(const fnr_event_contexts_t &event);
};

event_collector_t::event_collector_t(feature_context_ptr_t context)
    : context(context),
      event_counter(0),
      parser_result(new parser_result_t()),
      end_time(0)
{
    parser_result->bursts.push_back(parser_burst_t());
    local_times.push_back(local_time_vector_t());
    time_mappings.push_back(time_mapping_vector_t());
    current_burst = &parser_result->bursts.back();
}

void
event_collector_t::test_cancel()
{
    if ((event_counter++ %100) == 0)
	if (context)
	    if (context->feedback)
		// See if user has requested cancelling the analysis
		context->feedback->test_cancel();
}

static bool
sort_on_pc1(const sample_ptr_t &left,
	    const sample_ptr_t &right)
{
    if (left->begin.time < right->begin.time)
        return true;
    if (left->begin.time > right->begin.time)
        return false;
    if (left->line_size_log2 > right->line_size_log2)
        return true;
    if (left->line_size_log2 < right->line_size_log2)
        return false;
    if (left->begin.addr < right->begin.addr)
        return true;
    if (left->begin.addr > right->begin.addr)
        return false;

    // No two objects should be equal so we should reasonably not get
    // here. However, some older versions of libstdc++ compare objects with
    // themselves, and even make copies of objects and compare them with the
    // originals, so return false assuming that this is the case.
    return false;
}

parser_result_ptr_t
event_collector_t::get_result(const stream_container_header_t &stream_header)
{
    test_cancel();

    // There should be no sample end events without a corresponding sample
    // begin event left.
    assert(sample_end_map.size() == 0);

    // Convert all sample begin events without a corresponding sample end event
    // to dangling samples.
    for (begin_event_map_t::iterator
             i = sample_begin_map.begin(), e = sample_begin_map.end();
         i != e; i++)
        generate_dangling(i->second);
    sample_begin_map.clear();

    BOOST_FOREACH(parser_burst_t &burst, parser_result->bursts) {
	// The samples in samples_by_pc1 are originally in near pc2 order so we
	// need to sort them.
	sort(burst.samples_by_pc1.begin(), burst.samples_by_pc1.end(),
	     sort_on_pc1);

	// Store the length of the burst
	if (burst.begin && burst.end) {
	    burst.begin_time = burst.begin->time;
	    burst.length = burst.end->time - burst.begin->time;
	} else {
	    assert(parser_result->bursts.size() == 1);
	    assert(!burst.begin && !burst.end);
	    burst.begin_time = 0;
	    burst.length = end_time;
	}

	// Calculate the real sample period.
	if (!burst.samples_by_pc1.empty())
	    burst.sample_period = (double)burst.length /
		burst.samples_by_pc1.size();
	else
	    burst.sample_period = 0;
    }

    // Convert legacy time mapping events to current time mapping events.
    for (parser_burst_vector_t::size_type
             i = 0; i < parser_result->bursts.size(); i++) {
        parser_burst_t &burst = parser_result->bursts[i];
        local_time_vector_t &lts = local_times[i];
        time_mapping_vector_t &tms = time_mappings[i];

        if (!tms.empty()) {
            assert(lts.empty());

            map<fnr_thread_id_t, fnr_time_t> last_local_times;
            fnr_time_t last_global_time = 0;

            BOOST_FOREACH(const event_time_mapping_ptr_t &tm, tms) {
                BOOST_FOREACH(const fnr_event_time_mapping_t::time_event_t &te,
                              tm->time_events) {
                    if (te.global_time == last_global_time)
                        continue;
                    assert(te.global_time > last_global_time);
                    fnr_time_t local_time =
                        last_local_times[te.tid] +=
                        te.global_time - last_global_time;
                    last_global_time = te.global_time;

                    lts.push_back(
                        event_local_time_ptr_t(
                            new fnr_event_local_time_t(
                                te.tid, te.global_time, local_time)));
                }
            }
        }

        burst.time_map.reset(new time_map_t(thread_to_context.size()));
        BOOST_FOREACH(const event_local_time_ptr_t &lt, lts)
            burst.time_map->at(lt->tid.id).push_back(
                global_to_thread_t(lt->global_time, lt->local_time));
    }

    parser_result->contexts = thread_to_context;
    parser_result->fenris_features = fenris_features;
    parser_result->stream_header = stream_header;

    // Early releases of ThreadSpotter lacked the Fenris feature
    // field, but were MP sampled.
    if (stream_header.product == STREAM_CONTAINER_PRODUCT_VPEMP)
	parser_result->fenris_features |= FNR_Feature_MT;

    return parser_result;
}

void
event_collector_t::update_acquisition_end(fnr_time_t time)
{
    if (time > end_time)
	end_time = time;
}

void
event_collector_t::update_acquisition_end(const fnr_event_sample_t &sample)
{
    if (sample.end.empty())
	update_acquisition_end(sample.begin.time);
    else
	update_acquisition_end(sample.end.back().time);
}

fnr_thread_id_t
event_collector_t::convert_thread_id(fnr_thread_id_t thread)
{
    if (!(fenris_features & FNR_Feature_Virtual_Threads)) {
	const fnr_context_id_t context(thread.id);
	pair<thread_map_t::const_iterator, bool> insertion(
	    context_to_thread.insert(
		make_pair(context,
			  fnr_thread_id_t((uint32_t)thread_to_context.size()))));

	if (insertion.second)
	    thread_to_context.push_back(context);

	return insertion.first->second;
    } else
	return thread;
}

void
event_collector_t::generate_sample(const fnr_event_sample_t &sample)
{
    sample_ptr_t sample_ptr(new fnr_event_sample_t(sample));

    sample_ptr->begin.thread = convert_thread_id(sample_ptr->begin.thread);
    BOOST_FOREACH(fnr_sample_access_t &a, sample_ptr->end)
	a.thread = convert_thread_id(a.thread);

    current_burst->samples_by_pc1.push_back(sample_ptr);

    update_acquisition_end(sample);
}

void
event_collector_t::generate_sample(const fnr_event_sample_begin_t &begin,
                                   const fnr_event_sample_end_t &end)
{
    fnr_sample_access_t access_1(
	begin.pc,
	begin.address,
	begin.time,
	0, /* No pc1 call stack info. */
	convert_thread_id(begin.thread_id),
	begin.access_size,
	begin.kind);

    fnr_sample_access_vector_t access_2(
	1, fnr_sample_access_t(
	    end.pc,
	    end.address,
	    end.time,
	    0, /* No pc2 call stack info. */
	    convert_thread_id(end.thread_id),
	    end.access_size,
	    end.kind));

    sample_ptr_t sample(new fnr_event_sample_t(
			    access_1, access_2,
			    begin.pc_plus_one, 0,
			    line_size_to_log2(begin.line_size),
			    0));

    current_burst->samples_by_pc1.push_back(sample);

    update_acquisition_end(*sample);
}

void
event_collector_t::generate_dangling(const fnr_event_sample_begin_t &begin)
{
    fnr_sample_access_t access(
	begin.pc,
	begin.address,
	begin.time,
	0, /* No pc1 call stack info. */
	convert_thread_id(begin.thread_id),
	begin.access_size,
	begin.kind);

    sample_ptr_t sample(new fnr_event_sample_t(
			    access, fnr_sample_access_vector_t(),
			    begin.pc_plus_one, 0,
			    line_size_to_log2(begin.line_size),
			    0));

    current_burst->samples_by_pc1.push_back(sample);
    update_acquisition_end(*sample);
}

flow_status_t
event_collector_t::flow_init(fnr_features_t features)
{
    fenris_features = features;

    return FLOW_OK;
}

flow_status_t
event_collector_t::read(const fnr_event_read_t &event)
{
    throw feature_exception_t(
        "parser", "Got unexpected read event in sample file.");
}

flow_status_t
event_collector_t::write(const fnr_event_write_t &event)
{
    throw feature_exception_t(
        "parser", "Got unexpected write event in sample file.");
}

flow_status_t
event_collector_t::prefetch(const fnr_event_prefetch_t &event)
{
    throw feature_exception_t(
        "parser", "Got unexpected prefetch event in sample file.");
}

flow_status_t
event_collector_t::sample_begin(const fnr_event_sample_begin_t &event)
{
    test_cancel();

    const fnr_sample_count_t sid = event.sample_id;

    // Check if we already have a corresponding sample end event.
    end_event_map_t::iterator end_it = sample_end_map.find(sid);
    if (end_it != sample_end_map.end()) {
	// Found a matching sample end event. Generate a sample and remove
	// the sample end event from the map.
	generate_sample(event, end_it->second);
	sample_end_map.erase(end_it);
    }
    else
	// No matching sample end event. Add the sample begin event to the
	// map.
	sample_begin_map[sid] = event;
    return FLOW_OK;
}

flow_status_t
event_collector_t::sample_end(const fnr_event_sample_end_t &event)
{
    test_cancel();
    const fnr_sample_count_t sid = event.sample_id;

    update_acquisition_end(event.time);

    // Check if we already have a corresponding sample begin event.
    begin_event_map_t::iterator begin_it = sample_begin_map.find(sid);
    if (begin_it != sample_begin_map.end()) {
	// Found a matching sample begin event. Generate a sample and
	// remove the sample begin event from the map.
	generate_sample(begin_it->second, event);
	sample_begin_map.erase(begin_it);
    }
    else
	// No matching sample begin event. Add the sample end event to the
	// map.
	sample_end_map[sid] = event;
    return FLOW_OK;
}

flow_status_t
event_collector_t::sample_period(const fnr_event_sample_period_t &event)
{
    parser_result->nominal_sample_period = event.sample_period;
    return FLOW_OK;
}

flow_status_t
event_collector_t::acquisition_begin(const fnr_event_acquisition_begin_t &event)
{
    if(parser_result->process_info->acquisition_begin.get() != 0)
	throw feature_exception_t(
	    "parser", "Multiple acquisition begin events in sample file.");
    parser_result->process_info->acquisition_begin.reset(
	new fnr_event_acquisition_begin_t(event));
    return FLOW_OK;
}

flow_status_t
event_collector_t::acquisition_end(const fnr_event_acquisition_end_t &event)
{
    if (parser_result->process_info->acquisition_end.get() != 0)
	throw feature_exception_t(
	    "parser", "Multiple acquisition end events in sample file.");
    parser_result->process_info->acquisition_end.reset(
	new fnr_event_acquisition_end_t(event));
    return FLOW_OK;
}

flow_status_t
event_collector_t::mmap(const fnr_event_mmap_t &event)
{
    event_mmap_ptr_t mmap(new fnr_event_mmap_t(event));
    parser_result->process_info->mmaps.push_back(mmap);
    return FLOW_OK;
}

flow_status_t
event_collector_t::munmap(const fnr_event_munmap_t &event)
{
    /* Quietly ignore munmap events. */
    return FLOW_OK;
}

flow_status_t
event_collector_t::stack(const fnr_event_stack_t &event)
{
    test_cancel();

    fnr_event_stack_t *stack = new fnr_event_stack_t(event);
    event_ptr_t stack_event(stack);
    parser_result->event_store.push_back(stack_event);
    if (event.id >= parser_result->stacks.size())
	parser_result->stacks.resize(event.id * 2, NULL);
    parser_result->stacks[event.id] = stack;
    return FLOW_OK;
}

flow_status_t
event_collector_t::sample(const fnr_event_sample_t &event)
{
    test_cancel();

    generate_sample(event);
    return FLOW_OK;
}

flow_status_t
event_collector_t::time_mapping(const fnr_event_time_mapping_t &event)
{
    test_cancel();

    time_mappings.back().push_back(
	event_time_mapping_ptr_t(new fnr_event_time_mapping_t(event)));

    event_time_mapping_ptr_t &time(time_mappings.back().back());
    BOOST_FOREACH(fnr_event_time_mapping_t::time_event_t &te,
		  time->time_events)
	te.tid = convert_thread_id(te.tid);

    return FLOW_OK;
}

flow_status_t
event_collector_t::local_time(const fnr_event_local_time_t &event)
{
    local_times.back().push_back(
        event_local_time_ptr_t(new fnr_event_local_time_t(event)));
    return FLOW_OK;
}

flow_status_t
event_collector_t::burst_begin(const fnr_event_burst_begin_t &event)
{
    test_cancel();

    // We enter burst_begin with current_burst != NULL the first time
    // we see a burst_begin, burst_end sets current_burst to NULL, so
    // subsequent calls to burst_begin will be with current_burst ==
    // NULL.
    if (!current_burst) {
	parser_result->bursts.push_back(parser_burst_t());
        local_times.push_back(local_time_vector_t());
        time_mappings.push_back(time_mapping_vector_t());
	current_burst = &parser_result->bursts.back();
    } else
	parser_result->burst_sampled = true;

    current_burst->begin.reset(new fnr_event_burst_begin_t(event));

    return FLOW_OK;
}

flow_status_t
event_collector_t::burst_end(const fnr_event_burst_end_t &event)
{
    test_cancel();

    assert(current_burst);
    current_burst->end.reset(new fnr_event_burst_end_t(event));
    current_burst = NULL;

    return FLOW_OK;
}

flow_status_t
event_collector_t::contexts(const fnr_event_contexts_t &contexts)
{
    assert(thread_to_context.empty());

    thread_to_context = contexts.contexts;
    return FLOW_OK;
}

// ----- parser_t -------------------------------------------------------------

parser_result_cptr_t
parser_t::calculate(feature_context_ptr_t context,
		    const string &sample_file)
{
    event_collector_t event_collector(context);
    stream_container_header_t stream_header;

    try {
        // First check that the file actually contains samples.
        {
            stream_container_header_t header(
                header_collector_t(sample_file, context->warning_consumer).header);

            // Check that the stream contains samples.
            switch (header.data_type) {
            case STREAM_CONTAINER_DATA_SAMPLES_BINARY:
                break;
            case STREAM_CONTAINER_DATA_REPORT: {
                stringstream ss;
                ss << "The specified input file " << sample_file
                   << " is a report file, not a sample file.";
                context->warning_consumer->error(ss.str());
                throw feature_exception_t("parser", ss.str());
            }
            default: {
                stringstream ss;
                ss << "The specified input file " << sample_file
                   << " is not a sample file.";
                throw feature_exception_t("parser", ss.str());
            }
            }
        }

	flow_file_reader_t reader(
            sample_file.c_str(),
            &event_collector,
            context->warning_consumer);

	if (reader.dispatch() != STREAM_OK)
	    throw feature_exception_t(
		"parser_module",
		"failed to read samples file: '" + sample_file + "'");

	stream_header = reader.get_header();
    } catch (stream_exception_t se) {
	throw feature_exception_t("parser_module",
				  "Failed to read '" + sample_file +
				  "': " + se.what());
    } catch (flow_exception_t fe) {
	throw feature_exception_t("parser_module",
				  "Failed to read '" + sample_file +
				  "': " + fe.what());
    }


    parser_result_cptr_t result(event_collector.get_result(stream_header));
    const process_info_t &pi(*result->process_info);
    if (pi.acquisition_begin) {
	if (!pi.acquisition_end)
	    throw feature_exception_t("parser_module",
				      "Failed to read '" + sample_file +
				      "': Sample file is truncated.");
    } else
	context->warn(new feature_warning_t(
			  "parser_module",
			  "Sample file doesn't contain a sample acquisition "
			  "begin event, can't check if the file is truncated."));

    return result;
}

parser_t::parser_t(feature_context_ptr_t context,
                   const string &sample_file)
    : feature_t<parser_result_cptr_t>(context),
      sample_file(sample_file)
{
}

parser_result_cptr_t
parser_t::get() const
{
    return get_data();
}

string
parser_t::name() const
{
    return "parser_t(" + sample_file + ")";
}

parser_result_cptr_t
parser_t::calculate() const
{
    return calculate(context, sample_file);
}

END_NAMESPACE;
END_NAMESPACE;
