/// @file analysis/misc/parser.hh

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#ifndef PARSER_HH
#define PARSER_HH

#include "analysis/misc/process_info.hh"
#include "analysis/feature_base/feature_base.hh"
#include "library/stream/stream_container.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace std;

using boost::shared_ptr;

/// A global time and a corresponding thread local time.
struct global_to_thread_t {
    global_to_thread_t(fnr_time_t global_time,
                       fnr_time_t thread_time)
        : global_time(global_time),
          thread_time(thread_time)
        {}

    fnr_time_t global_time;
    fnr_time_t thread_time;
};

/// Vector of global time to thread time mappings.
typedef vector<global_to_thread_t> global_to_thread_vector_t;


/// Global to thread time mapping vectors, indexed by thread id.
typedef vector<global_to_thread_vector_t> time_map_t;
typedef shared_ptr<time_map_t> time_map_ptr_t;

/// Sampled data about a burst.
struct parser_burst_t {
    //@{
    /// Burst event pointers. May be NULL if the parsed file isn't a
    /// burst sample file.
    event_burst_begin_ptr_t begin;
    event_burst_end_ptr_t end;
    //@}

    //@{
    /// Vector with samples, may contain samples with an empty end
    /// vector (danglings). Samples are sorted in pc1 order.
    sample_vector_t samples_by_pc1;
    //@}

    /// Global to thread local time mappings, for each thread.
    time_map_ptr_t time_map;

    /// Start time (accesses)
    fnr_time_t begin_time;

    /// Length of the burst in accesses
    fnr_time_t length;

    /// Sample period in number of accesses. This is the real sample
    /// period, i.e., the number of accesses in the acquisition
    /// divided by the number of samples (including dangling).
    ///
    /// NOTE: All samples regardless of line size are used in the
    /// calculation, so if you, e.g., have samples with two line size
    /// in your sample file and have not filtered out one of them you
    /// will get about half the nominal sample period.
    double sample_period;
};

typedef vector<parser_burst_t> parser_burst_vector_t;
typedef parser_burst_vector_t::size_type parser_burst_id_t;

struct parser_result_t {
    friend class event_collector_t;
    friend class filter_module_t;

public:
    parser_result_t();

    /// Nominal sample period in number of accesses.
    fnr_time_t nominal_sample_period;

    /// Vector of sampling bursts. Contains at least one element.
    parser_burst_vector_t bursts;

    /// Process info.
    process_info_ptr_t process_info;

    /// Vector of stack event pointers.
    typedef vector<const fnr_event_stack_t *> stack_vector_t;

    /// Call stacks, indexed by call stack id. Index 0 contains an
    /// empty stack, to emulate an empty stack for pcs without call
    /// stack information.
    stack_vector_t stacks;

    /// Total number of replacement samples in all bursts.
    fnr_sample_count_t repl_sample_count() const;

    /// Total length of all bursts
    fnr_time_t total_length() const;

    /// True if this data originates from an MT sampling.
    bool mt_sampled() const {
	return fenris_features & FNR_Feature_MT;
    }

    /// Translate a virtual thread ID to a context ID
    fnr_context_id_t get_context_id(fnr_thread_id_t thread) const;

    /// Stream container header from the sample file.
    stream_container_header_t stream_header;

    /// Was the file burst sampled?
    bool burst_sampled;

    /// Available Fenris features
    fnr_features_t fenris_features;

    /// Vector mapping thread IDs to context IDs.
    fnr_context_id_vector_t contexts;

private:
    /// Vector of shared pointers to fenris events.
    typedef vector<event_ptr_t> fnr_event_store_t;

    /// This array holds all the fenris events referenced through
    /// stacks. It should not be used directly.
    fnr_event_store_t event_store;

    /// Fake empty stack a index 0 in the stacks array. It should not
    /// be used directly.
    fnr_event_stack_t empty_stack;
};

typedef shared_ptr<parser_result_t> parser_result_ptr_t;
typedef shared_ptr<const parser_result_t> parser_result_cptr_t;

class parser_t : public feature_t<parser_result_cptr_t> {
public:
    static parser_result_cptr_t calculate(
	feature_context_ptr_t context,
        const string &sample_file);

    parser_t(feature_context_ptr_t context,
             const string &sample_file);
    parser_result_cptr_t get() const;

protected:
    string name() const;
    parser_result_cptr_t calculate() const;

private:
    const string sample_file;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
