/// @file analysis/misc/sweep.cc

// Copyright (c) 2006-2008 Acumem AB.  All rights reserved.

#include "sweep.hh"

#include <boost/lexical_cast.hpp>

#include <math.h>
#include <assert.h>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace boost;
using namespace acumem::library;

sweep_iterator_t::reference
sweep_iterator_t::operator *() const
{
    switch (p.sweep_mode) {
    case SWEEP_INVALID:
	assert(p.sweep_mode != SWEEP_INVALID);
	break;

    case SWEEP_CACHE_SIZE:
    {
	fnr_cache_size_t c = (p.begin_cache_size << (step/p.subdivisions));
	int mod = step % p.subdivisions;
	if (mod != 0)
	    c = (fnr_cache_size_t)(c*exp(log(2.0)*(double) mod/(double) p.subdivisions));
	return sweep_point_t(c, p.begin_cache_line_size);
    }

    case SWEEP_CACHE_LINE_SIZE:
    {
	fnr_line_size_t l = (p.begin_cache_line_size << step);
	return sweep_point_t(p.begin_cache_size, l);
    }

    case SWEEP_NONE:
        return sweep_point_t(p.begin_cache_size, p.begin_cache_line_size);
    }

    /* NOTREACHED */
    return sweep_point_t(0, 0);
}

sweep_iterator_t &
sweep_iterator_t::operator++()
{
    assert(step != -1);
    if (++step == max_steps)
	step = -1;

    return *this;
}

bool
sweep_iterator_t::operator == (const sweep_iterator_t &i) const
{
    if ((step == -1 || i.step == -1) && step != i.step)
	return false;

    return (step == -1 && i.step == -1) ||
	(p.sweep_mode == i.p.sweep_mode &&
	 step == i.step &&
	 p.begin_cache_size == i.p.begin_cache_size &&
	 p.end_cache_size == i.p.end_cache_size &&
	 p.subdivisions == i.p.subdivisions &&
	 p.begin_cache_line_size == i.p.begin_cache_line_size &&
	 p.end_cache_line_size == i.p.end_cache_line_size);
}

sweep_iterator_t::sweep_iterator_t(const parameters_t &p_, const int size)
    : p(p_)
{
    step = 0;
    max_steps = size;
}

sweep_t::sweep_t() {
    parameters.sweep_mode = SWEEP_INVALID;
    parameters.begin_cache_size = 0;
    parameters.end_cache_size = 0;
    parameters.subdivisions = 0;
    parameters.begin_cache_line_size = 0;
    parameters.end_cache_line_size = 0;
}

sweep_t::sweep_t(fnr_cache_size_t begin_cache_size_,
		 fnr_cache_size_t end_cache_size_,
		 int subdivisions_,
		 fnr_line_size_t cache_line_size_)
{
    parameters.sweep_mode = SWEEP_CACHE_SIZE;
    parameters.begin_cache_size = begin_cache_size_;
    parameters.end_cache_size = end_cache_size_;
    parameters.subdivisions = subdivisions_;
    parameters.begin_cache_line_size = cache_line_size_;
    parameters.end_cache_line_size = cache_line_size_;
}

sweep_t::sweep_t(fnr_line_size_t begin_cache_line_size_,
		 fnr_line_size_t end_cache_line_size_,
		 fnr_cache_size_t cache_size_)
{
    parameters.sweep_mode = SWEEP_CACHE_LINE_SIZE;
    parameters.begin_cache_size = cache_size_;
    parameters.end_cache_size = cache_size_;
    parameters.subdivisions = 1;
    parameters.begin_cache_line_size = begin_cache_line_size_;
    parameters.end_cache_line_size = end_cache_line_size_;
}

sweep_t::sweep_t(fnr_cache_size_t cache_size_,
                 fnr_line_size_t line_size_)
{
    parameters.sweep_mode = SWEEP_NONE;
    parameters.begin_cache_size = cache_size_;
    parameters.end_cache_size = cache_size_;
    parameters.subdivisions = 1;
    parameters.begin_cache_line_size = line_size_;
    parameters.end_cache_line_size = line_size_;
}

string
sweep_t::name() const
{
    switch (parameters.sweep_mode)
    {
    case SWEEP_CACHE_LINE_SIZE:
	return "linesweep(" +
	    lexical_cast<string>(parameters.begin_cache_size) + "," +
	    lexical_cast<string>(parameters.begin_cache_line_size) + "," +
	    lexical_cast<string>(parameters.end_cache_line_size) + ")";

    case SWEEP_CACHE_SIZE:
	return "cachesweep(" +
	    lexical_cast<string>(parameters.begin_cache_size) + "," +
	    lexical_cast<string>(parameters.end_cache_size) + "," +
	    lexical_cast<string>(parameters.subdivisions) + "," +
	    lexical_cast<string>(parameters.begin_cache_line_size) + ")";

    case SWEEP_NONE:
        return "nonesweep(" +
	    lexical_cast<string>(parameters.begin_cache_size) + "," +
	    lexical_cast<string>(parameters.begin_cache_line_size) + ")";

    default:
	return "sweep()";
    }
}

int
sweep_t::size() const
{
    int size = 0;

    switch(parameters.sweep_mode) {
    case SWEEP_INVALID:
	assert(parameters.sweep_mode != SWEEP_INVALID);
	break;

    case SWEEP_CACHE_SIZE:
	for(fnr_cache_size_t c = parameters.begin_cache_size;
	    c < parameters.end_cache_size;
	    size += parameters.subdivisions, c *= 2)
	    /* empty */;
	break;

    case SWEEP_CACHE_LINE_SIZE:
	for(fnr_line_size_t l = parameters.begin_cache_line_size;
	    l < parameters.end_cache_line_size;
	    size ++, l *= 2)
	    /* empty */;
	break;

    case SWEEP_NONE:
        size = 0;
	break;
    }

    return size + 1;
}


END_NAMESPACE;
END_NAMESPACE;
