/// @file analysis/random_access/ra_lookup.cc

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#include "analysis/random_access/ra_lookup.hh"
#include <boost/lexical_cast.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace boost;
using namespace std;

// ----- ra_lookup_result_t ---------------------------------------------------

double
ra_lookup_result_t::random_ratio(locationx_t lx) const
{
    const lx_random_info_t::const_iterator it =
        lx_random.find(lx);
    return it == lx_random.end() ? 0.0 : it->second.random_ratio();
}

double
ra_lookup_result_t::random_ratio(const locationx_set_t &lx_set) const
{
    ra_lookup_info_t info;

    BOOST_FOREACH(const locationx_t &lx, lx_set) {
	const lx_random_info_t::const_iterator it =
	    lx_random.find(lx);
	if (it != lx_random.end())
	    info.add(it->second);
    }

    return info.random_ratio();
}

double
ra_lookup_result_t::random_ratio() const
{
    ra_lookup_info_t info;

    BOOST_FOREACH(const lx_random_info_t::value_type &v, lx_random)
	info.add(v.second);

    return info.random_ratio();
}

// ----- ra_lookup_t ----------------------------------------------------------

ra_lookup_result_cptr_t
ra_lookup_t::calculate(
    const locationx_mapper_result_cptr_t &lx_res,
    const fr_lookup_result_cptr_t &frl_res,
    const random_access_result_cptr_t &random_access)
{
    ra_lookup_result_ptr_t result(new ra_lookup_result_t());

    BOOST_FOREACH(const random_access_result_t::random_info_t::value_type &v,
		  random_access->random_info) {
	const locationx_specifier_t &lx_spec(v.first);
	const double random(v.second);
	const locationx_t lx(lx_res->get_locationx(lx_spec));

	result->lx_random[lx].sum_random += random;
    }

    locationx_id_t lx_size((locationx_id_t)lx_res->locationx_spec.size());
    for (locationx_id_t i = 0; i < lx_size; i++) {
	const locationx_t lx(i);
	const double fetches = frl_res->fr_by_lx(lx).by_kind().total_fetches();

	result->lx_random[lx].sum_fetches = fetches;
    }

    return result;
}


ra_lookup_t::ra_lookup_t(
    const feature_context_ptr_t &context,
    const string &sample_file,
    const fnr_stack_depth_t stack_depth,
    const fnr_line_size_t line_size,
    const fnr_line_size_t page_size,
    const thread_mapping_cptr_t &thread_mapping,
    const cpu_model_t &cpu_model,
    const locationx_mapper_cptr_t &lx_mapper,
    const fnr_cache_size_t cache_size,
    const repl_fr_engine_cptr_t &engine)
    : feature_t<ra_lookup_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      lx_mapper(lx_mapper),
      cache_size(cache_size),
      engine(engine)
{
}

ra_lookup_result_cptr_t
ra_lookup_t::get() const
{
    return get_data();
}

string
ra_lookup_t::name() const
{
    return "ra_lookup_t(" +
	lexical_cast<string>(sample_file) + "," +
        lexical_cast<string>(stack_depth) + "," +
	lexical_cast<string>(line_size) + "," +
	lexical_cast<string>(page_size) + "," +
        thread_mapping->name() + "," +
	cpu_model.name() + "," +
        lx_mapper->name() + "," +
	lexical_cast<string>(cache_size) + "," +
        engine->name() + ")";
}

ra_lookup_result_cptr_t
ra_lookup_t::calculate() const
{
    locationx_mapper_result_cptr_t lx_res(lx_mapper->get());

    fr_lookup_result_cptr_t frl_res(
        fr_lookup_t(context, sample_file, stack_depth, line_size, page_size,
                    thread_mapping, cpu_model, lx_mapper, cache_size, engine,
		    FRL_Type_All).get());

    random_access_result_cptr_t random_access(
	random_access_t(context, sample_file, stack_depth, line_size, page_size,
			thread_mapping, cpu_model, cache_size, engine).get());

    return calculate(lx_res, frl_res, random_access);
}

END_NAMESPACE;
END_NAMESPACE;
