/// @file analysis/random_access/ra_lookup.hh

// Copyright (c) 2008 Acumem AB.  All rights reserved.

#include "analysis/random_access/random_access.hh"

#ifndef RA_LOOKUP_HH
#define RA_LOOKUP_HH

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace std;

using boost::shared_ptr;

// ----- ra_lookup_result_t -------------------------------------------------

struct ra_lookup_info_t {
    ra_lookup_info_t()
	: sum_random(0), sum_fetches(0) {}

    // Return global random ratio.
    double random_ratio() const {
	return sum_fetches == 0.0 ? 0.0 : sum_random / sum_fetches;
    }

    void add(const ra_lookup_info_t &pi) {
	sum_random += pi.sum_random;
	sum_fetches += pi.sum_fetches;
    }

    double sum_random;
    double sum_fetches;
};

/// Result of the random access analysis module.
///
/// The random ratio is calculated as 1.0-(abs(pos-neg)/(pos+neg)), where pos
/// is the number of page size samples with positive stride and neg is the
/// number of page size samples with negative stride.
///
/// A ratio of 0.0 means that the accesses seem very regular, 1.0 means that
/// the accesses seem completely random. Anything above 0.2 means that the
/// accesses are pretty random.
struct ra_lookup_result_t {
    friend class ra_lookup_t;

    // Return the random ratio of one locationx.
    double random_ratio(locationx_t lx) const;

    // Return the average random ratio of a set of pcs, weighted by the number
    // of misses of each pc.
    double random_ratio(const locationx_set_t &lx_set) const;

    // Return global random ratio.
    double random_ratio() const;

private:
    ra_lookup_result_t() {};

    /// Maps pcs to their prefetch ratios. Pcs not in the map should be
    /// considered to have prefetch ratio 0.0.
    typedef map<locationx_t, ra_lookup_info_t> lx_random_info_t;
    lx_random_info_t lx_random;
};

typedef shared_ptr<ra_lookup_result_t> ra_lookup_result_ptr_t;
typedef shared_ptr<const ra_lookup_result_t> ra_lookup_result_cptr_t;

// ----- ra_lookup_t --------------------------------------------------------

class ra_lookup_t : public feature_t<ra_lookup_result_cptr_t>  {
public:
    static ra_lookup_result_cptr_t calculate(
        const locationx_mapper_result_cptr_t &lx_res,
        const fr_lookup_result_cptr_t &frl_res,
        const random_access_result_cptr_t &random_access);

    ra_lookup_t(const feature_context_ptr_t &context,
		const string &sample_file,
		const fnr_stack_depth_t stack_depth,
		const fnr_line_size_t line_size,
		const fnr_line_size_t page_size,
		const thread_mapping_cptr_t &thread_mapping,
		const cpu_model_t &cpu_model,
		const locationx_mapper_cptr_t &lx_mapper,
		const fnr_cache_size_t cache_size,
		const repl_fr_engine_cptr_t &engine);
    ra_lookup_result_cptr_t get() const;

protected:
    string name() const;
    ra_lookup_result_cptr_t calculate() const;

private:
    const string sample_file;
    const fnr_stack_depth_t stack_depth;
    const fnr_line_size_t line_size;
    const fnr_line_size_t page_size;
    const thread_mapping_cptr_t thread_mapping;
    const cpu_model_t cpu_model;
    const locationx_mapper_cptr_t lx_mapper;
    const fnr_cache_size_t cache_size;
    const repl_fr_engine_cptr_t engine;
};

END_NAMESPACE;
END_NAMESPACE;

#endif /* RA_LOOKUP_HH */
