/// @file analysis/random_access/random_access.cc

// Copyright (c) 2007-2008, 2010 Acumem AB.  All rights reserved.

// See doc/notes/random_access.txt for a description of the algorithm.

#include "analysis/random_access/random_access.hh"
#include "analysis/feature_base/warnings.hh"
#include <boost/lexical_cast.hpp>
#include <cmath>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace boost;
using namespace std;

/// PEAK_DEF is the number of percent of all positive resp. negative
/// samples that is required for a stride to be marked as a
/// PEAK. default 10%

#define PEAK_DEF 10

/// SUM_PEAK is the number of percent of all pos resp. neg. strides
/// that are marked as PEAKS, that is required to characterize the
/// situation as non-random. default 90%

#define SUM_PEAK 90

/// NUM_PEAK is the maximum number of pos. resp. negative peaks that
/// are allowed (to characterize the situation as non-random. default
/// 5
#define NUM_PEAK 5

// ----- random_access_t ------------------------------------------------------

static inline uint64_t identity(int64_t t) {
    return t;
}

/// Statistics per pc.
struct ra_stat_t {
    ra_stat_t()
        : pos_strides(0),
          neg_strides(0)
        {}

    /// Number of samples with the pc as pc2 and positive or negative stride,
    /// respectively.
    fnr_sample_count_t pos_strides;
    fnr_sample_count_t neg_strides;
    vector<int64_t> pstrides;
    vector<int64_t> nstrides;
};

typedef map<locationx_specifier_t, ra_stat_t> ra_stat_map_t;


random_access_result_cptr_t
random_access_t::calculate(
    const feature_context_ptr_t &context,
    const sample_mapping_result_cptr_t &sample_mapping,
    const locationx_mapper_result_cptr_t &lx_res,
    const fr_lookup_result_cptr_t &fr_lookup)
{
    // Prepare result object.
    random_access_result_ptr_t result(new random_access_result_t());

    // Can't do analysis if there are no page size samples.
    if (!sample_mapping->have_page_samples()) {
	context->warn(new line_size_missing_warning_t(
			  "random_access",
			  "No page size samples available "
			  "in sample file, skipping random "
			  "access analysis.",
			  4096));
	return result;
    }

    // Collect statistics about on-page strides of the pc2 seen in each sample.
    ra_stat_map_t ra_stat_map;
    BOOST_FOREACH(const sample_mapping_burst_t &smb, sample_mapping->bursts) {
	const mapped_repl_page_vector_t &page_samples(smb.repl_page_thread);
	BOOST_FOREACH(const mapped_repl_page_t &s, page_samples) {
	    const mapped_access_t &end(*s.end());
	    fnr_address_t addr1(s.start()->addr());
	    fnr_address_t addr2(end.addr());
	    int64_t stride(addr1.virtual_address - addr2.virtual_address);
	    ra_stat_t &ra_stat(ra_stat_map[locationx_specifier_t(end)]);

	    // We ignore samples that are begun and ended by accesses to the same
	    // address.
	    if (stride > 0) {
		ra_stat.pos_strides++;
		ra_stat.pstrides.push_back(stride);
	    }
	    if (stride < 0) {
		ra_stat.nstrides.push_back(-stride);
		ra_stat.neg_strides++;
	    }
	}
    }

    // Calculate a random ratio for each pc.
    BOOST_FOREACH(const ra_stat_map_t::value_type &v, ra_stat_map) {
        const locationx_specifier_t &loc = v.first;
        const ra_stat_t &ra_stat = v.second;

        double pos = ra_stat.pos_strides;
        double neg = ra_stat.neg_strides;
        if (neg + pos > 0) {
	    histogram_t<equal_size_histogram_t<1> > hist;
	    bin_vector_ptr_t phist =
		hist.get_histogram(ra_stat.pstrides.begin(),
				   ra_stat.pstrides.end(), identity);
	    bin_vector_ptr_t nhist =
		hist.get_histogram(ra_stat.nstrides.begin(),
				   ra_stat.nstrides.end(), identity);

	    uint64_t p_sum = 0;
	    uint64_t p_limit = ra_stat.pstrides.size() * PEAK_DEF / 100;
	    uint64_t n_sum = 0;
	    uint64_t n_limit = ra_stat.nstrides.size() * PEAK_DEF / 100;
	    int p_num = 0;
	    int n_num = 0;

	    BOOST_FOREACH(const bin_t &b, *phist.get()) {
		if (b.count >= p_limit) {
		    p_sum += b.count;
		    p_num ++;
		}
	    }

	    BOOST_FOREACH(const bin_t &b, *nhist.get()) {
		if (b.count >= n_limit) {
		    n_sum += b.count;
		    n_num ++;
		}
	    }

	    if (n_sum >= ra_stat.nstrides.size() * SUM_PEAK / 100 &&
		p_sum >= ra_stat.pstrides.size() * SUM_PEAK / 100 &&
		n_num <= NUM_PEAK &&
		p_num <= NUM_PEAK) {
		result->random_info[loc] = 0;
	    } else {
		double fetches = fr_lookup->fr_by_lx(
		    lx_res->get_locationx(loc)).by_kind().total_fetches();

		result->random_info[loc] = fetches * (
		    1.0 - fabs(pos - neg) / (pos + neg));
	    }
	}
    }

    return result;
}

random_access_t::random_access_t(
    const feature_context_ptr_t &context,
    const string &sample_file,
    const fnr_stack_depth_t stack_depth,
    const fnr_line_size_t line_size,
    const fnr_line_size_t page_size,
    const thread_mapping_cptr_t &thread_mapping,
    const cpu_model_t &cpu_model,
    const fnr_cache_size_t cache_size,
    const repl_fr_engine_cptr_t &engine)
    : feature_t<random_access_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      cache_size(cache_size),
      engine(engine)
{
}

random_access_result_cptr_t
random_access_t::get() const
{
    return get_data();
}

string
random_access_t::name() const
{
    return "random_access(" +
	lexical_cast<string>(sample_file) + "," +
        lexical_cast<string>(stack_depth) + "," +
	lexical_cast<string>(line_size) + "," +
	lexical_cast<string>(page_size) + "," +
	thread_mapping->name() + "," +
	cpu_model.name() + "," +
	lexical_cast<string>(cache_size) + "," +
        engine->name() + ")";
}

random_access_result_cptr_t
random_access_t::calculate() const
{
    sample_mapping_result_cptr_t sample_mapping_page(
	sample_mapping_t(context, sample_file, stack_depth, line_size,
			 page_size, thread_mapping, cpu_model).get());

    locationx_mapper_cptr_t lxm_full(
        new locationx_mapper_t(context, sample_file, stack_depth,
                               thread_mapping, true, true, true));

    locationx_mapper_result_cptr_t lxm_full_res = lxm_full->get();

    fr_lookup_result_cptr_t fr_lookup(
	fr_lookup_t(context, sample_file, stack_depth, line_size, page_size,
		    thread_mapping, cpu_model, lxm_full, cache_size, engine,
		    FRL_Type_All).get());

    return calculate(context, sample_mapping_page, lxm_full_res, fr_lookup);
}

END_NAMESPACE;
END_NAMESPACE;
