/// @file analysis/report/report.cc

// Copyright (c) 2010 Acumem AB.  All rights reserved.

/*
 *  This file is part of ThreadSpotter.
 *
 *  ThreadSpotter is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ThreadSpotter is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with ThreadSpotter.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "analysis/report/report.hh"

#include <boost/foreach.hpp>
#include <iostream>

using namespace boost;
using namespace std;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

rpt_report_t::rpt_report_t()
{
    mt_sampled = false;
    burst_sampled = false;
    begin_time = 0;
    end_time = 0;

    architecture =  RPT_Arch_i386;

    sample_count = 0;
    analysis_time = 0;
    autodetect_failed = false;
    num_cpus = 1;
    num_caches = 1;
    num_cache_levels = 1;
    cache_level = 1;
    cache_size = 65536;
    line_size = 64;
    swpf_active = false;
}

rpt_context_id_t
rpt_report_t::get_context_id(rpt_thread_id_t thread_id) const
{
    assert(thread_id < context_ids.size());
    return context_ids[thread_id];
}

rpt_address_t
rpt_report_t::get_address(rpt_address_id_t address_id) const
{
    assert(address_id < addresses.size());
    return addresses[address_id];
}

rpt_address_id_t
rpt_report_t::get_address_id(rpt_address_t addr) const
{
    rpt_address_vector_t::const_iterator i =
        lower_bound(addresses.begin(), addresses.end(), addr);
    assert(i != addresses.end());
    assert(*i == addr);
    return (rpt_address_id_t)(i - addresses.begin());
}

#if 0
const rpt_sliced_stats_t &
rpt_report_t::get_global_stats() const
{
    assert(global_stats < sliced_stats.size());
    return sliced_stats[global_stats];
}
#endif

rpt_sliced_stats_id_t
rpt_report_t::get_global_stats_id() const
{
    assert(global_stats < sliced_stats.size());
    return global_stats;
}

const rpt_stats_t &
rpt_report_t::get_sliced_stats_global(
    rpt_sliced_stats_id_t sliced_stats_id) const
{
    assert(sliced_stats_id < sliced_stats.size());
    return get_sliced_stats_global(sliced_stats[sliced_stats_id]);
}

const rpt_stats_t &
rpt_report_t::get_sliced_stats_global(
    const rpt_sliced_stats_t &sliced_stats) const
{
    assert(sliced_stats.global_stats < stats.size());
    return stats[sliced_stats.global_stats];
}

const rpt_stats_t &
rpt_report_t::get_sliced_stats_thread(
    rpt_sliced_stats_id_t sliced_stats_id,
    rpt_thread_id_t thread_id) const
{
    assert(sliced_stats_id < sliced_stats.size());
    return get_sliced_stats_thread(sliced_stats[sliced_stats_id], thread_id);
}

const rpt_stats_t &
rpt_report_t::get_sliced_stats_thread(
    const rpt_sliced_stats_t &sliced_stats,
    rpt_thread_id_t thread_id) const
{
    rpt_thread_stats_map_t::const_iterator i =
        sliced_stats.thread_stats.find(thread_id);
    assert(i != sliced_stats.thread_stats.end());
    assert(i->second < stats.size());
    return stats[i->second];
}

rpt_sliced_stats_id_t
rpt_report_t::get_location_sliced_stats_id(
    rpt_location_id_t location) const
{
    assert(location < location_stats.size());
    assert(location_stats[location] < sliced_stats.size());
    return location_stats[location];
}

const rpt_address_id_t &
rpt_report_t::get_location_pc_id(
    rpt_location_id_t location_id) const
{
    assert(location_id < locations.size());
    return locations[location_id].pc_id;
}

const rpt_trunc_stack_t &
rpt_report_t::get_location_trunc_stack(
    rpt_location_id_t location_id) const
{
    assert(location_id < locations.size());
    rpt_trunc_stack_id_t trunc_stack_id =
        locations[location_id].trunc_stack_id;
    assert(trunc_stack_id < trunc_stacks.size());
    return trunc_stacks[trunc_stack_id];
}

const rpt_full_stack_t &
rpt_report_t::get_location_full_stack(
    rpt_location_id_t location_id) const
{
    assert(location_id < locations.size());
    assert (location_id < full_stacks.size());

    return full_stacks[location_id];
}

#if 0
const rpt_sliced_stats_t &
rpt_report_t::get_location_sliced_stats(
    rpt_location_id_t location) const
{
    assert(location < location_stats.size());
    assert(location_stats[location] < sliced_stats.size());
    return sliced_stats[location_stats[location]];
}
#endif

const rpt_stats_t &
rpt_report_t::get_location_global_stats(
    rpt_location_id_t location) const
{
    return get_sliced_stats_global(get_sliced_stats(get_location_sliced_stats_id(location)));
}

rpt_access_classes_t
rpt_report_t::get_location_access_classes(
    rpt_location_id_t location) const
{
    assert(location < locations.size());
    assert(locations[location].pc_id < access_classes.size());
    return access_classes[locations[location].pc_id];
}

rpt_trunc_stack_id_t
rpt_report_t::get_trunc_stack_index(
    const rpt_trunc_stack_t &trunc_stack) const
{
    rpt_trunc_stacks_t::const_iterator i =
        lower_bound(trunc_stacks.begin(), trunc_stacks.end(), trunc_stack);
    assert(i != trunc_stacks.end());
    assert(*i == trunc_stack);
    return (rpt_trunc_stack_id_t)(i - trunc_stacks.begin());
}

const rpt_location_set_t &
rpt_report_t::get_location_set(
    rpt_location_set_id_t location_set_id) const
{
    assert(location_set_id < location_sets.size());
    return location_sets[location_set_id];
}

rpt_location_set_id_t
rpt_report_t::get_location_set_id(
    const rpt_location_set_t &location_set) const
{
    rpt_location_set_vector_t::const_iterator i =
        lower_bound(location_sets.begin(), location_sets.end(), location_set);
    assert(i != location_sets.end());
    assert(*i == location_set);
    return (rpt_location_set_id_t)(i - location_sets.begin());
}

rpt_sliced_stats_id_t
rpt_report_t::get_depgroup_stats_id(
    rpt_depgroup_id_t depgroup_id) const
{
    rpt_depgroup_map_t::const_iterator i = depgroups.find(depgroup_id);
    assert(i != depgroups.end());
    assert(i->second.location_set_id < location_set_stats.size());
    assert(location_set_stats[i->second.location_set_id] < sliced_stats.size());
    return location_set_stats[i->second.location_set_id];
}

#if 0
const rpt_sliced_stats_t &
rpt_report_t::get_depgroup_stats(
    rpt_depgroup_id_t depgroup_id) const
{
    rpt_depgroup_map_t::const_iterator i = depgroups.find(depgroup_id);
    assert(i != depgroups.end());
    assert(i->second.location_set_id < location_set_stats.size());
    assert(location_set_stats[i->second.location_set_id] < sliced_stats.size());
    return sliced_stats[location_set_stats[i->second.location_set_id]];
}
#endif

rpt_location_set_id_t
rpt_report_t::get_depgroup_location_set_id(
    rpt_depgroup_id_t depgroup_id) const {
    rpt_depgroup_map_t::const_iterator i = depgroups.find(depgroup_id);
    assert(i != depgroups.end());
    return i->second.location_set_id;
}

rpt_location_set_t
rpt_report_t::get_depgroup_location_set(
    rpt_depgroup_id_t depgroup_id) const {
    rpt_depgroup_map_t::const_iterator i = depgroups.find(depgroup_id);
    assert(i != depgroups.end());
    assert(i->second.location_set_id < location_sets.size());
    return location_sets[i->second.location_set_id];
}

rpt_depgroup_id_t
rpt_report_t::get_location_depgroup(rpt_location_id_t location) const
{
    BOOST_FOREACH(const rpt_depgroup_map_t::value_type &v, depgroups) {
        assert(v.second.location_set_id < location_sets.size());
        const rpt_location_set_t &locations =
            location_sets[v.second.location_set_id];
        if (locations.find(location) != locations.end())
            return v.first;
    }
    while (1)
        assert(0);
}

rpt_location_set_id_t
rpt_report_t::get_loop_incl_location_set_id(rpt_loop_id_t loop_id) const
{
    rpt_loop_map_t::const_iterator i = loops.find(loop_id);
    assert(i != loops.end());
    return i->second.incl_location_set_id;
}

rpt_location_set_t
rpt_report_t::get_loop_incl_location_set(rpt_loop_id_t loop_id) const
{
    return get_location_set(get_loop_incl_location_set_id(loop_id));
}

rpt_location_set_id_t
rpt_report_t::get_loop_excl_location_set_id(rpt_loop_id_t loop_id) const
{
    rpt_loop_map_t::const_iterator i = loops.find(loop_id);
    assert(i != loops.end());
    return i->second.excl_location_set_id;
}

rpt_location_set_t
rpt_report_t::get_loop_excl_location_set(rpt_loop_id_t loop_id) const
{
    return get_location_set(get_loop_excl_location_set_id(loop_id));
}

rpt_sliced_stats_id_t
rpt_report_t::get_loop_stats_id(
    rpt_loop_id_t loop_id) const
{
    rpt_loop_map_t::const_iterator i = loops.find(loop_id);
    assert(i != loops.end());
    assert(i->second.excl_location_set_id < location_set_stats.size());
    assert(location_set_stats[i->second.excl_location_set_id] < sliced_stats.size());
    return location_set_stats[i->second.excl_location_set_id];
}

#if 0
const rpt_sliced_stats_t &
rpt_report_t::get_loop_stats(
    rpt_loop_id_t loop_id) const
{
    rpt_loop_map_t::const_iterator i = loops.find(loop_id);
    assert(i != loops.end());
    assert(i->second.excl_location_set_id < location_set_stats.size());
    assert(location_set_stats[i->second.excl_location_set_id] < sliced_stats.size());
    return sliced_stats[location_set_stats[i->second.excl_location_set_id]];
}
#endif

const rpt_sliced_stats_t &
rpt_report_t::get_sliced_stats(rpt_sliced_stats_id_t id) const
{
    assert(id < sliced_stats.size());
    return sliced_stats[id];
}

rpt_binary_id_t
rpt_report_t::get_binary_id(
    const rpt_binary_t &binary) const
{
    rpt_binary_vector_t::const_iterator i =
        lower_bound(binaries.begin(), binaries.end(), binary);
    assert(i != binaries.end());
    assert(*i == binary);
    return (rpt_binary_id_t)(i - binaries.begin());
}

rpt_function_id_t
rpt_report_t::get_function_id(
    const rpt_function_t &function) const
{
    rpt_function_vector_t::const_iterator i =
        lower_bound(functions.begin(), functions.end(), function);
    assert(i != functions.end());
    assert(*i == function);
    return (rpt_function_id_t)(i - functions.begin());
}

rpt_symbol_id_t
rpt_report_t::get_symbol_id(
    const rpt_symbol_t &symbol) const
{
    rpt_symbol_vector_t::const_iterator i =
        lower_bound(symbols.begin(), symbols.end(), symbol);
    assert(i != symbols.end());
    assert(*i == symbol);
    return (rpt_symbol_id_t)(i - symbols.begin());
}

rpt_source_file_id_t
rpt_report_t::get_source_file_id(
    const rpt_source_file_t &source_file) const
{
#if 0
    cout << "get_souce_file_id: " << source_file.filename << endl;
    cout << "source_files:" << endl;
    BOOST_FOREACH(const rpt_source_file_vector_t::value_type &v, source_files)
	cout << "  " << v.filename << endl;
#endif

    rpt_source_file_vector_t::const_iterator i =
        lower_bound(source_files.begin(), source_files.end(), source_file);

    assert(i != source_files.end());
    assert(*i == source_file);
    return (rpt_source_file_id_t)(i - source_files.begin());
}

rpt_source_file_id_t
rpt_report_t::get_source_file_id(
    const std::string &filename) const
{
    return get_source_file_id(rpt_source_file_t(filename));
}

rpt_source_line_id_t
rpt_report_t::get_source_line_id(
    const rpt_source_line_t &source_line) const
{
    rpt_source_line_vector_t::const_iterator i =
        lower_bound(source_lines.begin(), source_lines.end(), source_line);
    assert(i != source_lines.end());
    assert(*i == source_line);
    return (rpt_source_line_id_t)(i - source_lines.begin());
}

rpt_debug_pc_id_t
rpt_report_t::get_debug_pc_id(
    rpt_address_id_t pc_id,
    bool return_address) const
{
    rpt_debug_pc_t debug_pc(pc_id, return_address);
    rpt_debug_pc_vector_t::const_iterator i =
        lower_bound(debug_pcs.begin(), debug_pcs.end(), debug_pc);
    assert(i != debug_pcs.end());
    assert(*i == debug_pc);
    return (rpt_debug_pc_id_t)(i - debug_pcs.begin());
}

END_NAMESPACE;
END_NAMESPACE;
