/// @file analysis/report/report.hh

// Copyright (c) 2010-2011 Acumem AB.  All rights reserved.

#ifndef REPORT_REPORT_HH
#define REPORT_REPORT_HH

#include "analysis/report/report_issues.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

/// Intermediate format report.
struct rpt_report_t {
    /// Default constructor
    rpt_report_t();

    /// Unique ID for this report.
    string uuid;

    // ----- Sampling info. ---------------------------------------------------

    /// Was the sample file MT sampled?
    bool mt_sampled;

    /// Was the sample file burst sampled?
    bool burst_sampled;

    /// Sampling begin time.
    rpt_real_time_t begin_time;

    /// Sampling end time.
    rpt_real_time_t end_time;

    /// Command line of the sampled process.
    vector<string> argv;

    /// User that did the sampling.
    string user;

    /// Sampled architecture.
    rpt_architecture_t architecture;

    /// Number of line samples in the sample file.
    rpt_sample_count_t sample_count;

    // ----- Analysis info. ---------------------------------------------------

    /// Absolute path to sample file.
    string sample_file;

    /// Analysis time.
    rpt_real_time_t analysis_time;

    /// Acurate command line.
    vector<string> analysis_argv;

    /// Report title, or empty if the default title should be used.
    string title;

    // ----- Hardware configuration. ------------------------------------------

    /// Is the model autodetected but unreliable, like when
    /// autodetecting the cpu and cache hierarchy in a virtualized
    /// environment.
    bool autodetect_failed;

    /// Processor model.
    string cpu_model;

    /// Number of processors.
    uint32_t num_cpus;

    /// Number of caches in the thread mapping.
    uint32_t num_caches;

    /// Number of cache levels.
    uint32_t num_cache_levels;

    /// Cache level.
    uint32_t cache_level;

    /// Analyzed cache size.
    rpt_size_t cache_size;

    /// Analyzed line size.
    rpt_size_t line_size;

    /// Replacement policy.
    string replacement_policy;

    /// Are any software prefetch instructions active?
    bool swpf_active;

    // ----- Program structure. -----------------------------------------------

    /// Context IDs, indexed by thread ID.
    rpt_context_id_vector_t context_ids;

    /// All statistics records, sorted in natural order.
    rpt_stats_vector_t stats;

    /// All sliced statistics records, sorted in natural order.
    rpt_sliced_stats_vector_t sliced_stats;

    /// All sweeps.
    rpt_sweep_vector_t sweeps;

    /// All statistics sweeps.
    rpt_stats_sweep_vector_t stats_sweeps;

    /// All addresses, sorted in natural order.
    rpt_address_vector_t addresses;

    /// Access classes corresponding to each address above.
    rpt_access_classes_vector_t access_classes;

    /// Truncated call stacks. Sorted in natural order, duplicates removed.
    rpt_trunc_stacks_t trunc_stacks;

    /// Full call stacks.
    rpt_full_stacks_t full_stacks;

    /// Sliced statistics ID for global statistics.
    rpt_sliced_stats_id_t global_stats;

    /// Global statistics sweep.
    rpt_stats_sweep_id_t global_stats_sweep;

    /// Locations, indexed by location ID.
    rpt_location_vector_t locations;

    /// Sliced statistics IDs, indexed by location ID.
    rpt_sliced_stats_id_vector_t location_stats;

    /// Location (ID) sets, sorted in natural order. We only include the
    /// location sets we have done a drill-down of.
    rpt_location_set_vector_t location_sets;

    /// Sliced statistics IDs, indexed by location set ID.
    rpt_sliced_stats_id_vector_t location_set_stats;

    /// Statistics sweep for each location set, indexed by location set ID.
    rpt_stats_sweep_id_vector_t location_set_stats_sweeps;

    /// Dependence groups indexed by ID. We only include the dependence groups
    /// that we have done a drill-down of.
    rpt_depgroup_map_t depgroups;

    /// Loops indexed by ID. We only include the loops that we have done a
    /// drill-down of.
    rpt_loop_map_t loops;

    /// Threads mapped to each cache, indexed by cache id.
    rpt_thread_id_set_vector_t cache_threads;

    // ----- Issues. ----------------------------------------------------------

    /// All issue pointers indexed by issue ID. Entries may by NULL if the file
    /// we loaded contains unknown issue types.
    rpt_issue_vector_t issues;

    // ----- Debug info and source stuff. -------------------------------------

    /// TODO
    rpt_binary_vector_t binaries;

    /// TODO
    rpt_symbol_vector_t symbols;

    /// TODO
    rpt_function_vector_t functions;

    /// TODO
    rpt_source_file_vector_t source_files;

    /// TODO
    rpt_source_line_vector_t source_lines;

    /// TODO
    rpt_debug_pc_vector_t debug_pcs;

    // ----- Accessor functions. ----------------------------------------------

    /// Return the context ID corresponding to the specified thread ID.
    rpt_context_id_t get_context_id(rpt_thread_id_t thread_id) const;

    /// Return the address corresponding to the specified address ID.
    rpt_address_t get_address(rpt_address_id_t address_id) const;

    /// Return the ID of the specified address.
    uint32_t get_address_id(rpt_address_t addr) const;

    /// TODO
    rpt_sliced_stats_id_t get_global_stats_id() const;
#if 0
    const rpt_sliced_stats_t &get_global_stats() const;
#endif

    /// TODO
    const rpt_stats_t &get_sliced_stats_global(
        rpt_sliced_stats_id_t sliced_stats_id) const;

    /// TODO
    const rpt_stats_t &get_sliced_stats_global(
        const rpt_sliced_stats_t &sliced_stats) const;

    /// TODO
    const rpt_stats_t &get_sliced_stats_thread(
        rpt_sliced_stats_id_t sliced_stats_id,
        rpt_thread_id_t thread_id) const;

    /// TODO
    const rpt_stats_t &get_sliced_stats_thread(
        const rpt_sliced_stats_t &sliced_stats,
        rpt_thread_id_t thread_id) const;

    /// TODO
    const rpt_address_id_t &get_location_pc_id(
        rpt_location_id_t location_id) const;

    /// TODO
    const rpt_trunc_stack_t &get_location_trunc_stack(
        rpt_location_id_t location_id) const;

    const rpt_full_stack_t &get_location_full_stack(
	rpt_location_id_t location_id) const;

    /// TODO: Return sliced statistics for the specified location.
    /// TODO: reference
    rpt_sliced_stats_id_t get_location_sliced_stats_id(
        rpt_location_id_t location) const;
#if 0
    const rpt_sliced_stats_t &get_location_sliced_stats(
        rpt_location_id_t location) const;
#endif

    /// TODO
    const rpt_stats_t &get_location_global_stats(
        rpt_location_id_t location) const;

    /// Return the access classes of the specified location.
    rpt_access_classes_t get_location_access_classes(
        rpt_location_id_t location) const;

    /// Return the index of the specified truncated stack.
    rpt_trunc_stack_id_t get_trunc_stack_index(
        const rpt_trunc_stack_t &trunc_stack) const;

    /// Return the location set corresponding to the specified location set ID.
    const rpt_location_set_t &get_location_set(
        rpt_location_set_id_t location_set_id) const;

    /// Return the ID of the specified location set.
    rpt_location_set_id_t get_location_set_id(
        const rpt_location_set_t &location_set) const;

    /// Return sliced statistics for the specified dependence group.
    rpt_sliced_stats_id_t get_depgroup_stats_id(
        rpt_depgroup_id_t depgroup_id) const;
#if 0
    const rpt_sliced_stats_t &get_depgroup_stats(
        rpt_depgroup_id_t depgroup_id) const;
#endif
    /// TODO
    rpt_location_set_id_t get_depgroup_location_set_id(
        rpt_depgroup_id_t depgroup_id) const;

    /// TODO
    rpt_location_set_t get_depgroup_location_set(
        rpt_depgroup_id_t depgroup_id) const;

    /// Return the ID of the dependence group containing the specified
    /// location.
    rpt_depgroup_id_t get_location_depgroup(rpt_location_id_t location) const;

    /// TODO
    rpt_location_set_id_t get_loop_incl_location_set_id(rpt_loop_id_t loop_id) const;

    /// Return the inclusive location set of the specified loop.
    rpt_location_set_t get_loop_incl_location_set(rpt_loop_id_t loop_id) const;

    /// TODO
    rpt_location_set_id_t get_loop_excl_location_set_id(rpt_loop_id_t loop_id) const;

    /// Return the exclusive location set of the specified loop.
    rpt_location_set_t get_loop_excl_location_set(rpt_loop_id_t loop_id) const;

    /// Return sliced statistics for exlusive location set of the specified
    /// loop.
    rpt_sliced_stats_id_t get_loop_stats_id(rpt_loop_id_t loop_id) const;
#if 0
    const rpt_sliced_stats_t &get_loop_stats(rpt_loop_id_t loop_id) const;
#endif

    /// Get the sliced_stats_t object for a dito id
    const rpt_sliced_stats_t &get_sliced_stats(
	rpt_sliced_stats_id_t id) const;

    /// TODO
    rpt_binary_id_t get_binary_id(
        const rpt_binary_t &binary) const;

    /// TODO
    rpt_function_id_t get_function_id(
        const rpt_function_t &function) const;

    /// TODO
    rpt_symbol_id_t get_symbol_id(
        const rpt_symbol_t &symbol) const;

    /// TODO
    rpt_source_file_id_t get_source_file_id(
        const rpt_source_file_t &source_file) const;
    rpt_source_file_id_t get_source_file_id(
	const std::string &filename) const;

    /// TODO
    rpt_source_line_id_t get_source_line_id(
        const rpt_source_line_t &source_line) const;

    /// TODO
    rpt_debug_pc_id_t get_debug_pc_id(
        rpt_address_id_t pc_id,
        bool return_address) const;
};

typedef shared_ptr<rpt_report_t> rpt_report_ptr_t;
typedef shared_ptr<const rpt_report_t> rpt_report_cptr_t;

END_NAMESPACE;
END_NAMESPACE;

#endif /* REPORT_HH */
