/// @file analysis/report/report_issues.hh

// Copyright (c) 2010 Acumem AB.  All rights reserved.

#ifndef REPORT_REPORT_ISSUES_HH
#define REPORT_REPORT_ISSUES_HH

#include "analysis/report/report_stats.hh"
#include "analysis/report/report_debug.hh"

#include <boost/shared_ptr.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

enum rpt_issue_type_t {
    Rpt_Issue_Prefetch,
    Rpt_Issue_Hot_Spot,
    Rpt_Issue_Utilization,
    Rpt_Issue_Loop_Order,
    Rpt_Issue_Random,
    Rpt_Issue_Loop_Fusion,
    Rpt_Issue_Block,
    Rpt_Issue_Comm_Spot,
    Rpt_Issue_Comm_Utilization,
    Rpt_Issue_False_Sharing,
    Rpt_Issue_Wnt_Possible,
    Rpt_Issue_Prefetchnta
};

typedef uint32_t rpt_issue_id_t;

// ----- rpt_issue_t ----------------------------------------------------------

struct rpt_issue_t {
    rpt_issue_t(rpt_issue_id_t issue_id)
        : issue_id(issue_id)
        {}

    virtual ~rpt_issue_t()
        {}

    /// TODO
    rpt_issue_id_t get_issue_id() const {
        return issue_id;
    }

    /// TODO
    virtual rpt_issue_type_t get_issue_type() const = 0;

    /// TODO
    virtual rpt_depgroup_id_t get_depgroup_id() const = 0;

private:
    rpt_issue_id_t issue_id;
};

typedef shared_ptr<rpt_issue_t> rpt_issue_ptr_t;
typedef shared_ptr<const rpt_issue_t> rpt_issue_cptr_t;

typedef vector<rpt_issue_ptr_t> rpt_issue_vector_t;

// ----- rpt_prefetch_issue_t -------------------------------------------------

struct rpt_prefetch_issue_t
    : public rpt_issue_t {

    enum problem_t {
        Unnecessary,
        Too_Far,
        Too_Close
    };

    rpt_prefetch_issue_t(
        rpt_issue_id_t issue_id)
        : rpt_issue_t(issue_id)
        {}

    rpt_prefetch_issue_t(
        rpt_issue_id_t issue_id,
        problem_t problem,
        rpt_location_id_t location,
        rpt_loop_id_t loop_id,
        rpt_depgroup_id_t depgroup_id,
        rpt_stat_fp_t cost,
        rpt_stat_fp_t fetch_ratio_,
        rpt_access_count_t median_rdist,
        const rpt_location_set_t &prev_locations,
        const rpt_location_set_t &next_locations)
        : rpt_issue_t(issue_id),
          problem(problem),
          location(location),
          loop_id(loop_id),
          depgroup_id(depgroup_id),
          cost(cost),
          fetch_ratio(fetch_ratio_),
          median_rdist(median_rdist),
          prev_locations(prev_locations),
          next_locations(next_locations)
        {}

    rpt_issue_type_t get_issue_type() const {
        return Rpt_Issue_Prefetch;
    }

    rpt_depgroup_id_t get_depgroup_id() const {
        return depgroup_id;
    }

    /// Type of problem.
    problem_t problem;

    /// Location of the prefetch instruction.
    rpt_location_id_t location;

    /// Loop containing the prefetch.
    rpt_loop_id_t loop_id;

    /// Dependency group containing the prefetch.
    rpt_depgroup_id_t depgroup_id;

    /// Cost of the issue in cache miss equivalents. Can be negative if the
    /// prefetch does more good than harm.
    rpt_stat_fp_t cost;

    /// Unnecessary - fetch ratio of prefetch instruction.
    /// Too_Far     - fetch ratio of following instruction.
    rpt_stat_fp_t fetch_ratio;

    /// Too_Close only.
    rpt_access_count_t median_rdist;

    /// Previous and next locations using the same data.
    rpt_location_set_t prev_locations;
    rpt_location_set_t next_locations;
};

// ----- rpt_hot_spot_issue_t -------------------------------------------------

struct rpt_hot_spot_issue_t
    : public rpt_issue_t {

    enum type_t {
        Fetch,
        Write_Back
    };

    rpt_hot_spot_issue_t(
        rpt_issue_id_t issue_id)
        : rpt_issue_t(issue_id)
        {}

    rpt_hot_spot_issue_t(
        rpt_issue_id_t issue_id,
        type_t type,
        rpt_loop_id_t loop_id,
        rpt_depgroup_id_t depgroup_id,
        const rpt_location_set_t &previous_writers)
        : rpt_issue_t(issue_id),
          type(type),
          loop_id(loop_id),
          depgroup_id(depgroup_id),
          previous_writers(previous_writers)
        {}

    rpt_issue_type_t get_issue_type() const {
        return Rpt_Issue_Hot_Spot;
    }

    rpt_depgroup_id_t get_depgroup_id() const {
        return depgroup_id;
    }

    /// Hot spot type.
    type_t type;

    /// Loop containing the hot spot.
    rpt_loop_id_t loop_id;

    /// Dependency group containing the hot spot.
    rpt_depgroup_id_t depgroup_id;

    // Previous locations writing to the same data.
    rpt_location_set_t previous_writers;
};

// ----- rpt_utilization_issue_t ----------------------------------------------

struct rpt_utilization_issue_t
    : public rpt_issue_t {

    enum type_t {
        Legacy,
        Fetch,
        Write_Back
    };

    rpt_utilization_issue_t(
        rpt_issue_id_t issue_id)
        : rpt_issue_t(issue_id)
        {}

    rpt_utilization_issue_t(
        rpt_issue_id_t issue_id,
        type_t type,
        rpt_loop_id_t loop_id,
        rpt_depgroup_id_t depgroup_id,
        const rpt_location_set_t &previous_writers)
        : rpt_issue_t(issue_id),
          type(type),
          loop_id(loop_id),
          depgroup_id(depgroup_id),
          previous_writers(previous_writers)
        {}

    rpt_issue_type_t get_issue_type() const {
        return Rpt_Issue_Utilization;
    }

    rpt_depgroup_id_t get_depgroup_id() const {
        return depgroup_id;
    }

    /// Utilization type.
    type_t type;

    /// Loop id.
    rpt_loop_id_t loop_id;

    /// Dependency group id.
    rpt_depgroup_id_t depgroup_id;

    /// Tracks last write to this dependency group
    rpt_location_set_t previous_writers;
};

// ----- rpt_loop_order_issue_t -----------------------------------------------

struct rpt_loop_order_issue_t
    : public rpt_issue_t {

    rpt_loop_order_issue_t(
        rpt_issue_id_t issue_id)
        : rpt_issue_t(issue_id)
        {}

    rpt_loop_order_issue_t(
        rpt_issue_id_t issue_id,
        rpt_loop_id_t loop_id,
        rpt_depgroup_id_t depgroup_id,
        const rpt_location_set_t &previous_writers)
        : rpt_issue_t(issue_id),
          loop_id(loop_id),
          depgroup_id(depgroup_id),
          previous_writers(previous_writers)
        {}

    rpt_issue_type_t get_issue_type() const {
        return Rpt_Issue_Loop_Order;
    }

    rpt_depgroup_id_t get_depgroup_id() const {
        return depgroup_id;
    }

    /// Loop id.
    rpt_loop_id_t loop_id;

    /// Dependency group id.
    rpt_depgroup_id_t depgroup_id;

    /// Tracks last write to this dependency group
    rpt_location_set_t previous_writers;
};

// ----- rpt_random_issue_t -----------------------------------------------

struct rpt_random_issue_t
    : public rpt_issue_t {

    rpt_random_issue_t(
        rpt_issue_id_t issue_id)
        : rpt_issue_t(issue_id)
        {}

    rpt_random_issue_t(
        rpt_issue_id_t issue_id,
        rpt_loop_id_t loop_id,
        rpt_depgroup_id_t depgroup_id,
        const rpt_location_set_t &previous_writers)
        : rpt_issue_t(issue_id),
          loop_id(loop_id),
          depgroup_id(depgroup_id),
          previous_writers(previous_writers)
        {}

    rpt_issue_type_t get_issue_type() const {
        return Rpt_Issue_Random;
    }

    rpt_depgroup_id_t get_depgroup_id() const {
        return depgroup_id;
    }

    /// Loop id.
    rpt_loop_id_t loop_id;

    /// Dependency group id.
    rpt_depgroup_id_t depgroup_id;

    /// Tracks last write to this dependency group
    rpt_location_set_t previous_writers;
};

// ----- rpt_loop_fusion_issue_t -----------------------------------------------

struct rpt_loop_fusion_issue_t
    : public rpt_issue_t {

    rpt_loop_fusion_issue_t(
        rpt_issue_id_t issue_id)
        : rpt_issue_t(issue_id)
        {}

    rpt_loop_fusion_issue_t(
        rpt_issue_id_t issue_id,
        rpt_loop_id_t loop1_id,
        rpt_loop_id_t loop2_id,
        rpt_location_set_t loop1_pcs,
        rpt_location_set_t loop2_pcs,
        rpt_depgroup_id_t depgroup1_id,
        rpt_depgroup_id_t depgroup2_id,
        bool move_loop2_up,
        bool move_loop1_down,
        rpt_location_id_vector_t loop1_down_pcs,
        rpt_location_set_vector_t loop1_down_blocks,
        rpt_location_id_vector_t loop2_up_pcs,
        rpt_location_set_vector_t loop2_up_blocks,
        rpt_location_set_t additional_pc1,
        rpt_location_set_t additional_pc2)
        : rpt_issue_t(issue_id),
          loop1_id(loop1_id),
          loop2_id(loop2_id),
          loop1_pcs(loop1_pcs),
          loop2_pcs(loop2_pcs),
          depgroup1_id(depgroup1_id),
          depgroup2_id(depgroup2_id),
          move_loop2_up(move_loop2_up),
          move_loop1_down(move_loop1_down),
          loop1_down_pcs(loop1_down_pcs),
          loop1_down_blocks(loop1_down_blocks),
          loop2_up_pcs(loop2_up_pcs),
          loop2_up_blocks(loop2_up_blocks),
          additional_pc1(additional_pc1),
          additional_pc2(additional_pc2)
        {}

    rpt_issue_type_t get_issue_type() const {
        return Rpt_Issue_Loop_Fusion;
    }

    rpt_depgroup_id_t get_depgroup_id() const {
        return depgroup2_id;
    }

    /// Loop ids.
    rpt_loop_id_t loop1_id;
    rpt_loop_id_t loop2_id;

    /// Instructions that are joined by a temporal or spatial sample.
    /// loop1_pcs and loop2_pcs are the exact start and end locations of
    /// joining samples.
    rpt_location_set_t loop1_pcs;
    rpt_location_set_t loop2_pcs;

    /// Dependence groups containing the loop1_pcs and loop2_pcs.
    rpt_depgroup_id_t depgroup1_id;
    rpt_depgroup_id_t depgroup2_id;

    /// Is it possible to move loop2 up into loop1, or vice versa, or both?
    bool move_loop2_up;
    bool move_loop1_down;

    /// Report blockers for each instruction in the loop separately.
    rpt_location_id_vector_t loop1_down_pcs;
    rpt_location_set_vector_t loop1_down_blocks;
    rpt_location_id_vector_t loop2_up_pcs;
    rpt_location_set_vector_t loop2_up_blocks;

    /// Report instructions involved in 1-to-n split-fusions and n-to-1
    /// split-fusions
    rpt_location_set_t additional_pc1;
    rpt_location_set_t additional_pc2;
};

// ----- rpt_block_issue_t ----------------------------------------------------

struct rpt_block_issue_t
    : public rpt_issue_t {

    enum type_t {
        Temporal = 1,
        Spatial = 2
    };

    rpt_block_issue_t(
        rpt_issue_id_t issue_id)
        : rpt_issue_t(issue_id)
        {}

    rpt_block_issue_t(
        rpt_issue_id_t issue_id,
        uint8_t type_mask,
        rpt_loop_id_t loop_id,
        rpt_depgroup_id_t depgroup_id,
        rpt_loop_id_t outer_loop_id,
        rpt_access_count_t outer_loop_iterations,
        rpt_loop_id_t next_outer_loop_id,
        rpt_access_count_t next_outer_loop_iterations,
        rpt_location_id_vector_t barrier_pcs,
        rpt_location_set_vector_t barrier_blocks,
        rpt_loop_id_t target_loop_id,
        rpt_access_count_t target_loop_iterations,
        const rpt_location_set_t &previous_writers,
        const rpt_location_set_t &next_writers)
        : rpt_issue_t(issue_id),
          type_mask(type_mask),
          loop_id(loop_id),
          depgroup_id(depgroup_id),
          outer_loop_id(outer_loop_id),
          outer_loop_iterations(outer_loop_iterations),
          next_outer_loop_id(next_outer_loop_id),
          next_outer_loop_iterations(next_outer_loop_iterations),
          barrier_pcs(barrier_pcs),
          barrier_blocks(barrier_blocks),
          target_loop_id(target_loop_id),
          target_loop_iterations(target_loop_iterations),
          previous_writers(previous_writers),
          next_writers(next_writers)
        {}

    rpt_issue_type_t get_issue_type() const {
        return Rpt_Issue_Block;
    }

    rpt_depgroup_id_t get_depgroup_id() const {
        return depgroup_id;
    }

    /// Blocking type; temporal, spatial or both..
    uint8_t type_mask;

    /// Loop id.
    rpt_loop_id_t loop_id;

    /// Dependency group id.
    rpt_depgroup_id_t depgroup_id;

    /// Outermost loop level where it is still possible to block with respect
    /// to the reuse instruction, and iterations of the outer loop per data
    /// reuse.
    rpt_loop_id_t outer_loop_id;
    rpt_access_count_t outer_loop_iterations;

    /// If barriers are detected, and if the loop hieararchy is complete, this
    /// is the next outer loop level, the one that contains the
    /// barriers. Otherwise the loop id is 0.
    rpt_loop_id_t next_outer_loop_id;
    rpt_access_count_t next_outer_loop_iterations;

    /// Blocking barriers.
    rpt_location_id_vector_t barrier_pcs;
    rpt_location_set_vector_t barrier_blocks;

    /// Target loop denotes the loop that has a length on par with the
    /// trigger reuse distance.
    rpt_loop_id_t target_loop_id;
    rpt_access_count_t target_loop_iterations;

    /// Previous and next writers of the data.
    rpt_location_set_t previous_writers;
    rpt_location_set_t next_writers;
};

// ----- rpt_comm_spot_issue_t ------------------------------------------------

struct rpt_comm_spot_issue_t
    : public rpt_issue_t {

    rpt_comm_spot_issue_t(
        rpt_issue_id_t issue_id)
        : rpt_issue_t(issue_id)
        {}

    rpt_comm_spot_issue_t(
        rpt_issue_id_t issue_id,
        rpt_loop_id_t loop_id,
        rpt_depgroup_id_t depgroup_id,
        rpt_stat_fp_t comm_lines,
        const rpt_location_set_t &producers,
        const rpt_location_set_t &previous_writers,
        const rpt_location_set_t &next_writers)
        : rpt_issue_t(issue_id),
          loop_id(loop_id),
          depgroup_id(depgroup_id),
          comm_lines(comm_lines),
          producers(producers),
          previous_writers(previous_writers),
          next_writers(next_writers)
        {}

    rpt_issue_type_t get_issue_type() const {
        return Rpt_Issue_Comm_Spot;
    }

    rpt_depgroup_id_t get_depgroup_id() const {
        return depgroup_id;
    }

    /// Loop id.
    rpt_loop_id_t loop_id;

    /// Dependency group id.
    rpt_depgroup_id_t depgroup_id;

    /// Communicated lines
    rpt_stat_fp_t comm_lines;

    /// List of data producers
    rpt_location_set_t producers;

    /// Previous and next writers of the data.
    rpt_location_set_t previous_writers;
    rpt_location_set_t next_writers;
};

// ----- rpt_comm_utilization_issue_t -----------------------------------------

struct rpt_comm_utilization_issue_t
    : public rpt_issue_t {

    rpt_comm_utilization_issue_t(
        rpt_issue_id_t issue_id)
        : rpt_issue_t(issue_id)
        {}

    rpt_comm_utilization_issue_t(
        rpt_issue_id_t issue_id,
        rpt_loop_id_t loop_id,
        rpt_depgroup_id_t depgroup_id,
        rpt_stat_fp_t utilization,
        rpt_stat_fp_t waste,
        const rpt_location_set_t &producers,
        const rpt_location_set_t &previous_writers,
        const rpt_location_set_t &next_writers)
        : rpt_issue_t(issue_id),
          loop_id(loop_id),
          depgroup_id(depgroup_id),
          utilization(utilization),
          waste(waste),
          producers(producers),
          previous_writers(previous_writers),
          next_writers(next_writers)
        {}

    rpt_issue_type_t get_issue_type() const {
        return Rpt_Issue_Comm_Utilization;
    }

    rpt_depgroup_id_t get_depgroup_id() const {
        return depgroup_id;
    }

    /// Loop id.
    rpt_loop_id_t loop_id;

    /// Dependency group id.
    rpt_depgroup_id_t depgroup_id;

    /// Communication utilization.
    rpt_stat_fp_t utilization;

    /// Communication waste.
    rpt_stat_fp_t waste;

    /// List of data producers
    rpt_location_set_t producers;

    /// Previous and next writers of the data.
    rpt_location_set_t previous_writers;
    rpt_location_set_t next_writers;
};

// ----- rpt_false_sharing_issue_t -----------------------------------------

struct rpt_false_sharing_issue_t
    : public rpt_issue_t {

    rpt_false_sharing_issue_t(
        rpt_issue_id_t issue_id)
        : rpt_issue_t(issue_id)
        {}

    rpt_false_sharing_issue_t(
        rpt_issue_id_t issue_id,
        rpt_loop_id_t loop_id,
        rpt_depgroup_id_t depgroup_id,
        rpt_stat_fp_t ratio,
        rpt_stat_fp_t waste,
        const rpt_location_set_t &offenders)
        : rpt_issue_t(issue_id),
          loop_id(loop_id),
          depgroup_id(depgroup_id),
          ratio(ratio),
          waste(waste),
          offenders(offenders)
        {}

    rpt_issue_type_t get_issue_type() const {
        return Rpt_Issue_False_Sharing;
    }

    rpt_depgroup_id_t get_depgroup_id() const {
        return depgroup_id;
    }

    /// Loop id.
    rpt_loop_id_t loop_id;

    /// Dependency group id.
    rpt_depgroup_id_t depgroup_id;

    /// False sharing ratio
    rpt_stat_fp_t ratio;

    /// False sharing waste
    rpt_stat_fp_t waste;

    /// List of possible false data consumers.
    rpt_location_set_t offenders;
};

// ----- rpt_wnt_possible_issue_t ---------------------------------------------

struct rpt_wnt_possible_issue_t
    : public rpt_issue_t {

    rpt_wnt_possible_issue_t(
        rpt_issue_id_t issue_id)
        : rpt_issue_t(issue_id)
        {}

    rpt_wnt_possible_issue_t(
        rpt_issue_id_t issue_id,
        rpt_loop_id_t loop_id,
        rpt_depgroup_id_t depgroup_id,
        const rpt_location_set_t &next_readers)
        : rpt_issue_t(issue_id),
          loop_id(loop_id),
          depgroup_id(depgroup_id),
          next_readers(next_readers)
        {}

    rpt_issue_type_t get_issue_type() const {
        return Rpt_Issue_Wnt_Possible;
    }

    rpt_depgroup_id_t get_depgroup_id() const {
        return depgroup_id;
    }

    /// Loop id.
    rpt_loop_id_t loop_id;

    /// Dependency group id.
    rpt_depgroup_id_t depgroup_id;

    // Next locations reading from the data.
    rpt_location_set_t next_readers;
};

// ----- rpt_prefetchnta_issue_t ---------------------------------------------

struct rpt_prefetchnta_issue_t
    : public rpt_issue_t {

    rpt_prefetchnta_issue_t(
        rpt_issue_id_t issue_id)
        : rpt_issue_t(issue_id)
        {}

    rpt_prefetchnta_issue_t(
        rpt_issue_id_t issue_id,
        rpt_loop_id_t loop_id,
        rpt_depgroup_id_t depgroup_id,
        rpt_stat_fp_t accesses,
        rpt_stat_fp_t fetches,
        rpt_stat_fp_t required_cache_size,
        const rpt_location_id_vector_t &pc1s,
        const rpt_stat_fp_vector_t &pc1_fractions,
        const rpt_stat_fp_vector_t &pc1_required_cache_sizes,
        const rpt_location_id_vector_t &pc2s,
        const rpt_stat_fp_vector_t &pc2_fractions)
        : rpt_issue_t(issue_id),
          loop_id(loop_id),
          depgroup_id(depgroup_id),
          accesses(accesses),
          fetches(fetches),
          required_cache_size(required_cache_size),
          pc1s(pc1s),
          pc1_fractions(pc1_fractions),
          pc1_required_cache_sizes(pc1_required_cache_sizes),
          pc2s(pc2s),
          pc2_fractions(pc2_fractions)
        {}

    rpt_issue_type_t get_issue_type() const {
        return Rpt_Issue_Prefetchnta;
    }

    rpt_depgroup_id_t get_depgroup_id() const {
        return depgroup_id;
    }

    /// Loop id.
    rpt_loop_id_t loop_id;

    /// Dependency group id.
    rpt_depgroup_id_t depgroup_id;

    ///TODO
    rpt_stat_fp_t accesses;

    ///TODO
    rpt_stat_fp_t fetches;

    ///TODO
    rpt_stat_fp_t required_cache_size;

    ///TODO
    rpt_location_id_vector_t pc1s;

    ///TODO
    rpt_stat_fp_vector_t pc1_fractions;

    ///TODO
    rpt_stat_fp_vector_t pc1_required_cache_sizes;

    ///TODO
    rpt_location_id_vector_t pc2s;

    ///TODO
    rpt_stat_fp_vector_t pc2_fractions;
};

END_NAMESPACE;
END_NAMESPACE;

#endif /* REPORT_ISSUES_HH */
