/// @file analysis/report/report_printer.cc

// Copyright (c) 2011 Acumem AB.  All rights reserved.

#include "analysis/report/report_printer.hh"
#include "library/lang/common_types.h"
#include "library/lang/am_assert.hh"

#include <boost/foreach.hpp>

#include <time.h>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

static const char *
bool_string(bool value)
{
    return value ? "yes" : "no";
}

static void
print_bool(FILE *f, const char *format, bool value)
{
    fprintf(f, format, bool_string(value));
}

static void
print_time(FILE *f, const char *format, rpt_real_time_t time)
{
    struct tm utctime;
    time_t t = time / 1000000000ULL;
    char buf[256];
#if defined(_WIN32)
    _gmtime64_s(&utctime, &t);
    strftime(buf, sizeof(buf), "%x %X", &utctime);
#else
    gmtime_r(&t, &utctime);
    strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S", &utctime);
#endif
    fprintf(f, format, buf);
}

static void
print_data(FILE *f, const char *data, uint32_t data_length)
{
    for (uint32_t i = 0; i < data_length; i++) {
        if (i % 16 == 0) {
            if (i != 0)
                fprintf(f, "\n");
            fprintf(f, "       ");
        }
        fprintf(f, " %02" PRIx8, (uint8_t)data[i]);
    }
    fprintf(f, "\n");
}

static void
print_location_set(FILE *f, const rpt_location_set_t &location_set)
{
    fprintf(f, " (");
    bool first = true;
    BOOST_FOREACH(rpt_location_id_t location_id, location_set) {
        fprintf(f, "%s%" PRIu32, first ? "" : " ", location_id);
        first = false;
    }
    fprintf(f, ")");
}

static void
print_location_id_vector(FILE *f,
                         const rpt_location_id_vector_t &location_id_vector)
{
    fprintf(f, " (");
    bool first = true;
    BOOST_FOREACH(rpt_location_id_t location_id, location_id_vector) {
        fprintf(f, "%s%" PRIu32, first ? "" : " ", location_id);
        first = false;
    }
    fprintf(f, ")");
}

static void
print_location_set_vector(FILE *f,
                          const rpt_location_set_vector_t &location_set_vector)
{
    fprintf(f, " (");
    BOOST_FOREACH(const rpt_location_set_t &location_set, location_set_vector)
        print_location_set(f, location_set);
    fprintf(f, ")");
}

static void
print_stat_fp_vector(FILE *f, const rpt_stat_fp_vector_t &stat_fp_vector)
{
    fprintf(f, " (");
    bool first = true;
    BOOST_FOREACH(rpt_stat_fp_t value, stat_fp_vector) {
        fprintf(f, "%s%f", first ? "" : " ", value);
        first = false;
    }
    fprintf(f, ")");
}

static void
print_uuid(FILE *f, const rpt_report_t &report)
{
    fprintf(f, "UUID: %s\n", report.uuid.c_str());
}

static void
print_sampling_info(FILE *f, const rpt_report_t &report)
{
    print_bool(f, "MT sampled: %s\n", report.mt_sampled);
    print_bool(f, "Burst sampled: %s\n", report.burst_sampled);
    print_time(f, "Sampling start time: %s\n", report.begin_time);
    print_time(f, "Sampling end time: %s\n", report.end_time);
    fprintf(f, "Command line:");
    BOOST_FOREACH(const string &arg, report.argv)
        fprintf(f, " %s", arg.c_str());
    fprintf(f, "\n");
    fprintf(f, "User: %s\n", report.user.c_str());
    switch (report.architecture) {
    case RPT_Arch_i386:
        fprintf(f, "Architecture: i386\n");
        break;
    case RPT_Arch_x86_64:
        fprintf(f, "Architecture: x86_64\n");
        break;
    case RPT_Arch_CPPemu_PPC:
        fprintf(f, "Architecture: CPPemu PPC\n");
        break;
    default:
        AM_ASSERT_MSG("Unsupported architecture.");
    }
    fprintf(f, "Sample count: %" PRIu32 "\n", report.sample_count);
}

static void
print_analysis_info(FILE *f, const rpt_report_t &report)
{
    fprintf(f, "Sample file: %s\n", report.sample_file.c_str());
    print_time(f, "Analysis time: %s\n", report.analysis_time);
    fprintf(f, "Analysis arguments:");
    BOOST_FOREACH(const string &arg, report.analysis_argv)
        fprintf(f, " %s", arg.c_str());
    fprintf(f, "\n");
    fprintf(f, "Title: %s\n", report.title.c_str());
}

static void
print_hardware_config(FILE *f, const rpt_report_t &report)
{
    fprintf(f, "CPU model: %s\n", report.cpu_model.c_str());
    fprintf(f, "Number of CPUs: %" PRIu32 "\n", report.num_cpus);
    fprintf(f, "Number of caches: %" PRIu32 "\n", report.num_caches);
    fprintf(f, "Number of cache levels: %" PRIu32 "\n", report.num_cache_levels);
    fprintf(f, "Cache level: %" PRIu32 "\n", report.cache_level);
    fprintf(f, "Cache size: %" PRIu64 "\n", report.cache_size);
    fprintf(f, "Line size: %" PRIu64 "\n", report.line_size);
    fprintf(f, "Replacement policy: %s\n", report.replacement_policy.c_str());
    print_bool(f, "Any software prefetches active: %s\n", report.swpf_active);
}

static void
print_addresses(FILE *f, const rpt_report_t &report)
{
    uint32_t num_addresses((uint32_t)report.addresses.size());
    fprintf(f, "Addresses, %" PRIu32 " in total:\n", num_addresses);
    for (uint32_t i = 0; i < num_addresses; i++)
        fprintf(f, "    ID %" PRIu32 ", address 0x%" PRIx64 "\n",
                i, report.addresses[i]);
}

static void
print_access_classes(FILE *f, const rpt_report_t &report)
{
    uint32_t num_addresses((uint32_t)report.access_classes.size());
    fprintf(f, "Access classes , for %" PRIu32 " addresses in total:\n", num_addresses);
    for (uint32_t i = 0; i < num_addresses; i++) {
        rpt_access_classes_t access_classes(report.access_classes[i]);
        fprintf(f, "    address ID %" PRIu32 ":", i);
        if (access_classes & RPT_Access_Class_RD)
            fprintf(f, " read");
        if (access_classes & RPT_Access_Class_WR)
            fprintf(f, " write");
        if (access_classes & RPT_Access_Class_PF)
            fprintf(f, " prefetch");
        if (access_classes & RPT_Access_Class_NTA)
            fprintf(f, " non-temporal");
        if (access_classes & RPT_Access_Class_Flush)
            fprintf(f, " flush");
        fprintf(f, "\n");
    }
}

static void
print_trunc_stacks(FILE *f, const rpt_report_t &report)
{
    uint32_t num_trunc_stacks((uint32_t)report.trunc_stacks.size());
    fprintf(f, "Truncated stacks, %" PRIu32 " in total:\n", num_trunc_stacks);
    for (uint32_t i = 0; i < num_trunc_stacks; i++) {
        fprintf(f, "    ID %" PRIu32 ": ", i);
        BOOST_FOREACH(rpt_address_id_t address_id, report.trunc_stacks[i])
            fprintf(f, " %" PRIu32, address_id);
        fprintf(f, "\n");
    }
}

static void
print_full_stacks(FILE *f, const rpt_report_t &report)
{
    uint32_t num_full_stacks((uint32_t)report.full_stacks.size());
    fprintf(f, "Full stacks, %" PRIu32 " in total:\n", num_full_stacks);
    for (uint32_t i = 0; i < num_full_stacks; i++) {
        fprintf(f, "    ID %" PRIu32 ": ", i);
        BOOST_FOREACH(const rpt_full_stack_frame_t &frame, report.full_stacks[i])
            fprintf(f, " (%" PRIu32 ", %f)", frame.pc_id, frame.fraction);
        fprintf(f, "\n");
    }
}

static void
print_locations(FILE *f, const rpt_report_t &report)
{
    uint32_t num_locations((uint32_t)report.locations.size());
    fprintf(f, "Locations, %" PRIu32 " in total:\n", num_locations);
    for (uint32_t i = 0; i < num_locations; i++) {
        const rpt_location_t &l(report.locations[i]);
        fprintf(f, "    ID %" PRIu32 ": %" PRIu32 ", %" PRIu32 "\n",
                i, l.pc_id, l.trunc_stack_id);
    }
}

static void
print_location_sets(FILE *f, const rpt_report_t &report)
{
    uint32_t num_location_sets((uint32_t)report.location_sets.size());
    fprintf(f, "Location sets, %" PRIu32 " in total:\n", num_location_sets);
    for (uint32_t i = 0; i < num_location_sets; i++) {
        fprintf(f, "    ID %" PRIu32 ":", i);
        BOOST_FOREACH(rpt_location_id_t l, report.location_sets[i])
            fprintf(f, " %" PRIu32, l);
        fprintf(f, "\n");
    }
}

static void
print_depgroups(FILE *f, const rpt_report_t &report)
{
    fprintf(f, "Dependency groups, %" PRIu32 " in total:\n",
            (uint32_t)report.depgroups.size());
    BOOST_FOREACH(const rpt_depgroup_map_t::value_type &v, report.depgroups)
        fprintf(f, "    ID %" PRIu32 ": %" PRIu32 "\n",
                v.first, v.second.location_set_id);
}

static void
print_loops(FILE *f, const rpt_report_t &report)
{
    fprintf(f, "Loops, %" PRIu32 " in total:\n",
            (uint32_t)report.loops.size());
    BOOST_FOREACH(const rpt_loop_map_t::value_type &v, report.loops) {
        const rpt_loop_t &loop(v.second);
        fprintf(f, "    ID %" PRIu32 ": %" PRIu32 " %" PRIu32 " (",
                v.first, loop.incl_location_set_id, loop.excl_location_set_id);
        bool first = true;
        BOOST_FOREACH(rpt_depgroup_id_t depgroup_id, loop.depgroups) {
            fprintf(f, "%s%" PRIu32, first ? "" : " ", depgroup_id);
            first = false;
        }
        fprintf(f, ")\n");
    }
}

static void
print_stats(FILE *f, const rpt_report_t &report)
{
    uint32_t num_stats((uint32_t)report.stats.size());
    fprintf(f, "Statistics, %" PRIu32 " in total:\n", num_stats);
    for (uint32_t i = 0; i < num_stats; i++) {
        const rpt_stats_t &stats(report.stats[i]);
        fprintf(f, "ID %" PRIu32 ":", i);
        fprintf(f, " %f", stats.get_accesses());
        fprintf(f, " %f", stats.get_cold_fetches());
        fprintf(f, " %f", stats.get_replacement_fetches());
        fprintf(f, " %f", stats.get_coherence_fetches());
        fprintf(f, " %f", stats.get_flush_fetches());
        fprintf(f, " %f", stats.get_cold_misses());
        fprintf(f, " %f", stats.get_replacement_misses());
        fprintf(f, " %f", stats.get_coherence_misses());
        fprintf(f, " %f", stats.get_flush_misses());
        fprintf(f, " %f", stats.get_upgrades());
        fprintf(f, " %f", stats.get_cold_write_backs());
        fprintf(f, " %f", stats.get_replacement_write_backs());
        fprintf(f, " %f", stats.get_coherence_write_backs());
        fprintf(f, " %f", stats.get_flush_write_backs());
        fprintf(f, " %f", stats.get_spatial_utilization().get_value());
        print_bool(f, " %s", stats.get_spatial_utilization().get_valid());
        fprintf(f, " %f", stats.get_fetch_utilization().get_value());
        print_bool(f, " %s", stats.get_fetch_utilization().get_valid());
        fprintf(f, " %f", stats.get_write_back_utilization().get_value());
        print_bool(f, " %s", stats.get_write_back_utilization().get_valid());
        fprintf(f, " %f", stats.get_communication_utilization().get_value());
        print_bool(f, " %s", stats.get_communication_utilization().get_valid());
        fprintf(f, " %f", stats.get_communicated_lines());
        fprintf(f, " %f", stats.get_hardware_prefetch_ratio());
        fprintf(f, " %f", stats.get_random_ratio());
        fprintf(f, " %f", stats.get_false_sharing_ratio().get_value());
        print_bool(f, " %s", stats.get_false_sharing_ratio().get_valid());
        fprintf(f, " %f", stats.get_util_corr_fetch_ratio());
        fprintf(f, " %f", stats.get_util_corr_write_back_ratio());
        fprintf(f, " %f", stats.get_avoidable_fetches());
        fprintf(f, "\n");
    }
}

static void
print_sliced_stats(FILE *f, const rpt_report_t &report)
{
    uint32_t num_sliced_stats((uint32_t)report.sliced_stats.size());
    fprintf(f, "Sliced statistics, %" PRIu32 " in total:\n", num_sliced_stats);
    for (uint32_t i = 0; i < num_sliced_stats; i++) {
        const rpt_sliced_stats_t ss(report.sliced_stats[i]);
        fprintf(f, "    ID %" PRIu32 ": %" PRIu32 " (",
                i, ss.global_stats);
        bool first = true;
        BOOST_FOREACH(const rpt_thread_stats_map_t::value_type &v,
                      ss.thread_stats) {
            fprintf(f, "%s(%" PRIu32 " %" PRIu32 ")",
                    first ? "" : " ", v.first, v.second);
            first = false;
        }
        fprintf(f, ")\n");
    }
}

static void
print_location_stats(FILE *f, const rpt_report_t &report)
{
    uint32_t num_location_stats((uint32_t)report.location_stats.size());
    fprintf(f, "Location sliced statistics, %" PRIu32 " in total:\n",
            num_location_stats);
    for (uint32_t i = 0; i < num_location_stats; i++)
        fprintf(f, "    location ID %" PRIu32 ": %" PRIu32 "\n",
                i, report.location_stats[i]);
}

static void
print_location_set_stats(FILE *f, const rpt_report_t &report)
{
    uint32_t num_location_set_stats((uint32_t)report.location_set_stats.size());
    fprintf(f, "Location set sliced statistics, %" PRIu32 " in total:\n",
            num_location_set_stats);
    for (uint32_t i = 0; i < num_location_set_stats; i++)
        fprintf(f, "    location set ID %" PRIu32 ": %" PRIu32 "\n",
                i, report.location_set_stats[i]);
}

static void
print_global_stats(FILE *f, const rpt_report_t &report)
{
    fprintf(f, "Global sliced statistics: %" PRIu32 "\n", report.global_stats);
}

static void
print_contexts(FILE *f, const rpt_report_t &report)
{
    uint32_t num_contexts((uint32_t)report.context_ids.size());
    fprintf(f, "Contexts, %" PRIu32 " in total:\n", num_contexts);
    for (uint32_t i = 0; i < num_contexts; i++)
        fprintf(f, "    ID %" PRIu32 ": %" PRIu32 "\n", i, report.context_ids[i]);
}

static void
print_sweeps(FILE *f, const rpt_report_t &report)
{
    uint32_t num_sweeps((uint32_t)report.sweeps.size());
    fprintf(f, "Sweeps, %" PRIu32 " in total:\n", num_sweeps);
    for (uint32_t i = 0; i < num_sweeps; i++) {
        fprintf(f, "    ID %" PRIu32 ":", i);
        BOOST_FOREACH(rpt_size_t cache_size, report.sweeps[i])
            fprintf(f, " %" PRIu64, cache_size);
        fprintf(f, "\n");
    }
}

static void
print_stats_sweeps(FILE *f, const rpt_report_t &report)
{
    uint32_t num_stats_sweeps((uint32_t)report.stats_sweeps.size());
    fprintf(f, "Sliced statistics sweeps, %" PRIu32 " in total:\n", num_stats_sweeps);
    for (uint32_t i = 0; i < num_stats_sweeps; i++) {
        const rpt_stats_sweep_t &stats_sweep(report.stats_sweeps[i]);
        fprintf(f, "    ID %" PRIu32 ": %" PRIu32 " (", i, stats_sweep.sweep_id);
        bool first = true;
        BOOST_FOREACH(rpt_sliced_stats_id_t ssi, stats_sweep.sliced_stats_ids) {
            fprintf(f, "%s%" PRIu32, first ? "" : " ", ssi);
            first = false;
        }
        fprintf(f, ")\n");
    }
}

static void
print_location_set_stats_sweeps(FILE *f, const rpt_report_t &report)
{
    uint32_t num_location_set_stats_sweeps(
        (uint32_t)report.location_set_stats_sweeps.size());
    fprintf(f, "Location set sliced statistics sweeps, %" PRIu32 " in total:\n",
            num_location_set_stats_sweeps);
    for (uint32_t i = 0; i < num_location_set_stats_sweeps; i++)
        fprintf(f, "    location set ID %" PRIu32 ": %" PRIu32 "\n",
                i, report.location_set_stats_sweeps[i]);
}

static void
print_global_stats_sweep(FILE *f, const rpt_report_t &report)
{
    fprintf(f, "Global sliced statistics sweep: %" PRIu32 "\n", report.global_stats_sweep);
}

static void
print_thread_mapping(FILE *f, const rpt_report_t &report)
{
    uint32_t num_cache_threads((uint32_t)report.cache_threads.size());
    fprintf(f, "Thread to cache mapping, for %" PRIu32 " caches:\n",
            num_cache_threads);
    for (uint32_t i = 0; i < num_cache_threads; i++) {
        const rpt_thread_id_set_t &threads(report.cache_threads[i]);
        fprintf(f, "    cache ID %" PRIu32 ":", i);
        BOOST_FOREACH(rpt_thread_id_t thread_id, threads)
            fprintf(f, " %" PRIu32, thread_id);
        fprintf(f, "\n");
    }
}

static void
print_binaries(FILE *f, const rpt_report_t &report)
{
    uint32_t num_binaries((uint32_t)report.binaries.size());
    fprintf(f, "Binaries, %" PRIu32 " in total:\n", num_binaries);
    for (uint32_t i = 0; i < num_binaries; i++) {
        const rpt_binary_t &binary(report.binaries[i]);
        fprintf(f, "    ID %" PRIu32 ": %s %" PRIx64 " %" PRIx64 "%d\n",
                i, binary.filename.c_str(), binary.base, binary.size,
                binary.debug_info_status);
    }
}

static void
print_symbols(FILE *f, const rpt_report_t &report)
{
    uint32_t num_symbols((uint32_t)report.symbols.size());
    fprintf(f, "Symbols, %" PRIu32 " in total:\n", num_symbols);
    for (uint32_t i = 0; i < num_symbols; i++) {
        const rpt_symbol_t &symbol(report.symbols[i]);
        fprintf(f, "    ID %" PRIu32 ": %s %" PRIx64 "\n",
                i, symbol.name.c_str(), symbol.base);
    }
}

static void
print_functions(FILE *f, const rpt_report_t &report)
{
    uint32_t num_functions((uint32_t)report.functions.size());
    fprintf(f, "Functions, %" PRIu32 " in total:\n", num_functions);
    for (uint32_t i = 0; i < num_functions; i++) {
        const rpt_function_t &function(report.functions[i]);
        fprintf(f, "    ID %" PRIu32 ": %s %" PRIx64 "\n",
                i, function.name.c_str(), function.base);
    }
}

static void
print_source_files(FILE *f, const rpt_report_t &report)
{
    uint32_t num_source_files((uint32_t)report.source_files.size());
    fprintf(f, "Source files, %" PRIu32 " in total:\n", num_source_files);
    for (uint32_t i = 0; i < num_source_files; i++) {
        const rpt_source_file_t &source_file(report.source_files[i]);
        fprintf(f, "    ID %" PRIu32 ": %s\n",
                i, source_file.filename.c_str());
        print_data(f, source_file.data.get(), source_file.data_length);
    }
}

static void
print_source_lines(FILE *f, const rpt_report_t &report)
{
    uint32_t num_source_lines((uint32_t)report.source_lines.size());
    fprintf(f, "Source lines, %" PRIu32 " in total:\n", num_source_lines);
    for (uint32_t i = 0; i < num_source_lines; i++) {
        const rpt_source_line_t &source_line(report.source_lines[i]);
        fprintf(f, "    ID %" PRIu32 ": %" PRIu32 " %" PRIu32 "\n",
                i, source_line.source_file_id, source_line.line);
    }
}

static void
print_debug_pcs(FILE *f, const rpt_report_t &report)
{
    uint32_t num_debug_pcs((uint32_t)report.debug_pcs.size());
    fprintf(f, "Debug pcs, %" PRIu32 " in total:\n", num_debug_pcs);
    for (uint32_t i = 0; i < num_debug_pcs; i++) {
        const rpt_debug_pc_t &debug_pc(report.debug_pcs[i]);
        fprintf(f, "    ID %" PRIu32 ": %" PRIu32 " %s %" PRIu32
                " %" PRIu32 " %" PRIu32 " %" PRIu32 "\n",
                i, debug_pc.pc_id, bool_string(debug_pc.return_address),
                debug_pc.binary_id, debug_pc.symbol_id, debug_pc.function_id,
                debug_pc.source_line_id);
    }
}

static void
print_issues(FILE *f, const rpt_report_t &report)
{
    uint32_t num_issues((uint32_t)report.issues.size());
    fprintf(f, "Issues, %" PRIu32 " in total:\n", num_issues);
    for (uint32_t issue_id = 0; issue_id < num_issues; issue_id++) {
        rpt_issue_cptr_t issue(report.issues[issue_id]);

        fprintf(f, "    ID %" PRIu32 ":", issue_id);

        if (issue)
            fprintf(f, " %" PRIu32, issue->get_issue_type());
        else {
            fprintf(f, " no issue\n");
            continue;
        }

        switch(issue->get_issue_type()) {
        case Rpt_Issue_Prefetch: {
            const rpt_prefetch_issue_t &i(
                *(rpt_prefetch_issue_t *)issue.get());
            fprintf(f, " %d %" PRIu32 " %" PRIu32 " %" PRIu32 " %f",
                    i.problem, i.location, i.loop_id, i.depgroup_id, i.cost);
            switch(i.problem) {
            case rpt_prefetch_issue_t::Unnecessary:
            case rpt_prefetch_issue_t::Too_Far:
                fprintf(f, " %f", i.fetch_ratio);
                break;
            case rpt_prefetch_issue_t::Too_Close:
                fprintf(f, " %" PRIu64, i.median_rdist);
                break;
            default:
                AM_ASSERT_MSG("Unsupported prefetch issue type.");
            }
            print_location_set(f, i.prev_locations);
            print_location_set(f, i.next_locations);
            fprintf(f, "\n");
        } break;
        case Rpt_Issue_Hot_Spot: {
            const rpt_hot_spot_issue_t &i(
                *(rpt_hot_spot_issue_t *)issue.get());
            fprintf(f, " %d %" PRIu32 " %" PRIu32,
                    i.type, i.loop_id, i.depgroup_id);
            print_location_set(f, i.previous_writers);
            fprintf(f, "\n");
        } break;
        case Rpt_Issue_Utilization: {
            const rpt_utilization_issue_t &i(
                *(rpt_utilization_issue_t *)issue.get());
            fprintf(f, " %d %" PRIu32 " %" PRIu32,
                    i.type, i.loop_id, i.depgroup_id);
            print_location_set(f, i.previous_writers);
            fprintf(f, "\n");
        } break;
        case Rpt_Issue_Loop_Order: {
            const rpt_loop_order_issue_t &i(
                *(rpt_loop_order_issue_t *)issue.get());
            fprintf(f, " %" PRIu32 " %" PRIu32,
                    i.loop_id, i.depgroup_id);
            print_location_set(f, i.previous_writers);
            fprintf(f, "\n");
        } break;
        case Rpt_Issue_Random: {
            const rpt_random_issue_t &i(
                *(rpt_random_issue_t *)issue.get());
            fprintf(f, " %" PRIu32 " %" PRIu32,
                    i.loop_id, i.depgroup_id);
            print_location_set(f, i.previous_writers);
            fprintf(f, "\n");
        } break;
        case Rpt_Issue_Loop_Fusion: {
            const rpt_loop_fusion_issue_t &i(
                *(rpt_loop_fusion_issue_t *)issue.get());
            fprintf(f, " %" PRIu32 " %" PRIu32,
                    i.loop1_id, i.loop2_id);
            print_location_set(f, i.loop1_pcs);
            print_location_set(f, i.loop2_pcs);
            fprintf(f, " %" PRIu32 " %" PRIu32 " %s %s",
                    i.depgroup1_id, i.depgroup2_id,
                    bool_string(i.move_loop2_up),
                    bool_string(i.move_loop1_down));
            print_location_id_vector(f, i.loop1_down_pcs);
            print_location_set_vector(f, i.loop1_down_blocks);
            print_location_id_vector(f, i.loop2_up_pcs);
            print_location_set_vector(f, i.loop2_up_blocks);
            print_location_set(f, i.additional_pc1);
            print_location_set(f, i.additional_pc2);
            fprintf(f, "\n");
        } break;
        case Rpt_Issue_Block: {
            const rpt_block_issue_t &i(
                *(rpt_block_issue_t *)issue.get());
            fprintf(f, " %" PRIu8 " %" PRIu32 " %" PRIu32
                    " %" PRIu32 " %" PRIu64 " %" PRIu32 " %" PRIu64,
                    i.type_mask, i.loop_id, i.depgroup_id,
                    i.outer_loop_id, i.outer_loop_iterations,
                    i.next_outer_loop_id, i.next_outer_loop_iterations);
            print_location_id_vector(f, i.barrier_pcs);
            print_location_set_vector(f, i.barrier_blocks);
            fprintf(f, " %" PRIu32 " %" PRIu64,
                    i.target_loop_id, i.target_loop_iterations);
            print_location_set(f, i.previous_writers);
            print_location_set(f, i.next_writers);
            fprintf(f, "\n");
        } break;
        case Rpt_Issue_Comm_Spot: {
            const rpt_comm_spot_issue_t &i(
                *(rpt_comm_spot_issue_t *)issue.get());
            fprintf(f, " %" PRIu32 " %" PRIu32 " %f",
                    i.loop_id, i.depgroup_id, i.comm_lines);
            print_location_set(f, i.producers);
            print_location_set(f, i.previous_writers);
            print_location_set(f, i.next_writers);
            fprintf(f, "\n");
        } break;
        case Rpt_Issue_Comm_Utilization: {
            const rpt_comm_utilization_issue_t &i(
                *(rpt_comm_utilization_issue_t *)issue.get());
            fprintf(f, " %" PRIu32 " %" PRIu32 " %f %f",
                    i.loop_id, i.depgroup_id, i.utilization, i.waste);
            print_location_set(f, i.producers);
            print_location_set(f, i.previous_writers);
            print_location_set(f, i.next_writers);
            fprintf(f, "\n");
        } break;
        case Rpt_Issue_False_Sharing: {
            const rpt_false_sharing_issue_t &i(
                *(rpt_false_sharing_issue_t *)issue.get());
            fprintf(f, " %" PRIu32 " %" PRIu32 " %f %f",
                    i.loop_id, i.depgroup_id, i.ratio, i.waste);
            print_location_set(f, i.offenders);
            fprintf(f, "\n");
        } break;
        case Rpt_Issue_Wnt_Possible: {
            const rpt_wnt_possible_issue_t &i(
                *(rpt_wnt_possible_issue_t *)issue.get());
            fprintf(f, " %" PRIu32 " %" PRIu32,
                    i.loop_id, i.depgroup_id);
            print_location_set(f, i.next_readers);
            fprintf(f, "\n");
        } break;
        case Rpt_Issue_Prefetchnta:{
            const rpt_prefetchnta_issue_t &i(
                *(rpt_prefetchnta_issue_t *)issue.get());
            fprintf(f, " %" PRIu32 " %" PRIu32 " %f %f %f",
                    i.loop_id, i.depgroup_id, i.accesses, i.fetches,
                    i.required_cache_size);
            print_location_id_vector(f, i.pc1s);
            print_stat_fp_vector(f, i.pc1_fractions);
            print_stat_fp_vector(f, i.pc1_required_cache_sizes);
            print_location_id_vector(f, i.pc2s);
            print_stat_fp_vector(f, i.pc2_fractions);
            fprintf(f, "\n");
        } break;
        default:
            AM_ASSERT_MSG("Unsupported issue type.");
        }
    }
}

void
print_report(FILE *f, const rpt_report_t &report)
{
    print_uuid(f, report);
    print_sampling_info(f, report);
    print_analysis_info(f, report);
    print_hardware_config(f, report);
    print_addresses(f, report);
    print_access_classes(f, report);
    print_trunc_stacks(f, report);
    print_full_stacks(f, report);
    print_locations(f, report);
    print_location_sets(f, report);
    print_depgroups(f, report);
    print_loops(f, report);
    print_stats(f, report);
    print_sliced_stats(f, report);
    print_location_stats(f, report);
    print_location_set_stats(f, report);
    print_global_stats(f, report);
    print_contexts(f, report);
    print_sweeps(f, report);
    print_stats_sweeps(f, report);
    print_location_set_stats_sweeps(f, report);
    print_global_stats_sweep(f, report);
    print_thread_mapping(f, report);
    print_binaries(f, report);
    print_symbols(f, report);
    print_functions(f, report);
    print_source_files(f, report);
    print_source_lines(f, report);
    print_debug_pcs(f, report);
    print_issues(f, report);
}

END_NAMESPACE;
END_NAMESPACE;
