/// @file analysis/report/report_stats.hh

// Copyright (c) 2010 Acumem AB.  All rights reserved.

#ifndef REPORT_REPORT_STATS_HH
#define REPORT_REPORT_STATS_HH

#include "analysis/report/report_types.hh"
#include "library/lang/compat.h"

#include <math.h>
#include <assert.h>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

/// Floating point type to use for the statistics.
typedef float rpt_stat_fp_t;

typedef vector<rpt_stat_fp_t> rpt_stat_fp_vector_t;

// ----- rpt_stat_fp_valid_t --------------------------------------------------

/// Statistics value and validity info.
struct rpt_stat_fp_valid_t {
    friend class report_encoder_t;
    friend class report_decoder_t;

    rpt_stat_fp_valid_t()
        {}

    rpt_stat_fp_valid_t(
        double value,
        bool valid)
        : value((rpt_stat_fp_t)value),
          valid(valid)
        {}

    bool operator==(const rpt_stat_fp_valid_t &that) const;
    bool operator<(const rpt_stat_fp_valid_t &that) const;
    bool operator>(const rpt_stat_fp_valid_t &that) const;

    double get_value() const {
        return value;
    }

    bool get_valid() const {
        return valid;
    }

private:
    rpt_stat_fp_t value;
    bool valid;
};

// ----- rpt_stats_t ----------------------------------------------------------

/// A statistics record.
///
/// Note that even though this type is used both for individual locations and
/// location sets, the contents differ depending in which it is. For individual
/// locations the miss statistics include prefetch misses, while for location
/// sets prefetches are considered to never miss. This mirrors the stats_t type
/// used in the analysis.
class rpt_stats_t {
    friend class report_encoder_t;
    friend class report_decoder_t;

public:
    rpt_stats_t()
        {}

    rpt_stats_t(double accesses,
                double cold_fetches,
                double replacement_fetches,
                double coherence_fetches,
                double flush_fetches,
                double cold_misses,
                double replacement_misses,
                double coherence_misses,
                double flush_misses,
                double upgrades,
                double cold_write_backs,
                double replacement_write_backs,
                double coherence_write_backs,
                double flush_write_backs,
                rpt_stat_fp_valid_t spatial_utilization,
                rpt_stat_fp_valid_t fetch_utilization,
                rpt_stat_fp_valid_t write_back_utilization,
                rpt_stat_fp_valid_t communication_utilization,
                double communicated_lines,
                double hardware_prefetch_ratio,
                double random_ratio,
                rpt_stat_fp_valid_t false_sharing_ratio,
                double util_corr_fetch_ratio,
                double util_corr_write_back_ratio,
                double avoidable_fetches)
        : accesses((rpt_stat_fp_t)accesses),
          cold_fetches((rpt_stat_fp_t)cold_fetches),
          replacement_fetches((rpt_stat_fp_t)replacement_fetches),
          coherence_fetches((rpt_stat_fp_t)coherence_fetches),
	  flush_fetches((rpt_stat_fp_t)flush_fetches),
          cold_misses((rpt_stat_fp_t)cold_misses),
          replacement_misses((rpt_stat_fp_t)replacement_misses),
          coherence_misses((rpt_stat_fp_t)coherence_misses),
	  flush_misses((rpt_stat_fp_t)flush_misses),
          upgrades((rpt_stat_fp_t)upgrades),
          cold_write_backs((rpt_stat_fp_t)cold_write_backs),
          replacement_write_backs((rpt_stat_fp_t)replacement_write_backs),
          coherence_write_backs((rpt_stat_fp_t)coherence_write_backs),
	  flush_write_backs((rpt_stat_fp_t)flush_write_backs),
          spatial_utilization(spatial_utilization),
          fetch_utilization(fetch_utilization),
          write_back_utilization(write_back_utilization),
          communication_utilization(communication_utilization),
          communicated_lines((rpt_stat_fp_t)communicated_lines),
          hardware_prefetch_ratio((rpt_stat_fp_t)hardware_prefetch_ratio),
          random_ratio((rpt_stat_fp_t)random_ratio),
          false_sharing_ratio(false_sharing_ratio),
          util_corr_fetch_ratio((rpt_stat_fp_t)util_corr_fetch_ratio),
          util_corr_write_back_ratio((rpt_stat_fp_t)util_corr_write_back_ratio),
          avoidable_fetches((rpt_stat_fp_t)avoidable_fetches)
        {
            assert(isfinite(accesses));
            assert(isfinite(cold_fetches));
            assert(isfinite(replacement_fetches));
            assert(isfinite(coherence_fetches));
            assert(isfinite(flush_fetches));
            assert(isfinite(cold_misses));
            assert(isfinite(replacement_misses));
            assert(isfinite(coherence_misses));
            assert(isfinite(flush_misses));
            assert(isfinite(upgrades));
            assert(isfinite(cold_write_backs));
            assert(isfinite(replacement_write_backs));
            assert(isfinite(coherence_write_backs));
            assert(isfinite(flush_write_backs));
            assert(isfinite(spatial_utilization.get_value()));
            assert(isfinite(fetch_utilization.get_value()));
            assert(isfinite(write_back_utilization.get_value()));
            assert(isfinite(communication_utilization.get_value()));
            assert(isfinite(communicated_lines));
            assert(isfinite(hardware_prefetch_ratio));
            assert(isfinite(random_ratio));
            assert(isfinite(false_sharing_ratio.get_value()));
            assert(isfinite(util_corr_fetch_ratio));
            assert(isfinite(util_corr_write_back_ratio));
            assert(isfinite(avoidable_fetches));
        }

    double get_accesses() const {
        return accesses;
    }

    double get_cold_fetches() const {
        return cold_fetches;
    }

    double get_replacement_fetches() const {
        return replacement_fetches;
    }

    double get_coherence_fetches() const {
        return coherence_fetches;
    }

    double get_flush_fetches() const {
        return flush_fetches;
    }

    double get_total_fetches() const {
        return
            cold_fetches +
            replacement_fetches +
            coherence_fetches +
            flush_fetches;
    }

    double get_cold_misses() const {
        return cold_misses;
    }

    double get_replacement_misses() const {
        return replacement_misses;
    }

    double get_coherence_misses() const {
        return coherence_misses;
    }

    double get_flush_misses() const {
        return flush_misses;
    }

    double get_total_misses() const {
        return
            cold_misses +
            replacement_misses +
            coherence_misses +
            flush_misses;
    }

    double get_upgrades() const {
        return upgrades;
    }

    double get_total_upgrades() const {
        return upgrades;
    }

    double get_cold_write_backs() const {
        return cold_write_backs;
    }

    double get_replacement_write_backs() const {
        return replacement_write_backs;
    }

    double get_coherence_write_backs() const {
        return coherence_write_backs;
    }

    double get_flush_write_backs() const {
        return flush_write_backs;
    }

    double get_total_write_backs() const {
        return
            cold_write_backs +
            replacement_write_backs +
            coherence_write_backs +
            flush_write_backs;
    }

    double get_cold_fetch_ratio() const {
        return accesses == 0.0 ? 0.0 : cold_fetches / accesses;
    }

    double get_replacement_fetch_ratio() const {
        return accesses == 0.0 ? 0.0 : get_replacement_fetches() / accesses;
    }

    double get_coherence_fetch_ratio() const {
        return accesses == 0.0 ? 0.0 : get_coherence_fetches() / accesses;
    }

    double get_flush_fetch_ratio() const {
        return accesses == 0.0 ? 0.0 : get_flush_fetches() / accesses;
    }

    double get_total_fetch_ratio() const {
        return accesses == 0.0 ? 0.0 : get_total_fetches() / accesses;
    }

    double get_cold_miss_ratio() const {
        return accesses == 0.0 ? 0.0 : get_cold_misses() / accesses;
    }

    double get_replacement_miss_ratio() const {
        return accesses == 0.0 ? 0.0 : get_replacement_misses() / accesses;
    }

    double get_coherence_miss_ratio() const {
        return accesses == 0.0 ? 0.0 : get_coherence_misses() / accesses;
    }

    double get_flush_miss_ratio() const {
        return accesses == 0.0 ? 0.0 : get_flush_misses() / accesses;
    }

    double get_total_miss_ratio() const {
        return accesses == 0.0 ? 0.0 : get_total_misses() / accesses;
    }

    double get_upgrade_ratio() const {
        return accesses == 0.0 ? 0.0 : get_upgrades() / accesses;
    }

    double get_total_upgrade_ratio() const {
        return accesses == 0.0 ? 0.0 : get_upgrades() / accesses;
    }

    double get_cold_write_back_ratio() const {
        return accesses == 0.0 ? 0.0 : get_cold_write_backs() / accesses;
    }

    double get_replacement_write_back_ratio() const {
        return accesses == 0.0 ? 0.0 : get_replacement_write_backs() / accesses;
    }

    double get_coherence_write_back_ratio() const {
        return accesses == 0.0 ? 0.0 : get_coherence_write_backs() / accesses;
    }

    double get_flush_write_back_ratio() const {
        return accesses == 0.0 ? 0.0 : get_flush_write_backs() / accesses;
    }

    double get_total_write_back_ratio() const {
        return accesses == 0.0 ? 0.0 : get_total_write_backs() / accesses;
    }

    rpt_stat_fp_valid_t get_spatial_utilization() const {
        return spatial_utilization;
    }

    rpt_stat_fp_valid_t get_fetch_utilization() const {
        return fetch_utilization;
    }

    rpt_stat_fp_valid_t get_write_back_utilization() const {
        return write_back_utilization;
    }

    rpt_stat_fp_valid_t get_communication_utilization() const {
        return communication_utilization;
    }

    double get_communicated_lines() const {
        return communicated_lines;
    }

    double get_hardware_prefetch_ratio() const {
        return hardware_prefetch_ratio;
    }

    double get_random_ratio() const {
        return random_ratio;
    }

    rpt_stat_fp_valid_t get_false_sharing_ratio() const {
        return false_sharing_ratio;
    }

    double get_util_corr_fetch_ratio() const {
        return util_corr_fetch_ratio;
    }

    double get_util_corr_write_back_ratio() const {
        return util_corr_write_back_ratio;
    }

    double get_avoidable_fetches() const {
        return avoidable_fetches;
    }

    bool operator==(const rpt_stats_t &that) const;

    bool operator<(const rpt_stats_t &that) const;

private:
    /// Number of accesses.
    rpt_stat_fp_t accesses;

    /// Fetches due to cold accesses.
    rpt_stat_fp_t cold_fetches;

    /// Fetches due to replacements.
    rpt_stat_fp_t replacement_fetches;

    /// Fetches due to coherence.
    rpt_stat_fp_t coherence_fetches;

    /// Fetches due to cache line flushes.
    rpt_stat_fp_t flush_fetches;

    /// Misses due to cold accesses.
    rpt_stat_fp_t cold_misses;

    /// Misses due to replacements.
    rpt_stat_fp_t replacement_misses;

    /// Misses due to coherence.
    rpt_stat_fp_t coherence_misses;

    /// Misses due to cache line flushes.
    rpt_stat_fp_t flush_misses;

    /// Upgrades.
    rpt_stat_fp_t upgrades;

    /// Write-backs due to cold misses.
    rpt_stat_fp_t cold_write_backs;

    /// Write-backs due to replacements.
    rpt_stat_fp_t replacement_write_backs;

    /// Write-backs due to coherence.
    rpt_stat_fp_t coherence_write_backs;

    /// Write-backs due to cache line flushes.
    rpt_stat_fp_t flush_write_backs;

    /// Spatial utilization.
    rpt_stat_fp_valid_t spatial_utilization;

    /// Fetch utilization.
    rpt_stat_fp_valid_t fetch_utilization;

    /// Write-back utilization.
    rpt_stat_fp_valid_t write_back_utilization;

    /// Communication utilization.
    rpt_stat_fp_valid_t communication_utilization;

    /// Communicated lines.
    rpt_stat_fp_t communicated_lines;

    /// Hardware prefetch ratio.
    rpt_stat_fp_t hardware_prefetch_ratio;

    /// Random ratio.
    rpt_stat_fp_t random_ratio;

    /// False sharing ratio.
    rpt_stat_fp_valid_t false_sharing_ratio;

    /// Utilization corrected fetch ratio.
    rpt_stat_fp_t util_corr_fetch_ratio;

    /// Utilization corrected write-back ratio.
    rpt_stat_fp_t util_corr_write_back_ratio;

    /// Avoidable fetches.
    rpt_stat_fp_t avoidable_fetches;
};

/// A set of statistics records.
typedef set<rpt_stats_t> rpt_stats_set_t;

/// A vector of statistics records, indexed by statistics ID.
typedef vector<rpt_stats_t> rpt_stats_vector_t;

/// A statistics ID.
typedef uint32_t rpt_stats_id_t;

/// Maps thread IDs to statistics IDs.
typedef map<rpt_thread_id_t, rpt_stats_id_t> rpt_thread_stats_map_t;

// ----- rpt_sliced_stats_t ---------------------------------------------------

/// Sliced statistics.
struct rpt_sliced_stats_t {
    rpt_sliced_stats_t()
        {}

    /// Global statistics.
    rpt_stats_id_t global_stats;

    /// Per-thread statistics indexed by thread ID.
    rpt_thread_stats_map_t thread_stats;

    bool operator==(const rpt_sliced_stats_t &that) const;

    bool operator<(const rpt_sliced_stats_t &that) const;
};

/// Set of sliced statistics.
typedef set<rpt_sliced_stats_t> rpt_sliced_stats_set_t;

/// Vector of sliced statistics, indexed by sliced statistics ID.
typedef vector<rpt_sliced_stats_t> rpt_sliced_stats_vector_t;

/// A sliced statistics ID.
typedef uint32_t rpt_sliced_stats_id_t;

/// A vector of sliced statistics IDs.
typedef vector<rpt_sliced_stats_id_t> rpt_sliced_stats_id_vector_t;

// ----- rpt_stats_sweep_t ----------------------------------------------------a

/// A sweep. Vector of cache sizes in a the sweep.
typedef vector<rpt_size_t> rpt_sweep_t;

/// Vector of sweeps.
typedef vector<rpt_sweep_t> rpt_sweep_vector_t;

/// A sweep ID.
typedef uint32_t rpt_sweep_id_t;

/// A statistics sweep.
struct rpt_stats_sweep_t {
    rpt_sweep_id_t sweep_id;
    rpt_sliced_stats_id_vector_t sliced_stats_ids;

    bool operator==(const rpt_stats_sweep_t &that) const;

    bool operator<(const rpt_stats_sweep_t &that) const;
};

/// Vector of statistics sweeps.
typedef vector<rpt_stats_sweep_t> rpt_stats_sweep_vector_t;

/// Set of statistics sweeps.
typedef set<rpt_stats_sweep_t> rpt_stats_sweep_set_t;

/// A statistics sweep ID.
typedef uint32_t rpt_stats_sweep_id_t;

/// Vector of statistics sweep IDs.
typedef vector<rpt_stats_sweep_id_t> rpt_stats_sweep_id_vector_t;

END_NAMESPACE;
END_NAMESPACE;

#endif /* REPORT_STATS_HH */
