/// @file analysis/report/report_types.hh

// Copyright (c) 2010 Acumem AB.  All rights reserved.

#ifndef REPORT_REPORT_TYPES_HH
#define REPORT_REPORT_TYPES_HH

#include "library/lang/common_macros.h"
#include "library/lang/common_types.h"

#include <vector>
#include <set>
#include <map>
#include <string>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;

/// Size in bytes.
typedef uint64_t rpt_size_t;

/// Sample count.
typedef uint32_t rpt_sample_count_t;

/// Access count.
typedef uint64_t rpt_access_count_t;

/// Real-time time. Number of nano-seconds since 1970-01-01 00:00 UTC.
typedef uint64_t rpt_real_time_t;

/// Sampled architecture.
typedef enum {
    RPT_Arch_i386,
    RPT_Arch_x86_64,
    RPT_Arch_CPPemu_PPC
} rpt_architecture_t;

/// An address.
typedef uint64_t rpt_address_t;

/// An address ID.
typedef uint32_t rpt_address_id_t;

/// A vector of addresses, indexed by address ID.
typedef vector<rpt_address_t> rpt_address_vector_t;

/// A set of addresses.
typedef set<rpt_address_t> rpt_address_set_t;

/// A vector of address IDs.
typedef vector<rpt_address_id_t> rpt_address_id_vector_t;

/// A set of address IDs.
typedef set<rpt_address_id_t> rpt_address_id_set_t;

/// Truncated call stack. Outermost frame first.
typedef rpt_address_id_vector_t rpt_trunc_stack_t;

/// Vector of truncated call stacks.
typedef vector<rpt_trunc_stack_t> rpt_trunc_stacks_t;

/// Truncated stack ID.
typedef uint32_t rpt_trunc_stack_id_t;

/// A full call stack frame. PC and the fraction of accesses at the
/// corresponding location that were taken when called through this stack
/// frame.
struct rpt_full_stack_frame_t {
    rpt_full_stack_frame_t()
        {}

    rpt_full_stack_frame_t(
        rpt_address_id_t pc_id,
        float fraction)
        : pc_id(pc_id),
          fraction(fraction)
        {}

    rpt_address_id_t pc_id;
    float fraction;
};

/// Full call stack. Excludes the frames in the truncated stack of the
/// corresponding location. Outermost frame first.
typedef vector<rpt_full_stack_frame_t> rpt_full_stack_t;

/// Vector of full call stacks, indexed by the corresponding location ID.
typedef vector<rpt_full_stack_t> rpt_full_stacks_t;

/// Info about a location.
struct rpt_location_t {
    rpt_location_t()
        {}

    rpt_location_t(
        rpt_address_id_t pc_id,
        rpt_trunc_stack_id_t trunc_stack_id)
        : pc_id(pc_id),
          trunc_stack_id(trunc_stack_id)
        {}

    rpt_address_id_t pc_id;
    rpt_trunc_stack_id_t trunc_stack_id;
};

/// Vector of location info records, indexed by location ID.
typedef vector<rpt_location_t> rpt_location_vector_t;

/// A location ID. We use the same location IDs as the analysis.
typedef uint32_t rpt_location_id_t;

/// A location (ID) set.
typedef set<rpt_location_id_t> rpt_location_set_t;

/// A location (ID) vector.
typedef vector<rpt_location_id_t> rpt_location_id_vector_t;

/// A location set ID.
typedef uint32_t rpt_location_set_id_t;

/// A vector of location (ID) sets.
typedef vector<rpt_location_set_t> rpt_location_set_vector_t;

/// A dependence group.
struct rpt_depgroup_t {
    rpt_depgroup_t()
        {}

    rpt_depgroup_t(rpt_location_set_id_t location_set_id)
        : location_set_id(location_set_id)
        {}

    rpt_location_set_id_t location_set_id;
};

/// A dependence group ID. We use the same dependence group IDs as the
/// analysis.
typedef uint32_t rpt_depgroup_id_t;

/// A dependence group (ID) set.
typedef set<rpt_depgroup_id_t> rpt_depgroup_set_t;

/// A dependence group (ID) vecotr.
typedef vector<rpt_depgroup_id_t> rpt_depgroup_vector_t;

/// A map from dependence group IDs to dependence groups.
typedef map<rpt_depgroup_id_t, rpt_depgroup_t> rpt_depgroup_map_t;

/// A loop.
struct rpt_loop_t {
    rpt_loop_t()
        {}

    rpt_loop_t(rpt_location_set_id_t incl_location_set_id,
               rpt_location_set_id_t excl_location_set_id,
               rpt_depgroup_set_t depgroups)
        : incl_location_set_id(incl_location_set_id),
          excl_location_set_id(excl_location_set_id),
          depgroups(depgroups)
        {}

    /// Locations in the loop, including locations in inner loops.
    rpt_location_set_id_t incl_location_set_id;

    /// Locations in the loop, excluding locations in inner loops.
    rpt_location_set_id_t excl_location_set_id;

    /// Dependence groups in the loop. TODO: Which are included?
    rpt_depgroup_set_t depgroups;
};

/// A loop ID. We use the same loop IDs as the analysis.
typedef uint32_t rpt_loop_id_t;

/// A map from loop IDs to loops.
typedef map<rpt_loop_id_t, rpt_loop_t> rpt_loop_map_t;

/// Access classes. These currently just mirror the Fenris access classes.
typedef uint8_t rpt_access_classes_t;
#define RPT_Access_Class_None  0x00
#define RPT_Access_Class_RD    0x01
#define RPT_Access_Class_WR    0x02
#define RPT_Access_Class_PF    0x04
#define RPT_Access_Class_NTA   0x08
#define RPT_Access_Class_Flush 0x10
#define RPT_Access_Class_Any   0xFF
#define RPT_Access_Classes_Pseudo \
    (RPT_Access_Class_PF | RPT_Access_Class_Flush)

/// Vector of access classes.
typedef vector<rpt_access_classes_t> rpt_access_classes_vector_t;

/// Our internal thread IDs, consecutive from 0 and up.
typedef uint32_t rpt_thread_id_t;

/// Vector of internal thread IDs.
typedef vector<rpt_thread_id_t> rpt_thread_id_vector_t;

/// Set of internal thread IDs.
typedef set<rpt_thread_id_t> rpt_thread_id_set_t;

/// Vector of sets of internal thread IDs.
typedef vector<rpt_thread_id_set_t> rpt_thread_id_set_vector_t;

/// Operating system context IDs.
typedef uint32_t rpt_context_id_t;

/// Vector of operating system context IDs, indexed by thread ID.
typedef vector<rpt_context_id_t> rpt_context_id_vector_t;

/// Cache ID.
typedef uint32_t rpt_cache_id_t;

END_NAMESPACE;
END_NAMESPACE;

#endif /* REPORT_TYPES_HH */
