/// @file analysis/report_codec/report_codec.cc

// Copyright (c) 2010 Acumem AB.  All rights reserved.

#include "analysis/report_codec/report_codec.hh"
#include "library/codec/codec_utils.hh"
#include "library/stream/stream_container.hh"

#include <boost/foreach.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

// ----- Version constants. ---------------------------------------------------

/// Current report_codec version.
#define REPORT_CODEC_VERSION 0

/// The minimum report codec version required to decode the current version,
/// possibly omitting some newer parts of the report.
#define REPORT_CODEC_VERSION_ENCODE_MINIMUM 0

/// The minimum report_codec version that the current version can decode.
#define REPORT_CODEC_VERSION_DECODE_MINIMUM 0

// ----- Header sizes. --------------------------------------------------------

/// File header size in bytes.
#define FILE_HEADER_SIZE 4

/// Packet header size in bytes.
#define PACKET_HEADER_SIZE 6

// ----- Packet types. --------------------------------------------------------

#define PACKET_SAMPLING_INFO              0
#define PACKET_ANALYSIS_INFO              1
#define PACKET_HARDWARE_CONFIG            2
#define PACKET_ADDRESSES                  3
#define PACKET_ACCESS_CLASSES             4
#define PACKET_TRUNC_STACKS               5
#define PACKET_FULL_STACKS                6
#define PACKET_LOCATIONS                  7
#define PACKET_LOCATION_SETS              8
#define PACKET_DEPGROUPS                  9
#define PACKET_LOOPS                     10
#define PACKET_STATS                     11
#define PACKET_SLICED_STATS              12
#define PACKET_LOCATION_STATS            13
#define PACKET_LOCATION_SET_STATS        14
#define PACKET_GLOBAL_STATS              15
#define PACKET_CONTEXTS                  16
#define PACKET_SWEEPS                    17
#define PACKET_STATS_SWEEPS              18
#define PACKET_LOCATION_SET_STATS_SWEEPS 19
#define PACKET_GLOBAL_STATS_SWEEP        20
#define PACKET_THREAD_MAPPING            21
#define PACKET_BINARIES                  22
#define PACKET_SYMBOLS                   23
#define PACKET_FUNCTIONS                 24
#define PACKET_SOURCE_FILES              25
#define PACKET_SOURCE_LINES              26
#define PACKET_DEBUG_PCS                 27
#define PACKET_ISSUE_COUNT               28
#define PACKET_PREFETCH_ISSUE            29
#define PACKET_HOT_SPOT_ISSUE            30
#define PACKET_UTILIZATION_ISSUE         31
#define PACKET_LOOP_ORDER_ISSUE          32
#define PACKET_RANDOM_ISSUE              33
#define PACKET_LOOP_FUSION_ISSUE         34
#define PACKET_BLOCK_ISSUE               35
#define PACKET_COMM_SPOT_ISSUE           36
#define PACKET_COMM_UTILIZATION_ISSUE    37
#define PACKET_FALSE_SHARING_ISSUE       38
#define PACKET_WNT_POSSIBLE_ISSUE        39
#define PACKET_PREFETCHNTA_ISSUE         40
#define PACKET_UUID                      41
#define PACKET_HARDWARE_CONFIG2          42

// ----- report_encoder_t -----------------------------------------------------

report_encoder_t::report_encoder_t(
    stream_t &stream)
    : stream(stream)
{
    codec_buffer_t buffer;
    binary_encode_uint16(&buffer, REPORT_CODEC_VERSION);
    binary_encode_uint16(&buffer, REPORT_CODEC_VERSION_ENCODE_MINIMUM);
    write_buffer(buffer);
}

void
report_encoder_t::encode(rpt_report_cptr_t report)
{
    encode_uuid(report);
    encode_sampling_info(report);
    encode_analysis_info(report);
    encode_hardware_config(report);
    encode_hardware_config2(report);
    encode_addresses(report);
    encode_access_classes(report);
    encode_trunc_stacks(report);
    encode_full_stacks(report);
    encode_locations(report);
    encode_location_sets(report);
    encode_depgroups(report);
    encode_loops(report);
    encode_stats(report);
    encode_sliced_stats(report);
    encode_location_stats(report);
    encode_location_set_stats(report);
    encode_global_stats(report);
    encode_contexts(report);
    encode_sweeps(report);
    encode_stats_sweeps(report);
    encode_location_set_stats_sweeps(report);
    encode_global_stats_sweep(report);
    encode_thread_mapping(report);
    encode_binaries(report);
    encode_symbols(report);
    encode_functions(report);
    encode_source_files(report);
    encode_source_lines(report);
    encode_debug_pcs(report);
    encode_issue_count(report);
    BOOST_FOREACH(rpt_issue_ptr_t issue, report->issues)
        switch(issue->get_issue_type()) {
        case Rpt_Issue_Prefetch:
            encode_prefetch_issue(report, issue);
            break;
        case Rpt_Issue_Hot_Spot:
            encode_hot_spot_issue(report, issue);
            break;
        case Rpt_Issue_Utilization:
            encode_utilization_issue(report, issue);
            break;
        case Rpt_Issue_Loop_Order:
            encode_loop_order_issue(report, issue);
            break;
        case Rpt_Issue_Random:
            encode_random_issue(report, issue);
            break;
        case Rpt_Issue_Loop_Fusion:
            encode_loop_fusion_issue(report, issue);
            break;
        case Rpt_Issue_Block:
            encode_block_issue(report, issue);
            break;
        case Rpt_Issue_Comm_Spot:
            encode_comm_spot_issue(report, issue);
            break;
        case Rpt_Issue_Comm_Utilization:
            encode_comm_utilization_issue(report, issue);
            break;
        case Rpt_Issue_False_Sharing:
            encode_false_sharing_issue(report, issue);
            break;
        case Rpt_Issue_Wnt_Possible:
            encode_wnt_possible_issue(report, issue);
            break;
        case Rpt_Issue_Prefetchnta:
            encode_prefetchnta_issue(report, issue);
            break;
        default:
            assert(0);
            while (1);
        }
    stream.flush();
}

void
report_encoder_t::write_buffer(codec_buffer_t &buffer)
{
    size_t size = buffer.size();
    stream.process_data(buffer.read(size), size);
}

void
report_encoder_t::write_packet(uint16_t type)
{
    binary_encode_uint16(&header_buffer,type);
    binary_encode_uint32(&header_buffer, (uint32_t) packet_buffer.size());
    header_buffer.assert_read(PACKET_HEADER_SIZE);
    write_buffer(header_buffer);
    packet_buffer.assert_read(packet_buffer.size());
    write_buffer(packet_buffer);
}

void
report_encoder_t::encode_location_id_vector(
    const rpt_location_id_vector_t &location_id_vector)
{
    binary_encode_uint32(&packet_buffer, (uint32_t) location_id_vector.size());
    BOOST_FOREACH(rpt_location_id_t location, location_id_vector)
        binary_encode_uint32(&packet_buffer, location);
}

void
report_encoder_t::encode_location_set(
    const rpt_location_set_t &location_set)
{
    binary_encode_uint32(&packet_buffer, (uint32_t) location_set.size());
    BOOST_FOREACH(rpt_location_id_t location, location_set)
        binary_encode_uint32(&packet_buffer, location);
}

void
report_encoder_t::encode_location_set_vector(
    const rpt_location_set_vector_t &location_set_vector)
{
    binary_encode_uint32(&packet_buffer, (uint32_t) location_set_vector.size());
    BOOST_FOREACH(const rpt_location_set_t &location_set, location_set_vector)
        encode_location_set(location_set);
}

void
report_encoder_t::encode_stat_fp_vector(
    const rpt_stat_fp_vector_t &stat_fp_vector)
{
    binary_encode_uint32(&packet_buffer, (uint32_t) stat_fp_vector.size());
    BOOST_FOREACH(rpt_stat_fp_t v, stat_fp_vector)
        binary_encode_float(&packet_buffer, v);
}

void
report_encoder_t::encode_uuid(rpt_report_cptr_t report)
{
    binary_encode_string32(&packet_buffer, report->uuid);
    write_packet(PACKET_UUID);
}

void
report_encoder_t::encode_sampling_info(rpt_report_cptr_t report)
{
    binary_encode_bool(&packet_buffer, report->mt_sampled);
    binary_encode_bool(&packet_buffer, report->burst_sampled);
    binary_encode_uint64(&packet_buffer, report->begin_time);
    binary_encode_uint64(&packet_buffer, report->end_time);
    binary_encode_string32_vector(&packet_buffer, report->argv);
    binary_encode_string32(&packet_buffer, report->user.c_str());
    binary_encode_uint32(&packet_buffer, report->architecture);
    binary_encode_uint32(&packet_buffer, report->sample_count);
    write_packet(PACKET_SAMPLING_INFO);
}

void
report_encoder_t::encode_analysis_info(rpt_report_cptr_t report)
{
    binary_encode_string32(&packet_buffer, report->sample_file.c_str());
    binary_encode_uint64(&packet_buffer, report->analysis_time);
    binary_encode_string32_vector(&packet_buffer, report->analysis_argv);
    binary_encode_string32(&packet_buffer, report->title);
    write_packet(PACKET_ANALYSIS_INFO);
}

void
report_encoder_t::encode_hardware_config(rpt_report_cptr_t report)
{
    binary_encode_string32(&packet_buffer, report->cpu_model);
    binary_encode_uint32(&packet_buffer, report->num_cpus);
    binary_encode_uint32(&packet_buffer, report->num_caches);
    binary_encode_uint32(&packet_buffer, report->num_cache_levels);
    binary_encode_uint32(&packet_buffer, report->cache_level);
    binary_encode_uint64(&packet_buffer, report->cache_size);
    binary_encode_uint64(&packet_buffer, report->line_size);
    binary_encode_string32(&packet_buffer, report->replacement_policy);
    binary_encode_bool(&packet_buffer, report->swpf_active);
    write_packet(PACKET_HARDWARE_CONFIG);
}

void
report_encoder_t::encode_hardware_config2(rpt_report_cptr_t report)
{
    binary_encode_bool(&packet_buffer, report->autodetect_failed);
    write_packet(PACKET_HARDWARE_CONFIG2);
}

void
report_encoder_t::encode_addresses(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->addresses.size());
    BOOST_FOREACH(rpt_address_t addr, report->addresses)
        binary_encode_uint64(&packet_buffer, addr);
    write_packet(PACKET_ADDRESSES);
}

void
report_encoder_t::encode_access_classes(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->access_classes.size());
    BOOST_FOREACH(uint8_t access_classes, report->access_classes)
        binary_encode_uint8(&packet_buffer, access_classes);
    write_packet(PACKET_ACCESS_CLASSES);
}

void
report_encoder_t::encode_trunc_stacks(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->trunc_stacks.size());
    BOOST_FOREACH(const rpt_trunc_stack_t &trunc_stack,
                  report->trunc_stacks) {
        binary_encode_uint16(&packet_buffer, (uint16_t)trunc_stack.size());
        BOOST_FOREACH(rpt_address_id_t address_id, trunc_stack)
            binary_encode_uint32(&packet_buffer, address_id);
    }
    write_packet(PACKET_TRUNC_STACKS);
}

void
report_encoder_t::encode_full_stacks(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->full_stacks.size());
    BOOST_FOREACH(const rpt_full_stack_t &full_stack, report->full_stacks) {
        binary_encode_uint16(&packet_buffer, (uint16_t)full_stack.size());
        BOOST_FOREACH(const rpt_full_stack_frame_t &full_stack_frame, full_stack) {
            binary_encode_uint32(&packet_buffer, full_stack_frame.pc_id);
            binary_encode_float(&packet_buffer, full_stack_frame.fraction);
        }
    }
    write_packet(PACKET_FULL_STACKS);
}

void
report_encoder_t::encode_locations(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->locations.size());
    BOOST_FOREACH(const rpt_location_t& location, report->locations) {
        binary_encode_uint32(&packet_buffer, location.pc_id);
        binary_encode_uint32(&packet_buffer, location.trunc_stack_id);
    }
    write_packet(PACKET_LOCATIONS);
}

void
report_encoder_t::encode_location_sets(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->location_sets.size());
    BOOST_FOREACH(const rpt_location_set_t &location_set,
                  report->location_sets) {
        binary_encode_uint32(&packet_buffer, (uint32_t)location_set.size());
        BOOST_FOREACH(rpt_location_id_t location_id, location_set)
            binary_encode_uint32(&packet_buffer, location_id);
    }
    write_packet(PACKET_LOCATION_SETS);
}

void
report_encoder_t::encode_depgroups(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->depgroups.size());
    BOOST_FOREACH(const rpt_depgroup_map_t::value_type &v, report->depgroups) {
        binary_encode_uint32(&packet_buffer, v.first);
        binary_encode_uint32(&packet_buffer, v.second.location_set_id);
    }
    write_packet(PACKET_DEPGROUPS);
}

void
report_encoder_t::encode_loops(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->loops.size());
    BOOST_FOREACH(const rpt_loop_map_t::value_type &v, report->loops) {
        binary_encode_uint32(&packet_buffer, v.first);
        binary_encode_uint32(&packet_buffer, v.second.incl_location_set_id);
        binary_encode_uint32(&packet_buffer, v.second.excl_location_set_id);
        binary_encode_uint32(&packet_buffer, (uint32_t)v.second.depgroups.size());
        BOOST_FOREACH(rpt_depgroup_id_t depgroup_id, v.second.depgroups)
            binary_encode_uint32(&packet_buffer, depgroup_id);
    }
    write_packet(PACKET_LOOPS);
}

void
report_encoder_t::encode_stats(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->stats.size());
    BOOST_FOREACH(const rpt_stats_t &stats, report->stats) {
        binary_encode_float(&packet_buffer, stats.accesses);
        binary_encode_float(&packet_buffer, stats.cold_fetches);
        binary_encode_float(&packet_buffer, stats.replacement_fetches);
        binary_encode_float(&packet_buffer, stats.coherence_fetches);
        binary_encode_float(&packet_buffer, stats.flush_fetches);
        binary_encode_float(&packet_buffer, stats.cold_misses);
        binary_encode_float(&packet_buffer, stats.replacement_misses);
        binary_encode_float(&packet_buffer, stats.coherence_misses);
        binary_encode_float(&packet_buffer, stats.flush_misses);
        binary_encode_float(&packet_buffer, stats.upgrades);
        binary_encode_float(&packet_buffer, stats.cold_write_backs);
        binary_encode_float(&packet_buffer, stats.replacement_write_backs);
        binary_encode_float(&packet_buffer, stats.coherence_write_backs);
        binary_encode_float(&packet_buffer, stats.flush_write_backs);
        binary_encode_float(&packet_buffer, stats.spatial_utilization.value);
        binary_encode_bool(&packet_buffer, stats.spatial_utilization.valid);
        binary_encode_float(&packet_buffer, stats.fetch_utilization.value);
        binary_encode_bool(&packet_buffer, stats.fetch_utilization.valid);
        binary_encode_float(&packet_buffer, stats.write_back_utilization.value);
        binary_encode_bool(&packet_buffer, stats.write_back_utilization.valid);
        binary_encode_float(&packet_buffer, stats.communication_utilization.value);
        binary_encode_bool(&packet_buffer, stats.communication_utilization.valid);
        binary_encode_float(&packet_buffer, stats.communicated_lines);
        binary_encode_float(&packet_buffer, stats.hardware_prefetch_ratio);
        binary_encode_float(&packet_buffer, stats.random_ratio);
        binary_encode_float(&packet_buffer, stats.false_sharing_ratio.value);
        binary_encode_bool(&packet_buffer, stats.false_sharing_ratio.valid);
        binary_encode_float(&packet_buffer, stats.util_corr_fetch_ratio);
        binary_encode_float(&packet_buffer, stats.util_corr_write_back_ratio);
        binary_encode_float(&packet_buffer, stats.avoidable_fetches);
    }
    write_packet(PACKET_STATS);
}

void
report_encoder_t::encode_sliced_stats(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->sliced_stats.size());
    BOOST_FOREACH(const rpt_sliced_stats_t &sliced_stats, report->sliced_stats) {
        binary_encode_uint32(&packet_buffer, sliced_stats.global_stats);
        binary_encode_uint32(&packet_buffer, (uint32_t)sliced_stats.thread_stats.size());
        BOOST_FOREACH(const rpt_thread_stats_map_t::value_type  &v, sliced_stats.thread_stats) {
            binary_encode_uint32(&packet_buffer, v.first);
            binary_encode_uint32(&packet_buffer, v.second);
        }
    }
    write_packet(PACKET_SLICED_STATS);
}

void
report_encoder_t::encode_location_stats(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->location_stats.size());
    BOOST_FOREACH(rpt_sliced_stats_id_t id, report->location_stats)
        binary_encode_uint32(&packet_buffer, id);
    write_packet(PACKET_LOCATION_STATS);
}

void
report_encoder_t::encode_location_set_stats(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->location_set_stats.size());
    BOOST_FOREACH(rpt_sliced_stats_id_t id, report->location_set_stats)
        binary_encode_uint32(&packet_buffer, id);
    write_packet(PACKET_LOCATION_SET_STATS);
}

void
report_encoder_t::encode_global_stats(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, report->global_stats);
    write_packet(PACKET_GLOBAL_STATS);
}

void
report_encoder_t::encode_contexts(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->context_ids.size());
    BOOST_FOREACH(rpt_context_id_t id, report->context_ids)
        binary_encode_uint32(&packet_buffer, id);
    write_packet(PACKET_CONTEXTS);
}

void
report_encoder_t::encode_sweeps(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->sweeps.size());
    BOOST_FOREACH(const rpt_sweep_t &sweep, report->sweeps) {
        binary_encode_uint32(&packet_buffer, (uint32_t)sweep.size());
        BOOST_FOREACH(rpt_size_t cache_size, sweep)
            binary_encode_uint64(&packet_buffer, cache_size);
    }
    write_packet(PACKET_SWEEPS);
}

void
report_encoder_t::encode_stats_sweeps(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->stats_sweeps.size());
    BOOST_FOREACH(const rpt_stats_sweep_t &stats_sweep, report->stats_sweeps) {
        binary_encode_uint32(&packet_buffer, stats_sweep.sweep_id);
        binary_encode_uint32(&packet_buffer, (uint32_t)stats_sweep.sliced_stats_ids.size());
        BOOST_FOREACH(rpt_sliced_stats_id_t sliced_stats_id, stats_sweep.sliced_stats_ids)
            binary_encode_uint32(&packet_buffer, sliced_stats_id);
    }
    write_packet(PACKET_STATS_SWEEPS);
}

void
report_encoder_t::encode_location_set_stats_sweeps(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->location_set_stats_sweeps.size());
    BOOST_FOREACH(const rpt_stats_sweep_id_t stats_sweep_id, report->location_set_stats_sweeps)
        binary_encode_uint32(&packet_buffer, stats_sweep_id);
    write_packet(PACKET_LOCATION_SET_STATS_SWEEPS);
}

void
report_encoder_t::encode_global_stats_sweep(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, report->global_stats_sweep);
    write_packet(PACKET_GLOBAL_STATS_SWEEP);
}

void
report_encoder_t::encode_thread_mapping(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->cache_threads.size());
    BOOST_FOREACH(const rpt_thread_id_set_t &threads, report->cache_threads) {
        binary_encode_uint32(&packet_buffer, (uint32_t)threads.size());
        BOOST_FOREACH(rpt_thread_id_t thread_id, threads)
            binary_encode_uint32(&packet_buffer, thread_id);
    }
    write_packet(PACKET_THREAD_MAPPING);
}

void
report_encoder_t::encode_binaries(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->binaries.size());
    BOOST_FOREACH(const rpt_binary_t &binary, report->binaries) {
        binary_encode_string32(&packet_buffer, binary.filename);
        binary_encode_uint64(&packet_buffer, binary.base);
        binary_encode_uint64(&packet_buffer, binary.size);
        binary_encode_uint8(&packet_buffer, binary.debug_info_status);
    }
    write_packet(PACKET_BINARIES);
}

void
report_encoder_t::encode_symbols(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->symbols.size());
    BOOST_FOREACH(const rpt_symbol_t &symbol, report->symbols) {
        binary_encode_string32(&packet_buffer, symbol.name);
        binary_encode_uint64(&packet_buffer, symbol.base);
    }
    write_packet(PACKET_SYMBOLS);
}

void
report_encoder_t::encode_functions(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->functions.size());
    BOOST_FOREACH(const rpt_function_t &function, report->functions) {
        binary_encode_string32(&packet_buffer, function.name);
        binary_encode_uint64(&packet_buffer, function.base);
    }
    write_packet(PACKET_FUNCTIONS);
}

void
report_encoder_t::encode_source_files(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->source_files.size());
    BOOST_FOREACH(const rpt_source_file_t &source_file, report->source_files) {
        binary_encode_string32(&packet_buffer, source_file.filename);
	const char *buf = "";
	if (source_file.data_length > 0)
	    buf = source_file.data.get();

        binary_encode_data32(
            &packet_buffer, buf, source_file.data_length);
    }
    write_packet(PACKET_SOURCE_FILES);
}

void
report_encoder_t::encode_source_lines(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->source_lines.size());
    BOOST_FOREACH(const rpt_source_line_t &source_line, report->source_lines) {
        binary_encode_uint32(&packet_buffer, source_line.source_file_id);
        binary_encode_uint32(&packet_buffer, source_line.line);
    }
    write_packet(PACKET_SOURCE_LINES);
}

void
report_encoder_t::encode_debug_pcs(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->debug_pcs.size());
    BOOST_FOREACH(const rpt_debug_pc_t &debug_pc, report->debug_pcs) {
        binary_encode_uint32(&packet_buffer, debug_pc.pc_id);
        binary_encode_bool(&packet_buffer, debug_pc.return_address);
        binary_encode_uint32(&packet_buffer, debug_pc.binary_id);
        binary_encode_uint32(&packet_buffer, debug_pc.symbol_id);
        binary_encode_uint32(&packet_buffer, debug_pc.function_id);
        binary_encode_uint32(&packet_buffer, debug_pc.source_line_id);
    }
    write_packet(PACKET_DEBUG_PCS);
}

void
report_encoder_t::encode_issue_count(rpt_report_cptr_t report)
{
    binary_encode_uint32(&packet_buffer, (uint32_t)report->issues.size());
    write_packet(PACKET_ISSUE_COUNT);
}

void
report_encoder_t::encode_prefetch_issue(rpt_report_cptr_t report,
                                        rpt_issue_cptr_t issue_ptr)
{
    rpt_prefetch_issue_t &issue(*(rpt_prefetch_issue_t *)issue_ptr.get());
    binary_encode_uint32(&packet_buffer, issue.get_issue_id());
    binary_encode_uint8(&packet_buffer, issue.problem);
    binary_encode_uint32(&packet_buffer, issue.location);
    binary_encode_uint32(&packet_buffer, issue.loop_id);
    binary_encode_uint32(&packet_buffer, issue.depgroup_id);
    binary_encode_float(&packet_buffer, issue.cost);
    binary_encode_float(&packet_buffer, issue.fetch_ratio);
    binary_encode_uint64(&packet_buffer, issue.median_rdist);
    encode_location_set(issue.prev_locations);
    encode_location_set(issue.next_locations);
    write_packet(PACKET_PREFETCH_ISSUE);
}

void
report_encoder_t::encode_hot_spot_issue(rpt_report_cptr_t report,
                                        rpt_issue_cptr_t issue_ptr)
{
    rpt_hot_spot_issue_t &issue(*(rpt_hot_spot_issue_t *)issue_ptr.get());
    binary_encode_uint32(&packet_buffer, issue.get_issue_id());
    binary_encode_uint8(&packet_buffer, issue.type);
    binary_encode_uint32(&packet_buffer, issue.loop_id);
    binary_encode_uint32(&packet_buffer, issue.depgroup_id);
    encode_location_set(issue.previous_writers);
    write_packet(PACKET_HOT_SPOT_ISSUE);
}

void
report_encoder_t::encode_utilization_issue(rpt_report_cptr_t report,
                                           rpt_issue_cptr_t issue_ptr)
{
    rpt_utilization_issue_t &issue(*(rpt_utilization_issue_t *)issue_ptr.get());
    binary_encode_uint32(&packet_buffer, issue.get_issue_id());
    binary_encode_uint8(&packet_buffer, issue.type);
    binary_encode_uint32(&packet_buffer, issue.loop_id);
    binary_encode_uint32(&packet_buffer, issue.depgroup_id);
    encode_location_set(issue.previous_writers);
    write_packet(PACKET_UTILIZATION_ISSUE);
}

void
report_encoder_t::encode_loop_order_issue(rpt_report_cptr_t report,
                                          rpt_issue_cptr_t issue_ptr)
{
    rpt_loop_order_issue_t &issue(*(rpt_loop_order_issue_t *)issue_ptr.get());
    binary_encode_uint32(&packet_buffer, issue.get_issue_id());
    binary_encode_uint32(&packet_buffer, issue.loop_id);
    binary_encode_uint32(&packet_buffer, issue.depgroup_id);
    encode_location_set(issue.previous_writers);
    write_packet(PACKET_LOOP_ORDER_ISSUE);
}

void
report_encoder_t::encode_random_issue(rpt_report_cptr_t report,
                                      rpt_issue_cptr_t issue_ptr)
{
    rpt_random_issue_t &issue(*(rpt_random_issue_t *)issue_ptr.get());
    binary_encode_uint32(&packet_buffer, issue.get_issue_id());
    binary_encode_uint32(&packet_buffer, issue.loop_id);
    binary_encode_uint32(&packet_buffer, issue.depgroup_id);
    encode_location_set(issue.previous_writers);
    write_packet(PACKET_RANDOM_ISSUE);
}

void
report_encoder_t::encode_loop_fusion_issue(rpt_report_cptr_t report,
                                           rpt_issue_cptr_t issue_ptr)
{
    rpt_loop_fusion_issue_t &issue(*(rpt_loop_fusion_issue_t *)issue_ptr.get());
    binary_encode_uint32(&packet_buffer, issue.get_issue_id());
    binary_encode_uint32(&packet_buffer, issue.loop1_id);
    binary_encode_uint32(&packet_buffer, issue.loop2_id);
    encode_location_set(issue.loop1_pcs);
    encode_location_set(issue.loop2_pcs);
    binary_encode_uint32(&packet_buffer, issue.depgroup1_id);
    binary_encode_uint32(&packet_buffer, issue.depgroup2_id);
    binary_encode_bool(&packet_buffer, issue.move_loop2_up);
    binary_encode_bool(&packet_buffer, issue.move_loop1_down);
    encode_location_id_vector(issue.loop1_down_pcs);
    encode_location_set_vector(issue.loop1_down_blocks);
    encode_location_id_vector(issue.loop2_up_pcs);
    encode_location_set_vector(issue.loop2_up_blocks);
    encode_location_set(issue.additional_pc1);
    encode_location_set(issue.additional_pc2);
    write_packet(PACKET_LOOP_FUSION_ISSUE);
}

void
report_encoder_t::encode_block_issue(rpt_report_cptr_t report,
                                     rpt_issue_cptr_t issue_ptr)
{
    rpt_block_issue_t &issue(*(rpt_block_issue_t *)issue_ptr.get());
    binary_encode_uint32(&packet_buffer, issue.get_issue_id());
    binary_encode_uint8(&packet_buffer, issue.type_mask);
    binary_encode_uint32(&packet_buffer, issue.loop_id);
    binary_encode_uint32(&packet_buffer, issue.depgroup_id);
    binary_encode_uint32(&packet_buffer, issue.outer_loop_id);
    binary_encode_uint64(&packet_buffer, issue.outer_loop_iterations);
    binary_encode_uint32(&packet_buffer, issue.next_outer_loop_id);
    binary_encode_uint64(&packet_buffer, issue.next_outer_loop_iterations);
    encode_location_id_vector(issue.barrier_pcs);
    encode_location_set_vector(issue.barrier_blocks);
    binary_encode_uint32(&packet_buffer, issue.target_loop_id);
    binary_encode_uint64(&packet_buffer, issue.target_loop_iterations);
    encode_location_set(issue.previous_writers);
    encode_location_set(issue.next_writers);
    write_packet(PACKET_BLOCK_ISSUE);
}

void
report_encoder_t::encode_comm_spot_issue(rpt_report_cptr_t report,
                                         rpt_issue_cptr_t issue_ptr)
{
    rpt_comm_spot_issue_t &issue(*(rpt_comm_spot_issue_t *)issue_ptr.get());
    binary_encode_uint32(&packet_buffer, issue.get_issue_id());
    binary_encode_uint32(&packet_buffer, issue.loop_id);
    binary_encode_uint32(&packet_buffer, issue.depgroup_id);
    binary_encode_float(&packet_buffer, issue.comm_lines);
    encode_location_set(issue.producers);
    encode_location_set(issue.previous_writers);
    encode_location_set(issue.next_writers);
    write_packet(PACKET_COMM_SPOT_ISSUE);
}

void
report_encoder_t::encode_comm_utilization_issue(rpt_report_cptr_t report,
                                                rpt_issue_cptr_t issue_ptr)
{
    rpt_comm_utilization_issue_t &issue(*(rpt_comm_utilization_issue_t *)issue_ptr.get());
    binary_encode_uint32(&packet_buffer, issue.get_issue_id());
    binary_encode_uint32(&packet_buffer, issue.loop_id);
    binary_encode_uint32(&packet_buffer, issue.depgroup_id);
    binary_encode_float(&packet_buffer, issue.utilization);
    binary_encode_float(&packet_buffer, issue.waste);
    encode_location_set(issue.producers);
    encode_location_set(issue.previous_writers);
    encode_location_set(issue.next_writers);
    write_packet(PACKET_COMM_UTILIZATION_ISSUE);
}

void
report_encoder_t::encode_false_sharing_issue(rpt_report_cptr_t report,
                                             rpt_issue_cptr_t issue_ptr)
{
    rpt_false_sharing_issue_t &issue(*(rpt_false_sharing_issue_t *)issue_ptr.get());
    binary_encode_uint32(&packet_buffer, issue.get_issue_id());
    binary_encode_uint32(&packet_buffer, issue.loop_id);
    binary_encode_uint32(&packet_buffer, issue.depgroup_id);
    binary_encode_float(&packet_buffer, issue.ratio);
    binary_encode_float(&packet_buffer, issue.waste);
    encode_location_set(issue.offenders);
    write_packet(PACKET_FALSE_SHARING_ISSUE);
}

void
report_encoder_t::encode_wnt_possible_issue(rpt_report_cptr_t report,
                                            rpt_issue_cptr_t issue_ptr)
{
    rpt_wnt_possible_issue_t &issue(*(rpt_wnt_possible_issue_t *)issue_ptr.get());
    binary_encode_uint32(&packet_buffer, issue.get_issue_id());
    binary_encode_uint32(&packet_buffer, issue.loop_id);
    binary_encode_uint32(&packet_buffer, issue.depgroup_id);
    encode_location_set(issue.next_readers);
    write_packet(PACKET_WNT_POSSIBLE_ISSUE);
}

void
report_encoder_t::encode_prefetchnta_issue(rpt_report_cptr_t report,
                                           rpt_issue_cptr_t issue_ptr)
{
    rpt_prefetchnta_issue_t &issue(*(rpt_prefetchnta_issue_t *)issue_ptr.get());
    binary_encode_uint32(&packet_buffer, issue.get_issue_id());
    binary_encode_uint32(&packet_buffer, issue.loop_id);
    binary_encode_uint32(&packet_buffer, issue.depgroup_id);
    binary_encode_float(&packet_buffer, issue.accesses);
    binary_encode_float(&packet_buffer, issue.fetches);
    binary_encode_float(&packet_buffer, issue.required_cache_size);
    encode_location_id_vector(issue.pc1s);
    encode_stat_fp_vector(issue.pc1_fractions);
    encode_stat_fp_vector(issue.pc1_required_cache_sizes);
    encode_location_id_vector(issue.pc2s);
    encode_stat_fp_vector(issue.pc2_fractions);
    write_packet(PACKET_PREFETCHNTA_ISSUE);
}

// ----- report_decoder_t -----------------------------------------------------

report_decoder_t::report_decoder_t(warning_consumer_ptr_t warning_consumer)
    : warning_consumer(warning_consumer),
      report(new rpt_report_t()),
      file_header_handled(false),
      waiting_for_header(true)
{
}

stream_status_t
report_decoder_t::process_data(const char *buffer_, size_t len)
{
    buffer.write(buffer_, len);

    if (!file_header_handled) {
        if (buffer.size() < FILE_HEADER_SIZE)
            return STREAM_OK;

        buffer.assert_read(FILE_HEADER_SIZE);

        uint16_t version = binary_decode_uint16(&buffer);
        uint16_t minimum_version = binary_decode_uint16(&buffer);

        if (version < REPORT_CODEC_VERSION_DECODE_MINIMUM)
            throw codec_exception_t("Report is too old to be read.", true);
        if (minimum_version > REPORT_CODEC_VERSION)
            throw codec_exception_t("Report is too new to be read.", true);

        file_header_handled = true;
    }

    while (1)
        if (waiting_for_header) {
            if (buffer.size() < PACKET_HEADER_SIZE)
                return STREAM_OK;

            buffer.assert_read(PACKET_HEADER_SIZE);

            packet_type = binary_decode_uint16(&buffer);
            packet_size = binary_decode_uint32(&buffer);

            waiting_for_header = false;
        }
        else {
            if (buffer.size() < packet_size)
                return STREAM_OK;

            buffer.assert_read(packet_size);

            switch (packet_type) {
            case PACKET_UUID:
                decode_uuid();
                break;
            case PACKET_SAMPLING_INFO:
                decode_sampling_info();
                break;
            case PACKET_ANALYSIS_INFO:
                decode_analysis_info();
                break;
            case PACKET_HARDWARE_CONFIG:
                decode_hardware_config();
                break;
            case PACKET_HARDWARE_CONFIG2:
                decode_hardware_config2();
                break;
            case PACKET_ADDRESSES:
                decode_addresses();
                break;
            case PACKET_ACCESS_CLASSES:
                decode_access_classes();
                break;
            case PACKET_TRUNC_STACKS:
                decode_trunc_stacks();
                break;
            case PACKET_FULL_STACKS:
                decode_full_stacks();
                break;
            case PACKET_LOCATIONS:
                decode_locations();
                break;
            case PACKET_LOCATION_SETS:
                decode_location_sets();
                break;
            case PACKET_DEPGROUPS:
                decode_depgroups();
                break;
            case PACKET_LOOPS:
                decode_loops();
                break;
            case PACKET_STATS:
                decode_stats();
                break;
            case PACKET_SLICED_STATS:
                decode_sliced_stats();
                break;
            case PACKET_LOCATION_STATS:
                decode_location_stats();
                break;
            case PACKET_LOCATION_SET_STATS:
                decode_location_set_stats();
                break;
            case PACKET_GLOBAL_STATS:
                decode_global_stats();
                break;
            case PACKET_CONTEXTS:
                decode_contexts();
                break;
            case PACKET_SWEEPS:
                decode_sweeps();
                break;
            case PACKET_STATS_SWEEPS:
                decode_stats_sweeps();
                break;
            case PACKET_LOCATION_SET_STATS_SWEEPS:
                decode_location_set_stats_sweeps();
                break;
            case PACKET_GLOBAL_STATS_SWEEP:
                decode_global_stats_sweep();
                break;
            case PACKET_THREAD_MAPPING:
                decode_thread_mapping();
                break;
            case PACKET_BINARIES:
                decode_binaries();
                break;
            case PACKET_SYMBOLS:
                decode_symbols();
                break;
            case PACKET_FUNCTIONS:
                decode_functions();
                break;
            case PACKET_SOURCE_FILES:
                decode_source_files();
                break;
            case PACKET_SOURCE_LINES:
                decode_source_lines();
                break;
            case PACKET_DEBUG_PCS:
                decode_debug_pcs();
                break;
            case PACKET_ISSUE_COUNT:
                decode_issue_count();
                break;
            case PACKET_PREFETCH_ISSUE:
                decode_prefetch_issue();
                break;
            case PACKET_HOT_SPOT_ISSUE:
                decode_hot_spot_issue();
                break;
            case PACKET_UTILIZATION_ISSUE:
                decode_utilization_issue();
                break;
            case PACKET_LOOP_ORDER_ISSUE:
                decode_loop_order_issue();
                break;
            case PACKET_RANDOM_ISSUE:
                decode_random_issue();
                break;
            case PACKET_LOOP_FUSION_ISSUE:
                decode_loop_fusion_issue();
                break;
            case PACKET_BLOCK_ISSUE:
                decode_block_issue();
                break;
            case PACKET_COMM_SPOT_ISSUE:
                decode_comm_spot_issue();
                break;
            case PACKET_COMM_UTILIZATION_ISSUE:
                decode_comm_utilization_issue();
                break;
            case PACKET_FALSE_SHARING_ISSUE:
                decode_false_sharing_issue();
                break;
            case PACKET_WNT_POSSIBLE_ISSUE:
                decode_wnt_possible_issue();
                break;
            case PACKET_PREFETCHNTA_ISSUE:
                decode_prefetchnta_issue();
                break;
            default: {
                stringstream ss;
                ss << "Ignoring unknown report part of type " << packet_type << ".";
                warning_consumer->info(ss.str());
                buffer.read(packet_size);
            } break;
            }

            waiting_for_header = true;
        }
}

stream_status_t
report_decoder_t::flush()
{
    return STREAM_OK;
}

rpt_report_cptr_t
report_decoder_t::get_report()
{
    return report;
}

void
report_decoder_t::decode_location_id_vector(
    rpt_location_id_vector_t &location_id_vector)
{
    location_id_vector.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_location_id_t &l, location_id_vector)
        l = binary_decode_uint32(&buffer);
}

void
report_decoder_t::decode_location_set(
    rpt_location_set_t &location_set)
{
    uint32_t num_locations = binary_decode_uint32(&buffer);
    for (uint32_t i = 0; i < num_locations; i++)
        location_set.insert(binary_decode_uint32(&buffer));
}

void
report_decoder_t::decode_location_set_vector(
    rpt_location_set_vector_t &location_set_vector)
{
    location_set_vector.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_location_set_t &ls, location_set_vector)
        decode_location_set(ls);
}

void
report_decoder_t::decode_stat_fp_vector(
    rpt_stat_fp_vector_t &stat_fp_vector)
{
    stat_fp_vector.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_stat_fp_t &v, stat_fp_vector)
        v = binary_decode_float(&buffer);
}

void
report_decoder_t::decode_uuid()
{
    report->uuid = binary_decode_string32(&buffer);
}

void
report_decoder_t::decode_sampling_info()
{
    report->mt_sampled = binary_decode_bool(&buffer);
    report->burst_sampled = binary_decode_bool(&buffer);
    report->begin_time = binary_decode_uint64(&buffer);
    report->end_time = binary_decode_uint64(&buffer);
    report->argv = binary_decode_string32_vector(&buffer);
    report->user = binary_decode_string32(&buffer);
    report->architecture = (rpt_architecture_t)binary_decode_uint32(&buffer);
    report->sample_count = binary_decode_uint32(&buffer);
}

void
report_decoder_t::decode_analysis_info()
{
    report->sample_file = binary_decode_string32(&buffer);
    report->analysis_time = binary_decode_uint64(&buffer);
    report->analysis_argv = binary_decode_string32_vector(&buffer);
    report->title = binary_decode_string32(&buffer);
}

void
report_decoder_t::decode_hardware_config()
{
    report->cpu_model = binary_decode_string32(&buffer);
    report->num_cpus = binary_decode_uint32(&buffer);
    report->num_caches = binary_decode_uint32(&buffer);
    report->num_cache_levels = binary_decode_uint32(&buffer);
    report->cache_level = binary_decode_uint32(&buffer);
    report->cache_size = binary_decode_uint64(&buffer);
    report->line_size = binary_decode_uint64(&buffer);
    report->replacement_policy = binary_decode_string32(&buffer);
    report->swpf_active = binary_decode_bool(&buffer);
}

void
report_decoder_t::decode_hardware_config2()
{
    report->autodetect_failed = binary_decode_bool(&buffer);
}

void
report_decoder_t::decode_addresses()
{
    report->addresses.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_address_t &addr, report->addresses)
        addr = binary_decode_uint64(&buffer);
}

void
report_decoder_t::decode_access_classes()
{
    report->access_classes.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(uint8_t &access_classes, report->access_classes)
        access_classes = binary_decode_uint8(&buffer);
}

void
report_decoder_t::decode_trunc_stacks()
{
    report->trunc_stacks.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_trunc_stack_t &trunc_stack, report->trunc_stacks) {
        trunc_stack.resize(binary_decode_uint16(&buffer));
        BOOST_FOREACH(rpt_address_id_t &addr_id, trunc_stack)
            addr_id = binary_decode_uint32(&buffer);
    }
}

void
report_decoder_t::decode_full_stacks()
{
    report->full_stacks.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_full_stack_t &full_stack, report->full_stacks) {
        full_stack.resize(binary_decode_uint16(&buffer));
        BOOST_FOREACH(rpt_full_stack_frame_t &full_stack_frame, full_stack) {
            full_stack_frame.pc_id = binary_decode_uint32(&buffer);
            full_stack_frame.fraction = binary_decode_float(&buffer);
        }
    }
}

void
report_decoder_t::decode_locations()
{
    report->locations.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_location_t &location, report->locations) {
        location.pc_id = binary_decode_uint32(&buffer);
        location.trunc_stack_id = binary_decode_uint32(&buffer);
    }
}

void
report_decoder_t::decode_location_sets()
{
    report->location_sets.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_location_set_t &location_set, report->location_sets) {
        uint32_t num_locations = binary_decode_uint32(&buffer);
        for (uint32_t i = 0; i < num_locations; i++)
            location_set.insert(binary_decode_uint32(&buffer));
    }
}

void
report_decoder_t::decode_depgroups()
{
    uint32_t num_depgroups = binary_decode_uint32(&buffer);
    for (uint32_t i = 0; i < num_depgroups; i++) {
        rpt_depgroup_id_t depgroup_id = binary_decode_uint32(&buffer);
        rpt_location_set_id_t location_set_id = binary_decode_uint32(&buffer);
        report->depgroups[depgroup_id] = rpt_depgroup_t(location_set_id);
    }
}

void
report_decoder_t::decode_loops()
{
    uint32_t num_loops = binary_decode_uint32(&buffer);
    for (uint32_t i = 0; i < num_loops; i++) {
        rpt_loop_id_t loop_id = binary_decode_uint32(&buffer);
        rpt_location_set_id_t incl_location_set_id = binary_decode_uint32(&buffer);
        rpt_location_set_id_t excl_location_set_id = binary_decode_uint32(&buffer);
        uint32_t num_depgroups = binary_decode_uint32(&buffer);
        rpt_depgroup_set_t depgroups;
        for (uint32_t i = 0; i < num_depgroups; i++)
            depgroups.insert(binary_decode_uint32(&buffer));
        report->loops[loop_id] = rpt_loop_t(
            incl_location_set_id,
            excl_location_set_id,
            depgroups);
    }
}

void
report_decoder_t::decode_stats()
{
    report->stats.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_stats_t &stats, report->stats) {
        stats.accesses = binary_decode_float(&buffer);
        stats.cold_fetches = binary_decode_float(&buffer);
        stats.replacement_fetches = binary_decode_float(&buffer);
        stats.coherence_fetches = binary_decode_float(&buffer);
        stats.flush_fetches = binary_decode_float(&buffer);
        stats.cold_misses = binary_decode_float(&buffer);
        stats.replacement_misses = binary_decode_float(&buffer);
        stats.coherence_misses = binary_decode_float(&buffer);
        stats.flush_misses = binary_decode_float(&buffer);
        stats.upgrades = binary_decode_float(&buffer);
        stats.cold_write_backs = binary_decode_float(&buffer);
        stats.replacement_write_backs = binary_decode_float(&buffer);
        stats.coherence_write_backs = binary_decode_float(&buffer);
        stats.flush_write_backs = binary_decode_float(&buffer);
        stats.spatial_utilization.value = binary_decode_float(&buffer);
        stats.spatial_utilization.valid = binary_decode_bool(&buffer);
        stats.fetch_utilization.value = binary_decode_float(&buffer);
        stats.fetch_utilization.valid = binary_decode_bool(&buffer);
        stats.write_back_utilization.value = binary_decode_float(&buffer);
        stats.write_back_utilization.valid = binary_decode_bool(&buffer);
        stats.communication_utilization.value = binary_decode_float(&buffer);
        stats.communication_utilization.valid = binary_decode_bool(&buffer);
        stats.communicated_lines = binary_decode_float(&buffer);
        stats.hardware_prefetch_ratio = binary_decode_float(&buffer);
        stats.random_ratio = binary_decode_float(&buffer);
        stats.false_sharing_ratio.value = binary_decode_float(&buffer);
        stats.false_sharing_ratio.valid = binary_decode_bool(&buffer);
        stats.util_corr_fetch_ratio = binary_decode_float(&buffer);
        stats.util_corr_write_back_ratio = binary_decode_float(&buffer);
        stats.avoidable_fetches = binary_decode_float(&buffer);
    }
}

void
report_decoder_t::decode_sliced_stats()
{
    report->sliced_stats.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_sliced_stats_t&sliced_stats, report->sliced_stats) {
        sliced_stats.global_stats = binary_decode_uint32(&buffer);
        uint32_t num_threads = binary_decode_uint32(&buffer);
        for (uint32_t i = 0; i < num_threads; i++)
            sliced_stats.thread_stats[binary_decode_uint32(&buffer)] =
                binary_decode_uint32(&buffer);
    }
}

void
report_decoder_t::decode_location_stats()
{
    report->location_stats.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_sliced_stats_id_t &id, report->location_stats)
        id = binary_decode_uint32(&buffer);
}

void
report_decoder_t::decode_location_set_stats()
{
    report->location_set_stats.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_sliced_stats_id_t &id, report->location_set_stats)
        id = binary_decode_uint32(&buffer);
}

void
report_decoder_t::decode_global_stats()
{
    report->global_stats = binary_decode_uint32(&buffer);
}

void
report_decoder_t::decode_contexts()
{
    report->context_ids.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_context_id_t &id, report->context_ids)
        id = binary_decode_uint32(&buffer);
}

void
report_decoder_t::decode_sweeps()
{
    report->sweeps.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_sweep_t &sweep, report->sweeps) {
        sweep.resize(binary_decode_uint32(&buffer));
        BOOST_FOREACH(rpt_size_t &cache_size, sweep)
            cache_size = binary_decode_uint64(&buffer);
    }
}

void
report_decoder_t::decode_stats_sweeps()
{
    report->stats_sweeps.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_stats_sweep_t &stats_sweep, report->stats_sweeps) {
        stats_sweep.sweep_id = binary_decode_uint32(&buffer);
        stats_sweep.sliced_stats_ids.resize(binary_decode_uint32(&buffer));
        BOOST_FOREACH(rpt_sliced_stats_id_t &sliced_stats_id, stats_sweep.sliced_stats_ids)
            sliced_stats_id = binary_decode_uint32(&buffer);
    }
}

void
report_decoder_t::decode_location_set_stats_sweeps()
{
    report->location_set_stats_sweeps.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_stats_sweep_id_t &stats_sweep_id, report->location_set_stats_sweeps)
        stats_sweep_id = binary_decode_uint32(&buffer);
}

void
report_decoder_t::decode_global_stats_sweep()
{
    report->global_stats_sweep = binary_decode_uint32(&buffer);
}

void
report_decoder_t::decode_thread_mapping()
{
    report->cache_threads.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_thread_id_set_t &threads, report->cache_threads) {
        uint32_t num_threads = binary_decode_uint32(&buffer);
        for (uint32_t i = 0; i < num_threads; i++)
            threads.insert(binary_decode_uint32(&buffer));
    }
}

void
report_decoder_t::decode_binaries()
{
    report->binaries.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_binary_t &binary, report->binaries) {
        binary.filename = binary_decode_string32(&buffer);
        binary.base = binary_decode_uint64(&buffer);
        binary.size = binary_decode_uint64(&buffer);
        binary.debug_info_status = (rpt_debug_info_status_t)binary_decode_uint8(&buffer);
    }
}

void
report_decoder_t::decode_symbols()
{
    report->symbols.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_symbol_t &symbol, report->symbols) {
        symbol.name = binary_decode_string32(&buffer);
        symbol.base = binary_decode_uint64(&buffer);
    }
}

void
report_decoder_t::decode_functions()
{
    report->functions.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_function_t &function, report->functions) {
        function.name = binary_decode_string32(&buffer);
        function.base = binary_decode_uint64(&buffer);
    }
}

void
report_decoder_t::decode_source_files()
{
    report->source_files.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_source_file_t &source_file, report->source_files) {
        source_file.filename = binary_decode_string32(&buffer);
        char *data;
        size_t data_length;
        binary_decode_data32(&buffer, &data, &data_length);
        source_file.data = char_array_ptr_t(data);
        source_file.data_length = (uint32_t) data_length;
    }
}

void
report_decoder_t::decode_source_lines()
{
    report->source_lines.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_source_line_t &source_line, report->source_lines) {
        source_line.source_file_id = binary_decode_uint32(&buffer);
        source_line.line = binary_decode_uint32(&buffer);
    }
}

void
report_decoder_t::decode_debug_pcs()
{
    report->debug_pcs.resize(binary_decode_uint32(&buffer));
    BOOST_FOREACH(rpt_debug_pc_t &debug_pc, report->debug_pcs) {
        debug_pc.pc_id = binary_decode_uint32(&buffer);
        debug_pc.return_address = binary_decode_bool(&buffer);
        debug_pc.binary_id = binary_decode_uint32(&buffer);
        debug_pc.symbol_id = binary_decode_uint32(&buffer);
        debug_pc.function_id = binary_decode_uint32(&buffer);
        debug_pc.source_line_id = binary_decode_uint32(&buffer);
    }
}

void
report_decoder_t::decode_issue_count()
{
    report->issues.resize(binary_decode_uint32(&buffer));
}

void
report_decoder_t::decode_prefetch_issue()
{
    rpt_issue_id_t issue_id = binary_decode_uint32(&buffer);
    rpt_prefetch_issue_t *issue(new rpt_prefetch_issue_t(issue_id));
    report->issues.at(issue_id).reset(issue);
    issue->problem = (rpt_prefetch_issue_t::problem_t)binary_decode_uint8(&buffer);
    issue->location = binary_decode_uint32(&buffer);
    issue->loop_id = binary_decode_uint32(&buffer);
    issue->depgroup_id = binary_decode_uint32(&buffer);
    issue->cost = binary_decode_float(&buffer);
    issue->fetch_ratio = binary_decode_float(&buffer);
    issue->median_rdist = binary_decode_uint64(&buffer);
    decode_location_set(issue->prev_locations);
    decode_location_set(issue->next_locations);
}

void
report_decoder_t::decode_hot_spot_issue()
{
    rpt_issue_id_t issue_id = binary_decode_uint32(&buffer);
    rpt_hot_spot_issue_t *issue(new rpt_hot_spot_issue_t(issue_id));
    report->issues.at(issue_id).reset(issue);
    issue->type = (rpt_hot_spot_issue_t::type_t)binary_decode_uint8(&buffer);
    issue->loop_id = binary_decode_uint32(&buffer);
    issue->depgroup_id = binary_decode_uint32(&buffer);
    decode_location_set(issue->previous_writers);
}

void
report_decoder_t::decode_utilization_issue()
{
    rpt_issue_id_t issue_id = binary_decode_uint32(&buffer);
    rpt_utilization_issue_t *issue(new rpt_utilization_issue_t(issue_id));
    report->issues.at(issue_id).reset(issue);
    issue->type = (rpt_utilization_issue_t::type_t)binary_decode_uint8(&buffer);
    issue->loop_id = binary_decode_uint32(&buffer);
    issue->depgroup_id = binary_decode_uint32(&buffer);
    decode_location_set(issue->previous_writers);
}

void
report_decoder_t::decode_loop_order_issue()
{
    rpt_issue_id_t issue_id = binary_decode_uint32(&buffer);
    rpt_loop_order_issue_t *issue(new rpt_loop_order_issue_t(issue_id));
    report->issues.at(issue_id).reset(issue);
    issue->loop_id = binary_decode_uint32(&buffer);
    issue->depgroup_id = binary_decode_uint32(&buffer);
    decode_location_set(issue->previous_writers);
}

void
report_decoder_t::decode_random_issue()
{
    rpt_issue_id_t issue_id = binary_decode_uint32(&buffer);
    rpt_random_issue_t *issue(new rpt_random_issue_t(issue_id));
    report->issues.at(issue_id).reset(issue);
    issue->loop_id = binary_decode_uint32(&buffer);
    issue->depgroup_id = binary_decode_uint32(&buffer);
    decode_location_set(issue->previous_writers);
}

void
report_decoder_t::decode_loop_fusion_issue()
{
    rpt_issue_id_t issue_id = binary_decode_uint32(&buffer);
    rpt_loop_fusion_issue_t *issue(new rpt_loop_fusion_issue_t(issue_id));
    report->issues.at(issue_id).reset(issue);

    issue->loop1_id = binary_decode_uint32(&buffer);
    issue->loop2_id = binary_decode_uint32(&buffer);
    decode_location_set(issue->loop1_pcs);
    decode_location_set(issue->loop2_pcs);
    issue->depgroup1_id = binary_decode_uint32(&buffer);
    issue->depgroup2_id = binary_decode_uint32(&buffer);
    issue->move_loop2_up = binary_decode_bool(&buffer);
    issue->move_loop1_down = binary_decode_bool(&buffer);
    decode_location_id_vector(issue->loop1_down_pcs);
    decode_location_set_vector(issue->loop1_down_blocks);
    decode_location_id_vector(issue->loop2_up_pcs);
    decode_location_set_vector(issue->loop2_up_blocks);
    decode_location_set(issue->additional_pc1);
    decode_location_set(issue->additional_pc2);
}

void
report_decoder_t::decode_block_issue()
{
    rpt_issue_id_t issue_id = binary_decode_uint32(&buffer);
    rpt_block_issue_t *issue(new rpt_block_issue_t(issue_id));
    report->issues.at(issue_id).reset(issue);
    issue->type_mask = (rpt_block_issue_t::type_t)binary_decode_uint8(&buffer);
    issue->loop_id = binary_decode_uint32(&buffer);
    issue->depgroup_id = binary_decode_uint32(&buffer);
    issue->outer_loop_id = binary_decode_uint32(&buffer);
    issue->outer_loop_iterations = binary_decode_uint64(&buffer);
    issue->next_outer_loop_id = binary_decode_uint32(&buffer);
    issue->next_outer_loop_iterations = binary_decode_uint64(&buffer);
    decode_location_id_vector(issue->barrier_pcs);
    decode_location_set_vector(issue->barrier_blocks);
    issue->target_loop_id = binary_decode_uint32(&buffer);
    issue->target_loop_iterations = binary_decode_uint64(&buffer);
    decode_location_set(issue->previous_writers);
    decode_location_set(issue->next_writers);
}

void
report_decoder_t::decode_comm_spot_issue()
{
    rpt_issue_id_t issue_id = binary_decode_uint32(&buffer);
    rpt_comm_spot_issue_t *issue(new rpt_comm_spot_issue_t(issue_id));
    report->issues.at(issue_id).reset(issue);
    issue->loop_id = binary_decode_uint32(&buffer);
    issue->depgroup_id = binary_decode_uint32(&buffer);
    issue->comm_lines = binary_decode_float(&buffer);
    decode_location_set(issue->producers);
    decode_location_set(issue->previous_writers);
    decode_location_set(issue->next_writers);
}

void
report_decoder_t::decode_comm_utilization_issue()
{
    rpt_issue_id_t issue_id = binary_decode_uint32(&buffer);
    rpt_comm_utilization_issue_t *issue(new rpt_comm_utilization_issue_t(issue_id));
    report->issues.at(issue_id).reset(issue);

    issue->loop_id = binary_decode_uint32(&buffer);
    issue->depgroup_id = binary_decode_uint32(&buffer);
    issue->utilization = binary_decode_float(&buffer);
    issue->waste = binary_decode_float(&buffer);
    decode_location_set(issue->producers);
    decode_location_set(issue->previous_writers);
    decode_location_set(issue->next_writers);
}

void
report_decoder_t::decode_false_sharing_issue()
{
    rpt_issue_id_t issue_id = binary_decode_uint32(&buffer);
    rpt_false_sharing_issue_t *issue(new rpt_false_sharing_issue_t(issue_id));
    report->issues.at(issue_id).reset(issue);
    issue->loop_id = binary_decode_uint32(&buffer);
    issue->depgroup_id = binary_decode_uint32(&buffer);
    issue->ratio = binary_decode_float(&buffer);
    issue->waste = binary_decode_float(&buffer);
    decode_location_set(issue->offenders);
}

void
report_decoder_t::decode_wnt_possible_issue()
{
    rpt_issue_id_t issue_id = binary_decode_uint32(&buffer);
    rpt_wnt_possible_issue_t *issue(new rpt_wnt_possible_issue_t(issue_id));
    report->issues.at(issue_id).reset(issue);
    issue->loop_id = binary_decode_uint32(&buffer);
    issue->depgroup_id = binary_decode_uint32(&buffer);
    decode_location_set(issue->next_readers);
}

void
report_decoder_t::decode_prefetchnta_issue()
{
    rpt_issue_id_t issue_id = binary_decode_uint32(&buffer);
    rpt_prefetchnta_issue_t *issue(new rpt_prefetchnta_issue_t(issue_id));
    report->issues.at(issue_id).reset(issue);

    issue->loop_id = binary_decode_uint32(&buffer);
    issue->depgroup_id = binary_decode_uint32(&buffer);
    issue->accesses = binary_decode_float(&buffer);
    issue->fetches = binary_decode_float(&buffer);
    issue->required_cache_size = binary_decode_float(&buffer);
    decode_location_id_vector(issue->pc1s);
    decode_stat_fp_vector(issue->pc1_fractions);
    decode_stat_fp_vector(issue->pc1_required_cache_sizes);
    decode_location_id_vector(issue->pc2s);
    decode_stat_fp_vector(issue->pc2_fractions);
}

END_NAMESPACE;
END_NAMESPACE;
