/// @file tools/report/conf.cc

// Copyright (c) 2007-2010,2012 Rogue Wave Software, Inc.  All rights reserved.

#include "conf.hh"

#include "analysis/fetch_ratio/windowing_adaptive.hh"
#include "analysis/fetch_ratio/windowing_static.hh"
#include "analysis/fetch_ratio/windowing_burst.hh"
#include "library/exe_path/exe_path.hh"
#include "library/lang/am_assert.hh"

#include <boost/program_options.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/tokenizer.hpp>

#include <iostream>
#include <time.h>

#if 0
#ifndef _WIN32
#include <unistd.h>
#else
#include <windows.h>
#include <stdio.h>
#include <tchar.h>
#endif
#endif

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);
BEGIN_NAMESPACE(report_generator);

using namespace acumem::analysis;
namespace fs = boost::filesystem;
namespace po = boost::program_options;

#define SAMPLES_PER_WINDOW 100

static cpudb_cpu_ptr_t
create_generic_cpu_model()
{
    cpudb_cpu_ptr_t cpu(
	new cpudb_cpu_t("generic",
			"Generic processor model",
			0, // family
			0, // model
			1, // cores
			1, // threads per core
			false) // prefetchw dirty?
	);
    cpu->caches.resize(2);

    cpu->caches[0].reset(
	new cpudb_cache_t("L1", // name
			  1, // level
			  1, // num cores
			  true, true, // data/instr
			  FNR_Prefetch_Class_Any,
                          true, // hwpf
			  65536, // size
			  64)); // line size

    cpu->caches[1].reset(
	new cpudb_cache_t("L2", // name
			  2, // level
			  1, // num cores
			  true, true, // data/instr
			  FNR_Prefetch_Class_Any,
                          true, // hwpf
			  2097152, // size
			  64)); // line size

    return cpu;
}

static cpudb_cpu_cptr_t
autodetect_cpu(const cpudb_t &cpudb, const warning_consumer_ptr_t &warning, bool &autodetect_failed)
{
    cpudb_cpu_cptr_t cpu(cpudb.autodetect_current_cpu(autodetect_failed));
    if (!cpu) {
	cpu = create_generic_cpu_model();
	warning->warning("Processor autodetection failed, "
                         "using a generic model.");
    }

    return cpu;
}

static void
print_cpu_usage(ostream &os, const cpudb_t &cpudb)
{
    os << "The processor should be specified like vendor/processor, "
	"use one of the vendors and processors from the list below. "
	"The current processor will be autodetected if this option is "
	"omitted or 'auto' is specified." << endl;

    BOOST_FOREACH(const cpudb_manufacturer_cptr_t &mfgr, cpudb.manufacturers) {
	const cpudb_cpu_vector_t &cpus(mfgr->get_cpus());
	os << endl << mfgr->id << " - " << mfgr->name << ":" << endl;
	BOOST_FOREACH(const cpudb_cpu_cptr_t &cpu, cpus) {
	    BOOST_FOREACH(const string &name, cpu->names)
		os << "  " << cpu->id << " - " << name << endl;
	}
    }
}

static void
general_help_hint(const warning_consumer_ptr_t &warning_consumer)
{
    warning_consumer->none("Use option -h or --help to get help.");
}


static bool
is_not_power_of_two(uint64_t val)
{
    return !(((val)&(val-1)) == 0);
}

#define RETURN(exit_, exit_code_) do {          \
    exit = (exit_);                             \
    exit_code = (exit_code_);                   \
    return;                                     \
} while (0)

conf_t::conf_t(bool &exit,
               int &exit_code,
               warning_consumer_ptr_t &warning_consumer,
	       feedback_ptr_t feedback,
               int argc, char **argv,
               bool feature_mt_advice,
               bool feature_multisocket,
               bool feature_unlimited_callstack,
               bool feature_blocking_expert,
	       bool feature_thread_binding,
               const string &version)
    : report_file("report.tsr"),
      percentage(1.0),
      cache_size(65536),
      line_size(64),
      replacement_policy("random"),
      num_cpus(0),
      cache_level(1),
      autodetect_failed(false),
      stack_depth(1),
      debug_symbol_level((int)Debug_Full),
      blocking_expert(false),
      num_caches(0),
      real_thread_id(false),
      new_algorithm(false),
      low_mem(false),
      verbose(false)
{
    string windowing_engine;
    fnr_sample_count_t samples_per_window(SAMPLES_PER_WINDOW);
    fnr_sample_count_t seed_size(SAMPLES_PER_WINDOW);
    fnr_sample_count_t growth_size(SAMPLES_PER_WINDOW / 4);
    double max_delta(0.25);
    string cpu_id;

    vector<string> groups;
    string cache_size_str;

    cpudb_t cpudb(warning_consumer);

    try {
	fs::path install_dir = acumem::exe_path::get_installation_directory();
	if (install_dir.empty()) {
	    warning_consumer->warning("Installation directory not found. "
				      "Could not load CPU database. "
				      "Results will be inaccurate.");
	} else {
	    fs::path cpudb_path = install_dir / "share" / PACKAGE / "cpudb.xml";
	    cpudb.load_from_file(cpudb_path.string());
	}
    } catch (const xml_exception_t &e) {
	stringstream ss;
	ss << "Failed to load CPU database. XML path: " << e.path()
	   << ". Reason: " << e.why() << " Results may not be accurate.";
	warning_consumer->warning(ss.str());
    } // Let all other exceptions be handled by outer catch clauses.

    // Our own command line parameters and the current date and time.
    for (int i = 0; i < argc; i++)
        analysis_argv.push_back(argv[i]);
    analysis_time = time(0) * 1000000000ULL;

    po::options_description opt_desc_general("Options");
    opt_desc_general.add_options()
	("help,h", "(optional) Print help message.")
	("version", "(optional) Print version information.")

	("in-file,i", po::value<string>(&sample_file),
	 "Specifies the input file.")


	("title,t", po::value<string>(&title),
	 "(optional) Title of the report.")

	("report,o", po::value<string>(&report_file),
	 "(optional) Name of the generated report file, defaults to "
         "'report.tsr'.")

	("percent,p", po::value<double>(&percentage),
	 "(optional) Percent of total fetches, upgrades or write-backs "
	 "required for advice to be reported. The default is 1.")

        ("verbose",
         "(optional) Show which filenames are tried when searching for debug "
         "information and source code.");

    if (feature_unlimited_callstack)
	opt_desc_general.add_options()
	    ("depth,d", po::value<fnr_stack_depth_t>(&stack_depth),
	     "(optional) Stack depth to use for separating issues caused by "
             "different calls to the same function. Default 1. Use 0 to merge "
             "all different call paths into a function for analysis.");
    else
	stack_depth = 0;

    opt_desc_general.add_options()
	("source-dir,s", po::value< vector<string> >(&source_directories),
	 "(optional) Additional directories to look for source code in.")

	("binary,b", po::value< vector<string> >(&extra_binaries),
	 "(optional) Specifies an additional binary containing debug "
         "information to use if the sampled binary with the same file name "
         "can not be found.")

	("debug-dir,D", po::value< vector<string> >(&debug_directories),
	 "(optional) Directories to resolve debug links in.")

	("debug-symbol-level", po::value<int>(&debug_symbol_level),
	 "(experimental) Balance debug symbol detail and processing speed.\n"
	 "0 - no debug symbols\n"
         "1 - line number\n"
	 "2 - +publ. symbols\n"
         "3 - +full debug info (default)")

	("c++filt",po::value<string>(&cppfilt)->implicit_value("c++filt"),
	 "(optional) De-mangle c++ symbols with the speficied program. If the "
         "option is given without a program it defaults to 'c++filt'.")

	("new-hwpf-algorithm",
	 "(experimental) Use a modified algorithm to estimate hardware "
	 "prefetcher activity")
	;

    if (feature_blocking_expert)
	opt_desc_general.add_options()
	    ("no-barriers",
	     "(optional, expert) Disregard potential code motion barriers "
	     "when deciding which blocking and "
	     "fusion issues to present.");

    po::options_description opt_desc_model("Processor model options");
    opt_desc_model.add_options()
	("cpu", po::value<string>(&cpu_id),
	 "(optional) Processor model use for the analysis. Use 'help' to get "
	 "a list of available CPUs. Default is to autodetect the processor "
         "model of the computer the report is being generated on.");

    if (feature_mt_advice && feature_multisocket)
	opt_desc_model.add_options()
	    ("number-of-cpus", po::value<unsigned int>(&num_cpus),
	     "(optional) Number of physical processors to include in the "
	     "analysis. Default is to autodetect the number of processors "
             "of the computer the report is being generated on.");
    else
        num_cpus = 1;

    opt_desc_model.add_options()
	("level", po::value<unsigned int>(&cache_level),
	 "(optional) Cache level to analyze. Defaults to the highest "
         "available level.")

	("cache-size,c", po::value<string>(&cache_size_str),
	 "(optional) Overrides the cache size from the selected CPU model.")

	("line-size,l", po::value<fnr_line_size_t>(&line_size),
	 "(optional) Overrides the line size from the selected CPU model.")

	("replacement,r", po::value<string>(&replacement_policy),
	 "(optional) Cache replacement policy. Must be 'random' or 'lru'. "
	 "Default: random.");

    if (feature_mt_advice)
	opt_desc_model.add_options()
	    ("number-of-caches,n", po::value<unsigned int>(&num_caches),
	     "(optional) Total number of caches to assign threads to. Should "
             "match the number of caches of the desired cache level for the "
	     "intended processor/architecture. 0 assigns a private cache to "
             "each thread. Defaults to what is specified by the processor "
             "model and cache level.");
    else
	num_caches = 1;

    if (feature_mt_advice && feature_multisocket && feature_thread_binding)
	opt_desc_model.add_options()
	    ("group", po::value<vector<string> >(&groups),
	     "(optional) Introduce a new group of threads sharing the same cache. "
	     "arg = thread-1,thread-2,...")

	    ("real-thread-id",
	     "(optional) With this options, the thread id:s specified with the --group "
	     "parameter should be "
	     "considered to be real thread id:s. Without this option, "
	     "virtual threads id:s 0..n are assigned to threads in the order "
	     "they appear in the sample file.");


    po::options_description opt_desc_visible("");
    opt_desc_visible.add(opt_desc_general)
	.add(opt_desc_model);


    po::options_description opt_desc_windowing("Windowing options");
    opt_desc_windowing.add_options()
	("windowing", po::value<string>(&windowing_engine),
	 "(optional) Window engine to use (static/adaptive/burst).")
	("samples-per-window,w", po::value<fnr_sample_count_t>(&samples_per_window),
	 "(optional) Samples per window, only applies to static windowing.")
	("window-seed", po::value<fnr_sample_count_t>(&seed_size),
	 "(optional) Size of seed window, applies to adaptive windowing.")
	("window-growth", po::value<fnr_sample_count_t>(&growth_size),
	 "(optional) Size of growth window, applies to adaptive windowing.")
	("window-delta", po::value<double>(&max_delta),
	 "(optional) Maximum growth window delta, applies to adaptive windowing.")
	;

    po::options_description opt_desc_invisible("Options");
    opt_desc_invisible.add_options()
#ifdef ENABLE_WORKSPACE_PROFILING
	("dump-workspace",
	 "(optional) Dump workspace contents to a file")
#endif
	("no-intro",
	 "(optional) Deprecated, has no effect.")
	("low-mem",
	 "(optional) Low memory optimizations.")
	("positional-args", "")
	;

    po::options_description opt_desc_all("Options");
    opt_desc_all.add(opt_desc_visible)
	.add(opt_desc_invisible)
	.add(opt_desc_windowing);

    po::positional_options_description pos_desc;
    pos_desc.add("positional-args", -1);

    po::variables_map vm;
    try {
	po::store(po::command_line_parser(argc, argv)
		  .options(opt_desc_all).positional(pos_desc).run(), vm);
	po::notify(vm);
    }
    catch (po::error &e) {
	warning_consumer->error(
            string("Command line parsing error: ") + e.what());
	general_help_hint(warning_consumer);
        RETURN(true, 1);
    }

    if (vm.count("help") != 0) {
	stringstream ss;
	ss << version << opt_desc_visible;
	warning_consumer->none(ss.str());
        RETURN(true, 0);
    }

    if (vm.count("version") != 0) {
	warning_consumer->none(version);
        RETURN(true, 0);
    }

#ifdef ENABLE_WORKSPACE_PROFILING
    dump_workspace = vm.count("dump-workspace") > 0;
#endif

    low_mem = vm.count("low-mem") > 0;

    // ------------------------------------------------------------
    // CPU model parameters
    // ------------------------------------------------------------
    if (vm.count("windowing") > 0) {
	if (windowing_engine == "static")
	    windowing.reset(
		new acumem::analysis::windowing_static_t(samples_per_window));
	else if (windowing_engine == "adaptive")
	    windowing.reset(
		new acumem::analysis::windowing_adaptive_t(
		    seed_size, growth_size, max_delta));
	else if (windowing_engine == "burst")
	    windowing.reset(new acumem::analysis::windowing_burst_t());
	else {
	    warning_consumer->error("Invalid window engine specified.");
	    general_help_hint(warning_consumer);
	    RETURN(true, 1);
	}
	windowing_burst = windowing;
    } else {
	windowing.reset(
	    new acumem::analysis::windowing_adaptive_t(
		seed_size, growth_size, max_delta));
	windowing_burst.reset(
	    new acumem::analysis::windowing_burst_t());
    }

    if (vm.count("cpu") > 0) {
	if (cpu_id == "help") {
	    stringstream ss;
	    print_cpu_usage(ss, cpudb);
	    warning_consumer->none(ss.str());
            RETURN(true, 0);
	}
        else if (cpu_id == "auto" || cpu_id == "auto/auto")
	    cpu = autodetect_cpu(cpudb, warning_consumer, autodetect_failed);
	else {
            pair<cpudb_manufacturer_cptr_t, cpudb_cpu_cptr_t> cpu_info(
                cpudb.get_cpu(cpu_id));
	    cpu_manufacturer = cpu_info.first;
            cpu = cpu_info.second;
	    if (!cpu) {
		warning_consumer->error(
                    "Invalid processor model specified. Use '--cpu help' to "
                    "get a list of valid processor models.");
                RETURN(true, 1);
	    }
	}
    } else
	cpu = autodetect_cpu(cpudb, warning_consumer, autodetect_failed);

    if (cpu && vm.count("level") > 0) {
	cache = cpu->get_data_cache(cache_level);
	if (!cache) {
	    warning_consumer->error(
                "The specified cache level does not exist.");
	    general_help_hint(warning_consumer);
            RETURN(true, 1);
	}
    } else if (cpu && !cpu->caches.empty()) {
	cache = cpu->caches.back();
	cache_level = cache->level;
    } else {
        warning_consumer->error(
            "The specified processor lacks valid cache definitions.");
	general_help_hint(warning_consumer);
        RETURN(true, 1);
    }

    if (vm.count("cache-size") > 0) {
	cache_size = fnr_parse_num_string(cache_size_str.c_str());
	if (cache_size == 0) {
	    warning_consumer->error("Invalid cache size specified.");
	    general_help_hint(warning_consumer);
	    RETURN(true, 1);
        }
    }
    else if (cache)
	cache_size = cache->size;

    if (vm.count("line-size") > 0) {
	if (is_not_power_of_two(line_size)) {
	    stringstream ss;
	    ss << "Invalid line size ("
	       << lexical_cast<string>(line_size)
	       << "). Must be a power of two.";
	    warning_consumer->error(ss.str());
	    general_help_hint(warning_consumer);
	    RETURN(true, 1);
	}
    }
    else if (cache)
	line_size = cache->line_size;

    if (num_cpus == 0) {
	try {
	    num_cpus = get_cpuid(warning_consumer)->num_packages();
	} catch (const cpuid_exception_t &e) {
	    stringstream ss;
	    ss << "Failed to get processor information: " << e.what()
	       << " Assuming 1 physical processor.";
	    warning_consumer->warning(ss.str());
	    num_cpus = 1;
	}
    }

    if (vm.count("number-of-caches") > 0) {
	// If the user overrides the number of caches we are fine
	// again. This is the info missing from the autodetection.
	autodetect_failed = false;
    }

    if (vm.count("number-of-caches") == 0 && cpu && cache) {
	num_caches = num_cpus * cpu->cores / cache->num_cores;
    }

    if (vm.count("group") > 0) {
	tm_cache_id_t c = 0;
	typedef boost::tokenizer<boost::char_separator<char> >
	    tokenizer;
	boost::char_separator<char> sep(",; ");

	BOOST_FOREACH(string g, groups) {
	    tokenizer tokens(g, sep);

	    BOOST_FOREACH(string s, tokens)
		explicit_mapping[lexical_cast<int>(s)] = c;

	    c++;
	}
    }

    if (vm.count("real-thread-id") != 0)
	real_thread_id = true;

    new_algorithm = (vm.count("new-hwpf-algorithm") > 0);

    if (groups.size() > num_caches) {
	stringstream ss;
	ss << "You have specified more thread groups ("
	   << groups.size()
	   << ") than the number of available caches (" << num_caches
	   << ")";
	if (cache && cpu) {
	    ss << " at the selected cache-level (" << cache->name
	       << ") for the specified processor (";

	    if (cpu->id == "auto") {
		bool first = true;
		BOOST_FOREACH(string s, cpu->names) {
		    if (!first) ss << " ";
		    first = false;
		    ss << boost::trim_copy(s);
		}
	    } else {
		ss << cpu->id;
	    }
	    ss << ")";
	    general_help_hint(warning_consumer);
	}
	warning_consumer->error(ss.str());
	RETURN(true, 1);
    }

    if (replacement_policy != "random" && replacement_policy != "lru") {
	warning_consumer->error("Invalid replacement policy specified");
	general_help_hint(warning_consumer);
	RETURN(true, 1);
    }

    if (vm.count("in-file") < 1) {
	warning_consumer->error("No input file specified.");
	general_help_hint(warning_consumer);
        RETURN(true, 1);
    }

    // ------------------------------------------------------------

    if (vm.count("no-barriers") != 0)
	blocking_expert = true;

    if (debug_symbol_level < 0 || debug_symbol_level > 3) {
	warning_consumer->warning("Invalid value for debug-symbol-level. "
                                  "Assuming level 0.");
	debug_symbol_level = Debug_None;
    }

    if (vm.count("positional-args") > 0) {
	warning_consumer->error("No non-option arguments allowed.");
	RETURN(true, 1);
    }

    if (cache_size < 16 * line_size) {
        warning_consumer->error(
            "The selected cache size must be at least 16 times larger than "
            "the selected line size.");
        RETURN(true, 1);
    }

    if (vm.count("verbose") > 0)
        verbose = true;

    if (vm.count("no-intro") > 0)
	warning_consumer->info("The --no-intro option has been deprecated and has no effect.");

    RETURN(false, 0);
}

END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;
