/// @file analysis/report_generation/debug_lookup.hh

// Copyright (c) 2010 Acumem AB.  All rights reserved.

#ifndef DEBUG_LOOKUP_HH
#define DEBUG_LOOKUP_HH

#include "library/sym/sym.hh"

#include <boost/foreach.hpp>
#include <boost/filesystem/path.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace boost::filesystem;
using namespace boost;

// This is direction to the symbol lookup (dwarf) telling what parts of symbol
// extraction to to, for instance whether to do the fairly expensive (for
// binaries with large debug section) recursive function name lookup.
enum debug_symbol_level_t {
    Debug_None,
    Debug_Lines,
    Debug_Pubnames,
    Debug_Full
};

struct file_info_t {
    file_info_t(const string &actual_filename,
                uint64_t address,
		uint64_t size,
		sym_t::debug_info_status_t debug_info_status = sym_t::File_Not_Found)
	: actual_filename(actual_filename),
          address(address),
          size(size),
          debug_info_status(debug_info_status)
	{}

    /// The filename of the binary we actually loaded debug info from.
    string actual_filename;
    sym_paddr_t address;
    uint64_t size;
    sym_t::debug_info_status_t debug_info_status;
};

typedef map<string, file_info_t> file_info_map_t;

typedef map<string, string> string_map_t;

struct debug_lookup_t {
    debug_lookup_t(string cppfilt,
                   bool verbose,
                   feedback_ptr_t feedback,
                   vector<string> source_directories,
                   vector<string> debug_directories,
                   vector<string> extra_binaries,
                   debug_symbol_level_t debug_symbol_level)
        : sym(cppfilt.size() ? sym_t::G_Demangle_Names : 0,
              cppfilt,
              verbose,
              feedback),
          extra_binaries(extra_binaries),
          verbose(verbose),
          debug_symbol_level(debug_symbol_level),
          feedback(feedback)
        {
            BOOST_FOREACH(const string &sd, source_directories)
                sym.add_source_dir(sd);
            BOOST_FOREACH(const string &dd, debug_directories)
                sym.add_debug_link_dir(dd);
#if !defined(_WIN32)
            sym.add_debug_link_dir("/usr/lib/debug");
#endif
        }

    void add_debug_file(const string &sym_filename,
			uint64_t load_address,
			uint64_t size) {
        if (debug_symbol_level == Debug_None)
            return;

        seen_filenames.insert(
            make_pair(sym_filename, file_info_t(sym_filename, load_address, size)));
    }

    void resolve_debug_files() {
        if (debug_symbol_level == Debug_None)
            return;

        int sym_flags;

        switch (debug_symbol_level) {
        case Debug_Full:
            sym_flags = sym_t::Full_Functions | sym_t::Line;
            break;
        case Debug_Pubnames:
            sym_flags = sym_t::Public_Functions | sym_t::Line;
            break;
        case Debug_Lines:
            sym_flags = sym_t::Line;
            break;
        case Debug_None:
            sym_flags = 0;
            break;
        default:
            assert(0);
        }

        BOOST_FOREACH(file_info_map_t::value_type &v, seen_filenames) {
            const string &filename = v.first;
            file_info_t &file_info = v.second;

            log_verbose("Looking for " + filename + ":");

            file_info.debug_info_status =
                sym.add_binary(filename, file_info.address, sym_flags);

            log_debug_file_lookup(filename, file_info.debug_info_status);

            if (file_info.debug_info_status == sym_t::File_Not_Found) {
                // Lookup an alternative matching file in the extra set.
                path p(filename);
                for (vector<string>::iterator
                         i = extra_binaries.begin(),
                         e = extra_binaries.end();
                     i != e; i++) {
                    path p2(*i);
                    if (p.filename() == p2.filename()) {
                        file_info.debug_info_status =
                            sym.add_binary(*i, file_info.address, sym_flags);

                        log_debug_file_lookup(*i, file_info.debug_info_status);

                        if (file_info.debug_info_status != sym_t::File_Not_Found) {
                            file_info.actual_filename = *i;
                            actual_to_nominal.insert(
                                make_pair(file_info.actual_filename, filename));
                            extra_binaries.erase(i);
                            break;
                        }
                    }
                }
            }

            log_debug_file_status(filename, file_info.debug_info_status);
        }

        // Ignore all extra binaries that didn't match any of the seen
        // binaries. We don't have any load address for these, and loading them
        // at their default address may collide with other binaries that were
        // actually loaded at those addresses.
        BOOST_FOREACH(string filename, extra_binaries)
            feedback->log(feedback_base_t::LOG_INFO,
                          "Warning: Ignoring extra binary " + filename + ", since "
                          "it did not match any binary in the sample file.");
    }

    sym_info_cptr_t lookup_pc(uint64_t pc) {
        return sym.lookup_sym_info(sym_paddr_t(pc));
    }


    string_cptr_t lookup_source_file(const string &filename) {
	return sym.lookup_source_file(filename);
    }
    
//private:
    void log_verbose(const string &message) {
        if (verbose)
            feedback->log(feedback_base_t::LOG_INFO, message);
    }


    void log_debug_file_lookup(
        const string &filename,
        sym_t::debug_info_status_t debug_info_status) {
        log_verbose("Tried " + filename +
                    (debug_info_status == sym_t::File_Not_Found ?
                     ", not found. " :
                     ", found!"));
    }

    void log_debug_file_status(
        const string &filename,
        sym_t::debug_info_status_t debug_info_status) {
        string space(max(0, (int)(50 - filename.size())), '.');
        string message = filename + ' ' + space + ' ' +
            sym_t::describe_debug_info_status(debug_info_status);
        feedback->log(feedback_base_t::LOG_INFO, message);
    }

    sym_t sym;

    vector<string> extra_binaries;

    string_map_t actual_to_nominal;

    bool verbose;

    int debug_symbol_level;

    file_info_map_t seen_filenames;

    feedback_ptr_t feedback;
};

typedef shared_ptr<debug_lookup_t> debug_lookup_ptr_t;
typedef shared_ptr<const debug_lookup_t> debug_lookup_cptr_t;

END_NAMESPACE;
END_NAMESPACE;

#endif /* DEBUG_LOOKUP_HH */
