/// @file analysis/report_generator/report_generator.cc

// Copyright (c) 2011 Acumem AB.  All rights reserved.

#include "report_generator.hh"
#include "conf.hh"

#include "analysis/advice/advice.hh"
#include "analysis/issues/issue.hh"
#include "analysis/fetch_ratio/repl_fr_random.hh"
#include "analysis/fetch_ratio/repl_fr_lru.hh"
#include "analysis/fetch_ratio/repl_fr.hh"
#include "analysis/exe_maps/exe_maps.hh"
#include "analysis/locationx/locationx.hh"
#include "analysis/locationa/locationa.hh"
#include "analysis/thread_mapping/thread_mapping_null.hh"
#include "analysis/thread_mapping/thread_mapping_rr.hh"
#include "analysis/thread_mapping/thread_mapping_legacy.hh"
#include "analysis/thread_mapping/cpu_model.hh"
#include "analysis/filter/filter_module.hh"
#include "library/lang/am_assert.hh"

#include <boost/uuid/uuid_generators.hpp>
#include <boost/uuid/uuid_io.hpp>

#include <fstream>

#if defined(_WIN32)
#include <windows.h>
#include <tchar.h>
#endif

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);
BEGIN_NAMESPACE(report_generator);

using namespace acumem::analysis::issues;


typedef map<location_set_t, drill_down_result_ptr_t> location_set_drill_down_t;

struct drill_down_info_t {
    location_set_drill_down_t drill_downs;
    set<depgroup_id_t> depgroup_ids;
    set<loop_id_t> loop_ids;
};

typedef shared_ptr<drill_down_info_t> drill_down_info_ptr_t;
typedef shared_ptr<const drill_down_info_t> drill_down_info_cptr_t;

typedef map<rpt_location_set_t, sliced_stats_cptr_t> location_set_stat_map_t;

static string
canonicalize_file_name(const string &f)
{
#ifdef _WIN32
    wstring ws;
    ws.assign(f.begin(), f.end());
    DWORD len = GetFullPathName(ws.c_str(), 0, 0, 0);
    TCHAR *buf = new TCHAR[len];
    GetFullPathName(ws.c_str(), len, buf, 0);
    string result(buf, buf + len);
    delete [] buf;
    return result;
#else
    char buf[PATH_MAX];

    if (0 == realpath(f.c_str(), buf))
	return f;

    return buf;
#endif
}

static void
print_model_info(feature_context_ptr_t context,
                 bool feature_sw_prefetch,
                 bool feature_mt_advice,
                 const thread_mapping_cptr_t &tm,
                 const cpu_model_t &cpu_model)
{
    thread_mapping_result_cptr_t tm_res(tm->get());

    if (feature_sw_prefetch &&
	!(cpu_model.active_pf & (~FNR_Prefetch_Class_Unknown)))
	context->warning_consumer->info(
            "No software prefetch instructions are active on this cache "
            "level. No software prefetch advice will be generated.");

    if (feature_mt_advice && tm_res->num_caches < 2)
	context->warning_consumer->info(
            "Only one data cache available at this level. Do not expect any "
            "communication related advice.");
}

static void
drill_down_location_set(
    drill_down_info_ptr_t &drill_down_info,
    const location_set_t &location_set)
{
    drill_down_info->drill_downs.insert(
        make_pair(location_set, drill_down_result_ptr_t()));
}

static void
drill_down_depgroup(
    drill_down_info_ptr_t &drill_down_info,
    advice_result_cptr_t advice_result,
    const depgroup_id_t depgroup_id)
{
    drill_down_info->depgroup_ids.insert(depgroup_id);

    drill_down_location_set(
        drill_down_info,
        advice_result->program_structure->depgroups[depgroup_id].pc_set());
}

static void
drill_down_loop(
    drill_down_info_ptr_t &drill_down_info,
    advice_result_cptr_t advice_result,
    const loop_id_t loop_id)
{
    // Loop 0 is the dummy loop containing all instructions that aren't part of
    // any real loop. We can't do a drill-down of that "loop".
    if (loop_id == 0)
        return;

    drill_down_info->loop_ids.insert(loop_id);

    drill_down_location_set(
        drill_down_info,
        advice_result->program_structure->loops[loop_id - 1].loop()->get_locations(false));
}

static void
do_drill_down(
    feature_context_ptr_t context,
    drill_down_result_ptr_t &global_drill_down_info,
    drill_down_info_ptr_t &drill_down_info,
    advisor_cptr_t advisor,
    configuration_cptr_t configuration,
    advice_result_cptr_t advice_result)
{
    // TODO: This is extremely ugly and error prone. We must predict what
    //       loops, dependency groups and other location sets the HTML
    //       generator will require drill-down information for.

    drill_down_info.reset(new drill_down_info_t());

    BOOST_FOREACH(const issues::hot_spot_issue_cptr_t &issue,
		  advice_result->hot_spot_issues) {
        drill_down_loop(drill_down_info, advice_result, issue->loop_id);
	drill_down_depgroup(drill_down_info, advice_result, issue->depgroup_id);
    }

    BOOST_FOREACH(const issues::loop_order_issue_cptr_t &issue,
		  advice_result->loop_order_issues) {
        drill_down_loop(drill_down_info, advice_result, issue->loop_id);
        drill_down_depgroup(drill_down_info, advice_result, issue->depgroup_id);
    }

    BOOST_FOREACH(const issues::utilization_issue_cptr_t &issue,
		  advice_result->utilization_issues) {
	drill_down_loop(drill_down_info, advice_result, issue->loop_id);
	drill_down_depgroup(drill_down_info, advice_result, issue->depgroup_id);
    }

    BOOST_FOREACH(const issues::random_access_issue_cptr_t &issue,
		  advice_result->random_access_issues) {
	drill_down_loop(drill_down_info, advice_result, issue->loop_id);
	drill_down_depgroup(drill_down_info, advice_result, issue->depgroup_id);
    }

    BOOST_FOREACH(const issues::block_issue_cptr_t &issue,
		  advice_result->block_issues) {
	drill_down_loop(drill_down_info, advice_result, issue->loop_id);
	drill_down_depgroup(drill_down_info, advice_result, issue->depgroup_id);
    }

    BOOST_FOREACH(const issues::loop_fusion_issue_cptr_t &issue,
		  advice_result->loop_fusion_issues) {
        drill_down_loop(drill_down_info, advice_result, issue->loop1_id);
        drill_down_loop(drill_down_info, advice_result, issue->loop2_id);
        drill_down_depgroup(drill_down_info, advice_result, issue->depgroup1_id);
    }

    BOOST_FOREACH(const issues::prefetch_issue_cptr_t &issue,
		  advice_result->prefetch_issues)
        // Prefetch issues don't directly require a drill-down of the loop, but
        // will cause the loop to be added to the loop list, which requires a
        // drill-down.
	drill_down_loop(drill_down_info, advice_result, issue->loop_id);

    BOOST_FOREACH(const issues::comm_spot_issue_cptr_t &issue,
                  advice_result->comm_spot_issues) {
        drill_down_loop(drill_down_info, advice_result, issue->loop->id);
	drill_down_depgroup(drill_down_info, advice_result, issue->depgroup_id);
    }

    BOOST_FOREACH(const issues::comm_utilization_issue_cptr_t &issue,
                  advice_result->comm_utilization_issues) {
        drill_down_loop(drill_down_info, advice_result, issue->loop->id);
	drill_down_depgroup(drill_down_info, advice_result, issue->depgroup_id);
    }

    BOOST_FOREACH(const issues::fs_issue_cptr_t &issue,
                  advice_result->fs_issues) {
        drill_down_loop(drill_down_info, advice_result, issue->loop->id);
	drill_down_depgroup(drill_down_info, advice_result, issue->depgroup_id);
    }

    BOOST_FOREACH(const issues::wnt_possible_issue_cptr_t &issue,
		  advice_result->wnt_possible_issues) {
        drill_down_loop(drill_down_info, advice_result, issue->loop->id);
	drill_down_depgroup(drill_down_info, advice_result, issue->depgroup_id);
    }

    BOOST_FOREACH(const issues::prefetchnta_issue_cptr_t &issue,
		  advice_result->prefetchnta_issues) {
        drill_down_loop(drill_down_info, advice_result, issue->loop->id);
	drill_down_depgroup(drill_down_info, advice_result, issue->depgroup_id);
    }

    // Do the actual drill-down of the collected location sets.
    drill_down_query_list_t query_list;
    query_list.push_back(drill_down_query_t(global_drill_down_info, NULL));
    BOOST_FOREACH(location_set_drill_down_t::value_type &v, drill_down_info->drill_downs)
        query_list.push_back(drill_down_query_t(v.second, &v.first));
    advisor->drill_down(*configuration, context, query_list);
}

static void
collect_location_set_stats(
    location_set_stat_map_t &location_set_stat_map,
    const location_set_t &location_set,
    const sliced_stats_cptr_t stats)
{
    location_set_stat_map_t::const_iterator i =
        location_set_stat_map.find(location_set);
    if (i == location_set_stat_map.end())
        location_set_stat_map.insert(make_pair(location_set, stats));
    else
        assert(*i->second == *stats);
}

static void
collect_depgroup_stats(
    location_set_stat_map_t &location_set_stat_map,
    advice_result_cptr_t advice_result,
    const depgroup_id_t depgroup_id)
{
    const prog_depgroup_t &depgroup =
        advice_result->program_structure->depgroups[depgroup_id];
    collect_location_set_stats(location_set_stat_map, depgroup.pc_set(), depgroup.stats());
}

static void
collect_loop_stats(
    location_set_stat_map_t &location_set_stat_map,
    advice_result_cptr_t advice_result,
    const loop_id_t loop_id)
{
    const prog_loop_t &loop =
        advice_result->program_structure->loops[loop_id - 1];
    collect_location_set_stats(location_set_stat_map, loop.loop()->get_locations(true), loop.stats());
    collect_location_set_stats(location_set_stat_map, loop.loop()->get_locations(false), loop.stats());
}

static void
collect_stats(location_set_stat_map_t &location_set_stat_map,
              advice_result_cptr_t advice_result)
{
    // Loop 0 is the dummy loop containing all instructions that aren't part of
    // any real loop. There is no entry for it in the loops vector, so it is
    // indexed by loop id - 1.
    for (rpt_loop_id_t loop_id = 1;
         loop_id < (rpt_loop_id_t)advice_result->program_structure->loops.size() + 1;
         loop_id++)
        collect_loop_stats(location_set_stat_map, advice_result, loop_id);

    for (rpt_depgroup_id_t depgroup_id = 0;
         depgroup_id < (rpt_depgroup_id_t)advice_result->program_structure->depgroups.size();
         depgroup_id++)
        collect_depgroup_stats(location_set_stat_map, advice_result, depgroup_id);
}

static rpt_stat_fp_valid_t
convert_value_valid(const result_valid_t &rv)
{
    return rpt_stat_fp_valid_t(rv.value(), rv.valid());
}

static rpt_stats_t
convert_stats(
    const stats_t &stats)
{
    return rpt_stats_t(
        stats.all_ratio.accesses(),
        stats.all_ratio.cold_fetches(),
        stats.all_ratio.replacement_fetches(),
        stats.all_ratio.coherence_fetches(),
        stats.all_ratio.flush_fetches(),
        stats.stall_ratio.cold_fetches(),
        stats.stall_ratio.replacement_fetches(),
        stats.stall_ratio.coherence_fetches(),
        stats.stall_ratio.flush_fetches(),
        stats.all_ratio.upgrades(),
        stats.all_ratio.cold_write_backs(),
        stats.all_ratio.replacement_write_backs(),
        stats.all_ratio.coherence_write_backs(),
        stats.all_ratio.flush_write_backs(),
        convert_value_valid(stats.spatial_util),
        convert_value_valid(stats.fetch_util),
        convert_value_valid(stats.write_back_util),
        convert_value_valid(stats.comm_util),
        stats.comm_lines,
        stats.hw_prefetch_ratio,
        stats.random_ratio,
        convert_value_valid(stats.fs_ratio),
        stats.util_corr_fr,
        stats.util_corr_wbr,
        stats.avoidable_fetches);
}

static rpt_stats_id_t
lookup_rpt_stats(
    const rpt_report_t &report,
    const rpt_stats_t &stats)
{
    rpt_stats_vector_t::const_iterator i =
        lower_bound(report.stats.begin(), report.stats.end(), stats);
    assert(i != report.stats.end());
    assert(*i == stats);
    return (rpt_stats_id_t)(i - report.stats.begin());
}

static rpt_sliced_stats_t
convert_sliced_stats(
    const rpt_report_t &report,
    const sliced_stats_t &sliced_stats)
{
    rpt_sliced_stats_t rpt_sliced_stats;

    rpt_sliced_stats.global_stats =
        lookup_rpt_stats(report, convert_stats(*sliced_stats.stats));
    BOOST_FOREACH(const thread_stats_t::value_type &v,
                  sliced_stats.thread_stats) {
        rpt_sliced_stats.thread_stats[v.first.id] =
            lookup_rpt_stats(report, convert_stats(*v.second));
    }

    return rpt_sliced_stats;
}

static void
collect_rpt_stats(
    rpt_stats_set_t &rpt_stats_set,
    const sliced_stats_t &sliced_stats)
{
    rpt_stats_set.insert(convert_stats(*sliced_stats.stats));
    BOOST_FOREACH(const thread_stats_t::value_type &v, sliced_stats.thread_stats)
        rpt_stats_set.insert(convert_stats(*v.second));
}

static void
collect_rpt_sliced_stats(
    rpt_sliced_stats_set_t &rpt_sliced_stats_set,
    const rpt_report_t &report,
    const sliced_stats_t &sliced_stats)
{
    rpt_sliced_stats_set.insert(convert_sliced_stats(report, sliced_stats));
}

static rpt_sliced_stats_id_t
lookup_sliced_stats(
    const rpt_report_t &report,
    const sliced_stats_t &sliced_stats)
{
    rpt_sliced_stats_t rpt_sliced_stats(
        convert_sliced_stats(report, sliced_stats));
    rpt_sliced_stats_vector_t::const_iterator i = lower_bound(
        report.sliced_stats.begin(), report.sliced_stats.end(),
        rpt_sliced_stats);
    assert(i != report.sliced_stats.end());
    assert(*i == rpt_sliced_stats);
    return (rpt_sliced_stats_id_t)(i - report.sliced_stats.begin());
}

static rpt_sweep_t
convert_sweep(
    drill_down_result_cptr_t drill_down)
{
    rpt_sweep_t sweep;

    BOOST_FOREACH(const cache_sweep_index_map_t::value_type &v,
                  drill_down->cache_range)
        sweep.push_back(v.first);

    return sweep;
}

static void
collect_sweep(
    rpt_report_t &report,
    drill_down_result_cptr_t drill_down)
{
    rpt_sweep_t sweep(convert_sweep(drill_down));

    /// TODO: For now we assume there will only be one sweep.
    if (report.sweeps.empty())
        report.sweeps.push_back(sweep);
    else
        assert(report.sweeps.size() == 1 &&
               report.sweeps.front() == sweep);
}

static rpt_sweep_id_t
lookup_sweep(const rpt_report_t &report,
             const rpt_sweep_t &sweep)
{
    rpt_sweep_vector_t::const_iterator i =
        lower_bound(report.sweeps.begin(), report.sweeps.end(), sweep);
    assert(i != report.sweeps.end());
    assert(*i == sweep);
    return (rpt_sweep_id_t)(i - report.sweeps.begin());
}

static void
collect_sweep_rpt_stats(
    rpt_stats_set_t &rpt_stats_set,
    drill_down_result_cptr_t drill_down)
{
    BOOST_FOREACH(const sliced_stats_cptr_t &sliced_stats, drill_down->stats)
        collect_rpt_stats(rpt_stats_set, *sliced_stats);
}

static void
collect_sweep_rpt_sliced_stats(
    rpt_sliced_stats_set_t &rpt_sliced_stats_set,
    const rpt_report_t &report,
    drill_down_result_cptr_t drill_down)
{
    BOOST_FOREACH(const sliced_stats_cptr_t &sliced_stats, drill_down->stats)
        collect_rpt_sliced_stats(rpt_sliced_stats_set, report, *sliced_stats);
}

static rpt_stats_sweep_t
convert_stats_sweep(
    const rpt_report_t &report,
    const drill_down_result_cptr_t &drill_down)
{
    rpt_stats_sweep_t stats_sweep;

    stats_sweep.sweep_id =
        lookup_sweep(report, convert_sweep(drill_down));
    BOOST_FOREACH(const sliced_stats_cptr_t &sliced_stats, drill_down->stats)
        stats_sweep.sliced_stats_ids.push_back(
            lookup_sliced_stats(
                report, *sliced_stats));

    return stats_sweep;
}

static void
collect_stats_sweeps(
    rpt_stats_sweep_set_t &rpt_stats_sweep_set,
    const rpt_report_t &report,
    const drill_down_result_cptr_t &drill_down)
{
    rpt_stats_sweep_set.insert(convert_stats_sweep(report, drill_down));
}

static rpt_stats_sweep_id_t
lookup_stats_sweep(
    const rpt_report_t &report,
    const drill_down_result_cptr_t &drill_down)
{
    rpt_stats_sweep_t stats_sweep(
        convert_stats_sweep(report, drill_down));
    rpt_stats_sweep_vector_t::const_iterator i = lower_bound(
        report.stats_sweeps.begin(), report.stats_sweeps.end(), stats_sweep);
    assert(i != report.stats_sweeps.end());
    assert(*i == stats_sweep);
    return (rpt_stats_sweep_id_t)(i - report.stats_sweeps.begin());
}

static void
collect_debug_pc(
    rpt_debug_pc_set_t &debug_pcs,
    rpt_address_id_t pc_id,
    bool return_address)
{
    debug_pcs.insert(rpt_debug_pc_t(pc_id, return_address));
}

static void
collect_location_debug_pcs(
    rpt_debug_pc_set_t &debug_pcs,
    const rpt_report_t &report,
    rpt_location_id_t location_id)
{
    collect_debug_pc(
        debug_pcs,
        report.get_location_pc_id(location_id),
        false);
    BOOST_FOREACH(rpt_address_id_t pc_id,
                  report.get_location_trunc_stack(location_id)) {
        assert(pc_id < report.addresses.size());
        collect_debug_pc(debug_pcs, pc_id, true);
    }


    BOOST_FOREACH(const rpt_full_stack_frame_t &frame,
                  report.get_location_full_stack(location_id)) {
        assert(frame.pc_id < report.addresses.size());
        collect_debug_pc(debug_pcs, frame.pc_id, true);
    }
}

static void
collect_location_set_debug_pcs(
    rpt_debug_pc_set_t &debug_pcs,
    const rpt_report_t &report,
    rpt_location_set_id_t location_set_id)
{
    BOOST_FOREACH(rpt_location_id_t location_id,
                  report.get_location_set(location_set_id))
        collect_location_debug_pcs(
            debug_pcs, report, location_id);
}

static void
collect_depgroup_debug_pcs(
    rpt_debug_pc_set_t &debug_pcs,
    const rpt_report_t &report,
    rpt_depgroup_id_t depgroup_id)
{
    collect_location_set_debug_pcs(
        debug_pcs,
        report,
        report.get_depgroup_location_set_id(depgroup_id));
}

static void
collect_loop_incl_debug_pcs(
    rpt_debug_pc_set_t &debug_pcs,
    const rpt_report_t &report,
    rpt_loop_id_t loop_id)
{
    if (loop_id != 0)
        collect_location_set_debug_pcs(
            debug_pcs,
            report,
            report.get_loop_incl_location_set_id(loop_id));
}

static void
collect_loop_excl_debug_pcs(
    rpt_debug_pc_set_t &debug_pcs,
    const rpt_report_t &report,
    rpt_loop_id_t loop_id)
{
    if (loop_id != 0)
        collect_location_set_debug_pcs(
            debug_pcs,
            report,
            report.get_loop_excl_location_set_id(loop_id));
}

static void
collect_debug_pcs(
    rpt_debug_pc_set_t &debug_pcs,
    bool feature_mt_advice,
    const rpt_report_t &report,
    advice_result_cptr_t advice_result)
{
    BOOST_FOREACH(const issues::hot_spot_issue_cptr_t &issue,
		  advice_result->hot_spot_issues) {
        collect_depgroup_debug_pcs(
            debug_pcs, report, issue->depgroup_id);
        collect_loop_incl_debug_pcs(
            debug_pcs, report, issue->loop_id);
        BOOST_FOREACH(rpt_location_id_t location_id,
                      issue->previous_writers)
            collect_location_debug_pcs(
                debug_pcs, report, location_id);
    }

    BOOST_FOREACH(const issues::loop_order_issue_cptr_t &issue,
		  advice_result->loop_order_issues) {
        collect_depgroup_debug_pcs(
            debug_pcs, report, issue->depgroup_id);
        collect_loop_incl_debug_pcs(
            debug_pcs, report, issue->loop_id);
        BOOST_FOREACH(rpt_location_id_t location_id,
                      issue->previous_writers)
            collect_location_debug_pcs(
                debug_pcs, report, location_id);
    }

    BOOST_FOREACH(const issues::utilization_issue_cptr_t &issue,
		  advice_result->utilization_issues) {
        collect_depgroup_debug_pcs(
            debug_pcs, report, issue->depgroup_id);
        collect_loop_incl_debug_pcs(
            debug_pcs, report, issue->loop_id);
        BOOST_FOREACH(rpt_location_id_t location_id,
                      issue->previous_writers)
            collect_location_debug_pcs(
                debug_pcs, report, location_id);
    }

    BOOST_FOREACH(const issues::random_access_issue_cptr_t &issue,
		  advice_result->random_access_issues) {
        collect_depgroup_debug_pcs(
            debug_pcs, report, issue->depgroup_id);
        collect_loop_incl_debug_pcs(
            debug_pcs, report, issue->loop_id);
        BOOST_FOREACH(rpt_location_id_t location_id,
                      issue->previous_writers)
            collect_location_debug_pcs(
                debug_pcs, report, location_id);
    }

    BOOST_FOREACH(const issues::block_issue_cptr_t &issue,
		  advice_result->block_issues) {
        collect_depgroup_debug_pcs(
            debug_pcs, report, issue->depgroup_id);
        collect_loop_incl_debug_pcs(
            debug_pcs, report, issue->loop_id);
        collect_loop_excl_debug_pcs(
            debug_pcs, report, issue->outer_loop_id);
        collect_loop_excl_debug_pcs(
            debug_pcs, report, issue->next_outer_loop_id);
        BOOST_FOREACH(const location_set_map_t::value_type &v,
                      issue->barriers) {
            collect_location_debug_pcs(
                debug_pcs, report, v.first);
            BOOST_FOREACH(rpt_location_id_t location_id, v.second)
                collect_location_debug_pcs(
                    debug_pcs, report, location_id);
        }
        collect_loop_excl_debug_pcs(
            debug_pcs, report, issue->target_loop_id);
        BOOST_FOREACH(rpt_location_id_t location_id,
                      issue->previous_writers)
            collect_location_debug_pcs(
                debug_pcs, report, location_id);
        BOOST_FOREACH(rpt_location_id_t location_id,
                      issue->next_writers)
            collect_location_debug_pcs(
                debug_pcs, report, location_id);
    }

    BOOST_FOREACH(const issues::loop_fusion_issue_cptr_t &issue,
		  advice_result->loop_fusion_issues) {
        collect_depgroup_debug_pcs(
            debug_pcs, report, issue->depgroup1_id);
        collect_depgroup_debug_pcs(
            debug_pcs, report, issue->depgroup2_id);
        collect_loop_incl_debug_pcs(
            debug_pcs, report, issue->loop1_id);
        collect_loop_incl_debug_pcs(
            debug_pcs, report, issue->loop2_id);
        BOOST_FOREACH(rpt_location_id_t location_id,
                      issue->loop1_pcs)
            collect_location_debug_pcs(
                debug_pcs, report, location_id);
        BOOST_FOREACH(rpt_location_id_t location_id,
                      issue->loop2_pcs)
            collect_location_debug_pcs(
                debug_pcs, report, location_id);
        BOOST_FOREACH(const location_set_map_t::value_type &v,
                      issue->block_for_moving_loop1_down) {
            collect_location_debug_pcs(
                debug_pcs, report, v.first);
            BOOST_FOREACH(rpt_location_id_t location_id, v.second)
                collect_location_debug_pcs(
                    debug_pcs, report, location_id);
        }
        BOOST_FOREACH(const location_set_map_t::value_type &v,
                      issue->block_for_moving_loop2_up) {
            collect_location_debug_pcs(
                debug_pcs, report, v.first);
            BOOST_FOREACH(rpt_location_id_t location_id, v.second)
                collect_location_debug_pcs(
                    debug_pcs, report, location_id);
        }
        BOOST_FOREACH(rpt_location_id_t location_id,
                      issue->additional_pc1)
            collect_location_debug_pcs(
                debug_pcs, report, location_id);
        BOOST_FOREACH(rpt_location_id_t location_id,
                      issue->additional_pc2)
            collect_location_debug_pcs(
                debug_pcs, report, location_id);
    }

    BOOST_FOREACH(const issues::prefetch_issue_cptr_t &issue,
		  advice_result->prefetch_issues) {
        collect_location_debug_pcs(
            debug_pcs, report, issue->pc);
        collect_depgroup_debug_pcs(
            debug_pcs, report, issue->depgroup_id);
        collect_loop_incl_debug_pcs(
            debug_pcs, report, issue->loop_id);
        BOOST_FOREACH(rpt_location_id_t location_id,
                      issue->prev_locations)
            collect_location_debug_pcs(
                debug_pcs, report, location_id);
        BOOST_FOREACH(rpt_location_id_t location_id,
                      issue->next_locations)
            collect_location_debug_pcs(
                debug_pcs, report, location_id);
    }

    if (feature_mt_advice) {
	BOOST_FOREACH(const issues::comm_spot_issue_cptr_t &issue,
		      advice_result->comm_spot_issues) {
            collect_depgroup_debug_pcs(
                debug_pcs, report, issue->depgroup_id);
            collect_loop_incl_debug_pcs(
                debug_pcs, report, issue->loop->id);
            BOOST_FOREACH(rpt_location_id_t location_id,
                          issue->previous_writers)
                collect_location_debug_pcs(
                    debug_pcs, report, location_id);
            BOOST_FOREACH(rpt_location_id_t location_id,
                          issue->next_writers)
                collect_location_debug_pcs(
                    debug_pcs, report, location_id);
            BOOST_FOREACH(rpt_location_id_t location_id,
                          issue->producers)
                collect_location_debug_pcs(
                    debug_pcs, report, location_id);
        }

	BOOST_FOREACH(const issues::comm_utilization_issue_cptr_t &issue,
		      advice_result->comm_utilization_issues) {
            collect_depgroup_debug_pcs(
                debug_pcs, report, issue->depgroup_id);
            collect_loop_incl_debug_pcs(
                debug_pcs, report, issue->loop->id);
            BOOST_FOREACH(rpt_location_id_t location_id,
                          issue->previous_writers)
                collect_location_debug_pcs(
                    debug_pcs, report, location_id);
            BOOST_FOREACH(rpt_location_id_t location_id,
                          issue->next_writers)
                collect_location_debug_pcs(
                    debug_pcs, report, location_id);
            BOOST_FOREACH(rpt_location_id_t location_id,
                          issue->producers)
                collect_location_debug_pcs(
                    debug_pcs, report, location_id);
        }

	BOOST_FOREACH(const issues::fs_issue_cptr_t &issue,
		      advice_result->fs_issues) {
            collect_depgroup_debug_pcs(
                debug_pcs, report, issue->depgroup_id);
            collect_loop_incl_debug_pcs(
                debug_pcs, report, issue->loop->id);
            BOOST_FOREACH(rpt_location_id_t location_id,
                          issue->offenders)
                collect_location_debug_pcs(
                    debug_pcs, report, location_id);
        }
    }

    BOOST_FOREACH(const issues::wnt_possible_issue_cptr_t &issue,
		  advice_result->wnt_possible_issues) {
        collect_depgroup_debug_pcs(
            debug_pcs, report, issue->depgroup_id);
        collect_loop_incl_debug_pcs(
            debug_pcs, report, issue->loop->id);
        BOOST_FOREACH(rpt_location_id_t location_id,
                      issue->next_readers)
            collect_location_debug_pcs(
                debug_pcs, report, location_id);
    }

    BOOST_FOREACH(const issues::prefetchnta_issue_cptr_t &issue,
		  advice_result->prefetchnta_issues) {
        collect_depgroup_debug_pcs(
            debug_pcs, report, issue->depgroup_id);
        collect_loop_incl_debug_pcs(
            debug_pcs, report, issue->loop->id);
        BOOST_FOREACH(location_fraction_t lf,
                      issue->pc1_fractions)
            collect_location_debug_pcs(
                debug_pcs, report, lf.location);
        BOOST_FOREACH(location_fraction_t lf,
                      issue->pc2_fractions)
            collect_location_debug_pcs(
                debug_pcs, report, lf.location);
    }
}

static rpt_binary_t
convert_binary(const sym_info_t &sym_info,
               rpt_address_t pc,
               debug_lookup_ptr_t debug_lookup)
{
    string filename(*sym_info.binary_file);
    string_map_t::const_iterator si =
        debug_lookup->actual_to_nominal.find(filename);
    if (si != debug_lookup->actual_to_nominal.end())
        filename = si->second;
    file_info_map_t::const_iterator i =
        debug_lookup->seen_filenames.find(filename);
    AM_ASSERT(i != debug_lookup->seen_filenames.end());
    const file_info_t &file_info(i->second);
    return rpt_binary_t(file_info.actual_filename,
                        file_info.address.paddr,
                        file_info.size,
                        (rpt_debug_info_status_t)file_info.debug_info_status);
}

static rpt_symbol_t
convert_symbol(const sym_info_t &sym_info, rpt_address_t pc) {
    return rpt_symbol_t(*sym_info.symbol_name, pc - sym_info.symbol_offset);
}

static rpt_function_t
convert_function(const sym_info_t &sym_info, rpt_address_t pc) {
    return rpt_function_t(*sym_info.function_name, pc - sym_info.function_offset);
}

static rpt_source_file_t
convert_source_file(const sym_info_t &sym_info) {
    return rpt_source_file_t(*sym_info.source_file);
}

static rpt_source_line_t
convert_source_line(const rpt_report_t &report, const sym_info_t &sym_info) {
    // Assuming that unknown source lines (-1) have been filtered out.
    assert(sym_info.source_line != -1);

    // Assuming that source lines are representable by uint32_t,
    // although they are represented by an int64_t
    assert(sym_info.source_line < (1LL<<32));

    return rpt_source_line_t(
        report.get_source_file_id(convert_source_file(sym_info)),
        (uint32_t) sym_info.source_line);
}

static void
convert_issues(rpt_report_t &report,
               advice_result_cptr_t advice_result,
               bool mt_advice)
{
    BOOST_FOREACH(const issues::hot_spot_issue_cptr_t &issue,
		  advice_result->hot_spot_issues) {
        rpt_issue_ptr_t rpt_issue(
            new rpt_hot_spot_issue_t(
                (rpt_issue_id_t)report.issues.size(),
                (rpt_hot_spot_issue_t::type_t)issue->type,
                issue->loop_id,
                issue->depgroup_id,
                issue->previous_writers));
        report.issues.push_back(rpt_issue);
    }

    BOOST_FOREACH(const issues::prefetch_issue_cptr_t &issue,
		  advice_result->prefetch_issues) {
        rpt_issue_ptr_t rpt_issue(
            new rpt_prefetch_issue_t(
                (rpt_issue_id_t)report.issues.size(),
                (rpt_prefetch_issue_t::problem_t)issue->problem,
                issue->pc,
                issue->loop_id,
                issue->depgroup_id,
                (rpt_stat_fp_t)issue->cost,
                (rpt_stat_fp_t)issue->fetch_ratio,
                issue->median_rdist,
                issue->prev_locations,
                issue->next_locations));
        report.issues.push_back(rpt_issue);
    }

    BOOST_FOREACH(const issues::utilization_issue_cptr_t &issue,
		  advice_result->utilization_issues) {
        rpt_issue_ptr_t rpt_issue(
            new rpt_utilization_issue_t(
                (rpt_issue_id_t)report.issues.size(),
                (rpt_utilization_issue_t::type_t)issue->type,
                issue->loop_id,
                issue->depgroup_id,
                issue->previous_writers));
        report.issues.push_back(rpt_issue);
    }

    BOOST_FOREACH(const issues::loop_order_issue_cptr_t &issue,
		  advice_result->loop_order_issues) {
        rpt_issue_ptr_t rpt_issue(
            new rpt_loop_order_issue_t(
                (rpt_issue_id_t)report.issues.size(),
                issue->loop_id,
                issue->depgroup_id,
                issue->previous_writers));
        report.issues.push_back(rpt_issue);
    }

    BOOST_FOREACH(const issues::random_access_issue_cptr_t &issue,
		  advice_result->random_access_issues) {
        rpt_issue_ptr_t rpt_issue(
            new rpt_random_issue_t(
                (rpt_issue_id_t)report.issues.size(),
                issue->loop_id,
                issue->depgroup_id,
                issue->previous_writers));
        report.issues.push_back(rpt_issue);
    }

    BOOST_FOREACH(const issues::loop_fusion_issue_cptr_t &issue,
		  advice_result->loop_fusion_issues) {
        rpt_location_id_vector_t loop1_down_pcs;
        rpt_location_set_vector_t loop1_down_blocks;
        BOOST_FOREACH(const location_set_map_t::value_type &v,
                      issue->block_for_moving_loop1_down) {
            loop1_down_pcs.push_back(v.first);
            loop1_down_blocks.push_back(v.second);
        }

        rpt_location_id_vector_t loop2_up_pcs;
        rpt_location_set_vector_t loop2_up_blocks;
        BOOST_FOREACH(const location_set_map_t::value_type &v,
                      issue->block_for_moving_loop2_up) {
            loop2_up_pcs.push_back(v.first);
            loop2_up_blocks.push_back(v.second);
        }

        rpt_issue_ptr_t rpt_issue(
            new rpt_loop_fusion_issue_t(
                (rpt_issue_id_t)report.issues.size(),
                issue->loop1_id,
                issue->loop2_id,
                issue->loop1_pcs,
                issue->loop2_pcs,
                issue->depgroup1_id,
                issue->depgroup2_id,
                issue->move_loop2_up,
                issue->move_loop1_down,
                loop1_down_pcs,
                loop1_down_blocks,
                loop2_up_pcs,
                loop2_up_blocks,
                issue->additional_pc1,
                issue->additional_pc2));
        report.issues.push_back(rpt_issue);
    }

    BOOST_FOREACH(const issues::block_issue_cptr_t &issue,
		  advice_result->block_issues) {
        rpt_location_id_vector_t barrier_pcs;
        rpt_location_set_vector_t barrier_blocks;
        BOOST_FOREACH(const location_set_map_t::value_type &v,
                      issue->barriers) {
            barrier_pcs.push_back(v.first);
            barrier_blocks.push_back(v.second);
        }

        rpt_issue_ptr_t rpt_issue(
            new rpt_block_issue_t(
                (rpt_issue_id_t)report.issues.size(),
                issue->block_type_mask,
                issue->loop_id,
                issue->depgroup_id,
                issue->outer_loop_id,
                issue->outer_loop_iterations,
                issue->next_outer_loop_id,
                issue->next_outer_loop_iterations,
                barrier_pcs,
                barrier_blocks,
                issue->target_loop_id,
                issue->target_loop_iterations,
                issue->previous_writers,
                issue->next_writers));
        report.issues.push_back(rpt_issue);
    }

    if (mt_advice) {
        BOOST_FOREACH(const issues::comm_spot_issue_cptr_t &issue,
                      advice_result->comm_spot_issues) {
            rpt_issue_ptr_t rpt_issue(
                new rpt_comm_spot_issue_t(
                    (rpt_issue_id_t)report.issues.size(),
                    issue->loop->id,
                    issue->depgroup_id,
                    (rpt_stat_fp_t)issue->comm_lines,
                    issue->producers,
                    issue->previous_writers,
                    issue->next_writers));
            report.issues.push_back(rpt_issue);
        }

        BOOST_FOREACH(const issues::comm_utilization_issue_cptr_t &issue,
                      advice_result->comm_utilization_issues) {
            rpt_issue_ptr_t rpt_issue(
                new rpt_comm_utilization_issue_t(
                    (rpt_issue_id_t)report.issues.size(),
                    issue->loop->id,
                    issue->depgroup_id,
                    (rpt_stat_fp_t)issue->utilization,
                    (rpt_stat_fp_t)issue->waste,
                    issue->producers,
                    issue->previous_writers,
                    issue->next_writers));
            report.issues.push_back(rpt_issue);
        }

        BOOST_FOREACH(const issues::fs_issue_cptr_t &issue,
                      advice_result->fs_issues) {
            rpt_issue_ptr_t rpt_issue(
                new rpt_false_sharing_issue_t(
                    (rpt_issue_id_t)report.issues.size(),
                    issue->loop->id,
                    issue->depgroup_id,
                    (rpt_stat_fp_t)issue->ratio,
                    (rpt_stat_fp_t)issue->waste,
                    issue->offenders));
            report.issues.push_back(rpt_issue);
        }
    }

    BOOST_FOREACH(const issues::wnt_possible_issue_cptr_t &issue,
                  advice_result->wnt_possible_issues) {
        rpt_issue_ptr_t rpt_issue(
            new rpt_wnt_possible_issue_t(
                (rpt_issue_id_t)report.issues.size(),
                issue->loop->id,
                issue->depgroup_id,
                issue->next_readers));
        report.issues.push_back(rpt_issue);
    }
    BOOST_FOREACH(const issues::prefetchnta_issue_cptr_t &issue,
                  advice_result->prefetchnta_issues) {
        rpt_location_id_vector_t pc1s;
        rpt_stat_fp_vector_t pc1_fractions;
        rpt_stat_fp_vector_t pc1_required_cache_sizes;
        BOOST_FOREACH(const location_fraction_t &lf, issue->pc1_fractions) {
            pc1s.push_back(lf.location);
            pc1_fractions.push_back(
		(rpt_stat_fp_t)lf.fraction);
            pc1_required_cache_sizes.push_back(
		(rpt_stat_fp_t)lf.required_cache_size);
        }

        rpt_location_id_vector_t pc2s;
        rpt_stat_fp_vector_t pc2_fractions;
        BOOST_FOREACH(const location_fraction_t &lf, issue->pc2_fractions) {
            pc2s.push_back(lf.location);
            pc2_fractions.push_back((rpt_stat_fp_t)lf.fraction);
        }

        rpt_issue_ptr_t rpt_issue(
            new rpt_prefetchnta_issue_t(
                (rpt_issue_id_t)report.issues.size(),
                issue->loop->id,
                issue->depgroup_id,
                (rpt_stat_fp_t)issue->accesses,
                (rpt_stat_fp_t)issue->fetches,
                (rpt_stat_fp_t)issue->required_cache_size,
                pc1s,
                pc1_fractions,
                pc1_required_cache_sizes,
                pc2s,
                pc2_fractions));
        report.issues.push_back(rpt_issue);
    }
}

rpt_report_cptr_t
generate_report(
    feature_context_ptr_t context,
    fnr_real_time_t analysis_time,
    const vector<string> &analysis_argv,
    const string &title,
    const string &sample_file,
    const vector<string> &extra_binaries,
    const vector<string> &debug_directories,
    const vector<string> &source_directories,
    debug_symbol_level_t debug_symbol_level,
    const string &cppfilt,
    double threshold_percentage,
    const string &replacement_policy,
    fnr_cache_size_t cache_size,
    fnr_line_size_t line_size,
    unsigned int num_caches,
    cpudb_manufacturer_cptr_t cpu_manufacturer,
    cpudb_cpu_cptr_t cpu,
    unsigned int num_cpus,
    cpudb_cache_cptr_t cache,
    unsigned int cache_level,
    const context_cache_map_t &explicit_mapping,
    bool real_thread_id,
    fnr_stack_depth_t stack_depth,
    windowing_engine_cptr_t windowing_non_burst,
    windowing_engine_cptr_t windowing_burst,
    bool feature_mt_advice,
    bool feature_sw_prefetch,
    bool feature_unlimited_callstack,
    bool feature_blocking_expert,
    bool feature_multisocket,
    bool new_hwpf_algorithm,
    bool low_mem,
    bool verbose,
    bool autodetect_failed)
{
    rpt_report_ptr_t report(new rpt_report_t());

    // Generate uuid.
    boost::uuids::random_generator uuid_generator;
    boost::uuids::uuid serial_number_uuid = uuid_generator();
    stringstream ss;
    ss << serial_number_uuid;
    report->uuid = ss.str();

    cpu_model_t cpu_model(cpu, cache);

    parser_result_cptr_t parser_result =
	parser_t(context, sample_file).get();

    thread_mapping_cptr_t thread_mapping;
    if (parser_result->mt_sampled()) {
	if (explicit_mapping.size() != 0)
	    thread_mapping.reset(
                new thread_mapping_explicit_context_t(
				     context,
				     sample_file,
				     num_caches,
                                     explicit_mapping,
				     real_thread_id));
        else if (num_caches == 0)
	    thread_mapping.reset(new thread_mapping_null_t(context,
							   sample_file));
	else
	    thread_mapping.reset(new thread_mapping_rr_t(context,
							 sample_file,
							 num_caches));
    }
    else
	thread_mapping.reset(new thread_mapping_legacy_t(context,
							 sample_file));

    // Print information about the analysis model, e.g., if software prefetches
    // are active on the selected level.
    print_model_info(context,
                     feature_sw_prefetch,
                     feature_mt_advice,
                     thread_mapping,
                     cpu_model);

    int major_step = context->feedback->add_major_step("Loading sample file and locating symbol information");
    context->feedback->select_major_step(major_step);

    windowing_engine_cptr_t windowing(parser_result->burst_sampled ?
				      windowing_burst :
				      windowing_non_burst);

    repl_fr_engine_cptr_t engine;
    if (replacement_policy == "random")
	engine.reset(
	    new repl_fr_random_t(context,
				 sample_file,
				 stack_depth,
				 thread_mapping,
				 windowing,
				 cpu_model));
    else if (replacement_policy == "lru")
	engine.reset(
	    new repl_fr_lru_t(context,
			      sample_file,
			      stack_depth,
			      thread_mapping,
			      windowing,
			      cpu_model));
    else {
	context->warning_consumer->error("Invalid replacement_policy.");
	return rpt_report_ptr_t();
    }

    locationa_result_cptr_t la_res(
	locationa_mapper_t(context,
			   sample_file,
			   stack_depth,
			   thread_mapping).get());

    locationx_mapper_cptr_t lx_mapper(
	new locationx_mapper_t(context,
			       sample_file,
			       stack_depth,
			       thread_mapping,
			       true, false, false));

    // Calculate upper and lower bound for diagrams.
    fnr_cache_size_t lower = cache_size;
    lower |= lower >> 1;
    lower |= lower >> 2;
    lower |= lower >> 4;
    lower |= lower >> 8;
    lower |= lower >> 16;
    lower |= lower >> 32;
    lower += 1;
    lower /= 16;
    fnr_cache_size_t upper = lower * 2048;

    // Set up basic advisor configuration.

    // Magic threshold 0.0001 indicates the total number of fetches from one
    // instruction divided by total number of accesses.  With a 100 cycle L2
    // fetch penalty, this would mean that we require an issue to contribute at
    // least 1% to the runtime to be displayed.
    configuration_cptr_t configuration(
        new configuration_t(
            sample_file,
            stack_depth,
            cache_size,
            line_size,
            thread_mapping,
            cpu_model,
            engine,
            windowing,
            sweep_t(lower, upper, 4, line_size),
            feature_sw_prefetch,
            feature_blocking_expert,
            feature_mt_advice,
            cache_level == cpu->num_cache_levels(),
            low_mem || la_res->locationas.size() > 10000,
            !new_hwpf_algorithm,
            cache->active_hwpf,
            threshold_percentage / 100,
            0.0001,
            0x8000 /* L1 cache size for prefetchnta analysis. */));

    advisor_cptr_t advisor(new advisor_t);

    advice_result_cptr_t advice_result(
        advisor->process(*configuration, context, Advisor_all));

    // Verify that we have samples for both the analyzed line size and 1-byte
    // line size, and that we have enough samples of the analyzed line size for
    // a reliable result.
    {
	// Check 32/64/... -sized lines
	parser_result_cptr_t line_samples =
	    filter_module_t(context, sample_file,
			    line_size).get();

	if (line_samples->repl_sample_count() == 0) {
	    stringstream s;
	    s << "No samples for the selected line size ("
	      << line_size
	      << " bytes) found in the sample file. Please re-sample "
	      << "the application with that line size enabled.";
	    context->warning_consumer->warning(s.str());
            return rpt_report_ptr_t();
	}

	// Check 1-byte sized lines
	parser_result_cptr_t byte_samples =
	    filter_module_t(context, sample_file, 1).get();

	fnr_sample_count_t byte_sample_count(
	    byte_samples->repl_sample_count());

	if (byte_sample_count == 0) {
	    context->warning_consumer->warning(
		"No samples for 1-byte line size found in the sample "
		"file. Samples for 1-byte line size are needed for "
		"the analysis. Please re-sample the application with "
		"1-byte line size enabled.");
            return rpt_report_ptr_t();
	}

	// Extract information from the sample file
	if (line_samples->process_info->acquisition_begin) {
	    const fnr_event_acquisition_begin_t &b =
		*line_samples->process_info->acquisition_begin;

	    BOOST_FOREACH(const std::string arg, b.argv)
		report->argv.push_back(arg);

	    report->user = b.user;
	    report->begin_time = b.time;
            report->architecture = (rpt_architecture_t)b.architecture;
	}

	if (line_samples->process_info->acquisition_end)
	    report->end_time =
                line_samples->process_info->acquisition_end->time;
    }

    // Perform drill-down.
    drill_down_result_ptr_t global_drill_down_info;
    drill_down_info_ptr_t drill_down_info;
    do_drill_down(context, global_drill_down_info, drill_down_info,
                  advisor, configuration, advice_result);

    // Get per-pc access info stuff.
    access_lookup_result_cptr_t access_lookup =
        access_lookup_t(context, sample_file, stack_depth).get();

    // Do call-stack stuff.
    stack_pc_map_result_cptr_t stack_pc_map =
	stack_pc_map_t(context, sample_file, stack_depth).get();

    // Add sampling info to the report.
    report->mt_sampled = parser_result->mt_sampled();
    report->burst_sampled = parser_result->burst_sampled;
    report->sample_count = advice_result->analysis_info->sample_count;

    // Add analysis info.
    report->sample_file = canonicalize_file_name(sample_file.c_str());
    report->analysis_time = analysis_time;
    report->analysis_argv = analysis_argv;
    report->title = title;

    // Generate a CPU model name.
    string cpu_model_name(cpu->id == "auto" ?
                          cpu->names.front() + " (auto)" :
                          (cpu_manufacturer ?
                           cpu_manufacturer->id + "/" + cpu->id :
                           cpu->id));

    // Look up the highest cache level.
    unsigned int num_cache_levels = cache_level;
    BOOST_FOREACH(const cpudb_cache_cptr_t &cache, cpu->caches)
	if (cache->data && cache->level > num_cache_levels)
	    num_cache_levels = cache->level;

    // Are any software prefetch instructions active?
    bool swpf_active = !!(cpu_model_t(cpu, cache).active_pf &
                          ~FNR_Prefetch_Class_Unknown);

    // Add hardware configuration.
    report->cpu_model = cpu_model_name;
    report->num_cpus = num_cpus;
    report->num_caches = advice_result->analysis_info->num_caches;
    report->num_cache_levels = num_cache_levels;
    report->autodetect_failed = autodetect_failed; // XXX FIXME
    report->cache_level = cache_level;
    report->cache_size = cache_size;
    report->line_size = line_size;
    report->replacement_policy = replacement_policy;
    report->swpf_active = swpf_active;

    // Collect cache to thread mapping for the intermediate format.
    report->cache_threads.resize(report->num_caches);
    BOOST_FOREACH(const thread_mapping_burst_t &b,
                  thread_mapping->get()->bursts) {
        BOOST_FOREACH(const tm_cache_map_t::value_type &v, b.get_cache_map(0))
            report->cache_threads[v.second].insert(v.first.id);
    }

    // Collect all PC addresses for the intermediate format.
    rpt_address_set_t address_set;
    BOOST_FOREACH(const trunc_stack_pc_id_map_t::value_type &v,
                  stack_pc_map->trunc_stack_pc_id_map) {
        address_set.insert(v.second.pc.virtual_address);
        BOOST_FOREACH(fnr_address_t addr, v.second.ret_addrs) {
            address_set.insert(addr.virtual_address);
	}
    }
    BOOST_FOREACH(const full_call_stack_map_t::value_type &v,
                  stack_pc_map->full_stack) {
        vector<const full_stack_node_t *> nodes;
        nodes.push_back(&v.second.node);
        do {
            const full_stack_node_t *node = nodes.back();
            address_set.insert(node->pc.virtual_address);
            nodes.pop_back();

            if (node->children != NULL) {
                BOOST_FOREACH(const full_stack_node_t &child_node,
                              *node->children)
                    nodes.push_back(&child_node);
            }
        } while (!nodes.empty());
    }
    report->addresses.reserve(address_set.size());
    BOOST_FOREACH(rpt_address_t addr, address_set)
        report->addresses.push_back(addr);

    // TODO: Update comment!
    // Convert locations and truncated call stacks to the intermediate format.
    rpt_address_vector_t location_pcs;
    set<rpt_address_id_vector_t> trunc_stack_set;
    typedef map<rpt_location_id_t, rpt_address_id_vector_t> lts_t;
    lts_t location_trunc_stacks;
    BOOST_FOREACH(const trunc_stack_pc_id_map_t::value_type &v,
                  stack_pc_map->trunc_stack_pc_id_map) {
        location_pcs.push_back(v.second.pc.virtual_address);
        rpt_trunc_stack_t trunc_stack;
        BOOST_FOREACH(fnr_address_t addr, v.second.ret_addrs)
            trunc_stack.push_back(
                report->get_address_id(addr.virtual_address));
        trunc_stack_set.insert(trunc_stack);
        location_trunc_stacks[v.first] = trunc_stack;
    }
    report->trunc_stacks.reserve(trunc_stack_set.size());
    BOOST_FOREACH(const rpt_trunc_stack_t &trunc_stack, trunc_stack_set)
        report->trunc_stacks.push_back(trunc_stack);
    report->locations.reserve(location_trunc_stacks.size());
    BOOST_FOREACH(const lts_t::value_type &v, location_trunc_stacks) {
        rpt_trunc_stacks_t::const_iterator i = lower_bound(
            report->trunc_stacks.begin(),
            report->trunc_stacks.end(),
            v.second);
        assert(i != report->trunc_stacks.end());
        assert(*i == v.second);
        report->locations.push_back(
            rpt_location_t(
                report->get_address_id(location_pcs[v.first]),
                (rpt_trunc_stack_id_t)(i - report->trunc_stacks.begin())));
    }

    // Convert locations to the intermediate format. TODO
    BOOST_FOREACH(const trunc_stack_pc_id_map_t::value_type &v,
                  stack_pc_map->trunc_stack_pc_id_map) {
        rpt_trunc_stack_t trunc_stack;
        BOOST_FOREACH(fnr_address_t addr, v.second.ret_addrs)
            trunc_stack.push_back(report->get_address_id(addr.virtual_address));
        trunc_stack_set.insert(trunc_stack);
    }

    // Convert the full call stacks to the intermediate format.
    report->full_stacks.reserve(stack_pc_map->full_stack.size());
    BOOST_FOREACH(const full_call_stack_map_t::value_type &v,
                  stack_pc_map->full_stack) {
        // Traverse the stack from the bottom up and collect the most common
        // call path. The frames are added in reverse order, innermost first.
        float fraction = 1.0;
        rpt_full_stack_t reverse_full_stack;
        const full_stack_node_t *node = &v.second.node;
        while (node != NULL && node->children != NULL) {
            const full_stack_node_t *best_child = NULL;
            int total_count = 0;
            BOOST_FOREACH(const full_stack_node_t &child_node, *node->children) {
                total_count += child_node.count;
                if (best_child == NULL ||
                    child_node.count > best_child->count)
                    best_child = &child_node;
            }
            if (best_child != NULL) {
                fraction = fraction * best_child->count / total_count;
                reverse_full_stack.push_back(
                    rpt_full_stack_frame_t(
			report->get_address_id(best_child->pc.virtual_address),
			fraction));
            }
            node = best_child;
        }

        // Allocate an array of exactly the needed size and add the stack
        // frames in the desired order, outermost first.
        rpt_full_stack_t full_stack;
        full_stack.reserve(reverse_full_stack.size());
        BOOST_REVERSE_FOREACH(const rpt_full_stack_frame_t &frame,
                              reverse_full_stack)
            full_stack.push_back(frame);
        report->full_stacks.push_back(full_stack);
    }

    // Collect access classes for all PCs for the intermediate format.
    report->access_classes.reserve(report->addresses.size());
    BOOST_FOREACH(rpt_address_t addr, report->addresses) {
        pc_access_info_map_t::const_iterator i =
            access_lookup->pc_access_info.find(fnr_address_t(addr));
        if (i == access_lookup->pc_access_info.end())
            // A pc that we never see an access from, e.g., one that we only
            // see in call stacks.
            report->access_classes.push_back(RPT_Access_Class_None);
        else
            report->access_classes.push_back(i->second.access_classes);
    }

    // Collect all location sets and corresponding statistics for in
    // location_set_stat_map.
    location_set_stat_map_t location_set_stat_map;
    collect_stats(location_set_stat_map, advice_result);

    // Collect location sets for the intermediate format.
    report->location_sets.reserve(location_set_stat_map.size());
    BOOST_FOREACH(const location_set_stat_map_t::value_type &v,
                  location_set_stat_map)
        report->location_sets.push_back(v.first);

    // Collect thread info for the intermediate format.
    BOOST_FOREACH(fnr_context_id_t context_id, parser_result->contexts)
        report->context_ids.push_back(context_id);

    // Collect sweeps.
    collect_sweep(*report, global_drill_down_info);
    BOOST_FOREACH(const location_set_drill_down_t::value_type &v,
                  drill_down_info->drill_downs)
        collect_sweep(*report, v.second);

    // ----- Statistics conversion begins. ------------------------------------

    // Phase 1: Collect all rpt_stats_t records.
    {
        rpt_stats_set_t rpt_stats_set;

        collect_rpt_stats(rpt_stats_set, *advice_result->global_statistics);
        BOOST_FOREACH(const prog_location_t &loc,
                      advice_result->program_structure->locations)
            collect_rpt_stats(rpt_stats_set, *loc.stats());
        BOOST_FOREACH(const location_set_stat_map_t::value_type &v,
                      location_set_stat_map)
            collect_rpt_stats(rpt_stats_set, *v.second);

        collect_sweep_rpt_stats(rpt_stats_set, global_drill_down_info);
        BOOST_FOREACH(const location_set_drill_down_t::value_type &v,
                      drill_down_info->drill_downs)
            collect_sweep_rpt_stats(rpt_stats_set, v.second);

        report->stats.reserve(rpt_stats_set.size());
        BOOST_FOREACH(const rpt_stats_t &stats, rpt_stats_set)
            report->stats.push_back(stats);
    }

    // Phase 2: Collect all rpt_sliced_stats_t records.
    {
        rpt_sliced_stats_set_t rpt_sliced_stats_set;

        collect_rpt_sliced_stats(
            rpt_sliced_stats_set, *report, *advice_result->global_statistics);
        BOOST_FOREACH(const prog_location_t &loc,
                      advice_result->program_structure->locations)
            collect_rpt_sliced_stats(
                rpt_sliced_stats_set, *report, *loc.stats());
        BOOST_FOREACH(const location_set_stat_map_t::value_type &v,
                      location_set_stat_map)
            collect_rpt_sliced_stats(
                rpt_sliced_stats_set, *report, *v.second);

        collect_sweep_rpt_sliced_stats(
            rpt_sliced_stats_set, *report, global_drill_down_info);
        BOOST_FOREACH(const location_set_drill_down_t::value_type &v,
                      drill_down_info->drill_downs)
            collect_sweep_rpt_sliced_stats(
                rpt_sliced_stats_set, *report, v.second);

        report->sliced_stats.reserve(rpt_sliced_stats_set.size());
        BOOST_FOREACH(const rpt_sliced_stats_t &sliced_stats,
                      rpt_sliced_stats_set)
            report->sliced_stats.push_back(sliced_stats);
    }

    // Phase 3: Lookup and save statistics references.
    {
        report->global_stats =
            lookup_sliced_stats(*report, *advice_result->global_statistics);
        report->location_stats.reserve(
            advice_result->program_structure->locations.size());
        BOOST_FOREACH(const prog_location_t &loc,
                      advice_result->program_structure->locations)
            report->location_stats.push_back(
                lookup_sliced_stats(*report, *loc.stats()));
        report->location_set_stats.reserve(
            location_set_stat_map.size());
        BOOST_FOREACH(const location_set_stat_map_t::value_type &v,
                      location_set_stat_map)
            report->location_set_stats.push_back(
                lookup_sliced_stats(*report, *v.second));
    }

    // Phase 4: Collect statistics sweeps.
    {
        rpt_stats_sweep_set_t rpt_stats_sweep_set;

        collect_stats_sweeps(
            rpt_stats_sweep_set, *report, global_drill_down_info);
        BOOST_FOREACH(const location_set_drill_down_t::value_type &v,
                      drill_down_info->drill_downs)
            collect_stats_sweeps(
                rpt_stats_sweep_set, *report, v.second);

        report->stats_sweeps.reserve(rpt_stats_sweep_set.size());
        BOOST_FOREACH(const rpt_stats_sweep_t &stats_sweep,
                      rpt_stats_sweep_set)
            report->stats_sweeps.push_back(stats_sweep);
    }

    // Phase 5: Lookup and save statistics sweep references.
    {
        report->global_stats_sweep =
            lookup_stats_sweep(*report, global_drill_down_info);
        report->location_set_stats_sweeps.reserve(
            report->location_sets.size());
        BOOST_FOREACH(const rpt_location_set_t &location_set,
                      report->location_sets) {
            location_set_drill_down_t::const_iterator i =
                drill_down_info->drill_downs.find(location_set);
            if (i == drill_down_info->drill_downs.end())
                report->location_set_stats_sweeps.push_back(-1);
            else
                report->location_set_stats_sweeps.push_back(
                    lookup_stats_sweep(*report, i->second));
        }
    }

    // ----- Statistics conversion ends. --------------------------------------

    // Collect dependence groups for the intermediate format.
    for (rpt_depgroup_id_t depgroup_id = 0;
         depgroup_id < (rpt_depgroup_id_t)advice_result->program_structure->depgroups.size();
         depgroup_id++) {
        const rpt_location_set_t &location_set =
            advice_result->program_structure->depgroups[depgroup_id].pc_set();
        rpt_location_set_vector_t::const_iterator i =
            lower_bound(report->location_sets.begin(), report->location_sets.end(), location_set);
        assert(i != report->location_sets.end());
        assert(*i == location_set);
        rpt_location_set_id_t location_set_id =
            (rpt_location_set_id_t)(i - report->location_sets.begin());
        report->depgroups[depgroup_id] = rpt_depgroup_id_t(location_set_id);
    }

    // Loop 0 is the dummy loop containing all instructions that aren't part of
    // any real loop. There is no entry for it in the loops vector, so it is
    // indexed by loop id - 1.
    for (rpt_loop_id_t loop_id = 1;
         loop_id < (rpt_loop_id_t)advice_result->program_structure->loops.size() + 1;
         loop_id++) {
        const prog_loop_t &loop = advice_result->program_structure->loops[loop_id - 1];

        rpt_location_set_id_t incl_location_set_id;
        {
            const rpt_location_set_t &location_set = loop.loop()->get_locations(true);
            rpt_location_set_vector_t::const_iterator i =
                lower_bound(report->location_sets.begin(), report->location_sets.end(), location_set);
            assert(i != report->location_sets.end());
            assert(*i == location_set);
            incl_location_set_id = (rpt_location_set_id_t)(i - report->location_sets.begin());
        }

        rpt_location_set_id_t excl_location_set_id;
        {
            const rpt_location_set_t &location_set = loop.loop()->get_locations(false);
            rpt_location_set_vector_t::const_iterator i =
                lower_bound(report->location_sets.begin(), report->location_sets.end(), location_set);
            assert(i != report->location_sets.end());
            assert(*i == location_set);
            excl_location_set_id = (rpt_location_set_id_t)(i - report->location_sets.begin());
        }

        rpt_depgroup_set_t loop_depgroups;
        BOOST_FOREACH(depgroup_id_t depgroup_id, loop.depgroup_ids())
            if (report->depgroups.find(depgroup_id) != report->depgroups.end())
                loop_depgroups.insert(depgroup_id);

        report->loops[loop_id] = rpt_loop_t(
            incl_location_set_id,
            excl_location_set_id,
            loop_depgroups);
    }

    BOOST_FOREACH(const string &extra_binary, extra_binaries) {
	context->feedback->log(feedback_base_t::LOG_INFO,
                               "Extra binary: " + extra_binary);
    }

    debug_lookup_ptr_t debug_lookup(
        new debug_lookup_t(
            cppfilt,
            verbose,
            context->feedback,
            source_directories,
            debug_directories,
            extra_binaries,
            debug_symbol_level));

    // Add binaries from the sample file.
    exe_maps_result_cptr_t exe_maps =
	exe_maps_t(context, configuration->sample_file).get();
    BOOST_FOREACH(const exe_map_t &exe_map, exe_maps->exe_maps)
	debug_lookup->add_debug_file(exe_map.filename,
                                     exe_map.address,
                                     exe_map.size);

    // Process all added files
    debug_lookup->resolve_debug_files();

    // ----- Debug info and source collection begins. -------------------------

    // Phase 1: Collect pcs that we need source code mapping for.
    {
        rpt_debug_pc_set_t debug_pcs;
        collect_debug_pcs(debug_pcs, feature_mt_advice, *report, advice_result);
        report->debug_pcs.reserve(debug_pcs.size());
        BOOST_FOREACH(const rpt_debug_pc_t &debug_pc, debug_pcs)
            report->debug_pcs.push_back(debug_pc);
    }

    // Phase 2: Collect binaries.
    {
        rpt_binary_set_t binaries;
        BOOST_FOREACH(const file_info_map_t::value_type &v,
                      debug_lookup->seen_filenames)
            binaries.insert(
                rpt_binary_t(v.second.actual_filename,
                             v.second.address.paddr,
                             v.second.size,
                             (rpt_debug_info_status_t)v.second.debug_info_status));
        BOOST_FOREACH(const rpt_binary_t &binary, binaries)
            report->binaries.push_back(binary);
    }

    // TODO: We do three sym lookups for each pc below. We could cache the
    //       lookup results if this turns out to be expensive.

    // Phase 3: Symbols, functions and source files.
    {
        rpt_symbol_set_t symbols;
        rpt_function_set_t functions;
        rpt_source_file_set_t source_files;

        BOOST_FOREACH(const rpt_debug_pc_t &debug_pc, report->debug_pcs) {
            rpt_address_t pc = report->get_address(debug_pc.pc_id);
            if (debug_pc.return_address)
                pc -= 1;
            sym_info_cptr_t sym_info = debug_lookup->lookup_pc(pc);

            if (sym_info->symbol_name)
                symbols.insert(rpt_symbol_t(*sym_info->symbol_name,
                                            pc - sym_info->symbol_offset));
            if (sym_info->function_name)
                functions.insert(rpt_function_t(*sym_info->function_name,
                                                pc - sym_info->function_offset));
            if (sym_info->source_file)
                source_files.insert(convert_source_file(*sym_info));
        }

        BOOST_FOREACH(const rpt_symbol_t &symbol, symbols)
            report->symbols.push_back(symbol);
        BOOST_FOREACH(const rpt_function_t &function, functions)
            report->functions.push_back(function);
        BOOST_FOREACH(const rpt_source_file_t &source_file, source_files)
            report->source_files.push_back(source_file);
    }

    // Phase 4: Collect source lines.
    {
        rpt_source_line_set_t source_lines;

        BOOST_FOREACH(const rpt_debug_pc_t &debug_pc, report->debug_pcs) {
            rpt_address_t pc = report->get_address(debug_pc.pc_id);
            if (debug_pc.return_address)
                pc -= 1;
            sym_info_cptr_t sym_info = debug_lookup->lookup_pc(pc);

            if (sym_info->source_line != -1)
                source_lines.insert(convert_source_line(*report, *sym_info));
        }

        BOOST_FOREACH(const rpt_source_line_t &source_line, source_lines)
            report->source_lines.push_back(source_line);
    }

    // Phase 5: Fill in pcs with the collected debug stuff.
    BOOST_FOREACH(rpt_debug_pc_t &debug_pc, report->debug_pcs) {
        rpt_address_t pc = report->get_address(debug_pc.pc_id);
        if (debug_pc.return_address)
            pc -= 1;
        sym_info_cptr_t sym_info = debug_lookup->lookup_pc(pc);

        if (sym_info->binary_file)
            debug_pc.binary_id =
                report->get_binary_id(convert_binary(*sym_info, pc, debug_lookup));
        if (sym_info->symbol_name)
            debug_pc.symbol_id =
                report->get_symbol_id(convert_symbol(*sym_info, pc));
        if (sym_info->function_name)
            debug_pc.function_id =
                report->get_function_id(convert_function(*sym_info, pc));
        if (sym_info->source_line != -1)
            debug_pc.source_line_id =
                report->get_source_line_id(
                    convert_source_line(*report, *sym_info));
    }

    // Phase 6: Read source files.
    BOOST_FOREACH(rpt_source_file_t &source_file, report->source_files) {
	string_cptr_t filename = debug_lookup->lookup_source_file(source_file.filename);
        ifstream is(filename->c_str(), ios::binary);
        if (is.fail())
	    context->feedback->log(feedback_base_t::LOG_INFO,
				   "Can't read source file: " + *filename);

        is.seekg(0, ios::end);
        streamoff data_length = is.tellg();
        if (data_length == -1)
            continue;
        is.seekg(0, ios::beg);
        char_array_ptr_t data(new char[data_length]);
        is.read(data.get(), data_length);
        is.close();
        if (is.fail())
            continue;

        source_file.data = data;
        source_file.data_length = (uint32_t)data_length;
    }

    // ----- Debug info and source collection ends. ---------------------------

    // Create issues.
    convert_issues(*report, advice_result, feature_mt_advice);

    return report;
}


rpt_report_cptr_t
generate_report(feature_context_ptr_t context,
		const conf_t &conf,
		bool feature_mt_advice,
		bool feature_sw_prefetch,
		bool feature_unlimited_callstack,
		bool feature_blocking_expert,
		bool feature_multisocket)
{
    return generate_report(context,
                conf.analysis_time,
                conf.analysis_argv,
                conf.title,
                conf.sample_file,
                conf.extra_binaries,
                conf.debug_directories,
                conf.source_directories,
                (debug_symbol_level_t)conf.debug_symbol_level,
                conf.cppfilt,
                conf.percentage,
                conf.replacement_policy,
                conf.cache_size,
                conf.line_size,
                conf.num_caches,
                conf.cpu_manufacturer,
                conf.cpu,
                conf.num_cpus,
                conf.cache,
                conf.cache_level,
                conf.explicit_mapping,
                conf.real_thread_id,
                conf.stack_depth,
                conf.windowing,
                conf.windowing_burst,
                feature_mt_advice,
                feature_sw_prefetch,
                feature_unlimited_callstack,
                feature_blocking_expert && conf.blocking_expert,
                feature_multisocket,
                conf.new_algorithm,
                conf.low_mem,
		conf.verbose,
                conf.autodetect_failed);
}

END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;
