/// @file analysis/statistics/statistics.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include "analysis/statistics/statistics.hh"
#include <boost/lexical_cast.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace boost;
using namespace std;

// ----- stats_t --------------------------------------------------------------

bool
stats_t::operator==(const stats_t &that) const
{
    return
        all_ratio == that.all_ratio &&
        stall_ratio == that.stall_ratio &&
        spatial_util == that.spatial_util &&
        fetch_util == that.fetch_util &&
        write_back_util == that.write_back_util &&
        wasted_space == that.wasted_space &&
        fetch_waste == that.fetch_waste &&
        write_back_waste == that.write_back_waste &&
        util_corr_fr == that.util_corr_fr &&
        util_corr_mr == that.util_corr_mr &&
        util_corr_wbr == that.util_corr_wbr &&
        avoidable_fetches == that.avoidable_fetches &&
        avoidable_misses == that.avoidable_misses &&
        avoidable_write_backs == that.avoidable_write_backs &&
        hw_prefetch_ratio == that.hw_prefetch_ratio &&
        random_ratio == that.random_ratio &&
        comm_lines == that.comm_lines &&
        comm_util == that.comm_util &&
        comm_waste == that.comm_waste &&
        fs_ratio == that.fs_ratio &&
        fs_waste == that.fs_waste;
}

// ----- statistics_result_t --------------------------------------------------

/*
  Comment regarding calculation of utilization corrected miss ratio.
  ------------------------------------------------------------------

  The basic formula that relates fetches, misses and prefetches

    fetches = misses + prefetches
    misses = fetches - prefetches
    misses = fetches * (1 - prefetches/fetches)

  Divide with number of acceses

    miss_ratio = fetch_ratio * (1 - prefetches / fetches)
               = fetch_ratio * (1 - prefetch_fraction)

  Now, the simplification is that when we fix the utilization
  problems, the quotient prefetches/fetches remain constant. This is
  not really accurate, as the access pattern will be more dense, and
  therefore more likely to trigger a prefetch stream.

    util.corr.miss_ratio = util.corr.fetch_ratio * (1 - prefetch_fraction)

  The error is that the prefetch_fraction is too low, and the
  util.corr.miss_ratio is likely calculated higher that it should be.

  prefetch_fraction can also be written

    prefetch_fraction = (prefetches/accesses) / (fetches/accesses)
                      = hw_prefetch_ratio / fetch_ratio.

  The slight source of confusion is that
  hw_prefetch_result::prefetch_ratio_rd_wr() gives prefetches/fetches,
  not prefetches/accesses.
*/

stats_cptr_t
statistics_result_t::get(const locationx_set_t &lx_set, bool include_all) const
{
    fr_data_t fr_data_fetches = fr_lookup->fr_by_lx_set(lx_set);
    fr_data_t fr_data_misses = include_all ?
        fr_lookup->fr_by_lx_set(lx_set) :
        fr_lookup_rd_wr->fr_by_lx_set(lx_set);
    double hw_prefetch_ratio = include_all ?
        hwpf_lookup->prefetch_ratio(lx_set) :
        hwpf_lookup_rd_wr->prefetch_ratio(lx_set);
    fr_data_t fr_data_wb(fr_lookup->fr_by_lx_set(lx_set));
    potential_result_t p_fetch_res(fetch_potential->get(lx_set));
    potential_result_t p_wb_res(write_back_potential ?
				write_back_potential->get(lx_set) :
				potential_result_t(0.0, 0.0));
    false_sharing_info_t fs_info(fs_res->get(lx_set));

    return stats_cptr_t(
	new stats_t(fr_data_fetches.by_kind(),
		    fr_data_misses.by_kind().multiply_fetches_upgrades(1.0 - hw_prefetch_ratio),
		    hw_prefetch_ratio,
		    (utilization_classic ? utilization_classic->get_utilization(lx_set) :
		     result_valid_t(0.0, false)),
		    (utilization_fetch ? utilization_fetch->get_utilization(lx_set) :
		     result_valid_t(0.0, false)),
		    (utilization_wb ? utilization_wb->get_utilization(lx_set) :
		     result_valid_t(0.0, false)),
		    (utilization_classic ? utilization_classic->get_wasted_space(lx_set) :
		     result_valid_t(0.0, false)),
		    (utilization_fetch ? utilization_fetch->get_wasted_space(lx_set) :
		     result_valid_t(0.0, false)),
		    (utilization_wb ? utilization_wb->get_wasted_space(lx_set) :
		     result_valid_t(0.0, false)),
		    p_fetch_res.util_corr_ratio,
		    p_fetch_res.util_corr_ratio * (1.0 - hw_prefetch_ratio),
		    p_wb_res.util_corr_ratio,
		    p_fetch_res.avoidable_count,
		    p_fetch_res.avoidable_count * (1.0 - hw_prefetch_ratio),
		    p_wb_res.avoidable_count,
		    ra_lookup->random_ratio(lx_set),
                    comm_res->get_info(lx_set).line_comm / line_size,
		    comm_res->utilization(lx_set),
		    comm_res->waste(lx_set),
		    result_valid_t(fs_info.valid, fs_info.ratio),
		    fs_info.waste));
}

stats_cptr_t
statistics_result_t::get(locationx_t lx, bool include_all) const
{
    locationx_set_t lx_set;
    lx_set.insert(lx);
    return get(lx_set, include_all);
}

stats_cptr_t
statistics_result_t::get(const location_set_t &pc_set, bool include_all) const
{
    locationx_set_t lx_set = lx_mapping->get_locationx_set(pc_set);
    return get(lx_set, include_all);
}

stats_cptr_t
statistics_result_t::get(location_t pc, bool include_all) const
{
    locationx_set_t lx_set = lx_mapping->get_locationx_set(pc);
    return get(lx_set, include_all);
}

statistics_result_t::statistics_result_t(
    fnr_line_size_t line_size,
    locationx_mapper_result_cptr_t lx_mapping,
    fr_lookup_result_cptr_t fr_lookup,
    fr_lookup_result_cptr_t fr_lookup_rd_wr,
    utilization_result_cptr_t utilization_classic,
    utilization_result_cptr_t utilization_fetch,
    utilization_result_cptr_t utilization_wb,
    potential_ptr_t fetch_potential,
    potential_ptr_t write_back_potential,
    hwpf_lookup_result_cptr_t hwpf_lookup,
    hwpf_lookup_result_cptr_t hwpf_lookup_rd_wr,
    ra_lookup_result_cptr_t ra_lookup,
    communication_result_cptr_t comm_res,
    false_sharing_result_cptr_t fs_res)
    : line_size(line_size),
      lx_mapping(lx_mapping),
      fr_lookup(fr_lookup),
      fr_lookup_rd_wr(fr_lookup_rd_wr),
      utilization_classic(utilization_classic),
      utilization_fetch(utilization_fetch),
      utilization_wb(utilization_wb),
      fetch_potential(fetch_potential),
      write_back_potential(write_back_potential),
      hwpf_lookup(hwpf_lookup),
      hwpf_lookup_rd_wr(hwpf_lookup_rd_wr),
      ra_lookup(ra_lookup),
      comm_res(comm_res),
      fs_res(fs_res)
{
}

// ----- statistics_t ---------------------------------------------------------

statistics_result_cptr_t
statistics_t::calculate(
    const fnr_line_size_t line_size,
    const locationx_mapper_result_cptr_t &lx_mapping,
    const fr_lookup_result_cptr_t &fr_lookup,
    const fr_lookup_result_cptr_t &fr_lookup_rd_wr,
    const utilization_result_cptr_t &utilization_classic,
    const utilization_result_cptr_t &utilization_fetch,
    const utilization_result_cptr_t &utilization_wb,
    const potential_ptr_t &fetch_potential,
    const potential_ptr_t &write_back_potential,
    const hwpf_lookup_result_cptr_t &hwpf_lookup,
    const hwpf_lookup_result_cptr_t &hwpf_lookup_rd_wr,
    const ra_lookup_result_cptr_t &ra_lookup,
    const communication_result_cptr_t &comm_res,
    const false_sharing_result_cptr_t &fs_res,
    const bool use_old_hwpf_algorithm,
    const bool hwpf_enabled)
{
    return statistics_result_ptr_t(
        new statistics_result_t(
            line_size,
            lx_mapping, fr_lookup, fr_lookup_rd_wr,
            utilization_classic, utilization_fetch, utilization_wb,
	    fetch_potential, write_back_potential,
	    hwpf_lookup, hwpf_lookup_rd_wr,
	    ra_lookup,
	    comm_res, fs_res));
}

statistics_t::statistics_t(const feature_context_ptr_t &context,
                           const string &sample_file,
                           const fnr_stack_depth_t stack_depth,
                           const fnr_line_size_t line_size,
                           const thread_mapping_cptr_t &thread_mapping,
			   const cpu_model_t &cpu_model,
                           const locationx_mapper_cptr_t &lx_mapper,
                           const fnr_cache_size_t cache_size,
                           const repl_fr_engine_cptr_t &engine,
			   const windowing_engine_cptr_t &windowing,
                           const fnr_line_size_t page_size,
                           const int pos_streams,
                           const int neg_streams,
                           const fnr_line_count_t prefetch_reach,
			   const bool use_old_hwpf_algorithm,
                           const bool hwpf_enabled)
    : feature_t<statistics_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      lx_mapper(lx_mapper),
      cache_size(cache_size),
      engine(engine),
      windowing(windowing),
      page_size(page_size),
      pos_streams(pos_streams),
      neg_streams(neg_streams),
      prefetch_reach(prefetch_reach),
      use_old_hwpf_algorithm(use_old_hwpf_algorithm),
      hwpf_enabled(hwpf_enabled)
{
}

statistics_result_cptr_t
statistics_t::get() const
{
    return get_data();
}

string
statistics_t::name() const
{
    return "statistics(" +
	lexical_cast<string>(sample_file) + "," +
        lexical_cast<string>(stack_depth) + "," +
	lexical_cast<string>(line_size) + "," +
	thread_mapping->name() + "," +
	cpu_model.name() + "," +
	lexical_cast<string>(cache_size) + "," +
        lx_mapper->name() + "," +
        engine->name() + "," +
	windowing->name() + "," +
	lexical_cast<string>(page_size) + "," +
	lexical_cast<string>(pos_streams) + "," +
	lexical_cast<string>(neg_streams) + "," +
	lexical_cast<string>(prefetch_reach) + "," +
	lexical_cast<string>(use_old_hwpf_algorithm) + "," +
        lexical_cast<string>(hwpf_enabled) + ")";
}

statistics_result_cptr_t
statistics_t::calculate() const
{
    parser_result_cptr_t parser(parser_t(context, sample_file).get());
    locationx_mapper_result_cptr_t lx_mapping(lx_mapper->get());

    fr_lookup_result_cptr_t fr_lookup(
	fr_lookup_t(context, sample_file, stack_depth, line_size, page_size,
		    thread_mapping, cpu_model, lx_mapper, cache_size, engine,
		    FRL_Type_All).get());

    fr_lookup_result_cptr_t fr_lookup_rd_wr(
	fr_lookup_t(context, sample_file, stack_depth, line_size, page_size,
		    thread_mapping, cpu_model, lx_mapper, cache_size, engine,
		    FRL_Type_Stalls).get());

    utilization_result_cptr_t utilization_classic;
    utilization_result_cptr_t utilization_fetch;
    utilization_result_cptr_t utilization_wb;
    potential_ptr_t write_back_potential;
    potential_ptr_t fetch_potential(
        new potential_t(context, sample_file, stack_depth, line_size,
			page_size, thread_mapping, cpu_model, lx_mapper,
			cache_size, engine, windowing,
			parser->mt_sampled() ? Utilization_Fetch : Utilization_Classic));

    if (parser->mt_sampled()) {
	utilization_fetch =
	    utilization_t(context, sample_file, stack_depth, line_size,
			  page_size, thread_mapping, cpu_model, lx_mapper,
			  cache_size, windowing, Utilization_Fetch).get();

	utilization_wb =
	    utilization_t(context, sample_file, stack_depth, line_size,
			  page_size, thread_mapping, cpu_model, lx_mapper,
			  cache_size, windowing, Utilization_Write_Back).get();

	write_back_potential.reset(
	    new potential_t(context, sample_file, stack_depth, line_size,
			    page_size, thread_mapping, cpu_model, lx_mapper,
			    cache_size, engine, windowing,
			    Utilization_Write_Back));
    } else {
	utilization_classic =
	    utilization_t(context, sample_file, stack_depth, line_size,
		      page_size, thread_mapping, cpu_model, lx_mapper,
		      cache_size, windowing, Utilization_Classic).get();
    }



    hwpf_lookup_result_cptr_t hwpf_lookup(
	hwpf_lookup_t(context, sample_file, stack_depth, line_size, page_size,
		      thread_mapping, cpu_model, lx_mapper, cache_size,
		      engine, windowing, pos_streams, neg_streams,
		      prefetch_reach, true, use_old_hwpf_algorithm, hwpf_enabled).get());

    hwpf_lookup_result_cptr_t hwpf_lookup_rd_wr(
	hwpf_lookup_t(context, sample_file, stack_depth, line_size, page_size,
		      thread_mapping, cpu_model, lx_mapper, cache_size,
		      engine, windowing, pos_streams, neg_streams,
		      prefetch_reach, false, use_old_hwpf_algorithm, hwpf_enabled).get());

    ra_lookup_result_cptr_t ra_lookup(
	ra_lookup_t(context, sample_file, stack_depth, line_size, page_size,
		    thread_mapping, cpu_model, lx_mapper, cache_size,
		    engine).get());

    communication_result_cptr_t comm_res(
	communication_t(context, sample_file, stack_depth, line_size,
			page_size, cache_size, thread_mapping,
			cpu_model, engine, lx_mapper).get());

    false_sharing_result_cptr_t fs_res(
	false_sharing_t(context, sample_file, stack_depth, line_size,
			page_size, cache_size, thread_mapping,
			cpu_model, engine, lx_mapper).get());

    return calculate(line_size, lx_mapping, fr_lookup, fr_lookup_rd_wr,
                     utilization_classic, utilization_fetch, utilization_wb,
		     fetch_potential, write_back_potential,
		     hwpf_lookup, hwpf_lookup_rd_wr,
		     ra_lookup,
		     comm_res, fs_res, use_old_hwpf_algorithm, hwpf_enabled);
}

END_NAMESPACE;
END_NAMESPACE;
