/// @file analysis/statistics/statistics.hh

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#ifndef STATISTICS_HH
#define STATISTICS_HH

#include "analysis/hw_prefetch/hwpf_lookup.hh"
#include "analysis/random_access/ra_lookup.hh"
#include "analysis/utilization/potential.hh"
#include "analysis/communication/false_sharing.hh"
#include "analysis/communication/communication.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace std;

using boost::shared_ptr;

// ----- stats_t --------------------------------------------------------------

/// Statistics. Global, for a pc set or for an individual pc.
///
/// Some of the fields have different semantics for individual pcs compared to
/// global and pc set statistics.
struct stats_t {
    stats_t(const stats_t &stats)
        : all_ratio(stats.all_ratio),
          stall_ratio(stats.stall_ratio),
	  spatial_util(stats.spatial_util),
          fetch_util(stats.fetch_util),
          write_back_util(stats.write_back_util),
	  wasted_space(stats.wasted_space),
	  fetch_waste(stats.fetch_waste),
	  write_back_waste(stats.write_back_waste),
          util_corr_fr(stats.util_corr_fr),
	  util_corr_mr(stats.util_corr_mr),
	  util_corr_wbr(stats.util_corr_wbr),
          avoidable_fetches(stats.avoidable_fetches),
	  avoidable_misses(stats.avoidable_misses),
	  avoidable_write_backs(stats.avoidable_write_backs),
          hw_prefetch_ratio(stats.hw_prefetch_ratio),
          random_ratio(stats.random_ratio),
          comm_lines(stats.comm_lines),
	  comm_util(stats.comm_util),
	  comm_waste(stats.comm_waste),
	  fs_ratio(stats.fs_ratio),
	  fs_waste(stats.fs_waste)
        {}

    stats_t(by_kind_t all_ratio,
            by_kind_t stall_ratio,
            double hw_prefetch_ratio,
	    result_valid_t spatial_util,
	    result_valid_t fetch_util,
	    result_valid_t write_back_util,
	    result_valid_t wasted_space,
	    result_valid_t fetch_waste,
	    result_valid_t write_back_waste,
            double util_corr_fr,
            double util_corr_mr,
            double util_corr_wbr,
            double avoidable_fetches,
	    double avoidable_misses,
	    double avoidable_write_backs,
            double random_ratio,
            double comm_lines,
	    result_valid_t comm_util,
	    double comm_waste,
	    result_valid_t fs_ratio,
	    double fs_waste)
        : all_ratio(all_ratio),
          stall_ratio(stall_ratio),
          spatial_util(spatial_util),
          fetch_util(fetch_util),
          write_back_util(write_back_util),
	  wasted_space(wasted_space),
	  fetch_waste(fetch_waste),
	  write_back_waste(write_back_waste),
          util_corr_fr((stat_fp_t)util_corr_fr),
          util_corr_mr((stat_fp_t)util_corr_mr),
          util_corr_wbr((stat_fp_t)util_corr_wbr),
          avoidable_fetches((stat_fp_t)avoidable_fetches),
          avoidable_misses((stat_fp_t)avoidable_misses),
          avoidable_write_backs((stat_fp_t)avoidable_write_backs),
          hw_prefetch_ratio((stat_fp_t)hw_prefetch_ratio),
          random_ratio((stat_fp_t)random_ratio),
          comm_lines((stat_fp_t)comm_lines),
	  comm_util(comm_util),
	  comm_waste((stat_fp_t)comm_waste),
	  fs_ratio(fs_ratio),
	  fs_waste((stat_fp_t)fs_waste)
        {}


    stats_t()
        : all_ratio(),
          stall_ratio(),
	  util_corr_fr(0),
          util_corr_mr(0),
          util_corr_wbr(0),
          avoidable_fetches(0),
          avoidable_misses(0),
          avoidable_write_backs(0),
          hw_prefetch_ratio(0),
          random_ratio(0),
          comm_lines(0.0),
	  comm_util(),
	  comm_waste(0),
	  fs_ratio(),
	  fs_waste(0)
        {}

    bool operator==(const stats_t &that) const;

    /// Fetch, upgrade and write-back ratios before compensating for the effect
    /// of software and hardware prefetches, i.e., "fetch ratio".
    by_kind_t all_ratio;

    /// Fetch, upgrade and write-back ratios after compensating for the effect
    /// of software and hardware prefetches, i.e., "miss ratio".
    by_kind_t stall_ratio;

    /// Spatial usage. Includes both reads, writes and prefetches.
    result_valid_t spatial_util;

    /// Fetch utilization
    result_valid_t fetch_util;

    /// Write-back utilization
    result_valid_t write_back_util;

    /// Wasted space. Includes both reads, writes and prefetches.
    result_valid_t wasted_space;

    /// Wasted bandwidth for fetches.
    result_valid_t fetch_waste;

    /// Wasted bandwidth for write-backs.
    result_valid_t write_back_waste;

    /// Fetch ratio if the fetch utilization of the code was increased to
    /// 100%. Includes both reads, writes and prefetches.
    stat_fp_t util_corr_fr;

    /// Miss ratio if the fetch utilization of the code was increased to
    /// 100%. Includes both reads, writes and prefetches.
    stat_fp_t util_corr_mr;

    /// Write-back ratio if the write-back utilization of the code was increased to
    /// 100%.
    stat_fp_t util_corr_wbr;

    /// Fetches that would be avoided if the fetch utilization of the code
    /// was increased to 100%. Includes both reads, writes and prefetches.
    stat_fp_t avoidable_fetches;

    /// Misses that would be avoided if the fetch utilization of the code
    /// was increased to 100%. Includes both reads, writes and prefetches.
    stat_fp_t avoidable_misses;

    /// Write-backs that would be avoided if the write-back utilization
    /// of the code was increased to 100%.
    stat_fp_t avoidable_write_backs;

    /// Hardware prefetch ratio. Includes only reads and writes for pc sets and
    /// global statistics. Includes reads, writes and prefetches for individual
    /// pcs.
    stat_fp_t hw_prefetch_ratio;

    /// Randomness ratio. Includes both reads, writes and prefetches.
    stat_fp_t random_ratio;

    /// Communicated cache lines.
    stat_fp_t comm_lines;
    /// Communication utilization.
    result_valid_t comm_util;
    /// Communication waste.
    stat_fp_t comm_waste;

    /// False-sharing ratio.
    result_valid_t fs_ratio;
    /// False sharing waste.
    stat_fp_t fs_waste;
};

typedef shared_ptr<stats_t> stats_ptr_t;
typedef shared_ptr<const stats_t> stats_cptr_t;

// ----- statistics_result_t --------------------------------------------------

class statistics_result_t {
    friend class statistics_t;

public:
    /// Return statistics for a lx set. If include_all is true,
    /// include contributions from instructions that wouldn't normally
    /// cause a pipeline stall if they miss, e.g. prefetches and
    /// flushes.
    stats_cptr_t get(const locationx_set_t &lx_set, bool include_all) const;

    /// Return statistics for a lx.  If include_all is true, include
    /// contributions from instructions that wouldn't normally cause a
    /// pipeline stall if they miss, e.g. prefetches and flushes.
    stats_cptr_t get(locationx_t lx, bool include_all) const;

    /// Return statistics for a pc set. If include_all is true,
    /// include contributions from instructions that wouldn't normally
    /// cause a pipeline stall if they miss, e.g. prefetches and
    /// flushes.
    stats_cptr_t get(const location_set_t &pc_set, bool include_all) const;

    /// Return statistics for a pc. If include_all is true, include
    /// contributions from instructions that wouldn't normally cause a
    /// pipeline stall if they miss, e.g. prefetches and flushes.
    stats_cptr_t get(location_t pc, bool include_all) const;

private:
    statistics_result_t(
        fnr_line_size_t line_size,
        locationx_mapper_result_cptr_t lx_mapping,
        fr_lookup_result_cptr_t fr_lookup,
        fr_lookup_result_cptr_t fr_lookup_rd_wr,
        utilization_result_cptr_t utilization_classic,
        utilization_result_cptr_t utilization_fetch,
        utilization_result_cptr_t utilization_wb,
        potential_ptr_t fetch_potential,
        potential_ptr_t write_back_potential,
        hwpf_lookup_result_cptr_t hwpf_lookup,
        hwpf_lookup_result_cptr_t hwpf_lookup_rd_wr,
        ra_lookup_result_cptr_t ra_lookup,
	communication_result_cptr_t comm_res,
	false_sharing_result_cptr_t fs_res);

    const fnr_line_size_t line_size;
    const locationx_mapper_result_cptr_t lx_mapping;
    const fr_lookup_result_cptr_t fr_lookup;
    const fr_lookup_result_cptr_t fr_lookup_rd_wr;
    const utilization_result_cptr_t utilization_classic;
    const utilization_result_cptr_t utilization_fetch;
    const utilization_result_cptr_t utilization_wb;
    const potential_ptr_t fetch_potential;
    const potential_ptr_t write_back_potential;
    const hwpf_lookup_result_cptr_t hwpf_lookup;
    const hwpf_lookup_result_cptr_t hwpf_lookup_rd_wr;
    const ra_lookup_result_cptr_t ra_lookup;
    const communication_result_cptr_t comm_res;
    const false_sharing_result_cptr_t fs_res;
};

typedef shared_ptr<statistics_result_t> statistics_result_ptr_t;
typedef shared_ptr<const statistics_result_t> statistics_result_cptr_t;

// ----- statistics_t ---------------------------------------------------------

/// Module for statistics calculation.
class statistics_t : public feature_t<statistics_result_cptr_t> {
public:
    static statistics_result_cptr_t calculate(
        const fnr_line_size_t line_size,
        const locationx_mapper_result_cptr_t &lx_mapping,
        const fr_lookup_result_cptr_t &fr_lookup,
        const fr_lookup_result_cptr_t &fr_lookup_rd_wr,
        const utilization_result_cptr_t &utilization_classic,
        const utilization_result_cptr_t &utilization_fetch,
        const utilization_result_cptr_t &utilization_wb,
        const potential_ptr_t &fetch_potential,
        const potential_ptr_t &write_back_potential,
        const hwpf_lookup_result_cptr_t &hwpf_lookup,
        const hwpf_lookup_result_cptr_t &hwpf_lookup_rd_wr,
        const ra_lookup_result_cptr_t &ra_lookup,
	const communication_result_cptr_t &comm_res,
	const false_sharing_result_cptr_t &fs_res,
	const bool use_old_hwpf_algorithm,
        const bool hwpf_enabled);

    statistics_t(const feature_context_ptr_t &context,
                 const string &sample_file,
                 const fnr_stack_depth_t stack_depth,
                 const fnr_line_size_t line_size,
                 const thread_mapping_cptr_t &thread_mapping,
		 const cpu_model_t &cpu_model,
                 const locationx_mapper_cptr_t &lx_mapper,
                 const fnr_cache_size_t cache_size,
                 const repl_fr_engine_cptr_t &engine,
		 const windowing_engine_cptr_t &windowing,
                 const fnr_line_size_t page_size,
                 const int pos_streams,
                 const int neg_streams,
                 const fnr_line_count_t prefetch_reach,
		 const bool use_old_hwpf_algorithm,
                 const bool hwpf_enabled);
    statistics_result_cptr_t get() const;

protected:
    string name() const;
    statistics_result_cptr_t calculate() const;

private:
    /// Sample file.
    const string sample_file;

    /// Call stack depth to use in the analysis.
    const fnr_stack_depth_t stack_depth;

    /// Cache line size.
    const fnr_line_size_t line_size;

    const thread_mapping_cptr_t thread_mapping;
    const cpu_model_t cpu_model;
    const locationx_mapper_cptr_t lx_mapper;

    /// Miss ratio calculation parameters.
    const fnr_cache_size_t cache_size;
    const repl_fr_engine_cptr_t engine;
    const windowing_engine_cptr_t windowing;

    /// Page size that the hardware prefetcher operates on.
    const fnr_line_size_t page_size;

    /// The number of prefetch streams with positive and negative strides,
    /// respectively.
    const int pos_streams;
    const int neg_streams;

    /// The number of sequential cache lines that are prefetched when a
    /// prefetch stream is triggered.
    const fnr_line_count_t prefetch_reach;

    /// True if we should use the old hwpf algorithm (which is fast but has
    /// drawbacks)
    const bool use_old_hwpf_algorithm;

    const bool hwpf_enabled;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
