/// @file analysis/statistics/stats_slicer.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include "analysis/statistics/stats_slicer.hh"
#include <boost/lexical_cast.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::analysis;
using namespace boost;

typedef map<fnr_thread_id_t, locationx_set_t> thread_lx_sets_t;

// ----- sliced_stats_t -------------------------------------------------------

bool
sliced_stats_t::operator==(const sliced_stats_t &that) const
{
    thread_stats_t::const_iterator this_i = thread_stats.begin();
    thread_stats_t::const_iterator that_i = that.thread_stats.begin();

    while (1) {
        if (this_i == thread_stats.end() &&
            that_i == that.thread_stats.end())
            break;
        if (this_i == thread_stats.end())
            return false;
        if (that_i == that.thread_stats.end())
            return false;
        if (!(this_i->first == that_i->first) ||
            !(*this_i->second == *that_i->second))
            return false;
        this_i++;
        that_i++;
    }

    return *stats == *that.stats;
}

// ----- stats_slicer_result_t ------------------------------------------------

stats_slicer_result_t::stats_slicer_result_t(
    locationx_mapper_result_cptr_t lx_mapping,
    statistics_result_cptr_t statistics)
    : lx_mapping(lx_mapping),
      statistics(statistics)
{
}

sliced_stats_cptr_t
stats_slicer_result_t::get(
    const locationx_set_t &lx_set,
    bool include_pfs) const
{
    // For all locations in the set 'locs', build a map from thread to
    // corresponding locationx
    thread_lx_sets_t thread_lx_sets;
    BOOST_FOREACH(locationx_t lx, lx_set)
        thread_lx_sets[lx_mapping->get_specifier(lx).thread()].insert(lx);

    sliced_stats_t *stats = new sliced_stats_t;

    // Get statistics for all threads
    stats->stats = statistics->get(lx_set, include_pfs);

    // Get statistics per thread
    BOOST_FOREACH(thread_lx_sets_t::value_type &v, thread_lx_sets)
        stats->thread_stats[v.first] = statistics->get(v.second, include_pfs);
    return sliced_stats_cptr_t(stats);
}

sliced_stats_cptr_t
stats_slicer_result_t::get(
    location_t loc,
    bool include_pfs) const
{
    location_set_t locs;
    locs.insert(loc);
    return get(locs, include_pfs);
}

// ----- stats_slicer_t -------------------------------------------------------

stats_slicer_result_cptr_t
stats_slicer_t::calculate(
    const locationx_mapper_result_cptr_t &lx_mapping,
    const statistics_result_cptr_t &statistics)
{
    stats_slicer_result_ptr_t result(
        new stats_slicer_result_t(lx_mapping, statistics));

    return result;
}

stats_slicer_t::stats_slicer_t(
    const feature_context_ptr_t &context,
    const string &sample_file,
    const fnr_stack_depth_t stack_depth,
    const fnr_line_size_t line_size,
    const thread_mapping_cptr_t &thread_mapping,
    const cpu_model_t &cpu_model,
    const fnr_cache_size_t cache_size,
    const repl_fr_engine_cptr_t &engine,
    const windowing_engine_cptr_t &windowing,
    const fnr_line_size_t page_size,
    const int pos_streams,
    const int neg_streams,
    const fnr_line_count_t prefetch_reach,
    const bool use_old_hwpf_algorithm,
    const bool hwpf_enabled)
    : feature_t<stats_slicer_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      cache_size(cache_size),
      engine(engine),
      windowing(windowing),
      page_size(page_size),
      pos_streams(pos_streams),
      neg_streams(neg_streams),
      prefetch_reach(prefetch_reach),
      use_old_hwpf_algorithm(use_old_hwpf_algorithm),
      hwpf_enabled(hwpf_enabled)
{
}

stats_slicer_result_cptr_t
stats_slicer_t::get() const
{
    return get_data();
}

string
stats_slicer_t::name() const
{
    return "stats_slicer(" +
	lexical_cast<string>(sample_file) + "," +
        lexical_cast<string>(stack_depth) + "," +
	lexical_cast<string>(line_size) + "," +
	thread_mapping->name() + "," +
	cpu_model.name() + "," +
	lexical_cast<string>(cache_size) + "," +
        engine->name() + "," +
	windowing->name() + "," +
	lexical_cast<string>(page_size) + "," +
	lexical_cast<string>(pos_streams) + "," +
	lexical_cast<string>(neg_streams) + "," +
	lexical_cast<string>(prefetch_reach) + "," +
	lexical_cast<string>(use_old_hwpf_algorithm) + "," +
        lexical_cast<string>(hwpf_enabled) + ")";
}

stats_slicer_result_cptr_t
stats_slicer_t::calculate() const
{
    locationx_mapper_cptr_t lx_mapper(
	new locationx_mapper_t(
            context, sample_file, stack_depth, thread_mapping, true, true,
            false));

    locationx_mapper_result_cptr_t lx_mapping(lx_mapper->get());

    statistics_result_cptr_t statistics(
	statistics_t(context, sample_file, stack_depth, line_size,
		     thread_mapping, cpu_model, lx_mapper, cache_size,
		     engine, windowing, page_size, pos_streams,
		     neg_streams, prefetch_reach,
		     use_old_hwpf_algorithm,
                     hwpf_enabled).get());

    return calculate(lx_mapping, statistics);
}

END_NAMESPACE;
END_NAMESPACE;
