/// @file analysis/statistics/stats_slicer.hh

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include "analysis/statistics/statistics.hh"

#ifndef STATS_SLICER_HH
#define STATS_SLICER_HH

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

// ----- sliced_stats_t -------------------------------------------------------

/// Statistics per thread.
typedef map<fnr_thread_id_t, stats_cptr_t> thread_stats_t;

/// Statistics, total and for each thread.
struct sliced_stats_t {
    /// Statistics including all threads.
    stats_cptr_t stats;

    /// Statistics per thread.
    thread_stats_t thread_stats;

    bool operator==(const sliced_stats_t &that) const;
};

/// Shared pointer to a sliced result
typedef shared_ptr<sliced_stats_t> sliced_stats_ptr_t;
typedef shared_ptr<const sliced_stats_t> sliced_stats_cptr_t;

/// Vector of sliced statistics.
typedef vector<sliced_stats_cptr_t> sliced_stats_vector_t;

// ----- stats_slicer_result_t ------------------------------------------------

class stats_slicer_result_t {
    friend class stats_slicer_t;

public:
    /// Return sliced statistics for a location set. If include_pfs is true
    /// fetches caused by software prefetches that are not hardware prefetched
    /// are counted as misses.
    template <class Cont>
    sliced_stats_cptr_t get(const Cont &locs, bool include_pfs) const {
	locationx_set_t lx_set = lx_mapping->get_locationx_set(locs);
	return get(lx_set, include_pfs);
    }

    /// Return sliced statistics for a location. If include_pfs is true fetches
    /// caused by software prefetches that are not hardware prefetched are
    /// counted as misses.
    sliced_stats_cptr_t get(location_t loc, bool include_pfs) const;

private:
    stats_slicer_result_t(
        locationx_mapper_result_cptr_t lx_mapping,
        statistics_result_cptr_t statistics);

    sliced_stats_cptr_t get(const locationx_set_t &lx_set, bool include_pfs) const;

    const locationx_mapper_result_cptr_t lx_mapping;
    const statistics_result_cptr_t statistics;
};

typedef shared_ptr<stats_slicer_result_t> stats_slicer_result_ptr_t;
typedef shared_ptr<const stats_slicer_result_t> stats_slicer_result_cptr_t;

// ----- stats_slicer_t -------------------------------------------------------

/// Module for statistics calculation.
class stats_slicer_t : public feature_t<stats_slicer_result_cptr_t> {
public:
    static stats_slicer_result_cptr_t calculate(
        const locationx_mapper_result_cptr_t &lx_mapping,
        const statistics_result_cptr_t &statistics);

    stats_slicer_t(const feature_context_ptr_t &context,
                   const string &sample_file,
                   const fnr_stack_depth_t stack_depth,
                   const fnr_line_size_t line_size,
                   const thread_mapping_cptr_t &thread_mapping,
		   const cpu_model_t &cpu_model,
                   const fnr_cache_size_t cache_size,
                   const repl_fr_engine_cptr_t &engine,
		   const windowing_engine_cptr_t &windowing,
                   const fnr_line_size_t page_size,
                   const int pos_streams,
                   const int neg_streams,
                   const fnr_line_count_t prefetch_reach,
		   const bool use_old_hwpf_algorithm,
                   const bool hwpf_enabled);
    stats_slicer_result_cptr_t get() const;

protected:
    string name() const;
    stats_slicer_result_cptr_t calculate() const;

private:
    /// Sample file.
    const string sample_file;

    /// Call stack depth to use in the analysis.
    const fnr_stack_depth_t stack_depth;

    /// Cache line size.
    const fnr_line_size_t line_size;

    /// Thread mapping.
    const thread_mapping_cptr_t thread_mapping;

    /// CPU model
    const cpu_model_t cpu_model;

    /// Miss ratio calculation parameters.
    const fnr_cache_size_t cache_size;
    const repl_fr_engine_cptr_t engine;

    /// Windowing configuration
    const windowing_engine_cptr_t windowing;

    /// Page size that the hardware prefetcher operates on.
    const fnr_line_size_t page_size;

    /// The number of prefetch streams with positive and negative strides,
    /// respectively.
    const int pos_streams;
    const int neg_streams;

    /// The number of sequential cache lines that are prefetched when a
    /// prefetch stream is triggered.
    const fnr_line_count_t prefetch_reach;

    /// Use the old algorithm for calculating the hw prefetch ratio
    const bool use_old_hwpf_algorithm;

    const bool hwpf_enabled;
};



END_NAMESPACE;
END_NAMESPACE;

#endif
