/// @file analysis/structure/dep_tree_search.cc

// Copyright (c) 2006-2009 Acumem AB.  All rights reserved.

#include "dep_tree_search.hh"
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>

using namespace acumem::library;
using namespace boost;
using namespace std;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

static inline location_set_t
make_location_set(location_t location)
{
    location_set_t locations;
    locations.insert(location);
    return locations;
}

// ----- dep_tree_search_result_t ---------------------------------------------

location_set_t
dep_tree_search_result_t::locations() const
{
    location_set_t s;
    BOOST_FOREACH(const location_time_map_t::value_type &v,
		  location_time_map)
	s.insert(v.first);
    return s;
}

// ----- CostPolicy classes - used in tree search routines --------------------

/// Search policy with total reuse distance as limiter
struct cost_rdist_t {
    fnr_time_t tot_rdist;

    cost_rdist_t(fnr_time_t tot_rdist)
	: tot_rdist(tot_rdist) {}

    cost_rdist_t next(fnr_time_t rdist) const {
	return cost_rdist_t(tot_rdist + rdist + 1);
    }

    template <typename INTTYPE>
    bool operator < (INTTYPE i) const {
	return tot_rdist < i;
    }

    template <typename INTTYPE>
    bool operator <= (INTTYPE i) const {
	return tot_rdist <= i;
    }

    template <typename INTTYPE>
    operator INTTYPE() const {
	return tot_rdist;
    }
};

// Search policy with number of jumps as limiter
struct cost_jumps_t {
    fnr_sample_count_t jumps;

    cost_jumps_t(fnr_sample_count_t jumps)
	: jumps(jumps) {}

    cost_jumps_t next(fnr_time_t rdist) const {
	return cost_jumps_t(jumps + 1);
    }

    template <typename INTTYPE>
    bool operator < (INTTYPE i) const {
	return jumps < i;
    }

    template <typename INTTYPE>
    bool operator <= (INTTYPE i) const {
	return jumps <= i;
    }

    template <typename INTTYPE>
    operator INTTYPE() const {
	return jumps;
    }
};

// ----- DirectionPolicy classes - used in tree search routines ---------------

/// Direction policy to follow samples in the pc1 -> pc2 direction
struct forward_direction_t {
    typedef per_pc2_t value_type;
    static const per_pc2_vector_t &next_pcs(const pc_info_t &pc_info)
	{ return pc_info.per_pc2; }
    static const location_t &location(const value_type &t)
	{ return t.pc2; }
};

/// Direction policy to follow samples in the pc2 -> pc1 direction
struct backward_direction_t {
    typedef per_pc1_t value_type;
    static const per_pc1_vector_t &next_pcs(const pc_info_t &pc_info)
	{ return pc_info.per_pc1; }
    static const location_t &location(const value_type &t)
	{ return t.pc1; }
};

// ----- SearchCriterion classes - used in tree search routines ---------------

struct search_criterion_t {
    enum action_t{no_action = 0,
		  log_action = 1,
		  recurse_action = 2,
		  log_recurse_action = 3};
};

/// Search for accesses matching a certain access mask
struct search_access_type_t : public search_criterion_t {
    search_access_type_t(access_lookup_result_cptr_t al_res,
			 const fnr_access_filter_t access_filter)
	: al_res(al_res),
	  access_filter(access_filter)
	{}

    action_t action(const location_t &location, int current_depth) const
	{
	    loc_access_info_map_t::const_iterator access_info_it =
		al_res->loc_access_info.find(location);

	    // The location that we look up must be known and possess
	    // a recorded access type.
	    assert(access_info_it != al_res->loc_access_info.end());

	    // If this access type matches the search policy, then
	    // report it, and terminate the search. Otherwise recurse.
	    return access_filter(access_info_it->second.access_classes)
		? log_action : recurse_action;
	}

    access_lookup_result_cptr_t al_res;
    fnr_access_filter_t access_filter;
};

/// Search for branches reaching a certain location.
///
/// Ticket #236 describes a flaw in loop detection which indirectly
/// seem to cause the program to spend a lot of time searching for a
/// dependency graph to a location. The max_depth parameter is a
/// kludge fix to limit the time spent here.
struct search_location_t : public search_criterion_t {
    search_location_t(const location_set_t &location_set, int max_depth)
	: location_set(location_set), max_depth(max_depth)
	{}

    action_t action(const location_t &location, int current_depth) const
	{
	    if (current_depth > max_depth)
		return no_action;

	    // If we have found our target, log it, otherwise indicate
	    // to continue search.
	    return (location_set.find(location) != location_set.end())
		? log_action : recurse_action;
	}

    location_set_t location_set;
    int max_depth;
};

// ----- Generic tree search routines ----------------------------------------

/// Maps visited locations to the lowest cost to visit it so far.
typedef map<location_t, fnr_time_t> visit_cost_map_t;

/// Search environment, packaged into a struct so we don't have to
/// pass all the arguments separately to search().
struct search_env_t {
    search_env_t(pc_lookup_result_cptr_t pcl_res,
		 dep_tree_search_result_ptr_t res)
	: pcl_res(pcl_res), res(res)
	{
	}

    visit_cost_map_t visited;
    pc_lookup_result_cptr_t pcl_res;
    dep_tree_search_result_ptr_t res;
};

/// Generic depth-first seach routine. Find locations that answer to a
/// certain criterion, following samples in a direction dictated by a
/// second policy.
template <typename CostPolicy,
	  typename DirectionPolicy,
	  typename SearchCriterion>
static void
search_i(search_env_t &se,
	 location_t location,
	 const CostPolicy &current_cost,
	 const DirectionPolicy &direction_policy,
	 const SearchCriterion &search_criterion,
	 int current_depth = 0)
{
    // See if there are samples from the current locations
    pc_info_map_t::const_iterator pc_info_it =
        se.pcl_res->pc_info.find(location);
    if  (pc_info_it == se.pcl_res->pc_info.end())
	return;

    // Iterate over all links from this location. It is important that
    // the list of links away from this location is sorted in
    // ascending order of rdist. See Dijkstra.
    BOOST_FOREACH(const typename DirectionPolicy::value_type &next,
		  DirectionPolicy::next_pcs(pc_info_it->second)) {

	const location_t &next_loc = DirectionPolicy::location(next);
	CostPolicy next_cost = current_cost.next(next.min_rdist);

	visit_cost_map_t::const_iterator visit = se.visited.find(next_loc);
	if (visit == se.visited.end() || next_cost < visit->second) {
	    se.visited[next_loc] = next_cost;

	    typename SearchCriterion::action_t action =
		search_criterion.action(next_loc, current_depth);

	    if (action & SearchCriterion::log_action)
		se.res->location_time_map[next_loc] = (fnr_time_t) next_cost;

	    if (action & SearchCriterion::recurse_action)
		search_i(se, next_loc, next_cost,
			 direction_policy, search_criterion, current_depth + 1);
	}
    }
}

/// Set up search parameters and bootstrap the recursive tree search
/// based on access types.
template <typename CostPolicy>
static dep_tree_search_result_cptr_t
search_access(access_lookup_result_cptr_t access_lookup,
              pc_lookup_result_cptr_t pc_lookup,
              fnr_access_filter_t access_filter,
	      location_t location,
	      bool backwards,
	      CostPolicy start_cost)
{
    dep_tree_search_result_ptr_t res(new dep_tree_search_result_t);
    search_env_t search_env(pc_lookup, res);
    search_access_type_t search_access_type(access_lookup, access_filter);

    if (backwards)
        search_i(search_env, location, start_cost,
                 backward_direction_t(), search_access_type);
    else
        search_i(search_env, location, start_cost,
                 forward_direction_t(), search_access_type);

    return res;
}

// ----- dep_tree_search_access_t ---------------------------------------------

dep_tree_search_access_t::dep_tree_search_access_t(
    const feature_context_ptr_t &context,
    const string &sample_file,
    const fnr_stack_depth_t stack_depth,
    const fnr_line_size_t line_size,
    const fnr_line_size_t page_size,
    const thread_mapping_cptr_t &thread_mapping,
    const cpu_model_t &cpu_model,
    const mapped_sample_type_t sample_type,
    const fnr_access_filter_t access_filter,
    const bool backwards,
    const bool use_rdist,
    const location_set_t &start_locations,
    const fnr_time_t max_rdist)
    : feature_t<dep_tree_search_data_ptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      sample_type(sample_type),
      access_filter(access_filter),
      backwards(backwards),
      use_rdist(use_rdist),
      start_locations(start_locations),
      max_rdist(max_rdist)
{
}

dep_tree_search_access_t::dep_tree_search_access_t(
    const feature_context_ptr_t &context,
    const string &sample_file,
    const fnr_stack_depth_t stack_depth,
    const fnr_line_size_t line_size,
    const fnr_line_size_t page_size,
    const thread_mapping_cptr_t &thread_mapping,
    const cpu_model_t &cpu_model,
    const mapped_sample_type_t sample_type,
    const fnr_access_filter_t access_filter,
    const bool backwards,
    const bool use_rdist,
    const location_t start_location,
    const fnr_time_t max_rdist)
    : feature_t<dep_tree_search_data_ptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      sample_type(sample_type),
      access_filter(access_filter),
      backwards(backwards),
      use_rdist(use_rdist),
      start_locations(make_location_set(start_location)),
      max_rdist(max_rdist)
{
}

string
dep_tree_search_access_t::name() const {
    // We omit start_locations and max_rdist, since results for different
    // start_locations and max_rdist are cached in the same data object.
    return "dep_tree_search_access_t(" + sample_file + "," +
	lexical_cast<string>(stack_depth) + "," +
	lexical_cast<string>(line_size) + "," +
	lexical_cast<string>(page_size) + "," +
	thread_mapping->name() + "," +
	cpu_model.name() + "," +
	lexical_cast<string>(sample_type) + "," +
	lexical_cast<string>((uintptr_t)access_filter) + "," +
	(backwards ? "backwards" : "forwards") +
	(use_rdist ? "rdist" : "jumps");
}

dep_tree_search_data_ptr_t
dep_tree_search_access_t::calculate() const
{
    return dep_tree_search_data_ptr_t(new dep_tree_search_data_t);
}

dep_tree_search_result_cptr_t
dep_tree_search_access_t::calculate(
    dep_tree_search_data_t *data,
    const access_lookup_result_cptr_t &access_lookup,
    const pc_lookup_result_cptr_t &pc_lookup,
    const fnr_access_filter_t access_filter,
    const bool backwards,
    const bool use_rdist,
    const location_set_t *start_locations,
    const fnr_time_t max_rdist)
{
    dep_tree_search_result_ptr_t result(
        new dep_tree_search_result_t());

    BOOST_FOREACH(location_t location, *start_locations) {
        // Use cached result if available.
        dep_tree_search_data_t::iterator i = data->find(location);
        if (i != data->end()) {
            BOOST_FOREACH(location_time_map_t::value_type &v, i->second)
                if (v.second < max_rdist)
                    result->location_time_map[v.first] = v.second;
            continue;
        }

	// Perform a search and insert results in 'data'

	dep_tree_search_result_cptr_t s_ret;
	if (use_rdist)
	    s_ret = search_access(access_lookup, pc_lookup, access_filter,
				  location, backwards, cost_rdist_t(0));
	else
	    s_ret = search_access(access_lookup, pc_lookup, access_filter,
                                  location, backwards, cost_jumps_t(0));

	location_time_map_t &ltm = (*data)[location];
	BOOST_FOREACH(const location_time_map_t::value_type &v,
		      s_ret->location_time_map) {
	    ltm[v.first] = v.second;
	    if (v.second <= max_rdist)
		result->location_time_map[v.first] = v.second;
	}
    }

    return result;
}

dep_tree_search_result_cptr_t
dep_tree_search_access_t::get()
{
    access_lookup_result_cptr_t access_lookup =
        access_lookup_t(context, sample_file, stack_depth).get();

    pc_lookup_result_cptr_t pc_lookup(
        pc_lookup_t(context, sample_file, stack_depth, line_size, page_size,
		    thread_mapping, cpu_model, sample_type).get());

    return calculate(&*get_data(), access_lookup, pc_lookup, access_filter,
                     backwards, use_rdist, &start_locations, max_rdist);
}

// ----------------------------------------------------------------------

/// Set up search parameters and bootstrap the recursive tree search
/// with criteria of reacting a specific location
template <typename CostPolicy>
static dep_tree_search_result_cptr_t
search_location(pc_lookup_result_cptr_t pc_lookup,
                location_t start_location,
		const location_set_t *end_locations,
		bool backwards,
		CostPolicy start_cost)
{
    dep_tree_search_result_ptr_t res(new dep_tree_search_result_t);
    search_env_t search_env(pc_lookup, res);

    // 30 is experimental value for depth-first max depth cutoff. It
    // seems to reduce the immense calculation time for 445.gobmk to
    // something manageable.  See ticket #236 regarding a rewrite of
    // the loop detection algorithm. When this is done, we could try
    // to remove this limitation.
    search_location_t search_location(*end_locations, 30);

    if (backwards)
        search_i(search_env, start_location, start_cost,
                 backward_direction_t(), search_location);
    else
        search_i(search_env, start_location, start_cost,
                 forward_direction_t(), search_location);

    return res;
}

// ----- dep_tree_search_location_t -------------------------------------------

dep_tree_search_location_t::dep_tree_search_location_t(
    const feature_context_ptr_t &context,
    const string &sample_file,
    const fnr_stack_depth_t stack_depth,
    const fnr_line_size_t line_size,
    const fnr_line_size_t page_size,
    const thread_mapping_cptr_t &thread_mapping,
    const cpu_model_t &cpu_model,
    const mapped_sample_type_t sample_type,
    const bool backwards,
    const location_set_t &start_locations,
    const location_set_t &end_locations,
    const fnr_time_t max_rdist)
    : feature_t<dep_tree_search_data_ptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      sample_type(sample_type),
      backwards(backwards),
      start_locations(start_locations),
      end_locations(end_locations),
      max_rdist(max_rdist)
{
}

dep_tree_search_location_t::dep_tree_search_location_t(
    const feature_context_ptr_t &context,
    const string &sample_file,
    const fnr_stack_depth_t stack_depth,
    const fnr_line_size_t line_size,
    const fnr_line_size_t page_size,
    const thread_mapping_cptr_t &thread_mapping,
    const cpu_model_t &cpu_model,
    const mapped_sample_type_t sample_type,
    const bool backwards,
    const location_t start_location,
    const location_set_t &end_locations,
    const fnr_time_t max_rdist)
    : feature_t<dep_tree_search_data_ptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      sample_type(sample_type),
      backwards(backwards),
      start_locations(make_location_set(start_location)),
      end_locations(end_locations),
      max_rdist(max_rdist)
{
}

string
dep_tree_search_location_t::name() const {
    // We omit start_locations, end_locations and max_rdist, since results for
    // different start_locations, end_locations and max_rdist are cached in the
    // same data object.
    return "dep_tree_search_location_t(" + sample_file + "," +
	lexical_cast<string>(stack_depth) + "," +
	lexical_cast<string>(line_size) + "," +
	lexical_cast<string>(page_size) + "," +
	thread_mapping->name() + "," +
	cpu_model.name() + "," +
	lexical_cast<string>(sample_type) + "," +
	(backwards ? "backwards" : "forwards");
}

// internal function to handle all access searches
dep_tree_search_result_cptr_t
dep_tree_search_location_t::calculate(
    dep_tree_search_data_t *data,
    const pc_lookup_result_cptr_t &pc_lookup,
    const bool backwards,
    const location_set_t *start_locations,
    const location_set_t *end_locations,
    const fnr_time_t max_rdist)
{
    dep_tree_search_result_ptr_t result(
        new dep_tree_search_result_t());

    BOOST_FOREACH(location_t location, *start_locations) {
        // Use cached result if available.
        dep_tree_search_data_t::iterator i = data->find(location);
        if (i != data->end()) {
            BOOST_FOREACH(location_time_map_t::value_type &v, i->second)
                if (v.second < max_rdist)
                    result->location_time_map[v.first] = v.second;
            continue;
        }

	// Perform a search and insert results in 'data' and update 'res'

	dep_tree_search_result_cptr_t s_ret =
	    search_location(pc_lookup, location, end_locations, backwards,
			    cost_rdist_t(0));

	location_time_map_t &ltm = (*data)[location];
	BOOST_FOREACH(const location_time_map_t::value_type &v,
		      s_ret->location_time_map) {
	    ltm[v.first] = v.second;
	    if (v.second < max_rdist)
		result->location_time_map[v.first] = v.second;
	}
    }

    return result;
}

dep_tree_search_result_cptr_t
dep_tree_search_location_t::get()
{
    pc_lookup_result_cptr_t pc_lookup(
        pc_lookup_t(context, sample_file, stack_depth, line_size, page_size,
		    thread_mapping, cpu_model, sample_type).get());

    return calculate(&*get_data(), pc_lookup, backwards, &start_locations,
                     &end_locations, max_rdist);
}

dep_tree_search_data_ptr_t
dep_tree_search_location_t::calculate() const
{
    return dep_tree_search_data_ptr_t(new dep_tree_search_data_t);
}

END_NAMESPACE;
END_NAMESPACE;
