/// @file analysis/structure/dep_tree_search.hh

// Copyright (c) 2006-2008 Acumem AB.  All rights reserved.

#ifndef DEP_TREE_SEARCH_HH
#define DEP_TREE_SEARCH_HH

#include "analysis/structure/access_lookup.hh"
#include "analysis/structure/pc_lookup.hh"
#include "analysis/filter/filter_module.hh"
#include "analysis/feature_base/feature_base.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace acumem::library;

using boost::shared_ptr;

// ----- dep_tree_search_result_t ---------------------------------------------

/// Search result.
typedef map<location_t, fnr_time_t> location_time_map_t;
typedef map<location_t, location_time_map_t> dep_tree_search_access_data_t;

struct dep_tree_search_result_t {
    /// The locations found by the search, and the minimum cost to reach them
    location_time_map_t location_time_map;
    location_set_t locations() const;
};

typedef shared_ptr<dep_tree_search_result_t> dep_tree_search_result_ptr_t;
typedef shared_ptr<const dep_tree_search_result_t> dep_tree_search_result_cptr_t;

// ----- dep_tree_search_access_t ---------------------------------------------

typedef map<location_t, location_time_map_t> dep_tree_search_data_t;
typedef shared_ptr<dep_tree_search_data_t> dep_tree_search_data_ptr_t;

/// Finds the previous or next locations accessing the same data with one of a
/// set of access types for a location set.
class dep_tree_search_access_t
    : public feature_t<dep_tree_search_data_ptr_t>
{
public:
    static dep_tree_search_result_cptr_t calculate(
        dep_tree_search_data_t *data,
        const access_lookup_result_cptr_t &access_lookup,
        const pc_lookup_result_cptr_t &pc_lookup,
        const fnr_access_filter_t access_filter,
        const bool backwards,
        const bool use_rdist,
        const location_set_t *start_locations,
        const fnr_time_t max_rdist);

    /// Find the previous or next locations with the specified access
    /// type touching the same data as any of the specified
    /// locations. Limit the search to locations that access the data
    /// within the specified number of accesses before or after it is
    /// accessed by a start location. This function is also used to
    /// use the number of jumps as a cost function (depending on the
    /// conf.use_rdist variable)
    dep_tree_search_access_t(const feature_context_ptr_t &context,
                             const string &sample_file,
                             const fnr_stack_depth_t stack_depth,
                             const fnr_line_size_t line_size,
                             const fnr_line_size_t page_size,
			     const thread_mapping_cptr_t &thread_mapping,
			     const cpu_model_t &cpu_model,
			     const mapped_sample_type_t sample_type,
                             const fnr_access_filter_t access_filter,
                             const bool backwards,
                             const bool use_rdist,
                             const location_set_t &start_locations,
                             const fnr_time_t max_rdist);

    /// Find the previous or next locations with the specified access
    /// type touching the same data as the specified location. Limit
    /// the search to locations that access the data within the
    /// specified number of accesses before or after it is accessed by
    /// a start location. This function is also used to use the number
    /// of jumps as a cost function (depending on the conf.use_rdist
    /// variable)
    dep_tree_search_access_t(const feature_context_ptr_t &context,
                             const string &sample_file,
                             const fnr_stack_depth_t stack_depth,
                             const fnr_line_size_t line_size,
                             const fnr_line_size_t page_size,
			     const thread_mapping_cptr_t &thread_mapping,
			     const cpu_model_t &cpu_model,
			     const mapped_sample_type_t sample_type,
                             const fnr_access_filter_t access_filter,
                             const bool backwards,
                             const bool use_rdist,
                             const location_t start_location,
                             const fnr_time_t max_rdist);

    dep_tree_search_result_cptr_t get();

protected:
    string name() const;
    dep_tree_search_data_ptr_t calculate() const;

private:
    /// Sample file.
    const string sample_file;

    /// Call stack depth to use in the analysis.
    const fnr_stack_depth_t stack_depth;

    /// This is the line size that we are interested in.
    const fnr_line_size_t line_size;

    /// Page size, required by sample_mapping_t.
    const fnr_line_size_t page_size;

    /// Thread to cache mappings.
    const thread_mapping_cptr_t thread_mapping;

    /// CPU model
    const cpu_model_t cpu_model;

    /// Sample type to use
    const mapped_sample_type_t sample_type;

    /// These are the access classes we are interested in.
    const fnr_access_filter_t access_filter;

    /// This is the direction we search in
    const bool backwards;

    /// This tells us whether we look at reuse distance or number of jumps
    const bool use_rdist;

    const location_set_t start_locations;
    const fnr_time_t max_rdist;
};

// ----- dep_tree_search_location_t -------------------------------------------

/// Follow a tree (backwards or forwards) from a starting location,
/// trying to reach target locations, subject to a maximum cost
class dep_tree_search_location_t
    : public feature_t<dep_tree_search_data_ptr_t>
{
public:
    static dep_tree_search_result_cptr_t calculate(
        dep_tree_search_data_t *data,
        const pc_lookup_result_cptr_t &pc_lookup,
        const bool backwards,
        const location_set_t *start_locations,
        const location_set_t *end_locations,
        const fnr_time_t max_rdist);

    /// Find which ones of the locations in the end_locations set are
    /// reachable from the locations in the start_locations set in
    /// less than max_rdist memory accesses, going forward/backward.
    dep_tree_search_location_t(const feature_context_ptr_t &context,
                               const string &sample_file,
                               const fnr_stack_depth_t stack_depth,
                               const fnr_line_size_t line_size,
                               const fnr_line_size_t page_size,
			       const thread_mapping_cptr_t &thread_mapping,
			       const cpu_model_t &cpu_model,
			       const mapped_sample_type_t sample_type,
                               const bool backwards,
                               const location_set_t &start_locations,
                               const location_set_t &end_locations,
                               const fnr_time_t max_rdist);

    /// Find which ones of the locations in the end_locations set are
    /// reachable from the 'start_location' in less than max_rdist
    /// memory accesses, going forward/backward.
    dep_tree_search_location_t(const feature_context_ptr_t &context,
                               const string &sample_file,
                               const fnr_stack_depth_t stack_depth,
                               const fnr_line_size_t line_size,
                               const fnr_line_size_t page_size,
			       const thread_mapping_cptr_t &thread_mapping,
			       const cpu_model_t &cpu_model,
			       const mapped_sample_type_t sample_type,
                               const bool backwards,
                               const location_t start_location,
                               const location_set_t &end_locations,
                               const fnr_time_t max_rdist);

    dep_tree_search_result_cptr_t get();

protected:
    string name() const;
    dep_tree_search_data_ptr_t calculate() const;

private:
    /// Sample file.
    const string sample_file;

    /// Call stack depth to use in the analysis.
    const fnr_stack_depth_t stack_depth;

    /// This is the line size that we are interested in.
    const fnr_line_size_t line_size;

    /// Page size, required by sample_mapping_t.
    const fnr_line_size_t page_size;

    /// Thread to cache mappings.
    const thread_mapping_cptr_t thread_mapping;

    /// CPU model
    const cpu_model_t cpu_model;

    /// Sample type to use
    const mapped_sample_type_t sample_type;

    /// This is the direction we search in
    const bool backwards;

    const location_set_t start_locations;
    const location_set_t end_locations;
    const fnr_time_t max_rdist;
};




END_NAMESPACE;
END_NAMESPACE;

#endif
