/// @file analysis/structure/dependence_group.cc

// Copyright (c) 2006-2009 Acumem AB.  All rights reserved.

/// Needed on Fedora 9. <boost/graph/adjacency_list.hpp> includes
/// <ext/hash_set> that is marked as deprecated in GCC 4.3.0.
#undef __DEPRECATED

#include "analysis/structure/dependence_group.hh"
#include "analysis/structure/pc_lookup.hh"
#include "analysis/filter/filter_module.hh"
#include "library/fenris/fenris.hh"
#include <boost/foreach.hpp>
#include <boost/format.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/connected_components.hpp>
#include <list>
#include <sstream>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace boost;
using namespace acumem::library;

// ----------------------------------------------------------------------

dependence_group_t::dependence_group_t(
    const feature_context_ptr_t &context,
    const string &sample_file,
    const fnr_stack_depth_t stack_depth,
    const fnr_line_size_t line_size,
    const fnr_line_size_t page_size,
    const thread_mapping_cptr_t &thread_mapping,
    const cpu_model_t &cpu_model,
    const mapped_sample_type_t sample_type,
    const int limit_factor)
    : feature_t<dependence_group_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      sample_type(sample_type),
      limit_factor(limit_factor),
      use_pc_set(false)
{
}

dependence_group_t::dependence_group_t(
    const feature_context_ptr_t &context,
    const string &sample_file,
    const fnr_stack_depth_t stack_depth,
    const fnr_line_size_t line_size,
    const fnr_line_size_t page_size,
    const thread_mapping_cptr_t &thread_mapping,
    const cpu_model_t &cpu_model,
    const mapped_sample_type_t sample_type,
    const int limit_factor,
    const location_set_t &pc_set)
    : feature_t<dependence_group_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      sample_type(sample_type),
      limit_factor(limit_factor),
      pc_set(pc_set),
      use_pc_set(true)
{
}

typedef adjacency_list<setS, vecS, undirectedS> dep_graph_t;
typedef dep_graph_t::vertex_descriptor dep_graph_vertex_t;

typedef map<location_t, dep_graph_vertex_t> pc_vertex_map_t;
typedef map<dep_graph_vertex_t, location_t> vertex_pc_map_t;

struct stuff_t {
    stuff_t() : max_count(0) {}

    dep_graph_t graph;

    pc_vertex_map_t pc_vertex_map;
    vertex_pc_map_t vertex_pc_map;

    typedef pair<dep_graph_vertex_t, dep_graph_vertex_t> edge_t;
    typedef map<edge_t, fnr_sample_count_t> edge_count_map_t;
    edge_count_map_t count;
    fnr_sample_count_t max_count;
};

static dep_graph_vertex_t
add_pc(stuff_t *stuff, location_t pc)
{
    dep_graph_vertex_t vertex;
    pc_vertex_map_t::const_iterator vertex_it = stuff->pc_vertex_map.find(pc);
    if (vertex_it == stuff->pc_vertex_map.end()) {
        vertex = add_vertex(stuff->graph);
        stuff->pc_vertex_map.insert(make_pair(pc, vertex));
        stuff->vertex_pc_map.insert(make_pair(vertex, pc));
    }
    else
        vertex = vertex_it->second;
    return vertex;
}

static void
add(stuff_t *stuff,
    location_t pc1,
    location_t pc2,
    fnr_sample_count_t count)
{
    dep_graph_vertex_t pc1_vertex = add_pc(stuff, pc1);
    dep_graph_vertex_t pc2_vertex = add_pc(stuff, pc2);

    fnr_sample_count_t &this_count =
        stuff->count[make_pair(pc1_vertex, pc2_vertex)];
    this_count += count;
    stuff->max_count = max(stuff->max_count, this_count);
}

static dependence_group_result_cptr_t
calculate_common(stuff_t *stuff, int limit_factor)
{
    dependence_group_result_ptr_t result(new dependence_group_result_t());

    // Add edges between all pc1s and pc2s with a large enough sample count.
    stuff->max_count /= limit_factor;
    for (stuff_t::edge_count_map_t::iterator
             ei = stuff->count.begin(), ee = stuff->count.end();
         ei != ee; ei++)
	if (ei->second > stuff->max_count)
	    add_edge(ei->first.first, ei->first.second, stuff->graph);

    // Find connected components (dependence groups) in the graph.
    if (num_vertices(stuff->graph) > 0) {
	vector<int> component(num_vertices(stuff->graph));
	int num = connected_components(stuff->graph, &component[0]);

	result->pc_sets.resize(num);

	for (dep_graph_vertex_t v = 0; v < component.size(); v++)
	    result->pc_sets[component[v]].insert(stuff->vertex_pc_map[v]);
    }

    return result;
}

dependence_group_result_cptr_t
dependence_group_t::calculate(
    const pc_lookup_result_cptr_t &pc_lookup,
    const int limit_factor)
{
    stuff_t stuff;

    // Add vertices for all pc1s and pc2s, and count the number of samples
    // between each pc1 and pc2.
    BOOST_FOREACH(const pc_info_map_t::value_type &v1, pc_lookup->pc_info) {
        location_t pc2 = v1.first;
        BOOST_FOREACH(const per_pc1_t &v2, v1.second.per_pc1) {
            location_t pc1 = v2.pc1;
            add(&stuff, pc1, pc2, (fnr_sample_count_t)v2.samples.size());
        }
    }

    return calculate_common(&stuff, limit_factor);
}

dependence_group_result_cptr_t
dependence_group_t::calculate(
    const pc_lookup_result_cptr_t &pc_lookup,
    const int limit_factor,
    const location_set_t &pc_set)
{
    stuff_t stuff;

    // Add vertices for all pc1s and pc2s, and count the number of samples
    // between each pc1 and pc2.
    BOOST_FOREACH(const location_t pc2, pc_set) {
        add_pc(&stuff, pc2);
        pc_info_map_t::const_iterator i = pc_lookup->pc_info.find(pc2);
        if (i != pc_lookup->pc_info.end()) {
            BOOST_FOREACH(const per_pc1_t &v, i->second.per_pc1) {
                const location_t pc1 = v.pc1;
                if (pc_set.find(pc1) != pc_set.end())
                    add(&stuff, pc1, pc2, (fnr_sample_count_t)v.samples.size());
            }
        }
    }

    return calculate_common(&stuff, limit_factor);
}

string
dependence_group_t::name() const
{
    stringstream pcs;
    if (use_pc_set)
	pcs << pc_set;
    else
	pcs << "all";

    return "dependence_group(" +
        sample_file + "," +
        lexical_cast<string>(stack_depth) + "," +
        lexical_cast<string>(line_size) + "," +
        lexical_cast<string>(page_size) + "," +
	thread_mapping->name() + "," +
	cpu_model.name() + "," +
	lexical_cast<string>(sample_type) + "," +
        lexical_cast<string>(limit_factor) + "," +
        pcs.str() + ")";
}

dependence_group_result_cptr_t
dependence_group_t::get()
{
    return get_data();
}

dependence_group_result_cptr_t
dependence_group_t::calculate() const
{
    // Get pc lookup.
    pc_lookup_result_cptr_t pc_lookup(
        pc_lookup_t(context, sample_file, stack_depth, line_size, page_size,
		    thread_mapping, cpu_model, sample_type).get());

    if (use_pc_set)
        return calculate(pc_lookup, limit_factor, pc_set);
    else
        return calculate(pc_lookup, limit_factor);
}

END_NAMESPACE;
END_NAMESPACE;
