/// @file analysis/structure/dependence_group.hh

// Copyright (c) 2006-2008 Acumem AB.  All rights reserved.

#ifndef DEPENDENCE_GROUP_HH
#define DEPENDENCE_GROUP_HH

#include "analysis/structure/pc_lookup.hh"
#include "analysis/call_stack/stack_pc_map.hh"
#include "analysis/thread_mapping/sample_mapping.hh"
#include "analysis/feature_base/feature_base.hh"
#include "library/lang/common_macros.h"
#include <boost/shared_ptr.hpp>
#include <set>
#include <vector>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace std;

using boost::shared_ptr;

// ----- dependence_group_result_t --------------------------------------------

/// This struct contains the result of dependence_group_t::get() - a
/// vector of pc sets, one set for each dependence group.
struct dependence_group_result_t {
    location_set_vector_t pc_sets;
};

typedef shared_ptr<dependence_group_result_t> dependence_group_result_ptr_t;
typedef shared_ptr<const dependence_group_result_t> dependence_group_result_cptr_t;

// ----- dependence_group_t ---------------------------------------------------

/// Module for calculating a set of dependence groups for a pc-set.
class dependence_group_t : public feature_t<dependence_group_result_cptr_t> {
public:
    static dependence_group_result_cptr_t calculate(
        const pc_lookup_result_cptr_t &pc_lookup,
        const int limit_factor);
    static dependence_group_result_cptr_t calculate(
        const pc_lookup_result_cptr_t &pc_lookup,
        const int limit_factor,
        const location_set_t &pc_set);

    dependence_group_t(const feature_context_ptr_t &context,
                       const string &sample_file,
                       const fnr_stack_depth_t stack_depth,
                       const fnr_line_size_t line_size,
                       const fnr_line_size_t page_size,
		       const thread_mapping_cptr_t &thread_mapping,
		       const cpu_model_t &cpu_model,
		       const mapped_sample_type_t sample_type,
                       const int limit_factor);
    dependence_group_t(const feature_context_ptr_t &context,
                       const string &sample_file,
                       const fnr_stack_depth_t stack_depth,
                       const fnr_line_size_t line_size,
                       const fnr_line_size_t page_size,
		       const thread_mapping_cptr_t &thread_mapping,
		       const cpu_model_t &cpu_model,
		       const mapped_sample_type_t sample_type,
                       const int limit_factor,
                       const location_set_t &pc_set);
    dependence_group_result_cptr_t get();

protected:
    string name() const;
    dependence_group_result_cptr_t calculate() const;

private:
    /// Sample file.
    const string sample_file;

    // Call stack depth to use in the analysis.
    const fnr_stack_depth_t stack_depth;

    /// Cache line size.
    const fnr_line_size_t line_size;

    /// Page size.
    const fnr_line_size_t page_size;

    /// Thread to cache mappings.
    const thread_mapping_cptr_t thread_mapping;

    /// CPU model
    const cpu_model_t cpu_model;

    /// Sample type to use
    const mapped_sample_type_t sample_type;

    /// Don't include edges with less samples than max/limit_factor.
    const int limit_factor;

    /// Tell which pc:s we are interested in
    const location_set_t pc_set;

    /// Use the pc_set, or take all addresses?
    const bool use_pc_set;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
