/// @file analysis/structure/loop_identification.cc

// Copyright (c) 2006-2009 Acumem AB.  All rights reserved.

// For a description of the algorithm see doc/notes/loop_identification.txt.

#include <boost/lexical_cast.hpp>
#include <boost/multi_index_container.hpp>
#include <boost/multi_index/member.hpp>
#include <boost/multi_index/identity.hpp>
#include <boost/multi_index/ordered_index.hpp>
#include <boost/multi_index/tag.hpp>
#include <boost/foreach.hpp>

#include "loop_identification.hh"

using namespace std;
using namespace boost;
using namespace boost::multi_index;
using namespace acumem::library;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

// ----- Comparison operators for cf_edge_t. ----------------------------------

bool
operator==(const cf_edge_t &x, const cf_edge_t &y) {
    return x.sample_count == y.sample_count && x.cf_node->id == y.cf_node->id;
}

bool
operator<(const cf_edge_t &x, const cf_edge_t &y) {
    return (x.sample_count < y.sample_count ||
            (x.sample_count == y.sample_count &&
             x.cf_node->id < y.cf_node->id));
}

// ----- cf_edge_map_t --------------------------------------------------------

/// Index tags for cf_edge_map_t.
struct by_node {};
struct by_count {};

/// Mapping between control flow nodes and edge counts.
typedef multi_index_container<
    cf_edge_t,
    indexed_by<
        ordered_unique<
            tag<by_node>,
            member<cf_edge_t, cf_node_ptr_t, &cf_edge_t::cf_node> >,
        ordered_unique<
            tag<by_count>,
            identity<cf_edge_t> >
    >
> cf_edge_map_t;

/// Indexes for cf_edge_map_t.
typedef cf_edge_map_t::index<by_node>::type by_node_t;
typedef cf_edge_map_t::index<by_count>::type by_count_t;

/// Function object that adds a sample count to a control flow edge.
struct add_edge_samples_op_t : public unary_function<void, cf_edge_t &> {
    add_edge_samples_op_t(double sample_count)
        : sample_count(sample_count) {}

    double sample_count;

    void operator()(cf_edge_t &edge) {
        edge.sample_count += sample_count;
    }
};

/// Vector of control flow edge maps, indexed by control flow node id.
typedef vector<cf_edge_map_t> cf_edge_map_vector_t;

// ----- env_t ----------------------------------------------------------------

/// Data needed during loop identification that can be free once we have the
/// result.
struct env_t {
    /// Automatically populates the cf_nodes, outermost_node, exits and entries
    /// vectors for all locations.
    env_t(loop_vector_t *loops,
          stack_pc_map_result_cptr_t spm_res);

    /// Resulting loop vector, indexed by loop id.
    loop_vector_t *loops;

    /// Control flow node vector, indexed by control flow node id.
    cf_node_vector_t cf_nodes;

    /// Vector with the outermost control flow node containing each location,
    /// indexed by node id.
    cf_node_vector_t outermost_node;

    /// Vector with a control flow edge map of edges leading from the node for
    /// each control flow node, indexed by control flow node id.
    cf_edge_map_vector_t exits;

    /// A set of all control flow nodes with edges to the node for each control
    /// flow node. Indexed by control flow node id. Needed for efficient
    /// control flow node replacements.
    vector<set<cf_node_ptr_t> > entries;

    /// Mapping between control flow nodes and their maximum exit sample count,
    /// for each control flow node not included in a loop and that we have not
    /// tried and failed to start a loop search from.
    cf_edge_map_t start_nodes;
};

env_t::env_t(loop_vector_t *loops,
             stack_pc_map_result_cptr_t spm_res)
    : loops(loops)
{
    for (location_t loc = 0; loc < spm_res->stack_id_pc_map.size(); loc++) {
        cf_node_ptr_t cf_node(new cf_node_t(loc, loc));
        cf_nodes.push_back(cf_node);
        outermost_node.push_back(cf_node);
        exits.push_back(cf_edge_map_t());
        entries.push_back(set<cf_node_ptr_t>());
    }
}

// ----- Control flow edge operations. ----------------------------------------

/// Add a control flow edge with the specified sample count between the
/// specified nodes, or if such an edge already exists add the sample count to
/// the edge.
static void
add_cf_edge(env_t *env, cf_node_ptr_t from, cf_node_ptr_t to, double sample_count)
{
    by_node_t *exits_by_node = &env->exits[from->id].get<by_node>();
    by_node_t::iterator it = exits_by_node->find(to);
    if (it == exits_by_node->end()) {
        exits_by_node->insert(cf_edge_t(to, sample_count));
        env->entries[to->id].insert(from);
    }
    else
        exits_by_node->modify(it, add_edge_samples_op_t(sample_count));
}

/// Remove the control flow edge between the specified control flow nodes.
static double
remove_cf_edge(env_t *env, cf_node_ptr_t from, cf_node_ptr_t to)
{
    by_node_t *exits_by_node = &env->exits[from->id].get<by_node>();
    by_node_t::iterator it = exits_by_node->find(to);
    assert(it != exits_by_node->end());
    double sample_count = it->sample_count;
    exits_by_node->erase(it);
    env->entries[to->id].erase(from);
    return sample_count;
}

/// Replace all control flow edges going to old_node with control flow edges
/// going to new_node.
static void
replace_cf_node(env_t *env, cf_node_ptr_t old_node, cf_node_ptr_t new_node)
{
    set<cf_node_ptr_t> *refs = &env->entries[old_node->id];
    while (!refs->empty()) {
        cf_node_ptr_t from = *refs->begin();
        double sample_count = remove_cf_edge(env, from, old_node);
        add_cf_edge(env, from, new_node, sample_count);
    }
}

// ----- loop_t ---------------------------------------------------------------

loop_t::loop_t(loop_id_t id)
    : id(id)
{
}

bool
loop_t::has_child_loop(loop_ptr_t loop, bool recursive) const
{
    BOOST_FOREACH(const cf_edge_t &edge, cf_edges)
        switch (edge.cf_node->type) {
        case CF_Node_Location:
            break;
        case CF_Node_Loop:
            if (edge.cf_node->loop == loop ||
                (recursive &&
                 edge.cf_node->loop->has_child_loop(loop, recursive)))
                return true;
            break;
        default:
            assert(0);
        }
    return false;
}

bool
loop_t::has_child_location(location_t location, bool recursive) const
{
    BOOST_FOREACH(const cf_edge_t &edge, cf_edges)
        switch (edge.cf_node->type) {
        case CF_Node_Location:
            if (edge.cf_node->location == location)
                return true;
            break;
        case CF_Node_Loop:
            if (recursive &&
                edge.cf_node->loop->has_child_location(location, recursive))
                return true;
            break;
        default:
            assert(0);
        }
    return false;
}

static void
collect_loops(const loop_t *loop, bool recursive, loop_vector_t *loops)
{
    BOOST_FOREACH(const cf_edge_t &edge, loop->cf_edges)
        switch (edge.cf_node->type) {
        case CF_Node_Location:
            break;
        case CF_Node_Loop:
            loops->push_back(edge.cf_node->loop);
            if (recursive)
                collect_loops(&*edge.cf_node->loop, recursive, loops);
            break;
        default:
            assert(0);
        }
}

loop_vector_t
loop_t::get_loops(bool recursive) const
{
    loop_vector_t loops;

    collect_loops(this, recursive, &loops);

    return loops;
}

static void
collect_locs(const loop_t *loop, bool recursive,
	     location_set_t *locs, const loop_t *exclude_loop)
{
    BOOST_FOREACH(const cf_edge_t &edge, loop->cf_edges)
        switch (edge.cf_node->type) {
        case CF_Node_Location:
            locs->insert(edge.cf_node->location);
            break;
        case CF_Node_Loop:
            if (recursive && ! (exclude_loop &&
				exclude_loop->id == edge.cf_node->loop->id))
		collect_locs(&*edge.cf_node->loop, recursive,
			     locs, exclude_loop);

	    break;
        default:
	    assert(0);
        }
}

location_set_t
loop_t::get_locations(bool recursive) const
{
    location_set_t locs;

    collect_locs(this, recursive, &locs, NULL);

    return locs;
}

location_set_t
loop_t::get_locations_excluding_loop(loop_ptr_t exclude_loop) const
{
    location_set_t locs;

    collect_locs(this, true, &locs,
		 exclude_loop.get() ? &*exclude_loop : NULL);

    return locs;
}

// ----- loop_identification_t ------------------------------------------------

/// Set the outermost node of the specified node and all of its child nodes
/// (recursively) to outermost_node.
static void
set_outermost_node(env_t *env, cf_node_ptr_t node,
                   cf_node_ptr_t outermost_node)
{
    env->outermost_node[node->id] = outermost_node;
    switch (node->type) {
    case CF_Node_Location:
        break;
    case CF_Node_Loop:
        BOOST_FOREACH(cf_edge_t &edge, node->loop->cf_edges)
            set_outermost_node(env, edge.cf_node, outermost_node);
        break;
    default:
        assert(0);
    }
}

/// Performs a back-tracking depth first search for loops, starting from the
/// specified node. Returns true if a loop was found, false it not.
///
/// visited   True for nodes that we have visited, indexed by node id.
/// trace     The currently followed code path.
/// in_trace  True for nodes on the currently followed code path, indexed by
///           node id.
static bool
loop_search(env_t *env,
            vector<bool> *visited,
            cf_edge_vector_t *trace,
            vector<bool> *in_trace,
            cf_node_ptr_t node)
{
    // Try all exits from the loop, the one most samples first.
    by_count_t *exits_by_count = &env->exits[node->id].get<by_count>();
    for (by_count_t::const_reverse_iterator
             ei = exits_by_count->rbegin(), ee = exits_by_count->rend();
         ei != ee; ei++) {
        // Add the edge to the currently followed code path
        trace->push_back(*ei);

        cf_node_ptr_t next_node = ei->cf_node;

        // Check if the exit completes a loop.
        if (in_trace->at(next_node->id)) {
            // Remove the followed code path up to the start of the loop.
            while (trace->front().cf_node != next_node)
                trace->erase(trace->begin());
            trace->erase(trace->begin());

            // Create the loop.
            loop_id_t new_loop_id = (loop_id_t)(env->loops->size() + 1);
            loop_ptr_t new_loop(new loop_t(new_loop_id));
            env->loops->push_back(new_loop);

            // Create a control flow node for the loop.
            cf_node_id_t new_node_id = (cf_node_id_t)env->cf_nodes.size();
            cf_node_ptr_t new_node(new cf_node_t(new_node_id, new_loop));
            env->cf_nodes.push_back(new_node);
            env->exits.push_back(cf_edge_map_t());
            env->entries.push_back(set<cf_node_ptr_t>());
            cf_edge_map_t *exits = &env->exits.back();
            env->outermost_node.push_back(new_node);

            // Iterate over all control flow edges in the loop and add them to
            // the loop.
            BOOST_FOREACH(cf_edge_t &edge, *trace)
                new_loop->cf_edges.push_back(edge);

            // Set this loop as the outermost loop of all its child nodes.
            set_outermost_node(env, new_node, new_node);

            // Replace all control flow edges leading to the child nodes of the
            // loop with control flow edges to the loop.
            BOOST_FOREACH(cf_edge_t &edge, *trace)
                replace_cf_node(env, edge.cf_node, new_node);

            // Remove the child nodes of this loop from the start nodes map.
            by_node_t *start_nodes_by_node = &env->start_nodes.get<by_node>();
            BOOST_FOREACH(cf_edge_t &edge, *trace)
                start_nodes_by_node->erase(edge.cf_node);

            // Add all exits of the child nodes that do not lead back to this
            // to this loop to this loop.
            BOOST_FOREACH(cf_edge_t &edge, *trace) {
                by_node_t *child_exits_by_node =
                    &env->exits[edge.cf_node->id].get<by_node>();
                while (!child_exits_by_node->empty()) {
                    by_node_t::iterator it = child_exits_by_node->begin();
                    if (it->cf_node != new_node)
                        add_cf_edge(env, new_node, it->cf_node, it->sample_count);
                    remove_cf_edge(env, edge.cf_node, it->cf_node);
                }
            }

            // We should try to start new loop searches from this loop if it
            // has any exits.
            if (!exits->empty())
                env->start_nodes.insert(
                    cf_edge_t(new_node,
                              exits->get<by_count>().rbegin()->sample_count));

            return true;
        }

        // Add the control flow node to the followed code path.
        (*in_trace)[ei->cf_node->id] = true;

        // Continue searching along the exit, unless we have already visited
        // the node it leads to.
        if (!visited->at(next_node->id) &&
            loop_search(env, visited, trace, in_trace, next_node))
            return true;

        // No loop found along this edge, remove it from the currently followed
        // code path.
        trace->pop_back();
        (*in_trace)[ei->cf_node->id] = false;
    }

    // Mark the node as visited.
    (*visited)[node->id] = true;

    return false;
}

/// Calculate the average loop length of the specified loop and all its child
/// loops recursively.
static void
calculate_average_loop_length(loop_ptr_t loop)
{
    double average_loop_length = 0.0;

    BOOST_FOREACH(cf_edge_t &edge, loop->cf_edges)
        switch (edge.cf_node->type) {
        case CF_Node_Location:
            average_loop_length += 1.0;
            break;
        case CF_Node_Loop:
            calculate_average_loop_length(edge.cf_node->loop);
            // Add average iterations in inner loop per iteration in outer loop
            // * average length of inner loop.
            average_loop_length +=
                edge.cf_node->loop->average_sample_count /
                loop->average_sample_count *
                edge.cf_node->loop->average_loop_length;
            break;
        default:
            assert(0);
        }

    loop->average_loop_length = average_loop_length;
}

#if 0
static void
sanity_check(env_t *env)
{
    cf_node_id_t location_count = env->cf_nodes.size() - env->loops->size();

    assert(env->cf_nodes.size() == env->outermost_node.size());
    assert(env->cf_nodes.size() == env->exits.size());
    assert(env->cf_nodes.size() == env->entries.size());

    for (cf_node_id_t i = 0; i < location_count; i++) {
        cf_node_ptr_t node = env->cf_nodes.at(i);
        assert(node->type == CF_Node_Location);
        assert(node->id == (cf_node_id_t)node->location);
    }

    for (cf_node_id_t i = location_count; i < (cf_node_id_t)env->cf_nodes.size(); i++) {
        cf_node_ptr_t node = env->cf_nodes.at(i);
        assert(node->type == CF_Node_Loop);
        assert(node->id + 1 == node->loop->id + location_count);
        for (cf_edge_vector_t::const_iterator
                 ei = node->loop->cf_edges.begin(),
                 ee = node->loop->cf_edges.end();
             ei != ee; ei++) {
            assert(env->outermost_node.at(ei->cf_node->id) ==
                   env->outermost_node.at(i));
        }
    }

    BOOST_FOREACH(cf_node_ptr_t &node, env->cf_nodes) {
        const by_node_t *exits_by_node =
            &env->exits.at(node->id).get<by_node>();
        assert(exits_by_node->empty() ||
               env->outermost_node.at(node->id)->id == node->id);
        for (cf_edge_map_t::const_iterator
                 ei = exits_by_node->begin(), ee = exits_by_node->end();
             ei != ee; ei++) {
            assert(ei->cf_node == env->outermost_node.at(ei->cf_node->id));
            set<cf_node_ptr_t> *s = &env->entries.at(ei->cf_node->id);
            assert(s->find(node) != s->end());
        }
    }

    for (cf_node_id_t i = 0; i < (cf_node_id_t)env->entries.size(); i++) {
        set<cf_node_ptr_t> *s = &env->entries.at(i);
        BOOST_FOREACH(cf_node_ptr_t node, *s) {
            by_node_t *exits_by_node = &env->exits.at(node->id).get<by_node>();
            assert(exits_by_node->find(env->cf_nodes.at(i)) != exits_by_node->end());
        }
    }

    for (cf_edge_map_t::const_iterator
             si = env->start_nodes.begin(), se = env->start_nodes.end();
         si != se; si++)
        assert(env->outermost_node.at(si->cf_node->id) == si->cf_node);
}
#endif

loop_identification_result_cptr_t
loop_identification_t::calculate(
    const parser_result_cptr_t &parser_result,
    const stack_pc_map_result_cptr_t &stack_pc_map)
{
    // Prepare result object.
    loop_identification_result_ptr_t result(new loop_identification_result_t());

    // Data structures needed during loop identification, that are no longer
    // needed once we have the result.
    env_t env(&result->loops, stack_pc_map);

    // Collect pc1 -> pc1+ transitions, using all line sizes and both samples
    // and danglings to get the best coverage we can. We will get different
    // numbers of samples for different line sizes, so we try to compensate by
    // counting the number of bytes accessed, remembering the largest access of
    // each pc and dividing the number bytes accessed by the largest access
    // size.
    //
    // Also count the number of distinct sampled line sizes.
    typedef map<location_t, double> pc1p_stat_map_t;
    typedef pair<fnr_access_size_t, pc1p_stat_map_t> pc1_stat_t;
    typedef map<location_t, pc1_stat_t> pc1_stat_map_t;
    pc1_stat_map_t pc1_stat_map;
    set<fnr_line_size_t> line_size_log2s;

    BOOST_FOREACH(const parser_burst_t &pb, parser_result->bursts) {
	BOOST_FOREACH(const event_sample_ptr_t sp, pb.samples_by_pc1) {
	    const fnr_event_sample_t *s(sp.get());

	    line_size_log2s.insert(s->line_size_log2);

	    location_t pc_1_loc = stack_pc_map->get_location(&s->begin);
	    location_t pc_1_plus_loc = stack_pc_map->get_pc_1_plus_location(s);

	    pc1_stat_t *pc1_stat = &pc1_stat_map[pc_1_loc];
	    pc1_stat->first = max(pc1_stat->first, s->begin.access_size);
	    pc1_stat->second[pc_1_plus_loc] += s->begin.access_size;

	    // TODO: Should this be done for every pc_end instead?
	    if (!s->end.empty()) {
		location_t pc_2_loc = stack_pc_map->get_location(&s->end[0]);
		pc1_stat_map_t::iterator pc2_stat_it = pc1_stat_map.find(pc_2_loc);
		if (pc2_stat_it != pc1_stat_map.end()) {
		    pc1_stat_t *pc2_stat = &pc2_stat_it->second;
		    pc2_stat->first = max(pc2_stat->first, s->end[0].access_size);
		}
	    }
	}
    }

    int num_line_sizes = (int)line_size_log2s.size();

    // Create control flow nodes for all locations, add their control flow
    // edges and add them to the start nodes map.
    for (pc1_stat_map_t::const_iterator
             pc1i = pc1_stat_map.begin(), pc1e = pc1_stat_map.end();
         pc1i != pc1e; pc1i++) {
        cf_node_ptr_t cf_node = env.cf_nodes[pc1i->first];
        by_count_t *exits_by_count = &env.exits[pc1i->first].get<by_count>();
        fnr_access_size_t access_size = pc1i->second.first;
        for (pc1p_stat_map_t::const_iterator
                 pc1pi = pc1i->second.second.begin(),
                 pc1pe = pc1i->second.second.end();
             pc1pi != pc1pe; pc1pi++)
            add_cf_edge(&env, env.cf_nodes[pc1i->first], env.cf_nodes[pc1pi->first],
                        pc1pi->second / access_size / num_line_sizes);
        env.start_nodes.insert(
            cf_edge_t(cf_node, exits_by_count->rbegin()->sample_count));
    }

    // Look up a suitable start node and try finding a loop starting from it,
    // until we have run out of suitable start nodes.
    by_count_t *start_nodes_by_count = &env.start_nodes.get<by_count>();
    by_node_t *start_nodes_by_node = &env.start_nodes.get<by_node>();

    by_count_t::const_reverse_iterator start_node_it;
    while ((start_node_it = start_nodes_by_count->rbegin()) !=
           start_nodes_by_count->rend()) {
        vector<bool> visited(env.cf_nodes.size());
        cf_edge_vector_t trace;
        vector<bool> in_trace(env.cf_nodes.size());
        cf_node_ptr_t start_node = start_node_it->cf_node;
        if (!loop_search(&env, &visited, &trace, &in_trace, start_node))
            start_nodes_by_node->erase(start_node_it->cf_node);
    }

    // Mark the top level loops.
    BOOST_FOREACH(cf_node_ptr_t node, env.cf_nodes)
        switch (node->type) {
        case CF_Node_Location:
            break;
        case CF_Node_Loop:
            node->loop->top_level = env.outermost_node[node->id] == node;
            break;
        default:
            assert(0);
        }

    // Calculate average sample count of all loops.
    BOOST_FOREACH(loop_ptr_t loop, *env.loops) {
        double sample_count = 0.0;
        BOOST_FOREACH(cf_edge_t &edge, loop->cf_edges)
            sample_count += edge.sample_count;
        loop->average_sample_count = sample_count / loop->cf_edges.size();
    }

    // Calculate average loop length of all loops.
    BOOST_FOREACH(loop_ptr_t loop, *env.loops)
        if (loop->top_level)
            calculate_average_loop_length(loop);

    return result;
}

loop_identification_t::loop_identification_t(
    const feature_context_ptr_t &context,
    const string &sample_file,
    const fnr_stack_depth_t stack_depth)
    : feature_t<loop_identification_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth)
{
}

loop_identification_result_cptr_t
loop_identification_t::get() const
{
    return get_data();
}

string
loop_identification_t::name() const
{
    return "loop_id(" +
        sample_file + "," +
        lexical_cast<string>(stack_depth) + ")";
}

loop_identification_result_cptr_t
loop_identification_t::calculate() const
{
    parser_result_cptr_t parser_result =
        parser_t(context, sample_file).get();

    stack_pc_map_result_cptr_t stack_pc_map =
        stack_pc_map_t(context, sample_file, stack_depth).get();

    return calculate(parser_result, stack_pc_map);
}

END_NAMESPACE;
END_NAMESPACE;
