/// @file analysis/structure/loop_identification.hh

// Copyright (c) 2006-2008 Acumem AB.  All rights reserved.

#ifndef LOOP_IDENTIFICATION_HH
#define LOOP_IDENTIFICATION_HH

#include "analysis/call_stack/stack_pc_map.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace acumem::library;

using boost::shared_ptr;

/// A shared pointer to a loop.
typedef shared_ptr<struct loop_t> loop_ptr_t;
typedef shared_ptr<const struct loop_t> loop_cptr_t;

// ----- Control flow representation. -----------------------------------------

/// A control flow node id. Locations are guaranteed to have the same control
/// flow node id as their location id.
typedef int cf_node_id_t;

/// Type of a control flow node, either a location or a loop.
enum cf_node_type_t {
    CF_Node_Location,
    CF_Node_Loop
};

/// A control flow node. Represents either a location or a loop.
struct cf_node_t {
    /// Creates a control flow node representing a location.
    cf_node_t(cf_node_id_t id, location_t location)
        : type(CF_Node_Location),
          id(id),
          location(location) {}

    /// Creates a control flow node representing a loop.
    cf_node_t(cf_node_id_t id, loop_ptr_t loop)
        : type(CF_Node_Loop),
          id(id),
          loop(loop) {}

    /// Node type, instruction or loop.
    cf_node_type_t type;

    /// Control flow node id.
    cf_node_id_t id;

    /// Only for locations: The location.
    location_t location;

    /// Only for loops: The loop.
    loop_ptr_t loop;
};

/// Shared pointer to a control flow node.
typedef shared_ptr<cf_node_t> cf_node_ptr_t;

/// A vector of control flow nodes.
typedef vector<cf_node_ptr_t> cf_node_vector_t;

/// An control flow edge to or from a control flow node.
struct cf_edge_t {
    cf_edge_t(cf_node_ptr_t cf_node, double sample_count)
        : cf_node(cf_node),
          sample_count(sample_count) {}

    /// Control flow node that this edge leads to or from.
    cf_node_ptr_t cf_node;

    /// Sample count along the edge. Based on all line sizes to improve
    /// accuracy, but compensated for access size and the number of sampled
    /// line sizes so that it is equivalent to the sample count of one line
    /// size.
    double sample_count;
};

/// Comparison operators for cf_edge_t that orders edges primarily by count
/// secondarily by control flow node id, to get a well defined sorting order.
bool operator==(const cf_edge_t &x, const cf_edge_t &y);
bool operator<(const cf_edge_t &x, const cf_edge_t &y);

/// A vector of control flow edges.
typedef vector<cf_edge_t> cf_edge_vector_t;

// ----- loop_t ---------------------------------------------------------------

/// A loop id. Loop id 0 is reserved for the dummy loop for instructions that
/// are not included in any real loop.
typedef int loop_id_t;

/// A vector of loops.
typedef vector<loop_ptr_t> loop_vector_t;

/// A loop.
struct loop_t {
    /// Creates a loop.
    loop_t(loop_id_t id);

    /// Id of the loop.
    loop_id_t id;

    /// True if the loop is not included in any other loop.
    bool top_level;

    /// Control flow edges making up the loop. Each edge contains a control
    /// flow node and the number of samples leading to bit. In pc1 -> pc1+
    /// order, beginning at an arbitrary node.
    cf_edge_vector_t cf_edges;

    /// Average sample count along the edges of the loop, excluding edges in
    /// inner loops. Based on all line sizes to improve accuracy, but
    /// compensated for access size and the number of sampled line sizes so
    /// that it is equivalent to the sample count of one line size.
    double average_sample_count;

    /// Average number of memory accesses per iteration of the loop, including
    /// accesses from inner loops.
    double average_loop_length;

    /// Returns true if the specified loop is an inner loop of this loop,
    /// recursively searching the inner loops of this loop if recursive is
    /// true.
    bool has_child_loop(loop_ptr_t loop, bool recursive) const;

    /// Returns true if the specified location is in this loop, recursively
    /// searching the inner loops of this loop if recursive is true.
    bool has_child_location(location_t location, bool recursive) const;

    /// Returns the inner loops of this loop, recursively including inner loops
    /// of this loop if recursive is true.
    loop_vector_t get_loops(bool recursive) const;

    /// Returns the locations in this loop, recursively including locations in
    /// inner loops of this loop if recursive is true.
    location_set_t get_locations(bool recursive) const;

    /// Special case of getting inclusive locations, used by the
    /// blocking code. Return all locations, except those below a
    /// certain loop.
    location_set_t get_locations_excluding_loop(loop_ptr_t exclude_loop) const;
};

// ----- loop_identification_result_t -----------------------------------------

/// Loop identification result.
struct loop_identification_result_t {
    /// Loops. Indexed by loop id - 1, since loop 0 (dummy loop for
    /// instructions that are not included in any real loop) is not included.
    loop_vector_t loops;
};

/// Shared pointer to a loop identification result.
typedef shared_ptr<loop_identification_result_t>
        loop_identification_result_ptr_t;
typedef shared_ptr<const loop_identification_result_t>
        loop_identification_result_cptr_t;

// ----- loop_identification_t ------------------------------------------------

/// Loop identification module.
class loop_identification_t
    : public feature_t<loop_identification_result_cptr_t> {
public:
    static loop_identification_result_cptr_t calculate(
        const parser_result_cptr_t &parser_result,
        const stack_pc_map_result_cptr_t &stack_pc_map);

    loop_identification_t(const feature_context_ptr_t &context,
                          const string &sample_file,
                          const fnr_stack_depth_t stack_depth);
    loop_identification_result_cptr_t get() const;

protected:
    string name() const;
    loop_identification_result_cptr_t calculate() const;

private:
    const string sample_file;
    const fnr_stack_depth_t stack_depth;
};

END_NAMESPACE;
END_NAMESPACE;

#endif /* LOOP_IDENTIFICATION_HH */
