/// @file analysis/structure/pc_lookup.cc

// Copyright (c) 2007-2009 Acumem AB.  All rights reserved.

#include <boost/lexical_cast.hpp>
#include <boost/foreach.hpp>

#include "analysis/filter/filter_module.hh"
#include "pc_lookup.hh"

using namespace acumem::library;
using namespace boost;
using namespace std;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

// ----- per_pc1_t ------------------------------------------------------------

bool operator==(const per_pc1_t &x, const per_pc1_t &y)
{
    return x.min_rdist == y.min_rdist && x.pc1 == y.pc1;
}

bool operator<(const per_pc1_t &x, const per_pc1_t &y)
{
    return (x.min_rdist < y.min_rdist ||
            (x.min_rdist == y.min_rdist && x.pc1 < y.pc1));
}

// ----- per_pc2_t ------------------------------------------------------------

bool operator==(const per_pc2_t &x, const per_pc2_t &y)
{
    return x.min_rdist == y.min_rdist && x.pc2 == y.pc2;
}

bool operator<(const per_pc2_t &x, const per_pc2_t &y)
{
    return (x.min_rdist < y.min_rdist ||
            (x.min_rdist == y.min_rdist && x.pc2 < y.pc2));
}

// ----- pc_lookup_t ----------------------------------------------------------

per_pc2_t *
find_per_pc2(per_pc2_vector_t &per_pc2, const location_t pc2_loc)
{
    BOOST_FOREACH(per_pc2_t &p, per_pc2) {
	if (p.pc2 == pc2_loc)
	    return &p;
    }

    return NULL;
}

per_pc1_t *
find_per_pc1(per_pc1_vector_t &per_pc1, const location_t pc1_loc)
{
    BOOST_FOREACH(per_pc1_t &p, per_pc1) {
            if (p.pc1 == pc1_loc)
                return &p;
    }

    return NULL;
}

template<class SampleType>
static void
add_samples(
    pc_lookup_result_ptr_t &result,
    const stack_pc_map_result_cptr_t &stack_pc_map,
    const vector<SampleType> &samples)
{
    BOOST_FOREACH(const SampleType &s, samples) {
	// Skip samples without pc_valid, only applicable to line
	// samples.
	if (!s.valid())
	    continue;

	const mapped_access_t &start(*s.start());
	const mapped_access_t &end(*s.end());
        fnr_time_t rdist(s.rdist());

        pc_info_t &pc1_info(result->pc_info[start.pc]);
        pc_info_t &pc2_info(result->pc_info[end.pc]);

        pc2_info.pc2_count++;

        bool sort_per_pc2(false);
        per_pc2_t *per_pc2(find_per_pc2(pc1_info.per_pc2, end.pc));
	if (!per_pc2) {
	    pc1_info.per_pc2.push_back(per_pc2_t(end.pc));
	    per_pc2 = &pc1_info.per_pc2.back();
	    sort_per_pc2 = true;
	}

        if (rdist < per_pc2->min_rdist) {
            per_pc2->min_rdist = rdist;
            sort_per_pc2 = true;
        }

        if (sort_per_pc2)
            sort(pc1_info.per_pc2.begin(), pc1_info.per_pc2.end());

	bool sort_per_pc1(false);
        per_pc1_t *per_pc1(find_per_pc1(pc2_info.per_pc1, start.pc));
	if (!per_pc1) {
	    pc2_info.per_pc1.push_back(per_pc1_t(start.pc));
	    per_pc1 = &pc2_info.per_pc1.back();
	    sort_per_pc1 = true;
	}

        if (rdist < per_pc1->min_rdist) {
            per_pc1->min_rdist = rdist;
            sort_per_pc1 = true;
        }

        if (rdist > per_pc1->max_rdist)
            per_pc1->max_rdist = rdist;

        per_pc1->samples.push_back(&s);

        if (sort_per_pc1)
            sort(pc2_info.per_pc1.begin(), pc2_info.per_pc1.end());
    }
}

pc_lookup_result_cptr_t
pc_lookup_t::calculate(
    stack_pc_map_result_cptr_t stack_pc_map,
    sample_mapping_result_cptr_t samples,
    mapped_sample_type_t sample_type)
{
    pc_lookup_result_ptr_t result(new pc_lookup_result_t());

    result->samples = samples;

    // Collect information from the samples.
    BOOST_FOREACH(const sample_mapping_burst_t &smb, samples->bursts) {
	// Collect information from the samples.
	switch (sample_type) {
	case Mapped_Byte:
	    add_samples(result, stack_pc_map, smb.repl_byte);
	    break;
	case Mapped_Line:
	    add_samples(result, stack_pc_map, smb.repl_line);
	    break;
	case Mapped_Page_Thread:
	    add_samples(result, stack_pc_map, smb.repl_page_thread);
	    break;
	case Mapped_Page_Cache:
	    add_samples(result, stack_pc_map, smb.repl_page_cache);
	    break;
	default:
	    assert(0);
	}
    }

    return result;
}

pc_lookup_t::pc_lookup_t(
    const feature_context_ptr_t &context,
    const string &sample_file,
    const fnr_stack_depth_t stack_depth,
    const fnr_line_size_t line_size,
    const fnr_line_size_t page_size,
    const thread_mapping_cptr_t &thread_mapping,
    const cpu_model_t &cpu_model,
    const mapped_sample_type_t sample_type)
    : feature_t<pc_lookup_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      sample_type(sample_type)
{
}

string
pc_lookup_t::name() const
{
    return "pc_lookup_t(" +
	lexical_cast<string>(sample_file) + "," +
        lexical_cast<string>(stack_depth) + "," +
	lexical_cast<string>(line_size) + "," +
	lexical_cast<string>(page_size) + "," +
	thread_mapping->name() + "," +
	cpu_model.name() + "," +
	lexical_cast<string>(sample_type) + ")";
}

pc_lookup_result_cptr_t
pc_lookup_t::get() const
{
    return get_data();
}

pc_lookup_result_cptr_t
pc_lookup_t::calculate() const
{
    stack_pc_map_result_cptr_t stack_pc_map =
        stack_pc_map_t(context, sample_file, stack_depth).get();

    sample_mapping_result_cptr_t sample_mapping(
        sample_mapping_t(context, sample_file, stack_depth,
			 line_size, page_size, thread_mapping,
			 cpu_model).get());

    return calculate(stack_pc_map, sample_mapping, sample_type);
}

END_NAMESPACE;
END_NAMESPACE;
