/// @file analysis/structure/pc_lookup.hh

// Copyright (c) 2007-2008 Acumem AB.  All rights reserved.

#ifndef PC_LOOKUP_HH
#define PC_LOOKUP_HH

#include <string>
#include <vector>

#include <boost/shared_ptr.hpp>

#include "library/lang/common_macros.h"
#include "library/lang/common_types.h"
#include "analysis/call_stack/stack_pc_map.hh"
#include "analysis/thread_mapping/sample_mapping.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace std;

using boost::shared_ptr;

// ----- per_pc1_t ------------------------------------------------------------

/// Aggregate information about all samples with a specific pc1 in the context
/// of a pc2.
struct per_pc1_t {
    per_pc1_t(location_t pc1)
	: pc1(pc1),
	  min_rdist((fnr_time_t)-1),
	  max_rdist(0) {}

    /// The pc1.
    location_t pc1;

    /// Minimum reuse distance between the pc1 and pc2.
    fnr_time_t min_rdist;

    /// Maximum reuse distance between the pc1 and pc2.
    fnr_time_t max_rdist;

    /// The samples between the pc1 and pc2.
    mapped_repl_base_cptr_vector_t samples;
};

/// Vector of aggregate information per pc1 for the samples with a certain pc2.
typedef vector<per_pc1_t> per_pc1_vector_t;

/// Comparison operators for per_pc1_t that sort objects primarily by min_rdist
/// and secondarily by pc1.
bool operator==(const per_pc1_t &x, const per_pc1_t &y);
bool operator<(const per_pc1_t &x, const per_pc1_t &y);

// ----- per_pc2_t ------------------------------------------------------------

/// Aggregate information about all samples with a specific pc2 in the context
/// of a pc1.
struct per_pc2_t {
    per_pc2_t(location_t pc2)
	: pc2(pc2),
	  min_rdist((fnr_time_t)-1) {}

    /// The pc2.
    location_t pc2;

    /// Minimum reuse distance between the pc1 and pc2.
    fnr_time_t min_rdist;
};

/// Vector of aggregate information per pc2 for the samples with a certain pc1.
typedef vector<per_pc2_t> per_pc2_vector_t;

/// Comparison operators for per_pc2_t that sort objects primarily by min_rdist
/// and secondarily by pc2.
bool operator==(const per_pc2_t &x, const per_pc2_t &y);
bool operator<(const per_pc2_t &x, const per_pc2_t &y);

// ----- pc_info_t ------------------------------------------------------------

/// Aggregate information about a pc.
struct pc_info_t {
    pc_info_t() : pc2_count(0) {}

    /// The number of samples where this pc is pc2.
    fnr_sample_count_t pc2_count;

    /// Aggregate information per pc1 for samples where this pc is pc2. Sorted
    /// by minimum reuse distance and pc1 location.
    per_pc1_vector_t per_pc1;

    /// Aggregate information per pc2 for samples where this pc is pc1. Sorted
    /// by minimum reuse distance and pc2 location.
    per_pc2_vector_t per_pc2;
};

typedef map<location_t, pc_info_t> pc_info_map_t;

// ----- pc_lookup_result_t ---------------------------------------------------

/// Lookup result.
struct pc_lookup_result_t {
    /// Aggregate information per pc.
    pc_info_map_t pc_info;

    /// Keep a smart pointer to the sample mapping results to make
    /// sure that it isn't destroyed while we still have live pointers
    /// to samples.
    sample_mapping_result_cptr_t samples;
};

typedef shared_ptr<pc_lookup_result_t> pc_lookup_result_ptr_t;
typedef shared_ptr<const pc_lookup_result_t> pc_lookup_result_cptr_t;

// ------ pc_lookup_t ---------------------------------------------------------

class pc_lookup_t : public feature_t<pc_lookup_result_cptr_t> {
public:
    static pc_lookup_result_cptr_t calculate(
        stack_pc_map_result_cptr_t stack_pc_map,
        sample_mapping_result_cptr_t samples,
	mapped_sample_type_t sample_type);

    pc_lookup_t(const feature_context_ptr_t &context,
                const string &sample_file,
                const fnr_stack_depth_t stack_depth,
                const fnr_line_size_t line_size,
                const fnr_line_size_t page_size,
		const thread_mapping_cptr_t &thread_mapping,
		const cpu_model_t &cpu_model,
		const mapped_sample_type_t sample_type);

    pc_lookup_result_cptr_t get() const;

protected:
    string name() const;
    pc_lookup_result_cptr_t calculate() const;

private:
    const string sample_file;
    const fnr_stack_depth_t stack_depth;
    const fnr_line_size_t line_size;
    const fnr_line_size_t page_size;
    const thread_mapping_cptr_t thread_mapping;
    const cpu_model_t cpu_model;
    const mapped_sample_type_t sample_type;
};

END_NAMESPACE;
END_NAMESPACE;

#endif /* PC_LOOKUP_HH */
