/// @file analysis/thread_mapping/coherence.cc

// Copyright (c) 2008 Acumem AB.  All rights reserved.
#include "coherence.hh"


BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

// --- coherence_bsm_t --------------------------------------------------

coherence_bsm_t::coherence_bsm_t(const fnr_address_t &byte_address)
    : byte_address(byte_address), state(BS_0)
{
}

void
coherence_bsm_t::local_access(const mapped_access_t &a)
{
    const uint64_t a_start(a.addr().virtual_address),
	a_end(a.addr().virtual_address + a.access_size());
    if (a_start <= byte_address.virtual_address &&
	byte_address.virtual_address < a_end &&
	a.is_proper())
	local_byte_access(a);
    else
	local_line_access(a);
}

void
coherence_bsm_t::local_byte_access(const mapped_access_t &a)
{
    const fnr_access_type_t at(a.access_type());

    // Line specific accesses, use local_line_access() instead.
    assert(!a.is_prefetch());
    assert(at != FNR_Access_Type_CLFLUSH);

    if (at == FNR_Access_Type_NTA_WR) {
	state = BS_NW;
	return;
    }

    switch (state) {
    case BS_0:
	switch (at) {
	case FNR_Access_Type_RD:
	case FNR_Access_Type_NTA_RD:
	    state = BS_R;
	    break;

	case FNR_Access_Type_WR:
	    state = BS_W;
	    break;

	case FNR_Access_Type_RDWR:
	    state = BS_RW;
	    break;

	default:
	    assert(0);
	}
	break;
    case BS_R:
	switch (at) {
	case FNR_Access_Type_RDWR:
	case FNR_Access_Type_WR:
	    state = BS_RW;
	    break;

	default:
	    break;
	}
	break;

    case BS_RW:
	break;

    case BS_W:
	break;

    case BS_M:
	switch (at) {
	case FNR_Access_Type_RDWR:
	case FNR_Access_Type_WR:
	    state = BS_W;
	    break;

	default:
	    break;
	}
	break;

    case BS_N:
    case BS_NW:
	// Should already have been handled
	assert(at != FNR_Access_Type_NTA_WR);
	state = BS_0;
	local_byte_access(a);
	break;
    }
}

void
coherence_bsm_t::local_line_access(const mapped_access_t &a)
{
    const fnr_access_type_t at(a.access_type());

    switch (at) {
    case FNR_Access_Type_NTA_WR:
	state = BS_N;
	break;
    case FNR_Access_Type_CLFLUSH:
	state = BS_0;
	break;

    default:
	switch (state) {
	case BS_0:
	case BS_R:
	case BS_RW:
	case BS_W:
	case BS_M:
	    break;

	case BS_N:
	case BS_NW:
	    state = BS_0;
	    local_line_access(a);
	    break;
	}
	break;
    }
}

void
coherence_bsm_t::line_invalidated()
{
    switch (state) {
    case BS_0:
    case BS_R:
    case BS_RW:
    case BS_W:
    case BS_M:
    case BS_N:
    case BS_NW:
	state = BS_0;
	break;
    }
}

void
coherence_bsm_t::line_downgraded()
{
    switch (state) {
    case BS_0:
    case BS_R:
    case BS_M:
    case BS_N:
    case BS_NW:
	break;

    case BS_W:
	state = BS_M;
	break;

    case BS_RW:
	state = BS_R;
	break;
    }
}


// --- coherence_line_t -------------------------------------------------------

coherence_line_state_change_t::coherence_line_state_change_t(
    const coherence_line_state_t old_state,
    const coherence_line_state_t new_state)
    : gained(new_state & ~old_state),
      lost(old_state & ~new_state)
{
}

coherence_line_t::coherence_line_t(const cpu_model_t &cpu_model,
                                   const tm_cache_id_t local_cache,
                                   const coherence_line_state_t line_state)
    : cpu_model(cpu_model),
      local_cache(local_cache),
      line_state(line_state)
{
}

coherence_line_state_change_t
coherence_line_t::access(const mapped_access_t &access)
{
    coherence_line_state_t old_state = line_state;

    if (access.cache == local_cache) {
        // A local access.
        if (access.access_type() == FNR_Access_Type_NTA_WR)
            line_state = Line_State_Non_Temporal;
        else if (access.is_flush())
            line_state = Line_State_None;
        else if (cpu_model.makes_modified(access))
            line_state = Line_State_Valid | Line_State_Exclusive | Line_State_Modified;
        else if (cpu_model.makes_exclusive(access)) {
            line_state &= ~Line_State_Non_Temporal;
            line_state |= Line_State_Valid | Line_State_Exclusive;
        } else {
            line_state &= ~Line_State_Non_Temporal;
            line_state |= Line_State_Valid;
        }
    }
    else {
        // A remote access.
        if (cpu_model.causes_invalidation(access))
            line_state = Line_State_None;
        else
            line_state &= ~(Line_State_Non_Temporal | Line_State_Exclusive | Line_State_Modified);
    }

    return coherence_line_state_change_t(old_state, line_state);
}

END_NAMESPACE;
END_NAMESPACE;
