/// @file analysis/thread_mapping/coherence.hh

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#ifndef COHERENCE_HH
#define COHERENCE_HH

#include "cpu_model.hh"
#include "mapped_sample.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace acumem::library;

// --- coherence_bsm_t --------------------------------------------------

class coherence_bsm_t {
public:
    enum state_t {
	BS_0 = 0,
	BS_R,
	BS_RW,
	BS_W,
	BS_M,
	BS_N,
	BS_NW
    };

    coherence_bsm_t(const fnr_address_t &byte_address);

    inline bool is_touched() const {
	switch (state) {
	case BS_R:
	case BS_RW:
	case BS_W:
	case BS_M:
	case BS_NW:
	    return true;

	case BS_0:
	case BS_N:
	    return false;
	}
	assert(0);
    }

    inline bool is_read() const {
	switch (state) {
	case BS_R:
	case BS_RW:
	    return true;

	case BS_0:
	case BS_W:
	case BS_M:
	case BS_N:
	case BS_NW:
	    return false;
	}
	assert(0);
    }

    inline bool is_written() const {
	switch (state) {
	case BS_RW:
	case BS_W:
	    return true;

	case BS_0:
	case BS_R:
	case BS_M:
	case BS_N:
	case BS_NW:
	    return false;
	}
	assert(0);
    }

    inline bool is_non_temporal() const {
	return state == BS_NW;
    }

    // Local access, either to the byte or elsewhere on the line.
    void local_access(const mapped_access_t &a);

    /// Line invalidated or flushed
    void line_invalidated();

    /// Line downgraded
    void line_downgraded();

private:
    /// This byte has been accessed
    void local_byte_access(const mapped_access_t &a);

    /// Some other byte has been accessed on the same line
    void local_line_access(const mapped_access_t &a);

    const fnr_address_t byte_address;
    state_t state;
};

// --- coherence_line_t -------------------------------------------------------

/// Cache line state components. Represented as a bitmap so that multiple
/// components can be represented at once.
typedef int coherence_line_state_t;

#define Line_State_None         0x0
#define Line_State_Valid        0x1
#define Line_State_Exclusive    0x2
#define Line_State_Modified     0x4
#define Line_State_Non_Temporal 0x8

/// Cache line state components gained and lost.
struct coherence_line_state_change_t {
    coherence_line_state_change_t(const coherence_line_state_t old_state,
                                  const coherence_line_state_t new_state);

    coherence_line_state_t gained;
    coherence_line_state_t lost;
};

/// Cache line state model.
class coherence_line_t {
public:
    coherence_line_t(const cpu_model_t &cpu_model,
                     const tm_cache_id_t local_cache,
                     const coherence_line_state_t line_state = Line_State_None);

    bool is_valid() const {
	return (line_state & Line_State_Valid) != 0;
    }

    bool is_exclusive() const {
	return (line_state & Line_State_Exclusive) != 0;
    }

    bool is_modified() const {
	return (line_state & Line_State_Modified) != 0;
    }

    bool is_non_temporal() const {
	return (line_state & Line_State_Non_Temporal) != 0;
    }

    coherence_line_state_change_t access(const mapped_access_t &access);

private:
    /// CPU model, queried for effects of access types.
    const cpu_model_t cpu_model;

    /// The cache in which we are tracking the cache line's state.
    const tm_cache_id_t local_cache;

    /// Current cache line state.
    coherence_line_state_t line_state;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
