/// @file analysis/thread_mapping/cpu_model.hh

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#ifndef CPU_MODEL_HH
#define CPU_MODEL_HH

#include <string>

#include "library/lang/common_macros.h"
#include "library/fenris/fnr_types.hh"
#include "library/cpudb/cpudb.hh"
#include "thread_mapping.hh" // tm_cache_id_t
#include "mapped_sample.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace acumem::library;

struct cpu_model_t {
    cpu_model_t();
    cpu_model_t(const cpudb_cpu_cptr_t &cpu,
		const cpudb_cache_cptr_t &cache);
    cpu_model_t(fnr_prefetch_classes_t active_pf,
		fnr_prefetch_classes_t coherence_pf,
		bool prefetchw_makes_modified);

    bool access_active(const tm_cache_id_t local_cache,
		       const mapped_access_t &a) const {
	if (a.cache == local_cache)
	    return access_active_local(*a.access);
	else
	    return access_active_foreign(*a.access);
    }

    bool access_active_local(const fnr_sample_access_t &a) const {
	fnr_prefetch_classes_t pfc(a.prefetch_classes());

	if (pfc == FNR_Prefetch_Class_None)
	    return true;
	else
	    return (pfc & active_pf) != 0;
    }

    bool access_active_foreign(const fnr_sample_access_t &a) const {
	fnr_prefetch_classes_t pfc(a.prefetch_classes());

	if (pfc == FNR_Prefetch_Class_None)
	    return true;
	else
	    return (pfc & coherence_pf) != 0;
    }

    //@{
    /// TODO: Document!
    bool makes_valid(const mapped_access_t &a) const {
	return makes_valid(a.access_type());
    }

    bool makes_valid(const fnr_sample_access_t &a) const {
	return makes_valid(a.access_type);
    }

    bool makes_valid(const fnr_access_type_t at) const {
	return
            at != FNR_Access_Type_CLFLUSH &&
            at != FNR_Access_Type_NTA_WR;
    }
    //@}

    //@{
    /// Check if an access makes the cache line exclusive in the local cache.
    bool makes_exclusive(const mapped_access_t &a) const {
	return makes_exclusive(a.access_type());
    }

    bool makes_exclusive(const fnr_sample_access_t &a) const {
	return makes_exclusive(a.access_type);
    }

    bool makes_exclusive(const fnr_access_type_t at) const {
	const fnr_access_classes_t ac(get_access_classes(at));
	return !!(ac & FNR_Access_Class_WR) &&
                at != FNR_Access_Type_NTA_WR;
    }
    //@}

    //@{
    /// Check if an access makes the cache line modified in the local cache,
    /// requiring a write-back. Assumes that the access is active on the
    /// current cache level.
    bool makes_modified(const mapped_access_t &a) const {
	return makes_modified(a.access_type());
    }

    bool makes_modified(const fnr_sample_access_t &a) const {
	return makes_modified(a.access_type);
    }

    bool makes_modified(const fnr_access_type_t at) const {
	if (at == FNR_Access_Type_PF_AMD_W)
	    return prefetchw_makes_modified;
	else {
	    const fnr_access_classes_t ac(get_access_classes(at));
	    return !!(ac & FNR_Access_Class_WR) &&
                at != FNR_Access_Type_NTA_WR;
	}
    }
    //@}

    //@{
    /// TODO: Document!
    bool makes_nontemporal(const mapped_access_t &a) const {
	return makes_nontemporal(a.access_type());
    }

    bool makes_nontemporal(const fnr_sample_access_t &a) const {
	return makes_nontemporal(a.access_type);
    }

    bool makes_nontemporal(const fnr_access_type_t at) const {
	return at == FNR_Access_Type_NTA_WR;
    }
    //@}

    //@{
    /// Check if an access causes the cache line to be invalidated in all
    /// remote caches.
    bool causes_invalidation(const mapped_access_t &a) const {
	return causes_invalidation(a.access_type());
    }

    bool causes_invalidation(const fnr_sample_access_t &a) const {
	return causes_invalidation(a.access_type);
    }

    bool causes_invalidation(const fnr_access_type_t at) const {
	const fnr_access_classes_t ac(get_access_classes(at));
	return !!(ac & (FNR_Access_Class_WR | FNR_Access_Class_Flush));
    }
    //@}

    /// Prefetches bringing data to the current cache level.
    fnr_prefetch_classes_t active_pf;
    /// Prefetches that are active and cause coherence actions.
    fnr_prefetch_classes_t coherence_pf;

    /// True if the prefetchw instruction fetches a cache line into the M
    /// state. False if the instruction fetches the line into the E state
    /// instead.
    bool prefetchw_makes_modified;

    string name() const;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
