/// @file analysis/thread_mapping/mapped_sample.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include "mapped_sample.hh"
#include "coherence.hh"
#include "cpu_model.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

// --- mapped_access_t --------------------------------------------------

mapped_access_t::mapped_access_t()
    : access(NULL),
      cache(0),
      pc(LOCATION_INVALID),
      local_time(0)
{
}

mapped_access_t::mapped_access_t(const time_mapping_burst_t &tmb,
				 const stack_pc_map_result_cptr_t &spm,
				 tm_cache_id_t cache,
				 const fnr_sample_access_t &access)
    : access(&access),
      cache(cache),
      pc(spm->get_location(&access)),
      local_time(tmb.get_local_time(cache, access.time))
{
}

// --- mapped_repl_byte_t -----------------------------------------------------

void
mapped_repl_byte_t::sanity_check(void) const
{
    // There should be at least one access in the accesses vector.
    assert(accesses.size() >= 1);

    // Check that the accesses in the accesses vector are in the correct order.
    for (mapped_access_vector_t::size_type i = 0; i < accesses.size() - 1; i++)
        assert(accesses[i].global_time() < accesses[i + 1].global_time());

    // Check that _start is the first access in the accesses vector, is a
    // proper access and comes before _inval and _end (_comm checked later).
    assert(_start == 0);
    assert(accesses[_start].is_proper());
    assert(_start < _end);

    // Check that _comm_use is proper, is a read, comes after _start and comes
    // before _inval and _end.
    BOOST_FOREACH(const mapped_access_index_t comm_use, _comm_use)
        if (comm_use != MAPPED_ACCESS_ILLEGAL_INDEX) {
            assert(accesses[comm_use].is_proper());
            assert(accesses[comm_use].is_read());
            assert(_start < comm_use);
            assert(comm_use < _end);
        }

    // Check that _end exists and is a proper access.
    assert(_end != MAPPED_ACCESS_ILLEGAL_INDEX);
    assert(accesses[_end].is_proper());
}

// --- mapped_repl_line_t -----------------------------------------------------

void
mapped_repl_line_t::sanity_check(void) const
{
    // There should be at least one access in the accesses vector.
    assert(accesses.size() >= 1);

    // Check that the accesses in the accesses vector are in the correct order.
    for (mapped_access_vector_t::size_type i = 0; i < accesses.size() - 1; i++)
        assert(accesses[i].global_time() < accesses[i + 1].global_time());

    // Check that _start is the first access in the accesses vector.
    assert(_start == 0);

    // Check that _start is the first access in the accesses vector and comes
    // before or is the same as all other accesses (doesn't check _comm, that
    // is checked in the next step).
    assert(_start <= _valid);
    assert(_start < _devalid);
    assert(_start <= _exclusive);
    assert(_start < _deexclusive);
    assert(_start <= _modified);
    assert(_start < _demodified);
    assert(_start <= _nontemporal);
    assert(_start < _denontemporal);
    assert(_start < _false_sharing);
    assert(_start < _end);

    // Check there is no devalid access without a valid access. Check that
    // valid comes before devalid.
    if (_valid != MAPPED_ACCESS_ILLEGAL_INDEX)
        assert(_valid < _devalid);
    else
        assert(_devalid == MAPPED_ACCESS_ILLEGAL_INDEX);

    // Check there is no denontemporal access without a nontemporal
    // access. Check that nontemporal comes before denontemporal.
    if (_nontemporal != MAPPED_ACCESS_ILLEGAL_INDEX)
        assert(_nontemporal < _denontemporal);
    else
        assert(_denontemporal == MAPPED_ACCESS_ILLEGAL_INDEX);

#if 0

    // Check that _comm comes after _start and comes before or is _inval.
    BOOST_FOREACH(const mapped_access_index_t comm, _comm)
        if (comm != MAPPED_ACCESS_ILLEGAL_INDEX) {
            assert(_start < comm);
            assert(comm <= _inval);
        }

    // Check _inval.
    if (_inval != MAPPED_ACCESS_ILLEGAL_INDEX) {
        assert(valid != MAPPED_ACCESS_ILLEGAL_INDEX);
        assert(valid < _inval);
    }

    // Check _deexclusive.
    if (_deexclusive != MAPPED_ACCESS_ILLEGAL_INDEX) {
        assert(_exclusive != MAPPED_ACCESS_ILLEGAL_INDEX);
        assert(_exclusive < _inval);
    }
#endif
}

// --- mapped_repl_page_t -----------------------------------------------------

mapped_repl_page_t::mapped_repl_page_t(fnr_address_t address,
                                       const mapped_access_t &start,
				       const mapped_access_t &end)
    : mapped_repl_base_t(address), _start(0), _end(1), accesses(2)
{
    accesses[0] = start;
    accesses[1] = end;
}

void
mapped_repl_page_t::sanity_check(void) const
{
    // There should be at least one access in the accesses vector.
    assert(accesses.size() >= 1);

    // Check that the accesses in the accesses vector are in the correct order.
    for (mapped_access_vector_t::size_type i = 0; i < accesses.size() - 1; i++)
        assert(accesses[i].global_time() < accesses[i + 1].global_time());

    // Check that _start is the first access in the accesses vector.
    assert(_start == 0);
}

END_NAMESPACE
END_NAMESPACE
