/// @file analysis/thread_mapping/mapped_sample.hh

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#ifndef MAPPED_SAMPLE_HH
#define MAPPED_SAMPLE_HH

#include "thread_mapping.hh"
#include "time_mapping.hh"
#include "analysis/call_stack/stack_pc_map.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace acumem::library;

// ----- mapped_access_t ----------------------------------------------------------

/// An access.
struct mapped_access_t {
    mapped_access_t();
    mapped_access_t(const time_mapping_burst_t &tmb,
		    const stack_pc_map_result_cptr_t &spm,
		    tm_cache_id_t cache,
		    const fnr_sample_access_t &access);

    bool is_write(bool include_pf = true, bool include_nta = true) const {
	const fnr_access_classes_t ac = access_classes();
	return ac & FNR_Access_Class_WR &&
	    (include_pf || !(ac & FNR_Access_Class_PF)) &&
	    (include_nta || !(ac & FNR_Access_Class_NTA));
    }

    bool is_read(bool include_pf = true, bool include_nta = true) const {
	const fnr_access_classes_t ac = access_classes();
	return ac & FNR_Access_Class_RD &&
	    (include_pf || !(ac & FNR_Access_Class_PF)) &&
	    (include_nta || !(ac & FNR_Access_Class_NTA));
    }

    bool is_prefetch() const {
	return (access_classes() & FNR_Access_Class_PF) != 0;
    }


    bool is_flush() const {
	return access_type() == FNR_Access_Type_CLFLUSH;
    }

    bool is_nta() const {
	return (access_classes() & FNR_Access_Class_NTA) != 0;
    }

    bool is_proper() const {
	return !(access_classes() & FNR_Access_Classes_Pseudo);
    }

    const inline fnr_thread_id_t &thread() const {
	return access->thread;
    }

    const inline fnr_address_t &addr() const {
	return access->addr;
    }

    const inline fnr_access_size_t &access_size() const {
	return access->access_size;
    }

    const inline fnr_time_t &global_time() const {
	return access->time;
    }

    const inline fnr_access_type_t &access_type() const {
	return access->access_type;
    }

    const inline fnr_access_classes_t access_classes() const {
	return access->access_classes();
    }

    const inline fnr_prefetch_classes_t prefetch_classes() const {
	return access->prefetch_classes();
    }

    const fnr_sample_access_t *access;

    /// Cache
    tm_cache_id_t cache;

    /// Location
    location_t pc;

    /// Cache local access time.
    tm_local_time_t local_time;
};

typedef std::vector<mapped_access_t> mapped_access_vector_t;


// ----- access indices -----------------------------------------------------------
typedef size_t mapped_access_index_t;
#define MAPPED_ACCESS_ILLEGAL_INDEX ((mapped_access_index_t)(-1))
typedef std::vector<mapped_access_index_t> mapped_access_index_vector_t;

// ----- mapped_repl_sample_t -----------------------------------------------------

#define MAPPED_ACCESS(n)				\
    const inline mapped_access_t *n() const {		\
	if (_ ## n != MAPPED_ACCESS_ILLEGAL_INDEX)	\
	    return &accesses[_ ## n];			\
	else						\
	    return NULL;				\
    }

#define MAPPED_ACCESS_VECTOR(n)						\
    const inline std::vector<const mapped_access_t *> n() const {	\
	const size_t num_accesses(_ ## n .size());			\
	std::vector<const mapped_access_t *> ret(num_accesses);		\
	for (size_t i = 0; i < num_accesses; i++)			\
	    ret[i] = &accesses[_ ## n [i]];				\
	return ret;							\
    }

/// Superclass of all mapped sample types. Allows the same windowing and window
/// replacement ratio code to be used for all sample types.
struct mapped_repl_base_t {
    virtual ~mapped_repl_base_t() {}

    mapped_repl_base_t(fnr_address_t address)
        : address(address)
        {}

    /// The last access making the sample valid. May be absent if the sample is
    /// never made valid.
    virtual const mapped_access_t *valid() const = 0;

    /// The end access of the sample.
    virtual const mapped_access_t *end() const = 0;

    /// Returns true if the end access is known to not cause a replacement in
    /// the cache, e.g., if it is a cache line flush.
    virtual bool known_nonreplacement() const {
        // There are no known non-replacements for byte and page samples.
        return false;
    }

    /// Returns true if the end access is known to cause a replacement in the
    /// cache, e.g., if the line has been flushed and the end access causes a
    /// fetch.
    virtual bool known_replacement() const {
        // There are no known replacements for byte and page samples.
        return false;
    }

    /// Local reuse distance of the sample.
    const tm_local_time_t rdist() const {
        return end()->local_time - valid()->local_time - 1;
    };

    /// We need the address of the sample only to get a total ordering of
    /// samples for sort(). It is not needed the analysis.
    fnr_address_t address;
};

/// Mapped byte sample. See the MT documentation (doc/mt/) for
/// details.
struct mapped_repl_byte_t
    : public mapped_repl_base_t
{
    mapped_repl_byte_t(fnr_address_t address)
	: mapped_repl_base_t(address),
          _start(MAPPED_ACCESS_ILLEGAL_INDEX),
	  _end(MAPPED_ACCESS_ILLEGAL_INDEX),
	  invalidated(false), downgraded(false), flushed(false) {}

    //@{
    /// Convenience accessor functions. MAPPED_ACCESS_ILLEGAL_INDEX is
    /// mapped to NULL. Only optional values should map to NULL.
    MAPPED_ACCESS(start)
    MAPPED_ACCESS_VECTOR(comm_use)
    MAPPED_ACCESS(end)
    //@}

    const mapped_access_t *valid() const {
        return &accesses[_start];
    }

    /// Sample start access, either a R, Rnt, W or Wnt. Always valid.
    mapped_access_index_t _start;
    /// Communication usage, first access from every foreign cache not
    /// preceeded by an invalidation, except for the cache causing the
    /// invalidation. Accesses up to the 2nd invalidation will be
    /// stored for the cache causing the 1st invalidation.
    mapped_access_index_vector_t _comm_use;
    /// Access terminating the sample. Must be from the same cache
    /// that started the sample. Always valid.
    mapped_access_index_t _end;

    /// Has a foreign cache invalidated this copy?
    bool invalidated;
    /// Has a foreign cache downgraded this copy?
    bool downgraded;
    /// Has this copy been flushed locally?
    bool flushed;

    /// List of all accesses in the sample.
    mapped_access_vector_t accesses;

    /// Check that the sample seems sane.
    void sanity_check(void) const;
};

struct mapped_repl_line_t
    : public mapped_repl_base_t
{
    mapped_repl_line_t(fnr_address_t address)
	: mapped_repl_base_t(address),
          _start(MAPPED_ACCESS_ILLEGAL_INDEX),
	  _valid(MAPPED_ACCESS_ILLEGAL_INDEX),
	  _devalid(MAPPED_ACCESS_ILLEGAL_INDEX),
	  _exclusive(MAPPED_ACCESS_ILLEGAL_INDEX),
	  _deexclusive(MAPPED_ACCESS_ILLEGAL_INDEX),
	  _modified(MAPPED_ACCESS_ILLEGAL_INDEX),
	  _demodified(MAPPED_ACCESS_ILLEGAL_INDEX),
	  _nontemporal(MAPPED_ACCESS_ILLEGAL_INDEX),
	  _denontemporal(MAPPED_ACCESS_ILLEGAL_INDEX),
	  _false_sharing(MAPPED_ACCESS_ILLEGAL_INDEX),
	  _end(MAPPED_ACCESS_ILLEGAL_INDEX) {}

    virtual bool known_nonreplacement() const {
        // TODO: Cleanup!

        // Cache line flushes can not cause replacements. Non-temporal stores
        // may cause replacements depending on processor model and cache level,
        // but we assume they do not for now.
        return
            end()->is_flush() ||
            (end()->is_nta() && end()->is_write());
    }

    virtual bool known_replacement() const {
        // TODO: Cleanup!

        // The sample will always cause a replacement if it is not a known
        // non-replacement and the line has either never been made valid or
        // been made non-valid.
        return
            !known_nonreplacement() &&
            (valid() == NULL || devalid() != NULL);
    };

    //@{
    /// Convenience accessor functions. MAPPED_ACCESS_ILLEGAL_INDEX is
    /// mapped to NULL. Only optional values should map to NULL.
    MAPPED_ACCESS(start)
    MAPPED_ACCESS(valid)
    MAPPED_ACCESS(devalid)
    MAPPED_ACCESS(exclusive)
    MAPPED_ACCESS(deexclusive)
    MAPPED_ACCESS(modified)
    MAPPED_ACCESS(demodified)
    MAPPED_ACCESS(nontemporal)
    MAPPED_ACCESS(denontemporal)
    MAPPED_ACCESS_VECTOR(comm)
    MAPPED_ACCESS(false_sharing)
    MAPPED_ACCESS(end);
    //@}

    /// Access starting the line sample. Always valid.
    mapped_access_index_t _start;

    /// The last access making the line valid.
    mapped_access_index_t _valid;
    /// The first access making the line non-valid.
    mapped_access_index_t _devalid;

    /// The last access making the line exclusive.
    mapped_access_index_t _exclusive;
    /// The first access after 'exclusive' making the line non-exclusive.
    mapped_access_index_t _deexclusive;

    /// The last access making the line modified.
    mapped_access_index_t _modified;
    /// The first access after 'modified' making the line non-modified.
    mapped_access_index_t _demodified;

    /// The last access making the line non-temporal.
    mapped_access_index_t _nontemporal;
    /// The first access after 'nontemporal' making the line non-non-temporal.
    mapped_access_index_t _denontemporal;

    /// First access to the line from every foreign cache not
    /// preceeded by any write, but including the first foreign write
    /// from any cache.
    mapped_access_index_vector_t _comm;
    /// Accesses causing false sharing. Vacant if false sharing does
    /// not occur.
    mapped_access_index_t _false_sharing;
    /// Access terminating the sample. Must be from the same cache
    /// that started the sample.
    ///
    /// \note There are restrictions on the end access types, see the
    /// mt design document for details.
    mapped_access_index_t _end;

    /// List of all accesses in the sample.
    mapped_access_vector_t accesses;

    /// Check that the sample seems sane.
    void sanity_check(void) const;
};

struct mapped_repl_page_t
    : public mapped_repl_base_t
{
    mapped_repl_page_t(fnr_address_t address,
                       const mapped_access_t &start,
		       const mapped_access_t &end);


    //@{
    /// Convenience accessor functions. MAPPED_ACCESS_ILLEGAL_INDEX is
    /// mapped to NULL. Only optional values should map to NULL.
    MAPPED_ACCESS(start)
    MAPPED_ACCESS(end)
    //@}

    const mapped_access_t *valid() const {
        return &accesses[_start];
    }

    /// Access starting the sample
    mapped_access_index_t _start;
    /// End access, may be in the same cache or thread depending on
    /// the source vector.
    mapped_access_index_t _end;

    /// List of all accesses in the sample.
    mapped_access_vector_t accesses;

    /// Check that the sample seems sane.
    void sanity_check(void) const;
};

typedef vector<const mapped_repl_base_t *> mapped_repl_base_cptr_vector_t;

typedef vector<mapped_repl_byte_t> mapped_repl_byte_vector_t;
typedef vector<mapped_repl_line_t> mapped_repl_line_vector_t;
typedef vector<mapped_repl_page_t> mapped_repl_page_vector_t;

// ----------------------------------------------------------------------

struct mapped_cold_t {
    mapped_cold_t(const mapped_access_t &end)
	: end(end) {}

    mapped_access_t end;
};

typedef vector<const mapped_cold_t *> mapped_cold_cptr_vector_t;
typedef vector<mapped_cold_t> mapped_cold_vector_t;

// ----------------------------------------------------------------------

enum mapped_sample_type_t {
    Mapped_Byte = 0,
    Mapped_Line,
    Mapped_Page_Thread,
    Mapped_Page_Cache
};

// ----------------------------------------------------------------------

#undef MAPPED_ACCESS
#undef MAPPED_ACCESS_VECTOR


END_NAMESPACE;
END_NAMESPACE;

#endif
