/// @file analysis/thread_mapping/sample_mapping.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include "sample_mapping.hh"
#include "coherence.hh"
#include "analysis/filter/filter_module.hh"
#include "analysis/feature_base/warnings.hh"

#include <boost/lexical_cast.hpp>
#include <boost/dynamic_bitset.hpp>

// TODO: Implement coherence handling of otherwise inactive prefetches.

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace boost;
using namespace acumem::library;

// --- Internal access type handling ------------------------------------

// Access types used during mapping, uses the same types as the
// specification. Types are bitmapped to allow easy masking.
enum access_type_index_t {
    // Byte specific accesses
    ATI_R  = 0,
    ATI_W,
    ATI_RW,
    ATI_RNT,
    ATI_WNT,
    // Line specific accesses
    ATI_PR,
    ATI_PW,
    ATI_FLUSH
};

enum access_type_t {
    // Byte specific accesses
    AT_R  = (1 << ATI_R),
    AT_W  = (1 << ATI_W),
    AT_RW = (1 << ATI_RW),
    AT_RNT = (1 << ATI_RNT),
    AT_WNT = (1 << ATI_WNT),
    // Line specific accesses
    AT_PR = (1 << ATI_PR),
    AT_PW = (1 << ATI_PW),
    AT_FLUSH = (1 << ATI_FLUSH)
};


static inline access_type_t
operator|(access_type_t lhs, access_type_t rhs)
{
    return (access_type_t)((unsigned int)lhs | (unsigned int)rhs);
}

static inline access_type_t
operator&(access_type_t lhs, access_type_t rhs)
{
    return (access_type_t)((unsigned int)lhs & (unsigned int)rhs);
}

static inline access_type_t
operator~(access_type_t at)
{
    return (access_type_t)(~(unsigned int)at);
}

/// Maps fnr_access_type_t to access_type_index_t.
static access_type_index_t
get_access_type_index(const cpu_model_t &cpu_model,
                      const fnr_access_type_t at)
{
    // Map between fnr_access_type_t and access_type_index_t.
    static const access_type_index_t fnr_access_to_ati[] = {
        ATI_R,     // FNR_Access_Type_RD
        ATI_W,     // FNR_Access_Type_WR
        ATI_PR,    // FNR_Access_Type_PF
        ATI_RW,    // FNR_Access_Type_RDWR
        ATI_PR,    // FNR_Access_Type_PF_AMD
        ATI_PW,    // FNR_Access_Type_PF_AMD_W
        ATI_PR,    // FNR_Access_Type_PF_Intel_NTA
        ATI_PR,    // FNR_Access_Type_PF_Intel_T0
        ATI_PR,    // FNR_Access_Type_PF_Intel_T1
        ATI_PR,    // FNR_Access_Type_PF_Intel_T2
        ATI_FLUSH, // FNR_Access_Type_CLFLUSH
        ATI_RNT,   // FNR_Access_Type_NTA_RD
        ATI_WNT    // FNR_Access_Type_NTA_WR
    };

    if (at == FNR_Access_Type_PF_AMD_W &&
        cpu_model.makes_modified(at))
        // A write prefetch that makes the line modified is a proper write for
        // sample mapping purposes.
        return ATI_W;
    else if (at <= FNR_Access_Type_NTA_WR)
	return fnr_access_to_ati[at];
    else {
	assert(0);
        while (1);
    }
}

static access_type_index_t
get_access_type_index(const cpu_model_t &cpu_model,
                      const mapped_access_t &a)
{
    return get_access_type_index(cpu_model, a.access_type());
}

static access_type_index_t
get_access_type_index(const cpu_model_t &cpu_model,
                      const fnr_sample_access_t &a)
{
    return get_access_type_index(cpu_model, a.access_type);
}

// ----------------------------------------------------------------------
// Sample termination tables
// ----------------------------------------------------------------------

/// Access types that that are started/terminated by a line access. Indexed by
/// access type index.
static const access_type_t termination_mask_line[] = {
    AT_R |                AT_RNT | AT_WNT | AT_PR |         AT_FLUSH, // AT_R
    AT_R | AT_W | AT_RW | AT_RNT | AT_WNT | AT_PR | AT_PW | AT_FLUSH, // AT_W
    AT_R | AT_W | AT_RW | AT_RNT | AT_WNT | AT_PR | AT_PW | AT_FLUSH, // AT_RW
    AT_R |                AT_RNT | AT_WNT | AT_PR |         AT_FLUSH, // AT_RNT
    AT_R | AT_W | AT_RW | AT_RNT | AT_WNT | AT_PR | AT_PW | AT_FLUSH, // AT_WNT
    AT_R |                AT_RNT | AT_WNT | AT_PR |         AT_FLUSH, // AT_PR
    AT_R |                AT_RNT | AT_WNT | AT_PR | AT_PW | AT_FLUSH, // AT_PW
    AT_R | AT_W | AT_RW | AT_RNT | AT_WNT | AT_PR | AT_PW | AT_FLUSH  // AT_FLUSH
};

/// Accesses that may terminate a byte sample. Indexed by starting
/// access.
static const access_type_t termination_mask_byte[] = {
    AT_R |                AT_RNT | AT_WNT, // AT_R
           AT_W |                  AT_WNT, // AT_W
    AT_R | AT_W | AT_RW | AT_RNT | AT_WNT, // AT_RW
    AT_R |                AT_RNT | AT_WNT, // AT_RNT
    AT_R | AT_W | AT_RW | AT_RNT | AT_WNT  // AT_WNT
};

// ----------------------------------------------------------------------

static bool
by_end(const mapped_repl_base_t &lhs,
       const mapped_repl_base_t &rhs)
{
    if (lhs.end()->global_time() < rhs.end()->global_time())
        return true;
    if (lhs.end()->global_time() > rhs.end()->global_time())
        return false;
    if (rhs.valid() == NULL)
        return false;
    if (lhs.valid() == NULL)
        return true;
    if (lhs.valid()->global_time() > rhs.valid()->global_time())
        return true;
    if (lhs.valid()->global_time() < rhs.valid()->global_time())
        return false;
    if (lhs.address < rhs.address)
        return true;
    if (lhs.address > rhs.address)
        return false;

    // No two objects should be equal so we should reasonably not get
    // here. However, some older versions of libstdc++ compare objects with
    // themselves, and even make copies of objects and compare them with the
    // originals, so return false assuming that this is the case.
    return false;
}

static bool
by_global_time(const mapped_access_t &lhs,
	       const mapped_access_t &rhs)
{
    if (lhs.global_time() < rhs.global_time())
        return true;
    if (lhs.global_time() > rhs.global_time())
        return false;
    if (lhs.access_size() > rhs.access_size())
        return true;
    if (lhs.access_size() < rhs.access_size())
        return false;
    if (lhs.addr() < rhs.addr())
        return true;
    if (lhs.addr() > rhs.addr())
        return false;

    // No two objects should be equal so we should reasonably not get
    // here. However, some older versions of libstdc++ compare objects with
    // themselves, and even make copies of objects and compare them with the
    // originals, so return false assuming that this is the case.
    return false;
}

/// Conditionally add an access to an access vector if the current
/// index is illegal. Returns the index of the added access, or
/// current_index if no access was added.
static mapped_access_index_t
cadd_access(mapped_access_vector_t &accesses,
	    mapped_access_index_t &current_index,
	    const mapped_access_t &a)
{
    if (current_index == MAPPED_ACCESS_ILLEGAL_INDEX) {
	current_index = accesses.size();
	accesses.push_back(a);
    }

    return current_index;
}

/// Find the first invalidation to a cache line in the cache
/// local_cache between access (begin, end].
///
/// \param begin iterator pointing to the firs access in the interval
/// \param end iterator pointing to the last access (exclusive)
/// \return The global time of the invalidation, or -1 if no
/// invalidation occurs.
static fnr_time_t
find_invalidation_time(const cpu_model_t &cpu_model,
		       const tm_cache_id_t local_cache,
		       const mapped_access_vector_t::const_iterator &begin,
		       const mapped_access_vector_t::const_iterator &end)
{
    coherence_line_t comm_lsm(cpu_model, local_cache, Line_State_Valid);
    BOOST_FOREACH(const mapped_access_t &a, make_pair(begin, end))
	if (comm_lsm.access(a).lost & Line_State_Valid)
	    return a.global_time();
    return (fnr_time_t)-1;
}

static bool
overlaps(const fnr_address_t a0, const fnr_access_size_t s0,
	 const fnr_address_t a1, const fnr_access_size_t s1)
{
    return
	(a0.virtual_address <= a1.virtual_address &&
	 a1.virtual_address < a0.virtual_address + s0) ||
	(a1.virtual_address <= a0.virtual_address &&
	 a0.virtual_address < a1.virtual_address + s1);
}

static bool
overlaps(const fnr_sample_access_t &a0,
	 const fnr_sample_access_t &a1)
{
    return overlaps(a0.addr, a0.access_size,
		    a1.addr, a1.access_size);
}

static void
add_unique_accesses(mapped_access_vector_t &accesses,
		    const time_mapping_burst_t &time_mappings,
		    const thread_mapping_burst_t &thread_mappings,
		    const stack_pc_map_result_cptr_t &spm_res,
		    const fnr_event_sample_t &byte_sample,
		    const mapped_access_vector_t &line_accesses)
{
    mapped_access_vector_t::const_iterator
	la(line_accesses.begin()),
	lae(line_accesses.end());

    fnr_sample_access_vector_t::const_iterator
	ba(byte_sample.end.begin()),
	bae(byte_sample.end.end());

    const tm_cache_map_t &cm(
	thread_mappings.get_cache_map(la->global_time()));

    while (la != lae || ba != bae) {
	if (la == lae) {
	    const fnr_sample_access_t &a(*(ba++));
		accesses.push_back(
		    mapped_access_t(time_mappings, spm_res,
				    cm.find(a.thread)->second,
				    a));
	} else if (ba == bae) {
	    const mapped_access_t &ma(*(la++));
		accesses.push_back(ma);
	} else {
	    const fnr_sample_access_t &byte_access(*ba);
	    const mapped_access_t &line_access(*la);

	    if (byte_access.time == line_access.global_time()) {
		ba++;
		la++;
		    accesses.push_back(line_access);
	    } else if (byte_access.time < line_access.global_time()) {
		ba++;
		    accesses.push_back(
			mapped_access_t(time_mappings, spm_res,
					cm.find(byte_access.thread)->second,
					byte_access));
	    } else {
		la++;
		    accesses.push_back(line_access);
	    }
	}
    }
}

// --- Translation functions - MT samples ----------------------------
static void
translate_byte_sample(mapped_repl_byte_vector_t &repl_byte,
		      const time_mapping_burst_t &time_mappings,
                      const thread_mapping_burst_t &thread_mappings,
                      const stack_pc_map_result_cptr_t &spm_res,
		      const cpu_model_t &cpu_model,
                      const fnr_event_sample_t &byte_sample,
		      const mapped_access_vector_t &line_accesses,
		      const tm_cache_id_t num_caches)
{
    mapped_repl_byte_t bs(byte_sample.begin.addr);

    mapped_access_vector_t accesses;
    add_unique_accesses(accesses, time_mappings, thread_mappings,
			spm_res, byte_sample, line_accesses);

    // Add the start access to the byte sample
    bs.accesses.push_back(accesses[0]);
    bs._start = 0;

    const mapped_access_t &begin(accesses[0]);

    assert(begin.cache == line_accesses[0].cache);
    assert(begin.local_time == line_accesses[0].local_time);

    // Setup the byte and line state machines
    coherence_line_t lsm(cpu_model, begin.cache);
    lsm.access(begin);

    bool do_comm(begin.is_write(false, false));
    vector<bool> comm_seen(num_caches, false);
    comm_seen[begin.cache] = true;
    tm_cache_id_t inval_cache(TM_CACHE_ID_ILLEGAL);
    fnr_time_t inval_cache_end_time((fnr_time_t)-1);

    // Accesses that may terminate the current sample
    access_type_t termination_mask(
	termination_mask_byte[
            get_access_type_index(cpu_model, byte_sample.begin)]);

    mapped_access_vector_t::const_iterator
	it(accesses.begin() + 1),
	it_end(accesses.end());
    for(; it != it_end; it++) {
	const mapped_access_t &ma(*it);

	if (!termination_mask)
	    break;

	// Update byte and line state
	const access_type_index_t end_type(
            get_access_type_index(cpu_model, ma));
	mapped_access_index_t mai(MAPPED_ACCESS_ILLEGAL_INDEX);

	const coherence_line_state_change_t state_change(lsm.access(ma));
	if (state_change.lost & Line_State_Valid) {
	    inval_cache = ma.cache;
	    if (ma.cache != begin.cache) {
		bs.invalidated = true;
		inval_cache_end_time = find_invalidation_time(
		    cpu_model, ma.cache, it, it_end);
	    } else {
		inval_cache_end_time = ma.global_time();
		bs.flushed = true;
	    }
	} else if (state_change.lost & Line_State_Exclusive)
	    bs.downgraded = true;


	// Only handle byte specific accesses
	if (!ma.is_proper())
	    continue;

	if (overlaps(byte_sample.begin, *ma.access)) {
	    if (ma.cache == begin.cache) {
		if (termination_mask & (1 << end_type)) {
		    bs._end = cadd_access(bs.accesses, mai, ma);
		    repl_byte.push_back(bs);

		    // Clear the comm list to prevent double counting of
		    // communication events.
		    bs._comm_use.clear();
		}

		// Remove all sample types terminated by this access from
		// the termination mask
		termination_mask = termination_mask &
		    ~termination_mask_byte[end_type];
	    } else if (do_comm &&
		       !comm_seen[ma.cache]) {
		comm_seen[ma.cache] = true;
		if (ma.global_time() < inval_cache_end_time &&
		    (inval_cache == TM_CACHE_ID_ILLEGAL ||
		     inval_cache == ma.cache) &&
		    ma.is_read()) {
		    bs._comm_use.push_back(bs.accesses.size());
		    bs.accesses.push_back(ma);
		}
	    }
	}
    }
}

static void
mark_used_bytes(dynamic_bitset<> &used_bytes,
		const mapped_access_t &access,
		const fnr_line_size_t line_size)
{
    const fnr_line_size_t line_mask(line_size - 1);
    const int byte_in_line((int)(access.addr().virtual_address & line_mask));

    for (int i = 0; i < access.access_size(); i++)
	used_bytes[byte_in_line + i] = true;
}

static bool
is_false_sharing(const fnr_line_size_t line_size,
                 const mapped_access_vector_t &line_accesses,
                 const mapped_access_index_t false_sharing,
                 const mapped_access_index_t end,
                 const coherence_line_t &current_line_state,
                 const bool was_killed)
{
    const tm_cache_id_t local(line_accesses[0].cache);
    const tm_cache_id_t remote(line_accesses[false_sharing].cache);

    coherence_line_t line_state(current_line_state);
    dynamic_bitset<> local_a((dynamic_bitset<>::size_type)line_size);
    dynamic_bitset<> local_w((dynamic_bitset<>::size_type)line_size);
    dynamic_bitset<> remote_a((dynamic_bitset<>::size_type)line_size);
    dynamic_bitset<> remote_w((dynamic_bitset<>::size_type)line_size);

    for (mapped_access_index_t i = 1; i < end; i++) {
        const mapped_access_t &access(line_accesses[i]);

        if (access.cache == remote) {
            if (access.is_write(false, true)) {
                mark_used_bytes(remote_a, access, line_size);
                mark_used_bytes(remote_w, access, line_size);
            }
            else if (access.is_read(false, true))
                mark_used_bytes(remote_a, access, line_size);
        }
    }

    if (remote_a.none())
        return true;

    for (mapped_access_index_t i = end; i < line_accesses.size(); i++) {
        const mapped_access_t &access(line_accesses[i]);

        if (access.cache == local) {
            if (access.is_write(false, true)) {
                mark_used_bytes(local_a, access, line_size);
                mark_used_bytes(local_w, access, line_size);
            }
            else if (access.is_read(false, true))
                mark_used_bytes(local_a, access, line_size);

            if (!(local_a & remote_w).none() ||
                !(local_w & remote_a).none())
                return false;

            coherence_line_state_change_t state_change =
                line_state.access(access);
            if ((state_change.gained &
                 (Line_State_Valid | Line_State_Exclusive |
                  Line_State_Modified | Line_State_Non_Temporal)) &&
                // Don't report false sharing just because the end access
                // causes a fetch/upgrade/write-back, it must be a later access
                // that causes it.
                i != end)
                return true;
        }
        else {
            coherence_line_state_change_t state_change =
                line_state.access(access);
            if (state_change.lost &
                (Line_State_Valid | Line_State_Non_Temporal))
                return true;
        }
    }

    // The sample was still live at the end of the sampling. If that was also
    // the end of the execution we can be sure that there is no overlap between
    // the accesses of the threads and hence this false sharing, but if the
    // user detached from the application there may be later accesses that
    // would cause overlap and this may not be false sharing. We are cautious
    // and avoid saying false sharing, since we are not sure. Should be an
    // unusual case in most applications, anyway.
    return false;
}

static void
translate_line_sample(mapped_repl_line_vector_t &repl_line,
		      const time_mapping_burst_t &time_mappings,
                      const thread_mapping_burst_t &thread_mappings,
                      const stack_pc_map_result_cptr_t &spm_res,
		      const cpu_model_t &cpu_model,
                      const mapped_access_vector_t &line_accesses,
		      const tm_cache_id_t num_caches,
		      const fnr_line_size_t line_size,
		      const bool was_killed,
                      const fnr_event_sample_t &sample)
{
    mapped_repl_line_t ls(sample.begin.addr);

    // Add the start access to the line sample
    ls.accesses.push_back(line_accesses[0]);
    ls._start = 0;
    const mapped_access_t &begin(line_accesses[0]);

    // Setup the line state machine
    coherence_line_t lsm(cpu_model, begin.cache);
    lsm.access(begin);
    if (lsm.is_valid())
	ls._valid = 0;
    if (lsm.is_exclusive())
        ls._exclusive = 0;
    if (lsm.is_modified())
        ls._modified = 0;
    if (lsm.is_non_temporal())
        ls._nontemporal = 0;

    // Access types that may terminate the current sample.
    access_type_t termination_mask(
	termination_mask_line[get_access_type_index(cpu_model, *ls.start())]);

    // Setup the seen_cache vector to keep track which caches cause
    // communication. We only report one communication event per cache.
    bool do_comm(begin.is_write(false, false));
    vector<bool> seen_cache(num_caches, false);

    for (mapped_access_index_t i = 1; i < line_accesses.size(); i++) {
        const mapped_access_t &ma(line_accesses[i]);

	if (!termination_mask)
	    break;

	const coherence_line_state_change_t state_change(lsm.access(ma));

	mapped_access_index_t mai(MAPPED_ACCESS_ILLEGAL_INDEX);
	if (ma.cache == begin.cache) {
            // Local access.

	    // Check for sample termination.
	    const access_type_index_t end_type(get_access_type_index(cpu_model, ma));
	    if (termination_mask & (1 << end_type)) {
		// pc_end
		ls._end = cadd_access(ls.accesses, mai, ma);
		repl_line.push_back(ls);
		mapped_repl_line_t &inserted_line(repl_line.back());

                // Check for false sharing.
                mapped_access_index_t fs_candidate;
                if (cpu_model.makes_valid(ma) &&
                    ls.devalid() != NULL &&
                    ls.devalid()->cache != ls.start()->cache)
                    fs_candidate = ls._devalid;
                else if (cpu_model.makes_exclusive(ma) &&
                         ls.deexclusive() != NULL &&
                         ls.deexclusive()->cache != ls.start()->cache)
                    fs_candidate = ls._deexclusive;
                else if (cpu_model.makes_modified(ma) &&
                         ls.demodified() != NULL &&
                         ls.demodified()->cache != ls.start()->cache)
                    fs_candidate = ls._demodified;
                else if (cpu_model.makes_nontemporal(ma) &&
                         ls.denontemporal() != NULL &&
                         ls.denontemporal()->cache != ls.start()->cache)
                    fs_candidate = ls._denontemporal;
                else
                    fs_candidate = MAPPED_ACCESS_ILLEGAL_INDEX;

                if (fs_candidate != MAPPED_ACCESS_ILLEGAL_INDEX &&
                    is_false_sharing(line_size, line_accesses, fs_candidate, i,
                                     lsm, was_killed))
		    inserted_line._false_sharing = fs_candidate;

		// Clear the comm list to prevent double counting of
		// communication events.
		ls._comm.clear();
	    }

	    // Remove all sample types terminated by this access from
	    // the termination mask
	    termination_mask = termination_mask &
		~termination_mask_line[end_type];

	    if (lsm.is_valid()) {
		ls._valid = cadd_access(ls.accesses, mai, ma);
		ls._devalid = MAPPED_ACCESS_ILLEGAL_INDEX;
                if (ls._nontemporal != MAPPED_ACCESS_ILLEGAL_INDEX &&
                    ls._denontemporal == MAPPED_ACCESS_ILLEGAL_INDEX)
                    ls._denontemporal = cadd_access(ls.accesses, mai, ma);
	    } else if (lsm.is_non_temporal()) {
                if (ls._valid != MAPPED_ACCESS_ILLEGAL_INDEX &&
                    ls._devalid == MAPPED_ACCESS_ILLEGAL_INDEX)
                    ls._devalid = cadd_access(ls.accesses, mai, ma);
                if (ls._exclusive != MAPPED_ACCESS_ILLEGAL_INDEX &&
                    ls._deexclusive == MAPPED_ACCESS_ILLEGAL_INDEX)
                    ls._deexclusive = cadd_access(ls.accesses, mai, ma);
                if (ls._modified != MAPPED_ACCESS_ILLEGAL_INDEX &&
                    ls._demodified == MAPPED_ACCESS_ILLEGAL_INDEX)
                    ls._demodified = cadd_access(ls.accesses, mai, ma);
                ls._nontemporal = cadd_access(ls.accesses, mai, ma);
                ls._denontemporal = MAPPED_ACCESS_ILLEGAL_INDEX;
                do_comm = false;
            } else {
		// We end up here if we encounter a CLFLUSH.
                if (ls._valid != MAPPED_ACCESS_ILLEGAL_INDEX &&
                    ls._devalid == MAPPED_ACCESS_ILLEGAL_INDEX)
                    ls._devalid = cadd_access(ls.accesses, mai, ma);
                if (ls._exclusive != MAPPED_ACCESS_ILLEGAL_INDEX &&
                    ls._deexclusive == MAPPED_ACCESS_ILLEGAL_INDEX)
                    ls._deexclusive = cadd_access(ls.accesses, mai, ma);
                if (ls._modified != MAPPED_ACCESS_ILLEGAL_INDEX &&
                    ls._demodified == MAPPED_ACCESS_ILLEGAL_INDEX)
                    ls._demodified = cadd_access(ls.accesses, mai, ma);
                if (ls._nontemporal != MAPPED_ACCESS_ILLEGAL_INDEX &&
                    ls._denontemporal == MAPPED_ACCESS_ILLEGAL_INDEX)
                    ls._denontemporal = cadd_access(ls.accesses, mai, ma);
                do_comm = false;
	    }

	    if (cpu_model.makes_exclusive(ma)) {
		ls._exclusive = cadd_access(ls.accesses, mai, ma);
		ls._deexclusive = MAPPED_ACCESS_ILLEGAL_INDEX;
            }

            if (cpu_model.makes_modified(ma)) {
                ls._modified = cadd_access(ls.accesses, mai, ma);
                ls._demodified = MAPPED_ACCESS_ILLEGAL_INDEX;
            }
	} else {
            // Remote access.

            if (do_comm && !seen_cache[ma.cache]) {
                seen_cache[ma.cache] = true;
                ls._comm.push_back(cadd_access(ls.accesses, mai, ma));
            }

            if (state_change.lost & Line_State_Valid) {
                assert(ls._valid != MAPPED_ACCESS_ILLEGAL_INDEX &&
                       ls._devalid == MAPPED_ACCESS_ILLEGAL_INDEX);
		ls._devalid = cadd_access(ls.accesses, mai, ma);
                do_comm = false;
            }

            if (state_change.lost & Line_State_Exclusive) {
                assert(ls._exclusive != MAPPED_ACCESS_ILLEGAL_INDEX &&
                       ls._deexclusive == MAPPED_ACCESS_ILLEGAL_INDEX);
		ls._deexclusive = cadd_access(ls.accesses, mai, ma);
	    }

            if (state_change.lost & Line_State_Modified) {
                assert(ls._modified != MAPPED_ACCESS_ILLEGAL_INDEX &&
                       ls._demodified == MAPPED_ACCESS_ILLEGAL_INDEX);
                ls._demodified = cadd_access(ls.accesses, mai, ma);
	    }

            if (state_change.lost & Line_State_Non_Temporal) {
                assert (ls._nontemporal != MAPPED_ACCESS_ILLEGAL_INDEX &&
                        ls._denontemporal == MAPPED_ACCESS_ILLEGAL_INDEX);
                ls._denontemporal = cadd_access(ls.accesses, mai, ma);
                do_comm = false;
            }
	}
    }
}

static void
translate_page_sample(mapped_repl_page_vector_t &repl_page_thread,
		      mapped_repl_page_vector_t &repl_page_cache,
		      const time_mapping_burst_t &time_mappings,
                      const thread_mapping_burst_t &thread_mappings,
                      const stack_pc_map_result_cptr_t &spm_res,
		      const cpu_model_t &cpu_model,
                      const fnr_event_sample_t &sample)
{
    // Ignore samples that started on an inactive PF
    if (!cpu_model.access_active_local(sample.begin))
	return;

    const tm_cache_map_t &cm(thread_mappings.get_cache_map(sample.begin.time));
    const mapped_access_t mapped_begin(time_mappings, spm_res,
				       cm.find(sample.begin.thread)->second,
				       sample.begin);
    const fnr_sample_access_t *end_cache(NULL);
    const fnr_sample_access_t *end_thread(NULL);

    // Find the first access to the page from the same cache and the
    // same thread.
    BOOST_FOREACH(const fnr_sample_access_t &a, sample.end) {
	if (cpu_model.access_active_local(a)) {
	    if (!end_thread &&
		a.thread == sample.begin.thread)
		end_thread = &a;

	    if (!end_cache &&
		mapped_begin.cache == cm.find(a.thread)->second)
		end_cache = &a;

	    if (end_thread && end_cache)
		break;
	}
    }

    if (end_cache)
	repl_page_cache.push_back(
	    mapped_repl_page_t(sample.begin.addr,
                               mapped_begin,
			       mapped_access_t(
				   time_mappings, spm_res,
				   mapped_begin.cache,
				   *end_cache)));
    if (end_thread)
	repl_page_thread.push_back(
	    mapped_repl_page_t(sample.begin.addr,
                               mapped_begin,
			       mapped_access_t(
				   time_mappings, spm_res,
				   mapped_begin.cache,
				   *end_thread)));
}

static void
translate_samples(feature_context_ptr_t &context,
		  sample_mapping_burst_t &result,
		  const time_mapping_burst_t &time_mappings,
		  const thread_mapping_burst_t &thread_mappings,
		  const stack_pc_map_result_cptr_t &spm_res,
		  const cpu_model_t &cpu_model,
		  const parser_burst_t &byte_samples,
		  const parser_burst_t &line_samples,
		  const tm_cache_id_t num_caches,
		  const fnr_line_size_t line_size)
{
    mapped_repl_byte_vector_t &repl_byte(result.repl_byte);
    mapped_repl_line_vector_t &repl_line(result.repl_line);

    const sample_vector_t::const_iterator bs_begin(
	byte_samples.samples_by_pc1.begin());
    const sample_vector_t::const_iterator bs_end(
	byte_samples.samples_by_pc1.end());
    sample_vector_t::const_iterator bs(bs_begin);

    BOOST_FOREACH(const sample_ptr_t &ls, line_samples.samples_by_pc1) {
	// The whole sample group is considered to be inactive if it
	// starts on an inactive prefetch.
	if (!cpu_model.access_active_local(ls->begin))
	    continue;

	// Setup a vector with all accesses to the line sample
	mapped_access_vector_t mapped_line_accesses;
	const tm_cache_map_t &lcm(
	    thread_mappings.get_cache_map(ls->begin.time));
	mapped_line_accesses.reserve(ls->end.size() + 1);
	mapped_line_accesses.push_back(
	    mapped_access_t(time_mappings, spm_res,
			    lcm.find(ls->begin.thread)->second,
			    ls->begin));
	const tm_cache_id_t local_cache(mapped_line_accesses[0].cache);
	BOOST_FOREACH(const fnr_sample_access_t &a, ls->end) {
	    mapped_access_t ma(time_mappings, spm_res,
			       lcm.find(a.thread)->second,
			       a);
	    if (cpu_model.access_active(local_cache, ma))
		mapped_line_accesses.push_back(ma);
	}

	// We have to revert to the last byte sample before the line
	// sample, this is a simple way of taking care of the case
	// where multiple line samples are started at the same global
	// time.
	while (bs != bs_begin &&
	       (bs == bs_end || (*bs)->begin.time >= ls->begin.time))
	    --bs;

	// Forward to the first byte sample starting at the same time,
	// or after, the line sample.
	while (bs != bs_end &&
	       (*bs)->begin.time < ls->begin.time)
	    ++bs;

	for (; bs != bs_end && (*bs)->begin.time == ls->begin.time; ++bs)
	    if (overlaps(ls->begin, (*bs)->begin) && (*bs)->begin.is_proper())
		translate_byte_sample(repl_byte,
				      time_mappings,
				      thread_mappings,
				      spm_res,
				      cpu_model,
				      **bs,
				      mapped_line_accesses,
				      num_caches);

	translate_line_sample(repl_line,
			      time_mappings,
			      thread_mappings,
			      spm_res,
			      cpu_model,
			      mapped_line_accesses,
			      num_caches,
			      line_size,
			      ls->flags & FNR_Sample_Flag_Killed,
                              *ls);
    }
}

// --- Translation functions - legacy samples ------------------------

static void
translate_byte_sample_legacy(mapped_repl_byte_vector_t &repl_byte,
			     const time_mapping_burst_t &time_mappings,
                             const stack_pc_map_result_cptr_t &spm_res,
                             const fnr_event_sample_t &sample)
{
    // We filter samples starting or ending with PFs to be able to
    // make the same assumptions about start/end accesses in both the
    // MT and ST analysis. We also don't want to translate danglings.
    if (sample.end.empty() ||
	!sample.begin.is_proper() || !sample.end[0].is_proper())
	return;

    mapped_repl_byte_t bs(sample.begin.addr);
    bs.accesses.reserve(2);
    bs.accesses.push_back(
	mapped_access_t(time_mappings, spm_res,
			0, sample.begin));
    bs.accesses.push_back(
	mapped_access_t(time_mappings, spm_res,
			0, sample.end[0]));
    bs._start = 0;
    bs._end = 1;

    repl_byte.push_back(bs);
}

static void
translate_line_sample_legacy(mapped_repl_line_vector_t &repl_line,
			     const time_mapping_burst_t &time_mappings,
                             const stack_pc_map_result_cptr_t &spm_res,
                             const fnr_event_sample_t &sample)
{
    // Don't translate danglings.
    if (sample.end.empty())
	return;

    const fnr_access_classes_t bac(sample.begin.access_classes());
    const fnr_access_classes_t eac(sample.end[0].access_classes());

    const bool begin_is_wnt((bac & FNR_Access_Class_NTA) &&
			    (bac & FNR_Access_Class_WR));
    const bool end_is_wnt((eac & FNR_Access_Class_NTA) &&
			  (eac & FNR_Access_Class_WR));

    mapped_repl_line_t ls(sample.begin.addr);
    ls.accesses.reserve(2);
    ls.accesses.push_back(
	mapped_access_t(time_mappings, spm_res,
			0, sample.begin));
    ls.accesses.push_back(
	mapped_access_t(time_mappings, spm_res,
			0, sample.end[0]));
    ls._start = 0;
    if (begin_is_wnt == end_is_wnt &&
	!(bac & FNR_Access_Class_Flush)) {
	ls._valid = 0;
	if (bac & FNR_Access_Class_WR)
	    ls._exclusive = 0;
    }

    ls._end = 1;

    repl_line.push_back(ls);
}


static void
translate_page_sample_legacy(mapped_repl_page_vector_t &repl_page_thread,
			     mapped_repl_page_vector_t &repl_page_cache,
			     const time_mapping_burst_t &time_mappings,
                             const stack_pc_map_result_cptr_t &spm_res,
                             const fnr_event_sample_t &sample)
{
    // Don't translate danglings.
    if (sample.end.empty())
	return;

    const mapped_access_t mapped_begin(time_mappings, spm_res,
				       0, sample.begin);
    const mapped_access_t mapped_end(time_mappings, spm_res,
				     0, sample.end[0]);

    repl_page_thread.push_back(
	mapped_repl_page_t(sample.begin.addr, mapped_begin, mapped_end));
    repl_page_cache.push_back(
	mapped_repl_page_t(sample.begin.addr, mapped_begin, mapped_end));
}

// --- Cold sample synthesis -----------------------------------------

static void
synthesize_cold_samples(mapped_cold_vector_t &cold,
			const time_mapping_burst_t &time_mappings,
			const thread_mapping_burst_t &thread_mappings,
			const stack_pc_map_result_cptr_t &spm_res,
			const cpu_model_t &cpu_model,
			const parser_burst_t &samples,
			const tm_cache_id_t num_caches,
			const bool burst_sampled,
			const bool byte_samples)
{
    typedef fnr_time_vector_t time_vector_t;
    typedef vector<time_vector_t> cache_time_vector_t;
    typedef vector<mapped_access_vector_t> cache_access_vector_t;

    cache_time_vector_t end_times(num_caches);
    cache_access_vector_t start_accesses(num_caches);
    fnr_sample_count_t danglings_left(0);

    // Count the number of danglings and get start and end times from
    // the repl. samples
    BOOST_FOREACH(const sample_ptr_t &sample, samples.samples_by_pc1) {
	if (sample->end.empty())
	    danglings_left++;

	// Skip inactive prefetches and skip non-byte accesses when
	// handling bytes.
	if (!cpu_model.access_active_local(sample->begin) ||
	    (byte_samples && !sample->begin.is_proper()))
	    continue;

	const tm_cache_map_t &cm(
	    thread_mappings.get_cache_map(sample->begin.time));
	const tm_cache_id_t cache(
	    cm.find(sample->begin.thread)->second);

	start_accesses[cache].push_back(
	    mapped_access_t(time_mappings, spm_res, cache, sample->begin));

	BOOST_FOREACH(const fnr_sample_access_t &a, sample->end) {
	    if (!cpu_model.access_active_local(a) ||
		(byte_samples && !a.is_proper()))
		continue;

	    if (cache == cm.find(a.thread)->second) {
		end_times[cache].push_back(a.time);
		break;
	    }
	}
    }

    // Sort the time vectors
    BOOST_FOREACH(mapped_access_vector_t &av, start_accesses)
	sort(av.begin(), av.end(), by_global_time);
    BOOST_FOREACH(time_vector_t &tv, end_times)
	sort(tv.begin(), tv.end());

    vector<fnr_sample_count_t> max_live(num_caches, 0);
    for (tm_cache_id_t cache = 0; cache < num_caches; cache++) {
	fnr_sample_count_t live(0);

	mapped_access_vector_t::const_iterator
	    bi(start_accesses[cache].begin()),
	    be(start_accesses[cache].end());

	time_vector_t::const_iterator
	    ei(end_times[cache].begin()),
	    ee(end_times[cache].end());

        while (bi != be) {
            const fnr_time_t etime(ei != ee ? *ei : (fnr_time_t)-1);
	    const fnr_time_t btime(bi->global_time());

            if (etime < btime) {
                live--;
                ei++;
            } else {
                if (++live > max_live[cache] &&
		    (!burst_sampled || danglings_left > 0)) {
		    danglings_left--;
                    cold.push_back(mapped_cold_t(*bi));

                    max_live[cache] = live;
                }
                bi++;
            }
	}
    }
}

// ----- sample_mapping_result_t ----------------------------------------------

bool
sample_mapping_result_t::have_byte_samples() const
{
    // MT samplings are always done with byte samples enabled. If we there are
    // no byte samples then that's because there were no proper accesses.
    if (mt_sampled())
        return true;

    BOOST_FOREACH(const sample_mapping_burst_t &b, bursts) {
	if (!b.repl_byte.empty() || !b.cold_byte.empty())
	    return true;
    }
    return false;
}

bool
sample_mapping_result_t::have_line_samples() const
{
    BOOST_FOREACH(const sample_mapping_burst_t &b, bursts) {
	if (!b.repl_line.empty() || !b.cold_line.empty())
	    return true;
    }
    return false;
}

bool
sample_mapping_result_t::have_page_samples() const
{
    BOOST_FOREACH(const sample_mapping_burst_t &b, bursts) {
	if (!b.repl_page_cache.empty() || !b.cold_page.empty())
	    return true;
    }
    return false;
}

// ----- sample_mapping_t -----------------------------------------------------

static void
calculate_burst(feature_context_ptr_t &context,
		sample_mapping_burst_t &result,
		const thread_mapping_burst_t &thread_mappings,
		const time_mapping_burst_t &time_mappings,
		const stack_pc_map_result_cptr_t &stack_pc_map,
		const parser_burst_t &byte_samples,
		const parser_burst_t &line_samples,
		const parser_burst_t &page_samples,
		const cpu_model_t &cpu_model,
		const tm_cache_id_t num_caches,
		const fnr_line_size_t line_size,
		const bool mt_sampled,
		const bool burst_sampled)
{
    // Translate replacement samples.
    if (mt_sampled) {
	translate_samples(context,
			  result,
			  time_mappings,
			  thread_mappings,
			  stack_pc_map,
			  cpu_model,
			  byte_samples,
			  line_samples,
			  num_caches,
			  line_size);

	BOOST_FOREACH(const sample_ptr_t &s, page_samples.samples_by_pc1)
	    translate_page_sample(result.repl_page_thread,
				  result.repl_page_cache,
				  time_mappings,
				  thread_mappings,
				  stack_pc_map,
				  cpu_model,
				  *s);

    } else {
	BOOST_FOREACH(const sample_ptr_t &s, byte_samples.samples_by_pc1)
	    translate_byte_sample_legacy(result.repl_byte,
					 time_mappings,
					 stack_pc_map,
					 *s);

	BOOST_FOREACH(const sample_ptr_t &s, line_samples.samples_by_pc1)
	    translate_line_sample_legacy(result.repl_line,
					 time_mappings,
					 stack_pc_map,
					 *s);

	BOOST_FOREACH(const sample_ptr_t &s, page_samples.samples_by_pc1)
	    translate_page_sample_legacy(result.repl_page_thread,
					 result.repl_page_cache,
					 time_mappings,
					 stack_pc_map,
					 *s);
    }

    synthesize_cold_samples(result.cold_byte,
			    time_mappings,
			    thread_mappings,
			    stack_pc_map,
			    cpu_model,
			    byte_samples,
			    num_caches,
			    burst_sampled,
			    true);

    synthesize_cold_samples(result.cold_line,
			    time_mappings,
			    thread_mappings,
			    stack_pc_map,
			    cpu_model,
			    line_samples,
			    num_caches,
			    burst_sampled,
			    false);

    synthesize_cold_samples(result.cold_page,
			    time_mappings,
			    thread_mappings,
			    stack_pc_map,
			    cpu_model,
			    page_samples,
			    num_caches,
			    burst_sampled,
			    false);


    // Sort the replacement samples, the cold samples are already
    // sorted.
    sort(result.repl_byte.begin(), result.repl_byte.end(),
	 by_end);
    sort(result.repl_line.begin(), result.repl_line.end(),
	 by_end);
    sort(result.repl_page_thread.begin(), result.repl_page_thread.end(),
	 by_end);
    sort(result.repl_page_cache.begin(), result.repl_page_cache.end(),
	 by_end);

    /// @todo Don't just copy the filter module's sample period. We need to
    /// consider the samples we throw away because they are started by a
    /// no-op prefetch or never terminated.

    // Calculate the actual sample period.
    result.sample_period = line_samples.sample_period;
}

sample_mapping_result_cptr_t
sample_mapping_t::calculate(feature_context_ptr_t context,
			    thread_mapping_result_cptr_t thread_mappings,
                            time_mapping_result_cptr_t time_mappings,
                            stack_pc_map_result_cptr_t stack_pc_map,
			    parser_result_cptr_t byte_samples,
			    parser_result_cptr_t line_samples,
			    parser_result_cptr_t page_samples,
			    const cpu_model_t &cpu_model,
			    fnr_line_size_t line_size)
{
    sample_mapping_result_ptr_t result(
	new sample_mapping_result_t(thread_mappings->num_caches,
				    byte_samples, line_samples, page_samples));
    result->bursts.resize(line_samples->bursts.size());

    thread_mapping_burst_vector_t::const_iterator it_thread(
	thread_mappings->bursts.begin());
    time_mapping_burst_vector_t::const_iterator it_time(
	time_mappings->bursts.begin());
    parser_burst_vector_t::const_iterator it_byte(byte_samples->bursts.begin()),
	it_line(line_samples->bursts.begin()),
	it_page(page_samples->bursts.begin());
    BOOST_FOREACH(sample_mapping_burst_t &burst, result->bursts) {
	context->feedback->test_cancel();

#ifdef DEBUGSTUFF
	const tm_cache_map_t &cm = it_thread->get_cache_map(0);
	BOOST_FOREACH(const tm_cache_map_t::value_type &v, cm) {
	    cout << "thread " << v.first << " -> " << "cache " << v.second << endl;
	}
#endif
	calculate_burst(context, burst,
			*it_thread++,
			*it_time++,
			stack_pc_map,
			*it_byte++, *it_line++, *it_page++,
			cpu_model, thread_mappings->num_caches,
			line_size,
			line_samples->mt_sampled(),
			line_samples->burst_sampled);
    }

    // Sanity checks.
    BOOST_FOREACH(const sample_mapping_burst_t &burst, result->bursts) {
        BOOST_FOREACH(const mapped_repl_byte_t &sample, burst.repl_byte)
            sample.sanity_check();
        BOOST_FOREACH(const mapped_repl_line_t &sample, burst.repl_line)
            sample.sanity_check();
        BOOST_FOREACH(const mapped_repl_page_t &sample, burst.repl_page_thread)
            sample.sanity_check();
        BOOST_FOREACH(const mapped_repl_page_t &sample, burst.repl_page_cache)
            sample.sanity_check();
    }

    return result;
}

sample_mapping_t::sample_mapping_t(feature_context_ptr_t context,
                                   const string &sample_file,
                                   fnr_stack_depth_t stack_depth,
                                   fnr_line_size_t line_size,
                                   fnr_line_size_t page_size,
                                   thread_mapping_cptr_t thread_mapping,
				   const cpu_model_t &cpu_model)
    : feature_t<sample_mapping_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model)
{
}

sample_mapping_result_cptr_t
sample_mapping_t::get() const
{
    return get_data();
}

string
sample_mapping_t::name() const
{
    return "sample_mapping_t(" +
        sample_file + "," +
        lexical_cast<string>(stack_depth) + "," +
        lexical_cast<string>(line_size) + "," +
        lexical_cast<string>(page_size) + "," +
        thread_mapping->name() + "," +
	cpu_model.name() + ")";
}

sample_mapping_result_cptr_t
sample_mapping_t::calculate() const
{
    return calculate(
	context,
        thread_mapping->get(),
        time_mapping_t(context, sample_file, thread_mapping).get(),
        stack_pc_map_t(context, sample_file, stack_depth).get(),
        filter_module_t(context, sample_file, 1).get(),
        filter_module_t(context, sample_file, line_size).get(),
        filter_module_t(context, sample_file, page_size).get(),
	cpu_model,
	line_size);
}

END_NAMESPACE;
END_NAMESPACE;
