/// @file analysis/thread_mapping/sample_mapping.hh

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#ifndef SAMPLE_MAPPING_HH
#define SAMPLE_MAPPING_HH

#include "mapped_sample.hh"
#include "cpu_model.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace acumem::library;

// ----- sample_mapping_result_t ----------------------------------------------

typedef vector<fnr_sample_count_t> sm_sample_index_vector_t;

/// Result from #calculate_sample_mapping
struct sample_mapping_burst_t {
    //@{
    /// Replacement samples sorted by pc2.
    mapped_repl_byte_vector_t repl_byte;
    mapped_repl_line_vector_t repl_line;
    mapped_repl_page_vector_t repl_page_thread;
    mapped_repl_page_vector_t repl_page_cache;
    //@}

    //@{
    /// Cold samples sorted by pc2.
    mapped_cold_vector_t cold_byte;
    mapped_cold_vector_t cold_line;
    mapped_cold_vector_t cold_page;
    //@}

    /// Actual sample period in number of accesses of the selected
    /// line size.
    double sample_period;
};

typedef vector<sample_mapping_burst_t> sample_mapping_burst_vector_t;

struct sample_mapping_result_t {
    sample_mapping_result_t(const tm_cache_id_t num_caches,
			    const parser_result_cptr_t &byte_samples,
			    const parser_result_cptr_t &line_samples,
			    const parser_result_cptr_t &page_samples)
	: num_caches(num_caches),
	  byte_samples(byte_samples),
	  line_samples(line_samples),
	  page_samples(page_samples) {}

    //@{
    /// Check if there are any samples (cold or replacement) of a
    /// given type.
    bool have_byte_samples() const;
    bool have_page_samples() const;
    bool have_line_samples() const;
    //@}

    /// Sample mapping result per burst.
    sample_mapping_burst_vector_t bursts;

    /// True if this data originates from an MT sampling.
    bool mt_sampled() const {
        return byte_samples->mt_sampled();
    }

    /// Total number of caches in the system
    tm_cache_id_t num_caches;

    //@{
    /// sm_access_t contains references into the parser result, so
    /// keep an active reference to prevent deallocation while the
    /// result is still in use.
    parser_result_cptr_t byte_samples;
    parser_result_cptr_t line_samples;
    parser_result_cptr_t page_samples;
    //@}
};

typedef boost::shared_ptr<sample_mapping_result_t> sample_mapping_result_ptr_t;
typedef boost::shared_ptr<const sample_mapping_result_t> sample_mapping_result_cptr_t;

// ----- sample_mapping_t -----------------------------------------------------

class sample_mapping_t : feature_t<sample_mapping_result_cptr_t> {
public:
    static sample_mapping_result_cptr_t calculate(
	feature_context_ptr_t context,
	thread_mapping_result_cptr_t thread_mappings,
        time_mapping_result_cptr_t time_mappings,
        stack_pc_map_result_cptr_t stack_pc_map,
        parser_result_cptr_t byte_samples,
        parser_result_cptr_t line_samples,
	parser_result_cptr_t page_samples,
	const cpu_model_t &cpu_model,
	fnr_line_size_t line_size);

    sample_mapping_t(feature_context_ptr_t context,
                     const string &sample_file,
                     fnr_stack_depth_t stack_depth,
                     fnr_line_size_t line_size,
                     fnr_line_size_t page_size,
                     thread_mapping_cptr_t thread_mapping,
		     const cpu_model_t &cpu_model);
    sample_mapping_result_cptr_t get() const;

protected:
    string name() const;
    sample_mapping_result_cptr_t calculate() const;

private:
    const string sample_file;
    const fnr_stack_depth_t stack_depth;
    const fnr_line_size_t line_size;
    const fnr_line_size_t page_size;
    const thread_mapping_cptr_t thread_mapping;
    const cpu_model_t cpu_model;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
