/// @file analysis/thread_mapping/thread_mapping.hh

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#ifndef THREAD_MAPPING_HH
#define THREAD_MAPPING_HH

#include "analysis/call_stack/stack_pc_map.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace std;

using boost::shared_ptr;


// ----- thread_mapping_result_t ----------------------------------------------

/// ID of a cache.
typedef uint32_t tm_cache_id_t;

#define TM_CACHE_ID_ILLEGAL ((tm_cache_id_t)-1)

/// Mapping between thread ids and cache ids. Cache ids are always
/// allocated in sequence starting from 0.
typedef map<fnr_thread_id_t, tm_cache_id_t> tm_cache_map_t;

/// Mapping between global start time (inclusive) and a cache map
/// (tm_cache_map_t). The end time is specified implicitly by the start time of
/// the next mapping.
typedef map<fnr_time_t, tm_cache_map_t> tm_time_cache_map_t;

struct thread_mapping_burst_t {
    /**
     * Get a reference to the cache mapings for a given time.
     */
    tm_cache_map_t &get_cache_map(fnr_time_t time);
    /**
     * Get a reference to the cache mapings for a given time.
     */
    const tm_cache_map_t &get_cache_map(fnr_time_t time) const;

    /**
     * Translate a thread id at a given global time into a cache id.
     *
     * \note Do this manually using the results from get_cache_map if
     * you expect to resolve multiple cache ids for the same global
     * time.
     */
    tm_cache_id_t get_cache_id(fnr_time_t time, fnr_thread_id_t tid) const;

    /**
     * Convenience function to get cache id for a Fenris access.
     *
     * \note Do this manually using the results from get_cache_map if
     * you expect to resolve multiple cache ids for the same global
     * time.
     */
    tm_cache_id_t get_cache_id(const fnr_sample_access_t &access) const {
	return get_cache_id(access.time, access.thread);
    };

    /**
     * Map between global start times (inclusive) and thread to cache
     * mappings.
     */
    tm_time_cache_map_t time_cache_map;
};

typedef vector<thread_mapping_burst_t> thread_mapping_burst_vector_t;

struct thread_mapping_result_t {
    thread_mapping_result_t(parser_burst_id_t num_bursts)
	: bursts(num_bursts) {}
    thread_mapping_burst_vector_t bursts;

    /**
     * (Maximum) Number of caches in the system. Some thread mapping
     * modules (e.g. rr) may declare a higher number of caches than
     * they actually use.
     */
    tm_cache_id_t num_caches;
};

typedef shared_ptr<thread_mapping_result_t> thread_mapping_result_ptr_t;
typedef shared_ptr<const thread_mapping_result_t> thread_mapping_result_cptr_t;

// ----- thread_mapping_t -----------------------------------------------------

class thread_mapping_t {
public:
    virtual ~thread_mapping_t() {};
    virtual thread_mapping_result_cptr_t get() const = 0;
    virtual string name() const = 0;
};

typedef shared_ptr<const thread_mapping_t> thread_mapping_cptr_t;

END_NAMESPACE;
END_NAMESPACE;

#endif
