/// @file analysis/thread_mapping/time_mapping.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#define __STDC_LIMIT_MACROS

#include "time_mapping.hh"
#include "library/am_mem/am_mem.h"

#include <boost/lexical_cast.hpp>

#include <cmath>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace boost;
using namespace acumem::library;

// ----- 128-bit math functions. ----------------------------------------------

struct u128_t {
    uint64_t hi;
    uint64_t lo;
};

// TODO: Could be implemented in assembler on x86-64!
static u128_t
umul(uint64_t a, uint64_t b)
{
    uint64_t a0 = a & 0xffffffff;
    uint64_t a1 = a >> 32;
    uint64_t b0 = b & 0xffffffff;
    uint64_t b1 = b >> 32;

    uint64_t c = a0 * b0;
    uint64_t d = a0 * b1;
    uint64_t e = a1 * b0;
    uint64_t f = a1 * b1;

    uint64_t c0 = c & 0xffffffff;
    uint64_t c1 = c >> 32;
    uint64_t d0 = d & 0xffffffff;
    uint64_t d1 = d >> 32;
    uint64_t e0 = e & 0xffffffff;
    uint64_t e1 = e >> 32;
    uint64_t f0 = f & 0xffffffff;
    uint64_t f1 = f >> 32;

    uint64_t g = c0;
    uint64_t h = c1 + d0 + e0;
    uint64_t i = d1 + e1 + f0 + (h >> 32);
    uint64_t j = f1 + (i >> 32);

    u128_t r = {(j << 32) | (i & 0xffffffff), (h << 32) | (g & 0xffffffff)};
    return r;
}

// TODO: Could be implemented in assembler on x86-64!
static uint64_t
udiv(u128_t a, uint64_t b)
{
    uint64_t r = 0;

    for (int i = 63; i >= 0; i--) {
        uint64_t c = b << i;
        uint64_t d = i == 0 ? 0 : b >> (64 - i);

        if (a.hi > d || (a.hi == d && a.lo >= c)) {
            r |= (uint64_t)1 << i;
            a.hi -= d + (c > a.lo ? 1 : 0);
            a.lo -= c;
        }
    }

    return r;
}

static uint64_t
max(uint64_t a, uint64_t b)
{
    return a > b ? a : b;
}

static uint64_t
min(uint64_t a, uint64_t b)
{
    return a < b ? a : b;
}

// ----- time_mapping_result_t ------------------------------------------------

fnr_time_t
time_mapping_burst_t::get_local_time(tm_cache_id_t cache,
                                     fnr_time_t global_time) const
{
    if (valid) {
        fnr_time_vector_t::const_iterator it =
            upper_bound(global_times.begin(), global_times.end(), global_time);

	if (it == global_times.begin())
	    return 0;
	else
	    return cache_times[cache][it - global_times.begin() - 1];
    }
    else
	return global_time;
}

tm_local_time_t
time_mapping_burst_t::get_last_local_time(tm_cache_id_t cache) const
{
    if (valid) {
        const fnr_time_vector_t &cts = cache_times[cache];
        return cts.empty() ? 0 : cts.back();
    } else
	return last_global_time;
}

// ----- time_mapping_t -------------------------------------------------------

typedef map<fnr_time_t, fnr_thread_id_t> time_tid_t;

static void
fixup_time_vector(fnr_time_vector_t &v)
{
    sort(v.begin(), v.end());

    fnr_time_t t1 = FNR_TIME_MAX;
    fnr_time_vector_t::size_type unique_times = 0;
    BOOST_FOREACH(fnr_time_t t, v) {
        if (t != t1)
            unique_times++;
        t1 = t;
    }

    fnr_time_vector_t w;
    w.reserve(unique_times);

    fnr_time_t t2 = FNR_TIME_MAX;
    BOOST_FOREACH(fnr_time_t t, v) {
        if (t != t2)
            w.push_back(t);
        t2 = t;
    }

    v.swap(w);
}

static void
handle_access(
    time_tid_t &time_tid,
    const fnr_sample_access_t &access)
{
    time_tid_t::const_iterator i = time_tid.find(access.time);
    if (i != time_tid.end())
        assert(i->second == access.thread);
    else
        time_tid.insert(make_pair(access.time, access.thread));
}

time_mapping_result_cptr_t
time_mapping_t::calculate(feature_context_ptr_t context,
			  parser_result_cptr_t parser_result,
                          thread_mapping_result_cptr_t thread_mappings)
{
    time_mapping_result_ptr_t result(
	new time_mapping_result_t(parser_result->bursts.size(),
				  thread_mappings->num_caches));

    uint32_t num_threads = (uint32_t)parser_result->contexts.size();

    time_mapping_burst_vector_t::iterator it_time(result->bursts.begin());
    thread_mapping_burst_vector_t::const_iterator it_tmb(
	thread_mappings->bursts.begin());
    BOOST_FOREACH(const parser_burst_t &burst, parser_result->bursts) {
	context->feedback->test_cancel();

	const thread_mapping_burst_t &tmb(*it_tmb++);

	time_mapping_burst_t &time_burst(*it_time++);

        // Maps global times to the thread that did the access at that
        // time. There may be time mapping events at global times that do not
        // correspond to an access in a sample where we don't know what thread
        // did the access, though.
        time_tid_t time_tid;

        // All global times where we have a time mapping, may correspond to an
        // access in a sample or not.
        fnr_time_vector_t &global_times(time_burst.global_times);

        // A vector of cache local times corresponding to the global time at
        // the same index on the global_times vector for each cache, indexed by
        // cache id.
	vector<fnr_time_vector_t> &cache_times(time_burst.cache_times);

        // Global times when all thread's local times are known.
        fnr_time_vector_t known_global_times;

        // Map from global time to thread time when the thread time is known,
        // for each thread.
        vector<fnr_time_map_t> thread_times(num_threads);

        // Fill in time_tid.
        BOOST_FOREACH(const event_sample_ptr_t &s, burst.samples_by_pc1) {
            handle_access(time_tid, s->begin);
            BOOST_FOREACH(const fnr_sample_access_t &a, s->end)
                handle_access(time_tid, a);
        }

        // At global time 0 all thread times are 0.
        global_times.push_back(0);
        for (uint32_t tid = 0; tid < num_threads; tid++)
            thread_times[tid][0] = 0;

        // Fill in global and thread times from the time mapping events.
        for (uint32_t tid = 0; tid < burst.time_map->size(); tid++) {
            BOOST_FOREACH(const global_to_thread_t &gtt, burst.time_map->at(tid)) {
                global_times.push_back(gtt.global_time);
                thread_times[tid][gtt.global_time] = gtt.thread_time;
            }
        }

        // Sort the global_times vector and remove duplicate elements.
        fixup_time_vector(global_times);

        // If the global time equals the sum of the thread times, then we know
        // all thread times.
        BOOST_FOREACH(fnr_time_t global_time, global_times) {
            fnr_time_t thread_time_sum = 0;
            for (uint32_t tid = 0; tid < num_threads; tid++) {
                const fnr_time_map_t &time_map = thread_times[tid];
                fnr_time_map_t::const_iterator i = time_map.upper_bound(global_time);
                assert(i != time_map.begin());
                thread_time_sum += (--i)->second;
            }
            if (global_time == thread_time_sum) {
                assert(known_global_times.empty() || global_time > known_global_times.back());
                known_global_times.push_back(global_time);

                for (uint32_t tid = 0; tid < num_threads; tid++) {
                    fnr_time_map_t &time_map = thread_times[tid];
                    fnr_time_map_t::const_iterator i = time_map.upper_bound(global_time);
                    assert(i != time_map.begin());
                }
            }
        }

        // Create mappings from global times to cache time 0 for all global
        // times and caches.
        BOOST_FOREACH(fnr_time_vector_t &cts, cache_times) {
            cts.reserve(global_times.size());
            cts.resize(global_times.size(), 0);
        }

        // If all thread times are not known at the last global time we need to
        // make them up, so that we can interpolate the times all the way to
        // the end. This should be unusual and only affect a few samples even
        // when it happends, so handle it in the simplest possible way.
        if (known_global_times.back() != global_times.back()) {
            assert(known_global_times.back() < global_times.back());

            fnr_time_map_t total_tt;
            for (fnr_time_vector_t::const_iterator
                     ii = lower_bound(global_times.begin(),
                                      global_times.end(),
                                      known_global_times.back()),
                     ei = global_times.end();
                 ii != ei; ii++) {
                fnr_time_t global_time = *ii;
                fnr_time_t tt = 0;
                for (uint32_t tid = 0; tid < num_threads; tid++) {
                    fnr_time_map_t::const_iterator ti =
                        thread_times[tid].upper_bound(global_time);
                    ti--;
                    tt += ti->second;
                }
                total_tt[global_time] = tt;
            }

            typedef set<pair<fnr_time_t, uint32_t> > tt_t;
            tt_t tt_set;

            for (uint32_t tid = 0; tid < num_threads; tid++)
                tt_set.insert(
                    make_pair(thread_times[tid].rbegin()->first, tid));

            const fnr_time_t end_global_time = global_times.back();
            fnr_time_t end_total_tt = total_tt.rbegin()->second;
            BOOST_FOREACH(const tt_t::value_type &v, tt_set) {
                uint32_t tid = v.second;
                fnr_time_t local_time = thread_times[tid][v.first];

                fnr_time_t time_diff = end_global_time - end_total_tt;
                local_time += time_diff;
                end_total_tt += time_diff;
                thread_times[tid][global_times.back()] = local_time;
            }

            assert(end_total_tt == end_global_time);

            known_global_times.push_back(global_times.back());
        }

        int allocated_times = 0;
        fnr_time_t *times = NULL;
        fnr_time_t *mintime = NULL;
        fnr_time_t *maxtime = NULL;
        fnr_time_t *global = NULL;

        uint64_t *rest = (uint64_t *)am_malloc(num_threads * sizeof(uint64_t));
        uint64_t *space = (uint64_t *)am_malloc(num_threads * sizeof(uint64_t));

        // Step through the known global times.
        for (unsigned int kgti = 0; kgti < known_global_times.size() - 1; kgti++) {
            // Find the index of the start and end of the interval in the global_times vector.
            int start_gti = (int)(lower_bound(
                global_times.begin(), global_times.end(),
                known_global_times[kgti]) - global_times.begin());
            int end_gti = (int)(lower_bound(
                global_times.begin(), global_times.end(),
                known_global_times[kgti + 1]) - global_times.begin());

            int num_times = end_gti - start_gti + 1;

            // TODO: Optimize the case that some threads have the same start
            // and end time.

            // Allocate memory.
            if (num_times > allocated_times) {
                times = (fnr_time_t *)am_realloc(
                    times, num_threads * num_times * sizeof(fnr_time_t));
                mintime = (fnr_time_t *)am_realloc(
                    mintime, num_threads * num_times * sizeof(fnr_time_t));
                maxtime = (fnr_time_t *)am_realloc(
                    maxtime, num_threads * num_times * sizeof(fnr_time_t));
                global = (fnr_time_t *)am_realloc(
                    global, num_times * sizeof(fnr_time_t));
                allocated_times = num_times;
            }

            // Fill in the global times.
            for (int i = 0; i < num_times; i++)
                global[i] = global_times[start_gti + i];

            // Initialize the min and max times.
            for (int i = 0; i < num_times; i++)
                for (uint32_t tid = 0; tid < num_threads; tid++) {
                    mintime[i * num_threads + tid] = 0;
                    maxtime[i * num_threads + tid] = UINT64_MAX;
                }

            // Fill in local times and whether they are known or not.
            for (uint32_t tid = 0; tid < num_threads; tid++) {
                const fnr_time_map_t &tt = thread_times[tid];

                for (int i = 0; i < num_times; i++) {
                    fnr_time_map_t::const_iterator tti = tt.lower_bound(global[i]);
                    if (tti == tt.end() || tti->first != global[i]) {
                        if (i == 0 || i == num_times - 1) {
                            // Special handling of first and last times in the interval.
                            uint64_t time = (--tti)->second;
                            times[i * num_threads + tid] = time;
                            mintime[i * num_threads + tid] = time;
                            maxtime[i * num_threads + tid] = time;
                        }
                        else
                            // Local time not known, guess that it is unchanged.
                            times[i * num_threads + tid] = times[(i - 1) * num_threads + tid];
                    }
                    else {
                        // Known local time.
                        uint64_t time = tti->second;
                        times[i * num_threads + tid] = time;
                        mintime[i * num_threads + tid] = time;
                        maxtime[i * num_threads + tid] = time;
                    }
                }
            }

            // If there are no unknown times there is no need for any
            // interpolation.
            if (num_times == 2)
                goto no_unknown;

            // Fill in minimum times.
            for (int i = 1; i < num_times; i++)
                for (uint32_t tid = 0; tid < num_threads; tid++)
                    mintime[i * num_threads + tid] = max(
                        mintime[i * num_threads + tid],
                        mintime[(i - 1) * num_threads + tid]);

            // Fill in maximum times.
            for (int i = num_times - 2; i >= 0; i--) {
                time_tid_t::iterator ii = time_tid.find(global[i + 1]);
                uint32_t access_tid =
                    ii == time_tid.end() ? UINT32_MAX : ii->second.id;
                for (uint32_t tid = 0; tid < num_threads; tid++)
                    if (tid == access_tid)
                        maxtime[i * num_threads + tid] =
                            min(maxtime[i * num_threads + tid],
                                times[(i + 1) * num_threads + tid] - 1);
                    else
                        maxtime[i * num_threads + tid] =
                            min(maxtime[i * num_threads + tid],
                                maxtime[(i + 1) * num_threads + tid]);
            }

            // Make sure the minimums don't increase faster than the global
            // times.
            for (int i = num_times - 2; i >= 0; i--) {
                uint64_t global_diff = global[i + 1] - global[i];
                for (uint32_t tid = 0; tid < num_threads; tid++)
                    if (mintime[(i + 1) * num_threads + tid] > global_diff)
                        mintime[i * num_threads + tid] =
                            max(mintime[i * num_threads + tid],
                                mintime[(i + 1) * num_threads + tid] - global_diff);
            }

            // Make sure the maximums don't increase faster than the global
            // times.
            for (int i = 1; i < num_times; i++) {
                uint64_t global_diff = global[i] - global[i - 1];
                for (uint32_t tid = 0; tid < num_threads; tid++)
                    maxtime[i * num_threads + tid] =
                        min(maxtime[i * num_threads + tid],
                            maxtime[(i - 1) * num_threads + tid] + global_diff);
            }

            // TODO: Do interpolation here!

            // Adjust the local time so the sum matches the global time.
            for (int i = 1; i < num_times - 1; i++) {
                uint64_t sum = 0;
                for (uint32_t tid = 0; tid < num_threads; tid++)
                    sum += times[i * num_threads + tid];
                if (sum < global[i]) {
                    uint64_t rest_left = global[i] - sum;
                    uint64_t space_left = 0;
                    for (uint32_t tid = 0; tid < num_threads; tid++)
                        if (maxtime[i * num_threads + tid] > times[i * num_threads + tid])
                            space_left += space[tid] = maxtime[i * num_threads + tid] - times[i * num_threads + tid];
                        else
                            space[tid] = 0;
                    for (uint32_t tid = 0; tid < num_threads; tid++) {
                        if (space_left == 0)
                            break;
                        uint64_t rest_part = udiv(umul(rest_left, space[tid]), space_left);
                        times[i * num_threads + tid] += rest_part;
                        rest_left -= rest_part;
                        space_left -= space[tid];
                    }
                    assert(rest_left == 0);
                    assert(space_left == 0);
                }
                if (sum > global[i]) {
                    uint64_t rest_left = sum - global[i];
                    uint64_t space_left = 0;
                    for (uint32_t tid = 0; tid < num_threads; tid++)
                        if (times[i * num_threads + tid] > mintime[i * num_threads + tid])
                            space_left += space[tid] = times[i * num_threads + tid] - mintime[i * num_threads + tid];
                        else
                            space[tid] = 0;
                    for (uint32_t tid = 0; tid < num_threads; tid++) {
                        if (space_left == 0)
                            break;
                        uint64_t rest_part = udiv(umul(rest_left, space[tid]), space_left);
                        times[i * num_threads + tid] -= rest_part;
                        rest_left -= rest_part;
                        space_left -= space[tid];
                    }
                    assert(rest_left == 0);
                    assert(space_left == 0);
                }
            }

            // Adjust local times so they aren't below their minimum.
            for (int i = 1; i < num_times - 1; i++) {
                uint64_t rest_left = 0;
                for (uint32_t tid = 0; tid < num_threads; tid++)
                    if (mintime[i * num_threads + tid] > times[i * num_threads + tid]) {
                        rest_left += mintime[i * num_threads + tid] - times[i * num_threads + tid];
                        times[i * num_threads + tid] = mintime[i * num_threads + tid];
                    }
                if (rest_left == 0)
                    continue;
                uint64_t space_left = 0;
                for (uint32_t tid = 0; tid < num_threads; tid++)
                    if (times[i * num_threads + tid] > mintime[i * num_threads + tid])
                        space_left += space[tid] = times[i * num_threads + tid] - mintime[i * num_threads + tid];
                    else
                        space[tid] = 0;
                for (uint32_t tid = 0; tid < num_threads; tid++) {
                    if (space_left == 0)
                        break;
                    uint64_t rest_part = udiv(umul(rest_left, space[tid]), space_left);
                    times[i * num_threads + tid] -= rest_part;
                    rest_left -= rest_part;
                    space_left -= space[tid];
                }
                assert(rest_left == 0);
                assert(space_left == 0);
            }

            // Adjust local times so they aren't above their maximum.
            for (int i = 1; i < num_times - 1; i++) {
                uint64_t rest_left = 0;
                for (uint32_t tid = 0; tid < num_threads; tid++)
                    if (times[i * num_threads + tid] > maxtime[i * num_threads + tid]) {
                        rest_left += times[i * num_threads + tid] - maxtime[i * num_threads + tid];
                        times[i * num_threads + tid] = maxtime[i * num_threads + tid];
                    }
                if (rest_left == 0)
                    continue;
                uint64_t space_left = 0;
                for (uint32_t tid = 0; tid < num_threads; tid++)
                    if (maxtime[i * num_threads + tid] > times[i * num_threads + tid])
                        space_left += space[tid] = maxtime[i * num_threads + tid] - times[i * num_threads + tid];
                    else
                        space[tid] = 0;
                for (uint32_t tid = 0; tid < num_threads; tid++) {
                    if (space_left == 0)
                        break;
                    uint64_t rest_part = udiv(umul(rest_left, space[tid]), space_left);
                    times[i * num_threads + tid] += rest_part;
                    rest_left -= rest_part;
                    space_left -= space[tid];
                }
                assert(rest_left == 0);
                assert(space_left == 0);
            }

            // Try to avoid making time go backwards, working forward from low
            // times to high.
            for (int i = 1; i < num_times - 1; i++)
                for (uint32_t tid = 0; tid < num_threads; tid++)
                    if (times[(i - 1) * num_threads + tid] > times[i * num_threads + tid])
                        times[i * num_threads + tid] = min(
                            times[(i - 1) * num_threads + tid],
                            maxtime[i * num_threads + tid]);

            // Try to avoid making time go backwards, working backward from
            // high times to low.
            for (int i = num_times - 2; i > 0; i--)
                for (uint32_t tid = 0; tid < num_threads; tid++)
                    if (times[(i + 1) * num_threads + tid] < times[i * num_threads + tid])
                        times[i * num_threads + tid] = max(
                            times[(i + 1) * num_threads + tid],
                            mintime[i * num_threads + tid]);

        no_unknown:
            // Verify the result.
#if 0 // The current interpolation doesn't guarantee that the row sums are
      // correct, so don't check them.
            for (int i = 0; i < num_times; i++) {
                uint64_t sum = 0;
                for (uint32_t tid = 0; tid < num_threads; tid++)
                    sum += times[i * num_threads + tid];
                assert(sum == global[i]);
            }
#endif
            for (int i = 1; i < num_times; i++)
                for (uint32_t tid = 0 ; tid < num_threads; tid++) {
                    time_tid_t::iterator ii = time_tid.find(global[i]);
                    if (ii != time_tid.end() && ii->second == tid)
                        assert(times[i * num_threads + tid] > times[(i - 1) * num_threads + tid]);
                    else
                        assert(times[i * num_threads + tid] >= times[(i - 1) * num_threads + tid]);
                }

            // Add each thread's local times to the corresponding cache's times.
            for (uint32_t tid = 0; tid < num_threads; tid++) {
                fnr_time_t last_thread_time = times[0 * num_threads + tid];
                for (int i = 1; i < num_times; i++) {
                    fnr_time_t global_time = global[i];
                    fnr_time_t thread_time = times[i * num_threads + tid];

                    if (thread_time > last_thread_time) {
                        const tm_cache_map_t &cache_map(tmb.get_cache_map(global_time));
                        const tm_cache_map_t::const_iterator ci(cache_map.find(tid));
                        assert(ci != cache_map.end());
                        const tm_cache_id_t cid(ci->second);

                        cache_times[cid][start_gti + i] += thread_time - last_thread_time;
                        last_thread_time = thread_time;
                    }
                }
            }
        }

        am_free(times);
        am_free(mintime);
        am_free(maxtime);
        am_free(global);

        am_free(rest);
        am_free(space);

        // Accumulate cache times.
        BOOST_FOREACH(fnr_time_vector_t &cts, cache_times) {
            fnr_time_t cache_time = 0;
            BOOST_FOREACH(fnr_time_t &ct, cts){
                cache_time += ct;
                ct = cache_time;
            }
        }

	time_burst.last_global_time = burst.begin_time + burst.length;
	time_burst.valid = parser_result->mt_sampled();
    }

    return result;
}

time_mapping_t::time_mapping_t(feature_context_ptr_t context,
                               const string &sample_file,
                               thread_mapping_cptr_t thread_mapping)
    : feature_t<time_mapping_result_cptr_t>(context),
      sample_file(sample_file),
      thread_mapping(thread_mapping)
{
}

time_mapping_result_cptr_t
time_mapping_t::get() const
{
    return get_data();
}

string
time_mapping_t::name() const
{
    return "time_mapping_t(" +
        sample_file + "," +
        thread_mapping->name() + ")";
}

time_mapping_result_cptr_t
time_mapping_t::calculate() const
{
    return calculate(context,
		     parser_t(context, sample_file).get(),
		     thread_mapping->get());
}

END_NAMESPACE;
END_NAMESPACE;
