/// @file analysis/thread_mapping/time_mapping.hh

// Copyright (c) 2008, 2010 Acumem AB.  All rights reserved.

#ifndef TIME_MAPPING_HH
#define TIME_MAPPING_HH

#include <vector>

#include "thread_mapping.hh"
#include "analysis/call_stack/stack_pc_map.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace std;
using namespace acumem::library;

// ----- time_mapping_result_t ------------------------------------------------

/// A cache local time, i.e., the number of memory accesses performed in a
/// cache. E.g., with a static thread to cache mapping this is the sum of the
/// local times of the threads mapped to that cache.
typedef fnr_time_t tm_local_time_t;

struct time_mapping_burst_t {
    time_mapping_burst_t(tm_cache_id_t num_caches)
	: cache_times(num_caches) {}

    /**
     * Get the cache local time for a given cache at a given global
     * time. The function returns the global time as a fall back if no
     * time mappings are available, i.e., the #valid flag is false.
     */
    tm_local_time_t get_local_time(tm_cache_id_t cache,
				   fnr_time_t global_time) const;

    /**
     * Returns the last seen cache local time.
     */
    tm_local_time_t get_last_local_time(tm_cache_id_t cache) const;

    /// Sorted vector of global times where we have time mappings;
    fnr_time_vector_t global_times;

    /// Vectors of cache local times indexed by cache id. The indexes in each
    /// time vector corresponds to the indexes in the global_times vector.
    vector<fnr_time_vector_t> cache_times;

    /**
     * Last global time reported by the parser.
     */
    fnr_time_t last_global_time;

    /**
     * The time mapping results are only valid if valid time mappings
     * are available (i.e., the sample file was produced with the MP
     * sampler/resampler).
     */
    bool valid;
};

typedef vector<time_mapping_burst_t> time_mapping_burst_vector_t;

struct time_mapping_result_t {
    time_mapping_result_t(parser_burst_id_t num_bursts,
			  tm_cache_id_t num_caches)
	: bursts(num_bursts, time_mapping_burst_t(num_caches)) {}

    time_mapping_burst_vector_t bursts;
};

typedef shared_ptr<time_mapping_result_t> time_mapping_result_ptr_t;
typedef shared_ptr<const time_mapping_result_t> time_mapping_result_cptr_t;

// ----- time_mapping_t -------------------------------------------------------

class time_mapping_t : public feature_t<time_mapping_result_cptr_t> {
public:
    static time_mapping_result_cptr_t calculate(
	feature_context_ptr_t context,
	parser_result_cptr_t parser_result,
        thread_mapping_result_cptr_t thread_mappings);

    time_mapping_t(feature_context_ptr_t context,
                   const string &sample_file,
                   thread_mapping_cptr_t thread_mapping);
    time_mapping_result_cptr_t get() const;

protected:
    string name() const;
    time_mapping_result_cptr_t calculate() const;

private:
    const string sample_file;
    const thread_mapping_cptr_t thread_mapping;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
