/// @file analysis/utilization/potential.cc

// Copyright (c) 2007-2008 Acumem AB.  All rights reserved.

#include "analysis/utilization/potential.hh"
#include "analysis/fetch_ratio/fr_lookup.hh"
#include <boost/lexical_cast.hpp>
#include <cmath>
#ifdef _WIN32
// Visual Studio 2005 math.h is missing log2
#define _USE_MATH_DEFINES
#include <math.h>
template <typename T>
inline T log2(T t)
{
    return log(t)/M_LN2;
}
#endif

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace boost;
using namespace std;

// ----- potential_result_t ---------------------------------------------------

potential_result_t::potential_result_t(double util_corr_ratio,
                                       double avoidable_count)
    : util_corr_ratio(util_corr_ratio),
      avoidable_count(avoidable_count)
{
}

// ----- potential_t ----------------------------------------------------------

potential_t::potential_t(const feature_context_ptr_t &context,
                         const string &sample_file,
                         const fnr_stack_depth_t stack_depth,
                         const fnr_line_size_t line_size,
                         const fnr_line_size_t page_size,
                         const thread_mapping_cptr_t &thread_mapping,
			 const cpu_model_t &cpu_model,
                         const locationx_mapper_cptr_t &lx_mapper,
                         const fnr_cache_size_t cache_size,
                         const repl_fr_engine_cptr_t &engine,
			 const windowing_engine_cptr_t &windowing,
			 const utilization_type_t type)
    : context(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      lx_mapper(lx_mapper),
      cache_size(cache_size),
      engine(engine),
      windowing(windowing),
      type(type)
{
}

potential_result_t
potential_t::get(const locationx_set_t &lx_set) const
{
    fr_lookup_result_cptr_t fr_res(
	fr_lookup_t(context, sample_file, stack_depth, line_size, page_size,
		    thread_mapping, cpu_model, lx_mapper, cache_size, engine,
		    FRL_Type_All).get());

    utilization_result_cptr_t su_res(
	utilization_t(context, sample_file, stack_depth, line_size, page_size,
		      thread_mapping, cpu_model, lx_mapper, cache_size,
		      windowing, type).get());

    fr_data_t fr_data(fr_res->fr_by_lx_set(lx_set));

    // get the spatial usage at the queried cache size
    double su = su_res->get_utilization(lx_set).value();

    // just to guard against nan
    if (su < 0.001)
	su = 0.001;

    // calculate the shift and bracketing cache sizes
    double log2_cache_size           = log2(cache_size / su);
    int log2_cache_size_low          = (int)(floor(log2_cache_size) + 0.5);
    double log2_cache_size_frac      = log2_cache_size - log2_cache_size_low;
    fnr_cache_size_t cache_size_low  = 1LL << log2_cache_size_low;
    fnr_cache_size_t cache_size_high = cache_size_low * 2;

    // Get data for the cache sizes and do the linear interpolation
    // and scaling
    fr_lookup_result_cptr_t fr_res_low =
        fr_lookup_t(context,
                    sample_file,
                    stack_depth,
                    line_size,
		    page_size,
                    thread_mapping,
		    cpu_model,
		    lx_mapper,
                    cache_size_low,
                    engine,
		    FRL_Type_All).get();
    fr_lookup_result_cptr_t fr_res_high =
        fr_lookup_t(context,
                    sample_file,
                    stack_depth,
                    line_size,
		    page_size,
                    thread_mapping,
		    cpu_model,
		    lx_mapper,
                    cache_size_high,
		    engine,
		    FRL_Type_All).get();

    // Get fetch ratio data for this lx set.
    double fr;
    double fr_low;
    double fr_high;

    switch (type) {
    case Utilization_Classic:
    case Utilization_Fetch:
	fr = fr_data.by_kind().total_fetch_ratio();
	fr_low = fr_res_low->fr_by_lx_set(lx_set).by_kind().total_fetch_ratio();
	fr_high = fr_res_high->fr_by_lx_set(lx_set).by_kind().total_fetch_ratio();
	break;
    case Utilization_Write_Back:
	fr = fr_data.by_kind().total_write_back_ratio();
	fr_low = fr_res_low->fr_by_lx_set(lx_set).by_kind().total_write_back_ratio();
	fr_high = fr_res_high->fr_by_lx_set(lx_set).by_kind().total_write_back_ratio();
	break;
    default:
	assert(0);
    }

    double su_high = utilization_t(
        context, sample_file, stack_depth, line_size, page_size,
        thread_mapping, cpu_model, lx_mapper, cache_size_high,
	windowing, type).get()->get_utilization(lx_set).value();

    double su_low = utilization_t(
        context, sample_file, stack_depth, line_size, page_size,
        thread_mapping, cpu_model, lx_mapper, cache_size_low,
	windowing, type).get()->get_utilization(lx_set).value();

    double su2 =
	su_low * (1 - log2_cache_size_frac) +
	su_high * log2_cache_size_frac;

    double util_corr_fr = (fr_low * (1 - log2_cache_size_frac) +
			   fr_high * log2_cache_size_frac) * su2;

    // Protect against the case that the utilization corrected fetch ratio is
    // greater than the fetch ratio. In that case, simply scale fr with
    // su. Don't bother with the scaling. Apparently there is a drop somewhere
    // earlier in the interval that we don't see, and the interpolation is to
    // crude.
    if (util_corr_fr > fr)
	util_corr_fr = fr * su;

    double avoidable_fetches = (fr - util_corr_fr) * fr_data.accesses();

    return potential_result_t(util_corr_fr, avoidable_fetches);

}

potential_result_t
potential_t::get(const locationx_t lx) const
{
    locationx_set_t lx_set;
    lx_set.insert(lx);
    return get(lx_set);
}

END_NAMESPACE;
END_NAMESPACE;
