/// @file analysis/utilization/potential.hh

// Copyright (c) 2007-2008 Acumem AB.  All rights reserved.

#ifndef POTENTIAL_HH
#define POTENTIAL_HH

#include "analysis/locationx/locationx.hh"
#include "analysis/utilization/utilization.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace std;

using boost::shared_ptr;

// ----- potential_result_t ---------------------------------------------------

struct potential_result_t {
    potential_result_t(double util_corr_ratio, double avoidable_count);

    double util_corr_ratio;
    double avoidable_count;
};

// ----- potential_t ----------------------------------------------------------

/**
 * Calculate how many misses could be removed if perfect
 * utilization could be achieved.
 * The goal is to have perfect spatial locality for each field for
 * each loop. If all memory holes are removed, then the desired
 * spatial usage is 100% for the target cache size. If not all
 * memory holes for this structure are removed, it will be
 * less. It is generally found by evaluating spatial usage at
 * infinity.
 *
 * \f{eqnarray*}
 *   S                  & = & \frac{Su(\infty)}{Su(cache\_size)}              \\
 *   util\_corr\_fr    & = & \frac{fr(cache\_size * S)}{S}                   \\
 *   avoidable\_fetches & = &
 *     \frac{(fr(cache\_size) - util\_corr\_fr) * \#samples}{sample\_rate}
 * \f}
 *
 * XXX We can't currently get a reliable measurement of
 * \f$ Su(\infty) \f$. We assume it to be 1.
 *
 * \f$ fr(cache\_size * S) == fr(\frac{cache\_size}{su(cache\_size)}) \f$
 * will involve calculating fr for a huge number of odd cache-sizes.
 * We will therefore calculate fr for even powers of two and
 * interpolate between them.
 */
class potential_t {
public:
    potential_t(const feature_context_ptr_t &context,
                const string &sample_file,
                const fnr_stack_depth_t stack_depth,
                const fnr_line_size_t line_size,
                const fnr_line_size_t page_size,
                const thread_mapping_cptr_t &thread_mapping,
		const cpu_model_t &cpu_model,
                const locationx_mapper_cptr_t &lx_mapper,
                const fnr_cache_size_t cache_size,
                const repl_fr_engine_cptr_t &engine,
		const windowing_engine_cptr_t &windowing,
		const utilization_type_t type);

    potential_result_t get(const locationx_set_t &lx_set) const;
    potential_result_t get(const locationx_t lx) const;

private:
    const feature_context_ptr_t context;
    const string sample_file;
    const fnr_stack_depth_t stack_depth;
    const fnr_line_size_t line_size;
    const fnr_line_size_t page_size;
    const thread_mapping_cptr_t thread_mapping;
    const cpu_model_t cpu_model;
    const locationx_mapper_cptr_t lx_mapper;
    const fnr_cache_size_t cache_size;
    const repl_fr_engine_cptr_t engine;
    const windowing_engine_cptr_t windowing;
    const utilization_type_t type;
};

typedef shared_ptr<potential_t> potential_ptr_t;

END_NAMESPACE;
END_NAMESPACE;

#endif
