/// @file analysis/utilization/utilization.cc

// Copyright (c) 2006-2009 Acumem AB.  All rights reserved.

#include "analysis/utilization/utilization.hh"
#include "analysis/fetch_ratio/repl_fr_random.hh"
#include <boost/lexical_cast.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace boost;
using namespace std;

struct utilization_classic_t {
    static inline bool numerator_filter(const mapped_repl_byte_t &bs) {
	return true;
    }

    static inline bool numerator_filter(const mapped_cold_t &bs) {
	return true;
    }

    static inline double add_line_repl(const by_kind_t &fetches) {
	return
            fetches.replacement_fetches() +
            fetches.coherence_fetches() +
            fetches.flush_fetches();
    }

    static inline double add_line_cold(const by_kind_t &fetches) {
	return fetches.cold_fetches();
    }

    static inline double add_sample(const mapped_repl_byte_t &bs,
				    double repl_fr) {
	return bs.invalidated || bs.flushed ? 1.0 : repl_fr;
    }
};

struct utilization_write_t {
    static inline bool numerator_filter(const mapped_repl_byte_t &bs) {
	return bs.end()->is_write();
    }

    static inline bool numerator_filter(const mapped_cold_t &bs) {
	return bs.end.is_write();
    }

    static inline double add_line_repl(const by_kind_t &fetches) {
        // TODO: Should probably do some special treatment of non-temporal
        //       stores. See ticket #1511, and also TODO below.
	return
            fetches.replacement_write_backs() +
            fetches.coherence_write_backs() +
            fetches.flush_write_backs();
    }

    static inline double add_line_cold(const by_kind_t &fetches) {
	return fetches.cold_write_backs();
    }

    static inline double add_sample(const mapped_repl_byte_t &bs,
				    double repl_fr) {
	if (bs.end()->is_nta())
            // TODO: Isn't this assumption wrong? A more reasonable assumption
            // would be that all bytes written by non-temporal stores are
            // misses, since that is how they are supposed too be used. This
            // seems to be a more reasonable assumption for line samples. See
            // ticket #1511.

	    // We assume that a sample ending on a Wnt always misses
	    // unless the start is a Wnt.
	    return bs.start()->is_write() && bs.start()->is_nta() ? 0.0 : 1.0;
	else
	    return (bs.invalidated ||
		    bs.downgraded ||
		    bs.flushed) ? 1.0 : repl_fr;
    }
};

struct utilization_read_t {
    static inline bool numerator_filter(const mapped_repl_byte_t &bs) {
	return bs.end()->is_read();
    }

    static inline bool numerator_filter(const mapped_cold_t &bs) {
	return bs.end.is_read();
    }

    static inline double add_line_repl(const by_kind_t &fetches) {
	return
            fetches.replacement_fetches() +
            fetches.coherence_fetches() +
            fetches.flush_fetches();
    }

    static inline double add_line_cold(const by_kind_t &fetches) {
	return fetches.cold_fetches();
    }

    static inline double add_sample(const mapped_repl_byte_t &bs,
				    double repl_fr) {
	// Apart from the obvious cases (foreign invalidations,
	// flushes), we assume that a sample starting at a Wnt always
	// misses on a subsequent read.
	return (bs.invalidated ||
		bs.flushed ||
		(bs.start()->is_write() && bs.start()->is_nta())) ?
	    1.0 : repl_fr;
    }
};

// ----- utilization_result_t -------------------------------------------------

utilization_datum_t
utilization_result_t::get_data(const locationx_set_t &lx_set) const
{
    utilization_datum_t d;

    BOOST_FOREACH(const locationx_t &lx, lx_set)
	d.add(data[lx.id]);

    return d;
}

utilization_datum_t
utilization_result_t::get_data(const locationx_t &lx) const
{
    return data[lx.id];
}

utilization_result_t::utilization_result_t(locationx_id_t lx_count,
					   fnr_line_size_t line_size)
    : data(lx_count), line_size(line_size)
{
}

result_valid_t
utilization_result_t::get_utilization(const locationx_set_t &lx_set) const
{
    utilization_datum_t d(get_data(lx_set));

    /// @todo XXX investigate if not num, den, cold_num and cold_den should be
    /// scaled with respect to their sample counts?
    return result_valid_t(d.valid(line_size), d.utilization());
}

result_valid_t
utilization_result_t::get_utilization(const locationx_t &lx) const
{
    utilization_datum_t d(get_data(lx));
    return result_valid_t(d.valid(line_size), d.utilization());
}

result_valid_t
utilization_result_t::get_utilization_no_cold(const locationx_set_t &lx_set) const
{
    utilization_datum_t d(get_data(lx_set));
    return result_valid_t(d.valid(line_size), d.utilization_no_cold());
}

result_valid_t
utilization_result_t::get_utilization_no_cold(const locationx_t &lx) const
{
    utilization_datum_t d(get_data(lx));
    return result_valid_t(d.valid(line_size), d.utilization_no_cold());
}

result_valid_t
utilization_result_t::get_wasted_space(const locationx_set_t &lx_set) const
{
    utilization_datum_t d(get_data(lx_set));

    /// @todo XXX investigate if not num, den, cold_num and cold_den should be
    /// scaled with respect to their sample counts?
    return result_valid_t(d.valid(line_size), d.wasted_space());
}

result_valid_t
utilization_result_t::get_wasted_space(const locationx_t &lx) const
{
    utilization_datum_t d(get_data(lx));
    return result_valid_t(d.valid(line_size), d.wasted_space());
}

// ----- utilization_t --------------------------------------------------------

template<class UtilizationType>
static utilization_result_cptr_t
internal_calculate(
    const sample_mapping_result_cptr_t &sample_mapping,
    const locationx_mapper_result_cptr_t &lx_mapping,
    const fr_lookup_result_cptr_t &ls_frl,
    const repl_fr_result_cptr_t &bs_fr,
    const fnr_line_size_t line_size)
{
    utilization_result_ptr_t result(
	new utilization_result_t(lx_mapping->max_id, line_size));

    // Q&A:
    //
    // Q: Why don't PF accesses get special treatment for line size samples?
    // A: Byte size samples are used to calculate how much data is used, line
    //    size samples are used to calculate how much data is brought into the
    //    cache. A prefetch doesn't use any data (special treatment for byte
    //    size samples), but it still brings the cache line into the cache like
    //    any other instruction (no special treatment for line size samples).

    // Sum up the parts of spatial usage formula.
    utilization_data_vector_t &udv(result->data);

    // Line samples
    for (locationx_id_t i = 0; i < lx_mapping->max_id; i++) {
	const locationx_t lx(i);
	const fr_data_t frd(ls_frl->fr_by_lx(lx));
	if (frd.samples() > 0) {
	    const by_kind_t fetches(frd.by_kind());
	    udv[i].add_line(line_size, frd.accesses(), frd.samples(),
			    UtilizationType::add_line_repl(fetches));
	    udv[i].add_line_cold(line_size, frd.accesses(), frd.samples(),
				 UtilizationType::add_line_cold(fetches));
	}
    }

    // Handle byte size samples
    repl_fr_burst_vector_t::const_iterator bs_fr_it(bs_fr->bursts.begin());
    BOOST_FOREACH(const sample_mapping_burst_t &smb, sample_mapping->bursts) {
	const ratio_vector_t &bs_repl_fr((bs_fr_it++)->val_to_end);
	const mapped_repl_byte_vector_t &byte_repl(smb.repl_byte);
	const mapped_cold_vector_t &byte_cold(smb.cold_byte);

	// Byte size replacement samples.
	fnr_sample_count_t num_byte_repl((fnr_sample_count_t)byte_repl.size());
	for (fnr_sample_count_t i = 0; i < num_byte_repl; i++) {
	    const mapped_repl_byte_t &bs(byte_repl[i]);
	    if (UtilizationType::numerator_filter(bs))
		udv[lx_mapping->get_locationx(*bs.end()).id].add_byte(
		    smb.sample_period,
		    UtilizationType::add_sample(bs, bs_repl_fr[i]));
	}

	// Byte size cold samples.
	fnr_sample_count_t num_byte_cold((fnr_sample_count_t)byte_cold.size());
	for (fnr_sample_count_t i = 0; i < num_byte_cold; i++) {
	    const mapped_cold_t &bs(byte_cold[i]);
	    if (UtilizationType::numerator_filter(bs))
		udv[lx_mapping->get_locationx(bs.end).id].add_byte_cold(
		    smb.sample_period);
	}
    }

    return result;
}

/* This is the "classic" utilization metric, denoted as "Cache Line
 * Utilization" in the MT document.
 */
utilization_result_cptr_t
utilization_t::calculate_classic(
    const sample_mapping_result_cptr_t &sample_mapping,
    const locationx_mapper_result_cptr_t &lx_mapping,
    const fr_lookup_result_cptr_t &ls_frl,
    const repl_fr_result_cptr_t &bs_fr,
    const fnr_line_size_t line_size)
{
    return internal_calculate<utilization_classic_t>(sample_mapping, lx_mapping,
						     ls_frl, bs_fr, line_size);
}

utilization_result_cptr_t
utilization_t::calculate_read(
    const sample_mapping_result_cptr_t &sample_mapping,
    const locationx_mapper_result_cptr_t &lx_mapping,
    const fr_lookup_result_cptr_t &ls_frl,
    const repl_fr_result_cptr_t &bs_fr,
    const fnr_line_size_t line_size)
{
    return internal_calculate<utilization_read_t>(sample_mapping, lx_mapping,
						  ls_frl, bs_fr, line_size);
}


utilization_result_cptr_t
utilization_t::calculate_write(
    const sample_mapping_result_cptr_t &sample_mapping,
    const locationx_mapper_result_cptr_t &lx_mapping,
    const fr_lookup_result_cptr_t &ls_frl,
    const repl_fr_result_cptr_t &bs_fr,
    const fnr_line_size_t line_size)
{
    return internal_calculate<utilization_write_t>(sample_mapping, lx_mapping,
						   ls_frl, bs_fr, line_size);
}

utilization_t::utilization_t(
    const feature_context_ptr_t &context,
    const string &sample_file,
    const fnr_stack_depth_t stack_depth,
    const fnr_line_size_t line_size,
    const fnr_line_size_t page_size,
    const thread_mapping_cptr_t &thread_mapping,
    const cpu_model_t &cpu_model,
    const locationx_mapper_cptr_t &lx_mapper,
    const fnr_cache_size_t cache_size,
    const windowing_engine_cptr_t &windowing,
    const utilization_type_t type)
    : feature_t<utilization_result_cptr_t>(context),
      sample_file(sample_file),
      stack_depth(stack_depth),
      line_size(line_size),
      page_size(page_size),
      thread_mapping(thread_mapping),
      cpu_model(cpu_model),
      lx_mapper(lx_mapper),
      cache_size(cache_size),
      windowing(windowing),
      type(type)
{
}

utilization_result_cptr_t
utilization_t::get(bool enable_cache) const
{
    return get_data(enable_cache);
}

string
utilization_t::name() const
{
    return "utilization(" +
        sample_file + "," +
        lexical_cast<string>(stack_depth) + "," +
        lexical_cast<string>(line_size) + "," +
        lexical_cast<string>(page_size) + "," +
        thread_mapping->name() + "," +
	cpu_model.name() + "," +
        lx_mapper->name() + "," +
        lexical_cast<string>(cache_size) + "," +
	windowing->name() + "," +
	lexical_cast<string>(type) + ")";
}

utilization_result_cptr_t
utilization_t::calculate() const
{
    sample_mapping_result_cptr_t sample_mapping(sample_mapping_t(
        context, sample_file, stack_depth, line_size, page_size,
	thread_mapping, cpu_model).get());
    locationx_mapper_result_cptr_t lx_mapping(lx_mapper->get());

    /// @todo We always calculate spatial usage using random replacement
    /// because of problems with spatial usage and LRU. See ticket #285.
    repl_fr_engine_cptr_t fr_engine(
	new repl_fr_random_t(context, sample_file, stack_depth, thread_mapping,
                             windowing, cpu_model));
    repl_fr_result_cptr_t bs_fr(repl_fr_t(
        context, sample_file, stack_depth, Mapped_Line, Mapped_Byte,
	line_size, page_size, cache_size, thread_mapping, cpu_model,
	fr_engine).get());

    fr_lookup_result_cptr_t frl_res =
        fr_lookup_t(context, sample_file, stack_depth, line_size, page_size,
                    thread_mapping, cpu_model, lx_mapper, cache_size,
                    fr_engine, FRL_Type_All).get();

    switch (type) {
    default:
	assert(0);
    case Utilization_Classic:
	return calculate_classic(sample_mapping, lx_mapping, frl_res,
				 bs_fr, line_size);
    case Utilization_Fetch:
	return calculate_read(sample_mapping, lx_mapping, frl_res,
			      bs_fr, line_size);
    case Utilization_Write_Back:
	return calculate_write(sample_mapping, lx_mapping, frl_res,
			       bs_fr, line_size);
    }
}

END_NAMESPACE;
END_NAMESPACE;
