/// @file analysis/utilization/utilization.hh

// Copyright (c) 2006-2009 Acumem AB.  All rights reserved.

#ifndef UTILIZATION_HH
#define UTILIZATION_HH

#include "analysis/fetch_ratio/repl_fr.hh"
#include "analysis/fetch_ratio/fr_lookup.hh"
#include "analysis/locationx/locationx.hh"
#include "analysis/misc/result_valid.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(analysis);

using namespace acumem::library;
using namespace std;

using boost::shared_ptr;

// ----- utilization_result_t -------------------------------------------------

struct utilization_datum_t {
    utilization_datum_t()
	: raw_wasted_space(0.0),
	  byte(0.0), byte_cold(0.0),
	  line(0.0), line_cold(0.0) {}

    void add(const utilization_datum_t &datum) {
	raw_wasted_space += datum.raw_wasted_space;
	byte += datum.byte;
	byte_cold += datum.byte_cold;
	line += datum.line;
	line_cold += datum.line_cold;
    }

    void add_byte(double sample_period, double fr) {
	raw_wasted_space -= (stat_fp_t)(sample_period * fr);
	byte += (stat_fp_t)fr;
    }

    void add_byte_cold(double sample_period) {
	raw_wasted_space -= (stat_fp_t)sample_period;
	byte_cold += 1;
    }

    void add_line(fnr_line_size_t line_size,
		  double access_count, double sample_count,
		  double fetches) {
	const double fr(fetches / access_count);
	raw_wasted_space += (stat_fp_t)(line_size * fetches);
	line += (stat_fp_t)(line_size * fr * sample_count);
    }

    void add_line_cold(fnr_line_size_t line_size,
		       double access_count, double sample_count,
		       double fetches) {
	const double fr(fetches / access_count);
	raw_wasted_space += (stat_fp_t)(line_size * fetches);
	line_cold += (stat_fp_t)(line_size * fr * sample_count);
    }

    /// This condition is somewhat arbitrary... The idea is that cold
    /// misses usually don't provide good data for the utilization
    /// calculation, unless there is a lot of them, having few line
    /// misses isn't neccessarily bad, as long as they aren't
    /// dominated by few cold misses.
    bool valid(fnr_line_size_t line_size) const {
	return (line > 5 * line_cold) || line_cold > 100 * line_size;
    }

    /// Spatial usage of the data set.
    double utilization() const {
	return min(1.0f, (byte + byte_cold) / (line + line_cold));
    }

    /// Spatial usage of the data set, excluding cold samples.
    double utilization_no_cold() const {
	return min(1.0f, byte / line);
    }

    /// Wasted space for the data set.
    double wasted_space() const {
	return max(0.0f, raw_wasted_space);
    }

    /// The wasted space is stored separately since this value needs
    /// to include the sample period, which might vary between bursts.
    stat_fp_t raw_wasted_space;

    stat_fp_t byte;
    stat_fp_t byte_cold;
    stat_fp_t line;
    stat_fp_t line_cold;
};

typedef vector<utilization_datum_t> utilization_data_vector_t;

struct utilization_result_t {
    utilization_result_t(locationx_id_t lx_count,
			 fnr_line_size_t line_size);

    //@{
    /// Data aggregation functions.
    utilization_datum_t get_data(const locationx_set_t &lx_set) const;
    utilization_datum_t get_data(const locationx_t &lx) const;
    //@}

    result_valid_t get_utilization(const locationx_set_t &lx_set) const;
    result_valid_t get_utilization(const locationx_t &lx) const;
    result_valid_t get_utilization_no_cold(const locationx_set_t &pc_set) const;
    result_valid_t get_utilization_no_cold(const locationx_t &lx) const;

    // Return the amount of wasted space in bytes.
    result_valid_t get_wasted_space(const locationx_set_t &lx_set) const;
    result_valid_t get_wasted_space(const locationx_t &lx) const;

    utilization_data_vector_t data;
    fnr_line_size_t line_size;
};

typedef shared_ptr<utilization_result_t> utilization_result_ptr_t;
typedef shared_ptr<const utilization_result_t> utilization_result_cptr_t;

// ----- utilization_t --------------------------------------------------------

enum utilization_type_t {
    Utilization_Classic = 0,
    Utilization_Fetch,
    Utilization_Write_Back
};

class utilization_t : feature_t<utilization_result_cptr_t> {
public:
    static utilization_result_cptr_t calculate_classic(
       const sample_mapping_result_cptr_t &sample_mapping,
       const locationx_mapper_result_cptr_t &lx_mapping,
       const fr_lookup_result_cptr_t &ls_frl,
       const repl_fr_result_cptr_t &bs_fr,
       const fnr_line_size_t line_size);

    static utilization_result_cptr_t calculate_read(
       const sample_mapping_result_cptr_t &sample_mapping,
       const locationx_mapper_result_cptr_t &lx_mapping,
       const fr_lookup_result_cptr_t &ls_frl,
       const repl_fr_result_cptr_t &bs_fr,
       const fnr_line_size_t line_size);

    static utilization_result_cptr_t calculate_write(
       const sample_mapping_result_cptr_t &sample_mapping,
       const locationx_mapper_result_cptr_t &lx_mapping,
       const fr_lookup_result_cptr_t &ls_frl,
       const repl_fr_result_cptr_t &bs_fr,
       const fnr_line_size_t line_size);

    utilization_t(const feature_context_ptr_t &context,
		  const string &sample_file,
		  const fnr_stack_depth_t stack_depth,
		  const fnr_line_size_t line_size,
		  const fnr_line_size_t page_size,
		  const thread_mapping_cptr_t &thread_mapping,
		  const cpu_model_t &cpu_model,
		  const locationx_mapper_cptr_t &lx_mapper,
		  const fnr_cache_size_t cache_size,
		  const windowing_engine_cptr_t &windowing,
		  const utilization_type_t type);
    utilization_result_cptr_t get(const bool enable_cache = true) const;

protected:
    string name() const;
    utilization_result_cptr_t calculate() const;

private:
    const string sample_file;
    const fnr_stack_depth_t stack_depth;
    const fnr_line_size_t line_size;
    const fnr_line_size_t page_size;
    const thread_mapping_cptr_t thread_mapping;
    const cpu_model_t cpu_model;
    const locationx_mapper_cptr_t lx_mapper;
    const fnr_cache_size_t cache_size;
    const windowing_engine_cptr_t windowing;
    const utilization_type_t type;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
