/** @file library/acquisition/acquisition.c */

/* Copyright (c) 2009-2010 Acumem AB.  All rights reserved. */

/*
 *  This file is part of ThreadSpotter.
 *
 *  ThreadSpotter is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ThreadSpotter is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with ThreadSpotter.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "acquisition.h"
#include "library/am_mem/am_mem.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#ifdef _WIN32
#define PATH_MAX 2048
#endif


struct acq_handle_t {
    FILE *f;
};

acq_handle_t *
acq_create(const char *filename)
{
    acq_handle_t *handle;
    FILE *f = fopen(filename, "wb");
    if (f == NULL)
	return NULL;
    handle = (acq_handle_t *) am_malloc(sizeof(acq_handle_t));
    handle->f = f;
    return handle;
}

int
acq_close(acq_handle_t *handle)
{
    int ret = fclose(handle->f);
    am_free(handle);
    return ret;
}

int
acq_fd(acq_handle_t *handle)
{
    return fileno(handle->f);
}

int
acq_write_begin(acq_handle_t *handle, const acq_begin_t *acq_begin)
{
    int i;

    fprintf(handle->f, "Version: %d\n", acq_begin->version);
    fprintf(handle->f, "Architecture: %s\n",
	    acq_begin->arch == Acq_I386 ? "i386" :
	    acq_begin->arch == Acq_X86_64 ? "x86_64" :
	    "unknown");
    fprintf(handle->f, "Begin time: %" PRId64"\n", acq_begin->start_time);
    fprintf(handle->f, "User: %s\n",
	    acq_begin->user ? acq_begin->user : "unknown");
    fprintf(handle->f, "Argc: %d\n", acq_begin->argc);
    for (i = 0; i < acq_begin->argc; i++) {
	fprintf(handle->f, "Arg: %d \"", (int) strlen(acq_begin->argv[i])+1);
	fwrite(acq_begin->argv[i], 1,
               strlen(acq_begin->argv[i]) + 1, handle->f);
	fprintf(handle->f, "\"\n");
    }
    return 1;
}


int
acq_write_end(acq_handle_t *handle, const acq_end_t *acq_end)
{
    fprintf(handle->f, "End time: %" PRId64 "\n", acq_end->end_time);
    return 1;
}

int
acq_write_burst_begin(acq_handle_t *handle, const acq_burst_begin_t *burst_begin)
{
    fprintf(handle->f, "Burst start: %d\n", burst_begin->burst_generation);
    fprintf(handle->f, "Burst start time: %" PRId64 "\n", burst_begin->burst_start_time);
    fprintf(handle->f, "Burst sampling length: %"PRId64 "\n", burst_begin->burst_sampling_length);
    fprintf(handle->f, "Burst cooldown length: %" PRId64 "\n", burst_begin->burst_cooldown_length);
    return 1;
}

int
acq_write_burst_end(acq_handle_t *handle, const acq_burst_end_t *burst_end)
{
    fprintf(handle->f, "Burst end: %d\n", burst_end->burst_generation);
    fprintf(handle->f, "Burst end time: %" PRId64 "\n", burst_end->burst_end_time);
    return 1;
}

struct string_node_t {
    struct string_node_t *next;
    char *s;
};

#ifdef _WIN32
#define snprintf _snprintf
#endif

int
acq_write_mapping(acq_handle_t *handle, const acq_mapping_t *mapping)
{
    int i;
    int len;
    int total_len = 0;
    struct string_node_t *head = 0, **iter;
    acq_mapping_entry_t *m;
    char buf[PATH_MAX + 0x100];


    iter = &head;
    for (i = 0; i < mapping->n_entries; i++) {
	*iter = (struct string_node_t *) am_malloc(sizeof (struct string_node_t));
	m = mapping->entries[i];

	len = snprintf(buf, PATH_MAX + 0x100,
		       "%" PRIx64 "-%" PRIx64" %c%c%c%c %" PRIx64
		       " %" PRIx32 ":%" PRIx32 " %" PRId64 " %s\n",
		       m->pr_vaddr,
		       m->pr_vaddr + m->pr_size,
		       m->flag_read ? 'r' : '-',
		       m->flag_write ? 'w' : '-',
		       m->flag_exec ? 'x' : '-',
		       m->flag_shared ? 's' : '-',
		       m->pr_offset,
		       (uint32_t) m->minor,
		       (uint32_t) m->major,
		       m->inode,
		       m->path);

	total_len += len;
	(*iter)->s = am_strdup(buf);
	(*iter)->next = NULL;
	iter = &((*iter)->next);
    }

    fprintf(handle->f, "Maps: %d \"", total_len);

    while (head) {
	struct string_node_t *next = head->next;
	fwrite(head->s, 1, strlen(head->s), handle->f);
	am_free(head->s);
	am_free(head);
	head = next;
    }

    fprintf(handle->f, "\"\n");
    return 1;
}
