/// @file library/atf_resampler/atf_resampler.cc

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#include "atf_resampler.hh"
#include "byte_watchpoint.hh"
#include "debug_log.hh"
#include "exceptions.hh"
#include "line_watchpoint.hh"
#include "page_watchpoint.hh"
#include "thread.hh"
#include "wp_group.hh"

#include "sampler/src/sampler_limits.h"
#include "library/random_lcg/random_lcg.h"
#include "library/am_mem/am_mem.h"
#include "library/lang/common_types.h"
#include "library/lang/tstring.hh"
#include "library/exe_path/exe_path.hh"

#include <sstream>
#include <set>
#include <boost/foreach.hpp>

#ifdef _WIN32
#include <windows.h> // MAX_PATH, DWORD
#else
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

using namespace std;
using namespace boost;
namespace fs = boost::filesystem;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

#include "sampler/src/atf_reader.c"

// This acts as a special value for a wp_group. Before the
// thread_t->waiting_for_pc_1_plus is dereferenced, its value is
// checked against this, which apparently is set if a watchpoint group
// is killed.
#define PC_1_PLUS_SAMPLE_KILLED ((acumem::library::wp_group_t *)this)

// ----------------------------------------------------------------------
// context_id -> thread_id mapping, maintaining effective sample period.

double
flow_atf_resampler_t::get_effective_sample_period(thread_t *thread)
{
    double ret = thread->effective_sample_period;

    if (atf_sample_count == 0)
        thread->effective_sample_period = (double) sample_period;
    else
        if (current_time <
            (fnr_time_t)atf_sample_count *
            sample_period)
            thread->effective_sample_period = (double) sample_period;
        else {
            thread->effective_sample_period =
                (double)current_time /
                atf_sample_count;
            if (thread->effective_sample_period > MAX_SAMPLE_PERIOD)
                thread->effective_sample_period = MAX_SAMPLE_PERIOD;
        }

    return ret;
}


// ----------------------------------------------------------------------
// thread

void
flow_atf_resampler_t::create_thread(fnr_context_id_t context_id,
                                    fnr_time_t global_time,
                                    bool new_thread)
{
    // Get map from global times to threads with this context_id.
    thread_time_map_t &thread_time_map = threads[context_id];

    // Check that no thread with the same context id has already be registered
    // at this time or later.
    assert(thread_time_map.lower_bound(global_time) == thread_time_map.end());

    // Assign the thread a new thread id if it is a new thread, otherwise find
    // the thread id it used last time.
    fnr_thread_id_t thread_id;
    if (new_thread) {
        thread_id = fnr_thread_id_t((uint32_t)all_contexts.size());
        all_contexts.push_back(context_id);
        thread_ids[context_id] = thread_id;
    }
    else {
        context_thread_map_t::const_iterator it = thread_ids.find(context_id);
        assert(it != thread_ids.end());
        thread_id = it->second;
    }

    // Create the thread and insert it into the map.
    thread_t *thread = new thread_t(context_id, thread_id);
    thread_time_map[global_time] = thread;

    // Calculate the thread's effective sample period.
    get_effective_sample_period(thread);

    // If the license only allows one thread, register the id of the first
    // thread we see so we can discard samples from all other threads.
    if (single_thread && !single_thread_seen) {
        single_thread_seen = true;
        single_thread_id = thread->thread_id;
    }
}

acumem::library::thread_t *
flow_atf_resampler_t::find_thread(fnr_context_id_t context_id,
                                  fnr_time_t global_time)
{
    thread_context_map_t::const_iterator context_it = threads.find(context_id);
    assert(context_it != threads.end());
    thread_time_map_t::const_iterator
	time_it = context_it->second.upper_bound(global_time);
    assert(time_it != context_it->second.begin());
    return (--time_it)->second;
}

// ----------------------------------------------------------------------
// temporary file handling

#ifdef _WIN32
FILE *
flow_atf_resampler_t::open_temp_file()
{
    FILE *f = 0;
    TCHAR filename_buf[MAX_PATH];
    DWORD retval = GetTempFileName(tmpdir.c_str(),
				   _T("atf_resampler_"),
				   0,
				   filename_buf);

    fs::path filename(filename_buf);
    if (retval == 0) {
	stringstream ss;
	ss << "Failed to create temporary file in directory \""
	   << tmpdir.string() << "\": error " << GetLastError()
	   << " retval: " << retval;
	throw resampler_exception_t(ss.str());
    }
    f = acumem::exe_path::fopen(filename, "w+b");
    if (!f) {
	stringstream ss;
	ss << "Failed to open temporary file: \"" << filename.string()
	   << "\": error " << GetLastError();
	fs::remove(filename);
	throw resampler_exception_t(ss.str());
    }

    // remember this file to be explicitly removed in dtor
    tempfiles.push_back(filename);

    return f;
}
#endif

#ifndef _WIN32
FILE *
flow_atf_resampler_t::open_temp_file()
{
    if (tmpdir.empty()) {
        FILE *f = tmpfile();
        if (f == NULL) {
            stringstream ss;
            ss << "Failed to create temporary file: error " << errno
	       << ", " << strerror(errno);
            throw resampler_exception_t(ss.str());
        }
        return f;
    }
    else {
	int fd;
	fs::path filename;
	for (int i=0; i<10; i++) {
	    filename = tmpdir /
		fs::unique_path("atf_resampler_%%%%%%%%%%%%%%%%");
	    fd = open(filename.c_str(),
		      O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
	    if (fd != -1)
		break;
	}
	
        if (fd == -1) {
            stringstream ss;
            ss << "Failed to create temporary file \"" << filename.string()
	       << "\": error " << errno << ", " << strerror(errno);
            throw resampler_exception_t(ss.str());
        }
        FILE *f = fdopen(fd, "w+b");
        if (f == NULL) {
            stringstream ss;
            ss << "Failed to open temporary file: error " << errno
	       << ", " << strerror(errno);
            throw resampler_exception_t(ss.str());
        }

	// remember this file to be explicitly removed in dtor
	tempfiles.push_back(filename);

        return f;
    }
}
#endif

void
flow_atf_resampler_t::open_temp_files()
{
    assert(meta_data_file == 0);
    assert(time_map_data_file == 0);

    stream_container_header_t header;

    assert(temp_fenris_file == 0);
    try {
        temp_fenris_file = open_temp_file();
        meta_data_file = open_temp_file();
        time_map_data_file = open_temp_file();
    } catch (...) {
	kill_temp_files();
	throw;
    }

    temp_file_writer = new flow_file_writer_t(temp_fenris_file, header, warning);

    if (temp_file_writer->flow_init(FNR_Feature_Temporary) != FLOW_OK) {
	kill_temp_files();
        throw resampler_exception_t("Error: failed to initialize next "
				    "flow component.");
    }
}

void
flow_atf_resampler_t::rewind_temp_files()
{
    fflush(temp_fenris_file);
    fflush(meta_data_file);
    fflush(time_map_data_file);
    fseek(temp_fenris_file, 0, SEEK_SET);
    fseek(meta_data_file, 0, SEEK_SET);
    fseek(time_map_data_file, 0, SEEK_SET);
}

void
flow_atf_resampler_t::kill_temp_files()
{
    if (meta_data_file)
	fclose(meta_data_file);
    meta_data_file = 0;

    if (time_map_data_file)
	fclose(time_map_data_file);
    time_map_data_file = 0;

    if (temp_fenris_file)
	fclose(temp_fenris_file);
    temp_fenris_file = 0;

    BOOST_FOREACH(const fs::path &file, tempfiles) {
	fs::remove(file);
    }
    tempfiles.clear();
}

// ----------------------------------------------------------------------------

void
flow_atf_resampler_t::add_time_mapping(fnr_time_t global_time,
                                       fnr_thread_id_t tid,
				       fnr_time_t local_time)
{
    int written;

    written = (int)fwrite(&global_time, sizeof(global_time), 1, time_map_data_file);

    if (written == 1)
        written = (int)fwrite(&tid, sizeof(tid), 1, time_map_data_file);

    if (written == 1)
        written = (int)fwrite(&local_time, sizeof(local_time), 1, time_map_data_file);

    if (written != 1) {
        stringstream ss;
        ss << "Write to temporary time map file failed; error " << errno << ", " << strerror(errno);
        const char *s = am_strdup(ss.str().c_str());
        throw resampler_exception_t(s);
    }
}

// ----------------------------------------------------------------------
// prepare and store wp_groups to a temporary file

static bool
cmp_access_by_time(const fnr_sample_access_t &a,
		   const fnr_sample_access_t &b)
{
    return a.time < b.time;
}

void
flow_atf_resampler_t::remove_duplicate_end_events(watchpoint_t *wp)
{
    fnr_sample_access_vector_t *end = &wp->event->end;

    sort(end->begin(), end->end(), cmp_access_by_time);

    fnr_sample_access_vector_t new_end;
    BOOST_FOREACH(const fnr_sample_access_t &a, *end) {
        if (new_end.empty() ||
            a.time != new_end.back().time)
            new_end.push_back(a);
    }

    end->swap(new_end);
}

fnr_stack_id_t
flow_atf_resampler_t::stack_id(stack_t *stack)
{
    /* Return 0 if we have no call stack information. */
    if (stack == NULL)
        return 0;

    /* Assign an id to the stack and enqueue a fenris event for it if
       it does not yet have an id. */
    if (stack->id == 0) {
        stack->id = next_stack_id++;

        fnr_event_t *event =
	    new fnr_event_stack_t(stack->id, stack->ret_addrs);
        undispatched_events.push_back(event);
    }

    /* Return the id. */
    return stack->id;
}

fnr_event_sample_t *
flow_atf_resampler_t::prepare_stack_sample_event(watchpoint_t *wp)
{
    // Here we update what we can, but notably we don't update the
    // time fields. They still hold the sampler times. The reason is
    // that the global times have not been sorted yet. We will
    // translate this when the temporary file is read back in again.

    fnr_event_sample_t *event(wp->event);
    event->begin.stack = stack_id(wp->pc_1_stack);
    event->pc_1_plus_stack = stack_id(wp->pc_1_plus_stack);

    remove_duplicate_end_events(wp);
    BOOST_FOREACH(fnr_sample_access_t &a, event->end) {
	time_stack_map_t::iterator s(wp->end_stacks.find(a.time));
	assert(s != wp->end_stacks.end());
	a.stack = stack_id(s->second);
    }
    return event;
}

void
flow_atf_resampler_t::write_wp_group_metadata(const wp_group_t *wp_group,
					      int32_t count)
{
    assert(meta_data_file != 0);

    size_t written;
    written = fwrite(&count, sizeof(count), 1, meta_data_file);

    if (written != 1)
	throw resampler_exception_t("Couldn't write to temporary file");

    assert(sizeof(wp_group->effective_sample_period) == 8);
    written = fwrite(&wp_group->effective_sample_period,
		     sizeof(wp_group->effective_sample_period),
		     1, meta_data_file);

    if (written != 1)
	throw resampler_exception_t("Couldn't write to temporary file");
}

static bool
another_thread(const fnr_event_sample_t *sample,
	       const fnr_thread_id_t single_thread_id)
{
    if (sample->begin.thread != single_thread_id)
        return true;

    BOOST_FOREACH(const fnr_sample_access_t &a, sample->end)
        if (a.thread != single_thread_id)
            return true;

    return false;
}

bool
flow_atf_resampler_t::single_thread_skip_sample(wp_group_t *wp_group) const
{
    if (!single_thread)
	return false;

    // Ignore the sample if only a single thread is allowed and this sample
    // includes an access from another thread.
    assert(single_thread_seen);

    bool skip_sample = false;
    BOOST_FOREACH(watchpoint_t *wp, wp_group->byte_wps)
	if (another_thread(wp->event, single_thread_id))
	    skip_sample = true;
    BOOST_FOREACH(watchpoint_vector_map_t::value_type &v,
		  wp_group->line_wps) {
	BOOST_FOREACH(watchpoint_t *wp, v.second)
	    if (another_thread(wp->event, single_thread_id))
		skip_sample = true;
    }
    BOOST_FOREACH(watchpoint_t *wp, wp_group->page_wps)
	if (another_thread(wp->event, single_thread_id))
	    skip_sample = true;

    return skip_sample;
}

void
flow_atf_resampler_t::write_group_to_temp_file(wp_group_t *wp_group)
{
    assert(wp_group != 0);
    assert(temp_file_writer != 0);

    // Check if we should discard this watchpoint group due to
    // accesses from forbidden threads.
    if (single_thread_skip_sample(wp_group))
	return;

    int32_t count = 0;
    BOOST_FOREACH(watchpoint_t *wp, wp_group->byte_wps) {
	temp_file_writer->process_event(*prepare_stack_sample_event(wp));
	count++;
    }

    BOOST_FOREACH(watchpoint_vector_map_t::value_type &v,
		  wp_group->line_wps) {
	BOOST_FOREACH(watchpoint_t *wp, v.second) {
	    temp_file_writer->process_event(*prepare_stack_sample_event(wp));
	    count++;
	}
    }
    BOOST_FOREACH(watchpoint_t *wp, wp_group->page_wps) {
	temp_file_writer->process_event(*prepare_stack_sample_event(wp));
	count++;
    }

    // Now write metadata, the number of samples, and the effective
    // sample period.
    write_wp_group_metadata(wp_group, count);
}


// ----------------------------------------------------------------------
// read and decimate wp_groups from temporary file

int32_t
flow_atf_resampler_t::read_wp_group_metadata(double *effective_sample_period)
{
    assert(meta_data_file != 0);

    size_t read;
    int32_t count;

    read = fread(&count, sizeof(count), 1, meta_data_file);
    if (read != 1)
	throw resampler_exception_t("Couldn't read from temporary file");

    read = fread(effective_sample_period, sizeof(double), 1, meta_data_file);
    if (read != 1)
	throw resampler_exception_t("Couldn't read from temporary file");

    return count;
}

class flow_phase2_t : public fenris_flow_by_type_default_t {
public:
    flow_phase2_t(flow_atf_resampler_t *parent)
	: parent(parent),
	  number_samples(0)
	{}

    virtual flow_status_t sample(const fnr_event_sample_t &event) {
	if (number_samples == 0) {
	    double effective_sample_period;
	    number_samples = parent->read_wp_group_metadata(&effective_sample_period);
	    keep = parent->decide_keep_event(effective_sample_period);

	}

	number_samples--;

	if (keep)
	    parent->keep_event(event);

	return FLOW_OK;
    }

private:
    flow_atf_resampler_t *parent;
    int number_samples;
    bool keep;
};

bool
flow_atf_resampler_t::decide_keep_event(double effective_sample_period)
{
    bool keep = true;

    // Don't decimate if we don't have a target count.
    if (target_sample_count > 0) {
	double r = (double) random_lcg_next(random_lcg.get()) / RANDOM_LCG_MAX;
	keep = r * current_time / target_sample_count < effective_sample_period;
    }

    if (keep)
	sample_count++;

    feedback->test_cancel();

    return keep;
}

void
flow_atf_resampler_t::mark_global_time(fnr_time_t global_time)
{
    seen_global_times.insert(global_time);
}

void
flow_atf_resampler_t::keep_event(const fnr_event_sample_t &event)
{
    fnr_event_sample_t *sample_event = new fnr_event_sample_t(event);

    mark_global_time(sample_event->begin.time);

    BOOST_FOREACH(fnr_sample_access_t &a, sample_event->end)
	mark_global_time(a.time);

    half_baked_samples.push_back(sample_event);
}

void
flow_atf_resampler_t::read_back_time_mapping()
{
    while (!feof(time_map_data_file)) {
        fnr_time_t sampler_time;
        fnr_thread_id_t tid;
        fnr_time_t local_time;

	fread(&sampler_time, sizeof(sampler_time), 1, time_map_data_file);
	fread(&tid, sizeof(tid), 1, time_map_data_file);
	fread(&local_time, sizeof(local_time), 1, time_map_data_file);

        if (seen_global_times.find(sampler_time) == seen_global_times.end())
            continue;

        global_tid_local_map_t::iterator stlm_i =
            global_tid_local_map.find(sampler_time);
        if (stlm_i == global_tid_local_map.end())
            stlm_i = global_tid_local_map.insert(
                make_pair(sampler_time, tid_local_map_t())).first;

        tid_local_map_t::iterator tlm_i = stlm_i->second.find(tid);
        if (tlm_i == stlm_i->second.end())
            stlm_i->second.insert(make_pair(tid, local_time));
        else {
            assert(local_time >= tlm_i->second);
            tlm_i->second = local_time;
        }
    }
}

void
flow_atf_resampler_t::enqueue_samples()
{
    BOOST_FOREACH(fnr_event_t *e, half_baked_samples)
	undispatched_events.push_back(e);

    half_baked_samples.clear();
}

void
flow_atf_resampler_t::enqueue_time_mappings()
{
    // Enqueue all the global time to thread local time events.
    BOOST_FOREACH(const global_tid_local_map_t::value_type &stlm_v,
                  global_tid_local_map) {
        fnr_time_t global_time = stlm_v.first;

        BOOST_FOREACH(const tid_local_map_t::value_type &tlm_v, stlm_v.second) {
            fnr_thread_id_t tid = tlm_v.first;
            fnr_time_t local_time = tlm_v.second;
            undispatched_events.push_back(
                new fnr_event_local_time_t(tid, global_time, local_time));
        }
    }

    // Clear all the time related structures so nothing spills over to the next
    // burst.
    seen_global_times.clear();
    global_tid_local_map.clear();
}

void
flow_atf_resampler_t::process_samples()
{
    // Dump all remaining unfinished wp_groups to the temp file, and
    // clear the wp_groups map.
    BOOST_FOREACH(wp_group_map_t::value_type &v, wp_groups) {
	wp_group_t *wp_group = v.second;
	write_group_to_temp_file(wp_group);
	delete wp_group;

	feedback->test_cancel();
    }
    wp_groups.clear();

    // close the temp file writer
    delete temp_file_writer;
    temp_file_writer = 0;

    // Create a fenris reader around the temporary files. It will use
    // flow_phase2_t to decimate the events, and call us back for each
    // events that should be kept.
    flow_phase2_t phase2(this);

    // flush and rewind the temporary files without closing them in
    // preparation for reading them again.
    rewind_temp_files();

    // Set up temp_file_reader.
    flow_file_reader_t temp_file_reader(temp_fenris_file, &phase2, warning);

    // Read back and decimate the samples.
    temp_file_reader.dispatch();

    // Read back and decimate the time mappings.
    read_back_time_mapping();

    // Enqueue the sample events.
    enqueue_samples();

    // Enqueue the time mapping events.
    enqueue_time_mappings();

    // Done with these files - close and delete.
    kill_temp_files();
}

// ----------------------------------------------------------------------
// atf decoding primitives

fnr_access_type_t
flow_atf_resampler_t::get_access_type(struct atf_packet *packet)
{
    switch (packet->header & 0x3) {
    case ATF_OP_MEM_RO:
        return FNR_Access_Type_RD;
    case ATF_OP_MEM_WO:
        return FNR_Access_Type_WR;
    case ATF_OP_MEM_RW:
        return FNR_Access_Type_RDWR;
    case ATF_OP_EXT:
        switch ((packet->header >> 2) & 0x7) {
        case ATF_XOP_MEM_PREFETCH:
            switch ((packet->header >> 5) & 0x7) {
            case ATF_PREFETCH_AMD:
                return FNR_Access_Type_PF_AMD;
            case ATF_PREFETCH_AMD_W:
                return FNR_Access_Type_PF_AMD_W;
            case ATF_PREFETCH_INTEL_NTA:
                return FNR_Access_Type_PF_Intel_NTA;
            case ATF_PREFETCH_INTEL_T0:
                return FNR_Access_Type_PF_Intel_T0;
            case ATF_PREFETCH_INTEL_T1:
                return FNR_Access_Type_PF_Intel_T1;
            case ATF_PREFETCH_INTEL_T2:
                return FNR_Access_Type_PF_Intel_T2;
            default:
                assert(0);
                while (1); // Avoid warning on Solaris.
            }
        case ATF_XOP_MEM_MOVNT:
            switch ((packet->header >> 5) & 0x1) {
            case 1:
                return FNR_Access_Type_NTA_RD;
            case 0:
                return FNR_Access_Type_NTA_WR;
            default:
                assert(0);
                while (1); // Avoid warning on Solaris.
            }
        case ATF_XOP_MEM_CLFLUSH:
            return FNR_Access_Type_CLFLUSH;
        default:
            assert(0);
            while (1); // Avoid warning on Solaris.
        }
    default:
        assert(0);
        while (1); // Avoid warning on Solaris.
    }
}

fnr_access_size_t
flow_atf_resampler_t::get_access_size(struct atf_packet *packet)
{
    switch (packet->header & 0x1f) {
    case ATF_OP_MEM_RO:
    case ATF_OP_MEM_RW:
    case ATF_OP_MEM_WO:
    case ATF_OP_EXT | (ATF_XOP_MEM_MOVNT << 2):
    case ATF_OP_EXT | (ATF_XOP_MEM_PREFETCH << 2):
    case ATF_OP_EXT | (ATF_XOP_MEM_CLFLUSH << 2):
        return packet->header >> 8;
    default:
        assert(0);
	return 0; // to keep the compiler happy
    }
}

fnr_context_id_t
flow_atf_resampler_t::get_context_id(struct atf_packet *packet)
{
    if ((packet->header & 0x1F) == (ATF_OP_EXT | (ATF_XOP_SHADOW << 2)))
        return packet->u.shadow.tid;
    else if (packet->header == (ATF_OP_EXT | (ATF_XOP_EXT << 2) |
				(ATF_XXOP_THREAD << 5)))
        return packet->u.thread.tid;
    else if (packet->header == (ATF_OP_EXT | (ATF_XOP_EXT << 2) |
				(ATF_XXOP_LOCAL_MACNT << 5)))
        return packet->u.local_macnt.tid;
    else if (packet->header == (ATF_OP_EXT | (ATF_XOP_EXT << 2) |
				(ATF_XXOP_THREAD_EXIT << 5)))
        return packet->u.thread_exit.tid;
    else
        return packet->u.mem.tid;
}

// ----------------------------------------------------------------------
// memory access state machine


void
flow_atf_resampler_t::handle_pc_1(thread_t *thread,
				  const fnr_sample_access_t *access)
{
    DEBUG_LOG_MEM(access, thread->local_time, "sample start");

    thread->last_local_begin_time = thread->local_time;

    fnr_access_type_t access_type = access->access_type;
    access_type_set_t access_type_set = get_access_type_set(access_type);

    watchpoint_t *page_wp = NULL;
    line_watchpoint_map_t line_wps;
    watchpoint_t *byte_wp = NULL;

    uint64_t start_addr = access->addr.virtual_address;
    uint64_t end_addr = start_addr + access->access_size;

    // Sweep over the bytes covered by the access and create watchpoints along
    // the way. This code must match what the sampler does.
    wp_group_t *wp_group = NULL;
    for (uint64_t addr = start_addr; addr != end_addr; addr++) {
        if (page_wp == NULL || (addr & (PAGE_SAMPLE_SIZE - 1)) == 0) {
            wp_group = new wp_group_t(this, thread, access);
            wp_groups[wp_group->id] = wp_group;
            page_wp = new page_watchpoint_t(this, thread, addr,
					    wp_group, access);
            wp_group->page_wps.push_back(page_wp);
        }

        BOOST_FOREACH(watchpoint_type_t *wp_type, line_watchpoint_types) {
            fnr_line_size_t line_size =
		(fnr_line_size_t)1 << wp_type->wp_size_log2;

            if (line_wps[line_size] == NULL || (addr & (line_size - 1)) == 0) {
                line_watchpoint_t *line_wp =
		    new line_watchpoint_t(this, thread, addr,
					  wp_type->wp_size_log2,
					  wp_group, access);
                line_wps[line_size] = line_wp;
                wp_group->line_wps[line_size].push_back(line_wp);
            }
        }

        if (!(access_type_set & ATS_CLASS_NO_DATA)) {
            byte_wp =
		new byte_watchpoint_t(this, thread, addr, wp_group,
				      &line_wps, access);
            wp_group->byte_wps.push_back(byte_wp);
        }
    }

    /* Insert the watchpoint into the waiting pc1+ list. */
    assert(thread->waiting_for_pc_1_plus == NULL);
    thread->waiting_for_pc_1_plus = wp_group;
}


bool
flow_atf_resampler_t::handle_pc_1_plus(thread_t *thread,
				       const fnr_sample_access_t *access)
{
    // Return false if not waiting for a pc1+ access.
    if (thread->waiting_for_pc_1_plus == NULL)
        return false;

    // The pc1+ access should always be the immediately following access. This
    // assertion should help detect if the resampler has gotten out of sync
    // with the sampler.
    assert(thread->last_local_begin_time + 1 == thread->local_time);

    if (thread->waiting_for_pc_1_plus != PC_1_PLUS_SAMPLE_KILLED) {
        thread->waiting_for_pc_1_plus->pc_1_plus(thread, access);

	if (thread->waiting_for_pc_1_plus->is_complete_lacking_pc_1_plus()) {
	    write_group_to_temp_file(thread->waiting_for_pc_1_plus);

	    wp_groups.erase(thread->waiting_for_pc_1_plus->id);
	    delete thread->waiting_for_pc_1_plus;
	}
    }

    thread->waiting_for_pc_1_plus = NULL;

    DEBUG_LOG_MEM(access, thread->local_time, "pc1+");

    return true;
}

bool
flow_atf_resampler_t::apply_access(watchpoint_type_t *wp_type,
				   watchpoint_vector_t *wps,
				   thread_t *thread,
				   const fnr_sample_access_t *access,
				   bool must_be_last)
{
    bool mid_end = false;

    uint64_t addr = access->addr.virtual_address;
    uint64_t size = access->access_size;

    fnr_line_count_t start_index = addr >> wp_type->wp_size_log2;
    fnr_line_count_t end_index =
	(addr + size + wp_type->wp_size - 1) >> wp_type->wp_size_log2;
    fnr_line_count_t index_mask =
	(fnr_line_count_t)-1 >> wp_type->wp_size_log2;

    // XXX Don't see the point for the mod operator in the loop
    // increment, now that we handle segment registers in the sampler.

    for (fnr_line_count_t index = start_index;
	 index != end_index;
	 index = ((index + 1) & index_mask)) {
	BOOST_FOREACH(watchpoint_t *wp, *wps) {
            if (wp->line_index == index) {
                switch (wp->apply_access(this, thread, access,
					 wp_type->wp_size_log2,
					 must_be_last)) {
                case 0: // No match.
                    break;
                case 1: // Mid access.
                    mid_end = true;
#if defined(DEBUG_LOG)
                    debug_log_wps.insert(
			debug_log_wp_t(wp->id, "mid",
				       index << wp_type->wp_size_log2,
				       wp_type->wp_size));
#endif
                    break;
                case 2: // Sample terminated.
                    mid_end = true;
#if defined(DEBUG_LOG)
                    debug_log_wps.insert(
			debug_log_wp_t(wp->id, "end",
				       index << wp_type->wp_size_log2,
				       wp_type->wp_size));
#endif
                    break;
                default:
                    assert(0);
                }
            }
        }
    }

    return mid_end;
}

void
flow_atf_resampler_t::handle_pc_2(thread_t *thread,
				  fnr_sample_count_t wp_group_id,
				  const fnr_sample_access_t *access,
				  bool must_be_last)
{
    bool mid_end = false;

#if defined(DEBUG_LOG)
    debug_log_wps.clear();
#endif

    wp_group_map_t::iterator wpgi = wp_groups.find(wp_group_id);
    assert(wpgi != wp_groups.end());
    wp_group_t *wp_group = wpgi->second;


    // Maintain a record of the last access seen for this watchpoint group.
    //
    // The flag 'must_be_last' may already be set from handle_mem, due
    // to its arriving out of order, globally speaking.
    //
    // XXX I don't know if it ever happens that this assignment of
    // 'must_be_last' ever is significant.
    if (access->time <= wp_group->last_sampler_time)
        must_be_last = true;
    else
        wp_group->last_sampler_time = access->time;

    // Let the byte watchpoint see this access. This in turn may cause
    // the line watchpoints to see the access as well if it is
    // relevant for the byte perspective. If it was one of the
    // interesting accesses, set the mid_end flag.
    if (apply_access(byte_watchpoint_type, &wp_group->byte_wps,
		     thread, access, must_be_last))
        mid_end = true;


    // Same here but for each line watchpoint
    BOOST_FOREACH(watchpoint_type_t *wpt, line_watchpoint_types)
        if (apply_access(wpt, &wp_group->line_wps[wpt->wp_size],
			 thread, access, must_be_last))
            mid_end = true;

    // And page
    if (apply_access(page_watchpoint_type, &wp_group->page_wps,
		     thread, access, must_be_last))
        mid_end = true;

#if defined(DEBUG_LOG)
    BOOST_FOREACH(const debug_log_wp_t &wp, debug_log_wps)
        DEBUG_LOG_WP(access->thread, wp_group_id, wp.id, wp.event_type,
		     wp.start, wp.length);

    DEBUG_LOG_MEM(access, thread->local_time,
		  mid_end ? "sample mid/end" : "sample last");
#endif

    // Check whether we are done with this watchpoint group, and if we
    // can write it to the temp file.
    if (mid_end && !wp_group->has_live_wps()) {
	if (wp_group->has_pc_1_plus) {
	    // complete groups are written to the temp file
	    write_group_to_temp_file(wp_group);
	    // and then we forget about it
	    delete wp_group;
	    wp_groups.erase(wpgi);
	} else {
	    // if it lacks pc_1_plus at this point, defer
	    // the temp file writing to the pc_1_plus event
	    wp_group->mark_complete_lacking_pc_1_plus();
	}
    }
}

// ----------------------------------------------------------------------
// Handle ATF packet

void
flow_atf_resampler_t::handle_mem(struct atf_packet *packet)
{
    fnr_time_t sampler_macnt = packet->u.mem.global_macnt;

    // Get the thread that the packet refers to.
    thread_t *thread = find_thread(get_context_id(packet), sampler_macnt);

    // Keep track of greatest seen local time.
    if (packet->u.mem.local_macnt > thread->greatest_local_time)
        thread->greatest_local_time = packet->u.mem.local_macnt;

    // Add time mapping.
    add_time_mapping(sampler_macnt, thread->thread_id,
                     packet->u.mem.local_macnt);

    // Check for the start of new bursts.
    if (burst_sampled && burst_generation == 0) {
        // Bump the burst generation counter.
        burst_generation = last_burst_generation + 1;

        // Enqueue a burst start event.
        burst_info_map_t::const_iterator
	    bii = burst_info.find(burst_generation);
        fnr_real_time_t start_time = 0;
        fnr_time_t sampling_time = 0;
        fnr_time_t cooldown_time = 0;
        if (bii != burst_info.end()) {
            const burst_info_t *bi = &bii->second;
            start_time = bi->start_time;
            sampling_time = bi->sampling_time;
            cooldown_time = bi->cooldown_time;
        }
        fnr_event_t *burst_begin_event = new fnr_event_burst_begin_t(
            start_time, 0, sampling_time, cooldown_time);
        undispatched_events.push_back(burst_begin_event);
    }

    // Events will be reported out-of-order from the sampler: After
    // the terminating end access, the last access from each thread is
    // also reported, and they must be treated in a special way during
    // the resampling. This is signified by the flag 'must_be_last'
    // which is propagated downwards.

    bool must_be_last = sampler_macnt < current_time;


    if (sampler_macnt > current_time)
        current_time = sampler_macnt;

    /* Update the thread's local time. */
    thread->local_time = packet->u.mem.local_macnt;

    // Create matching access struct.
    fnr_sample_access_t access(
        fnr_address_t(packet->u.mem.iaddr), // PC
        fnr_address_t(packet->u.mem.daddr), // Address
        sampler_macnt,                      // Time
        0,                                  // Stack id, calculated at dispatch
        thread->thread_id,                  // Thread
        get_access_size(packet),            // Access size
        get_access_type(packet));           // Access type

    if (packet->u.mem.group_id != 0)
        handle_pc_2(thread, packet->u.mem.group_id, &access, must_be_last);
    else
        if (!handle_pc_1_plus(thread, &access))
            handle_pc_1(thread, &access);

    thread->stack = thread->last_stack;
}

/* Returns true if the packet ends a burst. */
void
flow_atf_resampler_t::handle_kill(fnr_sample_count_t wp_group_id)
{
    wp_group_map_t::iterator it_wpg(wp_groups.find(wp_group_id));
    assert(it_wpg != wp_groups.end());
    wp_group_t *wp_group(it_wpg->second);

#if defined(DEBUG_LOG)
    list<unsigned int> byte_kills;
    BOOST_FOREACH(watchpoint_t *wp, wp_group->byte_wps)
        if (!wp->terminated)
            byte_kills.push_front(wp->id);
    BOOST_FOREACH(unsigned int wp_id, byte_kills)
        DEBUG_LOG_KILL(wp_group_id, wp_id);

    list<unsigned int> line_kills;
    BOOST_FOREACH(watchpoint_type_t *wpt, line_watchpoint_types) {
        BOOST_FOREACH(watchpoint_t *wp, wp_group->line_wps[wpt->wp_size])
            if (!wp->terminated)
                line_kills.push_front(wp->id);
    }
    BOOST_FOREACH(unsigned int wp_id, line_kills)
        DEBUG_LOG_KILL(wp_group_id, wp_id);

    list<unsigned int> page_kills;
    BOOST_FOREACH(watchpoint_t *wp, wp_group->page_wps)
        if (!wp->terminated)
            page_kills.push_front(wp->id);
    BOOST_FOREACH(unsigned int wp_id, page_kills)
        DEBUG_LOG_KILL(wp_group_id, wp_id);
#endif

    wp_group->set_sample_flag(FNR_Sample_Flag_Killed);

    // complete groups are written to the temp file
    write_group_to_temp_file(wp_group);
    // and then we forget about it
    delete wp_group;
    wp_groups.erase(it_wpg);

    BOOST_FOREACH(thread_context_map_t::value_type &v, threads) {
        BOOST_FOREACH(thread_time_map_t::value_type &t, v.second)
            if (t.second->waiting_for_pc_1_plus == wp_group)
                t.second->waiting_for_pc_1_plus = PC_1_PLUS_SAMPLE_KILLED;
    }
}

void
flow_atf_resampler_t::handle_stack(struct atf_packet *packet)
{
    fnr_time_t sampler_macnt = packet->u.shadow.global_macnt;

    thread_t *thread = find_thread(get_context_id(packet), sampler_macnt);
    int nrlost = packet->u.shadow.nrlost;
    int nrpop = packet->u.shadow.nrpop;
    int nrpush = packet->u.shadow.nrpush;

    stack_t *stack;
    uint64_t hash = 0;

    // Check if this is an absolute call stack.
    if (packet->u.shadow.absolute) {
        /* Construct the new stack and calculate its hash value. */
        stack = new stack_t(nrpush);
        for (int i = 0; i < nrpush; i++) {
            fnr_address_t pc(packet->u.shadow.pushes[i]);
            hash = stack->push(pc, hash);
        }
    }
    else {
        int old_depth = thread->last_stack == NULL ?
	    0 : (int)thread->last_stack->ret_addrs.size();
        assert(nrpop <= old_depth);

        /* The resampler reports frames as both lost and popped when it
           gets a stack overflow. Compensate for this by decresing nrpop
           by nrlost. */
        assert(nrpop >= nrlost || nrpop == old_depth);
        if (nrlost <= nrpop)
            nrpop -= nrlost;
        else
            nrpop = 0;

        int lost = nrlost > old_depth ? old_depth : nrlost;
        assert(lost + nrpop <= old_depth);
        int depth = old_depth - lost - nrpop + nrpush;

        /// @todo Always allocating a stack and freeing it if we find it in the
        /// hash table makes the code simpler, but seems wasteful.

        /* Construct the new stack and calculate its hash value. */
        stack = new stack_t(depth);
        for (int i = lost; i < old_depth - nrpop; i++)
            hash = stack->push(thread->last_stack->ret_addrs[i], hash);
        for (int i = 0; i < nrpush; i++) {
            fnr_address_t pc(packet->u.shadow.pushes[i]);
            hash = stack->push(pc, hash);
        }
    }

    /* Check if the stack is already in the hash table. In that case,
       use that stack and free our copy. */
    stack_t **hash_bucket = &stack_hash[hash & STACK_HASH_MASK];
    stack_t *hash_stack = *hash_bucket;
    while (hash_stack != NULL)
        if (*hash_stack == *stack) {
            delete stack;
            thread->stack = hash_stack;
            goto done;
        }
        else
            hash_stack = hash_stack->next;

    /* The stack was not found in the hash table. Insert our copy into
       the hash table and use it. */
    stack->next = *hash_bucket;
    *hash_bucket = stack;
    thread->stack = stack;

done:
    if (!packet->u.shadow.absolute)
        thread->last_stack = thread->stack;
}

void
flow_atf_resampler_t::handle_burst_end(struct atf_packet *packet)
{
    assert(burst_generation != 0);

    fnr_time_t sampler_macnt = packet->u.burst_end.global_macnt;

    // Enqueue all events from the current burst.
    process_samples();

    // Enqueue a burst end event.
    burst_info_map_t::const_iterator bii = burst_info.find(burst_generation);
    fnr_real_time_t end_time = 0;
    if (bii != burst_info.end()) {
	const burst_info_t *bi = &bii->second;
	end_time = bi->end_time;
    }
    fnr_event_t *burst_end_event = new fnr_event_burst_end_t(
        end_time, sampler_macnt);
    undispatched_events.push_back(burst_end_event);

    // Delete all thread objects, we will create new ones for the next burst.
    BOOST_FOREACH(thread_context_map_t::value_type &v, threads) {
        BOOST_FOREACH(thread_time_map_t::value_type &t, v.second)
            delete t.second;
    }
    threads.clear();

    // Reset the current time.
    current_time = 0;

    // Reset the burst generation counter.
    last_burst_generation = burst_generation;
    burst_generation = 0;

    // Prepare new temporary files.
    open_temp_files();
}

void
flow_atf_resampler_t::handle_thread(struct atf_packet *packet)
{
    fnr_time_t sampler_macnt = packet->u.thread.global_macnt;

    // Create the thread that the packet refers to.
    create_thread(get_context_id(packet), sampler_macnt,
                  !!(packet->u.thread.flags & ATF_THREAD_NEW));
}

void
flow_atf_resampler_t::handle_mmap(struct atf_packet *packet)
{
    fnr_time_t sampler_macnt = packet->u.mmap.global_macnt;

    // Convert mmap parameters to our internal format.
    fnr_mmap_attr_t attr =
	((packet->u.mmap.prot & ATF_MMAP_PROT_EXEC) ? FNR_Mmap_Attr_Execute : 0) |
	((packet->u.mmap.prot & ATF_MMAP_PROT_READ) ? FNR_Mmap_Attr_Read : 0) |
	((packet->u.mmap.prot & ATF_MMAP_PROT_WRITE) ? FNR_Mmap_Attr_Write : 0) |
	((packet->u.mmap.flags & ATF_MMAP_MAP_SHARED) ? FNR_Mmap_Attr_Shared : 0);

    char path[sizeof(packet->u.mmap.path) + 1];
    memcpy(path, packet->u.mmap.path, packet->u.mmap.path_length);
    path[packet->u.mmap.path_length] = '\0';

    fnr_event_t *mmap_event = new fnr_event_mmap_t(
	packet->u.mmap.start,
	packet->u.mmap.length,
	attr,
	packet->u.mmap.device,
	packet->u.mmap.inode,
	packet->u.mmap.offset,
	path,
	sampler_macnt);

    undispatched_events.push_back(mmap_event);
}

void
flow_atf_resampler_t::handle_munmap(struct atf_packet *packet)
{
    fnr_time_t sampler_macnt = packet->u.munmap.global_macnt;

    fnr_event_t *munmap_event = new fnr_event_munmap_t(
	packet->u.munmap.start,
	packet->u.munmap.length,
	sampler_macnt);

    undispatched_events.push_back(munmap_event);
}

void
flow_atf_resampler_t::handle_global_macnt(struct atf_packet *packet)
{
    fnr_time_t sampler_macnt = packet->u.global_macnt.global_macnt;

    // Update current sampler time.
    assert(sampler_macnt >= current_time);
    current_time = sampler_macnt;
}

void
flow_atf_resampler_t::handle_local_macnt(struct atf_packet *packet)
{
    fnr_context_id_t tid = get_context_id(packet);

    thread_t *thread = find_thread(tid, current_time);

    // Keep track of greatest seen local time.
    if (packet->u.local_macnt.local_macnt > thread->greatest_local_time)
        thread->greatest_local_time = packet->u.local_macnt.local_macnt;

    add_time_mapping(current_time, thread->thread_id, packet->u.local_macnt.local_macnt);
}

void
flow_atf_resampler_t::handle_thread_exit(struct atf_packet *packet)
{
    fnr_context_id_t tid = get_context_id(packet);

    thread_t *thread = find_thread(tid, current_time);

    add_time_mapping(packet->u.thread_exit.global_macnt,
                     thread->thread_id,
                     packet->u.thread_exit.local_macnt);
    mark_global_time(packet->u.thread_exit.global_macnt);
}

void
flow_atf_resampler_t::handle_packet(struct atf_packet *packet)
{
    switch (packet->header & 0x03) {
    case ATF_OP_MEM_RO:
    case ATF_OP_MEM_WO:
    case ATF_OP_MEM_RW:
	handle_mem(packet);
	break;

    case ATF_OP_EXT:
	switch ((packet->header >> 2) & 0x07) {
	case ATF_XOP_MEM_CLFLUSH:
	case ATF_XOP_MEM_PREFETCH:
	case ATF_XOP_MEM_MOVNT:
	    handle_mem(packet);
	    break;

	case ATF_XOP_SHADOW:
	    handle_stack(packet);
	    break;

	case ATF_XOP_MMAP:
	    handle_mmap(packet);
	    break;

	case ATF_XOP_MUNMAP:
	    handle_munmap(packet);
	    break;

	case ATF_XOP_MREMAP:
	    /* XXX: these are not yet handled, ignore them */
	    break;

	case ATF_XOP_EXT:
	    switch ((packet->header >> 5) & 0x07) {
	    case ATF_XXOP_KILLWP:
		handle_kill(packet->u.killwp.group_id);
		break;
	    case ATF_XXOP_BURST_END:
		handle_burst_end(packet);
		break;
            case ATF_XXOP_THREAD:
                handle_thread(packet);
                break;
            case ATF_XXOP_GLOBAL_MACNT:
                handle_global_macnt(packet);
                break;
            case ATF_XXOP_LOCAL_MACNT:
                handle_local_macnt(packet);
                break;
            case ATF_XXOP_THREAD_EXIT:
                handle_thread_exit(packet);
                break;
	    default:
		assert(0);
		break;
	    }
	}
	break;
    }
}

// ----------------------------------------------------------------------
// acquisition info file parser

int
flow_atf_resampler_t::get_num_line_sizes(
    const unsigned char *sampler_log2linesizes)
{
    int i;
    for (i = 0; i < 8; i++)
	if (sampler_log2linesizes[i] == 0xFF)
	    return i;

    return 8;
}

#define EXPECT(var, val)						\
    if ((var) != (val))							\
	throw resampler_exception_t("Syntax error in acquisition info file."  #var " != " #val );

void
flow_atf_resampler_t::read_acquisition_info(const struct atf_header *header)
{
    int res;
    int version;
    char arch_str[16];
    fnr_architecture_t arch;
    fnr_real_time_t begin_time, end_time;
    char user_str[1024];
    int argc;
    fnr_argv_t argv;
    int num_line_sizes = get_num_line_sizes(header->sampler_log2linesizes);
    fnr_line_size_log2_vector_t line_sizes;

    if (acq_file == NULL)
	return;

    EXPECT(fscanf(acq_file, "Version: %d\n", &version), 1);
    if (version != 1) {
	ostringstream oss;
	oss << "Error: incompatible acquisition info version: "
	    << version;
	throw resampler_exception_t(oss.str());
    }

    EXPECT(fscanf(acq_file, "Architecture: %15s\n", arch_str), 1);
    if (strcmp("i386", arch_str) == 0)
	arch = FNR_Arch_i386;
    else if (strcmp("x86_64", arch_str) == 0)
	arch = FNR_Arch_x86_64;
    else {
	ostringstream oss;
	oss << "Error: unsupported architecture: " << arch_str;
	throw resampler_exception_t(oss.str());
    }

    EXPECT(fscanf(acq_file, "Begin time: %" SCNu64 "\n", &begin_time), 1);
    EXPECT(fscanf(acq_file, "User: "), 0);
    fgets(user_str, 1023, acq_file);
    if (user_str[strlen(user_str)-1] == '\n')
	user_str[strlen(user_str)-1] = 0;
    EXPECT(fscanf(acq_file, "Argc: %d\n", &argc), 1);

    argv.reserve(argc);
    for (int i = 0; i < argc; i++) {
	int arg_len;
	EXPECT(fscanf(acq_file, "Arg: %d \"", &arg_len), 1);

	scoped_array<char> arg(new char[arg_len]);
	EXPECT(fread(arg.get(), 1, arg_len, acq_file), (unsigned int)arg_len);
	fscanf(acq_file, "\"\n");

	argv.push_back(string(arg.get()));
    }

    line_sizes.reserve(num_line_sizes);
    for (int i = 0; i < num_line_sizes; i++)
	line_sizes.push_back(header->sampler_log2linesizes[i]);

    begin_events.push_back(
	new fnr_event_acquisition_begin_t(
	    begin_time, arch, argv, user_str,
	    line_sizes));

    while (1) {
        int c = fgetc(acq_file);
        switch (c) {
        case 'M': {
            int maps_size;
            EXPECT(fscanf(acq_file, "aps: %d \"", &maps_size), 1);
            scoped_array<char> maps(new char[maps_size + 1]);
            EXPECT(fread(maps.get(), 1, maps_size, acq_file),
		   (size_t)maps_size);
            maps[maps_size] = 0;
            fscanf(acq_file, "\"\n");

            uint64_t begin, end;
            char r, w, x, s;
            uint64_t offset;
            int dev_major, dev_minor;
            uint64_t inode;
            char *maps_pos = maps.get();
            int consumed;
            while ((res = sscanf(maps_pos, "%" SCNx64 "-%" SCNx64
                                 " %c%c%c%c %" SCNx64 " %" SCNx32
                                 ":%" SCNx32 " %" SCNd64 "%n",
                                 &begin, &end, &r, &w, &x, &s,
                                 &offset, &dev_major, &dev_minor,
                                 &inode, &consumed))
                   != EOF) {
                EXPECT(res, 10);
                fnr_mmap_attr_t attr = (
		    ((r == 'r') ? FNR_Mmap_Attr_Read : 0) |
		    ((w == 'w') ? FNR_Mmap_Attr_Write : 0) |
		    ((x == 'x') ? FNR_Mmap_Attr_Execute : 0) |
		    ((s == 's') ? FNR_Mmap_Attr_Shared : 0));

                maps_pos += consumed;

                string filename;
                filename.reserve(maps_size);

                // Advance past leading white spaces
                while (*maps_pos == ' ')
                    maps_pos++;

                while (*maps_pos != '\n') {
                    if (*maps_pos == 0)
                        throw resampler_exception_t(
                            "Syntax error in acquisition info file.");
                    filename.append(1, *(maps_pos++));
                }

                begin_events.push_back(
                    new fnr_event_mmap_t(
                        begin,
                        end - begin,
                        attr,
                        (dev_major << 8) | dev_minor,
                        inode,
                        offset,
                        filename.c_str(),
			0)); // fenris calls this global time, assuming sampler_time
            }
        } break;
        case 'E': {
            EXPECT(fscanf(acq_file, "nd time: %" SCNu64 "\n", &end_time), 1);
            end_events.push_back(new fnr_event_acquisition_end_t(end_time));
            return;
        } break;
        case 'B': {
            char c2;
            EXPECT(fscanf(acq_file, "urst %c", &c2), 1);
            switch (c2) {
            case 's': {
                unsigned int burst_generation;
                uint64_t time;
                uint64_t burst_sampling_length;
                uint64_t burst_cooldown_length;
                EXPECT(fscanf(acq_file, "tart: %u\n", &burst_generation), 1);
                EXPECT(fscanf(acq_file, "Burst start time: %" SCNu64 "\n",
			      &time), 1);
                EXPECT(fscanf(acq_file, "Burst sampling length: %" SCNu64 "\n",
			      &burst_sampling_length), 1);
                EXPECT(fscanf(acq_file, "Burst cooldown length: %" SCNu64 "\n",
			      &burst_cooldown_length), 1);
                burst_info_t *bi = &burst_info[burst_generation];
                bi->start_time = time;
                bi->sampling_time = burst_sampling_length;
                bi->cooldown_time = burst_cooldown_length;
            } break;
            case 'e': {
                unsigned int burst_generation;
                uint64_t time;
                EXPECT(fscanf(acq_file, "nd: %u\n", &burst_generation), 1);
                EXPECT(fscanf(acq_file, "Burst end time: %" SCNu64 "\n",
			      &time), 1);
                burst_info_t *bi = &burst_info[burst_generation];
                bi->end_time = time;
            } break;
            default:
                throw resampler_exception_t("Syntax error in acquisition "
					    "info file.");
            }
        } break;
        case EOF:
            warning->warn(new generic_warning_t("Acquisition end time "
						"missing."));
            return;
        default:
            throw resampler_exception_t("Syntax error in acquisition "
					"info file.");
        }
    }
}

#undef EXPECT

// ----------------------------------------------------------------------
// top-level resampling routine

void
flow_atf_resampler_t::check_sample_count()
{
    if (sample_count < min_samples) {
        if (burst_sampled)
            warning->warn(new resampler_not_enough_bursts_warning_t(
                              sample_count,
                              min_samples,
                              desired_samples));
        else
            warning->warn(new resampler_not_enough_samples_warning_t(
                              sample_count,
                              min_samples,
                              desired_samples,
                              sample_period * sample_count / min_samples,
                              sample_period * sample_count / desired_samples));
    }
}

void
flow_atf_resampler_t::enqueue_end_events()
{
    BOOST_FOREACH(fnr_event_t *event, end_events)
        undispatched_events.push_back(event);

    undispatched_events.push_back(new fnr_event_contexts_t(all_contexts));
}

void
flow_atf_resampler_t::resample(fnr_sample_count_t req_target_sample_count)
{
    struct atf_header header;
    struct atf_reader reader;

    if (atf_read_init(atf_file, &reader, &header) < 0)
	throw resampler_exception_t("Failed to initialize atf reader.");

    read_acquisition_info(&header);

    /* Enqueue all acquisition info events, so that they become the
       first events in the fenris sample file. */
    BOOST_FOREACH(fnr_event_t *event, begin_events)
        undispatched_events.push_back(event);

    feedback->test_cancel();

    if (req_target_sample_count == 0)
        /* No sample count specified, decimate to the sample count
           from the ATF file, if any. */
        this->target_sample_count = header.sampler_count;
    else
        if (header.sampler_count > 0 &&
            req_target_sample_count > header.sampler_count) {
            /* Decimation to greater sample count than the ATF file
               was sampled for requested. */
	    ostringstream oss;
	    oss << "Error: Decimation to " << req_target_sample_count
		<< " samples requested, but the ATF file was only sampled with target "
		<< header.sampler_count << " samples.";
	    throw resampler_exception_t(oss.str());
        }
        else
            /* Requested decimation ok. */
            this->target_sample_count = req_target_sample_count;

    /// Check if the ATF file was burst sampled.
    if (header.flags & ATF_HEADER_FLAGS_BURST) {
        burst_sampled = 1;
        if (this->target_sample_count != 0)
            throw resampler_exception_t("Error: Decimation not supported "
                                        "with burst sampled ATF files.");
    }
    else
        burst_sampled = 0;

    /* Enqueue the sample period event now if we know the sample period. Needed
       to get it at the start of the sample file for burst sampled ATF
       files. */
    if (this->target_sample_count == 0)
        undispatched_events.push_back(
	    new fnr_event_sample_period_t(header.sampler_period));

    sample_period = header.sampler_period;
    atf_sample_count = header.sampler_count;

    /* Initialize the watchpoint types. */
    byte_watchpoint_type = new watchpoint_type_t(BYTE_SAMPLE_SIZE_LOG2);
    for (int i = 0; header.sampler_log2linesizes[i] != (unsigned char)-1; i++)
	line_watchpoint_types.push_back(
	    new watchpoint_type_t(header.sampler_log2linesizes[i]));

    page_watchpoint_type = new watchpoint_type_t(PAGE_SAMPLE_SIZE_LOG2);

    next_wp_group_id = 0;

#if defined(DEBUG_LOG)
    next_wp_id = 0;
#endif

    /* Perform the actual resampling, enqueueing sample and dangling
       events. */
    struct atf_packet packet;
    int ret;

    while ((ret = atf_read_packet(&reader, &packet)) == 1) {
        handle_packet(&packet);

	feedback->test_cancel();
    }

    if (ret == -1) {
	if (!feof(atf_file))
	    throw resampler_exception_t("Failed to read ATF packet. "
					"Sampler output might be corrupted.");
	else
	    warning->warn(new resampler_premature_eof_t());
    }

    /* Enqueue the sample period if we did not do it earlier (because we need
       to calculate the sample period). */
    if (this->target_sample_count != 0)
        undispatched_events.push_back(
	    new fnr_event_sample_period_t(
		max(current_time / this->target_sample_count,
		    (fnr_time_t)header.sampler_period)));

    // Enqueue the samples if we are not burst sampling, or if we have not
    // gotten a burst end event for the last burst (may happen, e.g., if the
    // sampled process is killed with a SIGKILL).
    if (!burst_sampled || burst_generation != 0) {
        process_samples();

        // Enqueue a burst end event. Use the last seen access time as the end
        // time, since we haven't gotten any burst end packet.
        if (burst_sampled) {
            burst_info_map_t::const_iterator bii =
		burst_info.find(burst_generation);
            fnr_real_time_t end_time = 0;
            if (bii != burst_info.end()) {
                const burst_info_t *bi = &bii->second;
                end_time = bi->end_time;
            }
            fnr_event_t *burst_end_event =
		new fnr_event_burst_end_t(end_time, current_time);
            undispatched_events.push_back(burst_end_event);
        }
    }

    // Warn if we have too few samples.
    check_sample_count();

    // Enqueue the acquisition end events last.
    enqueue_end_events();
}

// ----------------------------------------------------------------------
// cleaning up

void
flow_atf_resampler_t::free_wps()
{
    BOOST_FOREACH(wp_group_map_t::value_type &v, wp_groups)
        delete v.second;
    wp_groups.clear();
}

void
flow_atf_resampler_t::cleanup()
{
    free_wps();

    // Free the threads.
    BOOST_FOREACH(thread_context_map_t::value_type &v, threads) {
        BOOST_FOREACH(thread_time_map_t::value_type &t, v.second)
            delete t.second;
    }

    // Free any undispatched events.
    while (!undispatched_events.empty()) {
	fnr_event_t *event = undispatched_events.front();
	delete event;
	undispatched_events.pop_front();
    }

    // Free the access_types.
    delete byte_watchpoint_type;
    BOOST_FOREACH(watchpoint_type_t *wpt, line_watchpoint_types)
        delete wpt;
    delete page_watchpoint_type;

    // We must not try to free:
    // * Events in undispatched_events
    // * Watchpoints in the watchpoint hash tables
    // * Watchpoint groups in threads' waiting_for_pc1_plus

    // Free the call stack structures.
    for (unsigned int i = 0; i < STACK_HASH_SIZE; i++)
        while (stack_hash[i] != NULL) {
            stack_t *next = stack_hash[i]->next;
            delete stack_hash[i];
            stack_hash[i] = next;
        }

    if (atf_file && atf_file != stdin)
	fclose(atf_file);

    if (acq_file)
	fclose(acq_file);

    kill_temp_files();
}

// ----------------------------------------------------------------------
// public methods

flow_atf_resampler_t::flow_atf_resampler_t(const fs::path &atf_filename,
					   const fs::path &acq_filename,
					   fnr_sample_count_t req_target_sample_count,
                                           bool single_thread,
					   fenris_flow_t *flow_next,
					   warning_consumer_ptr_t warning,
					   feedback_ptr_t feedback,
                                           const fs::path &tmpdir)
    : tmpdir(tmpdir),
      temp_fenris_file(0),
      meta_data_file(0),
      time_map_data_file(0),
      temp_file_writer(0),
      flow_next(flow_next),
      single_thread(single_thread),
      single_thread_seen(false),
      current_time(0),
      byte_watchpoint_type(NULL),
      page_watchpoint_type(NULL),
      next_stack_id(1),
      random_lcg(random_lcg_new(1), random_lcg_delete),
      sample_count(0),
      min_samples(10000),
      desired_samples(req_target_sample_count == 0 ?
		      50000 : req_target_sample_count),
      warning(warning),
      feedback(feedback),
      burst_generation(0),
      last_burst_generation(0)
{
    if (flow_next->flow_init(FNR_Feature_MT |
			     FNR_Feature_Virtual_Threads) != FLOW_OK)
        throw resampler_exception_t("Error: failed to initialize next flow "
				    "component.");

    /* Initialize call stack handling. Start next_stack_id at 1, since
       0 is used to represent no id. */
    for (unsigned int i = 0; i < STACK_HASH_SIZE; i++)
        stack_hash[i] = NULL;

    acq_file = acq_filename.empty() ? NULL : acumem::exe_path::fopen(acq_filename, "rb");
    atf_file = atf_filename.empty() ? stdin : acumem::exe_path::fopen(atf_filename, "rb");

    try {
	if (!acq_filename.empty() && !acq_file) {
	    ostringstream oss;
	    oss << "Error: failed to open acquisition info file "
		<< acq_filename << ": " << strerror(errno);
	    throw resampler_exception_t(oss.str());
	}

	if (!atf_file) {
	    ostringstream oss;
	    oss << "Error: failed to open atf file "
		<< atf_filename << ": " << strerror(errno);
	    throw resampler_exception_t(oss.str());
	}

	// Open the temporary files used to store completed samples
	// between the two phases
	open_temp_files();

	resample(req_target_sample_count);
    } catch (...) {
	cleanup();
	throw;
    }
}

flow_atf_resampler_t::~flow_atf_resampler_t()
{
    cleanup();
}

flow_status_t
flow_atf_resampler_t::dispatch()
{
    flow_status_t status = FLOW_OK;
    /* Dequeue and dispatch all queued fenris events. */
    while (!undispatched_events.empty() && status == FLOW_OK) {
	fnr_event_t *event = undispatched_events.front();
        status = flow_next->process_event(*event);

	undispatched_events.pop_front();
        delete event;

	feedback->test_cancel();
    }

    return status;
}

END_NAMESPACE;
END_NAMESPACE;
