/// @file library/atf_resampler/atf_resampler.hh

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#ifndef FLOW_ATF_RESAMPLER_HH
#define FLOW_ATF_RESAMPLER_HH

#include "debug_log.hh"
#include "watchpoint_type.hh"

#include "library/fenris/flow.hh"
#include "library/fenris/flow_file_reader.hh"
#include "library/fenris/flow_file_writer.hh"
#include "library/random_lcg/random_lcg.h"
#include "library/warning/warning_consumer.hh"
#include "library/feedback/feedback.hh"

#include <boost/smart_ptr.hpp>
#include <boost/filesystem.hpp>

#include <map>
#include <deque>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

class thread_t;
struct stack_t;
struct watchpoint_t;
typedef std::vector<watchpoint_t*> watchpoint_vector_t;
class wp_group_t;


class flow_atf_resampler_t
{
public:
    /// Creates an ATF resampler reading from the specified file, or reading
    /// from stdin if filename is NULL.
    ///
    /// If req_target_sample_count is 0 all samples are included in
    /// the output. If req_target_sample_count is not 0 the samples
    /// are decimated so that samples corresponding to approximately
    /// the specified count of accesses remain.
    ///
    /// \param single_thread If true any samples containing accesses from any
    /// other thread than the first we see will be deleted.
    flow_atf_resampler_t(const boost::filesystem::path &atf_filename,
			 const boost::filesystem::path &acq_filename,
			 fnr_sample_count_t req_target_sample_count,
                         bool single_thread,
			 fenris_flow_t *flow_next,
			 warning_consumer_ptr_t warning,
			 feedback_ptr_t feedback,
                         const boost::filesystem::path &tmpdir);
    ~flow_atf_resampler_t();

    /// Dispatch the resulting events.
    flow_status_t dispatch();



private:

    /// wp_group_t uses these to initialize some values in its constructor.
    friend class wp_group_t;

    /// Return the next watchpoint group id. The atf_resampler is currently
    /// the keeper of various stuff, for instance this counter.
    fnr_sample_count_t get_next_wp_group_id()
	{
	    return ++next_wp_group_id;
	}

    /// Line watchpoint types.
    const watchpoint_type_vector_t &get_line_watchpoint_types()
	{
	    return line_watchpoint_types;
	}

    /// Return the effective sample period of the thread and update it for the
    /// next sample.
    double get_effective_sample_period(thread_t *thread);

private:
    // ----------------------------------------------------------------------

    /// @{ Thread and context maps

    /// Maps from context id and global time to thread.
    typedef map<fnr_time_t, thread_t *> thread_time_map_t;
    typedef map<fnr_context_id_t, thread_time_map_t> thread_context_map_t;

    /// Map from context id to thread id;
    typedef map<fnr_context_id_t, fnr_thread_id_t> context_thread_map_t;


    /// Map from context id to a map from global time to thread.
    thread_context_map_t threads;

    /// Map from context id to the last corresponding thread id.
    context_thread_map_t thread_ids;

    /// The context ids of all threads, in the order they were encountered.
    fnr_context_id_vector_t all_contexts;

    /// Register a thread with the specified id as having been instrumented for
    /// the first time at the specified time.
    void create_thread(fnr_context_id_t context_id,
                       fnr_time_t global_time,
                       bool new_thread);

    /// Find the thread that used the specified id at the specified time.
    thread_t *find_thread(fnr_context_id_t context_id,
                          fnr_time_t global_time);

    /// @}

    // ----------------------------------------------------------------------

    /// @{ Temporary file handling

    FILE * open_temp_file();

    /// Vector of filenames that must be explicitly removed
    std::vector<boost::filesystem::path> tempfiles;

    /// Temporary file directory, ignored on windows.
    boost::filesystem::path tmpdir;

    /// The temporary file used to store fenris events between phase 1
    /// and 2
    FILE *temp_fenris_file;

    /// The temporary file used to store event meta data between phase
    /// 1 and 2
    FILE *meta_data_file;

    /// The temporary file used to store event meta data between phase
    /// 1 and 2
    FILE *time_map_data_file;

    /// The writer of the temporary file.
    flow_file_writer_t *temp_file_writer;

    /// Create three temporary files, for samples, metadata and time mappings
    void open_temp_files();

    /// Close and rewind the temporary files
    void rewind_temp_files();

    /// Close and remove temporary files
    void kill_temp_files();

    /// @}

    // ------------------------------------------------------------------------

    /// Add mapping from global to thread local time.
    void add_time_mapping(fnr_time_t global_time,
                          fnr_thread_id_t tid,
                          fnr_time_t local_time);

    // ------------------------------------------------------------------------

    /// @{ Store wp_groups to temp file

    /// Remove duplicate end events from the watchpoint.
    void remove_duplicate_end_events(watchpoint_t *wp);

    /// Return the id of the specified call stack. Assign a stack id to the
    /// specified stack and enqueue a stack event for it, if it has not already
    /// been assigned an id.
    fnr_stack_id_t stack_id(stack_t *stack);

    /// Update the stack id and clean up the end list of watchpoints
    /// prior to writing them to temp file
    fnr_event_sample_t *prepare_stack_sample_event(watchpoint_t *wp);

    /// Write metadata regarding a wp_group to the meta data temp file
    void write_wp_group_metadata(const wp_group_t *wp_group, int32_t group);

    /// check if this sample should be skipped due to accesses from
    /// forbidden threads
    bool single_thread_skip_sample(wp_group_t *wp_group) const;

    /// Write an entire wp_group to the temp file
    void write_group_to_temp_file(wp_group_t *wp_group);

    /// @}

    // ----------------------------------------------------------------------

    /// @{ read and decimate wp_groups from temp files

    /// Read the number of samples in the next wp_group, and retrieve
    /// the effective sample_period.
    int32_t read_wp_group_metadata(double *effective_sample_period);

    /// the resampling is driven from a fenris flow. It makes
    /// callbacks into this class.
    friend class flow_phase2_t;

    /// This function is called from the flow_phase2_t, to decide
    /// whether the watchpoint group with the given effective sample
    /// period should be kept (true) or decimated (false)
    bool decide_keep_event(double effective_sample_period);

    /// Mark a global time that has been seen in an access that survived
    /// decimation or a thread exit.
    void mark_global_time(fnr_time_t global_time);

    /// This function is called from the flow_phase2_t, to save this
    /// decimated sample. It is stored in a temporary holding area
    /// until later.
    void keep_event(const fnr_event_sample_t &event);

    /// Read back the time mapping temporary file and add the time mappings
    /// that have survived decimation to global_tid_local_map.
    void read_back_time_mapping();

    /// Enqueuecurrently in the time fields with the corresponding global time.
    void enqueue_samples();

    /// Enqueue time mapping events.
    void enqueue_time_mappings();

    /// Read back the watchpoints from the temporary file and decimate the
    /// watchpoint groups. Read back the time events from the temporary file
    /// and decimate them. Enqueue the sample events and time events. Then free
    /// all the watchpoint and time data structures.
    void process_samples();

    /// @}



    // ----------------------------------------------------------------------
    /// @{ ATF decoding primitives

    /// Returns the access size of an ATF event.
    static fnr_access_size_t get_access_size(struct atf_packet *packet);

    /// Returns the access type of an ATF event.
    static fnr_access_type_t get_access_type(struct atf_packet *packet);

    /// Returns the context id of an ATF event.
    static fnr_context_id_t get_context_id(struct atf_packet *packet);

    /// @}

    // ----------------------------------------------------------------------
    /// @{ memory access state machine

    /// Creates watchpoints according to the access.
    void handle_pc_1(thread_t *thread, const fnr_sample_access_t *access);

    /// Checks if there are any watchpoints waiting for a pc1+
    /// access. If so adds the access as pc1+ access to the
    /// watchpoints and returns true, otherwise returns false.
    bool handle_pc_1_plus(thread_t *thread, const fnr_sample_access_t *access);

    /// Apply the access to all watchpoints of the specified type that it
    /// overlaps. Remove any watchpoints that are terminated, add the access to
    /// any watchpoints it is a mid access of.
    bool apply_access(watchpoint_type_t *wp_type,
                      watchpoint_vector_t *wps,
                      thread_t *thread,
                      const fnr_sample_access_t *access,
                      bool must_be_last);

    /// Handles an access to an existing watchpoint group.
    void handle_pc_2(thread_t *thread,
		     fnr_sample_count_t wp_group_id,
		     const fnr_sample_access_t *access,
		     bool must_be_last);
    /// @}

    // ----------------------------------------------------------------------
    /// @{ Handle ATF packets

    /// Memory packets
    void handle_mem(struct atf_packet *packet);

    /// Watchpoint kill packets. Sets the 'killed' flag on samples
    /// belonging to a killed watchpoint group.
    void handle_kill(fnr_sample_count_t group_id);

    /// Call stack packet, update the stack of the thread.
    void handle_stack(struct atf_packet *packet);

    /// Burst end packet
    void handle_burst_end(struct atf_packet *packet);

    /// Thread packet
    void handle_thread(struct atf_packet *packet);

    /// MMAP packets
    void handle_mmap(struct atf_packet *packet);

    /// MUNMAP packets
    void handle_munmap(struct atf_packet *packet);

    /// Global macnt packet
    void handle_global_macnt(struct atf_packet *packet);

    /// Local macnt packet
    void handle_local_macnt(struct atf_packet *packet);

    /// Thread exit packet
    void handle_thread_exit(struct atf_packet *packet);

    /// Handle an ATF packets.
    void handle_packet(struct atf_packet *packet);

    /// @}

    // ----------------------------------------------------------------------
    /// @{ Acquisition info file parser

    /// Return the number of line sizes.
    static int get_num_line_sizes(const unsigned char *sampler_log2linesizes);

    /// Read acquisition info from the specified file and add corresponding
    /// events to begin_events and end_events in the resampler.
    void read_acquisition_info(const struct atf_header *header);


    /// @}

    // ----------------------------------------------------------------------
    /// @{ top-level resampling routines

    /// Warn if we don't have enough samples for a reliable result.
    void check_sample_count();

    /// If there are pending end events, enqueue them.
    void enqueue_end_events();

    /// Read through the ATF file and perform the actual resampling.
    void resample(fnr_sample_count_t req_target_sample_count);


    /// @}

    // ----------------------------------------------------------------------
    /// @{ cleaning up

    /// Free all watchpoint groups.
    void free_wps();

    /// Final cleanup of everything
    void cleanup();

    /// @}


    // ----------------------------------------------------------------------
    /// @{ Private types and variables

    /// Vector of watchpoint groups.
    typedef map<fnr_sample_count_t, wp_group_t *> wp_group_map_t;

    /// The watchpoint groups.
    wp_group_map_t wp_groups;


    /// Event queues
    typedef deque<fnr_event_t*> event_queue_t;

    /// Queue of finished but not yet dispatched events.
    event_queue_t undispatched_events;

    /// Queue of half-baked sample events.
    event_queue_t half_baked_samples;


    /// Events from the acquisition info file to add to the beginning and end
    /// of the fenris file.
    typedef list<fnr_event_t*> event_list_t;
    event_list_t begin_events;
    event_list_t end_events;

    /// ATF file pointer.
    FILE *atf_file;

    /// Acquisition info file pointer.
    FILE *acq_file;

    /// Next flow node.
    fenris_flow_t *flow_next;

    /// Sample period of the ATF file.
    fnr_time_t sample_period;

    /// 0 if the ATF file was sampled with fixed sample period, a non-0 sample
    /// count if the ATF file was sampled with automatic sample period
    /// adjustment.
    fnr_sample_count_t atf_sample_count;

    /// We want samples corresponding to this number of memory accesses.
    fnr_sample_count_t target_sample_count;

    /// If true any samples containing accesses from any other thread than the
    /// first we see will be deleted.
    bool single_thread;
    bool single_thread_seen;
    fnr_thread_id_t single_thread_id;

    /// This is the last recorded combined time.
    fnr_time_t current_time;

    // TODO: We could merge seen_global_times, global_tid_local_map into one
    //       data structure to reduce the memory and processing overhead. They
    //       are all basically maps from global time to something else.

    /// Global times seen in accesses that survived decimation or thread exits.
    fnr_time_set_t seen_global_times;

    /// Mapping from global time to thread id to local time.
    ///
    /// When the time mappings are read from the temporary file, the mappings
    /// from global times that have survived decimation to thread ids and
    /// corresponding local times are stored here.
    typedef map<fnr_thread_id_t, fnr_time_t> tid_local_map_t;
    typedef map<fnr_time_t, tid_local_map_t> global_tid_local_map_t;
    global_tid_local_map_t global_tid_local_map;

    /// The byte watchpoint type.
    watchpoint_type_t *byte_watchpoint_type;

    /// Line watchpoint types.
    watchpoint_type_vector_t line_watchpoint_types;

    /// The page watchpoint type.
    watchpoint_type_t *page_watchpoint_type;

    /// Stack id to assign to the next call stack.
    fnr_stack_id_t next_stack_id;

    /// Number of buckets in the call stack hash table.
    static const uint64_t STACK_HASH_SIZE = 0x1000;

    /// Mask for the bits in the call stack hash table index.
    static const uint64_t STACK_HASH_MASK = STACK_HASH_SIZE - 1;

    /// Call stack hash table.
    stack_t *stack_hash[STACK_HASH_SIZE];

    /// Pseudo-random number generator, for repeatable pseudo-random numbers.
    shared_ptr<random_lcg_t> random_lcg;

    /// Number of samples enqueued for the sample file.
    fnr_sample_count_t sample_count;

    /// Minimum number of samples to form a reliable sampling.
    fnr_sample_count_t min_samples;

    /// Desired number of samples, for the purpose of producing a
    /// meaningful error message, unlike the member
    /// 'target_sample_count', which is used for actually performing
    /// the decimation.
    fnr_sample_count_t desired_samples;

    /// Pointer to a warning consumer.
    warning_consumer_ptr_t warning;

    /// Pointer to a feedback object
    feedback_ptr_t feedback;

    /// Next watchpoint group id to use. Used to match sample groups against
    /// sample groups in the sampler.
    fnr_sample_count_t next_wp_group_id;

#if defined(DEBUG_LOG)
    /// Next watchpoint id to use. Used to match watchpoints against the
    /// sampler's samples.
    fnr_sample_count_t next_wp_id;

    struct debug_log_wp_t {
        debug_log_wp_t(fnr_sample_count_t id,
                       const char *event_type,
                       uint64_t start,
                       fnr_access_size_t length)
            : id(id),
              event_type(event_type),
              start(start),
              length(length) {
        }

        fnr_sample_count_t id;
        const char *event_type;
        uint64_t start;
        fnr_access_size_t length;

        bool operator < (const debug_log_wp_t &other) const {
            if (this->start >> 12 != other.start >> 12)
                return this->start < other.start;

            if (this->length != other.length)
                return this->length < other.length;

            return this->id > other.id;
        }
    };

    typedef set<debug_log_wp_t> debug_log_wp_set_t;

    /// We need to report watchpoint mid accesses and terminations caused by
    /// the same access in ascending page order, descending size and reverse
    /// creation order to match the sampler when generating the debug log. We
    /// therefore buffer the messages in this set that is sorted in that order.
    debug_log_wp_set_t debug_log_wps;
#endif

    /// True if we are resampling a burst sampled ATF file.
    bool burst_sampled;

    /// The number of the current burst generation, or 0 if none.
    unsigned int burst_generation;

    /// The number of the last burst generation, or 0 if none.
    unsigned int last_burst_generation;

    /// Info about a burst.
    struct burst_info_t {
        burst_info_t()
            : start_time(0),
              sampling_time(0),
              cooldown_time(0),
              end_time(0)
            {}

        fnr_real_time_t start_time;
        fnr_time_t sampling_time;
        fnr_time_t cooldown_time;
        fnr_real_time_t end_time;
    };
    typedef map<unsigned int, burst_info_t> burst_info_map_t;

    /// Info about the bursts.
    burst_info_map_t burst_info;

};

END_NAMESPACE;
END_NAMESPACE;

#endif
