/// @file library/atf_resampler/byte_watchpoint.cc

// Copyright (c) 2009 Acumem AB.  All rights reserved.

#include "byte_watchpoint.hh"
#include "access_type.hh"

#include <boost/foreach.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

byte_watchpoint_t::byte_watchpoint_t(flow_atf_resampler_t *resampler,
				     thread_t *thread,
				     uint64_t wp_addr,
				     wp_group_t *wp_group,
				     line_watchpoint_map_t *line_wps,
				     const fnr_sample_access_t *access)
    : watchpoint_t(resampler, thread, wp_addr, BYTE_SAMPLE_SIZE_LOG2,
		   wp_group, access),
      line_wps(*line_wps),
      live_wp_types(byte_start_types[access->access_type])
{
    BOOST_FOREACH(line_watchpoint_map_t::value_type &v, *line_wps)
        v.second->live_byte_wps++;
}

int
byte_watchpoint_t::apply_access_impl(const fnr_sample_access_t *access,
				     stack_t *stack,
				     fnr_line_size_t line_size,
				     bool must_be_last)
{
    int ret = 0x0;
    access_type_set_t access_type_set =
        get_access_type_set(access->access_type);

    if (!terminated &&
        !(access_type_set & ATS_CLASS_NO_DATA)) {

	// This is only run for the first event of each thread,
	// and also the very last, finishing event of a byte
	// watchpoint (from the original thread). All other last
	// accesses of the threads will have the must_be_last flag
	// set.
        if (!must_be_last) {

	    // Check if this is the first of this access type for this
	    // thread. If so, flag this access as being
	    // interesting, and then watchpoint_t::apply_access
	    // will add this access to the finished fenris sample.
            if (!(seen_access_types[access->thread.id] & access_type_set)) {
                seen_access_types[access->thread.id] |= access_type_set;
                ret |= 0x1;
            }

	    // If this access is from the same thread as started the
	    // watchpoint, then its last access may be eligible to be
	    // booked for the line sample as well.

	    // XXX It is unclear why only the same tid, and not all
	    // threads' last acceesses are booked in this way.
	    //
	    // Also, it is appears as though the last accesses of the
	    // other threads will have must_be_last set, and therefore
	    // be excluded here by the condition at the top.

            if (access->thread == tid) {
                access_type_set_t new_live_wp_types =
                    live_wp_types & ~byte_end_types[access->access_type];

		// If we have seen all required access types (for the
		// original thread), then we are done with this
		// watchpoint, and we need to inform the corresponding
		// line watchpoint, so it can reduce its view of live
		// byte watchpoints.
		//
		// We also log this last byte access to the line. This
		// is likely to facilitate utilization calculation

		// XXX it is unclear why the "last" accesses on other
		// threads than the original one doesn't cause logging
		// to the line watchpoint.

                if (new_live_wp_types == 0) {
                    log_last_accesses(&last_access);
                    BOOST_FOREACH(line_watchpoint_map_t::value_type &v, line_wps) {
                        v.second->live_byte_wps--;
                        v.second->log_last_accesses(&v.second->last_access);
                    }

                    terminated = true;
                    return 0x3;
                }

                if (new_live_wp_types != live_wp_types) {
                    live_wp_types = new_live_wp_types;
                    ret |= 0x1;
                }
            }
        }

	// Always make note of the last access per thread and access type.
	// XXX what is this needed for?
        last_access[access->thread.id][access->access_type] =
	    make_pair(*access, stack);
    }

    return ret;
}

END_NAMESPACE;
END_NAMESPACE;
