/// @file library/atf_resampler/line_watchpoint.cc

// Copyright (c) 2009 Acumem AB.  All rights reserved.

#include "line_watchpoint.hh"

#include <boost/foreach.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

line_watchpoint_t::line_watchpoint_t(flow_atf_resampler_t *resampler,
				     thread_t *thread,
				     uint64_t wp_addr,
				     fnr_line_size_log2_t wp_size_log2,
				     wp_group_t *wp_group,
				     const fnr_sample_access_t *access)
    : watchpoint_t(resampler, thread, wp_addr, wp_size_log2, wp_group, access),
      touched_bytes(1 << wp_size_log2),
      live_byte_wps(0),
      live_wp_types(line_start_types[access->access_type]),
      start_access_type(access->access_type),
      maybe_inval(false),
      read_only(false)
{
}

int
line_watchpoint_t::apply_access_impl(const fnr_sample_access_t *access,
				     stack_t *stack,
				     fnr_line_size_t line_size,
				     bool must_be_last)
{
    int ret = 0x0;

    if (!terminated) {
        if (!must_be_last) {

	    // Maintain a record of seen access types per thread. Mark
	    // the first one of each, to make
	    // watchpoint_t::apply_access add an event to the sample
	    // for this access.

	    access_type_set_t access_type_set =
		get_access_type_set(access->access_type);
            if (!(seen_access_types[access->thread.id] & access_type_set)) {
                seen_access_types[access->thread.id] |= access_type_set;
                ret |= 0x1;
            }

	    // This watchpoint represents many parallel watches,
	    // starting and ending with different sets of access
	    // types. live_wp_types tell which types are waited for,
	    // and still adds information to the watchpoint. This
	    // mirrors the watchpoint mechanism that takes place in
	    // the sampler.

            if (live_wp_types != 0) {

                if (access->thread == tid) {

		    // accesses from the same thread are watched for
		    // the last access. The last access is marked to
		    // be recorded in the sample.

                    access_type_set_t new_live_wp_types =
                        access->access_type == start_access_type ?
                        0 :
                        (live_wp_types & ~line_end_types[access->access_type]);

                    if (new_live_wp_types != live_wp_types) {
                        if (new_live_wp_types == 0)
                            log_last_accesses(&last_access);
                        live_wp_types = new_live_wp_types;
                        ret |= 0x1;
                    }
                }
                else {
                    if (access_type_set & ATS_CLASS_MAYBE_INVALIDATION)
                        maybe_inval = true;

		    // Accesses from different threads are monitored
		    // for false sharing, as long as the watchpoint is
		    // not completed with respect to live_wp_types.

		    // Find or create a record for this thread.

                    false_sharing_map_t::iterator fsi =
			false_sharing.find(access->thread.id);
                    if (fsi == false_sharing.end())
                        fsi = false_sharing.insert(
			    make_pair(access->thread.id,
				      false_sharing_t(line_size))).first;
		    false_sharing_t &fs = fsi->second;

		    // If the access type involves actually using
		    // data, mark the bytes being accessed. If new
		    // info was recorded, mark this access to be
		    // recorded in the sample.

                    if (!(access_type_set & ATS_CLASS_NO_DATA)) {
                        uint64_t offset =
			    access->addr.virtual_address % line_size;
                        fnr_line_size_t length =
			    min(access->access_size,
				(fnr_access_size_t)(line_size - offset));
                        if (fs.touched_bytes.touch(offset, length,
				!!(access_type_set & ATS_CLASS_ALWAYS_UPGRADE)))
                            ret |= 0x1;
                    }
                }

                if (live_wp_types == 0 &&
                    !(access_type_set & ATS_CLASS_MAYBE_UPGRADE) &&
                    !maybe_inval)
                    read_only = true;
            }

	    // false_sharing is empty as long as there hasn't been any
	    // accesses from different threads. When we start see such
	    // accesses, we also gather information for the original
	    // thread.

            if (!false_sharing.empty()) {
                ret |= 0x1;

                if (live_wp_types == 0) {
                    if (access->thread == tid) {

                        if (read_only)
                            false_sharing.clear();

			// XXX Explain why a cache-line flush clears the
			// false sharing status.

                        if (access_type_set & ATS_CLASS_FLUSH)
                            false_sharing.clear();

			// Ordinary accesses updates the touched bytes

                        if (!(access_type_set & ATS_CLASS_NO_DATA)) {
                            uint64_t offset =
				access->addr.virtual_address % line_size;
                            fnr_line_size_t length =
				min(access->access_size,
				    (fnr_access_size_t)(line_size - offset));
                            if (touched_bytes.touch(
				    offset, length,
				    !!(access_type_set &
				       ATS_CLASS_ALWAYS_UPGRADE))) {
				// For accesses that cause new bits to
				// be set, check if they intersect any
				// accesses from different threads. If
				// so, these other accesses do not
				// constitute false sharing.

				// XXX explain the goto restart0

                            restart0:
                                BOOST_FOREACH(false_sharing_map_t::value_type &v, false_sharing)
                                    if (touched_bytes.covers(&v.second.touched_bytes)) {
                                        false_sharing.erase(v.first);
                                        goto restart0;
                                    }
                            }
                        }

			// XXX Explain what accesses are ALWAYS_UPGRADE and why
			// that clears what false sharing records.

                        if (access_type_set & ATS_CLASS_ALWAYS_UPGRADE) {
                        restart1:
                            BOOST_FOREACH(false_sharing_map_t::value_type &v, false_sharing)
                                if (v.second.downgrade) {
                                    false_sharing.erase(v.first);
                                    goto restart1;
                                }
                        }
                    }
                    else {

			// Access from a different thread.

			// XXX explain what we are doing here.

                        false_sharing_map_t::iterator fsi =
			    false_sharing.find(access->thread.id);
                        if (fsi != false_sharing.end()) {

			    // XXX Possible bug? erasing the iterator
			    // invalidates it. Yet the second test may
			    // access it.

                            if (access_type_set & ATS_CLASS_ALWAYS_INVALIDATION)
                                false_sharing.erase(fsi);

                            if (access_type_set & ATS_CLASS_ALWAYS_DOWNGRADE)
                                fsi->second.downgrade = true;
                        }
                    }
                }
            }

	    // Condition for terminating the watchpoint:
	    // * all relevant access types have been seen on the
	    //   original thread.
	    // * we have gathered enough info to decided whether there is
	    //   false sharing or not. This may be overly zealous, see #723
	    // * all related byte watchpoints have been terminated

            if (live_wp_types == 0 &&
                false_sharing.empty() &&
                live_byte_wps == 0) {
                terminated = true;
                return 0x3;
            }
        }


        last_access[access->thread.id][access->access_type] =
	    make_pair(*access, stack);
    }

    return ret;
}

END_NAMESPACE;
END_NAMESPACE;
