/// @file library/atf_resampler/line_watchpoint.hh

// Copyright (c) 2009 Acumem AB.  All rights reserved.

#ifndef LINE_WATCHPOINT_HH
#define LINE_WATCHPOINT_HH

#include "access_type.hh"
#include "watchpoint.hh"
#include "thread.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

/// Line watchpoint.
struct line_watchpoint_t : public watchpoint_t {

    line_watchpoint_t(flow_atf_resampler_t *resampler,
		      thread_t *thread,
		      uint64_t wp_addr,
		      fnr_line_size_log2_t wp_size_log2,
		      wp_group_t *wp_group,
		      const fnr_sample_access_t *access);


    /// Structure to record which bytes have been read respectively
    /// written in a cache line.
    struct touched_bytes_t {
	touched_bytes_t(fnr_line_size_t line_size)
	    : accessed((vector<bool>::size_type)line_size, false),
	      written((vector<bool>::size_type)line_size, false)
	    {}

	bool touch(uint64_t offset, fnr_line_size_t length, bool write) {
	    bool new_byte = false;
	    for (fnr_line_size_t i = offset; i < offset + length; i++) {
		if (!accessed[(vector<bool>::size_type)i]) {
		    accessed[(vector<bool>::size_type)i] = true;
		    new_byte = true;
		}
		if (write && !written[(vector<bool>::size_type)i]) {
		    written[(vector<bool>::size_type)i] = true;
		    new_byte = true;
		}
	    }
	    return new_byte;
	}

	bool covers(touched_bytes_t *tb) {
	    assert(written.size() == accessed.size() &&
		   written.size() == tb->accessed.size() &&
		   accessed.size() == tb->written.size());

	    for (vector<bool>::size_type i = 0; i < accessed.size(); i++) {
		if (tb->accessed[(vector<bool>::size_type)i] && !written[(vector<bool>::size_type)i])
		    return false;
		if (tb->written[(vector<bool>::size_type)i] && !accessed[(vector<bool>::size_type)i])
		    return false;
	    }
	    return true;
	}

	vector<bool> accessed;
	vector<bool> written;
    };

    /// structure that records whether we have detected a downgrade
    /// miss, and maintains data to judge if false sharing has
    /// occurred.
    ///
    /// XXX inaccurate name for this structure, since it deals with
    /// other things than false sharing.
    struct false_sharing_t {
	false_sharing_t(fnr_line_size_t line_size)
	    : downgrade(false),
	      touched_bytes(line_size) {}

	bool downgrade;
	touched_bytes_t touched_bytes;
    };

    /// For each thread, maintain a downgrade/false sharing collector
    typedef map<fnr_thread_id_t, false_sharing_t> false_sharing_map_t;
    false_sharing_map_t false_sharing;

    /// Once we have seen accesses from other threads, record accesses
    /// from the original thread here to be able to compare for a
    /// false sharing verdict.
    touched_bytes_t touched_bytes;

    /// Number of live byte watchpoints associated with the watchpoint.
    fnr_sample_count_t live_byte_wps;

    /// Live watchpoint access types.
    access_type_set_t live_wp_types;

    /// Access type that started the watchpoint.
    fnr_access_type_t start_access_type;

    /// Seen access types from each thread.
    thread_access_type_set_t seen_access_types;

    /// The last access of each type from each thread.
    last_access_t last_access;

    /// True if a thread that did not start the sample has done an access that
    /// may have invalidated the line.
    int maybe_inval;

    /// True if the end access may not have caused an upgrade and no thread
    /// that did not start the sample has done an access that may have
    /// invalidated the line.
    int read_only;

    int apply_access_impl(
	const fnr_sample_access_t *access,
	stack_t *stack,
	fnr_line_size_t line_size,
	bool must_be_last);
};

/// Maps a line size to a line watchpoint. Used by byte watchpoint to
/// map each byte watchpoint to all currently active line_watchpoints,
/// to be able to deposit extra events in the related sample.

typedef map<fnr_line_size_t, line_watchpoint_t *> line_watchpoint_map_t;

END_NAMESPACE;
END_NAMESPACE;

#endif
