/// @file library/atf_resampler/watchpoint.cc

// Copyright (c) 2009 Acumem AB.  All rights reserved.

#include "watchpoint.hh"

#include "debug_log.hh"
#include "thread.hh"
#include "wp_group.hh"

#include <boost/foreach.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

// Returns a copy of the supplied access with its address and size
// trimmed to the part that overlaps with the cache line.
static inline fnr_sample_access_t
trim_access(const fnr_sample_access_t *access,
            uint64_t line_addr,
            fnr_line_size_t line_size)
{
    uint64_t start = max(access->addr.virtual_address, line_addr);
    uint64_t end = min(access->addr.virtual_address + access->access_size,
                       line_addr + line_size);

    return fnr_sample_access_t(
        access->pc,
        start,
        access->time,
        access->stack,
        access->thread,
        (fnr_access_size_t)(end - start),
        access->access_type);
}

watchpoint_t::watchpoint_t(flow_atf_resampler_t *resampler,
			   thread_t *thread,
			   uint64_t wp_addr,
			   fnr_line_size_log2_t wp_size_log2,
			   wp_group_t *wp_group,
			   const fnr_sample_access_t *access)
    : wp_group(wp_group),
      tid(thread->thread_id),
      line_index(wp_addr >> wp_size_log2),
      pc_1_stack(thread->stack),
      pc_1_plus_stack(0),
      terminated(false)
{
    uint64_t line_addr = (uint64_t)line_index << wp_size_log2;
    fnr_line_size_t line_size = (fnr_line_size_t)1 << wp_size_log2;

    // Fill in the parts of the sample that we know from the pc1 access, the
    // rest is filled in when we get the pc1+, mid and end accesses.
    event = new fnr_event_sample_t(
        trim_access(access,
		    line_addr,
		    line_size),       // PC1 access
        fnr_sample_access_vector_t(), // PC2 access
        fnr_address_t(0),             // PC1+
        0,                            // PC1+ stack id, calculated at dispatch
        wp_size_log2,                 // Line size
	0);                           // Sample flags

#if defined(DEBUG_LOG)
    id = resampler->next_wp_id++;
#endif
    DEBUG_LOG_WP(tid, wp_group->id, id, "new", line_addr, line_size);
}

watchpoint_t::~watchpoint_t()
{
    delete event;
}

void
watchpoint_t::log_last_accesses(last_access_t *last_access)
{
    // Log last accesses, means that the very last access from each
    // thread and of each access type is added to the sample.

    // XXX Explain when this is called. (Is it when the watchpoint is
    // terminated?)

    // The outer loop loops over all threads
    // Log last accesses, means that the very last access from each
    // thread and of each access type is added to the sample.

    // XXX Explain when this is called. (Is it when the watchpoint is
    // terminated?)

    // The outer loop loops over all threads
    BOOST_FOREACH(last_access_t::value_type &la, *last_access) {
	// The inner loop loops over all access types
        BOOST_FOREACH(thread_last_access_t::value_type &tla, la.second) {
	    const fnr_sample_access_t &a(tla.second.first);
            event->end.push_back(a);
            end_stacks.insert(make_pair(a.time, tla.second.second));
        }
    }

    // XXX is it necessary to clear the last_access for any other reason
    // than saving memory?
    last_access->clear();
}

int
watchpoint_t::apply_access(flow_atf_resampler_t *resampler,
			   thread_t *thread,
			   const fnr_sample_access_t *access,
			   fnr_line_size_log2_t wp_size_log2,
			   bool must_be_last)
{
    uint64_t line_addr = (uint64_t)line_index << wp_size_log2;
    fnr_line_size_t line_size = (fnr_line_size_t)1 << wp_size_log2;

    fnr_sample_access_t trimmed_access(trim_access(access, line_addr,
						   line_size));

    int res = apply_access_impl(&trimmed_access, thread->stack,
				line_size, must_be_last);

    // If the apply_access_impl deemed the access to be interesting
    // for some reason, add it to the sample
    if (res & 0x1) {
        event->end.push_back(trimmed_access);
        end_stacks.insert(make_pair(trimmed_access.time, thread->stack));
    }

    // Magic return values:
    // 0 - access not interesting
    // 1 - access was logged
    // 2 - access was logged (presumably), but marks the end of the watchpoint

    return (res & 0x2) ? 2 : ((res & 0x1) ? 1 : 0);
}

void
watchpoint_t::pc_1_plus(thread_t *thread,
			const fnr_sample_access_t *access)
{
    /* Save the pc and stack of the pc1+ access. */
    pc_1_plus_stack = thread->stack;
    event->pc_1_plus = access->pc;
}

END_NAMESPACE;
END_NAMESPACE;
