/// @file library/atf_resampler/watchpoint.hh

// Copyright (c) 2009 Acumem AB.  All rights reserved.

#ifndef WATCHPOINT_HH
#define WATCHPOINT_HH

#include "stack.hh"

#include "library/lang/common_macros.h"
#include "library/lang/common_types.h"
#include "library/fenris/fenris.hh"
#include "sampler/src/sampler_limits.h"

#include <map>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

class flow_atf_resampler_t;
class thread_t;
class wp_group_t;
struct fnr_sample_access_t;


/// A watchpoint. This struct is owned by the access_t that it is
/// associated with.
struct watchpoint_t {
    watchpoint_t(flow_atf_resampler_t *resampler,
		 thread_t *thread,
		 uint64_t wp_addr,
		 fnr_line_size_log2_t wp_size_log2,
		 wp_group_t *wp_group,
		 const fnr_sample_access_t *access);

    virtual ~watchpoint_t();

    /// Corresponding watchpoint group.
    wp_group_t *wp_group;

    /// The sample event corresponding to the watchpoint.
    fnr_event_sample_t *event;

    /// Thread that started the watchpoint.
    fnr_thread_id_t tid;

    /// Line index, i.e., (virtual_address >> line_size_log2).
    fnr_line_count_t line_index;

    /// Call stacks.
    stack_t *pc_1_stack;
    stack_t *pc_1_plus_stack;
    time_stack_map_t end_stacks;

    /// Keep track of if the watchpoint has been terminated yet. Also set
    /// if the watchpoint has been killed because if caused too many
    /// watchpoint filter hits or logged too many memory accesses. If not
    /// the watchpoint is still in the watchpoint hash table.
    bool terminated;

    void log_last_accesses(last_access_t *last_access);

    /// Apply the specified access to the watchpoint. Returns 0 if the
    /// access did not affect the watchpoint, 1 of it was a mid access of
    /// the watchpoint or 2 if it terminated the watchpoint.
    int apply_access(flow_atf_resampler_t *resampler,
		     thread_t *thread,
		     const fnr_sample_access_t *access,
		     fnr_line_size_log2_t wp_size_log2,
		     bool must_be_last);

    /// Called by apply_access() to implement the semantics of the specific
    /// watchpoint type. The return value is a bitmask:
    ///
    /// bit 0: set if the access should be added to the watchpoint's
    ///        mid/end accesses
    /// bit 1: set if the watchpoint was terminated
    virtual int apply_access_impl(
	const fnr_sample_access_t *access,
	stack_t *stack,
	fnr_line_size_t line_size,
	bool must_be_last) = 0;

    /// Add the pc1+ access to the watchpoint.
    void pc_1_plus(
	thread_t *thread,
	const fnr_sample_access_t *access);

#if defined(DEBUG_LOG)
    /// Watchpoint id to match against the sampler's for debugging.
    fnr_sample_count_t id;
#endif
};

/// Vector of watchpoints.
typedef vector<watchpoint_t *> watchpoint_vector_t;

/// Maps a line size to a vector of watchpoints.
typedef map<fnr_line_size_t, watchpoint_vector_t> watchpoint_vector_map_t;

END_NAMESPACE;
END_NAMESPACE;

#endif
