/// @file library/atf_resampler/wp_group.cc

// Copyright (c) 2009 Acumem AB.  All rights reserved.

#include "wp_group.hh"

// Factor out the dependencies on atf_resampler. It needs it for the
// get_effective_sample_period(thread) call, as well as the
// line_watchpoint_types array and the next_wp_group_id member.
#include "atf_resampler.hh"
#include "thread.hh"

#include <boost/foreach.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

wp_group_t::wp_group_t(flow_atf_resampler_t *resampler,
		       thread_t *thread,
		       const fnr_sample_access_t *access)
    : resampler(resampler),
      id(resampler->get_next_wp_group_id()),
      start_access_type(get_access_type_set(access->access_type)),
      has_pc_1_plus(false),
      effective_sample_period(resampler->get_effective_sample_period(thread)),
      decimated(false),
      last_sampler_time(0),
      complete_lacking_pc_1_plus(false)
{
}

wp_group_t::~wp_group_t()
{
    BOOST_FOREACH(watchpoint_t *wp, byte_wps)
        delete wp;
    BOOST_FOREACH(const watchpoint_type_t *wp_type,
		  resampler->get_line_watchpoint_types()) {
        BOOST_FOREACH(watchpoint_t *wp, line_wps[wp_type->wp_size])
            delete wp;
    }
    BOOST_FOREACH(watchpoint_t *wp, page_wps)
        delete wp;
}

void
wp_group_t::set_sample_flag(fnr_sample_flags_t flags)
{
    BOOST_FOREACH(watchpoint_t *wp, byte_wps)
        wp->event->flags |= flags;

    BOOST_FOREACH(const watchpoint_type_t *wp_type,
		  resampler->get_line_watchpoint_types()) {
        BOOST_FOREACH(watchpoint_t *wp, line_wps[wp_type->wp_size])
            wp->event->flags |= flags;
    }
    BOOST_FOREACH(watchpoint_t *wp, page_wps)
        wp->event->flags |= flags;
}

void
wp_group_t::pc_1_plus(thread_t *thread,
		      const fnr_sample_access_t *access)
{
    assert(!has_pc_1_plus);
    has_pc_1_plus = true;

    // The pc1+ access should always be the immediately following access. This
    // assertion should help detect if the resampler has gotten out of sync
    // with the sampler.
    assert(thread->local_time == thread->last_local_begin_time + 1);

    BOOST_FOREACH(watchpoint_t *wp, byte_wps)
        wp->pc_1_plus(thread, access);
    BOOST_FOREACH(watchpoint_vector_map_t::value_type &v, line_wps) {
        BOOST_FOREACH(watchpoint_t *wp, v.second)
            wp->pc_1_plus(thread, access);
    }
    BOOST_FOREACH(watchpoint_t *wp, page_wps)
        wp->pc_1_plus(thread, access);
}

bool
wp_group_t::has_live_wps()
{
    BOOST_FOREACH(watchpoint_t *wp, byte_wps)
        if (!wp->terminated)
            return true;
    BOOST_FOREACH(const watchpoint_type_t *wp_type,
		  resampler->get_line_watchpoint_types()) {
        BOOST_FOREACH(watchpoint_t *wp, line_wps[wp_type->wp_size])
            if (!wp->terminated)
                return true;
    }
    BOOST_FOREACH(watchpoint_t *wp, page_wps)
        if (!wp->terminated)
            return true;
    return false;
}

END_NAMESPACE;
END_NAMESPACE;
