/// @file library/cache_sim/cachesim.hh

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#ifndef CACHE_SIM_HH
#define CACHE_SIM_HH

#include "library/lang/common_macros.h"
#include "library/lang/common_types.h"
#include "library/random_lcg/random_lcg.h"
#include "library/fenris/fnr_types.hh"
#include "library/fenris/fenris.hh"
#include "library/fenris/flow.hh"

#include <string>
#include <list>
#include <vector>
#include <map>
#include <set>
#include <bitset>
#include <boost/dynamic_bitset.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

using boost::dynamic_bitset;

// FOOTPRINT_CHUNK_SIZE determines the number bytes/lines to represent in each
// of the bitmap chunks used to keep track of the footprint.
#define FOOTPRINT_CHUNK_LOG2 16
#define FOOTPRINT_CHUNK_SIZE (1ULL << FOOTPRINT_CHUNK_LOG2)
#define FOOTPRINT_CHUNK_MASK (FOOTPRINT_CHUNK_SIZE - 1)

// Sparse bitmap used to keep track of the footprint with byte line size granularity.
typedef bitset<1ULL << FOOTPRINT_CHUNK_LOG2> footprint_bitmap_chunk_t;
typedef map<fnr_address_t, footprint_bitmap_chunk_t> footprint_bitmap_t;

struct cachesim_memop_t {
    fnr_address_t     address;
    fnr_access_size_t size;
};

class cachesim_request_interface_t {
public:
    virtual ~cachesim_request_interface_t();

    virtual void rd(const cachesim_memop_t &memop) = 0;
    virtual void rdx(const cachesim_memop_t &memop) = 0;
    virtual void wb(const cachesim_memop_t &memop) = 0;
};

class cachesim_action_interface_t {
public:
    virtual ~cachesim_action_interface_t();

    virtual void inv(const cachesim_memop_t &memop) = 0;
    virtual void cb(const cachesim_memop_t &memop) = 0;
    virtual void cb_inv(const cachesim_memop_t &memop) = 0;
};

struct cachesim_line_t {
    cachesim_line_t() {}

    // warning C4244: 'argument' : conversion from 'fnr_line_size_t' to 'dynamic_bitset<>::size_type', possible loss of data
    cachesim_line_t(bool _valid, bool _dirty, fnr_line_size_t line_size)
        : valid(_valid), dirty(_dirty), used_bytes((dynamic_bitset<>::size_type)line_size) {}

    uint32_t valid     : 1;
    uint32_t dirty     : 1;
    // add "uint32_t exclusive : 1;" here when MP support is added

    // Bitmap with one bit per byte that keeps track of which bytes have been
    // used for the spatial usage calculation.
    dynamic_bitset<> used_bytes;
};

class cachesim_conf_t {
public:
    cachesim_conf_t();
    cachesim_conf_t(fnr_cache_size_t size,
                    fnr_line_size_t line_size,
                    fnr_line_count_t assoc,
                    bool calc_footprint,
                    bool calc_spatial_usage);

    fnr_cache_size_t size;
    fnr_line_size_t line_size;
    fnr_line_count_t assoc;
    bool calc_footprint;
    bool calc_spatial_usage;
};

class cachesim_result_t {
public:
    cachesim_result_t();
    cachesim_result_t(
        uint64_t rd_total, uint64_t rd_repl_misses, uint64_t rd_cold_misses,
        uint64_t wr_total, uint64_t wr_repl_misses, uint64_t wr_cold_misses,
        uint64_t wb_incomming, uint64_t wb_outgoing,
        uint64_t fp_line, uint64_t fp_byte);

    uint64_t rd_total;
    uint64_t rd_repl_misses;
    uint64_t rd_cold_misses;
    uint64_t wr_total;
    uint64_t wr_repl_misses;
    uint64_t wr_cold_misses;
    uint64_t wb_incomming;
    uint64_t wb_outgoing;

    // Footprint in bytes calculated with line size granularity and byte
    // granularity, respectively.
    uint64_t fp_line;
    uint64_t fp_byte;

    // Keeps track of the number of used bytes used in the spatial usage
    // calculation.
    uint64_t used_bytes;
};

// Types keeping track of cache sets and cache
// lines. cachesim_line_map_vector_t contains a vector indexed by set index
// with one cache_line_map_t of for each set in the cache. The cache_line_map_t
// maps tags to cache lines for all valid lines in the set.
typedef map<fnr_address_t, cachesim_line_t> cachesim_line_map_t;
typedef vector<cachesim_line_map_t> cachesim_line_map_vector_t;

// The cachesimulator does not have support for MP simulations right
// now. It has support for multiple layers of caches. However, all
// caches have to use inclusion and has to be write-back.

// I have listed things that we might want to implement in the
// future. I have no idea which order we should implement these
// features. This is just a list of features that we might want to
// add. /zeffer

// ADD SUPPORT FOR MP
//
// this is to be done when we have implemented MP support in analysis
// module

// ADD SUPPORT FOR INCLUSION / EXCLUSION
//
// this is to be done if we figure out how to implement exclusion in
// the analysis module

// ADD SUPPORT FOR WRITE-THROUGH / WRITE-BACK
//
// this is to be done if we figure out how to implement write-through
// in the analysis module

class cachesim_base_t : public cachesim_request_interface_t,
			public cachesim_action_interface_t {
public:
    cachesim_base_t(fnr_cache_size_t size,
                    fnr_line_size_t line_size,
                    fnr_line_count_t assoc,
                    bool calc_footprint,
                    bool calc_spatial_usage);
    cachesim_base_t(fnr_cache_size_t size,
                    fnr_line_size_t line_size,
                    fnr_line_count_t assoc,
                    bool calc_footprint,
                    bool calc_spatial_usage,
		    string name);
    virtual ~cachesim_base_t();

private:
    void init();
    void rd_split(const cachesim_memop_t &memop);
    void rdx_split(const cachesim_memop_t &memop);

public:
    void                  rd(const cachesim_memop_t &memop);
    void                  rdx(const cachesim_memop_t &memop);
    void                  wb(const cachesim_memop_t &memop);
    void                  inv(const cachesim_memop_t &memop);
    void                  cb(const cachesim_memop_t &memop);
    void                  cb_inv(const cachesim_memop_t &memop);

    virtual void          dbgpr();

    void                  add_next(cachesim_request_interface_t *next);
    void                  add_prev(cachesim_action_interface_t *prev);

    void                  get_conf(cachesim_conf_t &conf);
    void                  get_counters(cachesim_result_t &res);
    void                  clear_counters();

protected:
    fnr_line_count_t      get_index(const cachesim_memop_t &memop);
    fnr_address_t         get_tag(const cachesim_memop_t &memop);
    cachesim_memop_t      get_addr_from_tag_and_index(fnr_line_count_t index,
                                                      fnr_address_t tag);
    void                  handle_victim(fnr_address_t tag,
                                        fnr_line_count_t index);
    cachesim_line_t      *insert_cache_line(cachesim_line_map_t *index_map,
                                            fnr_address_t tag,
                                            bool valid,  bool dirty);
    void                  update_used_bytes(const cachesim_memop_t *memop, cachesim_line_t *line);
    void                  handle_hit(const cachesim_memop_t *memop, cachesim_line_t *line);
    void                  handle_miss(const cachesim_memop_t *memop, cachesim_line_t *line);

    bool                  is_line_cold(const cachesim_memop_t &memop);
    fnr_access_size_t     are_bytes_cold(const cachesim_memop_t &memop);

    virtual fnr_address_t get_victim(const fnr_address_t &memop,
                                     fnr_line_count_t index) = 0;
    virtual void          replacement_update_after_miss(const fnr_address_t &memop, fnr_line_count_t index) = 0;
    virtual void          replacement_update_after_hit(const fnr_address_t &memop, fnr_line_count_t index) = 0;

    cachesim_conf_t       conf;

    fnr_line_count_t      indexes;
    int                   line_size_ln2;
    int                   tag_shift;

    cachesim_result_t     counters;
    string           name;

    cachesim_request_interface_t *next;	// next level towards the memory
    cachesim_action_interface_t  *prev;	// prev level towards the processor, change this one to a list when adding MP support

    // Vector indexed by set index with one map from tag to cache line per
    // set. Used to determine hit/miss.
    cachesim_line_map_vector_t v;

    // Sparse bitmap with one bit per cache line, used to calculate the
    // footprint with line granularity and to classify misses as cold or not.
    footprint_bitmap_t line_footprint;

    // Sparse bitmap with one bit per byte, used to calculate the footprint
    // with byte granularity.
    footprint_bitmap_t byte_footprint;
};

class cachesim_rand_t : public cachesim_base_t {
public:
    cachesim_rand_t(fnr_cache_size_t size,
                    fnr_line_size_t line_size,
                    fnr_line_count_t assoc,
                    bool calc_footprint,
                    bool calc_spatial_usage,
                    uint64_t seed);
    cachesim_rand_t(fnr_cache_size_t size,
                    fnr_line_size_t line_size,
                    fnr_line_count_t assoc,
                    bool calc_footprint,
                    bool calc_spatial_usage,
		    string name,
                    uint64_t seed);
    ~cachesim_rand_t();

private:
    void init();

    /// Deterministic random number generator.
    random_lcg_t *random_lcg;

public:

    void dbgpr();

protected:
    fnr_address_t get_victim(const fnr_address_t &memop,
                             fnr_line_count_t index);
    void          replacement_update_after_miss(const fnr_address_t &memop,
                                                fnr_line_count_t index);
    void          replacement_update_after_hit(const fnr_address_t &memop,
                                               fnr_line_count_t index);


    uint64_t                       seed; // seed for the random function
    vector<fnr_address_vector_t> rv; // random vector for fast victim search, one per index

private:
    fnr_address_t *victim_cache; // for internal use only
};

class cachesim_lru_t : public cachesim_base_t {
public:
    cachesim_lru_t(fnr_cache_size_t size,
                   fnr_line_size_t line_size,
                   fnr_line_count_t assoc,
                   bool calc_footprint,
                   bool calc_spatial_usage);
    cachesim_lru_t(fnr_cache_size_t size,
                   fnr_line_size_t line_size,
                   fnr_line_count_t assoc,
                   bool calc_footprint,
                   bool calc_spatial_usage,
		   string name);
private:
    void init();
public:

    void dbgpr();

protected:
    fnr_address_t get_victim(const fnr_address_t &memop,
                             fnr_line_count_t index);
    void          replacement_update_after_miss(const fnr_address_t &memop,
                                                fnr_line_count_t index);
    void          replacement_update_after_hit(const fnr_address_t &memop,
                                               fnr_line_count_t index);

    typedef vector<fnr_address_list_t> lru_vector_t;
    lru_vector_t ll; // lru list, one per index

    typedef map<fnr_address_t, fnr_address_list_t::iterator> address_map_t;
    vector<address_map_t> lm; // map to find entries in the lru list, one per index
};

class trace_to_cachesim_t : public fenris_flow_by_type_default_t {
public:
    trace_to_cachesim_t();
    ~trace_to_cachesim_t();

    flow_status_t read(const fnr_event_read_t &event);
    flow_status_t write(const fnr_event_write_t &event);

    void add_cache(cachesim_base_t *cache);
    void clear_cache_list();

protected:
    list<cachesim_base_t *> cache_list;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
