/// @file library/cache_sim/cachesim_base.cc

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#include "cachesim.hh"
#include "library/fenris/fnr_types.hh"

#include <iostream>
#include <iomanip>
#include <boost/format.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;
using namespace boost;

#define MIN(x, y) ((x) < (y) ? (x) : (y))

static bool
is_power_of_two(uint64_t val)
{
    return val != 0 && (val & (val - 1)) == 0;
}

static int
ln2(uint64_t val)
{
    assert(val != 0);

    int i = 1;

    while ((val >> i) != 0)
        i++;

    return i - 1;
}

cachesim_request_interface_t::~cachesim_request_interface_t()
{
}

cachesim_action_interface_t::~cachesim_action_interface_t()
{
}

cachesim_conf_t::cachesim_conf_t(fnr_cache_size_t size,
                                 fnr_line_size_t line_size,
                                 fnr_line_count_t assoc,
                                 bool calc_footprint,
                                 bool calc_spatial_usage)
    : size(size), line_size(line_size), assoc(assoc),
      calc_footprint(calc_footprint), calc_spatial_usage(calc_spatial_usage)
{
}

cachesim_conf_t::cachesim_conf_t()
    : size(0), line_size(0), assoc(0),
      calc_footprint(true), calc_spatial_usage(true)
{
}

cachesim_result_t::cachesim_result_t()
    : rd_total(0), rd_repl_misses(0), rd_cold_misses(0),
      wr_total(0), wr_repl_misses(0), wr_cold_misses(0),
      wb_incomming(0), wb_outgoing(0),
      fp_line(0), fp_byte(0)
{
}

cachesim_result_t::cachesim_result_t(
    uint64_t rd_total, uint64_t rd_repl_misses, uint64_t rd_cold_misses,
    uint64_t wr_total, uint64_t wr_repl_misses, uint64_t wr_cold_misses,
    uint64_t wb_incomming, uint64_t wb_outgoing,
    uint64_t fp_line, uint64_t fp_byte)
    : rd_total(rd_total), rd_repl_misses(rd_repl_misses), rd_cold_misses(rd_cold_misses),
      wr_total(wr_total), wr_repl_misses(wr_repl_misses), wr_cold_misses(wr_cold_misses),
      wb_incomming(wb_incomming), wb_outgoing(wb_outgoing),
      fp_line(fp_line), fp_byte(fp_byte)
{
}

void
cachesim_base_t::dbgpr()
{
    cout << name << ": size: " << conf.size << ", line-size: " << conf.line_size << ", assoc: " << conf.assoc << endl;
    cout << "indexes: " << indexes << ", line_size_ln2: " << line_size_ln2
	 << ", tag_shift: " << tag_shift << endl;

    cout << "rd_total: " << counters.rd_total
	 << ", rd_cold_misses: " << counters.rd_cold_misses
	 << ", rd_repl_misses: " << counters.rd_repl_misses << endl;
    cout << "wr_total: " << counters.wr_total
	 << ", wr_cold_misses: " << counters.wr_cold_misses
	 << ", wr_repl_misses: " << counters.wr_repl_misses << endl;
    cout << "wb_incomming: " << counters.wb_incomming
	 << ", wb_outgoing: " << counters.wb_outgoing << endl;

    cout << "index ";
    for (fnr_line_count_t i = 0; i<conf.assoc; i++) {
	if (i != 0)
	    cout << "               ";
	cout << "set" << i;
    }
    cout << endl;

    for (uint32_t i = 0; i<(uint32_t) indexes; i++) {
	cachesim_line_map_t &m = v[i];
	cachesim_line_map_t::const_iterator
	    map_iter = m.begin(),
	    map_end = m.end();

	cout << format("%1%") % io::group(setfill(' '), dec, setw(4), i);
	for (; map_iter != map_end; map_iter++) {
	    cout << format("  %1%:%2% v:%3% d:%4%")
		% io::group(setfill('0'), hex, setw(4), 0)
		% io::group(setfill('0'), hex, setw(8), map_iter->first.virtual_address)
		% io::group(dec, setw(1), map_iter->second.valid ? 1 : 0)
		% io::group(dec, setw(1), map_iter->second.dirty ? 1 : 0);
	}
	cout << endl;
    }
}

cachesim_base_t::~cachesim_base_t()
{
}

void
cachesim_base_t::init()
{
    // we assume powers of two for now
    assert(is_power_of_two(conf.size));
    assert(is_power_of_two(conf.line_size));
    assert(is_power_of_two(conf.assoc));

    indexes = (conf.size / conf.line_size) / conf.assoc;
    line_size_ln2 = ln2(conf.line_size);
    tag_shift = line_size_ln2 + ln2(indexes);

    v.reserve((uint32_t) indexes);
    for (uint32_t i = 0; i<(uint32_t) indexes; i++) {
	cachesim_line_map_t t;
	v.push_back(t);

	cachesim_line_map_t &m = v[i];
	fnr_address_t addr;

	for (fnr_line_count_t j = 0; j<conf.assoc; j++) {
	    addr.virtual_address = j;
            insert_cache_line(&m, addr, false, false);
	}
    }

    counters.used_bytes = 0;
}

cachesim_base_t::cachesim_base_t(fnr_cache_size_t size,
                                 fnr_line_size_t line_size,
                                 fnr_line_count_t assoc,
                                 bool calc_footprint,
                                 bool calc_spatial_usage,
                                 string name)
    : conf(size, line_size, assoc, calc_footprint, calc_spatial_usage),
      indexes(0), line_size_ln2(0), tag_shift(0),
      counters(), name(name), next(NULL), prev(NULL)
{
    init();
}

cachesim_base_t::cachesim_base_t(fnr_cache_size_t size,
                                 fnr_line_size_t line_size,
                                 fnr_line_count_t assoc,
                                 bool calc_footprint,
                                 bool calc_spatial_usage)
    : conf(size, line_size, assoc, calc_footprint, calc_spatial_usage),
      indexes(0), line_size_ln2(0), tag_shift(0),
      counters(), name(""), next(NULL), prev(NULL)
{
    init();
}

void
cachesim_base_t::add_next(cachesim_request_interface_t *next)
{
    this->next = next;
}

void
cachesim_base_t::add_prev(cachesim_action_interface_t *prev)
{
    this->prev = prev;
}

void
cachesim_base_t::get_conf(cachesim_conf_t &conf)
{
    conf.size = this->conf.size;
    conf.line_size = this->conf.line_size;
    conf.assoc = this->conf.assoc;
    conf.calc_footprint = this->conf.calc_footprint;
    conf.calc_spatial_usage = this->conf.calc_spatial_usage;
}

void
cachesim_base_t::get_counters(cachesim_result_t &res)
{
    res = counters;

    // Sum up the footprint from bitmaps.
    if (conf.calc_footprint) {
        for (footprint_bitmap_t::const_iterator
                 fbi = line_footprint.begin(), fbe = line_footprint.end();
             fbi != fbe; fbi++)
            res.fp_line += conf.line_size * fbi->second.count();
        for (footprint_bitmap_t::const_iterator
                 fbi = byte_footprint.begin(), fbe = byte_footprint.end();
             fbi != fbe; fbi++)
            res.fp_byte += fbi->second.count();
    }

    // Add used bytes from lines that have not been ejected.
    if (conf.calc_spatial_usage)
        for (cachesim_line_map_vector_t::const_iterator
                 i0 = v.begin(), e0 = v.end(); i0 != e0; i0++)
            for (cachesim_line_map_t::const_iterator
                     i1 = i0->begin(), e1 = i0->end(); i1 != e1; i1++)
                res.used_bytes += i1->second.used_bytes.count();
}

void
cachesim_base_t::clear_counters()
{
    counters.rd_total = 0;
    counters.rd_repl_misses = 0;
    counters.rd_cold_misses = 0;
    counters.wr_total = 0;
    counters.wr_repl_misses = 0;
    counters.wr_cold_misses = 0;
}

fnr_line_count_t
cachesim_base_t::get_index(const cachesim_memop_t &memop)
{
    uint64_t addr = memop.address.virtual_address;

    return ((addr >> line_size_ln2) & (indexes - 1));
}

fnr_address_t
cachesim_base_t::get_tag(const cachesim_memop_t &memop)
{
    fnr_address_t ret;

    ret.virtual_address = memop.address.virtual_address >> tag_shift;

    return ret;
}

cachesim_memop_t
cachesim_base_t::get_addr_from_tag_and_index(fnr_line_count_t index, fnr_address_t tag)
{
    cachesim_memop_t ret;

    // We're casting to fnr_access_size_t below, so we don't support line_sizes over 32bit.
    assert(conf.line_size <= 0xffffffffLL);

    ret.size = (fnr_access_size_t) conf.line_size;

    ret.address.virtual_address = (tag.virtual_address << tag_shift) | ((uint64_t) index << line_size_ln2);

    return ret;
}

// Returns whether the accessed line is cold and marks it as no longer cold.
bool
cachesim_base_t::is_line_cold(const cachesim_memop_t &memop)
{
    // Check that the access fits inside one cache line.
    assert(memop.address.virtual_address / conf.line_size ==
           (memop.address.virtual_address + memop.size - 1) / conf.line_size);

    fnr_line_count_t line = memop.address.virtual_address >> line_size_ln2;
    fnr_address_t chunk_index(line >> FOOTPRINT_CHUNK_LOG2);
    footprint_bitmap_chunk_t *chunk = &line_footprint[chunk_index];
    fnr_line_count_t index = line & FOOTPRINT_CHUNK_MASK;
    if (chunk->test((unsigned long)index))
        return 0;
    else {
        chunk->set((unsigned long)index);
        return 1;
    }
}

// Returns the number of accessed bytes that are cold and marks them as no
// longer cold.
fnr_access_size_t
cachesim_base_t::are_bytes_cold(const cachesim_memop_t &memop)
{
    uint64_t virtual_address = memop.address.virtual_address;

    // Check that the access fits inside one bitmap chunk.
    assert((virtual_address >> FOOTPRINT_CHUNK_LOG2) ==
           ((virtual_address + memop.size - 1) >> FOOTPRINT_CHUNK_LOG2));

    fnr_address_t chunk_index(virtual_address >> FOOTPRINT_CHUNK_LOG2);
    footprint_bitmap_chunk_t *chunk = &byte_footprint[chunk_index];
    fnr_access_size_t cold_count = 0;
    for (fnr_access_size_t ofs = 0; ofs < memop.size; ofs++) {
        uint64_t addr = (virtual_address + ofs) & FOOTPRINT_CHUNK_MASK;
	// warning C4244: 'argument' : conversion from 'fnr_line_count_t' to 'size_t', possible loss of data
        if (!chunk->test((unsigned long)addr)) {
            cold_count++;
            chunk->set((unsigned long)addr);
        }
    }
    return cold_count;
}

void
cachesim_base_t::handle_victim(fnr_address_t tag, fnr_line_count_t index)
{
    cachesim_line_map_t &m = v[(uint32_t) index];
    cachesim_line_map_t::iterator map_iter;

    // find victim (based on cache implementation and replacement policy)
    fnr_address_t victim = get_victim(tag, index);
    cachesim_memop_t victim_memop = get_addr_from_tag_and_index(index, victim);

    // check for prev levels in the cache hierarcy
    // we only model inclusion caches right now, hence, we have to invalidate
    if (prev)
	prev->inv(victim_memop);

    // get a iterator for the victim
    map_iter = m.find(victim);
    assert(map_iter != m.end());

    // Count the number of used bytes for the spatial usage calculation.
    if (map_iter->second.valid && conf.calc_spatial_usage)
        counters.used_bytes += map_iter->second.used_bytes.count();

    // check if the line is valid, dirty and if we have a next level
    if (map_iter->second.valid && map_iter->second.dirty) {
	counters.wb_outgoing += 1;
	if (next)
	    next->wb(victim_memop);
    }

    // remove the victim from the map
    m.erase(map_iter);
}

// Insert a cache line into the cache.
cachesim_line_t *
cachesim_base_t::insert_cache_line(cachesim_line_map_t *index_map,
                                   fnr_address_t tag, bool valid, bool dirty)
{
    cachesim_line_t line(valid, dirty,
                         conf.calc_spatial_usage ? conf.line_size : 0);
    (*index_map)[tag] = line;
    return &(*index_map)[tag];
}

// Update used bytes in the cache line for the spatial usage
// calculation.
void
cachesim_base_t::update_used_bytes(const cachesim_memop_t *memop, cachesim_line_t *line)
{
    fnr_line_size_t begin_offset = memop->address.virtual_address & ((1 << line_size_ln2) - 1);
    fnr_line_size_t end_offset = min(begin_offset + memop->size, (fnr_line_size_t)1 << line_size_ln2);

    assert(end_offset <= ((fnr_line_size_t)1 << line_size_ln2));

    for (fnr_line_size_t i = begin_offset; i < end_offset; i++)
        line->used_bytes.set((unsigned long)i, true);
}

void
cachesim_base_t::handle_hit(const cachesim_memop_t *memop, cachesim_line_t *line) {
    fnr_line_count_t index = get_index(*memop);
    fnr_address_t tag = get_tag(*memop);

    if (conf.calc_spatial_usage)
        update_used_bytes(memop, line);

    // update replacement priority (based on cache implementation and replacement policy)
    replacement_update_after_hit(tag, index);
}

void
cachesim_base_t::handle_miss(const cachesim_memop_t *memop, cachesim_line_t *line) {
    fnr_line_count_t index = get_index(*memop);
    fnr_address_t tag = get_tag(*memop);

    if (conf.calc_spatial_usage)
        update_used_bytes(memop, line);

    // update replacement priority (based on cache implementation and replacement policy)
    replacement_update_after_miss(tag, index);
}

// Read operation.
void
cachesim_base_t::rd_split(const cachesim_memop_t &memop)
{
    // Check that the access fits inside one cache line.
    assert(memop.address.virtual_address / conf.line_size ==
           (memop.address.virtual_address + memop.size - 1) / conf.line_size);

    fnr_line_count_t index = get_index(memop);
    fnr_address_t tag = get_tag(memop);
    cachesim_line_map_t &m = v[(uint32_t) index];
    cachesim_line_map_t::iterator map_iter;
    map_iter = m.find(tag);

    bool miss = true;
    if (map_iter != m.end() && map_iter->second.valid)
	miss = false;

    // Update the byte granularity footprint.
    if (conf.calc_footprint)
        are_bytes_cold(memop);

    counters.rd_total += 1;
    if (miss) {
	if (is_line_cold(memop))
	    counters.rd_cold_misses += 1;
        else
	    counters.rd_repl_misses += 1;

	handle_victim(tag, index);

	// read from next level in the cache hierarcy
	if (next)
	    next->rd(memop);

	// insert the new line into map
        cachesim_line_t *line = insert_cache_line(&m, tag, true, false);

        handle_miss(&memop, line);
    }
    else
        handle_hit(&memop, &map_iter->second);
}

// Write operation.
void
cachesim_base_t::rd(const cachesim_memop_t &memop)
{
    // Split reads that span multiple cache lines into multiple reads and call
    // rd_split for each.
    fnr_line_size_t line_size = conf.line_size;
    uint64_t access_begin = memop.address.virtual_address;
    uint64_t access_end = access_begin + memop.size;
    while (access_begin < access_end) {
        uint64_t next_access_begin =
            (access_begin + line_size) / line_size * line_size;
        cachesim_memop_t memop_split =
            {fnr_address_t(access_begin),
             (fnr_access_size_t)(MIN(access_end, next_access_begin) - access_begin)};
        rd_split(memop_split);
        if (next_access_begin < access_begin)
            break;
        access_begin = next_access_begin;
    }
}

void
cachesim_base_t::rdx_split(const cachesim_memop_t &memop)
{
    // Check that the access fits inside one cache line.
    assert(memop.address.virtual_address / conf.line_size ==
           (memop.address.virtual_address + memop.size - 1) / conf.line_size);

    fnr_line_count_t index = get_index(memop);
    fnr_address_t tag = get_tag(memop);
    cachesim_line_map_t &m = v[(uint32_t) index];
    cachesim_line_map_t::iterator map_iter;
    map_iter = m.find(tag);

    bool miss = true;
    if (map_iter != m.end() && map_iter->second.valid)
	miss = false;

    // Update the byte granularity footprint.
    if (conf.calc_footprint)
        are_bytes_cold(memop);

    counters.wr_total += 1;
    if (miss) {
	if (is_line_cold(memop))
	    counters.wr_cold_misses += 1;
        else
	    counters.wr_repl_misses += 1;

	handle_victim(tag, index);

	// read from next level in the cache hierarcy
	if (next)
	    next->rdx(memop);

	// insert the new line into map
        cachesim_line_t *line = insert_cache_line(&m, tag, true, true);

        handle_miss(&memop, line);
    } else {
	// make sure to set the dirty bit
	map_iter->second.dirty = 1;

        handle_hit(&memop, &map_iter->second);
    }
}

void
cachesim_base_t::rdx(const cachesim_memop_t &memop)
{
    // Split writes that span multiple cache lines into multiple writes and
    // call rdx_split for each.
    fnr_line_size_t line_size = conf.line_size;
    uint64_t access_begin = memop.address.virtual_address;
    uint64_t access_end = access_begin + memop.size;
    while (access_begin < access_end) {
        uint64_t next_access_begin =
            (access_begin + line_size) / line_size * line_size;
        cachesim_memop_t memop_split =
            {fnr_address_t(access_begin),
             (fnr_access_size_t)(MIN(access_end, next_access_begin) - access_begin)};
        rdx_split(memop_split);
        if (next_access_begin < access_begin)
            break;
        access_begin = next_access_begin;
    }
}

void
cachesim_base_t::wb(const cachesim_memop_t &memop)
{
    // TODO: How does this affect the footprint calculation?
    assert(!conf.calc_footprint);

    // TODO: How does this affect the spatial usage calculation?
    assert(!conf.calc_spatial_usage);

    // Check that the access fits inside one cache line.
    assert(memop.address.virtual_address / conf.line_size ==
           (memop.address.virtual_address + memop.size - 1) / conf.line_size);

    fnr_line_count_t index = get_index(memop);
    fnr_address_t tag = get_tag(memop);
    cachesim_line_map_t &m = v[(uint32_t) index];
    cachesim_line_map_t::iterator map_iter;
    map_iter = m.find(tag);

    bool miss = true;
    if (map_iter != m.end() && map_iter->second.valid)
	miss = false;

    counters.wb_incomming += 1;
    if (miss) {
	// this should not happen unless we have an exclusive cache
	assert(0); 		// for now
    } else {
	// make sure to set the dirty bit
	map_iter->second.dirty = 1;

	// update replacement priority (based on cache implementation and replacement policy)
	replacement_update_after_hit(tag, index);
    }
}

void
cachesim_base_t::inv(const cachesim_memop_t &memop)
{
    // TODO: How does this affect the footprint calculation?
    assert(!conf.calc_footprint);

    // TODO: How does this affect the spatial usage calculation?
    assert(!conf.calc_spatial_usage);

    // Check that the access fits inside one cache line.
    assert(memop.address.virtual_address / conf.line_size ==
           (memop.address.virtual_address + memop.size - 1) / conf.line_size);

    fnr_line_count_t index = get_index(memop);
    fnr_address_t tag = get_tag(memop);
    cachesim_line_map_t &m = v[(uint32_t) index];
    cachesim_line_map_t::iterator map_iter;
    map_iter = m.find(tag);

    // make sure to invalidate the line in this cache
    if (map_iter != m.end() && map_iter->second.valid)
	map_iter->second.valid = 0;

    // make sure to invalidate prev levels of the cache
    // make sure to change this code when adding support for multiple line sizes
    if (prev)
	prev->inv(memop);
}

void
cachesim_base_t::cb(const cachesim_memop_t &memop)
{
    // TODO: How does this affect the footprint calculation?
    assert(!conf.calc_footprint);

    // TODO: How does this affect the spatial usage calculation?
    assert(!conf.calc_spatial_usage);

    // Check that the access fits inside one cache line.
    assert(memop.address.virtual_address / conf.line_size ==
           (memop.address.virtual_address + memop.size - 1) / conf.line_size);

    assert(0);			// not implemented yet
}

void
cachesim_base_t::cb_inv(const cachesim_memop_t &memop)
{
    // TODO: How does this affect the footprint calculation?
    assert(!conf.calc_footprint);

    // TODO: How does this affect the spatial usage calculation?
    assert(!conf.calc_spatial_usage);

    // Check that the access fits inside one cache line.
    assert(memop.address.virtual_address / conf.line_size ==
           (memop.address.virtual_address + memop.size - 1) / conf.line_size);

    assert(0);			// not implemented yet
}

trace_to_cachesim_t::trace_to_cachesim_t()
{
}

trace_to_cachesim_t::~trace_to_cachesim_t()
{
}

void
trace_to_cachesim_t::add_cache(cachesim_base_t *cache)
{
    cache_list.push_back(cache);
}

void
trace_to_cachesim_t::clear_cache_list()
{
    cache_list.erase(cache_list.begin(), cache_list.end());
}

flow_status_t
trace_to_cachesim_t::read(const fnr_event_read_t &event)
{
    // Send the event to the first level of cache.

    // This is where we can implement a context_id splitter for the
    // multiprocessor cache simulator.

    cachesim_memop_t memop;
    list<cachesim_base_t *>::const_iterator list_iter = cache_list.begin();
    list<cachesim_base_t *>::const_iterator list_end = cache_list.end();

    memop.address = event.address;
    memop.size = event.size;

    for (; list_iter != list_end; list_iter++)
	(*list_iter)->rd(memop);

    return FLOW_OK;
}

flow_status_t
trace_to_cachesim_t::write(const fnr_event_write_t &event)
{
    // Send the event to the first level of cache.

    // This is where we can implement a context_id splitter for the
    // multiprocessor cache simulator.

    cachesim_memop_t memop;
    list<cachesim_base_t *>::const_iterator list_iter = cache_list.begin();
    list<cachesim_base_t *>::const_iterator list_end = cache_list.end();

    memop.address = event.address;
    memop.size = event.size;

    for (; list_iter != list_end; list_iter++)
	(*list_iter)->rdx(memop);

    return FLOW_OK;
}

END_NAMESPACE;
END_NAMESPACE;
