/// @file library/cache_sim/cachesim_lru.cc

// Copyright (c) 2006-2008, 2010 Acumem AB.  All rights reserved.

#include "cachesim.hh"

#include <iostream>
#include <iomanip>
#include <boost/format.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;
using namespace boost;

void
cachesim_lru_t::init()
{
    uint32_t indexes_ = (uint32_t) indexes;

    ll.reserve(indexes_);
    lm.reserve(indexes_);

    for (uint32_t i = 0; i<indexes; i++) {
	list<fnr_address_t> ll1;
	ll.push_back(ll1);
	list<fnr_address_t> &ll2 = ll[i];

	address_map_t lm1;
	lm.push_back(lm1);
	address_map_t &lm2 = lm[i];

	cachesim_line_map_t &m = v[i];
	cachesim_line_map_t::const_iterator map_iter;
	cachesim_line_map_t::const_iterator map_end;

	map_iter = m.begin();
	map_end = m.end();

	for (; map_iter != map_end; map_iter++) {
	    ll2.push_back(map_iter->first);
	    fnr_address_list_t::iterator list_iter = ll2.end();
	    list_iter--;
	    lm2[map_iter->first] = list_iter;
	}
    }
}

cachesim_lru_t::cachesim_lru_t(fnr_cache_size_t size,
                               fnr_line_size_t line_size,
                               fnr_line_count_t assoc,
                               bool calc_footprint,
                               bool calc_spatial_usage,
                               string name)
    : cachesim_base_t(
        size, line_size, assoc, calc_footprint, calc_spatial_usage, name)
{
    init();
}

cachesim_lru_t::cachesim_lru_t(fnr_cache_size_t size,
                               fnr_line_size_t line_size,
                               fnr_line_count_t assoc,
                               bool calc_footprint,
                               bool calc_spatial_usage)
    : cachesim_base_t(
        size, line_size, assoc, calc_footprint, calc_spatial_usage)
{
    init();
}

void
cachesim_lru_t::dbgpr()
{
    cachesim_base_t::dbgpr();

    cout << "index lru-list" << endl;
    for (uint32_t i = 0; i<(uint32_t)indexes; i++) {
	fnr_address_list_t &lru_list = ll[i];
	fnr_address_list_t::const_iterator list_iter = lru_list.begin();
	fnr_address_list_t::const_iterator list_end = lru_list.end();

	cout << format("%1%") % io::group(setfill(' '), dec, setw(4), i);
	for (; list_iter != list_end; list_iter++) {
	    cout << format("  %1%:%2%")
		% io::group(setfill('0'), hex, setw(4), 0/*list_iter->tag*/)
		% io::group(setfill('0'), hex, setw(8), list_iter->virtual_address);
	}
	cout << endl;
    }

    cout << "index lru-map" << endl;
    for (uint32_t i = 0; i<(uint32_t) indexes; i++) {
	address_map_t &lru_map = lm[i];
	address_map_t::const_iterator
	    map_iter = lru_map.begin(),
	    map_end = lru_map.end();

	cout << format("%1%") % io::group(setfill(' '), dec, setw(4), i);
	for (; map_iter != map_end; map_iter++) {
	    cout << format("  %1%:%2%:%3%:%4%")
		% io::group(setfill('0'), hex, setw(4), 0/*map_iter->first.tag*/)
		% io::group(setfill('0'), hex, setw(8), map_iter->first.virtual_address)
		% io::group(setfill('0'), hex, setw(4), 0/*map_iter->second->tag*/)
		% io::group(setfill('0'), hex, setw(8), map_iter->second->virtual_address);
	}
	cout << endl;
    }
}

fnr_address_t
cachesim_lru_t::get_victim(const fnr_address_t &memop, fnr_line_count_t index)
{
    list<fnr_address_t> &lru_list = ll[(uint32_t) index];
    list<fnr_address_t>::iterator iter = lru_list.begin();
    assert(iter != lru_list.end());
    return *iter;
}

void
cachesim_lru_t::replacement_update_after_miss(const fnr_address_t &memop, fnr_line_count_t index)
{
    address_map_t &lru_map = lm[(uint32_t) index];
    fnr_address_list_t &lru_list = ll[(uint32_t) index];

    // find victim in both list and map
    fnr_address_list_t::iterator lru_list_victim = lru_list.begin();
    address_map_t::iterator lru_map_victim;
    lru_map_victim = lru_map.find(*lru_list_victim);
    assert(lru_map_victim != lru_map.end());

    lru_list.erase(lru_list_victim); // remove the victim from the lru_list
    lru_map.erase(lru_map_victim); // remove the victim from the lru_map

    // insert new data in both list and map
    lru_list.push_back(memop);
    fnr_address_list_t::iterator lru_list_iter = lru_list.end();
    lru_list_iter--;
    lru_map[memop] = lru_list_iter;
}

void
cachesim_lru_t::replacement_update_after_hit(const fnr_address_t &memop, fnr_line_count_t index)
{
    address_map_t &lru_map = lm[(uint32_t) index];
    fnr_address_list_t &lru_list = ll[(uint32_t) index];

    // find the entry in the map
    address_map_t::iterator lru_iter;
    lru_iter = lru_map.find(memop);
    assert(lru_iter != lru_map.end());

    // remove the entry from the list and insert last
    fnr_address_list_t::iterator list_iter = lru_iter->second;
    lru_list.erase(list_iter);
    lru_list.push_back(memop);

    // get the list iterator and update the lru_map
    list_iter = lru_list.end();
    list_iter--;
    lru_iter->second = list_iter;
}

END_NAMESPACE;
END_NAMESPACE;
