/// @file library/codec/codec_utils.hh

// Copyright (c) 2010 Acumem AB.  All rights reserved.

#ifndef CODEC_UTILS_HH
#define CODEC_UTILS_HH

#ifdef __APPLE
#include <mach-o/dyld.h>
#endif

#include "library/codec/buffer.hh"
#include "library/lang/common_types.h"

#include <boost/foreach.hpp>

#ifndef _WIN32
#include <netinet/in.h>
#else
#include <winsock2.h>
#pragma comment(lib, "ws2_32")
#endif

#include <assert.h>
#include <limits.h>
#include <string.h>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

static inline uint64_t
htonll(uint64_t t)
{
#ifdef WORDS_BIGENDIAN
    return t;
#else
    return (((uint64_t)htonl((uint32_t) (t & 0xffffffff))) << 32) +
        htonl((uint32_t) (t >> 32));
#endif
}

static inline uint64_t
ntohll(uint64_t t)
{
    return htonll(t);
}

static inline uint16_t
to_big_endian_16(uint16_t i)
{
    return htons(i);
}

static inline uint16_t
from_big_endian_16(uint16_t i)
{
    return ntohs(i);
}

static inline uint32_t
to_big_endian_32(uint32_t i)
{
    return htonl(i);
}

static inline uint32_t
from_big_endian_32(uint32_t i)
{
    return ntohl(i);
}

static inline uint64_t
to_big_endian_64(uint64_t i)
{
    return htonll(i);
}

static inline uint64_t
from_big_endian_64(uint64_t i)
{
    return ntohll(i);
}

static inline void
binary_encode_uint8(codec_buffer_t *buffer, uint8_t i)
{
    buffer->write((char *)&i, 1);
}

static inline uint8_t
binary_decode_uint8(codec_buffer_t *buffer)
{
    return *(uint8_t *)buffer->read(sizeof(uint8_t));
}

static inline void
binary_encode_uint16(codec_buffer_t *buffer, uint16_t i)
{
    i = to_big_endian_16(i);
    buffer->write((char *)&i, 2);
}

static inline uint16_t
binary_decode_uint16(codec_buffer_t *buffer)
{
    return from_big_endian_16(*(uint16_t *)buffer->read(sizeof(uint16_t)));
}

static inline void
binary_encode_uint32(codec_buffer_t *buffer, uint32_t i)
{
    i = to_big_endian_32(i);
    buffer->write((char *)&i, 4);
}

static inline uint32_t
binary_decode_uint32(codec_buffer_t *buffer)
{
    return from_big_endian_32(*(uint32_t *)buffer->read(sizeof(uint32_t)));
}

static inline void
binary_encode_uint64(codec_buffer_t *buffer, uint64_t i)
{
    i = to_big_endian_64(i);
    buffer->write((char *)&i, 8);
}

static inline uint64_t
binary_decode_uint64(codec_buffer_t *buffer)
{
    return from_big_endian_64(*(uint64_t *)buffer->read(sizeof(uint64_t)));
}

static inline void
binary_encode_bool(codec_buffer_t *buffer, bool b)
{
    binary_encode_uint8(buffer, b ? 1 : 0);
}

static inline bool
binary_decode_bool(codec_buffer_t *buffer)
{
    return !!binary_decode_uint8(buffer);
}

static inline void
binary_encode_float(codec_buffer_t *buffer, float f)
{
    union {
        uint32_t i;
        float f;
    } b;
    b.f = f;
    b.i = to_big_endian_32(b.i);
    buffer->write((char *)&b.i, 4);
}

static inline float
binary_decode_float(codec_buffer_t *buffer)
{
    union {
        uint32_t i;
        float f;
    } b;
    b.i = from_big_endian_32(*(uint32_t *)buffer->read(sizeof(uint32_t)));
    return b.f;
}

static inline void
binary_encode_data16(codec_buffer_t *buffer,
                     const char *data,
                     size_t data_length)
{
    assert(data != NULL);
    assert(data_length <= UINT16_MAX);
    binary_encode_uint16(buffer, (uint16_t)data_length);
    buffer->write(data, data_length);
}

static inline void
binary_decode_data16(codec_buffer_t *buffer,
                     char **data,
                     size_t *data_length)
{
    *data_length = binary_decode_uint16(buffer);
    *data = new char[*data_length];
    memcpy(*data, buffer->read(*data_length), *data_length);
}

static inline void
binary_encode_data32(codec_buffer_t *buffer,
                     const char *data,
                     size_t data_length)
{
    assert(data != NULL);
    assert(data_length <= UINT32_MAX);
    binary_encode_uint32(buffer, (uint32_t)data_length);
    buffer->write(data, data_length);
}

static inline void
binary_decode_data32(codec_buffer_t *buffer,
                     char **data,
                     size_t *data_length)
{
    *data_length = binary_decode_uint32(buffer);
    *data = new char[*data_length];
    memcpy(*data, buffer->read(*data_length), *data_length);
}

static inline void
binary_encode_string16(codec_buffer_t *buffer,
                     const string &str)
{
    binary_encode_data16(buffer, str.c_str(), str.length() + 1);
}

static inline string
binary_decode_string16(codec_buffer_t *buffer)
{
    size_t cstr_length;
    char *cstr;
    binary_decode_data16(buffer, &cstr, &cstr_length);
    string str(cstr, cstr_length - 1);
    delete[] cstr;
    return str;
}

static inline void
binary_encode_string32(codec_buffer_t *buffer,
                     const string &str)
{
    binary_encode_data32(buffer, str.c_str(), str.length() + 1);
}

static inline string
binary_decode_string32(codec_buffer_t *buffer)
{
    size_t cstr_length;
    char *cstr;
    binary_decode_data32(buffer, &cstr, &cstr_length);
    string str(cstr, cstr_length - 1);
    delete[] cstr;
    return str;
}

static inline void
binary_encode_string16_vector(codec_buffer_t *buffer,
                              const vector<string> &strings)
{
    binary_encode_uint32(buffer, (uint32_t) strings.size());
    BOOST_FOREACH(const string &s, strings)
        binary_encode_string16(buffer, s.c_str());
}

static inline vector<string>
binary_decode_string16_vector(codec_buffer_t *buffer)
{
    vector<string> strings;
    uint32_t size = binary_decode_uint32(buffer);
    strings.reserve(size);
    for (uint32_t i = 0; i < size; i++)
        strings.push_back(binary_decode_string16(buffer));
    return strings;
}

static inline void
binary_encode_string32_vector(codec_buffer_t *buffer,
                              const vector<string> &strings)
{
    binary_encode_uint32(buffer, (uint32_t) strings.size());
    BOOST_FOREACH(const string &s, strings)
        binary_encode_string32(buffer, s.c_str());
}

static inline vector<string>
binary_decode_string32_vector(codec_buffer_t *buffer)
{
    vector<string> strings;
    uint32_t size = binary_decode_uint32(buffer);
    strings.reserve(size);
    for (uint32_t i = 0; i < size; i++)
        strings.push_back(binary_decode_string32(buffer));
    return strings;
}

END_NAMESPACE;
END_NAMESPACE;

#endif /* CODEC_UTILS_HH */
