/** @file library/control/control.c */

/* Copyright (c) 2007-2008 Acumem AB.  All rights reserved. */

#include <sys/types.h>
#ifndef _WIN32
#include <sys/socket.h>
#endif
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <assert.h>
#include <stdarg.h>

#include "library/am_mem/am_mem.h"
#include "control.h"

#ifdef _WIN32
#define va_copy(ap1, ap2) ((ap1) = (ap2))
#endif

#ifdef __linux__
/* On linux, use MSG_NOSIGNAL, MSG_DONTWAIT */
#define DEFAULT_SEND_FLAGS (MSG_NOSIGNAL)
#define DONT_WAIT_SEND_FLAGS (MSG_DONTWAIT)
#else
/* On Solaris and Windows, don't use MSG_NOSIGNAL */
#define DEFAULT_SEND_FLAGS (0)

#ifdef _WIN32
/* On Windows, don't use MSG_DONTWAIT */
#define DONT_WAIT_SEND_FLAGS (0)
#else
/* On Solaris, use MSG_DONTWAIT */
#define DONT_WAIT_SEND_FLAGS (MSG_DONTWAIT)
#endif
#endif

#define ENCODE(buf, offset, val, val_type) \
  *((val_type*)(&((char *)buf)[offset])) = val
#define ENCODE_DATA(buf, offset, val, val_type) \
  ENCODE(buf, (offset + CONTROL_PACKET_HEADER_LEN), val, val_type)

#define DECODE(buf, offset, val_type) \
  (*((val_type *)(&((char *)buf)[offset])))

#define SEND_HEAD(kind, size) \
  ssize_t ret; \
  int flags = DEFAULT_SEND_FLAGS | (block ? 0 : DONT_WAIT_SEND_FLAGS); \
  char *buf = am_malloc(CONTROL_PACKET_HEADER_LEN + (size)); \
  if (!buf) \
    return 0; \
  encode_header(buf, (kind), (size) + CONTROL_PACKET_HEADER_LEN);

#define SEND_TAIL(size) \
  ret = send(fd, buf, (size) + CONTROL_PACKET_HEADER_LEN, flags); \
  am_free(buf); \
  return ret;

typedef struct {
    am_control_kind_t kind;
    ssize_t (*decoder)(am_control_packet_t *packet, char *buf, ssize_t data_len);
} control_packet_decoder_t;


static const char* str_status[] = {
    // AM_Exit_Normal
    "No error",
    // AM_Exit_Usage,
    "Invalid command line option.",
    // AM_Exit_Licence_Invalid
    "Invalid license.",
    // AM_Exit_Licence_Expired
    "Licence has expired.",
    // AM_Exit_License_Exceeded,
    "License limitation exceeded.",
    // AM_Exit_OS_Unsupported
    "Unsupported operating system or kernel version.",
    // AM_Exit_PID
    "Attach failed, process ID does not exist.",
    // AM_Exit_Target_System
    "Attach failed, target process is a system process.",
    // AM_Exit_Target_Binary
    "Could not locate or read process binary file.",
    // AM_Exit_Target_Launch
    "Failed to launch target application.",
    // AM_Exit_Target_Attach
    "Failed to attach to target application.",
    // AM_Exit_Symbol
    "Failed do symbol lookup.",
    // AM_Exit_Unsupported_Arch,
    "Unsupported architecture in target application.",
    // AM_Exit_ACQ
    "Failed to create, or write to, acquisition info file.",
    // AM_Exit_Syscall
    "Unsupported syscall",
    // AM_Exit_Ctrl_Version
    "Control protocol version mismatch.",
    // AM_Exit_Ctrl_Protocol
    "Unhandled control protocol error.",
    // AM_Exit_Ctrl_Socket
    "Control socket error.",
    // AM_Exit_Unknown
    "An unexpected error has occured.",
    // AM_Exit_Max
    NULL
};


static ssize_t
decode_hello(am_control_packet_t *packet, char *buf, ssize_t data_len)
{
    if (data_len < 4)
	return -1;

    packet->u.hello.major = DECODE(buf, 0, uint16_t);
    packet->u.hello.minor = DECODE(buf, 2, uint16_t);

    return 4;
}

static ssize_t
decode_go(am_control_packet_t *packet, char *buf, ssize_t data_len)
{
    return 0;
}

static ssize_t
decode_detach(am_control_packet_t *packet, char *buf, ssize_t data_len)
{
    if(data_len < 1)
	return -1;

    packet->u.detach.type = DECODE(buf, 0, uint8_t);

    return 1;
}

static ssize_t
decode_hello_response(am_control_packet_t *packet, char *buf, ssize_t data_len)
{
    if (data_len < 4)
	return -1;

    packet->u.hello_response.major = DECODE(buf, 0, uint16_t);
    packet->u.hello_response.minor = DECODE(buf, 2, uint16_t);

    return 4;
}

static ssize_t
decode_message(am_control_packet_t *packet, char *buf, ssize_t data_len)
{
    int message_len = data_len - 1;

    if (data_len < 1)
	return -1;

    packet->u.message.level = DECODE(buf, 0, int8_t);
    // TODO: Handle errors here
    packet->u.message.message = am_malloc(message_len + 1);
    assert(packet->u.message.message != NULL);
    memcpy(packet->u.message.message, &buf[1], message_len);
    packet->u.message.message[message_len] = '\0';

    return data_len;
}

static ssize_t
decode_sampling_stopped(am_control_packet_t *packet, char *buf, ssize_t data_len)
{
    if(data_len < 1)
	return -1;

    packet->u.sampling_stopped.reason = DECODE(buf, 0, uint8_t);

    return 0;
}

static ssize_t
decode_exit(am_control_packet_t *packet, char *buf, ssize_t data_len)
{
    if(data_len < 2)
	return -1;

    packet->u.exit.status = DECODE(buf, 0, uint16_t);

    return 0;
}

static ssize_t
decode_protocol_mismatch(am_control_packet_t *packet, char *buf, ssize_t data_len)
{
    return 0;
}


static control_packet_decoder_t decoders[] = {
    { AM_Ctrl_Hello, &decode_hello },
    { AM_Ctrl_Go, &decode_go },
    { AM_Ctrl_Detach, &decode_detach },

    { AM_Ctrl_Hello_Response, &decode_hello_response },
    { AM_Ctrl_Message, &decode_message },
    { AM_Ctrl_Sampling_Stopped, &decode_sampling_stopped },
    { AM_Ctrl_Exit, &decode_exit },

    { AM_Ctrl_Protocol_Mismatch, &decode_protocol_mismatch },

    { 0, NULL }
};

static void
encode_header(char *buf, am_control_kind_t kind, uint32_t len)
{
    ENCODE(buf, 0, len, uint32_t);
    ENCODE(buf, 4, kind, uint16_t);
}


ssize_t
control_send_hello(int fd, int block, uint16_t major, uint16_t minor)
{
    SEND_HEAD(AM_Ctrl_Hello, 4);

    ENCODE_DATA(buf, 0, major, uint16_t);
    ENCODE_DATA(buf, 2, minor, uint16_t);

    SEND_TAIL(4);
}

ssize_t
control_send_go(int fd, int block)
{
    SEND_HEAD(AM_Ctrl_Go, 0);
    SEND_TAIL(0);
}

ssize_t
control_send_detach(int fd, int block, am_control_detach_type_t type)
{
    SEND_HEAD(AM_Ctrl_Detach, 1);

    ENCODE_DATA(buf, 0, type, uint8_t);

    SEND_TAIL(1);
}

ssize_t
control_send_hello_response(int fd, int block, uint16_t major, uint16_t minor)
{
    SEND_HEAD(AM_Ctrl_Hello_Response, 4);

    ENCODE_DATA(buf, 0, major, uint16_t);
    ENCODE_DATA(buf, 2, minor, uint16_t);

    SEND_TAIL(4);
}

ssize_t
control_send_message(int fd, int block, am_control_message_level_t level, char *message)
{
    int message_len = (int) strlen(message);
    SEND_HEAD(AM_Ctrl_Message, 1 + message_len);

    ENCODE_DATA(buf, 0, level, int8_t);
    memcpy(&buf[CONTROL_PACKET_HEADER_LEN + 1], message, message_len);

    SEND_TAIL(1 + message_len);
}

ssize_t
control_send_sampling_stopped(int fd, int block, am_control_sampling_stopped_reason_t reason)
{
    SEND_HEAD(AM_Ctrl_Sampling_Stopped, 1);

    ENCODE_DATA(buf, 0, reason, uint8_t);

    SEND_TAIL(1);
}

ssize_t
control_send_exit(int fd, int block, am_exit_status_t status)
{
    SEND_HEAD(AM_Ctrl_Exit, 2);

    ENCODE_DATA(buf, 0, status, uint16_t);

    SEND_TAIL(2);
}

ssize_t
control_send_protocol_mismatch(int fd, int block)
{
    SEND_HEAD(AM_Ctrl_Protocol_Mismatch, 0);
    SEND_TAIL(0);
}

ssize_t
control_printf(int fd, int block, am_control_message_level_t level, const char *format, ...)
{
    int ret;
    va_list ap;

    va_start(ap, format);
    ret = control_vprintf(fd, block, level, format, ap);
    va_end(ap);

    return ret;
}

ssize_t
control_vprintf(int fd, int block, am_control_message_level_t level, const char *format, va_list ap)
{
    ssize_t ret;
    int message_len;
    int flags = DEFAULT_SEND_FLAGS | (block ? 0 : DONT_WAIT_SEND_FLAGS);
    char *buf;
    va_list aq;

    /* Create a copy of the va_list to allow an additional invocation
     * of vsnprintf */
    va_copy(aq, ap);
    message_len = vsnprintf(NULL, 0, format, aq);
    va_end(aq);

    /* We need to allocate an extra byte for the zero termination from
     * vsnprintf. */
    buf = am_malloc(CONTROL_PACKET_HEADER_LEN + message_len + 2);
    if (!buf) {
	va_end(ap);
	return 0;
    }

    encode_header(buf, AM_Ctrl_Message, CONTROL_PACKET_HEADER_LEN + message_len + 1);
    ENCODE_DATA(buf, 0, level, int8_t);
    vsnprintf(&buf[CONTROL_PACKET_HEADER_LEN + 1], message_len + 1, format, ap);

    ret = send(fd, buf, CONTROL_PACKET_HEADER_LEN + message_len + 1, flags);
    am_free(buf);

    return ret;
}


ssize_t
control_decode_packet(am_control_packet_t *packet, void *buf, size_t buf_len)
{
    uint32_t packet_len;
    uint16_t packet_kind;
    int i;

    if (buf_len < CONTROL_PACKET_HEADER_LEN)
	return -1;

    packet_len = DECODE(buf, 0, uint32_t);
    packet_kind = DECODE(buf, 4, uint16_t);

    memset(packet, 0, sizeof(packet));
    packet->kind = packet_kind;
    for (i = 0; decoders[i].decoder; i++) {
	if (decoders[i].kind == packet_kind) {
	    int ret = decoders[i].decoder(packet, &((char *)buf)[6],
					  packet_len - CONTROL_PACKET_HEADER_LEN);
	    if (ret == -1)
		return -1;
	    else
		return ret + CONTROL_PACKET_HEADER_LEN;
	}
    }

    return 0;
}

ssize_t
control_receive_packet(int fd, int block, am_control_packet_t *packet)
{
    int ret;
    uint32_t message_len;
    void *message;
    const int flags = block ? MSG_WAITALL : DONT_WAIT_SEND_FLAGS;

    do {
	ret = recv(fd, (char *)&message_len, 4, flags | MSG_PEEK);
    } while (ret == -1 && errno == EINTR);
    if (ret == -1)
	return -1;
    else if (ret != 4)
	return -2;

    message = am_malloc(message_len);
    do {
	ret = recv(fd, message, message_len, flags);
    } while (ret == -1 && errno == EINTR);
    if (ret == -1)
	return -1;
    else if (ret != message_len)
	return -2;

    ret = control_decode_packet(packet, message, message_len);
    am_free(message);

    return ret;
}

void
control_free_packet_data(am_control_packet_t *packet)
{
    switch (packet->kind) {
    case AM_Ctrl_Message:
	if (packet->u.message.message)
	    am_free(packet->u.message.message);
	break;
    default:
	break;
    }
}

const char* control_exit_status_string(am_exit_status_t status)
{
    return str_status[status < AM_Exit_Max ? status : AM_Exit_Unknown];
}
