/** @file library/control/control.h */

/* Copyright (c) 2007-2008 Acumem AB.  All rights reserved. */

#ifndef CONTROL_H
#define CONTROL_H

#ifndef _WIN32
#include <stdint.h>
#include <unistd.h>
#else
#include <io.h>
#include <winsock2.h>
#endif


#include <stdarg.h>

#include "library/lang/common_macros.h"
#include "library/lang/common_types.h"

BEGIN_EXTERN_C;

#define AM_CTRL_MAJOR (1)
#define AM_CTRL_MINOR (0)

/**
 * \file
 *
 * ----------------------------------------------------------------------
 * Acumon control protocol version 1.0
 * ----------------------------------------------------------------------
 *
 * General packet format:
 *  0: unsigned 32 bit integer - packet length (including packet header)
 *  4: 2 byte - packet type
 *  6: Packet type specific data
 *
 * ----------------------------------------------------------------------
 * Protocol phases
 * ----------------------------------------------------------------------
 * 1. Handshake phase
 *
 * Valid packets: AM_Ctrl_Hello, AM_Ctrl_Hello_Response,
 *                AM_Ctrl_Protocol_Mismatch, AM_Ctrl_Exit
 *
 * A session starts with the controlling application issuing the
 * "Hello" message, the monitor must then respond with the "Hello
 * response" or the "Protocol mismatch" packet. The controlling
 * application may send a "Protocol mismatch" packet after receiving
 * the "Hello response" packet. Once the protocol mismatch message has
 * been sent, the connection should be considered invalid and be
 * closed.
 *
 * 2. Setup phase
 *
 * Valid packets: AM_Ctrl_Message, AM_Ctrl_Go, AM_Ctrl_Exit
 *
 * Phase largely unused at the moment. Sampling is not permitted to
 * start until the AM_Ctrl_Go message has been issued.
 *
 * 3. Sampling phase
 *
 * Valid packets: AM_Ctrl_Detach, AM_Ctrl_Message,
 *                AM_Ctrl_Sampling_Stopped, AM_Ctrl_Exit
 *
 * 4. Post sampling state
 * Valid packets: AM_Ctrl_Message, AM_Ctrl_Exit
 *
 * ----------------------------------------------------------------------
 * Packets
 * ----------------------------------------------------------------------
 *
 * (C) AM_Ctrl_Hello:
 *  0: unsigned 16 bit integer - major version
 *  2: unsigned 16 bit integer - minor version
 *
 * Sent from the controlling application to the monitor to initiate a
 * session.
 *
 * ----------------------------------------------------------------------
 * (M) AM_Ctrl_Hello_Response:
 *  0: unsigned 16 bit integer - major version
 *  2: unsigned 16 bit integer - minor version
 *
 * Sent from the monitor to the controlling application as an answer
 * to the "Hello" message.
 *
 * ----------------------------------------------------------------------
 * (C) AM_Ctrl_Go
 *
 * Starts sampling.
 *
 * ----------------------------------------------------------------------
 * (C) AM_Ctrl_Detach:
 *  0: 8 bit unsigned integer - Detach type
 *
 *  Detach type:
 *   0 - Normal, safe, detach
 *   1 - Force detach, but leave process in consistent state
 *   2 - Force detach now, no matter what.
 *
 * Message from controller to monitor, forces the monitor to detach
 * from the process monitored. The application will also be terminated
 * if it was started by the monitor.
 *
 * ----------------------------------------------------------------------
 * (M) AM_Ctrl_Message
 *  0: 8 bit signed integer - Message level
 *  1- : Message, not including terminating 0
 *
 * Message levels:
 *  -1   - Debug
 *   0   - Verbose
 *   10  - Info
 *   127 - Fatal errors
 *
 * Message sent from the monitor to the controlling application to
 * provide feedback to the user. Negative message levels are reserved
 * for debug messages. The message in the packet will not contain a
 * terminating 0, but the decoder will add this as a convenience.
 *
 * ----------------------------------------------------------------------
 * (M) AM_Ctrl_Sampling_Stopped
 *  0: 8 bit unsigned integer, reason.
 *
 * Reasons:
 *  0 - Application terminated normally
 *  1 - Detach requested
 *
 * Sent from the monitor when sampling stops.
 *
 * ----------------------------------------------------------------------
 * (M) AM_Ctrl_Exit
 *  0: 16 bit unsigned integer, exit status.
 *
 * Should be sent when acumon exits. Closing the socket without first
 * sending an exit message is considered to be an abnormal
 * termination. No further commnuication is allowed after sending
 * AM_Ctrl_Exit.
 *
 * ----------------------------------------------------------------------
 * (C,M) AM_Ctrl_Protocol_Mismatch
 *
 * Message sent from either the monitor or the controller to indicate
 * protocol mismatch.
 */

/**
 * Messsage kinds sent between the monitor and the controlling
 * application.
 */
typedef enum {
    /***********************************************************************
     * Messages originating from the controlling application
     ***********************************************************************/
    AM_Ctrl_Hello = 0,
    AM_Ctrl_Go,
    AM_Ctrl_Detach,

    /***********************************************************************
     * Messages originating from acumon
     ***********************************************************************/
    AM_Ctrl_Hello_Response = (1 << 14),
    AM_Ctrl_Message,
    AM_Ctrl_Sampling_Stopped,
    AM_Ctrl_Exit,

    /***********************************************************************
     * Messages originating from either end
     ***********************************************************************/
    AM_Ctrl_Protocol_Mismatch = (1 << 15)
} am_control_kind_t;

typedef enum {
    AM_Ctrl_Msg_Debug = -1,
    AM_Ctrl_Msg_Verbose = 0,
    AM_Ctrl_Msg_Info = 10,
    AM_Ctrl_Msg_Warn = 20,
    AM_Ctrl_Msg_Error = 127
} am_control_message_level_t;

typedef enum {
    AM_Ctrl_Detach_Normal = 0,
    AM_Ctrl_Detach_Force,
    AM_Ctrl_Detach_Force_Unsafe
} am_control_detach_type_t;

typedef enum {
    AM_Ctrl_Sampling_Stopped_Normal = 0,
    AM_Ctrl_Sampling_Stopped_Detached
} am_control_sampling_stopped_reason_t;

typedef enum {
    /// Normal exit
    AM_Exit_Normal = 0,
    /// Invalid parameter
    AM_Exit_Usage = 1,
    /// License invalid
    AM_Exit_License_Invalid,
    /// Licence expired
    AM_Exit_License_Expired,
    /// License limitation exceeded.
    AM_Exit_License_Exceeded,
    /// Unsupported OS or kernel version
    AM_Exit_OS_Unsupported,
    /// Process ID does not exist
    AM_Exit_PID,
    /// Target process is a system process
    AM_Exit_Target_System,
    /// Couldn't find or read process binary file.
    AM_Exit_Target_Binary,
    /// General launch failure
    AM_Exit_Target_Launch,
    /// General attach failure
    AM_Exit_Target_Attach,
    /// Symbol lookup failed
    AM_Exit_Symbol,
    /// Unsupported architecture, eg trying to sample 64 bit
    /// applications with 32 bit monitor.
    AM_Exit_Unsupported_Arch,
    /// Couldn't create, or write to, ACQ file
    AM_Exit_ACQ,
    /// Unsupported syscall
    AM_Exit_Syscall,
    /// Protocol mismatch in control protocol.
    AM_Exit_Ctrl_Version,
    /// Unhandled control protocol error.
    AM_Exit_Ctrl_Protocol,
    /// Control socket error, e.g. failed to open.
    AM_Exit_Ctrl_Socket,
    /// Error code used for uncategorized errors
    AM_Exit_Unknown,
    /// Place holder to get the highest defined error number.
    AM_Exit_Max
} am_exit_status_t;

typedef struct {
    am_control_kind_t kind;

    union {
	struct {
	    uint16_t major;
	    uint16_t minor;
	} hello;
	struct {
		char dummy; // C requires that a struct or union has at least one member
	} go;
	struct {
	    am_control_detach_type_t type;
	} detach;

	struct {
	    uint16_t major;
	    uint16_t minor;
	} hello_response;
	struct {
	    am_control_message_level_t level;
	    char *message;
	} message;
	struct {
	    am_control_sampling_stopped_reason_t reason;
	} sampling_stopped;
	struct {
	    am_exit_status_t status;
	} exit;

	struct {
		char dummy; // C requires that a struct or union has at least one member
	} protocol_mismatch;
    } u;
} am_control_packet_t;

#define CONTROL_PACKET_HEADER_LEN (6)

/* The send functions all return the number of characters sent. -1
 * indicates a send error, see send(2), a more detailed failure reason
 * is available in errno in such a case. A return value of 0 signifies
 * a local failure, such as not being able to allocate a buffer.
 *
 * Strictly positive return values indicate success.
 */
ssize_t control_send_hello(int fd, int block, uint16_t major, uint16_t minor);
ssize_t control_send_go(int fd, int block);
ssize_t control_send_detach(int fd, int block, am_control_detach_type_t type);

ssize_t control_send_hello_response(int fd, int block, uint16_t major, uint16_t minor);
ssize_t control_send_message(int fd, int block, am_control_message_level_t level, char *message);
ssize_t control_send_sampling_stopped(int fd, int block, am_control_sampling_stopped_reason_t reason);
ssize_t control_send_exit(int fd, int block, am_exit_status_t status);

ssize_t control_send_protocol_mismatch(int fd, int block);

/**
 * Packets that have been successfully decoded by
 * control_decode_packet have to be deallocated with
 * control_free_packet since some packet types contain variable length
 * data.
 *
 * Return values:
 * >0 - Number of bytes decoded.
 *  0 - Unknown packet
 * -1 - Buffer too small to hold the packet
 */
ssize_t control_decode_packet(am_control_packet_t *packet, void *buf, size_t buf_len);


/**
 * Return values:
 * >0 - Number of bytes decoded.
 *  0 - Unknown packet
 * -1 - IO Error, see errno
 * -2 - Other receive error, e.g. not enough data received.
 */
ssize_t control_receive_packet(int fd, int block, am_control_packet_t *packet);

/**
 * Note that this function will only free extra packet data allocated
 * by the decoder, not the packet as such!
 */
void control_free_packet_data(am_control_packet_t *packet);

/**
 * Convenience function for sending messages, uses the printf familly
 * internally.
 */
ssize_t control_printf(int fd, int block, am_control_message_level_t level, const char *format, ...);

ssize_t control_vprintf(int fd, int block, am_control_message_level_t level, const char *format, va_list ap);

/**
 * Returns a pointer to a string representation of the exit code.
 */
const char* control_exit_status_string(am_exit_status_t status);

END_EXTERN_C

#endif
