/// @file library/cpudb/cpudb.cc

// Copyright (c) 2008-2010,2012 Rogue Wave Software, Inc.  All rights reserved.

#include "cpudb.hh"

#include <string>

#include <boost/foreach.hpp>

#include "library/cpuid/cpuid.hh"
#include "library/cpuid/cpuid_intel.hh"
#include "library/cpuid/cpuid_amd.hh"

using namespace std;
using namespace boost;

#define GET_ATTR(node, name) \
    xml_utils_t::get_required_attr(node, "" # name, name)

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

static bool
by_cache_level(const cpudb_cache_cptr_t &lhs,
	       const cpudb_cache_cptr_t &rhs)
{
    return lhs->level < rhs->level;
}

cpudb_cache_t::cpudb_cache_t(const string &name,
			     unsigned int level,
			     unsigned int num_cores,
			     bool instructions,
			     bool data,
			     fnr_prefetch_classes_t active_prefetches,
                             bool active_hwpf,
			     fnr_cache_size_t size,
			     fnr_line_size_t line_size)
    : name(name),
      level(level),
      num_cores(num_cores),
      instructions(instructions), data(data),
      active_prefetches(active_prefetches),
      active_hwpf(active_hwpf),
      size(size), line_size(line_size)
{
}

cpudb_cache_t::cpudb_cache_t(xmlNodePtr node)
    : active_prefetches(FNR_Prefetch_Class_None),
      active_hwpf(true)
{
    if (!xmlStrEqual((const xmlChar*)"cache", node->name))
	throw xml_exception_t(node, "Invalid 'cache' element.");

    GET_ATTR(node, name);
    GET_ATTR(node, level);
    GET_ATTR(node, num_cores);
    GET_ATTR(node, instructions);
    GET_ATTR(node, data);
    GET_ATTR(node, size);
    GET_ATTR(node, line_size);

    for(xmlNodePtr n = node->children; n; n = n->next) {
	if(n->type == XML_ELEMENT_NODE) {
	    if(xmlStrEqual((const xmlChar*)"swpf", n->name)) {
		string id;
		GET_ATTR(n, id);

		fnr_prefetch_classes_t pf_class(
		    fnr_prefetch_class_by_name(id.c_str()));
		if(pf_class == FNR_Prefetch_Class_None)
		    throw xml_exception_t(n, "Unknown prefetch type '" + id + "'.");

		active_prefetches |= pf_class;
	    }
            else if(xmlStrEqual((const xmlChar*)"nohwpf", n->name))
                active_hwpf = false;
            else
		throw xml_exception_t(n, "Unexpected element in 'cache' node.");
	}
    }
}

cpudb_cache_t::cpudb_cache_t(const cpuid_cpu_info_t &cpu,
			     const cpuid_cache_level_t &cache)
    : level(cache.cache_level),
      num_cores(cache.threads_sharing * cpu.num_cores / cpu.num_threads),
      instructions(cache.instruction),
      data(cache.data),
      active_prefetches(FNR_Prefetch_Class_None),
      active_hwpf(true), // TODO: Correct?
      size(cache.size),
      line_size(cache.coherency_line_size)
{
    name = "L" + lexical_cast<string>(level);

    // TODO: Make sure that these defaults are sane.
    if (cpu.vendor == cpuid_cpu_intel_t::VENDOR_ID) {
	if (level == 1 && data)
	    active_prefetches =
		FNR_Prefetch_Class_Intel_NTA | FNR_Prefetch_Class_Intel_T0 |
		FNR_Prefetch_Class_Intel_T1 |  FNR_Prefetch_Class_Intel_T2;
    } else if (cpu.vendor == cpuid_cpu_amd_t::VENDOR_ID) {
	if (level == 1 && data)
	    active_prefetches =
		FNR_Prefetch_Class_Intel_NTA | FNR_Prefetch_Class_Intel_T0 |
		FNR_Prefetch_Class_Intel_T1 |  FNR_Prefetch_Class_Intel_T2 |
		FNR_Prefetch_Class_AMD | FNR_Prefetch_Class_AMD_W;
    } else
	active_prefetches = FNR_Prefetch_Class_Any;
}

cpudb_cache_t::cpudb_cache_t(const cpuid_cpu_info_t &cpu,
			     const cpuid_cache_level_t &cache,
			     const cpudb_cache_t &model_cache)
    : level(cache.cache_level),
      num_cores(cache.threads_sharing * cpu.num_cores / cpu.num_threads),
      instructions(cache.instruction),
      data(cache.data),
      active_prefetches(model_cache.active_prefetches),
      active_hwpf(model_cache.active_hwpf),
      size(cache.size),
      line_size(cache.coherency_line_size)
{
    name = "L" + lexical_cast<string>(level);
}

cpudb_cpu_t::cpudb_cpu_t(const string id,
			 const string &name,
			 unsigned int family,
			 unsigned int model,
			 unsigned int cores,
			 unsigned int threads,
			 bool prefetchw_dirty)
    : id(id),
      family(family),
      model(model),
      cores(cores),
      threads(threads),
      prefetchw_dirty(prefetchw_dirty),
      names(1, name)
{
}

cpudb_cpu_t::cpudb_cpu_t(xmlNodePtr node)
{
    if (!xmlStrEqual((const xmlChar*)"cpu", node->name))
	throw xml_exception_t(node, "Invalid 'cpu' element.");

    GET_ATTR(node, id);
    GET_ATTR(node, family);
    GET_ATTR(node, model);
    GET_ATTR(node, cores);
    GET_ATTR(node, threads);
    xml_utils_t::get_attr(node, "prefetchw_dirty", prefetchw_dirty, false);

    for(xmlNodePtr n = node->children; n; n = n->next) {
	if(n->type == XML_ELEMENT_NODE) {
	    if(xmlStrEqual((const xmlChar*)"name", n->name)) {
		string name(xml_utils_t::get_text(n));
		if (name.empty())
		    throw xml_exception_t(n, "CPU name must not be empty.");
		names.push_back(name);
	    } else if(xmlStrEqual((const xmlChar*)"cache", n->name)) {
		caches.push_back(
		    cpudb_cache_cptr_t(new cpudb_cache_t(n)));
	    } else
		throw xml_exception_t(n, "Unexpected element in 'cpu' node.");
	}
    }

    sort(caches.begin(), caches.end(), by_cache_level);
}

cpudb_cpu_t::cpudb_cpu_t(const cpuid_cpu_info_t &cpu,
			 const cpuid_cache_info_t &cache_info)
    : id("auto"),
      family(cpu.family),
      model(cpu.model),
      cores(cpu.num_cores),
      threads(cpu.num_threads / cpu.num_cores),
      names(1, cpu.brand_name)
{
    assert(threads >= 1);

    if(cpu.vendor == "AuthenticAMD")
	// TODO: Check if this is really the case
	prefetchw_dirty = (family < 0x10);
    else
	prefetchw_dirty = false;

    BOOST_FOREACH(const cpuid_cache_level_t &cache, cache_info.caches)
	caches.push_back(
	    cpudb_cache_cptr_t(new cpudb_cache_t(cpu, cache)));

    sort(caches.begin(), caches.end(), by_cache_level);
}

cpudb_cpu_t::cpudb_cpu_t(const cpuid_cpu_info_t &cpu,
			 const cpuid_cache_info_t &cache_info,
			 const cpudb_cpu_t &model_cpu)
    : id("auto"),
      family(cpu.family),
      model(cpu.model),
      cores(cpu.num_cores),
      threads(cpu.num_threads / cpu.num_cores),
      prefetchw_dirty(model_cpu.prefetchw_dirty),
      names(1, cpu.brand_name)
{
    assert(threads >= 1);

    BOOST_FOREACH(const cpuid_cache_level_t &cache, cache_info.caches) {
	if (cache.data) {
	    cpudb_cache_cptr_t model_cache(
		model_cpu.get_data_cache(cache.cache_level));
	    if (model_cache)
		caches.push_back(
		    cpudb_cache_cptr_t(
			new cpudb_cache_t(cpu, cache, *model_cache)));
	    else
		caches.push_back(
		    cpudb_cache_cptr_t(
			new cpudb_cache_t(cpu, cache)));
	}
    }

    sort(caches.begin(), caches.end(), by_cache_level);
}

cpudb_cache_cptr_t
cpudb_cpu_t::get_data_cache(unsigned int level) const
{
    BOOST_FOREACH(const cpudb_cache_cptr_t &cache, caches) {
	if (cache->data && cache->level == level)
	    return cache;
    }

    return cpudb_cache_cptr_t();
}

unsigned int
cpudb_cpu_t::num_cache_levels() const
{
    if (caches.empty())
	return 0;

    unsigned int levels(caches.back()->level);
    BOOST_FOREACH(const cpudb_cache_cptr_t &c, caches) {
	if (c->level > levels)
	    levels = c->level;
    }

    return levels;
}


cpudb_manufacturer_t::cpudb_manufacturer_t(xmlNodePtr node)
{
    if (!xmlStrEqual((const xmlChar*)"manufacturer", node->name))
	throw xml_exception_t(node, "Invalid 'manufacturer' node.");

    GET_ATTR(node, id);
    GET_ATTR(node, name);
    GET_ATTR(node, vendor_id);

    for(xmlNodePtr n = node->children; n; n = n->next) {
	if(n->type == XML_ELEMENT_NODE) {
	    if(xmlStrEqual((const xmlChar*)"cpu", n->name)) {
		cpudb_cpu_cptr_t cpu(new cpudb_cpu_t(n));
		try {
		    add_cpu(cpu);
		} catch (const cpudb_collision_t &) {
		    throw xml_exception_t(n,
					  "ID collision in cpu element with id='" +
					  cpu->id + "'.");
		}
	    } else
		throw xml_exception_t(n, "Unexpected element in 'manufacturer' node.");
	}
    }
}

cpudb_cpu_cptr_t
cpudb_manufacturer_t::cpu_by_name(const string &name) const
{
    string_cpu_map_t::const_iterator it(name_cpu_map.find(name));

    if (it != name_cpu_map.end())
	return it->second;
    else
	return cpudb_cpu_cptr_t();
}

cpudb_cpu_cptr_t
cpudb_manufacturer_t::cpu_by_id(const string &id) const
{
    string_cpu_map_t::const_iterator it(id_cpu_map.find(id));

    if (it != id_cpu_map.end())
	return it->second;
    else
	return cpudb_cpu_cptr_t();
}

cpudb_cpu_vector_t
cpudb_manufacturer_t::cpus_by_cpuid(unsigned int family, unsigned int model) const
{
    pair<cpuid_cpu_mmap_t::const_iterator,
	cpuid_cpu_mmap_t::const_iterator> range(
	    cpuid_cpu_mmap.equal_range(cpuid_t(family, model)));

    if (range.first != cpuid_cpu_mmap.end()) {
	cpudb_cpu_vector_t cpus;
	BOOST_FOREACH(const cpuid_cpu_mmap_t::value_type &v, range)
	    cpus.push_back(v.second);
	return cpus;
    } else
	return cpudb_cpu_vector_t();
}

void
cpudb_manufacturer_t::add_cpu(const cpudb_cpu_cptr_t &cpu)
{
    cpus.push_back(cpu);
    if (!id_cpu_map.insert(make_pair(cpu->id, cpu)).second)
	throw cpudb_collision_t();

    cpuid_cpu_mmap.insert(make_pair(
			      cpuid_t(cpu->family, cpu->model),
			      cpu));

    BOOST_FOREACH(const string &name, cpu->names)
	name_cpu_map.insert(make_pair(name, cpu));
}

cpudb_t::cpudb_t(warning_consumer_ptr_t warning_consumer)
  : warning_consumer(warning_consumer)
{
}

cpudb_t::~cpudb_t()
{
}

void
cpudb_t::load_from_file(const string &path)
{
    xml_doc_ptr_t doc(xml_utils_t::load_xml_file(path));
    xmlNodePtr root(xml_utils_t::get_root_element(doc));

    load(root);
}

void
cpudb_t::load(xmlNodePtr node)
{
    if (!xmlStrEqual((const xmlChar*)"cpudb", node->name))
	throw xml_exception_t(node, "Invalid 'cpudb' root node.");

    for(xmlNodePtr n = node->children; n; n = n->next) {
	if(n->type == XML_ELEMENT_NODE) {
	    if(xmlStrEqual((const xmlChar*)"manufacturer", n->name))
		 manufacturers.push_back(
		     cpudb_manufacturer_ptr_t(new cpudb_manufacturer_t(n)));
	    else
		throw xml_exception_t(n, "Unexpected element in 'cpudb' node.");
	}
    }

}

pair<cpudb_manufacturer_cptr_t, cpudb_cpu_cptr_t>
cpudb_t::get_cpu(const string &name) const
{
    size_t separator(name.find('/'));
    if (separator == string::npos ||
	separator + 1 >= name.size())
	return make_pair(cpudb_manufacturer_cptr_t(), cpudb_cpu_cptr_t());

    string
	mfgr_id(name, 0, separator),
	cpu_id(name, separator + 1);

    BOOST_FOREACH(const cpudb_manufacturer_cptr_t &mfgr, manufacturers) {
	if (mfgr->id == mfgr_id)
	    return make_pair(mfgr, mfgr->cpu_by_id(cpu_id));
    }

    return make_pair(cpudb_manufacturer_cptr_t(), cpudb_cpu_cptr_t());
}

cpudb_cpu_cptr_t
cpudb_t::autodetect_current_cpu(bool &autodetect_failed) const
{
    autodetect_failed = true;

    cpuid_cpu_cptr_t cpuid_cpu;
    try {
	cpuid_cpu = get_cpuid(warning_consumer);
	if (!cpuid_cpu) {
	    warning_consumer->warn(new generic_warning_t(
				       "Failed to autodetect the current CPU: "
				       "CPU vendor not recognized"));
	    return cpudb_cpu_cptr_t();
	}
    } catch (const cpuid_exception_t &e) {
	warning_consumer->warn(new generic_warning_t(
				   "Failed to autodetect the current CPU: "
				   + string(e.what())));
	return cpudb_cpu_cptr_t();
    }

    autodetect_failed = cpuid_cpu->get_autodetect_failed();

    cpuid_cpu_info_cptr_t cpu_info;
    cpuid_cache_info_cptr_t cache_info;

    try {
	cpu_info = cpuid_cpu->get_cpu_info();
	cache_info = cpuid_cpu->get_cache_info();
    } catch (const cpuid_exception_t &e) {
	warning_consumer->warn(new generic_warning_t(
				   "Failed to autodetect CPU info: "
				   + string(e.what())));
	return cpudb_cpu_cptr_t();
    }

    // Try to find a similar CPU in the CPU database using the vendor,
    // family and model from the autodetection.
    cpudb_cpu_cptr_t model_cpu(find_model_cpu(*cpu_info, *cache_info));

    // Create a CPU description using the data detected by cpuid and
    // the model. The model from the database will be used to describe
    // parameters that can't be probed using cpuid.
    if (model_cpu)
	return cpudb_cpu_cptr_t(new cpudb_cpu_t(*cpu_info,
						*cache_info,
						*model_cpu));
    else
	return cpudb_cpu_cptr_t(new cpudb_cpu_t(*cpu_info,
						*cache_info));
}

cpudb_manufacturer_cptr_t
cpudb_t::mfgr_by_cpuid(string vendor) const
{

    BOOST_FOREACH(const cpudb_manufacturer_cptr_t &mfgr, manufacturers) {
	if (mfgr->vendor_id == vendor)
	    return mfgr;
    }

    return cpudb_manufacturer_cptr_t();
}

cpudb_cpu_cptr_t
cpudb_t::find_model_cpu(const cpuid_cpu_info_t &cpu_info,
			const cpuid_cache_info_t &cache_info) const
{
    cpudb_manufacturer_cptr_t mfgr(mfgr_by_cpuid(cpu_info.vendor));
    if (!mfgr) {
	warning_consumer->warn(new generic_warning_t(
				   "Failed to autodetect a suitable model for "
				   "the running CPU. Some model parameters "
				   "may be inaccurate."));
	return cpudb_cpu_cptr_t();
    }

    cpudb_cpu_vector_t similar_cpus(
	mfgr->cpus_by_cpuid(cpu_info.family, cpu_info.model));
    if (similar_cpus.empty()) {
	warning_consumer->warn(new generic_warning_t(
				   "Failed to autodetect a suitable model for "
				   "the running CPU. Some model parameters "
				   "may be inaccurate."));
	return cpudb_cpu_cptr_t();
    }

    // Try to find a CPU with the same number of cache levels
    // and cores.
    const unsigned int cache_levels(cache_info.num_cache_levels());
    const unsigned int num_cores(cpu_info.num_cores);
    BOOST_FOREACH(const cpudb_cpu_cptr_t &cpu, similar_cpus) {
	if (cpu->num_cache_levels() == cache_levels &&
	    cpu->cores == num_cores)
	    return cpu;
    }

    // Use the first CPU with the same family and vendor as a
    // fallback.
    return similar_cpus.back();
}

END_NAMESPACE;
END_NAMESPACE;


