/// @file library/cpudb/cpudb.hh

// Copyright (c) 2008, 2010 Acumem AB.  All rights reserved.

#ifndef CPUDB_HH
#define CPUDB_HH

#include <map>
#include <vector>
#include <string>
#include <stdexcept>

#include <boost/shared_ptr.hpp>

#include "library/lang/common_macros.h"
#include "library/warning/warning_consumer.hh"
#include "library/fenris/fnr_types.hh"
#include "library/xml_utils/xml_utils.hh"
#include "library/cpuid/cpuid.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

using boost::shared_ptr;

struct cpudb_collision_t
    : public std::runtime_error
{
public:
    cpudb_collision_t()
	: runtime_error("cpudb collision")
	{}
};


/// Representation of a cache level for a processor defined in a cpu
/// database.
struct cpudb_cache_t
{
    cpudb_cache_t(const string &name,
		  unsigned int level,
		  unsigned int num_cores,
		  bool instructions,
		  bool data,
		  fnr_prefetch_classes_t active_prefetches,
                  bool active_hwpf,
		  fnr_cache_size_t size,
		  fnr_line_size_t line_size);

    cpudb_cache_t(xmlNodePtr node);

    cpudb_cache_t(const cpuid_cpu_info_t &cpu,
		  const cpuid_cache_level_t &cache);

    cpudb_cache_t(const cpuid_cpu_info_t &cpu,
		  const cpuid_cache_level_t &cache,
		  const cpudb_cache_t &model_cache);

    /// Human readable name of cache
    string name;

    /// Cache level
    unsigned int level;

    /// Number of cores sharing this cache
    unsigned int num_cores;

    /// True if the cache contains instructions
    bool instructions;

    /// True if the cache contains data
    bool data;

    /// Prefetches loading data into this cache.
    fnr_prefetch_classes_t active_prefetches;

    /// Hardware prefetching at this cache level?
    bool active_hwpf;

    /// Cache size in bytes
    fnr_cache_size_t size;

    /// Line size in bytes
    fnr_line_size_t line_size;
};

typedef shared_ptr<const cpudb_cache_t> cpudb_cache_cptr_t;
typedef vector<cpudb_cache_cptr_t> cpudb_cache_vector_t;

/// Representation of a processor in a cpu database.
struct cpudb_cpu_t
{
    cpudb_cpu_t(const string id,
		const string &name,
		unsigned int family,
		unsigned int model,
		unsigned int cores,
		unsigned int threads,
		bool prefetchw_dirty);

    cpudb_cpu_t(xmlNodePtr node);

    cpudb_cpu_t(const cpuid_cpu_info_t &cpu,
		const cpuid_cache_info_t &cache_info);

    cpudb_cpu_t(const cpuid_cpu_info_t &cpu,
		const cpuid_cache_info_t &cache_info,
		const cpudb_cpu_t &model_cpu);

    /// Get the cache structure representing a specific cache level.
    ///
    /// \return A pointer to the cache level structure or NULL on
    ///         error.
    cpudb_cache_cptr_t get_data_cache(unsigned int level) const;

    /// Return the number of available cache levels.
    unsigned int num_cache_levels() const;

    /// Identifier, used to specify the CPU from the command line.
    string id;

    /// Family id (from CPUID)
    unsigned int family;

    /// Model number (from CPUID)
    unsigned int model;

    /// Number of cores
    unsigned int cores;

    /// Threads per core
    unsigned int threads;

    /// True if prefetchw causes the cache line to be marked as dirty.
    bool prefetchw_dirty;

    /// Product name (one chip may be sold under several different
    /// names).
    vector<string> names;

    /// Vector of caches in this CPU, sorted by cache level.
    cpudb_cache_vector_t caches;
};

typedef shared_ptr<const cpudb_cpu_t> cpudb_cpu_cptr_t;
typedef shared_ptr<cpudb_cpu_t> cpudb_cpu_ptr_t;
typedef vector<cpudb_cpu_cptr_t> cpudb_cpu_vector_t;


/// Representation of a cpu manufacturer and all its
/// processors. Provides functions for looking up cpu models by name
/// and id.
class cpudb_manufacturer_t
{
public:
    cpudb_manufacturer_t(xmlNodePtr node);

    /// Identifier, used to specify the CPU from the command line.
    string id;

    /// Name of the manufacturer
    string name;

    /// The vendor ID as returned by CPUID, e.g. 'GenuineIntel',
    /// 'AuthenticAMD'.
    string vendor_id;

    /// @{
    /// Look up processor by name, id or family/model
    cpudb_cpu_cptr_t cpu_by_name(const string &name) const;
    cpudb_cpu_cptr_t cpu_by_id(const string &id) const;
    cpudb_cpu_vector_t cpus_by_cpuid(unsigned int family, unsigned int model) const;
    /// @}

    /// Add a CPU to the manufacturer entry.
    ///
    /// \throw cpudb_collision_t On ID or name collisions.
    void add_cpu(const cpudb_cpu_cptr_t &cpu);

    /// Returns all processors for this manufacturer
    const cpudb_cpu_vector_t &get_cpus() const {
	return cpus;
    }

private:
    struct cpuid_t
    {
	cpuid_t(unsigned int family, unsigned int model)
	    : family(family), model(model) {}

	bool operator<(const cpuid_t &rhs) const {
	    if (family == rhs.family)
		return  model < rhs.model;
	    else
		return family < rhs.family;
	}

	unsigned int family;
	unsigned int model;
    };

    typedef map<string, cpudb_cpu_cptr_t> string_cpu_map_t;
    typedef multimap<cpuid_t, cpudb_cpu_cptr_t> cpuid_cpu_mmap_t;

    /// Vector of CPUs by this manufacturer.
    cpudb_cpu_vector_t cpus;

    /// Map of CPUs with name as key, multiple names may map to the
    /// same CPU struct.
    string_cpu_map_t name_cpu_map;

    /// Map of CPUs with the CPU id as key.
    string_cpu_map_t id_cpu_map;

    /// Map of CPUs with the CPU id as key.
    cpuid_cpu_mmap_t cpuid_cpu_mmap;
};

typedef shared_ptr<cpudb_manufacturer_t> cpudb_manufacturer_ptr_t;
typedef shared_ptr<const cpudb_manufacturer_t> cpudb_manufacturer_cptr_t;

typedef vector<cpudb_manufacturer_ptr_t> cpudb_manufacturer_vector_t;


/// Representation of a cpu database. Provides lookup functions into
/// this database.
class cpudb_t
{
public:
    cpudb_t(warning_consumer_ptr_t warning_consumer);
    ~cpudb_t();

    void load_from_file(const string &path);
    void load(xmlNodePtr node);

    /// Get a CPU by name, where name is on the form "mfgr/id".
    ///
    /// \return NULL on error, a pointer to a structure representing
    ///         the CPU otherwise.
    pair<cpudb_manufacturer_cptr_t, cpudb_cpu_cptr_t>
    get_cpu(const string &name) const;

    /// Try to autodetect the properties of the current CPU, matching
    /// similar CPUs if necessary.
    ///
    /// \return NULL on error, a pointer to the cpudb_cpu_t
    ///         structure describing the CPU otherwise.
    cpudb_cpu_cptr_t autodetect_current_cpu(bool &autodetect_fail) const;

    /// Get the manufacturer with the specified vendor ID.
    ///
    /// \return NULL on error, pointer to the manufacturer otherwise.
    cpudb_manufacturer_cptr_t mfgr_by_cpuid(string vendor) const;

    cpudb_manufacturer_vector_t manufacturers;

private:
    cpudb_cpu_cptr_t find_model_cpu(const cpuid_cpu_info_t &cpu_info,
				    const cpuid_cache_info_t &cache_info) const;

    /// Pointer to the warning consumer. Should never be null.
    warning_consumer_ptr_t warning_consumer;
};

typedef shared_ptr<cpudb_t> cpudb_ptr_t;
typedef shared_ptr<const cpudb_t> cpudb_cptr_t;

END_NAMESPACE;
END_NAMESPACE;

#endif

