/// @file library/cpuid/cpuid.cc

// Copyright (c) 2007-2009 Acumem AB.  All rights reserved.

#include "cpuid_utils.hh"
#include "cpuid_intel.hh"
#include "cpuid_amd.hh"
#include "library/lang/am_assert.hh"

#include <boost/format.hpp>
#include <boost/foreach.hpp>

#include <sstream>
#include <set>

using namespace boost;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

unsigned int
cpuid_cache_info_t::num_cache_levels() const
{
    if (caches.empty())
	return 0;

    return caches.back().cache_level;
}

// ----------------------------------------------------------------------

void
cpuid_cpu_t::stupid_extra_constructor(warning_consumer_ptr_t warning_consumer)
{
    collect_apic_ids(warning_consumer);
    cpu_info = query_cpu_info();
    cache_info = query_cache_info();
}

cpuid_cpu_t::cpuid_cpu_t()
    : autodetect_failed(false)
{
    // Get the cpuid limit and the real vendor name.
    cpuid_regs_t regs_std(cpuid_eax(0x00000000));
    union {
        char chars[13];
        int regs[3];
    } v;
    cpuid_limit = (unsigned int)regs_std.eax;
    v.regs[0] = regs_std.ebx;
    v.regs[1] = regs_std.edx;
    v.regs[2] = regs_std.ecx;
    v.chars[12] = '\0';
    vendor = v.chars;

    // Get the extended cpuid limit.
    cpuid_regs_t regs_ext(cpuid_eax(0x80000000));
    cpuid_elimit = (unsigned int)regs_ext.eax;
}

void
cpuid_cpu_t::collect_apic_ids(warning_consumer_ptr_t warning_consumer)
{
    collect_apic_ids_impl();

    unsigned int thread_bits = get_thread_bits();
    unsigned int core_bits = get_core_bits();

    // Check that there are no duplicated APIC ids.
    for (cpuid_cpu_nr_t i = 0; i < apic_ids.size(); i++)
        for (cpuid_cpu_nr_t j = i + 1; j < apic_ids.size(); j++)
            if (apic_ids[i] == apic_ids[j]) {
                stringstream ss;
                ss << "Failed to auto-detect processors, duplicate APIC id "
                   << format("%1%") % io::group(setfill('0'), hex, setw(8), apic_ids[i])
                   << ". Assuming " << apic_ids.size() << " single-core, single-threaded "
                   << "processors.";
                warning_consumer->info(ss.str());
                fail_autodetect(warning_consumer);
                return;
            }

    // Collect number of threads per core and package.
    typedef map<unsigned int, unsigned int> c_t;
    c_t core_threads;
    c_t package_threads;
    BOOST_FOREACH(unsigned int apic_id, apic_ids) {
        core_threads[apic_id >> (thread_bits - core_bits)]++;
        package_threads[apic_id >> thread_bits]++;
    }

    // Check that each core has the same number of threads.
    unsigned int threads_per_core = core_threads.begin()->second;
    BOOST_FOREACH(const c_t::value_type &v, core_threads)
        if (v.second != threads_per_core) {
            stringstream ss;
            ss << "Failed to auto-detect processors, the number of threads "
               << "differs between cores, " << v.second << " vs "
               << threads_per_core << ". Assuming " << apic_ids.size()
               << " single-core, single-threaded processors.";
            warning_consumer->info(ss.str());
            fail_autodetect(warning_consumer);
            return;
        }

    // Check that each package has the same number of threads.
    unsigned int threads_per_package = package_threads.begin()->second;
    BOOST_FOREACH(const c_t::value_type &v, package_threads)
        if (v.second != threads_per_package) {
            stringstream ss;
            ss << "Failed to auto-detect processors, the number of threads "
               << "differs between packages, " << v.second << " vs "
               << threads_per_package << ". Assuming " << apic_ids.size()
               << " single-core, single-threaded processors.";
            warning_consumer->info(ss.str());
            fail_autodetect(warning_consumer);
            return;
        }

    // Check that the number of threads per package evenly divides the number
    // of OS processors.
    if (apic_ids.size() % threads_per_package != 0) {
            stringstream ss;
            ss << "Failed to auto-detect processors, the number of threads "
               << "per package does not evenly divide operating system "
               << "logical processors, " << apic_ids.size() << " / "
               << threads_per_package << ". Assuming " << apic_ids.size()
               << " single-core, single-threaded processors.";
            warning_consumer->info(ss.str());
            fail_autodetect(warning_consumer);
            return;
    }
}

cpuid_cpu_t::~cpuid_cpu_t()
{
}

string
cpuid_cpu_t::get_vendor() const
{
    return vendor;
}

cpuid_cpu_info_cptr_t
cpuid_cpu_t::get_cpu_info() const
{
    return cpu_info;
}

cpuid_cache_info_cptr_t
cpuid_cpu_t::get_cache_info() const
{
    return cache_info;
}

#define SAVE_STRING(s, regs)			\
    *((int*)(&(s)[0])) = (regs.eax);		\
    *((int*)(&(s)[4])) = (regs.ebx);		\
    *((int*)(&(s)[8])) = (regs.ecx);		\
    *((int*)(&(s)[12])) = (regs.edx);

string
cpuid_cpu_t::query_brand_name() const
{
    char brand[49];
    memset(brand, '\0', sizeof(brand));

    char *chunk(&brand[0]);
    for (unsigned int id = 0x80000002U;
	 id <= 0x80000004U && id <= cpuid_elimit;
	 id++) {
	cpuid_regs_t regs(cpuid_eax(id));
	SAVE_STRING(chunk, regs);
	chunk += 16;
    }

    return normalize_space(brand);
}

#undef SAVE_STRING

#ifdef _WIN32
bool isblank(char c) {
  return c == ' ' || c == '\t';
}
#endif

string
cpuid_cpu_t::normalize_space(const string &s)
{
    if (s.empty())
	return "";

    ostringstream ss;
    bool last_blank(isblank(s[0]));
    if (!last_blank)
	ss << s[0];

    for (string::size_type i = 1; i < s.length(); i++) {
	const bool this_blank(isblank(s[i]));

	if (!this_blank || !last_blank)
	    ss << s[i];

	last_blank = this_blank;
    }

    string ret(ss.str());

    if (!ret.empty() && isblank(*ret.rbegin()))
	return string(ret, 0, ret.length() - 1);
    else
	return ret;
}

unsigned int
cpuid_cpu_t::threads_per_package() const
{
    const unsigned int apic_mask((unsigned int)-1 << get_thread_bits());

    const cpuid_apic_id_t ref(apic_ids.front() & apic_mask);

    unsigned int num_threads(0);
    BOOST_FOREACH(cpuid_apic_id_t apic_id, apic_ids)
        if ((apic_id & apic_mask) == ref)
            num_threads++;

    return num_threads;
}

unsigned int
cpuid_cpu_t::cores_per_package() const
{
    const unsigned int apic_mask(
        ((unsigned int)-1 << get_thread_bits()) |
        ((1U << (get_thread_bits() - get_core_bits())) - 1));

    const cpuid_apic_id_t ref(apic_ids.front() & apic_mask);

    unsigned int num_cores(0);
    BOOST_FOREACH(cpuid_apic_id_t apic_id, apic_ids)
        if ((apic_id & apic_mask) == ref)
            num_cores++;

    return num_cores;
}

unsigned int
cpuid_cpu_t::threads_sharing(const unsigned int apic_mask) const
{
    const cpuid_apic_id_t ref(apic_ids.front() & apic_mask);

    unsigned int num_cores(0);
    BOOST_FOREACH(cpuid_apic_id_t apic_id, apic_ids)
        if ((apic_id & apic_mask) == ref)
            num_cores++;

    return num_cores;
}

unsigned int
cpuid_cpu_t::num_packages() const
{
    const unsigned int apic_mask((unsigned int)-1 << get_thread_bits());

    set<unsigned int> package_ids;
    BOOST_FOREACH(cpuid_apic_id_t apic_id, apic_ids)
        package_ids.insert(apic_id & apic_mask);

    return (unsigned int)package_ids.size();
}

const apic_id_vector_t &
cpuid_cpu_t::get_apic_ids() const
{
    return apic_ids;
}

void
cpuid_cpu_t::fail_autodetect(warning_consumer_ptr_t warning_consumer)
{
    autodetect_failed = true;
    warning_consumer->info("Processor auto-detection failure is often caused "
                           "by running in a virtualized environment. It is "
                           "recommended that you manually specify the "
                           "processor model and number of processors you want "
                           "to do the analysis for.");
    for (unsigned int c = 0; c < apic_ids.size(); c++)
        apic_ids[c] = c << get_thread_bits();
}

// ----------------------------------------------------------------------------

static cpuid_cpu_cptr_t cpuid;

cpuid_cpu_cptr_t
get_cpuid(warning_consumer_ptr_t warning_consumer)
{
    if (!cpuid) {
        string vendor(get_vendor_id());
        if (vendor == cpuid_cpu_intel_t::VENDOR_ID)
            cpuid = cpuid_cpu_cptr_t(new cpuid_cpu_intel_t(warning_consumer));
        if (vendor == cpuid_cpu_amd_t::VENDOR_ID)
            cpuid = cpuid_cpu_cptr_t(new cpuid_cpu_amd_t(warning_consumer));
    }
    return cpuid;
}

END_NAMESPACE;
END_NAMESPACE;
