/// @file library/cpuid/cpuid.hh

// Copyright (c) 2007-2009 Acumem AB.  All rights reserved.

#ifndef CPUID_HH
#define CPUID_HH

#include "library/warning/warning_consumer.hh"

#include <boost/shared_ptr.hpp>

#include <vector>
#include <string>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

using boost::shared_ptr;

typedef unsigned int cpuid_cpu_nr_t;

typedef unsigned int cpuid_apic_id_t;

typedef std::vector<cpuid_apic_id_t> apic_id_vector_t;

// ----------------------------------------------------------------------
// Exceptions
// ----------------------------------------------------------------------

struct cpuid_exception_t
    : public std::exception
{
};

struct cpuid_cpu_inactive_exception_t
    : public cpuid_exception_t
{
    char const* what() const throw() {
	return "CPUID query failed, CPU inactive.";
    }
};

struct cpuid_cpu_general_exception_t
    : public cpuid_exception_t
{
    ~cpuid_cpu_general_exception_t() throw() {}

    cpuid_cpu_general_exception_t(const char *where, int error) {
	char buf[256];
	error_string = "CPUID query failed. ";
	error_string += where;
	error_string += ": ";
	strerror_r(error, buf, sizeof buf);
	error_string += buf;
	error_string += ".";
    }
    char const* what() const throw() {
	return error_string.c_str();
    }

    std::string error_string;
};

// ----- Utility functions. ---------------------------------------------------

/// Extract the bit start to bit stop, inclusive, from val.
static inline unsigned int
bit_range(unsigned int val,
          unsigned int start,
          unsigned int stop)
{
    return (val >> start) & ((2 << (stop - start)) - 1);
}

/// Return the 2 logarithm of val rounded up.
static inline unsigned int
log2_ceil(const unsigned int val)
{
    unsigned int i;
    for (i = 0; val > (1U << i); i++)
	;
    return i;
}

// ----------------------------------------------------------------------
// CPU model information
// ----------------------------------------------------------------------

/// Information about a processor, but not its cache hieararchy.
struct cpuid_cpu_info_t {
    cpuid_cpu_info_t(const string &vendor,
		     int family, int model, int stepping,
		     const string &brand_name,
                     unsigned int num_threads,
                     unsigned int num_cores)
	: vendor(vendor),
	  family(family), model(model), stepping(stepping),
	  brand_name(brand_name),
          num_threads(num_threads),
          num_cores(num_cores)
        {}

    /// Vendor ID, e.g. GenuineIntel.
    string vendor;

    int family;
    int model;
    int stepping;

    /// Processor brand string as returned by Intel and AMD for
    /// CPUID.0x8000000[2-4].
    string brand_name;

    /// Threads per package.
    unsigned int num_threads;

    /// Cores per package.
    unsigned int num_cores;
};

typedef shared_ptr<cpuid_cpu_info_t> cpuid_cpu_info_ptr_t;
typedef shared_ptr<const cpuid_cpu_info_t> cpuid_cpu_info_cptr_t;


// ----------------------------------------------------------------------
// CPU cache information
// ----------------------------------------------------------------------

/// Information for a cache level.
struct cpuid_cache_level_t {
    cpuid_cache_level_t(int cache_level,
			bool data,
                        bool instruction,
			int ways_of_associativity,
			int physical_line_partition,
			int coherency_line_size,
			int size,
                        unsigned int threads_sharing)
	: cache_level(cache_level),
	  data(data),
          instruction(instruction),
	  ways_of_associativity(ways_of_associativity),
	  physical_line_partition(physical_line_partition),
	  coherency_line_size(coherency_line_size),
	  size(size),
          threads_sharing(threads_sharing)
        {}

    int cache_level;

    bool data;
    bool instruction;

    int ways_of_associativity;
    int physical_line_partition;
    int coherency_line_size;

    int size;

    unsigned int threads_sharing;
};

typedef vector<cpuid_cache_level_t> cpuid_cache_vector_t;


/// Collection of information for a hiearachy of caches.
struct cpuid_cache_info_t {
    unsigned int num_cache_levels() const;

    /// List of caches available for the CPU, sorted by cache level in
    /// ascending order.
    cpuid_cache_vector_t caches;
};

typedef shared_ptr<cpuid_cache_info_t> cpuid_cache_info_ptr_t;
typedef shared_ptr<const cpuid_cache_info_t> cpuid_cache_info_cptr_t;


// ----- cpuid_t --------------------------------------------------------------

/// A cpuid_t instance that can be queried for CPU information.
class cpuid_cpu_t {
public:
    /// Get the vendor ID of the CPU.
    string get_vendor() const;

    /// Returns CPU information.
    cpuid_cpu_info_cptr_t get_cpu_info() const;

    /// Returns cache information.
    cpuid_cache_info_cptr_t get_cache_info() const;

    unsigned int num_packages() const;

    bool get_autodetect_failed() const {
        return autodetect_failed;
    }

    /// Number of bits in the APIC ID that identify a thread within a package.
    virtual unsigned int get_thread_bits() const = 0;
    /// Number of bits in the APIC ID that identify a core within a package.
    virtual unsigned int get_core_bits() const = 0;
    /// Vector with the APIC IDs of all threads.
    const apic_id_vector_t &get_apic_ids() const;

protected:
    cpuid_cpu_t();
    virtual ~cpuid_cpu_t();

    /// Do stuff we would like to do in the constructor, but can't because C++
    /// sucks. Called back from the constructors of sub-classes.
    void stupid_extra_constructor(warning_consumer_ptr_t warning_consumer);

    /// Collect APIC IDs from all logical processors. Check that the APIC IDs
    /// are consistent. If not, print a warning and generate fake IDs.
    void collect_apic_ids(warning_consumer_ptr_t warning_consumer);

    /// Collect APIC IDs from all logical processors.
    virtual void collect_apic_ids_impl() = 0;

    /// Manufacturer-specific CPU information query.
    virtual cpuid_cpu_info_cptr_t query_cpu_info() const = 0;

    /// Manufacturer-specific cache information query.
    virtual cpuid_cache_info_cptr_t query_cache_info() const = 0;

    /// This function seems to be common between AMD and Intel.
    string query_brand_name() const;

    /// Utility function to remove duplicate white space characters and
    /// removes all blanks at the ends. Used by query_brand_name() to
    /// tidy up processor names.
    static string normalize_space(const string &s);

    unsigned int threads_per_package() const;
    unsigned int cores_per_package() const;
    unsigned int threads_sharing(const unsigned int apic_mask) const;

    /// APIC IDs from all logical processors.
    apic_id_vector_t apic_ids;

    /// Largest available CPUID function.
    unsigned int cpuid_limit;

    /// Largest available extended CPUID function.
    unsigned int cpuid_elimit;

    /// Vendor as detected by CPUID 0.
    string vendor;

private:
    /// CPU info.
    cpuid_cpu_info_cptr_t cpu_info;

    /// Cache info.
    cpuid_cache_info_cptr_t cache_info;

    /// True if the CPU auto detection failed.
    bool autodetect_failed;

    /// Called if we fail to autodetect the processor. Sets the
    /// autodetect_failed flag and creates faked APIC ids.
    void fail_autodetect(warning_consumer_ptr_t warning_consumer);
};

typedef shared_ptr<const cpuid_cpu_t> cpuid_cpu_cptr_t;

// ----------------------------------------------------------------------------

/// Get a cpuid_cpu_t instance.
cpuid_cpu_cptr_t get_cpuid(warning_consumer_ptr_t warning_consumer);

END_NAMESPACE;
END_NAMESPACE;

#endif
