/// @file library/cpuid/cpuid_amd.cc

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#include "library/cpuid/cpuid_amd.hh"
#include "library/cpuid/cpuid_utils.hh"
#include "library/lang/am_assert.hh"

#include <boost/foreach.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

static int l23_associativity_table[] = {
     0,  //0x0
     1,  //0x1
     2,  //0x2
     0,  //0x3
     4,  //0x4
     0,  //0x5
     8,  //0x6
     0,  //0x7
    16,  //0x8
     0,  //0x9
    32,  //0xA
    48,  //0xB
    64,  //0xC
    96,  //0xD
   128,  //0xE
     0,  //0xF
};

const char cpuid_cpu_amd_t::VENDOR_ID[] = "AuthenticAMD";


cpuid_cpu_amd_t::cpuid_cpu_amd_t(warning_consumer_ptr_t warning_consumer)
{
    stupid_extra_constructor(warning_consumer);
}

unsigned int
cpuid_cpu_amd_t::get_max_cores() const
{
    if(cpuid_elimit >= 0x80000008) {
	cpuid_regs_t regs(cpuid_eax(0x80000008));
	const unsigned int mnc(bit_range(regs.ecx, 12, 15));
	const unsigned int nc(bit_range(regs.ecx, 0, 7) + 1);
	return mnc > 0 ? (1 << mnc) : nc;
    } else
	return 1;
}

void
cpuid_cpu_amd_t::collect_apic_ids_impl() {
    // Collect all APIC ids.
    cpuid_cpu_nr_t max_cpus(cpuid_os_t::get_cpu_max());
    AM_ASSERT(max_cpus > 0);
    for (cpuid_cpu_nr_t c = 0; c < max_cpus; c++)
	if (cpuid_os_t::cpu_online(c)) {
            cpuid_regs_t regs(cpuid_cpu_eax(c, 0x00000001));
            apic_ids.push_back(bit_range(regs.ebx, 24, 31));
        }
}

cpuid_cpu_info_cptr_t
cpuid_cpu_amd_t::query_cpu_info() const
{
    const cpuid_regs_t model_info(cpuid_eax(0x00000001));

    const int stepping(bit_range(model_info.eax, 0, 3));
    const int base_family(bit_range(model_info.eax, 8, 11));
    const int ext_family(bit_range(model_info.eax, 20, 27));
    const int family(base_family +
		     (base_family == 0xF ? ext_family : 0));
    const int base_model(bit_range(model_info.eax, 4, 7));
    const int ext_model(bit_range(model_info.eax, 16, 19));
    const int model(base_model |
		    (base_family == 0xF ? (ext_model << 4) : 0));
    const string brand_name(query_brand_name());

    return cpuid_cpu_info_cptr_t(
	new cpuid_cpu_info_t(vendor,
			     family,
                             model,
                             stepping,
			     brand_name,
                             threads_per_package(),
                             cores_per_package()));
}

cpuid_cache_level_t
cpuid_cpu_amd_t::get_l1_cache(
    const int reg,
    const cpuid_apic_id_t owner_apic_id,
    const unsigned int apic_mask,
    const bool data,
    const bool instructions) const
{
    return cpuid_cache_level_t(
	1, // cache_level
	data,
	instructions,
	bit_range(reg, 16, 23), // ways_of_associativity
	bit_range(reg, 8, 15), // physical_line_partition
	bit_range(reg, 0, 7), // coherency_line_size
	bit_range(reg, 24, 31) * 1024,
        threads_sharing(apic_mask));
}

static bool
l23_cache_enabled(int reg)
{
    return bit_range(reg, 12, 15) != 0;
}

cpuid_cache_level_t
cpuid_cpu_amd_t::get_l2_cache(
    const int reg,
    const cpuid_apic_id_t owner_apic_id,
    const int apic_mask) const
{
    assert(l23_cache_enabled(reg));

    return cpuid_cache_level_t(
	2, // cache_level
	true, // data
	true, //instructions
	l23_associativity_table[bit_range(reg, 12, 15)], // ways_of_associativity
	bit_range(reg, 8, 11), // physical_line_partition
	bit_range(reg, 0, 7), // coherency_line_size
	bit_range(reg, 16, 31) * 1024,
        threads_sharing(apic_mask));
}

cpuid_cache_level_t
cpuid_cpu_amd_t::get_l3_cache(
    const int reg,
    const cpuid_apic_id_t owner_apic_id,
    const unsigned int apic_mask) const
{
    assert(l23_cache_enabled(reg));

    return cpuid_cache_level_t(
	3, // cache_level
	true, // data
	true, //instructions
	l23_associativity_table[bit_range(reg, 12, 15)], // ways_of_associativity
	bit_range(reg, 8, 11), // physical_line_partition (TODO: Check if this is right)
	bit_range(reg, 0, 7), // coherency_line_size
	bit_range(reg, 18, 31) * 1024 * 512,
        threads_sharing(apic_mask));
}

cpuid_cache_info_cptr_t
cpuid_cpu_amd_t::query_cache_info() const
{
    cpuid_cache_info_ptr_t cache_info(new cpuid_cache_info_t());

    // There is currently no way to know how the logical CPUs within a
    // physical CPU share caches on AMD processors. We make the
    // following assumptions:
    //  * L1 and L2 are core private, but may be shared between logical
    //    threads on the same core.
    //  * L3 is shared among all threads in the same physical CPU.

    const cpuid_regs_t model_info(cpuid_eax(0x00000001));
    const int apic_id(bit_range(model_info.ebx, 24, 31));
    const unsigned int core_bits(log2_ceil(get_max_cores()));

    if(cpuid_elimit >= 0x80000005) {
	const cpuid_regs_t regs(cpuid_eax(0x80000005));

	// L1 data cache
	cache_info->caches.push_back(
	    get_l1_cache(regs.ecx, apic_id, (unsigned)-1, true, false));

	// L1 instruction cache
	cache_info->caches.push_back(
	    get_l1_cache(regs.edx, apic_id, (unsigned)-1, false, true));
    }

    if(cpuid_elimit >= 0x80000006) {
	const cpuid_regs_t regs(cpuid_eax(0x80000006));

	// L2 cache
	if (l23_cache_enabled(regs.ecx))
	    cache_info->caches.push_back(
		get_l2_cache(regs.ecx, apic_id, (unsigned)-1));

	// L3 cache
	if (l23_cache_enabled(regs.edx))
	    cache_info->caches.push_back(
		get_l3_cache(regs.edx, apic_id, ~((1 << core_bits) - 1)));
    }

    return cache_info;
}

unsigned int
cpuid_cpu_amd_t::get_thread_bits() const
{
    return get_core_bits();
}

unsigned int
cpuid_cpu_amd_t::get_core_bits() const
{
    if(cpuid_limit >= 4) {
        return log2_ceil(get_max_cores());
    }
    else
        return 0;
}

END_NAMESPACE;
END_NAMESPACE;
