/// @file library/cpuid/cpuid_intel.cc

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#include "library/cpuid/cpuid_intel.hh"
#include "library/cpuid/cpuid_utils.hh"
#include "library/lang/am_assert.hh"

#include <boost/foreach.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

struct legacy_cache_descriptor_t {
    int id;
    int level;
    bool data;
    bool instruction;
    int ways_of_associativity;
    int line_size;
    int size;
};

static legacy_cache_descriptor_t legacy_cache_table[] = {
    { 0x06, 1, false,  true,  4, 32, 8   * 1024 },
    { 0x08, 1, false,  true,  4, 32, 16  * 1024 },
    { 0x0A, 1,  true, false,  2, 32, 8   * 1024 },
    { 0x0C, 1,  true, false,  4, 32, 16  * 1024 },

    { 0x22, 3,  true,  true,  4, 64, 512 * 1024 },
    { 0x23, 3,  true,  true,  8, 64, 1   * 1024 * 1024 },
    { 0x25, 3,  true,  true,  8, 64, 2   * 1024 * 1024 },
    { 0x29, 3,  true,  true,  8, 64, 4   * 1024 * 1024 },
    { 0x2C, 1,  true, false,  8, 64, 32  * 1024 },

    { 0x30, 1, false,  true,  8, 64, 32  * 1024 },
    { 0x39, 2,  true,  true,  4, 64, 128 * 1024 },
    { 0x3A, 2,  true,  true,  6, 64, 192 * 1024 },
    { 0x3B, 2,  true,  true,  2, 64, 128 * 1024 },
    { 0x3C, 2,  true,  true,  4, 64, 256 * 1024 },
    { 0x3D, 2,  true,  true,  6, 64, 384 * 1024 },
    { 0x3E, 2,  true,  true,  4, 64, 512 * 1024 },

    { 0x41, 2,  true,  true,  4, 32, 128 * 1024 },
    { 0x42, 2,  true,  true,  4, 32, 256 * 1024 },
    { 0x43, 2,  true,  true,  4, 32, 512 * 1024 },
    { 0x44, 2,  true,  true,  4, 32, 1   * 1024 * 1024 },
    { 0x45, 2,  true,  true,  4, 32, 2   * 1024 * 1024 },
    { 0x46, 3,  true,  true,  4, 64, 4   * 1024 * 1024 },
    { 0x47, 3,  true,  true,  8, 64, 8   * 1024 * 1024 },
    // TODO: We need to check for Intel Xeon MP Family F model 6 and
    // report level 3 instead in that case.
    { 0x49, 2,  true,  true, 16, 64, 4   * 1024 * 1024 },
    { 0x4A, 3,  true,  true, 12, 64, 6   * 1024 * 1024 },
    { 0x4B, 3,  true,  true, 16, 64, 8   * 1024 * 1024 },
    { 0x4C, 3,  true,  true, 12, 64, 12  * 1024 * 1024 },
    { 0x4D, 3,  true,  true, 16, 64, 16  * 1024 * 1024 },

    { 0x60, 1,  true, false,  8, 64, 16  * 1024 },
    { 0x66, 1,  true, false,  4, 64, 8   * 1024 },
    { 0x67, 1,  true, false,  4, 64, 16  * 1024 },
    { 0x68, 1,  true, false,  4, 64, 32  * 1024 },

    { 0x78, 2,  true,  true,  4, 64, 1   * 1024 * 1024 },
    { 0x79, 2,  true,  true,  8, 64, 128 * 1024 },
    { 0x7A, 2,  true,  true,  8, 64, 256 * 1024 },
    { 0x7B, 2,  true,  true,  8, 64, 512 * 1024 },
    { 0x7C, 2,  true,  true,  8, 64, 1   * 1024 * 1024 },
    { 0x7D, 2,  true,  true,  8, 64, 2   * 1024 * 1024 },
    { 0x7F, 2,  true,  true,  2, 64, 512 * 1024 },

    { 0x82, 2,  true,  true,  8, 32, 256 * 1024 },
    { 0x83, 2,  true,  true,  8, 32, 512 * 1024 },
    { 0x84, 2,  true,  true,  8, 32, 1   * 1024 * 1024 },
    { 0x85, 2,  true,  true,  8, 32, 2   * 1024 * 1024 },
    { 0x86, 2,  true,  true,  4, 64, 512 * 1024 },
    { 0x87, 2,  true,  true,  8, 64, 1   * 1024 * 1024 },

    { 0x00, 0, false, false,  0,  0, 0}
};

const char cpuid_cpu_intel_t::VENDOR_ID[] = "GenuineIntel";

cpuid_cpu_intel_t::cpuid_cpu_intel_t(warning_consumer_ptr_t warning_consumer)
    : supports_x2apic(false)
{
    if (cpuid_limit >= 0xb) {
        cpuid_regs_t regs(cpuid_eax_ecx(0xb, 0));
        if (regs.ebx != 0)
            supports_x2apic = true;
    }

    stupid_extra_constructor(warning_consumer);
}

void
cpuid_cpu_intel_t::collect_apic_ids_impl() {
    cpuid_cpu_nr_t max_cpus(cpuid_os_t::get_cpu_max());
    AM_ASSERT(max_cpus > 0);
    for (cpuid_cpu_nr_t c = 0; c < max_cpus; c++)
	if (cpuid_os_t::cpu_online(c)) {
            if (supports_x2apic) {
                cpuid_regs_t regs(cpuid_cpu_eax_ecx(c, 0xb, 0));
                apic_ids.push_back(regs.edx);
            }
            else {
                cpuid_regs_t regs(cpuid_cpu_eax(c, 0x00000001));
                apic_ids.push_back(bit_range(regs.ebx, 24, 31));
            }
        }
}

cpuid_cpu_info_cptr_t
cpuid_cpu_intel_t::query_cpu_info() const
{
    const cpuid_regs_t model_info(cpuid_eax(0x00000001));

    /* According to application note 485, it should be ok to add the
     * extended parts of the family and model unconditionally. This
     * seems to be in conflict with the instruction reference that
     * states that we have to check the family number.
     */
    const int basic_family(bit_range(model_info.eax, 8, 11));
    const int family(basic_family +
		     (basic_family == 0xF ? bit_range(model_info.eax, 20, 27) : 0));
    const int model(bit_range(model_info.eax, 4, 7) +
		    ((basic_family == 0xF || basic_family == 0x6) ?
		     bit_range(model_info.eax, 16, 19) << 4 : 0));
    const int stepping(bit_range(model_info.eax, 0, 3));
    const string brand_name(query_brand_name());

    return cpuid_cpu_info_cptr_t(
	new cpuid_cpu_info_t(vendor,
			     family,
                             model,
                             stepping,
			     brand_name,
                             threads_per_package(),
                             cores_per_package()));
}

static bool
cache_by_level(const cpuid_cache_level_t &lhs, const cpuid_cache_level_t &rhs)
{
    return lhs.cache_level < rhs.cache_level;
}

cpuid_cache_info_cptr_t
cpuid_cpu_intel_t::query_cache_info() const
{
    cpuid_cache_info_ptr_t ret;

    const cpuid_regs_t model_info(cpuid_eax(0x00000001));
    const int apic_id(bit_range(model_info.ebx, 24, 31));

    if(cpuid_limit >= 4) {
	ret = query_cache_info_deterministic(apic_id);
    } else if (cpuid_limit >= 2) {
	ret = query_cache_info_legacy(apic_id);
    } else
	return cpuid_cache_info_cptr_t();

    sort(ret->caches.begin(), ret->caches.end(), cache_by_level);
    return ret;
}

cpuid_cache_info_ptr_t
cpuid_cpu_intel_t::query_cache_info_deterministic(
    const cpuid_apic_id_t owner_apic_id) const
{
    cpuid_cache_info_ptr_t cache_info(new cpuid_cache_info_t());

    for(int i = 0; ; i++) {
	const cpuid_regs_t regs(cpuid_eax_ecx(0x00000004, i));
	const int cache_type(bit_range(regs.eax, 0, 4));

	if(cache_type == 0)
	    break;
	else if(cache_type < 4) {
	    const int ways_of_associativity(bit_range(regs.ebx, 22, 31) + 1);
	    const int coherency_line_size(bit_range(regs.ebx, 0, 11) + 1);
	    const int physical_line_partition(bit_range(regs.ebx, 12, 21) + 1);
	    const int num_sets(regs.ecx + 1);
	    const int size(num_sets
			   * physical_line_partition
			   * ways_of_associativity
			   * coherency_line_size);

	    const unsigned int max_threads(bit_range(regs.eax, 14, 25) + 1);
	    const unsigned int apic_mask(~((1 << log2_ceil(max_threads)) - 1));

	    cache_info->caches.push_back(
		cpuid_cache_level_t(
		    bit_range(regs.eax, 5, 7), // cache_level
		    (cache_type & 1) != 0, // data
                    (cache_type & 2) != 0, // instructions
		    ways_of_associativity,
		    physical_line_partition,
		    coherency_line_size,
		    size,
                    threads_sharing(apic_mask)));
	}
    }

    return cache_info;
}

void
cpuid_cpu_intel_t::add_legacy_cache(cpuid_cache_vector_t &caches,
                                    const cpuid_apic_id_t owner_apic_id,
                                    const int cache_type) const
{
    unsigned int apic_mask((unsigned int)-1);

    for(const legacy_cache_descriptor_t *d(&legacy_cache_table[0]);
	d->id != 0; d++) {
	if(d->id == cache_type) {
	    caches.push_back(cpuid_cache_level_t(
				 d->level,
				 d->data, d->instruction,
				 d->ways_of_associativity,
				 1, // physical_line_partition
				 d->line_size,
				 d->size,
                                 threads_sharing(apic_mask)));
	    return;
	}
    }
}

void
cpuid_cpu_intel_t::add_legacy_caches(cpuid_cache_vector_t &caches,
                                     const cpuid_apic_id_t owner_apic_id,
                                     const int reg,
                                     const int max_caches) const
{
    for(int i = 0; i < max_caches; i++) {
	const int cache_type = (reg >> (24 - i * 8)) & 0xFF;

	if(cache_type != 0)
	    add_legacy_cache(caches, owner_apic_id, cache_type);
    }
}

cpuid_cache_info_ptr_t
cpuid_cpu_intel_t::query_cache_info_legacy(
    const cpuid_apic_id_t owner_apic_id) const
{
    cpuid_cache_info_ptr_t cache_info(new cpuid_cache_info_t());

    int num_iterations = 1;

    for(int i = 0; i < num_iterations ; i++) {
	const cpuid_regs_t regs(cpuid_eax(0x00000002));

	num_iterations = bit_range(regs.eax, 0, 7);
	if(!(regs.eax & (1 << 31)))
	    add_legacy_caches(cache_info->caches,
			      owner_apic_id,
			      regs.eax, 3);
	if(!(regs.ebx & (1 << 31)))
	    add_legacy_caches(cache_info->caches,
			      owner_apic_id,
			      regs.ebx, 4);
	if(!(regs.ecx & (1 << 31)))
	    add_legacy_caches(cache_info->caches,
			      owner_apic_id,
			      regs.ecx, 4);
	if(!(regs.edx & (1 << 31)))
	    add_legacy_caches(cache_info->caches,
			      owner_apic_id,
			      regs.edx, 4);
    }

    return cache_info;
}

unsigned int
cpuid_cpu_intel_t::get_thread_bits() const
{
    if (supports_x2apic) {
        cpuid_regs_t regs(cpuid_eax_ecx(0xb, 1));
        return bit_range(regs.eax, 0, 4);
    }
    else {
        cpuid_regs_t regs(cpuid_eax(0x00000001));
        return log2_ceil(bit_range(regs.ebx, 16, 23));
    }
}

unsigned int
cpuid_cpu_intel_t::get_core_bits() const
{
    if (supports_x2apic) {
        cpuid_regs_t regs(cpuid_eax_ecx(0xb, 0));
        return get_thread_bits() - bit_range(regs.eax, 0, 4);
    }
    else {
        if(cpuid_limit >= 4) {
            cpuid_regs_t regs(cpuid_eax_ecx(0x00000004, 0));
            return log2_ceil(bit_range(regs.eax, 26, 31) + 1);
        }
        else
            return 0;
    }
}

END_NAMESPACE;
END_NAMESPACE;
