/// @file library/cpuid/cpuid_os_linux.cc

// Copyright (c) 2008,2012 Rogue Wave Software, Inc.  All rights reserved.

#include "cpuid_os.hh"
#include "cpuid.hh"

#include <errno.h>
#include <sched.h>
#include <unistd.h>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

int
cpuid_os_t::get_cpu_max()
{
    long cpu_max;

    // _SC_NPROCESSORS_CONF gives the amount of processors known to
    // the OS, but some may be disabled on the OS level by poking in
    // /sys/devices/system/cpu/cpu*/online.
    //
    // We have seen Bull systems turning off hyperthreading this way.
    //
    // _SC_NPROCESSORS_ONLN gives the number of enabled processors,
    // but there is no guarantee that the enabled ones are the ones
    // with the lower cpuid numbers.

    // We return the ONLN for now to handle the two cases of either
    // having all processors enabled, or if some are disabled, they
    // would have to be disabled from the top id, downwards, like on
    // Bull, where the disabled hyperthreads all occupied the top half
    // of the cpuid space.

    // It is expected to fail if processors are disabled anywhere else
    // than at the end.

    // ONLN is a non-standard glibc addition, so if it doesn't work we
    // will fall back to using CONF.
    
    cpu_max = sysconf(_SC_NPROCESSORS_ONLN);

    if (cpu_max < 0 && errno == EINVAL)
	cpu_max = sysconf(_SC_NPROCESSORS_CONF);

    if (cpu_max < 0)
	throw cpuid_cpu_general_exception_t("sysconf", errno);

    return cpu_max;
}

bool
cpuid_os_t::cpu_exists(unsigned int cpu_nr)
{
    return true;
}

bool
cpuid_os_t::cpu_online(unsigned int cpu_nr)
{
    return true;
}

void
cpuid_os_t::cpuid(cpuid_regs_t &regs)
{
    asm("cpuid;"
        : "=a"(regs.eax), "=b"(regs.ebx), "=c"(regs.ecx), "=d"(regs.edx)
        : "a"(regs.eax), "c"(regs.ecx));
}

void
cpuid_os_t::cpuid_cpu(const unsigned int cpu_nr, cpuid_regs_t &regs)
{
    int ret;
    cpu_set_t old_affinity;
    cpu_set_t cpu_set;
    CPU_ZERO(&cpu_set);
    CPU_SET(cpu_nr, &cpu_set);

    ret = sched_getaffinity(0, sizeof(cpu_set_t), &old_affinity);
    if (ret == -1)
	throw cpuid_cpu_general_exception_t("sched_getaffinity", errno);

    ret = sched_setaffinity(0, sizeof(cpu_set_t), &cpu_set);
    if (ret == -1) {
	if (errno == EINVAL)
	    throw cpuid_cpu_inactive_exception_t();
	else
	    throw cpuid_cpu_general_exception_t("sched_setaffinity", errno);
    }

    cpuid(regs);

    ret = sched_setaffinity(0, sizeof(cpu_set_t), &old_affinity);
    if (ret == -1)
	throw cpuid_cpu_general_exception_t("sched_getaffinity", errno);
}

END_NAMESPACE;
END_NAMESPACE;
