/// @file library/exe_path/exe_path.cc

// Copyright (c) 2010 Acumem AB.  All rights reserved.

#include "exe_path.hh"
#include "library/lang/tstring.hh"

#if defined(__APPLE__)
#include <mach-o/dyld.h>
#endif

#if defined(_WIN32)
#include <windows.h>
#include <io.h>
#include <share.h>
#else
#include <limits.h>
#include <stdio.h>
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <iostream>
#include <vector>
#include <boost/filesystem.hpp>

#define LICENSE_DIRNAME		"license"
#define BIN_DIRNAME 		"bin"
#define LIB_THREADSPOTTER_DIRNAME	"threadspotter"
#define LICENSEDIR_REGISTYKEY	"LicenseDir"
#define DOT_DIRNAME		".threadspotter"
#define APPDATA_DIRNAME		"ThreadSpotter"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(exe_path);

using namespace std;
namespace fs = boost::filesystem;

#if defined(__linux__)
#define SELF_LINK "/proc/self/exe"
#endif
#if defined(__sun__)
#define SELF_LINK "/proc/self/path/a.out"
#endif

fs::path
get_executable_filename()
{
#if defined(__linux__) || defined(__sun__)
    char exepath[PATH_MAX];
    ssize_t s = readlink(SELF_LINK, exepath, sizeof(exepath));
    if (s == -1 || s == sizeof(exepath))
  boost::throw_exception(
      boost::system::system_error(
    boost::system::error_code(
        (int) errno,
        boost::system::get_system_category()),
    "Error locating executable"));

    exepath[s] = 0;
    return fs::path(exepath);
#elif defined(__APPLE__)
    char exepath[PATH_MAX];
    uint32_t exepathSize = PATH_MAX;
    ssize_t s = _NSGetExecutablePath(exepath, &exepathSize);
    if (s == -1 || s == sizeof(exepath))
	boost::throw_exception(
	    boost::system::system_error(
		boost::system::error_code(
		    (int) errno,
		    boost::system::get_system_category()),
		"Error locating executable"));

    exepath[s] = 0;
    return fs::path(exepath);


#elif defined(_WIN32)
    char buffer[MAX_PATH];
    DWORD ret = GetModuleFileNameA(NULL, buffer, sizeof(buffer));

    if (ret == 0 || ret == sizeof(buffer))
	boost::throw_exception(
	    boost::system::system_error(
		boost::system::error_code(
		    (int) GetLastError(),
		    boost::system::get_system_category()),
		"Error locating executable"));

    return fs::path(buffer);

#else
#error Unsupported OS
#endif
}

fs::path
get_installation_directory()
{
    fs::path exe = get_executable_filename();
    fs::path exedir = exe.parent_path();
    fs::path exedir_parent = exedir.parent_path();

    // On linux, binaries are located in different places, some in bin/, and
    // some in lib/threadspotter/, or lib64/threadspotter/. We have to peel off
    // different number of directories to find the installation root.
    //
    // Return an empty path to signal that the binary is not installed in a
    // canonical place.

    if (exedir.filename() == BIN_DIRNAME)
	return exedir_parent;

    else if (exedir.filename() == LIB_THREADSPOTTER_DIRNAME)
	return exedir_parent.parent_path();

    else if (is_directory(exedir_parent / "share"))
	return exedir_parent;

    else
	return fs::path();
}

fs::path
get_app_license_file_directory()
{
#if defined(__linux__) || defined(__sun__) || defined(__APPLE__)
    fs::path installation_directory = get_installation_directory();
    if (installation_directory.empty())
	return fs::path();

    return installation_directory / LICENSE_DIRNAME;

#elif defined(_WIN32)

    // See scons.config.h for ACUMEM_PRODUCT_NAME, typically a name like:
    // "ThreadSpotter". However for historical reasons, we use
    // "Acumem ThreadSpotter", so we prepend the product name accordingly.
    HKEY hkey;
    LONG retval;
    retval = RegOpenKeyExA(HKEY_LOCAL_MACHINE,
			   "Software\\Acumem\\Acumem " ACUMEM_PRODUCT_NAME,
			   0, KEY_READ, &hkey);


    const char *error_message =
	ACUMEM_PRODUCT_NAME " is not properly registered";

    if (retval != ERROR_SUCCESS) {
	boost::throw_exception(
	    boost::system::system_error(
		boost::system::error_code(
		    (int) retval,
		    boost::system::get_system_category()),
		error_message));
    }

    DWORD size = 1024;
    DWORD s2 = size;
    char *buf = new char[size];
    while ((retval =
	    RegQueryValueExA(hkey, LICENSEDIR_REGISTYKEY,
			     NULL, NULL,
			     (LPBYTE) buf, &s2)) == ERROR_MORE_DATA) {
	delete [] buf;
	size *=2;
	s2 = size;
	buf = new char[size];
    }

    if (retval != ERROR_SUCCESS) {
	boost::throw_exception(
	    boost::system::system_error(
		boost::system::error_code(
		    (int) retval,
		    boost::system::get_system_category()),
		error_message));
    }

    // null may be included in the returned string, but must not
    // be part of the 'path' variable. Strip any trailing nul
    while (s2 > 0 && buf[s2-1] == 0)
	s2--;

    // In this case we don't want a trailing path at the end. A
    // trailing slash prevents FlexNet from automagically appending
    // "\*.lic" to the path.
    while (s2 > 0 && (buf[s2-1] == '\\'  || buf[s2-1] == '/'))
	s2--;

    fs::path license_file_dir(buf, buf+s2);
    delete [] buf;
    RegCloseKey(hkey);

    return license_file_dir;
#else
#error Unsupported OS
#endif
}

fs::path
get_user_license_file_directory()
{
#if defined(__linux__) || defined(__sun__) || defined(__APPLE__)
    const char *home = getenv("HOME");
    if (home != NULL)
	return fs::path(home) / DOT_DIRNAME / LICENSE_DIRNAME;

#elif defined(_WIN32)
    const char *appdata = getenv("APPDATA");
    if (appdata != NULL)
	return fs::path(appdata) / APPDATA_DIRNAME / LICENSE_DIRNAME;

#else
#error Unsupported OS
#endif

    return fs::path();
}

fs::path
get_temp_path()
{
    const char *tmpdir = getenv("THREADSPOTTER_TMPDIR");
    if (tmpdir != NULL)
	return fs::path(tmpdir);

    // This makes a OS dependent lookup, that may use environment variables
    // TMPDIR, TMP, TEMP, TEMPDIR (Linux)
    // GetTempPath (Windows)
    // and finally as a fallback: /tmp.
    //
    // Any error in doing this, we will fall back to using the current
    // directory.

    try {
	return fs::temp_directory_path();
    } catch (const fs::filesystem_error &) {
	// If the environment variable used by temp_directory_path
	// doesn't point to a directory, then silently default to
	// current dir.
	return fs::current_path();
    }
}

FILE *
fopen(const fs::path &path, const std::string &mode)
{
#ifdef _WIN32
    tstring_t m(mode.begin(), mode.end());
    return _tfopen(path.c_str(), m.c_str());
#else
    return ::fopen(path.c_str(), mode.c_str());
#endif
}

bool
create_file_fail_if_exists(const fs::path &path)
{
#ifdef _WIN32
    int fd;
    errno_t e = _tsopen_s(&fd, path.c_str(), _O_CREAT | _O_EXCL,
			  _SH_DENYNO, _S_IREAD | _S_IWRITE);
    if (e == 0) {
	_close(fd);
	return true;
    } else {
	// If the directory can't be found, then obviously, the
	// filename is unique.
	if (errno == ENOENT)
	    return true;

	// If the file already exist, indicate error
	if (errno == EEXIST)
	    return false;

	else {
	    // Other errors are thrown as exceptions
	    throw fs::filesystem_error(
		"creat", path,
		boost::system::error_code(errno,
					  boost::system::system_category()));
	}
    }
#else
    int fd = open(path.c_str(), O_CREAT | O_EXCL, 0777);
    if (fd != -1) {
	close(fd);
	return true;
    } else if (errno == EEXIST) {
	return false;
    } else {
	throw fs::filesystem_error(
	    "creat", path,
	    boost::system::error_code(errno,
				      boost::system::system_category()));
    }
#endif
}


END_NAMESPACE;
END_NAMESPACE;
