/// @file library/exe_path/path_substitute.cc

// Copyright (c) 2012 Rogue Wave Software.  All rights reserved.

#include "path_substitute.hh"
#include "exe_path.hh"

#include <stdlib.h>
#include <string>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>

#if defined(_WIN32)
#include <winsock2.h>
#include <windows.h>
#else
#include <sys/types.h>
#include <unistd.h>
#endif

namespace fs = boost::filesystem;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(exe_path);

using namespace std;

static string
get_hostname()
{
    char buf[256]; // sufficient both w/ respect to winsock and posix
    int ret = gethostname(buf, sizeof(buf));
    if (ret == -1)
	return string("unknown");

    buf[sizeof(buf)-1] = 0;
    return string(buf);
}

static string
get_pid()
{
#if defined(_WIN32)
    return boost::lexical_cast<string>(GetCurrentProcessId());
#else
    return boost::lexical_cast<string>(getpid());
#endif
}

static string
get_env(const string &env)
{
    char *p = getenv(env.c_str());
    if (p != NULL)
	return p;
    else
	return "";
}

static string
get_rank()
{
    // Copied from totalview:
    // debugger/src/iccdp_server/is_message_actions.cxx:get_rank

    const char *names[] =
    {
      "PMI_RANK",               /* MPICH 2 & derivatives */
      "OMPI_COMM_WORLD_RANK",   /* OpenMPI 1.3 */
      "OMPI_MCA_ns_nds_vpid",   /* OpenMPI 1.2 & derivatives */
      "PMI_ID",                 /* SLURM PMI, use this first */
      "SLURM_PROCID",           /* SLURM */
      "LAMRANK",                /* LAM */
      "MPI_RANKID",             /* HP MPI for Linux */
      "MP_CHILD",               /* IBM PE */
      "MP_RANK",                /* Sun CT */
      "MPIRUN_RANK",            /* MVAPICH 1.1 */
    };

    BOOST_FOREACH(const char *e, names) {
	const char *env = getenv(e);
	if (env)
	    return env;
    }
    return string();
}

static string
get_unique_string()
{
    string name = get_rank();
    if (name.length() != 0)
	return name;
    else
	return get_hostname() + "-" + get_pid();
}

static string
perform_substitutions(const string &s, bool *test_uniqueness, int *counter)
{
    string r;
    int index = 0;
    int len = (int) s.length();
    int start;

    while (index < len) {
	if (s[index] != '%') {
	    r += s[index];
	    index++;
	    continue;
	}

	if (++index >= len)
	    break;

	switch (s[index++]) {
	case '%':
	    r += '%';
	    break;
	case 'r':
	    r += get_unique_string();
	    break;
	case 'h':
	    r += get_hostname();
	    break;
	case 'p':
	    r += get_pid();
	    break;

	case 'u':
	    if (counter && *counter > 0)
		r += "." + boost::lexical_cast<string>(*counter);
	    *test_uniqueness = true;
	    break;

	case 'U':
	    if (counter)
		r += "." + boost::lexical_cast<string>(*counter);
	    *test_uniqueness = true;
	    break;

	case '{':
	    start = index;
	    while (index < len) {
		if (s[index++] == '}') {
		    r += get_env(s.substr(start, index-start-1));
		    break;
		}
	    }
	    break;
	}
    }
    return r;
}

fs::path
path_substitute_t::substitute(const fs::path &filename) const
{
    bool test_uniqueness = false;
    int counter = 0;

    fs::path ret;
    for (;;) {
	// Scan each component for % substitutions
	BOOST_FOREACH(fs::path p, filename)
	    ret /= perform_substitutions(p.string(),
					 &test_uniqueness,
					 &counter);

	if (!test_uniqueness)
	    break;
	if (create_file_fail_if_exists(ret))
	    break;
	
	counter ++;
	ret.clear();
    }

    return ret;
}

bool
path_substitute_t::includes_unique_marker(const fs::path &filename) const
{
    bool test_uniqueness = false;

    // Scan each component for % substitutions, collect uniqueness info
    BOOST_FOREACH(fs::path p, filename)
	perform_substitutions(p.string(),
			      &test_uniqueness,
			      NULL);

    return test_uniqueness;
}

END_NAMESPACE;
END_NAMESPACE;
