/// @file library/feedback/feedback.hh

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#ifndef FEEDBACK_HH
#define FEEDBACK_HH

#include "library/lang/common_macros.h"

#include <boost/shared_ptr.hpp>
#include <string>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

using boost::shared_ptr;

/**
 * Exception thrown to cancel processing after a user chooses to
 * cancel.
 */
class cancellation_exception_t {};

/**
 * Base class for feedback mechanism.
 *
 * The top level component derives a subclass of this class and passes
 * it through the configuration_t structure to the modules.
 *
 * The top module will call add_major_step() for each major invocation
 * it plans to do and then call select_major_step() when the step
 * starts to execute.  The feedback implementation may assume, for the
 * purpose of drawing progress bars etc., that all major steps will
 * execute in the order they were added.
 *
 * After a call to select_major_step(), the feedback handler must
 * assume that the number of minor steps in the selected major step is
 * 0 until a call to set_minor_steps() has occured. Major steps having
 * 0 minor steps are assumed to have a non deterministic progress, but
 * may still provide feedback by calling update().
 *
 * Feedback for a minor step is provided through the update call. The
 * completed minor_steps must be smaller or equal to the number of
 * minor sets defined using the set_minor_steps() call, if
 * set_minor_steps() hasn't been called the number of steps is assumed
 * to be 0.
 */
class feedback_base_t {
public:
    enum status_t {
	STATUS_UNKNOWN = 0,
	STATUS_RUNNING = 1,
	STATUS_COMPLETED,
	STATUS_FAILED,
	STATUS_CANCELED
    };

    enum log_type_t {
	LOG_DEBUG = -1,
	LOG_INFO = 0,
	LOG_WARNING = 10,
	LOG_ERROR = 20
    };

    /// Only tests and internal tools should use this class directly. Tools
    /// used by customers should use licensed_feedback_t, that also does
    /// license checks in test_cancel().
    feedback_base_t();
    virtual ~feedback_base_t();

    /**
     * This may be called by the top level component to request that
     * processing shall stop. The top module will call
     * update_status(status_t) with STATUS_CANCELED if the processing
     * was cancelled. The processing may however complete
     * successfully, or fail, despite the cancel request, in such a
     * case the appropriate status will be signaled.
     */
    void cancel();

    /// Modules can query the state of the cancel flag.
    bool query_cancel() const;
    /**
     * Conditionally throw a cancellation_exception_t if value of the
     * cancel flag is true. This should be called at regular intervals
     *
     * \throw cancellation_exception_t if cancel has been requested.
     */
    virtual void test_cancel() const throw(cancellation_exception_t);

    /// Resets the complete state of the feedback handler.
    virtual void reset();

    /**
     * Adds a major step.
     *
     * @param message name/Message to display while the step is executing.
     * @return ID of the step.
     */
    virtual int add_major_step(const string &message);

    /**
     * Advance the pointer to the next step. This is called by the
     * module, and should be implemented by a feedback
     * handler. Calling this function also sets the number of minor
     * steps to 0.
     *
     * @param step ID of the step
     */
    virtual void select_major_step(int step);

    /**
     * Sets the number of steps in the current major step. If this
     * function isn't called, or is called with steps set to zero,
     * the major step is assumed to have an unknown number of minor
     * steps.
     *
     * @param steps number of minor steps.
     */
    virtual void set_minor_steps(int steps);

    /**
     * Signals progress in major step. If the number of minor steps is
     * zero, the feedback module should signal "an unknown amount of
     * progress", e.g. by pulsing a progress bar.
     *
     * @param completed_minor_steps number of completed minor steps
     */
    virtual void update(int completed_minor_steps, const string &message = "");

    /**
     * Provides textual feedback to the user.
     *
     * @param type message severity.
     * @param message
     */
    virtual void log(log_type_t type, const string &message);

    /**
     * This function should be called by the top module whenever the
     * global status changes.
     *
     * @param status new global status.
     */
    virtual void update_status(status_t status);

private:
    bool cancel_;
};

typedef shared_ptr<feedback_base_t> feedback_ptr_t;

END_NAMESPACE;
END_NAMESPACE;

#endif
