/// @file library/fenris/codec_binary.cc

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#define __STDC_LIMIT_MACROS

#include "codec_binary.hh"
#include "library/codec/codec_utils.hh"
#include "library/version/version.h"

#include <boost/foreach.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

// ----- Utility functions. ---------------------------------------------------

static uint32_t
ones32(register uint32_t x)
{
    /* 32-bit recursive reduction using SWAR...
       but first step is mapping 2-bit values
       into sum of 2 1-bit values in sneaky way
    */
    x -= ((x >> 1) & 0x55555555);
    x = (((x >> 2) & 0x33333333) + (x & 0x33333333));
    x = (((x >> 4) + x) & 0x0f0f0f0f);
    x += (x >> 8);
    x += (x >> 16);
    return x & 0x0000003f;
}


static uint32_t
ones64(register uint64_t x)
{
    return ones32((uint32_t) x & 0xffffffff) + ones32((uint32_t) (x >> 32));
}

static uint32_t
floor_log2_64(register uint64_t x)
{
    x |= (x >> 1);
    x |= (x >> 2);
    x |= (x >> 4);
    x |= (x >> 8);
    x |= (x >> 16);
    x |= (x >> 32);
    return ones64(x >> 1);
}

#define log2(x) floor_log2_64(x)

// ----- Line size encoding and decoding. -------------------------------------

static inline uint8_t
encode_line_size(fnr_line_size_t line_size)
{
    return (uint8_t) (line_size == 0 ? 255U : log2(line_size));
}

static inline fnr_line_size_t
decode_line_size(uint8_t line_size_log2)
{
    return line_size_log2 == 255U ? 0 : (1LL << line_size_log2);
}

// ----- Status conversions. --------------------------------------------------

static stream_status_t
flow_to_stream_status(flow_status_t fs)
{
    switch (fs) {
    case FLOW_ERROR:
	return STREAM_ERROR;
    case FLOW_ABORT:
	return STREAM_ABORT;
    case FLOW_OK:
	return STREAM_OK;
    default:
	assert(0);
#ifdef _WIN32
	return STREAM_ERROR;
#endif
    }
}

static flow_status_t
stream_to_flow_status(stream_status_t ss)
{
    switch (ss) {
    case STREAM_ERROR:
	return FLOW_ERROR;
    case STREAM_ABORT:
	return FLOW_ABORT;
    case STREAM_OK:
	return FLOW_OK;
    default:
	assert(0);
#ifdef _WIN32
	return FLOW_ERROR;
#endif
    }
}

/* Address encoding and decoding in events compatible with version 0 of the
   fenris file format, 32-bit tag and 64-bit virtual address. */

static void
binary_encode_address_v0(codec_buffer_t *buffer,
                         const fnr_address_t address)
{
    binary_encode_uint32(buffer, 0);
    binary_encode_uint64(buffer, address.virtual_address);
}

static fnr_address_t
binary_decode_address_v0(codec_buffer_t *buffer)
{
    binary_decode_uint32(buffer);
    return fnr_address_t(binary_decode_uint64(buffer));
}

/* Address encoding and decoding in events compatible with version 1 of the
   fenris file format, 64-bit virtual address only. */

static void
binary_encode_address_v1(codec_buffer_t *buffer,
                         const fnr_address_t address)
{
    binary_encode_uint64(buffer, address.virtual_address);
}

static fnr_address_t
binary_decode_address_v1(codec_buffer_t *buffer)
{
    return fnr_address_t(binary_decode_uint64(buffer));
}

static void
binary_encode_access(codec_buffer_t *buffer,
		     const fnr_sample_access_t *a)
{
    binary_encode_address_v1(buffer, a->pc);
    binary_encode_address_v1(buffer, a->addr);
    binary_encode_uint64(buffer, a->time);
    binary_encode_uint32(buffer, a->stack);
    binary_encode_uint32(buffer, a->thread.id);
    binary_encode_uint16(buffer, a->access_size);
    binary_encode_uint8(buffer, a->access_type);
}

static void
binary_decode_access(codec_buffer_t *buffer,
		     fnr_sample_access_t *a)
{
    a->pc = binary_decode_address_v1(buffer);
    a->addr = binary_decode_address_v1(buffer);
    a->time = binary_decode_uint64(buffer);
    a->stack = binary_decode_uint32(buffer);
    a->thread.id = binary_decode_uint32(buffer);
    a->access_size = binary_decode_uint16(buffer);
    a->access_type = binary_decode_uint8(buffer);
}

// ----- fnr_codec_binary_context_t -------------------------------------------

#define CODE_VERSION_SELECT 9

const event_type_t fnr_codec_binary_context_t::events_v0[] = {
    {FNR_Event_Read, 0, &fnr_decoder_binary_t::event_read},
    {FNR_Event_Write, 1, &fnr_decoder_binary_t::event_write},
    {FNR_Event_Prefetch, 5, &fnr_decoder_binary_t::event_prefetch},
    {FNR_Event_Sample_Begin, 2, &fnr_decoder_binary_t::event_sample_begin},
    {FNR_Event_Sample_End, 3, &fnr_decoder_binary_t::event_sample_end},
    {FNR_Event_Sample_Period, 4, &fnr_decoder_binary_t::event_sample_period},
    {FNR_Event_Acquisition_Begin, 6, &fnr_decoder_binary_t::event_acquisition_begin},
    {FNR_Event_Acquisition_End, 7, &fnr_decoder_binary_t::event_acquisition_end},
    {FNR_Event_Mmap, 8, &fnr_decoder_binary_t::event_mmap},
    {(fnr_event_kind_t)-1, CODE_VERSION_SELECT, &fnr_decoder_binary_t::version_select},
    {(fnr_event_kind_t)-1, -1, 0}
};

const event_type_t fnr_codec_binary_context_t::events_v1[] = {
    {FNR_Event_Read, 0, &fnr_decoder_binary_t::event_read},
    {FNR_Event_Write, 1, &fnr_decoder_binary_t::event_write},
    {FNR_Event_Prefetch, 5, &fnr_decoder_binary_t::event_prefetch},
    {FNR_Event_Sample_Begin, 2, &fnr_decoder_binary_t::event_sample_begin},
    {FNR_Event_Sample_End, 3, &fnr_decoder_binary_t::event_sample_end},
    {FNR_Event_Sample_Period, 4, &fnr_decoder_binary_t::event_sample_period},
    {FNR_Event_Acquisition_Begin, 6, &fnr_decoder_binary_t::event_acquisition_begin},
    {FNR_Event_Acquisition_End, 7, &fnr_decoder_binary_t::event_acquisition_end},
    {FNR_Event_Mmap, 8, &fnr_decoder_binary_t::event_mmap},
    {(fnr_event_kind_t)-1, CODE_VERSION_SELECT, &fnr_decoder_binary_t::version_select},
    {FNR_Event_Stack, 10, &fnr_decoder_binary_t::event_stack},
    {(fnr_event_kind_t)-1, 11, &fnr_decoder_binary_t::event_sample_deprecated},
    {(fnr_event_kind_t)-1, 12, &fnr_decoder_binary_t::event_dangling_deprecated},
    {FNR_Event_Time_Mapping, 13, &fnr_decoder_binary_t::event_time_mapping},
    {FNR_Event_Local_Time, 19, &fnr_decoder_binary_t::event_local_time},
    {FNR_Event_Burst_Begin, 14, &fnr_decoder_binary_t::event_burst_begin},
    {FNR_Event_Burst_End, 15, &fnr_decoder_binary_t::event_burst_end},
    {FNR_Event_Sample, 16, &fnr_decoder_binary_t::event_sample},
    {FNR_Event_Contexts,  17, &fnr_decoder_binary_t::event_contexts},
    {FNR_Event_Munmap, 18, &fnr_decoder_binary_t::event_munmap},
    {(fnr_event_kind_t)-1, -1, 0}
};

fnr_codec_binary_context_t::fnr_codec_binary_context_t(
    warning_consumer_ptr_t warning_consumer)
    : warning_consumer(warning_consumer),
      version_major(0),
      version_minor(0),
      svn_version(""),
      features(FNR_Feature_None),
      events(NULL)
{
    events = events_v0;
}

void
fnr_codec_binary_context_t::set_version(
    int version_major,
    int version_minor,
    const string &svn_version,
    fnr_features_t features)
{
    switch (version_major) {
    case 0:
	events = events_v0;
        break;
    case 1:
    case 2: /* Version 2 doesn't change any event packet format, only the size
               field in the event header. */
    case 3: /* Version 3 doesn't change any events, only adds new time mapping
               events. */
	events = events_v1;
        break;
    default:
        throw codec_exception_t("Unsupport sample file version.", true);
    }

    this->version_major = version_major;
    this->version_minor = version_minor;
    this->svn_version = svn_version;
    this->features = features;
}

static const event_type_t *
binary_lookup_event(const event_type_t *events, fnr_event_kind_t kind)
{
    int i;
    for (i=0; events[i].decode != 0; i++) {
	if (events[i].kind == kind)
	    return &events[i];
    }

    return NULL;
}

static const event_type_t *
binary_lookup_code(const event_type_t *events, int code)
{
    int i;
    for (i=0; events[i].decode != 0; i++) {
	if (events[i].code == code)
	    return &events[i];
    }

    return NULL;
}

// ----- fnr_decoder_binary_t -------------------------------------------------

fnr_decoder_binary_t::fnr_decoder_binary_t(
    fenris_flow_t *flow_next,
    warning_consumer_ptr_t warning_consumer)
    : context(warning_consumer),
      flow_next(flow_next),
      warning_consumer(warning_consumer),
      flow_initialized(false),
      waiting_for_header(true)
{
}

stream_status_t
fnr_decoder_binary_t::process_data(const char *buf, size_t len)
{
    // Write the data to our internal buffer.
    buffer.write(buf, len);

    // Process events until we run out of data.
    while (1)
        if (waiting_for_header) {
            size_t size_size = context.version_major >= 2 ? 4 : 2;
            size_t header_size = 1 + size_size;

            if (buffer.size() < header_size)
                return STREAM_OK;

            buffer.assert_read(header_size);

            header.code = binary_decode_uint8(&buffer);
            switch (size_size) {
            case 2:
                header.size = binary_decode_uint16(&buffer);
                break;
            case 4:
                header.size = binary_decode_uint32(&buffer);
                break;
            default:
                assert(0);
            }

            waiting_for_header = false;
        }
        else {
            if (!flow_initialized && header.code != CODE_VERSION_SELECT)
                initialize_flow();

            if (buffer.size() < header.size)
                return STREAM_OK;

            waiting_for_header = true;

            buffer.assert_read(header.size);

            const event_type_t *event_type =
                binary_lookup_code(context.events, header.code);
            if (event_type) {
                event_ptr_t event((this->*event_type->decode)(header.size));

                if (header.code == CODE_VERSION_SELECT)
                    continue;

                flow_status_t flow_status = flow_next->process_event(*event);
                if (flow_status != FLOW_OK)
                    return flow_to_stream_status(flow_status);
            }
            else
                buffer.read(header.size);
        }
}

stream_status_t
fnr_decoder_binary_t::flush()
{
    return STREAM_OK;
}

fnr_event_t *
fnr_decoder_binary_t::event_read(size_t size)
{
    fnr_event_read_t *e = new fnr_event_read_t;
    e->pc = binary_decode_address_v0(&buffer);
    e->address = binary_decode_address_v0(&buffer);
    e->thread_id.id = binary_decode_uint32(&buffer);
    e->size = binary_decode_uint16(&buffer);
    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_write(size_t size)
{
    fnr_event_write_t *e = new fnr_event_write_t;
    e->pc = binary_decode_address_v0(&buffer);
    e->address = binary_decode_address_v0(&buffer);
    e->thread_id.id = binary_decode_uint32(&buffer);
    e->size = binary_decode_uint16(&buffer);
    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_prefetch(size_t size)
{
    fnr_event_prefetch_t *e = new fnr_event_prefetch_t;
    e->pc = binary_decode_address_v0(&buffer);
    e->address = binary_decode_address_v0(&buffer);
    e->thread_id.id = binary_decode_uint32(&buffer);
    e->size = binary_decode_uint16(&buffer);
    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_sample_begin(size_t size)
{
    fnr_event_sample_begin_t *e = new fnr_event_sample_begin_t;
    e->pc = binary_decode_address_v0(&buffer);
    e->address = binary_decode_address_v0(&buffer);
    e->time = binary_decode_uint64(&buffer);
    e->thread_id.id = binary_decode_uint32(&buffer);
    e->sample_id = (fnr_sample_count_t) binary_decode_uint64(&buffer);
    e->kind = binary_decode_uint8(&buffer);
    e->access_size = binary_decode_uint16(&buffer);
    e->line_size = decode_line_size(binary_decode_uint8(&buffer));
    e->pc_plus_one = binary_decode_address_v0(&buffer);
    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_sample_end(size_t size)
{
    fnr_event_sample_end_t *e = new fnr_event_sample_end_t;
    e->pc = binary_decode_address_v0(&buffer);
    e->address = binary_decode_address_v0(&buffer);
    e->time = binary_decode_uint64(&buffer);
    e->thread_id.id = binary_decode_uint32(&buffer);
    e->sample_id = (fnr_sample_count_t) binary_decode_uint64(&buffer);
    e->kind = binary_decode_uint8(&buffer);
    e->access_size = binary_decode_uint16(&buffer);
    e->rdist = binary_decode_uint64(&buffer);
    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_sample_period(size_t size)
{
    fnr_event_sample_period_t *e = new fnr_event_sample_period_t;
    e->sample_period = binary_decode_uint32(&buffer);
    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_acquisition_begin(size_t size)
{
    fnr_event_acquisition_begin_t *e = new fnr_event_acquisition_begin_t;

    size_t begin_size = buffer.size();

    e->time = binary_decode_uint64(&buffer);
    e->architecture = (fnr_architecture_t)binary_decode_uint32(&buffer);
    e->argv = binary_decode_string16_vector(&buffer);
    e->user = binary_decode_string16(&buffer);

    if (begin_size - buffer.size() < size) {
	uint8_t num_line_sizes = binary_decode_uint8(&buffer);
        e->line_sizes_log2.reserve(num_line_sizes);
        for (int i = 0; i < num_line_sizes; i++)
            e->line_sizes_log2.push_back(binary_decode_uint8(&buffer));
    }

    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_acquisition_end(size_t size)
{
    fnr_event_acquisition_end_t *e = new fnr_event_acquisition_end_t;
    e->time = binary_decode_uint64(&buffer);
    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_mmap(size_t size)
{
    fnr_event_mmap_t *e = new fnr_event_mmap_t();

    size_t begin_size = buffer.size();

    e->address = binary_decode_uint64(&buffer);
    e->size = binary_decode_uint64(&buffer);
    e->attr = binary_decode_uint32(&buffer);
    e->device = binary_decode_uint64(&buffer);
    e->inode = binary_decode_uint64(&buffer);
    e->offset = binary_decode_uint64(&buffer);
    e->filename = binary_decode_string16(&buffer);

    if (begin_size - buffer.size() < size)
        e->global_time = binary_decode_uint64(&buffer);
    else
	e->global_time = 0;

    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_munmap(size_t size)
{
    fnr_event_munmap_t *e = new fnr_event_munmap_t();
    e->address = binary_decode_uint64(&buffer);
    e->size = binary_decode_uint64(&buffer);
    e->global_time = binary_decode_uint64(&buffer);
    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_stack(size_t size)
{
    fnr_event_stack_t *e = new fnr_event_stack_t;

    e->id = binary_decode_uint32(&buffer);
    uint16_t depth = binary_decode_uint16(&buffer);
    e->ret_addrs.reserve(depth);
    for (uint16_t i = 0; i < depth; i++)
        e->ret_addrs.push_back(binary_decode_address_v1(&buffer));

    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_sample(size_t size)
{
    fnr_event_sample_t *e = new fnr_event_sample_t;

    binary_decode_access(&buffer, &e->begin);
    e->pc_1_plus = binary_decode_address_v1(&buffer);
    e->pc_1_plus_stack = binary_decode_uint32(&buffer);
    e->line_size_log2 = binary_decode_uint8(&buffer);
    e->flags = binary_decode_uint8(&buffer);
    uint32_t end_count = binary_decode_uint32(&buffer);
    e->end.resize(end_count);
    BOOST_FOREACH(fnr_sample_access_t &a, e->end)
	binary_decode_access(&buffer, &a);

    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_sample_deprecated(size_t size)
{
    fnr_event_sample_t *e = new fnr_event_sample_t;

    size_t begin_size = buffer.size();

    fnr_sample_access_t end;
    e->begin.addr = binary_decode_address_v1(&buffer);
    end.addr = binary_decode_address_v1(&buffer);
    e->begin.time = binary_decode_uint64(&buffer);
    end.time = binary_decode_uint64(&buffer);
    e->begin.pc = binary_decode_address_v1(&buffer);
    e->pc_1_plus = binary_decode_address_v1(&buffer);
    end.pc = binary_decode_address_v1(&buffer);
    e->begin.stack = binary_decode_uint32(&buffer);
    e->pc_1_plus_stack = binary_decode_uint32(&buffer);
    end.stack = binary_decode_uint32(&buffer);
    e->begin.thread.id = binary_decode_uint32(&buffer);
    end.thread.id = binary_decode_uint32(&buffer);
    e->begin.access_size = binary_decode_uint16(&buffer);
    end.access_size = binary_decode_uint16(&buffer);
    e->begin.access_type = binary_decode_uint8(&buffer);
    end.access_type = binary_decode_uint8(&buffer);
    e->line_size_log2 = binary_decode_uint8(&buffer);

    if (begin_size - buffer.size() < size) {
	// Old MP extension, should never occur in customer sample files.
	uint32_t end_count = binary_decode_uint32(&buffer);
        e->end.reserve(end_count);
        e->end.push_back(end);
        for (uint32_t i = 1; i < end_count; i++) {
            fnr_sample_access_t sa;
            sa.pc = binary_decode_address_v1(&buffer);
            sa.addr = binary_decode_address_v1(&buffer);
            sa.time = binary_decode_uint64(&buffer);
            sa.stack = binary_decode_uint32(&buffer);
            sa.thread.id = binary_decode_uint32(&buffer);
            sa.access_size = binary_decode_uint16(&buffer);
            sa.access_type = binary_decode_uint8(&buffer);
            e->end.push_back(sa);
	}
    }
    else {
	/* No MP extension. */
	e->end.reserve(1);
	e->end.push_back(end);
    }

    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_dangling_deprecated(size_t size)
{
    fnr_event_sample_t *e = new fnr_event_sample_t;
    e->begin.addr = binary_decode_address_v1(&buffer);
    e->begin.time = binary_decode_uint64(&buffer);
    e->begin.pc = binary_decode_address_v1(&buffer);
    e->pc_1_plus = binary_decode_address_v1(&buffer);
    e->begin.stack = binary_decode_uint32(&buffer);
    e->pc_1_plus_stack = binary_decode_uint32(&buffer);
    e->begin.thread.id = binary_decode_uint32(&buffer);
    e->begin.access_size = binary_decode_uint16(&buffer);
    e->begin.access_type = binary_decode_uint8(&buffer);
    e->line_size_log2 = binary_decode_uint8(&buffer);
    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_time_mapping(size_t size)
{
    fnr_event_time_mapping_t *e = new fnr_event_time_mapping_t;

    uint16_t event_count;

    event_count = binary_decode_uint16(&buffer);
    e->time_events.reserve(event_count);
    for (uint16_t i = 0; i < event_count; i++) {
        fnr_time_t global_time = binary_decode_uint64(&buffer);
        fnr_thread_id_t tid(binary_decode_uint32(&buffer));
        e->time_events.push_back(
            fnr_event_time_mapping_t::time_event_t(global_time, tid));
    }

    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_local_time(size_t size)
{
    uint32_t tid = binary_decode_uint32(&buffer);
    uint64_t global_time = binary_decode_uint64(&buffer);
    uint64_t local_time = binary_decode_uint64(&buffer);

    return new fnr_event_local_time_t(
        fnr_thread_id_t(tid), global_time, local_time);
}

fnr_event_t *
fnr_decoder_binary_t::event_burst_begin(size_t size)
{
    fnr_event_burst_begin_t *e = new fnr_event_burst_begin_t;
    e->real_time = binary_decode_uint64(&buffer);
    e->time = binary_decode_uint64(&buffer);
    e->sampling_time = binary_decode_uint64(&buffer);
    e->cooldown_time = binary_decode_uint64(&buffer);
    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_burst_end(size_t size)
{
    fnr_event_burst_end_t *e = new fnr_event_burst_end_t;
    e->real_time = binary_decode_uint64(&buffer);
    e->time = binary_decode_uint64(&buffer);
    return e;
}

fnr_event_t *
fnr_decoder_binary_t::event_contexts(size_t size)
{
    fnr_event_contexts_t *e = new fnr_event_contexts_t;

    uint32_t thread_count = binary_decode_uint32(&buffer);
    e->contexts.reserve(thread_count);
    for (uint32_t i = 0; i < thread_count; i++)
	e->contexts.push_back(binary_decode_uint32(&buffer));

    return e;
}

fnr_event_t *
fnr_decoder_binary_t::version_select(size_t size)
{
    uint32_t version_major;
    uint32_t version_minor = 0;
    uint64_t features = FNR_Feature_None;
    string svn_version = "unknown";

    assert(!flow_initialized);

    size_t begin_size = buffer.size();

    version_major = binary_decode_uint32(&buffer);

    // Older sample files may not have some of the fields below. If we are
    // reading such a file the defaults set above will be used instead.
    if (begin_size - buffer.size() < size)
        version_minor = binary_decode_uint32(&buffer);
    if (begin_size - buffer.size() < size)
        svn_version = binary_decode_string16(&buffer);
    if (begin_size - buffer.size() < size)
        features = binary_decode_uint64(&buffer);

    context.set_version(version_major, version_minor, svn_version, features);

    return NULL;
}

void
fnr_decoder_binary_t::initialize_flow()
{
    assert(!flow_initialized);
    flow_status_t flow_status = flow_next->flow_init(context.features);
    if (flow_status != FLOW_OK)
        throw codec_exception_t(
            "Failed to initialize flow.", flow_status == FLOW_ERROR);
    flow_initialized = true;
}

// ----- fnr_encoder_binary_t -------------------------------------------------

fnr_encoder_binary_t::fnr_encoder_binary_t(
    stream_t *stream_next,
    warning_consumer_ptr_t warning_consumer)
    : stream_next(stream_next),
      warning_consumer(warning_consumer),
      context(warning_consumer)
{
}

void
fnr_encoder_binary_t::write_event(int code)
{
    size_t size = event_buffer.size();

    binary_encode_uint8(&_buffer, code);
    if (context.version_major >= 2 && code != CODE_VERSION_SELECT)
        binary_encode_uint32(&_buffer, (uint32_t) size);
    else
        binary_encode_uint16(&_buffer, (uint16_t) size);
    _buffer.write(event_buffer.read(size), size);
}

flow_status_t
fnr_encoder_binary_t::flow_init(fnr_features_t features)
{
    int version_major = CODEC_BINARY_VERSION_CURRENT;
    int version_minor = 1;
    char svn_version[20];
    version_get_svnversion(svn_version, 20);
    context.set_version(version_major, version_minor, svn_version, features);

    // Encode and write the version select event.
    version_select();
    write_event(CODE_VERSION_SELECT);

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::process_event(const fnr_event_t &event)
{
    // Look up the event kind.
    const event_type_t *event_type =
        binary_lookup_event(context.events, event.get_kind());

    // Return an error for unknown event kinds.
    if (!event_type){
        warning_consumer->error("Unknown Fenris event kind.");
        return FLOW_ERROR;
    }

    // Encode the event.
    flow_status_t status = fenris_flow_by_type_t::process_event(event);
    if (status != FLOW_OK)
        return status;
    write_event(event_type->code);

    // Write the buffer to the stream.
    size_t size = _buffer.size();
    return stream_to_flow_status(
        stream_next->process_data(_buffer.read(size), size));
}

void
fnr_encoder_binary_t::version_select()
{
    binary_encode_uint32(&event_buffer, context.version_major);
    binary_encode_uint32(&event_buffer, context.version_minor);
    binary_encode_string16(&event_buffer, context.svn_version.c_str());
    binary_encode_uint64(&event_buffer, context.features);
}

flow_status_t
fnr_encoder_binary_t::read(const fnr_event_read_t &event)
{
    binary_encode_address_v0(&event_buffer, event.pc);
    binary_encode_address_v0(&event_buffer, event.address);
    binary_encode_uint32(&event_buffer, event.thread_id.id);
    binary_encode_uint16(&event_buffer, event.size);

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::write(const fnr_event_write_t &event)
{
    binary_encode_address_v0(&event_buffer, event.pc);
    binary_encode_address_v0(&event_buffer, event.address);
    binary_encode_uint32(&event_buffer, event.thread_id.id);
    binary_encode_uint16(&event_buffer, event.size);

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::prefetch(const fnr_event_prefetch_t &event)
{
    binary_encode_address_v0(&event_buffer, event.pc);
    binary_encode_address_v0(&event_buffer, event.address);
    binary_encode_uint32(&event_buffer, event.thread_id.id);
    binary_encode_uint16(&event_buffer, event.size);

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::sample_begin(const fnr_event_sample_begin_t &event)
{
    binary_encode_address_v0(&event_buffer, event.pc);
    binary_encode_address_v0(&event_buffer, event.address);
    binary_encode_uint64(&event_buffer, event.time);
    binary_encode_uint32(&event_buffer, event.thread_id.id);
    binary_encode_uint64(&event_buffer, event.sample_id);
    binary_encode_uint8(&event_buffer, event.kind);
    binary_encode_uint16(&event_buffer, event.access_size);
    binary_encode_uint8(&event_buffer, encode_line_size(event.line_size));
    binary_encode_address_v0(&event_buffer, event.pc_plus_one);

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::sample_end(const fnr_event_sample_end_t &event)
{
    binary_encode_address_v0(&event_buffer, event.pc);
    binary_encode_address_v0(&event_buffer, event.address);
    binary_encode_uint64(&event_buffer, event.time);
    binary_encode_uint32(&event_buffer, event.thread_id.id);
    binary_encode_uint64(&event_buffer, event.sample_id);
    binary_encode_uint8(&event_buffer, event.kind);
    binary_encode_uint16(&event_buffer, event.access_size);
    binary_encode_uint64(&event_buffer, event.rdist);

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::sample_period(const fnr_event_sample_period_t &event)
{
    /// Only 32 bits are currently encoded by the codec.
    assert(event.sample_period == (uint32_t)event.sample_period);

    /// XXX: This should be 64 bit
    binary_encode_uint32(&event_buffer, (uint32_t)event.sample_period);

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::acquisition_begin(const fnr_event_acquisition_begin_t &event)
{
    binary_encode_uint64(&event_buffer, event.time);
    binary_encode_uint32(&event_buffer, event.architecture);
    binary_encode_uint32(&event_buffer, (uint32_t) event.argv.size());

    BOOST_FOREACH(const string &arg, event.argv) {
        binary_encode_data16(&event_buffer, arg.c_str(), arg.length() + 1);
    }

    binary_encode_data16(&event_buffer, event.user.c_str(), event.user.length() + 1);

    binary_encode_uint8(&event_buffer, (uint8_t) event.line_sizes_log2.size());
    BOOST_FOREACH(fnr_line_size_log2_t ls, event.line_sizes_log2) {
	binary_encode_uint8(&event_buffer, ls);
    }

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::acquisition_end(const fnr_event_acquisition_end_t &event)
{
    binary_encode_uint64(&event_buffer, event.time);

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::mmap(const fnr_event_mmap_t &event)
{
    binary_encode_uint64(&event_buffer, event.address);
    binary_encode_uint64(&event_buffer, event.size);
    binary_encode_uint32(&event_buffer, event.attr);
    binary_encode_uint64(&event_buffer, event.device);
    binary_encode_uint64(&event_buffer, event.inode);
    binary_encode_uint64(&event_buffer, event.offset);
    binary_encode_string16(&event_buffer, event.filename.c_str());
    binary_encode_uint64(&event_buffer, event.global_time);

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::munmap(const fnr_event_munmap_t &event)
{
    binary_encode_uint64(&event_buffer, event.address);
    binary_encode_uint64(&event_buffer, event.size);
    binary_encode_uint64(&event_buffer, event.global_time);

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::stack(const fnr_event_stack_t &event)
{
    binary_encode_uint32(&event_buffer, event.id);
    binary_encode_uint16(&event_buffer, (uint16_t) event.ret_addrs.size());
    BOOST_FOREACH(const fnr_address_t &addr, event.ret_addrs)
        binary_encode_address_v1(&event_buffer, addr);

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::sample(const fnr_event_sample_t &event)
{
    binary_encode_access(&event_buffer, &event.begin);

    binary_encode_address_v1(&event_buffer, event.pc_1_plus);
    binary_encode_uint32(&event_buffer, event.pc_1_plus_stack);
    binary_encode_uint8(&event_buffer, event.line_size_log2);
    binary_encode_uint8(&event_buffer, event.flags);

    binary_encode_uint32(&event_buffer, (uint32_t)event.end.size());
    BOOST_FOREACH(const fnr_sample_access_t &a, event.end)
	binary_encode_access(&event_buffer, &a);

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::time_mapping(const fnr_event_time_mapping_t &event)
{
    binary_encode_uint16(&event_buffer, (uint16_t)event.time_events.size());
    BOOST_FOREACH(const fnr_event_time_mapping_t::time_event_t &te,
                  event.time_events) {
        binary_encode_uint64(&event_buffer, te.global_time);
        binary_encode_uint32(&event_buffer, te.tid.id);
    }

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::local_time(const fnr_event_local_time_t &event)
{
    binary_encode_uint32(&event_buffer, event.tid.id);
    binary_encode_uint64(&event_buffer, event.global_time);
    binary_encode_uint64(&event_buffer, event.local_time);

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::burst_begin(const fnr_event_burst_begin_t &event)
{
    binary_encode_uint64(&event_buffer, event.real_time);
    binary_encode_uint64(&event_buffer, event.time);
    binary_encode_uint64(&event_buffer, event.sampling_time);
    binary_encode_uint64(&event_buffer, event.cooldown_time);

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::burst_end(const fnr_event_burst_end_t &event)
{
    binary_encode_uint64(&event_buffer, event.real_time);
    binary_encode_uint64(&event_buffer, event.time);

    return FLOW_OK;
}

flow_status_t
fnr_encoder_binary_t::contexts(const fnr_event_contexts_t &event)
{
    binary_encode_uint32(&event_buffer, (uint32_t)event.contexts.size());

    BOOST_FOREACH(const fnr_context_id_t &c, event.contexts)
	binary_encode_uint32(&event_buffer, c);

    return FLOW_OK;
}

END_NAMESPACE;
END_NAMESPACE;
