/// @file library/fenris/codec_binary.hh

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#ifndef FNR_CODEC_BINARY_HH
#define FNR_CODEC_BINARY_HH

#include "library/fenris/flow.hh"
#include "library/codec/buffer.hh"
#include "library/stream/stream.hh"
#include "library/warning/warning_consumer.hh"
#include "library/lang/common_macros.h"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

#define CODEC_BINARY_VERSION_CURRENT      3

/* -------------------------------------------------------------------- */
/* context and version types */

class fnr_decoder_binary_t;

struct fenris_header_t {
    uint8_t code;
    size_t size;
};

struct event_type_t {
    fnr_event_kind_t kind;
    int code;
    fnr_event_t *(fnr_decoder_binary_t::*decode)(size_t size);
};


struct fnr_codec_binary_context_t {
    fnr_codec_binary_context_t(
        warning_consumer_ptr_t warning_consumer);
    void set_version(
        int version_major,
        int version_minor,
        const string &svn_version,
        fnr_features_t features);

    warning_consumer_ptr_t warning_consumer;

    int version_major;
    int version_minor;
    string svn_version;
    fnr_features_t features;

    const event_type_t *events;

    static const event_type_t events_v0[];
    static const event_type_t events_v1[];
};

class fnr_encoder_binary_t
    : public fenris_flow_by_type_t
{
public:
    fnr_encoder_binary_t(
        stream_t *stream_next,
        warning_consumer_ptr_t warning_consumer);

    flow_status_t flow_init(fnr_features_t features);
    flow_status_t process_event(const fnr_event_t &event);

private:
    void write_event(int code);

    void version_select();

    flow_status_t read(const fnr_event_read_t &event);
    flow_status_t write(const fnr_event_write_t &event);
    flow_status_t prefetch(const fnr_event_prefetch_t &event);
    flow_status_t sample_begin(const fnr_event_sample_begin_t &event);
    flow_status_t sample_end(const fnr_event_sample_end_t &event);
    flow_status_t sample_period(const fnr_event_sample_period_t &event);
    flow_status_t acquisition_begin(const fnr_event_acquisition_begin_t &event);
    flow_status_t acquisition_end(const fnr_event_acquisition_end_t &event);
    flow_status_t mmap(const fnr_event_mmap_t &event);
    flow_status_t munmap(const fnr_event_munmap_t &event);
    flow_status_t stack(const fnr_event_stack_t &event);
    flow_status_t sample(const fnr_event_sample_t &event);
    flow_status_t time_mapping(const fnr_event_time_mapping_t &event);
    flow_status_t local_time(const fnr_event_local_time_t &event);
    flow_status_t burst_begin(const fnr_event_burst_begin_t &event);
    flow_status_t burst_end(const fnr_event_burst_end_t &event);
    flow_status_t contexts(const fnr_event_contexts_t &event);

    stream_t *stream_next;
    warning_consumer_ptr_t warning_consumer;
    codec_buffer_t _buffer;
    codec_buffer_t event_buffer;
    fnr_codec_binary_context_t context;
    bool seen_flow_init;
};

class fnr_decoder_binary_t
    : public stream_t
{
    friend struct fnr_codec_binary_context_t;

public:
    fnr_decoder_binary_t(
        fenris_flow_t *flow_next,
        warning_consumer_ptr_t warning_consumer);

    stream_status_t process_data(const char *buffer, size_t len);
    stream_status_t flush();

private:
    fnr_event_t *version_select(size_t size);

    fnr_event_t *event_read(size_t size);
    fnr_event_t *event_write(size_t size);
    fnr_event_t *event_prefetch(size_t size);
    fnr_event_t *event_sample_begin(size_t size);
    fnr_event_t *event_sample_end(size_t size);
    fnr_event_t *event_sample_period(size_t size);
    fnr_event_t *event_acquisition_begin(size_t size);
    fnr_event_t *event_acquisition_end(size_t size);
    fnr_event_t *event_mmap(size_t size);
    fnr_event_t *event_munmap(size_t size);
    fnr_event_t *event_stack(size_t size);
    fnr_event_t *event_sample(size_t size);
    fnr_event_t *event_sample_deprecated(size_t size);
    fnr_event_t *event_dangling_deprecated(size_t size);
    fnr_event_t *event_time_mapping(size_t size);
    fnr_event_t *event_local_time(size_t size);
    fnr_event_t *event_burst_begin(size_t size);
    fnr_event_t *event_burst_end(size_t size);
    fnr_event_t *event_contexts(size_t size);

    void initialize_flow();

    fnr_codec_binary_context_t context;

    codec_buffer_t buffer;

    fenris_flow_t *flow_next;
    warning_consumer_ptr_t warning_consumer;

    bool flow_initialized;

    bool waiting_for_header;
    fenris_header_t header;
};


END_NAMESPACE;
END_NAMESPACE;

#endif
