/// @file library/fenris/fenris.hh

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#ifndef FNR_FENRIS_HH
#define FNR_FENRIS_HH

#include <stdio.h>
#include <boost/shared_ptr.hpp>

#include "library/lang/common_macros.h"
#include "fnr_types.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

using boost::shared_ptr;

const char *get_access_type_str(fnr_access_type_t access_type);

/* ---------------------------------------------------------------- */
/* event types                                                      */
/* ---------------------------------------------------------------- */

typedef enum {
    FNR_Event_Read,
    FNR_Event_Write,
    FNR_Event_Prefetch,
    FNR_Event_Sample_Begin,
    FNR_Event_Sample_End,
    FNR_Event_Sample_Period,
    FNR_Event_Acquisition_Begin,
    FNR_Event_Acquisition_End,
    FNR_Event_Mmap,
    FNR_Event_Munmap,
    FNR_Event_Stack,
    FNR_Event_Sample,
    FNR_Event_Time_Mapping,
    FNR_Event_Local_Time,
    FNR_Event_Burst_Begin,
    FNR_Event_Burst_End,
    FNR_Event_Contexts,
} fnr_event_kind_t;

struct fnr_event_t {
    virtual ~fnr_event_t() {}

    virtual fnr_event_kind_t get_kind() const = 0;
    virtual ostream &stream_put(ostream &os) const = 0;
    virtual bool operator==(const fnr_event_t &rhs) const = 0;

    inline bool operator!=(const fnr_event_t &rhs) const {
	return !(*this == rhs);
    }
};

struct fnr_event_read_t
    : public fnr_event_t
{
    fnr_event_read_t()
	: fnr_event_t() {}

    fnr_event_read_t(const fnr_address_t &pc,
		     const fnr_address_t &address,
		     const fnr_thread_id_t thread_id,
		     fnr_access_size_t size)
	: fnr_event_t(),
	  pc(pc), address(address), thread_id(thread_id), size(size) {}

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Read;
    }
    ostream &stream_put(ostream &os) const;
    bool operator==(const fnr_event_t &rhs) const;

    fnr_address_t     pc;
    fnr_address_t     address;
    fnr_thread_id_t   thread_id;
    fnr_access_size_t size;
};

struct fnr_event_write_t
    : public fnr_event_t
{
    fnr_event_write_t()
	: fnr_event_t() {}

    fnr_event_write_t(const fnr_address_t &pc,
		      const fnr_address_t &address,
		      const fnr_thread_id_t thread_id,
		      const fnr_access_size_t size)
	: fnr_event_t(),
	  pc(pc), address(address), thread_id(thread_id), size(size) {}

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Write;
    }
    ostream &stream_put(ostream &os) const;
    bool operator==(const fnr_event_t &rhs) const;

    fnr_address_t     pc;
    fnr_address_t     address;
    fnr_thread_id_t   thread_id;
    fnr_access_size_t size;
};

struct fnr_event_prefetch_t
    : public fnr_event_t
{
    fnr_event_prefetch_t()
	: fnr_event_t() {}

    fnr_event_prefetch_t(const fnr_address_t &pc,
			 const fnr_address_t &address,
			 const fnr_thread_id_t thread_id,
			 const fnr_access_size_t size)
	: fnr_event_t(),
	  pc(pc), address(address), thread_id(thread_id), size(size) {}

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Prefetch;
    }
    ostream &stream_put(ostream &os) const;
    bool operator==(const fnr_event_t &rhs) const;

    fnr_address_t     pc;
    fnr_address_t     address;
    fnr_thread_id_t   thread_id;
    fnr_access_size_t size;
};

struct fnr_event_sample_begin_t
    : public fnr_event_t
{
    fnr_event_sample_begin_t()
	: fnr_event_t() {}

    fnr_event_sample_begin_t(const fnr_time_t time,
			     const fnr_thread_id_t thread_id,
			     const fnr_sample_count_t sample_id,
			     const fnr_address_t &pc,
			     const fnr_address_t &pc_plus_one,
			     const fnr_address_t &address,
			     const fnr_access_type_t kind,
			     const fnr_access_size_t access_size,
			     const fnr_line_size_t line_size)
	: fnr_event_t(),
	  time(time), thread_id(thread_id), sample_id(sample_id), pc(pc),
	  pc_plus_one(pc_plus_one), address(address), kind(kind),
	  access_size(access_size), line_size(line_size) {}

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Sample_Begin;
    }
    ostream &stream_put(ostream &os) const;
    bool operator==(const fnr_event_t &rhs) const;

    fnr_time_t              time;

    /* (thread_id, sample_id) is a globally unique sample identifier */
    fnr_thread_id_t         thread_id;
    fnr_sample_count_t      sample_id;
    fnr_address_t           pc;
    fnr_address_t           pc_plus_one;
    fnr_address_t           address;
    fnr_access_type_t       kind;
    fnr_access_size_t       access_size;

    /* We use a fnr_line_size_t in memory. However, we only consume 8
     * bits when stored in the fenris format ( log2(line_size) ).
     * line_size eq to zero (0) corresponds to a watch point size eq
     * to the access size of the first memop. Zero doesn't work with
     * log2, and hence, the special constant 255 corresponds to a
     * line_size eq to zero in the fenris format.
     */
    fnr_line_size_t         line_size;
};

struct fnr_event_sample_end_t
    : public fnr_event_t
{
    fnr_event_sample_end_t()
	: fnr_event_t() {}

    fnr_event_sample_end_t(const fnr_time_t time,
			   const fnr_thread_id_t thread_id,
			   const fnr_sample_count_t sample_id,
			   const fnr_address_t &pc,
			   const fnr_address_t &address,
			   const fnr_access_type_t kind,
			   const fnr_access_size_t access_size,
			   const fnr_time_t rdist)
	: fnr_event_t(),
	  time(time), thread_id(thread_id), sample_id(sample_id),
	  pc(pc), address(address), kind(kind), access_size(access_size),
	  rdist(rdist) {}

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Sample_End;
    }
    ostream &stream_put(ostream &os) const;
    bool operator==(const fnr_event_t &rhs) const;

    fnr_time_t              time;
    fnr_thread_id_t         thread_id;
    fnr_sample_count_t      sample_id;
    fnr_address_t           pc;
    fnr_address_t           address;
    fnr_access_type_t       kind;
    fnr_access_size_t       access_size;
    fnr_time_t              rdist;
};

struct fnr_sample_access_t {
    fnr_sample_access_t() {}

    fnr_sample_access_t(const fnr_address_t &pc,
			const fnr_address_t &addr,
			const fnr_time_t time,
			const fnr_stack_id_t stack,
			const fnr_thread_id_t thread,
			const fnr_access_size_t access_size,
			const fnr_access_type_t access_type)
	: pc(pc), addr(addr), time(time), stack(stack),
	  thread(thread), access_size(access_size),
	  access_type(access_type) {}

    inline bool operator==(const fnr_sample_access_t &rhs) const {
	return
	    pc == rhs.pc &&
	    addr == rhs.addr &&
	    time == rhs.time &&
	    stack == rhs.stack &&
	    thread == rhs.thread &&
	    access_size == rhs.access_size &&
	    access_type == rhs.access_type;
    }

    /**
     * Is the access a read access?
     *
     * \note is_read and is_write are not mutually exclusive, an access
     * may be both a read and a write.
     *
     * \param include_pf Count read prefetches as reads if true.
     */
    bool is_read(bool include_pf = true) const;

    /**
     * Is the access a write access?
     *
     * \note is_read and is_write are not mutually exclusive, an access
     * may be both a read and a write.
     *
     * \param include_pf Count write prefetches as writes if true.
     */
    bool is_write(bool include_pf = true) const;

    /// True if the access is a prefetch.
    bool is_prefetch() const;

    /// True if the access actually reads or writes data.
    bool is_proper() const;

    inline fnr_access_classes_t access_classes() const {
	return get_access_classes(access_type);
    }

    inline fnr_prefetch_classes_t prefetch_classes() const {
	return get_prefetch_classes(access_type);
    }

    fnr_address_t           pc;              /* 8 bytes */
    fnr_address_t           addr;            /* 8 */
    fnr_time_t		    time;            /* 8 */
    fnr_stack_id_t          stack;           /* 4 */
    fnr_thread_id_t         thread;          /* 4 */
    fnr_access_size_t       access_size;     /* 2 */
    fnr_access_type_t       access_type;     /* 1 */
}; /* 35 bytes */

typedef vector<fnr_sample_access_t> fnr_sample_access_vector_t;

struct fnr_event_sample_t
    : public fnr_event_t
{
    fnr_event_sample_t()
	: fnr_event_t() {}

    fnr_event_sample_t(const fnr_sample_access_t &begin,
		       const fnr_sample_access_vector_t &end,
		       const fnr_address_t &pc_1_plus,
		       const fnr_stack_id_t pc_1_plus_stack,
		       const fnr_line_size_log2_t line_size_log2,
		       const fnr_sample_flags_t flags)
	: fnr_event_t(),
	  begin(begin), end(end), pc_1_plus(pc_1_plus),
	  pc_1_plus_stack(pc_1_plus_stack), line_size_log2(line_size_log2),
	  flags(flags) {}

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Sample;
    }
    ostream &stream_put(ostream &os) const;
    bool operator==(const fnr_event_t &rhs) const;

    /// Get the reuse distance
    inline fnr_time_t rdist() const {
	assert(end.begin() != end.end());
	return end[0].time - begin.time - 1;
    }

    fnr_sample_access_t        begin;
    fnr_sample_access_vector_t end;

    fnr_address_t              pc_1_plus;
    fnr_stack_id_t             pc_1_plus_stack;

    fnr_line_size_log2_t       line_size_log2;

    fnr_sample_flags_t         flags;
};

typedef shared_ptr<fnr_event_sample_t> sample_ptr_t;
typedef vector<sample_ptr_t> sample_vector_t;

struct fnr_event_sample_period_t
    : public fnr_event_t
{
    fnr_event_sample_period_t()
	: fnr_event_t() {}

    fnr_event_sample_period_t(const fnr_time_t sample_period)
	: fnr_event_t(),
          sample_period(sample_period) {
        /* Only 32 bits are currently encoded by the codec. */
        assert(sample_period == (uint32_t)sample_period);
    }

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Sample_Period;
    }
    ostream &stream_put(ostream &os) const;
    bool operator==(const fnr_event_t &rhs) const;

    /// Sample period. 64-bit type, but only 32 bits are currently used by the
    /// codec.
    fnr_time_t              sample_period;
};

typedef vector<string> fnr_argv_t;

/* Acquisition begin event. */
struct fnr_event_acquisition_begin_t
    : public fnr_event_t
{
    fnr_event_acquisition_begin_t()
	: fnr_event_t() {}

    fnr_event_acquisition_begin_t(const fnr_real_time_t time,
				  const fnr_architecture_t architecture,
				  const fnr_argv_t &argv,
				  const char *user,
				  fnr_line_size_log2_vector_t line_sizes_log2)
	: fnr_event_t(),
	  time(time), architecture(architecture), argv(argv), user(user),
	  line_sizes_log2(line_sizes_log2) {}

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Acquisition_Begin;
    }
    ostream &stream_put(ostream &os) const;
    bool operator==(const fnr_event_t &rhs) const;

    fnr_real_time_t time;
    fnr_architecture_t architecture;
    fnr_argv_t argv;
    string user;
    fnr_line_size_log2_vector_t line_sizes_log2;
};

/* Acquisition end event. */
struct fnr_event_acquisition_end_t
    : public fnr_event_t
{
    fnr_event_acquisition_end_t()
	: fnr_event_t() {}

    fnr_event_acquisition_end_t(const fnr_real_time_t &time)
	: fnr_event_t(),
          time(time) {}

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Acquisition_End;
    }
    ostream &stream_put(ostream &os) const;
    bool operator==(const fnr_event_t &rhs) const;

    fnr_real_time_t time;
};

/* Mmap event. */
struct fnr_event_mmap_t
    : public fnr_event_t
{
    fnr_event_mmap_t()
	: fnr_event_t() {}

    fnr_event_mmap_t(const uint64_t address,
		     const uint64_t size,
		     const uint32_t attr,
		     const uint64_t device,
		     const uint64_t inode,
		     const uint64_t offset,
		     const char *filename,
		     const fnr_time_t global_time)
	: fnr_event_t(),
	  address(address), size(size),
	  attr(attr), device(device), inode(inode),
	  filename(filename), offset(offset),
	  global_time(global_time) {}

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Mmap;
    }
    ostream &stream_put(ostream &os) const;
    bool operator==(const fnr_event_t &rhs) const;

    uint64_t address;
    uint64_t size;
    fnr_mmap_attr_t attr;
    uint64_t device;
    uint64_t inode;
    string filename;
    uint64_t offset;
    fnr_time_t global_time;
};

/* Munmap event. */
struct fnr_event_munmap_t
    : public fnr_event_t
{
    fnr_event_munmap_t()
	: fnr_event_t() {}

    fnr_event_munmap_t(const uint64_t address,
		       const uint64_t size,
		       const fnr_time_t global_time)
	: fnr_event_t(),
	  address(address), size(size),
	  global_time(global_time) {}

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Munmap;
    }
    ostream &stream_put(ostream &os) const;
    bool operator==(const fnr_event_t &rhs) const;

    uint64_t address;
    uint64_t size;
    fnr_time_t global_time;
};

/* Stack event. */
struct fnr_event_stack_t
    : public fnr_event_t
{
    fnr_event_stack_t()
	: fnr_event_t() {}

    fnr_event_stack_t(const fnr_stack_id_t id,
		      fnr_address_vector_t ret_addrs)
	: fnr_event_t(),
	  id(id), ret_addrs(ret_addrs) {}

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Stack;
    }
    ostream &stream_put(ostream &os) const;
    bool operator==(const fnr_event_t &rhs) const;

    fnr_stack_id_t id;
    /* Return addresses, outermost first. */
    fnr_address_vector_t ret_addrs;
};

/// Time mapping event.
///
/// Legacy event. No longer generated, but may be found in old sample files.
struct fnr_event_time_mapping_t
    : public fnr_event_t
{
    fnr_event_time_mapping_t()
	: fnr_event_t() {}

    ostream &stream_put(ostream &os) const;

    bool operator==(const fnr_event_t &rhs) const;

    struct time_event_t {
        time_event_t(fnr_time_t global_time,
                     fnr_thread_id_t tid)
            : global_time(global_time),
              tid(tid) {}

        fnr_time_t global_time;
        fnr_thread_id_t tid;
    };

    typedef vector<time_event_t> time_event_vector_t;

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Time_Mapping;
    }

    time_event_vector_t time_events;
};

/// Local time mapping event.
///
/// Tells the local time of a thread at a particular global time.
struct fnr_event_local_time_t
    : public fnr_event_t
{
    fnr_event_local_time_t(fnr_thread_id_t tid,
                           fnr_time_t global_time,
                           fnr_time_t local_time)
        : tid(tid),
          global_time(global_time),
          local_time(local_time)
        {}

    ostream &stream_put(ostream &os) const;

    bool operator==(const fnr_event_t &rhs) const;

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Local_Time;
    }

    const fnr_thread_id_t tid;
    const fnr_time_t global_time;
    const fnr_time_t local_time;
};

struct fnr_event_burst_begin_t
    : public fnr_event_t
{
    fnr_event_burst_begin_t()
	: fnr_event_t() {}

    fnr_event_burst_begin_t(
        fnr_real_time_t real_time,
        fnr_time_t time,
        fnr_time_t sampling_time,
        fnr_time_t cooldown_time)
	: fnr_event_t(),
          real_time(real_time),
          time(time),
          sampling_time(sampling_time),
          cooldown_time(cooldown_time)
        {}

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Burst_Begin;
    }
    ostream &stream_put(ostream &os) const;
    bool operator==(const fnr_event_t &rhs) const;

    /// Wall clock time when the burst started, or 0 if unknown;
    fnr_real_time_t real_time;

    /// Global memory access count when the burst started. Should currently
    /// always be 0.
    fnr_time_t time;

    /// The length of the sampling period, or 0 if unknown.
    fnr_time_t sampling_time;

    /// The length of the cooldown period, or 0 if unknown.
    fnr_time_t cooldown_time;
};

struct fnr_event_burst_end_t
    : public fnr_event_t
{
    fnr_event_burst_end_t()
	: fnr_event_t() {}

    fnr_event_burst_end_t(fnr_real_time_t real_time, fnr_time_t time)
	: fnr_event_t(),
          real_time(real_time), time(time) {}

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Burst_End;
    }
    ostream &stream_put(ostream &os) const;
    bool operator==(const fnr_event_t &rhs) const;

    /// Wall clock time when the burst ended, or 0 if unknown;
    fnr_real_time_t real_time;

    /// Global memory access count when the burst ended.
    fnr_time_t time;
};

struct fnr_event_contexts_t
    : public fnr_event_t
{
    fnr_event_contexts_t()
	: fnr_event_t() {}

    fnr_event_contexts_t(const fnr_context_id_vector_t &contexts)
	: fnr_event_t(),
          contexts(contexts) {}

    inline fnr_event_kind_t get_kind() const {
	return FNR_Event_Contexts;
    }
    ostream &stream_put(ostream &os) const;
    bool operator==(const fnr_event_t &rhs) const;

    fnr_context_id_vector_t contexts;
};

/* ---------------------------------------------------------------------- */

/* Convert an fnr_line_size_log2_t to an fnr_line_size_t. */
static inline fnr_line_size_t
line_size_from_log2(fnr_line_size_log2_t line_size_log2)
{
    if (line_size_log2 == 0xff)
        return 0;

    return (fnr_line_size_t)1 << line_size_log2;
}

/* Convert an fnr_line_size_t to an fnr_line_size_log2_t. */
static inline fnr_line_size_log2_t
line_size_to_log2(fnr_line_size_t line_size)
{
    if (line_size == 0)
        return 0xff;

    assert((line_size & (line_size - 1)) == 0);

    fnr_line_size_log2_t line_size_log2 = 0;
    while (line_size != 1) {
        line_size >>= 1;
        line_size_log2 += 1;
    }
    return line_size_log2;
}

/* ------------------------------------------------------------------------- */

ostream &operator<<(ostream &os, const fnr_event_t &event);
ostream &operator<<(ostream &os, const fnr_sample_access_t &event);

typedef shared_ptr<fnr_event_t> event_ptr_t;
typedef shared_ptr<fnr_event_read_t> event_read_ptr_t;
typedef shared_ptr<fnr_event_write_t> event_write_ptr_t;
typedef shared_ptr<fnr_event_prefetch_t> event_prefetch_ptr_t;
typedef shared_ptr<fnr_event_sample_begin_t> event_sample_begin_ptr_t;
typedef shared_ptr<fnr_event_sample_end_t> event_sample_end_ptr_t;
typedef shared_ptr<fnr_event_sample_period_t> event_sample_period_ptr_t;
typedef shared_ptr<fnr_event_acquisition_begin_t> event_acquisition_begin_ptr_t;
typedef shared_ptr<fnr_event_acquisition_end_t> event_acquisition_end_ptr_t;
typedef shared_ptr<fnr_event_mmap_t> event_mmap_ptr_t;
typedef shared_ptr<fnr_event_stack_t> event_stack_ptr_t;
typedef shared_ptr<fnr_event_sample_t> event_sample_ptr_t;
typedef shared_ptr<fnr_event_time_mapping_t> event_time_mapping_ptr_t;
typedef shared_ptr<fnr_event_local_time_t> event_local_time_ptr_t;
typedef shared_ptr<fnr_event_burst_begin_t> event_burst_begin_ptr_t;
typedef shared_ptr<fnr_event_burst_end_t> event_burst_end_ptr_t;

typedef vector<event_sample_ptr_t> sample_vector_t;
typedef vector<event_time_mapping_ptr_t> time_mapping_vector_t;
typedef vector<event_local_time_ptr_t> local_time_vector_t;

END_NAMESPACE;
END_NAMESPACE;

#endif
