/// @file library/fenris/flow.hh

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#ifndef FNR_FLOW_HH
#define FNR_FLOW_HH

#include "library/lang/common_macros.h"
#include "library/lang/common_types.h"
#include "fenris.hh"

#include <stdexcept>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

class flow_exception_t
    : public std::runtime_error {
public:
    flow_exception_t(const string &what_)
        : std::runtime_error(what_)
        {}
};

typedef enum {
    /// Flow state is OK
    FLOW_OK = 0,
    /// Abort was requested by a Flow node
    FLOW_ABORT,
    /// A Flow node returned a fatal error
    FLOW_ERROR
} flow_status_t;

class fenris_flow_t {
public:
    virtual ~fenris_flow_t() {}

    /// Initialize a flow. Called before any events arrive.
    virtual flow_status_t flow_init(fnr_features_t features) = 0;

    /// Process a Fenris event.
    virtual flow_status_t process_event(const fnr_event_t &event) = 0;
};

class fenris_flow_by_type_t
    : public fenris_flow_t {
public:
    virtual flow_status_t flow_init(fnr_features_t features) {
	return FLOW_OK;
    }

    /// Up cast the event and call an event specific function.
    virtual flow_status_t process_event(const fnr_event_t &event);

protected:
    virtual flow_status_t read(const fnr_event_read_t &event) = 0;
    virtual flow_status_t write(const fnr_event_write_t &event) = 0;
    virtual flow_status_t prefetch(const fnr_event_prefetch_t &event) = 0;
    virtual flow_status_t sample_begin(const fnr_event_sample_begin_t &event) = 0;
    virtual flow_status_t sample_end(const fnr_event_sample_end_t &event) = 0;
    virtual flow_status_t sample_period(const fnr_event_sample_period_t &event) = 0;
    virtual flow_status_t acquisition_begin(const fnr_event_acquisition_begin_t &event) = 0;
    virtual flow_status_t acquisition_end(const fnr_event_acquisition_end_t &event) = 0;
    virtual flow_status_t mmap(const fnr_event_mmap_t &event) = 0;
    virtual flow_status_t munmap(const fnr_event_munmap_t &event) = 0;
    virtual flow_status_t stack(const fnr_event_stack_t &event) = 0;
    virtual flow_status_t sample(const fnr_event_sample_t &event) = 0;
    virtual flow_status_t time_mapping(const fnr_event_time_mapping_t &event) = 0;
    virtual flow_status_t local_time(const fnr_event_local_time_t &event) = 0;
    virtual flow_status_t burst_begin(const fnr_event_burst_begin_t &event) = 0;
    virtual flow_status_t burst_end(const fnr_event_burst_end_t &event) = 0;
    virtual flow_status_t contexts(const fnr_event_contexts_t &event) = 0;
};

class fenris_flow_by_type_default_t
    : public fenris_flow_by_type_t {
protected:
    virtual flow_status_t read(const fnr_event_read_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t write(const fnr_event_write_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t prefetch(const fnr_event_prefetch_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t sample_begin(const fnr_event_sample_begin_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t sample_end(const fnr_event_sample_end_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t sample_period(const fnr_event_sample_period_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t acquisition_begin(const fnr_event_acquisition_begin_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t acquisition_end(const fnr_event_acquisition_end_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t mmap(const fnr_event_mmap_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t munmap(const fnr_event_munmap_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t stack(const fnr_event_stack_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t sample(const fnr_event_sample_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t time_mapping(const fnr_event_time_mapping_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t local_time(const fnr_event_local_time_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t burst_begin(const fnr_event_burst_begin_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t burst_end(const fnr_event_burst_end_t &event) {
	return FLOW_OK;
    }
    virtual flow_status_t contexts(const fnr_event_contexts_t &event) {
	return FLOW_OK;
    }
};

class fenris_flow_filter_t
    : public fenris_flow_by_type_t
{
public:
    fenris_flow_filter_t(fenris_flow_t *next);
    virtual ~fenris_flow_filter_t();

    virtual flow_status_t flow_init(fnr_features_t features);

protected:
    flow_status_t flow_to_next(const fnr_event_t &event);

    virtual flow_status_t read(const fnr_event_read_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t write(const fnr_event_write_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t prefetch(const fnr_event_prefetch_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t sample_begin(const fnr_event_sample_begin_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t sample_end(const fnr_event_sample_end_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t sample_period(const fnr_event_sample_period_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t acquisition_begin(const fnr_event_acquisition_begin_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t acquisition_end(const fnr_event_acquisition_end_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t mmap(const fnr_event_mmap_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t munmap(const fnr_event_munmap_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t stack(const fnr_event_stack_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t sample(const fnr_event_sample_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t time_mapping(const fnr_event_time_mapping_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t local_time(const fnr_event_local_time_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t burst_begin(const fnr_event_burst_begin_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t burst_end(const fnr_event_burst_end_t &event) {
	return flow_to_next(event);
    }
    virtual flow_status_t contexts(const fnr_event_contexts_t &event) {
	return flow_to_next(event);
    }

    fenris_flow_t *next;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
