/// @file library/fenris/fnr_types.cc

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#include "fnr_types.hh"

#include <sstream>
#include <ostream>
#include <iomanip>

#include <boost/format.hpp>
#include <boost/foreach.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;
using namespace boost;

struct name_prefetch_class_t {
    const char *name;
    const fnr_prefetch_classes_t pf_class;
};

/* Remember to update get_access_classes when updating this */
/* Prefetches should, in moost cases, be handled like reads/writes,
 * add set the read/write flag based on the effect on the cache
 * line. E.g. prefetchw should be considered a write on the cache line
 * level since it invalidates all other copies of the cache line and
 * sets the local state to M.
 */
fnr_access_classes_t fnr_access_type_class_map[] = {
    FNR_Access_Class_RD,                        // FNR_Access_Type_RD
    FNR_Access_Class_WR,                        // FNR_Access_Type_WR
    FNR_Access_Class_PF | FNR_Access_Class_RD,  // FNR_Access_Type_PF
    FNR_Access_Class_RD | FNR_Access_Class_WR,  // FNR_Access_Type_RDWR
    FNR_Access_Class_PF | FNR_Access_Class_RD,  // FNR_Access_Type_PF_AMD
    FNR_Access_Class_PF | FNR_Access_Class_WR,  // FNR_Access_Type_PF_AMD_W
    FNR_Access_Class_PF | FNR_Access_Class_RD |
    FNR_Access_Class_NTA,                       // FNR_Access_Type_PF_Intel_NTA
    FNR_Access_Class_PF | FNR_Access_Class_RD,  // FNR_Access_Type_PF_Intel_T0
    FNR_Access_Class_PF | FNR_Access_Class_RD,  // FNR_Access_Type_PF_Intel_T1
    FNR_Access_Class_PF | FNR_Access_Class_RD,  // FNR_Access_Type_PF_Intel_T2
    FNR_Access_Class_Flush,                     // FNR_Access_Type_CLFLUSH
    FNR_Access_Class_NTA | FNR_Access_Class_RD, // FNR_Access_Type_NTA_RD
    FNR_Access_Class_NTA | FNR_Access_Class_WR  // FNR_Access_Type_NTA_WR
};

/* Remember to update get_prefetch_classes when updating this */
fnr_prefetch_classes_t fnr_access_type_pf_class_map[] = {
    FNR_Prefetch_Class_None,                    // FNR_Access_Type_RD
    FNR_Prefetch_Class_None,                    // FNR_Access_Type_WR
    FNR_Prefetch_Class_Unknown,                 // FNR_Access_Type_PF
    FNR_Prefetch_Class_None,                    // FNR_Access_Type_RDWR
    FNR_Prefetch_Class_AMD,                     // FNR_Access_Type_PF_AMD
    FNR_Prefetch_Class_AMD_W,                   // FNR_Access_Type_PF_AMD_W
    FNR_Prefetch_Class_Intel_NTA,               // FNR_Access_Type_PF_Intel_NTA
    FNR_Prefetch_Class_Intel_T0,                // FNR_Access_Type_PF_Intel_T0
    FNR_Prefetch_Class_Intel_T1,                // FNR_Access_Type_PF_Intel_T1
    FNR_Prefetch_Class_Intel_T2,                // FNR_Access_Type_PF_Intel_T2
    FNR_Prefetch_Class_None,                    // FNR_Access_Type_CLFLUSH
    FNR_Prefetch_Class_None,                    // FNR_Access_Type_NTA_RD
    FNR_Prefetch_Class_None                     // FNR_Access_Type_NTA_WR
};

const name_prefetch_class_t prefetch_names[] = {
    { "prefetch", FNR_Prefetch_Class_AMD },
    { "prefetchw", FNR_Prefetch_Class_AMD_W },
    { "prefetchnta", FNR_Prefetch_Class_Intel_NTA },
    { "prefetcht0", FNR_Prefetch_Class_Intel_T0 },
    { "prefetcht1", FNR_Prefetch_Class_Intel_T1 },
    { "prefetcht2", FNR_Prefetch_Class_Intel_T2 },
    { NULL, FNR_Prefetch_Class_None }
};

uint64_t
fnr_parse_num_string(const char *s)
{
    uint64_t result = 0;
    assert(s != 0);

    while (*s && isdigit(*s)) {
        result *= 10;
        result += (*s - '0');
        s++;
    }

    if (*s == 'k' || *s == 'K')
        result *= 1024;
    else if (*s == 'm' || *s == 'M')
        result *= 1024 * 1024;
    else if (*s == 'g' || *s == 'G')
        result *= 1024 * 1024 * 1024;
    else if (*s == 't' || *s == 'T')
        result *= 1024LL * 1024LL * 1024LL * 1024LL;
    else if (*s == 'p' || *s == 'P')
        result *= 1024LL * 1024LL * 1024LL * 1024LL * 1024LL;

    return result;
}

static unsigned int
log2floor(uint64_t val)
{
    unsigned int l2;

    for(l2 = 0; val; l2++)
	val = val >> 1;
    l2--;

    return l2;
}

std::string
fnr_format_num_string(uint64_t size)
{
    static const char *suffixes[] = {"B", "kB", "MB", "GB", "TB", "PB", "EB"};
    std::ostringstream oss;

    if (size == 0) {
	oss << size;
    } else {
	unsigned int l2 = log2floor(size);
	unsigned int s = l2 / 10;

	oss << ((double)size) / (1 << (s * 10)) << " " << suffixes[s];
    }

    return oss.str();
}


fnr_prefetch_classes_t
fnr_prefetch_class_by_name(const char *s)
{
    if (!s)
	return FNR_Prefetch_Class_None;

    for (const name_prefetch_class_t *pf = prefetch_names; pf->name; pf++) {
	if (strcmp(pf->name, s) == 0)
	    return pf->pf_class;
    }

    return FNR_Prefetch_Class_None;
}


bool
fnr_access_filter_any(fnr_access_classes_t ac)
{
    return true;
}

bool
fnr_access_filter_rd(fnr_access_classes_t ac)
{
    return (ac & FNR_Access_Class_RD) &&
	!(ac & FNR_Access_Class_PF);
}

bool
fnr_access_filter_wr(fnr_access_classes_t ac)
{
    return (ac & FNR_Access_Class_WR) &&
	!(ac & FNR_Access_Class_PF) &&
	!(ac & FNR_Access_Class_NTA);
}

bool
fnr_access_filter_wr_wrnta(fnr_access_classes_t ac)
{
    return (ac & FNR_Access_Class_WR) &&
	!(ac & FNR_Access_Class_PF);
}

bool
fnr_access_filter_rw(fnr_access_classes_t ac)
{
    return fnr_access_filter_rd(ac) || fnr_access_filter_wr(ac);
}

bool
fnr_access_filter_pf(fnr_access_classes_t ac)
{
    return (ac & FNR_Access_Class_PF) != 0;
}


ostream &
operator <<(ostream &os, const fnr_address_t &a)
{
    os << format("0x%x") % a.virtual_address;

    return os;
}

ostream &
operator <<(ostream &os, const fnr_address_set_t &s)
{
    os << "(";
    bool first = true;
    BOOST_FOREACH(const fnr_address_t &i, s) {
	if (first)
	    first = false;
	else
	    os << ",";

	os << i;
    }
    os << ")";

    return os;
}

ostream &
operator <<(ostream &os, const fnr_address_vector_t &s)
{
    os << "(";
    bool first = true;
    BOOST_FOREACH(const fnr_address_t &i, s) {
	if (first)
	    first = false;
	else
	    os << ",";

	os << i;
    }
    os << ")";

    return os;
}

ostream &
operator <<(ostream &os, const pad_t<fnr_address_t> &a)
{
    os << format("0x%016x") % a.a.virtual_address;

    return os;
}

ostream &
operator <<(ostream &os, const pad_t<fnr_address_set_t> &s)
{
    os << "(";
    bool first = true;
    BOOST_FOREACH(const fnr_address_t &i, s.a) {
	if (first)
	    first = false;
	else
	    os << ",";

	os << pad(i);
    }
    os << ")";

    return os;
}

ostream &
operator <<(ostream &os, const pad_t<fnr_address_vector_t> &s)
{
    os << "(";
    bool first = true;
    BOOST_FOREACH(const fnr_address_t &i, s.a) {
	if (first)
	    first = false;
	else
	    os << ",";

	os << pad(i);
    }
    os << ")";

    return os;
}

ostream &
operator <<(ostream &os, const fnr_thread_id_t &thread)
{
    os << thread.id;
    return os;
}

END_NAMESPACE;
END_NAMESPACE;
