/// @file library/fenris/fnr_types.hh

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#ifndef FNR_TYPES_HH
#define FNR_TYPES_HH

#include "library/lang/common_macros.h"
#include "library/lang/common_types.h"

#include <boost/shared_ptr.hpp>

#include <ostream>
#include <string>
#include <list>
#include <map>
#include <set>
#include <vector>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

/// Bitfield signifying available sample file features
typedef uint64_t fnr_features_t;

#define FNR_Feature_None                  	0

/// MT Feature available in sample file
#define FNR_Feature_MT              		(1 << 0)

/// Sample file contains virtual thread IDs
#define FNR_Feature_Virtual_Threads		(1 << 1)

/// Temporary file, used by atf_resampler
#define FNR_Feature_Temporary			(1 << 2)



/// A thread id as used by the operating system. Operating systems may reuse
/// the same id for different threads.
typedef uint32_t fnr_context_id_t;

typedef vector<fnr_context_id_t> fnr_context_id_vector_t;

/// Analysis internal thread id. Every thread is assigned a unique internal
/// thread id, even if the operating system reuses thread ids.
struct fnr_thread_id_t {
    uint32_t id;

    fnr_thread_id_t() {}

    fnr_thread_id_t(uint32_t id)
        : id(id)
        {}

    bool operator==(const fnr_thread_id_t &rhs) const {
	return id == rhs.id;
    }

    bool operator!=(const fnr_thread_id_t &rhs) const {
	return id != rhs.id;
    }

    bool operator<(const fnr_thread_id_t &rhs) const {
	return id < rhs.id;
    }

    bool operator>(const fnr_thread_id_t &rhs) const {
	return id > rhs.id;
    }
};

ostream &operator<<(ostream &os, const fnr_thread_id_t &thread);

/// Access size.
typedef uint16_t fnr_access_size_t;

/// typically quite large
typedef uint64_t fnr_cache_size_t;

/// typically not so large, but when applied to TLB:s, the size may
/// get large for some systems
typedef uint64_t fnr_line_size_t;

/// Used both for numbering and counting cache lines in the cache and for
/// numbering and counting cache lines in memory. In the later case the count
/// may exceed what fits in 32 bits. */
typedef uint64_t fnr_line_count_t;

/// Memory access count. Used for sampler time, global time and thread time.
typedef uint64_t fnr_time_t;

#define FNR_TIME_MAX ((fnr_time_t)-1)

typedef set<fnr_time_t> fnr_time_set_t;
typedef vector<fnr_time_t> fnr_time_vector_t;
typedef map<fnr_time_t, fnr_time_t> fnr_time_map_t;

/// Used both for numbering (sample ids) and counting samples.
typedef uint32_t fnr_sample_count_t;

// Depth of a call stack.
typedef uint16_t fnr_stack_depth_t;

/// Instead of repeating a certain call stack every time it occurs we describe
/// each call stack once and assign it number, and then refer to it by that
/// number. 0 means no call stack is available.
typedef uint32_t fnr_stack_id_t;

/// Sample flags, see below.
typedef uint8_t fnr_sample_flags_t;

/// The sample was (prematurely) killed because of too many false
/// watchpoint filter hits or too many logged accesses.
#define FNR_Sample_Flag_Killed 0x01

/// This is the address
struct fnr_address_t {
    fnr_address_t(const uint64_t virtual_address = 0)
	: virtual_address(virtual_address) {}

    uint64_t virtual_address; /* virtual address from the processor */
};

typedef set<fnr_address_t> fnr_address_set_t;
typedef list<fnr_address_t> fnr_address_list_t;
typedef vector<fnr_address_t> fnr_address_vector_t;

/// Real-time time. Number of nano-seconds since 1970-01-01 00:00 UTC.
typedef uint64_t fnr_real_time_t;

/// Access types of fnr_access_type_t.
///
/// FNR_Access_Type_WR was used for both write and read-write access in older
/// sample files. New sample files should use FNR_Access_Type_RDWR for
/// read-write accesses.
///
/// FNR_Access_Type_PF is the generic prefetch type used before we started
/// saving the prefetch type. New sample files should use the more specific
/// prefetch types.
typedef uint8_t fnr_access_type_t;

#define FNR_Access_Type_RD             0
#define FNR_Access_Type_WR             1
#define FNR_Access_Type_PF             2
#define FNR_Access_Type_RDWR           3
#define FNR_Access_Type_PF_AMD         4
#define FNR_Access_Type_PF_AMD_W       5
#define FNR_Access_Type_PF_Intel_NTA   6
#define FNR_Access_Type_PF_Intel_T0    7
#define FNR_Access_Type_PF_Intel_T1    8
#define FNR_Access_Type_PF_Intel_T2    9
#define FNR_Access_Type_CLFLUSH        10
#define FNR_Access_Type_NTA_RD         11
#define FNR_Access_Type_NTA_WR         12

/// Bit mapped prefetch classes, bit defines below.
///
/// None    - the instruction is not a prefetch
/// Unknown - prefetch of unknown type, should only occur in very old sample
///           files
typedef uint16_t fnr_prefetch_classes_t;

#define FNR_Prefetch_Class_None       0x00
#define FNR_Prefetch_Class_Unknown    (1 << 0)
#define FNR_Prefetch_Class_AMD        (1 << 1)
#define FNR_Prefetch_Class_AMD_W      (1 << 2)
#define FNR_Prefetch_Class_Intel_NTA  (1 << 3)
#define FNR_Prefetch_Class_Intel_T0   (1 << 4)
#define FNR_Prefetch_Class_Intel_T1   (1 << 5)
#define FNR_Prefetch_Class_Intel_T2   (1 << 6)
#define FNR_Prefetch_Class_Any        0xFF

/// Set of access classes, bitmask of the defines below. The access classes are
/// not mutually exlusive, i.e., one access type can belong to multiple access
/// classes. All access types belong to at least one access class.
typedef uint8_t fnr_access_classes_t;

/// Empty access class set.
#define FNR_Access_Class_None  0x00

/// Access types belonging to the RD access class:
/// * read
/// * read-write
/// * read prefetches (all but prefetchw)
/// * non-temporal read
#define FNR_Access_Class_RD    0x01

/// Access types belonging to the WR access class:
/// * write
/// * read-write
/// * write prefetch (prefetchw)
/// * non-temporal write
///
/// Note that the WR access class does not imply that the instruction actually
/// writes any data or necessarily makes the line dirty (prefetchw). However,
/// it implies that the instruction makes the line exclusive and causes an
/// invalidation of the line in sibling caches.
#define FNR_Access_Class_WR    0x02

/// All prefetch access types belong to the PF access class.
#define FNR_Access_Class_PF    0x04

/// Access types belonging to the NTA access class:
/// * non-temporal read
/// * non-temporal write
/// * non-temporal prefetch
#define FNR_Access_Class_NTA   0x08

/// Access types belonging to the NTA access class:
/// * cache line flush
#define FNR_Access_Class_Flush 0x10

/// All access classes.
#define FNR_Access_Class_Any   0xFF

/// Accesses that belong to one of these classes affect the caching of the line
/// but do not actually touch any of the data in the line.
#define FNR_Access_Classes_Pseudo                       \
    (FNR_Access_Class_PF | FNR_Access_Class_Flush)

/// 2-logarithm of line size.
typedef uint8_t fnr_line_size_log2_t;

typedef vector<fnr_line_size_log2_t> fnr_line_size_log2_vector_t;

/// Processor architectures. Encoded as uint32_t in the fenris format.
typedef enum {
    FNR_Arch_i386,
    FNR_Arch_x86_64,
    FNR_Arch_CPPemu_PPC,
    FNR_Arch_None
} fnr_architecture_t;


/// MMAP attributes
typedef uint32_t fnr_mmap_attr_t;

//@{
/// Memory protection flags
#define FNR_Mmap_Attr_Read    0x00000001
#define FNR_Mmap_Attr_Write   0x00000002
#define FNR_Mmap_Attr_Execute 0x00000004
//@}

/// Memory visible to other processes, corresponds to MAP_SHARED.
#define FNR_Mmap_Attr_Shared  0x00000008


/* Printf formats for fenris types. */
#define FNR_TIME     PRId64
#define FNR_CID      PRId32
#define FNR_ADDR_VA  ".16" PRIx64
#define FNR_ASIZE    PRId32
#define FNR_LSIZE    PRId64
#define FNR_STACK_ID PRId32
#define FNR_SCOUNT   PRId32


/* ---------------------------------------------------------------- */
/* conversion routines                                              */
/* ---------------------------------------------------------------- */
extern fnr_access_classes_t fnr_access_type_class_map[];
extern fnr_prefetch_classes_t fnr_access_type_pf_class_map[];


/// convert strings such as 64k or 4T into the corresponding uint64_t
uint64_t fnr_parse_num_string(const char s[]);
/// Create a human readable size with a suitable suffix.
std::string fnr_format_num_string(uint64_t size);

/* Return the access classes corresponding to an access type. */
static inline fnr_access_classes_t
get_access_classes(fnr_access_type_t access_type)
{
    assert(access_type <= FNR_Access_Type_NTA_WR);
    return fnr_access_type_class_map[access_type];
}

/* Return the access classes corresponding to an access type. */
static inline fnr_prefetch_classes_t
get_prefetch_classes(fnr_access_type_t access_type)
{
    assert(access_type <= FNR_Access_Type_NTA_WR);
    return fnr_access_type_pf_class_map[access_type];
}

/// Lookup a prefetch access class by name.
/// \return FNR_Prefetch_Class_None on error.
fnr_prefetch_classes_t fnr_prefetch_class_by_name(const char *s);

/* ---------------------------------------------------------------- */
/* access class filters                                             */
/* ---------------------------------------------------------------- */
typedef bool (*fnr_access_filter_t)(fnr_access_classes_t ac);

/// Any access
bool fnr_access_filter_any(fnr_access_classes_t ac);
/// Read accesses, including NTA, but excluding PF.
bool fnr_access_filter_rd(fnr_access_classes_t ac);
/// Write accesses, excluding NTA and PF.
bool fnr_access_filter_wr(fnr_access_classes_t ac);
/// Write accesses, excluding PF.
bool fnr_access_filter_wr_wrnta(fnr_access_classes_t ac);
/// fnr_access_filter_rd() || fnr_access_filter_wr()
bool fnr_access_filter_rw(fnr_access_classes_t ac);
/// Prefetch accesses
bool fnr_access_filter_pf(fnr_access_classes_t ac);


/* ---------------------------------------------------------------- */
/* operators                                                        */
/* ---------------------------------------------------------------- */

inline
bool
operator < (const fnr_address_t &a,
	    const fnr_address_t &b)
{
    return a.virtual_address < b.virtual_address;
}

inline
bool
operator > (const fnr_address_t &a,
	    const fnr_address_t &b)
{
    return a.virtual_address > b.virtual_address;
}

inline
bool
operator == (const fnr_address_t &a,
	     const fnr_address_t &b)
{
    return a.virtual_address == b.virtual_address;
}

/// Output fnr_address_t as hex.

ostream & operator <<(ostream &os,
			   const fnr_address_t &s);
ostream & operator <<(ostream &os,
			   const fnr_address_set_t &s);
ostream & operator <<(ostream &os,
			   const fnr_address_vector_t &s);

/// Output fnr_address_t as hex padded to 16 digits.

template <typename T>
struct pad_t {
    pad_t(const T &a) : a(a) {}
    const T &a;
};

template <typename T>
pad_t<T> pad(const T &a) {
    return pad_t<T>(a);
}

ostream & operator <<(ostream &os,
		      const pad_t<fnr_address_t> &s);
ostream & operator <<(ostream &os,
		      const pad_t<fnr_address_set_t> &s);
ostream & operator <<(ostream &os,
		      const pad_t<fnr_address_vector_t> &s);


END_NAMESPACE;
END_NAMESPACE;

#endif
