/// @file library/license/license.hh

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#include "library/lang/common_types.h"
#include "library/warning/warning_consumer.hh"

#include <boost/enable_shared_from_this.hpp>

#include <set>
#include <vector>
#include <string>

#ifndef LICENSE_HH
#define LICENSE_HH

// As of 2011.2, we have only one SKU: ThreadSpotter. We also make a switch
// to FlexNet, removing the internal license tools.
//
// There will be three distinct feature lines:
//   FEATURE tsseats RW
//   FEATURE tssampling RW
//   FEATURE tsview RW
//
// The Version field will map to the support expiration date.
// The Expiration field will map to the license expiration date.
// The number of Licenses will be set to different numbers for these lines
//   tsuser: x
//   tstool: y
//   tsview: z
//
// HOSTID will be used for nodelocked seats.
//
// USE_GROUP will be used to ensure that at most x users can sample at
// most y processes, and at most z users can view reports at the same
// time.
//
// Differentiating and internal features will be enumerated in the
// VENDOR_STRING.
//

#include "library/warning/warning_consumer.hh"

#include <boost/weak_ptr.hpp>

// forward declarations
struct lm_handle;
struct config;
struct lm_err_info;

BEGIN_NAMESPACE(acumem)
BEGIN_NAMESPACE(license)

using namespace std;
using namespace acumem::library;

// ----- License constants. ---------------------------------------------------

/// How long before expiration a warning will be visible.
#define LIC_WARN_PERIOD 14


// ----- exceptions

/// Capture the contents of a LM_ERROR_INFO struct
class error_info_t
{
public:
    error_info_t(const char *msg, const struct lm_err_info *err_info);
    error_info_t(int maj_errno, const string &context);

    string msg;
    int maj_errno;
    int min_errno;
    int sys_errno;
    string feature;
    vector<string> lic_files;
    string context;
};

class license_exception_t : public std::runtime_error
{
public:
    license_exception_t(const char *msg,
			const struct lm_err_info *err_info)
	: runtime_error(""), error_info(msg, err_info) {}

    license_exception_t(int maj_errno, const string &context)
	: runtime_error(""), error_info(maj_errno, context) {}

    ~license_exception_t() throw()
	{
	}

    virtual const char* what() const throw()
	{
	    return error_info.msg.c_str();
	}
private:
    error_info_t error_info;
};



/// Accumulate flexnet errors in a list
class error_reporting_t
{
public:
    void add_error_code(int maj_errno);
    void add_error(const struct lm_err_info *err_info);
private:
    typedef vector<error_info_t> error_vector_t;
    error_vector_t error_vector;
};


// ----- Wrapping flexnet api:s

/// A FlexNet opaque job handle
typedef boost::shared_ptr<lm_handle> job_ptr_t;

/// Representation of a checked out token
class token_t;
typedef boost::shared_ptr<token_t> token_ptr_t;

/// The manager of checkouts and checkins and checks
class license_manager_t;
typedef boost::shared_ptr<license_manager_t> license_manager_ptr_t;


/// The central class which manages checking out licenses, and
/// channeling callbacks from the flexnet libraries
class flexnet_wrapper_t {
public:
    flexnet_wrapper_t();

    token_ptr_t checkout(char *feature, char *version, int num);
    token_ptr_t dry_run_checkout(char *feature, char *version);
    bool is_reconnecting() const;

protected:
    virtual void reconnect_call(const char *feature,
				int pass,
				int total_attempts,
				int interval);
    virtual void reconnect_done_call(const char *feature, int tries,
				     int total_attempts, int interval);

private:
    /// Where to print errors
    warning_consumer_ptr_t warning;

    // non-copyable. If you want to share one, use smart pointers
    flexnet_wrapper_t(const flexnet_wrapper_t &);

    /// Callback when a heartbeat failed
    static void private_reconnect_call(lm_handle *, char *, int pass,
					  int total_attempts,
					  int interval, void *);

    /// Callback when heartbeat works again
    static void private_reconnect_done_call(lm_handle *, char *, int tries,
					       int total_attempts,
					       int interval, void *);

    /// Encapsulation of a lm_handle
    job_ptr_t job;

    /// Flag that says whether we have a working license server connection
    bool reconnecting;
};

/// A tag to differentiate different constructors
class test_conf_t {};

/// A token_t is a checked out token for a feature.
class token_t {
public:
    token_t(job_ptr_t job, const char *feature, bool dry_run);
    ~token_t();
    void release();
    string get_vendor_string() const;
    string get_date() const;

private:
    bool dry_run;
    bool released;
    string feature;
    job_ptr_t job;
    string vendor_string;
    string date;
};


typedef boost::weak_ptr<token_t> token_wptr_t;



class feature_set_t {
public:

    feature_set_t()
	{}

    feature_set_t(license_manager_ptr_t license,
		  token_ptr_t user_token,
		  token_ptr_t app_token)
	: license(license),
	  user_token(user_token),
	  app_token(app_token)
	{
	    read_vendor_string(app_token);
	}

    bool is_dry_run() const {
	return !user_token;
    }

    void read_vendor_string(token_ptr_t token);

    /// option codes
    bool sw_prefetch() const;
    bool unlimited_callstack() const;
    bool blocking_expert() const;
    bool attach() const;
    bool unlimited_threads() const;
    bool api() const;
    bool burst() const;
    bool mt() const;
    bool multisocket() const;
    bool static_report() const;
    bool cmd_line() const;
    bool thread_binding() const;

    bool require_manifest_check() const;
    bool check_manifest_hash(const std::vector<unsigned char> &hash) const;

    // XXX move to private? Breaks runnerlib.
    license_manager_ptr_t license;

private:
    /// Test whether the subfeature code is found in the VENDOR_STRING
    /// of the toool token
    bool has_feature(const char *subfeature) const;
    set<string> features;
    token_ptr_t user_token;
    token_ptr_t app_token;
};

typedef boost::shared_ptr<feature_set_t> feature_set_ptr_t;

/// The application instantiates one of these and uses it as a high
/// level interface to acquiring and releasing tokens.
class license_manager_t : public boost::enable_shared_from_this<license_manager_t> {
public:
    license_manager_t(const warning_consumer_ptr_t &warning);

    enum license_kind_t {
	dry_run_kind,
	user_kind,
	tool_kind,
	view_kind
    };

    /// Dry-run - see if there are suitable licenses available, but do
    /// not consume a token. Populate a feature_set based on feature tstool.
    ///
    /// throws license_exception_t in case of errors, but not if there
    /// is no token
    feature_set_ptr_t dry_run();

    /// Get a token.
    /// XXX Return actual error code
    /// XXX Consider blocking, or polling with timeout, or balking
    ///
    /// throws license_exception_t in case of errors, but not if there
    /// is no token
    feature_set_ptr_t acquire(license_kind_t license_kind);

    /// Explicitly give up the token, even if checked out and held by someone
    void release(license_kind_t license_kind);

    /// Ensure that the license token is still there. This will ensure
    /// that the user_kind weak pointer is valid and that it has
    /// current heartbeats
    bool check_license(int timeout_seconds);

    /// Return remaining time before license expires, [days]. negative
    /// values for licenses that has expired. Permanent licenses return 9999
    int remaining_days() const;

    /// Return a textual representation of the expiration date
    string expiration_date() const;

private:
    /// Warning consumer, gets all error messages
    warning_consumer_ptr_t warning;

    /// Utility function to return the compile-time as a floating number string
    static void get_compile_time(char version[10]);

    /// Get me any held token so that I can query it for some common
    /// data. It will retrieve a locked token_ptr_t corresponding to
    /// the weak_ptr of either user_kind or dry_run_kind
    token_ptr_t any_token() const;

    flexnet_wrapper_t flexnet_wrapper;

    /// A collection of weak pointers to allocated tokens. The purpose
    /// of using weak pointers here is that as long as anyone keeps a
    /// copy of their token_ptr_t alive, the token is checked out and
    /// also retrievable from here. But as soon as all token_ptr_t are
    /// released, this token_wptr_t will go stale forcing a new
    /// checkout.
    typedef map<license_kind_t, token_wptr_t> token_map_t;
    token_map_t tokens;
};


END_NAMESPACE;
END_NAMESPACE;

#endif /* LICENSE_HH */
