/// @file library/stream/stream.hh

// Copyright (c) 2006, 2008 Acumem AB.  All rights reserved.

/**
 * Contains common declarations and helper functions for stream
 * nodes. Streams work analogous to UNIX pipes.
 *
 * Setup a stream by instanciating every node starting with the
 * consumer node and ending with the producer node. The constructor of
 * every producing node takes a stream data proccessing callback and a
 * data pointer as a parameter. These identify the next node in the stream.
 *
 * The callback may abort stream processing prematurely by returning
 * STREAM_ABORT, this return value will then be propagated to the
 * producer node. Returning STREAM_ERROR will also abort the stream,
 * but signifies an error condition.
 *
 * Streams are driven by the producer node, which provides a function
 * to produce data.
 *
 * Stream destruction is highly order sensitive and has to be done
 * from the producer node to the consumer node. This allows the
 * respective delete functions to flush their internal buffers and
 * pass buffered data to the next node in the stream.
 */

#ifndef FNR_STREAM_HH
#define FNR_STREAM_HH

#include "library/lang/common_macros.h"

#include <boost/lexical_cast.hpp>

#include <stdexcept>
#include <string>

#include <string.h>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace boost;
using namespace std;

typedef enum {
    /// Stream state is OK
    STREAM_OK = 0,
    /// Abort was requested by a stream node
    STREAM_ABORT,
    /// A stream node returned an error
    STREAM_ERROR
} stream_status_t;

class stream_exception_t
    : public std::runtime_error
{
public:
    stream_exception_t(const string &what_)
        : std::runtime_error(what_)
        {}
};

class stream_errno_exception_t
    : public stream_exception_t
{
public:
    stream_errno_exception_t(const string &message, int error)
        : stream_exception_t(
            message + " error "  + lexical_cast<string>(error) + ", " +
            strerror(error) + ".")
        {}
};

class stream_t
{
public:
    virtual ~stream_t() {};

    /**
     * Stream node data processing callback.
     *
     * \param buffer data to process, may never be NULL.
     * \param len length of #buffer.
     *
     * \return a stream_status_t.
     */
    virtual stream_status_t process_data(const char *buffer, size_t len) = 0;

    virtual stream_status_t flush() = 0;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
