/// @file library/stream/stream_container.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include "stream_container.hh"

#ifdef ENABLE_FNR_GZ
#include "stream_gz.hh"
#endif

#ifdef ENABLE_FNR_BZ2
#include "stream_bz2.hh"
#endif

#include "library/am_mem/am_mem.h"

#include <assert.h>
#include <iostream>

#define MIN(n, m) ((n) < (m) ? (n) : (m))

BEGIN_NAMESPACE(acumem)
BEGIN_NAMESPACE(library)

// ----- stream_container_header_t --------------------------------------------

ostream &
operator<<(ostream &os, const stream_container_header_t &header)
{
    os << "Container:" << endl
       << "  Version: " << (int)header.version << endl;
    if (header.version > 0) {
	os << "  Product: "
           << stream_container_product_name(header.product)
           << "(" << header.product << ")" << endl;
	os << "  Data type: "
           << stream_container_data_type_name(header.data_type)
           << "(" << header.data_type << ")" << endl;
	os << "  Compression: "
           << stream_container_compression_name(header.compression)
           << "(" << header.compression << ")" << endl;
    }
    return os;
}

// ----- stream_container_compression_t ---------------------------------------

const stream_container_compression_t stream_container_compressions[] = {
    STREAM_CONTAINER_COMPRESSION_NONE,
#ifdef ENABLE_FNR_GZ
	STREAM_CONTAINER_COMPRESSION_DEFLATE,
#endif
#ifdef ENABLE_FNR_BZ2
	STREAM_CONTAINER_COMPRESSION_BZIP2,
#endif

	STREAM_CONTAINER_COMPRESSION_ILLEGAL
	};

stream_container_writer_t::stream_container_writer_t(
    const stream_container_header_t &header,
    stream_t *stream_next,
    warning_consumer_ptr_t warning_consumer)
    : header(header), stream_next(stream_next), compression(NULL)
{
    if (header.version > 0)
	write_header();
    else {
	/* Header version 0 is the legacy mode, setup default header
	 * settings, but don't write them to the file.
	 */
	this->header.version = STREAM_CONTAINER_VERSION;
	this->header.product = STREAM_CONTAINER_PRODUCT_VPE;
	this->header.data_type = STREAM_CONTAINER_DATA_SAMPLES_BINARY;
	this->header.compression = STREAM_CONTAINER_COMPRESSION_DEFLATE;
    }

    switch (this->header.compression) {
    case STREAM_CONTAINER_COMPRESSION_DEFLATE:
#ifdef ENABLE_FNR_GZ
	compression = new stream_gz_compress_t(stream_next, warning_consumer);
	break;
#else
	throw stream_exception_t("stream_container_writer_t",
				 "Deflate compression unsupported.");
#endif

    case STREAM_CONTAINER_COMPRESSION_BZIP2:
#ifdef ENABLE_FNR_BZ2
	compression = new stream_bz2_compress_t(stream_next);
	break;
#else
	throw stream_exception_t("BZip2 compression unsupported.");
#endif

    case STREAM_CONTAINER_COMPRESSION_NONE:
	break;

    default:
    case STREAM_CONTAINER_COMPRESSION_ILLEGAL:
	throw stream_exception_t("Illegal compression specified.");
    }
}

stream_container_writer_t::~stream_container_writer_t()
{
    if (compression)
	delete compression;
}

void
stream_container_writer_t::write_header()
{
    char header_out[STREAM_CONTAINER_HEADER_SIZE];
    char *pos;

    memcpy(header_out, STREAM_CONTAINER_MAGIC, STREAM_CONTAINER_MAGIC_SIZE);
    pos = &header_out[STREAM_CONTAINER_MAGIC_SIZE];

    *(pos++) = (STREAM_CONTAINER_HEADER_SIZE >> 8) & 0xFF;
    *(pos++) = STREAM_CONTAINER_HEADER_SIZE & 0xFF;
    *(pos++) = header.version;
    *(pos++) = header.product;
    *(pos++) = header.data_type;
    *(pos++) = header.compression;

    stream_status_t status = stream_next->process_data(header_out,
						       STREAM_CONTAINER_HEADER_SIZE);
    if (status != STREAM_OK)
	throw stream_exception_t("Failed to write container header.");
}

stream_status_t
stream_container_writer_t::process_data(const char *buffer, size_t len)
{
    if (compression)
	return compression->process_data(buffer, len);
    else
	return stream_next->process_data(buffer, len);
}

stream_status_t
stream_container_writer_t::flush()
{
    if (compression)
	return compression->flush();
    else
	return stream_next->flush();
}

stream_container_reader_t::stream_container_reader_t(
    stream_t *stream_next,
    warning_consumer_ptr_t warning_consumer)
    : stream_next(stream_next),
      compression(NULL),
      header(),
      header_buf_in(0),
      warning_consumer(warning_consumer)
{
}

stream_container_reader_t::~stream_container_reader_t()
{
    /*
     * Seems like the stream was shorter than a header, output the
     * header cache to the stream if possible.
     */
    if (header_buf_in < STREAM_CONTAINER_HEADER_SIZE) {
	header.version = 0;
	stream_status_t status = process_header(header);

	if (status == STREAM_OK) {
	    try {
		setup_reader_from_header();
		get_next()->process_data(header_buf, header_buf_in);
	    } catch (...) {
		/// XXX: Handle errors
	    }
	}
    }

    if (compression)
	delete compression;
}

stream_status_t
stream_container_reader_t::process_data(const char *buffer, size_t len)
{
    if (header_buf_in < STREAM_CONTAINER_HEADER_SIZE) {
	int header_len_available;
	assert(len <= 0x7fffffff);
	header_len_available = MIN((int)(STREAM_CONTAINER_HEADER_SIZE - header_buf_in),
				   (int)len);

	memcpy(&header_buf[header_buf_in],
	       buffer,
	       header_len_available);

	header_buf_in += header_len_available;

	if (header_buf_in == STREAM_CONTAINER_HEADER_SIZE) {
	    /* Full header has arrived */
	    stream_status_t status;
	    /* Actual length of the header */
	    int header_len = decode_header();

	    if (header_len == -1) {
		return STREAM_ERROR;
	    } else {
		/* Fire the header callback to allow the callee to abort
		 * if data type mismatches etc.
		 */
		status = process_header(header);
		if (status != STREAM_OK)
		    return status;
	    }

	    try {
		setup_reader_from_header();
	    } catch (...) {
		/* XXX: More intelligent error handling */
		return STREAM_ERROR;
	    }

	    /* Output the data that has been cached in the header
	     * cache so far but not been used in the header.
	     */
	    if ((unsigned int)header_len < STREAM_CONTAINER_HEADER_SIZE) {
		status = get_next()->process_data(
		    &header_buf[header_len],
		    STREAM_CONTAINER_HEADER_SIZE - header_len);

		if (status != STREAM_OK)
		    return status;
	    }
	} else
	    return STREAM_OK;

	/* Output any outstanding data in the input buffer */
	if (len - header_len_available > 0)
	    return get_next()->process_data(
		&buffer[header_len_available],
		len - header_len_available);
	else
	    return STREAM_OK;

    } else
	return get_next()->process_data(buffer, len);
}

stream_status_t
stream_container_reader_t::flush()
{
    return get_next()->flush();
}

const stream_container_header_t &
stream_container_reader_t::get_header()
{
    return header;
}

int
stream_container_reader_t::decode_header()
{
    char *pos;
    uint16_t len;

    if (memcmp(header_buf, STREAM_CONTAINER_MAGIC, STREAM_CONTAINER_MAGIC_SIZE)) {
	header.version = 0;
	return 0;
    }

    pos = &header_buf[STREAM_CONTAINER_MAGIC_SIZE];

    len = ((unsigned char)(*pos++)) << 8;
    len |= (unsigned char)(*pos++);

    header.version = (uint8_t)(*pos++);
    header.product = (stream_container_product_t)((uint8_t)(*pos++));
    header.data_type = (stream_container_data_type_t)((uint8_t)(*pos++));
    header.compression = (stream_container_compression_t)((uint8_t)(*pos++));

    return header.version <= STREAM_CONTAINER_VERSION ? len : -1;
}

void
stream_container_reader_t::setup_reader_from_header()
{
    /* Version 0 is just means "no header". */
    if (header.version == 0)
	return;

    /* Setup compression */
    switch (header.compression) {
    case STREAM_CONTAINER_COMPRESSION_NONE:
	break;
    case STREAM_CONTAINER_COMPRESSION_DEFLATE:
#ifdef ENABLE_FNR_GZ
	compression = new stream_gz_decompress_t(
            stream_next, warning_consumer);
	break;
#else
	throw stream_exception_t("Deflate compression unsupported.");
#endif
    case STREAM_CONTAINER_COMPRESSION_BZIP2:
#ifdef ENABLE_FNR_BZ2
	compression = new stream_bz2_decompress_t(stream_next);
	break;
#else
	throw stream_exception_t("BZip2 compression unsupported.");
#endif
    default:
	throw stream_exception_t("Illegal compression specified.");
    }
}

stream_status_t
stream_container_reader_t::process_header(stream_container_header_t &header)
{
    if (header.version == 0) {
	header.version = STREAM_CONTAINER_VERSION;
	header.product = STREAM_CONTAINER_PRODUCT_VPE;
	header.data_type = STREAM_CONTAINER_DATA_SAMPLES_BINARY;
	header.compression = STREAM_CONTAINER_COMPRESSION_DEFLATE;
    }

    return STREAM_OK;
}

/* ---------------------------------------------------------------------- */

const char *
stream_container_product_name(stream_container_product_t product)
{
    switch (product) {
    case STREAM_CONTAINER_PRODUCT_VPE:
	return "SlowSpotter";
    case STREAM_CONTAINER_PRODUCT_VPEMP:
	return "ThreadSpotter";
    case STREAM_CONTAINER_PRODUCT_LITMUS:
	return "SpotLite";
    case STREAM_CONTAINER_PRODUCT_SS12:
	return "SlowSpotter 1.2";
    case STREAM_CONTAINER_PRODUCT_INTERNAL:
	return "Freja Internal";
    default:
	return "Unknown";
    }
}

const char *
stream_container_data_type_name(stream_container_data_type_t type)
{
    switch (type) {
    case STREAM_CONTAINER_DATA_SAMPLES_BINARY:
	return "Binary samples";
    default:
	return "Unknown";
    }
}

const char *
stream_container_compression_name(stream_container_compression_t compression)
{
    switch (compression) {
    case STREAM_CONTAINER_COMPRESSION_NONE:
	return "None";
    case STREAM_CONTAINER_COMPRESSION_DEFLATE:
	return "Deflate";
    case STREAM_CONTAINER_COMPRESSION_BZIP2:
	return "bzip2";
    default:
	return "Unknown";
    }
}

END_NAMESPACE;
END_NAMESPACE;
