/// @file library/stream/stream_container.hh

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#ifndef FNR_STREAM_CONTAINER_HH
#define FNR_STREAM_CONTAINER_HH

#include "library/lang/common_types.h"
#include "library/stream/stream.hh"
#include "library/warning/warning_consumer.hh"

/*
 * Header format:
 * +--------+---------+---------+---------+-----------+-------------+----------+
 * |"ACUMEM"| Length  | Version | Product | Data type | Compression | DATA ... |
 * | binary | Uint16  | Uint8   | Uint8   | Uint8     |	Uint8  	    |          |
 * +--------+---------+---------+---------+-----------+-------------+----------+
 *
 * All multiple byte integers are saved as big endian.
 *
 * magic       - "ACUMEM"
 * length      - Length of the header, including magic and length
 * version     - Version of the header format, will be increased
 *               every time a field is added.
 * product     - Product code
 * data type   - Type of data
 * compression - Compression algorithm to use
 *
 */

/** The most recent version supported by this header */
#define STREAM_CONTAINER_VERSION (1)

#define STREAM_CONTAINER_MAGIC "ACUMEM"
#define STREAM_CONTAINER_MAGIC_SIZE (sizeof(STREAM_CONTAINER_MAGIC) - 1)
#define STREAM_CONTAINER_HEADER_SIZE (STREAM_CONTAINER_MAGIC_SIZE + 6)

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

enum stream_container_product_t {
    STREAM_CONTAINER_PRODUCT_VPE = 0,
    STREAM_CONTAINER_PRODUCT_VPEMP = 1,
    STREAM_CONTAINER_PRODUCT_LITMUS = 2,
    /// SlowSpotter 1.2
    STREAM_CONTAINER_PRODUCT_SS12 = 3,
    /// Use this for internal tools, should never be seen in customer files.
    STREAM_CONTAINER_PRODUCT_INTERNAL = 255
};

enum stream_container_compression_t {
    /** No compression used */
    STREAM_CONTAINER_COMPRESSION_NONE = 0,
    /** Deflate from Zlib */
    STREAM_CONTAINER_COMPRESSION_DEFLATE = 1,
    /** BZip2 */
    STREAM_CONTAINER_COMPRESSION_BZIP2 = 2,

    /** A placeholder compression to signify end of lists etc. */
    STREAM_CONTAINER_COMPRESSION_ILLEGAL = -1
};

enum stream_container_data_type_t {
    /// File containing samples that can be decoded with codec binary.
    STREAM_CONTAINER_DATA_SAMPLES_BINARY = 0,
    /// Report in Hybris intermediate format.
    STREAM_CONTAINER_DATA_REPORT = 1
};

struct stream_container_header_t {
    stream_container_header_t()
	: version(STREAM_CONTAINER_VERSION),
	  product(STREAM_CONTAINER_PRODUCT_VPE),
	  data_type(STREAM_CONTAINER_DATA_SAMPLES_BINARY),
	  compression(STREAM_CONTAINER_COMPRESSION_DEFLATE)
	{}

    stream_container_header_t(
        uint8_t version,
        stream_container_product_t product,
        stream_container_data_type_t data_type,
        stream_container_compression_t compression)
        : version(version),
          product(product),
          data_type(data_type),
          compression(compression)
        {
        }

    /* ---------- VERSION 0 ---------- */
    /** Version of container format */
    uint8_t version;
    /* ---------- VERSION 1 ---------- */
    /** Product that generated the file */
    stream_container_product_t product;
    /** Type of data */
    stream_container_data_type_t data_type;
    /** Compression algorithm */
    stream_container_compression_t compression;
};

/// Output textual representation of the header.
ostream &operator<<(ostream &os, const stream_container_header_t &header);

/**
 * List of available compression algorithms, terminated with
 * STREAM_CONTAINER_COMPRESSION_ILLEGAL.
 */
extern const stream_container_compression_t stream_container_compressions[];

class stream_container_writer_t
    : public stream_t
{
public:
    /**
     * Create a new stream_container_writer_t instance. The container
     * writer will automatically compress its' input stream if compression
     * has been enabled in the header.
     *
     * To enable writing of legacy files without a header, use header
     * version number 0.
     *
     * \param header header data to write to file.
     * \param stream_next next stream node
     */
    stream_container_writer_t(
        const stream_container_header_t &header,
        stream_t *stream_next,
        warning_consumer_ptr_t warning_consumer);
    ~stream_container_writer_t();

    /**
     * Data processing callback for the stream container. Takes care of
     * stream compressino etc as specified in the header.
     *
     * \param buffer pointer to buffer with data to decompress, may not be NULL.
     * \param len length of #buffer.
     * \return Propagates return value from next node on error/abort, may
     *         fail with STREAM_ERROR, returns STREAM_OK on success.
     */
    stream_status_t process_data(const char *buffer, size_t len);

    stream_status_t flush();

private:
    /**
     * Writes the header to the output stream.
     *
     * \throw stream_exception_t on error.
     */
    void write_header();

    stream_container_header_t header;
    stream_t *stream_next;
    stream_t *compression;
};

class stream_container_reader_t
    : public stream_t
{
public:
    /**
     * Create a new stream_container_reader_t instance. The container
     * reader will automatically decompress its' input stream if compression
     * has been enabled in the header.
     *
     * A the process_header function may be overriden to do custom
     * header verification/manipulation. The function will be called
     * after a header has been read, or it has been determined that no
     * header exists in the begining of the stream.
     *
     * \param stream_next next stream node
     */
    stream_container_reader_t(
        stream_t *stream_next,
        warning_consumer_ptr_t warning_consumer);
    virtual ~stream_container_reader_t();

    /**
     * Data processing callback for the stream container. Takes care of
     * stream compression etc as specified in the header.
     *
     * \param buffer pointer to buffer with data to decompress, may not be NULL.
     * \param len length of #buffer.
     * \return Propagates return value from next node on error/abort, may
     *         fail with STREAM_ERROR, returns STREAM_OK on success.
     */
    stream_status_t process_data(const char *buffer, size_t len);

    stream_status_t flush();

    const stream_container_header_t &get_header();

protected:
    /**
     * Function called to do header verification or manipulation.
     *
     * If no valid header is available, the header pointed to by he
     * second parameter in the callback will have the special version
     * number 0. The function may in such a case "up convert" the
     * header to a recent version to enable decoding with the
     * parameters in the header, simplifies handling of legacy
     * files. The callback may return an error if the header contain
     * unacceptable data, e.g. incompatible product codes.
     *
     * \param header stream header, may be mainpulated to customize
     *   stream handling.
     */
    virtual stream_status_t process_header(stream_container_header_t &header);

private:
    /**
     * Decodes the header in header_buf. Returns the total length of the
     * header.
     *
     * The following return values have special meanings:
     *  0:  Failed to read magic (i e not a valid file)
     *  -1: Incompatible header version
     */
    int decode_header();

    /**
     * Setup reader parameters from header data.
     *
     * \throw stream_exception_t on error.
     */
    void setup_reader_from_header();

    inline stream_t *get_next() {
	return compression ? compression : stream_next;
    }
    stream_t *stream_next;
    stream_t *compression;
    stream_container_header_t header;

    char header_buf[STREAM_CONTAINER_HEADER_SIZE];
    unsigned int header_buf_in;

    warning_consumer_ptr_t warning_consumer;
};

/**
 * Returns a pointer to a constant character representation of the
 * product name represented by product.
 *
 * \param product product ID to translate.
 * \return a pointer to a NULL terminated character string,
 *   must not be freed by the caller.
 */
const char *stream_container_product_name(
    stream_container_product_t product);

/**
 * Returns a pointer to a constant character representation of the
 * data type represented by type.
 *
 * \param type type ID to translate.
 * \return a pointer to a NULL terminated character string,
 *   must not be freed by the caller.
 */
const char *stream_container_data_type_name(
    stream_container_data_type_t type);

/**
 * Returns a pointer to a constant character representation of the
 * compression represented by compression.
 *
 * \param compression compression ID to translate.
 * \return a pointer to a NULL terminated character string,
 *   must not be freed by the caller.
 */
const char *stream_container_compression_name(
    stream_container_compression_t compression);

END_NAMESPACE;
END_NAMESPACE;

#endif
