/// @file library/stream/stream_file.cc

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.


#include "stream_file.hh"

#include "library/am_mem/am_mem.h"
#include "library/lang/tstring.hh"
#include "library/exe_path/exe_path.hh"

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#define READ_BUFFER_SIZE 4096

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

namespace fs = boost::filesystem;

/* ---------------------------------------------------------------------- */
/* stream reader                                                          */
/* ---------------------------------------------------------------------- */
stream_file_reader_t::stream_file_reader_t(const fs::path &filename,
					   stream_t *stream_next)
    : stream_next(stream_next)
{
    if (!filename.empty()) {
	file = acumem::exe_path::fopen(filename, "rb");
	shall_close = true;
    }
    else {
	file = stdin;
	shall_close = false;
    }

    if (!file)
	throw stream_errno_exception_t(
            string("Failed to open file ") + filename.string() + ": ", errno);
}

stream_file_reader_t::stream_file_reader_t(FILE *f,
					   stream_t *stream_next)
    : stream_next(stream_next)
{
    file = f;

    shall_close = false;

    if (!file)
	throw stream_errno_exception_t("Failed to close file: ", errno);
}

stream_file_reader_t::~stream_file_reader_t()
{
    if (shall_close)
	fclose(file);
}

stream_status_t
stream_file_reader_t::dispatch()
{
    char dstbuf[READ_BUFFER_SIZE];
    size_t bytes;

    do {
	bytes = fread(dstbuf, 1, sizeof(dstbuf), file);

	if (bytes > 0) {
	    stream_status_t ret;

	    ret = stream_next->process_data(dstbuf, bytes);

	    if (ret != STREAM_OK)
		return ret;
	}
    } while (!ferror(file) && !feof(file));

    if (ferror(file) && !feof(file))
	return STREAM_ERROR;
    else {
	return stream_next->flush();
    }
}

/* ---------------------------------------------------------------------- */
/* stream writer                                                          */
/* ---------------------------------------------------------------------- */
stream_file_writer_t::stream_file_writer_t(const fs::path &filename)
{
    if (!filename.empty()) {
	file = acumem::exe_path::fopen(filename, "wb");
	shall_close = true;
    } else {
	file = stdout;
	shall_close = false;
    }

    if (!file)
	throw stream_errno_exception_t(
            string("Failed to open file ") + filename.string() + ": ", errno);
}

stream_file_writer_t::stream_file_writer_t(FILE *f)
{
    file = f;
    shall_close = false;

    if (!file)
	throw stream_errno_exception_t("Failed to close file: ", errno);
}

stream_file_writer_t::~stream_file_writer_t()
{
    if (shall_close)
	fclose(file);
}

stream_status_t
stream_file_writer_t::process_data(const char *buffer, size_t len)
{
    size_t written;

    written = fwrite(buffer, 1, len, file);

    if (written == 0 || written < len)
	return ferror(file) ? STREAM_ERROR : STREAM_OK;
    else
	return STREAM_OK;
}

stream_status_t
stream_file_writer_t::flush()
{
    if (fflush(file) == 0)
	return STREAM_OK;
    else
	return STREAM_ERROR;
}

END_NAMESPACE;
END_NAMESPACE;
