/// @file library/stream/stream_gz.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include "stream_gz.hh"
#include "stream.hh"
#include "library/am_mem/am_mem.h"

#include <stdio.h>
#include <assert.h>
#include <string.h>

#define DECOMPRESSION_BUFFER_SIZE 16384
#define COMPRESSION_BUFFER_SIZE 4096

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

static voidpf
my_alloc(voidpf opaque, uInt items, uInt size)
{
    return malloc(items * size);
}

static void
my_free(voidpf opaque, voidpf address)
{
    free(address);
}

stream_gz_compress_t::stream_gz_compress_t(
    stream_t *stream_next,
    warning_consumer_ptr_t warning_consumer)
    : stream_next(stream_next),
      warning_consumer(warning_consumer)
{
    memset(&zstream, 0, sizeof(z_stream));
    zstream.zalloc = my_alloc;
    zstream.zfree = my_free;

    int res = deflateInit(&zstream, Z_DEFAULT_COMPRESSION);
    if (res != Z_OK)
	throw stream_exception_t(
            string("Failed to initialize libz: deflateInit error ") +
            lexical_cast<string>(res) + ".");
}

stream_gz_compress_t::~stream_gz_compress_t()
{
    internal_deflate(NULL, 0, Z_FINISH);
    deflateEnd(&zstream);
}


stream_status_t
stream_gz_compress_t::process_data(const char *buffer, size_t len)
{
    return internal_deflate(buffer, len, Z_NO_FLUSH);
}

stream_status_t
stream_gz_compress_t::flush()
{
    stream_status_t status;

    status = internal_deflate(NULL, 0, Z_SYNC_FLUSH);
    if (status == STREAM_OK)
	return stream_next->flush();

    return status;
}

stream_status_t
stream_gz_compress_t::internal_deflate(const char *buffer, size_t len, int flush)
{
    int zresult;

    /* Make sure that there is no previous input that is not processed */
    assert(zstream.avail_in == 0);

    zstream.next_in = (unsigned char *)buffer;

    assert(len <= 0xffffffff);
    zstream.avail_in = (uInt) len;

    do {
	char compressed[COMPRESSION_BUFFER_SIZE];

        zstream.next_out = (unsigned char *)compressed;
        zstream.avail_out = COMPRESSION_BUFFER_SIZE;

        zresult = deflate(&zstream, flush);

        if (zresult == Z_OK ||
            zresult == Z_STREAM_END ||
            zresult == Z_BUF_ERROR) {
            size_t out_len = COMPRESSION_BUFFER_SIZE - zstream.avail_out;
            if (out_len > 0) {
                stream_status_t stream_status =
                    stream_next->process_data(compressed, out_len);
                if (stream_status != STREAM_OK)
                    return stream_status;
            }
        } else {
            stringstream ss;
            ss << "Error " << zresult << " when calling deflate().";
            warning_consumer->error(ss.str());
            return STREAM_ERROR;
        }

	/* Determine if we need to continue looping. zlib docs say continue:
	 *
	 * - if zstream.avail_out == 0 (presumably if flush is not set), or
	 * - if flush is set, for as long as zresult is not Z_STREAM_END.
	 *
	 * The academic doubt is how to handle avail_out != 0 &&
	 * zresult == Z_STREAM_END. Right now, zresult takes presedence.
	 */
    } while (zresult == Z_OK &&
             (zstream.avail_out == 0 ||
              flush == Z_FINISH));

    return STREAM_OK;
}

/* ---------------------------------------------------------------------- */
/* decompress */
stream_gz_decompress_t::stream_gz_decompress_t(
    stream_t *stream_next,
    warning_consumer_ptr_t warning_consumer)
    : stream_next(stream_next),
      warning_consumer(warning_consumer)
{
    memset(&zstream, 0, sizeof(z_stream));
    zstream.zalloc = my_alloc;
    zstream.zfree = my_free;

    int res = inflateInit(&zstream);
    if (res != Z_OK)
	throw stream_exception_t(
            string("Failed to initialize libz: inflateInit error ") +
            lexical_cast<string>(res) + ".");
}

stream_gz_decompress_t::~stream_gz_decompress_t()
{
    inflateEnd(&zstream);
}

stream_status_t
stream_gz_decompress_t::process_data(const char *buffer, size_t len)
{
    char out[DECOMPRESSION_BUFFER_SIZE];
    int zresult;

    assert(buffer);

    /* Make sure that there is no previous input that is not processed */
    assert(zstream.avail_in == 0);

    zstream.next_in = (unsigned char *)buffer;

	assert(len <= 0xffffffff);
    zstream.avail_in = (uInt) len;

    do {
        zstream.next_out = (unsigned char *)out;
        zstream.avail_out = DECOMPRESSION_BUFFER_SIZE;

        zresult = inflate(&zstream, Z_SYNC_FLUSH);
	/* input parameters inconsistent? */
	assert(zresult != Z_STREAM_ERROR);

	switch(zresult) {
	case Z_BUF_ERROR:
	    /* Buffer error */
	    /* The manual states that this can happen (when flush ==
	     * Z_SYNC_FLUSH) iff no progress is possible. This
	     * sometimes happens and is not fatal. One reason seems to
	     * be that inflate is called with zstream.avail_in == 0, which
	     * can happen if avail_out was 0 on the last call (inflate
	     * should in that case be called again to flush buffers
	     * etc). It seems reasonable to assume that "no progress
	     * possible" would mean that no output data is available,
	     * but the example code in zlib tries to write the output
	     * anyway, so we'll do that too. (It would probably be
	     * safe to just return STREAM_OK here instead...)
	     */
	    /* FALLTHROUGH */
	case Z_STREAM_END:
	    /* EOF */
	    /* FALLTHROUGH */
	case Z_OK:
	    /* Made progress, EOF not flagged */
	    if (zstream.avail_out < DECOMPRESSION_BUFFER_SIZE) {
		stream_status_t stream_status;
		stream_status = stream_next->process_data(
		    out,
		    DECOMPRESSION_BUFFER_SIZE - zstream.avail_out);
		if (stream_status != STREAM_OK)
		    return stream_status;
	    }
	    break;

	case Z_NEED_DICT:
	    /* We don't setup our own dictionary in when compressing,
	     * so this shouldn't happen. */
            warning_consumer->error("deflate() returned Z_NEED_DICT.");
	    return STREAM_ERROR;

	case Z_DATA_ERROR:
	    /* Input data corrupted. */
	    warning_consumer->error("deflate() returned Z_DATA_ERROR.");
	    return STREAM_ERROR;

	case Z_MEM_ERROR:
	    /* Out of memory */
	    warning_consumer->error("deflate() returned Z_MEM_ERROR.");
	    return STREAM_ERROR;
	}
    } while (zresult != Z_STREAM_END && zstream.avail_out == 0);

    return STREAM_OK;
}

stream_status_t
stream_gz_decompress_t::flush()
{
    return stream_next->flush();
}


END_NAMESPACE;
END_NAMESPACE;
