/// @file library/stream/stream_gz.hh

// Copyright (c) 2008 Acumem AB.  All rights reserved.

#ifndef FNR_STREAM_GZ_HH
#define FNR_STREAM_GZ_HH

#include "library/stream/stream.hh"
#include "library/warning/warning_consumer.hh"

#ifdef _WIN32
#define ZLIB_WINAPI
#endif
#include <zlib.h>

/* This is an adaptation of the zlib compression scheme (as implemented by
 * the zlib library) to the stream concept.
 *
 * It compresses or decompresses a byte stream according to the zlib deflate method.
 * This method is slightly different from the gzip method, and according to the documentation
 * in <zlib.h>, more suited for inter-process compression, while gzip is more directed towards
 * file compression.
 *
 * The same header says that the gzip format also is supported for memory operations,
 * and can then be added later if need arises.
 */

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

class stream_gz_compress_t
    : public stream_t
{
public:
    /**
     * Creates a stream compression node (deflate) that passes compressed
     * data to the next node.
     *
     * \param stream_next next node in the stream
     */
    stream_gz_compress_t(
        stream_t *stream_next,
        warning_consumer_ptr_t warning_consumer);
    ~stream_gz_compress_t();

    /**
     * Data processing callback for the stream_gz_compress_t. Compresses
     * data provided in #buffer and passes the compressed data to the next
     * node.
     *
     * \param buffer pointer to buffer with data to compress, may not be NULL.
     * \param len length of #buffer.
     * \return Propagates return value from next node on error/abort, may
     *         fail with STREAM_ERROR, returns STREAM_OK on success.
     */
    stream_status_t process_data(const char *buffer, size_t len);

    stream_status_t flush();

private:
    stream_status_t internal_deflate(const char *buffer, size_t len, int flush);

    stream_t *stream_next;
    z_stream zstream;

    warning_consumer_ptr_t warning_consumer;
};


class stream_gz_decompress_t
    : public stream_t
{
public:

    /**
     * Creates a stream decompression node (inflate) that passes
     * decompressed data to the node represented by callback and data.
     *
     * \param stream_next next node in the stream.
     */
    stream_gz_decompress_t(
        stream_t *stream_next,
        warning_consumer_ptr_t warning_consumer);
    ~stream_gz_decompress_t();

    /**
     * Data processing callback for the
     * stream_gz_decompress_t. Decompresses data provided in #buffer
     * and passes the compressed data to the next node.
     *
     * \param buffer pointer to buffer with data to decompress, may not be NULL.
     * \param len length of #buffer.
     * \return Propagates return value from next node on error/abort, may
     *         fail with STREAM_ERROR, returns STREAM_OK on success.
     */
    stream_status_t process_data(const char *buffer, size_t len);
    stream_status_t flush();

private:
    stream_status_t internal_deflate(const char *buffer, size_t len, int done);

    stream_t *stream_next;
    z_stream zstream;

    warning_consumer_ptr_t warning_consumer;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
