/// @file library/sym/cppfilt.cc

// Copyright (c) 2008, 2010 Acumem AB.  All rights reserved.

#include "cppfilt.hh"

#ifndef _WIN32
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <sys/wait.h>
#include <string.h>
#include <signal.h>
#endif

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;
using boost::shared_ptr;


BEGIN_NAMESPACE(detail);

#ifndef _WIN32

/// Represents a chunk of data to send
class chunk_t {
public:
    chunk_t(const char *b, size_t s);
    ~chunk_t();

    bool eof() const;
    const char *ptr() const;
    size_t avail() const;
    void advance(size_t s);
private:
    char *buf;
    size_t size;
    size_t read_pos;
};

typedef shared_ptr<chunk_t> chunk_ptr_t;
typedef list<chunk_ptr_t> chunk_list_t;


/// A manager of a queue of chunks. Knows how to extract complete lines.
class buffer_t {
public:
    buffer_t();

    void put(const char *b, size_t s);
    void put(const char *b);
    void put(const string &s);

    bool getline(string &line);

    bool eof() const;
    const char *ptr() const;
    size_t avail() const;
    void advance(size_t s);

private:
    chunk_list_t chunks;
    bool seen_new_line;
};


/// Bidirectional communication with an external process
class process_t {
public:
    process_t(const char *path, char *const argv[]);
    ~process_t();

    bool put(const string &s);
    bool getline(string &line);

private:
    bool work(bool flush_write, int timeout);
    void read_from_process_stdout();
    void write_to_process_stdin();
    static bool set_o_nonblock(int fd);

    pid_t pid;
    int stdin_fd;
    int stdout_fd;

    buffer_t to_process;
    buffer_t from_process;
};



inline
chunk_t::chunk_t(const char *b, size_t s)
{
    buf = new char[s];
    size = s;
    memcpy(buf, b, s);
    read_pos = 0;
}

inline
chunk_t::~chunk_t()
{
    delete [] buf;
}

inline
const char *
chunk_t::ptr() const
{
    return buf + read_pos;
}

inline
size_t
chunk_t::avail() const
{
    return size - read_pos;
}

inline
void
chunk_t::advance(size_t s)
{
    read_pos += s;
    assert(read_pos <= size);
}

inline
bool
chunk_t::eof() const
{
    return read_pos == size;
}

inline
buffer_t::buffer_t()
    : seen_new_line(false)
{
}

void
buffer_t::put(const char *b, size_t s)
{
    if (!seen_new_line && memchr(b, '\n', s) != 0)
	seen_new_line = true;

    chunks.push_back(chunk_ptr_t(new chunk_t(b, s)));
}

inline
void
buffer_t::put(const char *b)
{
    put(b, strlen(b));
}

void
buffer_t::put(const string &s)
{
    put(s.c_str(), s.size());
}

bool
buffer_t::getline(string &line)
{
    if (!seen_new_line)
	return false;

    size_t total = 0;
    chunk_list_t::iterator
	i = chunks.begin(), e = chunks.end();
    for (; i != e; i++) {
	const char *b = (*i)->ptr();
	size_t s = (*i)->avail();
	const char *p = (const char *) memchr(b, '\n', s);

	if (!p) {
	    // no newline here, just total this entire block.
	    total += (*i)->avail();
	} else {
	    // finally found a newline, copy out result.
	    total += (size_t)(p - b);

	    line.clear();
	    line.reserve(total);
	    chunk_list_t::iterator x = chunks.begin();
	    for (; x != i; x++)
		line.append((*x)->ptr(), (*x)->avail());

	    line.append(b, p);
	    (*i)->advance(p-b+1);
	    if ((*i)->eof())
		i++;

	    chunks.erase(chunks.begin(), i);
	    return true;
	}
    }

    seen_new_line = false;
    return false;
}

inline
bool
buffer_t::eof() const
{
    return chunks.size() == 0;
}

inline
const char *
buffer_t::ptr() const
{
    if (eof())
	return 0;
    return chunks.front()->ptr();
}

inline
size_t
buffer_t::avail() const
{
    if (eof())
	return 0;
    return chunks.front()->avail();
}

inline
void
buffer_t::advance(size_t s)
{
    chunk_t &c = *chunks.front();
    c.advance(s);
    if (c.eof())
	chunks.pop_front();
}



process_t::process_t(const char *path, char *const argv[])
{
    int control_pipe[2];
    int stdin_pipe[2];
    int stdout_pipe[2];

    pipe(control_pipe);
    fcntl(control_pipe[1], F_SETFD, FD_CLOEXEC);

    pipe(stdin_pipe);
    pipe(stdout_pipe);

    int saved_errno = 0;

    switch (pid = fork()) {
    case -1:
	throw process_exception_t("fork", errno);

    case 0: // child
	// stdin
	close(stdin_pipe[1]);	// Child doesn't write - close this
	dup2(stdin_pipe[0], 0);	// dup the read descriptor to stdin
	close(stdin_pipe[0]);	// Don't need the pipe here

	// stdout
	close(stdout_pipe[0]);
	dup2(stdout_pipe[1], 1);
	close(stdout_pipe[1]);

	// Control pipe - read end
	close(control_pipe[0]);

	execvp(path, argv);
	saved_errno = errno;
	write(control_pipe[1], &saved_errno, sizeof(saved_errno));
	exit(1);

    default: // parent
	close(stdin_pipe[0]);
	close(stdout_pipe[1]);
	stdin_fd = stdin_pipe[1];
	stdout_fd = stdout_pipe[0];
	set_o_nonblock(stdout_fd);
	set_o_nonblock(stdin_fd);

	close(control_pipe[1]);

	if (read(control_pipe[0], &saved_errno, sizeof(saved_errno)) > 0) {
	    // control_pipe indicates exec failed. Wait for child to
	    // die. Then throw exception

	    int status;
	    waitpid(pid, &status, 0);
	    close(stdin_fd);
	    close(stdout_fd);
	    close(control_pipe[0]);

	    throw process_exception_t("exec", saved_errno);
	}

	close(control_pipe[0]);
	break;
    }
}


process_t::~process_t()
{
    close(stdin_fd);
    close(stdout_fd);
    kill(pid, SIGKILL);
    int status;
    waitpid(pid, &status, 0);
}


bool
process_t::put(const string &s)
{
    to_process.put(s);
    return work(true, -1);
}

bool
process_t::getline(string &line)
{
    while(!from_process.getline(line))
	if (!work(false, -1))
	    return false;
    return true;
}

bool
process_t::work(bool flush_write, int timeout)
{
    while (!flush_write || !to_process.eof()) {
	pollfd f[2];
	f[0].fd = stdout_fd;
	f[0].events = POLLIN;
	f[0].revents = 0;
	f[1].fd = stdin_fd;
	f[1].events = POLLOUT;
	f[1].revents = 0;

	int r = poll(f, 2, timeout);

	if (r == -1) {
	    // Error
	    throw process_exception_t("poll", errno);
	}

	if (r == 0) {
	    // timeout
	    return false;
	}

	if ((f[1].revents | f[0].revents) & (POLLHUP | POLLERR)) {
	    /// XXX other exception here
	    throw process_exception_t("process died", 0);
	}

	if (f[1].revents & POLLOUT)
	    write_to_process_stdin();

	if (f[0].revents & POLLIN) {
	    read_from_process_stdout();
	    if (!flush_write)
		return true;
	}
    }

    return true;
}

void
process_t::read_from_process_stdout()
{
    int bytes;
    char buf[8192];

    do {
	bytes = read(stdout_fd, buf, sizeof(buf));
	if (bytes > 0)
	    from_process.put(buf, bytes);
    } while (bytes > 0);

    /// XXX different exception here
    if (bytes == 0)
	throw process_exception_t("read", 0);

    if (bytes < 0 && errno != EWOULDBLOCK)
	throw process_exception_t("read", errno);
}

void
process_t::write_to_process_stdin()
{
    while(!to_process.eof()) {
	const char *buf = to_process.ptr();
	size_t avail = to_process.avail();

	int bytes = write(stdin_fd, buf, avail);

	if (bytes < 0 && errno != EWOULDBLOCK)
	    throw process_exception_t("write", errno);

	to_process.advance(bytes);

	if (bytes != (int) avail)
	    return;
    }
}

bool
process_t::set_o_nonblock(int fd)
{
    int flags = 0;

    if ((flags = fcntl(fd, F_GETFL)) == -1)
	return false;

    if (!(flags & O_NONBLOCK) &&
	(fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1))
	return false;
    return true;
}

#endif // _WIN32
END_NAMESPACE;


cppfilt_t::cppfilt_t(const char *cppfilt_path)
{
#ifndef _WIN32
    if (cppfilt_path) {
	try {
	    char *const v[] = {
		(char *)cppfilt_path, 0};
	    cppfilt_process.reset(new detail::process_t(cppfilt_path, v));
	} catch (exception &e) {
	    cppfilt_process.reset();
	}
    }
#endif
}

string
cppfilt_t::translate(const string &symbol)
{
#ifndef _WIN32
    if (cppfilt_process) {
	cppfilt_process->put(symbol + '\n');
	string line;
	if (cppfilt_process->getline(line))
	    return line;

	// kill the child upon timeout errors
	cppfilt_process.reset();
    }
#endif
    return symbol;
}

END_NAMESPACE;
END_NAMESPACE;
