/// @file library/sym/debuglink.cc

// Copyright (c) 2009-2010 Acumem AB.  All rights reserved.

#include "debuglink.hh"

#include <boost/foreach.hpp>
#include <boost/filesystem.hpp>

#include <gelf.h>

#include <stdio.h>

using namespace std;

using boost::filesystem::path;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

struct elf_note_t {
    Elf64_Word namesz;
    Elf64_Word descsz;
    Elf64_Word type;

    const char *name;
    const char *desc;
};

typedef vector<elf_note_t> elf_note_vector_t;

/*
 * The note section is well defined for 32-bit binaries. 64-bit
 * binaries may have different formats, the GABI [1] states that all
 * fields should be 8 byte words and 8 byte aligned, while other
 * documentation [2] states that it should use 4 byte words with 4
 * byte alignment.
 *
 * GNU binutils use 4 byte words and 4 byte alignment for both 64 bit
 * and 32 bit targets. We'll assume that this is the case on all platforms for now.
 *
 * [1] http://sco.com/developers/gabi/latest/ch5.pheader.html#note_section
 * [2] http://www.netbsd.org/docs/kernel/elf-notes.html#note-format
 */
static bool
read_notes(elf_note_vector_t &notes, GElf_Shdr *shdr, Elf_Scn *scn)
{
    assert(shdr->sh_type == SHT_NOTE);
    Elf_Data *data(elf_rawdata(scn, NULL));

    const char *cdata((const char*)data->d_buf);
    const char *cdata_end(cdata + data->d_size);

    while (cdata < cdata_end) {
	elf_note_t note;

	// TODO: We assume that the ELF file has the same byte-order as
	// the host machine, this is currently always the case, but may
	// change in the future.
	note.namesz = ((Elf32_Word*)cdata)[0];
	note.descsz = ((Elf32_Word*)cdata)[1];
	note.type = ((Elf32_Word*)cdata)[2];

	const int name_len((note.namesz + 3) & ~3);
	const int desc_len((note.descsz + 3) & ~3);

	note.name = cdata + 4 * 3;
	note.desc = note.name + name_len;

	notes.push_back(note);

	cdata = note.desc + desc_len;
    }

    return true;
}

// ----------------------------------------------------------------------
// impl_gnu_debuglink_t
// ----------------------------------------------------------------------

struct impl_gnu_debuglink_t {
    impl_gnu_debuglink_t(Elf_Scn *scn);

    // Check if the section contains a valid debug link
    static bool is_debuglink(Elf *elf,
			     GElf_Shdr *shdr, Elf_Scn *scn);

    // Check if a debug file has a valid checksum
    bool verify_checksum(FILE *file) const;

    string get_filename() const {
	return filename;
    };

private:
    static const unsigned long crc32_table[0x100];
    // CRC32 function for GNU debuglink from GDB manual.
    static unsigned long crc32_part(const char *buf, size_t len,
				    unsigned long crc);

    string filename;
    uint32_t crc32;
};

const unsigned long impl_gnu_debuglink_t::crc32_table[0x100] = {
    0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419,
    0x706af48f, 0xe963a535, 0x9e6495a3, 0x0edb8832, 0x79dcb8a4,
    0xe0d5e91e, 0x97d2d988, 0x09b64c2b, 0x7eb17cbd, 0xe7b82d07,
    0x90bf1d91, 0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de,
    0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7, 0x136c9856,
    0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9,
    0xfa0f3d63, 0x8d080df5, 0x3b6e20c8, 0x4c69105e, 0xd56041e4,
    0xa2677172, 0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
    0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940, 0x32d86ce3,
    0x45df5c75, 0xdcd60dcf, 0xabd13d59, 0x26d930ac, 0x51de003a,
    0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423, 0xcfba9599,
    0xb8bda50f, 0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
    0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d, 0x76dc4190,
    0x01db7106, 0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f,
    0x9fbfe4a5, 0xe8b8d433, 0x7807c9a2, 0x0f00f934, 0x9609a88e,
    0xe10e9818, 0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
    0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e, 0x6c0695ed,
    0x1b01a57b, 0x8208f4c1, 0xf50fc457, 0x65b0d9c6, 0x12b7e950,
    0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3,
    0xfbd44c65, 0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2,
    0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb, 0x4369e96a,
    0x346ed9fc, 0xad678846, 0xda60b8d0, 0x44042d73, 0x33031de5,
    0xaa0a4c5f, 0xdd0d7cc9, 0x5005713c, 0x270241aa, 0xbe0b1010,
    0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
    0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17,
    0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad, 0xedb88320, 0x9abfb3b6,
    0x03b6e20c, 0x74b1d29a, 0xead54739, 0x9dd277af, 0x04db2615,
    0x73dc1683, 0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8,
    0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1, 0xf00f9344,
    0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb,
    0x196c3671, 0x6e6b06e7, 0xfed41b76, 0x89d32be0, 0x10da7a5a,
    0x67dd4acc, 0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
    0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252, 0xd1bb67f1,
    0xa6bc5767, 0x3fb506dd, 0x48b2364b, 0xd80d2bda, 0xaf0a1b4c,
    0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55, 0x316e8eef,
    0x4669be79, 0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
    0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f, 0xc5ba3bbe,
    0xb2bd0b28, 0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31,
    0x2cd99e8b, 0x5bdeae1d, 0x9b64c2b0, 0xec63f226, 0x756aa39c,
    0x026d930a, 0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
    0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38, 0x92d28e9b,
    0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21, 0x86d3d2d4, 0xf1d4e242,
    0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1,
    0x18b74777, 0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c,
    0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45, 0xa00ae278,
    0xd70dd2ee, 0x4e048354, 0x3903b3c2, 0xa7672661, 0xd06016f7,
    0x4969474d, 0x3e6e77db, 0xaed16a4a, 0xd9d65adc, 0x40df0b66,
    0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
    0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605,
    0xcdd70693, 0x54de5729, 0x23d967bf, 0xb3667a2e, 0xc4614ab8,
    0x5d681b02, 0x2a6f2b94, 0xb40bbe37, 0xc30c8ea1, 0x5a05df1b,
    0x2d02ef8d
};

impl_gnu_debuglink_t::impl_gnu_debuglink_t(Elf_Scn *scn)
{
    Elf_Data *data = elf_getdata(scn, NULL);
    filename.assign((const char *)data->d_buf);

    // The CRC is 4 byte aligned after the NULL terminated file name.
    int crc_offset((int)(filename.length() + 1));
    crc_offset = (crc_offset + 3) & ~3;
    // TODO: We assume that the ELF file has the same byte-order as
    // the host machine, this is currently always the case, but may
    // change in the future.
    crc32 = *(const uint32_t *)((const char *)data->d_buf + crc_offset);
}

bool
impl_gnu_debuglink_t::is_debuglink(Elf *elf,
				   GElf_Shdr *shdr, Elf_Scn *scn)
{
    size_t shstrndx;
    elf_getshstrndx(elf, &shstrndx);
    const char *name = elf_strptr(elf, shstrndx, (size_t)shdr->sh_name);

    return strcmp(".gnu_debuglink", name) == 0;
}


unsigned long
impl_gnu_debuglink_t::crc32_part(const char *buf, size_t len,
				 unsigned long crc) {
    for (const char *end = buf + len; buf != end; buf++)
	crc = crc32_table[(crc ^ *buf) & 0xff] ^ (crc >> 8);

    return crc;
}

bool
impl_gnu_debuglink_t::verify_checksum(FILE *file) const
{
    unsigned long crc(0xffffffff);
    char buf[1024];

    while (!feof(file)) {
	size_t len(fread(buf, 1, sizeof(buf), file));
	if (len < sizeof(buf) && ferror(file))
	    return false;

	crc = crc32_part(buf, len, crc);
    }

    return this->crc32 == (~crc & 0xffffffff);
}

// ----------------------------------------------------------------------
// impl_gnu_build_id_t
// ----------------------------------------------------------------------

#ifdef ELF_NOTE_GNU
#undef ELF_NOTE_GNU
#endif
#define ELF_NOTE_GNU "GNU"

#ifdef NT_GNU_BUILD_ID
#undef NT_GNU_BUILD_ID
#endif
#define NT_GNU_BUILD_ID	3

struct impl_gnu_build_id_t {
    impl_gnu_build_id_t(const elf_note_t &note);

    // Check if the section contains a valid debug link
    static bool is_debuglink(const elf_note_t &note);

    // Check if a debug file has a valid checksum
    bool verify_checksum(FILE *file) const {
	// The build id not contains no checksum.
	return true;
    }

    string get_filename() const {
	return filename;
    };

private:
    static string hexstring(const char *c, size_t len);
    string filename;
};


impl_gnu_build_id_t::impl_gnu_build_id_t(const elf_note_t &note)
{
    assert(is_debuglink(note));

    const path dir(hexstring(note.desc, 1));
    const path file(hexstring(note.desc + 1,
						 note.descsz - 1) +
				       ".debug");
    filename = (".build-id" / dir / file).string();
}

bool
impl_gnu_build_id_t::is_debuglink(const elf_note_t &note)
{
    return
	sizeof(ELF_NOTE_GNU) == note.namesz &&
	memcmp(ELF_NOTE_GNU, note.name, sizeof(ELF_NOTE_GNU)) == 0 &&
	note.type == NT_GNU_BUILD_ID &&
	note.descsz >= 3;
}

#ifdef _WIN32
#define snprintf _snprintf
#endif

string
impl_gnu_build_id_t::hexstring(const char *c, size_t len)
{
    char *hex = new char [len * 2 + 1];
    for (size_t i = 0; i < len; i++)
	snprintf(hex + 2 * i, 3, "%.2hhx", c[i]);

    string ret = hex;
    delete [] hex;
    return ret;
}


// ----------------------------------------------------------------------
// debuglink_finder_t
// ----------------------------------------------------------------------

template <class DebugLinkImpl>
static
bool
try_debuglink(const string &debug_file,
	      const DebugLinkImpl &link_impl)
{
    FILE *f(fopen(debug_file.c_str(), "r"));
    if (f) {
	bool checksum_ok(link_impl.verify_checksum(f));
	fclose(f);
	return checksum_ok;
    }

    return false;
}

template <class DebugLinkImpl>
static
debuglink_t
get_debuglink(const string &binary,
	      const vector<string> debug_dirs,
	      const DebugLinkImpl &link_impl)
{
    const path binary_path(binary);
    const path binary_dir(binary_path.parent_path());
    const path debug_path(link_impl.get_filename());

    const bool binary_absolute(binary_path.has_root_path());

#define TEST_PATH(p) do {			\
	const string path((p).string());	\
	if (try_debuglink(path, link_impl))	\
	    return debuglink_t(path, true);	\
    } while(0)

    // First check the user specified debug directories
    BOOST_FOREACH(const string &s_dir, debug_dirs) {
	const path debug_dir(s_dir);

	// Check for the data directly in the debug directory
	TEST_PATH(debug_dir / debug_path);

	// Check in debug_dir / `dirname binary`
	if (binary_absolute)
	    TEST_PATH(debug_dir / binary_dir / debug_path);
    }

    // Try the .debug directory in the binary's directory.
    TEST_PATH(binary_dir / ".debug" / debug_path);

    // Check in the binarys directory
    TEST_PATH(binary_dir / debug_path);

#undef TEST_PATH

    return debuglink_t(link_impl.get_filename(), false);
}

void
debuglink_finder_t::add_debug_dir(const string &dir)
{
    debug_dirs.push_back(dir);
}

debuglink_vector_t
debuglink_finder_t::find_links(const string &filename,
			       elf_ptr_t elf_part)
{
    debuglink_vector_t ret;
    Elf *elf(elf_part.get());
    Elf_Scn *scn(NULL);

    while((scn = elf_nextscn(elf, scn)) != 0) {
	GElf_Shdr shdr;
	if (!gelf_getshdr(scn, &shdr))
	    continue;

	if (shdr.sh_type == SHT_NOTE) {
	    elf_note_vector_t notes;
	    if (read_notes(notes, &shdr, scn)) {
		BOOST_FOREACH(const elf_note_t &note, notes) {
		    if (impl_gnu_build_id_t::is_debuglink(note))
			ret.push_back(
			    get_debuglink(filename,
					  debug_dirs,
					  impl_gnu_build_id_t(note)));
		}
	    }
	}

	if (impl_gnu_debuglink_t::is_debuglink(elf, &shdr, scn))
	    ret.push_back(
		get_debuglink(filename,
			      debug_dirs,
			      impl_gnu_debuglink_t(scn)));
    }

    return ret;
}

END_NAMESPACE;
END_NAMESPACE;
