/// @file library/sym/sym.hh

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#ifndef SYM_HH
#define SYM_HH

#include "cppfilt.hh"
#include "string_cache.hh"
#include "library/feedback/feedback.hh"

#include <libelf.h>

#include <map>
#include <vector>
#include <set>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

using boost::shared_ptr;

class sym_t;

// ----- sym_exception_t ------------------------------------------------------

class sym_exception_t : public std::exception {
};

// ----- Address types. -------------------------------------------------------

/// An address in a process's address space.
struct sym_paddr_t {
    uint64_t paddr;

    explicit sym_paddr_t()
        : paddr(0)
        {}

    explicit sym_paddr_t(uint64_t paddr)
        : paddr(paddr)
        {}
};

/// An address in a module's address space.
struct sym_maddr_t {
    uint64_t maddr;

    explicit sym_maddr_t()
        : maddr(0)
        {}

    explicit sym_maddr_t(uint64_t maddr)
        : maddr(maddr)
        {}
};

/// An offset between an address in a module's address space and the
/// corresponding address in a process's address space.
struct sym_mpoffs_t {
    uint64_t offs;

    explicit sym_mpoffs_t(uint64_t offs)
        : offs(offs)
        {}
};

static inline bool
operator==(sym_paddr_t paddr0, sym_paddr_t paddr1)
{
    return paddr0.paddr == paddr1.paddr;
}

static inline bool
operator<(sym_paddr_t paddr0, sym_paddr_t paddr1)
{
    return paddr0.paddr < paddr1.paddr;
}

static inline bool
operator<=(sym_paddr_t paddr0, sym_paddr_t paddr1)
{
    return paddr0.paddr <= paddr1.paddr;
}

static inline bool
operator>=(sym_paddr_t paddr0, sym_paddr_t paddr1)
{
    return paddr0.paddr >= paddr1.paddr;
}

static inline bool
operator==(sym_maddr_t maddr0, sym_maddr_t maddr1)
{
    return maddr0.maddr == maddr1.maddr;
}

static inline bool
operator<(sym_maddr_t maddr0, sym_maddr_t maddr1)
{
    return maddr0.maddr < maddr1.maddr;
}

static inline bool
operator>=(sym_maddr_t maddr0, sym_maddr_t maddr1)
{
    return maddr0.maddr >= maddr1.maddr;
}

static inline sym_mpoffs_t
operator-(sym_paddr_t paddr, sym_maddr_t maddr)
{
    return sym_mpoffs_t(paddr.paddr - maddr.maddr);
}

static inline sym_paddr_t
operator+(sym_maddr_t maddr, sym_mpoffs_t mpoffs)
{
    return sym_paddr_t(maddr.maddr + mpoffs.offs);
}

static inline sym_maddr_t
operator-(sym_paddr_t paddr, sym_mpoffs_t mpoffs)
{
    return sym_maddr_t(paddr.paddr - mpoffs.offs);
}

// ----- sym_paddr_range_t ----------------------------------------------------

/// An address range in a process's address space. Base is the first address in
/// the range and end is the first address outside the range. Base == end means
/// that the upper bound of a range is unknown.
class sym_paddr_range_t {
public:
    const sym_paddr_t base;
    const sym_paddr_t end;

    sym_paddr_range_t(sym_paddr_t base, sym_paddr_t end)
        : base(base), end(end)
        {}

    bool contains(sym_paddr_t addr) const {
        return addr >= base && (addr < end || base == end);
    }
};

// ----------------------------------------------------------------------

typedef shared_ptr<Elf> elf_ptr_t;

// ----- sym_info_t -----------------------------------------------------------

struct sym_info_t {
    sym_info_t()
        : source_line(-1)
        {}

    /// Path to binary/so/exe/dll, and offset. If binary_file is NULL both are
    /// invalid, otherwise both are valid.
    string_cptr_t binary_file;
    uint64_t binary_offset;

    /// ELF symbol / PECOFF export symbol, and offset from the symbol. If
    /// symbol_name is NULL both are invalid, otherwise both are valid.
    string_cptr_t symbol_name;
    uint64_t symbol_offset;

    /// Function according to debug info, and offset into the function. If
    /// function_name is NULL both are invalid, otherwise both are valid.
    string_cptr_t function_name;
    uint64_t function_offset;

    /// Source file according to debug info. NULL if unknown.
    string_cptr_t source_file;

    /// Source line according to debug info. -1 if unknown.
    int64_t source_line;
};

typedef shared_ptr<sym_info_t> sym_info_ptr_t;
typedef shared_ptr<const sym_info_t> sym_info_cptr_t;

typedef map<sym_paddr_t, sym_info_cptr_t> sym_info_cache_t;

// ----- sym_base_t -----------------------------------------------------------

class sym_base_t {
public:
    sym_t &sym;
    const sym_paddr_range_t range;

    sym_base_t(sym_t &sym, sym_paddr_t base, sym_paddr_t end)
	: sym(sym), range(base, end)
	{}

    virtual ~sym_base_t()
	{}

    virtual void lookup_sym_info(sym_paddr_t address,
                                 sym_info_ptr_t &sym_info) = 0;

protected:
    void cannot_cope()
	{
	    throw sym_exception_t();
	}
};

typedef shared_ptr<sym_base_t> sym_base_ptr_t;

typedef multimap<sym_paddr_t, sym_base_ptr_t> base_to_sym_t;

/// ----- sym_t ---------------------------------------------------------------

/// NOTE: None of the functions in sym_t may ever throw sym_exception_t.
class sym_t {
public:

    enum global_features_t {G_None = 0,
			    G_Demangle_Names = 1};

    enum feature_t {None = 0,
		    Line = 1,
		    Public_Functions = 2,
		    Full_Functions = 4};

    sym_t(int global_features = G_Demangle_Names,
	  const string &cppfilt="",
          bool verbose = false,
          feedback_ptr_t feedback = feedback_ptr_t());
    ~sym_t();

    enum debug_info_status_t {
        File_Not_Found,
        Error_Reading_File,
        No_Info_Found,
        Symbols_Found,
        Debug_Info_Found
    };

    static const char *describe_debug_info_status(
        debug_info_status_t debug_info_status);

    debug_info_status_t add_binary(const string &filename,
                                   sym_paddr_t load_offset,
                                   int features = Full_Functions | Line);

    sym_info_cptr_t lookup_sym_info(sym_paddr_t address);

    void add_debug_link_dir(const string &d) {
	debug_link_dirs.push_back(d);
    }

    void add_source_dir(const string &dir);
    string_cptr_t lookup_source_file(const string &filename);

    cppfilt_t cppfilt;

private:
    const bool verbose;
    const feedback_ptr_t feedback;

    void log_verbose(const string &message);

    bool file_exists(const string &filename);

    /// When we follow debug links, we still want to use the original filename
    /// in the results we return. We therefore need both actual_filename which
    /// the name of the ELF file we should open (which may a file found through
    /// a debug link), and original_filename which is the filename we should
    /// use in the results we return.
    debug_info_status_t process_elf_part(elf_ptr_t elf_file,
                                         elf_ptr_t elf_part,
                                         const string &actual_filename,
                                         const string &original_filename,
                                         sym_paddr_t load_address,
                                         int features,
                                         set<string> &processed_files);

    /// When we follow debug links, we still want to use the original filename
    /// in the results we return. We therefore need both actual_filename which
    /// the name of the ELF file we should open (which may a file found through
    /// a debug link), and original_filename which is the filename we should
    /// use in the results we return.
    debug_info_status_t process_elf_file(const string &actual_filename,
                                         const string &original_filename,
                                         sym_paddr_t load_address,
                                         int features,
                                         set<string> &processed_files);

#ifdef _WIN32
    debug_info_status_t process_pecoff_file(const string &filename,
                                            sym_paddr_t load_address,
                                            int features);
#endif

    /// Maps base addresses to sym_base_t objects.
    base_to_sym_t base_to_sym;

    vector<string> debug_link_dirs;

    // Source file lookup.
    vector<string> source_dirs;
    typedef map<string, string_cptr_t> source_file_cache_t;
    source_file_cache_t source_file_cache;

    /// Symbol info cache.
    sym_info_cache_t sym_info_cache;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
