/// @file library/sym/sym_base.cc

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#include "sym_dwarf.hh"
#include "sym_stabs.hh"
#include "sym_elf.hh"
#include "sym_binary.hh"

#include "debuglink.hh"

#ifdef _WIN32
#include "sym_pdb.hh"
#include "pecoff.hh"
#endif

#include <boost/foreach.hpp>

// for stat, according to stat(2) man page
#include <sys/types.h>
#include <sys/stat.h>
#ifndef _WIN32
#include <unistd.h>
#endif
#include <assert.h>
#include <stdio.h>

#include <iostream>
#include <iomanip>

#include <fcntl.h>
#ifdef _WIN32
#include <io.h>
#endif

#ifdef _WIN32
#ifndef S_ISREG
#define S_ISREG(x) (((x) & _S_IFREG) != 0)
#endif
#endif

#ifdef _WIN32
#pragma comment(lib, "libelf")
#endif

#include <libelf.h>
#include <gelf.h>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

// ----------------------------------------------------------------------

sym_t::sym_t(int global_features,
             const string &cppfilt,
             bool verbose,
             feedback_ptr_t feedback)
    : cppfilt((global_features & G_Demangle_Names) ?
              (!cppfilt.empty() ? cppfilt.c_str() : "c++filt") :
              ""),
      verbose(verbose),
      feedback(feedback)
{
}

sym_t::~sym_t()
{
}

sym_info_cptr_t
sym_t::lookup_sym_info(sym_paddr_t address)
{
    sym_info_cache_t::iterator cache_i = sym_info_cache.find(address);
    if (cache_i != sym_info_cache.end())
        return cache_i->second;

    sym_info_ptr_t s(new sym_info_t());

    base_to_sym_t::iterator i = base_to_sym.upper_bound(address);
    while (i != base_to_sym.begin() && (--i)->second->range.base <= address) {
        sym_base_ptr_t &sym = i->second;
        if (sym->range.contains(address))
            sym->lookup_sym_info(address, s);
    }

    sym_info_cache.insert(make_pair(address, s));

    return s;
}

sym_t::debug_info_status_t
sym_t::process_elf_part(elf_ptr_t elf_file,
                        elf_ptr_t elf_part,
                        const string &actual_filename,
                        const string &original_filename,
			sym_paddr_t load_address,
			int features,
                        set<string> &processed_files)
{
    // The file has obviously already been found, so the worst result we could
    // get is an error reading it.
    debug_info_status_t result = Error_Reading_File;

    debuglink_finder_t debug_finder(debug_link_dirs);
    debuglink_vector_t debug_links(
        debug_finder.find_links(actual_filename, elf_part));

    BOOST_FOREACH(const debuglink_t &dl, debug_links)
        if (dl.valid &&
            processed_files.find(dl.filename) == processed_files.end())
            result = max(result,
                         process_elf_file(dl.filename, original_filename,
                                          load_address, features,
                                          processed_files));

    sym_maddr_t base;
    sym_maddr_t end;

    GElf_Ehdr ehdr;
    GElf_Ehdr *ehdrp = gelf_getehdr(elf_part.get(), &ehdr);
    if (ehdrp != NULL)
        for (int i = 0; i < ehdr.e_phnum; i++) {
            GElf_Phdr phdr;
            GElf_Phdr *phdrp = gelf_getphdr(elf_part.get(), i, &phdr);
            if (phdrp != NULL &&
                phdr.p_type == PT_LOAD &&
                (phdr.p_flags & PF_X) != 0) {
                base = sym_maddr_t(phdr.p_vaddr);
                end = sym_maddr_t(phdr.p_vaddr + phdr.p_memsz);
                goto text_found;
            }
        }
    return result;

text_found:
    // Assume that the binary has not been relocated if load_address is 0.
    sym_mpoffs_t load_offset(load_address == sym_paddr_t(0) ?
                             sym_mpoffs_t(0) :
                             load_address - base);

    try {
        sym_base_ptr_t sym_binary(
            new sym_binary_t(
                *this, base + load_offset, end + load_offset, original_filename));

        base_to_sym.insert(make_pair(sym_binary->range.base, sym_binary));

        result = max(result, No_Info_Found);
    } catch (sym_exception_t &) {
        // TODO
    }

    try {
        sym_base_ptr_t sym_elf(
            new sym_elf_t(*this, base + load_offset, end + load_offset,
                          elf_part, load_offset));

        base_to_sym.insert(make_pair(sym_elf->range.base, sym_elf));

        result = max(result, Symbols_Found);
    } catch (sym_exception_t &) {
        // TODO
    }

    try {
        dwarf_debug_ptr_t dd(new dwarf_debug_elf_t(elf_file, elf_part, features));

        sym_base_ptr_t dwarf(
            new sym_dwarf_t(*this, base + load_offset, end + load_offset,
                            dd, load_offset, features));

        base_to_sym.insert(make_pair(dwarf->range.base, dwarf));

        result = max(result, Debug_Info_Found);
    } catch (sym_exception_t &) {
        // TODO
    }

    try {
        sym_base_ptr_t stabs(
            new sym_stabs_t(*this, base + load_offset, end + load_offset,
                            elf_part, load_offset, features));

        base_to_sym.insert(make_pair(stabs->range.base, stabs));

        result = max(result, Debug_Info_Found);
    } catch (sym_exception_t) {
        // TODO
    }

    return result;
}

sym_t::debug_info_status_t
sym_t::process_elf_file(const string &actual_filename,
                        const string &original_filename,
			sym_paddr_t load_address,
			int features,
                        set<string> &processed_files)
{
    processed_files.insert(actual_filename);

    int filedesc;

    // Make sure we are not using an incompatible libelf library.
    if (elf_version(EV_CURRENT) == EV_NONE) {
	cerr << "Error: Dynamic library libelf.so has wrong version." << endl;
        // TODO: Not an appropriate error code.
	return File_Not_Found;
    }

#if defined(__CYGWIN__) || defined(_WIN32)
    filedesc = _open(actual_filename.c_str(), O_RDONLY | O_BINARY);
#else
    filedesc = open(actual_filename.c_str(), O_RDONLY);
#endif

    if (filedesc == -1)
        // TODO: Should check if the file exists but is not readable and return
        //       a different error message in that case.
	return File_Not_Found;

    Elf_Cmd cmd;
    Elf *arf, *elf;
    cmd = ELF_C_READ;
    arf = elf_begin(filedesc, cmd, 0);
    elf_ptr_t elf_file(arf, elf_end);

    // The file has obviously already been found, so the worst result we could
    // get is an error reading it.
    debug_info_status_t result = Error_Reading_File;

    while ((elf = elf_begin(filedesc, cmd, arf)) != 0) {
	elf_ptr_t elf_part(elf, elf_end);

        result = max(result, process_elf_part(
                         elf_file, elf_part, actual_filename,
                         original_filename, load_address, features,
                         processed_files));

	cmd = elf_next(elf);
    }

    close(filedesc);

    return result;
};

#ifdef _WIN32
sym_t::debug_info_status_t
sym_t::process_pecoff_file(const string &filename,
			   sym_paddr_t load_address,
			   int features)
{
    sym_maddr_t base;
    sym_maddr_t end;

    LOADED_IMAGE image;
    if (!MapAndLoad(filename.c_str(), NULL, &image, FALSE, TRUE))
        return Error_Reading_File;
    assert(image.FileHeader != NULL);
    switch (image.FileHeader->FileHeader.Machine) {
    case IMAGE_FILE_MACHINE_I386: {
        // TODO: Is this correct?
        IMAGE_OPTIONAL_HEADER32 *oh =
            (IMAGE_OPTIONAL_HEADER32 *)&image.FileHeader->OptionalHeader;
        base = sym_maddr_t(oh->ImageBase);
        end = sym_maddr_t(oh->ImageBase + oh->SizeOfImage);
    } break;
    case IMAGE_FILE_MACHINE_AMD64: {
        // TODO: Is this correct?
        IMAGE_OPTIONAL_HEADER64 *oh =
            (IMAGE_OPTIONAL_HEADER64 *)&image.FileHeader->OptionalHeader;
        base = sym_maddr_t(oh->ImageBase);
        end = sym_maddr_t(oh->ImageBase + oh->SizeOfImage);
    } break;
    default:
        return Error_Reading_File;
    }
    if (!UnMapAndLoad(&image))
        return Error_Reading_File;

    // We have been able to read the file the worst result we can get is that
    // there  is neither symbols nor debug info.
    debug_info_status_t result = No_Info_Found;

    // Assume that the binary has not been relocated if load_address is 0.
    sym_mpoffs_t load_offset =
        (load_address == sym_paddr_t(0)) ? sym_mpoffs_t(0) : (load_address - base);

    try {
        sym_base_ptr_t sym_binary(
            new sym_binary_t(
                *this, base + load_offset, end + load_offset, filename));

        base_to_sym.insert(make_pair(sym_binary->range.base, sym_binary));

        result = max(result, No_Info_Found);
    } catch (sym_exception_t &) {
        // TODO
    }

    try {
        sym_pdb_t *p =
            new sym_pdb_t(*this, base + load_offset, end + load_offset,
                          filename, load_address, features);

        sym_base_ptr_t pdb(p);

        base_to_sym.insert(make_pair(pdb->range.base, pdb));

        // We may get either debug information or export symbol information
        // from sym_pdb_t, so ask it what kind we got.
        result = max(result, p->get_debug_info_status());
    } catch (sym_exception_t &) {
        // TODO
    }

    try {
        libdwarf_object_ptr_t object(new pecoff_t(filename));
        dwarf_debug_ptr_t dd(new dwarf_debug_object_t(object));

        sym_base_ptr_t dwarf(
            new sym_dwarf_t(*this, base + load_offset, end + load_offset,
                            dd, load_offset, features));

        base_to_sym.insert(make_pair(dwarf->range.base, dwarf));

        result = max(result, Debug_Info_Found);
    } catch (sym_exception_t &) {
        // TODO
    }

    // TODO: STABS support? The STABS implementation needs to be refactored to
    // be able to read pe-coff sections. Right now, sym_stabs_t is using elf
    // directly.

    return result;
}
#endif

const char *
sym_t::describe_debug_info_status(
    debug_info_status_t debug_info_status)
{
    switch (debug_info_status) {
    case File_Not_Found:
        return "File not found";
    case Error_Reading_File:
        return "Error reading file";
    case No_Info_Found:
        return "No debug info found";
    case Symbols_Found:
        return "Only symbols found";
    case Debug_Info_Found:
        return "Debug info found";
    default:
        while (1)
            assert(0);
    }
}

sym_t::debug_info_status_t
sym_t::add_binary(const string &filename,
                  sym_paddr_t load_address,
                  int features)
{
    // If ELF
    //    Loop over all parts
    //       Try sym_dwarf_t
    //          -> ELF + DWARF (done)
    //       Try sym_stabs_t
    //          -> ELF + STABS (done)
    //    else -> (no symbols)
    //
    // If PE-COFF
    //    Try sym_pdb_t
    //       -> PE-COFF + PDB (done)
    //    Try sym_dwarf_t
    //       -> PE-COFF + DWARF (done)
    //   [ Try sym_stabs_t              ] -> needs refactoring to get rid
    //   [    -> PE-COFF + STABS (done) ] -> of explicit elf calls
    //    else -> (no symbols)
    //
    // -> (no symbols)

    unsigned char elf[] = {0x7f, 'E', 'L', 'F'};
#ifdef _WIN32
    unsigned char pe[] =  {'M', 'Z'};
#endif

    FILE *f = fopen(filename.c_str(), "rb");
    if (f) {
	unsigned buf[4];
	size_t s = fread(buf, 1, 4, f);
	fclose(f);
	if (s >= 4 && memcmp(buf, elf, sizeof(elf)) == 0) {
            set<string> processed_files;
	    return process_elf_file(filename, filename, load_address,
                                    features, processed_files);
        }
#ifdef _WIN32
	else if (s >= 2 && memcmp(buf, pe, sizeof(pe)) == 0)
	    return process_pecoff_file(filename, load_address, features);
#endif
        else
            return Error_Reading_File;
    }
    else
        // TODO: Should check if the file exists but is not readable and return
        //       a different error message in that case.
        return File_Not_Found;
}

void
sym_t::add_source_dir(const string &dir)
{
    source_dirs.push_back(dir);
}

void
sym_t::log_verbose(const string &message)
{
    if (verbose && feedback)
        feedback->log(feedback_base_t::LOG_INFO, message);
}

bool
sym_t::file_exists(const string &filename)
{
    struct stat buf;
    int ret = stat(filename.c_str(), &buf);
    if (ret != -1 && S_ISREG(buf.st_mode)) {
        log_verbose("     Tried " + filename + ", found!");
        return true;
    }
    else {
        log_verbose("     Tried " + filename + ", not found.");
        return false;
    }
}

string_cptr_t
sym_t::lookup_source_file(const string &filename)
{
    source_file_cache_t::iterator i = source_file_cache.find(filename);
    if (i != source_file_cache.end())
        return i->second;

    log_verbose("     Looking for " + filename + ":");

    BOOST_FOREACH(const string &dir, source_dirs) {
        const char *q = filename.c_str();
        while (q && *q) {
            string test = dir + q;

            if (file_exists(test))
                return (source_file_cache[filename] =
                        string_cptr_t(new string(test)));

            const char *q1 = strchr(q + 1, '/');
            const char *q2 = strchr(q + 1, '\\');
            if (q1 == NULL)
                q = q2;
            else if (q2 == NULL)
                q = q1;
            else
                q = min(q1, q2);
        }
    }

    if (!file_exists(filename))
        log_verbose("     No file matching " + filename + " found.");

    return (source_file_cache[filename] = string_cptr_t(new string(filename)));
}

END_NAMESPACE;
END_NAMESPACE;
