/// @file library/sym/sym_dwarf.cc

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

#include "sym_dwarf.hh"

// Need to bring in this private header file to arrange for shared_ptr
// lifetime management of aranges.
#include <libdwarf/dwarf_arange.h>

// Need this header file to bring in some constants, like DW_TAG_subprogram.
#include <libdwarf/dwarf.h>

#include <boost/foreach.hpp>
#include <sstream>

#include <stdio.h>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;
using namespace boost;

// ----------------------------------------------------------------------

dwarf_debug_elf_t::dwarf_debug_elf_t(elf_ptr_t elf_file,
                                     elf_ptr_t elf_part,
				     int features)
    : elf_file(elf_file),
      elf_part(elf_part)
{
    int ret;
    Dwarf_Error err;
    ret = dwarf_elf_init(elf_part.get(), DW_DLC_READ, NULL, NULL, &dbg, &err);
    if (ret != DW_DLV_OK) {
	throw sym_exception_t();
    }
}

dwarf_debug_elf_t::~dwarf_debug_elf_t()
{
    Dwarf_Error err;
    dwarf_finish(dbg, &err);
}

// ----------------------------------------------------------------------

dwarf_debug_object_t::dwarf_debug_object_t(libdwarf_object_ptr_t obj)
    : obj(obj)
{
    Dwarf_Error err;
    int ret;
    if (!obj->init(&interface)) {
	throw sym_exception_t();
    }
    ret = dwarf_object_init(&interface, NULL, NULL, &dbg, &err);
    if (ret != DW_DLV_OK) {
	throw sym_exception_t();
    }
}

dwarf_debug_object_t::~dwarf_debug_object_t()
{
    Dwarf_Error err;
    dwarf_object_finish(dbg, &err);
}

/// Exception originating from a libdwarf function error.
class dwarf_exception_t : public sym_exception_t {
public:
    dwarf_exception_t(const Dwarf_Error &err, const char *file, int line) {
	stringstream ss;
	ss << "dwarf_exception_t, " << file
	   << " line " << line
	   << ": " << dwarf_errmsg(err);
	text = ss.str();
	error = (int)dwarf_errno(err);
    }

    ~dwarf_exception_t() throw() {}

    const char *what() const throw() {
        return text.c_str();
    }

    string text;
    int error;
};

// ----------------------------------------------------------------------

/// error_t is a wrapper for a Dwarf_Error, and handles throwing
/// exceptions upon serious errors.
struct error_t {
    error_t(dwarf_debug_ptr_t dbg)
	: dbg(dbg), err(0)
	{
	    // Make sure the dwarf context is valid
	    assert(dbg.get() != 0);
	}

    ~error_t()
	{
	    dealloc();
	}

    void dealloc()
	{
	    if (err) {
		dwarf_dealloc(*dbg, err, DW_DLA_ERROR);
		err = 0;
	    }
	}

    operator Dwarf_Error * ()
	{
	    dealloc();
	    return &err;
	}

    void raise(const char *file, int line) const
	{
	    throw dwarf_exception_t(err, file, line);
	}

    void check(int ret, const char *file, int line) const
	{
	    if (ret == DW_DLA_ERROR)
		raise(file, line);
	}

    int errno_code() const
	{
	    if (err == 0)
		return 0;
	    return (int)dwarf_errno(err); // XXX const_cast
	}

    dwarf_debug_ptr_t dbg;
    Dwarf_Error err;
};

// ----------------------------------------------------------------------
// Scope controlled deleters

class deleter_t {
public:
    deleter_t(Dwarf_Unsigned type, dwarf_debug_ptr_t dbg, void *p)
	: type(type), dbg(dbg), p(p)
	{
	}

    ~deleter_t()
	{
	    dwarf_dealloc(*dbg, p, type);
	}
private:
    Dwarf_Unsigned type;
    dwarf_debug_ptr_t dbg;
    void *p;
};

class deleter_globals_t {
public:
    deleter_globals_t(dwarf_debug_ptr_t dbg,
		      Dwarf_Global *globs,
		      Dwarf_Signed cnt)
	: dbg(dbg), globs(globs), cnt(cnt)
	{
	}

    ~deleter_globals_t()
	{
	    dwarf_globals_dealloc(*dbg, globs, cnt);
	}
private:
    dwarf_debug_ptr_t dbg;
    Dwarf_Global *globs;
    Dwarf_Signed cnt;
};

// ----------------------------------------------------------------------
// deleters for use in a shared_ptr

struct dwarf_srclines_deleter_t {
    dwarf_srclines_deleter_t(dwarf_debug_ptr_t dbg, Dwarf_Signed linecount)
	: dbg(dbg), linecount(linecount) {}

    void operator () (Dwarf_Line *linebuf) {
	dwarf_srclines_dealloc(*dbg, linebuf, linecount);
    }

    dwarf_debug_ptr_t dbg;
    Dwarf_Signed linecount;
};

struct dwarf_die_null_deleter_t {
    void operator() (Dwarf_Die die) {
    }
};

struct dwarf_die_deleter_t {
    dwarf_die_deleter_t(dwarf_debug_ptr_t dbg, Dwarf_Unsigned type)
	: dbg(dbg), type(type)
	{}

    void operator () (Dwarf_Die die)
	{
	    dwarf_dealloc(*dbg, die, type);
	}
    dwarf_debug_ptr_t dbg;
    Dwarf_Unsigned type;
};

// ----------------------------------------------------------------------

// XXX make sure that dbg and part_file life times are tied.
// dbg.

// XXX make sure line_numbers, aranges and globals are properly freed
// if the constructor throws an exceptions.

sym_dwarf_t::sym_dwarf_t(sym_t &sym,
                         sym_paddr_t base,
                         sym_paddr_t end,
			 dwarf_debug_ptr_t dbg,
			 sym_mpoffs_t load_offset,
			 int features)
    : sym_base_t(sym, base, end),
      dbg(dbg),
      load_offset(load_offset),
      features(features),
      function_cache(sym),
      statement_cache(sym)
{
    // loop over all cu
    Dwarf_Unsigned cu_header_length = 0;
    Dwarf_Unsigned abbrev_offset = 0;
    Dwarf_Half version_stamp = 0;
    Dwarf_Half address_size = 0;
    Dwarf_Unsigned next_cu_header = 0;
    error_t err(dbg);
    bool found_dwarf = false;
    int ret;

    try {
	while ((ret = dwarf_next_cu_header(*dbg,
					   &cu_header_length,
					   &version_stamp,
					   &abbrev_offset,
					   &address_size,
					   &next_cu_header,
					   err)) == DW_DLV_OK) {
	    if (features & sym_t::Line)
		if (register_line_numbers_this_cu())
		    found_dwarf = true;
	}

	// If we allow ad hoc full-function searches, we should
	// cache the aranges here.
	if (features & sym_t::Full_Functions)
	    if (register_aranges())
		found_dwarf = true;

	if (features & (sym_t::Full_Functions | sym_t::Public_Functions))
	    if (register_globals())
		found_dwarf = true;

    } catch (sym_exception_t) {
	// In case anything throws a dwarf error, and we let that
	// exception leave the constructor, we need to clean up the
	// arange_vector manually, as the destructor will not be
	// invoked.

	BOOST_FOREACH(arange_t &a, arange_vector) {
	    for (int i = 0; i < a.count; i++)
		dwarf_dealloc(*dbg, a.arange[i], DW_DLA_ARANGE);
	    dwarf_dealloc(*dbg, a.arange, DW_DLA_LIST);
	}

	throw;
    }

    if (!found_dwarf)
	cannot_cope();
}

sym_dwarf_t::~sym_dwarf_t()
{
    // clean up arange_vector
    BOOST_FOREACH(arange_t &a, arange_vector) {
	for (int i = 0; i < a.count; i++)
	    dwarf_dealloc(*dbg, a.arange[i], DW_DLA_ARANGE);
	dwarf_dealloc(*dbg, a.arange, DW_DLA_LIST);
    }
}

void
sym_dwarf_t::dwarf_finish()
{
    dbg.reset();
}

void
sym_dwarf_t::dwarf_debug_deleter(Dwarf_Debug d) throw()
{
    Dwarf_Error err;
    ::dwarf_finish(d, &err);
}

bool
sym_dwarf_t::register_globals()
{
    Dwarf_Signed cnt;
    Dwarf_Global *globs = 0;
    error_t err(dbg);
    int ret;
    ret = dwarf_get_globals(*dbg, &globs, &cnt, err);
    err.check(ret, __FILE__, __LINE__);
    if (ret != DW_DLV_OK)
	return false;

    deleter_globals_t deleter(dbg, globs, cnt);
    Dwarf_Off last_cu_header_offset = 0;
    string filename;

    for (int i=0; i<cnt; i++) {
	// find the cu die, the die, and register the function,

	Dwarf_Off cu_header_offset;
	ret = dwarf_global_cu_offset(globs[i], &cu_header_offset, err);
	err.check(ret, __FILE__, __LINE__);
	if (ret != DW_DLV_OK)
	    continue;

	// See if we need to look up the cu, or if it is the same as
	// last function.
	if (last_cu_header_offset != cu_header_offset) {
	    Dwarf_Off cu_offset;
	    ret = dwarf_get_cu_die_offset_given_cu_header_offset(
		*dbg, cu_header_offset, &cu_offset, err);
	    err.check(ret, __FILE__, __LINE__);
	    if (ret != DW_DLV_OK)
		continue;

	    Dwarf_Die cu_die = 0;
	    ret = dwarf_offdie(*dbg, cu_offset, &cu_die, err);
	    err.check(ret, __FILE__, __LINE__);
	    if (ret != DW_DLV_OK)
		continue;
	    deleter_t d2(DW_DLA_DIE, dbg, cu_die);

	    string f;
	    if (!get_filename(cu_die, f))
		continue;

	    last_cu_header_offset = cu_header_offset;
	    filename = f;
	}

	Dwarf_Off offset;
	ret = dwarf_global_die_offset(globs[i], &offset, err);
	err.check(ret, __FILE__, __LINE__);
	if (ret != DW_DLV_OK)
	    continue;

	Dwarf_Die die = 0;
	ret = dwarf_offdie(*dbg, offset, &die, err);
	err.check(ret, __FILE__, __LINE__);
	if (ret != DW_DLV_OK)
	    continue;
	deleter_t d(DW_DLA_DIE, dbg, die);

	address_range_t address_range;
	if (!get_address_range(die, address_range))
	    continue;

	string name;
	if (!get_subprogram_name_recursive(die, name))
	    continue;

	function_cache.add_function(address_range.start + load_offset,
                               address_range.end + load_offset,
                               name);
    }

    return true;
}

void
sym_dwarf_t::lookup_sym_info(sym_paddr_t address,
                             sym_info_ptr_t &sym_info)
{
    // Look for file and line in the cache.
    statement_ptr_t statement;
    if (statement_cache.lookup_statement(address, statement)) {
        if (statement->filename)
            sym_info->source_file = statement->filename;
        if (statement->row != -1)
            sym_info->source_line = statement->row;
    }

    // Look for function in the cache.
    function_ptr_t function;
    if (function_cache.lookup_function(address, function)) {
        sym_info->function_name = function->name;
        sym_info->function_offset = address.paddr - function->range.base.paddr;
    }
    else
        // No function found in the cache. Try to look it up in the debug info
        // if the full functions feature is enabled.
        if (features & sym_t::Full_Functions) {
            Dwarf_Die cu_die;
            if (find_cu_die(address - load_offset, cu_die) &&
                scan_functions(address, cu_die, function)) {
                sym_info->function_name = function->name;
                sym_info->function_offset = address.paddr - function->range.base.paddr;
            }
        }
}

bool
sym_dwarf_t::register_aranges()
{
    arange_t a;
    error_t err(dbg);
    int ret = dwarf_get_aranges(*dbg, &a.arange, &a.count, err);
    err.check(ret, __FILE__, __LINE__);

    if (ret != DW_DLV_OK)
	return false;

    arange_vector.push_back(a);
    return true;
}

bool
sym_dwarf_t::register_line_numbers_this_cu()
{
    Dwarf_Die cu_die;
    error_t err(dbg);

    int ret = dwarf_siblingof(*dbg, NULL, &cu_die, err);
    err.check(ret, __FILE__, __LINE__);

    string name;
    get_name(cu_die, name);

    Dwarf_Signed linecount;
    dwarf_line_ptr_t lines;
    Dwarf_Signed i;

    if (!get_srclines(cu_die, lines, linecount))
	return false;

    for (i = 0; i < linecount; i++) {
	Dwarf_Line line = lines[(ptrdiff_t)i];

	// DW_LINE_end_sequence terminates a line number
	// program. Don't add this address to the line mapping
	// database, but proceed to the next line instead.
	int end_sequence;
	ret = dwarf_lineendsequence(line, &end_sequence, err);
	err.check(ret, __FILE__, __LINE__);
	if (end_sequence)
	    continue;

	string filename;
	if (!get_linesrc(line, filename))
	    return false; // XXX continue ?

	Dwarf_Unsigned lineno;
	ret = dwarf_lineno(line, &lineno, err);
	err.check(ret, __FILE__, __LINE__);
	if (ret == DW_DLV_NO_ENTRY)
	    return false; // XXX continue ?

	Dwarf_Signed column;
	ret = dwarf_lineoff(line, &column, err);
	err.check(ret, __FILE__, __LINE__);
	if (ret == DW_DLV_NO_ENTRY)
	    column = -1LL;

	Dwarf_Addr pc;
	ret = dwarf_lineaddr(line, &pc, err);
	err.check(ret, __FILE__, __LINE__);
	if (ret == DW_DLV_NO_ENTRY)
	    return false; // XXX continue ?

	Dwarf_Addr pc2 = 0;
	if (i + 1 < linecount) {
	    ret = dwarf_lineaddr(lines[(ptrdiff_t)(i + 1)], &pc2, err);
	    err.check(ret, __FILE__, __LINE__);
	    // TODO: Is this a good fallback?
	    if (ret == DW_DLV_NO_ENTRY)
		pc2 = pc;
	} else {
	    ret = dwarf_highpc(cu_die, &pc2, err);
	    err.check(ret, __FILE__, __LINE__);
	    // TODO: Is this a good fallback?
	    if (ret == DW_DLV_NO_ENTRY)
		pc2 = pc;
	}
        statement_cache.add_statement(sym_maddr_t(pc) + load_offset,
                                      sym_maddr_t(pc2) + load_offset,
                                      filename,
                                      (int64_t)lineno,
                                      (int64_t)column);
    }
    return true;
}

bool
sym_dwarf_t::find_arange(sym_maddr_t addr,
			 Dwarf_Arange &ret_arange) const
{
    error_t err(dbg);
    BOOST_FOREACH(const arange_t &a, arange_vector) {
	int ret = dwarf_get_arange(a.arange, a.count,
				   addr.maddr, &ret_arange, err);
	err.check(ret, __FILE__, __LINE__);

	if (ret == DW_DLV_OK)
	    return true;
    }

    return false;
}

bool
sym_dwarf_t::find_cu_die(sym_maddr_t addr,
			 Dwarf_Die &ret_die) const
{
    // First, get the arange that covers the address
    Dwarf_Arange arange;
    if (!find_arange(addr, arange))
	return false;

    error_t err(dbg);

    // Find the offset of the compilation_unit die that is
    // associated with this arange.
    Dwarf_Off offset;
    int ret = dwarf_get_cu_die_offset(arange, &offset, err);
    err.check(ret, __FILE__, __LINE__);

    if (ret != DW_DLV_OK)
	return false;

    // Find the actual compilation unit die
    ret = dwarf_offdie(*dbg, offset, &ret_die, err);
    err.check(ret, __FILE__, __LINE__);

    if (ret == DW_DLV_OK)
	return true;

    return false;
}

bool
sym_dwarf_t::scan_functions(sym_paddr_t paddr,
			    Dwarf_Die cu_die,
			    function_ptr_t &ret_function)
{
    int ret;
    error_t err(dbg);

    // We use the cu offset as a key to our incremental workload map
    Dwarf_Off cu_off;
    ret = dwarf_dieoffset(cu_die, &cu_off, err);
    err.check(ret, __FILE__, __LINE__);

    cu_dwarf_die_map_t::iterator x = cu_dwarf_die_map.find(cu_off);

    // exhausted searching this compilation unit, no match, return false
    if (x != cu_dwarf_die_map.end() && x->second.size() == 0)
	return false;

    // never seen this cu, add a starting point
    if (x == cu_dwarf_die_map.end()) {
	x = cu_dwarf_die_map.insert(make_pair(cu_off,
					      dwarf_die_list_t())).first;
	x->second.push_front(
	    dwarf_die_ptr_t(cu_die, dwarf_die_null_deleter_t()));
    }

    // get directory (DW_AT_comp_dir) and filename
    // (DW_AT_name) from cu_die

    string filename;
    if (!get_filename(cu_die, filename)) {
	// if we can't get the filename, then we should not bother
	// with this cu. Clear the work list to avoid getting this far
	// in the future.
	x->second.clear();
	return false;
    }

    dwarf_die_list_t &queue = x->second;

    while(!queue.empty()) {
	dwarf_die_ptr_t dp = queue.front();
	queue.pop_front();

	Dwarf_Die die = dp.get();

	// Get tag
	Dwarf_Half tag;
	ret = dwarf_tag(die, &tag, err);
	err.check(ret, __FILE__, __LINE__);

	if (ret == DW_DLV_OK) {

	    // add sibling and child to the queue, in such a way as to
	    // have a post-order depth-first search
	    Dwarf_Die sibling;
	    ret = dwarf_siblingof(*dbg, die, &sibling, err);
	    if (ret == DW_DLV_ERROR) {
                // We could get this error when analyzing.
                if (err.errno_code() == DW_DLE_FIRST_DIE_NOT_CU)
                    continue;

		// We have seen the PGI fortran compiler emit dwarf
		// that causes DW_DLE_NEXT_DIE_PTR_NULL to be
		// indicated. It seems to work to neglect this error
		// code.
		if (err.errno_code() != DW_DLE_NEXT_DIE_PTR_NULL) {
		    err.raise(__FILE__, __LINE__);
		}
	    }
	    if (ret == DW_DLV_OK)
		queue.push_front(
		    dwarf_die_ptr_t(sibling,
				    dwarf_die_deleter_t(dbg, DW_DLA_DIE)));

	    Dwarf_Die child;
	    ret = dwarf_child(die, &child, err);
	    err.check(ret, __FILE__, __LINE__);
	    if (ret == DW_DLV_OK)
		queue.push_front(
		    dwarf_die_ptr_t(child,
				    dwarf_die_deleter_t(dbg, DW_DLA_DIE)));

	    // Now, looking at the current node, see if we are interested.
	    // If it is a subprogram, cache it. If it also intersects our
	    // query address, return the function object.
	    if (tag == DW_TAG_subprogram ||
		tag == DW_TAG_inlined_subroutine) { // XXX others?

		// locate low_pc and high_pc, and see that
		// addr is contained in this range
		address_range_t address_range;
		if (!get_address_range(die, address_range))
		    continue;

		// Find the name of this function by traversing back through
		// DW_AT_specification and DW_AT_abstract_origin attributes
		string name;
		if (!get_subprogram_name_recursive(die, name))
		    continue;

                function_ptr_t function =
                    function_cache.add_function(address_range.start + load_offset,
                                                address_range.end + load_offset,
                                                name);

		if (address_range.within(paddr - load_offset)) {
		    // Indicate a match, and postpone the rest of the search
		    ret_function = function;
		    return true;
		}
	    }
	}
    }

    // Indicate no match in this compilation unit. At this point the queue is
    // empty, and this causes the search not to be repeated.
    return false;
}

bool
sym_dwarf_t::get_subprogram_name_recursive(Dwarf_Die die,
					   string &name) const
{
    if (get_name(die, name))
	return true;

    Dwarf_Die spec;
    if (get_specification(die, spec) &&
        get_subprogram_name_recursive(spec, name))
        return true;

    if (get_abstract_origin(die, spec) &&
        get_subprogram_name_recursive(spec, name))
        return true;

    return false;
}

bool
sym_dwarf_t::get_name(Dwarf_Die die, string &name) const
{
    error_t err(dbg);
    char *n = 0;
    int ret = dwarf_diename(die, &n, err);
    err.check(ret, __FILE__, __LINE__);
    deleter_t d(DW_DLA_STRING, dbg, n);

    if (ret == DW_DLV_OK) {
	name = n;
	return true;
    }

    return false;
}

bool
sym_dwarf_t::get_filename(Dwarf_Die cu_die, string &filename) const
{
    int ret;
    error_t err(dbg);

    filename = "";

    // Get the compile directory.
    Dwarf_Attribute comp_dir_attr = 0;
    ret = dwarf_attr(cu_die, DW_AT_comp_dir, &comp_dir_attr, err);
    err.check(ret, __FILE__, __LINE__);

    if (ret == DW_DLV_OK) {
	deleter_t d(DW_DLA_ATTR, dbg, comp_dir_attr);
	char *comp_dir = 0;
	ret = dwarf_formstring(comp_dir_attr, &comp_dir, err);
	err.check(ret, __FILE__, __LINE__);
	if (ret == DW_DLV_OK) {
	    deleter_t d2(DW_DLA_STRING, dbg, comp_dir);
	    filename += comp_dir;
	    filename += "/";
	}
    }

    // Append the file name.
    string name;
    if (!get_name(cu_die, name))
	return false;

    filename += name;

    return true;
}

bool
sym_dwarf_t::get_specification(Dwarf_Die die, Dwarf_Die &ret_die) const
{
    Dwarf_Attribute attr = 0;
    error_t err(dbg);
    int ret = dwarf_attr(die, DW_AT_specification, &attr, err);
    err.check(ret, __FILE__, __LINE__);
    deleter_t d(DW_DLA_ATTR, dbg, attr);

    if (ret != DW_DLV_OK)
	return false;

    Dwarf_Off offset;
    ret = dwarf_global_formref(attr, &offset, err);
    err.check(ret, __FILE__, __LINE__);

    if (ret != DW_DLV_OK)
	return false;

    // Find the actual compilation unit die
    ret = dwarf_offdie(*dbg, offset, &ret_die, err);
    err.check(ret, __FILE__, __LINE__);

    if (ret != DW_DLV_OK)
	return false;

    return true;
}

bool
sym_dwarf_t::get_abstract_origin(Dwarf_Die die, Dwarf_Die &ret_die) const
{
    Dwarf_Attribute attr = 0;
    error_t err(dbg);
    int ret = dwarf_attr(die, DW_AT_abstract_origin, &attr, err);
    err.check(ret, __FILE__, __LINE__);
    deleter_t d(DW_DLA_ATTR, dbg, attr);

    if (ret != DW_DLV_OK)
	return false;

    Dwarf_Off offset = 0;
    ret = dwarf_global_formref(attr, &offset, err);
    err.check(ret, __FILE__, __LINE__);

    if (ret != DW_DLV_OK)
	return false;

    // Find the actual compilation unit die
    ret = dwarf_offdie(*dbg, offset, &ret_die, err);
    err.check(ret, __FILE__, __LINE__);

    if (ret != DW_DLV_OK)
	return false;

    return true;
}

bool
sym_dwarf_t::get_address_range(Dwarf_Die die,
			       address_range_t &address_range) const
{
    int ret;
    error_t err(dbg);
    Dwarf_Addr low_pc = 0;
    Dwarf_Addr high_pc = 0;

    ret = dwarf_lowpc(die, &low_pc, err);
    err.check(ret, __FILE__, __LINE__);
    if (ret != DW_DLV_OK)
	return false;
    ret = dwarf_highpc(die, &high_pc, err);
    err.check(ret, __FILE__, __LINE__);
    if (ret != DW_DLV_OK)
	return false;

    // According to dwarf3 section 2.17.2, the high_pc is the address of the first
    // location past the end of the range associated with the entity. Thus, they must never
    // be equal.
    if (low_pc == high_pc)
	high_pc++;

    address_range.start = sym_maddr_t(low_pc);
    address_range.end = sym_maddr_t(high_pc);

    return true;
}

bool
sym_dwarf_t::get_srclines(Dwarf_Die cu_die,
			  dwarf_line_ptr_t &lines_ptr,
			  Dwarf_Signed &linecount) const
{
    Dwarf_Line *lines;
    error_t err(dbg);
    int ret;

    ret = dwarf_srclines(cu_die, &lines, &linecount, err);
    err.check(ret, __FILE__, __LINE__);

    if (ret != DW_DLV_OK)
	return false;

    lines_ptr.reset(lines, dwarf_srclines_deleter_t(dbg, linecount));
    return true;
}

bool
sym_dwarf_t::get_linesrc(Dwarf_Line line, string &file) const
{
    error_t err(dbg);
    char *fn = 0;
    int ret = dwarf_linesrc(line, &fn, err);
    err.check(ret, __FILE__, __LINE__);
    if (ret == DW_DLV_NO_ENTRY)
	return false;
    deleter_t d(DW_DLA_STRING, dbg, fn);
    file = fn;
    return true;
}

END_NAMESPACE;
END_NAMESPACE;
