/// @file library/sym/sym_dwarf.hh

// Copyright (c) 2007-2008, 2010 Acumem AB.  All rights reserved.

#ifndef SYM_DWARF_HH
#define SYM_DWARF_HH

#include "libdwarf_object.hh"
#include "function_cache.hh"
#include "statement_cache.hh"

#include <libdwarf/libdwarf.h>
#include <boost/shared_array.hpp>
#include <list>


BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

using boost::shared_ptr;
using boost::shared_array;

// ----------------------------------------------------------------------

/// Wrapper around Dwarf_Debug, allowing custom initialization and
/// destruction of this pointer by means of subclassing.
class dwarf_debug_t {
public:
    dwarf_debug_t() {
	dbg = 0;
    }

    virtual ~dwarf_debug_t() {}

    operator Dwarf_Debug() {
	return dbg;
    }

protected:
    Dwarf_Debug dbg;
};

typedef boost::shared_ptr<dwarf_debug_t> dwarf_debug_ptr_t;

// ----------------------------------------------------------------------

/// Wrapper of Dwarf_Debug based on initialization based on Elf, using
/// libdwarf's specific elf init functions.
class dwarf_debug_elf_t : public dwarf_debug_t {
public:
    dwarf_debug_elf_t(elf_ptr_t elf_file, elf_ptr_t elf_part, int features);
    ~dwarf_debug_elf_t();
private:
    // The Dwarf_Debug will implicitly use elf_file and elf_part, so we must
    // keep references to them until we are sure that the Dwarf_Debug will not
    // be used anymore.
    elf_ptr_t elf_file;
    elf_ptr_t elf_part;
};

// ----------------------------------------------------------------------

/// Wrapper of Dwarf_Debug based on libdwarf's object oriented
/// interface. This is used for PE-COFF.
class dwarf_debug_object_t : public dwarf_debug_t {
public:
    dwarf_debug_object_t(libdwarf_object_ptr_t obj);
    ~dwarf_debug_object_t();

private:
    Dwarf_Obj_Access_Interface_s interface;
    libdwarf_object_ptr_t obj;
};


// ----------------------------------------------------------------------
// wrappers of various Dwarf structures, to help maintaining proper
// ownership and lifetime control.

typedef shared_array<Dwarf_Line> dwarf_line_ptr_t;
typedef pair<dwarf_line_ptr_t, Dwarf_Signed> srclines_result_t;

// ----------------------------------------------------------------------

/// Symbol extraction for DWARF-based debug sections.
class sym_dwarf_t : public sym_base_t {
public:
    sym_dwarf_t(sym_t &sym,
                sym_paddr_t base,
                sym_paddr_t end,
		dwarf_debug_ptr_t dbg,
		sym_mpoffs_t load_offset,
		int features);

    ~sym_dwarf_t();

    void lookup_sym_info(sym_paddr_t address, sym_info_ptr_t &sym_info);

private:
    /// Process the .debug_pubnames section and cache the addresses and
    /// names of all those functions.
    bool register_globals();

    /// Get the aranges and store them.
    bool register_aranges();

    bool get_srclines(Dwarf_Die cu_die,
		      dwarf_line_ptr_t &lines_ptr,
		      Dwarf_Signed &linecount) const;
    bool get_linesrc(Dwarf_Line line, string &file) const;

    bool register_line_numbers_this_cu();
    bool find_arange(sym_maddr_t addr,
		     Dwarf_Arange &ret_arange) const;
    bool find_cu_die(sym_maddr_t addr,
		     Dwarf_Die &ret_die) const;
    struct address_range_t {
        address_range_t()
            : start(0),
              end(0)
            {}

        address_range_t(sym_maddr_t start,
                        sym_maddr_t end)
            : start(start),
              end(end)
            {}

	bool within(sym_maddr_t address) const {
            return address >= start && address < end;
        }

        sym_maddr_t start;
        sym_maddr_t end;
    };

    bool get_address_range(Dwarf_Die die,
			   address_range_t &address_range) const;
    bool scan_functions(sym_paddr_t addr,
			Dwarf_Die cu_die,
			function_ptr_t &ret_function);
    bool get_name(Dwarf_Die die, string &name) const;
    bool get_filename(Dwarf_Die cu_die, string &filename) const;
    bool get_specification(Dwarf_Die die, Dwarf_Die &ret_die) const;
    bool get_abstract_origin(Dwarf_Die die, Dwarf_Die &ret_die) const;
    bool get_subprogram_name_recursive(Dwarf_Die die,
				       string &name) const;

    // Aranges. All contained addresses are unrelocated addresses,
    // corresponding to sym_maddr_t.
    struct arange_t {
	Dwarf_Signed count;
	Dwarf_Arange *arange;
    };

    typedef vector<arange_t> arange_vector_t;
    arange_vector_t arange_vector;


    // Map a Compilation unit die with it's incremental processing queue.
    typedef shared_ptr<Dwarf_Die_s> dwarf_die_ptr_t;
    typedef list<dwarf_die_ptr_t> dwarf_die_list_t;
    typedef map<Dwarf_Off, dwarf_die_list_t> cu_dwarf_die_map_t;
    mutable cu_dwarf_die_map_t cu_dwarf_die_map;


    // The dwarf context
    dwarf_debug_ptr_t dbg;

    // Invoked to delete the contents of the dwarf_debug_ptr_t
    static void dwarf_debug_deleter(Dwarf_Debug d) throw();

    // Prematurely Kill the dwarf context
    // XXX also release the elf_file_part, in case we start caching it.
    void dwarf_finish();

    /// Load offset.
    sym_mpoffs_t load_offset;

    // Keep the dwarf loaded and perform a lazy recursive function
    // search through if a function cannot be found in the
    // prepopulated globals table.

    // sym_t::feature_t bit mask
    int features;

    function_cache_t function_cache;
    statement_cache_t statement_cache;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
