/// @file library/sym/sym_elf.cc

// Copyright (c) 2010 Acumem AB.  All rights reserved.

#include "sym_elf.hh"

#include <libelf.h>
#include <gelf.h>

#include <boost/foreach.hpp>

#include <iostream>

#include <fcntl.h>
#include <stdio.h>

using namespace std;
using namespace boost;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

void
sym_elf_t::lookup_sym_info(sym_paddr_t address,
                           sym_info_ptr_t &sym_info)
{
    function_ptr_t function;
    if (function_cache.lookup_function(address, function)) {
        sym_info->symbol_name = function->name;
        sym_info->symbol_offset = address.paddr - function->range.base.paddr;
    }
}

static bool
process_syms(const char *sym_section_name,
             const char *str_section_name,
             Elf *elf,
             function_cache_t *function_cache,
             sym_mpoffs_t load_offset)
{
    GElf_Ehdr ehdr;
    if (gelf_getehdr(elf, &ehdr) == NULL)
        return false;

    size_t shstrndx;
    if (elf_getshstrndx(elf, &shstrndx) == 0)
        return false;

    Elf_Scn *sym_scn = NULL;
    GElf_Half str_idx = (GElf_Half)-1;

    for (GElf_Half i = 0; i < ehdr.e_shnum; i++) {
        Elf_Scn *scn = elf_getscn(elf, i);
        if (scn == NULL)
            continue;

        GElf_Shdr shdr;
        if (gelf_getshdr(scn, &shdr) == NULL ||
            shdr.sh_type == SHT_NULL)
            continue;

        const char *name = elf_strptr(elf, shstrndx, shdr.sh_name);
        if (name != NULL) {
            if (strcmp(name, sym_section_name) == 0)
                sym_scn = scn;
            if (strcmp(name, str_section_name) == 0)
                str_idx = i;
        }
    }

    if (sym_scn == NULL || str_idx == (GElf_Half)-1)
        return false;

    GElf_Shdr sym_shdr;
    if (gelf_getshdr(sym_scn, &sym_shdr) == NULL)
        return false;

    Elf_Data *sym_data = elf_getdata(sym_scn, NULL);
    if (sym_data == NULL)
        return false;

    bool found_symbols = false;

    if (sym_shdr.sh_size > 0 && sym_shdr.sh_entsize != 0) {
	GElf_Xword sym_count = sym_shdr.sh_size / sym_shdr.sh_entsize;

	for (GElf_Xword i = 0; i < sym_count; ++i) {
	    GElf_Sym symbol;
	    if (gelf_getsym(sym_data, (int)i, &symbol) == NULL)
		continue;

	    if (GELF_ST_TYPE(symbol.st_info) == STT_FUNC &&
		symbol.st_shndx != SHN_UNDEF) {
		sym_paddr_t base = sym_maddr_t(symbol.st_value) + load_offset;
		sym_paddr_t end = sym_paddr_t(base.paddr + symbol.st_size);
		string name(elf_strptr(elf, str_idx, symbol.st_name));
		function_cache->add_function(base, end, name);
		found_symbols = true;
	    }
	}
    }

    return found_symbols;
}

sym_elf_t::sym_elf_t(sym_t &sym,
                     sym_paddr_t base,
                     sym_paddr_t end,
                     elf_ptr_t elf_part,
                     sym_mpoffs_t load_offset)
    : sym_base_t(sym, base, end),
      function_cache(sym)
{
    bool found_syms = process_syms(".symtab", ".strtab", &*elf_part,
                                   &function_cache, load_offset);
    bool found_dynsyms = process_syms(".dynsym", ".dynstr", &*elf_part,
                                      &function_cache, load_offset);
    if (!found_syms && !found_dynsyms)
        cannot_cope();
}

END_NAMESPACE;
END_NAMESPACE;
