/// @file library/sym/sym_stabs.cc

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#define __STDC_LIMIT_MACROS

#include "sym_stabs.hh"

#include "library/lang/common_types.h"

#include <boost/foreach.hpp>

#include <libelf.h>
#include <gelf.h>
#include "stab.h"

#include <iostream>

#include <fcntl.h>
#include <stdio.h>

using namespace std;
using namespace boost;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

#ifdef __linux__
struct stab {
    unsigned n_strx; /* file String table index */
    unsigned char n_type; /* Stab type */
    char n_other; /* used by N_SLINE stab */
    short n_desc; /* Desc value */
    unsigned n_value; /* Offset or value */
};
#endif

typedef map<string, sym_maddr_t> func_address_map_t;

void
sym_stabs_t::lookup_sym_info(sym_paddr_t address,
                             sym_info_ptr_t &sym_info)
{
    // Look for file and line in the cache.
    statement_ptr_t statement;
    if (statement_cache.lookup_statement(address, statement)) {
        if (statement->filename)
            sym_info->source_file = statement->filename;
        if (statement->row != -1)
            sym_info->source_line = statement->row;
    }

    // Look for function in the cache.
    function_ptr_t function;
    if (function_cache.lookup_function(address, function)) {
        sym_info->function_name = function->name;
        sym_info->function_offset = address.paddr - function->range.base.paddr;
    }
}

static bool
handle_stab_section(function_cache_t &function_cache,
                    statement_cache_t &statement_cache,
		    Elf *elf,
		    Elf_Scn *scn,
		    Elf_Scn *stabstrscn,
		    sym_mpoffs_t load_offset,
		    func_address_map_t &func_address_map)
{
    Elf_Data *strdata = elf_getdata(stabstrscn, 0);
    if (strdata == 0)
	return false;

    char *str = (char *) strdata->d_buf;
    if (str == 0)
	return false;

    Elf_Data *stab_data = elf_getdata(scn, 0);
    if (stab_data == 0)
	return false;

    stab *s = (stab *) stab_data->d_buf;
    if (s == 0)
	return false;

    int num = (int)(stab_data->d_size / sizeof(stab));
    int current_strtable_offset = 0;
    int next_strtable_offset = 0;
    int current_file_ndx = 0;
    int current_function_ndx = 0;
    uint64_t current_offset = 0;

    // Work-around for GCC bug where no N_FUN stab is generated for some
    // functions (inlined functions?), but N_SLINE stabs are still
    // generated. This caused us to add the line references for such functions
    // to the function with the previous N_FUN.
    //
    // To avoid this we keep track of the last offset seen in an N_SLINE stab,
    // and since they are required to be strictly increasing we can conclude
    // that we have reached a new function when the offsets over from zero.
    uint64_t min_sline_value = UINT64_MAX;

    for (; num > 0; num--, s++) {
	switch (s->n_type) {
	case N_UNDF:
	    // N_UNDF tells how long the current string table range
	    // is. When advancing to the next file, adjust the offset
	    current_strtable_offset += next_strtable_offset;
	    next_strtable_offset = s->n_value;

	case N_SO:
	case N_SOL:
	    current_file_ndx = s->n_strx + current_strtable_offset;
	    break;

	case N_FUN: {
	    // XXX Handle N_OUTL for outlined functions, whatever that is?
	    current_function_ndx = s->n_strx + current_strtable_offset;
	    current_offset = s->n_value;
	    const char *colon = strchr(&str[current_function_ndx], ':');
	    if (colon == 0)
		colon = &str[current_function_ndx] +
		    strlen(&str[current_function_ndx]);
	    string funcname((const char *)&str[current_function_ndx], colon);

	    if (current_offset != 0) {
		// Only if we have some meaningful data here, see if
		// this function shall be added to the function map.

		sym_maddr_t &offset = func_address_map[funcname];
		if (offset == sym_maddr_t(0))
		    offset = sym_maddr_t(current_offset);
	    }
            else {
		// If the offset wasn't found in the N_FUN, get it from
		// the symbol table. Gcc puts the address in N_FUN, but
		// Sun Studio doesn't.
		if (current_offset == 0)
		    current_offset = func_address_map[funcname].maddr;
	    }
            min_sline_value = 0;
	    break;
	}

	case N_SLINE:
            if (s->n_value >= min_sline_value) {
                string filename(&str[current_file_ndx]);
                sym_paddr_t pc = sym_maddr_t(current_offset + s->n_value) + load_offset;
                statement_cache.add_statement(pc, pc, filename, s->n_desc, -1);
                min_sline_value = s->n_value + 1;
            }
            else
                min_sline_value = UINT64_MAX;
            break;
	}
    }

    return true;
}

// Find next section with a specific name, or 0 if none is found
static Elf_Scn *
get_section_by_name(Elf *elf, const char *section_name,
		    Elf_Scn *start_search = 0)
{
    size_t shstrndx;
    elf_getshstrndx(elf, &shstrndx);

    Elf_Scn *scn = start_search;
    while((scn = elf_nextscn(elf, scn)) != 0) {
	GElf_Shdr shdr;
	if (0 == gelf_getshdr(scn, &shdr))
	    continue;
	const char *name = elf_strptr(elf, shstrndx, (size_t)shdr.sh_name);
	if (strcmp(name, section_name) == 0)
	    return scn;
    }
    return 0;
}

// Find next section with a specific type, or 0 if none is found
static Elf_Scn *
get_section_by_type(Elf *elf, size_t type, Elf_Scn *start_search = 0)
{
    Elf_Scn *scn = start_search;
    while((scn = elf_nextscn(elf, scn)) != 0) {
	GElf_Shdr shdr;
	if (0 == gelf_getshdr(scn, &shdr))
	    continue;
	if (type == shdr.sh_type)
	    return scn;
    }
    return 0;
}

// dig out all function symbols from this section
static void
process_sym_section(Elf *elf, Elf_Scn *symbol_section,
		    func_address_map_t &func_address_map)
{
    GElf_Shdr shdr;
    gelf_getshdr(symbol_section, &shdr);

    Elf_Data *data = elf_getdata(symbol_section, NULL);
    int count = (int)(shdr.sh_size / shdr.sh_entsize);

    for (int ii = 0; ii < count; ++ii) {
        GElf_Sym symbol;
        gelf_getsym(data, ii, &symbol);

	if (GELF_ST_TYPE(symbol.st_info) == STT_FUNC &&
            symbol.st_shndx != SHN_UNDEF) {
	    func_address_map[elf_strptr(elf, shdr.sh_link, symbol.st_name)] =
		sym_maddr_t(symbol.st_value);
	}
    }
}

// Process an elf part and add all line info
static int
process_elf_file(function_cache_t &function_cache,
                 statement_cache_t &statement_cache,
                 Elf *elf,
                 sym_mpoffs_t load_offset)
{
    Elf_Scn *stabs = get_section_by_name(elf, ".stab");
    Elf_Scn *stabstr = get_section_by_name(elf, ".stabstr");
    if (stabs == 0 || stabstr == 0)
	return 0;

    // Get all function symbols and their respective start.
    func_address_map_t func_address_map;
    Elf_Scn *sym_section = 0;
    while ((sym_section = get_section_by_type(elf, SHT_SYMTAB,
					      sym_section)) != 0) {
	process_sym_section(elf, sym_section, func_address_map);
    }

    bool ret = handle_stab_section(function_cache, statement_cache, elf, stabs, stabstr,
				   load_offset, func_address_map);

    // Create functions mappings for all discovered functions
    BOOST_FOREACH(func_address_map_t::value_type &v, func_address_map)
	function_cache.add_function(v.second + load_offset,
                               v.second + load_offset,
                               v.first);

    return ret;
}

// ----------------------------------------------------------------------

sym_stabs_t::sym_stabs_t(sym_t &sym,
                         sym_paddr_t base,
                         sym_paddr_t end,
			 elf_ptr_t elf_part,
			 sym_mpoffs_t load_offset,
			 int features)
    : sym_base_t(sym, base, end),
      function_cache(sym),
      statement_cache(sym)
{
    if (!process_elf_file(function_cache, statement_cache, elf_part.get(), load_offset))
	cannot_cope();
}

END_NAMESPACE;
END_NAMESPACE;
