/// @file library/warning/warning_consumer.cc

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#include "warning_consumer.hh"

#include <boost/foreach.hpp>

#include <iostream>
#if !defined(_WIN32)
#include <syslog.h>
#endif /* !defined(_WIN32) */

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;
using namespace boost;

// ----------------------------------------------------------------------
// warning_consumer_base_t
// ----------------------------------------------------------------------
warning_consumer_base_t::warning_consumer_base_t()
{
}

warning_consumer_base_t::~warning_consumer_base_t()
{
}

void
warning_consumer_base_t::warn(const warning_base_t *warning)
{
    warning_const_ptr_t wp(warning);
    warn(wp);
}

void
warning_consumer_base_t::none(const string &message)
{
    warn(warning_const_ptr_t(new generic_warning_t(
                                 message, warning_base_t::Severity_None)));
}

void
warning_consumer_base_t::info(const string &message)
{
    warn(warning_const_ptr_t(new generic_warning_t(
                                 message, warning_base_t::Severity_Info)));
}

void
warning_consumer_base_t::warning(const string &message)
{
    warn(warning_const_ptr_t(new generic_warning_t(
                                 message, warning_base_t::Severity_Warning)));
}

void
warning_consumer_base_t::error(const string &message)
{
    warn(warning_const_ptr_t(new generic_warning_t(
                                 message, warning_base_t::Severity_Error)));
}

// ----------------------------------------------------------------------
// warning_consumer_null_t
// ----------------------------------------------------------------------
warning_consumer_null_t::warning_consumer_null_t()
{
}

warning_consumer_null_t::~warning_consumer_null_t()
{
}

void
warning_consumer_null_t::warn(const warning_const_ptr_t &warning)
{
}

// ----------------------------------------------------------------------
// warning_consumer_stderr_t
// ----------------------------------------------------------------------
warning_consumer_stderr_t::warning_consumer_stderr_t()
{
}

warning_consumer_stderr_t::~warning_consumer_stderr_t()
{
}

void
warning_consumer_stderr_t::warn(const warning_const_ptr_t &warning)
{
    switch (warning->severity()) {
    case warning_base_t::Severity_Info:
	cerr << "Info: ";
	break;
    case warning_base_t::Severity_Warning:
	cerr << "Warning: ";
	break;
    case warning_base_t::Severity_Error:
	cerr << "Error: ";
	break;
    case warning_base_t::Severity_None:
	break;
    default:
        assert(0);
    }

    cerr << warning->what() << endl;
}

// ----- warning_consumer_string_t --------------------------------------------

warning_consumer_string_t::warning_consumer_string_t()
{
}

warning_consumer_string_t::~warning_consumer_string_t()
{
}

string
warning_consumer_string_t::get_messages()
{
    string m = messages;
    messages.clear();
    return m;
}

void
warning_consumer_string_t::warn(const warning_const_ptr_t &warning)
{
    string prefix;
    switch (warning->severity()) {
    case warning_base_t::Severity_Info:
	prefix = "Info: ";
	break;
    case warning_base_t::Severity_Warning:
	prefix = "Warning: ";
	break;
    case warning_base_t::Severity_Error:
	prefix = "Error: ";
	break;
    case warning_base_t::Severity_None:
	break;
    default:
        assert(0);
    }

    messages += prefix + warning->what() + "\n";
}

// ----- warning_consumer_callback_t ------------------------------------------

warning_consumer_callback_t::warning_consumer_callback_t(
    output_callback_t callback)
    : callback(callback)
{
}

warning_consumer_callback_t::~warning_consumer_callback_t()
{
}

void
warning_consumer_callback_t::warn(const warning_const_ptr_t &warning)
{
    string prefix;
    switch (warning->severity()) {
    case warning_base_t::Severity_Info:
	prefix = "Info: ";
	break;
    case warning_base_t::Severity_Warning:
	prefix = "Warning: ";
	break;
    case warning_base_t::Severity_Error:
	prefix = "Error: ";
	break;
    case warning_base_t::Severity_None:
	break;
    default:
        assert(0);
    }

    callback((prefix + warning->what()).c_str());
}

// ----- warning_consumer_syslog_t --------------------------------------------

#if !defined(_WIN32)

warning_consumer_syslog_t::warning_consumer_syslog_t()
{
    openlog("threadspotter", LOG_PID, LOG_DAEMON);
}

warning_consumer_syslog_t::~warning_consumer_syslog_t()
{
}

void
warning_consumer_syslog_t::warn(const warning_const_ptr_t &warning)
{
    int prio = LOG_INFO;
    switch (warning->severity()) {
    case warning_base_t::Severity_Info:
        prio = LOG_INFO;
	break;
    case warning_base_t::Severity_Warning:
        prio = LOG_WARNING;
	break;
    case warning_base_t::Severity_Error:
        prio = LOG_ERR;
	break;
    case warning_base_t::Severity_None:
        prio = LOG_INFO;
	break;
    default:
        assert(0);
    }

    syslog(prio, warning->what().c_str());
}

#endif /* !defined(_WIN32) */

// ----- warning_consumer_splitter_t ------------------------------------------

void
warning_consumer_splitter_t::add_warning_consumer(
    const warning_consumer_ptr_t &warning_consumer)
{
    warning_consumers.insert(warning_consumer);
}

void
warning_consumer_splitter_t::remove_warning_consumer(
    const warning_consumer_ptr_t &warning_consumer)
{
    warning_consumers.erase(warning_consumer);
}

void
warning_consumer_splitter_t::warn(const warning_const_ptr_t &warning)
{
    BOOST_FOREACH(const warning_consumer_ptr_t &wc, warning_consumers)
        wc->warn(warning);
}

// ----- warning_consumer_locked_t --------------------------------------------

warning_consumer_locked_t::warning_consumer_locked_t(
    warning_consumer_ptr_t wc)
    : wc(wc)
{
}

warning_consumer_locked_t::~warning_consumer_locked_t()
{
}

void
warning_consumer_locked_t::warn(const warning_const_ptr_t &warning)
{
    lock_guard<mutex> l(m);
    wc->warn(warning);
}

END_NAMESPACE;
END_NAMESPACE;
