/// @file library/warning/warning_consumer.hh

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#ifndef WARNING_CONSUMER_HH
#define WARNING_CONSUMER_HH

#if defined(_WIN32)
// Needed to avoid the warning "code compiled as native, Found an intrisic not
// supported in managed code" when compiling boost::thread.
#pragma warning(disable: 4793)
#endif

#include "warning.hh"

#include <boost/shared_ptr.hpp>
#include <boost/thread.hpp>

#include <set>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

/**
 * Virtual base class for all warning consumers. Warning consumers
 * should inherit from this class and override the
 * warn(warning_const_ptr_t) function.
 */
class warning_consumer_base_t
{
public:
    warning_consumer_base_t();
    virtual ~warning_consumer_base_t();

    /**
     * Issue a warning. The callee is responsible for deleting the
     * class.
     */
    void warn(const warning_base_t *warning);

    // Issue a message.

    void none(const std::string &message);
    void info(const std::string &message);
    void warning(const std::string &message);
    void error(const std::string &message);

    /// Internal function, should not be called directly.
    virtual void warn(const warning_const_ptr_t &warning) = 0;
};

typedef boost::shared_ptr<warning_consumer_base_t> warning_consumer_ptr_t;

/**
 * This is the /dev/null of warning consumers. Eats warnings for
 * lunch.
 */
class warning_consumer_null_t
    : public warning_consumer_base_t
{
public:
    warning_consumer_null_t();
    ~warning_consumer_null_t();

protected:
    void warn(const warning_const_ptr_t &warning);
};

/**
 * Formats and displays warnings on standard error.
 */
class warning_consumer_stderr_t
    : public warning_consumer_base_t
{
public:
    warning_consumer_stderr_t();
    ~warning_consumer_stderr_t();

protected:
    void warn(const warning_const_ptr_t &warning);
};

/// Collects warning messages in a string.
class warning_consumer_string_t
    : public warning_consumer_base_t
{
public:
    warning_consumer_string_t();
    ~warning_consumer_string_t();

    string get_messages();

protected:
    void warn(const warning_const_ptr_t &warning);

private:
    string messages;
};

typedef boost::shared_ptr<warning_consumer_string_t> warning_consumer_string_ptr_t;

/// Calls callback with messages.
class warning_consumer_callback_t
    : public warning_consumer_base_t
{
public:
    typedef void (*output_callback_t)(const char *);

    warning_consumer_callback_t(output_callback_t callback);
    ~warning_consumer_callback_t();

protected:
    void warn(const warning_const_ptr_t &warning);

private:
    output_callback_t callback;
};

#if !defined(_WIN32)

/// Logs messages to syslog.
class warning_consumer_syslog_t
    : public warning_consumer_base_t
{
public:
    warning_consumer_syslog_t();
    ~warning_consumer_syslog_t();

protected:
    void warn(const warning_const_ptr_t &warning);
};

#endif /* !defined(_WIN32) */

/// Forwards warnings to other warning consumers.
class warning_consumer_splitter_t
    : public warning_consumer_base_t
{
public:
    void add_warning_consumer(
        const warning_consumer_ptr_t &warning_consumer);
    void remove_warning_consumer(
        const warning_consumer_ptr_t &warning_consumer);

protected:
    void warn(const warning_const_ptr_t &warning);

private:
    std::set<warning_consumer_ptr_t> warning_consumers;
};

/// Wraps around another warning consumer to make it thread safe.
class warning_consumer_locked_t
    : public warning_consumer_base_t
{
public:
    warning_consumer_locked_t(warning_consumer_ptr_t wc);
    ~warning_consumer_locked_t();

protected:
    void warn(const warning_const_ptr_t &warning);

private:
    const warning_consumer_ptr_t wc;
    boost::mutex m;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
