/// @file library/xml_utils/xml_utils.cc

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#include <libxml/parser.h>

#include <boost/shared_array.hpp>

#include "xml_utils.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

struct xml_free_functor_t {
    void operator()(void *addr) {
	if (addr)
	    xmlFree(addr);
    }
};

struct xml_free_doc_functor_t {
    void operator()(xmlDoc *addr) {
	if (addr)
	    xmlFreeDoc(addr);
    }
};

// --- xml_exception_t --------------------------------------------------

xml_exception_t::xml_exception_t(const xmlNodePtr node, const string &reason)
    : std::runtime_error(reason)
{
    boost::shared_array<xmlChar> p(xmlGetNodePath(node), xml_free_functor_t());
    _path = (const char *)p.get();
}

xml_exception_t::xml_exception_t(const string &path, const string &reason)
    : std::runtime_error(reason), _path(path)
{
}

xml_exception_t::~xml_exception_t() throw()
{
}

string
xml_exception_t::path() const throw()
{
    return _path;
}

string
xml_exception_t::why() const throw()
{
    return what();
}


// --- xml_utils_t ------------------------------------------------------

xml_doc_ptr_t
xml_utils_t::load_xml_file(const std::string &file)
{
    boost::shared_ptr<xmlDoc> doc(xmlParseFile(file.c_str()),
				  xml_free_doc_functor_t());
    if (!doc)
	throw xml_exception_t("/", "Couldn't open file.");
    else
	return doc;
}

xml_doc_ptr_t
xml_utils_t::load_xml_mem(const void *buffer, size_t size)
{
    boost::shared_ptr<xmlDoc> doc(xmlParseMemory((const char *)buffer, (int)size),
				  xml_free_doc_functor_t());
    if (!doc)
	throw xml_exception_t("/", "Couldn't open file.");
    else
	return doc;
}

xmlNodePtr
xml_utils_t::get_root_element(xml_doc_ptr_t doc)
{
    xmlNodePtr root(xmlDocGetRootElement(doc.get()));

    if (!root)
	throw xml_exception_t("/", "Failed to get root element.");
    else
	return root;
}

bool
xml_utils_t::get_attr(const xmlNodePtr node, const string &name,
		      string &var, const string def)
{
    boost::shared_array<xmlChar> v(xmlGetProp(node, (const xmlChar *)name.c_str()),
				   xml_free_functor_t());

    if (!v) {
	var = def;
	return false;
    } else {
	var = (const char *)v.get();
	return true;
    }
}

bool
xml_utils_t::get_attr(const xmlNodePtr node, const string &name,
		      bool &var, const bool def)
{
    string value;

    if (!get_attr(node, name, value)) {
	var = def;
	return false;
    } else {
	if (value == "true") {
	    var = true;
	    return true;
	} else if (value == "false") {
	    var = false;
	    return true;
	} else {
	    var = def;
	    return false;
	}
    }
}

string
xml_utils_t::get_text(const xmlNodePtr node)
{
    string text;
    append_text_nodes(node, text);
    return text;
}

string
xml_utils_t::get_name(const xmlNodePtr node)
{
    return string((const char *)node->name);
}

bool
xml_utils_t::append_text_nodes(const xmlNodePtr node, string &var)
{
    bool ret = false;

    for(xmlNodePtr n = node->children; n; n = n->next) {
	if(xmlNodeIsText(n) && n->content) {
	    var.append((const char*)n->content);
	    ret = true;
	}
    }

    return ret;
}

END_NAMESPACE;
END_NAMESPACE;
