/// @file library/xml_utils/xml_utils.hh

// Copyright (c) 2008 Acumem AB.  All rights reserved.

#ifndef XML_UTILS_HH
#define XML_UTILS_HH

#include <string>
#include <stdexcept>

#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>

#include <libxml/tree.h>

#include "library/lang/common_macros.h"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(library);

using namespace std;

using boost::shared_ptr;

typedef shared_ptr<xmlDoc> xml_doc_ptr_t;

class xml_exception_t
    : public std::runtime_error
{
public:
    xml_exception_t(const xmlNodePtr node, const string &reason);
    xml_exception_t(const string &path, const string &reason);
    virtual ~xml_exception_t() throw();

    virtual string path() const throw();
    virtual string why() const throw();

protected:
    string _path;
    string _why;
};

class xml_utils_t
{
public:
    static xml_doc_ptr_t load_xml_mem(const void *buffer, size_t size);
    static xml_doc_ptr_t load_xml_file(const string &file);

    static xmlNodePtr get_root_element(xml_doc_ptr_t doc);

    //@{
    /**
     * Retrieve an attribute for an element, or a default value of the
     * attribute can't be found.
     *
     * \param node Element containing the attribute
     * \param name Name of the attribute
     * \param var Reference to the destination variable.
     * \param def Default value if the attribute can't be found.
     * \return true if the attribute was found, false otherwise.
     */
    static bool get_attr(const xmlNodePtr node, const string &name,
			 string &var, const string def = "");
    static bool get_attr(const xmlNodePtr node, const string &name,
			 bool &var, const bool def = false);
    template <typename IntType>
    static bool get_attr(const xmlNodePtr node, const string &name,
			 IntType &var, const IntType def = 0) {
	string value;

	if (!get_attr(node, name, value)) {
	    var = def;
	    return false;
	} else {
	    try {
		var = boost::lexical_cast<IntType>(value);
		return true;
	    } catch (boost::bad_lexical_cast) {
		var = def;
		return false;
	    }
	}
    }
    //@}

    template<typename AttrType>
    static void get_required_attr(const xmlNodePtr node, const string &name,
				  AttrType &var) {
	if (!get_attr(node, name, var))
	    throw xml_exception_t(node,
				  "Failed to load attribute '" + name + "'.");
    }

    static string get_name(const xmlNodePtr node);

    static string get_text(const xmlNodePtr node);

protected:
    /**
     * Append all text() elements of a node to a text string.
     *
     * \param node Pointer to the element containing the text.
     * \param var Destination string.
     * \return True if node contains text() elements.
     */
    static bool append_text_nodes(const xmlNodePtr node, string &var);
};

END_NAMESPACE;
END_NAMESPACE;

#endif
