/** @file sampler/src/atf.h */

/* Copyright (c) 2006-2011 Rogue Wave Software, Inc.  All rights reserved. */

/*
 * Author: Mikael Pettersson.
 */
#ifndef ATF_H
#define ATF_H

#if defined(_WIN32) && !defined(__GNUC__)
#include <stddef.h>	/* uintptr_t */
#else
#include <stdint.h>	/* uintptr_t */
#endif

/*
 * ATF Revision 6.0 File header:
 *
 *  0: 4-byte magic number: ('A'<<24)|('T'<<16)|('F'<<8)|('6'<<0)
 *  4: 4-byte header size: 32
 *  8: 4-byte flags field:
 *     bits 0-7: architecture code; currently only 0x86 is allowed
 *     bits 8-15: sizeof(uintptr_t); must be 4 or 8
 *     bit 16: 1 [ATF_HEADER_FLAGS_BURST] if this is a burst sampling, 0 if not
 *     bit 17: 1 if memory access counts are compressed [ATF 5.1], 0 if not
 *     bits 18-31: reserved; must be zero
 * 12: 4-byte getpid()
 * 16: 4-byte sampler_period [see sampler.h]
 * 20: 8-byte sampler_log2linesizes [see sampler.h]
 * 28: 4-byte sampler_count
 *     If the file is sampled with increasing sample period this field
 *     contains the target sample count (not 0). If the file is sampled
 *     with constant sample period it contains 0.
 *
 * The data packets start immediately after the header.
 *
 * A data packet consists of a packet header followed by a payload.
 * The layout of the payload depends on the header.
 *
 * All uncompressed numbers are stored in producer-native byte order.
 *
 * Header: 1-byte or 2-byte Packet Header:
 *
 * The packet header is a 10-bit value interpreted like this:
 *
 * bits 1-0 == {00, 01, 10}
 *   Memory access packet
 * bits 1-0 == 11
 *   bits 4-2 == 000
 *     Clflush memory access packet
 *   bits 4-2 == 001
 *     Prefetch memory access packet
 *   bits 4-2 == 010
 *     Shadow stack packet
 *   bits 4-2 == 011
 *     MMAP packet
 *   bits 4-2 == 100
 *     MUNMAP packet
 *   bits 4-2 == 101
 *     MREMAP packet
 *   bits 4-2 == 110
 *     MOVNT packet
 *   bits 4-2 == 111
 *     bits 7-5 == 000
 *       KILLWP packet
 *     bits 7-5 == 001
 *       BURST_END packet
 *     bits 7-5 == 010
 *       THREAD packet
 *     bits 7-5 == 011
 *       GLOBAL_MACNT packet
 *     bits 7-5 == 100
 *       LOCAL_MACNT packet
 *     bits 7-5 == 101
 *       THREAD_EXIT packet
 *
 * Header bits above the opcode bits are opcode specific.
 *
 * The 10-bit header is encoded in 1 or 2 bytes. If the header value is less
 * than 0x1f it is encoded in 1 byte:
 *
 *    bits 4-0: bits 4-0 of the header value
 *    bits 7-5: reserved for compression flags
 *
 * If the header value is greater than or equal to 0x1f it is encoded in 2
 * bytes. First byte:
 *
 *    bits 1-0: bits 1-0 of the header value
 *    bits 4-2: 111
 *    bits 7-5: reserved for compression flags
 *
 * Second byte:
 *
 *    bits 7-0: bits 9-2 of the header value
 *
 * Memory Access Packets
 *
 * Segment 1: Packet Header:
 *    bits 1-0 == {00, 01, 10}: 2-bit access code
 *      bits 5-2: size code
 *      bits 9-6: reserved, must be zero
 *    bits 1-0 == 11: extended access code in bits 4-2
 *      bits 4-2 == {XOP_PREFETCH, XOP_CLFLUSH}
 *        bits 9-5: reserved, must be zero
 *      bits 4-2 == XOP_MOVNT
 *        bits 5: read (1) or write (0)
 *        bits 8-6: size code
 *        bit 9: reserved, must be zero
 *
 *    Compression flags:
 *    bit 5: thread id is implicit (0) or explicit (1)
 *    bit 6: thread id is uncompressed (0) or compressed (1)
 *    bit 7: instruction address is uncompressed (0) or compressed (1)
 *
 * Segment 2: Data Address: sizeof(uintptr_t) bytes.
 *
 * Segment 3: Instruction Address:
 *    If compression flag bit 7 is 0, the segment is sizeof(uintptr_t)
 *    bytes and contains an uncompressed instruction address.
 *
 *    If compression flag bit 7 is 1, the segment is 1 to 3 bytes long,
 *    and contains a compressed difference relative to the previous
 *    packet's instruction address. The instruction address before
 *    the first packet is defined to be zero.
 *
 *    The low-order bits of the first byte encode the length of the segment:
 *    bit 0 = 0: 1 byte containing a 7-bit signed integer
 *    bit 0 = 1, bit 1 = 0: 2 bytes containing a 14-bit signed integer
 *    bit 0 = 1, bit 1 = 1: 3 bytes containing a 22-bit signed integer
 *    A compressed difference is always stored in little-endian byte order.
 *
 * Segment 4: Local Memory Access Counter:
 *    This segment is 1, 2, 3, or 8 bytes long, and contains a compressed
 *    difference relative to the previous packet's local memory access
 *    counter. The local memory access counter before the first packet
 *    is defined to be zero.
 *
 *    The low-order bits of the first byte encode the length of the segment:
 *    bit 0 = 0: 1 byte containing a 7-bit signed integer
 *    bit 0 = 1, bit 1 = 0: 2 bytes containing a 14-bit signed integer
 *    bit 0 = 1, bit 1 = 1: bit 2 = 0: 3 bytes containing a 21-bit signed integer
 *    bit 0 = 1, bit 1 = 1: bit 2 = 1: 8 bytes containing a 61-bit signed integer
 *    A compressed difference is always stored in little-endian byte order.
 *
 * Segment 5: Global Memory Access Counter:
 *    This segment contains a compressed difference relative to the previous
 *    packet's global memory access counter, using the same representation as
 *    in the Local Memory Access Counter segment.
 *
 * Segment 6: Thread ID:
 *    If compression flag bit 5 is 0, this segment is empty, and the thread id
 *    is the same as that of the previous packet. The thread id before the
 *    first packet is defined to be zero.
 *
 *    If compression flag bit 5 is 1, the segment contains a compressed
 *    (bit 6 is 1) or uncompressed (bit 6 is 0) difference relative to the
 *    previous packet's thread id, using the same representation as in the
 *    Instruction Address segment.
 *
 * Shadow Stack Packets
 *
 * Segment 1: Packet Header:
 *    bits 1-0: 2-bit primary opcode == 11
 *    bits 4-2: 3-bit extended opcode == 010
 *    bit 5:    absolute (1) or relative (0) stack
 *    bits 9-5: reserved, must be zero
 *
 *    Compression flags:
 *    bit 5: thread id is implicit (0) or explicit (1)
 *    bit 6: thread id is uncompressed (0) or compressed (1)
 *    bit 7: NumberOfLost is implicit (0) or explicit (1)
 *
 *    Absolute stack packets are processed completely independently of the
 *    relative stack packets. They do not affect the stack maintained between
 *    relative stack packets, nor the "last index" or "clean
 *    index". NumberOfLost and NumberOfPops must be 0, and NumberOfPushes
 *    contains the number of return addresses on the stack.
 *
 * Segment 2: NumberOfPops: 1 byte.
 *
 * Segment 3: NumberOfPushes: 1 byte.
 *
 * Segment 4: NumberOfLost:
 *    If bit 7 of the header is 0, this segment is empty, and the value
 *    is zero. If bit 7 of the header is 1, this segment is 1 byte.
 *
 * Segment 5: Thread ID:
 *    This segment follows the same rules as the Thread ID segment in
 *    memory access packets.
 *
 * Segment 6: Pushes: sizeof(uintptr_t) times NumberOfPushes bytes.
 *
 * Segment 7: Global Memory Access Counter:
 *    This segment follows the sames rules as the Global Memory Access
 *    Counter segment in memory access packets.
 *
 * MMAP Packets:
 *
 * Segment 1: Packet Header:
 *    bits 1-0: 2-bit primary opcode == 11
 *    bits 4-2: 3-bit extended opcode == 011
 *    bits 9-5: reserved, must be zero
 *
 *    Compression flags:
 *    bit 5: offset is zero (0) or non-zero (1)
 *    bit 6: reserved, must be zero
 *    bit 7: reserved, must be zero
 *
 * Segment 2: SubHeader: 1 byte
 *    bits 2-0: 3-bit prot field
 *    bits 4-3: 2-bit flags field
 *    bits 7-5: reserved, must be zero
 *
 * Segment 3: Start: sizeof(uintptr_t) bytes
 *
 * Segment 4: Length: sizeof(uintptr_t) bytes
 *
 * Segment 5: Offset:
 *    If compression flag bit 5 is zero, this segment is empty, and the value
 *    is zero. Otherwise, this segment is 8 bytes long.
 *
 * Segment 6: Global Memory Access Counter:
 *    This segment follows the sames rules as the Global Memory Access
 *    Counter segment in memory access packets.
 *
 * Segment 7: Device:
 *    If bit 4 of the sub header is set (MAP_ANONYMOUS), this segment is empty,
 *    and the value is zero. Otherwise this segment is 8 bytes long.
 *
 * Segment 8: Inode:
 *    If bit 4 of the sub header is set (MAP_ANONYMOUS), this segment is empty,
 *    and the value is zero. Otherwise this segment is 8 bytes long.
 *
 * Segment 9: PathLength:
 *    If bit 4 of the sub header is set (MAP_ANONYMOUS), this segment is empty,
 *    and the value is zero. Otherwise this segment is 2 bytes long.
 *
 * Segment 10: Path:
 *    If bit 4 of the sub header is set (MAP_ANONYMOUS), this segment is empty.
 *    Otherwise it consists of as many bytes as indicated by path_length.
 *
 * MUNMAP Packets:
 *
 * Segment 1: Packet Header:
 *    bits 1-0: 2-bit primary opcode == 11
 *    bits 4-2: 3-bit extended opcode == 100
 *    bits 9-5: reserved, must be zero
 *
 *    Compression flags:
 *    bit 5: reserved, must be zero
 *    bit 6: reserved, must be zero
 *    bit 7: reserved, must be zero
 *
 * Segment 2: munmap data: sizeof(uintptr_t) times 2 bytes.
 *
 * Segment 3: Global Memory Access Counter:
 *    This segment follows the sames rules as the Global Memory Access
 *    Counter segment in memory access packets.
 *
 * MREMAP Packets:
 *
 * Segment 1: Packet Header:
 *    bits 1-0: 2-bit primary opcode == 11
 *    bits 4-2: 3-bit extended opcode == 101
 *    bits 9-5: reserved, must be zero
 *
 *    Compression flags:
 *    bit 5: reserved, must be zero
 *    bit 6: reserved, must be zero
 *    bit 7: reserved, must be zero
 *
 * Segment 2: mremap data: sizeof(uintptr_t) times 4 bytes.
 *
 * Segment 3: Global Memory Access Counter:
 *    This segment follows the sames rules as the Global Memory Access
 *    Counter segment in memory access packets.
 *
 * KILLWP Packets:
 *
 * Segment 1: Packet Header:
 *    bits 1-0: 2-bit primary opcode == 11
 *    bits 4-2: 3-bit extended opcode == 111
 *    bits 7-5: 3-bit extended extended opcode == 000
 *    bits 9-8: reserved, must be zero
 *
 *    Compression flags:
 *    bit 5: reserved, must be zero
 *    bit 6: reserved, must be zero
 *    bit 7: reserved, must be zero
 *
 * Segment 2: 4-byte watchpoint group id.
 *
 * BURST_END Packets:
 *
 * Segment 1: Packet Header:
 *    bits 1-0: 2-bit primary opcode == 11
 *    bits 4-2: 3-bit extended opcode == 111
 *    bits 7-5: 3-bit extended extended opcode == 001
 *    bits 9-8: reserved, must be zero
 *
 *    Compression flags:
 *    bit 5: reserved, must be zero
 *    bit 6: reserved, must be zero
 *    bit 7: reserved, must be zero
 *
 * Segment 2: Global Memory Access Counter:
 *    This segment follows the sames rules as the Global Memory Access
 *    Counter segment in memory access packets.
 *
 * THREAD Packets:
 *
 * Segment 1: Packet Header:
 *    bits 1-0: 2-bit primary opcode == 11
 *    bits 4-2: 3-bit extended opcode == 111
 *    bits 7-5: 3-bit extended extended opcode == 010
 *    bits 9-8: reserved, must be zero
 *
 *    Compression flags:
 *    bit 5: reserved, must be zero
 *    bit 6: reserved, must be zero
 *    bit 7: reserved, must be zero
 *
 * Segment 2: 4-byte thread id
 *
 * Segment 3: Global Memory Access Counter:
 *    This segment follows the sames rules as the Global Memory Access
 *    Counter segment in memory access packets.
 *
 * Segment 4: 1-byte flag field
 *    bit 0:   1 if this thread has not been instrumented before, otherwise 0
 *    bit 7-1: reserved, must be zero
 *
 * GLOBAL_MACNT Packets:
 *
 * Segment 1: Packet Header:
 *    bits 1-0: 2-bit primary opcode == 11
 *    bits 4-2: 3-bit extended opcode == 111
 *    bits 7-5: 3-bit extended extended opcode == 011
 *    bits 9-8: reserved, must be zero
 *
 *    Compression flags:
 *    bit 5: reserved, must be zero
 *    bit 6: reserved, must be zero
 *    bit 7: reserved, must be zero
 *
 * Segment 2: Global Memory Access Counter:
 *    This segment follows the sames rules as the Global Memory Access
 *    Counter segment in memory access packets.
 *
 * LOCAL_MACNT Packets:
 *
 * Segment 1: Packet Header:
 *    bits 1-0: 2-bit primary opcode == 11
 *    bits 4-2: 3-bit extended opcode == 111
 *    bits 7-5: 3-bit extended extended opcode == 100
 *    bits 9-8: reserved, must be zero
 *
 *    Compression flags:
 *    bit 5: reserved, must be zero
 *    bit 6: reserved, must be zero
 *    bit 7: reserved, must be zero
 *
 * Segment 2: 4-byte thread id
 *
 * Segment 3: Local Memory Access Counter:
 *    This segment follows the sames rules as the Local Memory Access
 *    Counter segment in memory access packets.
 *
 * THREAD_EXIT Packets:
 *
 * Segment 1: Packet Header:
 *    bits 1-0: 2-bit primary opcode == 11
 *    bits 4-2: 3-bit extended opcode == 111
 *    bits 7-5: 3-bit extended extended opcode == 101
 *    bits 9-8: reserved, must be zero
 *
 *    Compression flags:
 *    bit 5: reserved, must be zero
 *    bit 6: reserved, must be zero
 *    bit 7: reserved, must be zero
 *
 * Segment 2: 4-byte thread id
 *
 * Segment 3: Global Memory Access Counter:
 *    This segment follows the sames rules as the Global Memory Access
 *    Counter segment in memory access packets.
 *
 * Segment 4: Local Memory Access Counter:
 *    This segment follows the sames rules as the Local Memory Access
 *    Counter segment in memory access packets.
 */

struct atf_header {	/* file- and in-core format */
    unsigned int magic;
    unsigned int size;
    unsigned int flags;
    unsigned int pid;
    unsigned int sampler_period;
    unsigned char sampler_log2linesizes[8];
    unsigned int sampler_count;
};

/* atf_header.flags public bits (private bits are deliberately excluded) */
#define ATF_HEADER_FLAGS_BURST	(1<<16)

/* <net/if_arp.h> defines ATF_MAGIC... */
#define AM_ATF_MAGIC	(('A'<<24)|('T'<<16)|('F'<<8)|('6'<<0))

struct atf_packet {	/* in-core format only */
    /*
     * The in-core packet header has four componenets:
     * bits 1-0: 2-bit primary opcode (RO/WO/RW/EXT)
     * bits 4-2: if primary opcode is EXT then extended opcode, otherwise 0
     * bits 7-5: opcode qualifier
     * - if opcode is PREFETCH then bits 7-5 contain prefetch kind
     * - if opcode is MOVNT then bit 5 indicates WO (0) or RO (1)
     * - for other opcodes this field is undefined
     * bits 31-8: opcode-specific data
     * - if opcode denotes a memory packet (RW/WO/RW or EXT:PREFETCH/MOVNT/CLFLUSH)
     *   then bits 31-8 contain the memory access size in bytes
     * - for other opcodes this field is undefined
     */
    unsigned int header;
    union {
	struct {
	    uintptr_t daddr;
	    uintptr_t iaddr;
	    unsigned long long local_macnt;
	    unsigned long long global_macnt;
	    unsigned int tid;
	    unsigned int group_id;
	} mem;
	struct {
	    unsigned char nrpop;
	    unsigned char nrpush;
	    unsigned char nrlost;
	    unsigned int tid;
	    uintptr_t pushes[32];
	    int absolute;
	    unsigned long long global_macnt;
	} shadow;
	struct {
	    uintptr_t start;
	    size_t length;
	    unsigned long long offset;
	    unsigned char prot;
	    unsigned char flags;
	    unsigned long long global_macnt;
	    unsigned long long device;
	    unsigned long long inode;
	    unsigned short path_length;
	    char path[1024];
	} mmap;
	struct {
	    uintptr_t start;
	    size_t length;
	    unsigned long long global_macnt;
	} munmap;
	struct {
	    uintptr_t old_address;
	    uintptr_t new_address;
	    size_t old_size;
	    size_t new_size;
	    unsigned long long global_macnt;
	} mremap;
	struct {
	    unsigned int group_id;
	} killwp;
	struct {
	    unsigned long long global_macnt;
	} burst_end;
	struct {
	    unsigned int tid;
	    unsigned long long global_macnt;
	    unsigned char flags;
	} thread;
	struct {
	    unsigned long long global_macnt;
	} global_macnt;
	struct {
	    unsigned int tid;
	    unsigned long long local_macnt;
	} local_macnt;
	struct {
	    unsigned int tid;
	    unsigned long long global_macnt;
	    unsigned long long local_macnt;
	} thread_exit;
    } u;
};

/* atf_packet.header primary opcodes */
#define ATF_OP_MEM_RO		0
#define ATF_OP_MEM_WO		1
#define ATF_OP_MEM_RW		2
#define ATF_OP_EXT		3	/* extended opcode in higher bits */

/* atf_packet.header extended opcodes */
#define ATF_XOP_MEM_CLFLUSH	0
#define ATF_XOP_MEM_PREFETCH	1
#define ATF_XOP_SHADOW		2
#define ATF_XOP_MMAP		3
#define ATF_XOP_MUNMAP		4
#define ATF_XOP_MREMAP		5
#define ATF_XOP_MEM_MOVNT	6
#define ATF_XOP_EXT		7	/* extended opcode in higher bits */

/* atf_packet.header extended extended opcodes */
#define ATF_XXOP_KILLWP		0
#define ATF_XXOP_BURST_END      1
#define ATF_XXOP_THREAD         2
#define ATF_XXOP_GLOBAL_MACNT   3
#define ATF_XXOP_LOCAL_MACNT    4
#define ATF_XXOP_THREAD_EXIT    5

/* atf_packet.header prefetch kinds */
#define ATF_PREFETCH_AMD	0	/* 0f 0d /0 */
#define ATF_PREFETCH_AMD_W	1	/* 0f 0d /1 */
#define ATF_PREFETCH_INTEL_NTA	2	/* 0f 18 /0 */
#define ATF_PREFETCH_INTEL_T0	3	/* 0f 18 /1 */
#define ATF_PREFETCH_INTEL_T1	4	/* 0f 18 /2 */
#define ATF_PREFETCH_INTEL_T2	5	/* 0f 18 /3 */

/* atf_packet.u.mmap.prot bits */
#define ATF_MMAP_PROT_READ	0x1
#define ATF_MMAP_PROT_WRITE	0x2
#define ATF_MMAP_PROT_EXEC	0x4

/* atf_packet.u.mmap.flags bits */
#define ATF_MMAP_MAP_SHARED	0x1	/* if not set MAP_PRIVATE is implied */
#define ATF_MMAP_MAP_ANONYMOUS	0x2	/* path[] is irrelevant and unspecified */

/* atf_packet.u.thread.flags bits */
#define ATF_THREAD_NEW          0x1

#endif /* ATF_H */
