/** @file sampler/src/atf_reader.c */

/* Copyright (c) 2006-2011 Rogue Wave Software, Inc.  All rights reserved. */

/*
 * Author: Mikael Pettersson.
 */

/* Visual Studio defines _WIN32 and _WIN64 unlike mingw. */
#if !defined(__WINNT__) && defined(_WIN32)
#define __WINNT__
#endif

#if !defined(__x86_64__) && defined(_WIN64)
#define __x86_64__
#endif

#include <errno.h>
#if defined(__WINNT__)
#include <fcntl.h>	/* _O_BINARY */
#include <io.h>		/* _setmode() */
#endif /* __WINNT__ */
#include <stddef.h>	/* offsetof() */
#include <stdio.h>
#include <string.h>	/* strerror() */
#include "atf.h"
#include "atf_reader.h"
#include "decode.h"

#if defined(__x86_64__)
static inline void set_compat32(struct atf_reader *reader, int val)
{
    reader->compat32 = val;
}

static inline int get_compat32(const struct atf_reader *reader)
{
    return reader->compat32;
}
#else
static inline void set_compat32(struct atf_reader *reader, int val) { }
static inline int get_compat32(const struct atf_reader *reader) { return 0; }
#endif

static int atf_read_uintptrt(struct atf_reader *reader, uintptr_t *dst)
{
    int n;

    if (get_compat32(reader)) {
	unsigned int tmp;
	n = (int)fread(&tmp, sizeof tmp, 1, reader->fin);
	*dst = tmp;
    } else
	n = (int)fread(dst, sizeof *dst, 1, reader->fin);
    return n;
}

static int atf_read_uchar(struct atf_reader *reader, unsigned char *dst)
{
    return (int)fread(dst, sizeof *dst, 1, reader->fin);
}

static int atf_read_uint(struct atf_reader *reader, unsigned int *dst)
{
    return (int)fread(dst, sizeof *dst, 1, reader->fin);
}

static int atf_read_sizet(struct atf_reader *reader, size_t *dst)
{
    if (sizeof(size_t) != sizeof(uintptr_t)) {
	uintptr_t tmp;
	int n = atf_read_uintptrt(reader, &tmp);
	*dst = tmp;
	return n;
    }
    return atf_read_uintptrt(reader, (uintptr_t*)dst);
}

static int atf_read_ulonglong(struct atf_reader *reader, unsigned long long *dst)
{
    return (int)fread(dst, sizeof *dst, 1, reader->fin);
}

static int atf_read_macnt(
    struct atf_reader *reader, unsigned long long *value, unsigned long long *prev_value)
{
    int ch;
    unsigned char b[8];
    long long diff;

    if (reader->macnts_are_uncompressed)
	return atf_read_ulonglong(reader, value);

    b[0] = ch = fgetc(reader->fin);
    if (ch == EOF)
	return 0;
    if ((b[0] & 1) == 0) {
	diff = (signed char)b[0] >> 1;
    } else if (b[1] = ch = fgetc(reader->fin), ch == EOF) {
	return 0;
    } else if ((b[0] & 2) == 0) {
	diff = (((signed char)b[1] << 8) | b[0]) >> 2;
    } else if (b[2] = ch = fgetc(reader->fin), ch == EOF) {
	return 0;
    } else if ((b[0] & 4) == 0) {
	diff = (((signed char)b[2] << 16) | (b[1] << 8) | b[0]) >> 3;
    } else if (fread(&b[3], 5, 1, reader->fin) != 1) {
	return 0;
    } else {
	diff = (long long)(((unsigned long long)b[7] << 56) |
			   ((unsigned long long)b[6] << 48) |
			   ((unsigned long long)b[5] << 40) |
			   ((unsigned long long)b[4] << 32) |
			   ((unsigned int)b[3] << 24) |
			   ((unsigned int)b[2] << 16) |
			   ((unsigned int)b[1] << 8) |
			   ((unsigned int)b[0] << 0)) >> 3;
    }
    *value = *prev_value + diff;
    *prev_value = *value;
    return 1;
}

static int atf_read_global_macnt(struct atf_reader *reader, unsigned long long *value)
{
    return atf_read_macnt(reader, value, &reader->prev_global_macnt);
}

static int atf_read_local_macnt(struct atf_reader *reader, unsigned long long *value)
{
    return atf_read_macnt(reader, value, &reader->prev_local_macnt);
}

static int atf_read_compressed_uintptrt(
    struct atf_reader *reader, uintptr_t *value, uintptr_t *prev_value)
{
    int ch, ch2, ch3;
    int idiff;

    ch = fgetc(reader->fin);
    if (ch == EOF)
	return 0;
    if ((ch & 0x1) == 0) {
	idiff = ((signed char)ch >> 1);
    } else {
	ch2 = fgetc(reader->fin);
	if (ch2 == EOF)
	    return 0;
	if ((ch & 0x2) == 0) {
	    idiff = (((signed char)ch2 << 8) | (unsigned char)ch) >> 2;
	} else {
	    ch3 = fgetc(reader->fin);
	    if (ch3 == EOF)
		return 0;
	    idiff = (((signed char)ch3 << 16) | ((unsigned char)ch2 << 8) | (unsigned char)ch) >> 2;
	}
    }
    *value = *prev_value + idiff;
    if (get_compat32(reader))
	*value &= 0xffffffffUL;
    *prev_value = *value;
    return 1;
}

static int atf_read_compressed_uint(
    struct atf_reader *reader, unsigned int *value, unsigned int *prev_value)
{
    if (sizeof(unsigned int) != sizeof(uintptr_t)) {
	uintptr_t tmp_prev = *prev_value;
	uintptr_t tmp_val;
	int n = atf_read_compressed_uintptrt(reader, &tmp_val, &tmp_prev);
	*prev_value = (unsigned int)tmp_prev;
	*value = (unsigned int)tmp_val;
	return n;
    }
    return atf_read_compressed_uintptrt(reader, (uintptr_t*)value, (uintptr_t*)prev_value);
}

static int atf_read_compressible_uintptrt(
    struct atf_reader *reader,
    unsigned char flags,
    unsigned int hdr_compressed_bit,
    uintptr_t *value,
    uintptr_t *prev_value)
{
    if (!(flags & (1 << hdr_compressed_bit))) {
	if (atf_read_uintptrt(reader, value) != 1)
	    return 0;
	*prev_value = *value;
    } else {
	if (atf_read_compressed_uintptrt(reader, value, prev_value) != 1)
	    return 0;
    }
    return 1;
}

static int atf_read_tid(struct atf_reader *reader,
			unsigned char flags,
			unsigned int *value)
{
    if (!(flags & (1 << 5))) {
	*value = reader->prev_tid;
    } else if (!(flags & (1 << 6))) {
	if (atf_read_uint(reader, value) != 1)
	    return 0;
	reader->prev_tid = *value;
    } else {
	if (atf_read_compressed_uint(reader, value, &reader->prev_tid) != 1)
	    return 0;
    }
    return 1;
}

static int atf_read_nrlost(struct atf_reader *reader,
			   unsigned char flags,
			   unsigned char *value)
{
    if (!(flags & (1 << 7))) {
	*value = 0;
    } else {
	int ch = fgetc(reader->fin);
	if (ch == EOF)
	    return 0;
	*value = ch;
    }
    return 1;
}

int atf_read_init(FILE *fin, struct atf_reader *reader, struct atf_header *header)
{
#if defined(__WINNT__)
    /* If fin is stdin then it's usually in text translation mode,
       which emphatically doesn't work. Unconditionally reset fin
       to binary mode. */
    if (_setmode(_fileno(fin), _O_BINARY) == -1) {
	fprintf(stderr, "%s: unable to set FILE %p in binary mode: %s\n",
		__FUNCTION__, fin, strerror(errno));
	return -1;
    }
#endif /* __WINNT__ */

    reader->fin = fin;
    reader->prev_iaddr = 0;
    reader->prev_tid = 0;
    reader->prev_local_macnt = 0;
    reader->prev_global_macnt = 0;

    if (fread(header, sizeof *header, 1, fin) != 1) {
	fprintf(stderr, "%s: unable to read ATF header: %s\n",
		__FUNCTION__, feof(fin) ? "end-of-file" : strerror(errno));
	return -1;
    }
    if (header->magic != AM_ATF_MAGIC) {
	fprintf(stderr, "%s: input is not a native ATF file\n",
		__FUNCTION__);
	return -1;
    }
    if (header->size != sizeof *header) {
	fprintf(stderr, "%s: unsupported ATF header size: %u\n",
		__FUNCTION__, header->size);
	return -1;
    }
    if ((header->flags & 0xFFFC00FF) != 0x00000086)
	goto err_flags;

    reader->macnts_are_uncompressed = (header->flags & (1 << 17)) == 0;

    set_compat32(reader, 0);
    if (((header->flags >> 8) & 0xFF) != sizeof(uintptr_t)) {
	if (sizeof(uintptr_t) != sizeof(int) &&
	    ((header->flags >> 8) & 0xFF) == sizeof(int))
	    set_compat32(reader, 1);
	else
	    goto err_flags;
    }
    return 0;

 err_flags:
    fprintf(stderr, "%s: unsupported ATF flags: 0x%08x\n",
	    __FUNCTION__, header->flags);
    return -1;
}

static int atf_read_mem_packet(
    struct atf_reader *reader, unsigned int header, unsigned char flags, struct atf_packet *packet)
{
    unsigned int bytes;
    unsigned int size_code;

    if ((header & 0x03) == ATF_OP_EXT) {
	packet->header = header & 0xFF;
	if ((header & 0x1F) == (ATF_OP_EXT | (ATF_XOP_MEM_MOVNT << 2)))
	    /* EXT:2 | (MOVNT << 2):3 | (R << 5):1 | (size_code << 6):4 */
	    size_code = (header >> 6) & 0x0F;
	else
	    size_code = SIZE_CODE_1;
    } else {
	packet->header = header & 0x03;
	size_code = (header >> 2) & 0x0F;
    }
    bytes = decode_size(size_code);
    if (!bytes) {
	fprintf(stderr, "%s: Error: ATF memory packet with bogus size_code %u\n",
		__FUNCTION__, size_code);
	errno = EILSEQ;	/* was EPROTO, but WinNT does not have it */
	return -1;
    }
    packet->header |= (bytes << 8);

    if (atf_read_uintptrt(reader, &packet->u.mem.daddr) != 1)
	return 0;

    if (atf_read_compressible_uintptrt(reader, flags, 7, &packet->u.mem.iaddr, &reader->prev_iaddr) != 1)
	return 0;

    if (atf_read_local_macnt(reader, &packet->u.mem.local_macnt) != 1)
	return 0;

    if (atf_read_global_macnt(reader, &packet->u.mem.global_macnt) != 1)
	return 0;

    if (atf_read_tid(reader, flags, &packet->u.mem.tid) != 1)
	return 0;

    if (atf_read_uint(reader, &packet->u.mem.group_id) != 1)
	return 0;

    return 1;
}

static int atf_read_shadow_packet(
    struct atf_reader *reader, unsigned int header, unsigned char flags, struct atf_packet *packet)
{
    int ch;
    unsigned int i, nrpush;

    packet->u.shadow.absolute = !!(header & 0x20);

    packet->header = header & 0x1F;

    ch = fgetc(reader->fin);
    if (ch == EOF)
	return 0;
    packet->u.shadow.nrpop = ch;

    ch = fgetc(reader->fin);
    if (ch == EOF)
	return 0;
    nrpush = ch;
    packet->u.shadow.nrpush = nrpush;

    if (atf_read_nrlost(reader, flags, &packet->u.shadow.nrlost) != 1)
	return 0;

    if (atf_read_tid(reader, flags, &packet->u.shadow.tid) != 1)
	return 0;

    for (i = 0; i < nrpush; ++i)
	if (atf_read_uintptrt(reader, &packet->u.shadow.pushes[i]) != 1)
	    return 0;

    if (atf_read_global_macnt(reader, &packet->u.shadow.global_macnt) != 1)
	return 0;

    return 1;
}

static int atf_read_mmap_packet(
    struct atf_reader *reader, unsigned int header, unsigned char flags, struct atf_packet *packet)
{
    int ch;

    packet->header = header & 0x1F;

    ch = fgetc(reader->fin);
    if (ch == EOF)
	return 0;
    packet->u.mmap.prot = ch & 0x7;
    packet->u.mmap.flags = (ch >> 3) & 0x3;

    if (atf_read_uintptrt(reader, &packet->u.mmap.start) != 1 ||
	atf_read_sizet(reader, &packet->u.mmap.length) != 1)
	return 0;

    if (!(flags & (1 << 5)))
	packet->u.mmap.offset = 0;
    else if (fread(&packet->u.mmap.offset, 8, 1, reader->fin) != 1)
	return 0;

    if (atf_read_global_macnt(reader, &packet->u.mmap.global_macnt) != 1)
	return 0;

    if (packet->u.mmap.flags & ATF_MMAP_MAP_ANONYMOUS) {
	packet->u.mmap.device = 0;
	packet->u.mmap.inode = 0;
	packet->u.mmap.path_length = 0;
	packet->u.mmap.path[0] = '\0';
    } else {
	unsigned short len;

	if (atf_read_ulonglong(reader, &packet->u.mmap.device) != 1 ||
	    atf_read_ulonglong(reader, &packet->u.mmap.inode) != 1 ||
	    fread(&len, 2, 1, reader->fin) != 1)
	    return  0;

	if (len > 1024) {
	    fprintf(stderr, "%s: Error: ATF MMAP packet with invalid path_length %u\n",
		    __FUNCTION__, len);
	    return -1;
	}
	packet->u.mmap.path_length = len;
	if (fread(packet->u.mmap.path, len, 1, reader->fin) != 1)
	    return 0;
    }

    return 1;
}

static int atf_read_munmap_packet(
    struct atf_reader *reader, unsigned int header, unsigned char flags, struct atf_packet *packet)
{
    packet->header = header & 0x1F;

    if (atf_read_uintptrt(reader, &packet->u.munmap.start) != 1 ||
	atf_read_sizet(reader, &packet->u.munmap.length) != 1 ||
	atf_read_global_macnt(reader, &packet->u.munmap.global_macnt) != 1)
	return 0;

    return 1;
}

static int atf_read_mremap_packet(
    struct atf_reader *reader, unsigned int header, unsigned char flags, struct atf_packet *packet)
{
    packet->header = header & 0x1F;

    if (atf_read_uintptrt(reader, &packet->u.mremap.old_address) != 1 ||
	atf_read_uintptrt(reader, &packet->u.mremap.new_address) != 1 ||
	atf_read_sizet(reader, &packet->u.mremap.old_size) != 1 ||
	atf_read_sizet(reader, &packet->u.mremap.new_size) != 1 ||
	atf_read_global_macnt(reader, &packet->u.mremap.global_macnt) != 1)
	return 0;

    return 1;
}

static int atf_read_killwp_packet(
    struct atf_reader *reader, unsigned char header, unsigned char flags, struct atf_packet *packet)
{
    packet->header = header & 0xFF;

    if (atf_read_uint(reader, &packet->u.killwp.group_id) != 1)
	return 0;

    return 1;
}

static int atf_read_burst_end_packet(
    struct atf_reader *reader, unsigned char header, unsigned char flags, struct atf_packet *packet)
{
    packet->header = header & 0xFF;

    if (atf_read_global_macnt(reader, &packet->u.burst_end.global_macnt) != 1)
	return 0;

    return 1;
}

static int atf_read_thread_packet(
    struct atf_reader *reader, unsigned char header, unsigned char flags, struct atf_packet *packet)
{
    packet->header = header & 0xFF;

    if (atf_read_uint(reader, &packet->u.thread.tid) != 1 ||
	atf_read_global_macnt(reader, &packet->u.thread.global_macnt) != 1 ||
	atf_read_uchar(reader, &packet->u.thread.flags) != 1)
	return 0;

    return 1;
}

static int atf_read_global_macnt_packet(
    struct atf_reader *reader, unsigned char header, unsigned char flags, struct atf_packet *packet)
{
    packet->header = header & 0xFF;

    if (atf_read_global_macnt(reader, &packet->u.global_macnt.global_macnt) != 1)
	return 0;

    return 1;
}

static int atf_read_local_macnt_packet(
    struct atf_reader *reader, unsigned char header, unsigned char flags, struct atf_packet *packet)
{
    packet->header = header & 0xFF;

    if (atf_read_uint(reader, &packet->u.local_macnt.tid) != 1 ||
	atf_read_local_macnt(reader, &packet->u.local_macnt.local_macnt) != 1)
	return 0;

    return 1;
}

static int atf_read_thread_exit_packet(
    struct atf_reader *reader, unsigned char header, unsigned char flags, struct atf_packet *packet)
{
    packet->header = header & 0xFF;

    if (atf_read_uint(reader, &packet->u.thread_exit.tid) != 1 ||
	atf_read_global_macnt(reader, &packet->u.thread_exit.global_macnt) != 1 ||
	atf_read_local_macnt(reader, &packet->u.thread_exit.local_macnt) != 1)
	return 0;

    return 1;
}

int atf_read_packet(struct atf_reader *reader, struct atf_packet *packet)
{
    int ch;
    unsigned char flags;
    unsigned int header;
    int ret;

    ch = fgetc(reader->fin);
    if (ch == EOF)
	return 0;
    flags = ch & 0xe0;
    header = ch & 0x1f;
    if ((header & 0x1c) == 0x1c) {
	ch = fgetc(reader->fin);
	if (ch == EOF) {
	    fprintf(stderr, "%s: Error: EOF while reading ATF packet\n",
		    __FUNCTION__);
	    return -1;
	}
	header = (ch << 2) | (header & 0x03);
    }

    /* inspect opcode and call appropriate reader procedure */
    switch (header & 0x1F) {
      case ((ATF_XOP_SHADOW << 2) | ATF_OP_EXT):
	ret = atf_read_shadow_packet(reader, header, flags, packet);
	break;
      case ((ATF_XOP_MMAP << 2) | ATF_OP_EXT):
	ret = atf_read_mmap_packet(reader, header, flags, packet);
	break;
      case ((ATF_XOP_MUNMAP << 2) | ATF_OP_EXT):
	ret = atf_read_munmap_packet(reader, header, flags, packet);
	break;
      case ((ATF_XOP_MREMAP << 2) | ATF_OP_EXT):
	ret = atf_read_mremap_packet(reader, header, flags, packet);
	break;
      case ((ATF_XOP_EXT << 2) | ATF_OP_EXT):
	switch (header & 0xE0) {
	  case ATF_XXOP_KILLWP << 5:
	    ret = atf_read_killwp_packet(reader, header, flags, packet);
	    break;
	  case ATF_XXOP_BURST_END << 5:
	    ret = atf_read_burst_end_packet(reader, header, flags, packet);
	    break;
	  case ATF_XXOP_THREAD << 5:
	    ret = atf_read_thread_packet(reader, header, flags, packet);
	    break;
	  case ATF_XXOP_GLOBAL_MACNT << 5:
	    ret = atf_read_global_macnt_packet(reader, header, flags, packet);
	    break;
	  case ATF_XXOP_LOCAL_MACNT << 5:
	    ret = atf_read_local_macnt_packet(reader, header, flags, packet);
	    break;
	  case ATF_XXOP_THREAD_EXIT << 5:
	    ret = atf_read_thread_exit_packet(reader, header, flags, packet);
	    break;
	  default:
	    fprintf(stderr, "%s: Error: Unsupported ATF packet header: 0x%x\n",
		    __FUNCTION__, header);
	    return -1;
	}
	break;
      default:
	ret = atf_read_mem_packet(reader, header, flags, packet);
    }

    if (ret != 0)	/* 1 (ok) or -1 (already reported) */
	return ret;

    if (feof(reader->fin))
	fprintf(stderr, "%s: Error: EOF while reading ATF packet\n",
		__FUNCTION__);
    else
	fprintf(stderr, "%s: Error while reading ATF packet: %s\n",
		__FUNCTION__, strerror(errno));
    return -1;
}
