/** @file sampler/src/decode.h */

/* Copyright (c) 2006-2011 Rogue Wave Software, Inc.  All rights reserved. */

/*
 * Author: Mikael Pettersson.
 */
#ifndef DECODE_H
#define DECODE_H

#include "opcode.h"

/* honour DECODE_TARGET_${ISA} if set, otherwise set it to runtime platform */
#if defined(DECODE_TARGET_X86_64) || defined(DECODE_TARGET_I386)
#elif defined(__x86_64__)
#define DECODE_TARGET_X86_64
#elif defined(__i386__)
#define DECODE_TARGET_I386
#endif

struct opcode_descriptor {
    unsigned short opcode;
    unsigned short info;
    unsigned int opnds;
};

struct insn {
    struct opcode_descriptor desc;
    unsigned short flags;
    /*unsigned char offset_opcode:4;*/	/* offset to 1st opcode byte */ /* XXX: unused */
    unsigned char offset_modrm:4;	/* offset to modrm, or 0 */
    unsigned char length_modrm:4;	/* length of modrm, or 0 */
    unsigned char offset_imm:4;		/* offset to imm opnd, or 0 */
    unsigned char length_bytes:4;
};

/* flag bits for legacy prefixes */
#define PFX_66		0
#define PFX_OPND_SZ	PFX_66
#define PFX_ADDR_SZ	1
#define PFX_SEG		2	/* bits 2-4 */
#define PFX_LOCK	5
#define PFX_F3		6
#define PFX_REPE	PFX_F3
#define PFX_F2		7
#define PFX_REPNE	PFX_F2

/* PFX_SEG values (3 bits) */
#define PFX_SEG_NONE	0
#define PFX_SEG_CS	1
#define PFX_SEG_DS	2
#define PFX_SEG_ES	3
#define PFX_SEG_FS	4
#define PFX_SEG_GS	5
#define PFX_SEG_SS	6
#define PFX_SEG_MASK	0x7	/* 3 bits */

/* flags bits for a VEX prefix, VEX B/X/R/W use the same bits as REX B/X/R/W,
   VEX.pp is encoded in the legacy 66/F3/F2 prefix bits */
#define PFX_VEX		8
#define PFX_VEX_L	9
#define PFX_VEX_V1111	10
#define PFX_VEX_B	11	/* must equal PFX_REX_B */
#define PFX_VEX_X	12	/* must equal PFX_REX_X */
#define PFX_VEX_R	13	/* must equal PFX_REX_R */
#define PFX_VEX_W	14	/* must equal PFX_REX_W */

/* flag bits for a REX prefix; logically distinct from the legacy prefixes */
#if defined(DECODE_TARGET_X86_64)
#define PFX_REX_B	11
#define PFX_REX_X	12
#define PFX_REX_R	13
#define PFX_REX_W	14
#define PFX_REX		15
#endif

/* info: low 8 bits is a type code, high 8 bits are type-specific data */
#include "insn_info.h"

#define INSN_HAS_IMPLICIT_MEM_OPND	256

/* stored in bits 8-15 iff bits 0-7 == INSN_IS_PREFETCH */
#define INSN_PREFETCH_AMD	0	/* 0f 0d /0 */
#define INSN_PREFETCH_AMD_W	1	/* 0f 0d /1 */
#define INSN_PREFETCH_INTEL_NTA	2	/* 0f 18 /0 */
#define INSN_PREFETCH_INTEL_T0	3	/* 0f 18 /1 */
#define INSN_PREFETCH_INTEL_T1	4	/* 0f 18 /2 */
#define INSN_PREFETCH_INTEL_T2	5	/* 0f 18 /3 */

/* stored in bit 8 iff bits 0-7 == INSN_IS_MOVNT */
#define INSN_MOVNT_IS_RO	1	/* 1: RO (movntdqa), 0: WO (all others) */

static inline unsigned int insn_info_type(const struct insn *insn)
{
    return insn->desc.info & 0xFF;
}

int decode(const unsigned char *pc, struct insn *pinsn);

#if defined(DECODE_DFA) && defined(DECODE_GENERATE)
void decode_generate(void (*emitfn)(const unsigned char *, unsigned int));
#endif

/* packed size code (4 bits) */
typedef enum size_code {
    /* the order here is chosen to allow the five most
       common access sizes to be assigned values < 7 */
    SIZE_CODE_1,
    SIZE_CODE_2,
    SIZE_CODE_4,
    SIZE_CODE_8,
    SIZE_CODE_16,
    SIZE_CODE_6,
    SIZE_CODE_10,
    SIZE_CODE_14,
    SIZE_CODE_28,
    SIZE_CODE_32,
    SIZE_CODE_40,
    SIZE_CODE_98,
    SIZE_CODE_108,
    SIZE_CODE_512,
    SIZE_CODE_1024,
#if defined(DECODE_TARGET_I386)
    SIZE_CODE_5,
#endif
} size_code_t;

static inline unsigned int decode_size(unsigned int size_code)
{
    switch (size_code) {
      case SIZE_CODE_1:		return 1;
      case SIZE_CODE_2:		return 2;
      case SIZE_CODE_4:		return 4;
#if defined(DECODE_TARGET_I386)
      case SIZE_CODE_5:		return 5;
#endif
      case SIZE_CODE_6:		return 6;
      case SIZE_CODE_8:		return 8;
      case SIZE_CODE_10:	return 10;
      case SIZE_CODE_14:	return 14;
      case SIZE_CODE_16:	return 16;
      case SIZE_CODE_28:	return 28;
      case SIZE_CODE_32:	return 32;
      case SIZE_CODE_40:	return 40;
      case SIZE_CODE_98:	return 98;
      case SIZE_CODE_108:	return 108;
      case SIZE_CODE_512:	return 512;
      case SIZE_CODE_1024:	return 1024;
      default:			return 0;
    }
}

size_code_t insn_modrm_size_code(const struct insn *insn);

#endif /* DECODE_H */
