/** @file sampler/src/instrumenter_stub.h */

/* Copyright (c) 2007-2011 Rogue Wave Software, Inc.  All rights reserved. */

/*
 * The interface between the instrumenter stub and the instrumenter. Located in
 * the sampler directory to avoid dependencies on acumon in the sampler.
 *
 * All functions should be called using the instrumenter stack.
 */

#ifndef INSTRUMENTER_STUB_H
#define INSTRUMENTER_STUB_H

#include "am_stdint.h"
#if !defined(_MSC_VER)
#include <sys/types.h>  /* pid_t, not available when compiling with MSVC */
#endif
#include "am_registers.h"
#if defined(__WINNT__)
#include "windows_protocol.h"
#endif

#if !defined(__WINNT__)
/* Maximum length of the trace file prefix. */
#define TRACE_FILE_PREFIX_MAX 256
#endif

/* Must be identical for 32-bit and 64-bit target. */
struct sampler_params {

#if defined(__WINNT__)

    /* On windows, the bootstrap code may occur after all dll:s have
     * been loaded, but before they have been initialized.
     * Unfortunately, calling GetProcAddress from within the sampler
     * bootstrapper will invoke the dll initialization process, which
     * is then done on the sampler stack instead of the application
     * stack. That breaks at least cygwin applications.
     *
     * Instead we perform all GetProcAddress calls in the monitor on
     * behalf of the sampler. For this to work, all symbols must be in
     * one of the dll:s already loaded. Only kernel32 and ntdll are
     * guaranteed to be mapped into all processes, so we have to limit
     * ourselves to functions in those dlls. It probably wont work to
     * try to load additional libraries explicitly in the load_entry()
     * function, as that would probably cause the initialization of
     * all dlls to be performed.
     */

    uint64_t kernel32_IsWow64Process;
    uint64_t ntdll_NtContinue;
    uint64_t ntdll_NtFreeVirtualMemory;
    uint64_t ntdll_NtProtectVirtualMemory;
    uint64_t ntdll_NtAllocateVirtualMemory;
    uint64_t ntdll_LdrInitializeThunk;
    uint64_t ntdll_KiFastSystemCallRet;
    uint64_t ntdll_KiUserApcDispatcher;
    uint64_t ntdll_KiUserCallbackDispatcher;
    uint64_t ntdll_KiUserExceptionDispatcher;
    uint64_t ntdll_KiRaiseUserExceptionDispatcher;
#endif

    uint32_t sample_period;
    uint32_t sample_count;    /* 0 for constant sample period, non-0 for
				 automatic sample rate adjustment. */
    uint8_t log2linesizes[8];
    uint64_t bp;
#if defined(__WINNT__)
    union {
	void *hLogFile;		/* size varies */
	uint64_t pad;		/* ensure consistent size */
    } log_fd;
#else
    char trace_file_prefix[TRACE_FILE_PREFIX_MAX];
#endif

    /* Start and end of the memory area reserved for code allocation for 64-bit
       processes. Both are 0 for 32-bit processes. */
    uint64_t code_area_start;
    uint64_t code_area_end;

    /* The length of the burst sampling period in accesses from the start of
       the burst, or -1 (2^64-1) if burst sampling is disabled. */
    uint64_t burst_sampling_length;

    /* The length of the burst cooldown period in accesses from the start of
       the burst, or -1 (2^64-1) if burst sampling is disabled. */
    uint64_t burst_cooldown_length;

    /* Disable use of the application stack if non-0. */
    uint32_t safe_stack;

    /* Disable generation stack information if non-0. */
    uint32_t demo;

    /* Use stderr, use applications stderr instead of the acumon
       socket for logging if non-0. */
    uint32_t use_stderr;

    /* Padding to to avoid implicit tail padding on 64-bit. */
    uint32_t pad;
};
/* XXX: Acumon and Winmon still use the obsolete typedef */
typedef struct sampler_params sampler_params_t;

/*****************************************************************************
 *            Functions in the instrumenter stub called from the             *
 *  instrumenter loader, thread instrumenter and instrumenter unloader code. *
 *****************************************************************************/

/* Prototype for the load entry function of the instrumenter. */
typedef void (*load_entry_t)(
    const struct sampler_params *sampler_params,
    int socket_fd);

/* Prototype for the instrumentation entry function of the instrumenter. */
typedef void (*instrumentation_entry_t)(
    unsigned long stack_addr,
    am_registers_t *regs,
    int new_thread,
    unsigned int burst_generation);

/* Prototype for the unload entry function of the instrumenter. */
typedef void (*unload_entry_t)(void);

/*****************************************************************************
 *      Functions in the instrumenter stub called from the instrumenter.     *
 *****************************************************************************/

/* Handle a message to the thread from the monitor, by calling the
   corresponding callback. */
void handle_message(pid_t pid, pid_t tid);

/* Returns 0 if the thread should deinstrument, the new burst generation if the
   monitor chooses to cancel the deinstrumentation. */
unsigned int deinstrument_ready(pid_t pid, pid_t tid, const am_registers_t *regs, void *tls);

/* Called when the thread is ready to deinstrument. */
void deinstrument(pid_t pid, pid_t tid);

/* Called by the child thread when an instrumented thread executes a clone
   syscall. */
void cloned(pid_t pid, pid_t tid, uintptr_t stack_addr,
	    uintptr_t msg_counter_addr, uintptr_t noevpend_addr);

/* Called when an instrumented thread is about to do an exit syscall (but not
   for exit group syscalls). */
void exiting(pid_t pid, pid_t tid);

/* Notify the monitor that code has been generated at the specified address for
   an instruction the monitor has set a breakpoint on. */
void set_breakpoint(pid_t pid, pid_t tid, uint64_t addr);

/* Notify the monitor that code has been invalidated at the specified address
   for an instruction the monitor has set a breakpoint on. */
void inv_breakpoint(pid_t pid, pid_t tid, uint64_t addr);

/* Notify the monitor that the burst has ended. */
void send_burst_end(pid_t pid, pid_t tid);

/* Notify the monitor that we are about to protect a page. */
void send_protect_page(pid_t pid, pid_t tid, uint64_t addr);

/* Notify the monitor that we are about to unprotect a page. */
void send_unprotect_page(pid_t pid, pid_t tid, uint64_t addr);

/* Notify the monitor that the instrumenter has called am_abort().  */
void send_abort(const void *caller);

#if defined(__WINNT__)
/* Notify the monitor that we are about to patch some ntdll entries. */
void send_patches(unsigned int nrpatches, const struct monitor_patch_info *mpatches);

/* Notify the monitor that we are about to execute an unbalanced
   NtCallbackReturn. */
void send_callback_return(uintptr_t sp, uintptr_t bp);

/* Notify the monitor that another callback has been invoked immediately after
   an unbalanced NtCallbackReturn. */
void send_callback(void);

/* Notify the monitor of the addresses of the noevpend and monitor_msgs fields
   in the tls. */
void send_instrument_info(uintptr_t monitor_msgs_addr, uintptr_t noevpend_addr);

/* Notify the monitor that a spontaneously created thread
   has been instrumented. */
void send_spontaneous_instrument_done(void);

/* Notify the monitor that a spontaneously created thread
   has not been instrumented. */
void send_spontaneous_not_instrumented(void);

/* Look up addresses of ntdll/kernel32 functions. Limited to a static set of
   functions, known at compile time. The monitor performs the actual lookup.*/
void *lookup_address(const char *function_name);

#endif

/*****************************************************************************
 *      Functions in the instrumenter called from the instrumenter stub.     *
 *****************************************************************************/

#if !defined(__WINNT__)

/* Called to initialize the instrumenter state when the monitor has
   attached. However, it must not instrument any code since the page hijacked
   for the bootstrapping code may not be restored yet.

   The instrumenter must be ready to receive signals when this call
   returns.  */
typedef struct {
    uintptr_t msg_counter_addr;
    uintptr_t noevpend_addr;
} attach_info_t;

attach_info_t
monitor_attach_init(void *tls, int new_thread, unsigned int burst_generation);

/* Called to transfer control to instrumented code when the monitor has
   attached and the instrumenter state has been initialized.  */
__attribute__((__noreturn__)) __attribute__((regparm(2)))
void monitor_attach_boot(const am_registers_t *regs, int new_thread);

#endif /* !defined(__WINNT__) */

/* Called from handle_messages when a detach message has been received. */
void msg_deinstr(void);

/* Perform global once-only initialisation. Returns -1 on error, 0 if we should
   be disabled, +1 otherwise. */
int init_global(const struct sampler_params *sampler_params
#if defined(__WINNT__)
		, int has_pending_callback_returns
#endif
    );

/* Called from detach when the last thread detaches. Should free all global
   allocated resources. */
#if defined(__WINNT__)
void global_cleanup(int has_pending_callback_returns);
#else
void global_cleanup(void);
#endif

#endif /* INSTRUMENTER_STUB_H */
