/// @file tools/acurate/conf.cc

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#include "conf.hh"

#include "library/exe_path/exe_path.hh"
#include "library/version/product_version.h"
#include "library/version/copyright.h"
#include "library/lang/am_assert.hh"

#include <boost/program_options.hpp>
#include <boost/tokenizer.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/tuple/tuple.hpp>

#include "rng.hh"

#include <iostream>
#include <time.h>

#ifndef _WIN32
#include <unistd.h>
#else
#include <windows.h>
#include <stdio.h>
#include <tchar.h>
#endif

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace std;
using namespace acumem::library;
//using namespace acumem::analysis;

using boost::lexical_cast;

namespace po = boost::program_options;
namespace fs = boost::filesystem;

conf_t::conf_t(warning_consumer_ptr_t &warning_consumer,
	       feedback_ptr_t feedback)
    : warning_consumer(warning_consumer),
      feedback(feedback)
{
    installation_dir = acumem::exe_path::get_installation_directory();

    // XXX Also fix this value to something smaller:
    inactivity_timeout = 10;

    // 26^28 ~= 2^128
    cookie = "";
    random_number_generator_t rng;
    for(int i=0; i<28; i++)
	cookie += ('a' + (rng() % 26));

    server_port = 0;
    report_path = "acumem-report";
    feature_mt_advice = false;
    feature_sw_prefetch = false;
    feature_unlimited_callstack = false;
    feature_blocking_expert = false;
    feature_multisocket = false;
    intro_page = true;
    use_mr_scale = false;
    max_mr_scale = 1.0;
    allow_many_logins = false;
}

bool
conf_t::init(bool interactive, int argc, char **argv)
{
    bool ret = true;
    bool show_help = false;

    po::options_description opt_desc_visible("Options");
    po::options_description opt_desc_invisible("Options");

    opt_desc_visible.add_options()
	("help,h", "(optional) Print help message.")
	("version", "(optional) Print version information.")
	("in-file,i", po::value<string>(&report_file),
	 "Specifies the input report file.");

    opt_desc_invisible.add_options()
	("max-mr-scale", po::value<double>(&max_mr_scale),
	 "(experimental) Lock scale for mr series in graph to this value "
	 "(0-1). Default: auto.");
#if 0
	("no-intro",
	 "(optional) Don't prepare the initial introduction page.");
#endif

    if (static_report)
        opt_desc_visible.add_options()
            ("report,o", po::value<string>(&report_path),
             "(optional) Base name of top level file and subdirectory "
             "containing the report files, default 'acumem-report'.");
    else {
        opt_desc_visible.add_options()
            ("port", po::value<int>(&server_port),
             "(optional) Launch the web server on the specified port instead "
             "of selecting an arbitrary free port.")
            ("timeout", po::value<int>(&inactivity_timeout),
             "(optional) Terminate the web server when no activity has been "
             "seen from the web browser viewing the report for the specified "
             "number of seconds. The default is 10 seconds.")
            ("no-browser",
             "(optional) Do not launch a web browser viewing the report, just "
             "start the web server serving it.");
        opt_desc_invisible.add_options()
	    ("multiple-logins",
             "(optional) Define to allow multiple sessions to the same report")
            ;

        // This is a little fragile, html_front_t uses this field, which
        // in the server based report must be empty.
        report_path = "";
    }

    po::options_description opt_desc_all("Options");
    opt_desc_all
	.add(opt_desc_visible)
        .add(opt_desc_invisible);

    po::positional_options_description pos_desc;
    pos_desc.add("in-file", -1);

    po::variables_map vm;
    try {
	po::store(po::command_line_parser(argc, argv)
		  .options(opt_desc_all).positional(pos_desc).run(), vm);
	po::notify(vm);
    }
    catch (po::error &e) {
	warning_consumer->error(
            string("Command line parsing error: ") + e.what());

	ret = false;
    }

    if (vm.count("version") > 0) {
        printf("%s\n%s\n", PRODUCT_VERSION, COPYRIGHT);
        return false;
    }

    if (vm.count("help") != 0)
	show_help = true;

    if (vm.count("max-mr-scale") != 0)
	use_mr_scale = true;

    if (vm.count("no-intro") != 0)
	intro_page = false;

    if (vm.count("in-file") < 1) {
	if (!show_help)
	    warning_consumer->error("No input file specified.");
	ret = false;
    }

    if (vm.count("multiple-logins") != 0)
	allow_many_logins = true;

    no_browser = vm.count("no-browser") > 0;

    if (vm.count("positional-args") > 0) {
	if (!show_help)
	    warning_consumer->error("No non-option arguments allowed");
	ret = false;
    }

    if (interactive && !ret && !show_help)
	warning_consumer->none("Use --help to get detailed parameter info.");

    if (interactive && show_help) {
	stringstream ss;
	ss << PRODUCT_VERSION << endl << COPYRIGHT << endl;
	ss << opt_desc_visible;
	warning_consumer->none(ss.str());
    }

#ifdef _WIN32
    size_t last_slash_pos = report_path.find_last_of("/\\");
#else
    size_t last_slash_pos = report_path.find_last_of("/");
#endif

    if (last_slash_pos != string::npos)
        report_dir = report_path.substr(last_slash_pos + 1);
    else
        report_dir = report_path;

    return ret;
}

END_NAMESPACE;
END_NAMESPACE;
