/// @file tools/acurate/diagram.cc

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#ifdef _WIN32
// VisualC++ math libraries don't consider M_PI to be part of standard
// libraries and require a define to include them.
#define _USE_MATH_DEFINES
#endif

#include "diagram.hh"
#include "analysis/feature_base/exceptions.hh"

#include <gd/gd.h>
// the gd precompiled small font
#include <gd/gdfonts.h>
#include <boost/shared_ptr.hpp>
#include <boost/foreach.hpp>

#include <vector>
#include <set>
#include <map>
#include <string>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <stdio.h>
#include <math.h>
#include <errno.h>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace std;
using namespace boost;
using namespace acumem::analysis;

#undef MIN
#define MIN(n, m) ((n) < (m) ? (n) : (m))

#if defined(_WIN32)
/// MSVC doesn't have round() in math.h.
///
/// NOTE: This function handles negative values incorrectly (round() should
/// round away from zero).
static inline double
round(double v)
{
    if (v - floor(v) >= 0.5)
	return ceil(v);
    else
	return floor(v);
}
#endif

ostream &
operator<<(ostream &os, const color_t &c)
{
    os << "(A:" << c.a << ", R:" << c.r
       << ", G:" << c.g << ", B:" << c.b << ")";

    return os;
}


// image_t::internal_t, a pimpl idiom inner class

struct image_t::internal_t {
    int width;
    int height;
    gdImagePtr im;
    int foreground;
    int background;
    gdFontPtr small;
    bool antialiased;
    bool styled;
};

// --- image_t ----------------------------------------------------------

image_t::image_t(int width, int height, color_t c)
    : internal(new internal_t)
{
    internal->im = gdImageCreateTrueColor(width, height);
    internal->width = width;
    internal->height = height;
    gdImageSaveAlpha(internal->im, 1);
    set_blending(false);
    internal->foreground = gdImageColorResolve(internal->im, 0, 0, 0);
    internal->background = gdImageColorResolveAlpha(internal->im, c.r, c.g, c.b, c.a/2);
    gdImageFilledRectangle(internal->im, 0, 0, internal->width, internal->height, internal->background);
    internal->small = gdFontGetSmall();
    internal->antialiased = false;
    internal->styled = false;
}

image_t::~image_t()
{
    gdImageDestroy(internal->im);
}

void
image_t::set_blending(bool on)
{
    gdImageAlphaBlending(internal->im, on ? 1 : 0);
}

void
image_t::set_antialiased(bool a)
{
    internal->antialiased = a;
}

void
image_t::save_as_png(const string &filename)
{
    FILE *pngout = fopen(filename.c_str(), "wb");
    if (pngout == 0)
	throw feature_errno_t("acurate",
			      "Creating diagram: " + filename, errno);

    gdImagePng(internal->im, pngout);
    fclose(pngout);
}

static int
data_to_ostream(void *context, const char *buffer, int len)
{
    ostream &os = *(ostream*) context;
    os.write(buffer, len);
    return len;
}

void
image_t::save_as_png_ostream(std::ostream &os)
{
    gdSink mySink;
    mySink.context = (void *) &os;
    mySink.sink = data_to_ostream;
    gdImagePngToSink(internal->im, &mySink);
}

void
image_t::line(int x1, int y1, int x2, int y2)
{
    gdImageLine(internal->im, x1, y1, x2, y2, get_foreground());
}

void
image_t::set_foreground(color_t c)
{
    internal->foreground = gdImageColorResolveAlpha(internal->im, c.r, c.g, c.b, c.a/2);
}

void
image_t::line_style(const vector<int> &colors)
{
    const int &first = *colors.begin();
    gdImageSetStyle(internal->im, const_cast<int*>(&first), (int)colors.size());
}

int
image_t::get_foreground(bool disregard_style)
{
    if (!disregard_style && internal->styled)
	return gdStyled;

    if (internal->antialiased) {
	gdImageSetAntiAliased(internal->im, internal->foreground);
	return gdAntiAliased;
    } else
	return internal->foreground;
}

void
image_t::set_style(const line_style_t &style)
{
    vector<int> s;
    int color;

    set_foreground(style.color);
    set_thickness(style.thickness);

    color = get_foreground(true);
    switch (style.type) {
    case line_style_t::Type_Dotted:
	s.push_back(color);
	s.push_back(gdTransparent);
	internal->styled = true;
	break;
    case line_style_t::Type_Dashed:
	s.push_back(color);
	s.push_back(color);
	s.push_back(color);
	s.push_back(color);
	s.push_back(gdTransparent);
	s.push_back(gdTransparent);
	s.push_back(gdTransparent);
	s.push_back(gdTransparent);
	internal->styled = true;
	break;
    case line_style_t::Type_LongDash:
	s.push_back(color);
	s.push_back(color);
	s.push_back(color);
	s.push_back(color);
	s.push_back(color);
	s.push_back(color);
	s.push_back(color);
	s.push_back(color);
	s.push_back(gdTransparent);
	s.push_back(gdTransparent);
	s.push_back(gdTransparent);
	s.push_back(gdTransparent);
	s.push_back(gdTransparent);
	s.push_back(gdTransparent);
	s.push_back(gdTransparent);
	s.push_back(gdTransparent);
	internal->styled = true;
	break;
    default:
    case line_style_t::Type_Solid:
	internal->styled = false;
	s.push_back(color);
	break;
    }

    line_style(s);
}


void
image_t::set_thickness(int thickness)
{
    gdImageSetThickness(internal->im, thickness);
}


point_t
image_t::measure_text(const string &s)
{
    point_t res;
    res.x = (int)(s.length() * internal->small->w);
    res.y = internal->small->h;
    return res;
}

void
image_t::text(const string &s, int x, int y)
{
    gdImageString(internal->im, internal->small, x, y,
		  (unsigned char*) s.c_str(), internal->foreground);
}

void
image_t::rectangle(int x1, int y1, int x2, int y2)
{
    gdImageRectangle(internal->im, x1, y1, x2, y2, get_foreground());
}

void
image_t::filled_rectangle(int x1, int y1, int x2, int y2)
{
    gdImageFilledRectangle(internal->im, x1, y1, x2, y2, get_foreground());
}

void
image_t::filled_ellipse(int centerx, int centery,
			int diameterx, int diametery)
{
    gdImageFilledEllipse(internal->im, centerx, centery, diameterx, diametery,
			 get_foreground());
}

void
image_t::filled_polygon(const point_vector_t &point_vector)
{
    gdPoint *points = new gdPoint[point_vector.size()];
    int i=0;
    BOOST_FOREACH(const point_t &p, point_vector) {
	points[i].x = p.x;
	points[i].y = p.y;
	i++;
    }

    gdImageFilledPolygon(internal->im, points, (int) point_vector.size(),
			 get_foreground());

    delete[] points;
}


void
image_t::gradient_rectangle(int  x1, int y1, int x2, int y2,
			    color_t first, color_t last)
{
    float r = (float) first.r, g = (float) first.g,
	    b = (float) first.b, a = (float) first.a;
    int zones = y2-y1-1;
    if (zones < 1)
	zones = 1;
    float dr = (float) (last.r-first.r)/zones;
    float dg = (float) (last.g-first.g)/zones;
    float db = (float) (last.b-first.b)/zones;
    float da = (float) (last.a-first.a)/zones;

    for (int y=y1; y<y2; y++) {
	int color = gdImageColorResolveAlpha(internal->im, (int)r, (int)g,
					     (int)b, (int)a/2);
	r += dr; g += dg; b += db; a += da;

	gdImageLine(internal->im, x1, y, x2, y, color);
    }
}

void
image_t::set_pixel(int x1, int y1, const color_t &color)
{
    int c = gdImageColorResolveAlpha(internal->im, color.r, color.g, color.b, color.a/2);
    gdImageSetPixel(internal->im, x1, y1, c);
}

void
image_t::line_graph(int x1, int y1, int x2, int y2,
		    const group_data_t &g,
		    const diag_data_vector_t &points,
		    int start_point,
		    int end_point,
		    int max_points)
{
    const int num_points = (int) points.size();
    gdPointPtr poly = new gdPoint[num_points];

    double yscale = (double) (y2 - y1) / (double) (g.maxy - g.miny);
    double xscale = (double) (x2 - x1) / (double) (max_points - 1);

    diag_data_vector_t::size_type end = MIN(
	MIN(num_points, max_points),
	end_point + 1);
    for (diag_data_vector_t::size_type index = start_point, num = 0;
	 index < end;
	 index++, num++) {

	poly[num].x = (int)(xscale * index) + x1;
	poly[num].y = y2 - (int)(yscale * (points[index] - g.miny));
    }
    gdImageOpenPolygon(internal->im, poly, (int)(end - start_point), get_foreground());

    delete [] poly;
}

void
image_t::text_up(const string &s, int x, int y)
{
    gdImageStringUp(internal->im, internal->small, x, y,
		    (unsigned char*) s.c_str(), internal->foreground);
}


// ----------------------------------------------------------------------

diagram_t::diagram_t(int width, int height)
    : width(width), height(height)
{
    mark_x = 0;
}

void
diagram_t::set_sweep(const rpt_sweep_t &sweep, int every_n)
{
    this->sweep = sweep;
    this->every_n = every_n;
}

void
diagram_t::add_series(const diag_series_t &series)
{
    group_map_t::iterator gi = groups.find(series.group);
    if (gi == groups.end())
	return;

    series_vector.push_back(series);
    diag_series_t &s = series_vector.back();
    diag_group_t &g = gi->second;

    BOOST_FOREACH(double &v, s.data) {
	if (g.force_max && g.max_value < v)
	    v = g.max_value;
	if (g.force_min && g.min_value > v)
	    v = g.min_value;
    }
}

void
diagram_t::add_group(int group_id, const diag_group_t &g)
{
    groups[group_id] = g;
}

void
diagram_t::add_mark_x(double mark)
{
    mark_x = mark;
}

bool
diagram_t::get_range(int group_id, double &miny, double &maxy)
{
    group_map_t::iterator gi = groups.find(group_id);
    assert(gi != groups.end());
    if (gi->second.auto_range == false) {
	miny = gi->second.min_range;
	maxy = gi->second.max_range;
	return true;
    }

    bool found_first = false;
    BOOST_FOREACH(const diag_series_t &s, series_vector) {
	if (s.group == group_id && s.data.size() != 0) {
	    if (!found_first) {
		found_first = true;
		miny = maxy = s.data[0];
	    }

	    BOOST_FOREACH(double d, s.data) {
		if (d > maxy)
		    maxy = d;
		else
		    if (d < miny) miny = d;
	    }
	}
    }
    return found_first;
}

#if !defined(__linux)
static double pow10(double d)
{
    return pow(10.0, d);
}
#endif

void
diagram_t::find_good_range(const diag_group_t &g, double &miny,
			   double &maxy, double &scale)
{

    if (maxy>g.max_value) maxy = g.max_value;
    if (miny<g.min_value) miny = g.min_value;
    if (g.force_zero && miny > 0) miny = 0;
    if (maxy == 0) maxy = g.max_value; // to avoid log(0)

    double s = log10(maxy);
    double ticks = floor(s); // This is the base for the scale
    scale = pow10(ticks);

    double f = s - ticks;  // just the fraction of the log(max)

    // The new maxy is larger than old maxy, add some extra range to
    // the nearest 1, 1.2, 1.5, 2, 5 *10^?
    // Also adjust the scale so that 3-5 labels are visible.
    double extra;
    if (f < log10(1.2)) {
	extra = log10(1.2) - f;
	scale = scale / 2;
    } else if (f < log10(1.5)) {
	extra = log10(1.5) - f;
	scale = scale / 2;
    } else if (f < log10(3.0)) {
	extra = log10(3.0) - f;
	scale = scale / 2;
    } else if (f < log10(6.0)) {
	extra = log10(6.0) - f;
    } else {
	extra = log10(12.0) - f;
	scale = scale * 5;
    }

    maxy = pow10(s + extra);
}

void diagram_t::value_to_string(double x, char *buf)
{
    if (x < 1024.0)
	sprintf(buf, "%0.0lf", x);
    else if (x < 1024.0*1024.0)
	sprintf(buf, "%0.0lfk", x/1024.0);
    else if (x < 1024.0*1024.0*1024.0)
	sprintf(buf, "%0.0lfM", x/1024.0/1024.0);
    else
	sprintf(buf, "%0.0lfG", x/1024.0/1024.0/1024.0);
}


void
diagram_t::xgrid(image_t &image, int x1, int y1, int x2, int y2,
		 int divx, int every_n)
{
    double scale = (double) (x2 - x1) / (double) (divx - 1);
    for (int xi = 0; xi < divx; xi += every_n) {
	int x = (int) (scale * xi) + x1;
	image.line(x, y1, x, y2);
    }
}

void
diagram_t::xscale(image_t &image, int x1, int y1, int x2, int y2,
		  int divx, int every_n)
{
    double scale = (double) (x2 - x1) / (double) (divx - 1);

    for (int xi = 0; xi < divx; xi += every_n) {
	int x = (int) (scale * xi) + x1;

	char buf[10];
	value_to_string((double)sweep.at(xi), buf);
	image.text_up(buf, x-6, y2);
    }
}

void
diagram_t::ygrid(image_t &image, int x1, int y1, int x2, int y2,
		 const group_data_t &g)
{
    double scale = (double) (y2 - y1) / (g.maxy - g.miny);
    for(double y_value = g.miny; y_value <= g.maxy; y_value += g.scale) {
	int y = y2 - (int) (scale * y_value);
	image.line(x1, y, x2, y);
    }
}

void
diagram_t::yscale(image_t &image, int x1, int y1, int x2, int y2,
		  const group_data_t &g)
{
    // Simplifying assumptions.
    assert(g.miny == 0);
    assert(g.percent);

    // Multiply with 0.999 as a work-around for floating point rounding
    // problems.
    int scale_log10 = (int)floor(log10(g.scale) * 0.999);
    int decimals = max(0, -scale_log10 - (g.percent ? 2 : 0));

    double scale = (double) (y2-y1)/(g.maxy - g.miny);
    for(double y_value = g.miny; y_value <= g.maxy; y_value += g.scale) {
	int y = y2-(int) (scale*y_value);
	stringstream ss;
        ss << fixed << setprecision(decimals) << y_value * 100.0 << "%";

        string s = ss.str();
        while (s.length() < 6)
            s = " " + s;

        if (s.length() > 6) {
            if (y_value == g.miny)
                image.text("0.000%", (int)g.label_x, y -  g.label_height / 2);
            if (y_value + g.scale > g.maxy)
                image.text(s, (int)g.label_x, 1);
        }
        else
            image.text(s, (int)g.label_x, y -  g.label_height / 2);
    }
}

void
diagram_t::draw_mark(image_t &image, int x1, int y1, int x2, int y2,
		     int divx)
{
    double lower_log = log((double)sweep.front());
    double upper_log = log((double)sweep.back());
    double mark_log = log((double)mark_x);

    if (mark_log < lower_log || mark_log > upper_log)
        return;

    int x = (int)((mark_log - lower_log) / (upper_log - lower_log) * (x2 - x1) + x1);

    image.line(x, y1, x, y2);
}

image_ptr_t
diagram_t::draw()
{
    color_t white(255, 255, 255, 0);
    color_t black(0, 0, 0);
    color_t gray80(50, 50, 50, 0);
    color_t gray50(127, 127, 127, 0);
    color_t gray20(205, 205, 205, 0);
    color_t transp(255, 255, 255, 255);
    color_t ful(203, 225, 203, 0); // light green

    // Ugly hack. We "secretly" add some padding to the bottom of the image for
    // the legends.
    image_ptr_t dummy_image(new image_t(1, 1, white));
    int legend_row_height = 0;
    BOOST_FOREACH(const diag_series_t &series, series_vector) {
        int height = dummy_image->measure_text(series.name).y;
        if (height > legend_row_height)
            legend_row_height = height;
    }
    int legend_base_pad = 5;
    int legend_pad = (int)(legend_base_pad + legend_row_height * series_vector.size());

    // Create the image.
    image_ptr_t res (new image_t(width, height + legend_pad, white));
    image_t &image = *res.get();

    typedef map<int, group_data_t> group_data_map_t;
    group_data_map_t group_data_map;
    int left_margin = 0;
    int top_margin = 0;
    int right_margin = 5;

    // find the bottom margin
    int bottom_margin = 0;
    BOOST_FOREACH(rpt_size_t cache_size, sweep) {
	char buf[10];
	value_to_string((double)cache_size, buf);
	point_t x_label = image.measure_text(buf);
	bottom_margin = max(bottom_margin, x_label.x + 3);
    }

    group_map_t::iterator si = groups.begin(), se = groups.end();
    for (int groupcount = 0; si != se; groupcount++, si++) {
	group_data_t g;
	// Find out the max and min in the data set for the current group
	get_range(si->first, g.miny, g.maxy);

	// Now, rescale these to get a vertical range that is fixed to
	// a few different maximum numbers for the mantissa, and
	// adjust the scale accordingly.
	find_good_range(si->second, g.miny, g.maxy, g.scale);

        // The small font is a fixed-width font. Make sure the left margin is
        // wide enough for 6 characters.
	point_t p = image.measure_text("000000");

	g.label_width = p.x;
	g.label_height = p.y;
	top_margin = max(top_margin, p.y + 2);
	bottom_margin = max(bottom_margin, p.y/2);

	g.percent = si->second.percent;

	if (groups.size() != 2 || groupcount == 0) {
	    g.label_x = left_margin;
	    left_margin = left_margin + p.x + 4;
	} else {
	    g.label_x = width - right_margin - p.x;
	    right_margin = right_margin + p.x + 4;
	}

	g.grid_color = si->second.grid_color;
	group_data_map[si->first] = g;
    }

    color_t tmp(0xf0, 0xf0, 0xff);
    image.gradient_rectangle(left_margin, top_margin,
			     width-1-right_margin, height-1-bottom_margin,
			     tmp, tmp);
    image.set_foreground(black);
    image.rectangle(left_margin, top_margin,
		    width-1-right_margin, height-bottom_margin-1);


    int num_points = (int)sweep.size();

    image.set_foreground(gray20);
    xgrid(image, left_margin, top_margin,
	  width-1-right_margin, height-1-bottom_margin, num_points, every_n);

    image.set_foreground(black);
    xscale(image, left_margin, top_margin,
	   width-1-right_margin, height-1, num_points, every_n);

    // grid y
    BOOST_FOREACH(const group_data_map_t::value_type &v, group_data_map) {
	const group_data_t &g(v.second);

	// use a y-grid color that is somewhere between the series
	// color and white.
	color_t colgrid(g.grid_color);
	colgrid.r = (colgrid.r + 2*white.r)/3;
	colgrid.g = (colgrid.g + 2*white.g)/3;
	colgrid.b = (colgrid.b + 2*white.b)/3;
	colgrid.a = (colgrid.a + 2*white.a)/3;
	image.set_foreground(colgrid);
	ygrid(image, left_margin, top_margin,
	      width-1-right_margin, height-1-bottom_margin, g);

	image.set_foreground(g.grid_color);
	yscale(image, left_margin, top_margin,
	       width-1-right_margin, height-1-bottom_margin, g);
    }

    int legend_row = 0;

    image.set_antialiased(true);
    BOOST_FOREACH(const group_data_map_t::value_type &v, group_data_map) {
	const group_data_t &g(v.second);
	const int &group_id(v.first);

	BOOST_FOREACH(const diag_series_t &series, series_vector) {
	    if (series.group == group_id && series.data.size() != 0) {
		image.set_style(series.style);

                // Draw legend.
                int legend_y = height + legend_base_pad + legend_row++ * legend_row_height;
                int legend_line_y = legend_y + legend_row_height / 2;
                image.line(10, legend_line_y, 40, legend_line_y);
                image.text(series.name, 50, legend_y + 1);

		// First one line from index 0 to i->dash_from_index (if !-1)
		int max_points1 = (int)series.data.size();
		if (series.dash_from_index != -1) {
		    max_points1 = min(max_points1, series.dash_from_index);
		}

		image.line_graph(left_margin, top_margin,
				 width - 1 - right_margin,
				 height - 1 - bottom_margin,
				 g, series.data, 0, max_points1, num_points);

		// Then, potentially another dashed line from index,
		// continuing at the value of the last entry.
		if (series.dash_from_index != -1 &&
		    series.dash_from_index < (int)series.data.size()) {
		    // Setup the new style
		    line_style_t dash_style(series.style);
		    dash_style.type = line_style_t::Type_LongDash;
		    image.set_style(dash_style);

		    // Draw a dashed line from the last valid point
		    diag_data_vector_t d;
		    for (unsigned int xxx=0; xxx < series.data.size(); xxx++)
			d.push_back(series.data[series.dash_from_index]);

		    image.line_graph(left_margin, top_margin,
				     width - 1 - right_margin,
				     height - 1 - bottom_margin,
				     g, d, max_points1, (int)series.data.size(),
				     num_points);
		}
	    }
	}
    }
    image.set_antialiased(false);

    image.set_style(line_style_t(
			color_t(black),
			1,
			line_style_t::Type_Solid));
    image.rectangle(left_margin, top_margin,
		    width - 1 - right_margin,
		    height - 1 - bottom_margin);

    draw_mark(image, left_margin, top_margin,
	      width - 1 - right_margin,
	      height - 1 - bottom_margin, num_points);

    return res;
}

// ----------------------------------------------------------------------


#define GAUGE_SIZE_X 121
#define GAUGE_SIZE_Y 83

#define GAUGE_DIAMETER 120
#define GAUGE_CENTER_X 60
#define GAUGE_CENTER_Y 60

#define SCALE_OUTER_RADIUS 53
#define SCALE_INNER_RADIUS 40
#define SCALE_RADIANS 3.0
#define SCALE_START_ANGLE (M_PI / 2 - SCALE_RADIANS / 2)
#define SCALE_END_ANGLE (M_PI / 2 + SCALE_RADIANS / 2)

#define NEEDLE_CENTER_DIAMETER 24
#define NEEDLE_FRONT_LENGTH 55
#define NEEDLE_FRONT_RADIANS 0.04
#define NEEDLE_BACK_LENGTH 21
#define NEEDLE_BACK_RADIANS 5.8

color_t
gauge_t::mix_colors(const color_t &left, const color_t &right, double fraction0)
{
    double f0 = fraction0;
    double f1 = 1.0 - fraction0;


    return color_t((int)(left.r*f0 + right.r*f1),
		   (int)(left.g*f0 + right.g*f1),
		   (int)(left.b*f0 + right.b*f1),
		   (int)(left.a*f0 + right.a*f1));
}

color_t
gauge_t::scale_color(double fraction)
{
    color_t red(0xff, 0, 0);
    color_t yellow(0xff,0xff, 0);
    color_t green(0, 0xe0, 0);
    if (fraction < 0.5)
	return mix_colors(yellow, green, pow(fraction, 0.33) / pow(0.5, 0.33));
    else
	return mix_colors(red, yellow, pow(fraction - 0.5, 2) / pow(0.5, 2));
}


image_ptr_t
gauge_t::draw()
{
    color_t transparent(255, 255, 255, 255);
    color_t gray(0xe0, 0xe0, 0xe0);

    image_ptr_t res(new image_t(GAUGE_SIZE_X, GAUGE_SIZE_Y, transparent));
    image_t &image = *(res.get());

    // Draw the gauge background.
    image.set_foreground(gray);
    image.filled_ellipse(GAUGE_CENTER_X, GAUGE_CENTER_Y,
			 GAUGE_DIAMETER, GAUGE_DIAMETER);

    // Draw the scale. Too inefficient?
    for (int x = 0; x < GAUGE_SIZE_X; x++)
	for (int y = 0; y < GAUGE_SIZE_Y; y++) {
	    double rx = x - GAUGE_CENTER_X;
	    double ry = y - GAUGE_CENTER_Y;
	    double radius = sqrt(rx * rx + ry * ry);
	    double angle = acos((double)rx / radius);
	    if (radius <= SCALE_OUTER_RADIUS &&
		radius >= SCALE_INNER_RADIUS &&
		ry < 0 &&
		angle >= SCALE_START_ANGLE  &&
		angle <= SCALE_END_ANGLE)
		image.set_pixel(x, y, scale_color((SCALE_END_ANGLE - angle) / SCALE_RADIANS));
	}

    color_t needle_color(0x50, 0x50, 0x50);
    image.set_foreground(needle_color);
    image.set_antialiased(true);

    double needle_point_angle = SCALE_END_ANGLE - SCALE_RADIANS * needle_position;
    double needle_back_start_angle = needle_point_angle + NEEDLE_BACK_RADIANS / 2;
    double needle_back_end_angle = needle_point_angle - NEEDLE_BACK_RADIANS / 2;
    double needle_front_start_angle = needle_point_angle + NEEDLE_FRONT_RADIANS / 2;
    double needle_front_end_angle = needle_point_angle - NEEDLE_FRONT_RADIANS / 2;
    point_vector_t points;

    points.push_back(
	point_t((int)round(GAUGE_CENTER_X + NEEDLE_BACK_LENGTH * cos(needle_back_start_angle)),
		(int)round(GAUGE_CENTER_Y - NEEDLE_BACK_LENGTH * sin(needle_back_start_angle))));

    points.push_back(
	point_t((int)round(GAUGE_CENTER_X + NEEDLE_FRONT_LENGTH * cos(needle_front_start_angle)),
		(int)round(GAUGE_CENTER_Y - NEEDLE_FRONT_LENGTH * sin(needle_front_start_angle))));

    points.push_back(
	point_t((int)round(GAUGE_CENTER_X + NEEDLE_FRONT_LENGTH * cos(needle_front_end_angle)),
		(int)round(GAUGE_CENTER_Y - NEEDLE_FRONT_LENGTH * sin(needle_front_end_angle))));

    points.push_back(
	point_t((int)round(GAUGE_CENTER_X + NEEDLE_BACK_LENGTH * cos(needle_back_end_angle)),
		(int)round(GAUGE_CENTER_Y - NEEDLE_BACK_LENGTH * sin(needle_back_end_angle))));


    image.filled_polygon(points);
    image.filled_ellipse(GAUGE_CENTER_X, GAUGE_CENTER_Y,
		       NEEDLE_CENTER_DIAMETER, NEEDLE_CENTER_DIAMETER);

    return res;
}

END_NAMESPACE;
END_NAMESPACE;
