/// @file tools/acurate/diagram.hh

// Copyright (c) 2007-2009 Acumem AB.  All rights reserved.

#ifndef INCLUDED_DIAGRAM_HH
#define INCLUDED_DIAGRAM_HH

#include "analysis/report/report.hh"

#include <boost/shared_ptr.hpp>

#include <map>
#include <vector>
#include <string>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace acumem::analysis;
using namespace std;

// ----------------------------------------------------------------------
// Colors and styles
// ----------------------------------------------------------------------

struct color_t {
    color_t()
	: a(0), r(0), g(0), b(0) {}
    color_t(int r, int g, int b)
	: a(0), r(r), g(g), b(b) {}
    color_t(int r, int g, int b, int a)
	: a(a), r(r), g(g), b(b) {}
    int a;
    int r;
    int g;
    int b;

    color_t operator / (double f) {
	return color_t((int) (r/f), (int)(g/f), (int) (b/f), a);
    }
};

struct line_style_t {
    enum type_t {
	Type_Solid,
	Type_Dotted,
	Type_Dashed,
	Type_LongDash,
    };

    line_style_t()
	: color(), thickness(1), type(Type_Solid) {}

    line_style_t(const color_t &color,
		 const int thickness,
		 const type_t type)
	: color(color), thickness(thickness), type(type) {}

    color_t color;
    int thickness;
    type_t type;
};

// ----------------------------------------------------------------------
// Diagram series
// ----------------------------------------------------------------------

typedef vector<double> diag_data_vector_t;
typedef vector<bool> diag_valid_vector_t;

struct diag_series_data_t {
    void add(rpt_stat_fp_t value) {
        data.push_back(value);
    }

    void add(const rpt_stat_fp_valid_t &value_valid) {
        data.push_back(value_valid.get_value());
        valid.push_back(value_valid.get_valid());
    }

    diag_data_vector_t data;
    diag_valid_vector_t valid;
};

struct diag_series_t {
    diag_series_t(const std::string &name,
		  int group,
		  const diag_series_data_t &series,
		  line_style_t style)
	: name(name),
          group(group),
          data(series.data),
          style(style),
          dash_from_index(-1) {
        for (diag_valid_vector_t::size_type i = 0; i < series.valid.size(); i++)
            if (!series.valid[i]) {
                dash_from_index = (int)(i > 0 ? i - 1 : 0);
                break;
            }
    }

    std::string name;
    int group;
    diag_data_vector_t data;
    line_style_t style;
    int dash_from_index;
};

typedef std::vector<diag_series_t> diag_series_vector_t;

struct diag_group_t {
    diag_group_t()
	: max_value(0),
	  min_value(0),
	  percent(true),
	  force_zero(true),
	  force_max(true),
	  force_min(true),
	  auto_range(true),
	  min_range(0),
	  max_range(1),
	  grid_color()
	{}

    double max_value;
    double min_value;
    bool percent;
    bool force_zero;
    bool force_max;
    bool force_min;

    bool auto_range;
    double min_range;
    double max_range;

    color_t grid_color;
};

struct group_data_t {
    double miny;
    double maxy;
    double scale;
    double label_x;
    int label_width;
    int label_height;
    bool percent;

    color_t grid_color;
};


// ----------------------------------------------------------------------
// Drawing
// ----------------------------------------------------------------------

struct point_t {
    point_t(int x=0, int y=0) : x(x), y(y) {}
    point_t(const point_t &other) : x(other.x), y(other.y) {}

    const point_t & operator=(const point_t &other) {
	x = other.x;
	y = other.y;
	return *this;
    }

    int x;
    int y;
};

typedef std::vector<point_t> point_vector_t;

class image_t {
public:
    image_t(int width, int height, color_t c);

    ~image_t();

    void set_blending(bool on);

    void save_as_png(const std::string &filename);
    void save_as_png_ostream(std::ostream &os);

    void line(int x1, int y1, int x2, int y2);

    void set_foreground(color_t c);

    void set_style(const line_style_t &style);

    void line_style(const std::vector<int> &colors);

    point_t measure_text(const std::string &s);
    void text(const std::string &s, int x, int y);
    void text_up(const std::string &s, int x, int y);

    void rectangle(int x1, int y1, int x2, int y2);

    void filled_rectangle(int x1, int y1, int x2, int y2);

    void filled_ellipse(int centerx, int centery,
			int diameterx, int diametery);


    void gradient_rectangle(int  x1, int y1, int x2, int y2,
			    color_t first, color_t last);

    void set_pixel(int x, int y, const color_t &color);

    void line_graph(int x1, int y1, int x2, int y2,
		    const group_data_t &g,
		    const diag_data_vector_t &points,
		    int start_point,
		    int end_point,
		    int num_points);

    void filled_polygon(const point_vector_t &point_vector);

    void set_antialiased(bool a);
    void set_thickness(int thickness);

private:

    int get_foreground(bool disregard_style = false);

    struct internal_t;
    boost::shared_ptr<internal_t> internal;
};

typedef boost::shared_ptr<image_t> image_ptr_t;

class diagram_t {
public:
    diagram_t(int width, int height);

    void set_sweep(const rpt_sweep_t &sweep, int every_n);
    void add_series(const diag_series_t &series);
    void add_group(int group_id, const diag_group_t &g);
    void add_mark_x(double x_value);

    image_ptr_t draw();

private:

    bool get_range(int group_id, double &miny, double &maxy);
    void find_good_range(const diag_group_t &g, double &miny,
			 double &maxy, double &scale);

    void value_to_string(double x, char *buf);


    void xgrid(image_t &image, int x1, int y1, int x2, int y2,
	       int divx, int every_n);

    void xscale(image_t &image, int x1, int y1, int x2, int y2,
		int divx, int every_n);

    void ygrid(image_t &image, int x1, int y1, int x2, int y2,
	       const group_data_t &g);

    void yscale(image_t &image, int x1, int y1, int x2, int y2,
		const group_data_t &g);

    void draw_mark(image_t &image, int x1, int y1, int x2, int y2, int divx);

    int width;
    int height;
    rpt_sweep_t sweep;
    diag_series_vector_t series_vector;

    typedef std::map<int, diag_group_t> group_map_t;
    group_map_t groups;
    int every_n;
    double mark_x;
};

class gauge_t
{
public:
    gauge_t(double needle_position)
	: needle_position(bracket(needle_position, 0.0, 1.0))
	{}

    image_ptr_t draw();

private:
    static double bracket(double x, double l, double r) {
	if (x < l) return l;
	else if (x > r) return r;
	else return x;
    }

    color_t mix_colors(const color_t &left, const color_t &right, double fraction0);
    color_t scale_color(double fraction);
    double needle_position;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
