/// @file tools/acurate/drill_down.cc

// Copyright (c) 2007-2009 Acumem AB.  All rights reserved.

#include "drill_down.hh"
#include "diagram.hh"
#include "csv.hh"

#include <fstream>

#include <boost/foreach.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace std;
using namespace boost;
using namespace acumem::analysis;

static line_style_t style_fetch_ratio(
    color_t(0xE0, 0x00, 0x00),
    2,
    line_style_t::Type_Solid);

static line_style_t style_util_corr_fr(
    color_t(0xE0, 0x00, 0x00),
    2,
    line_style_t::Type_Dashed);

static line_style_t style_miss_ratio(
    color_t(0x80, 0x00, 0x00),
    2,
    line_style_t::Type_Solid);


static line_style_t style_write_back_ratio(
    color_t(0x00, 0x00, 0x00),
    2,
    line_style_t::Type_Solid);

static line_style_t style_util_corr_wbr(
    color_t(0x00, 0x00, 0x00),
    2,
    line_style_t::Type_Dashed);


static line_style_t style_spatial_util(
    color_t(0x00, 0x00, 0xCC),
    2,
    line_style_t::Type_Solid);

static line_style_t style_fetch_util(
    color_t(0xFF, 0x40, 0x00),
    2,
    line_style_t::Type_Solid);

static line_style_t style_write_back_util(
    color_t(0x00, 0x60, 0x00),
    2,
    line_style_t::Type_Solid);

static color_t grid_color(0, 0, 0);

// ----- Statistics collection function. --------------------------------------

enum series_kind_t {
    Series_Fetches,
    Series_Misses,
    Series_Fetch_Ratio,
    Series_Miss_Ratio,
    Series_Write_Back_Ratio,
    Series_Spatial_Util,
    Series_Fetch_Util,
    Series_Write_Back_Util,
    Series_Util_Corr_FR,
    Series_Util_Corr_WBR,
    Series_Avoidable_Fetches
};

static diag_series_data_t
get_series(
    const rpt_report_cptr_t &report,
    rpt_stats_sweep_id_t stats_sweep_id,
    series_kind_t series_kind)
{
    diag_series_data_t series;

    const rpt_stats_sweep_t &stats_sweep =
        report->stats_sweeps.at(stats_sweep_id);

    BOOST_FOREACH(rpt_sliced_stats_id_t sliced_stats_id,
                  stats_sweep.sliced_stats_ids) {
        const rpt_stats_t &stats(
            report->get_sliced_stats_global(sliced_stats_id));
        switch (series_kind) {
        case Series_Fetches:
            series.add((rpt_stat_fp_t)stats.get_total_fetches());
            break;
        case Series_Misses:
            series.add((rpt_stat_fp_t)stats.get_total_misses());
            break;
        case Series_Fetch_Ratio:
            series.add((rpt_stat_fp_t)stats.get_total_fetch_ratio());
            break;
        case Series_Miss_Ratio:
            series.add((rpt_stat_fp_t)stats.get_total_miss_ratio());
            break;
        case Series_Write_Back_Ratio:
            series.add((rpt_stat_fp_t)stats.get_total_write_back_ratio());
            break;
        case Series_Spatial_Util:
            series.add(stats.get_spatial_utilization());
            break;
        case Series_Fetch_Util:
            series.add(stats.get_fetch_utilization());
            break;
        case Series_Write_Back_Util:
            series.add(stats.get_write_back_utilization());
            break;
        case Series_Util_Corr_FR:
            series.add((rpt_stat_fp_t)stats.get_util_corr_fetch_ratio());
            break;
        case Series_Util_Corr_WBR:
            series.add((rpt_stat_fp_t)stats.get_util_corr_write_back_ratio());
            break;
        case Series_Avoidable_Fetches:
            series.add((rpt_stat_fp_t)stats.get_avoidable_fetches());
            break;
        default:
            assert(0);
        }
    }

    return series;
}

// ----- drill_down_t ---------------------------------------------------------

drill_down_t::drill_down_t(
    const rpt_report_cptr_t &report,
    rpt_stats_sweep_id_t stats_sweep_id)
    : report(report),
      stats_sweep_id(stats_sweep_id)
{
}

drill_down_t::~drill_down_t()
{
}

void
drill_down_t::setup_diagram(diagram_t &diagram)
{
    diagram.set_sweep(report->sweeps[0], 4);
    diagram.add_mark_x((double)report->cache_size);
}

void
drill_down_t::add_group(diagram_t &d, int group_id, const color_t &grid_color)
{
    diag_group_t g;
    g.max_value = 1;
    g.min_value = 0;
    g.percent = true;
    g.force_zero = true;
    g.force_min = true;
    g.force_max = true;
    g.auto_range = true;
    g.min_range = 0;
    g.max_range = 1;
    g.grid_color = grid_color;

    d.add_group(group_id, g);
}


void
drill_down_t::add_group(diagram_t &d, int group_id, const color_t &grid_color,
			double max_mr_scale)
{
    diag_group_t g2;
    g2.max_value = 1;
    g2.min_value = 0;
    g2.percent = true;
    g2.force_zero = true;
    g2.force_min = true;
    g2.force_max = true;
    g2.auto_range = false;
    g2.min_range = 0;
    g2.max_range = max_mr_scale;
    g2.grid_color = grid_color;

    d.add_group(group_id, g2);
}

void
drill_down_t::add_fetch_ratio(diagram_t &diagram, int group_id)
{
    diagram.add_series(
        diag_series_t(
            "Fetch ratio", group_id,
            get_series(report, stats_sweep_id,  Series_Fetch_Ratio),
            style_fetch_ratio));
}

void
drill_down_t::add_util_corr_fr(diagram_t &diagram, int group_id)
{
    diagram.add_series(
	diag_series_t(
            "Utilization corrected fetch ratio", group_id,
            get_series(report, stats_sweep_id, Series_Util_Corr_FR),
            style_util_corr_fr));
}

void
drill_down_t::add_miss_ratio(diagram_t &diagram, int group_id)
{
    diagram.add_series(
	diag_series_t(
            "Miss ratio", group_id,
            get_series(report, stats_sweep_id, Series_Miss_Ratio),
            style_miss_ratio));
}

void
drill_down_t::add_write_back_ratio(diagram_t &diagram, int group_id)
{
    diagram.add_series(
	diag_series_t(
            "Write-back ratio", group_id,
            get_series(report, stats_sweep_id, Series_Write_Back_Ratio),
            style_write_back_ratio));
}

void
drill_down_t::add_util_corr_wbr(diagram_t &diagram, int group_id)
{
    diagram.add_series(
	diag_series_t(
            "Utilization corrected write-back ratio", group_id,
            get_series(report, stats_sweep_id, Series_Util_Corr_WBR),
            style_util_corr_wbr));
}

void
drill_down_t::add_fetch_util(diagram_t &diagram, int group_id)
{
    diagram.add_series(
	diag_series_t(
            "Fetch utilization", group_id,
            get_series(report, stats_sweep_id, Series_Fetch_Util),
            style_fetch_util));
}

void
drill_down_t::add_write_back_util(diagram_t &diagram, int group_id)
{
    diagram.add_series(
	diag_series_t(
            "Write-back utilization", group_id,
            get_series(report, stats_sweep_id, Series_Write_Back_Util),
            style_write_back_util));
}

void
drill_down_t::add_spatial_util(diagram_t &diagram, int group_id)
{
    diagram.add_series(
	diag_series_t(
            "Cache line utilization", group_id,
            get_series(report, stats_sweep_id, Series_Spatial_Util),
            style_spatial_util));
}

void
drill_down_t::prepare_fr(ostream &os)
{
    diagram_t diagram(diagram_width(), diagram_height());

    setup_diagram(diagram);

    add_group(diagram, 1, grid_color);

    add_fetch_ratio(diagram, 1);
    add_util_corr_fr(diagram, 1);
    add_miss_ratio(diagram, 1);

    diagram.draw()->save_as_png_ostream(os);
}

void
drill_down_t::prepare_fr(ostream &os,
			 double max_scale)
{
    diagram_t diagram(diagram_width(), diagram_height());

    setup_diagram(diagram);

    add_group(diagram, 1, grid_color, max_scale);

    add_fetch_ratio(diagram, 1);
    add_util_corr_fr(diagram, 1);
    add_miss_ratio(diagram, 1);

    diagram.draw()->save_as_png_ostream(os);
}

void
drill_down_t::prepare_wb(ostream &os)
{
    diagram_t diagram(diagram_width(), diagram_height());

    setup_diagram(diagram);

    add_group(diagram, 1, grid_color);
    add_write_back_ratio(diagram, 1);
    add_util_corr_wbr(diagram, 1);

    diagram.draw()->save_as_png_ostream(os);
}

void
drill_down_t::prepare_wb(ostream &os,
                         double max_scale)
{
    diagram_t diagram(diagram_width(), diagram_height());

    setup_diagram(diagram);

    add_group(diagram, 1, grid_color, max_scale);
    add_write_back_ratio(diagram, 1);
    add_util_corr_wbr(diagram, 1);

    diagram.draw()->save_as_png_ostream(os);
}

void
drill_down_t::prepare_su(ostream &os)
{
    diagram_t diagram(diagram_width(), diagram_height());

    setup_diagram(diagram);

    add_group(diagram, 1, grid_color);

    if (report->mt_sampled) {
	add_fetch_util(diagram, 1);
	add_write_back_util(diagram, 1);
    } else
	add_spatial_util(diagram, 1);

    diagram.draw()->save_as_png_ostream(os);
}

// ----- drill_down_global_t --------------------------------------------------

drill_down_global_t::drill_down_global_t(
    const rpt_report_cptr_t &report,
    const string &file_base,
    double max_mr_scale)
    : drill_down_t(report, report->global_stats_sweep),
      file_base(file_base),
      max_mr_scale(max_mr_scale)
{
}

void
drill_down_global_t::write_files()
{
    {
	ofstream fr((file_base + "-fr.png").c_str(), ios::binary);
	prepare_fr(fr);
    }

    {
	ofstream fr((file_base + "-fr-100.png").c_str(), ios::binary);
	prepare_fr(fr, max_mr_scale);
    }

    {
	ofstream su((file_base + "-su.png").c_str(), ios::binary);
	prepare_su(su);
    }

    if (report->mt_sampled) {
	{
	    ofstream wb((file_base + "-wb.png").c_str(), ios::binary);
	    prepare_wb(wb);
	}

	{
	    ofstream wb((file_base + "-wb-100.png").c_str(), ios::binary);
	    prepare_wb(wb, max_mr_scale);
	}
    }

    report_csv();
}

void
drill_down_global_t::report_csv()
{
    const rpt_stats_sweep_t &stats_sweep =
        report->stats_sweeps.at(stats_sweep_id);

    csv_t csv(report, stats_sweep.sweep_id, "Cache", 0, ',');

    csv.add_series(
        "Fetch ratio",
        get_series(report, stats_sweep_id, Series_Fetch_Ratio).data);
    csv.add_series(
        "Fetches",
        get_series(report, stats_sweep_id, Series_Fetches).data);
    csv.add_series(
        "Miss ratio",
        get_series(report, stats_sweep_id, Series_Miss_Ratio).data);
    csv.add_series(
        "Misses",
        get_series(report, stats_sweep_id, Series_Misses).data);
    csv.add_series(
        "Spatial usage",
        get_series(report, stats_sweep_id, Series_Spatial_Util).data);
    csv.add_series(
        "Utilization corrected fetch ratio",
        get_series(report, stats_sweep_id, Series_Util_Corr_FR).data);
    csv.add_series(
        "Avoidable fetches",
        get_series(report, stats_sweep_id, Series_Avoidable_Fetches).data);

    csv.write_csv_file(file_base + ".csv");
}

// ----- drill_down_pc_set_t --------------------------------------------------

drill_down_pc_set_t::drill_down_pc_set_t(
    const rpt_report_cptr_t &report,
    const string &id,
    rpt_location_set_id_t location_set_id)
    : drill_down_t(report, report->location_set_stats_sweeps[location_set_id]),
      id(id)
{
}

void
drill_down_pc_set_t::write_files()
{
    {
	ofstream fr((id + "-fr.png").c_str(), ios::binary);
	prepare_fr(fr);
    }

    {
	ofstream su((id + "-su.png").c_str(), ios::binary);
	prepare_su(su);
    }

    if (report->mt_sampled) {
	{
	    ofstream wb((id + "-wb.png").c_str(), ios::binary);
	    prepare_wb(wb);
	}
    }
}

END_NAMESPACE;
END_NAMESPACE;
