/// @file tools/acurate/file_repos.cc

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#include "tools/acurate/file_repos.hh"
#include "library/lang/am_assert.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

// ----- fr_location_info_t ---------------------------------------------------

void
fr_location_info_t::add_reporter(location_reporter_ptr_t reporter)
{
    pc_info_ptr_t pi = pc_info.lock();
    line_info_ptr_t li = pi->line_info.lock();
    if (li.get()) {
	location_reporter_vector.push_back(reporter);
	li->num_reporters++;
    }
}

bool
fr_location_info_t::force() const
{
    BOOST_FOREACH(const location_reporter_ptr_t &r, location_reporter_vector) {
	if (r->force())
	    return true;
    }

    return false;
}

// ----- file_repos_t ---------------------------------------------------------

file_repos_t::file_repos_t(const rpt_report_t &report)
{
    // Create source line info for all known source lines.
    BOOST_FOREACH(const rpt_source_line_t &source_line, report.source_lines) {
        const rpt_source_file_t &source_file(
            report.source_files[source_line.source_file_id]);
        line_info_ptr_t li(
            new line_info_t(source_file.filename, source_line.line));
        file_map[li->filename][li->line] = li;
    }

    // Create pc info for all known pcs.
    for (rpt_debug_pc_id_t i = 0; i < report.debug_pcs.size(); i++) {
        const rpt_debug_pc_t &debug_pc(report.debug_pcs[i]);

        line_info_ptr_t line_info;
        if (debug_pc.source_line_id != RPT_NO_SOURCE_LINE) {
            const rpt_source_line_t &source_line(
                report.source_lines.at(debug_pc.source_line_id));
            const rpt_source_file_t &source_file(
                report.source_files.at(source_line.source_file_id));
            file_map_t::const_iterator fmi(file_map.find(source_file.filename));
            AM_ASSERT(fmi != file_map.end());
            const line_info_map_t &line_map(fmi->second);
            line_info_map_t::const_iterator lmi(line_map.find(source_line.line));
            AM_ASSERT(lmi != line_map.end());
            line_info = lmi->second;
        }

        rpt_address_t pc(report.get_address(debug_pc.pc_id));
        pc_info_ptr_t pc_info(new pc_info_t(pc, i, line_info));
        pc_info_index_t pc_info_index(pc, debug_pc.return_address);
        pc_info_map.insert(make_pair(pc_info_index, pc_info));

        if (line_info)
            line_info->pc_info_map.insert(make_pair(pc_info_index, pc_info));
    }

    // Create location info for all known locations with corresponding pc info.
    for (rpt_location_id_t i = 0; i < report.locations.size(); i++) {
        rpt_address_id_t pc_id(report.get_location_pc_id(i));
        rpt_address_t pc(report.get_address(pc_id));
        pc_info_map_t::const_iterator pcii(
            pc_info_map.find(pc_info_index_t(pc, false)));
        if  (pcii != pc_info_map.end()) {
            pc_info_ptr_t pc_info(pcii->second);
            fr_location_info_ptr_t location_info(
                new fr_location_info_t(i, pc_info));
            pc_info->location_info_map.insert(
                make_pair(i, location_info));
        }
    }
}

pc_info_ptr_t
file_repos_t::get_pc_info(
    const rpt_report_t &report,
    rpt_address_t pc,
    bool return_address) const
{
    pc_info_index_t pc_info_index(pc, return_address);
    pc_info_map_t::const_iterator pcii = pc_info_map.find(pc_info_index);
    AM_ASSERT(pcii != pc_info_map.end());
    return pcii->second;
}

fr_location_info_ptr_t
file_repos_t::get_line_info(
    const rpt_report_t &report,
    rpt_location_id_t location_id) const
{
    rpt_address_id_t pc_id(report.get_location_pc_id(location_id));
    rpt_address_t pc(report.get_address(pc_id));
    pc_info_ptr_t pc_info = get_pc_info(report, pc, false);
    location_info_map_t::const_iterator lii(
        pc_info->location_info_map.find(location_id));
    AM_ASSERT(lii != pc_info->location_info_map.end());
    return lii->second;
}

END_NAMESPACE;
END_NAMESPACE;
