/// @file tools/acurate/file_repos.hh

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#ifndef INCLUDED_FILE_REPOS_HH
#define INCLUDED_FILE_REPOS_HH

#include "tools/acurate/xml_file.hh"
#include "analysis/report/report.hh"

#include <boost/foreach.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>

#include <map>
#include <vector>
#include <string>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace acumem::analysis;
using namespace boost;
using namespace std;

struct pc_info_t;
typedef shared_ptr<pc_info_t> pc_info_ptr_t;
typedef weak_ptr<pc_info_t> pc_info_wptr_t;

struct line_info_t;
typedef shared_ptr<line_info_t> line_info_ptr_t;
typedef weak_ptr<line_info_t> line_info_wptr_t;

// ----- location_reporter_t --------------------------------------------------

class location_reporter_t {
public:
    location_reporter_t(int issue_number)
        : issue_number(issue_number)
        {}

    virtual ~location_reporter_t()
        {}

    /// The return value of this function is only used to avoid adding issues
    /// to a source code line multiple times in
    /// html_source_t::emit_terse_issue_line(). Issue types that may want to be
    /// reported twice on the same line, e.g., the first and second loop in a
    /// loop fusion issue, can override this function and return both
    /// issue_number and -issue_number.
    virtual int get_issue_number() const {
        return issue_number;
    }

    virtual void report(xml_file_t &xml) const = 0;

    virtual bool force() const = 0;

protected:
    const int issue_number;
};

typedef shared_ptr<location_reporter_t> location_reporter_ptr_t;

typedef vector<location_reporter_ptr_t> location_reporter_vector_t;

// ----- fr_location_info_t ---------------------------------------------------

struct fr_location_info_t {
    fr_location_info_t(
        rpt_location_id_t location_id,
        pc_info_ptr_t pc_info)
        : location_id(location_id),
          pc_info(pc_info)
        {}

    void add_reporter(location_reporter_ptr_t reporter);

    /// Returns true if force() on any of the location reporters returns true.
    bool force() const;

    // the location id for which this statistics pertains
    const rpt_location_id_t location_id;

    // link to the corresponding pc_info_t
    const pc_info_wptr_t pc_info;

    /// List of callbacks for source code annotation.
    location_reporter_vector_t location_reporter_vector;
};

typedef shared_ptr<fr_location_info_t> fr_location_info_ptr_t;
typedef map<rpt_location_id_t, fr_location_info_ptr_t> location_info_map_t;
typedef vector<fr_location_info_ptr_t> fr_location_info_vector_t;

// ----- pc_info_t ------------------------------------------------------------

struct pc_info_t {
    pc_info_t(rpt_address_t pc,
              rpt_debug_pc_id_t debug_pc_id,
              line_info_ptr_t line_info)
        : pc(pc),
          debug_pc_id(debug_pc_id),
          line_info(line_info)
        {}

    /// PC address.
    const rpt_address_t pc;

    // Debug info for this pc;
    const rpt_debug_pc_id_t debug_pc_id;

    // link to the corresponding line_info.
    const line_info_wptr_t line_info;

    // Map of locations recorded to end with this pc
    location_info_map_t location_info_map;
};

typedef pair<rpt_address_t, bool> pc_info_index_t;
typedef map<pc_info_index_t, pc_info_ptr_t> pc_info_map_t;

// ----- line_info_t ----------------------------------------------------------

struct line_info_t {
    line_info_t(
        const string filename,
        int line)
        : filename(filename),
          line(line),
          num_reporters(0) {}

    /// Source code reference.
    const string filename;
    const int line;

    /// Pcs that map to this file and line.o
    pc_info_map_t pc_info_map;

    /// Number of reporters associated with locations that map to this file and
    /// line.
    int num_reporters;
};

typedef map<int, line_info_ptr_t> line_info_map_t;
typedef map<string, line_info_map_t> file_map_t;

// ----- file_repos_t ---------------------------------------------------------

class file_repos_t {
public:
    file_repos_t(const rpt_report_t &report);

    // Populate the tables (if necessary) so much as to be able to
    // report out the info object for the supplied address.
    //
    // Returns 0 if a specific file/line cannot be found.
    pc_info_ptr_t get_pc_info(
        const rpt_report_t &report,
        rpt_address_t pc,
        bool return_address) const;

    fr_location_info_ptr_t get_line_info(
        const rpt_report_t &report,
        rpt_location_id_t location_id) const;

    struct by_pc_stack_t {
	by_pc_stack_t(const rpt_report_t &report)
	    : report(report)
            {}

        bool addr_id_less(rpt_address_id_t l, rpt_address_id_t r) const {
            return report.get_address(l) < report.get_address(r);
        }

	bool operator()(const fr_location_info_ptr_t &l,
                        const fr_location_info_ptr_t &r) const {
            rpt_address_id_t l_pc(report.get_location_pc_id(l->location_id));
            rpt_address_id_t r_pc(report.get_location_pc_id(r->location_id));
            if (addr_id_less(l_pc, r_pc))
                return true;
            if (addr_id_less(r_pc, l_pc))
                return false;

            const rpt_trunc_stack_t &l_stack(
                report.get_location_trunc_stack(l->location_id));
            const rpt_trunc_stack_t &r_stack(
                report.get_location_trunc_stack(r->location_id));
            for (uint32_t i = 0; i < r_stack.size(); i++)
                if (i == l_stack.size() ||
                    addr_id_less(l_stack[i], r_stack[i]))
                    return true;

            return false;
        }

        const rpt_report_t &report;
    };

    fr_location_info_vector_t get_line_info(
        const rpt_report_t &report,
        const rpt_location_set_t &locations) const {
	fr_location_info_vector_t res;
	BOOST_FOREACH(rpt_location_id_t  location, locations)
	    res.push_back(get_line_info(report, location));
	sort(res.begin(), res.end(), by_pc_stack_t(report));
	return res;
    }

    fr_location_info_vector_t get_depgroup_line_info(
        const rpt_report_t &report,
        rpt_depgroup_id_t depgroup_id) const {
        return get_line_info(
            report, report.get_depgroup_location_set(depgroup_id));
    }

    fr_location_info_vector_t get_loop_incl_line_info(
        const rpt_report_t &report,
        rpt_loop_id_t loop_id) const {
        if (loop_id != 0)
            return get_line_info(
                report, report.get_loop_incl_location_set(loop_id));
        else
            return fr_location_info_vector_t();
    }

    fr_location_info_vector_t get_loop_excl_line_info(
        const rpt_report_t &report,
        rpt_loop_id_t loop_id) const {
        if (loop_id != 0)
            return get_line_info(
                report, report.get_loop_excl_location_set(loop_id));
        else
            return fr_location_info_vector_t();
    }

    // Top of lookup structure.
    // filename -> line -> line info -> pc info -> location info
    file_map_t file_map;

    // Cache for already resolved pcs.
    pc_info_map_t pc_info_map;
};

typedef shared_ptr<file_repos_t> file_repos_ptr_t;
typedef shared_ptr<const file_repos_t> file_repos_cptr_t;

END_NAMESPACE;
END_NAMESPACE;

#endif
