/// @file tools/acurate/html_block.cc

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#include "block.hh"
#include "macros.hh"
#include "help_links.hh"
#include "file_png.hh"

#include "../diagram.hh"
#include "../file_repos.hh"
#include "../link_factory.hh"
#include "../loop_info.hh"
#include "../xml_file.hh"

#include <sstream>
#include <iomanip>
#include <set>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace std;
using namespace boost;
using namespace acumem::analysis;

html_block_t::html_block_t(html_context_t &html_context,
			   rpt_issue_id_t issue_id)
    : html_issue_t(html_context, issue_id)
{
}

summary_type_bitset_t
html_block_t::summary_type() const
{
    const rpt_report_t &report(*html_context.report);
    const rpt_block_issue_t &issue(get_issue());

    summary_type_bitset_t result;

    result.set(SUMMARY_BANDWIDTH);

    double hwpf_ratio(
        report.get_sliced_stats_global(
	    report.get_sliced_stats(
		report.get_depgroup_stats_id(
		    issue.depgroup_id))).get_hardware_prefetch_ratio());

    result.set(SUMMARY_LATENCY, hwpf_ratio < HWPF_HIGH_THRESHOLD);

    return result;
}

string
html_block_t::get_name() const
{
    const rpt_block_issue_t &issue(get_issue());

    if (issue.type_mask & rpt_block_issue_t::Temporal &&
	issue.type_mask & rpt_block_issue_t::Spatial)
	return TEXT_SPATIAL_TEMPORAL_BLOCK;
    else if (issue.type_mask & rpt_block_issue_t::Temporal)
	return TEXT_TEMPORAL_BLOCK;
    else if (issue.type_mask & rpt_block_issue_t::Spatial)
	return TEXT_SPATIAL_BLOCK;
    else
	return TEXT_BLOCK;
}

string
html_block_t::get_icon() const
{
    const rpt_block_issue_t &issue(get_issue());

    string icon;

    if (issue.type_mask & rpt_block_issue_t::Temporal &&
	issue.type_mask & rpt_block_issue_t::Spatial)
	icon = PNG_SPATIAL_TEMPORAL_BLOCK;
    else if (issue.type_mask & rpt_block_issue_t::Temporal)
	icon = PNG_TEMPORAL_BLOCK;
    else if (issue.type_mask & rpt_block_issue_t::Spatial)
	icon = PNG_SPATIAL_BLOCK;
    else
	icon = PNG_BLOCK;

    return html_context.link_factory->image_factory()->static_image_path(icon);
}

void
html_block_t::annotate_source()
{
    annotate_generic(get_issue().depgroup_id, false);
}

void
html_block_t::emit_issue_file(xml_file_t &issue_xml)
{
    const rpt_report_t &report(*html_context.report);
    const file_repos_t &file_repos(*html_context.file_repos);
    const rpt_block_issue_t &issue(get_issue());

    string type = get_name();
    string help;

    if (issue.type_mask & rpt_block_issue_t::Temporal &&
	issue.type_mask & rpt_block_issue_t::Spatial) {
	help = HELP_TEMP_SPAT_BLOCK;
    } else if (issue.type_mask & rpt_block_issue_t::Temporal) {
	help = HELP_TEMP_BLOCK;
    } else {
	help = HELP_SPAT_BLOCK;
    }

    fr_location_info_vector_t location_dep_info_vector =
	file_repos.get_depgroup_line_info(report, issue.depgroup_id);
    fr_location_info_vector_t location_loop_info_vector =
	file_repos.get_loop_incl_line_info(report, issue.loop_id);

    fr_location_info_vector_t location_backwards_info_vector =
	file_repos.get_line_info(report, issue.previous_writers);
    fr_location_info_vector_t location_forwards_info_vector =
	file_repos.get_line_info(report, issue.next_writers);

    html_context.interesting_depgroups.insert(issue.depgroup_id);
    html_context.interesting_loops.insert(issue.loop_id);

    XML_OBJECT(issue_xml) {
	T_DIV A_ID("sections") A_STYLE("display: none") {
	    emit_issue_header(issue_xml);

	    SECTION(h1, "d", "Issue #" + get_display_number() + ": " + type,
		    get_icon(), help, false, false);

	    // instruction group statistics
	    section_statistics(
                xml,
                html_context,
                "Statistics for instructions of this issue",
                HELP_ISSUE_STATISTICS,
                report.get_depgroup_stats_id(issue.depgroup_id),
                "d" + lexical_cast<string>(issue.depgroup_id),
                location_dep_info_vector,
                location_loop_info_vector,
                false, // initially visible
                false, // is_loop
                false, // show_avoidable
                "istat");

	    // instruction group instructions
	    section_instructions(xml, html_context,
				 "Instructions benefiting from blocking",
				 HELP_ISSUE_INSTRUCTIONS,
				 location_dep_info_vector,
				 location_loop_info_vector,
				 2, false, "iinstr");

	    if (location_backwards_info_vector.size() > 0) {
		section_instructions(xml, html_context,
				     "Instructions previously "
				     "writing to related data",
				     HELP_ISSUE_TRACKING,
				     location_backwards_info_vector,
				     location_loop_info_vector,
				     0, false, "iback");
	    }

	    if (location_forwards_info_vector.size() > 0) {
		section_instructions(xml, html_context,
				     "Next instructions to write "
				     "to related data",
				     HELP_ISSUE_TRACKING,
				     location_forwards_info_vector,
				     location_loop_info_vector,
				     0, false, "inext");
	    }


	    // -----------------------------------------------------

	    // Display the information specific to blocking:

	    string id = "ibarr";
	    SECTION(h2, id, "Loops and barriers related to this issue",
		    "", HELP_BLOCKING_INFO, true, true) {

		if (issue.outer_loop_id != 0) {
		    SECTION(h3, id +"a",
			    "Outermost loop without barriers",
			    "", HELP_BLOCKING_OUTER, true, true) {
			// First the feasible loop level.
			if (issue.outer_loop_id == issue.loop_id) {
			    // outer_loop (outermost feasible loop) same
			    // as issue loop. Just output a reference to
			    // the final loop section.

			    T_P {
				TEXT("The loop including the blocking "
				     "instruction is also the outermost "
				     "loop without barriers.");
				if (issue.next_outer_loop_id == 0 &&
				    issue.target_loop_id == 0) {
				    TEXT("There are no other identified loop "
					 "levels outside this level (which "
					 "is normal if the inner loop is "
					 "executed significantly more than "
					 "the outer loops). There may be "
					 "outer loops that would be a "
					 "suitable target, but they could "
					 "not be identified.");
				}
			    }

			} else {
			    // outer_loop. Output the exclusive loop
			    // instructions.
			    fr_location_info_vector_t location_outer_loop_info_vector;
                            location_outer_loop_info_vector =
                                file_repos.get_loop_excl_line_info(report, issue.outer_loop_id);

			    instruction_table(xml, html_context,
					      location_outer_loop_info_vector,
					      location_outer_loop_info_vector,
					      0, true);

			}
			T_P {
			    stringstream ss;
			    ss << "At this loop level, there would "
				"be approximately "
			       << issue.outer_loop_iterations
			       << " iterations per data reuse "
				"of the target instruction." << endl;
			    TEXT(ss.str());
			}
		    }
		}

		// Next: the loop level involving barriers
		if (issue.next_outer_loop_id != 0) {
		    SECTION(h3, id+"b",
			    "Next outer loop, introducing "
			    "barriers",
			    "", HELP_BLOCKING_NEXT_OUTER, true, true) {

			// Bizarre - should never occur. By definition the
			// feasible loop must not have barriers.
			assert(issue.outer_loop_id != 0 &&
			       issue.next_outer_loop_id != issue.outer_loop_id);

			// Also bizarre. We should never report issues for
			// blocking where the instruction loop have
			// barriers.
			assert(issue.next_outer_loop_id != issue.loop_id);

			// There is a loop level with identified barriers.
			// Display the loop instructions and barriers.

			fr_location_info_vector_t location_next_outer_loop_info_vector;
			location_next_outer_loop_info_vector =
			    file_repos.get_loop_excl_line_info(report, issue.next_outer_loop_id);

			instruction_table(xml, html_context,
					  location_next_outer_loop_info_vector,
					  location_next_outer_loop_info_vector,
					  0, true);
			T_P {
			    stringstream ss;
			    ss << "At this loop level, there would "
				"be approximately "
			       << issue.next_outer_loop_iterations
			       << " iterations per data reuse "
				"of the target instruction." << endl;
			    TEXT(ss.str());
			}

			if (issue.barrier_pcs.size()) {
			    T_P TEXT("The following table identifies which "
				     "instructions outside this loop writes "
				     "to data touched by the instructions of "
				     "the loop, thus making loop nesting "
				     "reorganizations difficult.");
			    motion_barrier_table(xml, html_context,
						 issue.barrier_pcs, issue.barrier_blocks,
						 location_next_outer_loop_info_vector);
			}
		    }
		}

		// Finally, the loop level which is about the size of
		// the trigger reuse distance.
		if (issue.target_loop_id != 0) {

		    SECTION(h3, id+"c",
			    "Optimal loop level to block at",
			    "", HELP_BLOCKING_TARGET, true, true) {

			// We have found the loop level which probably
			// carries the index for which to block

			if (issue.target_loop_id == issue.loop_id) {
			    // The instruction's own loop is the target
			    // loop. Odd, but not impossible. Output a
			    // reference to the loop section below.

			    T_P TEXT("The recommended optimal loop to block is "
				     "the instruction's own loop, or an "
				     "unidentified loop outside this loop.");

			} else if (issue.outer_loop_id != 0 &&
				   issue.target_loop_id ==
                                   issue.outer_loop_id) {
			    // same loop as the feasible loop. Output
			    // reference to feasible loop.

			    T_P TEXT("The recommended target loop to block is "
				     "the outermost loop, above.");

			} else if (issue.next_outer_loop_id != 0 &&
				   issue.target_loop_id ==
				   issue.next_outer_loop_id) {
			    // same loop as the barrier loop. Output
			    // reference to barrier loop.

			    T_P TEXT("The recommended target loop to block is "
				     "the loop introducing barriers, above");

			} else {
			    // Distinct outer loop. Output statistics and
			    // instructions.
			    fr_location_info_vector_t target_loop_info_vector =
				file_repos.get_loop_excl_line_info(report, issue.target_loop_id);

			    instruction_table(xml, html_context,
					      target_loop_info_vector,
					      target_loop_info_vector,
					      0, true);
			}

			T_P {
			    stringstream ss;
			    ss << "At this loop level, there would "
				"be approximately "
			       << issue.target_loop_iterations
			       << " iterations per data reuse "
				"of the target instruction." << endl;
			    TEXT(ss.str());
			}
		    }
		}
	    }

	    // -----------------------------------------------------
	    // Common section for describing the Loop

	    all_loop_sections(xml, html_context,
			      issue.loop_id, location_loop_info_vector);


	    // -----------------------------------------------------
	    // done
	    emit_issue_footer(html_context, issue_xml);

	}
    }
}


void
html_block_t::emit_summary(xml_file_t &xml,
			   summary_context_t &summary_context,
			   summary_type_t summary_type)
{
    const rpt_report_t &report(*html_context.report);
    const file_repos_t &file_repos(*html_context.file_repos);
    const rpt_block_issue_t &issue(get_issue());

    fr_location_info_vector_t location_dep_info_vector =
	file_repos.get_depgroup_line_info(report, issue.depgroup_id);
    fr_location_info_vector_t location_loop_info_vector =
	file_repos.get_loop_incl_line_info(report, issue.loop_id);

    add_summary_line(
        xml,
        summary_context,
        get_rel_file_name(),
        get_name(),
        get_icon(),
        location_dep_info_vector,
        location_loop_info_vector,
        report.get_depgroup_stats_id(issue.depgroup_id),
        lexical_cast<string>(issue.loop_id),
        summary_type);
}

END_NAMESPACE;
END_NAMESPACE;
