/// @file tools/acurate/html_false_sharing.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include "false_sharing.hh"
#include "macros.hh"
#include "help_links.hh"
#include "file_png.hh"

#include "../diagram.hh"
#include "../file_repos.hh"
#include "../link_factory.hh"
#include "../loop_info.hh"
#include "../xml_file.hh"

#include <sstream>
#include <iomanip>
#include <set>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace std;
using namespace boost;
using namespace acumem::analysis;

html_false_sharing_t::html_false_sharing_t(html_context_t &html_context,
					   rpt_issue_id_t issue_id)
    : html_issue_t(html_context, issue_id)
{
}

summary_type_bitset_t
html_false_sharing_t::summary_type() const
{
    summary_type_bitset_t result;

    result.set(SUMMARY_MT);

    return result;
}

string
html_false_sharing_t::get_icon() const
{
    string icon = PNG_FALSE_SHARING;
    return html_context.link_factory->image_factory()->static_image_path(icon);
}

void
html_false_sharing_t::annotate_source()
{
    annotate_generic(get_issue().depgroup_id, true);
}

void
html_false_sharing_t::emit_issue_file(xml_file_t &issue_xml)
{
    const rpt_report_t &report(*html_context.report);
    const file_repos_t &file_repos(*html_context.file_repos);
    const rpt_false_sharing_issue_t &issue(get_issue());

    fr_location_info_vector_t location_dep_info_vector =
	file_repos.get_depgroup_line_info(report, issue.depgroup_id);
    fr_location_info_vector_t location_loop_info_vector =
	file_repos.get_loop_incl_line_info(report, issue.loop_id);

    html_context.interesting_depgroups.insert(issue.depgroup_id);
    html_context.interesting_loops.insert(issue.loop_id);

    // Get information about false consumer instructions
    fr_location_info_vector_t offender_info_vector =
        file_repos.get_line_info(report, issue.offenders);

    XML_OBJECT(issue_xml) {
	T_DIV A_ID("sections") A_STYLE("display: none") {
	    emit_issue_header(issue_xml);

	    // Title
	    SECTION(h1, "d", "Issue #" + get_display_number() + ": " + get_name(),
		    get_icon(), HELP_FALSESHARING, false, false);

	    // instruction group statistics
	    section_statistics(
                xml,
                html_context,
                "Statistics for instructions of this issue",
                HELP_ISSUE_STATISTICS,
                report.get_depgroup_stats_id(issue.depgroup_id),
                "d" + lexical_cast<string>(issue.depgroup_id),
                location_dep_info_vector,
                location_loop_info_vector,
                false, // initially visible
                false, // is_loop
                true, // show_avoidable
                "istat");

	    // instruction group instructions
	    section_instructions(xml, html_context,
				 "Instructions involved in this issue",
				 HELP_ISSUE_INSTRUCTIONS,
				 location_dep_info_vector,
				 location_loop_info_vector,
				 2, false, "iinstr");

	    // -----------------------------------------------------
	    // Data consumers

	    if (offender_info_vector.size() > 0)
		section_instructions(xml, html_context,
				     "Instructions causing false "
				     "sharing of the cache line",
				     HELP_FS_CONSUMERS,
				     offender_info_vector,
				     location_loop_info_vector,
				     0, false, "ifs");

	    // -----------------------------------------------------
	    // Common section for describing the Loop

	    all_loop_sections(xml, html_context,
			      issue.loop_id, location_loop_info_vector);

	    // -----------------------------------------------------
	    // done

	    emit_issue_footer(html_context, issue_xml);


	}
    }
}

void
html_false_sharing_t::emit_summary(xml_file_t &xml,
				   summary_context_t &summary_context,
				   summary_type_t summary_type)
{
    const rpt_report_t &report(*html_context.report);
    const file_repos_t &file_repos(*html_context.file_repos);
    const rpt_false_sharing_issue_t &issue(get_issue());

    const fr_location_info_vector_t location_dep_info_vector =
	file_repos.get_depgroup_line_info(report, issue.depgroup_id);
    const fr_location_info_vector_t location_loop_info_vector =
	file_repos.get_loop_incl_line_info(report, issue.loop_id);

    add_summary_line(
        xml,
        summary_context,
        get_rel_file_name(),
        get_name(),
        get_icon(),
        location_dep_info_vector,
        location_loop_info_vector,
        report.get_depgroup_stats_id(issue.depgroup_id),
        lexical_cast<string>(issue.loop_id),
        summary_type);
}

END_NAMESPACE;
END_NAMESPACE;
