/// @file tools/acurate/html_front.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#ifdef _WIN32
// VisualC++ math libraries don't consider M_PI to be part of standard
// libraries and require a define to include them.
#define _USE_MATH_DEFINES
#endif

#include "front.hh"
#include "file_png.hh"
#include "help_links.hh"
#include "macros.hh"

#include "../diagram.hh"
#include "../link_factory.hh"
#include "../loop_info.hh"
#include "../product_data.hh"

#include "analysis/feature_base/exceptions.hh"

#include <gd/gd.h>
#include <boost/lexical_cast.hpp>
#include <boost/foreach.hpp>

#include <assert.h>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

#if defined(_WIN32)
/// MSVC doesn't have round() in math.h.
///
/// NOTE: This function handles negative values incorrectly (round() should
/// round away from zero).
static inline double
round(double v)
{
    if (v - floor(v) >= 0.5)
	return ceil(v);
    else
	return floor(v);
}
#endif

using namespace std;
using namespace boost;

string acumem_slowspotter_tm = product_data_t::product_name_utf8();
string read_more = "Read" NBSP "more...";
string see_manual = "Manual";

#define T_ROUNDED_DIV				\
    T_DIV A_CLASS("t")				\
	T_DIV A_CLASS("b")			\
	T_DIV A_CLASS("l")			\
	T_DIV A_CLASS("r")			\
	T_DIV A_CLASS("bl")			\
	T_DIV A_CLASS("br")			\
	T_DIV A_CLASS("tl")			\
	T_DIV A_CLASS("tr")

// ----------------------------------------------------------------------

class link_base_t {
public:
    virtual ~link_base_t() {}
    virtual void emit(xml_file_t &xml) {}
    virtual void emit_after(xml_file_t &xml) {}
};

typedef shared_ptr<link_base_t> link_ptr_t;
typedef vector<link_ptr_t> link_vector_t;

// ----------------------------------------------------------------------

class link_t : public link_base_t {
public:
    link_t(const string &href,
	   const string &text,
	   const string &target = "#blank")
	: href(href), text(text), target(target)
	{}

    virtual void emit(xml_file_t &xml)
	{
	    T_A A_HREF(href) {
		if (target != "") A_TARGET(target);
		TEXT_T(text);
	    }
	}

private:
    string href;
    string text;
    string target;
};


// ----------------------------------------------------------------------

class help_link_t : public link_base_t {
public:
    help_link_t(html_context_t &html_context,
		const string &help_section,
		const string &text)
	: help_url(html_context.link_factory->get_help_link(help_section)),
	  text(text)
	{}

    void emit(xml_file_t &xml)	{
	    T_A A_HREF(help_url)
		A_TARGET("help")
		A_ONCLICK("return popitup('" + help_url + "')")
		TEXT_T(text);
	}

private:
    string help_url;
    string text;
};

// ----------------------------------------------------------------------

class link_inline_t : public link_base_t {
public:
    link_inline_t(const string &inner_xml, const string &text)
	: inner_xml(inner_xml), text(text)
	{}

    void emit(xml_file_t &xml)
	{
	    id = xml.unique_id();
	    string help_toggle =
		"toggle('" + id + "');";
	    T_A A_HREF("javascript:void(0);") A_ONCLICK(help_toggle)
		TEXT_T(text);
	}

    void emit_after(xml_file_t &xml)
	{
	    assert(id != "");
	    T_DIV A_ID(id) A_STYLE("display:none")
		TEXT_T(inner_xml);
	}

private:
    string id;
    string inner_xml;
    string text;
};

// ----------------------------------------------------------------------

class link_section_t {
public:
    link_section_t()
	{}

    void add(link_base_t *l) {
	links.push_back(link_ptr_t(l));
    }

    /// XXX Sometimes this should bring up the help window, and we
    /// need special handling of that.
    void emit(xml_file_t &xml, const string &title = "", int columns = -1)
	{
	    string td_width;

	    if (columns == -1)
		columns = (int)links.size();
	    else
		td_width = lexical_cast<string>(100/columns) + "%";

	    if (links.size() == 1) {
		links.front()->emit(xml);
		links.front()->emit_after(xml);
	    } else {
		T_DIV A_CLASS("link") {
		    if (title.size())
			T_H4 TEXT(title);

		    int size = (int)((links.size() + columns - 1) / columns);
		    link_vector_t::const_iterator
			i = links.begin(), e = links.end();
		    T_TABLE {
			if (td_width != "") A_WIDTH("100%");
			for (int r=0; r<size; r++) T_TR {
				for (int c=0; c<columns; c++) T_TD {
					if (td_width != "") A_WIDTH(td_width);
					A_VALIGN("top") {
					    if (i != e) {
						(*i)->emit(xml);
						i++;
					    }
					}
				    }
			    }
		    }

		    BOOST_FOREACH(link_ptr_t l, links)
			l->emit_after(xml);

		}
	    }
	}

private:
    link_vector_t links;
};

// ----- Problem limits. ------------------------------------------------------

/// Limits for not reporting different kinds of problems for a cache level.
struct problem_limits_t {
    /// Higher fetch ratios result in a bandwidth problem.
    double fr_limit;

    /// Higher miss ratios result in a latency problem.
    double mr_limit;

    /// Lower utilizations result in a locality problem.
    double util_limit;
};

/// Problem limits for first-level caches.
static const problem_limits_t first_level_limits = {
    // Bandwidth to the next cache level should be plentiful. Allow up 10%
    // fetch ratio.
    0.1,

    // Assuming a cost of 10 cycles per first-level cache miss and 1 memory
    // access per cycle, we need a 2% miss ratio for a 20% performance hit.
    0.02,

    // Pretty arbitrary utilization limit.
    0.75
};

/// Problem limits for mid-level caches.
static const problem_limits_t mid_level_limits = {
    // Bandwidth to the next cache level should be quite plentiful. Allow up 5%
    // fetch ratio.
    0.05,

    // Assuming a cost of 30 cycles per mid-level cache miss and 1 memory
    // access per cycle, we need a 0.67% miss ratio for a 20% performance hit.
    0.0067,

    // Pretty arbitrary utilization limit.
    0.75
};

/// Problem limits for last-level caches.
static const problem_limits_t last_level_limits = {
    // Rough bandwidth limit calculation calculation:
    //
    // A 1066 MHz front-side bus gives us a bandwidth of
    // 1066 * 8 / 64 M lines/s = 133 M lines/s
    //
    // A 3000 MHz quad core processor doing 1 memory access per cycle per core
    // does a total of 3000 * 4 * 1 M accesses/s = 12000 M accesses/s
    //
    // To saturate the front-side bus we then need a fetch ratio of
    // 133 / 12000 = 1.1%
    0.04,

    // Assuming a cost of 100 cycles per last-level cache miss and 1 memory
    // access per cycle, we need a 0.2% miss ratio for a 20% performance hit.
    0.002,

    // Pretty arbitrary utilization limit.
    0.75
};


// ----------------------------------------------------------------------
// Verdict base class

class link_section_t;

class verdict_t {
public:
    virtual ~verdict_t() {}
    virtual string verdict() const = 0;
    virtual void add_links(link_section_t &ls) const {}
};

// ----------------------------------------------------------------------
// General problems

static string worst_behavior =
    "This application suffers from severe cache and memory bus related "
    "problems due to several different causes.";


// ----------------------------------------------------------------------
// Latency problems

static string prefetch_problem =
    "In addition to this, the application exhibits irregular memory "
    "access patterns "
    "suggesting that the hardware prefetcher will have trouble requesting "
    "data in advance. Thus, a large number of the memory accesses may cause "
    "the processor to come to a complete stop while waiting for memory "
    "transfers to complete.";

static string todo_util_pf =
    "If data can be packed tighter, if algorithms and data structures "
    "can be changed to access data more regularly, and if data can be "
    "organized to help the hardware prefetcher, then this application "
    "would run faster and scale much better on a wide range of "
    "multicore systems.";


// ----------------------------------------------------------------------
// Utilization problems

static string util_intro_low_fr =
    "Only a small portion of the data transferred over the memory bus, and "
    "loaded into the cache, is used by this application before being evicted "
    "from the cache again. This is disadvantageous in three respects:";

static string util_intro_high_fr =
    "This application transfers lots of data over the memory bus, but a large "
    "fraction of the data is not being used before it is evicted from the "
    "cache again. Typically, this means that data is accessed in a regular "
    "fashion, but that data structures are sparse. This is disadvantageous in "
    "three respects:";

static string util_reason_1 =
    "The cache is filled with lots of data that this application is not "
    "using. This essentially reduces the effective size of the cache since "
    "there is less space for actually used data. A smaller effective size "
    "makes the cache less effective and increases the number of cache misses.";

static string util_reason_2 =
    "Each time new data is brought into the cache, only a a fraction of it is "
    "useful. For that reason, new data has to be brought into the cache more "
    "often. This increases the number of times a processor will have to stall "
    "waiting for data.";

static string util_reason_3 =
    "The memory bus is used to transfer data which is not used. This "
    "increases the amount of data transferred across the memory bus, leading "
    "to more contention for the memory bus and seriously limiting scalability "
    "of this application on multicore architectures.";

static string util_in_short =
    "In short, too much unused data is transported between the memory and "
    "the caches, this limits application scalability, particularly on "
    "multicore systems where many cores share the same memory bus.";

static string util_todo =
    "If the used data can be packed tighter to minimize the bandwidth wasted "
    "transporting unused data between memory and caches, then this "
    "application would run faster and scale better on a wide range of "
    "multicore systems.";

// ----------------------------------------------------------------------

static string bad_access_problem =
    "This application is accessing data in "
    "such a way that some hardware prefetchers will not be able to work "
    "efficiently. This is related both to the data layout in memory, as well "
    "as the order in which algorithms are using that data.";

static string bad_access_problem2 =
    "There is no immediate problem with the use of the cache and memory bus, "
    "meaning that most of the data in every cache line is used before being "
    "replaced, and little bandwidth is wasted transported unused data over "
    "the memory bus. The main problem is the irregularity of the accesses.";

static string todo_pf =
    "If algorithms and data structures can be changed to access data "
    "more regularly, and if data can be organized to help the hardware "
    "prefetcher, then this application would run faster and scale "
    "better on a wide range of multicore systems.";

// ----------------------------------------------------------------------

class verdict_111_t : public verdict_t {
public:
    virtual string verdict() const {
	return "Avoid major processor stalls and a congested memory bus due "
	    "to poor data usage.";
    }

    virtual void add_links(link_section_t &ls) const
	{
	    stringstream e;
	    xml_file_t xml(e, Doctype_None);

	    T_P	TEXT(worst_behavior);
	    T_P TEXT(util_intro_low_fr);
	    T_UL {
		T_LI TEXT(util_reason_1);
		T_LI TEXT(util_reason_2);
		T_LI TEXT(util_reason_3);
	    }
	    T_P TEXT(prefetch_problem);

	    T_P T_STRONG TEXT("What needs to be done?");
	    T_P TEXT(todo_util_pf);

	    ls.add(new link_inline_t(e.str(), read_more));
	}
};

// ----------------------------------------------------------------------

class verdict_110_t : public verdict_t {
public:
    virtual string verdict() const {
	return "Avoid major processor stalls due to irregular access patterns";
    }

    virtual void add_links(link_section_t &ls) const
	{
	    stringstream e;
	    xml_file_t xml(e, Doctype_None);

	    T_P TEXT(bad_access_problem);
	    T_P TEXT(bad_access_problem2);

	    T_P T_STRONG TEXT("What needs to be done?");
	    T_P TEXT(todo_pf);

	    ls.add(new link_inline_t(e.str(), read_more));
	}
};

// ----------------------------------------------------------------------

class verdict_101_t : public verdict_t {
public:
    virtual string verdict() const {
	return "Avoid congesting the memory bus and get major "
	    "improvements in scalability through better data usage.";
    }

    virtual void add_links(link_section_t &ls) const
	{
	    stringstream e;
	    xml_file_t xml(e, Doctype_None);

	    T_P TEXT(util_intro_high_fr);
	    T_UL {
		T_LI TEXT(util_reason_1);
		T_LI TEXT(util_reason_2);
		T_LI TEXT(util_reason_3);
	    }
	    T_P TEXT(util_in_short);

	    T_P T_STRONG TEXT("What needs to be done?");
	    T_P TEXT(util_todo);

	    ls.add(new link_inline_t(e.str(), read_more));
	}
};

// ----------------------------------------------------------------------
static string text_122 =
    "This application congests the memory bus, even though "
    "its access patterns are regular and efficient. Since the "
    "application is heavily memory bound, even small changes to the "
    "application's access patterns may have a large effect.";

static string todo_blocking =
    "When data is already accessed in a regular and packed way, the next "
    "thing on the agenda is finding opportunities for reusing data while it "
    "remains in the cache. Doing this right ensures that the program will "
    "run efficiently on a wide range of architectures and cache hierarchies.";

static string todo_monitor_usage =
    "It is also valuable to see what behaviour this program has on "
    "different multicore systems, for different workloads and interleaved "
    "with other tasks.";

class verdict_100_t : public verdict_t {
public:
    virtual string verdict() const {
	return "Address general scalability and look for bandwidth "
	    "related optimizations";
    }

    virtual void add_links(link_section_t &ls) const
	{
	    stringstream e;
	    xml_file_t xml(e, Doctype_None);

	    T_P	TEXT(text_122);

	    T_P T_STRONG TEXT("What needs to be done?");
	    T_P {
		TEXT(todo_blocking);
		TEXT(" ");
		TEXT(todo_monitor_usage);
	    }

	    ls.add(new link_inline_t(e.str(), read_more));
	}
};

// ----------------------------------------------------------------------
static string low_fetch_ratio =
    "Given the current input set, this application is not putting high "
    "pressure on the memory bus, but other input sets may well have "
    "different characteristics and behave worse.";

static string todo_try_larger_sets =
    "It would be worth while to try other data sets to ensure that "
    "performance doesn't deteriorate with a larger working set.";

class verdict_011_t : public verdict_t {
public:
    virtual string verdict() const {
	return "Tune cache utilization to avoid processor stalls.";
    }

    virtual void add_links(link_section_t &ls) const
	{
	    stringstream e;
	    xml_file_t xml(e, Doctype_None);

	    T_P	{
		TEXT(low_fetch_ratio);
		TEXT(" ");
		TEXT(util_intro_low_fr);
	    }
	    T_UL {
		T_LI TEXT(util_reason_1);
		T_LI TEXT(util_reason_2);
		T_LI TEXT(util_reason_3);
	    }
	    T_P TEXT(prefetch_problem);

	    T_P T_STRONG TEXT("What needs to be done?");
	    T_P TEXT(todo_util_pf);
	    T_P TEXT(todo_try_larger_sets);

	    ls.add(new link_inline_t(e.str(), read_more));
	}
};

// ----------------------------------------------------------------------

static string bad_access_pattern =
    "This application is accessing data in "
    "such a way that some hardware prefetchers will not be able to work "
    "efficiently. This is related both to the data layout in memory, as well "
    "as the order in which algorithms are using that memory. There is no "
    "immediate "
    "problem with the cache and memory bus utilization, meaning that most of "
    "the data in every cache line is used before being replaced, and little "
    "bandwidth is wasted on the memory bus.";

class verdict_010_t : public verdict_t {
public:
    virtual string verdict() const {
	return "Tune hardware prefetching by rearranging memory "
	    "layouts and modifying algorithms.";
    }

    virtual void add_links(link_section_t &ls) const
	{
	    stringstream e;
	    xml_file_t xml(e, Doctype_None);

	    T_P	{
		TEXT(low_fetch_ratio);
	    }
	    T_P TEXT(bad_access_pattern);

	    T_P T_STRONG TEXT("What needs to be done?");
	    T_P TEXT(todo_pf);
	    T_P TEXT(todo_try_larger_sets);

	    ls.add(new link_inline_t(e.str(), read_more));
	}
};

// ----------------------------------------------------------------------

static string in_short_2 =
    "In short, unused data is transported between memory and caches, which "
    "could limit application scalability, particularly on multicore "
    "systems with small caches, systems where multiple processors share the "
    "same memory bus, or for larger data sets.";

class verdict_001_t : public verdict_t {
public:
    virtual string verdict() const {
	return "Tune bandwidth usage to ensure good scalability "
	    "for a large range of input sets.";
    }

    virtual void add_links(link_section_t &ls) const
	{
	    stringstream e;
	    xml_file_t xml(e, Doctype_None);

	    T_P	TEXT(low_fetch_ratio);
	    T_P	{
		TEXT(util_intro_high_fr);
	    }
	    T_UL {
		T_LI TEXT(util_reason_1);
		T_LI TEXT(util_reason_2);
		T_LI TEXT(util_reason_3);
	    }
	    T_P TEXT(in_short_2);

	    T_P T_STRONG TEXT("What needs to be done?");
	    T_P TEXT(util_todo);
	    T_P TEXT(todo_try_larger_sets);

	    ls.add(new link_inline_t(e.str(), read_more));
	}
};

// ----------------------------------------------------------------------

static string text_222 =
    "This application is fairly well-behaved. It uses little memory and "
    "its memory accesses seem to be fairly"
    "regular and packed.";

class verdict_000_t : public verdict_t {
public:
    virtual string verdict() const {
	return "Try more demanding data sets and ascertain that they also "
	    "are well behaved in terms of cache utilization and bandwidth "
	    "usage.";
    }

    virtual void add_links(link_section_t &ls) const
	{
	    stringstream e;
	    xml_file_t xml(e, Doctype_None);

	    T_P	TEXT(text_222);
	    T_P	TEXT(low_fetch_ratio);

	    T_P T_STRONG TEXT("What needs to be done?");
	    T_P TEXT(todo_try_larger_sets);

	    ls.add(new link_inline_t(e.str(), read_more));
	}
};

// ----------------------------------------------------------------------

static bool
have_issue_type(const rpt_report_t &report,
                rpt_issue_type_t issue_type)
{
    BOOST_FOREACH(const rpt_issue_ptr_t &issue, report.issues)
        if (issue && issue->get_issue_type() == issue_type)
            return true;
    return false;
}


application_verdict_t::application_verdict_t(html_context_t &html_context)
{
    const rpt_report_t &report(*html_context.report);

    const rpt_stats_t &global_stats(
        report.get_sliced_stats_global(
            report.get_sliced_stats(report.get_global_stats_id())));

    // Get some statistics.
    const double fetch_ratio(global_stats.get_total_fetch_ratio());
    const double miss_ratio(global_stats.get_total_miss_ratio());
    const double utilization(
        global_stats.get_fetch_utilization().get_valid() ?
        global_stats.get_fetch_utilization().get_value() :
	1.0);

    // Get the cache level and number of cache levels.
    unsigned int cache_level = html_context.report->cache_level;
    unsigned int num_cache_levels = html_context.report->num_cache_levels;


    // Get problem ratios.
    // Look up cache level dependent problem limits.
    const problem_limits_t *limits;
    if (cache_level == 1)
	limits = &first_level_limits;
    else if (cache_level == num_cache_levels)
	limits = &last_level_limits;
    else
	limits = &mid_level_limits;

    bandwidth_problem_ratio = fetch_ratio / limits->fr_limit;
    latency_problem_ratio = miss_ratio / limits->mr_limit;
    locality_problem_ratio =
	(1.0 - utilization) / (1.0 - limits->util_limit);


    // Scan for what advice there really is
    has_locality_advice =
	(have_issue_type(report, Rpt_Issue_Random) ||
	 have_issue_type(report, Rpt_Issue_Utilization) ||
	 have_issue_type(report, Rpt_Issue_Loop_Order));

    has_mt_advice =
	(have_issue_type(report, Rpt_Issue_False_Sharing) ||
	 have_issue_type(report, Rpt_Issue_Comm_Utilization));

    has_issues = has_bandwidth_advice = has_latency_advice = false;

    BOOST_FOREACH(const loop_info_map_t::value_type &v,
		  html_context.loop_registry->loop_infos) {
        const loop_info_ptr_t &loop = v.second;
        BOOST_FOREACH(rpt_depgroup_id_t depgroup_id, loop->depgroup_ids) {
	    depgroup_info_ptr_t &d =
		html_context.loop_registry->depgroup_infos[depgroup_id];

	    BOOST_FOREACH(html_issue_ptr_t i, d->issues) {
		has_issues = true;
		const summary_type_bitset_t summary_types(
                    i->summary_type());
		if (summary_types.test(SUMMARY_BANDWIDTH))
		    has_bandwidth_advice = true;
		if (summary_types.test(SUMMARY_LATENCY))
		    has_latency_advice = true;
	    }
	}
    }
}

bool
application_verdict_t::has_bandwidth_problem() const
{
    return bandwidth_problem_ratio >= 1.0;
}

bool
application_verdict_t::has_latency_problem() const
{
    return latency_problem_ratio >= 1.0;
}

bool
application_verdict_t::has_locality_problem() const
{
    // Don't complain about utilization unless there is a bandwidth or
    // latency problem.
    return
	(has_bandwidth_problem() || has_latency_problem()) &&
	locality_problem_ratio >= 1.0;
}

verdict_cptr_t
application_verdict_t::get_verdict() const
{
    if (has_bandwidth_problem()) {
	if (has_latency_problem()) {
	    if (has_locality_problem())
		return verdict_cptr_t(new verdict_111_t);
	    else
		return verdict_cptr_t(new verdict_110_t);
	} else {
	    if (has_locality_problem())
		return verdict_cptr_t(new verdict_101_t);
	    else
		return verdict_cptr_t(new verdict_100_t);
	}
    } else {
	if (has_latency_problem()) {
	    if (has_locality_problem())
		return verdict_cptr_t(new verdict_011_t);
	    else
		return verdict_cptr_t(new verdict_010_t);
	} else {
	    if (has_locality_problem())
		return verdict_cptr_t(new verdict_001_t);
	    else
		return verdict_cptr_t(new verdict_000_t);
	}
    }
}

void
application_verdict_t::render_fr_gauge(std::ostream &os) const
{
    gauge_t gauge(bandwidth_problem_ratio);
    gauge.draw()->save_as_png_ostream(os);
}

void
application_verdict_t::render_mr_gauge(std::ostream &os) const
{
    gauge_t gauge(latency_problem_ratio);
    gauge.draw()->save_as_png_ostream(os);
}

void
application_verdict_t::render_utilization_gauge(std::ostream &os) const
{
    gauge_t gauge(locality_problem_ratio);
    gauge.draw()->save_as_png_ostream(os);
}

void
application_verdict_t::render_mt_gauge(std::ostream &os) const
{
    gauge_t gauge(has_mt_advice ? 1.0 : 0.0);
    gauge.draw()->save_as_png_ostream(os);
}

// ----------------------------------------------------------------------

html_front_t::html_front_t(html_context_t &html_context,
			   const application_verdict_t &app_verdict)
    : html_context(html_context),
      app_verdict(app_verdict)
{
}

void
html_front_t::emit_warning(xml_file_t &xml)
{
    const rpt_report_t &report(*html_context.report);

    bool result_unreliable = html_context.result_unreliable();
    bool warn_virtualized = html_context.warn_virtualized();

    if (result_unreliable || warn_virtualized) {
	T_DIV A_CLASS("warning") {
	    T_H1 TEXT("Warning!");
	    T_UL {
		if (result_unreliable) T_LI {
		    T_P TEXT("These results are based on too few samples to be "
			     "reliable. The sample file contains " +
			     lexical_cast<string>(report.sample_count) +
			     " samples of the selected line size, but " +
			     lexical_cast<string>(MIN_RELIABLE_SAMPLE_COUNT) +
			     " samples are generally required for reliable results.");

		    T_P TEXT("Re-sample the application for a longer time "
			     "or with a shorter sample period for a "
			     "reliable result.");
		}

		if (warn_virtualized) T_LI {
		    T_P TEXT("CPU and cache hierarchy detection does not work "
			     "reliably in a virtualized environment. You need "
			     "to manually override the number of caches or "
			     "manually select a processor.");
		}
	    }
	}
    }
}

void
html_front_t::emit(xml_file_t &xml)
{
    T_DIV A_CLASS("overview") {
#if 0
	T_ROUNDED_DIV {
	    T_TABLE T_TR {
		T_TD A_VALIGN("top") {
		    emit_header(xml);
		    emit_warning(xml);
		}
		T_TD A_VALIGN("top") T_DIV A_ALIGN("right") emit_logo(xml);
	    }
	}
	T_P A_STYLE("height:0px") TEXT_T(NBSP);
	T_TABLE T_TR {
	    T_TD A_VALIGN("top") A_WIDTH("70%") T_ROUNDED_DIV emit_your_execution(xml);
	    T_TD A_VALIGN("top") {
		T_ROUNDED_DIV emit_what(xml);
		T_P A_STYLE("height: 0px") TEXT_T(NBSP);
		T_ROUNDED_DIV emit_resources(xml);
	    }
	}
#else
	emit_warning(xml);

	T_TABLE T_TR {
	    T_TD A_VALIGN("top") A_WIDTH("70%") {
		T_ROUNDED_DIV emit_header(xml);
		T_P A_STYLE("height:0px") TEXT_T(NBSP);
		T_ROUNDED_DIV emit_your_execution(xml);
	    }
	    T_TD A_VALIGN("top") {
		T_DIV A_ALIGN("right")
		    emit_logo(xml);
		T_P A_STYLE("height:90px") TEXT_T(NBSP);
		T_ROUNDED_DIV emit_what(xml);
		T_P A_STYLE("height:0px") TEXT_T(NBSP);
		T_ROUNDED_DIV emit_resources(xml);
	    }
	}
#endif
    }
}

// ----------------------------------------------------------------------

void
html_front_t::emit_header(xml_file_t &xml)
{
    T_DIV A_CLASS("explanation") {
	T_H3 TEXT(acumem_slowspotter_tm);

	T_P TEXT(acumem_slowspotter_tm +
		 " is a tool to quickly analyze an application "
		 "for a range of performance problems, particularly related "
		 "to multicore optimization.");

	stringstream explanation;
	xml_file_t x2(explanation, Doctype_None);
	XML_OBJECT(x2) NL {
	    T_P {
		TEXT(
		    "A modern processor has two types of resources that "
		    "are both "
		    "limited and highly responsible for "
		    "application performance:"
		    );
		T_UL {
		    T_LI {
			T_STRONG TEXT("The memory bus");
			TEXT(
			    ", which transports data between the memory "
			    "and the processor, and");
		    }
		    T_LI {
			T_STRONG TEXT("The caches");
			TEXT(", which store often used "
			    "data so each access doesn't have to go all "
			    "the way to "
			     "memory.");
		    }
		}
	    }

	    T_P {
		TEXT("A third resource, ");
		T_STRONG TEXT("the hardware prefetcher");
		TEXT(
		    ", is also important "
		    "for application performance, as it anticipates the "
		    "application's memory access patterns, and tries to fetch "
		    "data ahead of time. If it is successful, this hides "
		    "some of "
		    "the time it takes to access memory.");
	    }

	    T_P
		TEXT(
		    "It is of great importance for application performance "
		    "and scalability on multicore processors that these "
		    "resources are used efficiently.");

	    T_P TEXT(
		acumem_slowspotter_tm + " looks for abuse of these resources "
		" and points out the responsible slowspots.");
	}

	link_section_t links;
	links.add(new link_inline_t(explanation.str(), read_more));
	links.add(new help_link_t(html_context, HELP_INTRO, see_manual));
	links.emit(xml);

	T_FORM A_ACTION("main.html")
	    T_DIV A_ALIGN("center")
	    T_INPUT A_TYPE("submit") A_VALUE("Open the Report");
	T_P T_BR;
    }
}

// ----------------------------------------------------------------------

void
html_front_t::emit_logo(xml_file_t &xml)
{
    T_IMG
	A_SRC(html_context.link_factory->image_factory()->static_image_path(PNG_LOGO_SMALL))
	A_ALIGN("right") A_VALIGN("top");
}

// ----------------------------------------------------------------------------


void
html_front_t::emit_your_execution(xml_file_t &xml)
{
    verdict_cptr_t verdict = app_verdict.get_verdict();

    bool any_problem = false;

    T_DIV A_CLASS("execution") {
	T_H3 TEXT("Your application");

	T_P {
	    T_STRONG TEXT("Application:");

	    BOOST_FOREACH(const string &s, html_context.report->argv)
		TEXT(s);
	}

	T_TABLE {
	    T_TR NL {
		bool bw_problem = app_verdict.has_bandwidth_advice &&
		    app_verdict.has_bandwidth_problem();
		any_problem = any_problem || bw_problem;
		T_TD emit_problem_gauge(xml, PNG_BANDWIDTH_GAUGE);

		T_TD {
		    T_P T_STRONG TEXT("Memory Bandwidth");
		    T_P {
			TEXT("The memory bus transports data between the main "
			     "memory and the processor. The capacity of the "
			     "memory "
			     "bus is limited. Abuse of this resource limits "
			     "application scalability.");

			T_BR;
			help_link_t(html_context, HELP_BW,
				    see_manual + ": Bandwidth").emit(xml);
		    }
		}
	    }
	    T_TR NL {
		bool lat_problem = app_verdict.has_latency_advice &&
		    app_verdict.has_latency_problem();
		any_problem = any_problem || lat_problem;
		T_TD emit_problem_gauge(xml, PNG_LATENCY_GAUGE);
		T_TD {
		    T_P T_STRONG TEXT("Memory Latency");
		    T_P {
			TEXT("The regularity of the application's memory "
			     "accesses affects the efficiency of the hardware "
			     "prefetcher. "
			     "Irregular accesses causes cache misses, which "
			     "forces the processor to wait a lot "
			     "for data to arrive.");
			T_BR;
			help_link_t(html_context, HELP_MISS_RATIO,
				    see_manual + ": Cache misses").emit(xml);
			TEXT(" ");
			help_link_t(html_context, HELP_PREFETCH,
				    see_manual + ": Prefetching").emit(xml);
		    }
		}
	    }
	    T_TR NL {
		bool loc_problem = app_verdict.has_locality_advice &&
		    app_verdict.has_locality_problem();
		any_problem = any_problem || loc_problem;
		T_TD emit_problem_gauge(xml, PNG_LOCALITY_GAUGE);
		T_TD {
		    T_P T_STRONG TEXT("Data Locality");
		    T_P {
			TEXT("Failure to pay attention to data locality has "
			     "several negative effects. Caches will be filled "
			     "with unused data, and the memory bandwidth "
			     "will waste transporting unused data.");
			T_BR;
			help_link_t(html_context, HELP_LOCALITY,
				    see_manual + ": Locality").emit(xml);
		    }
		}
	    }
	    if (html_context.feature_mt_advice)
		T_TR NL {
		    bool mt_problem = app_verdict.has_mt_advice;
		    any_problem = any_problem || mt_problem;
		    T_TD emit_problem_gauge(xml, PNG_THREADING_GAUGE);
		    T_TD {
			T_P T_STRONG TEXT("Thread Communication / Interaction");
			T_P {
			    TEXT("Several threads contending over ownership of "
				 "data in their respective caches causes the "
				 "different processor cores to stall.");
			    T_BR;
			    help_link_t(html_context, HELP_THREADING,
					see_manual + ": Multithreading").emit(xml);
			}
		    }
		}
	}

	T_P TEXT("This means that your application shows opportunities to:");
	T_P T_STRONG TEXT(verdict->verdict());

	link_section_t links;
	verdict->add_links(links);
	links.emit(xml);

	if (!any_problem && app_verdict.has_issues)
	    T_P {
		TEXT("Despite the program not showing any symptoms of major "
		     "problems, there are some minor issues summarized in the "
		     "report.");
	    }
    }
}

void
html_front_t::emit_what(xml_file_t &xml)
{
    T_DIV A_CLASS("what") {
	T_H3 TEXT("Next Steps");
	T_P TEXT("The prepared report is divided into sections.");
	T_UL {
	    T_LI {
		TEXT("Select the tab ");
		T_STRONG TEXT("Summary");
		TEXT(" to see global statistics for the entire application.");
	    }

	    T_LI NL {
		TEXT("Select the tabs ");
		T_STRONG TEXT("Bandwidth Issues");
		if (html_context.feature_mt_advice)
		    TEXT(", ");
		else
		    TEXT(" and ");
		T_STRONG TEXT("Latency Issues");
		if (html_context.feature_mt_advice) {
		    TEXT(" and ");
		    T_STRONG TEXT("MT Issues");
		    TEXT(" to browse through the detected problems.");
		}
	    }

	    T_LI {
		TEXT("Select the tab ");
		T_STRONG TEXT("Loops");
		TEXT(" to browse through statistics and detected "
		     "problems loop by loop.");
	    }
	}

	T_P TEXT("The Issue and Source windows contain details and annotated "
		 "source code for the detected problems.");

	T_P A_ALIGN("center") T_IMG A_WIDTH("200") A_BORDER("1")
	    A_SRC(html_context.link_factory->get_help_link(
		      "images/report_layout_html.png"));
    }
}

// ----------------------------------------------------------------------

void
html_front_t::emit_resources(xml_file_t &xml)
{
    T_DIV A_CLASS("resources") {
	T_H3 TEXT("Resources");
	T_P T_STRONG TEXT("Manual");
	{
	    link_section_t links;
	    links.add(new help_link_t(html_context, HELP_MANUAL, "Table of Contents"));
	    links.add(new help_link_t(html_context, HELP_OVERVIEW, "Overview"));
	    links.add(new help_link_t(html_context, HELP_WORKFLOW, "Optimization Workflow"));
	    links.add(new help_link_t(html_context, HELP_CONCEPTS, "Concepts"));
	    links.add(new help_link_t(html_context, HELP_READING_REPORT, "Reading the Report"));
	    links.add(new help_link_t(html_context, HELP_ISSUE_REF, "Issue Reference"));
	    links.emit(xml, "", 2);
	}
	T_P T_STRONG TEXT("ParaTools Web Site");
	{
	    link_section_t links;
	    links.add(new link_t("http://www.paratools.com", "ParaTools Web Site"));
	    links.add(new link_t("http://www.paratools.org/threadspotter", "ThreadSpotter"));
	    links.emit(xml, "", 2);
	}
    }
}

// ----------------------------------------------------------------------

void
html_front_t::emit_problem_gauge(xml_file_t &xml, string png_name)
{
    // Calculate the path to the image.
    string image_path =
	html_context.link_factory->image_factory()->global_gauge_path(png_name);

    // Emit a reference to the image in the HTML file.
    A_WIDTH("80px") A_VALIGN("top") T_IMG A_SRC(image_path);

}

END_NAMESPACE;
END_NAMESPACE;
