/// @file tools/acurate/html_fusion.cc

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#include "fusion.hh"
#include "macros.hh"
#include "help_links.hh"
#include "file_png.hh"

#include "../diagram.hh"
#include "../file_repos.hh"
#include "../link_factory.hh"
#include "../loop_info.hh"
#include "../xml_file.hh"

#include <sstream>
#include <iomanip>
#include <set>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace std;
using namespace boost;
using namespace acumem::analysis;

// ----------------------------------------------------------------------

class fusion_reporter_t : public location_reporter_t {
public:
    fusion_reporter_t(html_fusion_t &parent,
                      int role,
                      int issue_number)
	: location_reporter_t(issue_number),
	  parent(parent),
          role(role)
	{
	}

    int get_issue_id() const {
        return role == 0 ? issue_number : -issue_number;
    }

    void report(xml_file_t &xml) const {
	parent.link_with_role(xml, role);
    }

    bool force() const {
        return true;
    }

    const html_fusion_t &parent;
    /// 0 for the first loop, 1 for the second loop.
    const int role;
};

// ----------------------------------------------------------------------

html_fusion_t::html_fusion_t(html_context_t &html_context,
			     rpt_issue_id_t issue_id)
    : html_issue_t(html_context, issue_id),
      expert(!(get_issue().move_loop1_down || get_issue().move_loop2_up) ||
             get_issue().additional_pc1.size() ||
             get_issue().additional_pc2.size())
{
}

summary_type_bitset_t
html_fusion_t::summary_type() const
{
    const rpt_report_t &report(*html_context.report);
    const rpt_loop_fusion_issue_t &issue(get_issue());

    summary_type_bitset_t result;

    result.set(SUMMARY_BANDWIDTH);

    double hwpf_ratio(
        report.get_sliced_stats_global(
	    report.get_sliced_stats(
		report.get_depgroup_stats_id(
		    issue.depgroup2_id))).get_hardware_prefetch_ratio());

    result.set(SUMMARY_LATENCY, hwpf_ratio < HWPF_HIGH_THRESHOLD);

    return result;
}

string
html_fusion_t::get_icon() const
{
    string icon = PNG_FUSION;
    return html_context.link_factory->image_factory()->static_image_path(icon);
}

void
html_fusion_t::annotate_source()
{
    const rpt_report_t &report(*html_context.report);
    const file_repos_t &file_repos(*html_context.file_repos);
    const rpt_loop_fusion_issue_t &issue(get_issue());
    rpt_issue_id_t issue_id = get_issue_id();

    fr_location_info_vector_t location_pc1_info_vector =
	file_repos.get_line_info(report, issue.loop1_pcs);
    fr_location_info_vector_t location_pc2_info_vector =
	file_repos.get_line_info(report, issue.loop2_pcs);

    BOOST_FOREACH(fr_location_info_ptr_t &p, location_pc1_info_vector)
        p->add_reporter(location_reporter_ptr_t(
                            new fusion_reporter_t(*this, 0, issue_id)));

    BOOST_FOREACH(fr_location_info_ptr_t &p, location_pc2_info_vector)
        p->add_reporter(location_reporter_ptr_t(
                            new fusion_reporter_t(*this, 1, issue_id)));
}

void
html_fusion_t::emit_issue_file(xml_file_t &issue_xml)
{
    const rpt_report_t &report(*html_context.report);
    const file_repos_t &file_repos(*html_context.file_repos);
    const rpt_loop_fusion_issue_t &issue(get_issue());

    fr_location_info_vector_t location_dep1_info_vector =
	file_repos.get_depgroup_line_info(report, issue.depgroup1_id);
    fr_location_info_vector_t location_dep2_info_vector =
	file_repos.get_depgroup_line_info(report, issue.depgroup2_id);
    fr_location_info_vector_t location_loop1_info_vector =
	file_repos.get_loop_incl_line_info(report, issue.loop1_id);
    fr_location_info_vector_t location_loop2_info_vector =
	file_repos.get_loop_incl_line_info(report, issue.loop2_id);

    fr_location_info_vector_t location_pc1_info_vector =
	file_repos.get_line_info(report, issue.loop1_pcs);
    fr_location_info_vector_t location_pc2_info_vector =
	file_repos.get_line_info(report, issue.loop2_pcs);

    fr_location_info_vector_t additional_pc1_info_vector =
	file_repos.get_line_info(report, issue.additional_pc1);
    fr_location_info_vector_t additional_pc2_info_vector =
	file_repos.get_line_info(report, issue.additional_pc2);

    string loop_id_1(lexical_cast<string>(issue.loop1_id));
    string loop_id_2(lexical_cast<string>(issue.loop2_id));

    html_context.interesting_depgroups.insert(issue.depgroup2_id);
    html_context.interesting_loops.insert(issue.loop1_id);
    html_context.interesting_loops.insert(issue.loop2_id);

    XML_OBJECT(issue_xml) {
	T_DIV A_ID("sections") A_STYLE("display: none") {
            emit_issue_header(issue_xml);

	    // Title
	    SECTION(h1, "d", "Issue #" + get_display_number() + ": " +
		    get_name(),
		    get_icon(), HELP_LOOP_FUSION, false, false) {
		T_P {
		    if (issue.move_loop1_down || issue.move_loop2_up) {
			TEXT("It may be possible to merge the bodies of loop #" +
			     loop_id_1 + " and loop #" + loop_id_2 + " by");
			if (issue.move_loop1_down) {
			    if (issue.move_loop2_up) {
				TEXT("moving the first loop down or the second loop up.");
			    } else {
				TEXT("moving the first loop down.");
			    }
			} else {
			    TEXT("moving the second loop up.");
			}
		    } else {
			TEXT("Loops cannot readily be merged. There are "
			     "potential code motion barriers that need "
			     "to be removed first.");
		    }
		}

		if (issue.additional_pc1.size() ||
		    issue.additional_pc2.size()) {
		    T_P {
			T_STRONG
			    TEXT("There are also other loops that could fuse "
				 "with either the first or second loop, and "
				 "thus interfere with "
				 "this suggested fusion.");
			TEXT("This means that to conclude this fusion, "
			     "entire loops or loop "
			     "bodies need to be duplicated, once for each "
			     "separate loop combination.");
		    }
		}
	    }

	    // interesting instructions for loop 1
	    // interesting instructions for loop 2
	    // statistics for loop 2
	    // motion barriers
	    // loop relationship
	    // loop 1 statistics
	    // loop 2 statistics


	    // instruction group statistics, bottom loop
            section_statistics(
                xml,
                html_context,
                "Statistics for fusible instruction group, second loop #" + loop_id_2,
                HELP_FUSION_STAT,
                report.get_depgroup_stats_id(issue.depgroup2_id),
                "d" + lexical_cast<string>(issue.depgroup2_id),
                location_dep2_info_vector,
                location_loop2_info_vector,
                false, // initially visible
                false, // is_loop
                false, // show_avoidable
                "istat1");

	    // instruction group instructions, top loop
	    section_instructions(xml, html_context,
				 "Fusible instruction group, first loop #" +
				 loop_id_1,
				 HELP_FUSION_GROUP,
                                 location_dep1_info_vector,
                                 location_loop1_info_vector,
				 2,
				 false, // visible
				 "iinstr1");

	    // instruction group instructions, bottom loop
	    section_instructions(xml, html_context,
				 "Fusible instruction group, second loop #" +
				 loop_id_2,
				 HELP_FUSION_GROUP,
                                 location_dep2_info_vector,
                                 location_loop2_info_vector,
				 2,
				 false,
				 "iinstr2");



	    if (!issue.move_loop1_down) {
		section_motion_barrier(xml, html_context,
				       "Potential barriers for moving the "
				       "body of the first loop #" + loop_id_1,
				       HELP_FUSION_BARRIER,
				       issue.loop1_down_pcs,
				       issue.loop1_down_blocks,
				       location_dep1_info_vector,
				       "imot1");
	    }

	    if (!issue.move_loop2_up) {
		section_motion_barrier(xml, html_context,
				       "Potential barriers for moving the "
				       "body of the second loop #" + loop_id_2,
				       HELP_FUSION_BARRIER,
				       issue.loop2_up_pcs,
				       issue.loop2_up_blocks,
				       location_dep2_info_vector,
				       "imot2");
	    }

	    if (issue.additional_pc1.size()) {

		section_instructions(xml, html_context,
				     "Other instructions that also could fuse "
				     "with second loop #" +
				     loop_id_2,
				     HELP_FUSION_EXPERT,
				     additional_pc1_info_vector,
				     location_loop2_info_vector,
				     2,
				     false,
				     "iadd2");
	    }

	    if (issue.additional_pc2.size()) {
		section_instructions(xml, html_context,
				     "Other instructions that also could fuse "
				     "with first loop #" +
				     loop_id_1,
				     HELP_FUSION_EXPERT,
				     additional_pc2_info_vector,
				     location_loop1_info_vector,
				     2,
				     false,
				     "iadd1");
	    }

	    // -----------------------------------------------------
	    // Common section for describing the Loops

	    all_loop_sections(xml, html_context,
			      issue.loop1_id, location_loop1_info_vector,
			      ", first loop #" + loop_id_1);

	    all_loop_sections(xml, html_context,
			      issue.loop2_id, location_loop2_info_vector,
			      ", second loop #" + loop_id_2, "2");

            emit_issue_footer(html_context, issue_xml);

	}
    }
}

void
html_fusion_t::emit_summary(xml_file_t &xml,
			    summary_context_t &summary_context,
			    summary_type_t summary_type)
{
    const rpt_report_t &report(*html_context.report);
    const file_repos_t &file_repos(*html_context.file_repos);
    const rpt_loop_fusion_issue_t &issue(get_issue());

    fr_location_info_vector_t location_to_dep =
	file_repos.get_depgroup_line_info(report, issue.depgroup2_id);
    fr_location_info_vector_t location_to_loop =
        file_repos.get_loop_excl_line_info(report, issue.loop2_id);

    add_summary_line(
        xml,
        summary_context,
        get_rel_file_name(),
        get_name(),
        get_icon(),
        location_to_dep,
        location_to_loop,
        report.get_depgroup_stats_id(issue.depgroup2_id),
        lexical_cast<string>(issue.loop2_id),
        summary_type);
}

void
html_fusion_t::link_with_role(xml_file_t &xml, int role) const
{
    T_A A_HREF(get_rel_file_name()) A_TARGET("info") {
	string type;
	string icon;
	switch (role) {
        case 0:
	    type = "1st loop";
	    icon = PNG_FUSION1;
            break;
        case 1:
	    type = "2nd loop";
	    icon = PNG_FUSION2;
            break;
        default:
            assert(0);
	    while(1);
	}

	icon =
	    html_context.link_factory->image_factory()->static_image_path(icon);

	string title = get_name() + " (" + type +
	    "), issue: " + get_display_number();
	emit_link_to_issue_no_source(xml, title, icon, *this, false);
    }
}

END_NAMESPACE;
END_NAMESPACE;
