/// @file tools/acurate/html_issue.cc

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#include "issue.hh"

#include "file_png.hh"
#include "help_links.hh"
#include "macros.hh"

#include "../link_factory.hh"
#include "../loop_info.hh"
#include "../product_data.hh"
#include "../xml_file.hh"

#include "library/license/license.hh"

#include <boost/lexical_cast.hpp>
#include <boost/filesystem.hpp>

#include <sstream>
#include <iomanip>
#include <time.h>
#ifndef _WIN32
#include <langinfo.h>
#endif

#include <cxxabi.h>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace std;
using namespace boost;
namespace fs = boost::filesystem;

// ----- emit_thread_popup_value ----------------------------------------------

#define TP_ACCESSES      "a"
#define TP_PERC_FETCHES  "f"
#define TP_PERC_MISSES   "m"
#define TP_PERC_BW       "bw"
#define TP_PERC_UPGRADE  "u"
#define TP_PERC_WB       "w"
#define TP_PERC_COMM     "c"
#define TP_UTILIZATION   "ut"
#define TP_FETCHUTIL     "fu"
#define TP_WBUTIL        "wu"
#define TP_HWPF          "hw"
#define TP_RANDOMNESS    "r"
#define TP_UPGRADE_RATIO "ur"
#define TP_UPGRADES      "up"
#define TP_COMM_RATIO    "cr"
#define TP_COMM_UTIL     "cu"
#define TP_FALSE_SHARING "fs"

/// Presents accesses.
class tp_accesses_t : public tp_field_t {
public:
    tp_accesses_t(double dummy = 0)
        {}

    string heading() const {
        return "Accesses";
    }

    string id() const {
        return "absolute";
    }

    bool total() const {
	return true;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        html_util_t::emit_scientific(xml, stats.get_accesses(), 0, false, true);
    }

    virtual string class_string() const {
	return TP_ACCESSES;
    }
};

/// Presents percentage of total fetches.
class tp_percentage_of_fetches_t : public tp_field_t {
    double total_fetches;

public:
    tp_percentage_of_fetches_t(double total_fetches)
        : total_fetches(total_fetches)
        {}

    string heading() const {
        return "% of fetches";
    }

    string id() const {
        return "percentage";
    }

    bool total() const {
	return true;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        html_util_t::emit_scientific(
            xml, stats.get_total_fetches(), total_fetches, true, false);
    }

    virtual string class_string() const { return TP_PERC_FETCHES; }
    virtual double value() const { return total_fetches; }
};

/// Presents percentage of total misses.
class tp_percentage_of_misses_t : public tp_field_t {
    double total_misses;

public:
    tp_percentage_of_misses_t(double total_misses)
        : total_misses(total_misses)
        {}

    string heading() const {
        return "% of misses";
    }

    string id() const {
        return "percentage";
    }

    bool total() const {
	return true;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        html_util_t::emit_scientific(
            xml, stats.get_total_misses(), total_misses, true, false);
    }

    virtual string class_string() const { return TP_PERC_MISSES; }
    virtual double value() const { return total_misses; }
};

/// Presents percentage of total bandwidth.
class tp_percentage_of_bandwidth_t : public tp_field_t {
    double total_bandwidth;

public:
    tp_percentage_of_bandwidth_t(double total_fetches,
				 double total_write_backs)
        : total_bandwidth(total_fetches + total_write_backs)
        {}

    tp_percentage_of_bandwidth_t(double total_bandwidth)
	: total_bandwidth(total_bandwidth)
	{}

    string heading() const {
        return "% of bandwidth";
    }

    string id() const {
        return "percentage";
    }

    bool total() const {
	return true;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
	const double wbs(stats.get_total_fetches() + stats.get_total_write_backs());

        html_util_t::emit_scientific(xml, wbs, total_bandwidth, true, false);
    }
    virtual string class_string() const { return TP_PERC_BW; }
    virtual double value() const { return total_bandwidth;}
};

/// Presents percentage of total upgrades.
class tp_percentage_of_upgrades_t : public tp_field_t {
    double total_upgrades;

public:
    tp_percentage_of_upgrades_t(double total_upgrades)
        : total_upgrades(total_upgrades)
        {}

    string heading() const {
        return "% of upgrades";
    }

    string id() const {
        return "percentage";
    }

    bool total() const {
	return true;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        html_util_t::emit_scientific(
            xml, stats.get_total_upgrades(), total_upgrades, true, false);
    }
    virtual string class_string() const { return TP_PERC_UPGRADE; }
    virtual double value() const { return total_upgrades; }
};

/// Presents percentage of total write_backs.
class tp_percentage_of_write_backs_t : public tp_field_t {
    double total_write_backs;

public:
    tp_percentage_of_write_backs_t(double total_write_backs)
        : total_write_backs(total_write_backs)
        {}

    string heading() const {
        return "% of write-backs";
    }

    string id() const {
        return "percentage";
    }

    bool total() const {
	return true;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        html_util_t::emit_scientific(
            xml, stats.get_total_write_backs(), total_write_backs, true, false);
    }
    virtual string class_string() const { return TP_PERC_WB; }
    virtual double value() const { return total_write_backs; }
};

/// Presents percentage of total communication.
class tp_percentage_of_comm_t : public tp_field_t {
    double total_comm;

public:
    tp_percentage_of_comm_t(double total_comm)
        : total_comm(total_comm)
        {}

    string heading() const {
        return "% of comm";
    }

    string id() const {
        return "percentage";
    }

    bool total() const {
	return true;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        html_util_t::emit_scientific(
            xml, stats.get_communicated_lines(), total_comm, true, false);
    }
    virtual string class_string() const { return TP_PERC_COMM; }
    virtual double value() const { return total_comm; }
};

/// Presents the old utilization value.
class tp_utilization_t : public tp_field_t {
public:
    tp_utilization_t(double dummy = 0)
        {}

    string heading() const {
        return "Utilization";
    }

    string id() const {
        return "percentage";
    }

    bool total() const {
	return false;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        html_util_t::emit_percent(xml, stats.get_spatial_utilization());
    }

    virtual string class_string() const { return TP_UTILIZATION; }
};

/// Presents fetch utilization.
class tp_fetch_utilization_t : public tp_field_t {
public:
    tp_fetch_utilization_t(double dummy = 0)
        {}

    string heading() const {
        return "Fetch utilization";
    }

    string id() const {
        return "percentage";
    }

    bool total() const {
	return false;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        html_util_t::emit_percent(xml, stats.get_fetch_utilization());
    }

    virtual string class_string() const { return TP_FETCHUTIL; }
};

/// Presents wb utilization.
class tp_wb_utilization_t : public tp_field_t {
public:
    tp_wb_utilization_t(double dummy = 0)
        {}

    string heading() const {
        return "Write-back utilization";
    }

    string id() const {
        return "percentage";
    }

    bool total() const {
	return false;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        html_util_t::emit_percent(xml, stats.get_write_back_utilization());
    }

    virtual string class_string() const { return TP_WBUTIL; }
};

/// Presents hardware prefetch probability.
class tp_hw_prefetch_t : public tp_field_t {
public:
    tp_hw_prefetch_t(double dummy = 0)
        {}

    string heading() const {
        return "HW prefetch probability";
    }

    string id() const {
        return "percentage";
    }

    bool total() const {
	return false;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        html_util_t::emit_percent(xml, stats.get_hardware_prefetch_ratio());
    }

    virtual string class_string() const { return TP_HWPF; }
};

/// Presents randomness.
class tp_randomness_t : public tp_field_t {
public:
    tp_randomness_t(double dummy = 0)
        {}

    string heading() const {
        return "Access randomness";
    }

    string id() const {
        return "text";
    }

    bool total() const {
	return false;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        if (stats.get_random_ratio() < html_util_t::RANDOM_LOW_THRESHOLD)
            TEXT("Low");
        else if (stats.get_random_ratio() < html_util_t::RANDOM_HIGH_THRESHOLD)
            TEXT("High");
        else
            TEXT("Very high");
    }
    virtual string class_string() const { return TP_RANDOMNESS; }
};

/// Presents upgrade ratio.
class tp_upgrade_ratio_t : public tp_field_t {
public:
    tp_upgrade_ratio_t(double dummy = 0)
        {}

    string heading() const {
        return "Upgrade ratio";
    }

    string id() const {
        return "percentage";
    }

    bool total() const {
	return false;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        html_util_t::emit_percent(xml, stats.get_upgrade_ratio());
    }

    virtual string class_string() const { return TP_UPGRADE_RATIO; }
};

/// Presents upgrades
class tp_upgrades_t : public tp_field_t {
public:
    tp_upgrades_t(double dummy = 0)
        {}

    string heading() const {
        return "Upgrades";
    }

    string id() const {
        return "absolute";
    }

    bool total() const {
	return true;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        html_util_t::emit_scientific(xml, stats.get_upgrades(), 0, false, true);
    }

    virtual string class_string() const { return TP_UPGRADES; }
};

/// Presents communication ratio.
class tp_comm_ratio_t : public tp_field_t {
public:
    tp_comm_ratio_t(double dummy = 0)
        {}

    string heading() const {
        return "Comm. ratio";
    }

    string id() const {
        return "percentage";
    }

    bool total() const {
	return false;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        html_util_t::emit_percent(xml,
				  stats.get_coherence_fetch_ratio() +
				  stats.get_upgrade_ratio());
    }

    virtual string class_string() const { return TP_COMM_RATIO; }
};

/// Presents communication utilization.
class tp_comm_utilization_t : public tp_field_t {
public:
    tp_comm_utilization_t(double dummy = 0)
        {}

    string heading() const {
        return "Comm. utilization";
    }

    string id() const {
        return "percentage";
    }

    bool total() const {
	return false;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        html_util_t::emit_percent(xml, stats.get_communication_utilization());
    }
    virtual string class_string() const { return TP_COMM_UTIL; }
};


/// Presents false sharing ratio
class tp_fs_ratio_t : public tp_field_t {
public:
    tp_fs_ratio_t(double dummy = 0)
        {}

    string heading() const {
        return "F-S. ratio";
    }

    string id() const {
        return "percentage";
    }

    bool total() const {
	return false;
    }

    void emit(xml_file_t &xml, const rpt_stats_t &stats) const {
        html_util_t::emit_percent(xml, stats.get_false_sharing_ratio());
    }
    virtual string class_string() const { return TP_FALSE_SHARING; }
};

template <typename T>
tp_field_ptr_t
create_tp_field(double v)
{
    return tp_field_ptr_t(new T(v));
}

struct create_tp_field_record_t {
    const char *tag;
    tp_field_ptr_t (*create_func)(double);
};

static const create_tp_field_record_t tp_field_funcs[] = {
    {TP_ACCESSES, 	&create_tp_field<tp_accesses_t>},
    {TP_PERC_FETCHES,	&create_tp_field<tp_percentage_of_fetches_t>},
    {TP_PERC_MISSES,	&create_tp_field<tp_percentage_of_misses_t>},
    {TP_PERC_BW,	&create_tp_field<tp_percentage_of_bandwidth_t>},
    {TP_PERC_UPGRADE,	&create_tp_field<tp_percentage_of_upgrades_t>},
    {TP_PERC_WB,	&create_tp_field<tp_percentage_of_write_backs_t>},
    {TP_PERC_COMM,	&create_tp_field<tp_percentage_of_comm_t>},
    {TP_UTILIZATION,	&create_tp_field<tp_utilization_t>},
    {TP_FETCHUTIL,	&create_tp_field<tp_fetch_utilization_t>},
    {TP_WBUTIL,		&create_tp_field<tp_wb_utilization_t>},
    {TP_HWPF,		&create_tp_field<tp_hw_prefetch_t>},
    {TP_RANDOMNESS,	&create_tp_field<tp_randomness_t>},
    {TP_UPGRADE_RATIO,	&create_tp_field<tp_upgrade_ratio_t>},
    {TP_UPGRADES,	&create_tp_field<tp_upgrades_t>},
    {TP_COMM_RATIO,	&create_tp_field<tp_comm_ratio_t>},
    {TP_COMM_UTIL,	&create_tp_field<tp_comm_utilization_t>},
    {TP_FALSE_SHARING,	&create_tp_field<tp_fs_ratio_t>},
    {NULL, NULL}
};

tp_field_ptr_t
tp_field_t::create(const string &tag, double value)
{
    const create_tp_field_record_t *p = tp_field_funcs;
    while (p->tag != NULL) {
	if (tag == p->tag)
	    return p->create_func(value);
	p++;
    }

    return tp_field_ptr_t();
}

string
tp_field_t::code() const
{
    return "'" + class_string() + "'," + lexical_cast<string>(value());
}

void
html_util_t::emit_thread_popup_table(html_context_t &html_context,
				     xml_file_t &xml,
				     rpt_sliced_stats_id_t stats_id,
				     const tp_field_t &field)
{
    const rpt_report_t &report(*html_context.report);
    const rpt_sliced_stats_t &stats = report.get_sliced_stats(stats_id);

    T_TABLE A_CLASS("summary") {
	T_COLGROUP {
	    T_COL;
	    T_COL;
	}
	T_THEAD {
	    T_TR {
		T_TH TEXT("Thread Id");
		T_TH TEXT(field.heading());
	    }
	}
	T_TBODY {
	    bool alt = false;
	    if (stats.thread_stats.size() > 1 || !html_context.feature_mt_advice)
		T_TR {
		    alt = !alt;
		    T_TD if (field.total())
			TEXT("Thread Total");
		    else
			TEXT("Thread Average");
		    T_TD A_ID(field.id()) field.emit(xml, report.get_sliced_stats_global(stats));
		}
	    if (html_context.feature_mt_advice) {
		BOOST_FOREACH(const rpt_thread_stats_map_t::value_type &v,
			      stats.thread_stats) {
		    T_TR {
			if (alt) A_CLASS("alternate");
			alt = !alt;

			T_TD html_util_t::emit_thread(html_context, xml, v.first);
			T_TD A_ID(field.id())
			    field.emit(xml, report.get_sliced_stats_thread(stats_id, v.first));
		    }
		}
	    }
	}
    }
}

/// Emit a value and an associated popup window with per-thread values.
void
html_util_t::emit_thread_popup_value(html_context_t &html_context,
				     xml_file_t &xml,
				     rpt_sliced_stats_id_t stats_id,
				     const tp_field_t &field,
				     bool left)
{
    html_context.link_factory->emit_sliced_statistics_link(xml, html_context,
							   stats_id, field);
}


// ----- emit_thread_popup_value_by_kind --------------------------------------


#define TPBK_FETCH_RATIO "fr"
#define TPBK_MISS_RATIO  "mr"
#define TPBK_WB_RATIO    "wr"

#define TPBK_FETCHES     "f"
#define TPBK_MISSES      "m"
#define TPBK_WRITE_BACKS "w"


/// Presents fetch ratio.
class tp_fetch_ratio_by_kind_t : public tp_field_by_kind_t {
public:
    tp_fetch_ratio_by_kind_t()
        {}

    string heading() const {
        return "Fetch ratio";
    }

    string id() const {
        return "percentage";
    }

    by_kind_value_t by_kind(const rpt_stats_t &stats) const {
        return by_kind_value_t(
            stats.get_cold_fetch_ratio(),
            stats.get_replacement_fetch_ratio(),
            stats.get_coherence_fetch_ratio(),
            stats.get_flush_fetch_ratio());
    }

    bool total() const {
	return false;
    }

    void emit(xml_file_t &xml, double value) const {
        html_util_t::emit_percent(xml, value);
    }

    string class_string() const { return TPBK_FETCH_RATIO; }
};

/// Presents miss ratio.
class tp_miss_ratio_by_kind_t : public tp_field_by_kind_t {
public:
    tp_miss_ratio_by_kind_t()
        {}

    string heading() const {
        return "Miss ratio";
    }

    string id() const {
        return "percentage";
    }

    by_kind_value_t by_kind(const rpt_stats_t &stats) const {
        return by_kind_value_t(
            stats.get_cold_miss_ratio(),
            stats.get_replacement_miss_ratio(),
            stats.get_coherence_miss_ratio(),
            stats.get_flush_miss_ratio());
    }

    bool total() const {
	return false;
    }

    void emit(xml_file_t &xml, double value) const {
        html_util_t::emit_percent(xml, value);
    }

    string class_string() const { return TPBK_MISS_RATIO; }
};


/// Presents write-back ratio.
class tp_write_back_ratio_by_kind_t : public tp_field_by_kind_t {
public:
    tp_write_back_ratio_by_kind_t()
        {}

    string heading() const {
        return "Write-back ratio";
    }

    string id() const {
        return "percentage";
    }

    by_kind_value_t by_kind(const rpt_stats_t &stats) const {
        return by_kind_value_t(
            stats.get_cold_write_back_ratio(),
            stats.get_replacement_write_back_ratio(),
            stats.get_coherence_write_back_ratio(),
            stats.get_flush_write_back_ratio());
    }

    bool total() const {
	return false;
    }

    void emit(xml_file_t &xml, double value) const {
        html_util_t::emit_percent(xml, value);
    }
    string class_string() const { return TPBK_WB_RATIO; }
};

/// Presents number of fetches.
class tp_fetches_by_kind_t : public tp_field_by_kind_t {
public:
    tp_fetches_by_kind_t()
        {}

    string heading() const {
        return "Fetches";
    }

    string id() const {
        return "absolute";
    }

    by_kind_value_t by_kind(const rpt_stats_t &stats) const {
        return by_kind_value_t(
            stats.get_cold_fetches(),
            stats.get_replacement_fetches(),
            stats.get_coherence_fetches(),
            stats.get_flush_fetches());
    }

    bool total() const {
	return true;
    }

    void emit(xml_file_t &xml, double value) const {
        html_util_t::emit_scientific(xml, value, 0, false, true);
    }

    string class_string() const { return TPBK_FETCHES; }
};

/// Presents number of misses.
class tp_misses_by_kind_t : public tp_field_by_kind_t {
public:
    tp_misses_by_kind_t()
        {}

    string heading() const {
        return "Misses";
    }

    string id() const {
        return "absolute";
    }

    by_kind_value_t by_kind(const rpt_stats_t &stats) const {
        return by_kind_value_t(
            stats.get_cold_misses(),
            stats.get_replacement_misses(),
            stats.get_coherence_misses(),
            stats.get_flush_misses());
    }

    bool total() const {
	return true;
    }

    void emit(xml_file_t &xml, double value) const {
        html_util_t::emit_scientific(xml, value, 0, false, true);
    }

    string class_string() const { return TPBK_MISSES; }
};

/// Presents number of write-backs.
class tp_write_backs_by_kind_t : public tp_field_by_kind_t {
public:
    tp_write_backs_by_kind_t()
        {}

    string heading() const {
        return "Write-backs";
    }

    string id() const {
        return "absolute";
    }

    by_kind_value_t by_kind(const rpt_stats_t &stats) const {
        return by_kind_value_t(
            stats.get_cold_write_backs(),
            stats.get_replacement_write_backs(),
            stats.get_coherence_write_backs(),
            stats.get_flush_write_backs());
    }

    bool total() const {
	return true;
    }

    void emit(xml_file_t &xml, double value) const {
        html_util_t::emit_scientific(xml, value, 0, false, true);
    }

    string class_string() const { return TPBK_WRITE_BACKS; }
};


template <typename T>
tp_field_by_kind_ptr_t
create_tp_field_by_kind()
{
    return tp_field_by_kind_ptr_t(new T);
}

struct create_tp_field_by_kind_record_t {
    const char *tag;
    tp_field_by_kind_ptr_t (*create_func)();
};

static const create_tp_field_by_kind_record_t tp_field_by_kind_funcs[] = {
    {TPBK_FETCH_RATIO, 	&create_tp_field_by_kind<tp_fetch_ratio_by_kind_t>},
    {TPBK_MISS_RATIO, 	&create_tp_field_by_kind<tp_miss_ratio_by_kind_t>},
    {TPBK_WB_RATIO, 	&create_tp_field_by_kind<tp_write_back_ratio_by_kind_t>},
    {TPBK_FETCHES, 	&create_tp_field_by_kind<tp_fetches_by_kind_t>},
    {TPBK_MISSES, 	&create_tp_field_by_kind<tp_misses_by_kind_t>},
    {TPBK_WRITE_BACKS, 	&create_tp_field_by_kind<tp_write_backs_by_kind_t>},
    {NULL, NULL}
};

tp_field_by_kind_ptr_t
tp_field_by_kind_t::create(const string &tag)
{
    const create_tp_field_by_kind_record_t *p = tp_field_by_kind_funcs;
    while (p->tag != NULL) {
	if (tag == p->tag)
	    return p->create_func();
	p++;
    }

    return tp_field_by_kind_ptr_t();
}



string
tp_field_by_kind_t::code() const
{
    return "'" + class_string() + "'";
}

void
html_util_t::emit_thread_popup_table_by_kind(html_context_t &html_context,
					     xml_file_t &xml,
					     rpt_sliced_stats_id_t stats_id,
					     const tp_field_by_kind_t &field)
{
    const rpt_report_t &report(*html_context.report);
    const rpt_sliced_stats_t &stats = report.get_sliced_stats(stats_id);
    const by_kind_value_t total_by_kind =
        field.by_kind(report.get_sliced_stats_global(stats));

    T_TABLE A_CLASS("summary") {
	T_COLGROUP {
	    T_COL;
	    T_COL;
	}
	T_THEAD {
	    T_TR {
		T_TH TEXT("Thread Id");
		T_TH TEXT("Total " + field.heading());
		T_TH TEXT("Uncategorized");
		T_TH TEXT("Replacement");
		if (html_context.feature_mt_advice)
		    T_TH TEXT("Coherence");
		T_TH TEXT("Flush");
	    }
	}
	T_TBODY {
	    bool alt = false;

	    if (stats.thread_stats.size() > 1 || !html_context.feature_mt_advice)
		T_TR {
		    alt = !alt;

		    T_TD TEXT(field.total() ? "Thread Total" : "Thread Average");

		    T_TD A_ID(field.id()) field.emit(xml, total_by_kind.total());
		    T_TD A_ID(field.id()) field.emit(xml, total_by_kind.cold);
		    T_TD A_ID(field.id()) field.emit(xml, total_by_kind.repl);
		    if (html_context.feature_mt_advice)
			T_TD A_ID(field.id()) field.emit(xml, total_by_kind.coh);
		    T_TD A_ID(field.id()) field.emit(xml, total_by_kind.flush);
		}

	    if (html_context.feature_mt_advice) {
		BOOST_FOREACH(const rpt_thread_stats_map_t::value_type &v,
			      stats.thread_stats) {
		    T_TR {
			if (alt) A_CLASS("alternate");
			alt = !alt;
			T_TD html_util_t::emit_thread(html_context, xml, v.first);
			const by_kind_value_t thread_by_kind =
			    field.by_kind(report.get_sliced_stats_thread(stats_id, v.first));
			T_TD A_ID(field.id()) field.emit(xml, thread_by_kind.total());
			T_TD A_ID(field.id()) field.emit(xml, thread_by_kind.cold);
			T_TD A_ID(field.id()) field.emit(xml, thread_by_kind.repl);
			T_TD A_ID(field.id()) field.emit(xml, thread_by_kind.coh);
			T_TD A_ID(field.id()) field.emit(xml, thread_by_kind.flush);
		    }
		}
	    }
	}
    }
}

/// Emit a value and an associated popup window with per-thread values.
void
html_util_t::emit_thread_popup_value_by_kind(html_context_t &html_context,
					     xml_file_t &xml,
					     rpt_sliced_stats_id_t stats_id,
					     const tp_field_by_kind_t &field,
					     bool left)
{
    html_context.link_factory->emit_sliced_statistics_link_by_kind(
	xml, html_context, stats_id, field);
}

// ----- html_util_t ----------------------------------------------------------

const double html_util_t::RANDOM_LOW_THRESHOLD = 0.2;
const double html_util_t::RANDOM_HIGH_THRESHOLD = 0.5;
const double html_util_t::HWPF_LOW_THRESHOLD = 0.1;
const double html_util_t::HWPF_HIGH_THRESHOLD = 0.4;

void
html_util_t::emit_thread(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_thread_id_t thread_id)
{
    stringstream s;
    s << html_context.report->get_context_id(thread_id);
    TEXT(s.str());
}

void
html_util_t::emit_accesses(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left)
{
    const tp_accesses_t field;
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_fetch_ratio(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left)
{
    const tp_fetch_ratio_by_kind_t field;
    emit_thread_popup_value_by_kind(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_miss_ratio(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left)
{
    const tp_miss_ratio_by_kind_t field;
    emit_thread_popup_value_by_kind(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_upgrade_ratio(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left)
{
    const tp_upgrade_ratio_t field;
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_write_back_ratio(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left)
{
    const tp_write_back_ratio_by_kind_t field;
    emit_thread_popup_value_by_kind(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_fetches(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left)
{
    const tp_fetches_by_kind_t field;
    emit_thread_popup_value_by_kind(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_misses(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left)
{
    const tp_misses_by_kind_t field;
    emit_thread_popup_value_by_kind(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_upgrades(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left)
{
    const tp_upgrades_t field;
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_write_backs(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left)
{
    const tp_write_backs_by_kind_t field;
    emit_thread_popup_value_by_kind(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_percentage_of_fetches(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    rpt_sliced_stats_id_t global_stats_id,
    bool left)
{
    const rpt_sliced_stats_t &global_stats =
	html_context.report->get_sliced_stats(global_stats_id);
    const tp_percentage_of_fetches_t field(
        html_context.report->get_sliced_stats_global(global_stats).get_total_fetches());
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_percentage_of_misses(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    rpt_sliced_stats_id_t global_stats_id,
    bool left)
{
    const rpt_sliced_stats_t &global_stats =
	html_context.report->get_sliced_stats(global_stats_id);
    const tp_percentage_of_misses_t field(
        html_context.report->get_sliced_stats_global(global_stats).get_total_misses());
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_percentage_of_bandwidth(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    rpt_sliced_stats_id_t global_stats_id,
    bool left)
{
    const rpt_sliced_stats_t &global_stats =
	html_context.report->get_sliced_stats(global_stats_id);
    const tp_percentage_of_bandwidth_t field(
        html_context.report->get_sliced_stats_global(global_stats).get_total_fetches(),
        html_context.report->get_sliced_stats_global(global_stats).get_total_write_backs());
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_percentage_of_upgrades(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    rpt_sliced_stats_id_t global_stats_id,
    bool left)
{
    const rpt_sliced_stats_t &global_stats =
	html_context.report->get_sliced_stats(global_stats_id);
    const tp_percentage_of_upgrades_t field(
        html_context.report->get_sliced_stats_global(global_stats).get_total_upgrades());
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_percentage_of_write_backs(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    rpt_sliced_stats_id_t global_stats_id,
    bool left)
{
    const rpt_sliced_stats_t &global_stats =
	html_context.report->get_sliced_stats(global_stats_id);
    const tp_percentage_of_write_backs_t field(
        html_context.report->get_sliced_stats_global(global_stats).get_total_write_backs());
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_percentage_of_comm(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    rpt_sliced_stats_id_t global_stats_id,
    bool left)
{
    const rpt_sliced_stats_t &global_stats =
	html_context.report->get_sliced_stats(global_stats_id);
    const tp_percentage_of_comm_t field(
        html_context.report->get_sliced_stats_global(global_stats).get_communicated_lines());
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_utilization(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left)
{
    const tp_utilization_t field;
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_fetch_utilization(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left)
{
    const tp_fetch_utilization_t field;
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_wb_utilization(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left)
{
    const tp_wb_utilization_t field;
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_hw_prefetch_ratio(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left) {
    const tp_hw_prefetch_t field;
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_random_ratio(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left) {
    const tp_randomness_t field;
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_comm_ratio(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left) {
    const tp_comm_ratio_t field;
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_comm_utilization(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left) {
    const tp_comm_utilization_t field;
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}

void
html_util_t::emit_fs_ratio(
    html_context_t &html_context,
    xml_file_t &xml,
    rpt_sliced_stats_id_t stats_id,
    bool left) {
    const tp_fs_ratio_t field;
    emit_thread_popup_value(html_context, xml, stats_id, field, left);
}


void
html_util_t::emit_head(
    xml_file_t &xml,
    const string &title,
    const string &script_path)
{
    T_HEAD {
        T_META;
        STYLE_FILE(script_path + "summary.css");
        STYLE_FILE(script_path + "hover_popup.css");
        SCRIPT_FILE(script_path + "table.js");
        SCRIPT_FILE(script_path + "nav.js");
	if (!title.empty())
	    T_TITLE TEXT(title);
    }
}

void
html_util_t::emit_issue_header(xml_file_t &xml,
			       bool warn_unreliable,
			       bool warn_virtualized)
{
    A_CLASS("issue");
    if (warn_unreliable || warn_unreliable) {
	T_DIV A_CLASS("warning") {
	    if (warn_unreliable) T_P {
		T_B TEXT("Warning!");
		TEXT("These results are based on too few samples to be reliable.");
	    }
	    if (warn_virtualized) {
		T_B TEXT("Warning!");
		TEXT("The autodetection of cache hiearchy does not work in "
		     "virtualized environments. You need to manually override "
		     "the number of caches.");
	    }

	}
    }
}


void
html_issue_t::emit_issue_header(xml_file_t &xml)
{
    html_util_t::emit_issue_header(xml,
				   html_context.result_unreliable(),
				   html_context.warn_virtualized());

}

void
html_util_t::emit_product_name_full(xml_file_t &xml)
{
    TEXT(product_data_t::product_name_utf8());
}

void
html_util_t::emit_issue_footer(html_context_t &html_context,
                               xml_file_t &xml)
{
    emit_copyright(xml, true);
    emit_terms_of_use(html_context, xml);
}

void
html_util_t::emit_terms_of_use(html_context_t &html_context,
                               xml_file_t &xml)
{
#if 0
    string condition = html_context.license->get_conditions();
    if (condition.size()) {
	T_P A_CLASS("smaller") TEXT("Terms of use: " + condition);
    }
#endif
}

void
html_util_t::emit_copyright(xml_file_t &xml, bool small_font)
{
    T_P {
	if (small_font)
	    A_CLASS("smaller");
	TEXT("Copyright (c) 2006-2012 Rogue Wave Software, Inc. All Rights Reserved.");
	T_BR;
	TEXT("Patents pending.");
    }
}

void
html_util_t::emit_time(xml_file_t &xml, fnr_real_time_t t)
{
    struct tm local_time;
    time_t start_time = t / 1000000000ULL;
    if (start_time != 0) {
	char buf[256];
#ifdef _WIN32
        _localtime64_s(&local_time, &start_time);
	strftime(buf, sizeof(buf), "%x %X", &local_time);
#else
        localtime_r(&start_time, &local_time);
	strftime(buf, sizeof(buf), nl_langinfo(D_T_FMT), &local_time);
#endif
	TEXT(buf);
    }
}

set<int>
html_util_t::get_highlight_lines(const fr_location_info_vector_t &loop_pcs,
				 const string &filename)
{
    set<int> highlight_lines;
    BOOST_FOREACH(const fr_location_info_ptr_t &loop_loc, loop_pcs) {
	pc_info_ptr_t loop_pi = loop_loc->pc_info.lock();
	line_info_ptr_t loop_li = loop_pi->line_info.lock();
	if (loop_li.get() && loop_li->filename == filename)
	    highlight_lines.insert(loop_li->line);
    }

    return highlight_lines;
}

string
html_util_t::get_highlight(const fr_location_info_vector_t &loop_pcs,
			   const string &filename)
{
    set<int> highlight_lines = get_highlight_lines(loop_pcs, filename);
    string highlight;
    BOOST_FOREACH(int x, highlight_lines) {
        if (!highlight.empty())
            highlight += ",";
        highlight = highlight + "%27" + lexical_cast<string>(x) + "%27";
    }
    return highlight;
}

void
html_util_t::link_call_to_source(html_context_t &html_context,
				 xml_file_t &xml,
				 const pc_info_ptr_t &pc_info,
				 const fr_location_info_vector_t &loop_pcs,
				 fnr_access_classes_t access_classes)
{
    const rpt_report_t &report(*html_context.report);

    stringstream s;

    string access_class_string;

    if (access_classes != 0 ) {
	access_class_string += " [";
	if (access_classes & FNR_Access_Class_NTA)
	    access_class_string += "NT ";
	if ((access_classes & FNR_Access_Class_RD) &&
	    !(access_classes & FNR_Access_Class_PF))
	    access_class_string += "R";
	if (access_classes & FNR_Access_Class_WR)
	    access_class_string += "W";
	if (access_classes & FNR_Access_Class_PF)
	    access_class_string += "PF";

	if (access_classes & FNR_Access_Class_Flush)
	    access_class_string = " [Flush";

	access_class_string += "]";
    }

    if (pc_info) {
        const rpt_debug_pc_t &debug_pc(
            report.debug_pcs.at(pc_info->debug_pc_id));
        if (debug_pc.binary_id != RPT_NO_BINARY) {
            const rpt_binary_t &binary = report.binaries.at(debug_pc.binary_id);
            s << fs::path(binary.filename).filename();
            if (debug_pc.function_id == RPT_NO_FUNCTION &&
                debug_pc.symbol_id == RPT_NO_SYMBOL)
                s << "+0x" << hex << (pc_info->pc - binary.base) << " ";
            else
                s << "!";
        }
        if (debug_pc.function_id != RPT_NO_FUNCTION) {
            const rpt_function_t &function = report.functions.at(debug_pc.function_id);
            s << function.name << "()"
              << "+0x" << hex << (pc_info->pc - function.base) << " ";
        } else if (debug_pc.symbol_id != RPT_NO_SYMBOL) {
            const rpt_symbol_t &symbol = report.symbols.at(debug_pc.symbol_id);
            int status;
            char *demangled_symbol_name = abi::__cxa_demangle(symbol.name.c_str(), 0, 0, &status);

            s << demangled_symbol_name
              << "+0x" << hex << (pc_info->pc - symbol.base) << " ";
        }
        s << "(" << html_util_t::format_address(pc_info->pc) << ")";
	s << access_class_string;
    }

    line_info_ptr_t line_info = pc_info->line_info.lock();
    if (line_info.get()) {

	string line_number_s = lexical_cast<string>(line_info->line);
	string file_name = line_info->filename;

	// take last part of the filename
	size_t pos = file_name.find_last_of("\\/");
	if (pos != file_name.npos)
	    file_name.erase(0, pos+1);

	s << ", " << file_name
	  << ":" << line_number_s;

	string highlight = get_highlight(loop_pcs, line_info->filename);
	int file_id = report.get_source_file_id(line_info->filename);

	T_A A_HREF("javascript:load_src_mark(" +
		   lexical_cast<string>(file_id) + "," + line_number_s +
		   ",%5B" + highlight + "%5D)")

	    TEXT(s.str());
    }
    else
	TEXT(s.str());
}

void
html_util_t::emit_link_to_issue_no_source(xml_file_t &xml,
					  const string &title,
					  const string &icon,
					  const html_issue_t &issue,
					  bool display_title)
{
    T_A A_HREF(get_issue_url_no_source(issue)) {
	A_TITLE(title) T_IMG A_SRC(icon) A_BORDER("0");
	if (display_title)
	    TEXT(" " + title);
    }
}

void
html_util_t::link_to_source(html_context_t &html_context,
			    xml_file_t &xml,
			    bool show_stack,
			    const fr_location_info_ptr_t &l,
			    const fr_location_info_vector_t &loop_pcs,
			    const string &div_id)
{
    const rpt_report_t &report(*html_context.report);
    const rpt_trunc_stack_t &trunc_stack(
        report.get_location_trunc_stack(l->location_id));

    if (show_stack && !trunc_stack.empty()) {
	// output the hidden stack
	T_DIV A_ID(div_id) A_STYLE("display:none") {
	    int indent = 0;
	    BOOST_FOREACH(rpt_address_id_t pc_id, trunc_stack) {
                rpt_address_t pc(report.get_address(pc_id));
		if (indent != 0)
		    T_BR;
		pc_info_ptr_t pc_info_stack =
		    html_context.file_repos->get_pc_info(report, pc, true);
		line_info_ptr_t line_info = pc_info_stack->line_info.lock();

		for(int i=0; i<indent; i++)
		    TEXTN_T(NBSP NBSP NBSP);
		link_call_to_source(html_context, xml, pc_info_stack,
                                    loop_pcs, 0);
		indent ++;
	    }
	    T_BR;
	}
    }

    pc_info_ptr_t p = l->pc_info.lock();

    rpt_access_classes_t access_classes =
        html_context.report->get_location_access_classes(l->location_id);

    return link_call_to_source(html_context, xml, p, loop_pcs,
                               access_classes);
}


void
html_util_t::link_to_file(html_context_t &html_context,
			  xml_file_t &xml,
			  const string &filename)
{
    int file_id = html_context.report->get_source_file_id(filename);
    T_A A_HREF("javascript:load_src_mark(" +
	       lexical_cast<string>(file_id) + ",%27%27,%5B%5D)") // ,'',[]
	TEXT(filename);
}

string
html_util_t::image_path(html_context_t &html_context,
			const string &image)
{
    return html_context.link_factory->image_factory()->static_image_path(image);
}

void
html_util_t::section_head(html_context_t &html_context,
			  xml_file_t &xml,
			  const string &id,
			  const string &title,
			  bool initial_hidden,
			  bool hideable,
			  const string &icon,
			  const string &help_section)
{

    if (help_section.size()) {
	string help_toggle =
	    "toggle('" + id + "help');"
	    "toggle('" + id + "help2');";
	A_ONMOUSEOVER(help_toggle);
	A_ONMOUSEOUT(help_toggle);
    }

    if (hideable) {
	string toggle="persist_toggle('" + id + "'); toggle('" +
	    id + "minus'); toggle('" + id + "plus');";
	T_IMG A_ID(id + "plus")
	    A_SRC(image_path(html_context, PNG_PLUS)) A_ONCLICK(toggle)
	    if (!initial_hidden) A_STYLE("display: none");

	T_IMG A_ID(id + "minus")
	    A_SRC(image_path(html_context, PNG_MINUS)) A_ONCLICK(toggle)
	    if (initial_hidden) A_STYLE("display: none");

	T_SPAN A_ONCLICK(toggle) {
	    TEXT(title);
	    // could add icon here also, but h1 items only carry icons
	    // right now
	}
    } else {
	TEXT(title);
	if (icon != "") T_IMG A_SRC(icon);
    }
    if (help_section.size()) {
	string help_url =
	    html_context.link_factory->get_help_link(help_section);
	T_A A_HREF(help_url) A_TARGET("help")
	    A_ID(id + "help") A_STYLE("display:none")
	    A_ONCLICK("return popitup('" + help_url + "')")
	    T_IMG A_SRC(image_path(html_context, PNG_QUESTION))
	    A_STYLE("border: 0px; vertical-align:middle")
	    A_TITLE("Explanation");
	T_IMG A_ID(id + "help2")
	    A_SRC(image_path(html_context, PNG_QUESTION_HOLLOW))
	    A_STYLE("border: 0px; vertical-align:middle")
	    A_TITLE("Explanation");
    }
}

void
html_util_t::emit_glossary_phrase(html_context_t &html_context,
				  xml_file_t &xml,
				  const string &phrase,
				  const string &help_section)
{
    string id = xml.unique_id();
    if (help_section.size()) {
	string help_toggle =
	    "toggle('" + id + "help');"
	    "toggle('" + id + "help2');";
	A_ONMOUSEOVER(help_toggle);
	A_ONMOUSEOUT(help_toggle);
    }

    TEXT(phrase);

    if (help_section.size()) {
	string help_url =
	    html_context.link_factory->get_help_link(help_section);
	T_A A_HREF(help_url) A_TARGET("help")
	    A_ID(id + "help") A_STYLE("display:none")
	    A_ONCLICK("return popitup('" + help_url + "')")
	    T_IMG A_SRC(image_path(html_context, PNG_QUESTION))
	    A_STYLE("border: 0px; vertical-align:middle")
		A_TITLE("Explanation");
	T_IMG A_ID(id + "help2")
	    A_SRC(image_path(html_context, PNG_QUESTION_HOLLOW))
	    A_STYLE("border: 0px; vertical-align:middle")
	    A_TITLE("Explanation");
    }
}

string
html_util_t::format_address(rpt_address_t address)
{
    stringstream ss;
    ss << "0x" << hex << address;
    return ss.str();
}

static void
insert_statistics_diagrams(xml_file_t &xml,
			   html_context_t &html_context,
			   const string &diagram,
			   int rows)
{
    const rpt_report_t &report(*html_context.report);

    // Split a diagram string of d16 into d and 16. First character is
    // d or l. The number is loop_id or depgroup_id.
    string kind = diagram.substr(0,1);
    int id = lexical_cast<int>(diagram.substr(1));

    T_TD A_VALIGN("top") A_ROWSPAN(lexical_cast<string>(rows)) {
	T_TABLE A_CLASS("summary") {
	    T_TR T_TH
		html_util_t::emit_glossary_phrase(html_context, xml, "Fetch/Miss ratio", HELP_DIAGRAM_FR);
	    T_TR T_TD T_IMG
		A_SRC(html_context.link_factory->get_diagram_link("fr",
								  kind, id));
	    if (report.mt_sampled) {
		T_TR T_TH
		    html_util_t::emit_glossary_phrase(html_context, xml, "Write-back ratio", HELP_DIAGRAM_WB);
		T_TR T_TD T_IMG
		    A_SRC(html_context.link_factory->get_diagram_link("wb",
								      kind,
								      id));
	    }
	    T_TR T_TH
		html_util_t::emit_glossary_phrase(html_context, xml, "Utilization", HELP_DIAGRAM_UTIL);
	    T_TR T_TD T_IMG
		A_SRC(html_context.link_factory->get_diagram_link("su",
								  kind, id));
	}
    }
}

void
html_util_t::statistics_table(xml_file_t &xml,
			      html_context_t &html_context,
			      rpt_sliced_stats_id_t stats_id,
			      const string &diagram,
			      const fr_location_info_vector_t &instructions,
			      const fr_location_info_vector_t &loop,
			      bool is_loop,
			      bool show_avoidable)
{
    const rpt_report_t &report(*html_context.report);

    rpt_sliced_stats_id_t global_stats_id = report.get_global_stats_id();

    bool display_worst_line = !is_loop;
    int rows = (7 +
                (display_worst_line ? 1 : 0) +
                (report.mt_sampled ? (html_context.feature_mt_advice ? 9 : 4) : 0) +
                (!report.burst_sampled ? 1 : 0));
    T_TABLE A_CLASS("summary") {
	if (!report.burst_sampled) {
	    T_TR {
                T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Accesses",
							  HELP_STAT_ACCESSES);
                T_TD emit_accesses(html_context, xml, stats_id, false);
		insert_statistics_diagrams(xml, html_context, diagram, rows);
	    }
	}
	T_TR {
	    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "% of misses",
						      HELP_STAT_PERCENT_MISSES);
	    T_TD emit_percentage_of_misses(
                html_context, xml, stats_id, global_stats_id, false);
	    if (report.burst_sampled)
		insert_statistics_diagrams(xml, html_context, diagram, rows);
	}

	if (report.mt_sampled) {
	    T_TR {
		T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "% of bandwidth",
							  HELP_STAT_PERCENT_BANDWIDTH);
		T_TD emit_percentage_of_bandwidth(
		    html_context, xml, stats_id, global_stats_id, false);
	    }
	}

	T_TR {
	    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "% of fetches",
						      HELP_STAT_PERCENT_FETCHES);
	    T_TD emit_percentage_of_fetches(
                html_context, xml, stats_id, global_stats_id, false);
	}
	if (report.mt_sampled) {
            T_TR {
                T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "% of write-backs",
							  HELP_STAT_PERCENT_WRITE_BACKS);
                T_TD emit_percentage_of_write_backs(
                    html_context, xml, stats_id, global_stats_id, false);
            }

            if (html_context.feature_mt_advice)
                T_TR {
                    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "% of upgrades",
                                                              HELP_STAT_PERCENT_UPGRADES);
                    T_TD emit_percentage_of_upgrades(
                        html_context, xml, stats_id, global_stats_id, false);
                }
        }
	T_TR {
	    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Miss ratio",
						      HELP_STAT_MISS_RATIO);
	    T_TD emit_miss_ratio(html_context, xml, stats_id, false);
	}
	T_TR {
	    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Fetch ratio",
						      HELP_STAT_FETCH_RATIO);
	    T_TD emit_fetch_ratio(html_context, xml, stats_id, false);
	}

	if (report.mt_sampled) {
            T_TR {
                T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Write-back ratio",
							  HELP_STAT_WRITE_BACK_RATIO);
                T_TD emit_write_back_ratio(html_context, xml, stats_id, false);
            }
            if (html_context.feature_mt_advice) {
                T_TR {
                    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Upgrade ratio",
                                                              HELP_STAT_UPGRADE_RATIO);
                    T_TD emit_upgrade_ratio(html_context, xml, stats_id, false);
                }
                T_TR {
                    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Communication ratio",
                                                              HELP_STAT_COMM_RATIO);
                    T_TD emit_comm_ratio(html_context, xml, stats_id, false);
                }
            }
	    T_TR {
		T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Fetch utilization",
							  HELP_STAT_FETCH_UTILIZATION);
		T_TD emit_fetch_utilization(html_context, xml, stats_id, false);
	    }
	    T_TR {
		T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Write-back utilization",
							  HELP_STAT_WB_UTILIZATION);
		T_TD emit_wb_utilization(html_context, xml, stats_id, false);
	    }
            if (html_context.feature_mt_advice) {
                T_TR {
                    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Communication utilization",
                                                              HELP_STAT_COMM_UTIL);
                    T_TD emit_comm_utilization(html_context, xml, stats_id, false);
                }
                T_TR {
                    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "False sharing ratio",
                                                              HELP_STAT_FS_RATIO);
                    T_TD emit_fs_ratio(html_context, xml, stats_id, false);
                }
            }
	} else {
	    T_TR {
		T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Utilization",
							  HELP_STAT_UTILIZATION);
		T_TD emit_utilization(html_context, xml, stats_id, false);
	    }
	}

	T_TR {
	    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "HW prefetch probability",
						      HELP_STAT_HW_PREF);
	    T_TD emit_hw_prefetch_ratio(html_context, xml, stats_id, false);
	}
	T_TR {
	    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Access randomness",
						      HELP_STAT_RANDOM);
	    T_TD emit_random_ratio(html_context, xml, stats_id, false);
	}

	if (display_worst_line) T_TR {
                T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Worst instruction",
							  HELP_STAT_WORST);
		T_TD {
		    fr_location_info_ptr_t worst =
			worst_instruction(html_context, instructions);
		    if (worst)
			link_to_source(html_context, xml, false, worst, loop);
		}
	    }
    }

    // It's no use displaying avoidable fetches if there are no read
    // operations. Such a condition would lead to 0/0 and ridiculus
    // values presented.
    bool has_read_access = false;
    BOOST_FOREACH(fr_location_info_ptr_t l, instructions) {
        rpt_access_classes_t access_classes =
            html_context.report->get_location_access_classes(l->location_id);
	if (access_classes & RPT_Access_Class_RD) {
	    has_read_access = true;
	    break;
	}
    }

    if (show_avoidable && has_read_access) {
	const rpt_sliced_stats_t &stats =
	    html_context.report->get_sliced_stats(stats_id);
	const rpt_sliced_stats_t &global_stats =
	    html_context.report->get_sliced_stats(global_stats_id);

	double local_gain =
	    report.get_sliced_stats_global(stats).get_avoidable_fetches() /
	    report.get_sliced_stats_global(stats).get_total_fetches();

	double global_gain =
	    report.get_sliced_stats_global(stats).get_avoidable_fetches() /
	    report.get_sliced_stats_global(global_stats).get_total_fetches();

	// sanity check
	if (local_gain < 10 && global_gain < 10) T_P NL {
	    TEXT("If the program was changed as to reach 100% "
		 "fetch utilization, fetches in this ");
	    if (is_loop) TEXT("loop");
	    else TEXT("instruction group");

	    TEXT(" would be reduced with ");
	    emit_percent(xml, local_gain);

	    TEXT(", and total number of fetches would be reduced with ");
	    emit_percent(xml, global_gain);
	    TEXT(".");
	}
    }
}

fr_location_info_ptr_t
html_util_t::worst_instruction(
    html_context_t &html_context,
    const fr_location_info_vector_t &instructions)
{
    // Start with worst_fetches negative, so that at least one instruction is
    // always reported if instructions is not empty.
    fr_location_info_ptr_t worst;
    double worst_fetches = -1.0;
    BOOST_FOREACH(const fr_location_info_ptr_t &p, instructions) {
        const rpt_stats_t &stats =
            html_context.report->get_location_global_stats(p->location_id);
        double fetches = stats.get_total_fetches();
        if (fetches > worst_fetches) {
            worst_fetches = fetches;
            worst = p;
        }
    }
    return worst;
}

void
html_util_t::section_statistics(xml_file_t &xml,
				html_context_t &html_context,
				const string &title,
				const string &help,
				rpt_sliced_stats_id_t stats_id,
				const string &diagram,
				const fr_location_info_vector_t &instructions,
				const fr_location_info_vector_t &loop,
				bool initially_visible,
				bool is_loop,
				bool show_avoidable,
				const string &id)
{
    string unique_id = id.length() ? id : xml.unique_id();
    SECTION(h2, unique_id, title, "", help, !initially_visible, true) {
	statistics_table(xml, html_context, stats_id,
			 diagram, instructions, loop,
			 is_loop, show_avoidable);
    }
}


struct stack_entry_t {
    stack_entry_t(rpt_address_t pc,
		  double fraction)
	: pc(pc), fraction(fraction)
	{}
    bool operator == (const stack_entry_t &s) const {
	return pc == s.pc;
    }
    rpt_address_t pc;
    double fraction;
};
typedef list<stack_entry_t> dominant_stack_t;

void
html_util_t::instruction_table(xml_file_t &xml,
			       html_context_t &html_context,
			       const fr_location_info_vector_t &instructions,
			       const fr_location_info_vector_t &loop,
			       int stat_level,
			       bool show_stack,
                               const extra_columns_t &extra_columns)
{
    const rpt_report_t &report(*html_context.report);

    rpt_sliced_stats_id_t global_stats_id = report.get_global_stats_id();

    T_TABLE A_CLASS("summary") {
	int colspan = 2;
	T_TR {
	    if (show_stack)
		T_TH TEXT("Stack");
	    else
		colspan--;
	    T_TH TEXT("Instruction");
	    if (stat_level > 0) {
		colspan += 3;
		T_TH TEXT("% of misses");
		T_TH TEXT("% of fetches");
		T_TH TEXT("Fetch ratio");
		if (stat_level > 1) {
		    if (report.mt_sampled) {
			colspan += 2;
			T_TH TEXT("Fetch utilization");
			T_TH TEXT("W-B Utilization");
		    } else {
			colspan++;
			T_TH TEXT("Utilization");
		    }
		}
	    }
            BOOST_FOREACH(const extra_column_t *ec, extra_columns) {
                ec->emit_heading(html_context, xml);
                colspan++;
            }
	}

	dominant_stack_t last_stack;
	bool first_address = false;
	string stack_id;
	string toggle_cmd;

        int line_number = 0;
	BOOST_FOREACH(const fr_location_info_ptr_t &p, instructions) {

	    // Currently emitting the truncated stack if it's different
	    // from the previously emitted stack.
	    //
	    // XXX. Instead we should emit only enough stack
	    // frames to differentiate from the previous stack.
	    // Seach backwards in the current stack for a match,
	    // and pop as necessary. At least one line needs to be
	    // output (properly indented), since it is difficult
	    // to visualize a series of pop:s otherwise.

	    if (show_stack) {
		dominant_stack_t this_stack;

                // First push the full stack...
                const rpt_full_stack_t &full_stack =
                    report.get_location_full_stack(p->location_id);
                BOOST_FOREACH(const rpt_full_stack_frame_t &frame, full_stack)
                    this_stack.push_back(
                        stack_entry_t(report.get_address(frame.pc_id),
                                      frame.fraction));

		// Then push the truncated stack...
		const rpt_trunc_stack_t &trunc_stack =
		    report.get_location_trunc_stack(p->location_id);
		BOOST_FOREACH(rpt_address_id_t pc_id, trunc_stack)
		    this_stack.push_back(
                        stack_entry_t(report.get_address(pc_id), 1.0));

		// Now see if the stack is different and thus if we
		// should output this_stack
		if (last_stack.size() == 0 || last_stack != this_stack) {
		    last_stack = this_stack;

		    if (this_stack.size() != 0) {
			first_address = true;
			stack_id = xml.unique_id();
			toggle_cmd="toggle('" + stack_id + "'); toggle('" +
			    stack_id+ "minus'); toggle('" + stack_id+ "plus');";
			T_TR A_ID(stack_id)
			    A_CLASS("alternate") A_STYLE("display:none")
			    T_TD A_COLSPAN(lexical_cast<string>(colspan))
			{
			    T_IMG A_ID(stack_id + "minus")
				A_SRC(image_path(html_context, PNG_MINUS))
				A_ONCLICK(toggle_cmd)
				A_STYLE("display: none");

			    int indent = 0;

			    BOOST_FOREACH(const stack_entry_t &stack_entry,
					  this_stack) {
				if (indent != 0) {
				    T_BR;
				    TEXTN_T(NBSP NBSP NBSP);
				}
				pc_info_ptr_t pc_info =
				    html_context.file_repos->get_pc_info(
                                        report,
					stack_entry.pc,
					true);
				line_info_ptr_t line_info =
				    pc_info->line_info.lock();

				for(int i=0; i<indent; i++)
				    TEXTN_T(NBSP NBSP NBSP);
				link_call_to_source(html_context, xml, pc_info,
                                                    loop, 0);

				if (stack_entry.fraction < 1.0) {
				    TEXT("[");
				    emit_percent(xml, stack_entry.fraction);
				    TEXT("]");
				}
				indent ++;
			    }
			}
		    }
		}
	    }

	    // emit the actual instruction
	    T_TR {
		if (show_stack) T_TD if (first_address) {
		    first_address = false;
		    T_IMG A_ID(stack_id + "plus")
			A_SRC(image_path(html_context, PNG_PLUS))
			A_ONCLICK(toggle_cmd);
		}
		T_TD
		    link_to_source(html_context, xml, false, p, loop);
		if (stat_level > 0) {
                    rpt_location_id_t loc = p->location_id;
                    rpt_sliced_stats_id_t location_stats_id =
			report.get_location_sliced_stats_id(loc);
                    rpt_access_classes_t access_classes =
                        report.get_location_access_classes(loc);


		    T_TD {
			// % of total misses. Software prefetch misses don't
			// count as misses.
			if (access_classes & RPT_Access_Class_PF)
			    emit_percent(xml, 0);
			else
			    emit_percentage_of_misses(
				html_context, xml, location_stats_id,
				global_stats_id, true);
		    }
		    T_TD {
			// % of total fetches
			emit_percentage_of_fetches(
                            html_context, xml, location_stats_id,
                            global_stats_id, true);
		    }
		    T_TD {
			// Fetch ratio.
			emit_fetch_ratio(html_context, xml, location_stats_id, true);
		    }
		    if (stat_level > 1) {
			depgroup_info_t &depgroup =
			    *html_context.loop_registry->depgroup_infos[
                                report.get_location_depgroup(loc)];
                        rpt_sliced_stats_id_t depgroup_stats_id =
                            report.get_depgroup_stats_id(depgroup.id);
			if (report.mt_sampled) {
			    T_TD
				emit_fetch_utilization(html_context, xml, depgroup_stats_id, true);
			    T_TD
                                emit_wb_utilization(html_context, xml, depgroup_stats_id, true);
			}
                        else
			    T_TD
				emit_utilization(html_context, xml, depgroup_stats_id, true);
		    }
		}
                BOOST_FOREACH(const extra_column_t *ec, extra_columns)
                    ec->emit_value(html_context, xml, line_number);
	    }
            line_number++;
	}
    }
}

void
html_util_t::section_instructions(xml_file_t &xml,
				  html_context_t &html_context,
				  const string &title,
				  const string &help,
				  const fr_location_info_vector_t &instructions,
				  const fr_location_info_vector_t &loop,
				  int stat_level,
				  bool initially_visible,
				  const string &id,
                                  const extra_columns_t &extra_columns)
{
    string unique_id = id.length() ? id : xml.unique_id();

    SECTION(h2, unique_id, title, "", help, !initially_visible, true) {
	instruction_table(xml, html_context, instructions,
			  loop, stat_level, true, extra_columns);
    }
}

void
html_util_t::all_loop_sections(xml_file_t &xml,
			       html_context_t &html_context,
			       rpt_loop_id_t loop_id,
			       fr_location_info_vector_t &location_loop_info_vector,
			       const string &extra_title,
			       const string &tag)
{
    if (location_loop_info_vector.size() == 0)
	return;

    // Loop statistics
    html_context.interesting_loops.insert(loop_id);
    section_statistics(
        xml,
        html_context,
        "Loop statistics" + extra_title,
        HELP_ISSUE_LOOP_STAT,
        html_context.report->get_loop_stats_id(loop_id),
        "l" + lexical_cast<string>(loop_id),
        location_loop_info_vector,
        location_loop_info_vector,
        false /* visible */,
        true /* is_loop */,
        false /* show_avoidable */,
        "lstat" + tag);

    // Loop instructions
    section_instructions(xml, html_context,
			 "Loop instructions" + extra_title,
			 HELP_ISSUE_LOOP_INSTRS,
			 location_loop_info_vector,
			 location_loop_info_vector,
			 2, false, "linstr" + tag);
}


void
html_util_t::motion_barrier_table(xml_file_t &xml,
				  html_context_t &html_context,
                                  const rpt_location_id_vector_t &move_pcs,
                                  const rpt_location_set_vector_t &move_blocks,
				  const fr_location_info_vector_t &loop)
{
    const rpt_report_t &report(*html_context.report);

    typedef map<rpt_location_id_t, rpt_location_set_t> inverted_map_t;
    inverted_map_t inverted;

    assert(move_pcs.size() == move_blocks.size());
    for (unsigned int i = 0; i < move_pcs.size(); i++) {
	const rpt_location_id_t &loc = move_pcs[i];
	const rpt_location_set_t &lset = move_blocks[i];
	BOOST_FOREACH(rpt_location_id_t l2, lset)
	    inverted[l2].insert(loc);
    }

    T_TABLE A_CLASS("summary") {
	T_TR {
	    T_TH A_COLSPAN("2") TEXT("Code motion barrier");
	    T_TH A_COLSPAN("2") TEXT("Loop instruction");
	}

	bool alt = false;
	BOOST_FOREACH(const inverted_map_t::value_type &as,
		      inverted) {

	    const rpt_location_set_t &new_set = as.second;

	    if (new_set.size()) {
		fr_location_info_vector_t barriers =
		    html_context.file_repos->get_line_info(report, new_set);

		T_TR A_STYLE("vertical-align: top;") {
		    if (alt) A_CLASS("alternate");

		    fr_location_info_ptr_t l =
			html_context.file_repos->get_line_info(report, as.first);

		    string call_stack_id = xml.unique_id();
		    string toggle_stack =
			"toggle('" + call_stack_id + "m');"
			"toggle('" + call_stack_id + "p');"
			"toggle('" + call_stack_id + "d');";

		    T_TD {
			T_IMG A_ID(call_stack_id + "m")
			    A_SRC(image_path(html_context, PNG_MINUS))
			    A_STYLE("display:none;")
			    A_ONCLICK(toggle_stack);
			T_IMG A_ID(call_stack_id + "p")
			    A_SRC(image_path(html_context, PNG_PLUS))
			    A_ONCLICK(toggle_stack);
		    }
		    T_TD link_to_source(html_context, xml, true, l, loop,
                                        call_stack_id + "d");

		    string loop_instr_id = xml.unique_id();
		    string toggle_loop_instr =
			"toggle('" + loop_instr_id + "m');"
			"toggle('" + loop_instr_id + "p');"
			"toggle('" + loop_instr_id + "d');";

		    T_TD {
			T_IMG A_ID(loop_instr_id + "m")
			    A_SRC(image_path(html_context, PNG_MINUS))
			    A_STYLE("display:none;")
			    A_ONCLICK(toggle_loop_instr);
			T_IMG A_ID(loop_instr_id + "p")
			    A_SRC(image_path(html_context, PNG_PLUS))
			    A_ONCLICK(toggle_loop_instr);
		    }

		    T_TD T_DIV A_ID(loop_instr_id + "d")
			A_STYLE("display:none;")
		    {
			BOOST_FOREACH(const fr_location_info_ptr_t &loc,
				      barriers) {
			    link_to_source(html_context, xml, false, loc,
                                           loop);
			    T_BR;
			}
		    }
		}

		alt = !alt;
	    }
	}
    }
}

void
html_util_t::section_motion_barrier(xml_file_t &xml,
				    html_context_t &html_context,
				    const string &title,
				    const string &help,
				    const rpt_location_id_vector_t &move_pcs,
                                    const rpt_location_set_vector_t &move_blocks,
				    const fr_location_info_vector_t &loop,
				    const string &id)
{
    string unique_id = id.length() ? id : xml.unique_id();
    SECTION(h2, unique_id, title, "", help, true, true) {
	motion_barrier_table(xml, html_context, move_pcs, move_blocks, loop);
    }
}


void
html_util_t::emit_percent(xml_file_t &xml, double v)
{
    stringstream t;
    t << fixed << setprecision(1) << v * 100.0 << "%";
    TEXT(t.str());
}

void
html_util_t::emit_percent(xml_file_t &xml, const rpt_stat_fp_valid_t &v)
{
    stringstream t;
    t << fixed << setprecision(1) << v.get_value() * 100.0 << "%";
    if (v.get_valid())
	TEXT(t.str());
    else
	T_SPAN A_CLASS("invalid")
	    A_TITLE("This value has a weak statistical base")
	    TEXT(t.str());
}

void
html_util_t::emit_scientific(xml_file_t &xml, double v,
                             double total, bool percent, bool absolute)
{
    if (absolute) {
	stringstream t;
	t << scientific << setprecision(2) << v;
	if (percent && total > v && total != 0)
	    t << " (" << fixed << setprecision(1) << v/total * 100.0 << "%)";
	TEXT(t.str());
    } else if (percent) {
	if (total == 0)
	    TEXT("---");
	else
	    emit_percent(xml, v/total);
    }
}

// ----------------------------------------------------------------------

summary_context_t::summary_context_t()
{
    alt = false;
}

// ----------------------------------------------------------------------

html_issue_t::html_issue_t(html_context_t &html_context,
                           rpt_issue_id_t issue_id)
    : html_context(html_context),
      issue_id(issue_id),
      display_number(lexical_cast<string>(issue_id + 1)),
      my_rel_file_name("issue-" + lexical_cast<string>(issue_id) + ".html"),
      my_issue_file_name(html_context.conf.report_path +
                         "/" + my_rel_file_name)
{
}

void
html_util_t::add_loop_link(xml_file_t &xml,
			   html_context_t &html_context,
			   const string &loop_id,
			   const fr_location_info_vector_t &loop_pcs)
{
    T_A A_HREF(get_loop_url(xml, html_context, loop_id, loop_pcs))
	TEXT(loop_id);
}

string
html_util_t::get_loop_url(xml_file_t &xml,
			  html_context_t &html_context,
			  const string &loop_id,
			  const fr_location_info_vector_t &loop_pcs)
{
    fr_location_info_ptr_t loc_info = worst_instruction(html_context, loop_pcs);
    string src_ref = "%27%27,0,%5B%5D"; // '',[]

    if (loc_info.get() != NULL) {
        pc_info_ptr_t pc_info = loc_info->pc_info.lock();
        line_info_ptr_t line_info = pc_info->line_info.lock();
        if (line_info.get()) {
            string line_number_s = lexical_cast<string>(line_info->line);
            string highlight = get_highlight(loop_pcs, line_info->filename);
	    int file_id = html_context.report->get_source_file_id(line_info->filename);
            src_ref = lexical_cast<string>(file_id) + "," + line_number_s +
                ",%5B" + highlight + "%5D";
        }
    }


    string loop_string = "javascript:load_loop(" +
        loop_id + "," + src_ref + ")";

    return loop_string;
}

void
html_util_t::add_issue_link(xml_file_t &xml,
			    html_context_t &html_context,
			    const string &issue_file_name,
			    const html_issue_t &issue,
			    const fr_location_info_vector_t &instructions,
			    const fr_location_info_vector_t &loop_pcs,
			    const string &icon)
{
    T_A A_HREF(get_issue_url(xml, html_context, issue, instructions, loop_pcs))
	TEXT(issue.get_display_number());
}

string
html_util_t::get_issue_url_no_source(const html_issue_t &issue)
{
    return "javascript:load_issue_no_source(" +
        lexical_cast<string>(issue.get_issue_id()) + ")";
}

string
html_util_t::get_issue_url(xml_file_t &xml,
			   html_context_t &html_context,
			   const html_issue_t &issue,
			   const fr_location_info_vector_t &instructions,
			   const fr_location_info_vector_t &loop_pcs)
{
    fr_location_info_ptr_t loc_info = worst_instruction(html_context, instructions);
    string src_ref = "%27%27,0,%5B%5D";

    if (loc_info.get() != NULL) {
        pc_info_ptr_t pc_info = loc_info->pc_info.lock();
        line_info_ptr_t line_info = pc_info->line_info.lock();
        if (line_info.get()) {
            string line_number_s = lexical_cast<string>(line_info->line);
            string highlight = get_highlight(loop_pcs, line_info->filename);
	    int file_id = html_context.report->get_source_file_id(line_info->filename);
            src_ref = lexical_cast<string>(file_id) + "," +
		line_number_s + ",%5B" + highlight + "%5D";
        }
    }

    string issue_string = "javascript:load_issue(" +
	lexical_cast<string>(issue.get_issue_id()) + "," + src_ref + ")";

    return issue_string;
}

bool
html_util_t::have_source(html_context_t &html_context,
                         const fr_location_info_vector_t &instructions)
{
    fr_location_info_ptr_t loc_info = worst_instruction(html_context, instructions);
    pc_info_ptr_t pc_info = loc_info->pc_info.lock();
    line_info_ptr_t line_info = pc_info->line_info.lock();
    return line_info.get() != NULL;
}

void
html_issue_t::insert_summary_header(html_context_t &html_context,
				    xml_file_t &xml,
				    summary_type_t summary_type)
{
    const rpt_report_t &report(*html_context.report);

#define COLUMN(title, sort)				\
    T_TH A_CLASS(string("table-sortable:") + sort)	\
	TEXT(title)

    T_THEAD T_TR {
	COLUMN("#", "numeric");
	COLUMN("", "default");
	COLUMN("Issue type", "default table-filterable");

	switch (summary_type) {
	case SUMMARY_BANDWIDTH:
	    if (report.mt_sampled)
		COLUMN("% of bandwidth", "scientific");
	    COLUMN("% of fetches", "scientific");
	    if (report.mt_sampled) {
		COLUMN("% of write-backs", "scientific");
		COLUMN("Fetch utilization", "numeric");
		COLUMN("Write-back utilization", "numeric");
	    } else
		COLUMN("Utilization", "numeric");
	    break;

	case SUMMARY_LATENCY:
	    COLUMN("% of misses", "scientific");
	    COLUMN("HW-Prefetch", "default");
	    COLUMN("Randomness", "default");
	    if (report.mt_sampled) {
		COLUMN("Fetch utilization", "numeric");
	    } else
		COLUMN("Utilization", "numeric");
	    break;

	case SUMMARY_MT:
	    COLUMN("% of communication", "scientific");
	    COLUMN("Communication utilization", "scientific");
	    COLUMN("False sharing", "scientific");
	    break;

	case SUMMARY_NT:
	    COLUMN("% of fetches", "scientific");
	    COLUMN("Required cache size", "scientific");
	    break;

	default:
	    assert(0);
	    while(1);
	}
    }
#undef COLUMN
}

void
html_issue_t::annotate_generic(rpt_depgroup_id_t depgroup_id, bool force)
{
    const rpt_report_t &report(*html_context.report);
    const file_repos_t &file_repos(*html_context.file_repos);

    fr_location_info_vector_t location_dep_info_vector =
	file_repos.get_depgroup_line_info(report, depgroup_id);

    BOOST_FOREACH(fr_location_info_ptr_t &p, location_dep_info_vector)
        p->add_reporter(
            location_reporter_ptr_t(
                new generic_location_reporter_t(*this, issue_id, force)));
}

void
html_issue_t::add_inferior(html_issue_ptr_t inferior)
{
    inferior_list.push_back(inferior);
}

void
html_issue_t::add_summary_line(xml_file_t &xml,
			       summary_context_t &summary_context,
			       const string &issue_file_name,
			       const string &title,
			       const string &icon,
			       const fr_location_info_vector_t &instructions,
			       const fr_location_info_vector_t &loop_pcs,
			       rpt_sliced_stats_id_t stats_id,
			       const string &loop_id,
			       summary_type_t summary_type,
                               // TODO: These two arguments is an ugly hack
                               //       only used by the prefetchnta issue.
                               double fetches,
                               double required_cache_size) const
{
    const rpt_report_t &report(*html_context.report);

    rpt_sliced_stats_id_t global_stats_id = report.get_global_stats_id();

    NL T_TR {
	if (html_util_t::have_source(html_context, instructions)) {
	    if (summary_context.alt)
		A_CLASS("alternate");
	} else {
	    A_TITLE("No source available");

	    if (summary_context.alt)
		A_CLASS("nosrc alternate");
	    else
		A_CLASS("nosrc");
	}
	summary_context.alt = !summary_context.alt;

	// issue-id
	T_TD NL
	    add_issue_link(xml, html_context, issue_file_name, *this,
                           instructions, loop_pcs, icon);

	T_TD NL
	    if (icon.size())
		T_A A_HREF(get_issue_url(xml, html_context, *this,
                                         instructions, loop_pcs))
		    T_IMG A_SRC(icon) A_BORDER("0");

	// title
	T_TD T_A A_HREF(get_issue_url(xml, html_context, *this, instructions,
                                      loop_pcs))
	    TEXT(title);

	switch (summary_type) {
	case SUMMARY_BANDWIDTH:
	    if (report.mt_sampled)
		T_TD emit_percentage_of_bandwidth(html_context, xml, stats_id,
						  global_stats_id,
						  false);

	    T_TD emit_percentage_of_fetches(html_context, xml, stats_id, global_stats_id, false);
	    if (report.mt_sampled) {
                T_TD emit_percentage_of_write_backs(html_context, xml, stats_id, global_stats_id, false);
		T_TD emit_fetch_utilization(html_context, xml, stats_id, true);
		T_TD emit_wb_utilization(html_context, xml, stats_id, true);
            }
            else
                T_TD emit_utilization(html_context, xml, stats_id, true);
            break;
	case SUMMARY_LATENCY:
	    T_TD emit_percentage_of_misses(html_context, xml, stats_id, global_stats_id, false);
	    T_TD emit_hw_prefetch_ratio(html_context, xml, stats_id, true);
	    T_TD emit_random_ratio(html_context, xml, stats_id, true);
	    if (report.mt_sampled)
		T_TD emit_fetch_utilization(html_context, xml, stats_id, true);
            else
                T_TD emit_utilization(html_context, xml, stats_id, true);
            break;
	case SUMMARY_MT:
            assert(html_context.feature_mt_advice);
	    T_TD emit_percentage_of_comm(html_context, xml, stats_id, global_stats_id, false);
	    T_TD emit_comm_utilization(html_context, xml, stats_id, true);
	    T_TD emit_fs_ratio(html_context, xml, stats_id, true);
	    break;

	case SUMMARY_NT: {
	    const rpt_sliced_stats_t &global_stats =
		report.get_sliced_stats(global_stats_id);

            double global_fetches =
                report.get_sliced_stats_global(global_stats).get_total_fetches();
	    T_TD emit_scientific(xml, fetches, global_fetches, true, false);
            T_TD TEXT(html_context.value_to_string(required_cache_size));
        } break;

        case SUMMARY_COUNT:
            assert(0);
            break;
	}
    }
}

void
html_issue_t::dump_inferior(xml_file_t &xml)
{
    if (inferior_list.size()) NL T_P {
	TEXT("This instruction group also shows symptoms of:");
	bool first = true;
	BOOST_FOREACH(html_issue_ptr_t inferior, inferior_list) {
	    if (first)
		TEXT(" ");
	    else
		TEXT(", ");
	    first = false;
	    inferior->link(xml, true, false);
	}
	TEXT(".");
    }
}

void
html_issue_t::link(xml_file_t &xml, bool text, bool do_link)
{
    string title(get_name());

    if (do_link) {
	title += ", issue: " + get_display_number();
	emit_link_to_issue_no_source(xml, title, get_icon(), *this, text);
    }
    else {
	A_TITLE(title) T_IMG A_SRC(get_icon()) A_BORDER("0");
	if (text)
            TEXT(" " + title);
    }
}

END_NAMESPACE;
END_NAMESPACE;
