/// @file tools/acurate/html_issue.hh

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#ifndef INCLUDED_HTML_ISSUE
#define INCLUDED_HTML_ISSUE

#include "../html_context.hh"

#include "analysis/report/report.hh"

#include "library/fenris/fnr_types.hh"

#include <boost/shared_ptr.hpp>

#include <list>
#include <vector>
#include <set>
#include <bitset>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace acumem::analysis;

class xml_file_t;

struct extra_column_t {
    virtual ~extra_column_t() {};

    virtual void emit_heading(html_context_t &html_context,
			      xml_file_t &xml) const = 0;
    virtual void emit_value(html_context_t &html_context,
                            xml_file_t &xml,
                            int line) const = 0;
};

// forward declaration
class tp_field_t;
typedef boost::shared_ptr<tp_field_t> tp_field_ptr_t;

/// Class used by emit_thread_popup_value to present a field in
/// sliced_stats_cptr_t in a thread popup window.
class tp_field_t {
public:
    virtual ~tp_field_t() {}

    /// Table heading for the value column.
    virtual std::string heading() const = 0;

    /// HTML ID attribute of the field.
    virtual std::string id() const = 0;

    /// Whether to print total (instead of average)
    virtual bool total() const = 0;

    /// Emit the textual representation of the field.
    virtual void emit(xml_file_t &xml, const rpt_stats_t &stats) const = 0;

    /// Output a string that can be parsed to construct a new object later
    virtual std::string code() const;

    virtual std::string class_string() const =0;
    virtual double value() const { return 0; }

    static tp_field_ptr_t create(const std::string &tag, double value);
};


// ----------------------------------------------------------------------

/// TODO
struct by_kind_value_t {
    by_kind_value_t(
        double cold,
        double repl,
        double coh,
        double flush)
        : cold(cold),
          repl(repl),
          coh(coh),
          flush(flush)
        {}

    const double cold;
    const double repl;
    const double coh;
    const double flush;

    double total() const {
        return cold + repl + coh + flush;
    }
};

class tp_field_by_kind_t;
typedef boost::shared_ptr<tp_field_by_kind_t> tp_field_by_kind_ptr_t;


/// Class used by emit_thread_popup_value to present a field in
/// sliced_stats_cptr_t in a thread popup window.
class tp_field_by_kind_t {
public:
    virtual ~tp_field_by_kind_t() {}

    /// Table heading for the value column.
    virtual string heading() const = 0;

    /// HTML ID attribute of the field.
    virtual string id() const = 0;

    /// Return the by_kind field.
    virtual by_kind_value_t by_kind(const rpt_stats_t &stats) const = 0;

    /// Whether to print "Total" (instead of "Average")
    virtual bool total() const = 0;

    /// Emit the textual representation of the field.
    virtual void emit(xml_file_t &xml, double value) const = 0;


    virtual std::string code() const;
    virtual std::string class_string() const =0;

    static tp_field_by_kind_ptr_t create(const std::string &tag);
};

// ----------------------------------------------------------------------

class html_issue_t;

typedef vector<const extra_column_t *> extra_columns_t;

struct html_util_t {

    static const double RANDOM_LOW_THRESHOLD;
    static const double RANDOM_HIGH_THRESHOLD;
    static const double HWPF_LOW_THRESHOLD;
    static const double HWPF_HIGH_THRESHOLD;


    /// Emit the table in the lower left hand corner
    static void emit_thread_popup_table(html_context_t &html_context,
					xml_file_t &xml,
					rpt_sliced_stats_id_t stats_id,
					const tp_field_t &field);

    static void emit_thread_popup_table_by_kind(html_context_t &html_context,
						xml_file_t &xml,
						rpt_sliced_stats_id_t stats_id,
						const tp_field_by_kind_t &field);

    /// Emit a value and a link to display the details in the lower
    /// left hand corner.
    static void emit_thread_popup_value(html_context_t &html_context,
					xml_file_t &xml,
					rpt_sliced_stats_id_t stats_id,
					const tp_field_t &field,
					bool left);

    static void emit_thread_popup_value_by_kind(html_context_t &html_context,
						xml_file_t &xml,
						rpt_sliced_stats_id_t stats_id,
						const tp_field_by_kind_t &field,
						bool left);

    /// Emit textual thread representation.
    static void emit_thread(
	html_context_t &html_context,
        xml_file_t &xml,
        rpt_thread_id_t thread_id);

    /// Emit number of accesses.
    static void emit_accesses(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit fetch ratio.
    static void emit_fetch_ratio(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit miss ratio.
    static void emit_miss_ratio(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit upgrade ratio.
    static void emit_upgrade_ratio(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit write-back ratio.
    static void emit_write_back_ratio(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit number of fetches.
    static void emit_fetches(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit number of misses.
    static void emit_misses(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit number of upgrades.
    static void emit_upgrades(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit number of write-backs.
    static void emit_write_backs(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit percentage of fetches.
    static void emit_percentage_of_fetches(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        rpt_sliced_stats_id_t global_stats_id,
        bool left);

    /// Emit percentage of misses.
    static void emit_percentage_of_misses(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        rpt_sliced_stats_id_t global_stats_id,
        bool left);

    /// Emit percentage of misses.
    static void emit_percentage_of_bandwidth(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        rpt_sliced_stats_id_t global_stats_id,
        bool left);

    /// Emit percentage of upgrades.
    static void emit_percentage_of_upgrades(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        rpt_sliced_stats_id_t global_stats_id,
        bool left);

    /// Emit percentage of write-backs.
    static void emit_percentage_of_write_backs(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        rpt_sliced_stats_id_t global_stats_id,
        bool left);

    /// Emit percentage of communication.
    static void emit_percentage_of_comm(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        rpt_sliced_stats_id_t global_stats_id,
        bool left);

    /// Emit fetch utilization
    static void emit_utilization(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit fetch utilization
    static void emit_fetch_utilization(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit write-back utilization
    static void emit_wb_utilization(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit hardware prefetch ratio.
    static void emit_hw_prefetch_ratio(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit random ratio.
    static void emit_random_ratio(
        html_context_t &html_context,
        xml_file_t &xml,
        rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit communication ratio
    static void emit_comm_ratio(
        html_context_t &html_context,
	xml_file_t &xml,
	rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit communication utilization
    static void emit_comm_utilization(
        html_context_t &html_context,
	xml_file_t &xml,
	rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit false sharing ratio
    static void emit_fs_ratio(
        html_context_t &html_context,
	xml_file_t &xml,
	rpt_sliced_stats_id_t stats_id,
        bool left);

    /// Emit html head section with the specified title. Script path prefixed
    /// to the names of stylesheet and javascript files.
    static void emit_head(
        xml_file_t &xml,
        const string &title = "",
        const string &script_path = "");

    // Emit header, should be the first thing in the body of the issue file.
    static void emit_issue_header(xml_file_t &xml,
				  bool warn_unreliable,
				  bool warn_virtualized);

    // Emit footer, copyright, terms of use
    static void emit_product_name_full(xml_file_t &xml);
    static void emit_issue_footer(html_context_t &html_context,
                                  xml_file_t &xml);
    static void emit_terms_of_use(html_context_t &html_context,
                                  xml_file_t &xml);
    static void emit_copyright(xml_file_t &xml, bool small_font);

    static void emit_time(xml_file_t &xml, acumem::library::fnr_real_time_t t);


    // Create a source code link
    static void link_call_to_source(html_context_t &html_context,
				    xml_file_t &xml,
				    const pc_info_ptr_t &pc_info,
				    const fr_location_info_vector_t &loop_pcs,
				    fnr_access_classes_t access_classes);

    static void link_to_source(html_context_t &html_context,
			       xml_file_t &xml,
			       bool show_stack,
			       const fr_location_info_ptr_t &p,
			       const fr_location_info_vector_t &loop_pcs,
			       const std::string &div_id = "");

    static void link_to_file(html_context_t &html_context,
			     xml_file_t &xml,
			     const std::string &filename);

    // ----------------------------------------------------------------
    // Major sections


    static void statistics_table(xml_file_t &xml,
				 html_context_t &html_context,
				 rpt_sliced_stats_id_t stats_id,
				 const std::string &diagram,
				 const fr_location_info_vector_t &instructions,
				 const fr_location_info_vector_t &loop,
				 bool is_loop,
				 bool show_avoidable);

    static void section_statistics(xml_file_t &xml,
				   html_context_t &html_context,
				   const std::string &title,
				   const std::string &help,
				   rpt_sliced_stats_id_t stats_id,
				   const std::string &diagram,
				   const fr_location_info_vector_t &instructions,
				   const fr_location_info_vector_t &loop,
				   bool initially_visible,
				   bool is_loop,
				   bool show_avoidable,
				   const std::string &id); // = ""

    static void instruction_table(xml_file_t &xml,
				  html_context_t &html_context,
				  const fr_location_info_vector_t &instructions,
				  const fr_location_info_vector_t &loop,
				  int stat_level,
				  bool show_stack,
                                  const extra_columns_t &extra_columns = extra_columns_t());

    static void section_instructions(xml_file_t &xml,
				     html_context_t &html_context,
				     const std::string &title,
				     const std::string &help,
				     const fr_location_info_vector_t &instructions,
				     const fr_location_info_vector_t &loop,
				     int stat_level, // 0, 1, 2
				     bool initially_visible,
				     const std::string &id,
                                     const extra_columns_t &extra_columns = extra_columns_t());

    static void all_loop_sections(xml_file_t &xml,
				  html_context_t &html_context,
				  rpt_loop_id_t loop_id,
				  fr_location_info_vector_t &location_loop_info_vector,
				  const std::string &extra_title = "",
				  const std::string &tag = "");

    static void motion_barrier_table(xml_file_t &xml,
				     html_context_t &html_context,
                                     const rpt_location_id_vector_t &move_pcs,
                                     const rpt_location_set_vector_t &move_blocks,
				     const fr_location_info_vector_t &loop);

    static void section_motion_barrier(xml_file_t &xml,
				       html_context_t &html_context,
				       const std::string &title,
				       const std::string &help,
                                       const rpt_location_id_vector_t &move_pcs,
                                       const rpt_location_set_vector_t &move_blocks,
				       const fr_location_info_vector_t &loop,
				       const std::string &id);

    static string image_path(html_context_t &html_context,
			     const std::string &image);

    static void section_head(html_context_t &html_context,
			     xml_file_t &xml,
			     const std::string &id,
			     const std::string &title,
			     bool initial_hidden,
			     bool hideable,
			     const std::string &icon,
			     const std::string &help_section);

    /// Returns the worst instruction in terms of number of fetches, or NULL if
    /// instructions is empty.
    static fr_location_info_ptr_t worst_instruction(
	html_context_t &html_context,
        const fr_location_info_vector_t &instructions);

    static std::set<int> get_highlight_lines(const fr_location_info_vector_t &loop_pcs,
					     const std::string &filename);
    static std::string get_highlight(const fr_location_info_vector_t &loop_pcs,
				     const std::string &filename);

    static void add_loop_link(xml_file_t &xml,
			      html_context_t &html_context,
			      const std::string &loop_id,
			      const fr_location_info_vector_t &loop_pcs);
    static std::string get_loop_url(xml_file_t &xml,
				    html_context_t &html_context,
				    const std::string &loop_id,
				    const fr_location_info_vector_t &loop_pcs);
    static void add_issue_link(xml_file_t &xml,
			       html_context_t &html_context,
			       const std::string &issue_file_name,
			       const html_issue_t &issue,
			       const fr_location_info_vector_t &instructions,
			       const fr_location_info_vector_t &loop_pcs,
			       const std::string &icon);

    static std::string get_issue_url_no_source(
	const html_issue_t &issue);

    static void emit_link_to_issue_no_source(xml_file_t &xml,
					     const std::string &title,
					     const std::string &icon,
					     const html_issue_t &issue,
					     bool display_title);

    static std::string get_issue_url(xml_file_t &xml,
				     html_context_t &html_context,
                                     const html_issue_t &issue,
				     const fr_location_info_vector_t &instructions,
				     const fr_location_info_vector_t &loop_pcs);

    static bool have_source(html_context_t &html_context,
                            const fr_location_info_vector_t &instructions);

    static void emit_percent(xml_file_t &xml, double v);
    static void emit_percent(xml_file_t &xml, const rpt_stat_fp_valid_t &v);
    static void emit_scientific(xml_file_t &xml, double v,
				double total, bool percent, bool absolute);
    static void emit_glossary_phrase(html_context_t &html_context,
				     xml_file_t &xml,
				     const std::string &phrase,
				     const std::string &help_section);

    /// Format an address for output.
    static string format_address(rpt_address_t address);
};


// Convenience macros for large commonly occurring sections
#define SECTION(tag, id, title, icon, help_section, initial_hidden, hideable) \
    T_A A_NAME(id) TEXT(""); TAG(tag)					\
    html_util_t::section_head(html_context, xml, id+string("d"), title,	\
			      initial_hidden,				\
			      hideable, icon, help_section);		\
    T_DIV A_ID(id + string("d")) A_CLASS(#tag "div")			\
	A_STYLE(initial_hidden ? "display: none" : "")


// ----------------------------------------------------------------------

typedef boost::shared_ptr<html_issue_t> html_issue_ptr_t;
typedef std::list<html_issue_ptr_t> html_issue_list_t;
typedef std::set<html_issue_ptr_t> html_issue_set_t;

enum summary_type_t {
    SUMMARY_LATENCY = 0,
    SUMMARY_BANDWIDTH,
    SUMMARY_MT,
    SUMMARY_NT,

    SUMMARY_COUNT
};

typedef std::bitset<SUMMARY_COUNT> summary_type_bitset_t;

struct summary_context_t {
    summary_context_t();
    bool alt;
};

class html_issue_t : public html_util_t {
public:
    virtual ~html_issue_t() {}

    html_issue_t(html_context_t &html_context,
                 rpt_issue_id_t issue_id);

    // Insert a header for an issue list
    static void insert_summary_header(html_context_t &html_context,
				      xml_file_t &xml,
				      summary_type_t summary_type);

    /// Issue ID in the intermediate format report.
    rpt_issue_id_t get_issue_id() const {
        return issue_id;
    }

    /// Issue number to display in output. Adds 1 to our internal ID, so that
    /// the issue numbering starts at 1.
    string get_display_number() const {
	return display_number;
    }

    string get_issue_file_name() const {
	return my_issue_file_name;
    }

    string get_rel_file_name() const {
	return my_rel_file_name;
    }

    // ----------------------------------------------------------------
    // To be implemented by descendants. Describing themselves for the
    // summary issues pane and source annotation.

    virtual summary_type_bitset_t summary_type() const = 0;

    // Insert a line in the top-level issuelist
    virtual void emit_summary(xml_file_t &xml,
			      summary_context_t &summary_context,
			      summary_type_t summary_type) =0;

    // Create an issue file (to be linked from the top-level issue list)
    virtual void emit_issue_file(xml_file_t &issue_xml) =0;

    void annotate_generic(rpt_depgroup_id_t depgroup_id, bool force);
    void emit_issue_header(xml_file_t &xml);
    virtual void annotate_source() =0;

    virtual void link(xml_file_t &xml,
		      bool text,
		      bool do_link);

    virtual string get_name() const = 0;
    virtual string get_icon() const = 0;

    void add_inferior(html_issue_ptr_t inferior);

protected:
    /// The all encompassing context beast.
    html_context_t &html_context;

    const rpt_issue_id_t issue_id;

    void add_summary_line(xml_file_t &xml,
			  summary_context_t &summary_context,
			  const std::string &issue_file_name,
			  const std::string &title,
			  const std::string &icon,
			  const fr_location_info_vector_t &instructions,
			  const fr_location_info_vector_t &loop_pcs,
			  rpt_sliced_stats_id_t stats_id,
			  const std::string &loop_id,
			  summary_type_t summary_type,
                          // TODO: These two arguments is an ugly hack only
                          //       used by the prefetchnta issue.
                          double fetches = 0.0,
                          double required_cache_size = 0.0) const;

    void report_issue() const;

    void dump_inferior(xml_file_t &xml);

    html_issue_list_t inferior_list;

private:
    string display_number;
    string my_rel_file_name;
    string my_issue_file_name;
};

class generic_location_reporter_t
    : public location_reporter_t
{
public:
    generic_location_reporter_t(html_issue_t &parent,
                                int issue_number,
                                bool do_force)
	: location_reporter_t(issue_number),
	  parent(parent),
          do_force(do_force)
        {}

    void report(xml_file_t &xml) const {
	parent.link(xml, false, true);
    }

    bool force() const {
	return do_force;
    }

private:
    html_issue_t &parent;
    bool do_force;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
