/// @file tools/acurate/html_loop.cc

// Copyright (c) 2009-2010 Acumem AB.  All rights reserved.

#include "loop.hh"
#include "macros.hh"
#include "help_links.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace std;
using namespace boost;
using namespace acumem::analysis;

#define COLUMN(title, sort)				\
    T_TH A_CLASS(string("table-sortable:") + sort)	\
	TEXT(title)

html_loop_t::html_loop_t(html_context_t &html_context,
			 const loop_info_ptr_t &loop)
    : html_context(html_context), loop(loop)
{
}

html_loop_t::~html_loop_t()
{
}

static void
get_issue_icons(set<string> &icons,
		html_context_t &html_context,
		loop_info_ptr_t &loop)
{
    BOOST_FOREACH(rpt_depgroup_id_t depgroup_id, loop->depgroup_ids) {
	depgroup_info_ptr_t &d =
	    html_context.loop_registry->depgroup_infos[depgroup_id];

	BOOST_FOREACH(html_issue_ptr_t &issue, d->issues)
	    icons.insert(issue->get_icon());
    }
}

void
html_loop_t::emit_summary(xml_file_t &xml,
			  loop_summary_context_t &summary_context)
{
    const rpt_report_t &report(*html_context.report);

    rpt_sliced_stats_id_t global_stats_id = report.get_global_stats_id();
    rpt_sliced_stats_id_t loop_stats_id =
	report.get_loop_stats_id(loop->loop_id);

    set<string> issue_icons;
    get_issue_icons(issue_icons, html_context, loop);

    // Don't show any loops without issues
    if (issue_icons.empty())
	return;

    T_TR {
	if (summary_context.alt)
	    A_CLASS("alternate");
	summary_context.alt = !summary_context.alt;

	string loop_s = lexical_cast<string>(loop->loop_id);
	fr_location_info_vector_t loop_pcs =
	    html_context.file_repos->get_line_info(
		report, report.get_loop_incl_location_set(loop->loop_id));

	T_TD add_loop_link(xml, html_context, loop_s, loop_pcs);

	T_TD emit_percentage_of_misses(
	    html_context, xml, loop_stats_id, global_stats_id, false);

	T_TD emit_percentage_of_fetches(
	    html_context, xml, loop_stats_id, global_stats_id, false);

	if (report.mt_sampled) {
	    T_TD emit_fetch_utilization(html_context, xml, loop_stats_id, true);
	    T_TD emit_wb_utilization(html_context, xml, loop_stats_id, true);
	} else
	    T_TD emit_utilization(html_context, xml, loop_stats_id, true);

	T_TD {
	    BOOST_FOREACH(const string &icon, issue_icons) {
		T_IMG A_SRC(icon);
	    }
	}
    }
}

void
html_loop_t::emit_depgroup_summary_line(xml_file_t &xml,
					const depgroup_info_ptr_t &d,
					int local_depgroup_id,
					bool alt)
{
    const rpt_report_t &report(*html_context.report);

    rpt_sliced_stats_id_t global_stats_id = report.get_global_stats_id();
    rpt_sliced_stats_id_t depgroup_stats_id =
        report.get_depgroup_stats_id(d->id);

    T_TR {
	if (alt)
	    A_CLASS("alternate");

	string target_id = "%27g" +
	    lexical_cast<string>(local_depgroup_id) + "%27";
	T_TD T_A A_HREF("javascript:show_and_jump(" +
			target_id + ")")
	    TEXT(lexical_cast<string>(local_depgroup_id));

	// misses
	T_TD emit_percentage_of_misses(
	    html_context, xml, depgroup_stats_id,
	    global_stats_id, false);

	// fetches
	T_TD emit_percentage_of_fetches(
	    html_context, xml, depgroup_stats_id,
	    global_stats_id, false);

	// utilization
	if (report.mt_sampled) {
	    T_TD emit_fetch_utilization(html_context, xml, depgroup_stats_id, false);
	    T_TD emit_wb_utilization(html_context, xml, depgroup_stats_id, false);
	} else
	    T_TD emit_utilization(html_context, xml, depgroup_stats_id, false);

	// hw prefetch
	T_TD emit_hw_prefetch_ratio(html_context, xml, depgroup_stats_id, true);

	// randomness
	T_TD emit_random_ratio(html_context, xml, depgroup_stats_id, true);

	T_TD {
	    BOOST_FOREACH(html_issue_ptr_t &issue, d->issues)
		T_IMG A_SRC(issue->get_icon()) A_VALIGN("middle");
	}
    }
}

void
html_loop_t::emit_depgroup_summary(xml_file_t &xml)
{
    const rpt_report_t &report(*html_context.report);

    SECTION(h2, "linstrg", "Instruction groups in this loop",
	    "", HELP_LOOP_GROUPS, true, true) {
	T_TABLE A_CLASS("summary table-autosort:2 "
			"table-stripeclass:alternate") {
	    T_THEAD {
		T_TR {
		    COLUMN("Group", "default");
		    COLUMN("% of misses", "scientific");
		    COLUMN("% of fetches", "scientific");
		    if (report.mt_sampled) {
			COLUMN("Fetch utilization", "numeric");
			COLUMN("Write-back utilization", "numeric");
		    } else
			COLUMN("Utilization", "numeric");
		    COLUMN("HW prefetch probability", "default");
		    COLUMN("Randomness", "default");
		    COLUMN("Issues", "default");
		}
	    }

	    T_TBODY {
		int count = 0;
		bool alt = false;
		BOOST_FOREACH(rpt_depgroup_id_t depgroup_id,
			      loop->depgroup_ids) {
		    depgroup_info_ptr_t &d(
			html_context.loop_registry->depgroup_infos[depgroup_id]);

		    count++;

		    if (!d->issues.empty()) {
			emit_depgroup_summary_line(xml, d, count, alt);
			alt = !alt;
		    }
		}
	    } // T_TBODY
	} // T_TABLE
    }
}

void
html_loop_t::emit_depgroup_section(xml_file_t &xml,
				   const depgroup_info_ptr_t &d,
				   const fr_location_info_vector_t &loop_locations,
				   int local_depgroup_id)
{
    const rpt_report_t &report(*html_context.report);

    rpt_sliced_stats_id_t depgroup_stats_id =
        report.get_depgroup_stats_id(d->id);

    fr_location_info_vector_t lcs =
	html_context.file_repos->get_depgroup_line_info(report, d->id);

    html_context.interesting_depgroups.insert(d->id);

    SECTION(h2,
	    "g" + lexical_cast<string>(local_depgroup_id),
	    "Instruction group " +
	    lexical_cast<string>(local_depgroup_id),
	    "", HELP_LOOP_INSTR_GROUP, true, true) {

	statistics_table(xml, html_context, depgroup_stats_id,
                         "d" + lexical_cast<string>(d->id),
                         lcs, lcs,
			 false, // is_loop
			 false); // show_avoidable

	// NL ("no linebreaks") is important to avoid
	// graphical clutter between the icon and the
	// description
	NL T_P {
	    TEXT("The following issues are detected for this "
		 "instruction group:");
	    T_UL {

		BOOST_FOREACH(html_issue_ptr_t p,
			      d->issues) T_LI
		    p->link(xml, true, true);
	    }
	}

	instruction_table(xml, html_context, lcs, loop_locations,
			  2, false);
    }
}

static void
emit_issue_list(html_context_t &html_context, xml_file_t &xml,
		html_issue_set_t &issues,
		summary_type_t summary_type)
{
    summary_context_t summary_context;

    T_TABLE A_CLASS("summary table-autosort:2 "
		    "table-stripeclass:alternate") {
	html_issue_t::insert_summary_header(html_context, xml, summary_type);

	T_TBODY {
	    BOOST_FOREACH(const html_issue_ptr_t &issue, issues)
		issue->emit_summary(xml, summary_context, summary_type);
	}
    }
}

void
html_loop_t::emit_issues_section(xml_file_t &xml)
{
    html_issue_set_t mt_issues, bw_issues, lat_issues, prefetch_issues;

    BOOST_FOREACH(rpt_depgroup_id_t depgroup_id, loop->depgroup_ids) {
	depgroup_info_ptr_t &depgroup =
	    html_context.loop_registry->depgroup_infos[depgroup_id];

	BOOST_FOREACH(html_issue_ptr_t &issue, depgroup->issues) {
	    const summary_type_bitset_t summary_types(
                issue->summary_type());
	    if (summary_types.test(SUMMARY_MT))
		mt_issues.insert(issue);

	    if (summary_types.test(SUMMARY_BANDWIDTH))
		bw_issues.insert(issue);

	    if (summary_types.test(SUMMARY_LATENCY))
		lat_issues.insert(issue);

	    if (summary_types.test(SUMMARY_NT))
		prefetch_issues.insert(issue);
	}
    }

    if (!bw_issues.empty()) {
	SECTION(h2, "lissues_bw", "Bandwidth issues related to this this loop",
		"", HELP_LOOP_ISSUES, true, true)
	    emit_issue_list(html_context, xml,
			    bw_issues, SUMMARY_BANDWIDTH);
    }

    if (!lat_issues.empty()) {
	SECTION(h2, "lissues_lat", "Latency issues related to this this loop",
		"", HELP_LOOP_ISSUES, true, true)
	    emit_issue_list(html_context, xml,
			    lat_issues, SUMMARY_LATENCY);
    }

    if (!mt_issues.empty()) {
	SECTION(h2, "lissues_mt", "MT issues related to this this loop",
		"", HELP_LOOP_ISSUES, true, true)
	    emit_issue_list(html_context, xml,
			    mt_issues, SUMMARY_MT);
    }
    if (!prefetch_issues.empty()) {
	SECTION(h2, "lissues_prefetch",
		"Prefetch issues related to this this loop",
		"", HELP_LOOP_ISSUES, true, true)
	    emit_issue_list(html_context, xml,
			    prefetch_issues, SUMMARY_NT);
    }
}

void
html_loop_t::emit_loop_file(xml_file_t &xml_loop)
{
    const rpt_report_t &report(*html_context.report);

    rpt_sliced_stats_id_t loop_stats_id =
        report.get_loop_stats_id(loop->loop_id);

    set<string> issue_icons;
    get_issue_icons(issue_icons, html_context, loop);

    // Don't show any loops without issues
    if (issue_icons.empty())
	return;

    string loop_s = lexical_cast<string>(loop->loop_id);

    XML_OBJECT(xml_loop) {
	T_DIV A_CLASS("issue") A_ID("sections") A_STYLE("display: none") {
	    fr_location_info_vector_t loop_locations;
	    string title;

	    if (loop_s == "0")
		title = "Instructions in incomplete loops";
	    else
		title = "Loop " + loop_s;
	    SECTION(h1, "i", title, "", HELP_LOOP, false, false);

	    if (loop_s != "0") {
		rpt_location_set_t loop_location_set =
		    report.get_loop_excl_location_set(loop->loop_id);
		loop_locations =
		    html_context.file_repos->get_line_info(report, loop_location_set);

		html_context.interesting_loops.insert(loop->loop_id);
		section_statistics(xml, html_context,
				   "Loop statistics",
				   HELP_LOOP_STAT,
				   loop_stats_id,
				   "l" + loop_s,
				   loop_locations,
				   loop_locations,
				   false,  // initially visible
				   true,   // is_loop
				   false,  // show_avoidable
				   "lstat");

		section_instructions(xml, html_context,
				     "Loop instructions",
				     HELP_LOOP_INSTR,
				     loop_locations,
				     loop_locations,
                                     2, false, "linstr");
	    }

	    emit_issues_section(xml);
	    emit_depgroup_summary(xml);

	    int local_depgroup_id = 0;
            BOOST_FOREACH(rpt_depgroup_id_t depgroup_id, loop->depgroup_ids) {
		depgroup_info_ptr_t &d =
		    html_context.loop_registry->depgroup_infos[depgroup_id];
		local_depgroup_id++;

		// Don't include dep groups without issues
		if (!d->issues.empty())
		    emit_depgroup_section(xml,
					  d,
					  loop_locations,
					  local_depgroup_id);
	    }

	    emit_copyright(xml, true);
	    emit_terms_of_use(html_context, xml);
	} // T_BODY
    } // HTML_XMLNS
}

END_NAMESPACE;
END_NAMESPACE;
