/// @file tools/acurate/html_prefetch.cc

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#include "prefetch.hh"
#include "macros.hh"
#include "help_links.hh"
#include "file_png.hh"

#include "../diagram.hh"
#include "../file_repos.hh"
#include "../link_factory.hh"
#include "../loop_info.hh"
#include "../xml_file.hh"

#include <sstream>
#include <iomanip>
#include <set>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace std;
using namespace boost;
using namespace acumem::analysis;

html_prefetch_t::html_prefetch_t(html_context_t &html_context,
                                 rpt_issue_id_t issue_id)
    : html_issue_t(html_context, issue_id)
{
}

summary_type_bitset_t
html_prefetch_t::summary_type() const
{
    const rpt_prefetch_issue_t &issue(get_issue());

    summary_type_bitset_t result;

    result.set(SUMMARY_BANDWIDTH,
	       issue.problem == rpt_prefetch_issue_t::Too_Far);

    result.set(SUMMARY_LATENCY,
	       issue.problem == rpt_prefetch_issue_t::Unnecessary ||
	       issue.problem == rpt_prefetch_issue_t::Too_Far ||
	       issue.problem == rpt_prefetch_issue_t::Too_Close);

    return result;
}

string
html_prefetch_t::get_name() const
{
    const rpt_prefetch_issue_t &issue(get_issue());

    switch (issue.problem) {
    case rpt_prefetch_issue_t::Unnecessary:
	return TEXT_PREFETCH_HITS;
    case rpt_prefetch_issue_t::Too_Far:
	return TEXT_PREFETCH_FAR;
    case rpt_prefetch_issue_t::Too_Close:
	return TEXT_PREFETCH_CLOSE;
    default:
        assert(0);
        while (1);
    }
}

string
html_prefetch_t::get_icon() const
{
    const rpt_prefetch_issue_t &issue(get_issue());

    string icon;

    switch (issue.problem) {
    case rpt_prefetch_issue_t::Unnecessary:
	icon = PNG_PREFETCH_HITS;
	break;
    case rpt_prefetch_issue_t::Too_Far:
	icon = PNG_PREFETCH_FAR;
	break;
    case rpt_prefetch_issue_t::Too_Close:
	icon = PNG_PREFETCH_CLOSE;
	break;
    default:
        assert(0);
        while (1);
    }

    return html_context.link_factory->image_factory()->static_image_path(icon);
}

void
html_prefetch_t::annotate_source()
{
    const rpt_report_t &report(*html_context.report);
    const file_repos_t &file_repos(*html_context.file_repos);
    const rpt_prefetch_issue_t &issue(get_issue());

    fr_location_info_ptr_t location_info =
	file_repos.get_line_info(report, issue.location);

    location_info->add_reporter(
	location_reporter_ptr_t(
            new generic_location_reporter_t(*this, get_issue_id(), true)));
}

void
html_prefetch_t::emit_issue_file(xml_file_t &issue_xml)
{
    const rpt_report_t &report(*html_context.report);
    const file_repos_t &file_repos(*html_context.file_repos);
    const rpt_prefetch_issue_t &issue(get_issue());

    string type(get_name());
    string help;
    string stat_help;
    switch (issue.problem) {
    case rpt_prefetch_issue_t::Unnecessary:
	help = HELP_PREF_UNNEC;
	stat_help = HELP_PREF_UNNEC_STAT;
	break;
    case rpt_prefetch_issue_t::Too_Far:
	help = HELP_PREF_DISTANT;
	stat_help = HELP_PREF_DISTANT_STAT;
	break;
    case rpt_prefetch_issue_t::Too_Close:
	help = HELP_PREF_CLOSE;
	stat_help = HELP_PREF_CLOSE_STAT;
	break;
    default:
        assert(0);
	while(1);
    }

    fr_location_info_ptr_t location_info =
        file_repos.get_line_info(report, issue.location);

    XML_OBJECT(issue_xml) {
	T_DIV A_ID("sections") A_STYLE("display: none") {
            emit_issue_header(issue_xml);

	    // Title
	    SECTION(h1, "d", "Issue #" + get_display_number() + ": " + type,
		    get_icon(), help, false, false);

	    SECTION(h2, "s", "Statistics",
		    "", stat_help,
		    true, true) {
		T_TABLE A_CLASS("summary") {
		    switch (issue.problem) {
		    case rpt_prefetch_issue_t::Unnecessary:
			T_TR {
			    T_TH TEXT("Fetch ratio of the prefetch instruction");
			    stringstream s;
			    s << issue.fetch_ratio * 100 << "%";
			    T_TD TEXT(s.str());
			}
			break;

		    case rpt_prefetch_issue_t::Too_Far:
			T_TR {
			    T_TH TEXT("Average fetch ratio of the instructions "
                                    "using the prefetched data");
			    stringstream s;
			    s << issue.fetch_ratio * 100 << "%";
			    T_TD TEXT(s.str());
			}
			break;

		    case rpt_prefetch_issue_t::Too_Close:
			T_TR {
			    T_TH TEXT("Median number of memory accesses to the "
                                    "instructions using the prefetched data");
			    T_TD TEXT(lexical_cast<string>(issue.median_rdist));
			}
			break;
		    default:
                        assert(0);
			while(1);
		    }
		}
	    }

            // Report the prefetch instruction.
            {
                fr_location_info_vector_t loc_info;
                loc_info.push_back(location_info);
                section_instructions(xml, html_context,
                                     "Prefetch instruction",
                                     HELP_PREF_INSTR,
                                     loc_info,
                                     loc_info,
                                     2,
				     false, // visible
				     "iinstr");
            }

            // Report other instructions related to the issue.
            {
                const char *rel_heading;
                const char *rel_help;
                const rpt_location_set_t *rel_locs;
                switch (issue.problem) {
                case rpt_prefetch_issue_t::Unnecessary:
                    rel_heading = "Instructions using the prefetched data "
                        "before the prefetch";
                    rel_help = HELP_PREF_PREV;
                    rel_locs = &issue.prev_locations;
                    break;
                case rpt_prefetch_issue_t::Too_Far:
                    rel_heading = "Instructions using the prefetched data "
                        "after the prefetch";
                    rel_help = HELP_PREF_NEXT_DISTANT;
                    rel_locs = &issue.next_locations;
                    break;
                case rpt_prefetch_issue_t::Too_Close:
                    rel_heading = "Instructions using the prefetched data "
                        "after the prefetch";
                    rel_help = HELP_PREF_NEXT_CLOSE;
                    rel_locs = &issue.next_locations;
                    break;
                default:
                    assert(0);
                    while (1); /* Avoid warning on Solaris. */
                }
                fr_location_info_vector_t rel_locs_info =
                    file_repos.get_line_info(report, *rel_locs);
                section_instructions(xml, html_context, rel_heading, rel_help,
                                     rel_locs_info, rel_locs_info, 0,
                                     false, "inext");
            }

            emit_issue_footer(html_context, issue_xml);
	}
    }
}

void
html_prefetch_t::emit_summary(xml_file_t &xml,
			      summary_context_t &summary_context,
			      summary_type_t summary_type)
{
    const rpt_report_t &report(*html_context.report);
    const file_repos_t &file_repos(*html_context.file_repos);
    const rpt_prefetch_issue_t &issue(get_issue());

    // Reports statistics for just the prefetch, not the dependence group.
    fr_location_info_ptr_t location_info =
	file_repos.get_line_info(report, issue.location);
    fr_location_info_vector_t location_dep_info_vector;
    location_dep_info_vector.push_back(location_info);
    fr_location_info_vector_t location_loop_info_vector =
	file_repos.get_line_info(
            report, report.get_loop_incl_location_set(issue.loop_id));

    add_summary_line(
        xml,
        summary_context,
        get_rel_file_name(),
        get_name(),
        get_icon(),
        location_dep_info_vector,
        location_loop_info_vector,
        report.get_loop_stats_id(issue.loop_id),
        lexical_cast<string>(issue.loop_id),
        summary_type);
}

END_NAMESPACE;
END_NAMESPACE;
