/// @file tools/acurate/html_prefetchnta.cc

// Copyright (c) 2009-2010 Acumem AB.  All rights reserved.

#include "prefetchnta.hh"
#include "macros.hh"
#include "help_links.hh"

#include "../diagram.hh"
#include "../file_repos.hh"
#include "../link_factory.hh"
#include "../loop_info.hh"
#include "../xml_file.hh"

#include <sstream>
#include <iomanip>
#include <set>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace std;
using namespace boost;
using namespace acumem::analysis;

struct fraction_column_t : public extra_column_t {
    fraction_column_t(
        const string &heading,
        const rpt_stat_fp_vector_t &fractions)
        : heading(heading),
          fractions(fractions)
        {}

    void emit_heading(html_context_t &html_context,
                      xml_file_t &xml) const {
        T_TH TEXT(heading);
    }

    void emit_value(html_context_t &html_context,
                    xml_file_t &xml,
                    int line) const {
        T_TD html_util_t::emit_percent(xml, fractions[line]);
    }

    const string heading;
    const rpt_stat_fp_vector_t &fractions;
};

struct required_size_column_t : public extra_column_t {
    required_size_column_t(
        const string &heading,
        const rpt_stat_fp_vector_t &required_cache_sizes)
        : heading(heading),
          required_cache_sizes(required_cache_sizes)
        {}

    void emit_heading(html_context_t &html_context,
                      xml_file_t &xml) const {
        T_TH TEXT(heading);
    }

    void emit_value(html_context_t &html_context,
                    xml_file_t &xml,
                    int line) const {
        T_TD TEXT(html_context.value_to_string(required_cache_sizes[line]));
    }

    const string heading;
    const rpt_stat_fp_vector_t &required_cache_sizes;
};

html_prefetchnta_t::html_prefetchnta_t(html_context_t &html_context,
				       rpt_issue_id_t issue_id)
    : html_issue_t(html_context, issue_id)
{
}

summary_type_bitset_t
html_prefetchnta_t::summary_type() const
{
    summary_type_bitset_t result;

    result.set(SUMMARY_NT);

    return result;
}

string
html_prefetchnta_t::get_icon() const
{
    string icon = PNG_PREFETCH_NT;
    return html_context.link_factory->image_factory()->static_image_path(icon);
}

void
html_prefetchnta_t::annotate_source()
{
    annotate_generic(get_issue().depgroup_id, true);
}

void
html_prefetchnta_t::emit_issue_file(xml_file_t &issue_xml)
{
    const rpt_report_t &report(*html_context.report);
    const file_repos_t &file_repos(*html_context.file_repos);
    const rpt_prefetchnta_issue_t &issue(get_issue());

    rpt_sliced_stats_id_t global_stats_id = report.get_global_stats_id();
    const rpt_sliced_stats_t &global_stats =
	report.get_sliced_stats(global_stats_id);

    rpt_location_set_t pc1s(issue.pc1s.begin(), issue.pc1s.end());
    rpt_location_set_t pc2s(issue.pc2s.begin(), issue.pc2s.end());

    fr_location_info_vector_t pc1_info =
        file_repos.get_line_info(report, pc1s);
    fr_location_info_vector_t pc2_info =
        file_repos.get_line_info(report, pc2s);
    fr_location_info_vector_t location_dep_info_vector =
	file_repos.get_depgroup_line_info(report, issue.depgroup_id);
    fr_location_info_vector_t location_loop_info_vector =
	file_repos.get_loop_incl_line_info(report, issue.loop_id);

    html_context.interesting_depgroups.insert(issue.depgroup_id);
    html_context.interesting_loops.insert(issue.loop_id);

    XML_OBJECT(issue_xml) {
	T_DIV A_ID("sections") A_STYLE("display: none") {
            emit_issue_header(issue_xml);

	    // Title
	    SECTION(h1, "d", "Issue #" + get_display_number() + ": " + get_name(),
                    get_icon(), HELP_PREFETCHNTA, false, false);
	    dump_inferior(xml);

            // Last instructions to touch the cache lines before they are
            // evicted.

            fraction_column_t pc1_fraction_column("% of non-temporal reuses", issue.pc1_fractions);
            required_size_column_t pc1_sd_column("Required cache size", issue.pc1_required_cache_sizes);
            extra_columns_t pc1_extra_columns;
            pc1_extra_columns.push_back(&pc1_fraction_column);
            pc1_extra_columns.push_back(&pc1_sd_column);

            SECTION(h2, "idata",
		    "Statistics for the reuses of the non-temporal data",
                    "", HELP_PREFETCHNTA_STATS, true, true) {
                T_TABLE A_CLASS("summary") {
                    T_TR {
                        T_TH A_ALIGN("left") {
                            TEXT("% fo fetches");
                        }
                        T_TD {
                            double global_fetches =
                                report.get_sliced_stats_global(global_stats).get_total_fetches();
                            emit_scientific(xml, issue.fetches, global_fetches, true, false);
                        }
                    }
                    T_TR {
                        T_TH A_ALIGN("left") {
                            TEXT("Fetch ratio");
                        }
                        T_TD {
                            emit_scientific(xml, issue.fetches, issue.accesses, true, false);
                        }
                    }
                    T_TR {
                        T_TH A_ALIGN("left") {
                            TEXT("Fetches");
                        }
                        T_TD {
                            emit_scientific(xml, issue.fetches, 0.0, false, true);
                        }
                    }
                }
            }

	    section_instructions(xml, html_context,
				 "Last instructions to touch the data before it is evicted",
				 HELP_PREFETCHNTA_BEFORE,
				 pc1_info,
				 location_loop_info_vector,
				 0,
				 false,
				 "last_before",
                                 pc1_extra_columns);

            // First instructions to touch the cache lines after they are
            // evicted.

            fraction_column_t pc2_fraction_column("% of non-temporal reuses", issue.pc2_fractions);
            extra_columns_t pc2_extra_columns;
            pc2_extra_columns.push_back(&pc2_fraction_column);

	    section_instructions(xml, html_context,
				 "First instructions to touch the data after it is evicted",
				 HELP_PREFETCHNTA_AFTER,
				 pc2_info,
				 location_loop_info_vector,
				 0,
				 false,
				 "first_after",
                                 pc2_extra_columns);

	    section_statistics(xml, html_context,
			       "Instruction group statistics",
			       HELP_ISSUE_STATISTICS,
			       report.get_depgroup_stats_id(issue.depgroup_id),
                               "d" + lexical_cast<string>(issue.depgroup_id),
			       location_dep_info_vector,
			       location_loop_info_vector,
			       false, // initially visible
			       false, // is_loop
			       false, // show_avoidable
			       "istat");

	    // instruction group instructions
	    section_instructions(xml, html_context,
				 "Instructions in instruction group",
				 HELP_ISSUE_INSTRUCTIONS,
				 location_dep_info_vector,
				 location_loop_info_vector,
				 2,
				 false,
				 "iinstr");

	    // -----------------------------------------------------
	    // Common section for describing the Loop

	    all_loop_sections(xml, html_context,
			      issue.loop_id, location_loop_info_vector);

	    // -----------------------------------------------------
	    // done

            emit_issue_footer(html_context, issue_xml);


	}
    }
}

void
html_prefetchnta_t::emit_summary(xml_file_t &xml,
                                 summary_context_t &summary_context,
                                 summary_type_t summary_type)
{
    const rpt_report_t &report(*html_context.report);
    const file_repos_t &file_repos(*html_context.file_repos);
    const rpt_prefetchnta_issue_t &issue(get_issue());

    rpt_location_set_t pc1s(issue.pc1s.begin(), issue.pc1s.end());

    const fr_location_info_vector_t location_instr_info_vector =
	file_repos.get_line_info(report, pc1s);
    const fr_location_info_vector_t location_loop_info_vector =
	file_repos.get_loop_incl_line_info(report, issue.loop_id);

    add_summary_line(
        xml,
        summary_context,
        get_rel_file_name(),
        get_name(),
        get_icon(),
        location_instr_info_vector,
        location_loop_info_vector,
        (rpt_sliced_stats_id_t)-1, // Ugly. Actually, this should never be used.
        lexical_cast<string>(issue.loop_id),
        summary_type,
        issue.fetches,
        issue.required_cache_size);
}

END_NAMESPACE;
END_NAMESPACE;
