/// @file tools/acurate/html_source.cc

// Copyright (c) 2009-2010 Acumem AB.  All rights reserved.

#include "source.hh"
#include "file_png.hh"
#include "help_links.hh"
#include "macros.hh"
#include "../link_factory.hh"
#include "../loop_info.hh"
#include "library/lang/am_assert.hh"

#include <boost/lexical_cast.hpp>

#include <sstream>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace std;
using namespace boost;

/// Sort statistics for a single source code line based on unique call
/// stack first, then pc within each call stack.
struct stack_then_pc_t {
    stack_then_pc_t(const rpt_report_t &report,
		    bool &different_stacks)
	: report(report),
	  different_stacks(different_stacks) {}

    const rpt_report_t report;
    bool &different_stacks;

    int compare_addrs(rpt_address_id_t l, rpt_address_id_t r) const {
        rpt_address_t l_addr(report.get_address(l));
        rpt_address_t r_addr(report.get_address(r));
        if (l_addr < r_addr)
            return -1;
        if (l_addr > r_addr)
            return 1;
        return 0;
    }

    /// Do a less than comparison of two stacks and pcs.
    bool operator()(const location_issue_number_t &l,
                    const location_issue_number_t &r) const {
        const rpt_trunc_stack_t &l_stack(
            report.get_location_trunc_stack(l.location->location_id));
        const rpt_trunc_stack_t &r_stack(
            report.get_location_trunc_stack(r.location->location_id));

        // Start by checking if the stacks differ.
        if (l_stack != r_stack) {
            different_stacks = true;

            // Do a lexigraphical comparison of the stacks.
            for (uint32_t i = 0; i < r_stack.size(); i++) {
                if (i == l_stack.size())
                    // l_stack is a prefix of r_stack, but r_stack is longer
                    // than l_stack, so l_stack is less than r_stack.
                    return true;

                int cmp = compare_addrs(l_stack[i], r_stack[i]);
                if (cmp < 0)
                    // The stacks were equal up to element i-1, but element i
                    // of l_stack is less than element i of r_stack, so l_stack
                    // is less than r_stack.
                    return true;
                if (cmp > 0)
                    // The stacks were equal up to element i-1, but element i
                    // of l_stack is greater than element i of r_stack, so
                    // l_stack is not less than r_stack.
                    return false;
            }

            if (l_stack.size() > r_stack.size())
                // r_stack is a prefix to l_stack, but l_stack is longer than
                // r_stack, so l_stack is not less than r_stack.
                return false;

            AM_ASSERT_MSG("Stacks that are not equal are equal.");
        }

        // The stacks were equal, so return true if l_pc is less than r_pc.
        rpt_address_id_t l_pc(report.get_location_pc_id(l.location->location_id));
        rpt_address_id_t r_pc(report.get_location_pc_id(r.location->location_id));
        return compare_addrs(l_pc, r_pc) < 0;
    }
};

// XXX Consider moving this function to the line_info_t class
std::string
html_source_t::get_function_names(html_context_t &html_context,
                                  const line_info_t &line_info)
{
    const rpt_report_t &report(*html_context.report);

    string names;

    BOOST_FOREACH(const acumem::acurate::pc_info_map_t::value_type &p,
		  line_info.pc_info_map) {
        const rpt_debug_pc_t &debug_pc(report.debug_pcs.at(p.second->debug_pc_id));
        if (debug_pc.function_id != RPT_NO_FUNCTION) {
            if (!names.empty())
                names += " ";
            names += report.functions.at(debug_pc.function_id).name;
        }
    }

    return names;
}


/// Display a source code line, just the line number and text but no
/// statistics
void
html_source_t::emit_simple_line(xml_file_t &xml,
				const string &row_s,
				const string &str)
{
    NL T_TR A_ID(row_s) {
	T_TD A_CLASS("line") T_A A_NAME(row_s) TEXTN(row_s + " ");
	T_TD A_CLASS("metric") TEXTN_T("&nbsp;");
	T_TD T_PRE TEXTN(str);
    }
}

/// Display a source code line, just the line number and text
/// including % of fetches statistics
void
html_source_t::emit_fr_source_line(xml_file_t &xml,
				   html_context_t &html_context,
				   const string &row_s,
				   const string &str,
				   double tot_fetches,
				   double global_fetches)
{
    string onclick =
	"toggle('V" + row_s + "'); "
	"toggle('S" + row_s + "'); "
	"toggle('P" + row_s + "'); "
	"toggle('M" + row_s + "'); ";

    NL T_TR A_ID(row_s) A_CLASS("advice") {
	T_TD A_CLASS("line") T_A A_NAME(row_s) {
	    TEXTN(row_s + " ");
	    T_IMG
		A_SRC(image_path(html_context, PNG_PLUS))
		A_ID("P" + row_s) A_ONCLICK(onclick);
	    T_IMG
		A_SRC(image_path(html_context, PNG_MINUS))
		A_STYLE("display:none")
		A_ID("M" + row_s) A_ONCLICK(onclick);
	}
	T_TD A_CLASS("metric") {
	    // XXX: Should use emit_percentage_of_fetches.
	    NL emit_scientific(xml, tot_fetches, global_fetches, true, false);
	}
	T_TD T_PRE TEXTN(str);
    }
}


void
html_source_t::emit_horizontal_rule_maybe(xml_file_t &xml,
					  int row, int &last_row)
{
    if (last_row != 0 && last_row != row - 1)
	NL T_TR T_TD A_COLSPAN("3") T_HR A_WIDTH("80%") A_COLOR("lightgray");
    last_row = row;
}

static void
emit_prefetch_help(xml_file_t &xml,
                   html_context_t &html_context,
                   const char *help_section)
{
    string help_url = html_context.link_factory->get_help_link(help_section);

    T_A A_HREF(help_url)
        A_TARGET("help")
        A_ONCLICK(string("return popitup('") + help_url + "')") TEXT("*");
}

void
html_source_t::emit_verbose_table(xml_file_t &xml,
				  html_context_t &html_context,
				  const std::string &row_s,
				  location_issue_vector_t &lines)
{
    const rpt_report_t &report(*html_context.report);

    rpt_sliced_stats_id_t global_stats_id = report.get_global_stats_id();

    NL T_TR A_STYLE("display:none;") A_ID("V" + row_s)
	A_CLASS("verbose") {
	T_TD A_CLASS("line") TEXTN(" ");
	T_TD A_COLSPAN("2")
	    T_TABLE A_CLASS("srcverb") {

	    T_THEAD {
		T_TD T_P TEXT("% of fetches");
		T_TD T_P TEXT("Miss ratio");
		T_TD T_P TEXT("Fetch ratio");
		T_TD T_P TEXT("WB ratio");
		if (report.mt_sampled) {
		    T_TD T_P TEXT("Fetch Util");
		    T_TD T_P TEXT("WB Util");
		} else {
		    T_TD T_P TEXT("Utilization");
		}
		T_TD T_P TEXT("PC");
		T_TD T_P TEXT("Type");
		T_TD T_P TEXT("Issues");
	    }

	    // Sort the advice in an unusual order, first stack,
	    // then pc.  At the same time, dig out whether there
	    // actually exist more than one unique stack.
	    bool different_stacks = false;
	    stack_then_pc_t stp(report, different_stacks);
	    sort(lines.begin(), lines.end(), stp);

	    rpt_trunc_stack_t last_stack;
	    bool alt = true;
	    int stack_context_count = 0;

	    BOOST_FOREACH(location_issue_number_t &x, lines) {
                rpt_location_id_t loc = x.location->location_id;

		// This is to find out if the stack has changed
		// from previous line, and in that case flip the
		// 'alt' variable to get a different color coding.
                const rpt_trunc_stack_t &stack =
                    report.trunc_stacks[
                        report.locations[loc].trunc_stack_id];
		if (stack != last_stack || stack_context_count == 0) {
		    alt = !alt;
		    last_stack = stack;
		    stack_context_count ++;
		}

		T_TR {
		    if (different_stacks)
			A_TITLE("Unique stack context: \r\n" +
				lexical_cast<string>(stack_context_count));

		    if (alt)
			A_CLASS("alt2");
		    else
			A_CLASS("alt1");

                    rpt_sliced_stats_id_t location_stats_id =
                        report.get_location_sliced_stats_id(loc);
                    rpt_access_classes_t access_classes =
                        report.get_location_access_classes(loc);

		    // column: miss/mr/su
		    T_TD T_P {
			emit_percentage_of_fetches(
			    html_context, xml, location_stats_id,
			    global_stats_id, false);
		    }
		    T_TD T_P {
			emit_miss_ratio(html_context, xml, location_stats_id, false);
			if (access_classes & RPT_Access_Class_PF)
			    emit_prefetch_help(xml, html_context, HELP_PREFETCH_MISS_RATIO);
		    }
		    T_TD T_P {
			emit_fetch_ratio(html_context, xml, location_stats_id, false);
		    }
		    T_TD T_P {
			emit_write_back_ratio(html_context, xml, location_stats_id, false);
		    }

		    depgroup_info_t &depgroup =
			*html_context.loop_registry->depgroup_infos[
			    report.get_location_depgroup(loc)];
                    rpt_sliced_stats_id_t depgroup_stats_id =
                        report.get_depgroup_stats_id(depgroup.id);
		    if (report.mt_sampled) {
			T_TD T_P {
			    emit_fetch_utilization(html_context, xml, depgroup_stats_id, false);
			}
			T_TD T_P {
			    emit_wb_utilization(html_context, xml, depgroup_stats_id, false);
			}
		    } else {
			T_TD T_P {
			    emit_utilization(html_context, xml, depgroup_stats_id, false);
			}
		    }

		    // column: pc address
		    stringstream s;
		    pc_info_ptr_t p = x.location->pc_info.lock();
		    if (p.get())
			s << html_util_t::format_address(p->pc) << " ";
		    T_TD T_P TEXT(s.str());

		    // column: access type
		    T_TD T_P {
			string access_class_string;
			if (access_classes & RPT_Access_Class_NTA)
			    access_class_string += "NT ";
			if ((access_classes & RPT_Access_Class_RD) &&
			    !(access_classes & RPT_Access_Class_PF))
			    access_class_string += "R";
			if (access_classes & RPT_Access_Class_WR)
			    access_class_string += "W";
			if (access_classes & RPT_Access_Class_PF)
			    access_class_string += "PF";
			if (access_classes & RPT_Access_Class_Flush)
			    access_class_string = "Flush";

			TEXT(access_class_string);
		    }

		    // column: advice
		    T_TD T_P A_VALIGN("middle")
			BOOST_FOREACH(shared_ptr<location_reporter_t> &reporter,
				      x.location->location_reporter_vector)
			    reporter->report(xml);
		}
	    }
	    TEXT(" "); // To force table to have some content at all
	}
    }
}


void
html_source_t::get_fetches_and_location_issues(html_context_t &html_context,
					       const line_info_t &line_info,
					       double &tot_fetches,
					       location_issue_vector_t &lines)
{
    const rpt_report_t &report(*html_context.report);

    const double global_fetches(
        report.get_sliced_stats_global(
	    report.get_sliced_stats(
		report.get_global_stats_id())).get_total_fetches());

    BOOST_FOREACH(const acumem::acurate::pc_info_map_t::value_type &p,
		  line_info.pc_info_map) {
	BOOST_FOREACH(location_info_map_t::value_type &l,
		      p.second->location_info_map) {
	    const rpt_location_id_t loc = l.second->location_id;
	    const double location_fetches(
                report.get_location_global_stats(loc).get_total_fetches());

	    // Only add those lines that contribute to
	    // something, or are prefetches (which respond
	    // to force()).
	    if (l.second->force() ||
		location_fetches > global_fetches * 0.001) {
		lines.push_back(location_issue_number_t(l.second, l.first));
		tot_fetches += location_fetches;
	    }
	}
    }
}

/// Terse section
void
html_source_t::emit_terse_issue_line(xml_file_t &xml,
				     html_context_t &html_context,
				     const std::string &row_s,
				     const line_info_t &line_info)
{
    const rpt_report_t &report(*html_context.report);

    const rpt_stats_t &global_stats(
        report.get_sliced_stats_global(
	    report.get_sliced_stats(
		report.get_global_stats_id())));

    const double global_fetches(global_stats.get_total_fetches());
    const double global_upgrades(global_stats.get_total_upgrades());
    const double global_write_backs(global_stats.get_total_write_backs());

    NL T_TR A_ID("S" + row_s) A_CLASS("terse") {
	T_TD A_CLASS("line") TEXTN(" ");
	T_TD A_VALIGN("top") A_COLSPAN("2") {
	    set<int> seen_issues;
	    BOOST_FOREACH(const acumem::acurate::pc_info_map_t::value_type &p,
			  line_info.pc_info_map) {
		BOOST_FOREACH(location_info_map_t::value_type &l,
			      p.second->location_info_map) {
                    const rpt_stats_t &location_stats(
                        report.get_sliced_stats_global(
			    report.get_sliced_stats(
				report.get_location_sliced_stats_id(
				    l.second->location_id))));

		    // Only add those lines that contribute to
		    // something, or are prefetches

		    if (l.second->force() ||
			location_stats.get_total_fetches() >
			global_fetches * 0.001 ||
			location_stats.get_total_upgrades() >
			global_upgrades * 0.001 ||
			location_stats.get_total_write_backs() >
                        global_write_backs * 0.001)
		    {
			BOOST_FOREACH(location_reporter_ptr_t &reporter,
				      l.second->location_reporter_vector) {
			    // Only report each issue once per line
			    rpt_issue_id_t issue_number = reporter->get_issue_number();
			    if (seen_issues.insert(issue_number).second)
				reporter->report(xml);
			}
		    }
		}
	    }
	    TEXT(" ");
	}
	T_TD;
    }
}


void
html_source_t::emit_source_line_info(html_context_t &html_context,
				     xml_file_t &xml,
				     const line_info_map_t &line_info_map,
				     int row,
				     const string &str,
				     bool source_unknown,
				     int &last_row)
{
    const rpt_report_t &report(*html_context.report);

    string row_s = lexical_cast<string>(row);
    string str_copy;

    // Add a horizontal rule to indicate non-contiguous source line blocks
    emit_horizontal_rule_maybe(xml, row, last_row);


    // Look up advice for this row
    line_info_map_t::const_iterator i = line_info_map.find(row);

    // Grab function names if source is not available
    if (source_unknown && i != line_info_map.end())
	str_copy = get_function_names(html_context, *i->second);
    else
	str_copy = str;

    // If there is no statistics for this line, or no advice just
    // output the line
    if (i == line_info_map.end() ||
	(i->second->num_reporters == 0)) {
	emit_simple_line(xml, row_s, str_copy);
	return;
    }

    const line_info_t &line_info = *i->second;

    // Find which (locations,issues) this source lines contains
    location_issue_vector_t lines;
    double tot_fetches = 0;
    get_fetches_and_location_issues(html_context, line_info,
				    tot_fetches, lines);

    // If there was no reportable statistics on this line, then we
    // shouldn't display anything except the text
    if (lines.empty()) {
	emit_simple_line(xml, row_s, str_copy);
	return;
    }

    // Emit a line with total fetch ratio
    const double global_fetches(
        report.get_sliced_stats_global(
	    report.get_sliced_stats(
		report.get_global_stats_id())).get_total_fetches());
    emit_fr_source_line(xml, html_context, row_s, str, tot_fetches,
			global_fetches);

    // Verbose section (which folds out)
    emit_verbose_table(xml, html_context, row_s, lines);

    // The alternative terse section with issue icons
    if (line_info.num_reporters)
	emit_terse_issue_line(xml, html_context, row_s, line_info);
}

void
html_source_t::emit_placeholder_lines(html_context_t &html_context,
				      xml_file_t &xml,
				      const line_info_map_t &line_info_map,
				      const std::string &filename)
{
    T_DIV A_CLASS("warning") T_P {
	T_B TEXT("File not found: ");
	TEXT(filename);
    }

    T_TABLE A_CLASS("source") {
	NL T_TR {
	    T_TH TEXT("Line");
	    T_TH A_WIDTH("60") TEXT("Fetches");
	    T_TH TEXT("Function");
	}

	int last_row = 0;
	BOOST_FOREACH(line_info_map_t::value_type v,
		      line_info_map) {
	    emit_source_line_info(html_context, xml, line_info_map,
				  v.first, "", true, last_row);
	}
    }
}

void
html_source_t::emit_source_lines(html_context_t &html_context,
				 xml_file_t &xml,
				 const line_info_map_t &line_info_map,
				 istream &source)
{
    T_TABLE A_CLASS("source") {
	NL T_TR {
	    T_TH TEXT("Line");
	    T_TH TEXT_T("%&nbsp;of&nbsp;fetches");
	    T_TH TEXT("Source");
	}

	int row = 1;
	string l;

	int last_row = 0;
	while (!source.eof()) {
	    getline(source, l);
	    emit_source_line_info(html_context, xml, line_info_map,
				  row, l, false, last_row);
	    row++;
	}
    }
}


void
html_source_t::write_annotated_file(xml_file_t &xml,
				    html_context_t &html_context,
				    const std::string &filename,
				    const line_info_map_t &line_info_map)
{
    const rpt_report_t &report(*html_context.report);

    int file_id = report.get_source_file_id(filename);

    NL T_H1 TEXT(filename);
    const rpt_source_file_t *source_file = NULL;
    if (file_id != -1) {
	source_file = &report.source_files[file_id];
    }

    if (source_file && source_file->data && source_file->data_length > 0) {
        istringstream is(string(source_file->data.get(),
                                source_file->data_length));
        emit_source_lines(html_context, xml, line_info_map, is);
    }
    else
        // If there is no source code file available, make up some
        // placeholder lines.
        emit_placeholder_lines(html_context, xml, line_info_map, filename);

    emit_copyright(xml, true);
    emit_terms_of_use(html_context, xml);
}

END_NAMESPACE;
END_NAMESPACE;
