/// @file tools/acurate/html_source.hh

// Copyright (c) 2009 Acumem AB.  All rights reserved.


#ifndef HTML_SOURCE_HH
#define HTML_SOURCE_HH

#include <string>
#include "issue.hh"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace std;


/// Grouping a location id with all the issues for that location. Used
/// to group issues for a source line based on their different call
/// stack contexts, a.k.a location.
struct location_issue_number_t {
    location_issue_number_t(fr_location_info_ptr_t location, int issue_number)
	: location(location),
	  issue_number(issue_number) {}
    fr_location_info_ptr_t location;
    int issue_number;
};

typedef std::vector<location_issue_number_t> location_issue_vector_t;


/// Emit annotated source code
class html_source_t : public html_util_t {
public:

    /// Write an html file with annotated source code
    static void write_annotated_file(xml_file_t &xml,
				     html_context_t &html_context,
				     const std::string &filename,
				     const line_info_map_t &line_info_map);

private:

    /// Get all function names for a given source file / line number
    static std::string get_function_names(html_context_t &html_context,
                                          const line_info_t &line_info);

    /// Summarize number of fetches and (location,issue number) per
    /// source file / line number
    static void get_fetches_and_location_issues(html_context_t &html_context,
						const line_info_t &line_info,
						double &tot_fetches,
						location_issue_vector_t &lines);


    // ----------------------------------------------------------------------
    // emitting components of a line

    /// Emit a line consisting of line number and source code line
    static void emit_simple_line(xml_file_t &xml,
				 const string &row_s,
				 const string &str);

    /// Emit a line consisting of line number, % of fetches, and source
    static void emit_fr_source_line(xml_file_t &xml,
				    html_context_t &html_context,
				    const string &row_s,
				    const string &str,
				    double tot_fetches,
				    double global_fetches);

    /// Emit a horizontal line to mark sections with noncontiguous
    /// line numbers
    static void emit_horizontal_rule_maybe(xml_file_t &xml,
					   int row, int &last_row);

    /// Emit a table for a source line with call-stack and pc sorted
    /// broken down statistics.
    static void emit_verbose_table(xml_file_t &xml,
				   html_context_t &html_context,
				   const std::string &row_s,
				   location_issue_vector_t &lines);

    /// Emit a line with issue icons
    static void emit_terse_issue_line(xml_file_t &xml,
				      html_context_t &html_context,
				      const std::string &row_s,
				      const line_info_t &line_info);


    // ----------------------------------------------------------------------
    // emitting a single line with annotations

    /// Emit a single line from a source file including all its annotations
    static void emit_source_line_info(html_context_t &html_context,
				      xml_file_t &xml,
				      const line_info_map_t &line_info_map,
				      int row,
				      const string &str,
				      bool source_unknown,
				      int &last_row);

    // ----------------------------------------------------------------------
    // emitting a file worth of lines

    /// Emit placeholder lines for all lines in a source file which
    /// could not be found.
    static void emit_placeholder_lines(html_context_t &html_context,
				       xml_file_t &xml,
				       const line_info_map_t &line_info_map,
				       const std::string &filename);

    /// Emit a file worth of annotated source code lines
    static void emit_source_lines(html_context_t &html_context,
				  xml_file_t &xml,
				  const line_info_map_t &line_info_map,
				  istream &source);
};

END_NAMESPACE;
END_NAMESPACE;

#endif
