/// @file tools/acurate/html_summary.cc

// Copyright (c) 2009-2010 Acumem AB.  All rights reserved.

#include "summary.hh"
#include "macros.hh"
#include "help_links.hh"
#include "file_png.hh"

#include "../consumer.hh"
#include "../link_factory.hh"

#include <boost/lexical_cast.hpp>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace boost;

void
html_summary_t::emit_summary_diagrams(xml_file_t &xml,
				      html_context_t &html_context,
				      int rows)
{
    const rpt_report_t &report(*html_context.report);
    image_factory_ptr_t image_factory = html_context.link_factory->image_factory();

    T_TD A_VALIGN("top") A_ROWSPAN(lexical_cast<string>(rows)) {
	T_TABLE A_CLASS("summary") {
	    T_TR T_TH A_COLSPAN("2")
		emit_glossary_phrase(html_context, xml, "Miss/Fetch ratio", HELP_DIAGRAM_FR);
	    T_TR T_TD A_COLSPAN("2") A_ALIGN("center") {
		T_IMG if (html_context.conf.use_mr_scale)
		    A_SRC(image_factory->global_diagram_path(PNG_GLOBAL_DIAGRAM_FR100));
		else
		    A_SRC(image_factory->global_diagram_path(PNG_GLOBAL_DIAGRAM_FR));
	    }

	    if (report.mt_sampled) {
		T_TR T_TH A_COLSPAN("2")
		    emit_glossary_phrase(html_context, xml, "Write-back ratio",
					 HELP_DIAGRAM_WB);

		T_TR T_TD A_COLSPAN("2") A_ALIGN("center") {
		    T_IMG if (html_context.conf.use_mr_scale)
			A_SRC(image_factory->global_diagram_path(PNG_GLOBAL_DIAGRAM_WB100));
		    else
			A_SRC(image_factory->global_diagram_path(PNG_GLOBAL_DIAGRAM_WB));
		}
	    }

	    T_TR T_TH A_COLSPAN("2")
		emit_glossary_phrase(html_context, xml, "Utilization", HELP_DIAGRAM_UTIL);

	    T_TR T_TD A_COLSPAN("2") A_ALIGN("center") {
		T_IMG A_SRC(image_factory->global_diagram_path(PNG_GLOBAL_DIAGRAM_SU));
	    }
	}
    }
}

void
html_summary_t::emit(xml_file_t &xml, html_context_t &html_context)
{
    const rpt_report_t &report(*html_context.report);

    rpt_sliced_stats_id_t global_stats_id = report.get_global_stats_id();

    bool result_unreliable = html_context.result_unreliable();
    bool warn_virtualized = html_context.warn_virtualized();

    if (result_unreliable || warn_virtualized) {
	T_DIV A_CLASS("warning") {
	    T_H1 TEXT("Warning!");
	    T_UL {
		if (result_unreliable) T_LI {
		    T_P TEXT("These results are based on too few samples to be "
			     "reliable. The sample file contains " +
			     lexical_cast<string>(report.sample_count) +
			     " samples of the selected line size, but " +
			     lexical_cast<string>(MIN_RELIABLE_SAMPLE_COUNT) +
			     " samples are generally required for reliable results.");

		    T_P TEXT("Re-sample the application for a longer time "
			     "or with a shorter sample period for a "
			     "reliable result.");
		}

		if (warn_virtualized) T_LI {
		    T_P TEXT("CPU and cache hierarchy detection does not work "
			     "reliably in a virtualized environment. You need "
			     "to manually override the number of caches or "
			     "manually select a processor.");
		}
	    }
	}
    }


    // Figure out how many rows there are (to give a proper row_span)
    int rows = 9;
    bool burst_sampled = report.burst_sampled;
    bool mt_sampled = report.mt_sampled;
    bool mt_advice = html_context.feature_mt_advice;
    bool sw_prefetch = html_context.feature_sw_prefetch;

    if (! burst_sampled)
	rows += 3; // absolute accesses, misses, fetches

    if (mt_sampled) {
	if (!burst_sampled) {
	    if (mt_advice)
		rows += 2; // Absolute write-backs and upgrades
	    else
		rows += 1; // Absolute write-backes
	}

	if (mt_advice)
	    rows += 8; // WB ratio, upgr ratio, comm ratio, fetch
		       // util, wb util, comm util, num cpu, num
		       // caches
	else
	    rows += 3; // WB ratio, fetch util, wb util

    } else
	rows += 1; // util

    if (sw_prefetch)
	rows += 1;


    T_TABLE A_CLASS("summary") {
	NL T_TR {
	    T_TH A_COLSPAN("2") TEXT("Global statistics");
	    emit_summary_diagrams(xml, html_context, rows);
	}

	if (!burst_sampled) {
	    NL T_TR {
		T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Accesses", HELP_SUMMARY_ACCESSES);
		T_TD emit_accesses(html_context, xml, global_stats_id, true);
	    }
	    NL T_TR {
		T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Misses", HELP_SUMMARY_MISSES);
		T_TD emit_misses(html_context, xml, global_stats_id, true);
	    }

	    NL T_TR {
		T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Fetches", HELP_SUMMARY_FETCHES);
		T_TD emit_fetches(html_context, xml, global_stats_id, true);
	    }

	    if (mt_sampled) {
		NL T_TR {
		    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Write-backs", HELP_SUMMARY_WRITE_BACKS);
		    T_TD emit_write_backs(html_context, xml, global_stats_id, true);
		}

		if (mt_advice)
		    NL T_TR {
			T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Upgrades", HELP_SUMMARY_UPGRADES);
			T_TD emit_upgrades(html_context, xml, global_stats_id, true);
		    }
	    }
	}

	NL T_TR {
	    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Miss ratio", HELP_SUMMARY_MISSRATIO);
	    T_TD emit_miss_ratio(html_context, xml, global_stats_id, true);
	}

	NL T_TR {
	    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Fetch ratio", HELP_SUMMARY_FETCHRATIO);
	    T_TD emit_fetch_ratio(html_context, xml, global_stats_id, true);
	}

	if (mt_sampled) {
	    NL T_TR {
		T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Write-back ratio", HELP_SUMMARY_WRITE_BACKRATIO);
		T_TD emit_write_back_ratio(html_context, xml, global_stats_id, true);
	    }

	    if (mt_advice) {
		NL T_TR {
		    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Upgrade ratio", HELP_SUMMARY_UPGRADERATIO);
		    T_TD emit_upgrade_ratio(html_context, xml, global_stats_id, true);
		}

		NL T_TR {
		    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Communication ratio", HELP_SUMMARY_COMMRATIO);
		    T_TD emit_comm_ratio(html_context, xml, global_stats_id, true);
		}
	    }
	}

	if (mt_sampled) {
	    NL T_TR {
		T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Fetch utilization", HELP_SUMMARY_FETCH_UTIL);
		T_TD emit_fetch_utilization(html_context, xml, global_stats_id, true);
	    }

	    NL T_TR {
		T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Write-back utilization", HELP_SUMMARY_WB_UTIL);
		T_TD emit_wb_utilization(html_context, xml, global_stats_id, true);
	    }

	    if (mt_advice)
		NL T_TR {
		    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Communication utilization",
							      HELP_SUMMARY_COMM_UTIL);
		    T_TD emit_comm_utilization(html_context, xml, global_stats_id, true);
		}
	} else {
	    NL T_TR {
		T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Utilization", HELP_SUMMARY_UTIL);
		T_TD emit_utilization(html_context, xml, global_stats_id, true);
	    }
	}

	T_TR T_TH A_COLSPAN("2") TEXT("Analysis parameters");

	NL T_TR {
	    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Processor model", HELP_SUMMARY_CPU_MODEL);
            T_TD TEXT(report.cpu_model);
	}

	if (mt_sampled) {
	    if (mt_advice) {
		NL T_TR {
		    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Number of CPUs", HELP_SUMMARY_NUM_CPUS);
		    T_TD TEXT(lexical_cast<string>(
				  report.num_cpus));
		}

		NL T_TR {
		    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Number of caches", HELP_SUMMARY_NUM_CACHES);
		    T_TD TEXT(lexical_cast<string>(
				  report.num_caches));
		}
	    }
	}

	NL T_TR {
	    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Cache level", HELP_SUMMARY_CACHE_LEVEL);
	    T_TD TEXT(lexical_cast<string>(
			  report.cache_level));
	}

	NL T_TR {
	    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Cache size", HELP_SUMMARY_CACHESIZE);
	    T_TD TEXT(html_context.value_to_string((double)report.cache_size));
	}

	NL T_TR {
	    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Line size", HELP_SUMMARY_LINESIZE);
	    T_TD TEXT(html_context.value_to_string((double)report.line_size));
	}

	T_TR {
	    T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Replacement policy",
						      HELP_SUMMARY_REPL);
	    T_TD TEXT(report.replacement_policy);
	}

	if (sw_prefetch) {
	    T_TR {
		T_TH A_ALIGN("left") emit_glossary_phrase(html_context, xml, "Software prefetches active",
							  HELP_SUMMARY_SWPF_ACTIVE);
		T_TD TEXT(report.swpf_active ? "Yes" : "No");
	    }
	}
    }

    emit_copyright(xml, true);
    emit_terms_of_use(html_context, xml);
}

END_NAMESPACE;
END_NAMESPACE;
