/// @file tools/acurate/html/topleft.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include "topleft.hh"

#include "about.hh"
#include "execution.hh"
#include "files.hh"
#include "help_links.hh"
#include "loop.hh"
#include "macros.hh"
#include "summary.hh"

#include "../link_factory.hh"
#include "../loop_info.hh"

#include <boost/bind.hpp>
#include <iostream>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace std;
using namespace boost;

html_topleft_t::html_topleft_t(html_context_t &html_context)
    : html_context(html_context)
{}

void
html_topleft_t::emit(xml_file_t &xml)
{
    tab_factory_ptr_t tab_factory = html_context.link_factory->tab_factory();

    tab_factory->begin(xml, "toptab");
    tab_factory->new_tab(xml, "Issues",
			 bind(&html_topleft_t::emit_issue_tab, this, _1));
    tab_factory->new_tab(xml, "Loops",
			 bind(&html_topleft_t::emit_loop_table, this, _1));
    tab_factory->new_tab(xml, "Summary",
			 bind(&html_topleft_t::emit_summary, this, _1));
    tab_factory->new_tab(xml, "Files",
			 bind(&html_topleft_t::emit_file_list, this, _1));
    tab_factory->new_tab(xml, "Execution",
			 bind(&html_topleft_t::emit_execution_statistics, this, _1));
    tab_factory->new_tab(xml, "About/Help",
			 bind(&html_topleft_t::emit_about, this, _1));
    tab_factory->end(xml);

}

void
html_topleft_t::emit_issue_tab(xml_file_t &xml)
{
    const rpt_report_t &report(*html_context.report);
    tab_factory_ptr_t tab_factory = html_context.link_factory->tab_factory();

    tab_factory->begin(xml, "issuetab");
    tab_factory->new_tab(xml, "Bandwidth Issues",
			 bind(&html_topleft_t::emit_bandwidth_issue_list, this, _1));

    tab_factory->new_tab(xml, "Latency Issues",
			 bind(&html_topleft_t::emit_latency_issue_list, this, _1));

    if (report.mt_sampled && html_context.feature_mt_advice)
	tab_factory->new_tab(xml, "Multi-Threading Issues",
			     bind(&html_topleft_t::emit_mt_issue_list, this, _1));

    tab_factory->new_tab(xml, "Pollution Issues",
			 bind(&html_topleft_t::emit_pollution_issue_list, this, _1));

    tab_factory->end(xml);
}


// ----------------------------------------------------------------------
// top left frame - bandwidth issue list

void
html_topleft_t::emit_bandwidth_issue_list(xml_file_t &xml)
{
    assert(html_context.loop_registry);

    if (html_context.result_unreliable())
	T_DIV A_CLASS("warning") T_P {
	    T_B TEXT("Warning!");
	    TEXT("These results are based on too few samples to be reliable.");
	}

    html_issue_list_t issues;
    BOOST_FOREACH(const loop_info_map_t::value_type &v,
		  html_context.loop_registry->loop_infos) {
        const loop_info_ptr_t &loop = v.second;
	BOOST_FOREACH(rpt_depgroup_id_t depgroup_id, loop->depgroup_ids) {
	    depgroup_info_ptr_t &d =
		html_context.loop_registry->depgroup_infos[depgroup_id];
	    BOOST_FOREACH(html_issue_ptr_t &i, d->issues)
		if (i->summary_type().test(SUMMARY_BANDWIDTH))
		    issues.push_back(i);
	}
    }

    if (issues.empty()) {
       T_P TEXT("No bandwidth issues found.");
       T_P {
	   emit_product_name_full(xml);
	   TEXT(" does not show issues causing "
		"less than 1% of the total fetches, upgrades or write-backs "
		"by default. Lowering this threshold may reveal additional "
		"issues. See ");
	   T_A A_TARGET("_blank") A_HREF(html_context.link_factory->get_help_link(HELP_THRESHOLD_GUI))
	       TEXT("the \"Ignore issues\" setting");
	   TEXT(" if you created the report using the graphical interface, "
		"or ");
           T_A A_TARGET("_blank") A_HREF(html_context.link_factory->get_help_link(HELP_THRESHOLD_CLI))
	       TEXT("the -p option");
	   TEXT(" if you created the report from the command line.");
       }
    }
    else
	T_TABLE A_CLASS("summary table-autosort:3 "
			"table-stripeclass:alternate "
			"table-autofilter") {
	    html_issue_t::insert_summary_header(html_context, xml,
						SUMMARY_BANDWIDTH);
	    T_TBODY {
		summary_context_t summary_context;
		BOOST_FOREACH(html_issue_ptr_t i, issues)
		    i->emit_summary(xml,
				    summary_context, SUMMARY_BANDWIDTH);
	    }
	}

    emit_copyright(xml, true);
    emit_terms_of_use(html_context, xml);
}


// ----------------------------------------------------------------------
// top left frame - latency issue list

void
html_topleft_t::emit_latency_issue_list(xml_file_t &xml)
{
    assert(html_context.loop_registry);

    if (html_context.result_unreliable())
	T_DIV A_CLASS("warning") T_P {
	    T_B TEXT("Warning!");
	    TEXT("These results are based on too few samples to be reliable.");
	}

    html_issue_list_t issues;
    BOOST_FOREACH(const loop_info_map_t::value_type &v,
		  html_context.loop_registry->loop_infos) {
        const loop_info_ptr_t &loop = v.second;
	BOOST_FOREACH(rpt_depgroup_id_t depgroup_id, loop->depgroup_ids) {
	    depgroup_info_ptr_t &d =
		html_context.loop_registry->depgroup_infos[depgroup_id];
	    BOOST_FOREACH(html_issue_ptr_t &i, d->issues)
		if (i->summary_type().test(SUMMARY_LATENCY))
		    issues.push_back(i);
	}
    }

    if (issues.empty()) {
	T_P TEXT("No latency issues found.");
	T_P {
	    emit_product_name_full(xml);
	    TEXT(" does not show issues causing "
		 "less than 1% of the total fetches, upgrades or write-backs "
		 "by default. Lowering this threshold may reveal additional "
		 "issues. See ");
	    T_A A_TARGET("_blank") A_HREF(html_context.link_factory->get_help_link(HELP_THRESHOLD_GUI))
		TEXT("the \"Ignore issues\" setting");
	    TEXT(" if you created the report using the graphical interface, "
		 "or ");
            T_A A_TARGET("_blank") A_HREF(html_context.link_factory->get_help_link(HELP_THRESHOLD_CLI))
		TEXT("the -p option");
	    TEXT(" if you created the report from the command line.");
	}
    }
    else
	T_TABLE A_CLASS("summary table-autosort:3 "
			"table-stripeclass:alternate "
			"table-autofilter") {
	    html_issue_t::insert_summary_header(html_context, xml,
						SUMMARY_LATENCY);
	    T_TBODY {
		summary_context_t summary_context;
		BOOST_FOREACH(html_issue_ptr_t i, issues)
		    i->emit_summary(xml,
				    summary_context, SUMMARY_LATENCY);
	    }
	}

    emit_copyright(xml, true);
    emit_terms_of_use(html_context, xml);
}


// ----------------------------------------------------------------------
// top left frame - mt issues

void
html_topleft_t::emit_mt_issue_list(xml_file_t &xml)
{
    assert(html_context.loop_registry);

    if (html_context.result_unreliable())
	T_DIV A_CLASS("warning") T_P {
	    T_B TEXT("Warning!");
	    TEXT("These results are based on too few samples to be reliable.");
	}

    html_issue_list_t issues;
    BOOST_FOREACH(const loop_info_map_t::value_type &v,
		  html_context.loop_registry->loop_infos) {
        const loop_info_ptr_t &loop = v.second;
	BOOST_FOREACH(rpt_depgroup_id_t depgroup_id, loop->depgroup_ids) {
	    depgroup_info_ptr_t &d =
		html_context.loop_registry->depgroup_infos[depgroup_id];
	    BOOST_FOREACH(html_issue_ptr_t &i, d->issues)
		if (i->summary_type().test(SUMMARY_MT))
		    issues.push_back(i);
	}
    }

    if (issues.empty()) {
	T_P TEXT("No multi-threading issues found.");
	T_P {
	    emit_product_name_full(xml);
	    TEXT(" does not show issues causing "
		 "less than 1% of the total fetches, upgrades or write-backs "
		 "by default. Lowering this threshold may reveal additional "
		 "issues. See ");
	    T_A A_TARGET("_blank") A_HREF(html_context.link_factory->get_help_link(HELP_THRESHOLD_GUI))
		TEXT("the \"Ignore issues\" setting");
	    TEXT(" if you created the report using the graphical interface, "
		 "or ");
            T_A A_TARGET("_blank") A_HREF(html_context.link_factory->get_help_link(HELP_THRESHOLD_CLI))
		TEXT("the -p option");
	    TEXT(" if you created the report from the command line.");
	}
    }
    else
	T_TABLE A_CLASS("summary table-autosort:3 "
			"table-stripeclass:alternate "
			"table-autofilter") {
	    html_issue_t::insert_summary_header(html_context, xml, SUMMARY_MT);
	    T_TBODY {
		summary_context_t summary_context;
		BOOST_FOREACH(html_issue_ptr_t i, issues)
		    i->emit_summary(xml,
				    summary_context, SUMMARY_MT);
	    }
	}

    emit_copyright(xml, true);
    emit_terms_of_use(html_context, xml);
}


// ----------------------------------------------------------------------
// top left frame - pollution issue list

void
html_topleft_t::emit_pollution_issue_list(xml_file_t &xml)
{
    assert(html_context.loop_registry);

    if (html_context.result_unreliable())
	T_DIV A_CLASS("warning") T_P {
	    T_B TEXT("Warning!");
	    TEXT("These results are based on too few samples to be reliable.");
	}

    html_issue_list_t issues;
    BOOST_FOREACH(const loop_info_map_t::value_type &v,
		  html_context.loop_registry->loop_infos) {
        const loop_info_ptr_t &loop = v.second;
	BOOST_FOREACH(rpt_depgroup_id_t depgroup_id, loop->depgroup_ids) {
	    depgroup_info_ptr_t &d =
		html_context.loop_registry->depgroup_infos[depgroup_id];
	    BOOST_FOREACH(html_issue_ptr_t &i, d->issues)
		if (i->summary_type().test(SUMMARY_NT))
		    issues.push_back(i);
	}
    }

    if (issues.empty()) {
	T_P TEXT("No pollution issues found.");
	T_P {
	    emit_product_name_full(xml);
	    TEXT(" does not show issues causing "
		 "less than 1% of the total fetches, upgrades or write-backs "
		 "by default. Lowering this threshold may reveal additional "
		 "issues. See ");
	    T_A A_TARGET("_blank") A_HREF(html_context.link_factory->get_help_link(HELP_THRESHOLD_GUI))
		TEXT("the \"Ignore issues\" setting");
	    TEXT(" if you created the report using the graphical interface, "
		 "or ");
            T_A A_TARGET("_blank") A_HREF(html_context.link_factory->get_help_link(HELP_THRESHOLD_CLI))
		TEXT("the -p option");
	    TEXT(" if you created the report from the command line.");
	}
    }
    else
	T_TABLE A_CLASS("summary table-autosort:3 "
			"table-stripeclass:alternate "
			"table-autofilter") {
	    html_issue_t::insert_summary_header(html_context, xml,
						SUMMARY_NT);
	    T_TBODY {
		summary_context_t summary_context;
		BOOST_FOREACH(html_issue_ptr_t i, issues)
		    i->emit_summary(xml,
				    summary_context, SUMMARY_NT);
	    }
	}

    emit_copyright(xml, true);
    emit_terms_of_use(html_context, xml);
}


// ----------------------------------------------------------------------
// top left frame - loop list

#define COLUMN(title, sort)				\
    T_TH A_CLASS(string("table-sortable:") + sort)	\
	TEXT(title)

void
html_topleft_t::emit_loop_table(xml_file_t &xml)
{
    assert(html_context.loop_registry);
    const rpt_report_t &report(*html_context.report);

    if (html_context.result_unreliable())
	T_DIV A_CLASS("warning") T_P {
	    T_B TEXT("Warning!");
	    TEXT("These results are based on too few samples to be reliable.");
	}

    T_TABLE A_CLASS("summary table-autosort:2 table-stripeclass:alternate") {
	T_THEAD {
	    T_TR {
		COLUMN("Loop", "numeric");
		COLUMN("% of misses", "scientific");
		COLUMN("% of fetches", "scientific");
		if (report.mt_sampled) {
		    COLUMN("Fetch utilization", "numeric");
		    COLUMN("Write-back utilization", "numeric");
		} else
		    COLUMN("Utilization", "numeric");
		COLUMN("Issues", "default");
	    }
	}

	loop_summary_context_t summary_context;
	T_TBODY BOOST_FOREACH(const loop_info_map_t::value_type &v,
                              html_context.loop_registry->loop_infos) {
            const loop_info_ptr_t &loop = v.second;
	    html_loop_t html_loop(html_context, loop);

            // Loop 0 is the dummy loop for instructions without a real loop.
            if (loop->loop_id == 0)
                continue;

	    html_loop.emit_summary(xml, summary_context);
	}
    }

    emit_copyright(xml, true);
    emit_terms_of_use(html_context, xml);
}


// ----------------------------------------------------------------------
// top left frame - summary
void
html_topleft_t::emit_summary(xml_file_t &xml) {
    html_summary_t::emit(xml, html_context);
}


// ----------------------------------------------------------------------
// top left frame - file list

void
html_topleft_t::emit_file_list(xml_file_t &xml)
{
    html_files_t::emit(xml, html_context);
}


// ----------------------------------------------------------------------
// top left frame - execution details

void
html_topleft_t::emit_execution_statistics(xml_file_t &xml)
{
    html_execution_t::emit(xml, html_context);
}


// ----------------------------------------------------------------------
// top left frame - about

void
html_topleft_t::emit_about(xml_file_t &xml)
{
    html_about_t::emit(xml, html_context);
}


END_NAMESPACE;
END_NAMESPACE;
