/// @file tools/acurate/html_advice_consumer.cc

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#ifdef _WIN32
// Disable warnings about "default alignment has changed by including
// a file".
//
// This is intentional and part of boost's method to manage alignment.
// It uses special include files that set and reset the alignment
// respectively.
#pragma warning(disable: 4103)
#endif

#include "html_advice_consumer.hh"
#include "report_runner.hh"

#include "consumer.hh"
#include "file_repos.hh"
#include "link_factory_static.hh"
#include "xml_file.hh"

#include "html/topleft.hh"
#include "html/help_links.hh"
#include "html/front.hh"
#include "html/issues.hh"
#include "html/loop.hh"
#include "html/source.hh"
#include "html/file_png.hh"

#include <sstream>
#include <fstream>
#include <iomanip>
#include <set>
#include <string>
#include <limits.h>
#include <boost/filesystem.hpp>

// html/macros.hh defines some fairly generic symbols such as
// 'T_B'. Macros tend to interefere with other libraries. Put this
// include as late as possible.
#include "html/macros.hh"

#include "product.h"

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace boost::filesystem;
using namespace acumem::analysis;
using namespace boost;
using namespace std;

// This file uses HTML composition macros from html/macros.hh. The tag
// macros are designed to introduce a scope, and therefore they can be
// followed by a single statement, or a several statements in curly
// braces:
//
// P TEXT("foo");
// ==>  <p>foo</p>
//
//
// P { IMG A_SRC("bar.gif") TEXT("image"); TEXT("foo"); }
// ==>  <p><img src="bar.gif">image</img>foo</p>
//
//
// P;
// ==>  <p/>


// ----------------------------------------------------------------------
// ctor, dtor
// ----------------------------------------------------------------------

html_advice_consumer_t::html_advice_consumer_t(
    html_context_t &html_context)
    : html_context(html_context)
{
    // Locate the installation directory.
    const path &installation_dir(html_context.conf.installation_dir);

    // Create the report directory.
    path report_dir = html_context.conf.report_path;
    path manual_dir = report_dir / "manual_html";
    path manual_images_dir = manual_dir / "images";

    if (exists(report_dir) && !is_directory(report_dir)) {
	throw feature_exception_t(
             "acurate",
            "Can't create directory " + html_context.conf.report_path +
            ". It exists but is not a directory.");
    }

    try {
	if (!exists(report_dir))
	    create_directories(report_dir);
	if (!exists(manual_dir))
	    create_directories(manual_dir);
	if (!exists(manual_images_dir))
	    create_directories(manual_images_dir);
    } catch (const boost::filesystem::filesystem_error &e) {
	throw feature_errno_t("acurate",
			      "Creating output directory " + html_context.conf.report_path,
			      e.code().value());
    }

    typedef pair<path, path> path_pair_t;
    vector<path_pair_t> paths;
    paths.push_back(
	make_pair(installation_dir / "share" / PACKAGE / "report", report_dir));

    paths.push_back(
	make_pair(installation_dir / "doc" / "manual_html", manual_dir));

    paths.push_back(
	make_pair(installation_dir / "doc" / "manual_html" / "images",
		  manual_images_dir));

    try {
	BOOST_FOREACH(path_pair_t p, paths) {
	    directory_iterator i(p.first), e;
	    for (; i != e; i++) {
		path from = i->path();
		path to = p.second / from.filename();
		if (exists(from) && is_regular_file(from)) {
		    if (exists(to))
			remove(to);
		    copy_file(from, to);
		}
	    }
	}
    } catch (const boost::filesystem::filesystem_error &error) {
	// This typically happens if the installation is incomplete
	stringstream ss;
	ss << "Can't copy static files to report "
	    "directory. Error: " << error.what();
	html_context.warning_consumer->warning(ss.str());
    }
}

// ----------------------------------------------------------------------

void
html_advice_consumer_t::write_report()
{
    // write the intro file.
    write_intro_file();

    // write frameset file
    write_frameset_file(html_context.conf.intro_page);

    // Write files with issues.
    write_issue_files();

    // Write the summary, loop and issue lists. Needs to be done after
    // the issue files have been written, since we output a list of
    // files which is constructed while outputting issues.
    write_left_frame();

    // Write all the loop details.
    write_loop_files();

    // Decide which drill-down diagrams to create
    collect_drill_down_info();

    // write files with annotated source code
    write_annotated_source_files();

    // Output the results for all drill down requests.
    BOOST_FOREACH(drill_down_ptr_t &dd, drill_down_list)
        dd->write_files();
}

// ----------------------------------------------------------------------
// drill down
// ----------------------------------------------------------------------

void
html_advice_consumer_t::write_issue_files()
{
    BOOST_FOREACH(const loop_info_map_t::value_type &v,
		  html_context.loop_registry->loop_infos) {
        const loop_info_ptr_t &loop = v.second;
	BOOST_FOREACH(rpt_depgroup_id_t depgroup_id, loop->depgroup_ids) {
	    depgroup_info_ptr_t &d =
		html_context.loop_registry->depgroup_infos[depgroup_id];
	    BOOST_FOREACH(html_issue_ptr_t &i, d->issues) {
                html_context.feedback->log(
                    feedback_base_t::LOG_INFO,
                    "Issue " + i->get_display_number() +
                    ": " + i->get_name());
		ofstream issue_file(i->get_issue_file_name().c_str());
		xml_file_t xml(issue_file, Doctype_Strict);
		HTML_XMLNS {
		    i->emit_head(xml);
		    T_BODY A_ONLOAD("unfold()")
			i->emit_issue_file(xml);
		}
	    }
	}
    }
}

void
html_advice_consumer_t::write_loop_files()
{
    BOOST_FOREACH(const loop_info_map_t::value_type &v,
		  html_context.loop_registry->loop_infos) {
	const loop_info_ptr_t &loop = v.second;
	html_loop_t html_loop(html_context, loop);

	// Loop 0 is the dummy loop for instructions without a real loop.
	if (loop->loop_id == 0)
	    continue;

	string loop_s = lexical_cast<string>(loop->loop_id);
	string file_name = html_context.conf.report_path + "/loop" +
	    loop_s + ".html";
	ofstream f(file_name.c_str());
	xml_file_t xml_loop(f, Doctype_Strict);
	XML_OBJECT(xml_loop) HTML_XMLNS {
	    emit_head(xml, "Loop " + loop_s);
	    T_BODY A_ONLOAD("unfold()")
		html_loop.emit_loop_file(xml_loop);
	}
    }
}

void
html_advice_consumer_t::write_annotated_source_files()
{
    // Dump all referenced files
    BOOST_FOREACH(const file_map_t::value_type &v,
                  html_context.file_repos->file_map) {
	const string &filename = v.first;
	const line_info_map_t &line_info_map = v.second;
	int file_id = html_context.report->get_source_file_id(filename);

	string html_file_name =
	    html_context.conf.report_path + "/source_" +
	    lexical_cast<string>(file_id) + ".html";
	ofstream f(html_file_name.c_str());
	xml_file_t xml(f, Doctype_Strict);

	HTML_XMLNS {
	    T_HEAD {
		T_META;
		T_TITLE TEXT(filename);
		STYLE_FILE("source.css");
		STYLE_FILE("hover_popup.css");
		SCRIPT_FILE("nav.js");

		NL SCRIPT_INLINE TEXT_T(
		    "function load() {\n"
		    "  var f = parent.document.getElementsByName('src')[0];\n"
		    "  if (f != null) f.loadComplete = true;\n"
		    "}\n"
		    "function unload() {\n"
		    "  parent.document.loaded_src = null;\n"
		    "}\n"
		    "function toggle(e) {\n"
		    "  elem = document.getElementById(e);\n"
		    "  if (elem == null) return;\n"
		    "  if (elem.style.display == \"none\")\n"
		    "  { elem.style.display=\"\"; }\n"
		    "  else\n"
		    "  { elem.style.display=\"none\"; }\n"
		    "}\n"
		    );
	    }
	    T_BODY A_ONLOAD("load()") A_ONUNLOAD("unload()")
		html_source_t::write_annotated_file(xml, html_context,
						    filename, line_info_map);
	}
    }
}

void
html_advice_consumer_t::collect_drill_down_info()
{
    const rpt_report_t &report(*html_context.report);

    // Create a global drill down query
    drill_down_list.push_back(
	drill_down_ptr_t(
	    new drill_down_global_t(
                html_context.report,
		html_context.conf.report_path + "/perf",
		html_context.conf.max_mr_scale)));

    // Add drill-down entries for all interesting loops and dependency groups.
    BOOST_FOREACH(const loop_info_map_t::value_type & v,
		  html_context.loop_registry->loop_infos) {
        const loop_info_ptr_t &loop = v.second;
	if (html_context.interesting_loops.find(loop->loop_id)
	    != html_context.interesting_loops.end()) {
	    string l = html_context.conf.report_path +
		"/l" + lexical_cast<string>(loop->loop_id);

	    location_set_t pc_set =
                html_context.report->get_loop_excl_location_set(loop->loop_id);

	    drill_down_list.push_back(
		drill_down_ptr_t(
		    new drill_down_pc_set_t(
                        html_context.report,
                        l,
                        report.get_location_set_id(pc_set))));
	}
    }

    BOOST_FOREACH(rpt_depgroup_id_t depgroup_id,
		  html_context.interesting_depgroups) {
	string l = html_context.conf.report_path + "/d" +
	    lexical_cast<string>(depgroup_id);

	drill_down_list.push_back(
	    drill_down_ptr_t(
		new drill_down_pc_set_t(
                    html_context.report,
                    l,
                    report.get_depgroup_location_set_id(depgroup_id))));
    }
}


// ----------------------------------------------------------------------
// frameset files
// ----------------------------------------------------------------------

void
html_advice_consumer_t::write_frameset_file(bool has_separate_intro_file)
{
    string path = html_context.conf.report_path + "/main.html";

    ofstream frameset(path.c_str());
    xml_file_t frameset_xml(frameset, Doctype_Frameset);
    XML_OBJECT(frameset_xml) HTML_XMLNS {
	T_HEAD {
	    T_META;
	    T_TITLE TEXT(html_context.report_title());
	    SCRIPT_FILE("nav.js");
	}
	T_FRAMESET A_COLS("*,*") {
	    T_FRAMESET A_ROWS("*,*,10%") {
		T_FRAME A_NAME("sum")
		    A_SRC("summary.html");
		T_FRAME A_NAME("info")
		    A_SRC("default-info.html");
		T_FRAME A_NAME("stat")
		    A_SRC("default-info.html");
	    }
	    T_FRAME A_NAME("src")
		A_SRC("default-src.html");
	}
    }

    ofstream placeholder1((html_context.conf.report_path +
			   "/default-info.html").c_str());
    xml_file_t pl1_xml(placeholder1, Doctype_Strict);
    XML_OBJECT(pl1_xml) HTML_XMLNS {
	emit_head(pl1_xml);
	T_BODY {
	    T_P ATTR("class", "center")
		TEXT("Select an issue or a loop in the issue or loop tables.");
	}
    }

    ofstream placeholder2((html_context.conf.report_path +
			   "/default-src.html").c_str());
    xml_file_t pl2_xml(placeholder2, Doctype_Strict);
    XML_OBJECT(pl2_xml) HTML_XMLNS {
	emit_head(pl2_xml);
	T_BODY {
	    T_P A_CLASS("center")
		TEXT("Select a file in the file table, or follow a "
		     "source code link from an issue or a loop description.");
	}
    }
}

// ----------------------------------------------------------------------
// intro page
// ----------------------------------------------------------------------
void
html_advice_consumer_t::write_intro_file()
{
    ofstream redirect((html_context.conf.report_dir + ".html").c_str());
    xml_file_t redirect_html(redirect, Doctype_Strict);
    string next =
	(html_context.conf.intro_page ? "/front.html" : "/main.html");
    XML_OBJECT(redirect_html) {
	HTML_XMLNS {
	    T_HEAD TAG(meta)
		ATTR("http-equiv", "refresh")
		ATTR("content", "0; URL=" +
		     html_context.conf.report_dir + next);
	    T_BODY {}
	}
    }


    ofstream intro((html_context.conf.report_dir + "/front.html").c_str());
    xml_file_t intro_html(intro, Doctype_Strict);
    XML_OBJECT(intro_html) {
	HTML_XMLNS {
            if (html_context.conf.intro_page) {
                T_HEAD {
                    T_META;
                    // XXX Empty TITLE; breaks IE
                    //TITLE;
                    SCRIPT_FILE("tabber.js");
                    SCRIPT_FILE("nav.js");

                    // Temporarily hide the "tabber" class so it does not
                    // "flash" on the page as plain HTML. After tabber runs,
                    // the class is changed to "tabberlive" and it will appear.
                    TEXT_T("<style type=\"text/css\">."
                           "tabber{display:none;}</style>");

                    STYLE_FILE("summary.css");
                    SCRIPT_FILE("table.js");

                    T_TITLE TEXT(html_context.report_title());
                }
                T_BODY {
		    application_verdict_t app_verdict(html_context);

                    html_front_t front(html_context, app_verdict);
                    front.emit(xml);

		    string report_path =
			html_context.conf.report_dir + "/";

		    {
			string image_path = report_path + PNG_BANDWIDTH_GAUGE;
			ofstream f(image_path.c_str(), ios::binary);
			app_verdict.render_fr_gauge(f);
		    }

		    {
			string image_path = report_path + PNG_LATENCY_GAUGE;
			ofstream f(image_path.c_str(), ios::binary);
			app_verdict.render_mr_gauge(f);
		    }

		    {
			string image_path = report_path + PNG_LOCALITY_GAUGE;
			ofstream f(image_path.c_str(), ios::binary);
			app_verdict.render_utilization_gauge(f);
		    }

		    {
			string image_path = report_path + PNG_THREADING_GAUGE;
			ofstream f(image_path.c_str(), ios::binary);
			app_verdict.render_mt_gauge(f);
		    }
                }
            }
	}
    }
}

// ----------------------------------------------------------------------
// top left frame
// ----------------------------------------------------------------------

void
html_advice_consumer_t::write_left_frame()
{
    ofstream left((html_context.conf.report_path + "/summary.html").c_str());
    xml_file_t left_html(left, Doctype_Strict);
    XML_OBJECT(left_html) {
	HTML_XMLNS {
	    T_HEAD {
		T_META;
		// XXX Empty TITLE; breaks IE
		//T_TITLE;
		SCRIPT_FILE("tabber.js");
		SCRIPT_FILE("nav.js");

		// Temporarily hide the "tabber" class so it does not
		// "flash" on the page as plain HTML. After tabber
		// runs, the class is changed to "tabberlive" and it
		// will appear.
		TEXT_T("<style type=\"text/css\">."
		       "tabber{display:none;}</style>");

		STYLE_FILE("summary.css");
		STYLE_FILE("hover_popup.css");
		SCRIPT_FILE("table.js");
	    }
	    T_BODY {
		html_topleft_t topleft(html_context);
		topleft.emit(xml);
	    }
	}
    }
}

const bool static_report(true);

void
run_report(html_context_t &html_context, report_url_t *ru)
{
    html_context.link_factory.reset(new link_factory_static_t);
    html_advice_consumer_t hac(html_context);
    hac.write_report();
}

END_NAMESPACE;
END_NAMESPACE;
