/// @file tools/acurate/html_context.hh

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

#ifndef INCLUDED_HTML_CONTEXT
#define INCLUDED_HTML_CONTEXT

#include "tools/acurate/conf.hh"
#include "tools/acurate/file_repos.hh"
#include "tools/acurate/consumer.hh"
#include "analysis/report/report.hh"
#include "library/feedback/feedback.hh"
#include "library/license/license.hh"
#include "library/warning/warning_consumer.hh"

#include <boost/shared_ptr.hpp>

#include <vector>
#include <set>
#include <string>
#include <iomanip>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace acumem::library;
using namespace acumem::license;

// Break circular dependency on loop_info.hh.
class loop_registry_t;
typedef boost::shared_ptr<loop_registry_t> loop_registry_ptr_t;

class link_factory_t;
typedef boost::shared_ptr<link_factory_t> link_factory_ptr_t;

/// Hold all data related to a report run. It is passed around to
/// those components that need some data, i.e., pretty much everyone.
class html_context_t {
public:
    html_context_t(const warning_consumer_ptr_t warning_consumer,
                   const feedback_ptr_t feedback,
                   const license_manager_ptr_t license,
                   const file_repos_cptr_t file_repos,
                   const loop_registry_ptr_t loop_registry,
                   const rpt_report_cptr_t report,
                   const conf_t &conf,
		   const link_factory_ptr_t link_factory,
                   const bool feature_mt_advice,
                   const bool feature_sw_prefetch,
                   const bool feature_unlimited_callstack,
                   const bool feature_blocking_expert,
                   const bool feature_multisocket)
	: license(license),
          warning_consumer(warning_consumer),
          feedback(feedback),
          file_repos(file_repos),
          loop_registry(loop_registry),
          report(report),
          conf(conf),
	  link_factory(link_factory),
          feature_mt_advice(feature_mt_advice),
          feature_sw_prefetch(feature_sw_prefetch),
          feature_unlimited_callstack(feature_unlimited_callstack),
          feature_blocking_expert(feature_blocking_expert),
          feature_multisocket(feature_multisocket)
	{}

    bool result_unreliable() const {
	return report->sample_count < MIN_RELIABLE_SAMPLE_COUNT;
    }

    bool warn_virtualized() const {
	return report->autodetect_failed;
    }

    /// Returns the report title.
    string report_title() const;

    static string value_to_string(double x) {
	stringstream ss;
	ss << fixed << setprecision(0);
	if (x < 1024.0)
	    ss << x;
	else if (x < 1024.0*1024.0)
	    ss << x/1024.0 << "k";
	else if (x < 1024.0*1024.0*1024.0)
	    ss << x/1024.0/1024.0 << "M";
	else
	    ss << x/1024.0/1024.0/1024.0 << "G";

	return ss.str();
    }

    /// License.
    const license_manager_ptr_t license;

    /// Warning consumer.
    const warning_consumer_ptr_t warning_consumer;

    /// Feedback object.
    const feedback_ptr_t feedback;

    const file_repos_cptr_t file_repos;

    const loop_registry_ptr_t loop_registry;

    const rpt_report_cptr_t report;

    const conf_t &conf;

    /// An abstract factory class for creating different kinds of HTML
    /// contents for various features, depending on the type of output
    link_factory_ptr_t link_factory;

    /// TODO: These shouldn't be needed, all license checks should be done
    ///       before generating the intermediate format report!
    ///
    /// Licensed features.
    bool feature_mt_advice;
    bool feature_sw_prefetch;
    bool feature_unlimited_callstack;
    bool feature_blocking_expert;
    bool feature_multisocket;

    std::set<rpt_depgroup_id_t> interesting_depgroups;
    std::set<rpt_loop_id_t> interesting_loops;
};

END_NAMESPACE;
END_NAMESPACE;

#endif
