/// @file tools/acurate/report_runner.cc

// Copyright (c) 2009-2010 Acumem AB.  All rights reserved.

#include "report_runner.hh"
#include "file_repos.hh"
#include "html_advice_consumer.hh"
#include "link_factory_static.hh"
#include "server/session.hh"
#include "server/webserver.hh"
#include "server/link_factory_ajax.hh"
#include "server/shutdownmanager.hh"

#include "analysis/report_codec/report_codec.hh"
#include "library/stream/stream_file.hh"
#include "library/stream/stream_container.hh"
#include "library/stream/header_collector.hh"
#include "library/lang/am_assert.hh"
#include "product.h"

#include <boost/shared_array.hpp>

#ifdef _WIN32
#include <windows.h>
#include <tchar.h>
#endif
#include <string>
#include <fstream>
#include <sstream>

#ifndef _WIN32
#include <unistd.h>
#define Sleep(x) usleep((x)*1000)
#endif

#include <iostream>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

using namespace std;

// ----------------------------------------------------------------------

bool
report_runner_t::check_container(const string &path) const
{
    stream_container_header_t header(
        header_collector_t(path, context->warning_consumer).header);

    // Check that the stream contains a report.
    switch (header.data_type) {
    case STREAM_CONTAINER_DATA_SAMPLES_BINARY: {
	stringstream ss;
	ss << "The specified input file " << path
           << " is a sample file, not a report file.";
	context->warning_consumer->error(ss.str());
	return false;
    }
    case STREAM_CONTAINER_DATA_REPORT:
        break;
    default: {
	stringstream ss;
	ss << "The specified input file " << path
           << " is not a report file.";
	context->warning_consumer->error(ss.str());
	return false;
    }
    }

    // Check that it is from a compatible product.
    if (!PRODUCT_STREAM_COMPATIBLE(header)) {
	stringstream ss;
	ss << "The selected sample file was produced by "
           << stream_container_product_name( header.product)
           << ", which is not compatible with " PRODUCT_NAME ".";
	context->warning_consumer->error(ss.str());
	return false;
    }

    return true;
}

bool
report_runner_t::parse_args(bool interactive,
                            bool feature_mt_advice,
                            bool feature_multisocket,
                            bool feature_sw_prefetch,
                            bool feature_unlimited_callstack,
                            bool feature_blocking_expert,
                            int argc, char **argv)
{
    assert(!arg_conf);

    arg_conf.reset(new conf_t(context->warning_consumer, context->feedback));
    conf_t &conf = *arg_conf.get();

    conf.feature_mt_advice = feature_mt_advice;
    conf.feature_multisocket = feature_multisocket;
    conf.feature_sw_prefetch = feature_sw_prefetch;
    conf.feature_unlimited_callstack = feature_unlimited_callstack;
    conf.feature_blocking_expert = feature_blocking_expert;

    if (!conf.init(interactive, argc, argv))
	return false;

    // Check that the stream container is compatible with this
    // product before calling the parser.
    if (!check_container(conf.report_file))
	return false;

    return true;
}


bool
report_runner_t::parse_args(bool interactive,
                            bool feature_mt_advice,
                            bool feature_multisocket,
                            bool feature_sw_prefetch,
                            bool feature_unlimited_callstack,
                            bool feature_blocking_expert,
                            const std::vector<std::string> &args)
{
    boost::shared_array<char *> argv(new char * [args.size()]);
    for (unsigned int i = 0; i < args.size(); i++) {
	argv[i] = const_cast<char *>(args[i].c_str());
    }

    return parse_args(interactive,
                      feature_mt_advice,
                      feature_multisocket,
                      feature_sw_prefetch,
                      feature_unlimited_callstack,
                      feature_blocking_expert,
                      (int)args.size(),
                      argv.get());
}

static bool
has_issue_type(const rpt_report_t &report,
               rpt_issue_type_t issue_type,
               html_issue_ptr_t issue)
{
    const rpt_issue_t &rpt_issue(*report.issues.at(issue->get_issue_id()));
    return rpt_issue.get_issue_type() == issue_type;
}

static void
hide_issues(const rpt_report_t &report,
            html_issue_list_t &issues,
	    rpt_issue_type_t superior_issue_type,
	    rpt_issue_type_t inferior_issue_type)
{
    BOOST_FOREACH(html_issue_ptr_t inferior, issues) {
        if (!has_issue_type(report, inferior_issue_type, inferior))
            continue;

        BOOST_FOREACH(html_issue_ptr_t superior, issues) {
            if (!has_issue_type(report, superior_issue_type, superior))
                continue;

            superior->add_inferior(inferior);
            issues.remove(inferior);

            hide_issues(
                report, issues, superior_issue_type, inferior_issue_type);
            return;
        }
    }
}

static void
hide_inferior_issues(html_context_t &html_context)
{
    // Hide less specific issue types.
    //
    // * hot-spot < poor utilization < loop order < random access
    // * comm-spot < comm utilization

    const rpt_report_t &report(*html_context.report);

    BOOST_FOREACH(depgroup_info_ptr_t &depgroup_info,
                  html_context.loop_registry->depgroup_infos) {
        hide_issues(report, depgroup_info->issues,
                    Rpt_Issue_Random, Rpt_Issue_Loop_Order);
        hide_issues(report, depgroup_info->issues,
                    Rpt_Issue_Random, Rpt_Issue_Utilization);
        hide_issues(report, depgroup_info->issues,
                    Rpt_Issue_Random, Rpt_Issue_Hot_Spot);


        hide_issues(report, depgroup_info->issues,
                    Rpt_Issue_Loop_Order, Rpt_Issue_Utilization);
        hide_issues(report, depgroup_info->issues,
                    Rpt_Issue_Loop_Order, Rpt_Issue_Hot_Spot);

        hide_issues(report, depgroup_info->issues,
                    Rpt_Issue_Utilization, Rpt_Issue_Hot_Spot);

        hide_issues(report, depgroup_info->issues,
                    Rpt_Issue_Comm_Utilization, Rpt_Issue_Comm_Spot);
    }
}

bool
report_runner_t::run(const license_manager_ptr_t license, report_url_t *ru)
{
    assert(arg_conf);
    conf_t &conf = *arg_conf.get();

    // Read the report from disk. Put these objects in a block so that the
    // report file is closed before we continue.
    rpt_report_cptr_t report;
    {
        check_container(conf.report_file);
        report_decoder_t in_decoder(context->warning_consumer);
        stream_container_reader_t in_container(
            &in_decoder,
            context->warning_consumer);
        stream_file_reader_t in_file(conf.report_file.c_str(), &in_container);
        in_file.dispatch();
        report = in_decoder.get_report();
    }

    file_repos_ptr_t file_repos(new file_repos_t(*report));

    loop_registry_ptr_t loop_registry(new loop_registry_t(*report));

    html_context_t html_context(
        conf.warning_consumer,
        conf.feedback,
        license,
        file_repos,
        loop_registry,
        report,
        conf,
        link_factory_ptr_t(),
        conf.feature_mt_advice,
        conf.feature_sw_prefetch,
        conf.feature_unlimited_callstack,
        conf.feature_blocking_expert,
        conf.feature_multisocket);

    // Add the issues.
    rpt_issue_id_t num_issues =
        (rpt_issue_id_t)html_context.report->issues.size();
    for (rpt_issue_id_t issue_id = 0; issue_id < num_issues; issue_id++)
        html_context.loop_registry->add_issue(html_context, issue_id);

    // Hide inferior issues.
    hide_inferior_issues(html_context);

    // Prepare source code annotation.
    BOOST_FOREACH(const loop_info_map_t::value_type &v,
		  html_context.loop_registry->loop_infos) {
        const loop_info_ptr_t &loop = v.second;
	BOOST_FOREACH(rpt_depgroup_id_t depgroup_id, loop->depgroup_ids) {
	    depgroup_info_ptr_t &d =
		html_context.loop_registry->depgroup_infos[depgroup_id];

	    BOOST_FOREACH(html_issue_ptr_t &i, d->issues)
		i->annotate_source();
	}
    }

    run_report(html_context, ru);

    return true;
}

report_runner_t::report_runner_t(workspace_ptr_t workspace,
                                 warning_consumer_ptr_t warning_consumer,
                                 feedback_ptr_t feedback)
    : context(new feature_context_t(workspace, warning_consumer, feedback))
{
}

report_runner_t::~report_runner_t()
{
    // XXX Kludge. See #1143. This is needed to break a shared_ptr cycle.
    // See issue #1143 and analysis/feature_base/context.hh for details.
    // Without this, we would have a leak.

    context->clear();
}

bool
report_runner_t::no_browser() const
{
    return arg_conf->no_browser;
}

END_NAMESPACE;
END_NAMESPACE;
