// adapted from boost

/* boost random_device.cpp implementation
 *
 * Copyright Jens Maurer 2000
 * Copyright Steven Watanabe 2010
 * Distributed under the Boost Software License, Version 1.0. (See
 * accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 *
 * $Id: random_device.cpp 60755 2010-03-22 00:45:06Z steven_watanabe $
 *
 */

#include "rng.hh"

#include <cstdio>
#include <string>
#include <stdexcept>  // std::invalid_argument

#ifdef _WIN32

#include <windows.h>
#include <wincrypt.h>

#pragma comment(lib, "Advapi32.lib")

#else

#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#endif

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

#ifdef _WIN32

class random_number_generator_t::impl {
public:
    impl()
	: provider(MS_DEF_PROV_A)
	{
	    char buffer[80];
	    DWORD type;
	    DWORD len;

	    // Find the type of the provider
	    for (DWORD i = 0; ; ++i) {
		len = sizeof(buffer);
		if (!CryptEnumProvidersA(i, NULL, 0, &type, buffer, &len))
		    error("Could not find provider name");
		if (provider == buffer)
		    break;
	    }

	    if (!CryptAcquireContextA(&hProv, NULL, provider.c_str(), type,
				     CRYPT_VERIFYCONTEXT | CRYPT_SILENT))
		error("Could not acquire CSP context");
	}

    ~impl()
	{
	    if (!CryptReleaseContext(hProv, 0))
		error("Could not release CSP context");
	}

    unsigned int next()
	{
	    unsigned int result;

	    if (!CryptGenRandom(
		    hProv, sizeof(result),
		    static_cast<BYTE*>(static_cast<void*>(&result))))
		error("error while reading");
	    return result;
	}

private:
  void error(const std::string & msg)
	{
	    DWORD error_code = GetLastError();
	    char buf[80];
	    DWORD num = FormatMessageA(
		FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		GetLastError(),
		0,
		buf,
		sizeof(buf),
		NULL);

	    throw std::invalid_argument("rng: " + msg +
					" Cryptopraphic Service Provider " + provider +
					": " + std::string(&buf[0], &buf[0] + num));
	}

    const std::string provider;
    HCRYPTPROV hProv;
};

#else
class random_number_generator_t::impl
{
public:
  impl()
      : path("/dev/urandom")
	{
	    fd = open(path.c_str(), O_RDONLY);
	    if (fd < 0)
		error("cannot open");
	}

    ~impl()
	{
	    if (close(fd) < 0)
		error("could not close");
	}

    unsigned int next()
	{
	    unsigned int result;
	    long sz = read(fd,
			   reinterpret_cast<char *>(&result),
			   sizeof(result));
	    if (sz == -1)
		error("error while reading");
	    else if (sz != sizeof(result)) {
		errno = 0;
		error("EOF while reading");
	    }
	    return result;
	}

private:
    void error(const std::string & msg)
	{
	    throw std::invalid_argument("rnd: " + msg +
					" random-number pseudo-device " + path +
					": " + strerror(errno));
	}

    const std::string path;
    int fd;
};

#endif


random_number_generator_t::random_number_generator_t()
  : pimpl(new impl)
{
}

random_number_generator_t::~random_number_generator_t()
{
    delete pimpl;
}

unsigned int random_number_generator_t::operator()()
{
    return pimpl->next();
}

END_NAMESPACE;
END_NAMESPACE;
