/// @file tools/acurate/server/launch.cc

// Copyright (c) 2011 Acumem AB.  All rights reserved.

#include "tools/acurate/server/webserver.hh"
#include "tools/acurate/server/session.hh"
#include "tools/acurate/server/shutdownmanager.hh"
#include "tools/acurate/report_runner.hh"

#include "library/lang/am_assert.hh"

#include "pion/PionException.hpp"

#if !defined(_WIN32)
#include <unistd.h>
#include <fcntl.h>
#include <wait.h>
#endif

#include <iostream>

using namespace std;
using namespace acumem::acurate::server;

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);

/// Launch a browser displaying the URL. Return true if successful, false if not.
static bool launch_browser(html_context_t &html_context, const string &url);

#if defined(_WIN32)

#include <shellapi.h>

static bool
launch_browser(html_context_t &html_context, const string &url)
{
    const char *env_browser = getenv("BROWSER");

    if (env_browser) {
	string cmd;
	if (env_browser[0] == '"')
	    cmd = env_browser;
	else
	    cmd = string("\"") + env_browser + "\"";
	cmd = cmd + string(" ") + url;

	PROCESS_INFORMATION processInformation;
	STARTUPINFOA startupInfo;
	memset(&processInformation, 0, sizeof(processInformation));
	memset(&startupInfo, 0, sizeof(startupInfo));
	startupInfo.cb = sizeof(startupInfo);

	BOOL result;

	result = CreateProcessA(NULL,
				const_cast<char *>(cmd.c_str()),
				NULL,
				NULL,
				FALSE,
				0,
				NULL,
				NULL,
				&startupInfo,
				&processInformation);

	if (!result) {
	    DWORD error = GetLastError();

	    stringstream ss;
	    ss << "Failed to launch web browser '" << env_browser
	       << "' specified by the BROWSER environment variable. Error: " << error;
	    html_context.warning_consumer->error(ss.str());

	    return false;
	}
	CloseHandle(processInformation.hProcess);
	CloseHandle(processInformation.hThread);

    } else {
	HINSTANCE hinstance = ShellExecuteA(NULL, "open", url.c_str(),
					    NULL, NULL, SW_SHOWNORMAL);
	if ((int) hinstance <= 32) {
	    stringstream ss;
	    ss << "Failed to launch default web browser";
	    html_context.warning_consumer->error(ss.str());
	    return false;
	}
    }
    return true;
}

#else

static const char *browser_commands[] = {
    "htmlview",                // Browser launch script on RedHat systems
    "sensible-browser",        // Browser launch script on Debian systems
    "gnome-www-browser",       // Default browser in Gnome on Debian systems
    "x-www-browser",           // Default browser in X on Debian systems
    "firefox",
    "mozilla",
    "mozilla-firefox",
    "/usr/sfw/bin/mozilla",    // Default location for mozilla on Solaris
    NULL
};

static bool
exec_browser(const string &browser,
             const string &url,
             string &error_message)
{
    int res;

    int pipe_fds[2];
    if (pipe(pipe_fds) == -1) {
        stringstream ss;
        ss << "Failed to create pipe: Error " << errno << ", "
           << strerror(errno) << ".";
        error_message =  ss.str();
        return false;
    }

    int fd_flags = fcntl(pipe_fds[1], F_GETFD);
    if (fd_flags == -1) {
        stringstream ss;
        ss << "Failed to get file descriptor flags: Error " << errno << ", "
           << strerror(errno) << ".";
        error_message =  ss.str();
        return false;
    }

    if (fcntl(pipe_fds[1], F_SETFD, fd_flags | FD_CLOEXEC) == -1) {
        stringstream ss;
        ss << "Failed to set file descriptor flags: Error " << errno << ", "
           << strerror(errno) << ".";
        error_message =  ss.str();
        return false;
    }

    pid_t pid = fork();
    AM_ASSERT(pid != -1);
    if (pid == 0) {
        res = close(pipe_fds[0]);
        AM_ASSERT(res == 0);

        // Create a new session id, so the web browser isn't killed if the user
        // presses ctrl-c to kill the web server.
        pid_t sid = setsid();
        AM_ASSERT(sid != -1);

        // Fork again, so that the main view process can do a waitpid() and we
        // avoid becoming a zombie process.
        pid_t pid2 = fork();
        assert(pid2 != -1);

        if (pid2 == 0) {
            const char * const argv[] = {browser.c_str(), url.c_str(), NULL};
            execvp(argv[0], (char * const *)argv);
            res = write(pipe_fds[1], &errno, sizeof(int));
            AM_ASSERT(res == sizeof(int));
            exit(1);
        }
        else
            exit(0);
    }
    else {
        res = close(pipe_fds[1]);
        AM_ASSERT(res == 0);

        int status;
        res = waitpid(pid, &status, 0);
        AM_ASSERT(res == pid && status == 0);

        int exec_errno;
        do {
            res = read(pipe_fds[0], &exec_errno, sizeof(int));
        } while (res == -1 && errno == EINTR);
        if (res != 0) {
            stringstream ss;
            ss << "Error " << exec_errno << ", "
               << strerror(exec_errno) << ".";
            error_message =  ss.str();
            return false;
        }
    }

    return true;
}

static bool
launch_browser(html_context_t &html_context, const string &url)
{
    const char *env_browser = getenv("BROWSER");

    if (env_browser) {
        string error;
        if (exec_browser(env_browser, url, error))
            return true;
        else {
            stringstream ss;
            ss << "Failed to launch web browser '" << env_browser
               << "' specified by the BROWSER environment variable.";
            html_context.warning_consumer->error(ss.str());
            html_context.warning_consumer->error(error);
        }
    }
    else {
        int i = 0;
        while (browser_commands[i] != NULL) {
            string error;
            if (exec_browser(browser_commands[i], url, error))
                return true;
            i++;
        }

        stringstream ss;
        ss << "Failed to launch all of the default web browsers. Please "
           << "specified by the browser to use with the BROWSER environment "
           << "variable.";
        html_context.warning_consumer->error(ss.str());
    }

    return false;
}

#endif

const bool static_report(false);

void
run_report(html_context_t &html_context, report_url_t *ru)
{
    // This is all very backwards. The fundamental challenge in
    // the current design is that the url_repository belongs in
    // the session, but the need to set up a common html_context
    // here is a bad match. It takes a link_factory which by the
    // current design takes an url_repository.  It would all be
    // easier if the url_repository could be populated before, and
    // would not need to be known by the link_factory_ajax.cc.

    shutdown_manager_t shutdown_manager;

    session_collection_t session_collection(html_context.license,
					    shutdown_manager);

    session_collection.allow_many_logins = html_context.conf.allow_many_logins;

    webserver_ptr_t webserver;

    url_repository_t url_repository;

    try {
        html_context.link_factory.reset(
            new link_factory_ajax_t(url_repository));

        // XXX "session" should come from the loaded report, being a
        // unique number for that report.
        session_ptr_t session_ptr(
            new session_t(html_context, url_repository, "session"));

        session_collection.session_vector.push_back(session_ptr);

        webserver = webserver_ptr_t(
            new webserver_t(html_context.conf.server_port,
                            session_collection,
                            html_context.conf.inactivity_timeout));

        webserver->add_session(session_collection.session_vector[0]);

        webserver->start();
    }
    catch (const pion::PionException &e) {
        throw std::runtime_error(
            string("Failed to start the built-in web server; ") + e.what());
    }

    string url = "http://localhost:" +
        lexical_cast<string>(session_collection.server_port) +
        "/login?pass=" + html_context.conf.cookie + "&url=/start";

    bool browser_ok = true;
    if (!html_context.conf.no_browser)
        browser_ok = launch_browser(html_context, url);

    // Tell whoever called us what the start url is.
    if (ru)
	ru->report_url(url);

    if (browser_ok)
        shutdown_manager.wait();

    // We have seen strange exceptions from inside
    // pion::net::TCPConnection::close() (10038). We have no idea what
    // they are, but since everything is going down anyway, we will
    // silently catch these problems.

    try {
	webserver->stop();
    } catch (boost::system::system_error &) {
	//cerr << "Error stopping report server: " << e.what() << endl;
	// Eat this error.

/*
 	KernelBase.dll!000007fefda3aa7d()
 	[Frames below may be incorrect and/or missing, no symbols loaded for KernelBase.dll]
 	msvcr90.dll!00000000738051c6()
>	view.exe!boost::asio::detail::throw_error(const boost::system::error_code & err={...})  Line 36 + 0x34 bytes	C++
 	view.exe!boost::asio::basic_socket<boost::asio::ip::tcp,boost::asio::stream_socket_service<boost::asio::ip::tcp> >::close()  Line 269	C++
 	view.exe!pion::net::TCPConnection::close()  Line 156	C++
 	view.exe!boost::_bi::list1<boost::arg<1> >::operator()<boost::_mfi::mf0<void,pion::net::TCPConnection>,boost::_bi::list1<boost::shared_ptr<pion::net::TCPConnection> & __ptr64> >(boost::_bi::type<void> __formal={...}, boost::_mfi::mf0<void,pion::net::TCPConnection> & f={...}, boost::_bi::list1<boost::shared_ptr<pion::net::TCPConnection> &> & a={...}, boost::_bi::type<void> __formal={...})  Line 254	C++
 	view.exe!boost::_bi::bind_t<void,boost::_mfi::mf0<void,pion::net::TCPConnection>,boost::_bi::list1<boost::arg<1> > >::operator()<boost::shared_ptr<pion::net::TCPConnection> >(boost::shared_ptr<pion::net::TCPConnection> & a1={...})  Line 33	C++
 	view.exe!std::for_each<std::_Tree<std::_Tset_traits<boost::shared_ptr<pion::net::TCPConnection>,std::less<boost::shared_ptr<pion::net::TCPConnection> >,std::allocator<boost::shared_ptr<pion::net::TCPConnection> >,0> >::iterator,boost::_bi::bind_t<void,boost::_mfi::mf0<void,pion::net::TCPConnection>,boost::_bi::list1<boost::arg<1> > > >(std::_Tree<std::_Tset_traits<boost::shared_ptr<pion::net::TCPConnection>,std::less<boost::shared_ptr<pion::net::TCPConnection> >,std::allocator<boost::shared_ptr<pion::net::TCPConnection> >,0> >::iterator _First={...}, std::_Tree<std::_Tset_traits<boost::shared_ptr<pion::net::TCPConnection>,std::less<boost::shared_ptr<pion::net::TCPConnection> >,std::allocator<boost::shared_ptr<pion::net::TCPConnection> >,0> >::iterator _Last={...}, boost::_bi::bind_t<void,boost::_mfi::mf0<void,pion::net::TCPConnection>,boost::_bi::list1<boost::arg<1> > > _Func={...})  Line 29 + 0x1a bytes	C++
 	view.exe!pion::net::TCPServer::stop(bool wait_until_finished=false)  Line 135	C++
 	view.exe!acumem::acurate::server::webserver_t::stop()  Line 758	C++
 	view.exe!acumem::acurate::run_report(acumem::acurate::html_context_t & html_context={...}, acumem::acurate::report_url_t * ru=0x0000000001f7cc30)  Line 208 + 0xd bytes	C++
 	view.exe!acumem::acurate::report_runner_t::run(const boost::shared_ptr<acumem::library::lic_license_t> * license=0x0000000001f44d70, acumem::acurate::report_url_t * ru=0x0000000000000000)  Line 263 + 0x15 bytes	C++
 	view.exe!main(int argc=3, char * * argv=0x0000007e157a2b73)  Line 130 + 0x66 bytes	C++
 	view.exe!__tmainCRTStartup()  Line 586 + 0x19 bytes	C
 	kernel32.dll!000000007770f56d()
 	ntdll.dll!0000000077842cc1()
*/


    }
}

END_NAMESPACE;
END_NAMESPACE;
