#include "webserver.hh"
#include "dyn_content.hh"
#include "file_service.hh"
#include "shutdownmanager.hh"

#include "../html/file_png.hh"
#include "../html/front.hh"
#include "../html/loop.hh"
#include "../html/macros.hh"
#include "../html/source.hh"
#include "../html/topleft.hh"

#include "../diagram.hh"
#include "../drill_down.hh"
#include "../html_context.hh"
#include "../link_factory.hh"
#include "../xml_file.hh"
#include "../report_runner.hh" // for report_url_t

#include "library/lang/am_assert.hh"

#include <pion/net/WebServer.hpp>
#include <pion/net/WebService.hpp>
#include <pion/net/HTTPAuth.hpp>
#include "auth.hh"

#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <boost/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/filesystem.hpp>

#include <iostream>

BEGIN_NAMESPACE(acumem);
BEGIN_NAMESPACE(acurate);
BEGIN_NAMESPACE(server);

using namespace std;
using namespace boost;
using namespace boost::filesystem;
using namespace boost::algorithm;
using namespace boost::system;
using namespace boost::asio;


// ----------------------------------------------------------------------

/// Local class, implementing a redirection from URL /start to either
/// the front page of a single loaded report, or to the table of
/// contents if there are several reports loaded.
class start_redirect_t : public pion::net::WebService {
public:
    start_redirect_t(const session_collection_t &session_collection)
	: session_collection(session_collection)
	{
	}

    void operator() (pion::net::HTTPRequestPtr &request,
		     pion::net::TCPConnectionPtr &tcp_conn);

private:

    void redirect(const std::string &redirection_url,
		  pion::net::HTTPRequestPtr &request,
		  pion::net::TCPConnectionPtr &tcp_conn);

    const session_collection_t &session_collection;
};


void
start_redirect_t::operator () (pion::net::HTTPRequestPtr &request,
			       pion::net::TCPConnectionPtr &tcp_conn)
{
    if (session_collection.session_vector.size() == 1) {
	// Exactly one report loaded - redirect to its front page or
	// main page depending on the configuration.
	session_ptr_t s = session_collection.session_vector.front();

	if (s->html_context.conf.intro_page)
	    redirect("/" + s->id + "/front.html", request, tcp_conn);
	else
	    redirect("/" + s->id + "/main.html", request, tcp_conn);
    }
    else
	// Zero or more than one report loaded - redirect to report table
	redirect("/", request, tcp_conn);
}

void
start_redirect_t::redirect(const std::string &redirection_url,
			   pion::net::HTTPRequestPtr &request,
			   pion::net::TCPConnectionPtr &tcp_conn)
{
    using namespace pion::net;
    static const std::string CONTENT =
	" <!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\""
	"\"http://www.w3.org/TR/1999/REC-html401-19991224/loose.dtd\">"
	"<HTML>"
	"<HEAD>"
	"<TITLE>Redirect</TITLE>"
	"<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; "
	"charset=ISO-8859-1\">"
	"</HEAD>"
	"<BODY><H1>302 Found.</H1></BODY>"
	"</HTML> ";
    HTTPResponseWriterPtr writer(
	HTTPResponseWriter::create(
	    tcp_conn, *request, boost::bind(&TCPConnection::finish, tcp_conn)));

    writer->getResponse().setStatusCode(HTTPTypes::RESPONSE_CODE_FOUND);
    writer->getResponse().setStatusMessage(HTTPTypes::RESPONSE_MESSAGE_FOUND);
    writer->getResponse().addHeader(HTTPTypes::HEADER_LOCATION, redirection_url);
    writer->writeNoCopy(CONTENT);
    writer->send();
}

// ----------------------------------------------------------------------

/// A wrapper around the pion webserver exposing the asio io_service
/// object, which is needed to schedule timers.
/// It also forces listening on a loopback address.
class my_server_t : public pion::net::WebServer {
public:
    my_server_t(int port)
	: pion::net::WebServer(
	    boost::asio::ip::tcp::endpoint(
		boost::asio::ip::address_v4::loopback(),
		port))
	{
	}

    // Make this function publicly available
    boost::asio::io_service & get_io_service() {
	return getIOService();
    }
};

// ----------------------------------------------------------------------

void
webserver_t::keep_alive(session_collection_t &session_collection,
			session_ptr_t session,
			const std::string &relative_url,
			std::ostream &os,
			std::string &mime_type)
{
    recharge_timer();
    if (session_collection.license->check_license(0)) {
	os << "license_ok();" << endl;
    } else {
	os << "license_error();" << endl;
    }
    mime_type = "text/javascript";
}

// XXX get_root_dir should use a session_collection, once the relevant
// parts from conf_t has been moved away
path
get_root_dir(session_ptr_t session)
{
    const path &install_dir = session->html_context.conf.installation_dir;
    return install_dir / "share" / "threadspotter" / "web" / "root";
}

void
webserver_t::title(session_collection_t &session_collection,
		   session_ptr_t session,
		   const std::string &relative_url,
		   std::ostream &os,
		   std::string &mime_type)
{
    if (!session_collection.license->check_license(0))
	throw 403;

    os << "document.title = '"
       << session->html_context.report_title()
       << "';" << endl;

    mime_type = "text/javascript";
}

void
webserver_t::topleft(session_collection_t &session_collection,
		     session_ptr_t session,
		     const std::string &relative_url,
		     std::ostream &os,
		     std::string &mime_type)
{
    if (!session_collection.license->check_license(0))
	throw 403;

    xml_file_t xml(os, Doctype_None);

    html_topleft_t topleft(session->html_context);
    topleft.emit(xml);

    mime_type = "text/html";
}

void
webserver_t::drill_down_table(session_collection_t &session_collection,
			      session_ptr_t session,
			      const std::string &relative_url,
			      std::ostream &os,
			      std::string &mime_type)
{
    if (!session_collection.license->check_license(0))
	throw 403;

    xml_file_t xml(os, Doctype_None);

    vector<string> components;
    split(components, relative_url, is_any_of("/"), token_compress_on);

    if (components.size() == 4 && components[0] == "stat") {
	rpt_sliced_stats_id_t stats_id =
	    lexical_cast<rpt_sliced_stats_id_t>(components[1]);
	double value = lexical_cast<double>(components[3]);
	tp_field_ptr_t f = tp_field_t::create(components[2], value);
	if (f) {
	    html_util_t::emit_thread_popup_table(session->html_context, xml,
						 stats_id, *f);
	}
    }
    mime_type = "text/html";
}

void
webserver_t::drill_down_table_by_kind(session_collection_t &session_collection,
				      session_ptr_t session,
				      const std::string &relative_url,
				      std::ostream &os,
				      std::string &mime_type)
{
    if (!session_collection.license->check_license(0))
	throw 403;

    xml_file_t xml(os, Doctype_None);

    vector<string> components;
    split(components, relative_url, is_any_of("/"), token_compress_on);

    if (components.size() == 3 && components[0] == "stat_by_kind") {
	rpt_sliced_stats_id_t stats_id =
	    lexical_cast<rpt_sliced_stats_id_t>(components[1]);
	tp_field_by_kind_ptr_t f = tp_field_by_kind_t::create(components[2]);
	if (f) {
	    html_util_t::emit_thread_popup_table_by_kind(
		session->html_context, xml, stats_id, *f);
	}
    }
    mime_type = "text/html";
}

void
webserver_t::display_issue(session_collection_t &session_collection,
			   session_ptr_t session,
			   const std::string &relative_url,
			   std::ostream &os,
			   std::string &mime_type)
{
    if (!session_collection.license->check_license(0))
	throw 403;

    // Compiling issues with lots of instructions and deep call stacks may take
    // a long time. Boost the timer so we don't time out while we are
    // generating the issue contents.
    recharge_timer(true);

    html_context_t &html_context(session->html_context);
    xml_file_t xml(os, Doctype_None);

    vector<string> components;
    split(components, relative_url, is_any_of("/"), token_compress_on);
    if (components.size() == 2 && components[0] == "issue") {
	string issue_number_s = components[1];

	html_issue_ptr_t issue;

	// Ugh! There should be a way to look up issues!

	BOOST_FOREACH(const loop_info_map_t::value_type &v,
		      html_context.loop_registry->loop_infos) {
	    const loop_info_ptr_t &loop = v.second;
	    BOOST_FOREACH(rpt_depgroup_id_t depgroup_id, loop->depgroup_ids) {
		depgroup_info_ptr_t &d =
		    html_context.loop_registry->depgroup_infos[depgroup_id];

		BOOST_FOREACH(html_issue_ptr_t &i, d->issues) {
		    if (lexical_cast<string>(i->get_issue_id()) == issue_number_s) {
			issue = i;
			goto found;
		    }
		}
	    }
	}

    found:
	if (issue) {
	    issue->emit_issue_file(xml);
	}
    }

    mime_type = "text/html";

    // Handling issues with lots of instructions and deep call stacks may take
    // a long time for the browser. Boost the timer again.
    recharge_timer(true);
}

void
webserver_t::display_loop(session_collection_t &session_collection,
			  session_ptr_t session,
			  const std::string &relative_url,
			  std::ostream &os,
			  std::string &mime_type)
{
    if (!session_collection.license->check_license(0))
	throw 403;

    // Compiling loops with lots of instructions and deep call stacks may take
    // a long time. Boost the timer so we don't time out while we are
    // generating the loop contents.
    recharge_timer(true);

    html_context_t &html_context(session->html_context);
    xml_file_t xml(os, Doctype_None);

    vector<string> components;

    split(components, relative_url, is_any_of("/"), token_compress_on);
    if (components.size() == 2 && components[0] == "loop") {
	string loop_id_s = components[1];
	int loop_id = lexical_cast<int>(loop_id_s);

	loop_info_map_t::const_iterator i =
	    html_context.loop_registry->loop_infos.find(loop_id);

	if (i != html_context.loop_registry->loop_infos.end()) {
	    html_loop_t html_loop(html_context, i->second);
	    html_loop.emit_loop_file(xml);
	}
    }
    mime_type = "text/html";

    // Handling loops with lots of instructions and deep call stacks may take
    // a long time for the browser. Boost the timer again.
    recharge_timer(true);
}

void
webserver_t::display_source(session_collection_t &session_collection,
			    session_ptr_t session,
			    const std::string &relative_url,
			    std::ostream &os,
			    std::string &mime_type)
{
    if (!session_collection.license->check_license(0))
	throw 403;

    // Compiling large source files may take a long time. Boost the timer so we
    // don't time out while we are generating the source file contents.
    recharge_timer(true);

    html_context_t &html_context(session->html_context);
    xml_file_t xml(os, Doctype_None);

    vector<string> components;
    split(components, relative_url, is_any_of("/"), token_compress_on);
    if (components.size() == 2 && components[0] == "source") {
	string source_number_s = components[1];

	int file_id = lexical_cast<int>(source_number_s);

	// XXX stupid translation back and forth between file_id and filename.
	string filename = html_context.report->source_files[file_id].filename;

	file_map_t::const_iterator i = html_context.file_repos->file_map.find(filename);
	if (i != html_context.file_repos->file_map.end()) {
	    const line_info_map_t &line_info_map = i->second;

	    html_source_t::write_annotated_file(xml, html_context, filename, line_info_map);
	} else {
	    T_P TEXT("XXX Can't find file in file_map");
	}
    }
    mime_type = "text/html";

    // Handling large source files may take a long time for the browser. Boost
    // the timer again.
    recharge_timer(true);
}

void
webserver_t::display_diagram(session_collection_t &session_collection,
			     session_ptr_t session,
			     const std::string &relative_url,
			     std::ostream &os,
			     std::string &mime_type)
{
    if (!session_collection.license->check_license(0))
	throw 403;

    html_context_t &html_context(session->html_context);
    const rpt_report_cptr_t report = html_context.report;
    vector<string> components;
    split(components, relative_url, is_any_of("/"), token_compress_on);
    // E.g diagram/fr/l/#
    if (components.size() == 4 && components[0] == "diagram") {
	string type = components[1]; // fr, wb, su
	string kind = components[2]; // l, d, global (=loop, depgroup, global)
	int id = lexical_cast<int>(components[3]); // loop_id, depgroup_id

	struct local_function_t {
	    static void prepare(std::ostream &os,
				drill_down_t &dd,
				const std::string &type,
				double max_mr_scale = 1.0)
		{
		    if (type == "fr")
			dd.prepare_fr(os);
		    else if (type == "wb")
			dd.prepare_wb(os);
		    else if (type == "su")
			dd.prepare_su(os);
		    else if (type == "fr-100")
			dd.prepare_fr(os, max_mr_scale);
		    else if (type == "wb-100")
			dd.prepare_wb(os, max_mr_scale);
		}
	};

	if (kind == "global") {
	    // type is something like perf-fr-100.png. Strip perf- and .png
	    string t = type.substr(5, type.length()-9);
	    double max_mr_scale = html_context.conf.max_mr_scale;
	    drill_down_global_t dd(report, "", max_mr_scale);
	    local_function_t::prepare(os, dd, t, max_mr_scale);
	} else {
	    // type is fr, wb or su.
	    rpt_location_set_t pc_set;
	    if (kind == "l")
		pc_set = report->get_loop_excl_location_set(id);
	    else if (kind == "d")
		pc_set = report->get_depgroup_location_set(id);

	    drill_down_pc_set_t dd(report,
				   "",
				   report->get_location_set_id(pc_set));

	    local_function_t::prepare(os, dd, type);
	}
    }
    mime_type = "image/png";
}

void
webserver_t::display_gauge(session_collection_t &session_collection,
			   session_ptr_t session,
			   const std::string &relative_url,
			   std::ostream &os,
			   std::string &mime_type)
{
    if (!session_collection.license->check_license(0))
	throw 403;

    html_context_t &html_context(session->html_context);
    application_verdict_t app_verdict(html_context);

    vector<string> components;
    split(components, relative_url, is_any_of("/"), token_compress_on);

    // E.g gauge/threading_gauge.png

    if (components.size() == 2 && components[0] == "gauge") {
	if (components[1] == PNG_BANDWIDTH_GAUGE) {
	    app_verdict.render_fr_gauge(os);
	} else if (components[1] == PNG_LATENCY_GAUGE) {
	    app_verdict.render_mr_gauge(os);
	} else if (components[1] == PNG_LOCALITY_GAUGE) {
	    app_verdict.render_utilization_gauge(os);
	} else if (components[1] == PNG_THREADING_GAUGE) {
	    app_verdict.render_mt_gauge(os);
	}
    }

    mime_type = "image/png";
}

void
webserver_t::front(session_collection_t &session_collection,
		   session_ptr_t session,
		   const std::string &relative_url,
		   std::ostream &os,
		   std::string &mime_type)
{
    if (!session_collection.license->check_license(0))
	throw 403;

    if (session.get() == NULL)
	throw 401; // or something.


    html_context_t &html_context(session->html_context);
    application_verdict_t app_verdict(html_context);

    xml_file_t xml(os, Doctype_None);
    html_front_t front(html_context, app_verdict);
    front.emit(xml);

    mime_type = "text/html";
}

void
webserver_t::sessions(session_collection_t &session_collection,
		      session_ptr_t session,
		      const std::string &relative_url,
		      std::ostream &os,
		      std::string &mime_type)
{
    if (!session_collection.license->check_license(0))
	throw 403;

    xml_file_t xml(os, Doctype_None);

    T_P {
	T_IMG A_SRC("images/rw-small.png");
    }

    T_H1 {
	TEXT("Loaded reports");
    }


    T_TABLE {
	T_TR {
	    T_TH TEXT("Report");
	    T_TH TEXT("");
	}

	BOOST_FOREACH(session_ptr_t session, session_collection.session_vector) {

	    T_TR {
		rpt_report_cptr_t report = session->html_context.report;

		string static_resource = "/" + session->id;

		T_TD {
                    // TODO: Get the sample file name from the report instead!
		    T_A A_HREF(static_resource + "/front.html")
			TEXT(report->sample_file);
		}

		T_TD {
		    bool br = false;
		    if (report->argv.size()) {
			TEXT("Application: " + report->argv[0]);
			br = true;
		    }

		    if (br)
			T_BR;
		    TEXT(session->html_context.report_title());
		    br = true;
     		}
	    }
	}
    }

    mime_type = "text/html";
}


void
webserver_t::add_session(session_ptr_t session)
{
    using boost::bind;

    std::string session_guid = session->id;
    std::string static_resource = "/" + session_guid;
    webserver->addService(static_resource, new pion::plugins::FileService);
    webserver->setServiceOption(static_resource,
				"directory",
				get_root_dir(session).string());

    // Require logged in users for this resource
    auth->addRestrict(static_resource);

    std::string dynamic_resource = static_resource + "/d";
    webserver->addService(dynamic_resource,
			  new dyn_content_t(session_collection, session));

    // Title
    session->url_repository.add_dispatcher(
	"title",
	boost::bind(&webserver_t::title, this, _1, _2, _3, _4, _5));

    // Front page
    session->url_repository.add_dispatcher(
	"front",
	boost::bind(&webserver_t::front, this, _1, _2, _3, _4, _5));

    // Top left frame
    session->url_repository.add_dispatcher(
	"top-left",
	boost::bind(&webserver_t::topleft, this, _1, _2, _3, _4, _5));

    // issue, loop details, source
    session->url_repository.add_dispatcher(
	"issue/",
	boost::bind(&webserver_t::display_issue, this, _1, _2, _3, _4, _5));

    session->url_repository.add_dispatcher(
	"loop/",
	boost::bind(&webserver_t::display_loop, this, _1, _2, _3, _4, _5));

    session->url_repository.add_dispatcher(
	"source/",
	boost::bind(&webserver_t::display_source, this, _1, _2, _3, _4, _5));

    // drilldown
    session->url_repository.add_dispatcher(
	"stat/",
	boost::bind(&webserver_t::drill_down_table, this, _1, _2, _3, _4, _5));

    session->url_repository.add_dispatcher(
	"stat_by_kind/",
	boost::bind(&webserver_t::drill_down_table_by_kind, this, _1, _2, _3, _4, _5));

    // diagrams
    session->url_repository.add_dispatcher(
	"diagram/",
	boost::bind(&webserver_t::display_diagram, this, _1, _2, _3, _4, _5));

    session->url_repository.add_dispatcher(
	"gauge/",
	boost::bind(&webserver_t::display_gauge, this, _1, _2, _3, _4, _5));
}

void
webserver_t::recharge_timer(bool initial_boost)
{
    inactivity_timer->cancel();

    int seconds = inactivity_timeout;
    if (initial_boost)
	seconds = max(seconds, 60);

    inactivity_timer->expires_from_now(
	boost::posix_time::seconds(seconds));
    inactivity_timer->async_wait(
	boost::bind(&webserver_t::timeout,
		    this,
		    placeholders::error));
}

void
webserver_t::timeout(const error_code &error)
{
    if (error == boost::asio::error::operation_aborted)
        return;
    session_collection.shutdown();
}


webserver_t::webserver_t(int port,
			 session_collection_t &session_collection,
			 int inactivity_timeout)
    : session_collection(session_collection),
      inactivity_timeout(inactivity_timeout)
{
    // XXX ugly, assumes that there is at least one session.
    // We should separate some things from the conf_t object.
    AM_ASSERT(session_collection.session_vector.size() == 1);
    session_ptr_t session = session_collection.session_vector[0];
    std::string cookie = session->html_context.conf.cookie;

    pion::PionLogger logger = PION_GET_LOGGER("threadspotter");
    PION_LOG_SETLEVEL_WARN(logger);

    try {

	using boost::bind;
	using namespace pion::net;

	PionUserManagerPtr user_manager(new PionUserManager);
	auth.reset(new auth_t(user_manager));
	auth->addUser("user", session->html_context.conf.cookie);

	// Disable extraneous logins
	if (!session_collection.allow_many_logins)
	    auth->setOption("single", "true");

	const path &install_dir = session->html_context.conf.installation_dir;

	path doc = install_dir / "doc" ;
	path html = install_dir / "share" / "threadspotter" / "web";

	my_server_t *my_server = new my_server_t(port);
	webserver.reset(my_server);

	// This is the home of "/" and "/index.html"
	path root_dir = get_root_dir(session);
	webserver->addService("/", new pion::plugins::FileService);
	webserver->setServiceOption("/", "directory", root_dir.string());
	webserver->setServiceOption("/", "file", (root_dir / "index.html").string());

	auth->addRestrict("/");
#if 0
	auth->addPermit("/login");
	auth->addPermit("login");
	auth->addPermit("/loginform");
	auth->addPermit("loginform");
#endif

	// This is the main start page
	webserver->addService("/start",
			      new start_redirect_t(session_collection));

	// Home of static files, css, images, js
	webserver->addService("/css", new pion::plugins::FileService);
	webserver->setServiceOption("/css", "directory", (html / "css").string());
	webserver->addService("/images", new pion::plugins::FileService);
	webserver->setServiceOption("/images", "directory", (html / "images").string());
	webserver->addService("/js", new pion::plugins::FileService);
	webserver->setServiceOption("/js", "directory", (html / "js").string());

	// Home of documentation files
	webserver->addService("/doc", new pion::plugins::FileService);
	webserver->setServiceOption("/doc", "directory", doc.string());

	// Login check
	// webserver->addService("/loginform", new login_check_t(session, "/"));

	// Dynamically constructed global data, typically the list of sessions
	webserver->addService("/d", new dyn_content_t(session_collection, session_ptr_t()));
	auth->addRestrict("/d");

	// Populate the url_repository with some global resources
	session_collection.global_url_repository.add_dispatcher(
	    "sessions",
	    boost::bind(&webserver_t::sessions, this, _1, _2, _3, _4, _5));

	session_collection.global_url_repository.add_dispatcher(
	    "ping",
	    boost::bind(&webserver_t::keep_alive, this, _1, _2, _3, _4, _5));

	inactivity_timer.reset(new boost::asio::deadline_timer(my_server->get_io_service()));


	webserver->setAuthentication(auth);

	started = false;
    } catch (pion::net::WebServer::WebServiceException &) {
	// File not found etc.
        throw;
    }
    // TODO: Doesn't compile on Linux!
    //PION_LOG_SETLEVEL_DEBUG(PION_GET_LOGGER("foo"));
}


webserver_t::~webserver_t()
{
    if (started) {
	stop();
    }
}

void
webserver_t::start()
{
    webserver->start();

#if !defined(_WIN32)
    // Set the close-on-exec flag in the TCP socket we listen on, so it does
    // not leak to other processes we spawn.
    int fd = webserver->m_tcp_acceptor.native();
    int flags = fcntl(fd, F_GETFD);
    AM_ASSERT(flags != -1);
    flags = fcntl(fd, F_SETFD, flags | FD_CLOEXEC);
    AM_ASSERT(flags != -1);
#endif

    recharge_timer(true);
    session_collection.server_port = webserver->getPort();
    started = true;
}

void
webserver_t::stop()
{
    if (started) {
	inactivity_timer->cancel();
	webserver->stop(true);
	started = false;
    }
}

END_NAMESPACE;
END_NAMESPACE;
END_NAMESPACE;
