/// @file tools/atf_resampler/atf_resampler.cc

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

// Converts an ATF sample stream into a fenris sample stream.


#include "library/atf_resampler/atf_resampler.hh"
#include "library/atf_resampler/exceptions.hh"
#include "library/exe_path/exe_path.hh"
#include "library/exe_path/path_substitute.hh"
#include "library/fenris/flow_file_writer.hh"
#include "library/license/license.hh"
#include "library/version/version.h"
#include "library/warning/warning.hh"
#include "library/feedback/licensed_feedback.hh"

#include "product.h"

#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>

#include <iostream>

#include <errno.h>

/* Bring in acumem svn version information */
BRING_IN_VERSIONINFO

// Enable testing parameters
//#define ATF_RESAMPLER_INTERNAL

using namespace std;
using namespace acumem::library;
using namespace acumem::license;
namespace po = boost::program_options;
namespace fs = boost::filesystem;


struct conf_t {
    conf_t()
        : sample_count(0),
	  legacy(false),
	  compression(PRODUCT_SMP_COMPRESSION),
	  product(PRODUCT_SMP_PRODUCT)
        {}

    fs::path in_filename;
    fs::path out_filename;
    fs::path acq_filename;
    fnr_sample_count_t sample_count;
    fs::path tmpdir;

    bool legacy;
    stream_container_compression_t compression;
    stream_container_product_t product;
};

static bool
parse_args(conf_t *conf, int argc, char **argv)
{
    po::options_description desc("Options");
#ifdef ATF_RESAMPLER_INTERNAL
    string compression;
    string product;
#endif

    conf->tmpdir = acumem::exe_path::get_temp_path();

    desc.add_options()
        ("help,h", "Print this help message.")
#ifdef ATF_RESAMPLER_INTERNAL
	("legacy,l",
	 "Create a headerless legacy file.")
	("compression,c", po::value<string>(&compression),
	 "Compression method to use.")
	("product,p", po::value<string>(&product),
	 "Product ID to use, see source for available product IDs.")
#endif
        ("in-file,i", po::value<fs::path>(&conf->in_filename),
         "Specifies the input file. If no file is specified stdin is used.")
        ("out-file,o", po::value<fs::path>(&conf->out_filename),
         "Specifies the output file. If no file is specified stdout is used. "
	 "Template substitutions apply, just like on the sampler command.")
        ("acquisition-info", po::value<fs::path>(&conf->acq_filename),
         "Specifies the optional acquisition info file.")
        ("sample-count", po::value<fnr_sample_count_t>(&conf->sample_count),
         "Decimate the samples to correspond to this number of memory "
         "accesses.")
	("tmpdir", po::value<fs::path>(&conf->tmpdir),
	 ("Store temporary files in this directory instead of the system "
	  "default temp directory or the one pointed out by the environment "
	  "variable THREADSPOTTER_TMPDIR"))
        ;

    po::variables_map vm;
    try {
        po::store(po::parse_command_line(argc, argv, desc), vm);
        po::notify(vm);
    }
    catch (po::error &e) {
        cerr << "Error: " << e.what() << endl;
        cerr << desc << endl;
        return true;
    }

    bool error = false;

    if (vm.count("help") > 0) {
        cerr << desc << endl;
        exit(0);
    }

    if (vm.count("in-file") > 1) {
        cerr << "Error: multiple input files specified" << endl;
        error = true;
    }

    if (vm.count("out-file") > 1) {
        cerr << "Error: multiple output files specified" << endl;
        error = true;
    }

    if (vm.count("acquisition-info") > 1) {
        cerr << "Error: multiple acquisition info files specified" << endl;
        error = true;
    }

#ifdef ATF_RESAMPLER_INTERNAL
    conf->legacy = vm.count("legacy") > 1;

    if (vm.count("compression") > 1) {
        cerr << "Error: compression specified multiple times" << endl;
        error = true;
    }

    if (vm.count("product") > 1) {
        cerr << "Error: product specified multiple times" << endl;
        error = true;
    }

    if (vm.count("compression") > 0) {
	if (compression == "none")
	    conf->compression = STREAM_CONTAINER_COMPRESSION_NONE;
	else if (compression == "deflate")
	    conf->compression = STREAM_CONTAINER_COMPRESSION_DEFLATE;
#ifdef ENABLE_FNR_BZ2
	else if (compression == "bzip2")
	    conf->compression = STREAM_CONTAINER_COMPRESSION_BZIP2;
#endif
	else {
	    cerr << "Error: illegal compression specified." << endl
		 << "Available compressions are:" << endl
		 << "  none    - No compression" << endl
		 << "  deflate - Zlib deflate" << endl;
#ifdef ENABLE_FNR_BZ2
	    cout << "  bzip2 - bzip2 (experimental)" << endl;
#endif
	    error = true;
	}
    }

    if (vm.count("product") > 0) {
	if (product == "vpe")
	    conf->product = STREAM_CONTAINER_PRODUCT_VPE;
	else if (product == "vpemp")
	    conf->product = STREAM_CONTAINER_PRODUCT_VPEMP;
	else if (product == "litmus")
	    conf->product = STREAM_CONTAINER_PRODUCT_LITMUS;
	else if (product == "ss12")
	    conf->product = STREAM_CONTAINER_PRODUCT_SS12;
	else if (product == "internal")
	    conf->product = STREAM_CONTAINER_PRODUCT_INTERNAL;
	else {
	    cerr << "Error: illegal product specified." << endl
		 << "Available products are:" << endl
		 << "  vpe      - Acumem VPE" << endl
		 << "  vpemp    - Acumem VPE MP" << endl
		 << "  litmus   - Acumem Litmus" << endl
		 << "  ss12     - Acumem SlowSpotter 1.2" << endl
		 << "  internal - For internal use only" << endl;
	    error = true;
	}
    }
#endif // defined(ATF_RESAMPLER_INTERNAL)

    return error;
}

int
main(int argc, char **argv)
{
    try {
	using namespace acumem::library;
	conf_t conf;
	warning_consumer_ptr_t wc(new warning_consumer_stderr_t());
	license_manager_ptr_t license(new license_manager_t(wc));
	
	feature_set_ptr_t fs = license->acquire(license_manager_t::tool_kind);
	if (!fs) {
	    cerr << "No license token available." << endl;
	    return 1;
	}
	
	feedback_ptr_t feedback(new licensed_feedback_t(license));
	
	/* Parse arguments. */
	if (parse_args(&conf, argc, argv))
	    return 1;
	
	/* Setup header */
	stream_container_header_t header;
	memset(&header, 0, sizeof(stream_container_header_t));
	header.version = conf.legacy ? 0 : STREAM_CONTAINER_VERSION;
	header.product = conf.product;
	header.data_type = STREAM_CONTAINER_DATA_SAMPLES_BINARY;
	header.compression = conf.compression;
	
	/* Open the input and output flows. */
	acumem::exe_path::path_substitute_t path_substitute;
	fs::path smp = path_substitute.substitute(conf.out_filename);
	flow_file_writer_t out_flow(smp, header, wc);

	flow_atf_resampler_t in_flow(
	    conf.in_filename,
	    conf.acq_filename,
	    conf.sample_count,
            !fs->unlimited_threads(),
	    &out_flow,
	    wc,
	    feedback,
            conf.tmpdir);

	/* Do the actual resampling. */
	if (in_flow.dispatch() != FLOW_OK) {
	    cerr << "Resampling failed, please make sure there " << endl
		 << "is enough disk space left for the operation." << endl;
	    return 1;
	}
    } catch (const resampler_exception_t &re) {
	cerr << "Resampling failed: " << re.what() << endl;
	return 1;
    } catch (const flow_exception_t &fe) {
	cerr << "Failed to create output file: " << fe.what() << endl;
        return 1;
    } catch (const stream_exception_t &se) {
	cerr << "Failed to create output file: " << se.what() << endl;
        return 1;
    } catch (const std::exception &e) {
        cerr << "Resampling failed: " << e.what() << endl;
        return 1;
    } catch (...) {
        cerr << "Unhandled internal error." << endl;
        return 1;
    }

    return 0;
}
