/// @file tools/cache_sim/cache_sim.cc

// Copyright (c) 2006-2008, 2010 Acumem AB.  All rights reserved.

/*
 *  This file is part of ThreadSpotter.
 *
 *  ThreadSpotter is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ThreadSpotter is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with ThreadSpotter.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "library/cache_sim/cachesim.hh"
#include "library/fenris/flow_file_reader.hh"
#include "library/warning/warning_consumer.hh"
#include "library/version/version.h"

#include <boost/program_options.hpp>

#ifndef _WIN32
#include <unistd.h>
#endif

#include <iostream>
#include <string>

using namespace std;
using namespace boost;
namespace po = boost::program_options;
using namespace acumem::library;

/* Bring in acumem svn version information */
BRING_IN_VERSIONINFO

static bool
is_power_of_two(uint64_t val)
{
    return val != 0 && (val & (val - 1)) == 0;
}

struct conf_t {
    conf_t();
    int parse_args(int argc, char **argv);

    string infile_name;
    fnr_line_size_t linesize_min;
    fnr_line_size_t linesize_max;
    fnr_cache_size_t cachesize_min;
    fnr_cache_size_t cachesize_max;
    bool calc_footprint;
    bool calc_spatial_usage;
    bool compact;
};

conf_t::conf_t()
{
    linesize_min = 0;
    linesize_max = 0;
    cachesize_min = 0;
    cachesize_max = 0;
    calc_footprint = false;
    calc_spatial_usage = false;
    compact = true;
}

int
conf_t::parse_args(int argc, char **argv)
{
    po::options_description desc("Allowed options");
    desc.add_options()
	("help,h", "produce help message")
	("verbose,v", "print lots of output")
	("in-file,i", po::value<string>(&infile_name),
	 "specifies the input file. stdin is used if empty")
	("min-linesize,l", po::value<fnr_line_size_t>(&linesize_min),
	 "minimum line size")
	("max-linesize,L", po::value<fnr_line_size_t>(&linesize_max),
	 "maximum line size")
	("min-cachesize,c", po::value<fnr_cache_size_t>(&cachesize_min),
	 "minimum cache size")
	("max-cachesize,C", po::value<fnr_cache_size_t>(&cachesize_max),
	 "maximum cache size")
	("footprint,f",
	 "calculate maximum footprint")
	("spatial-usage,u",
	 "calculate spatial usage")
    ;

    po::variables_map vm;
    try {
        po::store(po::parse_command_line(argc, argv, desc), vm);
        po::notify(vm);
    }
    catch (po::error &e) {
        cerr << "Error: " << e.what() << endl;
        cerr << desc << endl;
        return 1;
    }

    if (vm.count("help") != 0) {
	cerr << desc << endl;
        return 1;
    }

    if (vm.count("footprint") != 0) {
	calc_footprint = true;
    }

    if (vm.count("spatial-usage") != 0) {
	calc_spatial_usage = true;
    }

    if (vm.count("verbose") != 0) {
	compact = false;
    }

    if (! is_power_of_two(linesize_min)){
	cerr << "linesize_min must be a power of two" << endl;
	return 1;
    }

    if (! is_power_of_two(linesize_max)) {
	cerr << "linesize_max must be a power of two" << endl;
	return 1;
    }

    if (linesize_max < linesize_min) {
	cerr << "linesize_max is not allowed to be smaller than linesize_min" << endl;
	return 1;
    }

    if (! is_power_of_two(cachesize_min)) {
	cerr << "cachesize_min must be a power of two" << endl;
	return 1;
    }

    if (! is_power_of_two(cachesize_max)) {
	cerr << "cachesize_max must be a power of two" << endl;
	return 1;
    }

    if (cachesize_max < cachesize_min) {
	cerr << "cachesize_max is not allowed to be smaller than cachesize_min" << endl;
	return 1;
    }

    if (cachesize_min < linesize_max) {
	cerr << "cachesize_min is not allowed to be less than linesize_max" << endl;
	return 1;
    }

    return 0;
}

static void
print_header()
{
  cout <<
    "replacement\t"	"cache_size\t"		"line_size\t"	"assoc\t"
    "rd_total\t"	"rd_repl_misses\t"	"rd_cold_misses\t"
    "wr_total\t"	"wr_repl_misses\t"	"wr_cold_misses\t"
    "wb_incoming\t"	"wb_outgoing\t"
    "miss_ratio\t"	"miss_ratio_rd\t"	"miss_ratio_wr\t"
    "cold_miss_ratio\t"	"cold_miss_ratio_rd\t"	"cold_miss_ratio_wr\t"
    "repl_miss_ratio\t"	"repl_miss_ratio_rd\t"	"repl_miss_ratio_wr\t"
    "byte granularity footprint\t" "line granularity footprint\t"
    "used_space\t"	"wasted_space\t"	"spatial_usage\t"
       << endl;
}

static void
print_result(const char *replacement, cachesim_base_t *cachesim)
{
    cachesim_conf_t conf;
    cachesim->get_conf(conf);
    cout << "replacement: " << replacement
         << ", cache_size: " << conf.size
         << ", line_size: " << conf.line_size
         << ", assoc: " << conf.assoc << endl;

    cachesim_result_t res;
    cachesim->get_counters(res);
    cout << "\trd_total      : " << res.rd_total << endl;
    cout << "\trd_repl_misses: " << res.rd_repl_misses << endl;
    cout << "\trd_cold_misses: " << res.rd_cold_misses << endl;
    cout << "\twr_total      : " << res.wr_total << endl;
    cout << "\twr_repl_misses: " << res.wr_repl_misses << endl;
    cout << "\twr_cold_misses: " << res.wr_cold_misses << endl;
    cout << "\twb_incomming  : " << res.wb_incomming << endl;
    cout << "\twb_outgoing   : " << res.wb_outgoing << endl;
    cout << endl;

    cout << "\tmiss ratio         : "
         << ((double)(res.rd_repl_misses + res.rd_cold_misses
                      + res.wr_repl_misses + res.wr_cold_misses)
             / (double)(res.rd_total + res.wr_total)) << endl;
    cout << "\tmiss ratio rd      : "
         << ((double)(res.rd_repl_misses + res.rd_cold_misses)
             / (double)(res.rd_total + res.wr_total)) << endl;
    cout << "\tmiss ratio wr      : "
         << ((double)(res.wr_repl_misses + res.wr_cold_misses)
             / (double)(res.rd_total + res.wr_total)) << endl;
    cout << "\tcold miss ratio    : "
         << ((double)(res.rd_cold_misses + res.wr_cold_misses)
             / (double)(res.rd_total + res.wr_total)) << endl;
    cout << "\tcold miss ratio rd : "
         << ((double)(res.rd_cold_misses)
             / (double)(res.rd_total + res.wr_total)) << endl;
    cout << "\tcold miss ratio wr : "
         << ((double)(res.wr_cold_misses)
             / (double)(res.rd_total + res.wr_total)) << endl;
    cout << "\trepl miss ratio    : "
         << ((double)(res.rd_repl_misses + res.wr_repl_misses)
             / (double)(res.rd_total + res.wr_total)) << endl;
    cout << "\trepl miss ratio rd : "
         << ((double)(res.rd_repl_misses)
             / (double)(res.rd_total + res.wr_total)) << endl;
    cout << "\trepl miss ratio wr : "
         << ((double)(res.wr_repl_misses)
             / (double)(res.rd_total + res.wr_total)) << endl;
    cout << endl;

    if (conf.calc_footprint) {
        cout << "\tbyte granularity footprint : " << res.fp_byte << endl;
        cout << "\tline granularity footprint : " << res.fp_line << endl;
        cout << endl;
    }

    if (conf.calc_spatial_usage) {
        uint64_t su_num = res.used_bytes;
        uint64_t su_den = (conf.line_size
                           * (res.rd_repl_misses + res.rd_cold_misses
                              + res.wr_repl_misses + res.wr_cold_misses));
        cout << "\tused space    : " << su_num << endl;
        cout << "\twasted space  : " << su_den - su_num << endl;
        cout << "\tspatial usage : " << (double)su_num / su_den << endl;
        cout << endl;
    }
}

static void
print_result2(const char *replacement, cachesim_base_t *cachesim)
{
    const char *sep = "\t";
    cachesim_conf_t conf;
    cachesim->get_conf(conf);
    cout << replacement		<< sep
         << conf.size		<< sep
         << conf.line_size	<< sep
         << conf.assoc		<< sep;

    cachesim_result_t res;
    cachesim->get_counters(res);
    cout << res.rd_total << sep;
    cout << res.rd_repl_misses << sep;
    cout << res.rd_cold_misses << sep;
    cout << res.wr_total << sep;
    cout << res.wr_repl_misses << sep;
    cout << res.wr_cold_misses << sep;
    cout << res.wb_incomming << sep;
    cout << res.wb_outgoing << sep;

    cout << ((double)(res.rd_repl_misses + res.rd_cold_misses
                      + res.wr_repl_misses + res.wr_cold_misses)
             / (double)(res.rd_total + res.wr_total)) << sep;
    cout << ((double)(res.rd_repl_misses + res.rd_cold_misses)
             / (double)(res.rd_total + res.wr_total)) << sep;
    cout << ((double)(res.wr_repl_misses + res.wr_cold_misses)
             / (double)(res.rd_total + res.wr_total)) << sep;
    cout
         << ((double)(res.rd_cold_misses + res.wr_cold_misses)
             / (double)(res.rd_total + res.wr_total)) << sep;
    cout
         << ((double)(res.rd_cold_misses)
             / (double)(res.rd_total + res.wr_total)) << sep;
    cout
         << ((double)(res.wr_cold_misses)
             / (double)(res.rd_total + res.wr_total)) << sep;
    cout
         << ((double)(res.rd_repl_misses + res.wr_repl_misses)
             / (double)(res.rd_total + res.wr_total)) << sep;
    cout
         << ((double)(res.rd_repl_misses)
             / (double)(res.rd_total + res.wr_total)) << sep;
    cout
         << ((double)(res.wr_repl_misses)
             / (double)(res.rd_total + res.wr_total)) << sep;

    if (conf.calc_footprint) {
        cout << res.fp_byte << sep;
        cout << res.fp_line << sep;
    }
    else
        cout << 0 << sep << 0 << sep;

    if (conf.calc_spatial_usage) {
        uint64_t su_num = res.used_bytes;
        uint64_t su_den = (conf.line_size
                           * (res.rd_repl_misses + res.rd_cold_misses
                              + res.wr_repl_misses + res.wr_cold_misses));
        cout << su_num << sep;
        cout << su_den - su_num << sep;
        cout << (double)su_num / (double)su_den << endl;
    }
    else
        cout << 0 << sep << 0 << sep << 0 << sep;
}

int
main(int argc, char *argv[])
{
    conf_t conf;

    if (0 != conf.parse_args(argc, argv))
	return 1;

    trace_to_cachesim_t trace2sim;

    warning_consumer_ptr_t warning_consumer(
        new warning_consumer_stderr_t());

    flow_file_reader_t in(
        conf.infile_name == "" ? NULL : conf.infile_name.c_str(),
        &trace2sim,
        warning_consumer);

    list<cachesim_rand_t *> rand_list;
    list<cachesim_lru_t *> lru_list;

    for (fnr_cache_size_t size = conf.cachesize_min; size <= conf.cachesize_max; size *= 2) {
	for (fnr_line_size_t line_size = conf.linesize_min; line_size <= conf.linesize_max; line_size *= 2) {
	    fnr_line_count_t assoc = size / line_size;

	    cachesim_rand_t *rand = new cachesim_rand_t(
                size, line_size, assoc,
                conf.calc_footprint, conf.calc_spatial_usage, 0);
	    trace2sim.add_cache(rand);
	    rand_list.push_back(rand);

	    cachesim_lru_t *lru = new cachesim_lru_t(
                size, line_size, assoc,
                conf.calc_footprint, conf.calc_spatial_usage);
	    trace2sim.add_cache(lru);
	    lru_list.push_back(lru);
	}
    }

    // read all events in file
    in.dispatch();

    trace2sim.clear_cache_list();

    if (conf.compact)
      print_header();

    for (list<cachesim_rand_t *>::const_iterator
             ri = rand_list.begin(), re = rand_list.end();
         ri != re; ri++) {
	if (conf.compact)
	    print_result2("random", *ri);
	else
	    print_result("random", *ri);
	delete *ri;
    }

    for (list<cachesim_lru_t *>::const_iterator
             li = lru_list.begin(), le = lru_list.end();
         li != le; li++) {
	if (conf.compact)
	    print_result2("lru", *li);
	else
	    print_result("lru", *li);
        delete *li;
    }

    return 0;
}
