/// @file tools/container_tools/cnt_cat.cc

// Copyright (c) 2008 Acumem AB.  All rights reserved.

// Print information about a stream container, or dump the
// decompressed contents.

#include "library/stream/stream_file.hh"
#include "library/stream/stream_container.hh"
#include "library/version/version.h"

#include <boost/program_options.hpp>

#include <fstream>
#include <iostream>
#include <memory>

/* Bring in acumem svn version information */
BRING_IN_VERSIONINFO


using namespace acumem::library;
using namespace std;
using namespace boost;

namespace po = boost::program_options;

class stream_aborter_t
    : public stream_t
{
public:
    stream_aborter_t()
	: stream_t() {}

    stream_status_t process_data(const char *buffer, size_t len) {
	return STREAM_ABORT;
    }

    stream_status_t flush() {
	return STREAM_OK;
    }
};

struct conf_t {
    conf_t()
	: in_file(""), out_file(""), dump_data(false) {}
    string                  in_file;
    string                  out_file;
    bool                    dump_data;
};

static void
parse_args(conf_t &conf, int argc, char **argv)
{
    po::options_description desc("Allowed options");
    desc.add_options()
	("help,h", "produce help message")
	("data,d", "dump container data stream")
	("input-file,i", po::value<string>(&conf.in_file),
	 "input file, 'stdin' is used by default.")
	("output-file,o", po::value<string>(&conf.out_file),
	 "output file when dumping data, 'stdout' is used by default.")
	;

    po::positional_options_description p_desc;
    p_desc.add("input-file", 1);
    p_desc.add("output-file", 1);

    po::variables_map vm;
    po::store(
	po::command_line_parser(argc, argv).options(desc).positional(p_desc).run(),
	vm);
    po::notify(vm);

    if (vm.count("help") != 0) {
	cout << "Usage: " << argv[0] << " [OPTION]... [IN FILE] [OUT FILE]" << endl
	     << endl
	     << "Prints information about an Freja stream container." << endl
	     << endl;
	cout << desc << endl;
	exit(0);
    }

    if (conf.in_file == "-")
	conf.in_file = "";
    if (conf.out_file == "-")
	conf.out_file = "";

    conf.dump_data = vm.count("data") > 0;
}

int
main(int argc, char *argv[])
{
    conf_t conf;

    parse_args(conf, argc, argv);

    try {
        warning_consumer_ptr_t warning_consumer(
            new warning_consumer_stderr_t());

	auto_ptr<stream_t> out(
	    conf.dump_data ?
	    (stream_t*)(
		new stream_file_writer_t(
		    conf.out_file == "" ? NULL : conf.out_file.c_str()))
	    : (stream_t*)(new stream_aborter_t()));;
	stream_container_reader_t container(out.get(), warning_consumer);
    	stream_file_reader_t in(
            conf.in_file == "" ? NULL : conf.in_file.c_str(),
            &container);

	in.dispatch();
	if (conf.dump_data)
	    cerr << container.get_header();
	else
	    cout << container.get_header();
    } catch (stream_exception_t se) {
	cerr << "Caught stream exception:" << endl
	     << se.what() << endl;
	exit(1);
    }

    return 0;
}
