/// @file tools/container_tools/cnt_dog.cc

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

// Creates an Acumem stream container.

#include "library/stream/stream_file.hh"
#include "library/stream/stream_container.hh"
#include "library/version/version.h"

#include <boost/program_options.hpp>

#include <fstream>
#include <iostream>
#include <memory>

/* Bring in acumem svn version information */
BRING_IN_VERSIONINFO


using namespace acumem::library;
using namespace std;
using namespace boost;

namespace po = boost::program_options;

struct conf_t {
    conf_t()
	: in_file(""), out_file(""),
	  legacy(false),
	  compression(STREAM_CONTAINER_COMPRESSION_NONE),
	  product(STREAM_CONTAINER_PRODUCT_INTERNAL)
	{}
    string                  in_file;
    string                  out_file;

    bool legacy;
    stream_container_compression_t compression;
    stream_container_product_t product;
};

static void
parse_args(conf_t &conf, int argc, char **argv)
{
    string compression;
    string product;

    po::options_description desc("Allowed options");
    desc.add_options()
	("help,h", "produce help message")
	("input-file,i", po::value<string>(&conf.in_file),
	 "input file, (default: stdin)")
	("output-file,o", po::value<string>(&conf.out_file),
	 "output file when dumping data (default: stdout)")
	("legacy,l",
	 "Create a headerless legacy file. Forces compression to deflate.")
	("compression,c", po::value<string>(&compression),
	 "Compression method to use (default: none)")
	("product,p", po::value<string>(&product),
	 "Product ID to use, see source for available product IDs (default: internal)")
	;

    po::positional_options_description p_desc;
    p_desc.add("input-file", 1);
    p_desc.add("output-file", 1);

    po::variables_map vm;
    po::store(
	po::command_line_parser(argc, argv).options(desc).positional(p_desc).run(),
	vm);
    po::notify(vm);

    if (vm.count("help") != 0) {
	cout << "Usage: " << argv[0] << " [OPTION]... [IN FILE] [OUT FILE]" << endl
	     << endl
	     << "Encapsulates data inside an Freja stream container." << endl
	     << endl;
	cout << desc << endl;
	exit(0);
    }

    if (conf.in_file == "-")
	conf.in_file = "";
    if (conf.out_file == "-")
	conf.out_file = "";

    if (vm.count("legacy")) {
	conf.legacy = true;
#ifdef ENABLE_FNR_GZ
	conf.compression = STREAM_CONTAINER_COMPRESSION_DEFLATE;
#else
	cerr << "Warning: zlib compression not enabled at compile " << endl
	     << "time, can't set compression to deflate." << endl;
#endif
    }

    if (vm.count("compression")) {
	if (compression == "none")
	    conf.compression = STREAM_CONTAINER_COMPRESSION_NONE;
#ifdef ENABLE_FNR_GZ
	else if (compression == "deflate")
	    conf.compression = STREAM_CONTAINER_COMPRESSION_DEFLATE;
#endif
#ifdef ENABLE_FNR_BZ2
	else if (compression == "bzip2")
	    conf.compression = STREAM_CONTAINER_COMPRESSION_BZIP2;
#endif
	else {
	    cerr << "Error: illegal compression specified." << endl
		 << "Available compressions are:" << endl
		 << "  none    - No compression" << endl
#ifdef ENABLE_FNR_GZ
		 << "  deflate - Zlib deflate" << endl;
#endif
#ifdef ENABLE_FNR_BZ2
	    cout << "  bzip2 - bzip2 (experimental)" << endl;
#endif

	    exit(1);
	}
    }

    if (vm.count("product")) {
	if (product == "vpe")
	    conf.product = STREAM_CONTAINER_PRODUCT_VPE;
	else if (product == "vpemp")
	    conf.product = STREAM_CONTAINER_PRODUCT_VPEMP;
	else if (product == "litmus")
	    conf.product = STREAM_CONTAINER_PRODUCT_LITMUS;
	else if (product == "ss12")
	    conf.product = STREAM_CONTAINER_PRODUCT_SS12;
	else if (product == "internal")
	    conf.product = STREAM_CONTAINER_PRODUCT_INTERNAL;
	else {
	    cerr << "Error: illegal product specified." << endl
		 << "Available products are:" << endl
		 << "  vpe      - Acumem VPE" << endl
		 << "  vpemp    - Acumem VPE MP" << endl
		 << "  litmus   - Acumem Litmus" << endl
		 << "  ss12     - Acumem SlowSpotter 1.2" << endl
		 << "  internal - For internal use only" << endl;
	    exit(1);
	}
    }

    if (conf.legacy &&
	conf.compression != STREAM_CONTAINER_COMPRESSION_DEFLATE)
	cerr << "Warning: Legacy specified, but compression is not 'deflate', " << endl
	     << "output file will will not be a valid Fenris file." << endl;
}

int main(int argc, char *argv[])
{
    using namespace acumem::library;
    conf_t conf;

    parse_args(conf, argc, argv);

    try {
        warning_consumer_ptr_t warning_consumer(
            new warning_consumer_stderr_t());

	/* Setup header */
	stream_container_header_t header;
	memset(&header, 0, sizeof(stream_container_header_t));
	header.version = conf.legacy ? 0 : STREAM_CONTAINER_VERSION;
	header.product = conf.product;
	header.data_type = STREAM_CONTAINER_DATA_SAMPLES_BINARY;
	header.compression = conf.compression;

	stream_file_writer_t out(
	    conf.out_file == "" ? NULL : conf.out_file.c_str());
	stream_container_writer_t container(header, &out, warning_consumer);
    	stream_file_reader_t in(conf.in_file == "" ? NULL : conf.in_file.c_str(),
				&container);

	in.dispatch();
    } catch (stream_exception_t se) {
	cerr << "Caught stream exception:" << endl
	     << se.what() << endl;
	exit(1);
    }

    return 0;
}
