/// @file tools/env_helper/env_helper.c

// Copyright (c) 2007, 2009 Acumem AB.  All rights reserved.

/* When running a program in an xterm the LD_LIBRARY_PATH environment variable
   may be cleared, because xterm is usually installed as setuid. This program
   is run as a wrapper around the program to restore LD_LIBRARY_PATH. It sets
   LD_LIBRARY_PATH to the argv[1] and then does an execvp on &argv[2]. */

/* Since the Solaris xterm version doesn't support -hold, we have to
   wait here until the xterm is closed. */

#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <assert.h>

int
main(int argc, char ** argv)
{
    if (argc < 3) {
        fprintf(stderr, "env_helper: too few arguments\n");
        return 1;
    }

    if (strcmp("(null)", argv[1]) == 0) {
        if (unsetenv("LD_LIBRARY_PATH") != 0) {
            fprintf(stderr, "env_helper: failed to unset LD_LIBRARY_PATH\n");
            return 1;
        }
    }
    else
        if (setenv("LD_LIBRARY_PATH", argv[1], 1) != 0) {
            fprintf(stderr, "env_helper: failed to set LD_LIBRARY_PATH\n");
            return 1;
        }

    struct sigaction sa_old;
    struct sigaction sa_new = {
	.sa_handler = SIG_IGN,
	.sa_flags = 0
    };
    sigemptyset(&sa_new.sa_mask);

    int ret = sigaction(SIGINT, &sa_new, &sa_old);
    assert(ret == 0);

    ret = fork();
    if (ret == -1) {
	fprintf(stderr, "env_helper: failed to fork, %s\n",
		strerror(errno));
    } else if (ret == 0) {
	ret = sigaction(SIGINT, &sa_old, NULL);
	assert(ret == 0);

	execvp(argv[2], &argv[2]);
	fprintf(stderr, "env_helper: failed to execute '%s', %s\n",
		argv[2], strerror(errno));
    } else {
	int status;
	if(waitpid(ret, &status, 0) == -1) {
	    fprintf(stderr, "env_helper: failed to wait for child, '%s'\n",
		    strerror(errno));
	} else if(WIFEXITED(status)) {
	    fprintf(stderr, "%s exited with status %i.\n",
		    argv[2], WEXITSTATUS(status));
	} else if(WIFSIGNALED(status)) {
	    fprintf(stderr, "%s caught an unhandled signal: %i.\n",
		    argv[2], WTERMSIG(status));
	} else {
	    fprintf(stderr, "env_helper: unhandled status from wait.\n");
	}
    }

    ret = sigaction(SIGINT, &sa_old, NULL);
    assert(ret == 0);

    fprintf(stderr, "You may now close the xterm.\n");
    while (1)
	sleep(100);
}
