/// @file tools/mapping_remover/mapping_remover.c

// Copyright (c) 2011 Acumem AB.  All rights reserved.

#define PRODUCT NO_PRODUCT

#include "product.h"
#include "library/fenris/flow_file_writer.hh"
#include "library/fenris/flow_file_reader.hh"
#include "library/stream/stream_container.hh"
#include "library/warning/warning_consumer.hh"
#include "library/lang/am_assert.hh"

#include <boost/program_options.hpp>
#include <boost/foreach.hpp>

#include <iostream>

namespace po = boost::program_options;
using namespace acumem::library;
using namespace std;

struct conf_t {
    string in_filename;
    string out_filename;
    vector<string> exceptions;
};

static conf_t
parse_args(int argc, char **argv)
{
    conf_t conf;

    po::options_description opt_desc_visible("Options");
    opt_desc_visible.add_options()
        ("help,h", "Print this help message.")
        ("in-file,i", po::value<string>(&conf.in_filename), "Input file.")
        ("out-file,o", po::value<string>(&conf.out_filename), "Output file.")
        ("exception,e", po::value<vector<string> >(&conf.exceptions))
        ;

    po::options_description opt_desc_invisible("Options");
    opt_desc_invisible.add_options()
        ("positional-args", "")
        ;

    po::options_description opt_desc_all("Options");
    opt_desc_all.add(opt_desc_visible).add(opt_desc_invisible);

    po::positional_options_description pos_desc;
    pos_desc.add("positional-args", -1);

    po::variables_map vm;
    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(opt_desc_all).positional(pos_desc).run(), vm);
        po::notify(vm);
    }
    catch (po::unknown_option &e) {
        cerr << "Error: " << e.what() << endl;
        cerr << opt_desc_visible << endl;
        exit(1);
    }

    if (vm.count("help") > 0) {
        cerr << opt_desc_visible << endl;
        exit(0);
    }

    if (vm.count("in-file") > 1) {
        cerr << "Error: multiple input files specified" << endl;
        exit(1);
    }

    if (vm.count("out-file") > 1) {
        cerr << "Error: multiple output files specified" << endl;
        exit(1);
    }

    if (vm.count("positional-args") > 0) {
        cerr << "Error: no positional arguments allowed" << endl;
        exit(1);
    }

    return conf;
}

class mapping_remover_flow_t
    : public fenris_flow_filter_t
{
    set<string> exceptions;

public:
    mapping_remover_flow_t(fenris_flow_t *next,
                           const vector<string> &exceptions)
        : fenris_flow_filter_t(next)
        {
            BOOST_FOREACH(const string &e, exceptions)
                this->exceptions.insert(e);
        }

    flow_status_t mmap(const fnr_event_mmap_t &event) {
        if (exceptions.find(event.filename) == exceptions.end())
            return FLOW_OK;
        else
            return fenris_flow_filter_t::flow_to_next(event);
    }
};

int
main(int argc, char **argv)
{
    try {
        conf_t conf(parse_args(argc, argv));

        warning_consumer_ptr_t warning_consumer(
            new warning_consumer_stderr_t());

        stream_container_header_t header;
        memset(&header, 0, sizeof(stream_container_header_t));
        header.version = STREAM_CONTAINER_VERSION;
        header.product = STREAM_CONTAINER_PRODUCT_SS12;
        header.data_type = STREAM_CONTAINER_DATA_SAMPLES_BINARY;
        header.compression = STREAM_CONTAINER_COMPRESSION_DEFLATE;

        flow_file_writer_t out_flow(
            conf.out_filename.c_str(), header, warning_consumer);

        mapping_remover_flow_t mapping_remover_flow(&out_flow, conf.exceptions);

        flow_file_reader_t in_flow(
            conf.in_filename.c_str(), &mapping_remover_flow, warning_consumer);

        AM_ASSERT(in_flow.dispatch() == STREAM_OK);
    }
    catch (stream_exception_t &e) {
        cerr << e.what() << endl;
    }

    return 0;
}
