/// @file tools/miss_ratio/miss_ratio.cc

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

// Prints miss ratios for a sample file and cache configuration.

#include "analysis/feature_base/exceptions.hh"
#include "analysis/structure/loop_identification.hh"
#include "analysis/thread_mapping/thread_mapping_null.hh"
#include "analysis/thread_mapping/thread_mapping_legacy.hh"
#include "analysis/filter/filter_module.hh"
#include "analysis/fetch_ratio/fr_lookup.hh"
#include "analysis/fetch_ratio/repl_fr_random.hh"
#include "analysis/fetch_ratio/repl_fr_lru.hh"
#include "analysis/utilization/utilization.hh"
#include "analysis/hw_prefetch/hwpf_lookup.hh"
#include "analysis/thread_mapping/thread_mapping_auto.hh"
#include "library/version/version.h"
#include <boost/foreach.hpp>
#include <boost/program_options.hpp>

#include "analysis/fetch_ratio/windowing_adaptive.hh"
#include "analysis/fetch_ratio/windowing_static.hh"
#include "analysis/fetch_ratio/windowing_burst.hh"

#include <list>
#include <iostream>

#define SAMPLES_PER_WINDOW 100

using namespace acumem::analysis;
using namespace acumem::library;
namespace po = boost::program_options;
using namespace boost;
using namespace std;

/* Bring in acumem svn version information */
BRING_IN_VERSIONINFO

struct conf_t {
    conf_t()
        : sample_file(""), line_size(32),
          engine("random"),
          show_loops(false), stack_depth(0),
	  page_size(4096), pos_streams(16), neg_streams(4), prefetch_reach(2),
	  effectiveness(1), hwpf_enabled(true) {}

    string sample_file;
    vector<fnr_cache_size_t> cache_size;
    fnr_line_size_t line_size;
    string engine;
    acumem::analysis::windowing_engine_cptr_t windowing;
    acumem::analysis::windowing_engine_cptr_t windowing_burst;
    bool show_loops;
    fnr_stack_depth_t stack_depth;

    int page_size;
    int pos_streams;
    int neg_streams;
    int prefetch_reach;
    double effectiveness;
    bool use_old_hwpf_algorithm;
    bool hwpf_enabled;
};

static bool
parse_args(conf_t *conf, int argc, char **argv)
{
    string windowing_engine;
    fnr_sample_count_t samples_per_window(SAMPLES_PER_WINDOW);
    fnr_sample_count_t seed_size(SAMPLES_PER_WINDOW);
    fnr_sample_count_t growth_size(SAMPLES_PER_WINDOW / 4);
    double max_delta(0.25);

    vector<string> cache_sizes;

    po::options_description opt_desc_general("General options");
    opt_desc_general.add_options()
        ("help,h", "Print this help message.")
        ("in-file,i", po::value<string>(&conf->sample_file),
         "Sample file. If not specified stdin is used.")
        ("cache-size,c",
	 po::value<vector<string> >(&cache_sizes),
         "Cache size in bytes. If not specified 65536 bytes is used. If "
	 "specified once, that single size is reported. If specified "
	 "twice, a range of sizes are reported. Can be specified using "
	 "k, M, G, T, P suffixes, e.g. 64k")
        ("line-size,l", po::value<fnr_line_size_t>(&conf->line_size),
         "Line size in bytes. If not specified 32 bytes is used.")
        ("repl-alg,r", po::value<string>(&conf->engine),
         "Replacement algorithm. If not specified random replacement is used.")
        ("show-loops", "Show exlusive percentage of all misses per loop.")
        ("stack-depth",
         po::value<fnr_stack_depth_t>(&conf->stack_depth),
         "Call stack depth to use in the analysis.")
        ;

    po::options_description opt_desc_windowing("Windowing options");
    opt_desc_windowing.add_options()
	("windowing", po::value<string>(&windowing_engine),
	 "Window engine to use (static/adaptive/burst).")
	("samples-per-window,w", po::value<fnr_sample_count_t>(&samples_per_window),
	 "Samples per window, only applies to static windowing.")
	("window-seed", po::value<fnr_sample_count_t>(&seed_size),
	 "Size of seed window, applies to adaptive windowing.")
	("window-growth", po::value<fnr_sample_count_t>(&growth_size),
	 "Size of growth window, applies to adaptive windowing.")
	("window-delta", po::value<double>(&max_delta),
	 "Maximum growth window delta, applies to adaptive windowing.")
	("new-algorithm", "Use new hwpf algorithm")
        ;

    po::options_description opt_desc_invisible("Options");
    opt_desc_invisible.add_options()
        ("positional-args", "")
        ;

    po::options_description opt_desc_visible("Options");
    opt_desc_visible.add(opt_desc_general).add(opt_desc_windowing);


    po::options_description opt_desc_all("Options");
    opt_desc_all.add(opt_desc_visible).add(opt_desc_invisible);

    po::positional_options_description pos_desc;
    pos_desc.add("positional-args", -1);

    po::variables_map vm;
    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(opt_desc_all).positional(pos_desc).run(), vm);
        po::notify(vm);
    }
    catch (po::error &e) {
        cerr << "Error: " << e.what() << endl;
        cerr << opt_desc_visible << endl;
        return true;
    }

    bool error = false;

    if (vm.count("help") > 0)
        cerr << opt_desc_visible << endl;

    if (vm.count("show-loops") > 0)
        conf->show_loops = true;

    if (vm.count("cache-size") == 0)
	conf->cache_size.push_back(65536);
    else if (cache_sizes.size() > 2) {
	cerr << "Error: -s may be given 0, 1 or 2 times" << endl;
	error = true;
    } else if (cache_sizes.size() == 2) {
	string last_s = cache_sizes.back();
	string first_s = cache_sizes.front();

	fnr_cache_size_t last = fnr_parse_num_string(last_s.c_str());
	fnr_cache_size_t first = fnr_parse_num_string(first_s.c_str());

	if (last < first) {
	    swap(last, first);
	}
	conf->cache_size.clear();
	conf->cache_size.push_back(first);

	first *= 2;
	while (first < last) {
	    conf->cache_size.push_back(first);
	    first *= 2;
	}
	conf->cache_size.push_back(last);
    } else if (cache_sizes.size() == 1)
	conf->cache_size.push_back(fnr_parse_num_string(cache_sizes.front().c_str()));

    if (vm.count("windowing") > 0) {
	if (windowing_engine == "static")
	    conf->windowing.reset(
		new acumem::analysis::windowing_static_t(samples_per_window));
	else if (windowing_engine == "adaptive")
	    conf->windowing.reset(
		new acumem::analysis::windowing_adaptive_t(
		    seed_size, growth_size, max_delta));
	else if (windowing_engine == "burst")
	    conf->windowing.reset(new acumem::analysis::windowing_burst_t());
	else {
	    cerr << "Error: Illegal window engine specified." << endl;
	    error = true;
	}
	conf->windowing_burst = conf->windowing;
    } else {
	conf->windowing.reset(
	    new acumem::analysis::windowing_adaptive_t(
		seed_size, growth_size, max_delta));
	conf->windowing_burst.reset(
	    new acumem::analysis::windowing_burst_t());
    }

    if (vm.count("positional-args") > 0) {
        cerr << "Error: no non-option arguments allowed" << endl;
        error = true;
    }

    conf->use_old_hwpf_algorithm = vm.count("new-algorithm") == 0;

    return error;
}

int
main(int argc, char **argv)
{
    conf_t conf;

    if (parse_args(&conf, argc, argv))
        return 1;

    feature_context_ptr_t context(new feature_context_t());

    try {
	parser_result_cptr_t parser_result =
	    parser_t(context, conf.sample_file).get();
	stack_pc_map_result_cptr_t stack_pc_map =
	    stack_pc_map_t(context, conf.sample_file, conf.stack_depth).get();
	thread_mapping_cptr_t thread_mapping(
	    new thread_mapping_legacy_t(context, conf.sample_file));
	windowing_engine_cptr_t windowing(parser_result->burst_sampled ?
					  conf.windowing_burst :
					  conf.windowing);
	locationx_mapper_cptr_t lxm_global(
	    new locationx_mapper_t(context,
				   conf.sample_file,
				   conf.stack_depth,
				   thread_mapping,
				   false, false, false));
	locationx_mapper_cptr_t lxm(
	    new locationx_mapper_t(context,
				   conf.sample_file,
				   conf.stack_depth,
				   thread_mapping,
				   true, false, false));
	locationx_mapper_result_cptr_t lxm_global_res(lxm_global->get());
	locationx_mapper_result_cptr_t lxm_res(lxm->get());
	locationx_t lx_global(0);

	BOOST_FOREACH(fnr_cache_size_t cache_size, conf.cache_size) {
	    repl_fr_engine_cptr_t engine;
	    cout << "replacement: " << conf.engine
		 << ", cache_size: " << cache_size
		 << ", line_size: " << conf.line_size << endl;
	    cout << endl;

	    if (conf.engine == "random")
		engine.reset(
		    new repl_fr_random_t(context,
					 conf.sample_file,
					 conf.stack_depth,
					 thread_mapping,
					 windowing,
					 cpu_model_t()));
	    else if (conf.engine == "lru")
		engine.reset(
		    new repl_fr_lru_t(context,
				      conf.sample_file,
				      conf.stack_depth,
				      thread_mapping,
				      windowing,
				      cpu_model_t()));
	    else
		assert(0);

            // Get hardware prefetch ratio.
            hwpf_lookup_result_cptr_t hwpf_res = hwpf_lookup_t(
                context, conf.sample_file, conf.stack_depth, conf.line_size,
                conf.page_size, thread_mapping, cpu_model_t(), lxm_global,
		cache_size, engine, windowing, conf.pos_streams,
		conf.neg_streams, conf.prefetch_reach, false,
		conf.use_old_hwpf_algorithm, conf.hwpf_enabled).get();

            // Spatial usage for this cache size
            utilization_result_cptr_t su_res = utilization_t(
                context, conf.sample_file, conf.stack_depth, conf.line_size,
		conf.page_size, thread_mapping, cpu_model_t(), lxm_global,
		cache_size, windowing, Utilization_Classic).get();

	    // Get fr data for the current cache size
	    fr_lookup_result_cptr_t frl_rd_res =
		fr_lookup_t(context,
			    conf.sample_file,
			    conf.stack_depth,
			    conf.line_size,
			    conf.page_size,
			    thread_mapping,
			    cpu_model_t(),
			    lxm_global,
			    cache_size,
			    engine,
			    FRL_Type_Read_No_Prefetch).get();

	    fr_lookup_result_cptr_t frl_wr_res =
		fr_lookup_t(context,
			    conf.sample_file,
			    conf.stack_depth,
			    conf.line_size,
			    conf.page_size,
			    thread_mapping,
			    cpu_model_t(),
			    lxm_global,
			    cache_size,
			    engine,
			    FRL_Type_Write).get();

	    fr_lookup_result_cptr_t frl_pf_res =
		fr_lookup_t(context,
			    conf.sample_file,
			    conf.stack_depth,
			    conf.line_size,
			    conf.page_size,
			    thread_mapping,
			    cpu_model_t(),
			    lxm_global,
			    cache_size,
			    engine,
			    FRL_Type_Prefetch).get();

	    fr_lookup_result_cptr_t frl_rd_wr_res =
		fr_lookup_t(context,
			    conf.sample_file,
			    conf.stack_depth,
			    conf.line_size,
			    conf.page_size,
			    thread_mapping,
			    cpu_model_t(),
			    lxm_global,
			    cache_size,
			    engine,
			    FRL_Type_Stalls).get();

	    fr_data_t global_rd_wr = frl_rd_wr_res->fr_by_lx(lx_global);
	    double repl_fr = global_rd_wr.by_kind().replacement_fetch_ratio();

	    // Fraction of fetches that are because of hw prefetching
	    double pf_frac = hwpf_res->prefetch_ratio(lx_global);
	    double prefetch_rate = repl_fr * pf_frac;

	    // fr data per access type
	    fr_data_t global_rd = frl_rd_res->fr_by_lx(lx_global);
	    fr_data_t global_wr = frl_wr_res->fr_by_lx(lx_global);
	    fr_data_t global_pf = frl_pf_res->fr_by_lx(lx_global);

	    cout << "  rw samples:                  "
                 << global_rd_wr.samples() << endl;
	    cout << "  rw total number of accesses: "
                 << global_rd_wr.accesses() << endl;
	    cout << "  rw repl+cold miss rate:      "
                 << global_rd_wr.by_kind().total_fetch_ratio() << endl;
	    cout << "  rw repl miss rate:           "
                 << global_rd_wr.by_kind().replacement_fetch_ratio() << endl;
	    cout << "  rw cold miss rate:           "
                 << global_rd_wr.by_kind().cold_fetch_ratio() << endl << endl;
	    cout << "  rd repl+cold miss rate:      "
                 << (global_rd.by_kind().total_fetches() /
                     global_rd_wr.accesses()) << endl;
	    cout << "  rd repl miss rate:           "
                 << (global_rd.by_kind().replacement_fetches() /
                     global_rd_wr.accesses()) << endl;
	    cout << "  rd cold miss rate:           "
                 << (global_rd.by_kind().cold_fetches() /
                     global_rd_wr.accesses()) << endl << endl;
	    cout << "  wr repl+cold miss rate:      "
                 << (global_wr.by_kind().total_fetches() /
                     global_rd_wr.accesses()) << endl;
	    cout << "  wr repl miss rate:           "
                 << (global_wr.by_kind().replacement_fetches() /
                     global_rd_wr.accesses()) << endl;
	    cout << "  wr cold miss rate:           "
                 << (global_wr.by_kind().cold_fetches() /
                     global_rd_wr.accesses()) << endl << endl;
	    cout << "  pf samples:                  "
                 << global_pf.samples() << endl;
	    cout << "  pf total number of accesses: "
                 << global_pf.accesses() << endl;
	    cout << "  pf repl+cold miss rate:      "
                 << global_pf.by_kind().total_fetches() << endl;
	    cout << "  pf repl miss rate:           "
                 << global_pf.by_kind().replacement_fetches() << endl;
	    cout << "  pf cold miss rate:           "
                 << global_pf.by_kind().cold_fetches() << endl << endl;
	    cout << "  prefetch ratio:              "
                 << prefetch_rate << endl;
	    cout << "  spatial usage:               "
                 << su_res->get_utilization(lx_global).value() << endl;
	    cout << "  su - no-cold:                "
                 << su_res->get_utilization_no_cold(lx_global).value() << endl;
	    cout << "  wasted space:                "
                 << su_res->get_wasted_space(lx_global).value() << endl;

	    if (conf.show_loops) {
		cout << "  Loops (% of all fetches excl):" << endl;

		double all_fetches = global_rd_wr.by_kind().total_fetches();
		double total_loop_fetches = 0.0;

		loop_identification_result_cptr_t li_res = loop_identification_t(
		    context, conf.sample_file, conf.stack_depth).get();

		fr_lookup_result_cptr_t frl_res =
		    fr_lookup_t(context,
				conf.sample_file,
				conf.stack_depth,
				conf.line_size,
				conf.page_size,
				thread_mapping,
				cpu_model_t(),
				lxm,
				cache_size,
				engine,
				FRL_Type_Stalls).get();

		for (loop_vector_t::const_iterator
			 li = li_res->loops.begin(), le = li_res->loops.end();
		     li != le; li++) {
		    loop_ptr_t loop = *li;
		    fr_data_t loop_fr = frl_res->fr_by_lx_set(
			lxm_res->get_locationx_set(loop->get_locations(false)));
		    double loop_fetches = loop_fr.by_kind().total_fetches();
		    total_loop_fetches += loop_fetches;
		    cout << "\t" << loop->id << ": " << (loop_fetches * 100.0 / all_fetches) << endl;
		}

		cout << "\tTotal: " << (total_loop_fetches * 100.0 / all_fetches) << endl;

		cout << endl;
	    }
	}

    } catch (feature_exception_t fe) {
	cerr << "Feature excetion: " << fe.what() << endl;
	return 1;
    }

    return 0;
}
