/// @file tools/permutations/permutations.cc

// Copyright (c) 2009-2010 Acumem AB.  All rights reserved.

#include "analysis/statistics/statistics.hh"
#include "analysis/fetch_ratio/fr_lookup.hh"
#include "analysis/fetch_ratio/repl_fr_random.hh"
#include "analysis/fetch_ratio/windowing_static.hh"
#include "analysis/locationx/locationx.hh"
#include "analysis/thread_mapping/thread_mapping_explicit.hh"
#include "analysis/feature_base/exceptions.hh"

#include <boost/program_options.hpp>

using namespace acumem::analysis;
using namespace acumem::library;

namespace po = boost::program_options;

using namespace std;

struct conf_t {
    string in_filename;
};

static bool
parse_args(conf_t &conf, int argc, char **argv)
{
    po::options_description opt_desc_visible("Options");
    opt_desc_visible.add_options()
        ("help,h", "Print this help message.")
        ("in-file,i", po::value<string>(&conf.in_filename),
         "Specifies the input file.")
        ;

    po::options_description opt_desc_invisible("Options");
    opt_desc_invisible.add_options()
        ("positional-args", "")
        ;

    po::options_description opt_desc_all("Options");
    opt_desc_all.add(opt_desc_visible).add(opt_desc_invisible);

    po::positional_options_description pos_desc;
    pos_desc.add("positional-args", -1);

    po::variables_map vm;
    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(opt_desc_all).positional(pos_desc).run(), vm);
        po::notify(vm);
    }
    catch (po::unknown_option &e) {
        cerr << "Error: " << e.what() << endl;
        cerr << opt_desc_visible << endl;
        return true;
    }

    if (vm.count("help") > 0) {
        cerr << opt_desc_visible << endl;
        exit(0);
    }

    bool error = false;

    if (vm.count("in-file") == 0) {
        cerr << "Error: no input file specified" << endl;
        error = true;
    }

    if (vm.count("positional-args") > 0) {
        cerr << "Error: no positional arguments allowed" << endl;
        error = true;
    }

    return error;
}

int
main(int argc, char **argv)
{
    conf_t conf;

    if (parse_args(conf, argc, argv))
        return 1;

    try {
        int samples_per_window(200);
        int stack_depth(0);
        int line_size(64);
        int page_size(4096);
        int cache_size(2048 * 1024);
        int pos_streams(16);
        int neg_streams(8);
        int prefetch_reach(2);
        bool use_location = false;
        bool use_thread = false;
        bool use_cache = false;
        bool use_old_hwpf_algorithm = true;
        bool hwpf_enabled = true;

        vector<tm_cache_map_t> cmv;
        for (int i = 0; i < 0x80; i++) {
            int n = (i & 0x55) + ((i >> 1) & 0x55);
            n = (n & 0x33) + ((n >> 2) & 0x33);
            n = (n & 0x0f) + ((n >> 4) & 0x0f);
            if (n != 4)
                continue;
            tm_cache_map_t cm;
            for (int j = 0; j < 8; j++)
                cm[j] = (i >> j) & 1;
            cmv.push_back(cm);
        }

        BOOST_FOREACH(tm_cache_map_t &cm, cmv) {
            cpu_model_t cpu_model;

            feature_context_ptr_t context(
                new feature_context_t());

            windowing_engine_cptr_t windowing(
                new windowing_static_t(samples_per_window));

            thread_mapping_cptr_t thread_mapping(
                new thread_mapping_explicit_t(context,
                                              conf.in_filename,
                                              false,
                                              cm));

            locationx_mapper_cptr_t locationx_mapper(
                new locationx_mapper_t(context,
                                       conf.in_filename,
                                       stack_depth,
                                       thread_mapping,
                                       use_location,
                                       use_thread,
                                       use_cache));

            repl_fr_engine_cptr_t engine(
                new repl_fr_random_t(context,
                                     conf.in_filename,
                                     stack_depth,
                                     thread_mapping,
                                     windowing,
                                     cpu_model));

            statistics_result_cptr_t statistics_result =
                statistics_t(context,
                             conf.in_filename,
                             stack_depth,
                             line_size,
                             thread_mapping,
                             cpu_model,
                             locationx_mapper,
                             cache_size,
                             engine,
                             windowing,
                             page_size,
                             pos_streams,
                             neg_streams,
                             prefetch_reach,
                             use_old_hwpf_algorithm,
                             hwpf_enabled).get();

            locationx_mapper_result_cptr_t locationx_mapping =
                locationx_mapper->get();

            locationx_t all_locationx =
                locationx_mapping->get_locationx(
                    locationx_specifier_t(LOCATION_INVALID, 0, 0));

            stats_cptr_t all_stats = statistics_result->get(all_locationx, true);

            printf("(");
            BOOST_FOREACH(tm_cache_map_t::value_type &v, cm)
                if (v.second == 0)
                    printf("%d ", v.first.id);
            printf(") (");
            BOOST_FOREACH(tm_cache_map_t::value_type &v, cm)
                if (v.second == 1)
                    printf("%d ", v.first.id);
            printf(") ");

            printf("mr %f, fr %f, ur %f, wr %f\n",
                   all_stats->stall_ratio.total_fetch_ratio(),
                   all_stats->all_ratio.total_fetch_ratio(),
                   all_stats->all_ratio.total_upgrade_ratio(),
                   all_stats->all_ratio.total_write_back_ratio());

            context->clear();
        }

        return 0;
    } catch (feature_exception_t &t) {
        cerr << "Error: " << t.what() << endl;
	return 1;
    }
}
