/// @file tools/pnta/pnta.cc

// Copyright (c) 2010 Acumem AB.  All rights reserved.

#include "analysis/fetch_ratio/repl_fr_line.hh"
#include "analysis/fetch_ratio/repl_fr_random.hh"
#include "analysis/fetch_ratio/windowing_static.hh"
#include "analysis/thread_mapping/thread_mapping_rr.hh"

#include <boost/program_options.hpp>

using namespace acumem::analysis;
using namespace acumem::library;

namespace po = boost::program_options;

using namespace std;

struct conf_t {
    string in_filename;
};

static bool
parse_args(conf_t &conf, int argc, char **argv)
{
    po::options_description opt_desc_visible("Options");
    opt_desc_visible.add_options()
        ("help,h", "Print this help message.")
        ("in-file,i", po::value<string>(&conf.in_filename),
         "Specifies the input file.")
        ;

    po::options_description opt_desc_invisible("Options");
    opt_desc_invisible.add_options()
        ("positional-args", "")
        ;

    po::options_description opt_desc_all("Options");
    opt_desc_all.add(opt_desc_visible).add(opt_desc_invisible);

    po::positional_options_description pos_desc;
    pos_desc.add("positional-args", -1);

    po::variables_map vm;
    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(opt_desc_all).positional(pos_desc).run(), vm);
        po::notify(vm);
    }
    catch (po::unknown_option &e) {
        cerr << "Error: " << e.what() << endl;
        cerr << opt_desc_visible << endl;
        return true;
    }

    if (vm.count("help") > 0) {
        cerr << opt_desc_visible << endl;
        exit(0);
    }

    if (vm.count("in-file") == 0) {
        cerr << "Error: no input file specified" << endl;
        return true;
    }

    if (vm.count("positional-args") > 0) {
        cerr << "Error: no positional arguments allowed" << endl;
        return true;
    }

    return false;
}

class af_t {
public:
    double accesses;
    double fetches;

    af_t()
        : accesses(0.0),
          fetches(0.0)
        {}
};

int
main(int argc, char **argv)
{
    conf_t conf;

    if (parse_args(conf, argc, argv))
        return 1;

    int samples_per_window(200);
    int stack_depth(0);
    fnr_line_size_t line_size(64);
    fnr_line_size_t page_size(4096);
    fnr_cache_size_t l1_cache_size(64 * 1024);
    fnr_cache_size_t l3_cache_size(1024 * 1024);

    feature_context_ptr_t context(
        new feature_context_t());

    cpu_model_t cpu_model;

    windowing_engine_cptr_t windowing(
        new windowing_static_t(samples_per_window));

    thread_mapping_cptr_t thread_mapping(
        new thread_mapping_rr_t(context,
                                conf.in_filename,
                                1));

    sample_mapping_result_cptr_t sm_res(
        sample_mapping_t(context,
                         conf.in_filename,
                         stack_depth,
                         line_size,
                         page_size,
                         thread_mapping,
                         cpu_model).get());

    repl_fr_engine_cptr_t engine(
        new repl_fr_random_t(context,
                             conf.in_filename,
                             stack_depth,
                             thread_mapping,
                             windowing,
                             cpu_model));

    repl_fr_line_result_cptr_t l1_fr_res(
        repl_fr_line_t(context,
                       conf.in_filename,
                       stack_depth,
                       line_size,
                       page_size,
                       l1_cache_size,
                       thread_mapping,
                       cpu_model,
                       engine).get());

    repl_fr_line_result_cptr_t l3_fr_res(
        repl_fr_line_t(context,
                       conf.in_filename,
                       stack_depth,
                       line_size,
                       page_size,
                       l3_cache_size,
                       thread_mapping,
                       cpu_model,
                       engine).get());

    map<fnr_address_t, double> pc1_l1_fetches;
    map<fnr_address_t, double> pc1_nt_fetches;

    typedef map<fnr_address_t, af_t> pc_af_t;
    typedef map<fnr_address_t, pc_af_t> pc_pc_af_t;

    pc_pc_af_t pc1_pc2_af;

    typedef set<fnr_address_t> pc_set_t;
    typedef vector<pc_set_t> pc_set_vector_t;

    pc_set_t pcs;

    double total_l3_fetches = 0;

    assert(sm_res->bursts.size() == l1_fr_res->bursts.size());
    assert(sm_res->bursts.size() == l3_fr_res->bursts.size());
    for (unsigned int b = 0; b < sm_res->bursts.size(); b++) {
        const mapped_repl_line_vector_t &samples = sm_res->bursts[b].repl_line;
        const double sample_period = sm_res->bursts[b].sample_period;
        const ratio_vector_t &l1_frs = l1_fr_res->bursts[b].val_to_end;
        const ratio_vector_t &l3_frs = l3_fr_res->bursts[b].val_to_end;
        assert(samples.size() == l1_frs.size());
        assert(samples.size() == l3_frs.size());
        for (unsigned int s = 0; s < samples.size(); s++) {
            const mapped_access_t *valid = samples[s].valid();
            if (valid == NULL)
                continue;
            const mapped_access_t *end = samples[s].end();
            assert(end != NULL);

            fnr_address_t pc1 = valid->access->pc;
            fnr_address_t pc2 = end->access->pc;

            double l1_fetches = l1_frs[s] * sample_period;
            double l3_fetches = l3_frs[s] * sample_period;

            pcs.insert(pc1);

            pc1_l1_fetches[pc1] += l1_fetches;
            pc1_nt_fetches[pc1] += l3_fetches;

            af_t &af12 = pc1_pc2_af[pc1][pc2];

            af12.accesses += sample_period;
            af12.fetches += l3_fetches;

            total_l3_fetches += l3_fetches;
        }
    }

#define THRESHOLD 0.1

    typedef map<fnr_address_t, pc_set_t> pc_set_map_t;

    pc_set_map_t connections;
    BOOST_FOREACH(const pc_pc_af_t::value_type &v, pc1_pc2_af) {
        fnr_address_t pc1 = v.first;
        BOOST_FOREACH(const pc_af_t::value_type &w, v.second) {
            fnr_address_t pc2 = w.first;

            if (w.second.fetches / w.second.accesses <= THRESHOLD) {
                connections[pc1].insert(pc2);
                connections[pc2].insert(pc1);
            }
        }
    }

    pc_set_vector_t groups;
    while (!pcs.empty()) {
        pc_set_t group;
        pc_set_t search;

        group.insert(*pcs.begin());
        search.insert(*pcs.begin());
        pcs.erase(pcs.begin());

        while (!search.empty()) {
            fnr_address_t pc = *search.begin();
            search.erase(search.begin());

            BOOST_FOREACH(fnr_address_t npc, connections[pc]) {
                pc_set_t::iterator pcsi = pcs.find(npc);
                if (pcsi == pcs.end())
                    continue;
                group.insert(npc);
                search.insert(npc);
                pcs.erase(pcsi);
            }
        }

        groups.push_back(group);
    }

    BOOST_FOREACH(pc_set_vector_t::value_type &v, groups) {
        double l1_fetches = 0.0;
        double nt_fetches = 0.0;
        BOOST_FOREACH(const pc_set_t::value_type &w, v)
            if (pc1_nt_fetches[w] / pc1_l1_fetches[w] > 0.8) {
                l1_fetches += pc1_l1_fetches[w];
                nt_fetches += pc1_nt_fetches[w];
            }
        if (nt_fetches / total_l3_fetches < 0.01)
            continue;
#if 0
        BOOST_FOREACH(const pc_set_t::value_type &w, v)
            printf("0x%lx ", w.virtual_address);
        printf(" -> ");
#endif
        BOOST_FOREACH(const pc_set_t::value_type &w, v)
            if (pc1_nt_fetches[w] / pc1_l1_fetches[w] > 0.8)
                printf("0x%llx ", (long long)w.virtual_address);
        printf(" (%f%% of fetches, %f%% fetch ratio)\n", 100.0 * nt_fetches / total_l3_fetches, 100.0 * nt_fetches / l1_fetches);
    }

    return 0;
}
