/// @file tools/report/report.cc

// Copyright (c) 2011 Acumem AB.  All rights reserved.

/*
 *  This file is part of ThreadSpotter.
 *
 *  ThreadSpotter is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ThreadSpotter is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with ThreadSpotter.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "report_feedback.hh"
#include "analysis/report_generator/conf.hh"
#include "analysis/report_generator/report_generator.hh"
#include "analysis/report_codec/report_codec.hh"
#include "analysis/feature_base/exceptions.hh"
#include "library/stream/stream_file.hh"
#include "library/license/license.hh"
#include "library/version/product_version.h"
#include "library/version/copyright.h"
#include "library/version/version.h"
#include "product.h"

#include <exception>

#if defined(__linux__) || defined(__sun__)
#include <sys/resource.h>
#endif

using namespace acumem::report;
using namespace acumem::analysis;
using namespace acumem::analysis::report_generator;

BRING_IN_VERSIONINFO

int
main(int argc, char **argv)
{
    try {
        warning_consumer_ptr_t wc(new warning_consumer_stderr_t());
        license_manager_ptr_t license(new license_manager_t(wc));

	feature_set_ptr_t fs = license->acquire(license_manager_t::tool_kind);

	if (!fs) {
	    cerr << "No license token available." << endl;
	    exit(1);
	}

	if (!fs->cmd_line()) {
	    cerr << "Your license does not include command line tools." << endl;
	    return false;
	}

        workspace_ptr_t workspace(new workspace_t());
	feedback_ptr_t feedback(new report_feedback_t(license));
        feature_context_ptr_t context(
            new feature_context_t(workspace, wc, feedback));

        bool exit;
        int exit_code;
        conf_t conf(exit, exit_code,
                    wc, feedback,
                    argc, argv,
                    fs->mt(),
                    fs->multisocket(),
                    fs->unlimited_callstack(),
                    fs->blocking_expert(),
		    fs->thread_binding(),
                    PRODUCT_VERSION "\n" COPYRIGHT);

        if (exit)
            return exit_code;

        rpt_report_cptr_t report(
            generate_report(context,
			    conf,
			    fs->mt(),
			    fs->sw_prefetch(),
			    fs->unlimited_callstack(),
			    fs->blocking_expert() && conf.blocking_expert,
			    fs->multisocket()));

        // Return false if the report generation failed.
        if (!report)
            return 1;

        // Write the report to disk.
        stream_file_writer_t out_file(conf.report_file.c_str());
        stream_container_header_t out_header(
            STREAM_CONTAINER_VERSION,
            PRODUCT_REPORT_PRODUCT,
            STREAM_CONTAINER_DATA_REPORT,
            STREAM_CONTAINER_COMPRESSION_DEFLATE);
        stream_container_writer_t out_container(
            out_header, &out_file, wc);
        report_encoder_t out_encoder(out_container);
        out_encoder.encode(report);

    } catch (const feature_exception_t &t) {
        cerr << "Error: " << t.what() << endl;
	return 1;

    } catch (const stream_exception_t &se) {
        cerr << "Error: " << se.what() << endl;
	return 1;

    } catch (const bad_alloc &) {
        cerr << "Error: Out of memory." << endl;
#if defined(__linux__) || defined(__sun__)
        struct rlimit rlim;
        int res = getrlimit(RLIMIT_AS, &rlim);
        assert(res == 0);
        if (rlim.rlim_cur != (rlim_t)RLIM_INFINITY ||
            rlim.rlim_max != (rlim_t)RLIM_INFINITY)
            cerr << endl
                 << "A resource limit has been set on the available virtual "
                 << "address space. You may want to try to increase or remove "
                 << "it, e.g., \"ulimit -S -v unlimited\"." << endl;
#endif
        return 1;

    } catch (const std::exception &re) {
	cerr << re.what() << endl;
	return 1;
    }

    return 0;
}
