

/*
 *  This file is part of ThreadSpotter.
 *
 *  ThreadSpotter is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ThreadSpotter is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with ThreadSpotter.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "library/exe_path/exe_path.hh"
#include "library/version/product_version.h"
#include "library/version/copyright.h"
#include "library/version/version.h"
#include "library/am_mem/am_mem.h"
#include "library/lang/am_assert.hh"

#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>

#include <sys/wait.h>
#include <stdlib.h>

BRING_IN_VERSIONINFO;

using namespace boost;
using namespace std;

namespace po = boost::program_options;
namespace fs = boost::filesystem;

// Burst sampling parameters.
static const int64_t num_bursts = 100;
static const int64_t burst_sampling = 100000000;
static int64_t burst_cooldown = 1000000000;

struct conf_t {
    conf_t()
        : launch(false),
          attach(false),
          sample_file("sample.smp"),
          sample_count("50000"),
          sample_period("100000"),
          line_sizes("64"),
          no_decimate(false),
          fixed_sample_period(false),
          burst_runtime(0),
          sampling_delay(NULL),
          sampling_duration(NULL),
          start_at_function(NULL),
          start_at_address(NULL),
          start_at_ignore(NULL),
          stop_at_function(NULL),
          stop_at_address(NULL),
          stop_at_ignore(NULL),
          preallocation(NULL),
          safe_stack(false),
          use_target_stderr(false),
          unsupported_attach(false),
	  sample_generation(0)
        {}

    bool launch;
    vector<string> launch_args;

    bool attach;
    string attach_pid;

    fs::path tmpdir;
    fs::path sample_file;

    string sample_count;
    string sample_period;
    string line_sizes;
    bool no_decimate;
    bool fixed_sample_period;

    /// Non-0 if burst sampling, 0 if not.
    int64_t burst_runtime;

    // These contain the option value if specified, or NULL if not specified.

    char *sampling_delay;
    char *sampling_duration;

    char *start_at_function;
    char *start_at_address;
    char *start_at_ignore;
    char *stop_at_function;
    char *stop_at_address;
    char *stop_at_ignore;

    char *preallocation;

    bool safe_stack;
    bool use_target_stderr;
    bool unsupported_attach;

    int sample_generation;
};

enum parse_args_result_t {
    Parse_Args_Ok,
    Parse_Args_Exit_Ok,
    Parse_Args_Exit_Error
};

static parse_args_result_t
parse_args(int argc, char **argv, conf_t &conf)
{
    // Replace the first -r we find in the command line with -- so that the
    // rest of the command line is interpreted as positional arguments, and set
    // the launch configuration variable.
    char **arg = argv;
    while (*arg != NULL)
        if (strcmp(*(arg++), "-r") == 0) {
            conf.launch = true;
            arg[-1][1] = '-';
            break;
        }

    string burst_runtime;
    string burst_quality;

    string sampling_delay;
    string sampling_duration;

    string start_at_function;
    string start_at_address;
    string start_at_ignore;
    string stop_at_function;
    string stop_at_address;
    string stop_at_ignore;

    string preallocation;

    conf.tmpdir = acumem::exe_path::get_temp_path();

    po::options_description opt_desc_fundamental("Basic options");
    opt_desc_fundamental.add_options()
        ("version", "Print version information.")

        ("help", "Print this help message.")

        ("run,r",
	 "Start and sample app with args. Must be the last argument.")

        ("pid,p", po::value<string>(&conf.attach_pid),
         "Sample the already running process with the specified pid.")

        ("sample-file,o", po::value<fs::path>(&conf.sample_file),
         "Sample file name, sample.smp by default. Path subject to "
	 "substitutions:\n"
	 "  %r     \tunique value per rank\n"
	 "  %h     \thostname\n"
	 "  %p     \tprocess id\n"
	 "  %{ENV} \tcontents of environment variable ENV\n")

        ("tmpdir", po::value<fs::path>(&conf.tmpdir),
	 ("Store temporary files in this directory instead of the system "
	  "default temp directory or the one pointed out by the environment "
	  "variable THREADSPOTTER_TMPDIR"))
	;


    po::options_description opt_desc_sampler("Sampler control options");
    opt_desc_sampler.add_options()
        ("sample-count,c", po::value<string>(&conf.sample_count),
         "Number of samples to collect. Default is 50000.")

        ("sample-period,s", po::value<string>(&conf.sample_period),
         "Sample period in number of memory accesses. Default is 100000.")

        ("line-sizes,l", po::value<string>(&conf.line_sizes),
         "Line sizes to sample separated by commas, e.g., 32,64,128 to sample "
         "both 32, 64 and 128 byte line sizes. Default is 64-byte line size.")

        ("no-decimate,n", "Do not decimate the sample file.")

        ("fixed-sample-period,f", "Use fixed sample period instead of "
         "automatic sample period adjustment.")

        ("burst,b", po::value<string>(&burst_runtime),
         "Enable burst sampling, assuming the specified estimated "
         "uninstrumented run in minutes.")

        ("burst-quality,q", po::value<string>(&burst_quality),
         "Burst sampling quality setting, valid values are 'fast', 'normal' "
         "and 'detailed'.")

        ("sampling-delay,d", po::value<string>(&sampling_delay),
         "Start sampling after the specified number of seconds.")

        ("sampling-duration,t", po::value<string>(&sampling_duration),
         "Stop sampling after the specified number of seconds.")

        ("start-at-function", po::value<string>(&start_at_function),
         "Start sampling when the specified function is executed.")

        ("start-at-address", po::value<string>(&start_at_address),
         "Start sampling when the instruction at the specified address is "
         "executed.")

        ("start-at-ignore", po::value<string>(&start_at_ignore),
         "Ignore the specified first number times the sampling start "
         "location, set with --start-at-function or --start-at-address, is "
         "reached.")

        ("stop-at-function", po::value<string>(&stop_at_function),
         "Stop sampling when the specified function is executed.")

        ("stop-at-address", po::value<string>(&stop_at_address),
         "Stop sampling when the instruction at the specified address is "
         "executed.")

        ("stop-at-ignore", po::value<string>(&stop_at_ignore),
         "Ignore the specified first number times the sampling stop location, "
         "set with --stop-at-function or --stop-at-address, is reached.")

        ("preallocation", po::value<string>(&preallocation),
         "Reduces the the amount of virtual memory reserved to the sampler to "
         "the specified number of megabytes. Only available on 64-bit "
         "installations.")

        ("safe-stack", "Disable optimization that may cause problems with "
         "programs with non-standard stack handling.")

        ("use-target-stderr", "Allow the sampler to log error messages and "
         "warnings to the target application's stderr.")

        ("unsupported-attach", "Allows sampling of other users applications "
         "when runnig as root, but also opens security holes.")

	("generation,g", po::value<int>(&conf.sample_generation),
	 "Whether to sample the invoked process (=0) or its children (=1) "
	 "etc. In some MPI environments the target application is "
	 "bootstrapped using an intermediate process which should not be "
	 "sampled.")
	;

    po::options_description opt_desc_invisible("Options");
    opt_desc_invisible.add_options()
        ("launch-args", po::value<vector<string> >(&conf.launch_args), "");

    po::options_description opt_desc_all("Options");
    opt_desc_all
	.add(opt_desc_fundamental)
	.add(opt_desc_sampler)
	.add(opt_desc_invisible)
	;

    po::positional_options_description opt_desc_pos;
    opt_desc_pos.add("launch-args", -1);

    po::variables_map vm;
    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(opt_desc_all).positional(opt_desc_pos).run(), vm);
        po::notify(vm);
    }
    catch (po::error &e) {
        fprintf(stderr, "Error parsing command line: %s\n", e.what());
        return Parse_Args_Exit_Error;
    }

    if (vm.count("help") > 0) {
        stringstream ss;
        ss << opt_desc_fundamental;
	ss << endl;
	ss << opt_desc_sampler;
        printf("%s\n%s\n", PRODUCT_VERSION, COPYRIGHT);
        printf("Usage: sample [OPTION]... [app args...]\n%s",
               ss.str().c_str());
        return Parse_Args_Exit_Ok;
    }

    if (vm.count("version") > 0) {
        printf("%s\n%s\n", PRODUCT_VERSION, COPYRIGHT);
        return Parse_Args_Exit_Ok;
    }

    conf.attach = vm.count("pid") > 0;
    bool sample_count = vm.count("sample-count") > 0;
    bool sample_period = vm.count("sample-period") > 0;
    conf.no_decimate = vm.count("no-decimate") > 0;
    conf.fixed_sample_period = vm.count("fixed-sample-period") > 0;

    if (vm.count("burst") > 0) {
        char *endptr;
        conf.burst_runtime = strtol(burst_runtime.c_str(), &endptr, 10);
        if (*endptr != '\0') {
            fprintf(stderr, "Error: %s is not a decimal integer.",
                    burst_runtime.c_str());
            return Parse_Args_Exit_Error;
        }
        if (conf.burst_runtime < 5) {
            fprintf(stderr, "Error: Burst sampling can not be used for "
                    "runs shorter than 5 minutes.\n");
            return Parse_Args_Exit_Error;
        }
        if (conf.burst_runtime > 1000000000) {
            fprintf(stderr, "Error: Estimated run time is too large.\n");
            return Parse_Args_Exit_Error;
        }
        conf.sample_period = "200000";
    }

    if (vm.count("burst-quality") > 0) {
        if (conf.burst_runtime == 0) {
            fprintf(stderr,
                    "Error: --burst-quality specified without --burst.\n");
            return Parse_Args_Exit_Error;
        }

        if (burst_quality == "fast")
            burst_cooldown /= 2;
        else if (burst_quality == "normal")
            ;
        else if (burst_quality == "detailed")
            burst_cooldown *= 2;
        else {
            fprintf(stderr, "Error: Invalid burst sampling quality level "
                    "'%s'.\n", burst_quality.c_str());
            return Parse_Args_Exit_Error;
        }
    }

    if (vm.count("sampling-delay") > 0)
        conf.sampling_delay = am_strdup(sampling_delay.c_str());
    if (vm.count("sampling-duration") > 0)
        conf.sampling_duration = am_strdup(sampling_duration.c_str());

    if (vm.count("start-at-function") > 0)
        conf.start_at_function = am_strdup(start_at_function.c_str());
    if (vm.count("start-at-address") > 0)
        conf.start_at_address = am_strdup(start_at_address.c_str());
    if (vm.count("start-at-ignore") > 0)
        conf.start_at_ignore = am_strdup(start_at_ignore.c_str());
    if (vm.count("stop-at-function") > 0)
        conf.stop_at_function = am_strdup(stop_at_function.c_str());
    if (vm.count("stop-at-address") > 0)
        conf.stop_at_address = am_strdup(stop_at_address.c_str());
    if (vm.count("stop-at-ignore") > 0)
        conf.stop_at_ignore = am_strdup(stop_at_ignore.c_str());

    if (vm.count("preallocation") > 0)
        conf.preallocation = am_strdup(preallocation.c_str());

    conf.safe_stack = vm.count("safe-stack") > 0;
    conf.use_target_stderr = vm.count("use-target-stderr") > 0;
    conf.unsupported_attach = vm.count("unsupported-attach") > 0;

    if ((conf.sampling_delay != NULL ? 1 : 0) +
        (conf.start_at_function != NULL ? 1 : 0) +
        (conf.start_at_address != NULL ? 1 : 0) > 1) {
        fprintf(stderr,  "Error: Only one of --sampling-delay, "
                "--start-at-function and --start-at-address may be "
                "specified.\n");
        return Parse_Args_Exit_Error;
    }

    if (conf.start_at_ignore != NULL &&
        conf.start_at_function == NULL &&
        conf.start_at_address == NULL) {
        fprintf(stderr,  "Error: --start-at-ignore specified without "
                "--start-at-function or --start-at-address.\n");
        return Parse_Args_Exit_Error;
    }

    if ((conf.sampling_duration != NULL ? 1 : 0) +
        (conf.stop_at_function != NULL ? 1 : 0) +
        (conf.stop_at_address != NULL ? 1 : 0) > 1) {
        fprintf(stderr,  "Error: Only one of --sampling-duration, "
                "--stop-at-function and --stop-at-address may be "
                "specified.\n");
        return Parse_Args_Exit_Error;
    }

    if (conf.stop_at_ignore != NULL &&
        conf.stop_at_function == NULL &&
        conf.stop_at_address == NULL) {
        fprintf(stderr,  "Error: --stop-at-ignore specified without "
                "--stop-at-function or --stop-at-address.\n");
        return Parse_Args_Exit_Error;
    }

    if (!conf.launch && !conf.attach) {
        fprintf(stderr,  "Error: Neither --run nor --pid specified.\n");
        return Parse_Args_Exit_Error;
    }

    if (conf.launch && conf.attach) {
        fprintf(stderr,  "Error: Both --run and --pid specified.\n");
        return Parse_Args_Exit_Error;
    }

    if (!conf.launch && !conf.launch_args.empty()) {
        fprintf(stderr,
                "Error: Non-option arguments specified without --run.\n");
        return Parse_Args_Exit_Error;
    }

    if (conf.launch && conf.launch_args.empty()) {
        fprintf(stderr,
                "Error: --run specified without application and arguments.\n");
        return Parse_Args_Exit_Error;
    }

    if (sample_count && conf.no_decimate) {
        fprintf(stderr, "Error: --sample-count and --no-decimate can not be "
                "used at the same time.\n");
        return Parse_Args_Exit_Error;
    }

    if (conf.burst_runtime != 0 && sample_period) {
        fprintf(stderr, "Error: --burst and --sample-period can not be used "
                "at the same time.\n");
        return Parse_Args_Exit_Error;
    }

    return Parse_Args_Ok;
}

static bool
exec_binary(const vector<string> &argv)
{
    pid_t pid = fork();
    if (pid == -1) {
        fprintf(stderr, "Error: fork() failed: error %d, %s.\n",
                errno, strerror(errno));
        return false;
    }

    // TODO: Determine if the exec failed or if acumon failed.
    if (pid == 0) {
        int argc_ = argv.size();
        char *argv_[argc_];
        for (int i = 0; i < argc_; i++)
            argv_[i] = am_strdup(argv[i].c_str());
        argv_[argc_] = NULL;

        execvp(argv_[0], argv_);
        fprintf(stderr, "Error: Failed to execute %s: error %d, %s.\n",
                argv_[0], errno, strerror(errno));
        exit(127);
    }
    else {
        int status;

        signal(SIGINT, SIG_IGN);

        int res = waitpid(pid, &status, 0);

        signal(SIGINT, SIG_DFL);

        if (res != pid) {
            fprintf(stderr, "Error: waitpid() failed: error %d, %s.\n",
                    errno, strerror(errno));
            return false;
        }
        if (status != 0) {
            if (WIFEXITED(status))
                fprintf(stderr, "Error: %s exited with status %d.\n",
                        argv[0].c_str(), WEXITSTATUS(status));
            else if (WIFSIGNALED(status))
                fprintf(stderr, "Error: %s was terminated by signal %d; %s.\n",
                        argv[0].c_str(), WTERMSIG(status),
                        strsignal(WTERMSIG(status)));
            else
                fprintf(stderr, "Error: %s was terminated for unknown reason, "
                        "wait returned %d.\n", argv[0].c_str(), status);
            return false;
        }
        return true;
    }
}

int
main(int argc, char **argv)
{
    conf_t conf;
    switch (parse_args(argc, argv, conf)) {
    case Parse_Args_Ok:
        break;
    case Parse_Args_Exit_Ok:
        return 0;
    case Parse_Args_Exit_Error:
        return 1;
    default:
        AM_ASSERT(0);
    }

    fs::path sample_path;
    try {
	sample_path = acumem::exe_path::get_executable_filename();
    } catch (const std::exception &e) {
	fprintf(stderr, "%s\n", e.what());
	return 2;
    }

    fs::path install_path(fs::path(sample_path).parent_path().parent_path());
    fs::path acumon_path(install_path / "lib" / "threadspotter" / "acumon");
    fs::path atf_resampler_path(
        install_path / "lib" / "threadspotter" / "atf_resampler");

    const string pid_str(lexical_cast<string>(getpid()));
    fs::path atf_path(conf.tmpdir / (string("x") + pid_str + ".atf"));
    fs::path acq_path(conf.tmpdir / (string("x") + pid_str + ".acq"));

    vector<string> acumon_argv;
    acumon_argv.push_back(acumon_path.string());
    acumon_argv.push_back("--atf-file-prefix");
    acumon_argv.push_back(atf_path.string());
    acumon_argv.push_back("--acquisition-info");
    acumon_argv.push_back(acq_path.string());
    if (conf.launch) {
        bool first = true;
        BOOST_FOREACH(const string &arg, conf.launch_args) {
            acumon_argv.push_back(first ? "--binary" : "--argument");
            acumon_argv.push_back(arg);
            first = false;
        }
    }
    if (conf.attach) {
        acumon_argv.push_back("--pid");
        acumon_argv.push_back(conf.attach_pid);
    }
    if (!conf.fixed_sample_period && conf.burst_runtime == 0) {
        acumon_argv.push_back("--sample-count");
        acumon_argv.push_back(conf.sample_count);
    }
    acumon_argv.push_back("--sample-period");
    acumon_argv.push_back(conf.sample_period);
    acumon_argv.push_back("--line-sizes");
    acumon_argv.push_back(conf.line_sizes);
    if (conf.burst_runtime != 0) {
        int64_t burst_time = (burst_sampling + burst_cooldown) / 1000000;
        int64_t burst_interval =
            (conf.burst_runtime * 60000 / num_bursts) - burst_time;
        acumon_argv.push_back("--burst-sampling");
        acumon_argv.push_back(lexical_cast<string>(burst_sampling));
        acumon_argv.push_back("--burst-cooldown");
        acumon_argv.push_back(lexical_cast<string>(burst_cooldown));
        acumon_argv.push_back("--burst-interval");
        acumon_argv.push_back(lexical_cast<string>(burst_interval));
    }
    if (conf.sampling_delay != NULL) {
        acumon_argv.push_back("--sampling-delay");
        acumon_argv.push_back(conf.sampling_delay);
    }
    if (conf.sampling_duration != NULL) {
        acumon_argv.push_back("--sampling-duration");
        acumon_argv.push_back(conf.sampling_duration);
    }
    if (conf.start_at_function != NULL) {
        acumon_argv.push_back("--start-at-function");
        acumon_argv.push_back(conf.start_at_function);
    }
    if (conf.start_at_address != NULL) {
        acumon_argv.push_back("--start-at-address");
        acumon_argv.push_back(conf.start_at_address);
    }
    if (conf.start_at_ignore != NULL) {
        acumon_argv.push_back("--start-at-ignore");
        acumon_argv.push_back(conf.start_at_ignore);
    }
    if (conf.stop_at_function != NULL) {
        acumon_argv.push_back("--stop-at-function");
        acumon_argv.push_back(conf.stop_at_function);
    }
    if (conf.stop_at_address != NULL) {
        acumon_argv.push_back("--stop-at-address");
        acumon_argv.push_back(conf.stop_at_address);
    }
    if (conf.stop_at_ignore != NULL) {
        acumon_argv.push_back("--stop-at-ignore");
        acumon_argv.push_back(conf.stop_at_ignore);
    }
    if (conf.preallocation != NULL) {
        acumon_argv.push_back("--preallocation");
        acumon_argv.push_back(conf.preallocation);
    }
    if (conf.safe_stack)
        acumon_argv.push_back("--safe-stack");
    if (conf.use_target_stderr)
        acumon_argv.push_back("--use-target-stderr");
    if (conf.unsupported_attach)
        acumon_argv.push_back("--unsupported-attach");

    // We currently have just one parameter in the sample interface
    // that gets mapped to both of these.
    if (conf.sample_generation > 0) {
	acumon_argv.push_back("--min-generation");
	acumon_argv.push_back(lexical_cast<string>(conf.sample_generation));
	acumon_argv.push_back("--max-generation");
	acumon_argv.push_back(lexical_cast<string>(conf.sample_generation));
    }

    acumon_argv.push_back("--tmpdir");
    acumon_argv.push_back(conf.tmpdir.string());

    if (!exec_binary(acumon_argv)) {
        fprintf(stderr, "Error: Sampling failed.\n");
        return 2;
    }

    int num_sample_files;
    for (num_sample_files = 0; ; num_sample_files++) {
        const string num(lexical_cast<string>(num_sample_files));
        fs::path atf(atf_path.string() + "." + num);
        fs::path acq(acq_path.string() + "." + num);
        if (!exists(atf) || !exists(atf))
            break;
    }

    if (num_sample_files == 0) {
	// Suppress this error output when the user knows he is
	// interested in a particular fork generation. Rationale in
	// TS-2239.
	if (conf.sample_generation == 0)
	    fprintf(stderr, "Error: No sample file was generated.\n");
        return 2;
    }

    for (int i = 0; i < num_sample_files; i++) {
        fprintf(stderr, "Post-processing sample file: %d/%d, please wait...\n",
                i + 1, num_sample_files);

        const string num(lexical_cast<string>(i));
        fs::path atf(atf_path.string() + "." + num);
        fs::path acq(acq_path.string() + "." + num);
	fs::path smp = conf.sample_file;

	// Ensure that the target directory exists
	fs::path parent_path = smp.parent_path();
	if (!parent_path.empty())
	    fs::create_directories(parent_path);

	// Last component can't be a directory
	if (smp.filename() == ".")
	    smp = parent_path / "sample.smp";

        vector<string> ar_argv;
        ar_argv.push_back(atf_resampler_path.string());
	ar_argv.push_back("--tmpdir");
	ar_argv.push_back(conf.tmpdir.string());
        ar_argv.push_back("-i");
        ar_argv.push_back(atf.string());
        ar_argv.push_back("--acquisition-info");
        ar_argv.push_back(acq.string());
        ar_argv.push_back("-o");
        ar_argv.push_back(smp.string());
        if (!conf.no_decimate && conf.burst_runtime == 0) {
            ar_argv.push_back("--sample-count");
            ar_argv.push_back(conf.sample_count);
        }

        if (!exec_binary(ar_argv)) {
            fprintf(stderr, "Error: Sample file post processing failed.\n");
            return 3;
        }

        system::error_code ec;
        if (!remove(atf))
            fprintf(stderr, "Warning: Failed to remove temporary file %s: "
                    "error %d, %s.\n", atf.string().c_str(), ec.value(),
                    ec.message().c_str());
        if (!remove(acq))
            fprintf(stderr, "Warning: Failed to remove temporary file %s: "
                    "error %d, %s.\n", acq.string().c_str(), ec.value(),
                    ec.message().c_str());
    }

    return 0;
}
