/// @file tools/sample_file_stats/sample_file_stats.cc

// Copyright (c) 2006-2010 Acumem AB.  All rights reserved.

// Prints some statistics about a sample file.

#define __STDC_FORMAT_MACROS

#include "library/fenris/flow_file_reader.hh"
#include "library/warning/warning_consumer.hh"
#include "library/version/version.h"

#include <boost/program_options.hpp>
#include <boost/multi_index_container.hpp>
#include <boost/multi_index/member.hpp>
#include <boost/multi_index/ordered_index.hpp>
#include <boost/multi_index/tag.hpp>
#include <boost/foreach.hpp>

#include <time.h>
#include <iostream>
#ifndef _WIN32
#include <inttypes.h>
#include <langinfo.h>
#endif

/* Bring in acumem svn version information */
BRING_IN_VERSIONINFO

using namespace std;
using namespace ::boost;
using namespace ::boost::multi_index;
namespace po = boost::program_options;
using namespace acumem::library;

struct conf_t {
    conf_t() : in_filename("") {}

    string in_filename;
};

static bool
parse_args(conf_t *conf, int argc, char **argv)
{
    po::options_description opt_desc_visible("Options");
    opt_desc_visible.add_options()
        ("help,h", "Print this help message.")
        ("in-file,i", po::value<string>(&conf->in_filename),
         "Specifies the input file. If no file is specified stdin is used.")
        ;

    po::options_description opt_desc_invisible("Options");
    opt_desc_invisible.add_options()
        ("positional-args", "")
        ;

    po::options_description opt_desc_all("Options");
    opt_desc_all.add(opt_desc_visible).add(opt_desc_invisible);

    po::positional_options_description pos_desc;
    pos_desc.add("positional-args", -1);

    po::variables_map vm;
    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(opt_desc_all).positional(pos_desc).run(), vm);
        po::notify(vm);
    }
    catch (po::unknown_option &e) {
        cerr << "Error: " << e.what() << endl;
        cerr << opt_desc_visible << endl;
        return true;
    }

    bool error = false;

    if (vm.count("help") > 0) {
        cerr << opt_desc_visible << endl;
        exit(0);
    }

    if (vm.count("in-file") > 1) {
        cerr << "Error: multiple input files specified" << endl;
        error = true;
    }

    if (vm.count("positional-args") > 0) {
        cerr << "Error: no positional arguments allowed" << endl;
        error = true;
    }

    return error;
}

typedef map<fnr_access_type_t, fnr_sample_count_t> access_type_counter_t;
typedef map<fnr_event_kind_t, access_type_counter_t> event_kind_counter_t;

typedef pair<fnr_address_t, fnr_time_t> pc_count_t;
struct by_pc {};
struct by_count {};
typedef multi_index_container<
    pc_count_t,
    indexed_by<
        ordered_unique<
            tag<by_pc>,
            member<pc_count_t, fnr_address_t, &pc_count_t::first> >,
        ordered_non_unique<
            tag<by_count>,
            member<pc_count_t, fnr_time_t, &pc_count_t::second> >
    >
> pc_counter_t;
typedef pc_counter_t::index<by_pc>::type pc_counter_by_pc_t;
typedef pc_counter_t::index<by_count>::type pc_counter_by_count_t;


struct inc_pc_counter_op_t : public unary_function<void, pc_count_t &> {
    void operator()(pc_count_t &count) {
        count.second++;
    }
} inc_pc_counter_op;

static void
inc_pc_counter(pc_counter_t &pc_counters, fnr_address_t pc)
{
    pc_counter_by_pc_t &pc_counter_by_pc = pc_counters.get<by_pc>();
    pc_counter_by_pc_t::iterator it = pc_counter_by_pc.find(pc);
    if (it == pc_counter_by_pc.end())
        pc_counter_by_pc.insert(pc_count_t(pc, 1));
    else
        pc_counter_by_pc.modify(it, inc_pc_counter_op);
}

struct line_size_stat_t {
    line_size_stat_t(): dangling_count(0) {}

    event_kind_counter_t event_counts;
    pc_counter_t begin_pc_counts;
    pc_counter_t end_pc_counts;
    fnr_time_set_t begin_times;
    fnr_time_set_t end_times;
    fnr_sample_count_t dangling_count;

    access_type_counter_t access_type_counts;
};

typedef map<unsigned int, fnr_real_time_t> burst_time_t;

// Capture data to present a thread communication breakdown:
// thread,type -> thread,type,count (or actually t1,typ1,t2,typ2 -> count
class communication_statistics_t {
public:

    void add(fnr_thread_id_t t0,
	     fnr_thread_id_t t1,
	     fnr_access_type_t type0,
	     fnr_access_type_t type1)
	{
	    comm_set[comm_entry_t(t0, t1, type0, type1)]++;
	}

    struct comm_entry_t {
	comm_entry_t(fnr_thread_id_t t0,
	     fnr_thread_id_t t1,
	     fnr_access_type_t type0,
	     fnr_access_type_t type1)
	    : t0(t0), t1(t1), type0(type0), type1(type1) {}

	fnr_thread_id_t t0;
	fnr_thread_id_t t1;
	fnr_access_type_t type0;
	fnr_access_type_t type1;

	bool operator < (const comm_entry_t &t) const {
	    if (t0 < t.t0) return true;
	    if (t0 > t.t0) return false;

	    if (t1 < t.t1) return true;
	    if (t1 > t.t1) return false;

	    if (type0 < t.type0) return true;
	    if (type0 > t.type0) return false;

	    if (type1 < t.type1) return true;
	    return false;
	}
    };

    typedef std::map<comm_entry_t, int> comm_set_t;
    comm_set_t comm_set;
};


class sample_file_stats_t
    : public fenris_flow_by_type_default_t
{
public:
    sample_file_stats_t();

    void count_danglings(fnr_line_size_t line_size,
			 line_size_stat_t &line_size_stat);

    void print_line_size_stats(fnr_line_size_t line_size,
			       const line_size_stat_t &line_size_stat);

    flow_status_t sample_begin(const fnr_event_sample_begin_t &event);
    flow_status_t sample_end(const fnr_event_sample_end_t &event);
    flow_status_t sample_period(const fnr_event_sample_period_t &event);
    flow_status_t acquisition_begin(const fnr_event_acquisition_begin_t &event);
    flow_status_t acquisition_end(const fnr_event_acquisition_end_t &event);
    flow_status_t mmap(const fnr_event_mmap_t &event);
    flow_status_t stack(const fnr_event_stack_t &event);
    flow_status_t sample(const fnr_event_sample_t &event);
    flow_status_t burst_begin(const fnr_event_burst_begin_t &event);
    flow_status_t burst_end(const fnr_event_burst_end_t &event);

    fnr_time_t end_time;
    map<fnr_line_size_t, line_size_stat_t> line_size_stats;

    map<fnr_sample_count_t, fnr_event_sample_begin_t> pending_begins;
    map<fnr_sample_count_t, fnr_event_sample_end_t> pending_ends;

    fnr_sample_count_t stack_count;

    vector<fnr_sample_count_t> time_buckets;
    fnr_time_t time_bucket_width;

    unsigned int burst_count;
    burst_time_t burst_begins;
    burst_time_t burst_ends;
    communication_statistics_t communication;

private:
    void time_bucket_add(fnr_time_t time);
    void add_sample_begin_stats(const fnr_event_sample_begin_t &begin);
    void add_sample_end_stats(const fnr_event_sample_begin_t &begin,
			      const fnr_event_sample_end_t &end);
    void add_sample_stats(const fnr_event_sample_t &sample);
};

sample_file_stats_t::sample_file_stats_t()
    : end_time(0),
      stack_count(0),
      burst_count(0)
{
    time_buckets.resize(20, 0);
    time_bucket_width = 1;
}

flow_status_t
sample_file_stats_t::sample_begin(const fnr_event_sample_begin_t &event)
{
    add_sample_begin_stats(event);

    map<fnr_sample_count_t, fnr_event_sample_end_t>::iterator
	end_iter = pending_ends.find(event.sample_id);
    if (end_iter != pending_ends.end()) {
	add_sample_end_stats(event, end_iter->second);
	pending_ends.erase(end_iter);
    } else
	pending_begins[event.sample_id] = event;

    return FLOW_OK;
}

flow_status_t
sample_file_stats_t::sample_end(const fnr_event_sample_end_t &event)
{
    map<fnr_sample_count_t, fnr_event_sample_begin_t>::iterator
	begin_iter = pending_begins.find(event.sample_id);
    if (begin_iter != pending_begins.end()) {
	add_sample_end_stats(begin_iter->second, event);
	pending_begins.erase(begin_iter);
    } else
	pending_ends[event.sample_id] = event;
    return FLOW_OK;
}

flow_status_t
sample_file_stats_t::sample_period(const fnr_event_sample_period_t &event) {
    cout << "Nominal sample period: " << event.sample_period << endl;
    return FLOW_OK;
}

static string
fnr_time_string(fnr_real_time_t t)
{
    struct tm local_time;
    time_t start_time = t / 1000000000ULL;
    if (start_time != 0) {
	char buf[256];
#ifdef _WIN32
        _localtime64_s(&local_time, &start_time);
	strftime(buf, sizeof(buf), "%x %X", &local_time);
#else
        localtime_r(&start_time, &local_time);
	strftime(buf, sizeof(buf), nl_langinfo(D_T_FMT), &local_time);
#endif
	return buf;
    }
    return "?";
}

flow_status_t
sample_file_stats_t::acquisition_begin(const fnr_event_acquisition_begin_t &event)
{
    cout << "Acquisition begin: " << fnr_time_string(event.time) << endl;
    switch (event.architecture) {
    case FNR_Arch_i386:
	cout << "Architecture: i386" << endl;
	break;
    case FNR_Arch_x86_64:
	cout << "Architecture: x86_64" << endl;
	break;
    default:
	assert(0);
    }
    cout << "Command line:";
    BOOST_FOREACH(const std::string &arg, event.argv)
	cout << " " << arg;
    cout << endl;
    cout << "User: " << event.user << endl;
    return FLOW_OK;
}

flow_status_t
sample_file_stats_t::acquisition_end(const fnr_event_acquisition_end_t &event)
{
    cout << "Acquisition end: " << fnr_time_string(event.time) << endl;
    return FLOW_OK;
}

flow_status_t
sample_file_stats_t::mmap(const fnr_event_mmap_t &event)
{
    printf("Mmap: %" PRIx64 "-%" PRIx64 " %s\n", event.address,
	   event.address + event.size,
	   event.filename.c_str());
    return FLOW_OK;
}

flow_status_t
sample_file_stats_t::stack(const fnr_event_stack_t &event)
{
    stack_count++;
    return FLOW_OK;
}

flow_status_t
sample_file_stats_t::sample(const fnr_event_sample_t &event)
{
    add_sample_stats(event);
    return FLOW_OK;
}

flow_status_t
sample_file_stats_t::burst_begin(const fnr_event_burst_begin_t &event)
{
    burst_begins[++burst_count] = event.real_time;
    return FLOW_OK;
}

flow_status_t
sample_file_stats_t::burst_end(const fnr_event_burst_end_t &event)
{
    burst_ends[burst_count] = event.real_time;
    return FLOW_OK;
}

void
sample_file_stats_t::time_bucket_add(fnr_time_t time)
{
    while (time >= time_buckets.size() * time_bucket_width) {
        for (vector<fnr_sample_count_t>::size_type i = 0;
             i < time_buckets.size(); i++)
            if (i < time_buckets.size() / 2)
                time_buckets[i] =
                    time_buckets[i * 2] + time_buckets[i * 2 + 1];
            else
                time_buckets[i] = 0;
        time_bucket_width *= 2;
    }

    time_buckets[(int)(time / time_bucket_width)]++;
}

void
sample_file_stats_t::add_sample_begin_stats(const fnr_event_sample_begin_t &begin)
{
    line_size_stat_t &line_size_stat = line_size_stats[begin.line_size];
    line_size_stat.event_counts[FNR_Event_Sample_Begin][begin.kind]++;
    line_size_stat.begin_times.insert(begin.time);
    inc_pc_counter(line_size_stat.begin_pc_counts, begin.pc);
    if (begin.time > end_time)
        end_time = begin.time;
    time_bucket_add(begin.time);
}

void
sample_file_stats_t::add_sample_end_stats(const fnr_event_sample_begin_t &begin,
					  const fnr_event_sample_end_t &end)
{
    line_size_stat_t &line_size_stat = line_size_stats[begin.line_size];
    line_size_stat.event_counts[FNR_Event_Sample_End][end.kind]++;
    line_size_stat.end_times.insert(end.time);
    inc_pc_counter(line_size_stat.end_pc_counts, end.pc);
    if (end.time > end_time)
        end_time = end.time;
}

void
sample_file_stats_t::add_sample_stats(const fnr_event_sample_t &sample)
{
    fnr_line_size_t line_size = line_size_from_log2(sample.line_size_log2);
    line_size_stat_t &line_size_stat = line_size_stats[line_size];
    line_size_stat.event_counts[FNR_Event_Sample_Begin][sample.begin.access_type]++;
    line_size_stat.begin_times.insert(sample.begin.time);
    inc_pc_counter(line_size_stat.begin_pc_counts, sample.begin.pc);
    time_bucket_add(sample.begin.time);

    if (!sample.end.empty()) {
	line_size_stat.event_counts[FNR_Event_Sample_End][sample.end[0].access_type]++;
	line_size_stat.end_times.insert(sample.end[0].time);
	inc_pc_counter(line_size_stat.end_pc_counts, sample.end[0].pc);
	if (sample.end[0].time > end_time)
	    end_time = sample.end[0].time;
    }

    line_size_stat.access_type_counts[sample.begin.access_type]++;
    BOOST_FOREACH(const fnr_sample_access_t &end, sample.end)
        line_size_stat.access_type_counts[end.access_type]++;

    fnr_thread_id_t last_thread = sample.begin.thread;
    fnr_access_type_t last_type = sample.begin.access_type;
    BOOST_FOREACH(const fnr_sample_access_t &end, sample.end) {
	communication.add(last_thread, end.thread, last_type, end.access_type);
	last_thread = end.thread;
	last_type = end.access_type;
    }
}

void
sample_file_stats_t::count_danglings(fnr_line_size_t line_size,
				     line_size_stat_t &line_size_stat)
{
    for (map<fnr_sample_count_t, fnr_event_sample_begin_t>::iterator
             di = pending_begins.begin(), de = pending_begins.end();
         di != de; di++)
        if (di->second.line_size == line_size)
            line_size_stat.dangling_count++;
    assert(pending_ends.size() == 0);
}

void
sample_file_stats_t::print_line_size_stats(fnr_line_size_t line_size,
					   const line_size_stat_t &line_size_stat)
{
    fnr_sample_count_t begin_events = 0;
    fnr_sample_count_t end_events = 0;

    BOOST_FOREACH(const event_kind_counter_t::value_type &ek,
                  line_size_stat.event_counts)
        switch (ek.first) {
        case FNR_Event_Sample_Begin:
            BOOST_FOREACH(const access_type_counter_t::value_type &at,
                          ek.second)
                begin_events += at.second;
            break;
        case FNR_Event_Sample_End:
            BOOST_FOREACH(const access_type_counter_t::value_type &at,
                          ek.second)
                end_events += at.second;
            break;
        default:
            assert(0);
        }

    cout << "Statitistics for line size " << line_size << ":" << endl;
    cout << "  Sample begin events:  " << begin_events << endl;
    cout << "  Sample end events:    " << end_events << endl;
    cout << "  Sample begin times:   "
         << line_size_stat.begin_times.size() << endl;
    cout << "  Sample end times:     "
         << line_size_stat.end_times.size() << endl;
    cout << "  Acquisition accesses: " << end_time << endl;
    cout << "  Real sample period:   "
         << (double)end_time / line_size_stat.begin_times.size() << endl;
    cout << "  Dangling samples:     " << line_size_stat.dangling_count << endl;
    cout << "  Top sample begin pcs:" << endl;
    {
        const pc_counter_by_count_t &pc_counters_by_count =
            line_size_stat.begin_pc_counts.get<by_count>();
        pc_counter_by_count_t::reverse_iterator
            ci = pc_counters_by_count.rbegin(),
            ce = pc_counters_by_count.rend();
        for (int i = 0; ci != ce && i < 10; ci++, i++)
            cout << "    0x" << hex << ci->first.virtual_address
                 << ", " << dec << ci->second << " samples" << endl;
    }
    {
        cout << "  Top sample end pcs:" << endl;
        const pc_counter_by_count_t &pc_counters_by_count =
            line_size_stat.end_pc_counts.get<by_count>();
        pc_counter_by_count_t::reverse_iterator
            ci = pc_counters_by_count.rbegin(),
            ce = pc_counters_by_count.rend();
        for (int i = 0; ci != ce && i < 10; ci++, i++)
            cout << "    0x" << hex << ci->first.virtual_address
                 << ", " << dec << ci->second << " samples" << endl;
    }

    // Burst stuff.
    if (burst_count > 0) {
        cout << "  Burst begin events: " << burst_begins.size() << endl;
        cout << "  Burst end events: " << burst_ends.size() << endl;
        unsigned int bursts = (int)min(burst_begins.size(), burst_ends.size());
        fnr_real_time_t total_time = 0;
        for (unsigned int i = 1; i <= bursts; i++) {
            burst_time_t::const_iterator bi = burst_begins.find(i);
            burst_time_t::const_iterator ei = burst_ends.find(i);
            assert(bi != burst_begins.end());
            assert(ei != burst_ends.end());
            total_time += ei->second - bi->second;
        }
        cout << "  Average burst length: "
	     << (total_time / 1e9 / bursts) << " s" << endl;
    }

    // Access types.
    cout << "  Access type counts:" << endl;
    BOOST_FOREACH(const access_type_counter_t::value_type &v,
                  line_size_stat.access_type_counts)
        cout << "    " << get_access_type_str(v.first)
	     << ": " << v.second << endl;
}

int
main(int argc, char **argv)
{
    conf_t conf;

    /* Parse arguments. */
    if (parse_args(&conf, argc, argv))
        return 1;

    warning_consumer_ptr_t warning_consumer(
        new warning_consumer_stderr_t());

    /* Read the samples. */
    sample_file_stats_t sfs;
    flow_file_reader_t flow(
        conf.in_filename == "" ? NULL :  conf.in_filename.c_str(),
        &sfs,
        warning_consumer);
    if (flow.dispatch() == STREAM_ERROR) {
	cerr << "Failed to read input file." << endl;
	return 1;
    }

    /* Print container information */
    cout << flow.get_header();

    /* Count dangling samples. */
    for (std::map<fnr_line_size_t, line_size_stat_t>::iterator
             lssi = sfs.line_size_stats.begin(), lsse = sfs.line_size_stats.end();
         lssi != lsse; lssi++)
        sfs.count_danglings(lssi->first, lssi->second);

    /* Print stack statistics. */
    cout << "Stacks: " <<  sfs.stack_count << endl;

    /* Print statistics for each line size. */
    for (map<fnr_line_size_t, line_size_stat_t>::const_iterator
             lssi = sfs.line_size_stats.begin(), lsse = sfs.line_size_stats.end();
         lssi != lsse; lssi++)
        sfs.print_line_size_stats(lssi->first, lssi->second);

    cout << "Sample distribution over time (" << sfs.time_bucket_width
         << " accesses per bucket):" << endl;
    BOOST_FOREACH(fnr_sample_count_t c, sfs.time_buckets)
        cout << c << endl;


    cout << endl;

    cout << "Communication patterns:" << endl;

    int total = 0;
    BOOST_FOREACH(communication_statistics_t::comm_set_t::value_type &v,
		  sfs.communication.comm_set) {
	total += v.second;
    }

    BOOST_FOREACH(communication_statistics_t::comm_set_t::value_type &v,
		  sfs.communication.comm_set) {
	cout << v.first.t0 << " -> " << v.first.t1 << "  "
	     << get_access_type_str(v.first.type0) << "-"
	     << get_access_type_str(v.first.type1) << "   : "
	     << v.second << "  ("
	     << 100.0 * (double) v.second / (double)total << "%)" << endl;
    }

    return 0;
}
