/// @file tools/sample_gui/cpu_selector.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include "cpu_selector.hh"
#include "library/version/product_version.h"

#include <boost/foreach.hpp>

using namespace std;
using namespace acumem::library;


static cpudb_cpu_ptr_t
create_generic_cpu_model()
{
    cpudb_cpu_ptr_t cpu(
	new cpudb_cpu_t("auto",
			"Generic processor model",
			0, // family
			0, // model
			1, // cores
			1, // threads per core
			false) // prefetchw dirty?
	);

    cpu->caches.resize(2);

    cpu->caches[0].reset(
	new cpudb_cache_t("L1", // name
			  1, // level
			  1, // num cores
			  true, true, // data/instr
			  FNR_Prefetch_Class_Any,
                          true, // hwpf
			  65536, // size
			  64)); // line size

    cpu->caches[1].reset(
	new cpudb_cache_t("L2", // name
			  2, // level
			  1, // num cores
			  true, true, // data/instr
			  FNR_Prefetch_Class_Any,
                          true, // hwpf
			  2097152, // size
			  64)); // line size

    return cpu;
}

cpu_selector_t::cpu_selector_t(const acumem::library::cpudb_cptr_t &cpudb,
			       bool feature_mt_advice,
			       bool feature_multisocket,
			       GtkWidget *parent_win)
    : cpudb(cpudb),
      feature_mt_advice(feature_mt_advice),
      feature_multisocket(feature_multisocket)
{
    autodetected_cpu = cpudb->autodetect_current_cpu(autodetect_failed);
    if (!autodetected_cpu)
        autodetected_cpu = create_generic_cpu_model();

    GtkWidget *w_dialog = gtk_dialog_new_with_buttons(
        "Select processor model - " PRODUCT_VERSION,
        GTK_WINDOW(parent_win),
        GTK_DIALOG_MODAL,
	GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
        GTK_STOCK_OK, GTK_RESPONSE_OK,
        NULL);

    dialog = GTK_DIALOG(w_dialog);
    gtk_window_set_default_size(GTK_WINDOW(dialog), 400, 600);
    gtk_box_set_spacing(GTK_BOX(dialog->vbox), 6);
    // We don't want the OK button to be sensitive by default,
    // on_cpu_changed will make it sensitive once a CPU has been
    // selected.
    gtk_dialog_set_response_sensitive(dialog, GTK_RESPONSE_OK, false);

    GtkWidget *label = gtk_label_new("Select processor model to use in the analysis");
    gtk_box_pack_start(
        GTK_BOX(dialog->vbox), label,
	false, false, 0);


    GtkCellRenderer *mfgr_renderer = gtk_cell_renderer_text_new();
    GtkCellRenderer *model_renderer = gtk_cell_renderer_text_new();
    if (gtk_check_version(2, 6, 0) == NULL)
	g_object_set(G_OBJECT(model_renderer),
		     "ellipsize", PANGO_ELLIPSIZE_END,
		     NULL);

    /* Setup the mfgr tree view and its associated list store */
    list_store_mfgr = gtk_list_store_new(N_MFGR_COLUMNS,
					 G_TYPE_POINTER,  /* Data */
					 G_TYPE_STRING    /* Name */
	);

    mfgr_list = GTK_COMBO_BOX(
	gtk_combo_box_new_with_model(GTK_TREE_MODEL(list_store_mfgr)));

    g_signal_connect(G_OBJECT(mfgr_list), "changed",
                     G_CALLBACK(on_mfgr_changed),
                     this);

    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(mfgr_list),
			       mfgr_renderer,
			       false); // expand

    gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(mfgr_list),
				  mfgr_renderer,
				  "text", COLUMN_MFGR_NAME);

    gtk_box_pack_start(
	GTK_BOX(dialog->vbox), GTK_WIDGET(mfgr_list),
	false, true, 0);

    /* Setup the CPU tree view and its associated list store */
    list_store_cpu = gtk_list_store_new(N_CPU_COLUMNS,
					G_TYPE_POINTER,  /* Data */
					G_TYPE_STRING,   /* ID */
					G_TYPE_STRING    /* Name */
	);

    cpu_list = GTK_TREE_VIEW(gtk_tree_view_new_with_model(GTK_TREE_MODEL(list_store_cpu)));

    GtkTreeSelection *cpu_selection = gtk_tree_view_get_selection(cpu_list);
    gtk_tree_selection_set_mode(cpu_selection, GTK_SELECTION_BROWSE);
    g_signal_connect(G_OBJECT(cpu_selection), "changed",
                     G_CALLBACK(on_cpu_changed),
                     this);

    GtkTreeViewColumn *column_name =
	gtk_tree_view_column_new_with_attributes("Processor Name",
						 model_renderer,
						 "text", COLUMN_CPU_NAME,
						 NULL);

    GtkTreeViewColumn *column_id =
	gtk_tree_view_column_new_with_attributes("Analysis model",
						 model_renderer,
						 "text", COLUMN_CPU_ID,
						 NULL);

    gtk_tree_view_column_set_sort_column_id(column_id, COLUMN_CPU_ID);
    gtk_tree_view_column_set_sort_column_id(column_name, COLUMN_CPU_NAME);

    g_object_set(G_OBJECT(column_id),
		 "resizable", true,
		 "expand", false,
		 NULL);
    g_object_set(G_OBJECT(column_name),
		 "resizable", true,
		 "expand", true,
		 NULL);

    gtk_tree_view_append_column(GTK_TREE_VIEW(cpu_list), column_name);
    gtk_tree_view_append_column(GTK_TREE_VIEW(cpu_list), column_id);

    gtk_tree_view_set_search_column(GTK_TREE_VIEW(cpu_list), COLUMN_CPU_NAME);
    gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(list_store_cpu),
					 COLUMN_CPU_NAME, GTK_SORT_ASCENDING);
    GtkWidget *sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_box_pack_start(
	GTK_BOX(dialog->vbox), sw,
	true, true, 0);
    gtk_container_add(GTK_CONTAINER(sw), GTK_WIDGET(cpu_list));



    if (feature_mt_advice && feature_multisocket) {
	// Create physical CPUs widgets
	GtkWidget *ncpu_hbox = gtk_hbox_new(false, 0);
	GtkLabel *ncpu_label = GTK_LABEL(gtk_label_new("Physical processors:"));
	gtk_widget_show(GTK_WIDGET(ncpu_label));
	//gtk_misc_set_alignment(GTK_MISC(ncpu_label), 0.0, 0.5);
	gtk_box_pack_start(GTK_BOX(ncpu_hbox), GTK_WIDGET(ncpu_label),
			   false, true, 0);

	ncpu_sbutton = GTK_SPIN_BUTTON(gtk_spin_button_new_with_range(1, 16, 1));
	try {
            warning_consumer_ptr_t wc(new warning_consumer_null_t());
	    gtk_spin_button_set_value(ncpu_sbutton, get_cpuid(wc)->num_packages());
	} catch (cpuid_exception_t e) {
	    gtk_spin_button_set_value(ncpu_sbutton, 1);
	}

	gtk_widget_show(GTK_WIDGET(ncpu_sbutton));
	gtk_box_pack_start(GTK_BOX(ncpu_hbox), GTK_WIDGET(ncpu_sbutton),
			   true, true, 0);

	gtk_box_pack_start(GTK_BOX(dialog->vbox), ncpu_hbox,
			   false, true, 0);
    }

    // Add all cpu manufacturers to the combo box, set the first (auto
    // detected) as active.
    append_mfgr_auto();
    BOOST_FOREACH(const cpudb_manufacturer_cptr_t &mfgr, cpudb->manufacturers)
	append_mfgr(mfgr);
    gtk_combo_box_set_active(mfgr_list, 0);
    update_cpu_list();
}

cpu_selector_t::~cpu_selector_t() {
}

bool
cpu_selector_t::show() {
    while(1) {
	gtk_widget_show_all(GTK_WIDGET(dialog));
	gint response = gtk_dialog_run(dialog);

	switch (response) {
	case GTK_RESPONSE_OK:
	    gtk_widget_hide(GTK_WIDGET(dialog));
	    return true;

	case GTK_RESPONSE_CANCEL:
	case GTK_RESPONSE_DELETE_EVENT:
	    gtk_widget_hide(GTK_WIDGET(dialog));
	    return false;

	default:
	    assert(0);
	}
    }
}

string
cpu_selector_t::get_selected_id()
{
    GtkTreeSelection *sel = gtk_tree_view_get_selection(cpu_list);
    GtkTreeIter it_mfgr, it_cpu;

    if (!gtk_combo_box_get_active_iter(mfgr_list, &it_mfgr) ||
	!gtk_tree_selection_get_selected(sel, NULL, &it_cpu))
	return "auto";

    const cpudb_manufacturer_t *mfgr;
    const cpudb_cpu_t *cpu;

    gtk_tree_model_get(GTK_TREE_MODEL(list_store_mfgr), &it_mfgr,
		       COLUMN_MFGR_DATA, &mfgr,
		       -1);
    gtk_tree_model_get(GTK_TREE_MODEL(list_store_cpu), &it_cpu,
		       COLUMN_CPU_DATA, &cpu,
		       -1);

    if (mfgr)
	return mfgr->id + "/" + cpu->id;
    else
	return "auto";
}

const cpudb_cpu_t *
cpu_selector_t::get_selected_cpu()
{
    GtkTreeSelection *sel = gtk_tree_view_get_selection(cpu_list);
    GtkTreeIter it;

    if (!gtk_tree_selection_get_selected(sel, NULL, &it))
	return autodetected_cpu.get();

    const cpudb_cpu_t *cpu;

    gtk_tree_model_get(GTK_TREE_MODEL(list_store_cpu), &it,
		       COLUMN_CPU_DATA, &cpu,
		       -1);
    return cpu;
}

string
cpu_selector_t::get_selected_display_name()
{
    GtkTreeSelection *sel = gtk_tree_view_get_selection(cpu_list);
    GtkTreeIter it;

    if (!gtk_tree_selection_get_selected(sel, NULL, &it))
	return autodetected_cpu->names.back();

    GValue val = {0};
    gtk_tree_model_get_value(GTK_TREE_MODEL(list_store_cpu),
			     &it,
			     COLUMN_CPU_NAME,
			     &val);
    string ret(g_value_get_string(&val));
    g_value_unset(&val);

    return ret;
}

unsigned int
cpu_selector_t::get_num_cpus() const
{
    if (feature_mt_advice && feature_multisocket)
	return (unsigned int)gtk_spin_button_get_value(ncpu_sbutton);
    else
	return 1;
}


void
cpu_selector_t::update_cpu_list()
{
    GtkTreeIter iter;

    gtk_list_store_clear(list_store_cpu);
    if (!gtk_combo_box_get_active_iter(mfgr_list, &iter))
	return;

    const cpudb_manufacturer_t *mfgr;
    gtk_tree_model_get(GTK_TREE_MODEL(list_store_mfgr),
		       &iter,
		       COLUMN_MFGR_DATA, &mfgr,
		       -1);

    if (mfgr) {
	BOOST_FOREACH(const cpudb_cpu_cptr_t &cpu, mfgr->get_cpus())
	    append_cpu(cpu);
    } else
	append_cpu(autodetected_cpu);

}


void
cpu_selector_t::on_mfgr_changed(GtkComboBox *widget, gpointer user_data)
{
    cpu_selector_t *_this((cpu_selector_t *)user_data);

    _this->update_cpu_list();
}

void
cpu_selector_t::on_cpu_changed(GtkTreeSelection *selection, gpointer user_data)
{
    cpu_selector_t *_this((cpu_selector_t *)user_data);
    bool selection_empty(
	gtk_tree_selection_count_selected_rows(selection) == 0);

    gtk_dialog_set_response_sensitive(_this->dialog,
				      GTK_RESPONSE_OK,
				      !selection_empty);
}

void
cpu_selector_t::append_mfgr(const cpudb_manufacturer_cptr_t &mfgr)
{
    GtkTreeIter iter;
    gtk_list_store_append(list_store_mfgr, &iter);
    gtk_list_store_set(list_store_mfgr, &iter,
		       COLUMN_MFGR_DATA, mfgr.get(),
		       COLUMN_MFGR_NAME, mfgr->name.c_str(),
		       -1);
}

void
cpu_selector_t::append_mfgr_auto()
{
    GtkTreeIter iter;
    gtk_list_store_append(list_store_mfgr, &iter);
    gtk_list_store_set(list_store_mfgr, &iter,
		       COLUMN_MFGR_DATA, NULL,
		       COLUMN_MFGR_NAME, "Autodetected processor",
		       -1);
}

void
cpu_selector_t::append_cpu(const cpudb_cpu_cptr_t &cpu)
{
    BOOST_FOREACH(const string &name, cpu->names) {
	GtkTreeIter iter;
	gtk_list_store_append(list_store_cpu, &iter);
	gtk_list_store_set(list_store_cpu, &iter,
			   COLUMN_CPU_DATA, cpu.get(),
			   COLUMN_CPU_ID, cpu->id.c_str(),
			   COLUMN_CPU_NAME, name.c_str(),
			   -1);
    }
}
