/// @file tools/sample_gui/pid_selector.cc

// Copyright (c) 2007-2009 Acumem AB.  All rights reserved.

#include "pid_selector.hh"
#include "library/version/product_version.h"

#include <assert.h>
#include <errno.h>
#include <string.h>

using namespace std;

pid_selector_base_t::pid_selector_base_t(GtkWidget *parent_win)
{
    dialog = gtk_dialog_new_with_buttons(
        "Select process - " PRODUCT_VERSION,
        GTK_WINDOW(parent_win),
        GTK_DIALOG_MODAL,
	GTK_STOCK_REFRESH, 1,
	GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
        GTK_STOCK_OK, GTK_RESPONSE_OK,
        NULL);
    gtk_window_set_default_size(GTK_WINDOW(dialog), 400, 600);
    gtk_box_set_spacing(GTK_BOX(GTK_DIALOG(dialog)->vbox), 6);
    // We don't want the OK button to be sensitive by default,
    // on_cpu_changed will make it sensitive once a CPU has been
    // selected.
    gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog),
				      GTK_RESPONSE_OK, false);

    GtkWidget *label = gtk_label_new("Select process to sample.");
    gtk_box_pack_start(
        GTK_BOX(GTK_DIALOG(dialog)->vbox), label,
	false, false, 0);

    GtkWidget *sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_box_pack_start(
	GTK_BOX(GTK_DIALOG(dialog)->vbox), sw,
	true, true, 0);

    list_store = gtk_list_store_new (N_COLUMNS,
				     G_TYPE_INT,      /* PID */
				     G_TYPE_STRING    /* App name */
				     );
    list = GTK_TREE_VIEW(gtk_tree_view_new_with_model(GTK_TREE_MODEL(list_store)));
    GtkTreeSelection *selection = gtk_tree_view_get_selection(list);
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);
    g_signal_connect(G_OBJECT(selection), "changed",
                     G_CALLBACK(on_pid_changed),
                     this);

    GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
    if (gtk_check_version(2, 6, 0) == NULL)
	g_object_set(G_OBJECT(renderer),
		     "ellipsize", PANGO_ELLIPSIZE_END,
		     NULL);

    GtkTreeViewColumn *column_pid =
	gtk_tree_view_column_new_with_attributes("PID", renderer, "text", COLUMN_PID, NULL);
    GtkTreeViewColumn *column_name =
	gtk_tree_view_column_new_with_attributes("Name", renderer, "text", COLUMN_CMDLINE, NULL);

    gtk_tree_view_column_set_sort_column_id(column_pid, COLUMN_PID);
    gtk_tree_view_column_set_sort_column_id(column_name, COLUMN_CMDLINE);


    g_object_set(G_OBJECT(column_pid),
		 "resizable", true,
		 "expand", false,
		 NULL);
    g_object_set(G_OBJECT(column_name),
		 "resizable", true,
		 "expand", true,
		 NULL);

    gtk_tree_view_append_column(GTK_TREE_VIEW(list), column_pid);
    gtk_tree_view_append_column(GTK_TREE_VIEW(list), column_name);

    gtk_tree_view_set_search_column(GTK_TREE_VIEW(list), COLUMN_CMDLINE);
    gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(list_store),
					 COLUMN_PID, GTK_SORT_ASCENDING);

    gtk_container_add(GTK_CONTAINER(sw), GTK_WIDGET(list));
}

pid_selector_base_t::~pid_selector_base_t() {
}

bool
pid_selector_base_t::show() {
    if(!update_list())
	return 0;

    while(1) {
	gtk_widget_show_all(dialog);
	gint response = gtk_dialog_run(GTK_DIALOG(dialog));

	switch (response) {
	case GTK_RESPONSE_OK: {
	    GtkTreeSelection *sel = gtk_tree_view_get_selection(list);
	    if(gtk_tree_selection_get_selected(sel, NULL, NULL))
		return true;
	    else
		break;
	}

	case GTK_RESPONSE_CANCEL:
	case GTK_RESPONSE_DELETE_EVENT:
	    gtk_widget_hide(dialog);
	    return false;

	case 1:
	    // TODO: Check return value
	    update_list();
	    break;
	default:
	    assert(0);
	}
    }
}

pid_t
pid_selector_base_t::get_selected_pid()
{
    GtkTreeSelection *sel = gtk_tree_view_get_selection(list);
    GtkTreeIter it;
    if(gtk_tree_selection_get_selected(sel, NULL, &it)) {
	GValue val = {0};
	gtk_tree_model_get_value(GTK_TREE_MODEL(list_store), &it, COLUMN_PID, &val);
	if(G_VALUE_HOLDS(&val, G_TYPE_INT)) {
	    pid_t ret = g_value_get_int(&val);
	    g_value_unset(&val);
	    gtk_widget_hide(dialog);
	    return ret;
	}
	g_value_unset(&val);
    }

    gtk_widget_hide(dialog);
    return 0;
}

void
pid_selector_base_t::display_error(const int errnum, const char *msg)
{
    GtkWidget *dlg_error = gtk_message_dialog_new(GTK_WINDOW(dialog),
						  GTK_DIALOG_MODAL,
						  GTK_MESSAGE_ERROR,
						  GTK_BUTTONS_OK,
						  "%s\n%s",
						  msg,
						  strerror(errnum));
    gtk_dialog_run(GTK_DIALOG(dlg_error));
    gtk_widget_destroy(dlg_error);
}


bool
pid_selector_base_t::is_str_int(const char *txt)
{
    for(; *txt; txt++)
	if(!isdigit(*txt))
	    return 0;
    return 1;
}


void
pid_selector_base_t::append_process(int pid, const char *cmd)
{
  GtkTreeIter iter;
  gtk_list_store_append(list_store, &iter);
  gtk_list_store_set(list_store, &iter,
		     COLUMN_PID, pid,
		     COLUMN_CMDLINE, cmd,
		     -1);
}

void
pid_selector_base_t::on_pid_changed(GtkTreeSelection *selection, gpointer user_data)
{
    pid_selector_base_t *_this((pid_selector_base_t *)user_data);
    bool selection_empty(
	gtk_tree_selection_count_selected_rows(selection) == 0);

    gtk_dialog_set_response_sensitive(GTK_DIALOG(_this->dialog),
				      GTK_RESPONSE_OK,
				      !selection_empty);
}
