/// @file tools/sample_gui/sample_gui.cc

// Copyright (c) 2007-2010 Acumem AB.  All rights reserved.

// TODO: Check that a report directory has been selected
// TODO: Check that a report name has been selected
// TODO: Check that debug info and and source have been entered, if selected.
// TODO: Don't keep the xterm open by default?
// TODO: Use event driven I/O instead of timeout.
// TODO: Fix all the memory leaks.
// TODO: Add choice of terminal (try gnome-terminal)
// TODO: Cancel sampling from dialog
// TODO: Cancel report from dialog
// TODO: Check that the user has entered a correct binary
// TODO: Check that the user has entered a correct pid

// TODO: Dialog text gets selected.

#include "library/version/product_version.h"
#include "library/version/copyright.h"
#include "library/version/version.h"
#include "library/lang/am_assert.hh"
#include "library/exe_path/exe_path.hh"

#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <assert.h>

#include <iostream>
#include <string>
#include <vector>

#include <boost/program_options.hpp>
#include <boost/foreach.hpp>
#include <boost/filesystem.hpp>

#include "library/license/license.hh"

#include "library/warning/warning_consumer.hh"

#include "line_size_selector.hh"

#include "paths.hh"
#include "web_browser.hh"

#include "cpu_selector.hh"

#if defined(__sun__)
#include "pid_selector_solaris.hh"
typedef pid_selector_solaris_t pid_selector_t;

#define kill killpg

#elif defined(__linux__)

#include "pid_selector_linux.hh"
typedef pid_selector_linux_t pid_selector_t;

#endif

#include "product.h"

BRING_IN_VERSIONINFO;

#define STRINGIFY_(x) #x
#define STRINGIFY(x) STRINGIFY_(x)

#define COL_SPACING 2
#define ROW_SPACING 0

using namespace acumem::library;
using namespace acumem::license;
using namespace std;
namespace po = boost::program_options;
namespace fs = boost::filesystem;

// Shorthand for common layout parameters.
#define ATT_NONE        ((GtkAttachOptions)0)
#define ATT_FILL        ((GtkAttachOptions)GTK_FILL)
#define ATT_EXPAND_FILL ((GtkAttachOptions)(GTK_EXPAND | GTK_FILL))

struct browse_t;

/// Maps a widget to the widgets that must be active for the widget to be
/// sensitive.
typedef map<GtkWidget *, set<GtkWidget *> > sensitivity_map_t;

// Global paths variable, ugly, but simplifies things...  Note: Has to
// be a pointer since we have to guarantee that Glib is initialized
// before the contstructor of paths_t executes.
paths_t *paths;




struct conf_t {
    conf_t();

    void set_exit(int exit_code_);

    // If the application should exit immediately after parsing its arguments
    // exit is set to true and exit code to an appropriate exit code.
    bool exit;
    int exit_code;
    fs::path tmpdir;

    string program;
    vector<string> arguments;
};

conf_t::conf_t()
    : exit(false)
{
}

void
conf_t::set_exit(int exit_code_)
{
    exit = true;
    exit_code = exit_code_;
}

static void
parse_args(conf_t *conf, int argc, char **argv)
{
    string usage(string("Usage: ") + argv[0] +
                 " [option]... [-- program [argument]...]\n\n");

    conf->tmpdir = acumem::exe_path::get_temp_path();

    po::options_description opt_desc("Options");
    opt_desc.add_options()
	("tmpdir", po::value<fs::path>(&conf->tmpdir),
	 ("Directory to store temporary files in. Default: " + conf->tmpdir.string()).c_str())
        ("help,h", "Print this help message.")
        ("version", "Print version information.")
        ;

    po::options_description opt_desc_invisible("Options");
    opt_desc_invisible.add_options()
        ("program", po::value<string>(&conf->program))
        ("argument", po::value<vector<string> >(&conf->arguments)->composing())
        ;

    po::positional_options_description pos_desc;
    pos_desc.add("program", 1);
    pos_desc.add("argument", -1);

    po::options_description opt_desc_all("Options");
    opt_desc_all
	.add(opt_desc)
        .add(opt_desc_invisible);

    po::variables_map vm;
    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(opt_desc_all)
                  .positional(pos_desc)
                  .run(), vm);
        po::notify(vm);
    }
    catch (po::error &e) {
        cerr << "Error: " << e.what() << endl;
        cout << usage;
        cerr << opt_desc << endl;
        conf->set_exit(1);
        return;
    }

    if (vm.count("help") > 0) {
        cout << usage;
        cout << opt_desc << endl;
        conf->set_exit(0);
        return;
    }

    if (vm.count("version") > 0) {
        cout << PRODUCT_VERSION << endl << COPYRIGHT << endl;
        conf->set_exit(0);
        return;
    }
}


/// Removes leading and trailing whitespace from the string.
static string
trim(string *s)
{
    string::size_type start = 0;
    string::size_type end = s->size();

    while (start < end && isspace(s->at(start)))
        start++;

    while (start < end && isspace(s->at(end - 1)))
        end--;

    return s->substr(start, end - start);
}

static const char *burst_quality_levels[] = {
    "fast",
    "normal",
    "detailed",
    NULL
};

static const int default_burst_quality_level = 1;

static const char *report_line_sizes[] = {
    "2",
    "4",
    "8",
    "16",
    "32",
    "64",
    "128",
    "256",
    "512",
    "1024",
    "2048",
    NULL
};

static const int default_report_line_size = 5;

static void
append_combo_box_text_vector(GtkComboBox *combo, const char **text_vector)
{
    for (const char **text = text_vector; *text; text++)
	gtk_combo_box_append_text(combo, *text);
}

// License feature flags;
static bool show_depth_setting = false;
static bool show_blocking_expert_mode = false;
static bool show_attach_options = false;
static bool feature_mt_advice = false;
static bool feature_multisocket = false;

class app_t {
public:
    app_t(const conf_t *conf);
    ~app_t();

    void create_sampling_options_dialog();
    void create_report_options_dialog();

    void update_cache_levels();

    int get_selected_cache_level();

    const conf_t *conf;

    // Main window and layout table.
    GtkWidget *main_win;
    GtkWidget *main_table;

    // Menu.
    GtkWidget *menu_bar;
    GtkWidget *file_item;
    GtkWidget *file_menu;
    GtkWidget *quit_item;

    // Sample source label.
    GtkWidget *smp_src_label;

    // Sample source application radio button.
    GtkWidget *smp_src_app_rbutton;

    // Sample source application launch radio button.
    GtkWidget *smp_src_app_launch_rbutton;

    // Program input widgets.
    GtkWidget *prg_label;
    GtkWidget *prg_entry;
    GtkWidget *prg_browse_button;

    // Argument input widgets.
    GtkWidget *arg_label;
    GtkWidget *arg_entry;

    // Working directory input widgets.
    GtkWidget *wd_label;
    GtkWidget *wd_entry;
    GtkWidget *wd_browse_button;

    // Sample source application attach radio button.
    GtkWidget *smp_src_app_attach_rbutton;

    // Pid input widgets.
    GtkWidget *pid_label;
    GtkWidget *pid_entry;
    GtkWidget *pid_select_button;

    // Advanced sampling options button.
    GtkWidget *smp_adv_button;

    // Advanced sampling options window and layout table.
    GtkWidget *smp_adv_dialog;
    GtkWidget *smp_adv_table;

    // Sampling start label.
    GtkWidget *start_label;

    // Start sampling immediately input widgets.
    GtkWidget *start_imm_rbutton;

    // Start sampling after delay input widgets.
    GtkWidget *start_del_rbutton;
    GtkWidget *start_del_sbutton;
    GtkWidget *start_del_unit_label;

    // Start sampling at function input widgets.
    GtkWidget *start_fun_rbutton;
    GtkWidget *start_fun_entry;
    GtkWidget *start_fun_label2;
    GtkWidget *start_fun_sbutton;
    GtkWidget *start_fun_label3;

    // Start sampling at address input widgets.
    GtkWidget *start_addr_rbutton;
    GtkWidget *start_addr_entry;
    GtkWidget *start_addr_label2;
    GtkWidget *start_addr_sbutton;
    GtkWidget *start_addr_label3;

    // Sampling stop label.
    GtkWidget *stop_label;

    // Stop sampling immediately input widgets.
    GtkWidget *stop_imm_rbutton;

    // Stop sampling after delay input widgets.
    GtkWidget *stop_dur_rbutton;
    GtkWidget *stop_dur_sbutton;
    GtkWidget *stop_dur_unit_label;

    // Stop sampling at function input widgets.
    GtkWidget *stop_fun_rbutton;
    GtkWidget *stop_fun_entry;
    GtkWidget *stop_fun_label2;
    GtkWidget *stop_fun_sbutton;
    GtkWidget *stop_fun_label3;

    // Stop sampling at address input widgets.
    GtkWidget *stop_addr_rbutton;
    GtkWidget *stop_addr_entry;
    GtkWidget *stop_addr_label2;
    GtkWidget *stop_addr_sbutton;
    GtkWidget *stop_addr_label3;

    // Sampling line sizes input widgets.
    GtkWidget *smp_lsz_label;
    GtkWidget *smp_lsz_val_label;
    GtkWidget *smp_lsz_select_button;

    // Sample period input widgets.
    GtkWidget *smp_per_label;
    GtkWidget *smp_per_sbutton;
    GtkWidget *smp_per_unit_label;

    // Sampling sample file input widgets.
    GtkWidget *smp_smp_file_cbutton;
    GtkWidget *smp_smp_file_entry;
    GtkWidget *smp_smp_file_browse_button;

    // Safe stack input widgets.
    GtkWidget *safe_stk_cbutton;

    // Burst mode input widgets.
    GtkWidget *burst_cbutton;
    GtkWidget *burst_len_sbutton;
    GtkWidget *burst_len_unit_label;
    GtkWidget *burst_quality_combo;
    GtkWidget *burst_quality_label;

    // Sample button.
    GtkWidget *smp_button;

    // Sample source file radio button.
    GtkWidget *smp_src_file_rbutton;

    // Read sample file input widgets.
    GtkWidget *read_smp_file_label;
    GtkWidget *read_smp_file_entry;
    GtkWidget *read_smp_file_browse_button;

    // Report generation label.
    GtkWidget *rpt_label;

    // Sample source application radio button.
    GtkWidget *rpt_gen_rbutton;

    // Report directory input widgets.
    GtkWidget *rpt_file_label;
    GtkWidget *rpt_file_entry;
    GtkWidget *rpt_file_browse_button;

    // CPU model
    GtkWidget *cm_label;
    GtkWidget *cm_cpu;
    GtkWidget *cm_select;

    // Cache level selection
    enum {
	COLUMN_LEVEL_ID,
	COLUMN_LEVEL_NAME,
	N_LEVEL_COLUMNS
    };
    GtkWidget *cls_label;
    GtkWidget *cls_level;
    GtkListStore *cls_level_store;

    // Web browser launch input widgets.
    GtkWidget *www_cbutton;

    // Advanced report options button.
    GtkWidget *rpt_adv_button;

    // Advanced report options expander and layout table.
    GtkWidget *rpt_adv_dialog;
    GtkWidget *rpt_adv_table;

    // Cache size override
    GtkWidget *rpt_cache_size_cbutton;
    GtkWidget *rpt_cache_size_entry;
    GtkWidget *rpt_cache_size_unit_label;

    // Line size override
    GtkWidget *rpt_line_size_cbutton;
    GtkWidget *rpt_line_size_combo;
    GtkWidget *rpt_line_size_unit_label;

    // Number of caches override
    GtkWidget *rpt_num_caches_cbutton;
    GtkWidget *rpt_num_caches_sbutton;

    // Number of CPUs override
    GtkWidget *rpt_num_cpus_cbutton;
    GtkWidget *rpt_num_cpus_sbutton;

    // Replacement policy input widgets.
    GtkWidget *rpt_repl_label;
    GtkWidget *rpt_repl_combo;

    // Call stack depth input widgets.
    GtkWidget *stk_label;
    GtkWidget *stk_sbutton;
    GtkWidget *stk_unit_label;

    // Debug info file input widgets.
    GtkWidget *dbg_cbutton;
    GtkWidget *dbg_entry;
    GtkWidget *dbg_browse_button;

    // Additional debug directory input widgets.
    GtkWidget *dbgd_cbutton;
    GtkWidget *dbgd_entry;
    GtkWidget *dbgd_browse_button;

    // Source directory input widgets.
    GtkWidget *src_cbutton;
    GtkWidget *src_entry;
    GtkWidget *src_browse_button;

    // c++filt input widgets.
    GtkWidget *cppfilt_cbutton;
    GtkWidget *cppfilt_entry;
    GtkWidget *cppfilt_browse_button;

    // Issue threshold input widgets.
    GtkWidget *thr_label;
    GtkWidget *thr_sbutton;
    GtkWidget *thr_unit_label;

    // Report title input widgets.
    GtkWidget *ttl_cbutton;
    GtkWidget *ttl_entry;

    // Blocking expert mode button
    GtkWidget *blocking_expert_cbutton;

    // Verbose button
    GtkWidget *verbose_cbutton;

    // View existing report radio button.
    GtkWidget *rpt_file_rbutton;

    // View existing report file input widgets.
    GtkWidget *report_file_label;
    GtkWidget *report_file_entry;
    GtkWidget *report_file_browse_button;

    // Report button.
    GtkWidget *rpt_button;

    // Column spacers.
    GtkWidget *col0_spc;
    GtkWidget *col1_spc;
    GtkWidget *col2_spc;

#if 0
    // Skip-intro-page input widgets.
    GtkWidget *skip_intro_cbutton;
#endif

    // Sampling line size selector.
    line_size_selector_t *line_sizes;

    // Process selector
    pid_selector_t *pid_selector;

    // CPU model selector
    cpu_selector_t *cpu_selector;

    // Browse button bindings.
    vector<browse_t *> browse_bindings;

    // Web browser interface object
    web_browser_t *browser;

    // Sensitivity dependencies.
    sensitivity_map_t sensitivity_deps;
    sensitivity_map_t inverse_sensitivity_deps;

    void add_sensitivity_dep(GtkWidget *dependent, GtkWidget *dependee);
    void add_inverse_sensitivity_dep(GtkWidget *dependent, GtkWidget *dependee);
};

void
app_t::add_sensitivity_dep(GtkWidget *dependent, GtkWidget *dependee)
{
    sensitivity_deps[dependent].insert(dependee);
}

void
app_t::add_inverse_sensitivity_dep(GtkWidget *dependent, GtkWidget *dependee)
{
    inverse_sensitivity_deps[dependent].insert(dependee);
}

static void
error_dialog(GtkWidget *parent_win, string message)
{
    GtkWidget *dialog = gtk_message_dialog_new(
        GTK_WINDOW(parent_win), GTK_DIALOG_MODAL,
        GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
        message.c_str());
    gtk_window_set_title(GTK_WINDOW(dialog), PRODUCT_VERSION);
    gtk_widget_show(dialog);
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
}

#if 0
static void
info_dialog(GtkWidget *parent_win, string message)
{
    GtkWidget *dialog = gtk_message_dialog_new(
        GTK_WINDOW(parent_win), GTK_DIALOG_MODAL,
        GTK_MESSAGE_INFO, GTK_BUTTONS_OK,
        message.c_str());
    gtk_window_set_title(GTK_WINDOW(dialog), PRODUCT_VERSION);
    gtk_widget_show(dialog);
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
}
#endif

static void
warning_dialog(GtkWidget *parent_win, string message)
{
    GtkWidget *dialog = gtk_message_dialog_new(
        GTK_WINDOW(parent_win), GTK_DIALOG_MODAL,
        GTK_MESSAGE_WARNING, GTK_BUTTONS_OK,
        message.c_str());
    gtk_window_set_title(GTK_WINDOW(dialog), PRODUCT_VERSION);
    gtk_widget_show(dialog);
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
}

struct browse_t {
    browse_t(app_t *app,
             string title,
             GtkWidget *entry,
             GtkFileChooserAction action)
        : app(app),
          title(title),
          entry(entry),
          action(action) {}

    app_t *app;
    string title;
    GtkWidget *entry;
    GtkFileChooserAction action;
};

// Enable/disable widgets when the radio/check buttons are toggled.
static void
update_sensitivities(GtkWidget *widget,
                     gpointer data)
{
    app_t *app = (app_t *)data;

    const char *go_label;
    if (gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->rpt_file_rbutton)))
        go_label = "View report";
    else
        if (gtk_toggle_button_get_active(
                GTK_TOGGLE_BUTTON(app->smp_src_app_rbutton)))
            go_label = "Sample application and generate report";
        else
            go_label = "Generate report from sample file";
    gtk_button_set_label(GTK_BUTTON(app->rpt_button), go_label);

    BOOST_FOREACH(sensitivity_map_t::value_type dep, app->sensitivity_deps) {
        bool sensitive = true;
        BOOST_FOREACH(GtkWidget *dependee, dep.second)
            if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(dependee)))
                sensitive = false;
        gtk_widget_set_sensitive(dep.first, sensitive);
    }

    BOOST_FOREACH(sensitivity_map_t::value_type dep, app->inverse_sensitivity_deps) {
        bool sensitive = true;
        BOOST_FOREACH(GtkWidget *dependee, dep.second)
            if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(dependee)))
                sensitive = false;
        gtk_widget_set_sensitive(dep.first, sensitive);
    }
}

// Return the current working directory.
static string
get_working_directory(void)
{
    size_t size = 1;
    char *wd = NULL;
    char *res;
    do {
        size *= 2;
        wd = (char *)realloc(wd, size);
        assert(wd != NULL);
        res = getcwd(wd, size);
    } while (res == NULL && errno == ERANGE);

    assert(res != NULL);

    string wds = string(wd);
    free(wd);

    return wds;
}

// Exit when the main window is destroyed.
static void
main_win_destroy(GtkWidget *widget,
                 gpointer data)
{
    gtk_main_quit();
}

// Exit when the quit menu item is activated.
static void
quit_item_activate(GtkWidget *widget,
                   gpointer data)
{
    gtk_main_quit();
}

// Select cache button clicked
static void
cm_select_clicked(GtkWidget *widget,
		  gpointer data)
{
    app_t *app = (app_t *)data;
    if (app->cpu_selector->show()) {
	string name(
	    app->cpu_selector->get_selected_display_name());
	gtk_label_set_text(GTK_LABEL(app->cm_cpu),
			   name.c_str());
	app->update_cache_levels();
    }
}

// Signal handler for various widgets that change the analysis line
// size.
//  Widget:                 Signal:
//  ---------------------   -------
//  cls_level               changed
//  rpt_line_size_cbutton   toggled
//  rpt_line_size_combo     changed
static void
update_active_line_size(GtkWidget *widget, gpointer data)
{
    app_t *app = (app_t *)data;
    using namespace acumem::library;
    if (gtk_toggle_button_get_active(
	    GTK_TOGGLE_BUTTON(app->rpt_line_size_cbutton))) {
	int ls_index = gtk_combo_box_get_active(GTK_COMBO_BOX(app->rpt_line_size_combo));

	if (ls_index > 0)
	    app->line_sizes->set_analysis_line_size(
		atoi(report_line_sizes[ls_index]));
    } else {
	int level = app->get_selected_cache_level();

	if (level != -1) {
	    const cpudb_cpu_t *cpu = app->cpu_selector->get_selected_cpu();
	    cpudb_cache_cptr_t cache = cpu->get_data_cache(level);

	    app->line_sizes->set_analysis_line_size(cache->line_size);
	}
    }
}

// Advanced sampling options button is clicked.
static void
smp_adv_button_clicked(GtkWidget *widget,
                       gpointer data)
{
    app_t *app = (app_t *)data;
    gtk_label_set_text(GTK_LABEL(app->smp_lsz_val_label),
                       app->line_sizes->line_sizes().c_str());
    gtk_dialog_run(GTK_DIALOG(app->smp_adv_dialog));
    gtk_widget_hide(app->smp_adv_dialog);
}

// Advanced sampling options button is clicked.
static void
rpt_adv_button_clicked(GtkWidget *widget,
                       gpointer data)
{
    app_t *app = (app_t *)data;
    gtk_dialog_run(GTK_DIALOG(app->rpt_adv_dialog));
    gtk_widget_hide(app->rpt_adv_dialog);
}

// Open file browser when a browse button is clicked.
static void
browse_button_clicked(GtkWidget *widget,
                      gpointer data)
{
    browse_t *browse = (browse_t *)data;
    GtkWidget *chooser = gtk_file_chooser_dialog_new(
        (browse->title + " - " PRODUCT_VERSION).c_str(), GTK_WINDOW(browse->app->main_win),
        browse->action,
        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
        GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
        NULL);
    gtk_widget_show(chooser);
    gtk_window_set_modal(GTK_WINDOW(chooser), TRUE);
    // TODO: Should only set the path if it is absolute and exists.
    string path = gtk_entry_get_text(GTK_ENTRY(browse->entry));
    if (path[0] != '/')
        path = get_working_directory() + path;
    gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(chooser), path.c_str());
    gint response = gtk_dialog_run(GTK_DIALOG(chooser));
    switch (response) {
    case GTK_RESPONSE_ACCEPT:
        gtk_entry_set_text(GTK_ENTRY(browse->entry),
                           gtk_file_chooser_get_filename(
                               GTK_FILE_CHOOSER(chooser)));
        break;
    case GTK_RESPONSE_CANCEL:
    case GTK_RESPONSE_DELETE_EVENT:
        break;
    default:
        assert(0);
    }
    gtk_widget_destroy(chooser);
}

// Bind a browse button to a window and an entry field.
static void
bind_browse_button(app_t *app,
                   GtkWidget *button,
                   GtkWidget *entry,
                   string title,
                   GtkFileChooserAction action)
{
    browse_t *browse_binding = new browse_t(app, title, entry, action);
    app->browse_bindings.push_back(browse_binding);

    g_signal_connect(G_OBJECT(button), "clicked",
                     G_CALLBACK(browse_button_clicked),
                     browse_binding);
}


struct check_child_t {
    check_child_t()
        : dialog(NULL),
          pipe(-1),
          exit(false)
        {}

    pid_t pid;
    GtkWidget *dialog;
    int pipe;
    string output;
    bool exit;
    bool error_exit;
};

static gboolean
check_child(gpointer data)
{
    check_child_t *cc = (check_child_t *)data;

    if (cc->pipe != -1) {
        ssize_t num;
        char buf[256];
        while ((num = read(cc->pipe, buf, sizeof(buf))) > 0)
            cc->output.append(buf, num);
        assert(num != -1 || errno == EAGAIN);
    }

    int status;
    int pid = waitpid(cc->pid, &status, WNOHANG);
    switch (pid) {
    case -1:
        assert(errno == EINTR);
        return TRUE;
    case 0:
        return TRUE;
    default:
        assert(pid == cc->pid);
        cc->exit = true;
        cc->error_exit = status != 0;
        gtk_dialog_response(GTK_DIALOG(cc->dialog), GTK_RESPONSE_ACCEPT);
        return FALSE;
    }
}

/* Get a sample file, either a user-specified file or one created by sampling
   an application. Returns true if successful, false if not. If successful,
   sets sample_file to the name of the sample file, and sets remove_sample_file
   to true if the sample file should be removed once it has been used, or false
   if not. */
static bool
sample(app_t *app, string *out_sample_file, bool *out_remove_sample_file)
{
    /* If the user has specified a sample file to be read, we just return
       it. It should not be removed. */
    if (gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->smp_src_file_rbutton))) {
        *out_sample_file = gtk_entry_get_text(GTK_ENTRY(app->read_smp_file_entry));
        *out_remove_sample_file = false;
        return true;
    }

    string sample_file;
    bool remove_sample_file;
    if (gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->smp_smp_file_cbutton))) {
        sample_file = gtk_entry_get_text(
            GTK_ENTRY(app->smp_smp_file_entry));
        remove_sample_file = false;
    }
    else {
        stringstream ss;
        ss << app->conf->tmpdir.string() << "/threadspotter-" << getpid() << ".smp";
        sample_file = ss.str();
        remove_sample_file = true;
    }

    struct stat stat_buf;
    if (stat(sample_file.c_str(), &stat_buf) == 0) {
        GtkWidget *dialog = gtk_message_dialog_new(
            GTK_WINDOW(app->main_win), GTK_DIALOG_MODAL,
            GTK_MESSAGE_QUESTION, GTK_BUTTONS_OK_CANCEL,
            ("Sample file '" + sample_file +
             "' already exists.\n\n Overwrite it?").c_str());
        gint response = gtk_dialog_run(GTK_DIALOG(dialog));
        gtk_widget_destroy(dialog);
        switch (response) {
        case GTK_RESPONSE_OK:
            break;
        case GTK_RESPONSE_CANCEL:
        case GTK_RESPONSE_DELETE_EVENT:
            return false;
        default:
            assert(0);
        }
    }

    /* Message to show in the dialog box while sampling. */
    const char *dialog_msg;

    /* We need to propagate LD_LIBRARY_PATH through env_helper. */
    const char *ld_library_path = getenv("LD_LIBRARY_PATH");
    if (ld_library_path == NULL)
        ld_library_path = "(null)";

    /* Arguments to the execvp call. */
    vector<string> args;
    args.push_back(paths->xterm);
    args.push_back("-T");
    args.push_back(PRODUCT_NAME " - Sampling application");
    args.push_back("-e");
    args.push_back(paths->env_helper);
    args.push_back(string(ld_library_path));
    args.push_back(paths->sample_script);

    args.push_back("--tmpdir");
    args.push_back(app->conf->tmpdir.string());

    if (gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->start_del_rbutton))) {
        args.push_back("-d");
        args.push_back(gtk_entry_get_text(GTK_ENTRY(app->start_del_sbutton)));
    }

    if (gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->start_fun_rbutton))) {
        args.push_back("--start-at-function");
        args.push_back(gtk_entry_get_text(GTK_ENTRY(app->start_fun_entry)));
        args.push_back("--start-at-ignore");
        stringstream ss;
        ss << gtk_spin_button_get_value(GTK_SPIN_BUTTON(app->start_fun_sbutton)) - 1;
        args.push_back(ss.str().c_str());
    }

    if (gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->start_addr_rbutton))) {
        args.push_back("--start-at-address");
        args.push_back(gtk_entry_get_text(GTK_ENTRY(app->start_addr_entry)));
        args.push_back("--start-at-ignore");
        stringstream ss;
        ss << gtk_spin_button_get_value(GTK_SPIN_BUTTON(app->start_addr_sbutton)) - 1;
        args.push_back(ss.str().c_str());
    }

    if (gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->stop_dur_rbutton))) {
        args.push_back("-t");
	string s = gtk_entry_get_text(GTK_ENTRY(app->stop_dur_sbutton));
        args.push_back(s);
    }

    if (gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->stop_fun_rbutton))) {
        args.push_back("--stop-at-function");
        args.push_back(gtk_entry_get_text(GTK_ENTRY(app->stop_fun_entry)));
        args.push_back("--stop-at-ignore");
        stringstream ss;
        ss << gtk_spin_button_get_value(GTK_SPIN_BUTTON(app->stop_fun_sbutton)) - 1;
        args.push_back(ss.str().c_str());
    }

    if (gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->stop_addr_rbutton))) {
        args.push_back("--stop-at-address");
        args.push_back(gtk_entry_get_text(GTK_ENTRY(app->stop_addr_entry)));
        args.push_back("--stop-at-ignore");
        stringstream ss;
        ss << gtk_spin_button_get_value(GTK_SPIN_BUTTON(app->stop_addr_sbutton)) - 1;
        args.push_back(ss.str().c_str());
    }

    if (gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->safe_stk_cbutton)))
        args.push_back("--safe-stack");

    if (!gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->burst_cbutton))) {
        args.push_back("-s");
        args.push_back(gtk_entry_get_text(GTK_ENTRY(app->smp_per_sbutton)));
    }

    if (gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->burst_cbutton))) {
        args.push_back("-b");
        args.push_back(gtk_entry_get_text(GTK_ENTRY(app->burst_len_sbutton)));

	int q_level = gtk_combo_box_get_active(GTK_COMBO_BOX(app->burst_quality_combo));
	assert(q_level >= 0);

	args.push_back("-q");
        args.push_back(burst_quality_levels[q_level]);
    }

    args.push_back("-l");
    args.push_back(app->line_sizes->line_sizes());

    args.push_back("-o");
    args.push_back(sample_file);

    if (app->line_sizes->num_line_sizes() > MAX_LINE_SIZES) {
        stringstream ss;
        ss << "At most " STRINGIFY(MAX_LINE_SIZES) " line sizes may be "
            "sampled at the same time,\n"
           << app->line_sizes->num_line_sizes() << " line sizes are selected.";
        error_dialog(app->main_win, ss.str().c_str());
        return false;
    }

    /* Check that everything is ok before we start sampling. We must do before
       we fork, since there is no good way to report errors from the child
       process. */

    if (gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->burst_cbutton))) {
        int burst_len = atoi(gtk_entry_get_text(GTK_ENTRY(app->burst_len_sbutton)));
        if (burst_len < 5) {
            error_dialog(app->main_win,
                         string("Burst sampling can not be used for shorter "
                                "runs than 5 minutes."));
            return false;
        }
    }

    if (!show_attach_options || gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->smp_src_app_launch_rbutton))) {
        /* We are about to launch and sample an application. */
        dialog_msg = "Sampling the application.\n\n"
            "Close the xterm window once the application has exited and "
            "the post-processing of the sample file has finished.";

        string cmd = gtk_entry_get_text(GTK_ENTRY(app->prg_entry));
        cmd = trim(&cmd);

        args.push_back("-r");
        args.push_back(cmd);

        string a(gtk_entry_get_text(GTK_ENTRY(app->arg_entry)));
        while (a.size() != 0) {
            string::size_type e = a.find_first_of(" ", 0);
            switch (e) {
            case string::npos:
                args.push_back(a);
                a.clear();
                break;
            case 0:
                a.erase(0, 1);
                break;
            default:
                args.push_back(a.substr(0, e));
                a.erase(0, e + 1);
                break;
            }
        }

        /* Change to the specified working directory. */
        int res = chdir(gtk_entry_get_text(GTK_ENTRY(app->wd_entry)));
        if (res != 0) {
            error_dialog(app->main_win,
                         string("Failed to change working directory to:\n") +
                         gtk_entry_get_text(GTK_ENTRY(app->wd_entry)) +
                         "\n\n" + strerror(errno));
            return false;
        }

        /* Check that a program has been specified. */
        if (cmd.empty()) {
            error_dialog(app->main_win, string("No program specified."));
            return false;
        }

        /* Check that the specified program exists. */
        gchar *abs_cmd = g_find_program_in_path(cmd.c_str());
        if (abs_cmd == NULL) {
            error_dialog(app->main_win,
                         "No program '" + cmd +"' found in search path.");
            return false;
        }
        g_free(abs_cmd);
    }
    else {
        /* We are about to sample a running application. */
        dialog_msg = "Sampling the application.\n\n"
            "Close the xterm window once you have stopped the sampling and "
            "the post-processing of the sample file has finished.";

        string pid_str(gtk_entry_get_text(GTK_ENTRY(app->pid_entry)));

        if (pid_str.empty()) {
            error_dialog(app->main_win, string("No process id specified."));
            return false;
        }

        char *end_ptr;
        long pid = strtol(pid_str.c_str(), &end_ptr, 10);
        while (isspace(*end_ptr))
            end_ptr++;
        if (*end_ptr != '\0' || pid < 0) {
            error_dialog(app->main_win, "Process id '" + pid_str +
                         "' is not a positive integer.");
            return false;
        }


        if (kill(pid, 0)) {
            error_dialog(app->main_win,
                         "There is no process with id '" + pid_str + "', or "
                         "you do not have permission to sample it.");
            return false;
        }

        args.push_back("-p");
        args.push_back(pid_str);
    }

    check_child_t cc;
    cc.pid = fork();
    if (cc.pid == 0) {
        char **argv = (char **)malloc((args.size() + 1) * sizeof(char *));
        assert(argv != NULL);
        int i = 0;
        for (vector<string>::const_iterator
                 argi = args.begin(), arge = args.end();
             argi != arge; argi++, i++)
            argv[i] = strdup(argi->c_str());
        argv[i] = NULL;

        execvp(argv[0], argv);
        assert(0);
    }
    else {
        cc.dialog = gtk_message_dialog_new(
            GTK_WINDOW(app->main_win), GTK_DIALOG_MODAL,
            GTK_MESSAGE_INFO, GTK_BUTTONS_NONE,
            dialog_msg);
        // TODO: Not available in GTK 2.4.
        // gtk_window_set_deletable(GTK_WINDOW(cc.dialog), FALSE);
        gtk_widget_show(cc.dialog);
        g_timeout_add(100, check_child, &cc);
        while (!cc.exit)
            gtk_dialog_run(GTK_DIALOG(cc.dialog));
        gtk_widget_destroy(cc.dialog);
    }

    *out_sample_file = sample_file;
    *out_remove_sample_file = remove_sample_file;
    return true;
}

static void
view_report(app_t *app, const string& report_file)
{
    int res;

    struct stat stat_buf;
    res = stat(report_file.c_str(), &stat_buf);
    if (res != 0 || !S_ISREG(stat_buf.st_mode)) {
        error_dialog(app->main_win, "Report file '" + report_file +
                     "' does not exist or is not readable.");
        return;
    }

    int pipe_fds[2];
    if (pipe(pipe_fds) == -1) {
        stringstream ss;
        ss << "Failed to create pipe: Error " << errno << ", "
           << strerror(errno) << ".";
        error_dialog(app->main_win, ss.str());
        return;
    }

    int fd_flags = fcntl(pipe_fds[1], F_GETFD);
    if (fd_flags == -1) {
        stringstream ss;
        ss << "Failed to get file descriptor flags: Error " << errno << ", "
           << strerror(errno) << ".";
        error_dialog(app->main_win, ss.str());
        return;
    }

    if (fcntl(pipe_fds[1], F_SETFD, fd_flags | FD_CLOEXEC) == -1) {
        stringstream ss;
        ss << "Failed to set file descriptor flags: Error " << errno << ", "
           << strerror(errno) << ".";
        error_dialog(app->main_win, ss.str());
        return;
    }

    pid_t pid = fork();
    if (pid == 0) {
        res = close(pipe_fds[0]);
        AM_ASSERT(res == 0);

        pid_t sid = setsid();
        AM_ASSERT(sid != -1);

        // Fork again, so that the main sample_gui process can do a waitpid()
        // and we avoid becoming a zombie process.
        pid_t pid2 = fork();
        assert(pid2 != -1);

        if (pid2 == 0) {
            const char * const argv[] = {paths->view.c_str(), "-i", report_file.c_str(), NULL};
            execvp(argv[0], (char * const *)argv);
            res = write(pipe_fds[1], &errno, sizeof(int));
            AM_ASSERT(res == sizeof(int));
            exit(1);
        }
        else
            exit(0);
    }
    else {
        res = close(pipe_fds[1]);
        AM_ASSERT(res == 0);

        int status;
        res = waitpid(pid, &status, 0);
        AM_ASSERT(res == pid && status == 0);

        int exec_errno;
        do {
            res = read(pipe_fds[0], &exec_errno, sizeof(int));
        } while (res == -1 && errno == EINTR);
        if (res != 0) {
            stringstream ss;
            ss << "Error " << exec_errno << ", "
               << strerror(exec_errno) << ".";
            error_dialog(app->main_win, ss.str());
            return;
        }
    }
}

static void
report(app_t *app) {
    int res;

    if (gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->rpt_file_rbutton))) {
        view_report(app, gtk_entry_get_text(GTK_ENTRY(app->report_file_entry)));
        return;
    }

    string sample_file;
    bool remove_sample_file;
    if (!sample(app, &sample_file, &remove_sample_file))
        return;

    // Check that the sample file is a readable file.
    struct stat stat_buf;
    res = stat(sample_file.c_str(), &stat_buf);
    if (res != 0 || !S_ISREG(stat_buf.st_mode)) {
        error_dialog(app->main_win, "Sample file '" + sample_file +
                     "' does not exist or is not readable.");
        goto done;
    }

    {
    string report_file(gtk_entry_get_text(
                           GTK_ENTRY(app->rpt_file_entry)));

    bool verbose = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(app->verbose_cbutton));

    int out_pipe_fds[2];
    res = pipe(out_pipe_fds);
    assert(res == 0);
    int in_pipe_fds[2];
    res = pipe(in_pipe_fds);
    assert(res == 0);

    check_child_t cc;
    cc.pid = fork();
    if (cc.pid == 0) {
        close(STDIN_FILENO);
        close(STDOUT_FILENO);
        close(STDERR_FILENO);
        res = dup2(in_pipe_fds[0], STDIN_FILENO);
        assert(res != -1);
        res = dup2(out_pipe_fds[1], STDOUT_FILENO);
        assert(res != -1);
        res = dup2(out_pipe_fds[1], STDERR_FILENO);
        assert(res != -1);

        vector<string> args;
        args.push_back(paths->report);

        args.push_back(string("-i"));
        args.push_back(sample_file);

        args.push_back(string("-o"));
        args.push_back(report_file.c_str());

	args.push_back("--cpu");
	args.push_back(app->cpu_selector->get_selected_id());

	int cache_level = app->get_selected_cache_level();
	assert(cache_level != -1);
	args.push_back("--level");
	args.push_back(
	    boost::lexical_cast<string>(cache_level));

	if (feature_mt_advice && feature_multisocket) {
	    args.push_back("--number-of-cpus");
	    if (gtk_toggle_button_get_active(
		    GTK_TOGGLE_BUTTON(app->rpt_num_cpus_cbutton))) {
		args.push_back(string(gtk_entry_get_text(
					  GTK_ENTRY(app->rpt_num_cpus_sbutton))));
	    } else {
		args.push_back(boost::lexical_cast<string>(
				   app->cpu_selector->get_num_cpus()));
	    }
	    
	    if (gtk_toggle_button_get_active(
		    GTK_TOGGLE_BUTTON(app->rpt_num_caches_cbutton))) {
		args.push_back(string("--number-of-caches"));
		args.push_back(string(gtk_entry_get_text(
					  GTK_ENTRY(app->rpt_num_caches_sbutton))));
	    }
	}

        if (gtk_toggle_button_get_active(
                GTK_TOGGLE_BUTTON(app->rpt_cache_size_cbutton))) {
            args.push_back(string("-c"));
            args.push_back(string(gtk_entry_get_text(
                                      GTK_ENTRY(app->rpt_cache_size_entry))));
        }

        if (gtk_toggle_button_get_active(
                GTK_TOGGLE_BUTTON(app->rpt_line_size_cbutton))) {
	    int ls_index = gtk_combo_box_get_active(GTK_COMBO_BOX(app->rpt_line_size_combo));
	    assert(ls_index >= 0);

            args.push_back(string("-l"));
	    args.push_back(report_line_sizes[ls_index]);
        }

	if (show_depth_setting) {
	    args.push_back("-d");
	    args.push_back(gtk_entry_get_text(GTK_ENTRY(app->stk_sbutton)));
	}

        args.push_back("-p");
        args.push_back(gtk_entry_get_text(GTK_ENTRY(app->thr_sbutton)));

        args.push_back(string("-r"));
        switch (gtk_combo_box_get_active(GTK_COMBO_BOX(app->rpt_repl_combo))) {
        case 0:
            args.push_back(string("random"));
            break;
        case 1:
            args.push_back(string("lru"));
            break;
        default:
            assert(0);
        }

        if (gtk_toggle_button_get_active(
                GTK_TOGGLE_BUTTON(app->dbg_cbutton))) {
            args.push_back(string("-b"));
            args.push_back(string(gtk_entry_get_text(
                                      GTK_ENTRY(app->dbg_entry))));
        }

        if (gtk_toggle_button_get_active(
                GTK_TOGGLE_BUTTON(app->dbgd_cbutton))) {
            args.push_back(string("-D"));
            args.push_back(string(gtk_entry_get_text(
                                      GTK_ENTRY(app->dbgd_entry))));
        }

        if (gtk_toggle_button_get_active(
                GTK_TOGGLE_BUTTON(app->src_cbutton))) {
            args.push_back(string("-s"));
            args.push_back(string(gtk_entry_get_text(
                                      GTK_ENTRY(app->src_entry))));
        }

        if (gtk_toggle_button_get_active(
                GTK_TOGGLE_BUTTON(app->cppfilt_cbutton))) {
	    string entry(gtk_entry_get_text(GTK_ENTRY(app->cppfilt_entry)));
	    if (entry.size() == 0)
		args.push_back(string("--c++filt"));
	    else
		args.push_back(string("--c++filt=") + entry);
	}

        if (gtk_toggle_button_get_active(
                GTK_TOGGLE_BUTTON(app->src_cbutton))) {
            args.push_back(string("-s"));
            args.push_back(string(gtk_entry_get_text(
                                      GTK_ENTRY(app->src_entry))));
        }

        if (gtk_toggle_button_get_active(
                GTK_TOGGLE_BUTTON(app->ttl_cbutton))) {
            args.push_back(string("-t"));
            args.push_back(string(gtk_entry_get_text(
                                      GTK_ENTRY(app->ttl_entry))));
        }

	if (show_blocking_expert_mode)
	    if (gtk_toggle_button_get_active(
		    GTK_TOGGLE_BUTTON(app->blocking_expert_cbutton))) {
		args.push_back(string("--no-barriers"));
	    }

        if (verbose)
            args.push_back(string("--verbose"));

#if 0
	if (gtk_toggle_button_get_active(
		GTK_TOGGLE_BUTTON(app->skip_intro_cbutton)))
	    args.push_back("--no-intro");
#endif

        char **argv = (char **)malloc((args.size() + 1) * sizeof(char *));
        assert(argv != NULL);
        int i = 0;
        for (vector<string>::const_iterator
                 argi = args.begin(), arge = args.end();
             argi != arge; argi++, i++)
            argv[i] = strdup(argi->c_str());
        argv[i] = NULL;

        execvp(argv[0], argv);
        assert(0);
    }
    else {
        int res = fcntl(out_pipe_fds[0], F_SETFL, O_NONBLOCK);
        assert(res != -1);

        cc.pipe = out_pipe_fds[0];
        cc.dialog = gtk_message_dialog_new(
            GTK_WINDOW(app->main_win), GTK_DIALOG_MODAL,
            GTK_MESSAGE_INFO, GTK_BUTTONS_NONE,
            "Generating the report, please wait.");
        // TODO: Not available in GTK 2.4.
        // gtk_window_set_deletable(GTK_WINDOW(cc.dialog), FALSE);
        gtk_widget_show(cc.dialog);
        g_timeout_add(100, check_child, &cc);
        while (!cc.exit)
            gtk_dialog_run(GTK_DIALOG(cc.dialog));
        gtk_widget_destroy(cc.dialog);

        if (cc.error_exit || verbose) {
            const char *title = cc.error_exit ? "Report generation failed" : "Report generation output";
            GtkTextBuffer *buffer = gtk_text_buffer_new(NULL);
            gtk_text_buffer_insert_at_cursor(buffer, cc.output.c_str(), cc.output.size());
            GtkWidget *view = gtk_text_view_new_with_buffer(buffer);
            gtk_text_view_set_editable(GTK_TEXT_VIEW(view), false);
            gtk_widget_set_size_request(GTK_WIDGET(view), 600, 400);
            GtkWidget *scrolled = gtk_scrolled_window_new(NULL, NULL);
            gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
                                           GTK_POLICY_AUTOMATIC,
                                           GTK_POLICY_AUTOMATIC);
            gtk_container_add(GTK_CONTAINER(scrolled), view);
            GtkWidget *dialog = gtk_dialog_new_with_buttons(
                title, GTK_WINDOW(app->main_win), GTK_DIALOG_MODAL, GTK_STOCK_OK, GTK_RESPONSE_OK, NULL);
            gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), scrolled);
            gtk_widget_show_all(dialog);
            gtk_dialog_run(GTK_DIALOG(dialog));
            gtk_widget_destroy(dialog);

            if (cc.error_exit)
                goto done;
        }
    }

    if (gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->www_cbutton))) {
        view_report(app, report_file);
    }
    }

done:
    if (remove_sample_file)
        unlink(sample_file.c_str());
}

// Run the sampler when the run button is clicked.
static void
sample_button_clicked(GtkWidget *widget,
                      gpointer data)
{
    app_t *app = (app_t *)data;

    if (!gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(app->smp_smp_file_cbutton))) {
        error_dialog(app->main_win, "No sample file name selected.");
        return;
    }

    string sample_file;
    bool remove_sample_file;
    sample(app, &sample_file, &remove_sample_file);
}

// Generate report when the report button is clicked.
static void
rpt_button_clicked(GtkWidget *widget,
                   gpointer data)
{
    report((app_t *)data);
}

// Run the sampler when the run button is clicked.
static void
smp_lsz_button_clicked(GtkWidget *widget,
                       gpointer data)
{
    app_t *app = (app_t *)data;
    app->line_sizes->show();
    gtk_label_set_text(GTK_LABEL(app->smp_lsz_val_label),
                       app->line_sizes->line_sizes().c_str());
}

static void
pid_select_button_clicked(GtkWidget *widget,
			  gpointer data)
{
    app_t *app = (app_t *)data;
    if(app->pid_selector->show()) {
	ostringstream spid;
	pid_t pid = app->pid_selector->get_selected_pid();

	spid << pid;
	gtk_entry_set_text(GTK_ENTRY(app->pid_entry), spid.str().c_str());
    }
}

static bool
check_license(const license_manager_ptr_t license)
{
    feature_set_ptr_t fs;

    try {
	fs = license->dry_run();
    } catch (const std::exception &e) {
	error_dialog(NULL, e.what());
    }
    if (!fs)
	return false;

    show_depth_setting = fs->unlimited_callstack();
    show_blocking_expert_mode =	fs->blocking_expert();
    show_attach_options = fs->attach();
    feature_mt_advice =	fs->mt();
    feature_multisocket = fs->multisocket();

    return true;
}

int
main(int argc, char **argv)
{
    // Initialize GTK.
    gtk_init(&argc, &argv);

    // Parse arguments.
    conf_t conf;
    parse_args(&conf, argc, argv);
    if (conf.exit)
        return conf.exit_code;

    {
        warning_consumer_string_ptr_t wc(new warning_consumer_string_t());

	license_manager_ptr_t license(new license_manager_t(wc));

        const bool license_ok(check_license(license));

        const string messages(wc->get_messages());
        if (!messages.empty())
            warning_dialog(NULL, messages);

	if (!license_ok)
            return 1;
    }

    paths = new paths_t();
    if(paths->sample_script.empty()) {
	error_dialog(NULL,
		     "Failed to locate the sample tool, installation is corrupted.");
	return 1;
    }
    if(paths->report.empty()) {
	error_dialog(NULL,
		     "Failed to locate the report tool, installation is corrupted.");
	return 1;
    }
    if(paths->env_helper.empty()) {
	error_dialog(NULL,
		     "Failed to locate the env_helper application, installation is corrupted.");
	return 1;
    }
    if(paths->xterm.empty()) {
	error_dialog(NULL,
		     "Failed to locate xterm, please make sure xterm is availiable in your path.");
	return 1;
    }

    app_t app(&conf);

    // Set default icon for all windows.
    gtk_window_set_default_icon_from_file(
        paths->rw_icon32x32.c_str(),
        NULL);

    // Show the window. We wait until here to do it so that the window has
    // gotten its correct size before it is shown, helping the window manager
    // to position it sensibly.
    gtk_widget_show(app.main_win); // TODO: Clean up!

    // Run the program.
    gtk_main();

    delete paths;
    return 0;
}

void
app_t::create_sampling_options_dialog()
{
    guint row = 0;

    // Create advanced sampling options expander and layout table.
    smp_adv_dialog = gtk_dialog_new_with_buttons(
        "Advanced sampling settings - " PRODUCT_VERSION,
        GTK_WINDOW(main_win),
        GTK_DIALOG_MODAL,
        GTK_STOCK_OK, GTK_RESPONSE_OK,
        NULL);

    smp_adv_table = gtk_table_new(0, 0, FALSE);
    gtk_widget_show(smp_adv_table);
    gtk_table_set_col_spacings(GTK_TABLE(smp_adv_table), COL_SPACING);
    gtk_table_set_row_spacings(GTK_TABLE(smp_adv_table), ROW_SPACING);
     gtk_box_pack_start_defaults(
        GTK_BOX(GTK_DIALOG(smp_adv_dialog)->vbox), smp_adv_table);

    // Create the sampling start label.
    start_label = gtk_label_new("Start sampling");
    gtk_widget_show(start_label);
    gtk_misc_set_alignment(GTK_MISC(start_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), start_label, 0, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the start sampling immediately input widgets.
    start_imm_rbutton = gtk_radio_button_new_with_label(NULL, "Immediately");
    gtk_widget_show(start_imm_rbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), start_imm_rbutton, 1, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the start sampling after delay input widgets.
    start_del_rbutton = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(start_imm_rbutton), "After");
    gtk_widget_show(start_del_rbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), start_del_rbutton, 1, 2, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    start_del_sbutton = gtk_spin_button_new_with_range(
        1, 1ULL << 32, 1);
    start_del_sbutton = start_del_sbutton;
    gtk_widget_show(start_del_sbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), start_del_sbutton, 2, 3, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(start_del_sbutton, start_del_rbutton);

    start_del_unit_label = gtk_label_new("seconds");
    gtk_widget_show(start_del_unit_label);
    gtk_misc_set_alignment(GTK_MISC(start_del_unit_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), start_del_unit_label, 3, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the start sampling at function input widgets.
    start_fun_rbutton = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(start_imm_rbutton), "When function");
    gtk_widget_show(start_fun_rbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), start_fun_rbutton, 1, 2, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    start_fun_entry = gtk_entry_new();
    gtk_entry_set_text(GTK_ENTRY(start_fun_entry), "main");
    gtk_widget_show(start_fun_entry);
    gtk_table_attach(GTK_TABLE(smp_adv_table), start_fun_entry, 2, 3, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(start_fun_entry, start_fun_rbutton);

    start_fun_label2 = gtk_label_new("has been called");
    gtk_widget_show(start_fun_label2);
    gtk_misc_set_alignment(GTK_MISC(start_fun_label2), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), start_fun_label2, 3, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    start_fun_sbutton = gtk_spin_button_new_with_range(
        1, 1ULL << 32, 1);
    start_fun_sbutton = start_fun_sbutton;
    gtk_widget_show(start_fun_sbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), start_fun_sbutton, 4, 5, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(start_fun_sbutton, start_fun_rbutton);

    start_fun_label3 = gtk_label_new("times");
    gtk_widget_show(start_fun_label3);
    gtk_misc_set_alignment(GTK_MISC(start_fun_label3), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), start_fun_label3, 5, 6, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the start sampling at address input widgets.
    start_addr_rbutton = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(start_imm_rbutton), "When address");
    gtk_widget_show(start_addr_rbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), start_addr_rbutton, 1, 2, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    start_addr_entry = gtk_entry_new();
    gtk_widget_show(start_addr_entry);
    gtk_table_attach(GTK_TABLE(smp_adv_table), start_addr_entry, 2, 3, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(start_addr_entry, start_addr_rbutton);

    start_addr_label2 = gtk_label_new("has been executed");
    gtk_widget_show(start_addr_label2);
    gtk_misc_set_alignment(GTK_MISC(start_addr_label2), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), start_addr_label2, 3, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    start_addr_sbutton = gtk_spin_button_new_with_range(
        1, 1ULL << 32, 1);
    start_addr_sbutton = start_addr_sbutton;
    gtk_widget_show(start_addr_sbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), start_addr_sbutton, 4, 5, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(start_addr_sbutton, start_addr_rbutton);

    start_addr_label3 = gtk_label_new("times");
    gtk_widget_show(start_addr_label3);
    gtk_misc_set_alignment(GTK_MISC(start_addr_label3), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), start_addr_label3, 5, 6, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the sampling stop label.
    stop_label = gtk_label_new("Stop sampling");
    gtk_widget_show(stop_label);
    gtk_misc_set_alignment(GTK_MISC(stop_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), stop_label, 0, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the stop sampling immediately input widgets.
    stop_imm_rbutton = gtk_radio_button_new_with_label(NULL, "When the program terminates");
    gtk_widget_show(stop_imm_rbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), stop_imm_rbutton, 1, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the stop sampling after duration input widgets.
    stop_dur_rbutton = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(stop_imm_rbutton), "After");
    gtk_widget_show(stop_dur_rbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), stop_dur_rbutton, 1, 2, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    stop_dur_sbutton = gtk_spin_button_new_with_range(
        1, 1ULL << 32, 1);
    stop_dur_sbutton = stop_dur_sbutton;
    gtk_widget_show(stop_dur_sbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), stop_dur_sbutton, 2, 3, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(stop_dur_sbutton, stop_dur_rbutton);

    stop_dur_unit_label = gtk_label_new("seconds");
    gtk_widget_show(stop_dur_unit_label);
    gtk_misc_set_alignment(GTK_MISC(stop_dur_unit_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), stop_dur_unit_label, 3, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the stop sampling at function input widgets.
    stop_fun_rbutton = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(stop_imm_rbutton), "When function");
    gtk_widget_show(stop_fun_rbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), stop_fun_rbutton, 1, 2, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    stop_fun_entry = gtk_entry_new();
    gtk_entry_set_text(GTK_ENTRY(stop_fun_entry), "main");
    gtk_widget_show(stop_fun_entry);
    gtk_table_attach(GTK_TABLE(smp_adv_table), stop_fun_entry, 2, 3, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(stop_fun_entry, stop_fun_rbutton);

    stop_fun_label2 = gtk_label_new("has been called");
    gtk_widget_show(stop_fun_label2);
    gtk_misc_set_alignment(GTK_MISC(stop_fun_label2), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), stop_fun_label2, 3, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    stop_fun_sbutton = gtk_spin_button_new_with_range(
        1, 1ULL << 32, 1);
    stop_fun_sbutton = stop_fun_sbutton;
    gtk_widget_show(stop_fun_sbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), stop_fun_sbutton, 4, 5, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(stop_fun_sbutton, stop_fun_rbutton);

    stop_fun_label3 = gtk_label_new("times");
    gtk_widget_show(stop_fun_label3);
    gtk_misc_set_alignment(GTK_MISC(stop_fun_label3), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), stop_fun_label3, 5, 6, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the stop sampling at address input widgets.
    stop_addr_rbutton = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(stop_imm_rbutton), "When address");
    gtk_widget_show(stop_addr_rbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), stop_addr_rbutton, 1, 2, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    stop_addr_entry = gtk_entry_new();
    gtk_widget_show(stop_addr_entry);
    gtk_table_attach(GTK_TABLE(smp_adv_table), stop_addr_entry, 2, 3, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(stop_addr_entry, stop_addr_rbutton);

    stop_addr_label2 = gtk_label_new("has been executed");
    gtk_widget_show(stop_addr_label2);
    gtk_misc_set_alignment(GTK_MISC(stop_addr_label2), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), stop_addr_label2, 3, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    stop_addr_sbutton = gtk_spin_button_new_with_range(
        1, 1ULL << 32, 1);
    stop_addr_sbutton = stop_addr_sbutton;
    gtk_widget_show(stop_addr_sbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), stop_addr_sbutton, 4, 5, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(stop_addr_sbutton, stop_addr_rbutton);

    stop_addr_label3 = gtk_label_new("times");
    gtk_widget_show(stop_addr_label3);
    gtk_misc_set_alignment(GTK_MISC(stop_addr_label3), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), stop_addr_label3, 5, 6, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the sampling line sizes input widgets.
    smp_lsz_label = gtk_label_new("Line sizes");
    gtk_widget_show(smp_lsz_label);
    gtk_misc_set_alignment(GTK_MISC(smp_lsz_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), smp_lsz_label, 0, 2, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    smp_lsz_val_label = gtk_label_new(line_sizes->line_sizes().c_str());
    gtk_widget_show(smp_lsz_val_label);
    gtk_misc_set_alignment(GTK_MISC(smp_lsz_val_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), smp_lsz_val_label, 2, 3, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);

    smp_lsz_select_button = gtk_button_new_with_label("Select...");
    gtk_widget_show(smp_lsz_select_button);
    gtk_table_attach(GTK_TABLE(smp_adv_table), smp_lsz_select_button, 3, 4, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    g_signal_connect(G_OBJECT(smp_lsz_select_button), "clicked",
                     G_CALLBACK(smp_lsz_button_clicked), this);

    row++;

    // Create the burst sampling input widgets.
    burst_cbutton = gtk_check_button_new_with_label(
        "Burst sampling, estimated run time");
    gtk_widget_show(burst_cbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), burst_cbutton, 0, 2, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    burst_len_sbutton = gtk_spin_button_new_with_range(
        0, 100000000, 1);
    gtk_widget_show(burst_len_sbutton);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(burst_len_sbutton), 60);
    gtk_table_attach(GTK_TABLE(smp_adv_table), burst_len_sbutton, 2, 3, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(burst_len_sbutton, burst_cbutton);

    burst_len_unit_label = gtk_label_new("minutes");
    gtk_widget_show(burst_len_unit_label);
    gtk_misc_set_alignment(GTK_MISC(burst_len_unit_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), burst_len_unit_label, 3, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    burst_quality_combo = gtk_combo_box_new_text();
    gtk_widget_show(burst_quality_combo);
    append_combo_box_text_vector(GTK_COMBO_BOX(burst_quality_combo), burst_quality_levels);
    gtk_combo_box_set_active(GTK_COMBO_BOX(burst_quality_combo), default_burst_quality_level);
    gtk_table_attach(GTK_TABLE(smp_adv_table), burst_quality_combo, 4, 5, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(burst_quality_combo, burst_cbutton);

    burst_quality_label = gtk_label_new("quality");
    gtk_widget_show(burst_quality_label);
    gtk_misc_set_alignment(GTK_MISC(burst_quality_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), burst_quality_label, 5, 6, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the sample period input widgets.
    smp_per_label = gtk_label_new("Initial sample period");
    gtk_widget_show(smp_per_label);
    gtk_misc_set_alignment(GTK_MISC(smp_per_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), smp_per_label, 0, 2, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    smp_per_sbutton = gtk_spin_button_new_with_range(
        0, 100000000, 1);
    gtk_widget_show(smp_per_sbutton);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(smp_per_sbutton), 100000);
    gtk_table_attach(GTK_TABLE(smp_adv_table), smp_per_sbutton, 2, 3, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_inverse_sensitivity_dep(smp_per_sbutton, burst_cbutton);

    smp_per_unit_label = gtk_label_new("accesses");
    gtk_widget_show(smp_per_unit_label);
    gtk_misc_set_alignment(GTK_MISC(smp_per_unit_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(smp_adv_table), smp_per_unit_label, 3, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the safe stack input widgets.
    safe_stk_cbutton = gtk_check_button_new_with_label(
        "Safe stack handling");
    gtk_widget_show(safe_stk_cbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), safe_stk_cbutton, 0, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the sampling sample file input widgets.
    smp_smp_file_cbutton = gtk_check_button_new_with_label(
        "Save sample file as");
    gtk_widget_show(smp_smp_file_cbutton);
    gtk_table_attach(GTK_TABLE(smp_adv_table), smp_smp_file_cbutton, 0, 2, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    smp_smp_file_entry = gtk_entry_new();
    gtk_widget_show(smp_smp_file_entry);
    gtk_entry_set_text(GTK_ENTRY(smp_smp_file_entry),
                       (get_working_directory() + "/sample.smp").c_str());
    gtk_table_attach(GTK_TABLE(smp_adv_table), smp_smp_file_entry, 2, 3, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(smp_smp_file_entry, smp_smp_file_cbutton);

    smp_smp_file_browse_button = gtk_button_new_with_label("Browse...");
    gtk_widget_show(smp_smp_file_browse_button);
    gtk_table_attach(GTK_TABLE(smp_adv_table), smp_smp_file_browse_button, 3, 4, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    bind_browse_button(this, smp_smp_file_browse_button, smp_smp_file_entry,
                       string("Select the name of the created sample file"),
                       GTK_FILE_CHOOSER_ACTION_SAVE);
    add_sensitivity_dep(smp_smp_file_browse_button, smp_smp_file_cbutton);

    row++;
}

void
app_t::create_report_options_dialog()
{
    guint row = 0;

    // Create advanced report options expander and layout table.
    rpt_adv_dialog = gtk_dialog_new_with_buttons(
        "Advanced report settings - " PRODUCT_VERSION,
        GTK_WINDOW(main_win),
        GTK_DIALOG_MODAL,
        GTK_STOCK_OK, GTK_RESPONSE_OK,
        NULL);

    rpt_adv_table = gtk_table_new(0, 0, FALSE);
    gtk_widget_show(rpt_adv_table);
    gtk_table_set_col_spacings(GTK_TABLE(rpt_adv_table), COL_SPACING);
    gtk_table_set_row_spacings(GTK_TABLE(rpt_adv_table), ROW_SPACING);
    gtk_box_pack_start_defaults(
        GTK_BOX(GTK_DIALOG(rpt_adv_dialog)->vbox), rpt_adv_table);

    // Create the cache size override widgets
    rpt_cache_size_cbutton = gtk_check_button_new_with_label("Cache size");
    gtk_widget_show(rpt_cache_size_cbutton);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), rpt_cache_size_cbutton,
		     0, 1, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    rpt_cache_size_entry = gtk_entry_new();
    gtk_entry_set_text(GTK_ENTRY(rpt_cache_size_entry), "64k");
    gtk_widget_show(rpt_cache_size_entry);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), rpt_cache_size_entry,
		     1, 2, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);

    rpt_cache_size_unit_label = gtk_label_new("bytes");
    gtk_misc_set_alignment(GTK_MISC(rpt_cache_size_unit_label), 0.0, 0.5);
    gtk_widget_show(rpt_cache_size_unit_label);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), rpt_cache_size_unit_label,
		     2, 3, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    add_sensitivity_dep(rpt_cache_size_entry, rpt_cache_size_cbutton);

    row++;

    // Create line size override widget
    rpt_line_size_cbutton = gtk_check_button_new_with_label("Line size");
    gtk_widget_show(rpt_line_size_cbutton);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), rpt_line_size_cbutton,
		     0, 1, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    g_signal_connect(G_OBJECT(rpt_line_size_cbutton), "toggled",
                     G_CALLBACK(update_active_line_size), this);

    rpt_line_size_combo = gtk_combo_box_new_text();
    gtk_widget_show(rpt_line_size_combo);
    append_combo_box_text_vector(GTK_COMBO_BOX(rpt_line_size_combo), report_line_sizes);
    gtk_combo_box_set_active(GTK_COMBO_BOX(rpt_line_size_combo), default_report_line_size);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), rpt_line_size_combo,
		     1, 2, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    g_signal_connect(G_OBJECT(rpt_line_size_combo), "changed",
                     G_CALLBACK(update_active_line_size), this);

    rpt_line_size_unit_label = gtk_label_new("bytes");
    gtk_misc_set_alignment(GTK_MISC(rpt_line_size_unit_label), 0.0, 0.5);
    gtk_widget_show(rpt_line_size_unit_label);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), rpt_line_size_unit_label,
		     2, 3, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    add_sensitivity_dep(rpt_line_size_combo, rpt_line_size_cbutton);

    row++;

    if (feature_mt_advice) {
	// Create number of caches override widget
	rpt_num_caches_cbutton = gtk_check_button_new_with_label("Number of caches");
	gtk_widget_show(rpt_num_caches_cbutton);
	gtk_table_attach(GTK_TABLE(rpt_adv_table), rpt_num_caches_cbutton,
			 0, 1, row, row + 1,
			 ATT_FILL, ATT_NONE, 0, 0);

	rpt_num_caches_sbutton = gtk_spin_button_new_with_range(
	    0, 256, 1);
	gtk_widget_show(rpt_num_caches_sbutton);
	gtk_table_attach(GTK_TABLE(rpt_adv_table), rpt_num_caches_sbutton,
			 1, 2, row, row + 1,
			 ATT_EXPAND_FILL, ATT_NONE, 0, 0);

	add_sensitivity_dep(rpt_num_caches_sbutton, rpt_num_caches_cbutton);

	row++;
	// Create number of cpus override widget
	rpt_num_cpus_cbutton = gtk_check_button_new_with_label("Number of CPUs");
	gtk_widget_show(rpt_num_cpus_cbutton);
	gtk_table_attach(GTK_TABLE(rpt_adv_table), rpt_num_cpus_cbutton,
			 0, 1, row, row + 1,
			 ATT_FILL, ATT_NONE, 0, 0);

	rpt_num_cpus_sbutton = gtk_spin_button_new_with_range(
	    0, 256, 1);
	gtk_widget_show(rpt_num_cpus_sbutton);
	gtk_table_attach(GTK_TABLE(rpt_adv_table), rpt_num_cpus_sbutton,
			 1, 2, row, row + 1,
			 ATT_EXPAND_FILL, ATT_NONE, 0, 0);

	add_sensitivity_dep(rpt_num_cpus_sbutton, rpt_num_cpus_cbutton);

	row++;
    }

    // Create the replacement policy input widgets.
    rpt_repl_label = gtk_label_new("Replacement policy");
    gtk_widget_show(rpt_repl_label);
    gtk_misc_set_alignment(GTK_MISC(rpt_repl_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), rpt_repl_label, 0, 1, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    rpt_repl_combo = gtk_combo_box_new_text();
    gtk_widget_show(rpt_repl_combo);
    gtk_combo_box_append_text(GTK_COMBO_BOX(rpt_repl_combo), "Random");
    gtk_combo_box_append_text(GTK_COMBO_BOX(rpt_repl_combo), "LRU");
    gtk_combo_box_set_active(GTK_COMBO_BOX(rpt_repl_combo), 0);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), rpt_repl_combo, 1, 2, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    if (show_depth_setting) {
	// Create the call stack depth input widgets.
	stk_label = gtk_label_new("Call stack depth");
	gtk_widget_show(stk_label);
	gtk_misc_set_alignment(GTK_MISC(stk_label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(rpt_adv_table), stk_label, 0, 1, row, row + 1,
			 ATT_FILL, ATT_NONE, 0, 0);

	stk_sbutton = gtk_spin_button_new_with_range(
	    0, 16, 1);
	gtk_widget_show(stk_sbutton);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(stk_sbutton), 1);
	gtk_table_attach(GTK_TABLE(rpt_adv_table), stk_sbutton, 1, 2, row, row + 1,
			 ATT_FILL, ATT_NONE, 0, 0);

	stk_unit_label = gtk_label_new("frames");
	gtk_widget_show(stk_unit_label);
	gtk_misc_set_alignment(GTK_MISC(stk_unit_label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(rpt_adv_table), stk_unit_label, 2, 3, row, row + 1,
			 ATT_FILL, ATT_NONE, 0, 0);

        row++;
    }

    // Create the debug info file input widgets.
    dbg_cbutton = gtk_check_button_new_with_label(
	"Read debug info from");
    gtk_widget_show(dbg_cbutton);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), dbg_cbutton, 0, 1, row, row + 1,
		     ATT_FILL, ATT_NONE, 0, 0);

    dbg_entry = gtk_entry_new();
    gtk_widget_show(dbg_entry);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), dbg_entry, 1, 2, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);

    dbg_browse_button = gtk_button_new_with_label("Browse...");
    gtk_widget_show(dbg_browse_button);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), dbg_browse_button,
		     2, 3, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    bind_browse_button(this, dbg_browse_button, dbg_entry,
                       string("Select binary to read debug info from"),
                       GTK_FILE_CHOOSER_ACTION_OPEN);

    add_sensitivity_dep(dbg_entry, dbg_cbutton);
    add_sensitivity_dep(dbg_browse_button, dbg_cbutton);

    row++;

    // Additional debug directory
    dbgd_cbutton = gtk_check_button_new_with_label(
	"Additional debug directories");
    gtk_widget_show(dbgd_cbutton);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), dbgd_cbutton, 0, 1, row, row + 1,
		     ATT_FILL, ATT_NONE, 0, 0);

    dbgd_entry = gtk_entry_new();
    gtk_widget_show(dbgd_entry);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), dbgd_entry, 1, 2, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);

    dbgd_browse_button = gtk_button_new_with_label("Browse...");
    gtk_widget_show(dbgd_browse_button);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), dbgd_browse_button,
		     2, 3, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    bind_browse_button(this, dbgd_browse_button, dbgd_entry,
                       string("Select directory to read debug info from"),
                       GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);

    add_sensitivity_dep(dbgd_entry, dbgd_cbutton);
    add_sensitivity_dep(dbgd_browse_button, dbgd_cbutton);

    row++;

    // Create the source dir input widgets.
    src_cbutton = gtk_check_button_new_with_label(
        "Lookup source in");
    gtk_widget_show(src_cbutton);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), src_cbutton, 0, 1, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    src_entry = gtk_entry_new();
    gtk_widget_show(src_entry);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), src_entry, 1, 2, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);

    src_browse_button = gtk_button_new_with_label("Browse...");
    gtk_widget_show(src_browse_button);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), src_browse_button, 2, 3, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    bind_browse_button(this, src_browse_button, src_entry,
                       string("Select directory to read source code from"),
                       GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);

    add_sensitivity_dep(src_entry, src_cbutton);
    add_sensitivity_dep(src_browse_button, src_cbutton);

    row++;

    // Create the c++filt input widgets.
    cppfilt_cbutton = gtk_check_button_new_with_label(
        "C++ symbol demangler");
    gtk_widget_show(cppfilt_cbutton);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), cppfilt_cbutton, 0, 1, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    cppfilt_entry = gtk_entry_new();
    gtk_widget_show(cppfilt_entry);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), cppfilt_entry, 1, 2, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);

    cppfilt_browse_button = gtk_button_new_with_label("Browse...");
    gtk_widget_show(cppfilt_browse_button);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), cppfilt_browse_button, 2, 3, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    bind_browse_button(this, cppfilt_browse_button, cppfilt_entry,
                       string("Select c++ symbol demangler filter program"),
                       GTK_FILE_CHOOSER_ACTION_OPEN);

    add_sensitivity_dep(cppfilt_entry, cppfilt_cbutton);
    add_sensitivity_dep(cppfilt_browse_button, cppfilt_cbutton);

    row++;


    // Create the issue threshold input widgets.
    thr_label = gtk_label_new("Ignore issue involving less than");
    gtk_widget_show(thr_label);
    gtk_misc_set_alignment(GTK_MISC(thr_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), thr_label, 0, 1, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    thr_sbutton = gtk_spin_button_new_with_range(
        0, 100, 1);
    gtk_widget_show(thr_sbutton);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(thr_sbutton), 1);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), thr_sbutton, 1, 2, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    thr_unit_label = gtk_label_new("% of all fetches");
    gtk_widget_show(thr_unit_label);
    gtk_misc_set_alignment(GTK_MISC(thr_unit_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), thr_unit_label, 2, 3, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the report title input widgets.
    ttl_cbutton = gtk_check_button_new_with_label("Report title");
    gtk_widget_show(ttl_cbutton);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), ttl_cbutton, 0, 1, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    ttl_entry = gtk_entry_new();
    gtk_widget_show(ttl_entry);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), ttl_entry, 1, 3, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    add_sensitivity_dep(ttl_entry, ttl_cbutton);

    row++;

#if 0
    // Create the skip-intro input widgets.
    skip_intro_cbutton = gtk_check_button_new_with_label("Skip intro page");
    gtk_widget_show(skip_intro_cbutton);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), skip_intro_cbutton, 0, 1, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;
#endif

    if (show_blocking_expert_mode) {
	// Create the blocking_expert input widgets.
	blocking_expert_cbutton = gtk_check_button_new_with_label("Block/Fusion expert mode");
	gtk_widget_show(blocking_expert_cbutton);
	gtk_table_attach(GTK_TABLE(rpt_adv_table), blocking_expert_cbutton, 0, 3, row, row + 1,
			 ATT_FILL, ATT_NONE, 0, 0);

        row++;
    }

    verbose_cbutton = gtk_check_button_new_with_label("Verbose report generation output");
    gtk_widget_show(verbose_cbutton);
    gtk_table_attach(GTK_TABLE(rpt_adv_table), verbose_cbutton, 0, 3, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;
}

app_t::app_t(const conf_t *conf)
    : conf(conf)
{
    using namespace acumem::library;

    try {
	browser = new web_browser_t();
    } catch (...) {
	browser = NULL;
    }

    warning_consumer_string_ptr_t wc(new warning_consumer_string_t());

    cpudb_ptr_t cpudb(new cpudb_t(wc));

    try {
	cpudb->load_from_file(paths->cpudb);
    } catch (xml_exception_t e) {
	stringstream ss;
	ss << "Error: Failed to load CPU database." << endl
	   << "  XML path: " << e.path() << endl
	   << "  Reason: " << e.why() << endl;

	warning_dialog(NULL, ss.str());
    }

    string messages = wc->get_messages();
    if (!messages.empty())
        warning_dialog(NULL, messages);

    // Create the main window and layout table.
    main_win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(main_win), PRODUCT_VERSION);
    g_signal_connect(G_OBJECT(main_win), "destroy",
                     G_CALLBACK(main_win_destroy), NULL);

    main_table = gtk_table_new(0, 0, FALSE);
    gtk_table_set_col_spacings(GTK_TABLE(main_table), COL_SPACING);
    gtk_table_set_row_spacings(GTK_TABLE(main_table), ROW_SPACING);
    gtk_widget_show(main_table);
    gtk_container_add(GTK_CONTAINER(main_win), main_table);

    // Initialize the line size selection dialog.
    line_sizes = new line_size_selector_t(main_win);

    // Initialize the pid selection dialog.
    pid_selector = new pid_selector_t(main_win);

    // Initialize the cpu selection dialog.
    cpu_selector = new cpu_selector_t(cpudb,
				      feature_mt_advice,
				      feature_multisocket,
				      main_win);


    // Create the advanced settings dialogs.
    create_sampling_options_dialog();
    create_report_options_dialog();

    // Keeps track of the next row number in the main table.
    guint row = 0;

    // Create the menu.
    menu_bar = gtk_menu_bar_new();
    gtk_widget_show(menu_bar);
    gtk_table_attach(GTK_TABLE(main_table), menu_bar, 0, 6, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);

    file_item = gtk_menu_item_new_with_mnemonic("_File");
    gtk_widget_show(file_item);
    gtk_menu_bar_append(GTK_MENU_BAR(menu_bar), file_item);

    file_menu = gtk_menu_new();
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(file_item), file_menu);

    quit_item = gtk_menu_item_new_with_mnemonic("_Quit");
    gtk_widget_show(quit_item);
    gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), quit_item);
    g_signal_connect(G_OBJECT(quit_item), "activate",
                     G_CALLBACK(quit_item_activate), NULL);

    row++;

    // Create sample source label.
    smp_src_label = gtk_label_new("<b>Sample source</b>");
    gtk_widget_show(smp_src_label);
    gtk_misc_set_alignment(GTK_MISC(smp_src_label), 0.0, 0.5);
    gtk_label_set_use_markup(GTK_LABEL(smp_src_label), TRUE);
    gtk_table_attach(GTK_TABLE(main_table), smp_src_label, 0, 6, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);

    row++;

    // Create sample source application radio button.
    smp_src_app_rbutton = gtk_radio_button_new_with_label(
        NULL, "Sample application");
    gtk_widget_show(smp_src_app_rbutton);
    gtk_table_attach(GTK_TABLE(main_table), smp_src_app_rbutton, 1, 6, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);

    row++;

    // Create sample source application launch radio button.
    smp_src_app_launch_rbutton = gtk_radio_button_new_with_label(
	NULL, "Launch application");
    if (show_attach_options) {

	gtk_widget_show(smp_src_app_launch_rbutton);
	gtk_table_attach(GTK_TABLE(main_table), smp_src_app_launch_rbutton, 2, 6, row, row + 1,
			 ATT_EXPAND_FILL, ATT_NONE, 0, 0);
	add_sensitivity_dep(smp_src_app_launch_rbutton, smp_src_app_rbutton);

    row++;
    }

    // Create the program input widgets.
    prg_label = gtk_label_new("Program");
    gtk_widget_show(prg_label);
    gtk_misc_set_alignment(GTK_MISC(prg_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(main_table), prg_label, 3, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(prg_label, smp_src_app_rbutton);
    add_sensitivity_dep(prg_label, smp_src_app_launch_rbutton);

    prg_entry = gtk_entry_new();
    gtk_widget_show(prg_entry);
    gtk_table_attach(GTK_TABLE(main_table), prg_entry, 4, 5, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(prg_entry, smp_src_app_rbutton);
    add_sensitivity_dep(prg_entry, smp_src_app_launch_rbutton);

    prg_browse_button = gtk_button_new_with_label("Browse...");
    gtk_widget_show(prg_browse_button);
    gtk_table_attach(GTK_TABLE(main_table), prg_browse_button, 5, 6, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    bind_browse_button(this, prg_browse_button, prg_entry,
                       string("Select binary to sample"),
                       GTK_FILE_CHOOSER_ACTION_OPEN);
    add_sensitivity_dep(prg_browse_button, smp_src_app_rbutton);
    add_sensitivity_dep(prg_browse_button, smp_src_app_launch_rbutton);

    row++;

    // Create the argument input widgets.
    arg_label = gtk_label_new("Arguments");
    gtk_widget_show(arg_label);
    gtk_misc_set_alignment(GTK_MISC(arg_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(main_table), arg_label, 3, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(arg_label, smp_src_app_rbutton);
    add_sensitivity_dep(arg_label, smp_src_app_launch_rbutton);

    arg_entry = gtk_entry_new();
    gtk_widget_show(arg_entry);
    gtk_table_attach(GTK_TABLE(main_table), arg_entry, 4, 6, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(arg_entry, smp_src_app_rbutton);
    add_sensitivity_dep(arg_entry, smp_src_app_launch_rbutton);

    row++;

    // Create the working directory input.
    wd_label = gtk_label_new("Working directory");
    gtk_widget_show(wd_label);
    gtk_misc_set_alignment(GTK_MISC(wd_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(main_table), wd_label, 3, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(wd_label, smp_src_app_rbutton);
    add_sensitivity_dep(wd_label, smp_src_app_launch_rbutton);

    wd_entry = gtk_entry_new();
    gtk_widget_show(wd_entry);
    gtk_entry_set_text(GTK_ENTRY(wd_entry), get_working_directory().c_str());
    gtk_table_attach(GTK_TABLE(main_table), wd_entry, 4, 5, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(wd_entry, smp_src_app_rbutton);
    add_sensitivity_dep(wd_entry, smp_src_app_launch_rbutton);

    wd_browse_button = gtk_button_new_with_label("Browse...");
    gtk_widget_show(wd_browse_button);
    gtk_table_attach(GTK_TABLE(main_table), wd_browse_button, 5, 6, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    bind_browse_button(this, wd_browse_button, wd_entry,
                       string("Select working directory"),
                       GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);
    add_sensitivity_dep(wd_browse_button, smp_src_app_rbutton);
    add_sensitivity_dep(wd_browse_button, smp_src_app_launch_rbutton);

    row++;

    if (show_attach_options) {
	// Create sample source application attach radio button.
	smp_src_app_attach_rbutton = gtk_radio_button_new_with_label_from_widget(
	    GTK_RADIO_BUTTON(smp_src_app_launch_rbutton), "Attach to running application");
	gtk_widget_show(smp_src_app_attach_rbutton);
	gtk_table_attach(GTK_TABLE(main_table), smp_src_app_attach_rbutton, 2, 6, row, row + 1,
			 ATT_EXPAND_FILL, ATT_NONE, 0, 0);
	add_sensitivity_dep(smp_src_app_attach_rbutton, smp_src_app_rbutton);

	row++;

	// Create the pid input widgets.
	pid_label = gtk_label_new("Pid");
	gtk_widget_show(pid_label);
	gtk_misc_set_alignment(GTK_MISC(pid_label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(main_table), pid_label, 3, 4, row, row + 1,
			 ATT_FILL, ATT_NONE, 0, 0);
	add_sensitivity_dep(pid_label, smp_src_app_rbutton);
	add_sensitivity_dep(pid_label, smp_src_app_attach_rbutton);

	pid_entry = gtk_entry_new();
	gtk_widget_show(pid_entry);
	gtk_table_attach(GTK_TABLE(main_table), pid_entry, 4, 5, row, row + 1,
			 ATT_EXPAND_FILL, ATT_NONE, 0, 0);
	add_sensitivity_dep(pid_entry, smp_src_app_rbutton);
	add_sensitivity_dep(pid_entry, smp_src_app_attach_rbutton);

	pid_select_button = gtk_button_new_with_label("Select...");
	gtk_widget_show(pid_select_button);
	gtk_table_attach(GTK_TABLE(main_table), pid_select_button, 5, 6, row, row + 1,
			 ATT_EXPAND_FILL, ATT_NONE, 0, 0);
	g_signal_connect(G_OBJECT(pid_select_button), "clicked",
			 G_CALLBACK(pid_select_button_clicked), this);
	add_sensitivity_dep(pid_select_button, smp_src_app_rbutton);
	add_sensitivity_dep(pid_select_button, smp_src_app_attach_rbutton);

	row++;
    }

    // Create advanced sampling options button.
    smp_adv_button = gtk_button_new_with_label(
        "Advanced sampling settings...");
    gtk_widget_show(smp_adv_button);
    gtk_table_attach(GTK_TABLE(main_table), smp_adv_button, 2, 6, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    g_signal_connect(G_OBJECT(smp_adv_button), "clicked",
                     G_CALLBACK(smp_adv_button_clicked), this);
    add_sensitivity_dep(smp_adv_button, smp_src_app_rbutton);

    row++;


    // Create the sample button.
    smp_button = gtk_button_new_with_label("Sample application");
    gtk_widget_show(smp_button);
    gtk_table_attach(GTK_TABLE(main_table), smp_button, 2, 6, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    g_signal_connect(G_OBJECT(smp_button), "clicked",
                     G_CALLBACK(sample_button_clicked), this);
    add_sensitivity_dep(smp_button, smp_src_app_rbutton);

    row++;

    // Create sample source file radio button.
    smp_src_file_rbutton = gtk_radio_button_new_with_label_from_widget(
        GTK_RADIO_BUTTON(smp_src_app_rbutton), "Read sample file");
    gtk_widget_show(smp_src_file_rbutton);
    gtk_table_attach(GTK_TABLE(main_table), smp_src_file_rbutton, 1, 6, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the read sample file input widgets.
    read_smp_file_label = gtk_label_new("Sample file");
    gtk_widget_show(read_smp_file_label);
    gtk_misc_set_alignment(GTK_MISC(read_smp_file_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(main_table), read_smp_file_label, 2, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(read_smp_file_label, smp_src_file_rbutton);

    read_smp_file_entry = gtk_entry_new();
    gtk_entry_set_text(GTK_ENTRY(read_smp_file_entry),
                       (string(get_working_directory()) +
                        "/sample.smp").c_str());
    gtk_widget_show(read_smp_file_entry);
    gtk_table_attach(GTK_TABLE(main_table), read_smp_file_entry, 4, 5, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(read_smp_file_entry, smp_src_file_rbutton);

    read_smp_file_browse_button = gtk_button_new_with_label("Browse...");
    gtk_widget_show(read_smp_file_browse_button);
    gtk_table_attach(GTK_TABLE(main_table), read_smp_file_browse_button, 5, 6, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    bind_browse_button(this, read_smp_file_browse_button, read_smp_file_entry,
                       string("Select sample file to read"),
                       GTK_FILE_CHOOSER_ACTION_SAVE);
    add_sensitivity_dep(read_smp_file_browse_button, smp_src_file_rbutton);

    row++;

    // Create report generation label.
    rpt_label = gtk_label_new("<b>Report generation</b>");
    gtk_widget_show(rpt_label);
    gtk_misc_set_alignment(GTK_MISC(rpt_label), 0.0, 0.5);
    gtk_label_set_use_markup(GTK_LABEL(rpt_label), TRUE);
    gtk_table_attach(GTK_TABLE(main_table), rpt_label, 0, 6, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);

    row++;

    rpt_gen_rbutton = gtk_radio_button_new_with_label(NULL, "Generate report");
    gtk_widget_show(rpt_gen_rbutton);
    gtk_table_attach(GTK_TABLE(main_table), rpt_gen_rbutton, 1, 6, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the report directory input.
    rpt_file_label = gtk_label_new("Report file");
    gtk_widget_show(rpt_file_label);
    gtk_misc_set_alignment(GTK_MISC(rpt_file_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(main_table), rpt_file_label, 2, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(rpt_file_label, rpt_gen_rbutton);

    rpt_file_entry = gtk_entry_new();
    gtk_widget_show(rpt_file_entry);
    gtk_entry_set_text(GTK_ENTRY(rpt_file_entry), "report.tsr");
    gtk_table_attach(GTK_TABLE(main_table), rpt_file_entry, 4, 5, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(rpt_file_entry, rpt_gen_rbutton);

    rpt_file_browse_button = gtk_button_new_with_label("Browse...");
    gtk_widget_show(rpt_file_browse_button);
    gtk_table_attach(GTK_TABLE(main_table), rpt_file_browse_button, 5, 6, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    bind_browse_button(this, rpt_file_browse_button, rpt_file_entry,
                       string("Select the name of the created report file"),
                       GTK_FILE_CHOOSER_ACTION_SAVE);
    add_sensitivity_dep(rpt_file_browse_button, rpt_gen_rbutton);

    row++;

    // Create the CPU model selection widgets
    cm_label = gtk_label_new("Processor model:");
    gtk_widget_show(cm_label);
    gtk_misc_set_alignment(GTK_MISC(cm_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(main_table), cm_label, 2, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(cm_label, rpt_gen_rbutton);

    cm_cpu = gtk_label_new(cpu_selector->get_selected_display_name().c_str());
    if (gtk_check_version(2, 6, 0) == NULL)
	g_object_set(G_OBJECT(cm_cpu),
		     "ellipsize", PANGO_ELLIPSIZE_MIDDLE,
		     NULL);
    gtk_widget_show(cm_cpu);
    gtk_misc_set_alignment(GTK_MISC(cm_cpu), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(main_table), cm_cpu, 4, 5, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(cm_cpu, rpt_gen_rbutton);

    cm_select = gtk_button_new_with_label("Select...");
    gtk_widget_show(cm_select);
    gtk_table_attach(GTK_TABLE(main_table), cm_select, 5, 6, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    g_signal_connect(G_OBJECT(cm_select), "clicked",
                     G_CALLBACK(cm_select_clicked), this);
    add_sensitivity_dep(cm_select, rpt_gen_rbutton);

    row++;

    // Create cache level selection widgets
    GtkCellRenderer *level_renderer = gtk_cell_renderer_text_new();
    cls_label = gtk_label_new("Cache level:");
    gtk_widget_show(cls_label);
    gtk_misc_set_alignment(GTK_MISC(cls_label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(main_table), cls_label, 2, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(cls_label, rpt_gen_rbutton);

    cls_level_store = gtk_list_store_new(N_LEVEL_COLUMNS,
					 G_TYPE_INT,
					 G_TYPE_STRING);
    cls_level = gtk_combo_box_new_with_model(GTK_TREE_MODEL(cls_level_store));
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(cls_level),
			       level_renderer,
			       false);
    gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(cls_level),
				  level_renderer,
				  "text", COLUMN_LEVEL_NAME);
    gtk_widget_show(cls_level);
    gtk_table_attach(GTK_TABLE(main_table), cls_level, 4, 5, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    g_signal_connect(G_OBJECT(cls_level), "changed",
                     G_CALLBACK(update_active_line_size), this);
    update_cache_levels();
    add_sensitivity_dep(cls_level, rpt_gen_rbutton);

    row++;

    // Create the web browser launch input widgets.
    www_cbutton = gtk_check_button_new_with_label(
        "Launch report viewer");
    gtk_widget_show(www_cbutton);
    if (browser) {
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(www_cbutton), TRUE);
    } else {
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(www_cbutton), FALSE);
	gtk_widget_set_sensitive(GTK_WIDGET(www_cbutton), FALSE);
    }
    gtk_table_attach(GTK_TABLE(main_table), www_cbutton, 2, 6, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(www_cbutton, rpt_gen_rbutton);

    row++;

    // Create advanced report options button.
    rpt_adv_button = gtk_button_new_with_label(
        "Advanced report settings...");
    gtk_widget_show(rpt_adv_button);
    gtk_table_attach(GTK_TABLE(main_table), rpt_adv_button, 2, 6, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    g_signal_connect(G_OBJECT(rpt_adv_button), "clicked",
                     G_CALLBACK(rpt_adv_button_clicked), this);
    add_sensitivity_dep(rpt_adv_button, rpt_gen_rbutton);

    row++;

    rpt_file_rbutton = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(rpt_gen_rbutton), "View existing report");
    gtk_widget_show(rpt_file_rbutton);
    gtk_table_attach(GTK_TABLE(main_table), rpt_file_rbutton, 1, 6, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    row++;

    // Create the report_file file input widgets.
    report_file_label = gtk_label_new("Report file");
    gtk_widget_show(report_file_label);
    gtk_table_attach(GTK_TABLE(main_table), report_file_label, 1, 4, row, row + 1,
                     ATT_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(report_file_label, rpt_file_rbutton);

    report_file_entry = gtk_entry_new();
    gtk_widget_show(report_file_entry);
    gtk_entry_set_text(GTK_ENTRY(report_file_entry),
                       (get_working_directory() + "/report.tsr").c_str());
    gtk_table_attach(GTK_TABLE(main_table), report_file_entry, 4, 5, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    add_sensitivity_dep(report_file_entry, rpt_file_rbutton);

    report_file_browse_button = gtk_button_new_with_label("Browse...");
    gtk_widget_show(report_file_browse_button);
    gtk_table_attach(GTK_TABLE(main_table), report_file_browse_button, 5, 6, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    bind_browse_button(this, report_file_browse_button, report_file_entry,
                       string("Select report file to view"),
                       GTK_FILE_CHOOSER_ACTION_SAVE);
    add_sensitivity_dep(report_file_browse_button, rpt_file_rbutton);

    row++;

    // Create the report button.
    rpt_button = gtk_button_new_with_label("Generate report");
    gtk_widget_show(rpt_button);
    gtk_table_attach(GTK_TABLE(main_table), rpt_button, 1, 6, row, row + 1,
                     ATT_EXPAND_FILL, ATT_NONE, 0, 0);
    g_signal_connect(G_OBJECT(rpt_button), "clicked",
                     G_CALLBACK(rpt_button_clicked), this);

    // Create column spacers. This is really ugly, but I have not found any
    // better way to set a minimum width for a table column.
    col0_spc = gtk_label_new("    ");
    gtk_widget_show(col0_spc);
    gtk_table_attach(GTK_TABLE(main_table), col0_spc, 0, 1, 0, 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    col1_spc = gtk_label_new("    ");
    gtk_widget_show(col1_spc);
    gtk_table_attach(GTK_TABLE(main_table), col1_spc, 1, 2, 0, 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    col2_spc = gtk_label_new("    ");
    gtk_widget_show(col2_spc);
    gtk_table_attach(GTK_TABLE(main_table), col2_spc, 2, 3, 0, 1,
                     ATT_FILL, ATT_NONE, 0, 0);

    // Fill in the program and arguments entry widgets with any argument from
    // the command line.
    gtk_entry_set_text(GTK_ENTRY(prg_entry), conf->program.c_str());
    string args;
    for (vector<string>::const_iterator
             ai = conf->arguments.begin(), ae = conf->arguments.end();
         ai != ae; ai++) {
        if (!args.empty())
            args += " ";
        args += *ai;
    }
    gtk_entry_set_text(GTK_ENTRY(arg_entry), args.c_str());

    // Register update_sensitivities callback for all widgets that control the
    // sensitivity of other widgets.
    set<GtkWidget *> dependee_widgets;
    BOOST_FOREACH(sensitivity_map_t::value_type dep, sensitivity_deps) {
        BOOST_FOREACH(GtkWidget *dependee, dep.second)
            dependee_widgets.insert(dependee);
    }
    BOOST_FOREACH(sensitivity_map_t::value_type dep, inverse_sensitivity_deps) {
        BOOST_FOREACH(GtkWidget *dependee, dep.second)
            dependee_widgets.insert(dependee);
    }
    BOOST_FOREACH(GtkWidget *dependee, dependee_widgets)
        g_signal_connect(G_OBJECT(dependee), "toggled",
                         G_CALLBACK(update_sensitivities), this);

    // Set up initial widget sensitivities.
    update_sensitivities(NULL, this);
}

app_t::~app_t()
{
    delete line_sizes;
    delete pid_selector;
    delete cpu_selector;

    if (browser)
	delete browser;

    for (vector<browse_t *>::iterator
             bi = browse_bindings.begin(), be = browse_bindings.end();
         bi != be; bi++)
        delete *bi;
}

void
app_t::update_cache_levels()
{
    using namespace acumem::library;
    const cpudb_cpu_t *cpu = cpu_selector->get_selected_cpu();
    const unsigned int num_cpus = cpu_selector->get_num_cpus();

    gtk_list_store_clear(cls_level_store);
    GtkTreeIter iter;
    bool iter_valid(false);
    BOOST_FOREACH(const cpudb_cache_cptr_t &cache, cpu->caches) {
	if (cache->data) {
	    stringstream ss;

	    ss << cache->name << " (";
	    if (num_cpus > 1 || cpu->cores > cache->num_cores) {
		const int caches_per_chip = cpu->cores / cache->num_cores;
		ss << num_cpus * caches_per_chip << "x";
	    }
	    ss << fnr_format_num_string(cache->size) << ")";

	    gtk_list_store_append(cls_level_store, &iter);
	    gtk_list_store_set(cls_level_store, &iter,
			       COLUMN_LEVEL_ID, cache->level,
			       COLUMN_LEVEL_NAME, ss.str().c_str(),
			       -1);
	    iter_valid = true;
	}
    }

    if (iter_valid)
	gtk_combo_box_set_active_iter(GTK_COMBO_BOX(cls_level), &iter);
}

int
app_t::get_selected_cache_level()
{
    GtkTreeIter it;

    if (!gtk_combo_box_get_active_iter(GTK_COMBO_BOX(cls_level), &it))
	return -1;

    gint level;
    gtk_tree_model_get(GTK_TREE_MODEL(cls_level_store), &it,
		       COLUMN_LEVEL_ID, &level,
		       -1);

    return level;
}
