/// @file tools/sample_gui/web_browser.cc

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#include <boost/foreach.hpp>
#include <glib.h>

#include "web_browser.hh"

static const char *browser_commands[] = {
    "htmlview",                // Browser launch script on RedHat systems
    "sensible-browser",        // Browser launch script on Debian systems
    "gnome-www-browser",       // Default browser in Gnome on Debian systems
    "x-www-browser",           // Default browser in X on Debian systems
    "firefox",
    "mozilla",
    "mozilla-firefox",
    "/usr/sfw/bin/mozilla",    // Default location for mozilla on Solaris
    NULL
};


// --- Utility functions and wrappers for broken glibmm functions -------

/**
 * Wrapper around g_spawn_async, the Glibmm wrapper is broken and
 * doesn't pass optional string parameters correctly.
 */
static void
spawn_async(const std::vector<std::string> &argv)
{
    GError *error = NULL;

    const char **argpv = new const char*[argv.size() + 1];

    for (size_t i = 0; i < argv.size(); i++) {
	argpv[i] = argv[i].c_str();
    }
    argpv[argv.size()] = NULL;

    /* This isa pretty much what Glib::spawn_async would do, with the
     * execption of the buggy handling of CWD. The wrapper uses the
     * same ugly const_cast, so we assume that it is safe. We don't
     * use Glib::spawn_async since it doesn't allow inheriting the CWD
     * due to a bug in the library.
     */
    g_spawn_async(NULL, // cwd
		  const_cast<char**>(argpv), // argv
		  NULL, // envp
		  GSpawnFlags(0),    // flags
		  NULL, // child_setup
		  NULL, // user_data
		  NULL, // child_pid
		  &error);

    delete[] argpv;

    if (error) {
	std::string error_message(error->message);
	g_error_free(error);

	throw web_browser_launch_failed_exception_t(
	    "Failed to launch browser: " +
	    error_message);
    }
}

static std::string
glib_getenv(const std::string &variable, bool &found)
{
    const gchar *value = g_getenv(variable.c_str());
    if (value) {
	std::string ret(value);
	found = true;
	return ret;
    } else {
	found = false;
	return "";
    }
}

static std::string
glib_find_program_in_path(const std::string &program)
{
    gchar *path = g_find_program_in_path(program.c_str());
    if (path) {
	std::string ret(path);
	g_free(path);
	return ret;
    } else
	return "";
}

static std::vector<std::string>
glib_shell_parse_argv(const std::string &command_line)
{
    gint argc;
    gchar **argv;

    if (g_shell_parse_argv(command_line.c_str(), &argc, &argv, NULL)) {
	std::vector<std::string> ret;
	ret.reserve(argc);
	for (gint i = 0; i < argc; i++)
	    ret.push_back(argv[i]);
	g_strfreev(argv);
	return ret;
    } else
	return std::vector<std::string>();
}

static std::vector<std::string>
split(const std::string &str, const char c)
{
    std::vector<std::string> res;

    size_t start, end;
    start = 0;

    while (start < str.length()) {
	end = str.find(c, start);
	if (end == std::string::npos)
	    end = str.length();

	res.push_back(str.substr(start, end - start));

	start = end + 1;
    }

    return res;
}

// --- web_browser_t ----------------------------------------------------

web_browser_t::web_browser_t()
{
    if (!lookup_in_environ() &&
	!lookup_known())
	throw web_browser_not_found_exception_t();
}


bool
web_browser_t::lookup_in_environ()
{
    bool found_env;
    std::string browser_env(glib_getenv("BROWSER", found_env));
    if(!found_env || browser_env.empty())
	return false;

    std::vector<std::string> browsers(split(browser_env, ':'));

    BOOST_FOREACH(const std::string &browser, browsers) {
	std::vector<std::string> argv(
	    glib_shell_parse_argv(browser));
	if (argv.empty())
	    continue;

	// Try to locate the binary
	binary = glib_find_program_in_path(argv[0]);
	if (!binary.empty()) {
	    arguments.resize(argv.size() - 1);
	    copy(++argv.begin(),
		 argv.end(),
		 arguments.begin());

	    return true;
	}
    }

    return false;
}

bool
web_browser_t::lookup_known()
{
    // If no environment variable specifies a working browser, try to
    // look for a known browser in the users PATH.
    for(int i = 0; browser_commands[i]; i++) {
	binary = glib_find_program_in_path(browser_commands[i]);
	if(!binary.empty()) {
	    // Pass the URL as the first and only argument
	    arguments.resize(1);
	    arguments[0] = "%s";
	    return true;
	}
    }
    return false;
}

std::string
web_browser_t::setup_argument(const std::string &arg,
			      const std::string &url,
			      bool &url_substituted)
{
    size_t res_len(0);

    // Calculate argument length
    for (size_t i = 0; i < arg.length(); i++) {
	// Skip the next character if we encounter a '%' and add the
	// substitution length to res_len.
	if (arg[i] == '%' && i < arg.length() - 1) {
	    switch (arg[++i]) {
	    case 's':
		res_len += url.length();
		break;

	    case '%':
		res_len++;
		break;

	    default:
		break;
	    }
	} else
	    res_len++;
    }

    // Do the substitutions and save the results in res.
    std::string res(res_len, '\0');
    for (size_t res_pos = 0, i = 0;
	 i < arg.length();
	 i++) {

	if (arg[i] == '%' && i < arg.length() - 1) {
	    switch (arg[++i]) {
	    case 's':
		// Substitute %s with the URL
		res.replace(res_pos, url.length(), url);
		res_pos += url.length();
		url_substituted |= true;
		break;

	    case '%':
		// Substitute %% with '%'
		res[res_pos++] = '%';
		break;

	    default:
		// Unknown substitutions are be skipped
		break;
	    }
	} else
	    res[res_pos++] = arg[i];
    }

    return res;
}


void
web_browser_t::open_url(const std::string &url) const
{
    std::vector<std::string> argv;
    bool url_substituted(false);
    argv.reserve(arguments.size() + 2);
    argv.push_back(setup_argument(binary, url, url_substituted));

    BOOST_FOREACH(const std::string &arg, arguments)
	argv.push_back(setup_argument(arg, url, url_substituted));

    if (!url_substituted)
	argv.push_back(url);

    spawn_async(argv);
}
