/// @file tools/sampler_sim/filter.cc

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#include "sampler_sim.hh"

#include <boost/foreach.hpp>

#include <assert.h>

filter_t::filter_t(unsigned int num_entries,
                   unsigned int granularity,
                   bool separate_threads,
                   bool separate_access_types)
    : num_entries(num_entries),
      granularity(granularity),
      separate_threads(separate_threads),
      separate_access_types(separate_access_types)
{
    read_access_types.insert(Access_Type_RD);
    write_access_types.insert(Access_Type_WR);
    write_access_types.insert(Access_Type_RDWR);
    write_access_types.insert(Access_Type_PF_AMD);
    write_access_types.insert(Access_Type_PF_AMD_W);
    write_access_types.insert(Access_Type_PF_Intel_NTA);
    write_access_types.insert(Access_Type_PF_Intel_T0);
    write_access_types.insert(Access_Type_PF_Intel_T1);
    write_access_types.insert(Access_Type_PF_Intel_T2);
    all_access_types.insert(Access_Type_RD);
    all_access_types.insert(Access_Type_WR);
    all_access_types.insert(Access_Type_RDWR);
    all_access_types.insert(Access_Type_PF_AMD);
    all_access_types.insert(Access_Type_PF_AMD_W);
    all_access_types.insert(Access_Type_PF_Intel_NTA);
    all_access_types.insert(Access_Type_PF_Intel_T0);
    all_access_types.insert(Access_Type_PF_Intel_T1);
    all_access_types.insert(Access_Type_PF_Intel_T2);
}

void
filter_t::add(wp_t *wp)
{
    unsigned long first_index = wp->address / granularity;
    unsigned long last_index = (wp->address + wp->size - 1) / granularity;
    for (unsigned long index = first_index;
         index >= first_index && index <= last_index;
         index++)
        entries[index % num_entries].insert(wp);
}

void
filter_t::remove(wp_t *wp)
{
    unsigned long first_index = wp->address / granularity;
    unsigned long last_index = (wp->address + wp->size - 1) / granularity;
    for (unsigned long index = first_index;
         index >= first_index && index <= last_index;
         index++)
        entries[index % num_entries].erase(wp);
}

bool
filter_t::check(
    unsigned long long access_id,
    unsigned long address,
    unsigned int size,
    access_type_t access_type,
    unsigned long access_tid,
    tid_set_t *all_tids)
{
    tid_set_t t;
    t.insert(access_tid);
    tid_set_t *tids;
    access_type_set_t *access_types;
    events_t log_events;

    if (separate_threads)
        tids = &t;
    else
        tids = all_tids;

    if (separate_access_types)
        if (read_access_types.find(access_type) != read_access_types.end())
            access_types = &read_access_types;
        else if (write_access_types.find(access_type) != write_access_types.end())
            access_types = &write_access_types;
        else {
            assert(0);
            while (1); // Avoid warning on Solaris.
        }
    else
        access_types = &all_access_types;

    unsigned long first_index = address / granularity;
    unsigned long last_index = (address + size - 1) / granularity;
    for (unsigned long index = first_index;
         index >= first_index && index <= last_index;
         index++) {
        filter_entry_t *e = &entries[index % num_entries];
        for (filter_entry_t::iterator wpi = e->begin(), wpe = e->end();
             wpi != wpe; wpi++) {
            wp_t *wp = *wpi;

            BOOST_FOREACH(unsigned long tid, *tids) {
                BOOST_FOREACH(access_type_t access_type, *access_types)
                    if (wp->event(access_id, &log_events, tid, access_type, 0, LINE_SIZE, false))
                        return true;
            }
        }
    }

    return false;
}
