/// @file tools/sampler_sim/sampler_sim.cc

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#include "sampler_sim.hh"
#include "library/random_lcg/random_lcg.h"
#include "sampler/src/atf_reader.c"
#include <boost/foreach.hpp>
#include <boost/program_options.hpp>
#include <iostream>

using namespace boost;
namespace po = boost::program_options;

struct conf_t {
    conf_t()
        : separate_threads(false),
          separate_access_types(false),
          filter_entries(1U << 20),
          filter_granularity(1U << 9),
          sample_period(1000) {
    }

    string atf_file;
    bool atf_stdin;

    bool st_sampler;
    bool mp_sampler;
    bool ww_sampler;
    bool fs_sampler;
    bool lt_sampler;

    bool separate_threads;
    bool separate_access_types;

    unsigned int filter_entries;
    unsigned int filter_granularity;

    unsigned long sample_period;

    bool debug;
};

static bool
parse_args(conf_t *conf, int argc, char **argv)
{
    po::options_description opt_desc_visible("Options");
    opt_desc_visible.add_options()
        ("help,h", "Print this help message.")
        ("atf-file", po::value<string>(&conf->atf_file),
         "Read ATF trace from file.")
        ("stdin", "Read ATF trace from stdin.")
        ("st-sampler", "Single threaded sampler.")
        ("mp-sampler", "StatCache MP sampler.")
        ("ww-sampler", "Write-write MP sampler.")
        ("fs-sampler", "False sharing MT sampler.")
        ("lt-sampler", "Line tracing MT sampler.")
        ("separate-threads", "Separate watchpoint filters for different threads.")
        ("separate-access-types", "Separate watchpoint filters for reads and writes.")
        ("filter-entries", po::value<unsigned int>(&conf->filter_entries),
         "Number of watchpoint filter entries, 1M by default.")
        ("filter-granularity", po::value<unsigned int>(&conf->filter_granularity),
         "Watchpoint filter granularity in bytes, 512 by default.")
        ("sample-period", po::value<unsigned long>(&conf->sample_period),
         "Sample period, 1000 by default.")
        ("debug", "Enable debug printouts.")
        ;

    po::options_description opt_desc_invisible("Options");
    opt_desc_invisible.add_options()
        ("positional-args", "")
        ;

    po::options_description opt_desc_all("Options");
    opt_desc_all.add(opt_desc_visible).add(opt_desc_invisible);

    po::positional_options_description pos_desc;
    pos_desc.add("positional-args", -1);

    po::variables_map vm;
    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(opt_desc_all).positional(pos_desc).run(), vm);
        po::notify(vm);
    }
    catch (po::unknown_option &e) {
        cerr << "Error: " << e.what() << endl;
        cerr << opt_desc_visible << endl;
        return true;
    }

    if (vm.count("help") > 0) {
        cerr << opt_desc_visible << endl;
        exit(0);
    }

    if (vm.count("atf-file") + vm.count("stdin") < 1) {
        cerr << "Error: no input specified" << endl;
        return true;
    }

    if (vm.count("atf-file") + vm.count("stdin") > 1) {
        cerr << "Error: multiple inputs specified" << endl;
        return true;
    }

    conf->atf_stdin = vm.count("stdin") > 0;

    if (vm.count("st-sampler") +
        vm.count("mp-sampler") +
        vm.count("ww-sampler") +
        vm.count("fs-sampler") +
        vm.count("lt-sampler")
        < 1) {
        cerr << "Error: no sampler type specified" << endl;
        return true;
    }

    if (vm.count("st-sampler") +
        vm.count("mp-sampler") +
        vm.count("ww-sampler") +
        vm.count("fs-sampler") +
        vm.count("lt-sampler")
        > 1) {
        cerr << "Error: multiple sampler types specified" << endl;
        return true;
    }

    conf->st_sampler = vm.count("st-sampler") > 0;
    conf->mp_sampler = vm.count("mp-sampler") > 0;
    conf->ww_sampler = vm.count("ww-sampler") > 0;
    conf->fs_sampler = vm.count("fs-sampler") > 0;
    conf->lt_sampler = vm.count("lt-sampler") > 0;

    conf->separate_threads = vm.count("separate-threads") > 0;
    conf->separate_access_types = vm.count("separate-access-types") > 0;

    conf->debug = vm.count("debug") > 0;

    if (vm.count("positional-args") > 0) {
        cerr << "Error: no positional arguments allowed" << endl;
        return true;
    }

    return false;
}

random_lcg_t *lcg;

void
lcg_srandom(int seed)
{
    lcg = random_lcg_new(seed);
}

void
lcg_free()
{
    random_lcg_delete(lcg);
}

uint32_t
lcg_random()
{
    return random_lcg_next(lcg);
}

unsigned int
access_size(const struct atf_packet *packet)
{
    switch (packet->header & 0x1f) {
    case ATF_OP_MEM_RO:
    case ATF_OP_MEM_RW:
    case ATF_OP_MEM_WO:
    case ATF_OP_EXT | (ATF_XOP_MEM_MOVNT << 2):
    case ATF_OP_EXT | (ATF_XOP_MEM_PREFETCH << 2):
    case ATF_OP_EXT | (ATF_XOP_MEM_CLFLUSH << 2):
        return packet->header >> 8;
    default:
        assert(0);
        while (1); // Avoid warning on Solaris.
    }
}

access_type_t
access_type(const struct atf_packet *packet)
{
    switch (packet->header & 0x1f) {
    case ATF_OP_EXT | (ATF_XOP_MEM_CLFLUSH << 2): /* XXX: loses info */
    case ATF_OP_MEM_RO:
        return Access_Type_RD;
    case ATF_OP_MEM_WO:
        return Access_Type_WR;
    case ATF_OP_MEM_RW:
        return Access_Type_RDWR;
    case ATF_OP_EXT | (ATF_XOP_MEM_PREFETCH << 2):
        switch ((packet->header >> 5) & 0x7) {
        case ATF_PREFETCH_AMD:
            return Access_Type_PF_AMD;
        case ATF_PREFETCH_AMD_W:
            return Access_Type_PF_AMD_W;
        case ATF_PREFETCH_INTEL_NTA:
            return Access_Type_PF_Intel_NTA;
        case ATF_PREFETCH_INTEL_T0:
            return Access_Type_PF_Intel_T0;
        case ATF_PREFETCH_INTEL_T1:
            return Access_Type_PF_Intel_T1;
        case ATF_PREFETCH_INTEL_T2:
            return Access_Type_PF_Intel_T2;
        default:
            assert(0);
            while (1); // Avoid warning on Solaris.
        }
    case ATF_OP_EXT | (ATF_XOP_MEM_MOVNT << 2) | (1 << 5):
        return Access_Type_RD_NT;
    case ATF_OP_EXT | (ATF_XOP_MEM_MOVNT << 2) | (0 << 5):
        return Access_Type_WR_NT;
    default:
        assert(0);
        while (1); // Avoid warning on Solaris.
    }
}

bool
is_prefetch(access_type_t access_type)
{
    switch (access_type) {
    case Access_Type_RD:
    case Access_Type_WR:
    case Access_Type_RDWR:
        return false;
    case Access_Type_PF_AMD:
    case Access_Type_PF_AMD_W:
    case Access_Type_PF_Intel_NTA:
    case Access_Type_PF_Intel_T0:
    case Access_Type_PF_Intel_T1:
    case Access_Type_PF_Intel_T2:
        return true;
    default:
        assert(0);
        while (1); // Avoid warning on Solaris.
    }
}

bool
is_write(access_type_t access_type)
{
    switch (access_type) {
    case Access_Type_WR:
    case Access_Type_RDWR:
    case Access_Type_WR_NT:
        return true;
    case Access_Type_RD:
    case Access_Type_PF_AMD:
    case Access_Type_PF_AMD_W:
    case Access_Type_PF_Intel_NTA:
    case Access_Type_PF_Intel_T0:
    case Access_Type_PF_Intel_T1:
    case Access_Type_PF_Intel_T2:
    case Access_Type_RD_NT:
    case Access_Type_F:
        return false;
    default:
        assert(0);
        while (1); // Avoid warning on Solaris.
    }
}

bool
is_write_or_write_prefetch(access_type_t access_type)
{
    switch (access_type) {
    case Access_Type_WR:
    case Access_Type_RDWR:
    case Access_Type_PF_AMD_W:
    case Access_Type_WR_NT:
        return true;
    case Access_Type_RD:
    case Access_Type_PF_AMD:
    case Access_Type_PF_Intel_NTA:
    case Access_Type_PF_Intel_T0:
    case Access_Type_PF_Intel_T1:
    case Access_Type_PF_Intel_T2:
    case Access_Type_RD_NT:
    case Access_Type_F:
        return false;
    default:
        assert(0);
        while (1); // Avoid warning on Solaris.
    }
}

bool
is_write_nt(access_type_t access_type)
{
    switch (access_type) {
    case Access_Type_WR_NT:
        return true;
    case Access_Type_RD:
    case Access_Type_WR:
    case Access_Type_RDWR:
    case Access_Type_PF_AMD:
    case Access_Type_PF_AMD_W:
    case Access_Type_PF_Intel_NTA:
    case Access_Type_PF_Intel_T0:
    case Access_Type_PF_Intel_T1:
    case Access_Type_PF_Intel_T2:
    case Access_Type_RD_NT:
    case Access_Type_F:
        return false;
    default:
        assert(0);
        while (1); // Avoid warning on Solaris.
    }
}

bool
is_flush(access_type_t access_type)
{
    switch (access_type) {
    case Access_Type_F:
        return true;
    case Access_Type_RD:
    case Access_Type_WR:
    case Access_Type_RDWR:
    case Access_Type_PF_AMD:
    case Access_Type_PF_AMD_W:
    case Access_Type_PF_Intel_NTA:
    case Access_Type_PF_Intel_T0:
    case Access_Type_PF_Intel_T1:
    case Access_Type_PF_Intel_T2:
    case Access_Type_RD_NT:
    case Access_Type_WR_NT:
        return false;
    default:
        assert(0);
        while (1); // Avoid warning on Solaris.
    }
}

bool
is_always_invalidation(access_type_t access_type)
{
    switch (access_type) {
    case Access_Type_WR:
    case Access_Type_RDWR:
    case Access_Type_WR_NT:
    case Access_Type_F:
        return true;
    case Access_Type_RD:
    case Access_Type_PF_AMD:
    case Access_Type_PF_AMD_W:
    case Access_Type_PF_Intel_NTA:
    case Access_Type_PF_Intel_T0:
    case Access_Type_PF_Intel_T1:
    case Access_Type_PF_Intel_T2:
    case Access_Type_RD_NT:
        return false;
    default:
        assert(0);
        while (1); // Avoid warning on Solaris.
    }
}

bool
is_maybe_invalidation(access_type_t access_type)
{
    switch (access_type) {
    case Access_Type_WR:
    case Access_Type_RDWR:
    case Access_Type_PF_AMD_W:
    case Access_Type_WR_NT:
    case Access_Type_F:
        return true;
    case Access_Type_RD:
    case Access_Type_PF_AMD:
    case Access_Type_PF_Intel_NTA:
    case Access_Type_PF_Intel_T0:
    case Access_Type_PF_Intel_T1:
    case Access_Type_PF_Intel_T2:
    case Access_Type_RD_NT:
        return false;
    default:
        assert(0);
        while (1); // Avoid warning on Solaris.
    }
}

bool
is_always_downgrade(access_type_t access_type)
{
    switch (access_type) {
    case Access_Type_RD:
    case Access_Type_RD_NT:
        return true;
    case Access_Type_WR:
    case Access_Type_RDWR:
    case Access_Type_PF_AMD:
    case Access_Type_PF_AMD_W:
    case Access_Type_PF_Intel_NTA:
    case Access_Type_PF_Intel_T0:
    case Access_Type_PF_Intel_T1:
    case Access_Type_PF_Intel_T2:
    case Access_Type_WR_NT:
    case Access_Type_F:
        return false;
    default:
        assert(0);
        while (1); // Avoid warning on Solaris.
    }
}

bool
access_type_seen(access_type_t access_type,
                 unsigned int *access_types,
                 bool side_effects)
{
    unsigned int seen_mask;
    unsigned int see_mask;

    switch (access_type) {
    case Access_Type_WR:
    case Access_Type_RDWR:
        seen_mask = 1U << Access_Type_WR;
        see_mask = (unsigned)-1;
        break;
    case Access_Type_RD:
    case Access_Type_PF_AMD:
    case Access_Type_PF_AMD_W:
    case Access_Type_PF_Intel_NTA:
    case Access_Type_PF_Intel_T0:
    case Access_Type_PF_Intel_T1:
    case Access_Type_PF_Intel_T2:
        seen_mask = 1U << access_type;
        see_mask = seen_mask;
        break;
    default:
        assert(0);
        while (1); // Avoid warning on Solaris.
    }

    bool seen = !!(*access_types & seen_mask);
    if (side_effects)
        *access_types |= see_mask;
    return seen;
}

static const char *
access_type_name(access_type_t access_type)
{
    switch (access_type) {
    case Access_Type_RD:
        return "read";
    case Access_Type_WR:
        return "write";
    case Access_Type_RDWR:
        return "read-write";
    case Access_Type_PF_AMD:
        return "prefetch";
    case Access_Type_PF_AMD_W:
        return "prefetchw";
    case Access_Type_PF_Intel_NTA:
        return "prefetchnta";
    case Access_Type_PF_Intel_T0:
        return "prefetcht0";
    case Access_Type_PF_Intel_T1:
        return "prefetcht1";
    case Access_Type_PF_Intel_T2:
        return "prefetcht2";
    default:
        assert(0);
        while (1); // Avoid warning on Solaris.
    }
}

wp_t::wp_t(
    unsigned long address,
    unsigned int size,
    unsigned long long wp_id,
    unsigned long long wp_group_id)
    : address(address),
      size(size),
      wp_id(wp_id),
      wp_group_id(wp_group_id),
      hits(0),
      events(0)
{
}

wp_t::~wp_t()
{
}

sampler_t::sampler_t(unsigned long sample_period,
                     filter_t *filter)
    : sample_period(sample_period),
      filter(filter),
      accesses(0),
      groups(0),
      watchpoints(0),
      hits(0),
      events(0),
      kills(0)
{
}

sampler_t::~sampler_t()
{
}

bool
sampler_t::sample_access()
{
    return ((double)lcg_random() / RANDOM_LCG_MAX) < (1.0 / sample_period);
}

void
sampler_t::access(unsigned long address,
                  unsigned int size,
                  access_type_t access_type,
                  unsigned long tid,
                  bool debug)
{
    unsigned long long access_id = accesses++;

    events_t log_events;

    set<unsigned long long> kill_groups;

    if (all_tids.find(tid) == all_tids.end())
        all_tids.insert(tid);

    if (filter->check(access_id, address, size, access_type, tid, &all_tids)) {
        hits++;

        for (unsigned long next_laddress, laddress = address;
             laddress >= address && laddress <= address + size - 1;
             laddress = next_laddress) {

            next_laddress = (laddress + LINE_SIZE) / LINE_SIZE * LINE_SIZE;
            unsigned int lsize = min(next_laddress, address + size) - laddress;

            for (unsigned long byte_index = laddress;
                 byte_index >= laddress && byte_index <= laddress + lsize - 1;
                 byte_index++) {
                wp_vector_map_t::iterator byte_wpi = byte_wps.find(byte_index);
                if (byte_wpi != byte_wps.end()) {
                    wp_vector_t *wps = &byte_wpi->second;
                    for (wp_vector_t::size_type i = 0; i < wps->size(); i++) {
                        wp_t *wp = wps->at(i);
                        if (++(wp->hits) == HIT_LIMIT)
                            kill_groups.insert(wp->wp_group_id);
                        if (wp->event(access_id, &log_events, tid, access_type, 0, 1, true)) {
                            if (++(wp->events) == EVENT_LIMIT)
                                kill_groups.insert(wp->wp_group_id);

                            if (debug)
                                printf("Event from byte WP %lld on %s\n", wp->wp_id, access_type_name(access_type));
                        }
                        if (wp->remove(tid, access_type)) {
                            if (debug)
                                printf("Ended byte WP %lld on %s\n", wp->wp_id, access_type_name(access_type));

                            filter->remove(wp);
                            delete wp;
                            wps->at(i) = wps->back();
                            wps->pop_back();
                            i--;
                        }
                    }
                }
            }

            unsigned long line_index = laddress / LINE_SIZE;
            wp_vector_map_t::iterator line_wpi = line_wps.find(line_index);
            if (line_wpi != line_wps.end()) {
                wp_vector_t *wps = &line_wpi->second;
                for (wp_vector_t::size_type i = 0; i < wps->size(); i++) {
                    wp_t *wp = wps->at(i);
                    if (++(wp->hits) == HIT_LIMIT)
                        kill_groups.insert(wp->wp_group_id);
                    if (wp->event(access_id, &log_events, tid, access_type, laddress % LINE_SIZE, lsize, true)) {
                        if (++(wp->events) == EVENT_LIMIT)
                            kill_groups.insert(wp->wp_group_id);

                        if (debug)
                            printf("Event from line WP %lld on %s\n", wp->wp_id, access_type_name(access_type));
                    }
                    if (wp->remove(tid, access_type)) {
                        if (debug)
                            printf("Ended line WP %lld on %s\n", wp->wp_id, access_type_name(access_type));

                        filter->remove(wp);
                        delete wp;
                        wps->at(i) = wps->back();
                        wps->pop_back();
                        i--;
                    }
                }
            }
        }

        for (unsigned long next_paddress, paddress = address;
             paddress >= address && paddress <= address + size - 1;
             paddress = next_paddress) {

            next_paddress = (paddress + PAGE_SIZE) / PAGE_SIZE * PAGE_SIZE;
            unsigned int psize = min(next_paddress, address + size) - paddress;

            unsigned long page_index = paddress / PAGE_SIZE;
            wp_vector_map_t::iterator page_wpi = page_wps.find(page_index);
            if (page_wpi != page_wps.end()) {
                wp_vector_t *wps = &page_wpi->second;
                for (wp_vector_t::size_type i = 0; i < wps->size(); i++) {
                    wp_t *wp = wps->at(i);
                    if (++(wp->hits) == HIT_LIMIT)
                        kill_groups.insert(wp->wp_group_id);
                    if (wp->event(access_id, &log_events, tid, access_type, paddress % PAGE_SIZE, psize, true)) {
                        if (++(wp->events) == EVENT_LIMIT)
                            kill_groups.insert(wp->wp_group_id);

                        if (debug)
                            printf("Event from page WP %lld on %s\n", wp->wp_id, access_type_name(access_type));
                    }
                    if (wp->remove(tid, access_type)) {
                        if (debug)
                            printf("Ended page WP %lld on %s\n", wp->wp_id, access_type_name(access_type));

                        filter->remove(wp);
                        delete wp;
                        wps->at(i) = wps->back();
                        wps->pop_back();
                        i--;
                    }
                }
            }
        }
    }

    if (sample_access()) {
        events += 2;

        unsigned long long wp_group_id = groups++;

        for (unsigned long next_paddress, paddress = address;
             paddress >= address && paddress <= address + size - 1;
             paddress = next_paddress) {

            next_paddress = (paddress + PAGE_SIZE) / PAGE_SIZE * PAGE_SIZE;

            unsigned long page_index = paddress / PAGE_SIZE;
            wp_t *page_wp = sample_page_access(page_index, access_type, tid, watchpoints++, wp_group_id);
            page_wps[page_index].push_back(page_wp);

            if (debug)
                printf("Started page WP %lld on %s\n", page_wp->wp_id, access_type_name(access_type));
        }

        for (unsigned long next_laddress, laddress = address;
             laddress >= address && laddress <= address + size - 1;
             laddress = next_laddress) {

            next_laddress = (laddress + LINE_SIZE) / LINE_SIZE * LINE_SIZE;
            unsigned int lsize = min(next_laddress, address + size)  - laddress;

            unsigned long line_index = laddress / LINE_SIZE;
            wp_t *line_wp = sample_line_access(line_index, access_type, tid, watchpoints++, wp_group_id);
            line_wps[line_index].push_back(line_wp);

            if (debug)
                printf("Started line WP %lld on %s\n", line_wp->wp_id, access_type_name(access_type));

            for (unsigned long byte_index = laddress;
                 byte_index >= laddress && byte_index <= laddress + lsize - 1;
                 byte_index++) {
                wp_t *byte_wp = sample_byte_access(byte_index, access_type, tid, line_wp, watchpoints++, wp_group_id);
                if (byte_wp != NULL) {
                    byte_wps[byte_index].push_back(byte_wp);

                    if (debug)
                        printf("Started byte WP %lld on %s\n", byte_wp->wp_id, access_type_name(access_type));
                }
            }
        }
    }

    events += log_events.size();

    BOOST_FOREACH(unsigned long long wp_group_id, kill_groups) {
        kills++;

        BOOST_FOREACH(wp_vector_map_t::value_type &v, byte_wps)
            for (wp_vector_t::size_type i = 0; i < v.second.size(); i++) {
                wp_t *wp = v.second.at(i);
                if (wp->wp_group_id == wp_group_id) {
                    if (debug)
                        printf("Killed byte WP %lld\n", wp->wp_id);

                    filter->remove(wp);
                    delete wp;
                    v.second.at(i) = v.second.back();
                    v.second.pop_back();
                    i--;
                }
        }

        BOOST_FOREACH(wp_vector_map_t::value_type &v, line_wps)
            for (wp_vector_t::size_type i = 0; i < v.second.size(); i++) {
                wp_t *wp = v.second.at(i);
                if (wp->wp_group_id == wp_group_id) {
                    if (debug)
                        printf("Killed line WP %lld\n", wp->wp_id);

                    filter->remove(wp);
                    delete wp;
                    v.second.at(i) = v.second.back();
                    v.second.pop_back();
                    i--;
                }
        }

        BOOST_FOREACH(wp_vector_map_t::value_type &v, page_wps)
            for (wp_vector_t::size_type i = 0; i < v.second.size(); i++) {
                wp_t *wp = v.second.at(i);
                if (wp->wp_group_id == wp_group_id) {
                    if (debug)
                        printf("Killed page WP %lld\n", wp->wp_id);

                    filter->remove(wp);
                    delete wp;
                    v.second.at(i) = v.second.back();
                    v.second.pop_back();
                    i--;
                }
        }
    }
}

int
main(int argc, char **argv)
{
    conf_t conf;

    if (parse_args(&conf, argc, argv))
        return 1;

    FILE *f;

    if (conf.atf_stdin)
        f = stdin;
    else
        if ((f = fopen(conf.atf_file.c_str(), "r")) == NULL) {
            cerr << "Error: failed to open ATF file " << conf.atf_file << endl;
            return 1;
        }

    filter_t filter(conf.filter_entries,
                    conf.filter_granularity,
                    conf.separate_threads,
                    conf.separate_access_types);

    sampler_t *sampler;
    if (conf.st_sampler)
        sampler = new st_sampler_t(conf.sample_period, &filter);
    else if (conf.mp_sampler)
        sampler = new mp_sampler_t(conf.sample_period, &filter);
    else if (conf.ww_sampler)
        sampler = new ww_sampler_t(conf.sample_period, &filter);
    else if (conf.fs_sampler)
        sampler = new fs_sampler_t(conf.sample_period, &filter);
    else if (conf.lt_sampler)
        sampler = new lt_sampler_t(conf.sample_period, &filter);
    else {
        assert(0);
        while (1); // Avoid warning on Solaris.
    }

    struct atf_header header;
    struct atf_reader reader;

    if (atf_read_init(f, &reader, &header) < 0) {
        cerr << "Error: error reading ATF file " << conf.atf_file << endl;
        return 1;
    }

    lcg_srandom(0);

    int atf_ret;
    struct atf_packet packet;

    //unsigned long num_packets = 0;
    while ((atf_ret = atf_read_packet(&reader, &packet)) == 1) {
        //if (++num_packets % 1000 == 0)
        //cout << "Processed " << num_packets << " packets." << endl;

        // Handle memory mapping packets.
        switch (packet.header & 0x1F) {
        case (ATF_OP_EXT|(ATF_XOP_MMAP<<2)):
        case (ATF_OP_EXT|(ATF_XOP_MUNMAP<<2)):
        case (ATF_OP_EXT|(ATF_XOP_MREMAP<<2)):
            continue;
        }

        // Handle call stack packets.
        if ((packet.header & 0xFF) == (ATF_OP_EXT|(ATF_XOP_SHADOW<<2)))
            continue;

        // Handle memory access and prefetch packets.
        sampler->access((unsigned long)packet.u.mem.daddr,
                        access_size(&packet),
                        access_type(&packet),
                        packet.u.mem.tid,
                        conf.debug);
    }

    if (atf_ret == -1)
        cerr << "Error: error reading ATF file " << conf.atf_file << endl;

    if (!conf.atf_stdin)
        if (fclose(f) != 0) {
            cerr << "Error: failed to close ATF file " << conf.atf_file << endl;
            return 1;
        }

    cout << "Accesses:         " << sampler->accesses << endl
         << "Groups:           " << sampler->groups << endl
         << "Watchpoints:      " << sampler->watchpoints << endl
         << "Hits:             " << sampler->hits << endl
         << "Events:           " << sampler->events << endl
         << "Kills:            " << sampler->kills << endl;

    delete sampler;

	lcg_free();

    return 0;
}
