/// @file tools/sampler_sim/sampler_sim.hh

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#include <assert.h>
#include <set>
#include <map>
#include <vector>
#include <boost/foreach.hpp>

using namespace std;

#define LINE_SIZE 64UL
#define PAGE_SIZE 4096UL

#define HIT_LIMIT 1000000
#define EVENT_LIMIT 1000

typedef enum {
    Access_Type_RD,
    Access_Type_WR,
    Access_Type_RDWR,
    Access_Type_PF_AMD,
    Access_Type_PF_AMD_W,
    Access_Type_PF_Intel_NTA,
    Access_Type_PF_Intel_T0,
    Access_Type_PF_Intel_T1,
    Access_Type_PF_Intel_T2,
    Access_Type_RD_NT,
    Access_Type_WR_NT,
    Access_Type_F,
    Num_Access_Types
} access_type_t;

bool is_prefetch(access_type_t access_type);
bool is_write(access_type_t access_type);
bool is_write_or_write_prefetch(access_type_t access_type);
bool is_write_nt(access_type_t access_type);
bool is_flush(access_type_t access_type);
bool is_always_invalidation(access_type_t access_type);
bool is_maybe_invalidation(access_type_t access_type);
bool is_always_downgrade(access_type_t access_type);

bool access_type_seen(access_type_t access_type,
                      unsigned int *access_types,
                      bool side_effects);

/// Pair of watchpoint group id and access id.
typedef pair<unsigned long long, unsigned long long> event_t;
typedef set<event_t> events_t;

struct wp_t {
    unsigned long address;
    unsigned int size;

    unsigned long long wp_id;
    unsigned long long wp_group_id;

    unsigned int hits;
    unsigned int events;

    wp_t(
        unsigned long address,
        unsigned int size,
        unsigned long long wp_id,
        unsigned long long wp_group_id);

    virtual ~wp_t();

    virtual bool event(
        unsigned long long access_id,
        events_t *events,
        unsigned long access_tid,
        access_type_t access_type,
        unsigned int offset,
        unsigned int size,
        bool side_effects) = 0;

    virtual bool remove(
        unsigned long access_tid,
        access_type_t access_type) = 0;
};

typedef set<access_type_t> access_type_set_t;

typedef set<unsigned long> tid_set_t;

struct filter_t {
    unsigned int num_entries;
    unsigned int granularity;
    bool separate_threads;
    bool separate_access_types;

    typedef set<wp_t *> filter_entry_t;
    typedef map<unsigned int, filter_entry_t> filter_entries_t;
    filter_entries_t entries;

    access_type_set_t read_access_types;
    access_type_set_t write_access_types;
    access_type_set_t all_access_types;

    filter_t(unsigned int num_entries,
             unsigned int granularity,
             bool separate_threads,
             bool separate_access_types);

    void add(wp_t *wp);
    void remove(wp_t *wp);
    bool check(
        unsigned long long access_id,
        unsigned long address,
        unsigned int size,
        access_type_t access_type,
        unsigned long access_tid,
        tid_set_t *all_tids);
};

typedef map<unsigned int, unsigned int> tid_access_types_t;

struct sampler_t {
    typedef vector<wp_t *> wp_vector_t;
    typedef map<unsigned long, wp_vector_t> wp_vector_map_t;

    unsigned long sample_period;
    filter_t *filter;

    wp_vector_map_t page_wps;
    wp_vector_map_t line_wps;
    wp_vector_map_t byte_wps;

    unsigned long long accesses;
    unsigned long long groups;
    unsigned long long watchpoints;
    unsigned long long hits;
    unsigned long long events;
    unsigned long long kills;

    tid_set_t all_tids;

    sampler_t(unsigned long sample_period,
              filter_t *filter);
    virtual ~sampler_t();

    bool sample_access();


    virtual wp_t *sample_page_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        unsigned long long wp_id,
        unsigned long long wp_group_id) = 0;

    virtual wp_t *sample_line_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        unsigned long long wp_id,
        unsigned long long wp_group_id) = 0;

    virtual wp_t *sample_byte_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        wp_t *line_wp,
        unsigned long long wp_id,
        unsigned long long wp_group_id) = 0;

    void access(unsigned long address,
                unsigned int size,
                access_type_t access_type,
                unsigned long access_tid,
                bool debug);
};

struct st_sampler_t : public sampler_t {
    struct st_wp_t : public wp_t {
        st_wp_t(
            unsigned long address,
            unsigned int size,
            unsigned long long wp_id,
            unsigned long long wp_group_id);

        bool event(
            unsigned long long access_id,
            events_t *events,
            unsigned long access_tid,
            access_type_t access_type,
            unsigned int offset,
            unsigned int size,
            bool side_effects);

        bool remove(
            unsigned long access_tid,
            access_type_t access_type);
    };

    st_sampler_t(unsigned long sample_period,
                 filter_t *filter);


    wp_t *sample_page_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        unsigned long long wp_id,
        unsigned long long wp_group_id);

    wp_t *sample_line_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        unsigned long long wp_id,
        unsigned long long wp_group_id);

    wp_t *sample_byte_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        wp_t *line_wp,
        unsigned long long wp_id,
        unsigned long long wp_group_id);
};

struct mp_sampler_t : public sampler_t {
    struct page_wp_t : public wp_t {
        unsigned long tid;
        unsigned int seen_access_types;

        page_wp_t(
            unsigned long tid,
            unsigned long address,
            unsigned long long wp_id,
            unsigned long long wp_group_id);

        bool event(
            unsigned long long access_id,
            events_t *events,
            unsigned long access_tid,
            access_type_t access_type,
            unsigned int offset,
            unsigned int size,
            bool side_effects);

        bool remove(
            unsigned long access_tid,
            access_type_t access_type);
    };

    struct line_wp_t : public wp_t {
        unsigned long tid;
        tid_access_types_t seen_access_types;
        bool shadow;
        unsigned int live_byte_wps;

        line_wp_t(
            unsigned long tid,
            unsigned long address,
            unsigned long long wp_id,
            unsigned long long wp_group_id);

        bool event(
            unsigned long long access_id,
            events_t *events,
            unsigned long access_tid,
            access_type_t access_type,
            unsigned int offset,
            unsigned int size,
            bool side_effects);

        bool remove(
            unsigned long access_tid,
            access_type_t access_type);
    };

    struct byte_wp_t : public wp_t {
        unsigned long tid;
        tid_access_types_t seen_access_types;
        line_wp_t *line_wp;

        byte_wp_t(
            unsigned long tid,
            line_wp_t *line_wp,
            unsigned long address,
            unsigned long long wp_id,
            unsigned long long wp_group_id);

        ~byte_wp_t();

        bool event(
            unsigned long long access_id,
            events_t *events,
            unsigned long access_tid,
            access_type_t access_type,
            unsigned int offset,
            unsigned int size,
            bool side_effects);

        bool remove(
            unsigned long access_tid,
            access_type_t access_type);
    };

    mp_sampler_t(unsigned long sample_period,
                 filter_t *filter);

    wp_t *sample_page_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        unsigned long long wp_id,
        unsigned long long wp_group_id);

    wp_t *sample_line_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        unsigned long long wp_id,
        unsigned long long wp_group_id);

    wp_t *sample_byte_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        wp_t *line_wp,
        unsigned long long wp_id,
        unsigned long long wp_group_id);
};

struct ww_sampler_t : public sampler_t {
    struct page_wp_t : public wp_t {
        unsigned long tid;
        unsigned int seen_access_types;

        page_wp_t(
            unsigned long tid,
            unsigned long address,
            unsigned long long wp_id,
            unsigned long long wp_group_id);

        bool event(
            unsigned long long access_id,
            events_t *events,
            unsigned long access_tid,
            access_type_t access_type,
            unsigned int offset,
            unsigned int size,
            bool side_effects);

        bool remove(
            unsigned long access_tid,
            access_type_t access_type);
    };

    struct line_wp_t : public wp_t {
        unsigned long tid;
        tid_access_types_t seen_access_types;
        bool want_invalidations;
        bool write;
        bool shadow;
        unsigned int live_byte_wps;

        line_wp_t(
            unsigned long tid,
            bool write,
            unsigned long address,
            unsigned long long wp_id,
            unsigned long long wp_group_id);

        bool event(
            unsigned long long access_id,
            events_t *events,
            unsigned long access_tid,
            access_type_t access_type,
            unsigned int offset,
            unsigned int size,
            bool side_effects);

        bool remove(
            unsigned long access_tid,
            access_type_t access_type);
    };

    struct byte_wp_t : public wp_t {
        unsigned long tid;
        tid_access_types_t seen_access_types;
        bool want_invalidations;
        line_wp_t *line_wp;

        byte_wp_t(
            unsigned long tid,
            line_wp_t *line_wp,
            unsigned long address,
            unsigned long long wp_id,
            unsigned long long wp_group_id);

        ~byte_wp_t();

        bool event(
            unsigned long long access_id,
            events_t *events,
            unsigned long access_tid,
            access_type_t access_type,
            unsigned int offset,
            unsigned int size,
            bool side_effects);

        bool remove(
            unsigned long access_tid,
            access_type_t access_type);
    };

    ww_sampler_t(unsigned long sample_period,
                 filter_t *filter);

    wp_t *sample_page_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        unsigned long long wp_id,
        unsigned long long wp_group_id);

    wp_t *sample_line_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        unsigned long long wp_id,
        unsigned long long wp_group_id);

    wp_t *sample_byte_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        wp_t *line_wp,
        unsigned long long wp_id,
        unsigned long long wp_group_id);
};

struct fs_sampler_t : public sampler_t {
    struct first_last_t {
        /// Pair of access type and thread id.
        typedef pair<access_type_t, unsigned int> last_combo_t;

        /// Set of access type - thread id pairs.
        typedef set<last_combo_t> seen_t;

        /// Map from access type - thread id pairs to access ids.
        typedef map<last_combo_t, unsigned long long> last_t;

        bool first;

        seen_t seen;

        last_t last;

        first_last_t(
            bool first);

        bool record(
            unsigned long long access_id,
            unsigned int access_tid,
            access_type_t access_type,
            bool already_logged);

        void log(
            events_t *events,
            unsigned long long wp_group_id,
            bool side_effects);
    };

    struct touched_bytes_t {
        char touched[LINE_SIZE];

        touched_bytes_t();

        bool report_event(
            unsigned long offset,
            unsigned int size,
            bool side_effects);

        bool covers(
            touched_bytes_t &b);
    };

    struct fs_state_t {
        touched_bytes_t touched_bytes;
        bool downgrade;

        fs_state_t()
            : downgrade(false) {
        }
    };

    typedef map<unsigned long, fs_state_t> tid_fs_state_t;

    struct page_wp_t : public wp_t {
        unsigned long start_tid;
        access_type_t start_access_type;
        tid_access_types_t seen_access_types;

        page_wp_t(
            unsigned long access_tid,
            access_type_t access_type,
            unsigned long address,
            unsigned long long wp_id,
            unsigned long long wp_group_id);

        bool event(
            unsigned long long access_id,
            events_t *events,
            unsigned long access_tid,
            access_type_t access_type,
            unsigned int offset,
            unsigned int size,
            bool side_effects);

        bool remove(
            unsigned long access_tid,
            access_type_t access_type);
    };

    struct line_wp_t : public wp_t {
        unsigned long start_tid;
        access_type_t start_access_type;
        unsigned int live_wp_types;
        unsigned int live_byte_wps;
        first_last_t last;
        touched_bytes_t touched_bytes;
        tid_fs_state_t tid_fs_state;

        line_wp_t(
            unsigned long access_tid,
            access_type_t access_type,
            unsigned long address,
            unsigned long long wp_id,
            unsigned long long wp_group_id);

        bool event(
            unsigned long long access_id,
            events_t *events,
            unsigned long access_tid,
            access_type_t access_type,
            unsigned int offset,
            unsigned int size,
            bool side_effects);

        bool remove(
            unsigned long access_tid,
            access_type_t access_type);
    };

    struct byte_wp_t : public wp_t {
        unsigned long start_tid;
        unsigned int live_wp_types;
        first_last_t first_last;
        line_wp_t *line_wp;

        byte_wp_t(
            unsigned long access_tid,
            access_type_t access_type,
            line_wp_t *line_wp,
            unsigned long address,
            unsigned long long wp_id,
            unsigned long long wp_group_id);
        ~byte_wp_t();

        bool event(
            unsigned long long access_id,
            events_t *events,
            unsigned long access_tid,
            access_type_t access_type,
            unsigned int offset,
            unsigned int size,
            bool side_effects);

        bool remove(
            unsigned long access_tid,
            access_type_t access_type);
    };

    fs_sampler_t(unsigned long sample_period,
                 filter_t *filter);

    wp_t *sample_page_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        unsigned long long wp_id,
        unsigned long long wp_group_id);

    wp_t *sample_line_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        unsigned long long wp_id,
        unsigned long long wp_group_id);

    wp_t *sample_byte_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        wp_t *line_wp,
        unsigned long long wp_id,
        unsigned long long wp_group_id);
};

struct lt_sampler_t : public sampler_t {
    struct page_wp_t : public wp_t {
        unsigned long start_tid;
        access_type_t start_access_type;
        tid_access_types_t seen_access_types;

        page_wp_t(
            unsigned long access_tid,
            access_type_t access_type,
            unsigned long address,
            unsigned long long wp_id,
            unsigned long long wp_group_id);

        bool event(
            unsigned long long access_id,
            events_t *events,
            unsigned long access_tid,
            access_type_t access_type,
            unsigned int offset,
            unsigned int size,
            bool side_effects);

        bool remove(
            unsigned long access_tid,
            access_type_t access_type);
    };

    struct line_wp_t : public wp_t {
        unsigned long start_tid;
        access_type_t start_access_type;
        unsigned int live_wp_types;
        unsigned int live_byte_wps;

        line_wp_t(
            unsigned long access_tid,
            access_type_t access_type,
            unsigned long address,
            unsigned long long wp_id,
            unsigned long long wp_group_id);

        bool event(
            unsigned long long access_id,
            events_t *events,
            unsigned long access_tid,
            access_type_t access_type,
            unsigned int offset,
            unsigned int size,
            bool side_effects);

        bool remove(
            unsigned long access_tid,
            access_type_t access_type);
    };

    struct byte_wp_t : public wp_t {
        unsigned long start_tid;
        unsigned int live_wp_types;
        line_wp_t *line_wp;

        byte_wp_t(
            unsigned long access_tid,
            access_type_t access_type,
            line_wp_t *line_wp,
            unsigned long address,
            unsigned long long wp_id,
            unsigned long long wp_group_id);

        ~byte_wp_t();

        bool event(
            unsigned long long access_id,
            events_t *events,
            unsigned long access_tid,
            access_type_t access_type,
            unsigned int offset,
            unsigned int size,
            bool side_effects);

        bool remove(
            unsigned long access_tid,
            access_type_t access_type);
    };

    lt_sampler_t(unsigned long sample_period,
                 filter_t *filter);

    wp_t *sample_page_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        unsigned long long wp_id,
        unsigned long long wp_group_id);

    wp_t *sample_line_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        unsigned long long wp_id,
        unsigned long long wp_group_id);

    wp_t *sample_byte_access(
        unsigned long index,
        access_type_t access_type,
        unsigned long access_tid,
        wp_t *line_wp,
        unsigned long long wp_id,
        unsigned long long wp_group_id);
};
