/// @file tools/sampler_sim/sampler_sim_fs.cc

// Copyright (c) 2008-2009 Acumem AB.  All rights reserved.

#include "sampler_sim.hh"
#include <boost/foreach.hpp>
#include <assert.h>

#define A_R     0x01
#define A_W     0x02
#define A_RW    0x04
#define A_Pr    0x08
#define A_Pw    0x10
#define A_Rnt   0x20
#define A_Wnt   0x40
#define A_F     0x80

#define L_Start_R   (0 | A_R |              A_Pr |        A_Rnt | A_Wnt | A_F)
#define L_Start_W   (0 | A_R | A_W | A_RW | A_Pr | A_Pw | A_Rnt | A_Wnt | A_F)
#define L_Start_RW  (0 | A_R | A_W | A_RW | A_Pr | A_Pw | A_Rnt | A_Wnt | A_F)
#define L_Start_Pr  (0 | A_R |              A_Pr |        A_Rnt | A_Wnt | A_F)
#define L_Start_Pw  (0 | A_R |              A_Pr | A_Pw | A_Rnt | A_Wnt | A_F)
#define L_Start_Rnt (0 | A_R |              A_Pr |        A_Rnt | A_Wnt | A_F)
#define L_Start_Wnt (0 | A_R | A_W | A_RW | A_Pr | A_Pw | A_Rnt | A_Wnt | A_F)
#define L_Start_F   (0 | A_R | A_W | A_RW | A_Pr | A_Pw | A_Rnt | A_Wnt | A_F)

#define L_End_R     (0 | A_R |              A_Pr |        A_Rnt | A_Wnt | A_F)
#define L_End_W     (0 | A_R | A_W | A_RW | A_Pr | A_Pw | A_Rnt | A_Wnt | A_F)
#define L_End_RW    (0 | A_R | A_W | A_RW | A_Pr | A_Pw | A_Rnt | A_Wnt | A_F)
#define L_End_Pr    (0                                                       )
#define L_End_Pw    (0                                                       )
#define L_End_Rnt   (0 | A_R |              A_Pr |        A_Rnt | A_Wnt | A_F)
#define L_End_Wnt   (0 | A_R | A_W | A_RW | A_Pr | A_Pw | A_Rnt | A_Wnt | A_F)
#define L_End_F     (0 | A_R | A_W | A_RW | A_Pr | A_Pw | A_Rnt | A_Wnt | A_F)

#define B_Start_R   (0 | A_R |              A_Rnt | A_Wnt)
#define B_Start_W   (0 |       A_W |                A_Wnt)
#define B_Start_RW  (0 | A_R | A_W | A_RW | A_Rnt | A_Wnt)
#define B_Start_Rnt (0 | A_R |              A_Rnt | A_Wnt)
#define B_Start_Wnt (0 | A_R | A_W | A_RW | A_Rnt | A_Wnt)

#define B_End_R     (0 | A_R |              A_Rnt | A_Wnt)
#define B_End_W     (0 |       A_W |                A_Wnt)
#define B_End_RW    (0 | A_R | A_W | A_RW | A_Rnt | A_Wnt)
#define B_End_Rnt   (0 | A_R |              A_Rnt | A_Wnt)
#define B_End_Wnt   (0 | A_R | A_W | A_RW | A_Rnt | A_Wnt)

static map<access_type_t, unsigned int> line_start_types;
static map<access_type_t, unsigned int> line_end_types;
static map<access_type_t, unsigned int> byte_start_types;
static map<access_type_t, unsigned int> byte_end_types;

static unsigned int
get_types(const map<access_type_t, unsigned int> &types, access_type_t access_type)
{
    map<access_type_t, unsigned int>::const_iterator i = types.find(access_type);
    assert(i != types.end());
    return i->second;
}

fs_sampler_t::first_last_t::first_last_t(
    bool first)
    : first(first)
{
}

bool
fs_sampler_t::first_last_t::record(
    unsigned long long access_id,
    unsigned int access_tid,
    access_type_t access_type,
    bool already_logged)
{
    bool event = false;
    last_combo_t combo = make_pair(access_type, access_tid);

    if (first &&
        !already_logged &&
        seen.find(combo) == seen.end()) {
        event = true;
        already_logged = true;
    }
    seen.insert(combo);

    if (already_logged)
        last.erase(combo);
    else
        last[combo] = access_id;

    return event;
}

void
fs_sampler_t::first_last_t::log(
    events_t *events,
    unsigned long long wp_group_id,
    bool side_effects)
{
    BOOST_FOREACH(last_t::value_type v, last)
        events->insert(make_pair(wp_group_id, v.second));
    if (side_effects)
        last.clear();
}

fs_sampler_t::touched_bytes_t::touched_bytes_t()
{
    memset(touched, 0, sizeof(touched));
}

bool
fs_sampler_t::touched_bytes_t::report_event(
    unsigned long offset,
    unsigned int size,
    bool side_effects) {
    bool report = false;
    for (unsigned int i = offset; i < offset + size; i++)
        if (touched[i] == 0) {
            report = true;
            if (side_effects)
                touched[i] = 1;
        }
    return report;
}

bool
fs_sampler_t::touched_bytes_t::covers(
    touched_bytes_t &b) {
    for (unsigned int i = 0; i < LINE_SIZE; i++)
        if (b.touched[i] && !touched[i])
            return false;
    return true;
}

fs_sampler_t::page_wp_t::page_wp_t(
    unsigned long access_tid,
    access_type_t access_type,
    unsigned long address,
    unsigned long long wp_id,
    unsigned long long wp_group_id)
    : wp_t(address, PAGE_SIZE, wp_id, wp_group_id),
      start_tid(access_tid),
      start_access_type(access_type)
{
}

bool
fs_sampler_t::page_wp_t::event(
    unsigned long long access_id,
    events_t *events,
    unsigned long access_tid,
    access_type_t access_type,
    unsigned int offset,
    unsigned int size,
    bool side_effects)
{
    // Log unseen access types from each thread until we have seen a read or
    // write access from the thread (writes set all access type to seen).
    bool hit =
        !access_type_seen(Access_Type_RD, &seen_access_types[access_tid], false) &&
        !access_type_seen(access_type, &seen_access_types[access_tid], side_effects);
    if (hit)
        events->insert(make_pair(wp_group_id, access_id));
    return hit;
}

bool
fs_sampler_t::page_wp_t::remove(
    unsigned long access_tid,
    access_type_t access_type)
{
    // Terminate the watchpoint on non-prefetch accesses (or prefetch accesses
    // of the same type that started the sample) from the thread itself.
    return
        access_tid == start_tid &&
        (!is_prefetch(access_type) ||
         access_type == start_access_type);
}

fs_sampler_t::line_wp_t::line_wp_t(
    unsigned long access_tid,
    access_type_t access_type,
    unsigned long address,
    unsigned long long wp_id,
    unsigned long long wp_group_id)
    : wp_t(address, LINE_SIZE, wp_id, wp_group_id),
      start_tid(access_tid),
      start_access_type(access_type),
      live_wp_types(get_types(line_start_types, access_type)),
      live_byte_wps(0),
      last(false)
{
}

bool
fs_sampler_t::line_wp_t::event(
    unsigned long long access_id,
    events_t *events,
    unsigned long access_tid,
    access_type_t access_type,
    unsigned int offset,
    unsigned int size,
    bool side_effects)
{
    bool hit = false;

    if (live_wp_types != 0) {
        hit = true;
        if (side_effects) {
            if (access_tid == start_tid) {
                if (access_type == start_access_type)
                    live_wp_types = 0;
                else
                    live_wp_types &= ~get_types(line_end_types, access_type);

                if (live_wp_types == 0)
                    touched_bytes.report_event(offset, size, side_effects);
            }
            else
                tid_fs_state[access_tid].touched_bytes.report_event(offset, size, side_effects);
        }
    }
    else if (!tid_fs_state.empty()) {
        hit = true;
        if (side_effects) {
            if (is_flush(access_type) ||
                is_write_nt(access_type))
                tid_fs_state.clear();

            set<unsigned long> remove_tids;

            if (access_tid == start_tid) {
                touched_bytes.report_event(offset, size, side_effects);

                BOOST_FOREACH(tid_fs_state_t::value_type &s, tid_fs_state) {
                    fs_state_t &fs = s.second;

                    if ((is_write_or_write_prefetch(access_type) && fs.downgrade) ||
                        touched_bytes.covers(fs.touched_bytes))
                        remove_tids.insert(s.first);
                }
            }
            else
                if (tid_fs_state.find(access_tid) != tid_fs_state.end()) {
                    if (is_always_invalidation(access_type))
                        remove_tids.insert(access_tid);
                    if (is_always_downgrade(access_type))
                        tid_fs_state[access_tid].downgrade = true;
                }

            BOOST_FOREACH(unsigned long tid, remove_tids)
                tid_fs_state.erase(tid_fs_state.find(tid));
        }
    }

    if (last.record(access_id, access_tid, access_type, hit))
        hit = true;

    if (hit)
        events->insert(make_pair(wp_group_id, access_id));
    return hit;
}

bool
fs_sampler_t::line_wp_t::remove(
    unsigned long access_tid,
    access_type_t access_type)
{
    return
        live_wp_types == 0 &&
        tid_fs_state.empty() &&
        live_byte_wps == 0;
}

fs_sampler_t::byte_wp_t::byte_wp_t(
    unsigned long access_tid,
    access_type_t access_type,
    line_wp_t *line_wp,
    unsigned long address,
    unsigned long long wp_id,
    unsigned long long wp_group_id)
    : wp_t(address, 1, wp_id, wp_group_id),
      start_tid(access_tid),
      live_wp_types(get_types(byte_start_types, access_type)),
      first_last(true),
      line_wp(line_wp)
{
    line_wp->live_byte_wps++;
}

fs_sampler_t::byte_wp_t::~byte_wp_t()
{
    line_wp->live_byte_wps--;
}

bool
fs_sampler_t::byte_wp_t::event(
    unsigned long long access_id,
    events_t *events,
    unsigned long access_tid,
    access_type_t access_type,
    unsigned int offset,
    unsigned int size,
    bool side_effects)
{
    bool hit = false;

    /* Log the last accesses to the line and the byte if the byte sample is
       going to be terminated. */
    if (access_tid == start_tid &&
        !is_prefetch(access_type) &&
        (live_wp_types & ~get_types(byte_end_types, access_type)) == 0) {
        hit = true;
        line_wp->last.log(events, wp_group_id, side_effects);
        first_last.log(events, wp_group_id, side_effects);
    }
    else
        hit = first_last.record(access_id, access_tid, access_type, false);
    if (hit)
        events->insert(make_pair(wp_group_id, access_id));
    return hit;
}

bool
fs_sampler_t::byte_wp_t::remove(
    unsigned long access_tid,
    access_type_t access_type)
{
    if (access_tid == start_tid &&
        !is_prefetch(access_type))
        live_wp_types &= ~get_types(byte_end_types, access_type);

    // Terminate the watchpoint when all watchpoint types have been terminated.
    return live_wp_types == 0;
}

fs_sampler_t::fs_sampler_t(unsigned long sample_period,
                           filter_t *filter)
    : sampler_t(sample_period, filter)
{
    line_start_types[Access_Type_RD]           = L_Start_R;
    line_start_types[Access_Type_WR]           = L_Start_W;
    line_start_types[Access_Type_RDWR]         = L_Start_RW;
    line_start_types[Access_Type_PF_AMD]       = L_Start_Pr;
    line_start_types[Access_Type_PF_AMD_W]     = L_Start_Pw;
    line_start_types[Access_Type_PF_Intel_NTA] = L_Start_Pr;
    line_start_types[Access_Type_PF_Intel_T0]  = L_Start_Pr;
    line_start_types[Access_Type_PF_Intel_T1]  = L_Start_Pr;
    line_start_types[Access_Type_PF_Intel_T2]  = L_Start_Pr;

    line_end_types[Access_Type_RD]             = L_End_R;
    line_end_types[Access_Type_WR]             = L_End_W;
    line_end_types[Access_Type_RDWR]           = L_End_RW;
    line_end_types[Access_Type_PF_AMD]         = L_End_Pr;
    line_end_types[Access_Type_PF_AMD_W]       = L_End_Pw;
    line_end_types[Access_Type_PF_Intel_NTA]   = L_End_Pr;
    line_end_types[Access_Type_PF_Intel_T0]    = L_End_Pr;
    line_end_types[Access_Type_PF_Intel_T1]    = L_End_Pr;
    line_end_types[Access_Type_PF_Intel_T2]    = L_End_Pr;

    byte_start_types[Access_Type_RD]           = B_Start_R;
    byte_start_types[Access_Type_WR]           = B_Start_W;
    byte_start_types[Access_Type_RDWR]         = B_Start_RW;

    byte_end_types[Access_Type_RD]             = B_End_R;
    byte_end_types[Access_Type_WR]             = B_End_W;
    byte_end_types[Access_Type_RDWR]           = B_End_RW;
}

wp_t *
fs_sampler_t::sample_page_access(
    unsigned long index,
    access_type_t access_type,
    unsigned long access_tid,
    unsigned long long wp_id,
    unsigned long long wp_group_id)
{
    wp_t *wp = new page_wp_t(access_tid, access_type, index * PAGE_SIZE, wp_id, wp_group_id);
    filter->add(wp);
    return wp;
}

wp_t *
fs_sampler_t::sample_line_access(
    unsigned long index,
    access_type_t access_type,
    unsigned long access_tid,
    unsigned long long wp_id,
    unsigned long long wp_group_id)
{
    wp_t *wp = new line_wp_t(access_tid, access_type, index * LINE_SIZE, wp_id, wp_group_id);
    filter->add(wp);
    return wp;
}

wp_t *
fs_sampler_t::sample_byte_access(
    unsigned long index,
    access_type_t access_type,
    unsigned long access_tid,
    wp_t *line_wp,
    unsigned long long wp_id,
    unsigned long long wp_group_id)
{
    if (!is_prefetch(access_type)) {
        line_wp_t *mp_line_wp = (line_wp_t *)line_wp;

        wp_t *wp = new byte_wp_t(access_tid, access_type, mp_line_wp, index, wp_id, wp_group_id);
        filter->add(wp);
        return wp;
    }
    else
        return NULL;
}
