/// @file tools/sampler_sim/sampler_sim_ww.cc

// Copyright (c) 2008 Acumem AB.  All rights reserved.

#include "sampler_sim.hh"

#include <assert.h>

ww_sampler_t::ww_sampler_t(unsigned long sample_period,
                           filter_t *filter)
    : sampler_t(sample_period, filter)
{
}

ww_sampler_t::page_wp_t::page_wp_t(
    unsigned long tid,
    unsigned long address,
    unsigned long long wp_id,
    unsigned long long wp_group_id)
    : wp_t(address, PAGE_SIZE, wp_id, wp_group_id),
      tid(tid),
      seen_access_types(0)
{
}

bool
ww_sampler_t::page_wp_t::event(
    unsigned long long access_id,
    events_t *events,
    unsigned long access_tid,
    access_type_t access_type,
    unsigned int offset,
    unsigned int size,
    bool side_effects)
{
    // Log only unseen access types from the thread itself.
    bool hit =
        access_tid == tid &&
        !access_type_seen(access_type, &seen_access_types, side_effects);
    if (hit)
        events->insert(make_pair(0, access_id));
    return hit;
}

bool
ww_sampler_t::page_wp_t::remove(
    unsigned long access_tid,
    access_type_t access_type)
{
    // Terminate the watchpoint on non-prefetch accesses from the thread
    // itself.
    return access_tid == tid && !is_prefetch(access_type);
}

ww_sampler_t::line_wp_t::line_wp_t(
    unsigned long tid,
    bool write,
    unsigned long address,
    unsigned long long wp_id,
    unsigned long long wp_group_id)
    : wp_t(address, LINE_SIZE, wp_id, wp_group_id),
      tid(tid),
      want_invalidations(true),
      write(write),
      shadow(false),
      live_byte_wps(0)
{
}

bool
ww_sampler_t::line_wp_t::event(
    unsigned long long access_id,
    events_t *events,
    unsigned long access_tid,
    access_type_t access_type,
    unsigned int offset,
    unsigned int size,
    bool side_effects)
{
    bool hit = false;
    if (access_tid != tid) {
        if (!access_type_seen(access_type, &seen_access_types[access_tid], side_effects) ||
            (is_maybe_invalidation(access_type) && !shadow))
            // Log unseen access types from other threads, and all writes if
            // the watchpoint isn't a shadow.
            hit = true;
    }
    else
        // Log all accesses from the thread itself if the watchpoint is not a
        // shadow.
        if (!shadow) {
            hit = true;
            if (side_effects) {
                if ((write && is_write(access_type)) ||
                    (!write && !is_prefetch(access_type)))
                    // Turn write watchpoints into shadows on write accesses,
                    // turn non-write watchpoints into shadows on any
                    // non-prefetch access.
                    shadow = true;
                else
                    // We need invalidations from other threads to calculate
                    // the good old replacement miss ratio.
                    want_invalidations = true;
            }
        }
    if (hit)
        events->insert(make_pair(0, access_id));
    return hit;
}

bool
ww_sampler_t::line_wp_t::remove(
    unsigned long access_tid,
    access_type_t access_type)
{
    // Terminate the watchpoint when all associated byte watchpoints have been
    // terminated.
    return live_byte_wps == 0;
}

ww_sampler_t::byte_wp_t::byte_wp_t(
    unsigned long tid,
    line_wp_t *line_wp,
    unsigned long address,
    unsigned long long wp_id,
    unsigned long long wp_group_id)
    : wp_t(address, 1, wp_id, wp_group_id),
      tid(tid),
      want_invalidations(true),
      line_wp(line_wp)
{
    line_wp->live_byte_wps++;
}

ww_sampler_t::byte_wp_t::~byte_wp_t()
{
    line_wp->live_byte_wps--;
}

bool
ww_sampler_t::byte_wp_t::event(
    unsigned long long access_id,
    events_t *events,
    unsigned long access_tid,
    access_type_t access_type,
    unsigned int offset,
    unsigned int size,
    bool side_effects)
{
    bool hit = false;

    if (access_tid != tid) {
        // Log unseen access types from other threads.
        if (!access_type_seen(access_type, &seen_access_types[access_tid], side_effects))
            hit = true;

        // Log invalidations, if wanted.
        if (want_invalidations && is_maybe_invalidation(access_type)) {
            hit = true;
            if (side_effects && is_always_invalidation(access_type))
                want_invalidations = false;
        }
    }
    else {
        if (side_effects)
            // We need invalidations from other threads to calculate
            // the good old replacement miss ratio.
            want_invalidations = true;

        // Log all accesses from the thread itself.
        hit = true;
    }

    if (hit)
        events->insert(make_pair(0, access_id));
    return hit;
}

bool
ww_sampler_t::byte_wp_t::remove(
    unsigned long access_tid,
    access_type_t access_type)
{
    // Terminate write watchpoints on write accesses, terminate
    // non-write watchpoints on any non-prefetch access.
    return (access_tid == tid &&
            ((line_wp->write && is_write(access_type)) ||
             (!line_wp->write && !is_prefetch(access_type))));
}

wp_t *
ww_sampler_t::sample_page_access(
    unsigned long index,
    access_type_t access_type,
    unsigned long access_tid,
    unsigned long long wp_id,
    unsigned long long wp_group_id)
{
    wp_t *wp = new page_wp_t(access_tid, index * PAGE_SIZE, wp_id, wp_group_id);
    filter->add(wp);
    return wp;
}

wp_t *
ww_sampler_t::sample_line_access(
    unsigned long index,
    access_type_t access_type,
    unsigned long access_tid,
    unsigned long long wp_id,
    unsigned long long wp_group_id)
{
    wp_t *wp = new line_wp_t(access_tid, is_write_or_write_prefetch(access_type), index * LINE_SIZE, wp_id, wp_group_id);
    filter->add(wp);
    return wp;
}

wp_t *
ww_sampler_t::sample_byte_access(
    unsigned long index,
    access_type_t access_type,
    unsigned long access_tid,
    wp_t *line_wp,
    unsigned long long wp_id,
    unsigned long long wp_group_id)
{
    line_wp_t *mp_line_wp = (line_wp_t *)line_wp;

    wp_t *wp = new byte_wp_t(access_tid, mp_line_wp, index, wp_id, wp_group_id);
    filter->add(wp);
    return wp;
}
