/// @file tools/smpmerge/smpmerge.cc

// Copyright (c) 2009-2010 Acumem AB.  All rights reserved.

// NOTE: This tool is currently not built. It has not been updated to handle
//       the new "local time" and "global time" Fenris events.

// TODO: Separate addresses from different sample files.

#include "library/fenris/flow_file_reader.hh"
#include "library/fenris/flow_file_writer.hh"

#include <boost/program_options.hpp>
#include <boost/foreach.hpp>

#include <map>
#include <iostream>

using namespace acumem::library;

namespace po = boost::program_options;

struct u128_t {
    uint64_t hi;
    uint64_t lo;
};

static u128_t
umul(uint64_t a, uint64_t b)
{
    uint64_t a0 = a & 0xffffffff;
    uint64_t a1 = a >> 32;
    uint64_t b0 = b & 0xffffffff;
    uint64_t b1 = b >> 32;

    uint64_t c = a0 * b0;
    uint64_t d = a0 * b1;
    uint64_t e = a1 * b0;
    uint64_t f = a1 * b1;

    uint64_t c0 = c & 0xffffffff;
    uint64_t c1 = c >> 32;
    uint64_t d0 = d & 0xffffffff;
    uint64_t d1 = d >> 32;
    uint64_t e0 = e & 0xffffffff;
    uint64_t e1 = e >> 32;
    uint64_t f0 = f & 0xffffffff;
    uint64_t f1 = f >> 32;

    uint64_t g = c0;
    uint64_t h = c1 + d0 + e0;
    uint64_t i = d1 + e1 + f0 + (h >> 32);
    uint64_t j = f1 + (i >> 32);

    u128_t r = {(j << 32) | (i & 0xffffffff), (h << 32) | (g & 0xffffffff)};
    return r;
}

static uint64_t
udiv(u128_t a, uint64_t b)
{
    uint64_t r = 0;

    for (int i = 63; i >= 0; i--) {
        uint64_t c = b << i;
        uint64_t d = i == 0 ? 0 : b >> (64 - i);

        if (a.hi > d || (a.hi == d && a.lo >= c)) {
            r |= (uint64_t)1 << i;
            a.hi -= d + (c > a.lo ? 1 : 0);
            a.lo -= c;
        }
    }

    return r;
}

struct conf_t {
    vector<string> in_filenames;
    string out_filename;
};

static bool
parse_args(conf_t &conf, int argc, char **argv)
{
    po::options_description opt_desc_visible("Options");
    opt_desc_visible.add_options()
        ("help,h", "Print this help message.")
        ("in-file,i", po::value<vector<string> >(&conf.in_filenames),
         "Specifies an input file.")
        ("out-file,o", po::value<string>(&conf.out_filename),
         "Specifies the output file.")
        ;

    po::options_description opt_desc_invisible("Options");
    opt_desc_invisible.add_options()
        ("positional-args", "")
        ;

    po::options_description opt_desc_all("Options");
    opt_desc_all.add(opt_desc_visible).add(opt_desc_invisible);

    po::positional_options_description pos_desc;
    pos_desc.add("positional-args", -1);

    po::variables_map vm;
    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(opt_desc_all).positional(pos_desc).run(), vm);
        po::notify(vm);
    }
    catch (po::unknown_option &e) {
        cerr << "Error: " << e.what() << endl;
        cerr << opt_desc_visible << endl;
        return true;
    }

    if (vm.count("help") > 0) {
        cerr << opt_desc_visible << endl;
        exit(0);
    }

    bool error = false;

    if (vm.count("in-file") == 0) {
        cerr << "Error: no input file specified" << endl;
        error = true;
    }

    if (vm.count("out-file") == 0) {
        cerr << "Error: no output file specified" << endl;
        error = true;
    }

    if (vm.count("positional-args") > 0) {
        cerr << "Error: no positional arguments allowed" << endl;
        error = true;
    }

    return error;
}

class file_events_t : public fenris_flow_by_type_t {
public:
    virtual flow_status_t flow_init(fnr_features_t features) {
        this->features = features;
        return FLOW_OK;
    }

protected:
    virtual flow_status_t read(const fnr_event_read_t &event) {
        assert(0);
    }
    virtual flow_status_t write(const fnr_event_write_t &event) {
        assert(0);
    }
    virtual flow_status_t prefetch(const fnr_event_prefetch_t &event) {
        assert(0);
    }
    virtual flow_status_t sample_begin(const fnr_event_sample_begin_t &event) {
        assert(0);
    }
    virtual flow_status_t sample_end(const fnr_event_sample_end_t &event) {
        assert(0);
    }
    virtual flow_status_t sample_period(const fnr_event_sample_period_t &event) {
        sample_period_events.push_back(event);
        return FLOW_OK;
    }
    virtual flow_status_t acquisition_begin(const fnr_event_acquisition_begin_t &event) {
        acquisition_begin_events.push_back(event);
        return FLOW_OK;
    }
    virtual flow_status_t acquisition_end(const fnr_event_acquisition_end_t &event) {
        acquisition_end_events.push_back(event);
        return FLOW_OK;
    }
    virtual flow_status_t mmap(const fnr_event_mmap_t &event) {
        mmap_events.push_back(event);
        return FLOW_OK;
    }
    virtual flow_status_t munmap(const fnr_event_munmap_t &event) {
        munmap_events.push_back(event);
        return FLOW_OK;
    }
    virtual flow_status_t stack(const fnr_event_stack_t &event) {
        stack_events.push_back(event);
        return FLOW_OK;
    }
    virtual flow_status_t sample(const fnr_event_sample_t &event) {
        sample_events.push_back(event);
        return FLOW_OK;
    }
    virtual flow_status_t time_mapping(const fnr_event_time_mapping_t &event) {
        time_mapping_events.push_back(event);
        return FLOW_OK;
    }
    virtual flow_status_t local_time(const fnr_event_local_time_t &event) {
        // TODO: Handle this event type!
        assert(0);
    }
    virtual flow_status_t burst_begin(const fnr_event_burst_begin_t &event) {
        assert(0);
    }
    virtual flow_status_t burst_end(const fnr_event_burst_end_t &event) {
        assert(0);
    }
    virtual flow_status_t contexts(const fnr_event_contexts_t &event) {
        contexts_events.push_back(event);
        return FLOW_OK;
    }

public:
    fnr_features_t features;
    vector<fnr_event_sample_period_t> sample_period_events;
    vector<fnr_event_acquisition_begin_t> acquisition_begin_events;
    vector<fnr_event_acquisition_end_t> acquisition_end_events;
    vector<fnr_event_mmap_t> mmap_events;
    vector<fnr_event_munmap_t> munmap_events;
    vector<fnr_event_stack_t> stack_events;
    vector<fnr_event_sample_t> sample_events;
    vector<fnr_event_time_mapping_t> time_mapping_events;
    vector<fnr_event_contexts_t> contexts_events;
};

static fnr_time_t
global_2_at(fnr_time_t global_1, fnr_time_t global_end_1, fnr_time_t global_end_2) {
    fnr_time_t global_2 = udiv(umul(global_end_2, global_1), global_end_1);
    return global_2;
}

static fnr_time_t
global_2_before(fnr_time_t global_1, fnr_time_t global_end_1, fnr_time_t global_end_2) {
    fnr_time_t global_2 = udiv(umul(global_end_2, global_1), global_end_1);
    if (udiv(umul(global_end_1, global_2), global_end_2) == global_1)
        return global_2 - 1;
    else
        return global_2;
}

static fnr_time_t
merged_time_at(fnr_time_t global, unsigned int in_file, unsigned int num_in_file, const fnr_time_vector_t &end_times)
{
    fnr_time_t merged = global;
    for (unsigned int i = 0; i < num_in_file; i++) {
        if (i < in_file)
            merged += global_2_at(global, end_times[in_file], end_times[i]);
        if (i > in_file)
            merged += global_2_before(global, end_times[in_file], end_times[i]);
    }
    return merged;
}

int
main(int argc, char **argv)
{
    warning_consumer_ptr_t wc(new warning_consumer_stderr_t());

    conf_t conf;

    if (parse_args(conf, argc, argv))
        return 1;

    unsigned int num_in_files = conf.in_filenames.size();

    vector<file_events_t> file_events(num_in_files);

    // Read the input files.
    for (unsigned int i = 0; i < num_in_files; i++) {
        file_events_t &events = file_events[i];

        flow_file_reader_t file_reader(conf.in_filenames[i].c_str(), &events, wc);

        file_reader.dispatch();

        assert(events.features == (FNR_Feature_MT | FNR_Feature_Virtual_Threads));

        stream_container_header_t header = file_reader.get_header();
        assert(header.version == 1);
        assert(header.data_type == STREAM_CONTAINER_DATA_SAMPLES_BINARY);
    }

    // Create and initialize the output file.
    stream_container_header_t header(
        1,
        STREAM_CONTAINER_PRODUCT_INTERNAL,
        STREAM_CONTAINER_DATA_SAMPLES_BINARY,
        STREAM_CONTAINER_COMPRESSION_DEFLATE);

    flow_file_writer_t file_writer(conf.out_filename.c_str(), header, wc);

    file_writer.flow_init(FNR_Feature_MT | FNR_Feature_Virtual_Threads);

    // Gather values for the acquisition begin event.
    fnr_architecture_t architecture = FNR_Arch_None;
    fnr_argv_t merged_argv;
    string user;
    fnr_line_size_log2_vector_t line_sizes_log2;

    for (unsigned int i = 0; i < num_in_files; i++) {
        file_events_t &events = file_events[i];
        assert(events.acquisition_begin_events.size() == 1);
        fnr_event_acquisition_begin_t &acq_begin = events.acquisition_begin_events[0];

        if (i == 0) {
            architecture = acq_begin.architecture;
            line_sizes_log2 = acq_begin.line_sizes_log2;
        }
        else {
            assert(acq_begin.architecture == architecture);
            assert(acq_begin.line_sizes_log2 == line_sizes_log2);
            merged_argv.push_back("---");
            user += " --- ";
        }
        merged_argv.insert(merged_argv.end(), acq_begin.argv.begin(), acq_begin.argv.end());
        merged_argv.push_back("(");
        assert(events.contexts_events.size() == 1);
        BOOST_FOREACH(fnr_context_id_t context, events.contexts_events[0].contexts) {
            stringstream ss;
            ss << context;
            merged_argv.push_back(ss.str());
        }
        merged_argv.push_back(")");
        user += acq_begin.user;
    }

    // Write the acquisition begin event.
    file_writer.process_event(
        fnr_event_acquisition_begin_t(
            0,
            architecture,
            merged_argv,
            user.c_str(),
            line_sizes_log2));

    // Gather the acquisition lengths of each input file and calculate the
    // length of the output file.
    fnr_real_time_t merged_acq_length = (fnr_real_time_t)-1;
    vector<fnr_real_time_t> acq_lengths;

    for (unsigned int i = 0; i < num_in_files; i++) {
        file_events_t &events = file_events[i];
        assert(events.acquisition_begin_events.size() == 1);
        fnr_event_acquisition_begin_t &acq_begin = events.acquisition_begin_events[0];
        assert(events.acquisition_end_events.size() == 1);
        fnr_event_acquisition_end_t &acq_end = events.acquisition_end_events[0];

        fnr_real_time_t length = acq_end.time - acq_begin.time;
        acq_lengths.push_back(length);
        if (length < merged_acq_length)
            merged_acq_length = length;
    }

    // Gather the number of accessses in each input file.
    fnr_time_vector_t end_times;

    for (unsigned int i = 0; i < num_in_files; i++) {
        file_events_t &events = file_events[i];

        fnr_time_t end_time = 0;

        BOOST_FOREACH(const fnr_event_sample_t &sample, events.sample_events) {
            fnr_time_t last_time = sample.end.empty() ? sample.begin.time : sample.end.back().time;
            if (last_time > end_time)
                end_time = last_time;
        }

        assert(end_time > 0);

        end_times.push_back(end_time);
    }

    // Find the greatest sample period of the input files.
    fnr_time_t merged_sample_period = 0;

    for (unsigned int i = 0; i < num_in_files; i++) {
        file_events_t &events = file_events[i];

        assert(events.sample_period_events.size() == 1);
        if (events.sample_period_events[0].sample_period > merged_sample_period)
            merged_sample_period = events.sample_period_events[0].sample_period;
    }

    // Write the sample period event.
    file_writer.process_event(
        fnr_event_sample_period_t(
            merged_sample_period));

    // Decimate samples to get the same sample period in all files.
    for (unsigned int i = 0; i < num_in_files; i++) {
        file_events_t &events = file_events[i];

        // Calculate the ratio of samples to be kept.
        assert(events.sample_period_events.size() == 1);
        double ratio = (double)events.sample_period_events[0].sample_period / (double)merged_sample_period;

        // Create a set of all sample begin times.
        fnr_time_set_t begin_times;
        BOOST_FOREACH(fnr_event_sample_t &sample, events.sample_events)
            begin_times.insert(sample.begin.time);

        // Decimate the sample begin times.
        fnr_time_set_t kept_begin_times;
        BOOST_FOREACH(fnr_time_t begin_time, begin_times)
            if ((double)rand() / (double)RAND_MAX <= ratio)
                kept_begin_times.insert(begin_time);

        // Decimate the samples.
        unsigned int t = 0;
        for (unsigned int s = 0; s < events.sample_events.size(); s++)
            if (kept_begin_times.find(events.sample_events[s].begin.time) != kept_begin_times.end())
                events.sample_events[t++] = events.sample_events[s];
        events.sample_events.resize(t);
    }

    // Merge stacks and write them to the output file.
    fnr_stack_id_t stack_id = 1;
    for (unsigned int i = 0; i < num_in_files; i++) {
        file_events_t &events = file_events[i];

        map<fnr_stack_id_t, fnr_stack_id_t> stacks;
        stacks[0] = 0;

        BOOST_FOREACH(fnr_event_stack_t &stack, events.stack_events) {
            stacks[stack.id] = stack_id;
            stack.id = stack_id++;
            file_writer.process_event(stack);
        }

        BOOST_FOREACH(fnr_event_sample_t &sample, events.sample_events) {
            sample.begin.stack = stacks[sample.begin.stack];
            sample.pc_1_plus_stack = stacks[sample.pc_1_plus_stack];
            BOOST_FOREACH(fnr_sample_access_t &access, sample.end)
                access.stack = stacks[access.stack];
        }
    }

    // Merge contexts and write them to the output file.
    fnr_context_id_vector_t total_contexts;
    for (unsigned int i = 0; i < num_in_files; i++) {
        file_events_t &events = file_events[i];

        uint32_t thread_id_offset = total_contexts.size();

        // Merge context events.
        assert(events.contexts_events.size() == 1);
        fnr_event_contexts_t &contexts = events.contexts_events[0];
        total_contexts.insert(total_contexts.end(), contexts.contexts.begin(), contexts.contexts.end());

        // Update accesses.
        BOOST_FOREACH(fnr_event_sample_t &sample, events.sample_events) {
            sample.begin.thread.id += thread_id_offset;
            BOOST_FOREACH(fnr_sample_access_t &access, sample.end)
                access.thread.id += thread_id_offset;
        }

        // Update time mappings.
        BOOST_FOREACH(fnr_event_time_mapping_t &time_mapping, events.time_mapping_events) {
            BOOST_FOREACH(fnr_event_time_mapping_t::time_event_t &time_event, time_mapping.time_events)
                time_event.tid.id += thread_id_offset;
        }
    }

    file_writer.process_event(
        fnr_event_contexts_t(
            total_contexts));

    // Truncate the number of accesses from each input file to the length of
    // the output file and calculate the number of accesses in the output file.
    fnr_time_t merged_end_time = 0;

    for (unsigned int i = 0; i < num_in_files; i++) {
        end_times[i] = udiv(umul(end_times[i], merged_acq_length), acq_lengths[i]);
        merged_end_time += end_times[i];
    }

    // Translate times in accesses, mmap and munmap events to merged time.
    for (unsigned int i = 0; i < num_in_files; i++) {
        file_events_t &events = file_events[i];

        // Accesses...
        for (unsigned int s = 0; s < events.sample_events.size(); s++) {
            fnr_event_sample_t &sample = events.sample_events[s];
            sample.begin.time = merged_time_at(sample.begin.time, i, num_in_files, end_times);
            BOOST_FOREACH(fnr_sample_access_t &access, sample.end)
                access.time = merged_time_at(access.time, i, num_in_files, end_times);
        }

        // Mmaps...
        BOOST_FOREACH(fnr_event_mmap_t &mmap, events.mmap_events)
            mmap.global_time = merged_time_at(mmap.global_time, i, num_in_files, end_times);

        // Munmaps...
        BOOST_FOREACH(fnr_event_munmap_t &munmap, events.munmap_events)
            munmap.global_time = merged_time_at(munmap.global_time, i, num_in_files, end_times);
    }

    // Remove any accesses, mmap and munmap events after the merged end time.
    for (unsigned int i = 0; i < num_in_files; i++) {
        file_events_t &events = file_events[i];

        // Remove accesses beginning after the merged end time...
        unsigned int t = 0;
        for (unsigned int s = 0; s < events.sample_events.size(); s++)
            if (events.sample_events[s].begin.time <= merged_end_time)
                events.sample_events[t++] = events.sample_events[s];
        events.sample_events.resize(t);

        // Remove end accesses after the merged end time...
        BOOST_FOREACH(fnr_event_sample_t &sample, events.sample_events)
            for (unsigned int a = 0; a < sample.end.size(); a++)
                if (sample.end[a].time > merged_end_time) {
                    sample.end.resize(a);
                    break;
                }

        // Mmaps...
        for (unsigned int m = 0; m < events.mmap_events.size(); m++) {
            fnr_event_mmap_t &mmap = events.mmap_events[m];
            if (mmap.global_time > merged_end_time)
                events.mmap_events.erase(events.mmap_events.begin() + m--);
        }

        // Munmaps...
        for (unsigned int m = 0; m < events.munmap_events.size(); m++) {
            fnr_event_munmap_t &munmap = events.munmap_events[m];
            if (munmap.global_time > merged_end_time)
                events.munmap_events.erase(events.munmap_events.begin() + m--);
        }
    }

    // Write the sample events.
    for (unsigned int i = 0; i < num_in_files; i++) {
        file_events_t &events = file_events[i];

        BOOST_FOREACH(fnr_event_sample_t &sample, events.sample_events)
            file_writer.process_event(sample);
    }

    // Write the mmap events.
    for (unsigned int i = 0; i < num_in_files; i++) {
        file_events_t &events = file_events[i];

        BOOST_FOREACH(fnr_event_mmap_t &mmap, events.mmap_events)
            file_writer.process_event(mmap);
    }

    // Write the munmap events.
    for (unsigned int i = 0; i < num_in_files; i++) {
        file_events_t &events = file_events[i];

        BOOST_FOREACH(fnr_event_munmap_t &munmap, events.munmap_events)
            file_writer.process_event(munmap);
    }

    // Write the time mappings.
    typedef map<fnr_time_t, fnr_thread_id_t> time_tid_map_t;
    time_tid_map_t times;

    for (unsigned int i = 0; i < num_in_files; i++) {
        file_events_t &events = file_events[i];

        BOOST_FOREACH(fnr_event_sample_t &sample, events.sample_events) {
            times[sample.begin.time] = sample.begin.thread;
            BOOST_FOREACH(fnr_sample_access_t &access, sample.end)
                times[access.time] = access.thread;
        }
    }

    fnr_event_time_mapping_t time_event;
    BOOST_FOREACH(time_tid_map_t::value_type &time_tid, times) {
        if (time_event.time_events.size() == 1000) {
            file_writer.process_event(time_event);
            time_event.time_events.clear();
        }
        time_event.time_events.push_back(fnr_event_time_mapping_t::time_event_t(time_tid.first, time_tid.second));
    }
    if (time_event.time_events.size() > 0) {
        file_writer.process_event(time_event);
        time_event.time_events.clear();
    }

    // Write the acquisition end event.
    file_writer.process_event(
        fnr_event_acquisition_end_t(
            merged_acq_length));

    return 0;
}
