/// @file tools/stat_sweep/stat_sweep.cc

// Copyright (c) 2008-2010 Acumem AB.  All rights reserved.

#include <boost/foreach.hpp>
#include <boost/program_options.hpp>
#include <iostream>
#include <fstream>

#include "analysis/thread_mapping/thread_mapping_auto.hh"
#include "analysis/thread_mapping/thread_mapping_rr.hh"
#include "analysis/fetch_ratio/repl_fr_random.hh"
#include "analysis/fetch_ratio/repl_fr_lru.hh"
#include "analysis/fetch_ratio/windowing_static.hh"
#include "analysis/fetch_ratio/windowing_burst.hh"
#include "analysis/fetch_ratio/windowing_adaptive.hh"
#include "analysis/advice/advice.hh"
#include "analysis/feature_base/exceptions.hh"

using namespace std;
using namespace boost;
namespace po = boost::program_options;
using namespace acumem::library;
using namespace acumem::analysis;

#define SAMPLES_PER_WINDOW 100

struct conf_t {
    conf_t();
    bool init(int argc, char **argv);

    std::string sample_file;

    string engine;

    unsigned int num_caches;

    windowing_engine_cptr_t windowing;
    windowing_engine_cptr_t windowing_burst;

    fnr_cache_size_t cache_min;
    fnr_cache_size_t cache_max;
    int subdivisions;
    fnr_line_size_t line_size;

    bool dump_workspace;
    std::string workspace_dump_file;
    bool use_old_hwpf_algorithm;
    bool hwpf_enabled;
};

conf_t::conf_t()
    : sample_file(), engine("random"),
      num_caches(0),
      cache_min(8192), cache_max(16777216), subdivisions(4),
      line_size(64),
      dump_workspace(false),
      hwpf_enabled(true)
{
}

bool
conf_t::init(int argc, char **argv)
{
    string windowing_engine;
    fnr_sample_count_t samples_per_window(SAMPLES_PER_WINDOW);
    fnr_sample_count_t seed_size(SAMPLES_PER_WINDOW);
    fnr_sample_count_t growth_size(SAMPLES_PER_WINDOW / 4);
    double max_delta(0.25);

    po::options_description opt_desc_common("Options");
    opt_desc_common.add_options()
	("help,h", "Print help message.")
	("sample-file,s", po::value<string>(&sample_file),
	 "Sample file to analyze.")
	("engine,e", po::value<string>(&engine),
	 "Miss ratio engine.")
	("numberf-of-caches,n", po::value<unsigned int>(&num_caches),
	 "Total number of caches to assign threads to. Default is one "
	 "private cache per thread. Should match the number of caches "
	 "of the desired cache level for the intended processor/architecture.");

    po::options_description opt_desc_sweep("Sweep options");
    opt_desc_sweep.add_options()
	("cache-min,c", po::value<fnr_cache_size_t>(&cache_min),
	 "Sweep start.")
	("cache-max,C", po::value<fnr_cache_size_t>(&cache_max),
	 "Sweep end.")
	("line-size,l", po::value<fnr_cache_size_t>(&line_size),
	 "Line size.");

    po::options_description opt_desc_windowing("Windowing options");
    opt_desc_windowing.add_options()
	("windowing", po::value<string>(&windowing_engine),
	 "Window engine to use (static/adaptive/burst).")
	("samples-per-window,w", po::value<fnr_sample_count_t>(&samples_per_window),
	 "Samples per window, only applies to static windowing.")
	("window-seed", po::value<fnr_sample_count_t>(&seed_size),
	 "Size of seed window, applies to adaptive windowing.")
	("window-growth", po::value<fnr_sample_count_t>(&growth_size),
	 "Size of growth window, applies to adaptive windowing.")
	("window-delta", po::value<double>(&max_delta),
	 "Maximum growth window delta, applies to adaptive windowing.")
	("new-algorithm", "Use new hwpf algorithm")
        ;

    po::options_description opt_desc_debug("Debug options");
    opt_desc_debug.add_options()
	("dump-workspace", po::value<string>(&workspace_dump_file),
	 "Dump workspace contents to a file")
	;

    po::options_description opt_desc_all("Options");
    opt_desc_all.add(opt_desc_common)
	.add(opt_desc_sweep)
	.add(opt_desc_windowing)
	.add(opt_desc_debug);

    po::positional_options_description pos_desc;
    pos_desc.add("positional-args", -1);

    po::variables_map vm;
    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(opt_desc_all).positional(pos_desc).run(), vm);
        po::notify(vm);
    } catch (po::error &e) {
        cerr << "Error: " << e.what() << endl;
        cerr << opt_desc_all << endl;
        return false;
    }

    if (vm.count("help") > 0) {
	cout << opt_desc_all << endl;
	return false;
    }

    if (vm.count("windowing") > 0) {
	if (windowing_engine == "static")
	    windowing.reset(
		new acumem::analysis::windowing_static_t(samples_per_window));
	else if (windowing_engine == "adaptive")
	    windowing.reset(
		new acumem::analysis::windowing_adaptive_t(
		    seed_size, growth_size, max_delta));
	else if (windowing_engine == "burst")
	    windowing.reset(new acumem::analysis::windowing_burst_t());
	else {
	    cerr << "Error: Illegal window engine specified." << endl;
	    return false;
	}
	windowing_burst = windowing;
    } else {
	windowing.reset(
	    new acumem::analysis::windowing_adaptive_t(
		seed_size, growth_size, max_delta));
	windowing_burst.reset(
	    new acumem::analysis::windowing_burst_t());
    }

    dump_workspace = vm.count("dump-workspace") > 0;

    use_old_hwpf_algorithm = vm.count("new-algorithm") == 0;
    return true;
}

void
output_csv(ostream &os, const drill_down_result_t &result)
{
    os << "\"cache size\",\"accesses\","
       << "\"fetch rate\",\"miss rate\",\"fetches\",\"misses\","
       << "\"spatial usage\"" << endl;

    BOOST_FOREACH(const cache_sweep_index_map_t::value_type &cs,
		  result.cache_range) {
	const stats_t &stats(*result.stats[cs.second]->stats);
	os << cs.first << ","
	   << stats.all_ratio.accesses() << ","
	   << stats.all_ratio.total_fetch_ratio() << ","
	   << stats.stall_ratio.total_fetch_ratio() << ","
	   << stats.all_ratio.total_fetches() << ","
	   << stats.stall_ratio.total_fetches() << ","
	   << stats.spatial_util.value() << endl;
    }
}


int
main(int argc, char *argv[])
{
    conf_t conf;

    if (!conf.init(argc, argv))
	return 1;

    try {
	feature_context_ptr_t context(new feature_context_t());

        parser_result_cptr_t parser_result(
	    parser_t(context, conf.sample_file).get());


	thread_mapping_cptr_t thread_mapping;
	if (conf.num_caches == 0)
	    thread_mapping.reset(new thread_mapping_auto_t(context,
							   conf.sample_file,
							   false));
	else
	    thread_mapping.reset(new thread_mapping_rr_t(context,
							 conf.sample_file,
							 conf.num_caches));

	repl_fr_engine_cptr_t engine;
	if (conf.engine == "random")
	    engine.reset(
		new repl_fr_random_t(context,
				     conf.sample_file,
				     0,
				     thread_mapping,
				     conf.windowing,
				     cpu_model_t()));
	else if (conf.engine == "lru")
	    engine.reset(
		new repl_fr_lru_t(context,
				  conf.sample_file,
				  0,
				  thread_mapping,
				  conf.windowing,
				  cpu_model_t()));
	else {
	    cerr << "Invalid FR engine specified." << endl;
	    return 1;
	}

	configuration_t c(conf.sample_file,
			  0,
			  conf.cache_min,
			  conf.line_size,
			  thread_mapping,
			  cpu_model_t(),
			  engine,
			  (parser_result->burst_sampled ?
			   conf.windowing_burst :
			   conf.windowing),
			  sweep_t(conf.cache_min,
				  conf.cache_max,
				  conf.subdivisions,
				  conf.line_size),
			  true, true, true, true, false,
			  conf.use_old_hwpf_algorithm,
                          conf.hwpf_enabled,
			  0.01,
                          0.0001,
                          0x8000 /* L1 cache size for prefetchnta analysis. */);
	drill_down_result_ptr_t result;
	advisor_t a;
	drill_down_query_t q(result, NULL);
	a.drill_down(c, context, q);

	output_csv(cout, *result);

	if(conf.dump_workspace) {
	    ofstream wdump(conf.workspace_dump_file.c_str());
	    context->workspace_stack->dump(wdump);
	}

	return 0;
    } catch (feature_exception_t &t) {
        cerr << "Error: " << t.what() << endl;
	return 1;
    }
}
